(function (global, factory) {
  typeof exports === 'object' && typeof module !== 'undefined' ? module.exports = factory(require('geotiff')) :
  typeof define === 'function' && define.amd ? define(['geotiff'], factory) :
  (global = typeof globalThis !== 'undefined' ? globalThis : global || self, global.ol = factory(global.geotiff));
})(this, (function (geotiff) { 'use strict';

  /**
   * @module ol/AssertionError
   */

  /** @type {Object<number, string>} */
  const messages = {
    1: 'The view center is not defined',
    2: 'The view resolution is not defined',
    3: 'The view rotation is not defined',
    4: '`image` and `src` cannot be provided at the same time',
    5: '`imgSize` must be set when `image` is provided',
    7: '`format` must be set when `url` is set',
    8: 'Unknown `serverType` configured',
    9: '`url` must be configured or set using `#setUrl()`',
    10: 'The default `geometryFunction` can only handle `Point` geometries',
    11: '`options.featureTypes` must be an Array',
    12: '`options.geometryName` must also be provided when `options.bbox` is set',
    13: 'Invalid corner',
    14: 'Invalid color',
    15: 'Tried to get a value for a key that does not exist in the cache',
    16: 'Tried to set a value for a key that is used already',
    17: '`resolutions` must be sorted in descending order',
    18: 'Either `origin` or `origins` must be configured, never both',
    19: 'Number of `tileSizes` and `resolutions` must be equal',
    20: 'Number of `origins` and `resolutions` must be equal',
    22: 'Either `tileSize` or `tileSizes` must be configured, never both',
    24: 'Invalid extent or geometry provided as `geometry`',
    25: 'Cannot fit empty extent provided as `geometry`',
    26: 'Features must have an id set',
    27: 'Features must have an id set',
    28: '`renderMode` must be `"hybrid"` or `"vector"`',
    30: 'The passed `feature` was already added to the source',
    31: 'Tried to enqueue an `element` that was already added to the queue',
    32: 'Transformation matrix cannot be inverted',
    33: 'Invalid units',
    34: 'Invalid geometry layout',
    36: 'Unknown SRS type',
    37: 'Unknown geometry type found',
    38: '`styleMapValue` has an unknown type',
    39: 'Unknown geometry type',
    40: 'Expected `feature` to have a geometry',
    41: 'Expected an `ol/style/Style` or an array of `ol/style/Style.js`',
    42: 'Question unknown, the answer is 42',
    43: 'Expected `layers` to be an array or a `Collection`',
    47: 'Expected `controls` to be an array or an `ol/Collection`',
    48: 'Expected `interactions` to be an array or an `ol/Collection`',
    49: 'Expected `overlays` to be an array or an `ol/Collection`',
    50: '`options.featureTypes` should be an Array',
    51: 'Either `url` or `tileJSON` options must be provided',
    52: 'Unknown `serverType` configured',
    53: 'Unknown `tierSizeCalculation` configured',
    55: 'The {-y} placeholder requires a tile grid with extent',
    56: 'mapBrowserEvent must originate from a pointer event',
    57: 'At least 2 conditions are required',
    59: 'Invalid command found in the PBF',
    60: 'Missing or invalid `size`',
    61: 'Cannot determine IIIF Image API version from provided image information JSON',
    62: 'A `WebGLArrayBuffer` must either be of type `ELEMENT_ARRAY_BUFFER` or `ARRAY_BUFFER`',
    64: 'Layer opacity must be a number',
    66: '`forEachFeatureAtCoordinate` cannot be used on a WebGL layer if the hit detection logic has not been enabled. This is done by providing adequate shaders using the `hitVertexShader` and `hitFragmentShader` properties of `WebGLPointsLayerRenderer`',
    67: 'A layer can only be added to the map once. Use either `layer.setMap()` or `map.addLayer()`, not both',
    68: 'A VectorTile source can only be rendered if it has a projection compatible with the view projection',
  };

  /**
   * Error object thrown when an assertion failed. This is an ECMA-262 Error,
   * extended with a `code` property.
   * See https://developer.mozilla.org/en-US/docs/Web/JavaScript/Reference/Global_Objects/Error.
   */
  class AssertionError extends Error {
    /**
     * @param {number} code Error code.
     */
    constructor(code) {
      const message = messages[code];

      super(message);

      /**
       * Error code. The meaning of the code can be found on
       * https://openlayers.org/en/latest/doc/errors/ (replace `latest` with
       * the version found in the OpenLayers script's header comment if a version
       * other than the latest is used).
       * @type {number}
       * @deprecated ol/AssertionError and error codes will be removed in v8.0
       * @api
       */
      this.code = code;

      /**
       * @type {string}
       */
      this.name = 'AssertionError';

      // Re-assign message, see https://github.com/Rich-Harris/buble/issues/40
      this.message = message;
    }
  }

  /**
   * @module ol/events/Event
   */

  /**
   * @classdesc
   * Stripped down implementation of the W3C DOM Level 2 Event interface.
   * See https://www.w3.org/TR/DOM-Level-2-Events/events.html#Events-interface.
   *
   * This implementation only provides `type` and `target` properties, and
   * `stopPropagation` and `preventDefault` methods. It is meant as base class
   * for higher level events defined in the library, and works with
   * {@link module:ol/events/Target~Target}.
   */
  class BaseEvent {
    /**
     * @param {string} type Type.
     */
    constructor(type) {
      /**
       * @type {boolean}
       */
      this.propagationStopped;

      /**
       * @type {boolean}
       */
      this.defaultPrevented;

      /**
       * The event type.
       * @type {string}
       * @api
       */
      this.type = type;

      /**
       * The event target.
       * @type {Object}
       * @api
       */
      this.target = null;
    }

    /**
     * Prevent default. This means that no emulated `click`, `singleclick` or `doubleclick` events
     * will be fired.
     * @api
     */
    preventDefault() {
      this.defaultPrevented = true;
    }

    /**
     * Stop event propagation.
     * @api
     */
    stopPropagation() {
      this.propagationStopped = true;
    }
  }

  /**
   * @param {Event|import("./Event.js").default} evt Event
   */
  function stopPropagation(evt) {
    evt.stopPropagation();
  }

  /**
   * @param {Event|import("./Event.js").default} evt Event
   */
  function preventDefault(evt) {
    evt.preventDefault();
  }

  /**
   * @module ol/ObjectEventType
   */

  /**
   * @enum {string}
   */
  var ObjectEventType = {
    /**
     * Triggered when a property is changed.
     * @event module:ol/Object.ObjectEvent#propertychange
     * @api
     */
    PROPERTYCHANGE: 'propertychange',
  };

  /**
   * @typedef {'propertychange'} Types
   */

  /**
   * @module ol/Disposable
   */

  /**
   * @classdesc
   * Objects that need to clean up after themselves.
   */
  class Disposable {
    constructor() {
      /**
       * The object has already been disposed.
       * @type {boolean}
       * @protected
       */
      this.disposed = false;
    }

    /**
     * Clean up.
     */
    dispose() {
      if (!this.disposed) {
        this.disposed = true;
        this.disposeInternal();
      }
    }

    /**
     * Extension point for disposable objects.
     * @protected
     */
    disposeInternal() {}
  }

  /**
   * @module ol/array
   */

  /**
   * Performs a binary search on the provided sorted list and returns the index of the item if found. If it can't be found it'll return -1.
   * https://github.com/darkskyapp/binary-search
   *
   * @param {Array<*>} haystack Items to search through.
   * @param {*} needle The item to look for.
   * @param {Function} [comparator] Comparator function.
   * @return {number} The index of the item if found, -1 if not.
   */
  function binarySearch$1(haystack, needle, comparator) {
    let mid, cmp;
    comparator = comparator || numberSafeCompareFunction;
    let low = 0;
    let high = haystack.length;
    let found = false;

    while (low < high) {
      /* Note that "(low + high) >>> 1" may overflow, and results in a typecast
       * to double (which gives the wrong results). */
      mid = low + ((high - low) >> 1);
      cmp = +comparator(haystack[mid], needle);

      if (cmp < 0.0) {
        /* Too low. */
        low = mid + 1;
      } else {
        /* Key found or too high */
        high = mid;
        found = !cmp;
      }
    }

    /* Key not found. */
    return found ? low : ~low;
  }

  /**
   * Compare function for array sort that is safe for numbers.
   * @param {*} a The first object to be compared.
   * @param {*} b The second object to be compared.
   * @return {number} A negative number, zero, or a positive number as the first
   *     argument is less than, equal to, or greater than the second.
   */
  function numberSafeCompareFunction(a, b) {
    return a > b ? 1 : a < b ? -1 : 0;
  }

  /**
   * {@link module:ol/tilegrid/TileGrid~TileGrid#getZForResolution} can use a function
   * of this type to determine which nearest resolution to use.
   *
   * This function takes a `{number}` representing a value between two array entries,
   * a `{number}` representing the value of the nearest higher entry and
   * a `{number}` representing the value of the nearest lower entry
   * as arguments and returns a `{number}`. If a negative number or zero is returned
   * the lower value will be used, if a positive number is returned the higher value
   * will be used.
   * @typedef {function(number, number, number): number} NearestDirectionFunction
   * @api
   */

  /**
   * @param {Array<number>} arr Array in descending order.
   * @param {number} target Target.
   * @param {number|NearestDirectionFunction} direction
   *    0 means return the nearest,
   *    > 0 means return the largest nearest,
   *    < 0 means return the smallest nearest.
   * @return {number} Index.
   */
  function linearFindNearest(arr, target, direction) {
    const n = arr.length;
    if (arr[0] <= target) {
      return 0;
    } else if (target <= arr[n - 1]) {
      return n - 1;
    } else {
      let i;
      if (direction > 0) {
        for (i = 1; i < n; ++i) {
          if (arr[i] < target) {
            return i - 1;
          }
        }
      } else if (direction < 0) {
        for (i = 1; i < n; ++i) {
          if (arr[i] <= target) {
            return i;
          }
        }
      } else {
        for (i = 1; i < n; ++i) {
          if (arr[i] == target) {
            return i;
          } else if (arr[i] < target) {
            if (typeof direction === 'function') {
              if (direction(target, arr[i - 1], arr[i]) > 0) {
                return i - 1;
              } else {
                return i;
              }
            } else if (arr[i - 1] - target < target - arr[i]) {
              return i - 1;
            } else {
              return i;
            }
          }
        }
      }
      return n - 1;
    }
  }

  /**
   * @param {Array<*>} arr Array.
   * @param {number} begin Begin index.
   * @param {number} end End index.
   */
  function reverseSubArray(arr, begin, end) {
    while (begin < end) {
      const tmp = arr[begin];
      arr[begin] = arr[end];
      arr[end] = tmp;
      ++begin;
      --end;
    }
  }

  /**
   * @param {Array<VALUE>} arr The array to modify.
   * @param {!Array<VALUE>|VALUE} data The elements or arrays of elements to add to arr.
   * @template VALUE
   */
  function extend$2(arr, data) {
    const extension = Array.isArray(data) ? data : [data];
    const length = extension.length;
    for (let i = 0; i < length; i++) {
      arr[arr.length] = extension[i];
    }
  }

  /**
   * @param {Array<VALUE>} arr The array to modify.
   * @param {VALUE} obj The element to remove.
   * @template VALUE
   * @return {boolean} If the element was removed.
   */
  function remove$1(arr, obj) {
    const i = arr.indexOf(obj);
    const found = i > -1;
    if (found) {
      arr.splice(i, 1);
    }
    return found;
  }

  /**
   * @param {Array|Uint8ClampedArray} arr1 The first array to compare.
   * @param {Array|Uint8ClampedArray} arr2 The second array to compare.
   * @return {boolean} Whether the two arrays are equal.
   */
  function equals$3(arr1, arr2) {
    const len1 = arr1.length;
    if (len1 !== arr2.length) {
      return false;
    }
    for (let i = 0; i < len1; i++) {
      if (arr1[i] !== arr2[i]) {
        return false;
      }
    }
    return true;
  }

  /**
   * Sort the passed array such that the relative order of equal elements is preserved.
   * See https://en.wikipedia.org/wiki/Sorting_algorithm#Stability for details.
   * @param {Array<*>} arr The array to sort (modifies original).
   * @param {!function(*, *): number} compareFnc Comparison function.
   * @api
   */
  function stableSort(arr, compareFnc) {
    const length = arr.length;
    const tmp = Array(arr.length);
    let i;
    for (i = 0; i < length; i++) {
      tmp[i] = {index: i, value: arr[i]};
    }
    tmp.sort(function (a, b) {
      return compareFnc(a.value, b.value) || a.index - b.index;
    });
    for (i = 0; i < arr.length; i++) {
      arr[i] = tmp[i].value;
    }
  }

  /**
   * @param {Array<*>} arr The array to test.
   * @param {Function} [func] Comparison function.
   * @param {boolean} [strict] Strictly sorted (default false).
   * @return {boolean} Return index.
   */
  function isSorted(arr, func, strict) {
    const compare = func || numberSafeCompareFunction;
    return arr.every(function (currentVal, index) {
      if (index === 0) {
        return true;
      }
      const res = compare(arr[index - 1], currentVal);
      return !(res > 0 || (strict && res === 0));
    });
  }

  /**
   * @module ol/functions
   */

  /**
   * Always returns true.
   * @return {boolean} true.
   */
  function TRUE() {
    return true;
  }

  /**
   * Always returns false.
   * @return {boolean} false.
   */
  function FALSE() {
    return false;
  }

  /**
   * A reusable function, used e.g. as a default for callbacks.
   *
   * @return {void} Nothing.
   */
  function VOID() {}

  /**
   * Wrap a function in another function that remembers the last return.  If the
   * returned function is called twice in a row with the same arguments and the same
   * this object, it will return the value from the first call in the second call.
   *
   * @param {function(...any): ReturnType} fn The function to memoize.
   * @return {function(...any): ReturnType} The memoized function.
   * @template ReturnType
   */
  function memoizeOne(fn) {
    let called = false;

    /** @type {ReturnType} */
    let lastResult;

    /** @type {Array<any>} */
    let lastArgs;

    let lastThis;

    return function () {
      const nextArgs = Array.prototype.slice.call(arguments);
      if (!called || this !== lastThis || !equals$3(nextArgs, lastArgs)) {
        called = true;
        lastThis = this;
        lastArgs = nextArgs;
        lastResult = fn.apply(this, arguments);
      }
      return lastResult;
    };
  }

  /**
   * @template T
   * @param {function(): (T | Promise<T>)} getter A function that returns a value or a promise for a value.
   * @return {Promise<T>} A promise for the value.
   */
  function toPromise(getter) {
    function promiseGetter() {
      let value;
      try {
        value = getter();
      } catch (err) {
        return Promise.reject(err);
      }
      if (value instanceof Promise) {
        return value;
      }
      return Promise.resolve(value);
    }
    return promiseGetter();
  }

  /**
   * @module ol/obj
   */

  /**
   * Removes all properties from an object.
   * @param {Object} object The object to clear.
   */
  function clear$2(object) {
    for (const property in object) {
      delete object[property];
    }
  }

  /**
   * Determine if an object has any properties.
   * @param {Object} object The object to check.
   * @return {boolean} The object is empty.
   */
  function isEmpty$1(object) {
    let property;
    for (property in object) {
      return false;
    }
    return !property;
  }

  /**
   * @module ol/events/Target
   */

  /**
   * @typedef {EventTarget|Target} EventTargetLike
   */

  /**
   * @classdesc
   * A simplified implementation of the W3C DOM Level 2 EventTarget interface.
   * See https://www.w3.org/TR/2000/REC-DOM-Level-2-Events-20001113/events.html#Events-EventTarget.
   *
   * There are two important simplifications compared to the specification:
   *
   * 1. The handling of `useCapture` in `addEventListener` and
   *    `removeEventListener`. There is no real capture model.
   * 2. The handling of `stopPropagation` and `preventDefault` on `dispatchEvent`.
   *    There is no event target hierarchy. When a listener calls
   *    `stopPropagation` or `preventDefault` on an event object, it means that no
   *    more listeners after this one will be called. Same as when the listener
   *    returns false.
   */
  class Target extends Disposable {
    /**
     * @param {*} [target] Default event target for dispatched events.
     */
    constructor(target) {
      super();

      /**
       * @private
       * @type {*}
       */
      this.eventTarget_ = target;

      /**
       * @private
       * @type {Object<string, number>}
       */
      this.pendingRemovals_ = null;

      /**
       * @private
       * @type {Object<string, number>}
       */
      this.dispatching_ = null;

      /**
       * @private
       * @type {Object<string, Array<import("../events.js").Listener>>}
       */
      this.listeners_ = null;
    }

    /**
     * @param {string} type Type.
     * @param {import("../events.js").Listener} listener Listener.
     */
    addEventListener(type, listener) {
      if (!type || !listener) {
        return;
      }
      const listeners = this.listeners_ || (this.listeners_ = {});
      const listenersForType = listeners[type] || (listeners[type] = []);
      if (!listenersForType.includes(listener)) {
        listenersForType.push(listener);
      }
    }

    /**
     * Dispatches an event and calls all listeners listening for events
     * of this type. The event parameter can either be a string or an
     * Object with a `type` property.
     *
     * @param {import("./Event.js").default|string} event Event object.
     * @return {boolean|undefined} `false` if anyone called preventDefault on the
     *     event object or if any of the listeners returned false.
     * @api
     */
    dispatchEvent(event) {
      const isString = typeof event === 'string';
      const type = isString ? event : event.type;
      const listeners = this.listeners_ && this.listeners_[type];
      if (!listeners) {
        return;
      }

      const evt = isString ? new BaseEvent(event) : /** @type {Event} */ (event);
      if (!evt.target) {
        evt.target = this.eventTarget_ || this;
      }
      const dispatching = this.dispatching_ || (this.dispatching_ = {});
      const pendingRemovals =
        this.pendingRemovals_ || (this.pendingRemovals_ = {});
      if (!(type in dispatching)) {
        dispatching[type] = 0;
        pendingRemovals[type] = 0;
      }
      ++dispatching[type];
      let propagate;
      for (let i = 0, ii = listeners.length; i < ii; ++i) {
        if ('handleEvent' in listeners[i]) {
          propagate = /** @type {import("../events.js").ListenerObject} */ (
            listeners[i]
          ).handleEvent(evt);
        } else {
          propagate = /** @type {import("../events.js").ListenerFunction} */ (
            listeners[i]
          ).call(this, evt);
        }
        if (propagate === false || evt.propagationStopped) {
          propagate = false;
          break;
        }
      }
      if (--dispatching[type] === 0) {
        let pr = pendingRemovals[type];
        delete pendingRemovals[type];
        while (pr--) {
          this.removeEventListener(type, VOID);
        }
        delete dispatching[type];
      }
      return propagate;
    }

    /**
     * Clean up.
     */
    disposeInternal() {
      this.listeners_ && clear$2(this.listeners_);
    }

    /**
     * Get the listeners for a specified event type. Listeners are returned in the
     * order that they will be called in.
     *
     * @param {string} type Type.
     * @return {Array<import("../events.js").Listener>|undefined} Listeners.
     */
    getListeners(type) {
      return (this.listeners_ && this.listeners_[type]) || undefined;
    }

    /**
     * @param {string} [type] Type. If not provided,
     *     `true` will be returned if this event target has any listeners.
     * @return {boolean} Has listeners.
     */
    hasListener(type) {
      if (!this.listeners_) {
        return false;
      }
      return type
        ? type in this.listeners_
        : Object.keys(this.listeners_).length > 0;
    }

    /**
     * @param {string} type Type.
     * @param {import("../events.js").Listener} listener Listener.
     */
    removeEventListener(type, listener) {
      const listeners = this.listeners_ && this.listeners_[type];
      if (listeners) {
        const index = listeners.indexOf(listener);
        if (index !== -1) {
          if (this.pendingRemovals_ && type in this.pendingRemovals_) {
            // make listener a no-op, and remove later in #dispatchEvent()
            listeners[index] = VOID;
            ++this.pendingRemovals_[type];
          } else {
            listeners.splice(index, 1);
            if (listeners.length === 0) {
              delete this.listeners_[type];
            }
          }
        }
      }
    }
  }

  /**
   * @module ol/events/EventType
   */

  /**
   * @enum {string}
   * @const
   */
  var EventType$1 = {
    /**
     * Generic change event. Triggered when the revision counter is increased.
     * @event module:ol/events/Event~BaseEvent#change
     * @api
     */
    CHANGE: 'change',

    /**
     * Generic error event. Triggered when an error occurs.
     * @event module:ol/events/Event~BaseEvent#error
     * @api
     */
    ERROR: 'error',

    BLUR: 'blur',
    CLEAR: 'clear',
    CONTEXTMENU: 'contextmenu',
    CLICK: 'click',
    DBLCLICK: 'dblclick',
    DRAGENTER: 'dragenter',
    DRAGOVER: 'dragover',
    DROP: 'drop',
    FOCUS: 'focus',
    KEYDOWN: 'keydown',
    KEYPRESS: 'keypress',
    LOAD: 'load',
    RESIZE: 'resize',
    TOUCHMOVE: 'touchmove',
    WHEEL: 'wheel',
  };

  /**
   * @module ol/events
   */

  /**
   * Key to use with {@link module:ol/Observable.unByKey}.
   * @typedef {Object} EventsKey
   * @property {ListenerFunction} listener Listener.
   * @property {import("./events/Target.js").EventTargetLike} target Target.
   * @property {string} type Type.
   * @api
   */

  /**
   * Listener function. This function is called with an event object as argument.
   * When the function returns `false`, event propagation will stop.
   *
   * @typedef {function((Event|import("./events/Event.js").default)): (void|boolean)} ListenerFunction
   * @api
   */

  /**
   * @typedef {Object} ListenerObject
   * @property {ListenerFunction} handleEvent HandleEvent listener function.
   */

  /**
   * @typedef {ListenerFunction|ListenerObject} Listener
   */

  /**
   * Registers an event listener on an event target. Inspired by
   * https://google.github.io/closure-library/api/source/closure/goog/events/events.js.src.html
   *
   * This function efficiently binds a `listener` to a `this` object, and returns
   * a key for use with {@link module:ol/events.unlistenByKey}.
   *
   * @param {import("./events/Target.js").EventTargetLike} target Event target.
   * @param {string} type Event type.
   * @param {ListenerFunction} listener Listener.
   * @param {Object} [thisArg] Object referenced by the `this` keyword in the
   *     listener. Default is the `target`.
   * @param {boolean} [once] If true, add the listener as one-off listener.
   * @return {EventsKey} Unique key for the listener.
   */
  function listen(target, type, listener, thisArg, once) {
    if (thisArg && thisArg !== target) {
      listener = listener.bind(thisArg);
    }
    if (once) {
      const originalListener = listener;
      listener = function () {
        target.removeEventListener(type, listener);
        originalListener.apply(this, arguments);
      };
    }
    const eventsKey = {
      target: target,
      type: type,
      listener: listener,
    };
    target.addEventListener(type, listener);
    return eventsKey;
  }

  /**
   * Registers a one-off event listener on an event target. Inspired by
   * https://google.github.io/closure-library/api/source/closure/goog/events/events.js.src.html
   *
   * This function efficiently binds a `listener` as self-unregistering listener
   * to a `this` object, and returns a key for use with
   * {@link module:ol/events.unlistenByKey} in case the listener needs to be
   * unregistered before it is called.
   *
   * When {@link module:ol/events.listen} is called with the same arguments after this
   * function, the self-unregistering listener will be turned into a permanent
   * listener.
   *
   * @param {import("./events/Target.js").EventTargetLike} target Event target.
   * @param {string} type Event type.
   * @param {ListenerFunction} listener Listener.
   * @param {Object} [thisArg] Object referenced by the `this` keyword in the
   *     listener. Default is the `target`.
   * @return {EventsKey} Key for unlistenByKey.
   */
  function listenOnce(target, type, listener, thisArg) {
    return listen(target, type, listener, thisArg, true);
  }

  /**
   * Unregisters event listeners on an event target. Inspired by
   * https://google.github.io/closure-library/api/source/closure/goog/events/events.js.src.html
   *
   * The argument passed to this function is the key returned from
   * {@link module:ol/events.listen} or {@link module:ol/events.listenOnce}.
   *
   * @param {EventsKey} key The key.
   */
  function unlistenByKey(key) {
    if (key && key.target) {
      key.target.removeEventListener(key.type, key.listener);
      clear$2(key);
    }
  }

  /**
   * @module ol/Observable
   */

  /***
   * @template {string} Type
   * @template {Event|import("./events/Event.js").default} EventClass
   * @template Return
   * @typedef {(type: Type, listener: (event: EventClass) => ?) => Return} OnSignature
   */

  /***
   * @template {string} Type
   * @template Return
   * @typedef {(type: Type[], listener: (event: Event|import("./events/Event").default) => ?) => Return extends void ? void : Return[]} CombinedOnSignature
   */

  /**
   * @typedef {'change'|'error'} EventTypes
   */

  /***
   * @template Return
   * @typedef {OnSignature<EventTypes, import("./events/Event.js").default, Return> & CombinedOnSignature<EventTypes, Return>} ObservableOnSignature
   */

  /**
   * @classdesc
   * Abstract base class; normally only used for creating subclasses and not
   * instantiated in apps.
   * An event target providing convenient methods for listener registration
   * and unregistration. A generic `change` event is always available through
   * {@link module:ol/Observable~Observable#changed}.
   *
   * @fires import("./events/Event.js").default
   * @api
   */
  class Observable extends Target {
    constructor() {
      super();

      this.on =
        /** @type {ObservableOnSignature<import("./events").EventsKey>} */ (
          this.onInternal
        );

      this.once =
        /** @type {ObservableOnSignature<import("./events").EventsKey>} */ (
          this.onceInternal
        );

      this.un = /** @type {ObservableOnSignature<void>} */ (this.unInternal);

      /**
       * @private
       * @type {number}
       */
      this.revision_ = 0;
    }

    /**
     * Increases the revision counter and dispatches a 'change' event.
     * @api
     */
    changed() {
      ++this.revision_;
      this.dispatchEvent(EventType$1.CHANGE);
    }

    /**
     * Get the version number for this object.  Each time the object is modified,
     * its version number will be incremented.
     * @return {number} Revision.
     * @api
     */
    getRevision() {
      return this.revision_;
    }

    /**
     * @param {string|Array<string>} type Type.
     * @param {function((Event|import("./events/Event").default)): ?} listener Listener.
     * @return {import("./events.js").EventsKey|Array<import("./events.js").EventsKey>} Event key.
     * @protected
     */
    onInternal(type, listener) {
      if (Array.isArray(type)) {
        const len = type.length;
        const keys = new Array(len);
        for (let i = 0; i < len; ++i) {
          keys[i] = listen(this, type[i], listener);
        }
        return keys;
      } else {
        return listen(this, /** @type {string} */ (type), listener);
      }
    }

    /**
     * @param {string|Array<string>} type Type.
     * @param {function((Event|import("./events/Event").default)): ?} listener Listener.
     * @return {import("./events.js").EventsKey|Array<import("./events.js").EventsKey>} Event key.
     * @protected
     */
    onceInternal(type, listener) {
      let key;
      if (Array.isArray(type)) {
        const len = type.length;
        key = new Array(len);
        for (let i = 0; i < len; ++i) {
          key[i] = listenOnce(this, type[i], listener);
        }
      } else {
        key = listenOnce(this, /** @type {string} */ (type), listener);
      }
      /** @type {Object} */ (listener).ol_key = key;
      return key;
    }

    /**
     * Unlisten for a certain type of event.
     * @param {string|Array<string>} type Type.
     * @param {function((Event|import("./events/Event").default)): ?} listener Listener.
     * @protected
     */
    unInternal(type, listener) {
      const key = /** @type {Object} */ (listener).ol_key;
      if (key) {
        unByKey(key);
      } else if (Array.isArray(type)) {
        for (let i = 0, ii = type.length; i < ii; ++i) {
          this.removeEventListener(type[i], listener);
        }
      } else {
        this.removeEventListener(type, listener);
      }
    }
  }

  /**
   * Listen for a certain type of event.
   * @function
   * @param {string|Array<string>} type The event type or array of event types.
   * @param {function((Event|import("./events/Event").default)): ?} listener The listener function.
   * @return {import("./events.js").EventsKey|Array<import("./events.js").EventsKey>} Unique key for the listener. If
   *     called with an array of event types as the first argument, the return
   *     will be an array of keys.
   * @api
   */
  Observable.prototype.on;

  /**
   * Listen once for a certain type of event.
   * @function
   * @param {string|Array<string>} type The event type or array of event types.
   * @param {function((Event|import("./events/Event").default)): ?} listener The listener function.
   * @return {import("./events.js").EventsKey|Array<import("./events.js").EventsKey>} Unique key for the listener. If
   *     called with an array of event types as the first argument, the return
   *     will be an array of keys.
   * @api
   */
  Observable.prototype.once;

  /**
   * Unlisten for a certain type of event.
   * @function
   * @param {string|Array<string>} type The event type or array of event types.
   * @param {function((Event|import("./events/Event").default)): ?} listener The listener function.
   * @api
   */
  Observable.prototype.un;

  /**
   * Removes an event listener using the key returned by `on()` or `once()`.
   * @param {import("./events.js").EventsKey|Array<import("./events.js").EventsKey>} key The key returned by `on()`
   *     or `once()` (or an array of keys).
   * @api
   */
  function unByKey(key) {
    if (Array.isArray(key)) {
      for (let i = 0, ii = key.length; i < ii; ++i) {
        unlistenByKey(key[i]);
      }
    } else {
      unlistenByKey(/** @type {import("./events.js").EventsKey} */ (key));
    }
  }

  /**
   * @module ol/util
   */

  /**
   * @return {never} Any return.
   */
  function abstract() {
    throw new Error('Unimplemented abstract method.');
  }

  /**
   * Counter for getUid.
   * @type {number}
   * @private
   */
  let uidCounter_ = 0;

  /**
   * Gets a unique ID for an object. This mutates the object so that further calls
   * with the same object as a parameter returns the same value. Unique IDs are generated
   * as a strictly increasing sequence. Adapted from goog.getUid.
   *
   * @param {Object} obj The object to get the unique ID for.
   * @return {string} The unique ID for the object.
   * @api
   */
  function getUid(obj) {
    return obj.ol_uid || (obj.ol_uid = String(++uidCounter_));
  }

  /**
   * OpenLayers version.
   * @type {string}
   */
  const VERSION = 'latest';

  /**
   * @module ol/Object
   */

  /**
   * @classdesc
   * Events emitted by {@link module:ol/Object~BaseObject} instances are instances of this type.
   */
  class ObjectEvent extends BaseEvent {
    /**
     * @param {string} type The event type.
     * @param {string} key The property name.
     * @param {*} oldValue The old value for `key`.
     */
    constructor(type, key, oldValue) {
      super(type);

      /**
       * The name of the property whose value is changing.
       * @type {string}
       * @api
       */
      this.key = key;

      /**
       * The old value. To get the new value use `e.target.get(e.key)` where
       * `e` is the event object.
       * @type {*}
       * @api
       */
      this.oldValue = oldValue;
    }
  }

  /***
   * @template Return
   * @typedef {import("./Observable").OnSignature<import("./Observable").EventTypes, import("./events/Event.js").default, Return> &
   *    import("./Observable").OnSignature<import("./ObjectEventType").Types, ObjectEvent, Return> &
   *    import("./Observable").CombinedOnSignature<import("./Observable").EventTypes|import("./ObjectEventType").Types, Return>} ObjectOnSignature
   */

  /**
   * @classdesc
   * Abstract base class; normally only used for creating subclasses and not
   * instantiated in apps.
   * Most non-trivial classes inherit from this.
   *
   * This extends {@link module:ol/Observable~Observable} with observable
   * properties, where each property is observable as well as the object as a
   * whole.
   *
   * Classes that inherit from this have pre-defined properties, to which you can
   * add your owns. The pre-defined properties are listed in this documentation as
   * 'Observable Properties', and have their own accessors; for example,
   * {@link module:ol/Map~Map} has a `target` property, accessed with
   * `getTarget()` and changed with `setTarget()`. Not all properties are however
   * settable. There are also general-purpose accessors `get()` and `set()`. For
   * example, `get('target')` is equivalent to `getTarget()`.
   *
   * The `set` accessors trigger a change event, and you can monitor this by
   * registering a listener. For example, {@link module:ol/View~View} has a
   * `center` property, so `view.on('change:center', function(evt) {...});` would
   * call the function whenever the value of the center property changes. Within
   * the function, `evt.target` would be the view, so `evt.target.getCenter()`
   * would return the new center.
   *
   * You can add your own observable properties with
   * `object.set('prop', 'value')`, and retrieve that with `object.get('prop')`.
   * You can listen for changes on that property value with
   * `object.on('change:prop', listener)`. You can get a list of all
   * properties with {@link module:ol/Object~BaseObject#getProperties}.
   *
   * Note that the observable properties are separate from standard JS properties.
   * You can, for example, give your map object a title with
   * `map.title='New title'` and with `map.set('title', 'Another title')`. The
   * first will be a `hasOwnProperty`; the second will appear in
   * `getProperties()`. Only the second is observable.
   *
   * Properties can be deleted by using the unset method. E.g.
   * object.unset('foo').
   *
   * @fires ObjectEvent
   * @api
   */
  class BaseObject extends Observable {
    /**
     * @param {Object<string, *>} [values] An object with key-value pairs.
     */
    constructor(values) {
      super();

      /***
       * @type {ObjectOnSignature<import("./events").EventsKey>}
       */
      this.on;

      /***
       * @type {ObjectOnSignature<import("./events").EventsKey>}
       */
      this.once;

      /***
       * @type {ObjectOnSignature<void>}
       */
      this.un;

      // Call {@link module:ol/util.getUid} to ensure that the order of objects' ids is
      // the same as the order in which they were created.  This also helps to
      // ensure that object properties are always added in the same order, which
      // helps many JavaScript engines generate faster code.
      getUid(this);

      /**
       * @private
       * @type {Object<string, *>}
       */
      this.values_ = null;

      if (values !== undefined) {
        this.setProperties(values);
      }
    }

    /**
     * Gets a value.
     * @param {string} key Key name.
     * @return {*} Value.
     * @api
     */
    get(key) {
      let value;
      if (this.values_ && this.values_.hasOwnProperty(key)) {
        value = this.values_[key];
      }
      return value;
    }

    /**
     * Get a list of object property names.
     * @return {Array<string>} List of property names.
     * @api
     */
    getKeys() {
      return (this.values_ && Object.keys(this.values_)) || [];
    }

    /**
     * Get an object of all property names and values.
     * @return {Object<string, *>} Object.
     * @api
     */
    getProperties() {
      return (this.values_ && Object.assign({}, this.values_)) || {};
    }

    /**
     * @return {boolean} The object has properties.
     */
    hasProperties() {
      return !!this.values_;
    }

    /**
     * @param {string} key Key name.
     * @param {*} oldValue Old value.
     */
    notify(key, oldValue) {
      let eventType;
      eventType = `change:${key}`;
      if (this.hasListener(eventType)) {
        this.dispatchEvent(new ObjectEvent(eventType, key, oldValue));
      }
      eventType = ObjectEventType.PROPERTYCHANGE;
      if (this.hasListener(eventType)) {
        this.dispatchEvent(new ObjectEvent(eventType, key, oldValue));
      }
    }

    /**
     * @param {string} key Key name.
     * @param {import("./events.js").Listener} listener Listener.
     */
    addChangeListener(key, listener) {
      this.addEventListener(`change:${key}`, listener);
    }

    /**
     * @param {string} key Key name.
     * @param {import("./events.js").Listener} listener Listener.
     */
    removeChangeListener(key, listener) {
      this.removeEventListener(`change:${key}`, listener);
    }

    /**
     * Sets a value.
     * @param {string} key Key name.
     * @param {*} value Value.
     * @param {boolean} [silent] Update without triggering an event.
     * @api
     */
    set(key, value, silent) {
      const values = this.values_ || (this.values_ = {});
      if (silent) {
        values[key] = value;
      } else {
        const oldValue = values[key];
        values[key] = value;
        if (oldValue !== value) {
          this.notify(key, oldValue);
        }
      }
    }

    /**
     * Sets a collection of key-value pairs.  Note that this changes any existing
     * properties and adds new ones (it does not remove any existing properties).
     * @param {Object<string, *>} values Values.
     * @param {boolean} [silent] Update without triggering an event.
     * @api
     */
    setProperties(values, silent) {
      for (const key in values) {
        this.set(key, values[key], silent);
      }
    }

    /**
     * Apply any properties from another object without triggering events.
     * @param {BaseObject} source The source object.
     * @protected
     */
    applyProperties(source) {
      if (!source.values_) {
        return;
      }
      Object.assign(this.values_ || (this.values_ = {}), source.values_);
    }

    /**
     * Unsets a property.
     * @param {string} key Key name.
     * @param {boolean} [silent] Unset without triggering an event.
     * @api
     */
    unset(key, silent) {
      if (this.values_ && key in this.values_) {
        const oldValue = this.values_[key];
        delete this.values_[key];
        if (isEmpty$1(this.values_)) {
          this.values_ = null;
        }
        if (!silent) {
          this.notify(key, oldValue);
        }
      }
    }
  }

  /**
   * @module ol/CollectionEventType
   */

  /**
   * @enum {string}
   */
  var CollectionEventType = {
    /**
     * Triggered when an item is added to the collection.
     * @event module:ol/Collection.CollectionEvent#add
     * @api
     */
    ADD: 'add',
    /**
     * Triggered when an item is removed from the collection.
     * @event module:ol/Collection.CollectionEvent#remove
     * @api
     */
    REMOVE: 'remove',
  };

  /**
   * @module ol/Collection
   */

  /**
   * @enum {string}
   * @private
   */
  const Property$5 = {
    LENGTH: 'length',
  };

  /**
   * @classdesc
   * Events emitted by {@link module:ol/Collection~Collection} instances are instances of this
   * type.
   * @template T
   */
  class CollectionEvent extends BaseEvent {
    /**
     * @param {import("./CollectionEventType.js").default} type Type.
     * @param {T} element Element.
     * @param {number} index The index of the added or removed element.
     */
    constructor(type, element, index) {
      super(type);

      /**
       * The element that is added to or removed from the collection.
       * @type {T}
       * @api
       */
      this.element = element;

      /**
       * The index of the added or removed element.
       * @type {number}
       * @api
       */
      this.index = index;
    }
  }

  /***
   * @template T
   * @template Return
   * @typedef {import("./Observable").OnSignature<import("./Observable").EventTypes, import("./events/Event.js").default, Return> &
   *   import("./Observable").OnSignature<import("./ObjectEventType").Types|'change:length', import("./Object").ObjectEvent, Return> &
   *   import("./Observable").OnSignature<'add'|'remove', CollectionEvent<T>, Return> &
   *   import("./Observable").CombinedOnSignature<import("./Observable").EventTypes|import("./ObjectEventType").Types|
   *     'change:length'|'add'|'remove',Return>} CollectionOnSignature
   */

  /**
   * @typedef {Object} Options
   * @property {boolean} [unique=false] Disallow the same item from being added to
   * the collection twice.
   */

  /**
   * @classdesc
   * An expanded version of standard JS Array, adding convenience methods for
   * manipulation. Add and remove changes to the Collection trigger a Collection
   * event. Note that this does not cover changes to the objects _within_ the
   * Collection; they trigger events on the appropriate object, not on the
   * Collection as a whole.
   *
   * @fires CollectionEvent
   *
   * @template T
   * @api
   */
  class Collection extends BaseObject {
    /**
     * @param {Array<T>} [array] Array.
     * @param {Options} [options] Collection options.
     */
    constructor(array, options) {
      super();

      /***
       * @type {CollectionOnSignature<T, import("./events").EventsKey>}
       */
      this.on;

      /***
       * @type {CollectionOnSignature<T, import("./events").EventsKey>}
       */
      this.once;

      /***
       * @type {CollectionOnSignature<T, void>}
       */
      this.un;

      options = options || {};

      /**
       * @private
       * @type {boolean}
       */
      this.unique_ = !!options.unique;

      /**
       * @private
       * @type {!Array<T>}
       */
      this.array_ = array ? array : [];

      if (this.unique_) {
        for (let i = 0, ii = this.array_.length; i < ii; ++i) {
          this.assertUnique_(this.array_[i], i);
        }
      }

      this.updateLength_();
    }

    /**
     * Remove all elements from the collection.
     * @api
     */
    clear() {
      while (this.getLength() > 0) {
        this.pop();
      }
    }

    /**
     * Add elements to the collection.  This pushes each item in the provided array
     * to the end of the collection.
     * @param {!Array<T>} arr Array.
     * @return {Collection<T>} This collection.
     * @api
     */
    extend(arr) {
      for (let i = 0, ii = arr.length; i < ii; ++i) {
        this.push(arr[i]);
      }
      return this;
    }

    /**
     * Iterate over each element, calling the provided callback.
     * @param {function(T, number, Array<T>): *} f The function to call
     *     for every element. This function takes 3 arguments (the element, the
     *     index and the array). The return value is ignored.
     * @api
     */
    forEach(f) {
      const array = this.array_;
      for (let i = 0, ii = array.length; i < ii; ++i) {
        f(array[i], i, array);
      }
    }

    /**
     * Get a reference to the underlying Array object. Warning: if the array
     * is mutated, no events will be dispatched by the collection, and the
     * collection's "length" property won't be in sync with the actual length
     * of the array.
     * @return {!Array<T>} Array.
     * @api
     */
    getArray() {
      return this.array_;
    }

    /**
     * Get the element at the provided index.
     * @param {number} index Index.
     * @return {T} Element.
     * @api
     */
    item(index) {
      return this.array_[index];
    }

    /**
     * Get the length of this collection.
     * @return {number} The length of the array.
     * @observable
     * @api
     */
    getLength() {
      return this.get(Property$5.LENGTH);
    }

    /**
     * Insert an element at the provided index.
     * @param {number} index Index.
     * @param {T} elem Element.
     * @api
     */
    insertAt(index, elem) {
      if (index < 0 || index > this.getLength()) {
        throw new Error('Index out of bounds: ' + index);
      }
      if (this.unique_) {
        this.assertUnique_(elem);
      }
      this.array_.splice(index, 0, elem);
      this.updateLength_();
      this.dispatchEvent(
        new CollectionEvent(CollectionEventType.ADD, elem, index)
      );
    }

    /**
     * Remove the last element of the collection and return it.
     * Return `undefined` if the collection is empty.
     * @return {T|undefined} Element.
     * @api
     */
    pop() {
      return this.removeAt(this.getLength() - 1);
    }

    /**
     * Insert the provided element at the end of the collection.
     * @param {T} elem Element.
     * @return {number} New length of the collection.
     * @api
     */
    push(elem) {
      if (this.unique_) {
        this.assertUnique_(elem);
      }
      const n = this.getLength();
      this.insertAt(n, elem);
      return this.getLength();
    }

    /**
     * Remove the first occurrence of an element from the collection.
     * @param {T} elem Element.
     * @return {T|undefined} The removed element or undefined if none found.
     * @api
     */
    remove(elem) {
      const arr = this.array_;
      for (let i = 0, ii = arr.length; i < ii; ++i) {
        if (arr[i] === elem) {
          return this.removeAt(i);
        }
      }
      return undefined;
    }

    /**
     * Remove the element at the provided index and return it.
     * Return `undefined` if the collection does not contain this index.
     * @param {number} index Index.
     * @return {T|undefined} Value.
     * @api
     */
    removeAt(index) {
      if (index < 0 || index >= this.getLength()) {
        return undefined;
      }
      const prev = this.array_[index];
      this.array_.splice(index, 1);
      this.updateLength_();
      this.dispatchEvent(
        /** @type {CollectionEvent<T>} */ (
          new CollectionEvent(CollectionEventType.REMOVE, prev, index)
        )
      );
      return prev;
    }

    /**
     * Set the element at the provided index.
     * @param {number} index Index.
     * @param {T} elem Element.
     * @api
     */
    setAt(index, elem) {
      const n = this.getLength();
      if (index >= n) {
        this.insertAt(index, elem);
        return;
      }
      if (index < 0) {
        throw new Error('Index out of bounds: ' + index);
      }
      if (this.unique_) {
        this.assertUnique_(elem, index);
      }
      const prev = this.array_[index];
      this.array_[index] = elem;
      this.dispatchEvent(
        /** @type {CollectionEvent<T>} */ (
          new CollectionEvent(CollectionEventType.REMOVE, prev, index)
        )
      );
      this.dispatchEvent(
        /** @type {CollectionEvent<T>} */ (
          new CollectionEvent(CollectionEventType.ADD, elem, index)
        )
      );
    }

    /**
     * @private
     */
    updateLength_() {
      this.set(Property$5.LENGTH, this.array_.length);
    }

    /**
     * @private
     * @param {T} elem Element.
     * @param {number} [except] Optional index to ignore.
     */
    assertUnique_(elem, except) {
      for (let i = 0, ii = this.array_.length; i < ii; ++i) {
        if (this.array_[i] === elem && i !== except) {
          throw new AssertionError(58);
        }
      }
    }
  }

  /**
   * @module ol/TileState
   */

  /**
   * @enum {number}
   */
  var TileState = {
    IDLE: 0,
    LOADING: 1,
    LOADED: 2,
    /**
     * Indicates that tile loading failed
     * @type {number}
     */
    ERROR: 3,
    EMPTY: 4,
  };

  /**
   * @module ol/easing
   */

  /**
   * Start slow and speed up.
   * @param {number} t Input between 0 and 1.
   * @return {number} Output between 0 and 1.
   * @api
   */
  function easeIn(t) {
    return Math.pow(t, 3);
  }

  /**
   * Start fast and slow down.
   * @param {number} t Input between 0 and 1.
   * @return {number} Output between 0 and 1.
   * @api
   */
  function easeOut(t) {
    return 1 - easeIn(1 - t);
  }

  /**
   * Start slow, speed up, and then slow down again.
   * @param {number} t Input between 0 and 1.
   * @return {number} Output between 0 and 1.
   * @api
   */
  function inAndOut(t) {
    return 3 * t * t - 2 * t * t * t;
  }

  /**
   * Maintain a constant speed over time.
   * @param {number} t Input between 0 and 1.
   * @return {number} Output between 0 and 1.
   * @api
   */
  function linear(t) {
    return t;
  }

  /**
   * Start slow, speed up, and at the very end slow down again.  This has the
   * same general behavior as {@link module:ol/easing.inAndOut}, but the final
   * slowdown is delayed.
   * @param {number} t Input between 0 and 1.
   * @return {number} Output between 0 and 1.
   * @api
   */
  function upAndDown(t) {
    if (t < 0.5) {
      return inAndOut(2 * t);
    } else {
      return 1 - inAndOut(2 * (t - 0.5));
    }
  }

  /**
   * @module ol/Tile
   */

  /**
   * A function that takes an {@link module:ol/Tile~Tile} for the tile and a
   * `{string}` for the url as arguments. The default is
   * ```js
   * source.setTileLoadFunction(function(tile, src) {
   *   tile.getImage().src = src;
   * });
   * ```
   * For more fine grained control, the load function can use fetch or XMLHttpRequest and involve
   * error handling:
   *
   * ```js
   * import TileState from 'ol/TileState';
   *
   * source.setTileLoadFunction(function(tile, src) {
   *   const xhr = new XMLHttpRequest();
   *   xhr.responseType = 'blob';
   *   xhr.addEventListener('loadend', function (evt) {
   *     const data = this.response;
   *     if (data !== undefined) {
   *       tile.getImage().src = URL.createObjectURL(data);
   *     } else {
   *       tile.setState(TileState.ERROR);
   *     }
   *   });
   *   xhr.addEventListener('error', function () {
   *     tile.setState(TileState.ERROR);
   *   });
   *   xhr.open('GET', src);
   *   xhr.send();
   * });
   * ```
   *
   * @typedef {function(Tile, string): void} LoadFunction
   * @api
   */

  /**
   * {@link module:ol/source/Tile~TileSource} sources use a function of this type to get
   * the url that provides a tile for a given tile coordinate.
   *
   * This function takes an {@link module:ol/tilecoord~TileCoord} for the tile
   * coordinate, a `{number}` representing the pixel ratio and a
   * {@link module:ol/proj/Projection~Projection} for the projection  as arguments
   * and returns a `{string}` representing the tile URL, or undefined if no tile
   * should be requested for the passed tile coordinate.
   *
   * @typedef {function(import("./tilecoord.js").TileCoord, number,
   *           import("./proj/Projection.js").default): (string|undefined)} UrlFunction
   * @api
   */

  /**
   * @typedef {Object} Options
   * @property {number} [transition=250] A duration for tile opacity
   * transitions in milliseconds. A duration of 0 disables the opacity transition.
   * @property {boolean} [interpolate=false] Use interpolated values when resampling.  By default,
   * the nearest neighbor is used when resampling.
   * @api
   */

  /**
   * @classdesc
   * Base class for tiles.
   *
   * @abstract
   */
  class Tile extends Target {
    /**
     * @param {import("./tilecoord.js").TileCoord} tileCoord Tile coordinate.
     * @param {import("./TileState.js").default} state State.
     * @param {Options} [options] Tile options.
     */
    constructor(tileCoord, state, options) {
      super();

      options = options ? options : {};

      /**
       * @type {import("./tilecoord.js").TileCoord}
       */
      this.tileCoord = tileCoord;

      /**
       * @protected
       * @type {import("./TileState.js").default}
       */
      this.state = state;

      /**
       * An "interim" tile for this tile. The interim tile may be used while this
       * one is loading, for "smooth" transitions when changing params/dimensions
       * on the source.
       * @type {Tile}
       */
      this.interimTile = null;

      /**
       * A key assigned to the tile. This is used by the tile source to determine
       * if this tile can effectively be used, or if a new tile should be created
       * and this one be used as an interim tile for this new tile.
       * @type {string}
       */
      this.key = '';

      /**
       * The duration for the opacity transition.
       * @type {number}
       */
      this.transition_ =
        options.transition === undefined ? 250 : options.transition;

      /**
       * Lookup of start times for rendering transitions.  If the start time is
       * equal to -1, the transition is complete.
       * @type {Object<string, number>}
       */
      this.transitionStarts_ = {};

      /**
       * @type {boolean}
       */
      this.interpolate = !!options.interpolate;
    }

    /**
     * @protected
     */
    changed() {
      this.dispatchEvent(EventType$1.CHANGE);
    }

    /**
     * Called by the tile cache when the tile is removed from the cache due to expiry
     */
    release() {
      if (this.state === TileState.ERROR) {
        // to remove the `change` listener on this tile in `ol/TileQueue#handleTileChange`
        this.setState(TileState.EMPTY);
      }
    }

    /**
     * @return {string} Key.
     */
    getKey() {
      return this.key + '/' + this.tileCoord;
    }

    /**
     * Get the interim tile most suitable for rendering using the chain of interim
     * tiles. This corresponds to the  most recent tile that has been loaded, if no
     * such tile exists, the original tile is returned.
     * @return {!Tile} Best tile for rendering.
     */
    getInterimTile() {
      if (!this.interimTile) {
        //empty chain
        return this;
      }
      let tile = this.interimTile;

      // find the first loaded tile and return it. Since the chain is sorted in
      // decreasing order of creation time, there is no need to search the remainder
      // of the list (all those tiles correspond to older requests and will be
      // cleaned up by refreshInterimChain)
      do {
        if (tile.getState() == TileState.LOADED) {
          // Show tile immediately instead of fading it in after loading, because
          // the interim tile is in place already
          this.transition_ = 0;
          return tile;
        }
        tile = tile.interimTile;
      } while (tile);

      // we can not find a better tile
      return this;
    }

    /**
     * Goes through the chain of interim tiles and discards sections of the chain
     * that are no longer relevant.
     */
    refreshInterimChain() {
      if (!this.interimTile) {
        return;
      }

      let tile = this.interimTile;

      /**
       * @type {Tile}
       */
      let prev = this;

      do {
        if (tile.getState() == TileState.LOADED) {
          //we have a loaded tile, we can discard the rest of the list
          //we would could abort any LOADING tile request
          //older than this tile (i.e. any LOADING tile following this entry in the chain)
          tile.interimTile = null;
          break;
        } else if (tile.getState() == TileState.LOADING) {
          //keep this LOADING tile any loaded tiles later in the chain are
          //older than this tile, so we're still interested in the request
          prev = tile;
        } else if (tile.getState() == TileState.IDLE) {
          //the head of the list is the most current tile, we don't need
          //to start any other requests for this chain
          prev.interimTile = tile.interimTile;
        } else {
          prev = tile;
        }
        tile = prev.interimTile;
      } while (tile);
    }

    /**
     * Get the tile coordinate for this tile.
     * @return {import("./tilecoord.js").TileCoord} The tile coordinate.
     * @api
     */
    getTileCoord() {
      return this.tileCoord;
    }

    /**
     * @return {import("./TileState.js").default} State.
     */
    getState() {
      return this.state;
    }

    /**
     * Sets the state of this tile. If you write your own {@link module:ol/Tile~LoadFunction tileLoadFunction} ,
     * it is important to set the state correctly to {@link module:ol/TileState~ERROR}
     * when the tile cannot be loaded. Otherwise the tile cannot be removed from
     * the tile queue and will block other requests.
     * @param {import("./TileState.js").default} state State.
     * @api
     */
    setState(state) {
      if (this.state !== TileState.ERROR && this.state > state) {
        throw new Error('Tile load sequence violation');
      }
      this.state = state;
      this.changed();
    }

    /**
     * Load the image or retry if loading previously failed.
     * Loading is taken care of by the tile queue, and calling this method is
     * only needed for preloading or for reloading in case of an error.
     * @abstract
     * @api
     */
    load() {
      abstract();
    }

    /**
     * Get the alpha value for rendering.
     * @param {string} id An id for the renderer.
     * @param {number} time The render frame time.
     * @return {number} A number between 0 and 1.
     */
    getAlpha(id, time) {
      if (!this.transition_) {
        return 1;
      }

      let start = this.transitionStarts_[id];
      if (!start) {
        start = time;
        this.transitionStarts_[id] = start;
      } else if (start === -1) {
        return 1;
      }

      const delta = time - start + 1000 / 60; // avoid rendering at 0
      if (delta >= this.transition_) {
        return 1;
      }
      return easeIn(delta / this.transition_);
    }

    /**
     * Determine if a tile is in an alpha transition.  A tile is considered in
     * transition if tile.getAlpha() has not yet been called or has been called
     * and returned 1.
     * @param {string} id An id for the renderer.
     * @return {boolean} The tile is in transition.
     */
    inTransition(id) {
      if (!this.transition_) {
        return false;
      }
      return this.transitionStarts_[id] !== -1;
    }

    /**
     * Mark a transition as complete.
     * @param {string} id An id for the renderer.
     */
    endTransition(id) {
      if (this.transition_) {
        this.transitionStarts_[id] = -1;
      }
    }
  }

  /**
   * @module ol/DataTile
   */

  /**
   * Data that can be used with a DataTile.  For increased browser compatibility, use
   * Uint8Array instead of Uint8ClampedArray where possible.
   * @typedef {Uint8Array|Uint8ClampedArray|Float32Array|DataView} Data
   */

  /**
   * @typedef {Object} Options
   * @property {import("./tilecoord.js").TileCoord} tileCoord Tile coordinate.
   * @property {function(): Promise<Data>} loader Data loader.
   * @property {number} [transition=250] A duration for tile opacity
   * transitions in milliseconds. A duration of 0 disables the opacity transition.
   * @property {boolean} [interpolate=false] Use interpolated values when resampling.  By default,
   * the nearest neighbor is used when resampling.
   * @property {import('./size.js').Size} [size=[256, 256]] Tile size.
   * @api
   */

  class DataTile extends Tile {
    /**
     * @param {Options} options Tile options.
     */
    constructor(options) {
      const state = TileState.IDLE;

      super(options.tileCoord, state, {
        transition: options.transition,
        interpolate: options.interpolate,
      });

      /**
       * @type {function(): Promise<Data>}
       * @private
       */
      this.loader_ = options.loader;

      /**
       * @type {Data}
       * @private
       */
      this.data_ = null;

      /**
       * @type {Error}
       * @private
       */
      this.error_ = null;

      /**
       * @type {import('./size.js').Size}
       * @private
       */
      this.size_ = options.size || [256, 256];
    }

    /**
     * Get the tile size.
     * @return {import('./size.js').Size} Tile size.
     */
    getSize() {
      return this.size_;
    }

    /**
     * Get the data for the tile.
     * @return {Data} Tile data.
     * @api
     */
    getData() {
      return this.data_;
    }

    /**
     * Get any loading error.
     * @return {Error} Loading error.
     * @api
     */
    getError() {
      return this.error_;
    }

    /**
     * Load not yet loaded URI.
     * @api
     */
    load() {
      if (this.state !== TileState.IDLE && this.state !== TileState.ERROR) {
        return;
      }
      this.state = TileState.LOADING;
      this.changed();

      const self = this;
      this.loader_()
        .then(function (data) {
          self.data_ = data;
          self.state = TileState.LOADED;
          self.changed();
        })
        .catch(function (error) {
          self.error_ = error;
          self.state = TileState.ERROR;
          self.changed();
        });
    }
  }

  /**
   * @module ol/asserts
   */

  /**
   * @param {*} assertion Assertion we expected to be truthy.
   * @param {number} errorCode Error code.
   */
  function assert(assertion, errorCode) {
    if (!assertion) {
      throw new AssertionError(errorCode);
    }
  }

  /**
   * @module ol/Feature
   */

  /**
   * @typedef {typeof Feature|typeof import("./render/Feature.js").default} FeatureClass
   */

  /**
   * @typedef {Feature|import("./render/Feature.js").default} FeatureLike
   */

  /***
   * @template Return
   * @typedef {import("./Observable").OnSignature<import("./Observable").EventTypes, import("./events/Event.js").default, Return> &
   *   import("./Observable").OnSignature<import("./ObjectEventType").Types|'change:geometry', import("./Object").ObjectEvent, Return> &
   *   import("./Observable").CombinedOnSignature<import("./Observable").EventTypes|import("./ObjectEventType").Types
   *     |'change:geometry', Return>} FeatureOnSignature
   */

  /***
   * @template Geometry
   * @typedef {Object<string, *> & { geometry?: Geometry }} ObjectWithGeometry
   */

  /**
   * @classdesc
   * A vector object for geographic features with a geometry and other
   * attribute properties, similar to the features in vector file formats like
   * GeoJSON.
   *
   * Features can be styled individually with `setStyle`; otherwise they use the
   * style of their vector layer.
   *
   * Note that attribute properties are set as {@link module:ol/Object~BaseObject} properties on
   * the feature object, so they are observable, and have get/set accessors.
   *
   * Typically, a feature has a single geometry property. You can set the
   * geometry using the `setGeometry` method and get it with `getGeometry`.
   * It is possible to store more than one geometry on a feature using attribute
   * properties. By default, the geometry used for rendering is identified by
   * the property name `geometry`. If you want to use another geometry property
   * for rendering, use the `setGeometryName` method to change the attribute
   * property associated with the geometry for the feature.  For example:
   *
   * ```js
   *
   * import Feature from 'ol/Feature';
   * import Polygon from 'ol/geom/Polygon';
   * import Point from 'ol/geom/Point';
   *
   * const feature = new Feature({
   *   geometry: new Polygon(polyCoords),
   *   labelPoint: new Point(labelCoords),
   *   name: 'My Polygon',
   * });
   *
   * // get the polygon geometry
   * const poly = feature.getGeometry();
   *
   * // Render the feature as a point using the coordinates from labelPoint
   * feature.setGeometryName('labelPoint');
   *
   * // get the point geometry
   * const point = feature.getGeometry();
   * ```
   *
   * @api
   * @template {import("./geom/Geometry.js").default} [Geometry=import("./geom/Geometry.js").default]
   */
  class Feature extends BaseObject {
    /**
     * @param {Geometry|ObjectWithGeometry<Geometry>} [geometryOrProperties]
     *     You may pass a Geometry object directly, or an object literal containing
     *     properties. If you pass an object literal, you may include a Geometry
     *     associated with a `geometry` key.
     */
    constructor(geometryOrProperties) {
      super();

      /***
       * @type {FeatureOnSignature<import("./events").EventsKey>}
       */
      this.on;

      /***
       * @type {FeatureOnSignature<import("./events").EventsKey>}
       */
      this.once;

      /***
       * @type {FeatureOnSignature<void>}
       */
      this.un;

      /**
       * @private
       * @type {number|string|undefined}
       */
      this.id_ = undefined;

      /**
       * @type {string}
       * @private
       */
      this.geometryName_ = 'geometry';

      /**
       * User provided style.
       * @private
       * @type {import("./style/Style.js").StyleLike}
       */
      this.style_ = null;

      /**
       * @private
       * @type {import("./style/Style.js").StyleFunction|undefined}
       */
      this.styleFunction_ = undefined;

      /**
       * @private
       * @type {?import("./events.js").EventsKey}
       */
      this.geometryChangeKey_ = null;

      this.addChangeListener(this.geometryName_, this.handleGeometryChanged_);

      if (geometryOrProperties) {
        if (
          typeof (
            /** @type {?} */ (geometryOrProperties).getSimplifiedGeometry
          ) === 'function'
        ) {
          const geometry = /** @type {Geometry} */ (geometryOrProperties);
          this.setGeometry(geometry);
        } else {
          /** @type {Object<string, *>} */
          const properties = geometryOrProperties;
          this.setProperties(properties);
        }
      }
    }

    /**
     * Clone this feature. If the original feature has a geometry it
     * is also cloned. The feature id is not set in the clone.
     * @return {Feature<Geometry>} The clone.
     * @api
     */
    clone() {
      const clone = /** @type {Feature<Geometry>} */ (
        new Feature(this.hasProperties() ? this.getProperties() : null)
      );
      clone.setGeometryName(this.getGeometryName());
      const geometry = this.getGeometry();
      if (geometry) {
        clone.setGeometry(/** @type {Geometry} */ (geometry.clone()));
      }
      const style = this.getStyle();
      if (style) {
        clone.setStyle(style);
      }
      return clone;
    }

    /**
     * Get the feature's default geometry.  A feature may have any number of named
     * geometries.  The "default" geometry (the one that is rendered by default) is
     * set when calling {@link module:ol/Feature~Feature#setGeometry}.
     * @return {Geometry|undefined} The default geometry for the feature.
     * @api
     * @observable
     */
    getGeometry() {
      return /** @type {Geometry|undefined} */ (this.get(this.geometryName_));
    }

    /**
     * Get the feature identifier.  This is a stable identifier for the feature and
     * is either set when reading data from a remote source or set explicitly by
     * calling {@link module:ol/Feature~Feature#setId}.
     * @return {number|string|undefined} Id.
     * @api
     */
    getId() {
      return this.id_;
    }

    /**
     * Get the name of the feature's default geometry.  By default, the default
     * geometry is named `geometry`.
     * @return {string} Get the property name associated with the default geometry
     *     for this feature.
     * @api
     */
    getGeometryName() {
      return this.geometryName_;
    }

    /**
     * Get the feature's style. Will return what was provided to the
     * {@link module:ol/Feature~Feature#setStyle} method.
     * @return {import("./style/Style.js").StyleLike|undefined} The feature style.
     * @api
     */
    getStyle() {
      return this.style_;
    }

    /**
     * Get the feature's style function.
     * @return {import("./style/Style.js").StyleFunction|undefined} Return a function
     * representing the current style of this feature.
     * @api
     */
    getStyleFunction() {
      return this.styleFunction_;
    }

    /**
     * @private
     */
    handleGeometryChange_() {
      this.changed();
    }

    /**
     * @private
     */
    handleGeometryChanged_() {
      if (this.geometryChangeKey_) {
        unlistenByKey(this.geometryChangeKey_);
        this.geometryChangeKey_ = null;
      }
      const geometry = this.getGeometry();
      if (geometry) {
        this.geometryChangeKey_ = listen(
          geometry,
          EventType$1.CHANGE,
          this.handleGeometryChange_,
          this
        );
      }
      this.changed();
    }

    /**
     * Set the default geometry for the feature.  This will update the property
     * with the name returned by {@link module:ol/Feature~Feature#getGeometryName}.
     * @param {Geometry|undefined} geometry The new geometry.
     * @api
     * @observable
     */
    setGeometry(geometry) {
      this.set(this.geometryName_, geometry);
    }

    /**
     * Set the style for the feature to override the layer style.  This can be a
     * single style object, an array of styles, or a function that takes a
     * resolution and returns an array of styles. To unset the feature style, call
     * `setStyle()` without arguments or a falsey value.
     * @param {import("./style/Style.js").StyleLike} [style] Style for this feature.
     * @api
     * @fires module:ol/events/Event~BaseEvent#event:change
     */
    setStyle(style) {
      this.style_ = style;
      this.styleFunction_ = !style ? undefined : createStyleFunction(style);
      this.changed();
    }

    /**
     * Set the feature id.  The feature id is considered stable and may be used when
     * requesting features or comparing identifiers returned from a remote source.
     * The feature id can be used with the
     * {@link module:ol/source/Vector~VectorSource#getFeatureById} method.
     * @param {number|string|undefined} id The feature id.
     * @api
     * @fires module:ol/events/Event~BaseEvent#event:change
     */
    setId(id) {
      this.id_ = id;
      this.changed();
    }

    /**
     * Set the property name to be used when getting the feature's default geometry.
     * When calling {@link module:ol/Feature~Feature#getGeometry}, the value of the property with
     * this name will be returned.
     * @param {string} name The property name of the default geometry.
     * @api
     */
    setGeometryName(name) {
      this.removeChangeListener(this.geometryName_, this.handleGeometryChanged_);
      this.geometryName_ = name;
      this.addChangeListener(this.geometryName_, this.handleGeometryChanged_);
      this.handleGeometryChanged_();
    }
  }

  /**
   * Convert the provided object into a feature style function.  Functions passed
   * through unchanged.  Arrays of Style or single style objects wrapped
   * in a new feature style function.
   * @param {!import("./style/Style.js").StyleFunction|!Array<import("./style/Style.js").default>|!import("./style/Style.js").default} obj
   *     A feature style function, a single style, or an array of styles.
   * @return {import("./style/Style.js").StyleFunction} A style function.
   */
  function createStyleFunction(obj) {
    if (typeof obj === 'function') {
      return obj;
    } else {
      /**
       * @type {Array<import("./style/Style.js").default>}
       */
      let styles;
      if (Array.isArray(obj)) {
        styles = obj;
      } else {
        assert(typeof (/** @type {?} */ (obj).getZIndex) === 'function', 41); // Expected an `import("./style/Style.js").Style` or an array of `import("./style/Style.js").Style`
        const style = /** @type {import("./style/Style.js").default} */ (obj);
        styles = [style];
      }
      return function () {
        return styles;
      };
    }
  }

  /**
   * @module ol/has
   */

  const ua =
    typeof navigator !== 'undefined' && typeof navigator.userAgent !== 'undefined'
      ? navigator.userAgent.toLowerCase()
      : '';

  /**
   * User agent string says we are dealing with Firefox as browser.
   * @type {boolean}
   */
  const FIREFOX = ua.includes('firefox');

  /**
   * User agent string says we are dealing with Safari as browser.
   * @type {boolean}
   */
  const SAFARI = ua.includes('safari') && !ua.includes('chrom');

  /**
   * https://bugs.webkit.org/show_bug.cgi?id=237906
   * @type {boolean}
   */
  const SAFARI_BUG_237906 =
    SAFARI &&
    (ua.includes('version/15.4') ||
      /cpu (os|iphone os) 15_4 like mac os x/.test(ua));

  /**
   * User agent string says we are dealing with a WebKit engine.
   * @type {boolean}
   */
  const WEBKIT = ua.includes('webkit') && !ua.includes('edge');

  /**
   * User agent string says we are dealing with a Mac as platform.
   * @type {boolean}
   */
  const MAC = ua.includes('macintosh');

  /**
   * The ratio between physical pixels and device-independent pixels
   * (dips) on the device (`window.devicePixelRatio`).
   * @const
   * @type {number}
   * @api
   */
  const DEVICE_PIXEL_RATIO =
    typeof devicePixelRatio !== 'undefined' ? devicePixelRatio : 1;

  /**
   * The execution context is a worker with OffscreenCanvas available.
   * @const
   * @type {boolean}
   */
  const WORKER_OFFSCREEN_CANVAS =
    typeof WorkerGlobalScope !== 'undefined' &&
    typeof OffscreenCanvas !== 'undefined' &&
    self instanceof WorkerGlobalScope; //eslint-disable-line

  /**
   * Image.prototype.decode() is supported.
   * @type {boolean}
   */
  const IMAGE_DECODE =
    typeof Image !== 'undefined' && Image.prototype.decode;

  /**
   * @type {boolean}
   */
  const PASSIVE_EVENT_LISTENERS = (function () {
    let passive = false;
    try {
      const options = Object.defineProperty({}, 'passive', {
        get: function () {
          passive = true;
        },
      });

      window.addEventListener('_', null, options);
      window.removeEventListener('_', null, options);
    } catch (error) {
      // passive not supported
    }
    return passive;
  })();

  /**
   * @module ol/transform
   */

  /**
   * An array representing an affine 2d transformation for use with
   * {@link module:ol/transform} functions. The array has 6 elements.
   * @typedef {!Array<number>} Transform
   * @api
   */

  /**
   * Collection of affine 2d transformation functions. The functions work on an
   * array of 6 elements. The element order is compatible with the [SVGMatrix
   * interface](https://developer.mozilla.org/en-US/docs/Web/API/SVGMatrix) and is
   * a subset (elements a to f) of a 3×3 matrix:
   * ```
   * [ a c e ]
   * [ b d f ]
   * [ 0 0 1 ]
   * ```
   */

  /**
   * @private
   * @type {Transform}
   */
  const tmp_ = new Array(6);

  /**
   * Create an identity transform.
   * @return {!Transform} Identity transform.
   */
  function create$2() {
    return [1, 0, 0, 1, 0, 0];
  }

  /**
   * Resets the given transform to an identity transform.
   * @param {!Transform} transform Transform.
   * @return {!Transform} Transform.
   */
  function reset(transform) {
    return set(transform, 1, 0, 0, 1, 0, 0);
  }

  /**
   * Multiply the underlying matrices of two transforms and return the result in
   * the first transform.
   * @param {!Transform} transform1 Transform parameters of matrix 1.
   * @param {!Transform} transform2 Transform parameters of matrix 2.
   * @return {!Transform} transform1 multiplied with transform2.
   */
  function multiply(transform1, transform2) {
    const a1 = transform1[0];
    const b1 = transform1[1];
    const c1 = transform1[2];
    const d1 = transform1[3];
    const e1 = transform1[4];
    const f1 = transform1[5];
    const a2 = transform2[0];
    const b2 = transform2[1];
    const c2 = transform2[2];
    const d2 = transform2[3];
    const e2 = transform2[4];
    const f2 = transform2[5];

    transform1[0] = a1 * a2 + c1 * b2;
    transform1[1] = b1 * a2 + d1 * b2;
    transform1[2] = a1 * c2 + c1 * d2;
    transform1[3] = b1 * c2 + d1 * d2;
    transform1[4] = a1 * e2 + c1 * f2 + e1;
    transform1[5] = b1 * e2 + d1 * f2 + f1;

    return transform1;
  }

  /**
   * Set the transform components a-f on a given transform.
   * @param {!Transform} transform Transform.
   * @param {number} a The a component of the transform.
   * @param {number} b The b component of the transform.
   * @param {number} c The c component of the transform.
   * @param {number} d The d component of the transform.
   * @param {number} e The e component of the transform.
   * @param {number} f The f component of the transform.
   * @return {!Transform} Matrix with transform applied.
   */
  function set(transform, a, b, c, d, e, f) {
    transform[0] = a;
    transform[1] = b;
    transform[2] = c;
    transform[3] = d;
    transform[4] = e;
    transform[5] = f;
    return transform;
  }

  /**
   * Set transform on one matrix from another matrix.
   * @param {!Transform} transform1 Matrix to set transform to.
   * @param {!Transform} transform2 Matrix to set transform from.
   * @return {!Transform} transform1 with transform from transform2 applied.
   */
  function setFromArray(transform1, transform2) {
    transform1[0] = transform2[0];
    transform1[1] = transform2[1];
    transform1[2] = transform2[2];
    transform1[3] = transform2[3];
    transform1[4] = transform2[4];
    transform1[5] = transform2[5];
    return transform1;
  }

  /**
   * Transforms the given coordinate with the given transform returning the
   * resulting, transformed coordinate. The coordinate will be modified in-place.
   *
   * @param {Transform} transform The transformation.
   * @param {import("./coordinate.js").Coordinate|import("./pixel.js").Pixel} coordinate The coordinate to transform.
   * @return {import("./coordinate.js").Coordinate|import("./pixel.js").Pixel} return coordinate so that operations can be
   *     chained together.
   */
  function apply(transform, coordinate) {
    const x = coordinate[0];
    const y = coordinate[1];
    coordinate[0] = transform[0] * x + transform[2] * y + transform[4];
    coordinate[1] = transform[1] * x + transform[3] * y + transform[5];
    return coordinate;
  }

  /**
   * Applies rotation to the given transform.
   * @param {!Transform} transform Transform.
   * @param {number} angle Angle in radians.
   * @return {!Transform} The rotated transform.
   */
  function rotate$2(transform, angle) {
    const cos = Math.cos(angle);
    const sin = Math.sin(angle);
    return multiply(transform, set(tmp_, cos, sin, -sin, cos, 0, 0));
  }

  /**
   * Applies scale to a given transform.
   * @param {!Transform} transform Transform.
   * @param {number} x Scale factor x.
   * @param {number} y Scale factor y.
   * @return {!Transform} The scaled transform.
   */
  function scale$3(transform, x, y) {
    return multiply(transform, set(tmp_, x, 0, 0, y, 0, 0));
  }

  /**
   * Creates a scale transform.
   * @param {!Transform} target Transform to overwrite.
   * @param {number} x Scale factor x.
   * @param {number} y Scale factor y.
   * @return {!Transform} The scale transform.
   */
  function makeScale(target, x, y) {
    return set(target, x, 0, 0, y, 0, 0);
  }

  /**
   * Applies translation to the given transform.
   * @param {!Transform} transform Transform.
   * @param {number} dx Translation x.
   * @param {number} dy Translation y.
   * @return {!Transform} The translated transform.
   */
  function translate$1(transform, dx, dy) {
    return multiply(transform, set(tmp_, 1, 0, 0, 1, dx, dy));
  }

  /**
   * Creates a composite transform given an initial translation, scale, rotation, and
   * final translation (in that order only, not commutative).
   * @param {!Transform} transform The transform (will be modified in place).
   * @param {number} dx1 Initial translation x.
   * @param {number} dy1 Initial translation y.
   * @param {number} sx Scale factor x.
   * @param {number} sy Scale factor y.
   * @param {number} angle Rotation (in counter-clockwise radians).
   * @param {number} dx2 Final translation x.
   * @param {number} dy2 Final translation y.
   * @return {!Transform} The composite transform.
   */
  function compose(transform, dx1, dy1, sx, sy, angle, dx2, dy2) {
    const sin = Math.sin(angle);
    const cos = Math.cos(angle);
    transform[0] = sx * cos;
    transform[1] = sy * sin;
    transform[2] = -sx * sin;
    transform[3] = sy * cos;
    transform[4] = dx2 * sx * cos - dy2 * sx * sin + dx1;
    transform[5] = dx2 * sy * sin + dy2 * sy * cos + dy1;
    return transform;
  }

  /**
   * Creates a composite transform given an initial translation, scale, rotation, and
   * final translation (in that order only, not commutative). The resulting transform
   * string can be applied as `transform` property of an HTMLElement's style.
   * @param {number} dx1 Initial translation x.
   * @param {number} dy1 Initial translation y.
   * @param {number} sx Scale factor x.
   * @param {number} sy Scale factor y.
   * @param {number} angle Rotation (in counter-clockwise radians).
   * @param {number} dx2 Final translation x.
   * @param {number} dy2 Final translation y.
   * @return {string} The composite css transform.
   * @api
   */
  function composeCssTransform(dx1, dy1, sx, sy, angle, dx2, dy2) {
    return toString$3(compose(create$2(), dx1, dy1, sx, sy, angle, dx2, dy2));
  }

  /**
   * Invert the given transform.
   * @param {!Transform} source The source transform to invert.
   * @return {!Transform} The inverted (source) transform.
   */
  function invert(source) {
    return makeInverse(source, source);
  }

  /**
   * Invert the given transform.
   * @param {!Transform} target Transform to be set as the inverse of
   *     the source transform.
   * @param {!Transform} source The source transform to invert.
   * @return {!Transform} The inverted (target) transform.
   */
  function makeInverse(target, source) {
    const det = determinant(source);
    assert(det !== 0, 32); // Transformation matrix cannot be inverted

    const a = source[0];
    const b = source[1];
    const c = source[2];
    const d = source[3];
    const e = source[4];
    const f = source[5];

    target[0] = d / det;
    target[1] = -b / det;
    target[2] = -c / det;
    target[3] = a / det;
    target[4] = (c * f - d * e) / det;
    target[5] = -(a * f - b * e) / det;

    return target;
  }

  /**
   * Returns the determinant of the given matrix.
   * @param {!Transform} mat Matrix.
   * @return {number} Determinant.
   */
  function determinant(mat) {
    return mat[0] * mat[3] - mat[1] * mat[2];
  }

  /**
   * @type {HTMLElement}
   * @private
   */
  let transformStringDiv;

  /**
   * A rounded string version of the transform.  This can be used
   * for CSS transforms.
   * @param {!Transform} mat Matrix.
   * @return {string} The transform as a string.
   */
  function toString$3(mat) {
    const transformString = 'matrix(' + mat.join(', ') + ')';
    if (WORKER_OFFSCREEN_CANVAS) {
      return transformString;
    }
    const node =
      transformStringDiv || (transformStringDiv = document.createElement('div'));
    node.style.transform = transformString;
    return node.style.transform;
  }

  /**
   * @module ol/extent/Relationship
   */

  /**
   * Relationship to an extent.
   * @enum {number}
   */
  var Relationship = {
    UNKNOWN: 0,
    INTERSECTING: 1,
    ABOVE: 2,
    RIGHT: 4,
    BELOW: 8,
    LEFT: 16,
  };

  /**
   * @module ol/extent
   */

  /**
   * An array of numbers representing an extent: `[minx, miny, maxx, maxy]`.
   * @typedef {Array<number>} Extent
   * @api
   */

  /**
   * Extent corner.
   * @typedef {'bottom-left' | 'bottom-right' | 'top-left' | 'top-right'} Corner
   */

  /**
   * Build an extent that includes all given coordinates.
   *
   * @param {Array<import("./coordinate.js").Coordinate>} coordinates Coordinates.
   * @return {Extent} Bounding extent.
   * @api
   */
  function boundingExtent(coordinates) {
    const extent = createEmpty();
    for (let i = 0, ii = coordinates.length; i < ii; ++i) {
      extendCoordinate(extent, coordinates[i]);
    }
    return extent;
  }

  /**
   * @param {Array<number>} xs Xs.
   * @param {Array<number>} ys Ys.
   * @param {Extent} [dest] Destination extent.
   * @private
   * @return {Extent} Extent.
   */
  function _boundingExtentXYs(xs, ys, dest) {
    const minX = Math.min.apply(null, xs);
    const minY = Math.min.apply(null, ys);
    const maxX = Math.max.apply(null, xs);
    const maxY = Math.max.apply(null, ys);
    return createOrUpdate$2(minX, minY, maxX, maxY, dest);
  }

  /**
   * Return extent increased by the provided value.
   * @param {Extent} extent Extent.
   * @param {number} value The amount by which the extent should be buffered.
   * @param {Extent} [dest] Extent.
   * @return {Extent} Extent.
   * @api
   */
  function buffer$1(extent, value, dest) {
    if (dest) {
      dest[0] = extent[0] - value;
      dest[1] = extent[1] - value;
      dest[2] = extent[2] + value;
      dest[3] = extent[3] + value;
      return dest;
    } else {
      return [
        extent[0] - value,
        extent[1] - value,
        extent[2] + value,
        extent[3] + value,
      ];
    }
  }

  /**
   * Creates a clone of an extent.
   *
   * @param {Extent} extent Extent to clone.
   * @param {Extent} [dest] Extent.
   * @return {Extent} The clone.
   */
  function clone(extent, dest) {
    if (dest) {
      dest[0] = extent[0];
      dest[1] = extent[1];
      dest[2] = extent[2];
      dest[3] = extent[3];
      return dest;
    } else {
      return extent.slice();
    }
  }

  /**
   * @param {Extent} extent Extent.
   * @param {number} x X.
   * @param {number} y Y.
   * @return {number} Closest squared distance.
   */
  function closestSquaredDistanceXY(extent, x, y) {
    let dx, dy;
    if (x < extent[0]) {
      dx = extent[0] - x;
    } else if (extent[2] < x) {
      dx = x - extent[2];
    } else {
      dx = 0;
    }
    if (y < extent[1]) {
      dy = extent[1] - y;
    } else if (extent[3] < y) {
      dy = y - extent[3];
    } else {
      dy = 0;
    }
    return dx * dx + dy * dy;
  }

  /**
   * Check if the passed coordinate is contained or on the edge of the extent.
   *
   * @param {Extent} extent Extent.
   * @param {import("./coordinate.js").Coordinate} coordinate Coordinate.
   * @return {boolean} The coordinate is contained in the extent.
   * @api
   */
  function containsCoordinate(extent, coordinate) {
    return containsXY(extent, coordinate[0], coordinate[1]);
  }

  /**
   * Check if one extent contains another.
   *
   * An extent is deemed contained if it lies completely within the other extent,
   * including if they share one or more edges.
   *
   * @param {Extent} extent1 Extent 1.
   * @param {Extent} extent2 Extent 2.
   * @return {boolean} The second extent is contained by or on the edge of the
   *     first.
   * @api
   */
  function containsExtent(extent1, extent2) {
    return (
      extent1[0] <= extent2[0] &&
      extent2[2] <= extent1[2] &&
      extent1[1] <= extent2[1] &&
      extent2[3] <= extent1[3]
    );
  }

  /**
   * Check if the passed coordinate is contained or on the edge of the extent.
   *
   * @param {Extent} extent Extent.
   * @param {number} x X coordinate.
   * @param {number} y Y coordinate.
   * @return {boolean} The x, y values are contained in the extent.
   * @api
   */
  function containsXY(extent, x, y) {
    return extent[0] <= x && x <= extent[2] && extent[1] <= y && y <= extent[3];
  }

  /**
   * Get the relationship between a coordinate and extent.
   * @param {Extent} extent The extent.
   * @param {import("./coordinate.js").Coordinate} coordinate The coordinate.
   * @return {import("./extent/Relationship.js").default} The relationship (bitwise compare with
   *     import("./extent/Relationship.js").Relationship).
   */
  function coordinateRelationship(extent, coordinate) {
    const minX = extent[0];
    const minY = extent[1];
    const maxX = extent[2];
    const maxY = extent[3];
    const x = coordinate[0];
    const y = coordinate[1];
    let relationship = Relationship.UNKNOWN;
    if (x < minX) {
      relationship = relationship | Relationship.LEFT;
    } else if (x > maxX) {
      relationship = relationship | Relationship.RIGHT;
    }
    if (y < minY) {
      relationship = relationship | Relationship.BELOW;
    } else if (y > maxY) {
      relationship = relationship | Relationship.ABOVE;
    }
    if (relationship === Relationship.UNKNOWN) {
      relationship = Relationship.INTERSECTING;
    }
    return relationship;
  }

  /**
   * Create an empty extent.
   * @return {Extent} Empty extent.
   * @api
   */
  function createEmpty() {
    return [Infinity, Infinity, -Infinity, -Infinity];
  }

  /**
   * Create a new extent or update the provided extent.
   * @param {number} minX Minimum X.
   * @param {number} minY Minimum Y.
   * @param {number} maxX Maximum X.
   * @param {number} maxY Maximum Y.
   * @param {Extent} [dest] Destination extent.
   * @return {Extent} Extent.
   */
  function createOrUpdate$2(minX, minY, maxX, maxY, dest) {
    if (dest) {
      dest[0] = minX;
      dest[1] = minY;
      dest[2] = maxX;
      dest[3] = maxY;
      return dest;
    } else {
      return [minX, minY, maxX, maxY];
    }
  }

  /**
   * Create a new empty extent or make the provided one empty.
   * @param {Extent} [dest] Extent.
   * @return {Extent} Extent.
   */
  function createOrUpdateEmpty(dest) {
    return createOrUpdate$2(Infinity, Infinity, -Infinity, -Infinity, dest);
  }

  /**
   * @param {import("./coordinate.js").Coordinate} coordinate Coordinate.
   * @param {Extent} [dest] Extent.
   * @return {Extent} Extent.
   */
  function createOrUpdateFromCoordinate(coordinate, dest) {
    const x = coordinate[0];
    const y = coordinate[1];
    return createOrUpdate$2(x, y, x, y, dest);
  }

  /**
   * @param {Array<import("./coordinate.js").Coordinate>} coordinates Coordinates.
   * @param {Extent} [dest] Extent.
   * @return {Extent} Extent.
   */
  function createOrUpdateFromCoordinates(coordinates, dest) {
    const extent = createOrUpdateEmpty(dest);
    return extendCoordinates(extent, coordinates);
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   * @param {Extent} [dest] Extent.
   * @return {Extent} Extent.
   */
  function createOrUpdateFromFlatCoordinates(
    flatCoordinates,
    offset,
    end,
    stride,
    dest
  ) {
    const extent = createOrUpdateEmpty(dest);
    return extendFlatCoordinates(extent, flatCoordinates, offset, end, stride);
  }

  /**
   * @param {Array<Array<import("./coordinate.js").Coordinate>>} rings Rings.
   * @param {Extent} [dest] Extent.
   * @return {Extent} Extent.
   */
  function createOrUpdateFromRings(rings, dest) {
    const extent = createOrUpdateEmpty(dest);
    return extendRings(extent, rings);
  }

  /**
   * Determine if two extents are equivalent.
   * @param {Extent} extent1 Extent 1.
   * @param {Extent} extent2 Extent 2.
   * @return {boolean} The two extents are equivalent.
   * @api
   */
  function equals$2(extent1, extent2) {
    return (
      extent1[0] == extent2[0] &&
      extent1[2] == extent2[2] &&
      extent1[1] == extent2[1] &&
      extent1[3] == extent2[3]
    );
  }

  /**
   * Determine if two extents are approximately equivalent.
   * @param {Extent} extent1 Extent 1.
   * @param {Extent} extent2 Extent 2.
   * @param {number} tolerance Tolerance in extent coordinate units.
   * @return {boolean} The two extents differ by less than the tolerance.
   */
  function approximatelyEquals(extent1, extent2, tolerance) {
    return (
      Math.abs(extent1[0] - extent2[0]) < tolerance &&
      Math.abs(extent1[2] - extent2[2]) < tolerance &&
      Math.abs(extent1[1] - extent2[1]) < tolerance &&
      Math.abs(extent1[3] - extent2[3]) < tolerance
    );
  }

  /**
   * Modify an extent to include another extent.
   * @param {Extent} extent1 The extent to be modified.
   * @param {Extent} extent2 The extent that will be included in the first.
   * @return {Extent} A reference to the first (extended) extent.
   * @api
   */
  function extend$1(extent1, extent2) {
    if (extent2[0] < extent1[0]) {
      extent1[0] = extent2[0];
    }
    if (extent2[2] > extent1[2]) {
      extent1[2] = extent2[2];
    }
    if (extent2[1] < extent1[1]) {
      extent1[1] = extent2[1];
    }
    if (extent2[3] > extent1[3]) {
      extent1[3] = extent2[3];
    }
    return extent1;
  }

  /**
   * @param {Extent} extent Extent.
   * @param {import("./coordinate.js").Coordinate} coordinate Coordinate.
   */
  function extendCoordinate(extent, coordinate) {
    if (coordinate[0] < extent[0]) {
      extent[0] = coordinate[0];
    }
    if (coordinate[0] > extent[2]) {
      extent[2] = coordinate[0];
    }
    if (coordinate[1] < extent[1]) {
      extent[1] = coordinate[1];
    }
    if (coordinate[1] > extent[3]) {
      extent[3] = coordinate[1];
    }
  }

  /**
   * @param {Extent} extent Extent.
   * @param {Array<import("./coordinate.js").Coordinate>} coordinates Coordinates.
   * @return {Extent} Extent.
   */
  function extendCoordinates(extent, coordinates) {
    for (let i = 0, ii = coordinates.length; i < ii; ++i) {
      extendCoordinate(extent, coordinates[i]);
    }
    return extent;
  }

  /**
   * @param {Extent} extent Extent.
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   * @return {Extent} Extent.
   */
  function extendFlatCoordinates(
    extent,
    flatCoordinates,
    offset,
    end,
    stride
  ) {
    for (; offset < end; offset += stride) {
      extendXY(extent, flatCoordinates[offset], flatCoordinates[offset + 1]);
    }
    return extent;
  }

  /**
   * @param {Extent} extent Extent.
   * @param {Array<Array<import("./coordinate.js").Coordinate>>} rings Rings.
   * @return {Extent} Extent.
   */
  function extendRings(extent, rings) {
    for (let i = 0, ii = rings.length; i < ii; ++i) {
      extendCoordinates(extent, rings[i]);
    }
    return extent;
  }

  /**
   * @param {Extent} extent Extent.
   * @param {number} x X.
   * @param {number} y Y.
   */
  function extendXY(extent, x, y) {
    extent[0] = Math.min(extent[0], x);
    extent[1] = Math.min(extent[1], y);
    extent[2] = Math.max(extent[2], x);
    extent[3] = Math.max(extent[3], y);
  }

  /**
   * This function calls `callback` for each corner of the extent. If the
   * callback returns a truthy value the function returns that value
   * immediately. Otherwise the function returns `false`.
   * @param {Extent} extent Extent.
   * @param {function(import("./coordinate.js").Coordinate): S} callback Callback.
   * @return {S|boolean} Value.
   * @template S
   */
  function forEachCorner(extent, callback) {
    let val;
    val = callback(getBottomLeft(extent));
    if (val) {
      return val;
    }
    val = callback(getBottomRight(extent));
    if (val) {
      return val;
    }
    val = callback(getTopRight(extent));
    if (val) {
      return val;
    }
    val = callback(getTopLeft(extent));
    if (val) {
      return val;
    }
    return false;
  }

  /**
   * Get the size of an extent.
   * @param {Extent} extent Extent.
   * @return {number} Area.
   * @api
   */
  function getArea$1(extent) {
    let area = 0;
    if (!isEmpty(extent)) {
      area = getWidth(extent) * getHeight(extent);
    }
    return area;
  }

  /**
   * Get the bottom left coordinate of an extent.
   * @param {Extent} extent Extent.
   * @return {import("./coordinate.js").Coordinate} Bottom left coordinate.
   * @api
   */
  function getBottomLeft(extent) {
    return [extent[0], extent[1]];
  }

  /**
   * Get the bottom right coordinate of an extent.
   * @param {Extent} extent Extent.
   * @return {import("./coordinate.js").Coordinate} Bottom right coordinate.
   * @api
   */
  function getBottomRight(extent) {
    return [extent[2], extent[1]];
  }

  /**
   * Get the center coordinate of an extent.
   * @param {Extent} extent Extent.
   * @return {import("./coordinate.js").Coordinate} Center.
   * @api
   */
  function getCenter(extent) {
    return [(extent[0] + extent[2]) / 2, (extent[1] + extent[3]) / 2];
  }

  /**
   * Get a corner coordinate of an extent.
   * @param {Extent} extent Extent.
   * @param {Corner} corner Corner.
   * @return {import("./coordinate.js").Coordinate} Corner coordinate.
   */
  function getCorner(extent, corner) {
    let coordinate;
    if (corner === 'bottom-left') {
      coordinate = getBottomLeft(extent);
    } else if (corner === 'bottom-right') {
      coordinate = getBottomRight(extent);
    } else if (corner === 'top-left') {
      coordinate = getTopLeft(extent);
    } else if (corner === 'top-right') {
      coordinate = getTopRight(extent);
    } else {
      assert(false, 13); // Invalid corner
    }
    return coordinate;
  }

  /**
   * @param {Extent} extent1 Extent 1.
   * @param {Extent} extent2 Extent 2.
   * @return {number} Enlarged area.
   */
  function getEnlargedArea(extent1, extent2) {
    const minX = Math.min(extent1[0], extent2[0]);
    const minY = Math.min(extent1[1], extent2[1]);
    const maxX = Math.max(extent1[2], extent2[2]);
    const maxY = Math.max(extent1[3], extent2[3]);
    return (maxX - minX) * (maxY - minY);
  }

  /**
   * @param {import("./coordinate.js").Coordinate} center Center.
   * @param {number} resolution Resolution.
   * @param {number} rotation Rotation.
   * @param {import("./size.js").Size} size Size.
   * @param {Extent} [dest] Destination extent.
   * @return {Extent} Extent.
   */
  function getForViewAndSize(center, resolution, rotation, size, dest) {
    const [x0, y0, x1, y1, x2, y2, x3, y3] = getRotatedViewport(
      center,
      resolution,
      rotation,
      size
    );
    return createOrUpdate$2(
      Math.min(x0, x1, x2, x3),
      Math.min(y0, y1, y2, y3),
      Math.max(x0, x1, x2, x3),
      Math.max(y0, y1, y2, y3),
      dest
    );
  }

  /**
   * @param {import("./coordinate.js").Coordinate} center Center.
   * @param {number} resolution Resolution.
   * @param {number} rotation Rotation.
   * @param {import("./size.js").Size} size Size.
   * @return {Array<number>} Linear ring representing the viewport.
   */
  function getRotatedViewport(center, resolution, rotation, size) {
    const dx = (resolution * size[0]) / 2;
    const dy = (resolution * size[1]) / 2;
    const cosRotation = Math.cos(rotation);
    const sinRotation = Math.sin(rotation);
    const xCos = dx * cosRotation;
    const xSin = dx * sinRotation;
    const yCos = dy * cosRotation;
    const ySin = dy * sinRotation;
    const x = center[0];
    const y = center[1];
    return [
      x - xCos + ySin,
      y - xSin - yCos,
      x - xCos - ySin,
      y - xSin + yCos,
      x + xCos - ySin,
      y + xSin + yCos,
      x + xCos + ySin,
      y + xSin - yCos,
      x - xCos + ySin,
      y - xSin - yCos,
    ];
  }

  /**
   * Get the height of an extent.
   * @param {Extent} extent Extent.
   * @return {number} Height.
   * @api
   */
  function getHeight(extent) {
    return extent[3] - extent[1];
  }

  /**
   * @param {Extent} extent1 Extent 1.
   * @param {Extent} extent2 Extent 2.
   * @return {number} Intersection area.
   */
  function getIntersectionArea(extent1, extent2) {
    const intersection = getIntersection(extent1, extent2);
    return getArea$1(intersection);
  }

  /**
   * Get the intersection of two extents.
   * @param {Extent} extent1 Extent 1.
   * @param {Extent} extent2 Extent 2.
   * @param {Extent} [dest] Optional extent to populate with intersection.
   * @return {Extent} Intersecting extent.
   * @api
   */
  function getIntersection(extent1, extent2, dest) {
    const intersection = dest ? dest : createEmpty();
    if (intersects$3(extent1, extent2)) {
      if (extent1[0] > extent2[0]) {
        intersection[0] = extent1[0];
      } else {
        intersection[0] = extent2[0];
      }
      if (extent1[1] > extent2[1]) {
        intersection[1] = extent1[1];
      } else {
        intersection[1] = extent2[1];
      }
      if (extent1[2] < extent2[2]) {
        intersection[2] = extent1[2];
      } else {
        intersection[2] = extent2[2];
      }
      if (extent1[3] < extent2[3]) {
        intersection[3] = extent1[3];
      } else {
        intersection[3] = extent2[3];
      }
    } else {
      createOrUpdateEmpty(intersection);
    }
    return intersection;
  }

  /**
   * @param {Extent} extent Extent.
   * @return {number} Margin.
   */
  function getMargin(extent) {
    return getWidth(extent) + getHeight(extent);
  }

  /**
   * Get the size (width, height) of an extent.
   * @param {Extent} extent The extent.
   * @return {import("./size.js").Size} The extent size.
   * @api
   */
  function getSize(extent) {
    return [extent[2] - extent[0], extent[3] - extent[1]];
  }

  /**
   * Get the top left coordinate of an extent.
   * @param {Extent} extent Extent.
   * @return {import("./coordinate.js").Coordinate} Top left coordinate.
   * @api
   */
  function getTopLeft(extent) {
    return [extent[0], extent[3]];
  }

  /**
   * Get the top right coordinate of an extent.
   * @param {Extent} extent Extent.
   * @return {import("./coordinate.js").Coordinate} Top right coordinate.
   * @api
   */
  function getTopRight(extent) {
    return [extent[2], extent[3]];
  }

  /**
   * Get the width of an extent.
   * @param {Extent} extent Extent.
   * @return {number} Width.
   * @api
   */
  function getWidth(extent) {
    return extent[2] - extent[0];
  }

  /**
   * Determine if one extent intersects another.
   * @param {Extent} extent1 Extent 1.
   * @param {Extent} extent2 Extent.
   * @return {boolean} The two extents intersect.
   * @api
   */
  function intersects$3(extent1, extent2) {
    return (
      extent1[0] <= extent2[2] &&
      extent1[2] >= extent2[0] &&
      extent1[1] <= extent2[3] &&
      extent1[3] >= extent2[1]
    );
  }

  /**
   * Determine if an extent is empty.
   * @param {Extent} extent Extent.
   * @return {boolean} Is empty.
   * @api
   */
  function isEmpty(extent) {
    return extent[2] < extent[0] || extent[3] < extent[1];
  }

  /**
   * @param {Extent} extent Extent.
   * @param {Extent} [dest] Extent.
   * @return {Extent} Extent.
   */
  function returnOrUpdate(extent, dest) {
    if (dest) {
      dest[0] = extent[0];
      dest[1] = extent[1];
      dest[2] = extent[2];
      dest[3] = extent[3];
      return dest;
    } else {
      return extent;
    }
  }

  /**
   * @param {Extent} extent Extent.
   * @param {number} value Value.
   */
  function scaleFromCenter(extent, value) {
    const deltaX = ((extent[2] - extent[0]) / 2) * (value - 1);
    const deltaY = ((extent[3] - extent[1]) / 2) * (value - 1);
    extent[0] -= deltaX;
    extent[2] += deltaX;
    extent[1] -= deltaY;
    extent[3] += deltaY;
  }

  /**
   * Determine if the segment between two coordinates intersects (crosses,
   * touches, or is contained by) the provided extent.
   * @param {Extent} extent The extent.
   * @param {import("./coordinate.js").Coordinate} start Segment start coordinate.
   * @param {import("./coordinate.js").Coordinate} end Segment end coordinate.
   * @return {boolean} The segment intersects the extent.
   */
  function intersectsSegment(extent, start, end) {
    let intersects = false;
    const startRel = coordinateRelationship(extent, start);
    const endRel = coordinateRelationship(extent, end);
    if (
      startRel === Relationship.INTERSECTING ||
      endRel === Relationship.INTERSECTING
    ) {
      intersects = true;
    } else {
      const minX = extent[0];
      const minY = extent[1];
      const maxX = extent[2];
      const maxY = extent[3];
      const startX = start[0];
      const startY = start[1];
      const endX = end[0];
      const endY = end[1];
      const slope = (endY - startY) / (endX - startX);
      let x, y;
      if (!!(endRel & Relationship.ABOVE) && !(startRel & Relationship.ABOVE)) {
        // potentially intersects top
        x = endX - (endY - maxY) / slope;
        intersects = x >= minX && x <= maxX;
      }
      if (
        !intersects &&
        !!(endRel & Relationship.RIGHT) &&
        !(startRel & Relationship.RIGHT)
      ) {
        // potentially intersects right
        y = endY - (endX - maxX) * slope;
        intersects = y >= minY && y <= maxY;
      }
      if (
        !intersects &&
        !!(endRel & Relationship.BELOW) &&
        !(startRel & Relationship.BELOW)
      ) {
        // potentially intersects bottom
        x = endX - (endY - minY) / slope;
        intersects = x >= minX && x <= maxX;
      }
      if (
        !intersects &&
        !!(endRel & Relationship.LEFT) &&
        !(startRel & Relationship.LEFT)
      ) {
        // potentially intersects left
        y = endY - (endX - minX) * slope;
        intersects = y >= minY && y <= maxY;
      }
    }
    return intersects;
  }

  /**
   * Apply a transform function to the extent.
   * @param {Extent} extent Extent.
   * @param {import("./proj.js").TransformFunction} transformFn Transform function.
   * Called with `[minX, minY, maxX, maxY]` extent coordinates.
   * @param {Extent} [dest] Destination extent.
   * @param {number} [stops] Number of stops per side used for the transform.
   * By default only the corners are used.
   * @return {Extent} Extent.
   * @api
   */
  function applyTransform(extent, transformFn, dest, stops) {
    let coordinates = [];
    if (stops > 1) {
      const width = extent[2] - extent[0];
      const height = extent[3] - extent[1];
      for (let i = 0; i < stops; ++i) {
        coordinates.push(
          extent[0] + (width * i) / stops,
          extent[1],
          extent[2],
          extent[1] + (height * i) / stops,
          extent[2] - (width * i) / stops,
          extent[3],
          extent[0],
          extent[3] - (height * i) / stops
        );
      }
    } else {
      coordinates = [
        extent[0],
        extent[1],
        extent[2],
        extent[1],
        extent[2],
        extent[3],
        extent[0],
        extent[3],
      ];
    }
    transformFn(coordinates, coordinates, 2);
    const xs = [];
    const ys = [];
    for (let i = 0, l = coordinates.length; i < l; i += 2) {
      xs.push(coordinates[i]);
      ys.push(coordinates[i + 1]);
    }
    return _boundingExtentXYs(xs, ys, dest);
  }

  /**
   * Modifies the provided extent in-place to be within the real world
   * extent.
   *
   * @param {Extent} extent Extent.
   * @param {import("./proj/Projection.js").default} projection Projection
   * @return {Extent} The extent within the real world extent.
   */
  function wrapX$2(extent, projection) {
    const projectionExtent = projection.getExtent();
    const center = getCenter(extent);
    if (
      projection.canWrapX() &&
      (center[0] < projectionExtent[0] || center[0] >= projectionExtent[2])
    ) {
      const worldWidth = getWidth(projectionExtent);
      const worldsAway = Math.floor(
        (center[0] - projectionExtent[0]) / worldWidth
      );
      const offset = worldsAway * worldWidth;
      extent[0] -= offset;
      extent[2] -= offset;
    }
    return extent;
  }

  /**
   * Fits the extent to the real world
   *
   * If the extent does not cross the anti meridian, this will return the extent in an array
   * If the extent crosses the anti meridian, the extent will be sliced, so each part fits within the
   * real world
   *
   *
   * @param {Extent} extent Extent.
   * @param {import("./proj/Projection.js").default} projection Projection
   * @return {Array<Extent>} The extent within the real world extent.
   */
  function wrapAndSliceX(extent, projection) {
    if (projection.canWrapX()) {
      const projectionExtent = projection.getExtent();

      if (!isFinite(extent[0]) || !isFinite(extent[2])) {
        return [[projectionExtent[0], extent[1], projectionExtent[2], extent[3]]];
      }

      wrapX$2(extent, projection);
      const worldWidth = getWidth(projectionExtent);

      if (getWidth(extent) > worldWidth) {
        // the extent wraps around on itself
        return [[projectionExtent[0], extent[1], projectionExtent[2], extent[3]]];
      } else if (extent[0] < projectionExtent[0]) {
        // the extent crosses the anti meridian, so it needs to be sliced
        return [
          [extent[0] + worldWidth, extent[1], projectionExtent[2], extent[3]],
          [projectionExtent[0], extent[1], extent[2], extent[3]],
        ];
      } else if (extent[2] > projectionExtent[2]) {
        // the extent crosses the anti meridian, so it needs to be sliced
        return [
          [extent[0], extent[1], projectionExtent[2], extent[3]],
          [projectionExtent[0], extent[1], extent[2] - worldWidth, extent[3]],
        ];
      }
    }

    return [extent];
  }

  /**
   * @module ol/proj/Units
   */

  /**
   * @typedef {'radians' | 'degrees' | 'ft' | 'm' | 'pixels' | 'tile-pixels' | 'us-ft'} Units
   * Projection units.
   */

  /**
   * See http://duff.ess.washington.edu/data/raster/drg/docs/geotiff.txt
   * @type {Object<number, Units>}
   */
  const unitByCode = {
    '9001': 'm',
    '9002': 'ft',
    '9003': 'us-ft',
    '9101': 'radians',
    '9102': 'degrees',
  };

  /**
   * @param {number} code Unit code.
   * @return {Units} Units.
   */
  function fromCode(code) {
    return unitByCode[code];
  }

  /**
   * @typedef {Object} MetersPerUnitLookup
   * @property {number} radians Radians
   * @property {number} degrees Degrees
   * @property {number} ft  Feet
   * @property {number} m Meters
   * @property {number} us-ft US feet
   */

  /**
   * Meters per unit lookup table.
   * @const
   * @type {MetersPerUnitLookup}
   * @api
   */
  const METERS_PER_UNIT$1 = {
    // use the radius of the Normal sphere
    'radians': 6370997 / (2 * Math.PI),
    'degrees': (2 * Math.PI * 6370997) / 360,
    'ft': 0.3048,
    'm': 1,
    'us-ft': 1200 / 3937,
  };

  /**
   * @module ol/proj/Projection
   */

  /**
   * @typedef {Object} Options
   * @property {string} code The SRS identifier code, e.g. `EPSG:4326`.
   * @property {import("./Units.js").Units} [units] Units. Required unless a
   * proj4 projection is defined for `code`.
   * @property {import("../extent.js").Extent} [extent] The validity extent for the SRS.
   * @property {string} [axisOrientation='enu'] The axis orientation as specified in Proj4.
   * @property {boolean} [global=false] Whether the projection is valid for the whole globe.
   * @property {number} [metersPerUnit] The meters per unit for the SRS.
   * If not provided, the `units` are used to get the meters per unit from the {@link METERS_PER_UNIT}
   * lookup table.
   * @property {import("../extent.js").Extent} [worldExtent] The world extent for the SRS.
   * @property {function(number, import("../coordinate.js").Coordinate):number} [getPointResolution]
   * Function to determine resolution at a point. The function is called with a
   * `number` view resolution and a {@link module:ol/coordinate~Coordinate} as arguments, and returns
   * the `number` resolution in projection units at the passed coordinate. If this is `undefined`,
   * the default {@link module:ol/proj.getPointResolution} function will be used.
   */

  /**
   * @classdesc
   * Projection definition class. One of these is created for each projection
   * supported in the application and stored in the {@link module:ol/proj} namespace.
   * You can use these in applications, but this is not required, as API params
   * and options use {@link module:ol/proj~ProjectionLike} which means the simple string
   * code will suffice.
   *
   * You can use {@link module:ol/proj.get} to retrieve the object for a particular
   * projection.
   *
   * The library includes definitions for `EPSG:4326` and `EPSG:3857`, together
   * with the following aliases:
   * * `EPSG:4326`: CRS:84, urn:ogc:def:crs:EPSG:6.6:4326,
   *     urn:ogc:def:crs:OGC:1.3:CRS84, urn:ogc:def:crs:OGC:2:84,
   *     http://www.opengis.net/gml/srs/epsg.xml#4326,
   *     urn:x-ogc:def:crs:EPSG:4326
   * * `EPSG:3857`: EPSG:102100, EPSG:102113, EPSG:900913,
   *     urn:ogc:def:crs:EPSG:6.18:3:3857,
   *     http://www.opengis.net/gml/srs/epsg.xml#3857
   *
   * If you use [proj4js](https://github.com/proj4js/proj4js), aliases can
   * be added using `proj4.defs()`. After all required projection definitions are
   * added, call the {@link module:ol/proj/proj4.register} function.
   *
   * @api
   */
  class Projection {
    /**
     * @param {Options} options Projection options.
     */
    constructor(options) {
      /**
       * @private
       * @type {string}
       */
      this.code_ = options.code;

      /**
       * Units of projected coordinates. When set to `TILE_PIXELS`, a
       * `this.extent_` and `this.worldExtent_` must be configured properly for each
       * tile.
       * @private
       * @type {import("./Units.js").Units}
       */
      this.units_ = /** @type {import("./Units.js").Units} */ (options.units);

      /**
       * Validity extent of the projection in projected coordinates. For projections
       * with `TILE_PIXELS` units, this is the extent of the tile in
       * tile pixel space.
       * @private
       * @type {import("../extent.js").Extent}
       */
      this.extent_ = options.extent !== undefined ? options.extent : null;

      /**
       * Extent of the world in EPSG:4326. For projections with
       * `TILE_PIXELS` units, this is the extent of the tile in
       * projected coordinate space.
       * @private
       * @type {import("../extent.js").Extent}
       */
      this.worldExtent_ =
        options.worldExtent !== undefined ? options.worldExtent : null;

      /**
       * @private
       * @type {string}
       */
      this.axisOrientation_ =
        options.axisOrientation !== undefined ? options.axisOrientation : 'enu';

      /**
       * @private
       * @type {boolean}
       */
      this.global_ = options.global !== undefined ? options.global : false;

      /**
       * @private
       * @type {boolean}
       */
      this.canWrapX_ = !!(this.global_ && this.extent_);

      /**
       * @private
       * @type {function(number, import("../coordinate.js").Coordinate):number|undefined}
       */
      this.getPointResolutionFunc_ = options.getPointResolution;

      /**
       * @private
       * @type {import("../tilegrid/TileGrid.js").default}
       */
      this.defaultTileGrid_ = null;

      /**
       * @private
       * @type {number|undefined}
       */
      this.metersPerUnit_ = options.metersPerUnit;
    }

    /**
     * @return {boolean} The projection is suitable for wrapping the x-axis
     */
    canWrapX() {
      return this.canWrapX_;
    }

    /**
     * Get the code for this projection, e.g. 'EPSG:4326'.
     * @return {string} Code.
     * @api
     */
    getCode() {
      return this.code_;
    }

    /**
     * Get the validity extent for this projection.
     * @return {import("../extent.js").Extent} Extent.
     * @api
     */
    getExtent() {
      return this.extent_;
    }

    /**
     * Get the units of this projection.
     * @return {import("./Units.js").Units} Units.
     * @api
     */
    getUnits() {
      return this.units_;
    }

    /**
     * Get the amount of meters per unit of this projection.  If the projection is
     * not configured with `metersPerUnit` or a units identifier, the return is
     * `undefined`.
     * @return {number|undefined} Meters.
     * @api
     */
    getMetersPerUnit() {
      return this.metersPerUnit_ || METERS_PER_UNIT$1[this.units_];
    }

    /**
     * Get the world extent for this projection.
     * @return {import("../extent.js").Extent} Extent.
     * @api
     */
    getWorldExtent() {
      return this.worldExtent_;
    }

    /**
     * Get the axis orientation of this projection.
     * Example values are:
     * enu - the default easting, northing, elevation.
     * neu - northing, easting, up - useful for "lat/long" geographic coordinates,
     *     or south orientated transverse mercator.
     * wnu - westing, northing, up - some planetary coordinate systems have
     *     "west positive" coordinate systems
     * @return {string} Axis orientation.
     * @api
     */
    getAxisOrientation() {
      return this.axisOrientation_;
    }

    /**
     * Is this projection a global projection which spans the whole world?
     * @return {boolean} Whether the projection is global.
     * @api
     */
    isGlobal() {
      return this.global_;
    }

    /**
     * Set if the projection is a global projection which spans the whole world
     * @param {boolean} global Whether the projection is global.
     * @api
     */
    setGlobal(global) {
      this.global_ = global;
      this.canWrapX_ = !!(global && this.extent_);
    }

    /**
     * @return {import("../tilegrid/TileGrid.js").default} The default tile grid.
     */
    getDefaultTileGrid() {
      return this.defaultTileGrid_;
    }

    /**
     * @param {import("../tilegrid/TileGrid.js").default} tileGrid The default tile grid.
     */
    setDefaultTileGrid(tileGrid) {
      this.defaultTileGrid_ = tileGrid;
    }

    /**
     * Set the validity extent for this projection.
     * @param {import("../extent.js").Extent} extent Extent.
     * @api
     */
    setExtent(extent) {
      this.extent_ = extent;
      this.canWrapX_ = !!(this.global_ && extent);
    }

    /**
     * Set the world extent for this projection.
     * @param {import("../extent.js").Extent} worldExtent World extent
     *     [minlon, minlat, maxlon, maxlat].
     * @api
     */
    setWorldExtent(worldExtent) {
      this.worldExtent_ = worldExtent;
    }

    /**
     * Set the getPointResolution function (see {@link module:ol/proj.getPointResolution}
     * for this projection.
     * @param {function(number, import("../coordinate.js").Coordinate):number} func Function
     * @api
     */
    setGetPointResolution(func) {
      this.getPointResolutionFunc_ = func;
    }

    /**
     * Get the custom point resolution function for this projection (if set).
     * @return {function(number, import("../coordinate.js").Coordinate):number|undefined} The custom point
     * resolution function (if set).
     */
    getPointResolutionFunc() {
      return this.getPointResolutionFunc_;
    }
  }

  /**
   * @module ol/proj/epsg3857
   */

  /**
   * Radius of WGS84 sphere
   *
   * @const
   * @type {number}
   */
  const RADIUS$1 = 6378137;

  /**
   * @const
   * @type {number}
   */
  const HALF_SIZE = Math.PI * RADIUS$1;

  /**
   * @const
   * @type {import("../extent.js").Extent}
   */
  const EXTENT$2 = [-HALF_SIZE, -HALF_SIZE, HALF_SIZE, HALF_SIZE];

  /**
   * @const
   * @type {import("../extent.js").Extent}
   */
  const WORLD_EXTENT = [-180, -85, 180, 85];

  /**
   * Maximum safe value in y direction
   * @const
   * @type {number}
   */
  const MAX_SAFE_Y = RADIUS$1 * Math.log(Math.tan(Math.PI / 2));

  /**
   * @classdesc
   * Projection object for web/spherical Mercator (EPSG:3857).
   */
  class EPSG3857Projection extends Projection {
    /**
     * @param {string} code Code.
     */
    constructor(code) {
      super({
        code: code,
        units: 'm',
        extent: EXTENT$2,
        global: true,
        worldExtent: WORLD_EXTENT,
        getPointResolution: function (resolution, point) {
          return resolution / Math.cosh(point[1] / RADIUS$1);
        },
      });
    }
  }

  /**
   * Projections equal to EPSG:3857.
   *
   * @const
   * @type {Array<import("./Projection.js").default>}
   */
  const PROJECTIONS$1 = [
    new EPSG3857Projection('EPSG:3857'),
    new EPSG3857Projection('EPSG:102100'),
    new EPSG3857Projection('EPSG:102113'),
    new EPSG3857Projection('EPSG:900913'),
    new EPSG3857Projection('http://www.opengis.net/def/crs/EPSG/0/3857'),
    new EPSG3857Projection('http://www.opengis.net/gml/srs/epsg.xml#3857'),
  ];

  /**
   * Transformation from EPSG:4326 to EPSG:3857.
   *
   * @param {Array<number>} input Input array of coordinate values.
   * @param {Array<number>} [output] Output array of coordinate values.
   * @param {number} [dimension] Dimension (default is `2`).
   * @return {Array<number>} Output array of coordinate values.
   */
  function fromEPSG4326(input, output, dimension) {
    const length = input.length;
    dimension = dimension > 1 ? dimension : 2;
    if (output === undefined) {
      if (dimension > 2) {
        // preserve values beyond second dimension
        output = input.slice();
      } else {
        output = new Array(length);
      }
    }
    for (let i = 0; i < length; i += dimension) {
      output[i] = (HALF_SIZE * input[i]) / 180;
      let y = RADIUS$1 * Math.log(Math.tan((Math.PI * (+input[i + 1] + 90)) / 360));
      if (y > MAX_SAFE_Y) {
        y = MAX_SAFE_Y;
      } else if (y < -MAX_SAFE_Y) {
        y = -MAX_SAFE_Y;
      }
      output[i + 1] = y;
    }
    return output;
  }

  /**
   * Transformation from EPSG:3857 to EPSG:4326.
   *
   * @param {Array<number>} input Input array of coordinate values.
   * @param {Array<number>} [output] Output array of coordinate values.
   * @param {number} [dimension] Dimension (default is `2`).
   * @return {Array<number>} Output array of coordinate values.
   */
  function toEPSG4326(input, output, dimension) {
    const length = input.length;
    dimension = dimension > 1 ? dimension : 2;
    if (output === undefined) {
      if (dimension > 2) {
        // preserve values beyond second dimension
        output = input.slice();
      } else {
        output = new Array(length);
      }
    }
    for (let i = 0; i < length; i += dimension) {
      output[i] = (180 * input[i]) / HALF_SIZE;
      output[i + 1] =
        (360 * Math.atan(Math.exp(input[i + 1] / RADIUS$1))) / Math.PI - 90;
    }
    return output;
  }

  /**
   * @module ol/proj/epsg4326
   */

  /**
   * Semi-major radius of the WGS84 ellipsoid.
   *
   * @const
   * @type {number}
   */
  const RADIUS = 6378137;

  /**
   * Extent of the EPSG:4326 projection which is the whole world.
   *
   * @const
   * @type {import("../extent.js").Extent}
   */
  const EXTENT$1 = [-180, -90, 180, 90];

  /**
   * @const
   * @type {number}
   */
  const METERS_PER_UNIT = (Math.PI * RADIUS) / 180;

  /**
   * @classdesc
   * Projection object for WGS84 geographic coordinates (EPSG:4326).
   *
   * Note that OpenLayers does not strictly comply with the EPSG definition.
   * The EPSG registry defines 4326 as a CRS for Latitude,Longitude (y,x).
   * OpenLayers treats EPSG:4326 as a pseudo-projection, with x,y coordinates.
   */
  class EPSG4326Projection extends Projection {
    /**
     * @param {string} code Code.
     * @param {string} [axisOrientation] Axis orientation.
     */
    constructor(code, axisOrientation) {
      super({
        code: code,
        units: 'degrees',
        extent: EXTENT$1,
        axisOrientation: axisOrientation,
        global: true,
        metersPerUnit: METERS_PER_UNIT,
        worldExtent: EXTENT$1,
      });
    }
  }

  /**
   * Projections equal to EPSG:4326.
   *
   * @const
   * @type {Array<import("./Projection.js").default>}
   */
  const PROJECTIONS = [
    new EPSG4326Projection('CRS:84'),
    new EPSG4326Projection('EPSG:4326', 'neu'),
    new EPSG4326Projection('urn:ogc:def:crs:OGC:1.3:CRS84'),
    new EPSG4326Projection('urn:ogc:def:crs:OGC:2:84'),
    new EPSG4326Projection('http://www.opengis.net/def/crs/OGC/1.3/CRS84'),
    new EPSG4326Projection('http://www.opengis.net/gml/srs/epsg.xml#4326', 'neu'),
    new EPSG4326Projection('http://www.opengis.net/def/crs/EPSG/0/4326', 'neu'),
  ];

  /**
   * @module ol/proj/projections
   */

  /**
   * @type {Object<string, import("./Projection.js").default>}
   */
  let cache = {};

  /**
   * Clear the projections cache.
   */
  function clear$1() {
    cache = {};
  }

  /**
   * Get a cached projection by code.
   * @param {string} code The code for the projection.
   * @return {import("./Projection.js").default} The projection (if cached).
   */
  function get$4(code) {
    return (
      cache[code] ||
      cache[code.replace(/urn:(x-)?ogc:def:crs:EPSG:(.*:)?(\w+)$/, 'EPSG:$3')] ||
      null
    );
  }

  /**
   * Add a projection to the cache.
   * @param {string} code The projection code.
   * @param {import("./Projection.js").default} projection The projection to cache.
   */
  function add$2(code, projection) {
    cache[code] = projection;
  }

  /**
   * @module ol/proj/transforms
   */

  /**
   * @private
   * @type {!Object<string, Object<string, import("../proj.js").TransformFunction>>}
   */
  let transforms = {};

  /**
   * Clear the transform cache.
   */
  function clear() {
    transforms = {};
  }

  /**
   * Registers a conversion function to convert coordinates from the source
   * projection to the destination projection.
   *
   * @param {import("./Projection.js").default} source Source.
   * @param {import("./Projection.js").default} destination Destination.
   * @param {import("../proj.js").TransformFunction} transformFn Transform.
   */
  function add$1(source, destination, transformFn) {
    const sourceCode = source.getCode();
    const destinationCode = destination.getCode();
    if (!(sourceCode in transforms)) {
      transforms[sourceCode] = {};
    }
    transforms[sourceCode][destinationCode] = transformFn;
  }

  /**
   * Unregisters the conversion function to convert coordinates from the source
   * projection to the destination projection.  This method is used to clean up
   * cached transforms during testing.
   *
   * @param {import("./Projection.js").default} source Source projection.
   * @param {import("./Projection.js").default} destination Destination projection.
   * @return {import("../proj.js").TransformFunction} transformFn The unregistered transform.
   */
  function remove(source, destination) {
    const sourceCode = source.getCode();
    const destinationCode = destination.getCode();
    const transform = transforms[sourceCode][destinationCode];
    delete transforms[sourceCode][destinationCode];
    if (isEmpty$1(transforms[sourceCode])) {
      delete transforms[sourceCode];
    }
    return transform;
  }

  /**
   * Get a transform given a source code and a destination code.
   * @param {string} sourceCode The code for the source projection.
   * @param {string} destinationCode The code for the destination projection.
   * @return {import("../proj.js").TransformFunction|undefined} The transform function (if found).
   */
  function get$3(sourceCode, destinationCode) {
    let transform;
    if (sourceCode in transforms && destinationCode in transforms[sourceCode]) {
      transform = transforms[sourceCode][destinationCode];
    }
    return transform;
  }

  /**
   * @module ol/math
   */

  /**
   * Takes a number and clamps it to within the provided bounds.
   * @param {number} value The input number.
   * @param {number} min The minimum value to return.
   * @param {number} max The maximum value to return.
   * @return {number} The input number if it is within bounds, or the nearest
   *     number within the bounds.
   */
  function clamp(value, min, max) {
    return Math.min(Math.max(value, min), max);
  }

  /**
   * Returns the square of the closest distance between the point (x, y) and the
   * line segment (x1, y1) to (x2, y2).
   * @param {number} x X.
   * @param {number} y Y.
   * @param {number} x1 X1.
   * @param {number} y1 Y1.
   * @param {number} x2 X2.
   * @param {number} y2 Y2.
   * @return {number} Squared distance.
   */
  function squaredSegmentDistance(x, y, x1, y1, x2, y2) {
    const dx = x2 - x1;
    const dy = y2 - y1;
    if (dx !== 0 || dy !== 0) {
      const t = ((x - x1) * dx + (y - y1) * dy) / (dx * dx + dy * dy);
      if (t > 1) {
        x1 = x2;
        y1 = y2;
      } else if (t > 0) {
        x1 += dx * t;
        y1 += dy * t;
      }
    }
    return squaredDistance$1(x, y, x1, y1);
  }

  /**
   * Returns the square of the distance between the points (x1, y1) and (x2, y2).
   * @param {number} x1 X1.
   * @param {number} y1 Y1.
   * @param {number} x2 X2.
   * @param {number} y2 Y2.
   * @return {number} Squared distance.
   */
  function squaredDistance$1(x1, y1, x2, y2) {
    const dx = x2 - x1;
    const dy = y2 - y1;
    return dx * dx + dy * dy;
  }

  /**
   * Solves system of linear equations using Gaussian elimination method.
   *
   * @param {Array<Array<number>>} mat Augmented matrix (n x n + 1 column)
   *                                     in row-major order.
   * @return {Array<number>} The resulting vector.
   */
  function solveLinearSystem(mat) {
    const n = mat.length;

    for (let i = 0; i < n; i++) {
      // Find max in the i-th column (ignoring i - 1 first rows)
      let maxRow = i;
      let maxEl = Math.abs(mat[i][i]);
      for (let r = i + 1; r < n; r++) {
        const absValue = Math.abs(mat[r][i]);
        if (absValue > maxEl) {
          maxEl = absValue;
          maxRow = r;
        }
      }

      if (maxEl === 0) {
        return null; // matrix is singular
      }

      // Swap max row with i-th (current) row
      const tmp = mat[maxRow];
      mat[maxRow] = mat[i];
      mat[i] = tmp;

      // Subtract the i-th row to make all the remaining rows 0 in the i-th column
      for (let j = i + 1; j < n; j++) {
        const coef = -mat[j][i] / mat[i][i];
        for (let k = i; k < n + 1; k++) {
          if (i == k) {
            mat[j][k] = 0;
          } else {
            mat[j][k] += coef * mat[i][k];
          }
        }
      }
    }

    // Solve Ax=b for upper triangular matrix A (mat)
    const x = new Array(n);
    for (let l = n - 1; l >= 0; l--) {
      x[l] = mat[l][n] / mat[l][l];
      for (let m = l - 1; m >= 0; m--) {
        mat[m][n] -= mat[m][l] * x[l];
      }
    }
    return x;
  }

  /**
   * Converts radians to to degrees.
   *
   * @param {number} angleInRadians Angle in radians.
   * @return {number} Angle in degrees.
   */
  function toDegrees(angleInRadians) {
    return (angleInRadians * 180) / Math.PI;
  }

  /**
   * Converts degrees to radians.
   *
   * @param {number} angleInDegrees Angle in degrees.
   * @return {number} Angle in radians.
   */
  function toRadians(angleInDegrees) {
    return (angleInDegrees * Math.PI) / 180;
  }

  /**
   * Returns the modulo of a / b, depending on the sign of b.
   *
   * @param {number} a Dividend.
   * @param {number} b Divisor.
   * @return {number} Modulo.
   */
  function modulo(a, b) {
    const r = a % b;
    return r * b < 0 ? r + b : r;
  }

  /**
   * Calculates the linearly interpolated value of x between a and b.
   *
   * @param {number} a Number
   * @param {number} b Number
   * @param {number} x Value to be interpolated.
   * @return {number} Interpolated value.
   */
  function lerp(a, b, x) {
    return a + x * (b - a);
  }

  /**
   * Returns a number with a limited number of decimal digits.
   * @param {number} n The input number.
   * @param {number} decimals The maximum number of decimal digits.
   * @return {number} The input number with a limited number of decimal digits.
   */
  function toFixed(n, decimals) {
    const factor = Math.pow(10, decimals);
    return Math.round(n * factor) / factor;
  }

  /**
   * Rounds a number to the nearest integer value considering only the given number
   * of decimal digits (with rounding on the final digit).
   * @param {number} n The input number.
   * @param {number} decimals The maximum number of decimal digits.
   * @return {number} The nearest integer.
   */
  function round(n, decimals) {
    return Math.round(toFixed(n, decimals));
  }

  /**
   * Rounds a number to the next smaller integer considering only the given number
   * of decimal digits (with rounding on the final digit).
   * @param {number} n The input number.
   * @param {number} decimals The maximum number of decimal digits.
   * @return {number} The next smaller integer.
   */
  function floor(n, decimals) {
    return Math.floor(toFixed(n, decimals));
  }

  /**
   * Rounds a number to the next bigger integer considering only the given number
   * of decimal digits (with rounding on the final digit).
   * @param {number} n The input number.
   * @param {number} decimals The maximum number of decimal digits.
   * @return {number} The next bigger integer.
   */
  function ceil(n, decimals) {
    return Math.ceil(toFixed(n, decimals));
  }

  /**
   * @module ol/string
   */

  /**
   * @param {number} number Number to be formatted
   * @param {number} width The desired width
   * @param {number} [precision] Precision of the output string (i.e. number of decimal places)
   * @return {string} Formatted string
   */
  function padNumber(number, width, precision) {
    const numberString =
      precision !== undefined ? number.toFixed(precision) : '' + number;
    let decimal = numberString.indexOf('.');
    decimal = decimal === -1 ? numberString.length : decimal;
    return decimal > width
      ? numberString
      : new Array(1 + width - decimal).join('0') + numberString;
  }

  /**
   * Adapted from https://github.com/omichelsen/compare-versions/blob/master/index.js
   * @param {string|number} v1 First version
   * @param {string|number} v2 Second version
   * @return {number} Value
   */
  function compareVersions(v1, v2) {
    const s1 = ('' + v1).split('.');
    const s2 = ('' + v2).split('.');

    for (let i = 0; i < Math.max(s1.length, s2.length); i++) {
      const n1 = parseInt(s1[i] || '0', 10);
      const n2 = parseInt(s2[i] || '0', 10);

      if (n1 > n2) {
        return 1;
      }
      if (n2 > n1) {
        return -1;
      }
    }

    return 0;
  }

  /**
   * @module ol/coordinate
   */

  /**
   * An array of numbers representing an xy coordinate. Example: `[16, 48]`.
   * @typedef {Array<number>} Coordinate
   * @api
   */

  /**
   * A function that takes a {@link module:ol/coordinate~Coordinate} and
   * transforms it into a `{string}`.
   *
   * @typedef {function((Coordinate|undefined)): string} CoordinateFormat
   * @api
   */

  /**
   * Add `delta` to `coordinate`. `coordinate` is modified in place and returned
   * by the function.
   *
   * Example:
   *
   *     import {add} from 'ol/coordinate';
   *
   *     const coord = [7.85, 47.983333];
   *     add(coord, [-2, 4]);
   *     // coord is now [5.85, 51.983333]
   *
   * @param {Coordinate} coordinate Coordinate.
   * @param {Coordinate} delta Delta.
   * @return {Coordinate} The input coordinate adjusted by
   * the given delta.
   * @api
   */
  function add(coordinate, delta) {
    coordinate[0] += +delta[0];
    coordinate[1] += +delta[1];
    return coordinate;
  }

  /**
   * Calculates the point closest to the passed coordinate on the passed circle.
   *
   * @param {Coordinate} coordinate The coordinate.
   * @param {import("./geom/Circle.js").default} circle The circle.
   * @return {Coordinate} Closest point on the circumference.
   */
  function closestOnCircle(coordinate, circle) {
    const r = circle.getRadius();
    const center = circle.getCenter();
    const x0 = center[0];
    const y0 = center[1];
    const x1 = coordinate[0];
    const y1 = coordinate[1];

    let dx = x1 - x0;
    const dy = y1 - y0;
    if (dx === 0 && dy === 0) {
      dx = 1;
    }
    const d = Math.sqrt(dx * dx + dy * dy);

    const x = x0 + (r * dx) / d;
    const y = y0 + (r * dy) / d;

    return [x, y];
  }

  /**
   * Calculates the point closest to the passed coordinate on the passed segment.
   * This is the foot of the perpendicular of the coordinate to the segment when
   * the foot is on the segment, or the closest segment coordinate when the foot
   * is outside the segment.
   *
   * @param {Coordinate} coordinate The coordinate.
   * @param {Array<Coordinate>} segment The two coordinates
   * of the segment.
   * @return {Coordinate} The foot of the perpendicular of
   * the coordinate to the segment.
   */
  function closestOnSegment(coordinate, segment) {
    const x0 = coordinate[0];
    const y0 = coordinate[1];
    const start = segment[0];
    const end = segment[1];
    const x1 = start[0];
    const y1 = start[1];
    const x2 = end[0];
    const y2 = end[1];
    const dx = x2 - x1;
    const dy = y2 - y1;
    const along =
      dx === 0 && dy === 0
        ? 0
        : (dx * (x0 - x1) + dy * (y0 - y1)) / (dx * dx + dy * dy || 0);
    let x, y;
    if (along <= 0) {
      x = x1;
      y = y1;
    } else if (along >= 1) {
      x = x2;
      y = y2;
    } else {
      x = x1 + along * dx;
      y = y1 + along * dy;
    }
    return [x, y];
  }

  /**
   * Returns a {@link module:ol/coordinate~CoordinateFormat} function that can be
   * used to format
   * a {Coordinate} to a string.
   *
   * Example without specifying the fractional digits:
   *
   *     import {createStringXY} from 'ol/coordinate';
   *
   *     const coord = [7.85, 47.983333];
   *     const stringifyFunc = createStringXY();
   *     const out = stringifyFunc(coord);
   *     // out is now '8, 48'
   *
   * Example with explicitly specifying 2 fractional digits:
   *
   *     import {createStringXY} from 'ol/coordinate';
   *
   *     const coord = [7.85, 47.983333];
   *     const stringifyFunc = createStringXY(2);
   *     const out = stringifyFunc(coord);
   *     // out is now '7.85, 47.98'
   *
   * @param {number} [fractionDigits] The number of digits to include
   *    after the decimal point. Default is `0`.
   * @return {CoordinateFormat} Coordinate format.
   * @api
   */
  function createStringXY(fractionDigits) {
    return (
      /**
       * @param {Coordinate} coordinate Coordinate.
       * @return {string} String XY.
       */
      function (coordinate) {
        return toStringXY(coordinate, fractionDigits);
      }
    );
  }

  /**
   * @param {string} hemispheres Hemispheres.
   * @param {number} degrees Degrees.
   * @param {number} [fractionDigits] The number of digits to include
   *    after the decimal point. Default is `0`.
   * @return {string} String.
   */
  function degreesToStringHDMS(hemispheres, degrees, fractionDigits) {
    const normalizedDegrees = modulo(degrees + 180, 360) - 180;
    const x = Math.abs(3600 * normalizedDegrees);
    const decimals = fractionDigits || 0;

    let deg = Math.floor(x / 3600);
    let min = Math.floor((x - deg * 3600) / 60);
    let sec = toFixed(x - deg * 3600 - min * 60, decimals);

    if (sec >= 60) {
      sec = 0;
      min += 1;
    }

    if (min >= 60) {
      min = 0;
      deg += 1;
    }

    let hdms = deg + '\u00b0';
    if (min !== 0 || sec !== 0) {
      hdms += ' ' + padNumber(min, 2) + '\u2032';
    }
    if (sec !== 0) {
      hdms += ' ' + padNumber(sec, 2, decimals) + '\u2033';
    }
    if (normalizedDegrees !== 0) {
      hdms += ' ' + hemispheres.charAt(normalizedDegrees < 0 ? 1 : 0);
    }

    return hdms;
  }

  /**
   * Transforms the given {@link module:ol/coordinate~Coordinate} to a string
   * using the given string template. The strings `{x}` and `{y}` in the template
   * will be replaced with the first and second coordinate values respectively.
   *
   * Example without specifying the fractional digits:
   *
   *     import {format} from 'ol/coordinate';
   *
   *     const coord = [7.85, 47.983333];
   *     const template = 'Coordinate is ({x}|{y}).';
   *     const out = format(coord, template);
   *     // out is now 'Coordinate is (8|48).'
   *
   * Example explicitly specifying the fractional digits:
   *
   *     import {format} from 'ol/coordinate';
   *
   *     const coord = [7.85, 47.983333];
   *     const template = 'Coordinate is ({x}|{y}).';
   *     const out = format(coord, template, 2);
   *     // out is now 'Coordinate is (7.85|47.98).'
   *
   * @param {Coordinate} coordinate Coordinate.
   * @param {string} template A template string with `{x}` and `{y}` placeholders
   *     that will be replaced by first and second coordinate values.
   * @param {number} [fractionDigits] The number of digits to include
   *    after the decimal point. Default is `0`.
   * @return {string} Formatted coordinate.
   * @api
   */
  function format(coordinate, template, fractionDigits) {
    if (coordinate) {
      return template
        .replace('{x}', coordinate[0].toFixed(fractionDigits))
        .replace('{y}', coordinate[1].toFixed(fractionDigits));
    } else {
      return '';
    }
  }

  /**
   * @param {Coordinate} coordinate1 First coordinate.
   * @param {Coordinate} coordinate2 Second coordinate.
   * @return {boolean} The two coordinates are equal.
   */
  function equals$1(coordinate1, coordinate2) {
    let equals = true;
    for (let i = coordinate1.length - 1; i >= 0; --i) {
      if (coordinate1[i] != coordinate2[i]) {
        equals = false;
        break;
      }
    }
    return equals;
  }

  /**
   * Rotate `coordinate` by `angle`. `coordinate` is modified in place and
   * returned by the function.
   *
   * Example:
   *
   *     import {rotate} from 'ol/coordinate';
   *
   *     const coord = [7.85, 47.983333];
   *     const rotateRadians = Math.PI / 2; // 90 degrees
   *     rotate(coord, rotateRadians);
   *     // coord is now [-47.983333, 7.85]
   *
   * @param {Coordinate} coordinate Coordinate.
   * @param {number} angle Angle in radian.
   * @return {Coordinate} Coordinate.
   * @api
   */
  function rotate$1(coordinate, angle) {
    const cosAngle = Math.cos(angle);
    const sinAngle = Math.sin(angle);
    const x = coordinate[0] * cosAngle - coordinate[1] * sinAngle;
    const y = coordinate[1] * cosAngle + coordinate[0] * sinAngle;
    coordinate[0] = x;
    coordinate[1] = y;
    return coordinate;
  }

  /**
   * Scale `coordinate` by `scale`. `coordinate` is modified in place and returned
   * by the function.
   *
   * Example:
   *
   *     import {scale as scaleCoordinate} from 'ol/coordinate';
   *
   *     const coord = [7.85, 47.983333];
   *     const scale = 1.2;
   *     scaleCoordinate(coord, scale);
   *     // coord is now [9.42, 57.5799996]
   *
   * @param {Coordinate} coordinate Coordinate.
   * @param {number} scale Scale factor.
   * @return {Coordinate} Coordinate.
   */
  function scale$2(coordinate, scale) {
    coordinate[0] *= scale;
    coordinate[1] *= scale;
    return coordinate;
  }

  /**
   * @param {Coordinate} coord1 First coordinate.
   * @param {Coordinate} coord2 Second coordinate.
   * @return {number} Squared distance between coord1 and coord2.
   */
  function squaredDistance(coord1, coord2) {
    const dx = coord1[0] - coord2[0];
    const dy = coord1[1] - coord2[1];
    return dx * dx + dy * dy;
  }

  /**
   * @param {Coordinate} coord1 First coordinate.
   * @param {Coordinate} coord2 Second coordinate.
   * @return {number} Distance between coord1 and coord2.
   */
  function distance(coord1, coord2) {
    return Math.sqrt(squaredDistance(coord1, coord2));
  }

  /**
   * Calculate the squared distance from a coordinate to a line segment.
   *
   * @param {Coordinate} coordinate Coordinate of the point.
   * @param {Array<Coordinate>} segment Line segment (2
   * coordinates).
   * @return {number} Squared distance from the point to the line segment.
   */
  function squaredDistanceToSegment(coordinate, segment) {
    return squaredDistance(coordinate, closestOnSegment(coordinate, segment));
  }

  /**
   * Format a geographic coordinate with the hemisphere, degrees, minutes, and
   * seconds.
   *
   * Example without specifying fractional digits:
   *
   *     import {toStringHDMS} from 'ol/coordinate';
   *
   *     const coord = [7.85, 47.983333];
   *     const out = toStringHDMS(coord);
   *     // out is now '47° 58′ 60″ N 7° 50′ 60″ E'
   *
   * Example explicitly specifying 1 fractional digit:
   *
   *     import {toStringHDMS} from 'ol/coordinate';
   *
   *     const coord = [7.85, 47.983333];
   *     const out = toStringHDMS(coord, 1);
   *     // out is now '47° 58′ 60.0″ N 7° 50′ 60.0″ E'
   *
   * @param {Coordinate} coordinate Coordinate.
   * @param {number} [fractionDigits] The number of digits to include
   *    after the decimal point. Default is `0`.
   * @return {string} Hemisphere, degrees, minutes and seconds.
   * @api
   */
  function toStringHDMS(coordinate, fractionDigits) {
    if (coordinate) {
      return (
        degreesToStringHDMS('NS', coordinate[1], fractionDigits) +
        ' ' +
        degreesToStringHDMS('EW', coordinate[0], fractionDigits)
      );
    } else {
      return '';
    }
  }

  /**
   * Format a coordinate as a comma delimited string.
   *
   * Example without specifying fractional digits:
   *
   *     import {toStringXY} from 'ol/coordinate';
   *
   *     const coord = [7.85, 47.983333];
   *     const out = toStringXY(coord);
   *     // out is now '8, 48'
   *
   * Example explicitly specifying 1 fractional digit:
   *
   *     import {toStringXY} from 'ol/coordinate';
   *
   *     const coord = [7.85, 47.983333];
   *     const out = toStringXY(coord, 1);
   *     // out is now '7.8, 48.0'
   *
   * @param {Coordinate} coordinate Coordinate.
   * @param {number} [fractionDigits] The number of digits to include
   *    after the decimal point. Default is `0`.
   * @return {string} XY.
   * @api
   */
  function toStringXY(coordinate, fractionDigits) {
    return format(coordinate, '{x}, {y}', fractionDigits);
  }

  /**
   * Modifies the provided coordinate in-place to be within the real world
   * extent. The lower projection extent boundary is inclusive, the upper one
   * exclusive.
   *
   * @param {Coordinate} coordinate Coordinate.
   * @param {import("./proj/Projection.js").default} projection Projection.
   * @return {Coordinate} The coordinate within the real world extent.
   */
  function wrapX$1(coordinate, projection) {
    if (projection.canWrapX()) {
      const worldWidth = getWidth(projection.getExtent());
      const worldsAway = getWorldsAway(coordinate, projection, worldWidth);
      if (worldsAway) {
        coordinate[0] -= worldsAway * worldWidth;
      }
    }
    return coordinate;
  }
  /**
   * @param {Coordinate} coordinate Coordinate.
   * @param {import("./proj/Projection.js").default} projection Projection.
   * @param {number} [sourceExtentWidth] Width of the source extent.
   * @return {number} Offset in world widths.
   */
  function getWorldsAway(coordinate, projection, sourceExtentWidth) {
    const projectionExtent = projection.getExtent();
    let worldsAway = 0;
    if (
      projection.canWrapX() &&
      (coordinate[0] < projectionExtent[0] || coordinate[0] > projectionExtent[2])
    ) {
      sourceExtentWidth = sourceExtentWidth || getWidth(projectionExtent);
      worldsAway = Math.floor(
        (coordinate[0] - projectionExtent[0]) / sourceExtentWidth
      );
    }
    return worldsAway;
  }

  /**
   * @module ol/sphere
   */

  /**
   * Object literal with options for the {@link getLength} or {@link getArea}
   * functions.
   * @typedef {Object} SphereMetricOptions
   * @property {import("./proj.js").ProjectionLike} [projection='EPSG:3857']
   * Projection of the  geometry.  By default, the geometry is assumed to be in
   * Web Mercator.
   * @property {number} [radius=6371008.8] Sphere radius.  By default, the
   * [mean Earth radius](https://en.wikipedia.org/wiki/Earth_radius#Mean_radius)
   * for the WGS84 ellipsoid is used.
   */

  /**
   * The mean Earth radius (1/3 * (2a + b)) for the WGS84 ellipsoid.
   * https://en.wikipedia.org/wiki/Earth_radius#Mean_radius
   * @type {number}
   */
  const DEFAULT_RADIUS = 6371008.8;

  /**
   * Get the great circle distance (in meters) between two geographic coordinates.
   * @param {Array} c1 Starting coordinate.
   * @param {Array} c2 Ending coordinate.
   * @param {number} [radius] The sphere radius to use.  Defaults to the Earth's
   *     mean radius using the WGS84 ellipsoid.
   * @return {number} The great circle distance between the points (in meters).
   * @api
   */
  function getDistance(c1, c2, radius) {
    radius = radius || DEFAULT_RADIUS;
    const lat1 = toRadians(c1[1]);
    const lat2 = toRadians(c2[1]);
    const deltaLatBy2 = (lat2 - lat1) / 2;
    const deltaLonBy2 = toRadians(c2[0] - c1[0]) / 2;
    const a =
      Math.sin(deltaLatBy2) * Math.sin(deltaLatBy2) +
      Math.sin(deltaLonBy2) *
        Math.sin(deltaLonBy2) *
        Math.cos(lat1) *
        Math.cos(lat2);
    return 2 * radius * Math.atan2(Math.sqrt(a), Math.sqrt(1 - a));
  }

  /**
   * Get the cumulative great circle length of linestring coordinates (geographic).
   * @param {Array} coordinates Linestring coordinates.
   * @param {number} radius The sphere radius to use.
   * @return {number} The length (in meters).
   */
  function getLengthInternal(coordinates, radius) {
    let length = 0;
    for (let i = 0, ii = coordinates.length; i < ii - 1; ++i) {
      length += getDistance(coordinates[i], coordinates[i + 1], radius);
    }
    return length;
  }

  /**
   * Get the spherical length of a geometry.  This length is the sum of the
   * great circle distances between coordinates.  For polygons, the length is
   * the sum of all rings.  For points, the length is zero.  For multi-part
   * geometries, the length is the sum of the length of each part.
   * @param {import("./geom/Geometry.js").default} geometry A geometry.
   * @param {SphereMetricOptions} [options] Options for the
   * length calculation.  By default, geometries are assumed to be in 'EPSG:3857'.
   * You can change this by providing a `projection` option.
   * @return {number} The spherical length (in meters).
   * @api
   */
  function getLength(geometry, options) {
    options = options || {};
    const radius = options.radius || DEFAULT_RADIUS;
    const projection = options.projection || 'EPSG:3857';
    const type = geometry.getType();
    if (type !== 'GeometryCollection') {
      geometry = geometry.clone().transform(projection, 'EPSG:4326');
    }
    let length = 0;
    let coordinates, coords, i, ii, j, jj;
    switch (type) {
      case 'Point':
      case 'MultiPoint': {
        break;
      }
      case 'LineString':
      case 'LinearRing': {
        coordinates = /** @type {import("./geom/SimpleGeometry.js").default} */ (
          geometry
        ).getCoordinates();
        length = getLengthInternal(coordinates, radius);
        break;
      }
      case 'MultiLineString':
      case 'Polygon': {
        coordinates = /** @type {import("./geom/SimpleGeometry.js").default} */ (
          geometry
        ).getCoordinates();
        for (i = 0, ii = coordinates.length; i < ii; ++i) {
          length += getLengthInternal(coordinates[i], radius);
        }
        break;
      }
      case 'MultiPolygon': {
        coordinates = /** @type {import("./geom/SimpleGeometry.js").default} */ (
          geometry
        ).getCoordinates();
        for (i = 0, ii = coordinates.length; i < ii; ++i) {
          coords = coordinates[i];
          for (j = 0, jj = coords.length; j < jj; ++j) {
            length += getLengthInternal(coords[j], radius);
          }
        }
        break;
      }
      case 'GeometryCollection': {
        const geometries =
          /** @type {import("./geom/GeometryCollection.js").default} */ (
            geometry
          ).getGeometries();
        for (i = 0, ii = geometries.length; i < ii; ++i) {
          length += getLength(geometries[i], options);
        }
        break;
      }
      default: {
        throw new Error('Unsupported geometry type: ' + type);
      }
    }
    return length;
  }

  /**
   * Returns the spherical area for a list of coordinates.
   *
   * [Reference](https://trs.jpl.nasa.gov/handle/2014/40409)
   * Robert. G. Chamberlain and William H. Duquette, "Some Algorithms for
   * Polygons on a Sphere", JPL Publication 07-03, Jet Propulsion
   * Laboratory, Pasadena, CA, June 2007
   *
   * @param {Array<import("./coordinate.js").Coordinate>} coordinates List of coordinates of a linear
   * ring. If the ring is oriented clockwise, the area will be positive,
   * otherwise it will be negative.
   * @param {number} radius The sphere radius.
   * @return {number} Area (in square meters).
   */
  function getAreaInternal(coordinates, radius) {
    let area = 0;
    const len = coordinates.length;
    let x1 = coordinates[len - 1][0];
    let y1 = coordinates[len - 1][1];
    for (let i = 0; i < len; i++) {
      const x2 = coordinates[i][0];
      const y2 = coordinates[i][1];
      area +=
        toRadians(x2 - x1) *
        (2 + Math.sin(toRadians(y1)) + Math.sin(toRadians(y2)));
      x1 = x2;
      y1 = y2;
    }
    return (area * radius * radius) / 2.0;
  }

  /**
   * Get the spherical area of a geometry.  This is the area (in meters) assuming
   * that polygon edges are segments of great circles on a sphere.
   * @param {import("./geom/Geometry.js").default} geometry A geometry.
   * @param {SphereMetricOptions} [options] Options for the area
   *     calculation.  By default, geometries are assumed to be in 'EPSG:3857'.
   *     You can change this by providing a `projection` option.
   * @return {number} The spherical area (in square meters).
   * @api
   */
  function getArea(geometry, options) {
    options = options || {};
    const radius = options.radius || DEFAULT_RADIUS;
    const projection = options.projection || 'EPSG:3857';
    const type = geometry.getType();
    if (type !== 'GeometryCollection') {
      geometry = geometry.clone().transform(projection, 'EPSG:4326');
    }
    let area = 0;
    let coordinates, coords, i, ii, j, jj;
    switch (type) {
      case 'Point':
      case 'MultiPoint':
      case 'LineString':
      case 'MultiLineString':
      case 'LinearRing': {
        break;
      }
      case 'Polygon': {
        coordinates = /** @type {import("./geom/Polygon.js").default} */ (
          geometry
        ).getCoordinates();
        area = Math.abs(getAreaInternal(coordinates[0], radius));
        for (i = 1, ii = coordinates.length; i < ii; ++i) {
          area -= Math.abs(getAreaInternal(coordinates[i], radius));
        }
        break;
      }
      case 'MultiPolygon': {
        coordinates = /** @type {import("./geom/SimpleGeometry.js").default} */ (
          geometry
        ).getCoordinates();
        for (i = 0, ii = coordinates.length; i < ii; ++i) {
          coords = coordinates[i];
          area += Math.abs(getAreaInternal(coords[0], radius));
          for (j = 1, jj = coords.length; j < jj; ++j) {
            area -= Math.abs(getAreaInternal(coords[j], radius));
          }
        }
        break;
      }
      case 'GeometryCollection': {
        const geometries =
          /** @type {import("./geom/GeometryCollection.js").default} */ (
            geometry
          ).getGeometries();
        for (i = 0, ii = geometries.length; i < ii; ++i) {
          area += getArea(geometries[i], options);
        }
        break;
      }
      default: {
        throw new Error('Unsupported geometry type: ' + type);
      }
    }
    return area;
  }

  /**
   * Returns the coordinate at the given distance and bearing from `c1`.
   *
   * @param {import("./coordinate.js").Coordinate} c1 The origin point (`[lon, lat]` in degrees).
   * @param {number} distance The great-circle distance between the origin
   *     point and the target point.
   * @param {number} bearing The bearing (in radians).
   * @param {number} [radius] The sphere radius to use.  Defaults to the Earth's
   *     mean radius using the WGS84 ellipsoid.
   * @return {import("./coordinate.js").Coordinate} The target point.
   */
  function offset(c1, distance, bearing, radius) {
    radius = radius || DEFAULT_RADIUS;
    const lat1 = toRadians(c1[1]);
    const lon1 = toRadians(c1[0]);
    const dByR = distance / radius;
    const lat = Math.asin(
      Math.sin(lat1) * Math.cos(dByR) +
        Math.cos(lat1) * Math.sin(dByR) * Math.cos(bearing)
    );
    const lon =
      lon1 +
      Math.atan2(
        Math.sin(bearing) * Math.sin(dByR) * Math.cos(lat1),
        Math.cos(dByR) - Math.sin(lat1) * Math.sin(lat)
      );
    return [toDegrees(lon), toDegrees(lat)];
  }

  /**
   * @module ol/proj
   */

  let showCoordinateWarning = true;

  /**
   * @param {boolean} [disable = true] Disable console info about `useGeographic()`
   */
  function disableCoordinateWarning(disable) {
    const hide = disable === undefined ? true : disable;
    showCoordinateWarning = !hide;
  }

  /**
   * @param {Array<number>} input Input coordinate array.
   * @param {Array<number>} [output] Output array of coordinate values.
   * @param {number} [dimension] Dimension.
   * @return {Array<number>} Output coordinate array (new array, same coordinate
   *     values).
   */
  function cloneTransform(input, output, dimension) {
    if (output !== undefined) {
      for (let i = 0, ii = input.length; i < ii; ++i) {
        output[i] = input[i];
      }
      output = output;
    } else {
      output = input.slice();
    }
    return output;
  }

  /**
   * @param {Array<number>} input Input coordinate array.
   * @param {Array<number>} [output] Output array of coordinate values.
   * @param {number} [dimension] Dimension.
   * @return {Array<number>} Input coordinate array (same array as input).
   */
  function identityTransform(input, output, dimension) {
    if (output !== undefined && input !== output) {
      for (let i = 0, ii = input.length; i < ii; ++i) {
        output[i] = input[i];
      }
      input = output;
    }
    return input;
  }

  /**
   * Add a Projection object to the list of supported projections that can be
   * looked up by their code.
   *
   * @param {Projection} projection Projection instance.
   * @api
   */
  function addProjection(projection) {
    add$2(projection.getCode(), projection);
    add$1(projection, projection, cloneTransform);
  }

  /**
   * @param {Array<Projection>} projections Projections.
   */
  function addProjections(projections) {
    projections.forEach(addProjection);
  }

  /**
   * Fetches a Projection object for the code specified.
   *
   * @param {ProjectionLike} projectionLike Either a code string which is
   *     a combination of authority and identifier such as "EPSG:4326", or an
   *     existing projection object, or undefined.
   * @return {Projection|null} Projection object, or null if not in list.
   * @api
   */
  function get$2(projectionLike) {
    return typeof projectionLike === 'string'
      ? get$4(/** @type {string} */ (projectionLike))
      : /** @type {Projection} */ (projectionLike) || null;
  }

  /**
   * Get the resolution of the point in degrees or distance units.
   * For projections with degrees as the unit this will simply return the
   * provided resolution. For other projections the point resolution is
   * by default estimated by transforming the `point` pixel to EPSG:4326,
   * measuring its width and height on the normal sphere,
   * and taking the average of the width and height.
   * A custom function can be provided for a specific projection, either
   * by setting the `getPointResolution` option in the
   * {@link module:ol/proj/Projection~Projection} constructor or by using
   * {@link module:ol/proj/Projection~Projection#setGetPointResolution} to change an existing
   * projection object.
   * @param {ProjectionLike} projection The projection.
   * @param {number} resolution Nominal resolution in projection units.
   * @param {import("./coordinate.js").Coordinate} point Point to find adjusted resolution at.
   * @param {import("./proj/Units.js").Units} [units] Units to get the point resolution in.
   * Default is the projection's units.
   * @return {number} Point resolution.
   * @api
   */
  function getPointResolution(projection, resolution, point, units) {
    projection = get$2(projection);
    let pointResolution;
    const getter = projection.getPointResolutionFunc();
    if (getter) {
      pointResolution = getter(resolution, point);
      if (units && units !== projection.getUnits()) {
        const metersPerUnit = projection.getMetersPerUnit();
        if (metersPerUnit) {
          pointResolution =
            (pointResolution * metersPerUnit) / METERS_PER_UNIT$1[units];
        }
      }
    } else {
      const projUnits = projection.getUnits();
      if ((projUnits == 'degrees' && !units) || units == 'degrees') {
        pointResolution = resolution;
      } else {
        // Estimate point resolution by transforming the center pixel to EPSG:4326,
        // measuring its width and height on the normal sphere, and taking the
        // average of the width and height.
        const toEPSG4326 = getTransformFromProjections(
          projection,
          get$2('EPSG:4326')
        );
        if (toEPSG4326 === identityTransform && projUnits !== 'degrees') {
          // no transform is available
          pointResolution = resolution * projection.getMetersPerUnit();
        } else {
          let vertices = [
            point[0] - resolution / 2,
            point[1],
            point[0] + resolution / 2,
            point[1],
            point[0],
            point[1] - resolution / 2,
            point[0],
            point[1] + resolution / 2,
          ];
          vertices = toEPSG4326(vertices, vertices, 2);
          const width = getDistance(vertices.slice(0, 2), vertices.slice(2, 4));
          const height = getDistance(vertices.slice(4, 6), vertices.slice(6, 8));
          pointResolution = (width + height) / 2;
        }
        const metersPerUnit = units
          ? METERS_PER_UNIT$1[units]
          : projection.getMetersPerUnit();
        if (metersPerUnit !== undefined) {
          pointResolution /= metersPerUnit;
        }
      }
    }
    return pointResolution;
  }

  /**
   * Registers transformation functions that don't alter coordinates. Those allow
   * to transform between projections with equal meaning.
   *
   * @param {Array<Projection>} projections Projections.
   * @api
   */
  function addEquivalentProjections(projections) {
    addProjections(projections);
    projections.forEach(function (source) {
      projections.forEach(function (destination) {
        if (source !== destination) {
          add$1(source, destination, cloneTransform);
        }
      });
    });
  }

  /**
   * Registers transformation functions to convert coordinates in any projection
   * in projection1 to any projection in projection2.
   *
   * @param {Array<Projection>} projections1 Projections with equal
   *     meaning.
   * @param {Array<Projection>} projections2 Projections with equal
   *     meaning.
   * @param {TransformFunction} forwardTransform Transformation from any
   *   projection in projection1 to any projection in projection2.
   * @param {TransformFunction} inverseTransform Transform from any projection
   *   in projection2 to any projection in projection1..
   */
  function addEquivalentTransforms(
    projections1,
    projections2,
    forwardTransform,
    inverseTransform
  ) {
    projections1.forEach(function (projection1) {
      projections2.forEach(function (projection2) {
        add$1(projection1, projection2, forwardTransform);
        add$1(projection2, projection1, inverseTransform);
      });
    });
  }

  /**
   * Clear all cached projections and transforms.
   */
  function clearAllProjections() {
    clear$1();
    clear();
  }

  /**
   * @param {Projection|string|undefined} projection Projection.
   * @param {string} defaultCode Default code.
   * @return {Projection} Projection.
   */
  function createProjection(projection, defaultCode) {
    if (!projection) {
      return get$2(defaultCode);
    } else if (typeof projection === 'string') {
      return get$2(projection);
    } else {
      return /** @type {Projection} */ (projection);
    }
  }

  /**
   * Creates a {@link module:ol/proj~TransformFunction} from a simple 2D coordinate transform
   * function.
   * @param {function(import("./coordinate.js").Coordinate): import("./coordinate.js").Coordinate} coordTransform Coordinate
   *     transform.
   * @return {TransformFunction} Transform function.
   */
  function createTransformFromCoordinateTransform(coordTransform) {
    return (
      /**
       * @param {Array<number>} input Input.
       * @param {Array<number>} [output] Output.
       * @param {number} [dimension] Dimension.
       * @return {Array<number>} Output.
       */
      function (input, output, dimension) {
        const length = input.length;
        dimension = dimension !== undefined ? dimension : 2;
        output = output !== undefined ? output : new Array(length);
        for (let i = 0; i < length; i += dimension) {
          const point = coordTransform(input.slice(i, i + dimension));
          const pointLength = point.length;
          for (let j = 0, jj = dimension; j < jj; ++j) {
            output[i + j] = j >= pointLength ? input[i + j] : point[j];
          }
        }
        return output;
      }
    );
  }

  /**
   * Registers coordinate transform functions to convert coordinates between the
   * source projection and the destination projection.
   * The forward and inverse functions convert coordinate pairs; this function
   * converts these into the functions used internally which also handle
   * extents and coordinate arrays.
   *
   * @param {ProjectionLike} source Source projection.
   * @param {ProjectionLike} destination Destination projection.
   * @param {function(import("./coordinate.js").Coordinate): import("./coordinate.js").Coordinate} forward The forward transform
   *     function (that is, from the source projection to the destination
   *     projection) that takes a {@link module:ol/coordinate~Coordinate} as argument and returns
   *     the transformed {@link module:ol/coordinate~Coordinate}.
   * @param {function(import("./coordinate.js").Coordinate): import("./coordinate.js").Coordinate} inverse The inverse transform
   *     function (that is, from the destination projection to the source
   *     projection) that takes a {@link module:ol/coordinate~Coordinate} as argument and returns
   *     the transformed {@link module:ol/coordinate~Coordinate}. If the transform function can only
   *     transform less dimensions than the input coordinate, it is supposeed to return a coordinate
   *     with only the length it can transform. The other dimensions will be taken unchanged from the
   *     source.
   * @api
   */
  function addCoordinateTransforms(source, destination, forward, inverse) {
    const sourceProj = get$2(source);
    const destProj = get$2(destination);
    add$1(
      sourceProj,
      destProj,
      createTransformFromCoordinateTransform(forward)
    );
    add$1(
      destProj,
      sourceProj,
      createTransformFromCoordinateTransform(inverse)
    );
  }

  /**
   * Transforms a coordinate from longitude/latitude to a different projection.
   * @param {import("./coordinate.js").Coordinate} coordinate Coordinate as longitude and latitude, i.e.
   *     an array with longitude as 1st and latitude as 2nd element.
   * @param {ProjectionLike} [projection] Target projection. The
   *     default is Web Mercator, i.e. 'EPSG:3857'.
   * @return {import("./coordinate.js").Coordinate} Coordinate projected to the target projection.
   * @api
   */
  function fromLonLat(coordinate, projection) {
    disableCoordinateWarning();
    return transform(
      coordinate,
      'EPSG:4326',
      projection !== undefined ? projection : 'EPSG:3857'
    );
  }

  /**
   * Transforms a coordinate to longitude/latitude.
   * @param {import("./coordinate.js").Coordinate} coordinate Projected coordinate.
   * @param {ProjectionLike} [projection] Projection of the coordinate.
   *     The default is Web Mercator, i.e. 'EPSG:3857'.
   * @return {import("./coordinate.js").Coordinate} Coordinate as longitude and latitude, i.e. an array
   *     with longitude as 1st and latitude as 2nd element.
   * @api
   */
  function toLonLat(coordinate, projection) {
    const lonLat = transform(
      coordinate,
      projection !== undefined ? projection : 'EPSG:3857',
      'EPSG:4326'
    );
    const lon = lonLat[0];
    if (lon < -180 || lon > 180) {
      lonLat[0] = modulo(lon + 180, 360) - 180;
    }
    return lonLat;
  }

  /**
   * Checks if two projections are the same, that is every coordinate in one
   * projection does represent the same geographic point as the same coordinate in
   * the other projection.
   *
   * @param {Projection} projection1 Projection 1.
   * @param {Projection} projection2 Projection 2.
   * @return {boolean} Equivalent.
   * @api
   */
  function equivalent(projection1, projection2) {
    if (projection1 === projection2) {
      return true;
    }
    const equalUnits = projection1.getUnits() === projection2.getUnits();
    if (projection1.getCode() === projection2.getCode()) {
      return equalUnits;
    } else {
      const transformFunc = getTransformFromProjections(projection1, projection2);
      return transformFunc === cloneTransform && equalUnits;
    }
  }

  /**
   * Searches in the list of transform functions for the function for converting
   * coordinates from the source projection to the destination projection.
   *
   * @param {Projection} sourceProjection Source Projection object.
   * @param {Projection} destinationProjection Destination Projection
   *     object.
   * @return {TransformFunction} Transform function.
   */
  function getTransformFromProjections(
    sourceProjection,
    destinationProjection
  ) {
    const sourceCode = sourceProjection.getCode();
    const destinationCode = destinationProjection.getCode();
    let transformFunc = get$3(sourceCode, destinationCode);
    if (!transformFunc) {
      transformFunc = identityTransform;
    }
    return transformFunc;
  }

  /**
   * Given the projection-like objects, searches for a transformation
   * function to convert a coordinates array from the source projection to the
   * destination projection.
   *
   * @param {ProjectionLike} source Source.
   * @param {ProjectionLike} destination Destination.
   * @return {TransformFunction} Transform function.
   * @api
   */
  function getTransform(source, destination) {
    const sourceProjection = get$2(source);
    const destinationProjection = get$2(destination);
    return getTransformFromProjections(sourceProjection, destinationProjection);
  }

  /**
   * Transforms a coordinate from source projection to destination projection.
   * This returns a new coordinate (and does not modify the original).
   *
   * See {@link module:ol/proj.transformExtent} for extent transformation.
   * See the transform method of {@link module:ol/geom/Geometry~Geometry} and its
   * subclasses for geometry transforms.
   *
   * @param {import("./coordinate.js").Coordinate} coordinate Coordinate.
   * @param {ProjectionLike} source Source projection-like.
   * @param {ProjectionLike} destination Destination projection-like.
   * @return {import("./coordinate.js").Coordinate} Coordinate.
   * @api
   */
  function transform(coordinate, source, destination) {
    const transformFunc = getTransform(source, destination);
    return transformFunc(coordinate, undefined, coordinate.length);
  }

  /**
   * Transforms an extent from source projection to destination projection.  This
   * returns a new extent (and does not modify the original).
   *
   * @param {import("./extent.js").Extent} extent The extent to transform.
   * @param {ProjectionLike} source Source projection-like.
   * @param {ProjectionLike} destination Destination projection-like.
   * @param {number} [stops] Number of stops per side used for the transform.
   * By default only the corners are used.
   * @return {import("./extent.js").Extent} The transformed extent.
   * @api
   */
  function transformExtent(extent, source, destination, stops) {
    const transformFunc = getTransform(source, destination);
    return applyTransform(extent, transformFunc, undefined, stops);
  }

  /**
   * Transforms the given point to the destination projection.
   *
   * @param {import("./coordinate.js").Coordinate} point Point.
   * @param {Projection} sourceProjection Source projection.
   * @param {Projection} destinationProjection Destination projection.
   * @return {import("./coordinate.js").Coordinate} Point.
   */
  function transformWithProjections(
    point,
    sourceProjection,
    destinationProjection
  ) {
    const transformFunc = getTransformFromProjections(
      sourceProjection,
      destinationProjection
    );
    return transformFunc(point);
  }

  /**
   * @type {Projection|null}
   */
  let userProjection = null;

  /**
   * Set the projection for coordinates supplied from and returned by API methods.
   * This includes all API methods except for those interacting with tile grids.
   * @param {ProjectionLike} projection The user projection.
   * @api
   */
  function setUserProjection(projection) {
    userProjection = get$2(projection);
  }

  /**
   * Clear the user projection if set.
   * @api
   */
  function clearUserProjection() {
    userProjection = null;
  }

  /**
   * Get the projection for coordinates supplied from and returned by API methods.
   * Note that this method is not yet a part of the stable API.  Support for user
   * projections is not yet complete and should be considered experimental.
   * @return {Projection|null} The user projection (or null if not set).
   * @api
   */
  function getUserProjection() {
    return userProjection;
  }

  /**
   * Use geographic coordinates (WGS-84 datum) in API methods.  This includes all API
   * methods except for those interacting with tile grids.
   * @api
   */
  function useGeographic() {
    setUserProjection('EPSG:4326');
  }

  /**
   * Return a coordinate transformed into the user projection.  If no user projection
   * is set, the original coordinate is returned.
   * @param {Array<number>} coordinate Input coordinate.
   * @param {ProjectionLike} sourceProjection The input coordinate projection.
   * @return {Array<number>} The input coordinate in the user projection.
   */
  function toUserCoordinate(coordinate, sourceProjection) {
    if (!userProjection) {
      return coordinate;
    }
    return transform(coordinate, sourceProjection, userProjection);
  }

  /**
   * Return a coordinate transformed from the user projection.  If no user projection
   * is set, the original coordinate is returned.
   * @param {Array<number>} coordinate Input coordinate.
   * @param {ProjectionLike} destProjection The destination projection.
   * @return {Array<number>} The input coordinate transformed.
   */
  function fromUserCoordinate(coordinate, destProjection) {
    if (!userProjection) {
      if (
        showCoordinateWarning &&
        !equals$1(coordinate, [0, 0]) &&
        coordinate[0] >= -180 &&
        coordinate[0] <= 180 &&
        coordinate[1] >= -90 &&
        coordinate[1] <= 90
      ) {
        showCoordinateWarning = false;
        // eslint-disable-next-line no-console
        console.warn(
          'Call useGeographic() from ol/proj once to work with [longitude, latitude] coordinates.'
        );
      }
      return coordinate;
    }
    return transform(coordinate, userProjection, destProjection);
  }

  /**
   * Return an extent transformed into the user projection.  If no user projection
   * is set, the original extent is returned.
   * @param {import("./extent.js").Extent} extent Input extent.
   * @param {ProjectionLike} sourceProjection The input extent projection.
   * @return {import("./extent.js").Extent} The input extent in the user projection.
   */
  function toUserExtent(extent, sourceProjection) {
    if (!userProjection) {
      return extent;
    }
    return transformExtent(extent, sourceProjection, userProjection);
  }

  /**
   * Return an extent transformed from the user projection.  If no user projection
   * is set, the original extent is returned.
   * @param {import("./extent.js").Extent} extent Input extent.
   * @param {ProjectionLike} destProjection The destination projection.
   * @return {import("./extent.js").Extent} The input extent transformed.
   */
  function fromUserExtent(extent, destProjection) {
    if (!userProjection) {
      return extent;
    }
    return transformExtent(extent, userProjection, destProjection);
  }

  /**
   * Return the resolution in user projection units per pixel. If no user projection
   * is set, or source or user projection are missing units, the original resolution
   * is returned.
   * @param {number} resolution Resolution in input projection units per pixel.
   * @param {ProjectionLike} sourceProjection The input projection.
   * @return {number} Resolution in user projection units per pixel.
   */
  function toUserResolution(resolution, sourceProjection) {
    if (!userProjection) {
      return resolution;
    }
    const sourceUnits = get$2(sourceProjection).getUnits();
    const userUnits = userProjection.getUnits();
    return sourceUnits && userUnits
      ? (resolution * METERS_PER_UNIT$1[sourceUnits]) / METERS_PER_UNIT$1[userUnits]
      : resolution;
  }

  /**
   * Return the resolution in user projection units per pixel. If no user projection
   * is set, or source or user projection are missing units, the original resolution
   * is returned.
   * @param {number} resolution Resolution in user projection units per pixel.
   * @param {ProjectionLike} destProjection The destination projection.
   * @return {number} Resolution in destination projection units per pixel.
   */
  function fromUserResolution(resolution, destProjection) {
    if (!userProjection) {
      return resolution;
    }
    const sourceUnits = get$2(destProjection).getUnits();
    const userUnits = userProjection.getUnits();
    return sourceUnits && userUnits
      ? (resolution * METERS_PER_UNIT$1[userUnits]) / METERS_PER_UNIT$1[sourceUnits]
      : resolution;
  }

  /**
   * Creates a safe coordinate transform function from a coordinate transform function.
   * "Safe" means that it can handle wrapping of x-coordinates for global projections,
   * and that coordinates exceeding the source projection validity extent's range will be
   * clamped to the validity range.
   * @param {Projection} sourceProj Source projection.
   * @param {Projection} destProj Destination projection.
   * @param {function(import("./coordinate.js").Coordinate): import("./coordinate.js").Coordinate} transform Transform function (source to destiation).
   * @return {function(import("./coordinate.js").Coordinate): import("./coordinate.js").Coordinate} Safe transform function (source to destiation).
   */
  function createSafeCoordinateTransform(sourceProj, destProj, transform) {
    return function (coord) {
      let transformed, worldsAway;
      if (sourceProj.canWrapX()) {
        const sourceExtent = sourceProj.getExtent();
        const sourceExtentWidth = getWidth(sourceExtent);
        coord = coord.slice(0);
        worldsAway = getWorldsAway(coord, sourceProj, sourceExtentWidth);
        if (worldsAway) {
          // Move x to the real world
          coord[0] = coord[0] - worldsAway * sourceExtentWidth;
        }
        coord[0] = clamp(coord[0], sourceExtent[0], sourceExtent[2]);
        coord[1] = clamp(coord[1], sourceExtent[1], sourceExtent[3]);
        transformed = transform(coord);
      } else {
        transformed = transform(coord);
      }
      if (worldsAway && destProj.canWrapX()) {
        // Move transformed coordinate back to the offset world
        transformed[0] += worldsAway * getWidth(destProj.getExtent());
      }
      return transformed;
    };
  }

  /**
   * Add transforms to and from EPSG:4326 and EPSG:3857.  This function is called
   * by when this module is executed and should only need to be called again after
   * `clearAllProjections()` is called (e.g. in tests).
   */
  function addCommon() {
    // Add transformations that don't alter coordinates to convert within set of
    // projections with equal meaning.
    addEquivalentProjections(PROJECTIONS$1);
    addEquivalentProjections(PROJECTIONS);
    // Add transformations to convert EPSG:4326 like coordinates to EPSG:3857 like
    // coordinates and back.
    addEquivalentTransforms(
      PROJECTIONS,
      PROJECTIONS$1,
      fromEPSG4326,
      toEPSG4326
    );
  }

  addCommon();

  /**
   * @module ol/geom/flat/transform
   */

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   * @param {import("../../transform.js").Transform} transform Transform.
   * @param {Array<number>} [dest] Destination.
   * @return {Array<number>} Transformed coordinates.
   */
  function transform2D(
    flatCoordinates,
    offset,
    end,
    stride,
    transform,
    dest
  ) {
    dest = dest ? dest : [];
    let i = 0;
    for (let j = offset; j < end; j += stride) {
      const x = flatCoordinates[j];
      const y = flatCoordinates[j + 1];
      dest[i++] = transform[0] * x + transform[2] * y + transform[4];
      dest[i++] = transform[1] * x + transform[3] * y + transform[5];
    }
    if (dest && dest.length != i) {
      dest.length = i;
    }
    return dest;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   * @param {number} angle Angle.
   * @param {Array<number>} anchor Rotation anchor point.
   * @param {Array<number>} [dest] Destination.
   * @return {Array<number>} Transformed coordinates.
   */
  function rotate(
    flatCoordinates,
    offset,
    end,
    stride,
    angle,
    anchor,
    dest
  ) {
    dest = dest ? dest : [];
    const cos = Math.cos(angle);
    const sin = Math.sin(angle);
    const anchorX = anchor[0];
    const anchorY = anchor[1];
    let i = 0;
    for (let j = offset; j < end; j += stride) {
      const deltaX = flatCoordinates[j] - anchorX;
      const deltaY = flatCoordinates[j + 1] - anchorY;
      dest[i++] = anchorX + deltaX * cos - deltaY * sin;
      dest[i++] = anchorY + deltaX * sin + deltaY * cos;
      for (let k = j + 2; k < j + stride; ++k) {
        dest[i++] = flatCoordinates[k];
      }
    }
    if (dest && dest.length != i) {
      dest.length = i;
    }
    return dest;
  }

  /**
   * Scale the coordinates.
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   * @param {number} sx Scale factor in the x-direction.
   * @param {number} sy Scale factor in the y-direction.
   * @param {Array<number>} anchor Scale anchor point.
   * @param {Array<number>} [dest] Destination.
   * @return {Array<number>} Transformed coordinates.
   */
  function scale$1(
    flatCoordinates,
    offset,
    end,
    stride,
    sx,
    sy,
    anchor,
    dest
  ) {
    dest = dest ? dest : [];
    const anchorX = anchor[0];
    const anchorY = anchor[1];
    let i = 0;
    for (let j = offset; j < end; j += stride) {
      const deltaX = flatCoordinates[j] - anchorX;
      const deltaY = flatCoordinates[j + 1] - anchorY;
      dest[i++] = anchorX + sx * deltaX;
      dest[i++] = anchorY + sy * deltaY;
      for (let k = j + 2; k < j + stride; ++k) {
        dest[i++] = flatCoordinates[k];
      }
    }
    if (dest && dest.length != i) {
      dest.length = i;
    }
    return dest;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   * @param {number} deltaX Delta X.
   * @param {number} deltaY Delta Y.
   * @param {Array<number>} [dest] Destination.
   * @return {Array<number>} Transformed coordinates.
   */
  function translate(
    flatCoordinates,
    offset,
    end,
    stride,
    deltaX,
    deltaY,
    dest
  ) {
    dest = dest ? dest : [];
    let i = 0;
    for (let j = offset; j < end; j += stride) {
      dest[i++] = flatCoordinates[j] + deltaX;
      dest[i++] = flatCoordinates[j + 1] + deltaY;
      for (let k = j + 2; k < j + stride; ++k) {
        dest[i++] = flatCoordinates[k];
      }
    }
    if (dest && dest.length != i) {
      dest.length = i;
    }
    return dest;
  }

  /**
   * @module ol/geom/Geometry
   */

  /**
   * @typedef {'XY' | 'XYZ' | 'XYM' | 'XYZM'} GeometryLayout
   * The coordinate layout for geometries, indicating whether a 3rd or 4th z ('Z')
   * or measure ('M') coordinate is available.
   */

  /**
   * @typedef {'Point' | 'LineString' | 'LinearRing' | 'Polygon' | 'MultiPoint' | 'MultiLineString' | 'MultiPolygon' | 'GeometryCollection' | 'Circle'} Type
   * The geometry type.  One of `'Point'`, `'LineString'`, `'LinearRing'`,
   * `'Polygon'`, `'MultiPoint'`, `'MultiLineString'`, `'MultiPolygon'`,
   * `'GeometryCollection'`, or `'Circle'`.
   */

  /**
   * @type {import("../transform.js").Transform}
   */
  const tmpTransform$1 = create$2();

  /**
   * @classdesc
   * Abstract base class; normally only used for creating subclasses and not
   * instantiated in apps.
   * Base class for vector geometries.
   *
   * To get notified of changes to the geometry, register a listener for the
   * generic `change` event on your geometry instance.
   *
   * @abstract
   * @api
   */
  class Geometry extends BaseObject {
    constructor() {
      super();

      /**
       * @private
       * @type {import("../extent.js").Extent}
       */
      this.extent_ = createEmpty();

      /**
       * @private
       * @type {number}
       */
      this.extentRevision_ = -1;

      /**
       * @protected
       * @type {number}
       */
      this.simplifiedGeometryMaxMinSquaredTolerance = 0;

      /**
       * @protected
       * @type {number}
       */
      this.simplifiedGeometryRevision = 0;

      /**
       * Get a transformed and simplified version of the geometry.
       * @abstract
       * @param {number} revision The geometry revision.
       * @param {number} squaredTolerance Squared tolerance.
       * @param {import("../proj.js").TransformFunction} [transform] Optional transform function.
       * @return {Geometry} Simplified geometry.
       */
      this.simplifyTransformedInternal = memoizeOne(function (
        revision,
        squaredTolerance,
        transform
      ) {
        if (!transform) {
          return this.getSimplifiedGeometry(squaredTolerance);
        }
        const clone = this.clone();
        clone.applyTransform(transform);
        return clone.getSimplifiedGeometry(squaredTolerance);
      });
    }

    /**
     * Get a transformed and simplified version of the geometry.
     * @abstract
     * @param {number} squaredTolerance Squared tolerance.
     * @param {import("../proj.js").TransformFunction} [transform] Optional transform function.
     * @return {Geometry} Simplified geometry.
     */
    simplifyTransformed(squaredTolerance, transform) {
      return this.simplifyTransformedInternal(
        this.getRevision(),
        squaredTolerance,
        transform
      );
    }

    /**
     * Make a complete copy of the geometry.
     * @abstract
     * @return {!Geometry} Clone.
     */
    clone() {
      return abstract();
    }

    /**
     * @abstract
     * @param {number} x X.
     * @param {number} y Y.
     * @param {import("../coordinate.js").Coordinate} closestPoint Closest point.
     * @param {number} minSquaredDistance Minimum squared distance.
     * @return {number} Minimum squared distance.
     */
    closestPointXY(x, y, closestPoint, minSquaredDistance) {
      return abstract();
    }

    /**
     * @param {number} x X.
     * @param {number} y Y.
     * @return {boolean} Contains (x, y).
     */
    containsXY(x, y) {
      const coord = this.getClosestPoint([x, y]);
      return coord[0] === x && coord[1] === y;
    }

    /**
     * Return the closest point of the geometry to the passed point as
     * {@link module:ol/coordinate~Coordinate coordinate}.
     * @param {import("../coordinate.js").Coordinate} point Point.
     * @param {import("../coordinate.js").Coordinate} [closestPoint] Closest point.
     * @return {import("../coordinate.js").Coordinate} Closest point.
     * @api
     */
    getClosestPoint(point, closestPoint) {
      closestPoint = closestPoint ? closestPoint : [NaN, NaN];
      this.closestPointXY(point[0], point[1], closestPoint, Infinity);
      return closestPoint;
    }

    /**
     * Returns true if this geometry includes the specified coordinate. If the
     * coordinate is on the boundary of the geometry, returns false.
     * @param {import("../coordinate.js").Coordinate} coordinate Coordinate.
     * @return {boolean} Contains coordinate.
     * @api
     */
    intersectsCoordinate(coordinate) {
      return this.containsXY(coordinate[0], coordinate[1]);
    }

    /**
     * @abstract
     * @param {import("../extent.js").Extent} extent Extent.
     * @protected
     * @return {import("../extent.js").Extent} extent Extent.
     */
    computeExtent(extent) {
      return abstract();
    }

    /**
     * Get the extent of the geometry.
     * @param {import("../extent.js").Extent} [extent] Extent.
     * @return {import("../extent.js").Extent} extent Extent.
     * @api
     */
    getExtent(extent) {
      if (this.extentRevision_ != this.getRevision()) {
        const extent = this.computeExtent(this.extent_);
        if (isNaN(extent[0]) || isNaN(extent[1])) {
          createOrUpdateEmpty(extent);
        }
        this.extentRevision_ = this.getRevision();
      }
      return returnOrUpdate(this.extent_, extent);
    }

    /**
     * Rotate the geometry around a given coordinate. This modifies the geometry
     * coordinates in place.
     * @abstract
     * @param {number} angle Rotation angle in radians.
     * @param {import("../coordinate.js").Coordinate} anchor The rotation center.
     * @api
     */
    rotate(angle, anchor) {
      abstract();
    }

    /**
     * Scale the geometry (with an optional origin).  This modifies the geometry
     * coordinates in place.
     * @abstract
     * @param {number} sx The scaling factor in the x-direction.
     * @param {number} [sy] The scaling factor in the y-direction (defaults to sx).
     * @param {import("../coordinate.js").Coordinate} [anchor] The scale origin (defaults to the center
     *     of the geometry extent).
     * @api
     */
    scale(sx, sy, anchor) {
      abstract();
    }

    /**
     * Create a simplified version of this geometry.  For linestrings, this uses
     * the [Douglas Peucker](https://en.wikipedia.org/wiki/Ramer-Douglas-Peucker_algorithm)
     * algorithm.  For polygons, a quantization-based
     * simplification is used to preserve topology.
     * @param {number} tolerance The tolerance distance for simplification.
     * @return {Geometry} A new, simplified version of the original geometry.
     * @api
     */
    simplify(tolerance) {
      return this.getSimplifiedGeometry(tolerance * tolerance);
    }

    /**
     * Create a simplified version of this geometry using the Douglas Peucker
     * algorithm.
     * See https://en.wikipedia.org/wiki/Ramer-Douglas-Peucker_algorithm.
     * @abstract
     * @param {number} squaredTolerance Squared tolerance.
     * @return {Geometry} Simplified geometry.
     */
    getSimplifiedGeometry(squaredTolerance) {
      return abstract();
    }

    /**
     * Get the type of this geometry.
     * @abstract
     * @return {Type} Geometry type.
     */
    getType() {
      return abstract();
    }

    /**
     * Apply a transform function to the coordinates of the geometry.
     * The geometry is modified in place.
     * If you do not want the geometry modified in place, first `clone()` it and
     * then use this function on the clone.
     * @abstract
     * @param {import("../proj.js").TransformFunction} transformFn Transform function.
     * Called with a flat array of geometry coordinates.
     */
    applyTransform(transformFn) {
      abstract();
    }

    /**
     * Test if the geometry and the passed extent intersect.
     * @abstract
     * @param {import("../extent.js").Extent} extent Extent.
     * @return {boolean} `true` if the geometry and the extent intersect.
     */
    intersectsExtent(extent) {
      return abstract();
    }

    /**
     * Translate the geometry.  This modifies the geometry coordinates in place.  If
     * instead you want a new geometry, first `clone()` this geometry.
     * @abstract
     * @param {number} deltaX Delta X.
     * @param {number} deltaY Delta Y.
     * @api
     */
    translate(deltaX, deltaY) {
      abstract();
    }

    /**
     * Transform each coordinate of the geometry from one coordinate reference
     * system to another. The geometry is modified in place.
     * For example, a line will be transformed to a line and a circle to a circle.
     * If you do not want the geometry modified in place, first `clone()` it and
     * then use this function on the clone.
     *
     * @param {import("../proj.js").ProjectionLike} source The current projection.  Can be a
     *     string identifier or a {@link module:ol/proj/Projection~Projection} object.
     * @param {import("../proj.js").ProjectionLike} destination The desired projection.  Can be a
     *     string identifier or a {@link module:ol/proj/Projection~Projection} object.
     * @return {Geometry} This geometry.  Note that original geometry is
     *     modified in place.
     * @api
     */
    transform(source, destination) {
      /** @type {import("../proj/Projection.js").default} */
      const sourceProj = get$2(source);
      const transformFn =
        sourceProj.getUnits() == 'tile-pixels'
          ? function (inCoordinates, outCoordinates, stride) {
              const pixelExtent = sourceProj.getExtent();
              const projectedExtent = sourceProj.getWorldExtent();
              const scale = getHeight(projectedExtent) / getHeight(pixelExtent);
              compose(
                tmpTransform$1,
                projectedExtent[0],
                projectedExtent[3],
                scale,
                -scale,
                0,
                0,
                0
              );
              transform2D(
                inCoordinates,
                0,
                inCoordinates.length,
                stride,
                tmpTransform$1,
                outCoordinates
              );
              return getTransform(sourceProj, destination)(
                inCoordinates,
                outCoordinates,
                stride
              );
            }
          : getTransform(sourceProj, destination);
      this.applyTransform(transformFn);
      return this;
    }
  }

  /**
   * @module ol/geom/SimpleGeometry
   */

  /**
   * @classdesc
   * Abstract base class; only used for creating subclasses; do not instantiate
   * in apps, as cannot be rendered.
   *
   * @abstract
   * @api
   */
  class SimpleGeometry extends Geometry {
    constructor() {
      super();

      /**
       * @protected
       * @type {import("./Geometry.js").GeometryLayout}
       */
      this.layout = 'XY';

      /**
       * @protected
       * @type {number}
       */
      this.stride = 2;

      /**
       * @protected
       * @type {Array<number>}
       */
      this.flatCoordinates = null;
    }

    /**
     * @param {import("../extent.js").Extent} extent Extent.
     * @protected
     * @return {import("../extent.js").Extent} extent Extent.
     */
    computeExtent(extent) {
      return createOrUpdateFromFlatCoordinates(
        this.flatCoordinates,
        0,
        this.flatCoordinates.length,
        this.stride,
        extent
      );
    }

    /**
     * @abstract
     * @return {Array<*> | null} Coordinates.
     */
    getCoordinates() {
      return abstract();
    }

    /**
     * Return the first coordinate of the geometry.
     * @return {import("../coordinate.js").Coordinate} First coordinate.
     * @api
     */
    getFirstCoordinate() {
      return this.flatCoordinates.slice(0, this.stride);
    }

    /**
     * @return {Array<number>} Flat coordinates.
     */
    getFlatCoordinates() {
      return this.flatCoordinates;
    }

    /**
     * Return the last coordinate of the geometry.
     * @return {import("../coordinate.js").Coordinate} Last point.
     * @api
     */
    getLastCoordinate() {
      return this.flatCoordinates.slice(
        this.flatCoordinates.length - this.stride
      );
    }

    /**
     * Return the {@link import("./Geometry.js").GeometryLayout layout} of the geometry.
     * @return {import("./Geometry.js").GeometryLayout} Layout.
     * @api
     */
    getLayout() {
      return this.layout;
    }

    /**
     * Create a simplified version of this geometry using the Douglas Peucker algorithm.
     * @param {number} squaredTolerance Squared tolerance.
     * @return {SimpleGeometry} Simplified geometry.
     */
    getSimplifiedGeometry(squaredTolerance) {
      if (this.simplifiedGeometryRevision !== this.getRevision()) {
        this.simplifiedGeometryMaxMinSquaredTolerance = 0;
        this.simplifiedGeometryRevision = this.getRevision();
      }
      // If squaredTolerance is negative or if we know that simplification will not
      // have any effect then just return this.
      if (
        squaredTolerance < 0 ||
        (this.simplifiedGeometryMaxMinSquaredTolerance !== 0 &&
          squaredTolerance <= this.simplifiedGeometryMaxMinSquaredTolerance)
      ) {
        return this;
      }

      const simplifiedGeometry =
        this.getSimplifiedGeometryInternal(squaredTolerance);
      const simplifiedFlatCoordinates = simplifiedGeometry.getFlatCoordinates();
      if (simplifiedFlatCoordinates.length < this.flatCoordinates.length) {
        return simplifiedGeometry;
      } else {
        // Simplification did not actually remove any coordinates.  We now know
        // that any calls to getSimplifiedGeometry with a squaredTolerance less
        // than or equal to the current squaredTolerance will also not have any
        // effect.  This allows us to short circuit simplification (saving CPU
        // cycles) and prevents the cache of simplified geometries from filling
        // up with useless identical copies of this geometry (saving memory).
        this.simplifiedGeometryMaxMinSquaredTolerance = squaredTolerance;
        return this;
      }
    }

    /**
     * @param {number} squaredTolerance Squared tolerance.
     * @return {SimpleGeometry} Simplified geometry.
     * @protected
     */
    getSimplifiedGeometryInternal(squaredTolerance) {
      return this;
    }

    /**
     * @return {number} Stride.
     */
    getStride() {
      return this.stride;
    }

    /**
     * @param {import("./Geometry.js").GeometryLayout} layout Layout.
     * @param {Array<number>} flatCoordinates Flat coordinates.
     */
    setFlatCoordinates(layout, flatCoordinates) {
      this.stride = getStrideForLayout(layout);
      this.layout = layout;
      this.flatCoordinates = flatCoordinates;
    }

    /**
     * @abstract
     * @param {!Array<*>} coordinates Coordinates.
     * @param {import("./Geometry.js").GeometryLayout} [layout] Layout.
     */
    setCoordinates(coordinates, layout) {
      abstract();
    }

    /**
     * @param {import("./Geometry.js").GeometryLayout|undefined} layout Layout.
     * @param {Array<*>} coordinates Coordinates.
     * @param {number} nesting Nesting.
     * @protected
     */
    setLayout(layout, coordinates, nesting) {
      /** @type {number} */
      let stride;
      if (layout) {
        stride = getStrideForLayout(layout);
      } else {
        for (let i = 0; i < nesting; ++i) {
          if (coordinates.length === 0) {
            this.layout = 'XY';
            this.stride = 2;
            return;
          } else {
            coordinates = /** @type {Array} */ (coordinates[0]);
          }
        }
        stride = coordinates.length;
        layout = getLayoutForStride(stride);
      }
      this.layout = layout;
      this.stride = stride;
    }

    /**
     * Apply a transform function to the coordinates of the geometry.
     * The geometry is modified in place.
     * If you do not want the geometry modified in place, first `clone()` it and
     * then use this function on the clone.
     * @param {import("../proj.js").TransformFunction} transformFn Transform function.
     * Called with a flat array of geometry coordinates.
     * @api
     */
    applyTransform(transformFn) {
      if (this.flatCoordinates) {
        transformFn(this.flatCoordinates, this.flatCoordinates, this.stride);
        this.changed();
      }
    }

    /**
     * Rotate the geometry around a given coordinate. This modifies the geometry
     * coordinates in place.
     * @param {number} angle Rotation angle in counter-clockwise radians.
     * @param {import("../coordinate.js").Coordinate} anchor The rotation center.
     * @api
     */
    rotate(angle, anchor) {
      const flatCoordinates = this.getFlatCoordinates();
      if (flatCoordinates) {
        const stride = this.getStride();
        rotate(
          flatCoordinates,
          0,
          flatCoordinates.length,
          stride,
          angle,
          anchor,
          flatCoordinates
        );
        this.changed();
      }
    }

    /**
     * Scale the geometry (with an optional origin).  This modifies the geometry
     * coordinates in place.
     * @param {number} sx The scaling factor in the x-direction.
     * @param {number} [sy] The scaling factor in the y-direction (defaults to sx).
     * @param {import("../coordinate.js").Coordinate} [anchor] The scale origin (defaults to the center
     *     of the geometry extent).
     * @api
     */
    scale(sx, sy, anchor) {
      if (sy === undefined) {
        sy = sx;
      }
      if (!anchor) {
        anchor = getCenter(this.getExtent());
      }
      const flatCoordinates = this.getFlatCoordinates();
      if (flatCoordinates) {
        const stride = this.getStride();
        scale$1(
          flatCoordinates,
          0,
          flatCoordinates.length,
          stride,
          sx,
          sy,
          anchor,
          flatCoordinates
        );
        this.changed();
      }
    }

    /**
     * Translate the geometry.  This modifies the geometry coordinates in place.  If
     * instead you want a new geometry, first `clone()` this geometry.
     * @param {number} deltaX Delta X.
     * @param {number} deltaY Delta Y.
     * @api
     */
    translate(deltaX, deltaY) {
      const flatCoordinates = this.getFlatCoordinates();
      if (flatCoordinates) {
        const stride = this.getStride();
        translate(
          flatCoordinates,
          0,
          flatCoordinates.length,
          stride,
          deltaX,
          deltaY,
          flatCoordinates
        );
        this.changed();
      }
    }
  }

  /**
   * @param {number} stride Stride.
   * @return {import("./Geometry.js").GeometryLayout} layout Layout.
   */
  function getLayoutForStride(stride) {
    let layout;
    if (stride == 2) {
      layout = 'XY';
    } else if (stride == 3) {
      layout = 'XYZ';
    } else if (stride == 4) {
      layout = 'XYZM';
    }
    return /** @type {import("./Geometry.js").GeometryLayout} */ (layout);
  }

  /**
   * @param {import("./Geometry.js").GeometryLayout} layout Layout.
   * @return {number} Stride.
   */
  function getStrideForLayout(layout) {
    let stride;
    if (layout == 'XY') {
      stride = 2;
    } else if (layout == 'XYZ' || layout == 'XYM') {
      stride = 3;
    } else if (layout == 'XYZM') {
      stride = 4;
    }
    return /** @type {number} */ (stride);
  }

  /**
   * @param {SimpleGeometry} simpleGeometry Simple geometry.
   * @param {import("../transform.js").Transform} transform Transform.
   * @param {Array<number>} [dest] Destination.
   * @return {Array<number>} Transformed flat coordinates.
   */
  function transformGeom2D(simpleGeometry, transform, dest) {
    const flatCoordinates = simpleGeometry.getFlatCoordinates();
    if (!flatCoordinates) {
      return null;
    } else {
      const stride = simpleGeometry.getStride();
      return transform2D(
        flatCoordinates,
        0,
        flatCoordinates.length,
        stride,
        transform,
        dest
      );
    }
  }

  /**
   * @module ol/geom/flat/closest
   */

  /**
   * Returns the point on the 2D line segment flatCoordinates[offset1] to
   * flatCoordinates[offset2] that is closest to the point (x, y).  Extra
   * dimensions are linearly interpolated.
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset1 Offset 1.
   * @param {number} offset2 Offset 2.
   * @param {number} stride Stride.
   * @param {number} x X.
   * @param {number} y Y.
   * @param {Array<number>} closestPoint Closest point.
   */
  function assignClosest(
    flatCoordinates,
    offset1,
    offset2,
    stride,
    x,
    y,
    closestPoint
  ) {
    const x1 = flatCoordinates[offset1];
    const y1 = flatCoordinates[offset1 + 1];
    const dx = flatCoordinates[offset2] - x1;
    const dy = flatCoordinates[offset2 + 1] - y1;
    let offset;
    if (dx === 0 && dy === 0) {
      offset = offset1;
    } else {
      const t = ((x - x1) * dx + (y - y1) * dy) / (dx * dx + dy * dy);
      if (t > 1) {
        offset = offset2;
      } else if (t > 0) {
        for (let i = 0; i < stride; ++i) {
          closestPoint[i] = lerp(
            flatCoordinates[offset1 + i],
            flatCoordinates[offset2 + i],
            t
          );
        }
        closestPoint.length = stride;
        return;
      } else {
        offset = offset1;
      }
    }
    for (let i = 0; i < stride; ++i) {
      closestPoint[i] = flatCoordinates[offset + i];
    }
    closestPoint.length = stride;
  }

  /**
   * Return the squared of the largest distance between any pair of consecutive
   * coordinates.
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   * @param {number} max Max squared delta.
   * @return {number} Max squared delta.
   */
  function maxSquaredDelta(flatCoordinates, offset, end, stride, max) {
    let x1 = flatCoordinates[offset];
    let y1 = flatCoordinates[offset + 1];
    for (offset += stride; offset < end; offset += stride) {
      const x2 = flatCoordinates[offset];
      const y2 = flatCoordinates[offset + 1];
      const squaredDelta = squaredDistance$1(x1, y1, x2, y2);
      if (squaredDelta > max) {
        max = squaredDelta;
      }
      x1 = x2;
      y1 = y2;
    }
    return max;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<number>} ends Ends.
   * @param {number} stride Stride.
   * @param {number} max Max squared delta.
   * @return {number} Max squared delta.
   */
  function arrayMaxSquaredDelta(
    flatCoordinates,
    offset,
    ends,
    stride,
    max
  ) {
    for (let i = 0, ii = ends.length; i < ii; ++i) {
      const end = ends[i];
      max = maxSquaredDelta(flatCoordinates, offset, end, stride, max);
      offset = end;
    }
    return max;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<Array<number>>} endss Endss.
   * @param {number} stride Stride.
   * @param {number} max Max squared delta.
   * @return {number} Max squared delta.
   */
  function multiArrayMaxSquaredDelta(
    flatCoordinates,
    offset,
    endss,
    stride,
    max
  ) {
    for (let i = 0, ii = endss.length; i < ii; ++i) {
      const ends = endss[i];
      max = arrayMaxSquaredDelta(flatCoordinates, offset, ends, stride, max);
      offset = ends[ends.length - 1];
    }
    return max;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   * @param {number} maxDelta Max delta.
   * @param {boolean} isRing Is ring.
   * @param {number} x X.
   * @param {number} y Y.
   * @param {Array<number>} closestPoint Closest point.
   * @param {number} minSquaredDistance Minimum squared distance.
   * @param {Array<number>} [tmpPoint] Temporary point object.
   * @return {number} Minimum squared distance.
   */
  function assignClosestPoint(
    flatCoordinates,
    offset,
    end,
    stride,
    maxDelta,
    isRing,
    x,
    y,
    closestPoint,
    minSquaredDistance,
    tmpPoint
  ) {
    if (offset == end) {
      return minSquaredDistance;
    }
    let i, squaredDistance;
    if (maxDelta === 0) {
      // All points are identical, so just test the first point.
      squaredDistance = squaredDistance$1(
        x,
        y,
        flatCoordinates[offset],
        flatCoordinates[offset + 1]
      );
      if (squaredDistance < minSquaredDistance) {
        for (i = 0; i < stride; ++i) {
          closestPoint[i] = flatCoordinates[offset + i];
        }
        closestPoint.length = stride;
        return squaredDistance;
      } else {
        return minSquaredDistance;
      }
    }
    tmpPoint = tmpPoint ? tmpPoint : [NaN, NaN];
    let index = offset + stride;
    while (index < end) {
      assignClosest(
        flatCoordinates,
        index - stride,
        index,
        stride,
        x,
        y,
        tmpPoint
      );
      squaredDistance = squaredDistance$1(x, y, tmpPoint[0], tmpPoint[1]);
      if (squaredDistance < minSquaredDistance) {
        minSquaredDistance = squaredDistance;
        for (i = 0; i < stride; ++i) {
          closestPoint[i] = tmpPoint[i];
        }
        closestPoint.length = stride;
        index += stride;
      } else {
        // Skip ahead multiple points, because we know that all the skipped
        // points cannot be any closer than the closest point we have found so
        // far.  We know this because we know how close the current point is, how
        // close the closest point we have found so far is, and the maximum
        // distance between consecutive points.  For example, if we're currently
        // at distance 10, the best we've found so far is 3, and that the maximum
        // distance between consecutive points is 2, then we'll need to skip at
        // least (10 - 3) / 2 == 3 (rounded down) points to have any chance of
        // finding a closer point.  We use Math.max(..., 1) to ensure that we
        // always advance at least one point, to avoid an infinite loop.
        index +=
          stride *
          Math.max(
            ((Math.sqrt(squaredDistance) - Math.sqrt(minSquaredDistance)) /
              maxDelta) |
              0,
            1
          );
      }
    }
    if (isRing) {
      // Check the closing segment.
      assignClosest(
        flatCoordinates,
        end - stride,
        offset,
        stride,
        x,
        y,
        tmpPoint
      );
      squaredDistance = squaredDistance$1(x, y, tmpPoint[0], tmpPoint[1]);
      if (squaredDistance < minSquaredDistance) {
        minSquaredDistance = squaredDistance;
        for (i = 0; i < stride; ++i) {
          closestPoint[i] = tmpPoint[i];
        }
        closestPoint.length = stride;
      }
    }
    return minSquaredDistance;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<number>} ends Ends.
   * @param {number} stride Stride.
   * @param {number} maxDelta Max delta.
   * @param {boolean} isRing Is ring.
   * @param {number} x X.
   * @param {number} y Y.
   * @param {Array<number>} closestPoint Closest point.
   * @param {number} minSquaredDistance Minimum squared distance.
   * @param {Array<number>} [tmpPoint] Temporary point object.
   * @return {number} Minimum squared distance.
   */
  function assignClosestArrayPoint(
    flatCoordinates,
    offset,
    ends,
    stride,
    maxDelta,
    isRing,
    x,
    y,
    closestPoint,
    minSquaredDistance,
    tmpPoint
  ) {
    tmpPoint = tmpPoint ? tmpPoint : [NaN, NaN];
    for (let i = 0, ii = ends.length; i < ii; ++i) {
      const end = ends[i];
      minSquaredDistance = assignClosestPoint(
        flatCoordinates,
        offset,
        end,
        stride,
        maxDelta,
        isRing,
        x,
        y,
        closestPoint,
        minSquaredDistance,
        tmpPoint
      );
      offset = end;
    }
    return minSquaredDistance;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<Array<number>>} endss Endss.
   * @param {number} stride Stride.
   * @param {number} maxDelta Max delta.
   * @param {boolean} isRing Is ring.
   * @param {number} x X.
   * @param {number} y Y.
   * @param {Array<number>} closestPoint Closest point.
   * @param {number} minSquaredDistance Minimum squared distance.
   * @param {Array<number>} [tmpPoint] Temporary point object.
   * @return {number} Minimum squared distance.
   */
  function assignClosestMultiArrayPoint(
    flatCoordinates,
    offset,
    endss,
    stride,
    maxDelta,
    isRing,
    x,
    y,
    closestPoint,
    minSquaredDistance,
    tmpPoint
  ) {
    tmpPoint = tmpPoint ? tmpPoint : [NaN, NaN];
    for (let i = 0, ii = endss.length; i < ii; ++i) {
      const ends = endss[i];
      minSquaredDistance = assignClosestArrayPoint(
        flatCoordinates,
        offset,
        ends,
        stride,
        maxDelta,
        isRing,
        x,
        y,
        closestPoint,
        minSquaredDistance,
        tmpPoint
      );
      offset = ends[ends.length - 1];
    }
    return minSquaredDistance;
  }

  /**
   * @module ol/geom/flat/deflate
   */

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {import("../../coordinate.js").Coordinate} coordinate Coordinate.
   * @param {number} stride Stride.
   * @return {number} offset Offset.
   */
  function deflateCoordinate(flatCoordinates, offset, coordinate, stride) {
    for (let i = 0, ii = coordinate.length; i < ii; ++i) {
      flatCoordinates[offset++] = coordinate[i];
    }
    return offset;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<import("../../coordinate.js").Coordinate>} coordinates Coordinates.
   * @param {number} stride Stride.
   * @return {number} offset Offset.
   */
  function deflateCoordinates(
    flatCoordinates,
    offset,
    coordinates,
    stride
  ) {
    for (let i = 0, ii = coordinates.length; i < ii; ++i) {
      const coordinate = coordinates[i];
      for (let j = 0; j < stride; ++j) {
        flatCoordinates[offset++] = coordinate[j];
      }
    }
    return offset;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<Array<import("../../coordinate.js").Coordinate>>} coordinatess Coordinatess.
   * @param {number} stride Stride.
   * @param {Array<number>} [ends] Ends.
   * @return {Array<number>} Ends.
   */
  function deflateCoordinatesArray(
    flatCoordinates,
    offset,
    coordinatess,
    stride,
    ends
  ) {
    ends = ends ? ends : [];
    let i = 0;
    for (let j = 0, jj = coordinatess.length; j < jj; ++j) {
      const end = deflateCoordinates(
        flatCoordinates,
        offset,
        coordinatess[j],
        stride
      );
      ends[i++] = end;
      offset = end;
    }
    ends.length = i;
    return ends;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<Array<Array<import("../../coordinate.js").Coordinate>>>} coordinatesss Coordinatesss.
   * @param {number} stride Stride.
   * @param {Array<Array<number>>} [endss] Endss.
   * @return {Array<Array<number>>} Endss.
   */
  function deflateMultiCoordinatesArray(
    flatCoordinates,
    offset,
    coordinatesss,
    stride,
    endss
  ) {
    endss = endss ? endss : [];
    let i = 0;
    for (let j = 0, jj = coordinatesss.length; j < jj; ++j) {
      const ends = deflateCoordinatesArray(
        flatCoordinates,
        offset,
        coordinatesss[j],
        stride,
        endss[i]
      );
      if (ends.length === 0) {
        ends[0] = offset;
      }
      endss[i++] = ends;
      offset = ends[ends.length - 1];
    }
    endss.length = i;
    return endss;
  }

  /**
   * @module ol/geom/flat/simplify
   */

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   * @param {number} squaredTolerance Squared tolerance.
   * @param {boolean} highQuality Highest quality.
   * @param {Array<number>} [simplifiedFlatCoordinates] Simplified flat
   *     coordinates.
   * @return {Array<number>} Simplified line string.
   */
  function simplifyLineString(
    flatCoordinates,
    offset,
    end,
    stride,
    squaredTolerance,
    highQuality,
    simplifiedFlatCoordinates
  ) {
    simplifiedFlatCoordinates =
      simplifiedFlatCoordinates !== undefined ? simplifiedFlatCoordinates : [];
    if (!highQuality) {
      end = radialDistance(
        flatCoordinates,
        offset,
        end,
        stride,
        squaredTolerance,
        simplifiedFlatCoordinates,
        0
      );
      flatCoordinates = simplifiedFlatCoordinates;
      offset = 0;
      stride = 2;
    }
    simplifiedFlatCoordinates.length = douglasPeucker(
      flatCoordinates,
      offset,
      end,
      stride,
      squaredTolerance,
      simplifiedFlatCoordinates,
      0
    );
    return simplifiedFlatCoordinates;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   * @param {number} squaredTolerance Squared tolerance.
   * @param {Array<number>} simplifiedFlatCoordinates Simplified flat
   *     coordinates.
   * @param {number} simplifiedOffset Simplified offset.
   * @return {number} Simplified offset.
   */
  function douglasPeucker(
    flatCoordinates,
    offset,
    end,
    stride,
    squaredTolerance,
    simplifiedFlatCoordinates,
    simplifiedOffset
  ) {
    const n = (end - offset) / stride;
    if (n < 3) {
      for (; offset < end; offset += stride) {
        simplifiedFlatCoordinates[simplifiedOffset++] = flatCoordinates[offset];
        simplifiedFlatCoordinates[simplifiedOffset++] =
          flatCoordinates[offset + 1];
      }
      return simplifiedOffset;
    }
    /** @type {Array<number>} */
    const markers = new Array(n);
    markers[0] = 1;
    markers[n - 1] = 1;
    /** @type {Array<number>} */
    const stack = [offset, end - stride];
    let index = 0;
    while (stack.length > 0) {
      const last = stack.pop();
      const first = stack.pop();
      let maxSquaredDistance = 0;
      const x1 = flatCoordinates[first];
      const y1 = flatCoordinates[first + 1];
      const x2 = flatCoordinates[last];
      const y2 = flatCoordinates[last + 1];
      for (let i = first + stride; i < last; i += stride) {
        const x = flatCoordinates[i];
        const y = flatCoordinates[i + 1];
        const squaredDistance = squaredSegmentDistance(x, y, x1, y1, x2, y2);
        if (squaredDistance > maxSquaredDistance) {
          index = i;
          maxSquaredDistance = squaredDistance;
        }
      }
      if (maxSquaredDistance > squaredTolerance) {
        markers[(index - offset) / stride] = 1;
        if (first + stride < index) {
          stack.push(first, index);
        }
        if (index + stride < last) {
          stack.push(index, last);
        }
      }
    }
    for (let i = 0; i < n; ++i) {
      if (markers[i]) {
        simplifiedFlatCoordinates[simplifiedOffset++] =
          flatCoordinates[offset + i * stride];
        simplifiedFlatCoordinates[simplifiedOffset++] =
          flatCoordinates[offset + i * stride + 1];
      }
    }
    return simplifiedOffset;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<number>} ends Ends.
   * @param {number} stride Stride.
   * @param {number} squaredTolerance Squared tolerance.
   * @param {Array<number>} simplifiedFlatCoordinates Simplified flat
   *     coordinates.
   * @param {number} simplifiedOffset Simplified offset.
   * @param {Array<number>} simplifiedEnds Simplified ends.
   * @return {number} Simplified offset.
   */
  function douglasPeuckerArray(
    flatCoordinates,
    offset,
    ends,
    stride,
    squaredTolerance,
    simplifiedFlatCoordinates,
    simplifiedOffset,
    simplifiedEnds
  ) {
    for (let i = 0, ii = ends.length; i < ii; ++i) {
      const end = ends[i];
      simplifiedOffset = douglasPeucker(
        flatCoordinates,
        offset,
        end,
        stride,
        squaredTolerance,
        simplifiedFlatCoordinates,
        simplifiedOffset
      );
      simplifiedEnds.push(simplifiedOffset);
      offset = end;
    }
    return simplifiedOffset;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<Array<number>>} endss Endss.
   * @param {number} stride Stride.
   * @param {number} squaredTolerance Squared tolerance.
   * @param {Array<number>} simplifiedFlatCoordinates Simplified flat
   *     coordinates.
   * @param {number} simplifiedOffset Simplified offset.
   * @param {Array<Array<number>>} simplifiedEndss Simplified endss.
   * @return {number} Simplified offset.
   */
  function douglasPeuckerMultiArray(
    flatCoordinates,
    offset,
    endss,
    stride,
    squaredTolerance,
    simplifiedFlatCoordinates,
    simplifiedOffset,
    simplifiedEndss
  ) {
    for (let i = 0, ii = endss.length; i < ii; ++i) {
      const ends = endss[i];
      const simplifiedEnds = [];
      simplifiedOffset = douglasPeuckerArray(
        flatCoordinates,
        offset,
        ends,
        stride,
        squaredTolerance,
        simplifiedFlatCoordinates,
        simplifiedOffset,
        simplifiedEnds
      );
      simplifiedEndss.push(simplifiedEnds);
      offset = ends[ends.length - 1];
    }
    return simplifiedOffset;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   * @param {number} squaredTolerance Squared tolerance.
   * @param {Array<number>} simplifiedFlatCoordinates Simplified flat
   *     coordinates.
   * @param {number} simplifiedOffset Simplified offset.
   * @return {number} Simplified offset.
   */
  function radialDistance(
    flatCoordinates,
    offset,
    end,
    stride,
    squaredTolerance,
    simplifiedFlatCoordinates,
    simplifiedOffset
  ) {
    if (end <= offset + stride) {
      // zero or one point, no simplification possible, so copy and return
      for (; offset < end; offset += stride) {
        simplifiedFlatCoordinates[simplifiedOffset++] = flatCoordinates[offset];
        simplifiedFlatCoordinates[simplifiedOffset++] =
          flatCoordinates[offset + 1];
      }
      return simplifiedOffset;
    }
    let x1 = flatCoordinates[offset];
    let y1 = flatCoordinates[offset + 1];
    // copy first point
    simplifiedFlatCoordinates[simplifiedOffset++] = x1;
    simplifiedFlatCoordinates[simplifiedOffset++] = y1;
    let x2 = x1;
    let y2 = y1;
    for (offset += stride; offset < end; offset += stride) {
      x2 = flatCoordinates[offset];
      y2 = flatCoordinates[offset + 1];
      if (squaredDistance$1(x1, y1, x2, y2) > squaredTolerance) {
        // copy point at offset
        simplifiedFlatCoordinates[simplifiedOffset++] = x2;
        simplifiedFlatCoordinates[simplifiedOffset++] = y2;
        x1 = x2;
        y1 = y2;
      }
    }
    if (x2 != x1 || y2 != y1) {
      // copy last point
      simplifiedFlatCoordinates[simplifiedOffset++] = x2;
      simplifiedFlatCoordinates[simplifiedOffset++] = y2;
    }
    return simplifiedOffset;
  }

  /**
   * @param {number} value Value.
   * @param {number} tolerance Tolerance.
   * @return {number} Rounded value.
   */
  function snap(value, tolerance) {
    return tolerance * Math.round(value / tolerance);
  }

  /**
   * Simplifies a line string using an algorithm designed by Tim Schaub.
   * Coordinates are snapped to the nearest value in a virtual grid and
   * consecutive duplicate coordinates are discarded.  This effectively preserves
   * topology as the simplification of any subsection of a line string is
   * independent of the rest of the line string.  This means that, for examples,
   * the common edge between two polygons will be simplified to the same line
   * string independently in both polygons.  This implementation uses a single
   * pass over the coordinates and eliminates intermediate collinear points.
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   * @param {number} tolerance Tolerance.
   * @param {Array<number>} simplifiedFlatCoordinates Simplified flat
   *     coordinates.
   * @param {number} simplifiedOffset Simplified offset.
   * @return {number} Simplified offset.
   */
  function quantize(
    flatCoordinates,
    offset,
    end,
    stride,
    tolerance,
    simplifiedFlatCoordinates,
    simplifiedOffset
  ) {
    // do nothing if the line is empty
    if (offset == end) {
      return simplifiedOffset;
    }
    // snap the first coordinate (P1)
    let x1 = snap(flatCoordinates[offset], tolerance);
    let y1 = snap(flatCoordinates[offset + 1], tolerance);
    offset += stride;
    // add the first coordinate to the output
    simplifiedFlatCoordinates[simplifiedOffset++] = x1;
    simplifiedFlatCoordinates[simplifiedOffset++] = y1;
    // find the next coordinate that does not snap to the same value as the first
    // coordinate (P2)
    let x2, y2;
    do {
      x2 = snap(flatCoordinates[offset], tolerance);
      y2 = snap(flatCoordinates[offset + 1], tolerance);
      offset += stride;
      if (offset == end) {
        // all coordinates snap to the same value, the line collapses to a point
        // push the last snapped value anyway to ensure that the output contains
        // at least two points
        // FIXME should we really return at least two points anyway?
        simplifiedFlatCoordinates[simplifiedOffset++] = x2;
        simplifiedFlatCoordinates[simplifiedOffset++] = y2;
        return simplifiedOffset;
      }
    } while (x2 == x1 && y2 == y1);
    while (offset < end) {
      // snap the next coordinate (P3)
      const x3 = snap(flatCoordinates[offset], tolerance);
      const y3 = snap(flatCoordinates[offset + 1], tolerance);
      offset += stride;
      // skip P3 if it is equal to P2
      if (x3 == x2 && y3 == y2) {
        continue;
      }
      // calculate the delta between P1 and P2
      const dx1 = x2 - x1;
      const dy1 = y2 - y1;
      // calculate the delta between P3 and P1
      const dx2 = x3 - x1;
      const dy2 = y3 - y1;
      // if P1, P2, and P3 are colinear and P3 is further from P1 than P2 is from
      // P1 in the same direction then P2 is on the straight line between P1 and
      // P3
      if (
        dx1 * dy2 == dy1 * dx2 &&
        ((dx1 < 0 && dx2 < dx1) || dx1 == dx2 || (dx1 > 0 && dx2 > dx1)) &&
        ((dy1 < 0 && dy2 < dy1) || dy1 == dy2 || (dy1 > 0 && dy2 > dy1))
      ) {
        // discard P2 and set P2 = P3
        x2 = x3;
        y2 = y3;
        continue;
      }
      // either P1, P2, and P3 are not colinear, or they are colinear but P3 is
      // between P3 and P1 or on the opposite half of the line to P2.  add P2,
      // and continue with P1 = P2 and P2 = P3
      simplifiedFlatCoordinates[simplifiedOffset++] = x2;
      simplifiedFlatCoordinates[simplifiedOffset++] = y2;
      x1 = x2;
      y1 = y2;
      x2 = x3;
      y2 = y3;
    }
    // add the last point (P2)
    simplifiedFlatCoordinates[simplifiedOffset++] = x2;
    simplifiedFlatCoordinates[simplifiedOffset++] = y2;
    return simplifiedOffset;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<number>} ends Ends.
   * @param {number} stride Stride.
   * @param {number} tolerance Tolerance.
   * @param {Array<number>} simplifiedFlatCoordinates Simplified flat
   *     coordinates.
   * @param {number} simplifiedOffset Simplified offset.
   * @param {Array<number>} simplifiedEnds Simplified ends.
   * @return {number} Simplified offset.
   */
  function quantizeArray(
    flatCoordinates,
    offset,
    ends,
    stride,
    tolerance,
    simplifiedFlatCoordinates,
    simplifiedOffset,
    simplifiedEnds
  ) {
    for (let i = 0, ii = ends.length; i < ii; ++i) {
      const end = ends[i];
      simplifiedOffset = quantize(
        flatCoordinates,
        offset,
        end,
        stride,
        tolerance,
        simplifiedFlatCoordinates,
        simplifiedOffset
      );
      simplifiedEnds.push(simplifiedOffset);
      offset = end;
    }
    return simplifiedOffset;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<Array<number>>} endss Endss.
   * @param {number} stride Stride.
   * @param {number} tolerance Tolerance.
   * @param {Array<number>} simplifiedFlatCoordinates Simplified flat
   *     coordinates.
   * @param {number} simplifiedOffset Simplified offset.
   * @param {Array<Array<number>>} simplifiedEndss Simplified endss.
   * @return {number} Simplified offset.
   */
  function quantizeMultiArray(
    flatCoordinates,
    offset,
    endss,
    stride,
    tolerance,
    simplifiedFlatCoordinates,
    simplifiedOffset,
    simplifiedEndss
  ) {
    for (let i = 0, ii = endss.length; i < ii; ++i) {
      const ends = endss[i];
      const simplifiedEnds = [];
      simplifiedOffset = quantizeArray(
        flatCoordinates,
        offset,
        ends,
        stride,
        tolerance,
        simplifiedFlatCoordinates,
        simplifiedOffset,
        simplifiedEnds
      );
      simplifiedEndss.push(simplifiedEnds);
      offset = ends[ends.length - 1];
    }
    return simplifiedOffset;
  }

  /**
   * @module ol/geom/flat/inflate
   */

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   * @param {Array<import("../../coordinate.js").Coordinate>} [coordinates] Coordinates.
   * @return {Array<import("../../coordinate.js").Coordinate>} Coordinates.
   */
  function inflateCoordinates(
    flatCoordinates,
    offset,
    end,
    stride,
    coordinates
  ) {
    coordinates = coordinates !== undefined ? coordinates : [];
    let i = 0;
    for (let j = offset; j < end; j += stride) {
      coordinates[i++] = flatCoordinates.slice(j, j + stride);
    }
    coordinates.length = i;
    return coordinates;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<number>} ends Ends.
   * @param {number} stride Stride.
   * @param {Array<Array<import("../../coordinate.js").Coordinate>>} [coordinatess] Coordinatess.
   * @return {Array<Array<import("../../coordinate.js").Coordinate>>} Coordinatess.
   */
  function inflateCoordinatesArray(
    flatCoordinates,
    offset,
    ends,
    stride,
    coordinatess
  ) {
    coordinatess = coordinatess !== undefined ? coordinatess : [];
    let i = 0;
    for (let j = 0, jj = ends.length; j < jj; ++j) {
      const end = ends[j];
      coordinatess[i++] = inflateCoordinates(
        flatCoordinates,
        offset,
        end,
        stride,
        coordinatess[i]
      );
      offset = end;
    }
    coordinatess.length = i;
    return coordinatess;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<Array<number>>} endss Endss.
   * @param {number} stride Stride.
   * @param {Array<Array<Array<import("../../coordinate.js").Coordinate>>>} [coordinatesss]
   *     Coordinatesss.
   * @return {Array<Array<Array<import("../../coordinate.js").Coordinate>>>} Coordinatesss.
   */
  function inflateMultiCoordinatesArray(
    flatCoordinates,
    offset,
    endss,
    stride,
    coordinatesss
  ) {
    coordinatesss = coordinatesss !== undefined ? coordinatesss : [];
    let i = 0;
    for (let j = 0, jj = endss.length; j < jj; ++j) {
      const ends = endss[j];
      coordinatesss[i++] =
        ends.length === 1 && ends[0] === offset
          ? []
          : inflateCoordinatesArray(
              flatCoordinates,
              offset,
              ends,
              stride,
              coordinatesss[i]
            );
      offset = ends[ends.length - 1];
    }
    coordinatesss.length = i;
    return coordinatesss;
  }

  /**
   * @module ol/geom/flat/area
   */

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   * @return {number} Area.
   */
  function linearRing(flatCoordinates, offset, end, stride) {
    let twiceArea = 0;
    let x1 = flatCoordinates[end - stride];
    let y1 = flatCoordinates[end - stride + 1];
    for (; offset < end; offset += stride) {
      const x2 = flatCoordinates[offset];
      const y2 = flatCoordinates[offset + 1];
      twiceArea += y1 * x2 - x1 * y2;
      x1 = x2;
      y1 = y2;
    }
    return twiceArea / 2;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<number>} ends Ends.
   * @param {number} stride Stride.
   * @return {number} Area.
   */
  function linearRings(flatCoordinates, offset, ends, stride) {
    let area = 0;
    for (let i = 0, ii = ends.length; i < ii; ++i) {
      const end = ends[i];
      area += linearRing(flatCoordinates, offset, end, stride);
      offset = end;
    }
    return area;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<Array<number>>} endss Endss.
   * @param {number} stride Stride.
   * @return {number} Area.
   */
  function linearRingss$1(flatCoordinates, offset, endss, stride) {
    let area = 0;
    for (let i = 0, ii = endss.length; i < ii; ++i) {
      const ends = endss[i];
      area += linearRings(flatCoordinates, offset, ends, stride);
      offset = ends[ends.length - 1];
    }
    return area;
  }

  /**
   * @module ol/geom/LinearRing
   */

  /**
   * @classdesc
   * Linear ring geometry. Only used as part of polygon; cannot be rendered
   * on its own.
   *
   * @api
   */
  class LinearRing extends SimpleGeometry {
    /**
     * @param {Array<import("../coordinate.js").Coordinate>|Array<number>} coordinates Coordinates.
     *     For internal use, flat coordinates in combination with `layout` are also accepted.
     * @param {import("./Geometry.js").GeometryLayout} [layout] Layout.
     */
    constructor(coordinates, layout) {
      super();

      /**
       * @private
       * @type {number}
       */
      this.maxDelta_ = -1;

      /**
       * @private
       * @type {number}
       */
      this.maxDeltaRevision_ = -1;

      if (layout !== undefined && !Array.isArray(coordinates[0])) {
        this.setFlatCoordinates(
          layout,
          /** @type {Array<number>} */ (coordinates)
        );
      } else {
        this.setCoordinates(
          /** @type {Array<import("../coordinate.js").Coordinate>} */ (
            coordinates
          ),
          layout
        );
      }
    }

    /**
     * Make a complete copy of the geometry.
     * @return {!LinearRing} Clone.
     * @api
     */
    clone() {
      return new LinearRing(this.flatCoordinates.slice(), this.layout);
    }

    /**
     * @param {number} x X.
     * @param {number} y Y.
     * @param {import("../coordinate.js").Coordinate} closestPoint Closest point.
     * @param {number} minSquaredDistance Minimum squared distance.
     * @return {number} Minimum squared distance.
     */
    closestPointXY(x, y, closestPoint, minSquaredDistance) {
      if (minSquaredDistance < closestSquaredDistanceXY(this.getExtent(), x, y)) {
        return minSquaredDistance;
      }
      if (this.maxDeltaRevision_ != this.getRevision()) {
        this.maxDelta_ = Math.sqrt(
          maxSquaredDelta(
            this.flatCoordinates,
            0,
            this.flatCoordinates.length,
            this.stride,
            0
          )
        );
        this.maxDeltaRevision_ = this.getRevision();
      }
      return assignClosestPoint(
        this.flatCoordinates,
        0,
        this.flatCoordinates.length,
        this.stride,
        this.maxDelta_,
        true,
        x,
        y,
        closestPoint,
        minSquaredDistance
      );
    }

    /**
     * Return the area of the linear ring on projected plane.
     * @return {number} Area (on projected plane).
     * @api
     */
    getArea() {
      return linearRing(
        this.flatCoordinates,
        0,
        this.flatCoordinates.length,
        this.stride
      );
    }

    /**
     * Return the coordinates of the linear ring.
     * @return {Array<import("../coordinate.js").Coordinate>} Coordinates.
     * @api
     */
    getCoordinates() {
      return inflateCoordinates(
        this.flatCoordinates,
        0,
        this.flatCoordinates.length,
        this.stride
      );
    }

    /**
     * @param {number} squaredTolerance Squared tolerance.
     * @return {LinearRing} Simplified LinearRing.
     * @protected
     */
    getSimplifiedGeometryInternal(squaredTolerance) {
      const simplifiedFlatCoordinates = [];
      simplifiedFlatCoordinates.length = douglasPeucker(
        this.flatCoordinates,
        0,
        this.flatCoordinates.length,
        this.stride,
        squaredTolerance,
        simplifiedFlatCoordinates,
        0
      );
      return new LinearRing(simplifiedFlatCoordinates, 'XY');
    }

    /**
     * Get the type of this geometry.
     * @return {import("./Geometry.js").Type} Geometry type.
     * @api
     */
    getType() {
      return 'LinearRing';
    }

    /**
     * Test if the geometry and the passed extent intersect.
     * @param {import("../extent.js").Extent} extent Extent.
     * @return {boolean} `true` if the geometry and the extent intersect.
     * @api
     */
    intersectsExtent(extent) {
      return false;
    }

    /**
     * Set the coordinates of the linear ring.
     * @param {!Array<import("../coordinate.js").Coordinate>} coordinates Coordinates.
     * @param {import("./Geometry.js").GeometryLayout} [layout] Layout.
     * @api
     */
    setCoordinates(coordinates, layout) {
      this.setLayout(layout, coordinates, 1);
      if (!this.flatCoordinates) {
        this.flatCoordinates = [];
      }
      this.flatCoordinates.length = deflateCoordinates(
        this.flatCoordinates,
        0,
        coordinates,
        this.stride
      );
      this.changed();
    }
  }

  /**
   * @module ol/geom/Point
   */

  /**
   * @classdesc
   * Point geometry.
   *
   * @api
   */
  class Point extends SimpleGeometry {
    /**
     * @param {import("../coordinate.js").Coordinate} coordinates Coordinates.
     * @param {import("./Geometry.js").GeometryLayout} [layout] Layout.
     */
    constructor(coordinates, layout) {
      super();
      this.setCoordinates(coordinates, layout);
    }

    /**
     * Make a complete copy of the geometry.
     * @return {!Point} Clone.
     * @api
     */
    clone() {
      const point = new Point(this.flatCoordinates.slice(), this.layout);
      point.applyProperties(this);
      return point;
    }

    /**
     * @param {number} x X.
     * @param {number} y Y.
     * @param {import("../coordinate.js").Coordinate} closestPoint Closest point.
     * @param {number} minSquaredDistance Minimum squared distance.
     * @return {number} Minimum squared distance.
     */
    closestPointXY(x, y, closestPoint, minSquaredDistance) {
      const flatCoordinates = this.flatCoordinates;
      const squaredDistance = squaredDistance$1(
        x,
        y,
        flatCoordinates[0],
        flatCoordinates[1]
      );
      if (squaredDistance < minSquaredDistance) {
        const stride = this.stride;
        for (let i = 0; i < stride; ++i) {
          closestPoint[i] = flatCoordinates[i];
        }
        closestPoint.length = stride;
        return squaredDistance;
      } else {
        return minSquaredDistance;
      }
    }

    /**
     * Return the coordinate of the point.
     * @return {import("../coordinate.js").Coordinate} Coordinates.
     * @api
     */
    getCoordinates() {
      return !this.flatCoordinates ? [] : this.flatCoordinates.slice();
    }

    /**
     * @param {import("../extent.js").Extent} extent Extent.
     * @protected
     * @return {import("../extent.js").Extent} extent Extent.
     */
    computeExtent(extent) {
      return createOrUpdateFromCoordinate(this.flatCoordinates, extent);
    }

    /**
     * Get the type of this geometry.
     * @return {import("./Geometry.js").Type} Geometry type.
     * @api
     */
    getType() {
      return 'Point';
    }

    /**
     * Test if the geometry and the passed extent intersect.
     * @param {import("../extent.js").Extent} extent Extent.
     * @return {boolean} `true` if the geometry and the extent intersect.
     * @api
     */
    intersectsExtent(extent) {
      return containsXY(extent, this.flatCoordinates[0], this.flatCoordinates[1]);
    }

    /**
     * @param {!Array<*>} coordinates Coordinates.
     * @param {import("./Geometry.js").GeometryLayout} [layout] Layout.
     * @api
     */
    setCoordinates(coordinates, layout) {
      this.setLayout(layout, coordinates, 0);
      if (!this.flatCoordinates) {
        this.flatCoordinates = [];
      }
      this.flatCoordinates.length = deflateCoordinate(
        this.flatCoordinates,
        0,
        coordinates,
        this.stride
      );
      this.changed();
    }
  }

  /**
   * @module ol/geom/flat/contains
   */

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   * @param {import("../../extent.js").Extent} extent Extent.
   * @return {boolean} Contains extent.
   */
  function linearRingContainsExtent(
    flatCoordinates,
    offset,
    end,
    stride,
    extent
  ) {
    const outside = forEachCorner(
      extent,
      /**
       * @param {import("../../coordinate.js").Coordinate} coordinate Coordinate.
       * @return {boolean} Contains (x, y).
       */
      function (coordinate) {
        return !linearRingContainsXY(
          flatCoordinates,
          offset,
          end,
          stride,
          coordinate[0],
          coordinate[1]
        );
      }
    );
    return !outside;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   * @param {number} x X.
   * @param {number} y Y.
   * @return {boolean} Contains (x, y).
   */
  function linearRingContainsXY(
    flatCoordinates,
    offset,
    end,
    stride,
    x,
    y
  ) {
    // https://geomalgorithms.com/a03-_inclusion.html
    // Copyright 2000 softSurfer, 2012 Dan Sunday
    // This code may be freely used and modified for any purpose
    // providing that this copyright notice is included with it.
    // SoftSurfer makes no warranty for this code, and cannot be held
    // liable for any real or imagined damage resulting from its use.
    // Users of this code must verify correctness for their application.
    let wn = 0;
    let x1 = flatCoordinates[end - stride];
    let y1 = flatCoordinates[end - stride + 1];
    for (; offset < end; offset += stride) {
      const x2 = flatCoordinates[offset];
      const y2 = flatCoordinates[offset + 1];
      if (y1 <= y) {
        if (y2 > y && (x2 - x1) * (y - y1) - (x - x1) * (y2 - y1) > 0) {
          wn++;
        }
      } else if (y2 <= y && (x2 - x1) * (y - y1) - (x - x1) * (y2 - y1) < 0) {
        wn--;
      }
      x1 = x2;
      y1 = y2;
    }
    return wn !== 0;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<number>} ends Ends.
   * @param {number} stride Stride.
   * @param {number} x X.
   * @param {number} y Y.
   * @return {boolean} Contains (x, y).
   */
  function linearRingsContainsXY(
    flatCoordinates,
    offset,
    ends,
    stride,
    x,
    y
  ) {
    if (ends.length === 0) {
      return false;
    }
    if (!linearRingContainsXY(flatCoordinates, offset, ends[0], stride, x, y)) {
      return false;
    }
    for (let i = 1, ii = ends.length; i < ii; ++i) {
      if (
        linearRingContainsXY(flatCoordinates, ends[i - 1], ends[i], stride, x, y)
      ) {
        return false;
      }
    }
    return true;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<Array<number>>} endss Endss.
   * @param {number} stride Stride.
   * @param {number} x X.
   * @param {number} y Y.
   * @return {boolean} Contains (x, y).
   */
  function linearRingssContainsXY(
    flatCoordinates,
    offset,
    endss,
    stride,
    x,
    y
  ) {
    if (endss.length === 0) {
      return false;
    }
    for (let i = 0, ii = endss.length; i < ii; ++i) {
      const ends = endss[i];
      if (linearRingsContainsXY(flatCoordinates, offset, ends, stride, x, y)) {
        return true;
      }
      offset = ends[ends.length - 1];
    }
    return false;
  }

  /**
   * @module ol/geom/flat/interiorpoint
   */

  /**
   * Calculates a point that is likely to lie in the interior of the linear rings.
   * Inspired by JTS's com.vividsolutions.jts.geom.Geometry#getInteriorPoint.
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<number>} ends Ends.
   * @param {number} stride Stride.
   * @param {Array<number>} flatCenters Flat centers.
   * @param {number} flatCentersOffset Flat center offset.
   * @param {Array<number>} [dest] Destination.
   * @return {Array<number>} Destination point as XYM coordinate, where M is the
   * length of the horizontal intersection that the point belongs to.
   */
  function getInteriorPointOfArray(
    flatCoordinates,
    offset,
    ends,
    stride,
    flatCenters,
    flatCentersOffset,
    dest
  ) {
    let i, ii, x, x1, x2, y1, y2;
    const y = flatCenters[flatCentersOffset + 1];
    /** @type {Array<number>} */
    const intersections = [];
    // Calculate intersections with the horizontal line
    for (let r = 0, rr = ends.length; r < rr; ++r) {
      const end = ends[r];
      x1 = flatCoordinates[end - stride];
      y1 = flatCoordinates[end - stride + 1];
      for (i = offset; i < end; i += stride) {
        x2 = flatCoordinates[i];
        y2 = flatCoordinates[i + 1];
        if ((y <= y1 && y2 <= y) || (y1 <= y && y <= y2)) {
          x = ((y - y1) / (y2 - y1)) * (x2 - x1) + x1;
          intersections.push(x);
        }
        x1 = x2;
        y1 = y2;
      }
    }
    // Find the longest segment of the horizontal line that has its center point
    // inside the linear ring.
    let pointX = NaN;
    let maxSegmentLength = -Infinity;
    intersections.sort(numberSafeCompareFunction);
    x1 = intersections[0];
    for (i = 1, ii = intersections.length; i < ii; ++i) {
      x2 = intersections[i];
      const segmentLength = Math.abs(x2 - x1);
      if (segmentLength > maxSegmentLength) {
        x = (x1 + x2) / 2;
        if (linearRingsContainsXY(flatCoordinates, offset, ends, stride, x, y)) {
          pointX = x;
          maxSegmentLength = segmentLength;
        }
      }
      x1 = x2;
    }
    if (isNaN(pointX)) {
      // There is no horizontal line that has its center point inside the linear
      // ring.  Use the center of the the linear ring's extent.
      pointX = flatCenters[flatCentersOffset];
    }
    if (dest) {
      dest.push(pointX, y, maxSegmentLength);
      return dest;
    } else {
      return [pointX, y, maxSegmentLength];
    }
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<Array<number>>} endss Endss.
   * @param {number} stride Stride.
   * @param {Array<number>} flatCenters Flat centers.
   * @return {Array<number>} Interior points as XYM coordinates, where M is the
   * length of the horizontal intersection that the point belongs to.
   */
  function getInteriorPointsOfMultiArray(
    flatCoordinates,
    offset,
    endss,
    stride,
    flatCenters
  ) {
    let interiorPoints = [];
    for (let i = 0, ii = endss.length; i < ii; ++i) {
      const ends = endss[i];
      interiorPoints = getInteriorPointOfArray(
        flatCoordinates,
        offset,
        ends,
        stride,
        flatCenters,
        2 * i,
        interiorPoints
      );
      offset = ends[ends.length - 1];
    }
    return interiorPoints;
  }

  /**
   * @module ol/geom/flat/segments
   */

  /**
   * This function calls `callback` for each segment of the flat coordinates
   * array. If the callback returns a truthy value the function returns that
   * value immediately. Otherwise the function returns `false`.
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   * @param {function(import("../../coordinate.js").Coordinate, import("../../coordinate.js").Coordinate): T} callback Function
   *     called for each segment.
   * @return {T|boolean} Value.
   * @template T
   */
  function forEach(flatCoordinates, offset, end, stride, callback) {
    let ret;
    offset += stride;
    for (; offset < end; offset += stride) {
      ret = callback(
        flatCoordinates.slice(offset - stride, offset),
        flatCoordinates.slice(offset, offset + stride)
      );
      if (ret) {
        return ret;
      }
    }
    return false;
  }

  /**
   * @module ol/geom/flat/intersectsextent
   */

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   * @param {import("../../extent.js").Extent} extent Extent.
   * @return {boolean} True if the geometry and the extent intersect.
   */
  function intersectsLineString(
    flatCoordinates,
    offset,
    end,
    stride,
    extent
  ) {
    const coordinatesExtent = extendFlatCoordinates(
      createEmpty(),
      flatCoordinates,
      offset,
      end,
      stride
    );
    if (!intersects$3(extent, coordinatesExtent)) {
      return false;
    }
    if (containsExtent(extent, coordinatesExtent)) {
      return true;
    }
    if (coordinatesExtent[0] >= extent[0] && coordinatesExtent[2] <= extent[2]) {
      return true;
    }
    if (coordinatesExtent[1] >= extent[1] && coordinatesExtent[3] <= extent[3]) {
      return true;
    }
    return forEach(
      flatCoordinates,
      offset,
      end,
      stride,
      /**
       * @param {import("../../coordinate.js").Coordinate} point1 Start point.
       * @param {import("../../coordinate.js").Coordinate} point2 End point.
       * @return {boolean} `true` if the segment and the extent intersect,
       *     `false` otherwise.
       */
      function (point1, point2) {
        return intersectsSegment(extent, point1, point2);
      }
    );
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<number>} ends Ends.
   * @param {number} stride Stride.
   * @param {import("../../extent.js").Extent} extent Extent.
   * @return {boolean} True if the geometry and the extent intersect.
   */
  function intersectsLineStringArray(
    flatCoordinates,
    offset,
    ends,
    stride,
    extent
  ) {
    for (let i = 0, ii = ends.length; i < ii; ++i) {
      if (
        intersectsLineString(flatCoordinates, offset, ends[i], stride, extent)
      ) {
        return true;
      }
      offset = ends[i];
    }
    return false;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   * @param {import("../../extent.js").Extent} extent Extent.
   * @return {boolean} True if the geometry and the extent intersect.
   */
  function intersectsLinearRing(
    flatCoordinates,
    offset,
    end,
    stride,
    extent
  ) {
    if (intersectsLineString(flatCoordinates, offset, end, stride, extent)) {
      return true;
    }
    if (
      linearRingContainsXY(
        flatCoordinates,
        offset,
        end,
        stride,
        extent[0],
        extent[1]
      )
    ) {
      return true;
    }
    if (
      linearRingContainsXY(
        flatCoordinates,
        offset,
        end,
        stride,
        extent[0],
        extent[3]
      )
    ) {
      return true;
    }
    if (
      linearRingContainsXY(
        flatCoordinates,
        offset,
        end,
        stride,
        extent[2],
        extent[1]
      )
    ) {
      return true;
    }
    if (
      linearRingContainsXY(
        flatCoordinates,
        offset,
        end,
        stride,
        extent[2],
        extent[3]
      )
    ) {
      return true;
    }
    return false;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<number>} ends Ends.
   * @param {number} stride Stride.
   * @param {import("../../extent.js").Extent} extent Extent.
   * @return {boolean} True if the geometry and the extent intersect.
   */
  function intersectsLinearRingArray(
    flatCoordinates,
    offset,
    ends,
    stride,
    extent
  ) {
    if (!intersectsLinearRing(flatCoordinates, offset, ends[0], stride, extent)) {
      return false;
    }
    if (ends.length === 1) {
      return true;
    }
    for (let i = 1, ii = ends.length; i < ii; ++i) {
      if (
        linearRingContainsExtent(
          flatCoordinates,
          ends[i - 1],
          ends[i],
          stride,
          extent
        )
      ) {
        if (
          !intersectsLineString(
            flatCoordinates,
            ends[i - 1],
            ends[i],
            stride,
            extent
          )
        ) {
          return false;
        }
      }
    }
    return true;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<Array<number>>} endss Endss.
   * @param {number} stride Stride.
   * @param {import("../../extent.js").Extent} extent Extent.
   * @return {boolean} True if the geometry and the extent intersect.
   */
  function intersectsLinearRingMultiArray(
    flatCoordinates,
    offset,
    endss,
    stride,
    extent
  ) {
    for (let i = 0, ii = endss.length; i < ii; ++i) {
      const ends = endss[i];
      if (
        intersectsLinearRingArray(flatCoordinates, offset, ends, stride, extent)
      ) {
        return true;
      }
      offset = ends[ends.length - 1];
    }
    return false;
  }

  /**
   * @module ol/geom/flat/reverse
   */

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   */
  function coordinates(flatCoordinates, offset, end, stride) {
    while (offset < end - stride) {
      for (let i = 0; i < stride; ++i) {
        const tmp = flatCoordinates[offset + i];
        flatCoordinates[offset + i] = flatCoordinates[end - stride + i];
        flatCoordinates[end - stride + i] = tmp;
      }
      offset += stride;
      end -= stride;
    }
  }

  /**
   * @module ol/geom/flat/orient
   */

  /**
   * Is the linear ring oriented clockwise in a coordinate system with a bottom-left
   * coordinate origin? For a coordinate system with a top-left coordinate origin,
   * the ring's orientation is clockwise when this function returns false.
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   * @return {boolean} Is clockwise.
   */
  function linearRingIsClockwise(flatCoordinates, offset, end, stride) {
    // https://stackoverflow.com/q/1165647/clockwise-method#1165943
    // https://github.com/OSGeo/gdal/blob/master/gdal/ogr/ogrlinearring.cpp
    let edge = 0;
    let x1 = flatCoordinates[end - stride];
    let y1 = flatCoordinates[end - stride + 1];
    for (; offset < end; offset += stride) {
      const x2 = flatCoordinates[offset];
      const y2 = flatCoordinates[offset + 1];
      edge += (x2 - x1) * (y2 + y1);
      x1 = x2;
      y1 = y2;
    }
    return edge === 0 ? undefined : edge > 0;
  }

  /**
   * Determines if linear rings are oriented.  By default, left-hand orientation
   * is tested (first ring must be clockwise, remaining rings counter-clockwise).
   * To test for right-hand orientation, use the `right` argument.
   *
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<number>} ends Array of end indexes.
   * @param {number} stride Stride.
   * @param {boolean} [right] Test for right-hand orientation
   *     (counter-clockwise exterior ring and clockwise interior rings).
   * @return {boolean} Rings are correctly oriented.
   */
  function linearRingsAreOriented(
    flatCoordinates,
    offset,
    ends,
    stride,
    right
  ) {
    right = right !== undefined ? right : false;
    for (let i = 0, ii = ends.length; i < ii; ++i) {
      const end = ends[i];
      const isClockwise = linearRingIsClockwise(
        flatCoordinates,
        offset,
        end,
        stride
      );
      if (i === 0) {
        if ((right && isClockwise) || (!right && !isClockwise)) {
          return false;
        }
      } else {
        if ((right && !isClockwise) || (!right && isClockwise)) {
          return false;
        }
      }
      offset = end;
    }
    return true;
  }

  /**
   * Determines if linear rings are oriented.  By default, left-hand orientation
   * is tested (first ring must be clockwise, remaining rings counter-clockwise).
   * To test for right-hand orientation, use the `right` argument.
   *
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<Array<number>>} endss Array of array of end indexes.
   * @param {number} stride Stride.
   * @param {boolean} [right] Test for right-hand orientation
   *     (counter-clockwise exterior ring and clockwise interior rings).
   * @return {boolean} Rings are correctly oriented.
   */
  function linearRingssAreOriented(
    flatCoordinates,
    offset,
    endss,
    stride,
    right
  ) {
    for (let i = 0, ii = endss.length; i < ii; ++i) {
      const ends = endss[i];
      if (!linearRingsAreOriented(flatCoordinates, offset, ends, stride, right)) {
        return false;
      }
      if (ends.length) {
        offset = ends[ends.length - 1];
      }
    }
    return true;
  }

  /**
   * Orient coordinates in a flat array of linear rings.  By default, rings
   * are oriented following the left-hand rule (clockwise for exterior and
   * counter-clockwise for interior rings).  To orient according to the
   * right-hand rule, use the `right` argument.
   *
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<number>} ends Ends.
   * @param {number} stride Stride.
   * @param {boolean} [right] Follow the right-hand rule for orientation.
   * @return {number} End.
   */
  function orientLinearRings(
    flatCoordinates,
    offset,
    ends,
    stride,
    right
  ) {
    right = right !== undefined ? right : false;
    for (let i = 0, ii = ends.length; i < ii; ++i) {
      const end = ends[i];
      const isClockwise = linearRingIsClockwise(
        flatCoordinates,
        offset,
        end,
        stride
      );
      const reverse =
        i === 0
          ? (right && isClockwise) || (!right && !isClockwise)
          : (right && !isClockwise) || (!right && isClockwise);
      if (reverse) {
        coordinates(flatCoordinates, offset, end, stride);
      }
      offset = end;
    }
    return offset;
  }

  /**
   * Orient coordinates in a flat array of linear rings.  By default, rings
   * are oriented following the left-hand rule (clockwise for exterior and
   * counter-clockwise for interior rings).  To orient according to the
   * right-hand rule, use the `right` argument.
   *
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<Array<number>>} endss Array of array of end indexes.
   * @param {number} stride Stride.
   * @param {boolean} [right] Follow the right-hand rule for orientation.
   * @return {number} End.
   */
  function orientLinearRingsArray(
    flatCoordinates,
    offset,
    endss,
    stride,
    right
  ) {
    for (let i = 0, ii = endss.length; i < ii; ++i) {
      offset = orientLinearRings(
        flatCoordinates,
        offset,
        endss[i],
        stride,
        right
      );
    }
    return offset;
  }

  /**
   * Return a two-dimensional endss
   * @param {Array<number>} flatCoordinates Flat coordinates
   * @param {Array<number>} ends Linear ring end indexes
   * @return {Array<Array<number>>} Two dimensional endss array that can
   * be used to contruct a MultiPolygon
   */
  function inflateEnds(flatCoordinates, ends) {
    const endss = [];
    let offset = 0;
    let prevEndIndex = 0;
    for (let i = 0, ii = ends.length; i < ii; ++i) {
      const end = ends[i];
      // classifies an array of rings into polygons with outer rings and holes
      if (!linearRingIsClockwise(flatCoordinates, offset, end, 2)) {
        endss.push(ends.slice(prevEndIndex, i + 1));
      } else {
        if (endss.length === 0) {
          continue;
        }
        endss[endss.length - 1].push(ends[prevEndIndex]);
      }
      prevEndIndex = i + 1;
      offset = end;
    }
    return endss;
  }

  /**
   * @module ol/geom/Polygon
   */

  /**
   * @classdesc
   * Polygon geometry.
   *
   * @api
   */
  class Polygon extends SimpleGeometry {
    /**
     * @param {!Array<Array<import("../coordinate.js").Coordinate>>|!Array<number>} coordinates
     *     Array of linear rings that define the polygon. The first linear ring of the
     *     array defines the outer-boundary or surface of the polygon. Each subsequent
     *     linear ring defines a hole in the surface of the polygon. A linear ring is
     *     an array of vertices' coordinates where the first coordinate and the last are
     *     equivalent. (For internal use, flat coordinates in combination with
     *     `layout` and `ends` are also accepted.)
     * @param {import("./Geometry.js").GeometryLayout} [layout] Layout.
     * @param {Array<number>} [ends] Ends (for internal use with flat coordinates).
     */
    constructor(coordinates, layout, ends) {
      super();

      /**
       * @type {Array<number>}
       * @private
       */
      this.ends_ = [];

      /**
       * @private
       * @type {number}
       */
      this.flatInteriorPointRevision_ = -1;

      /**
       * @private
       * @type {import("../coordinate.js").Coordinate}
       */
      this.flatInteriorPoint_ = null;

      /**
       * @private
       * @type {number}
       */
      this.maxDelta_ = -1;

      /**
       * @private
       * @type {number}
       */
      this.maxDeltaRevision_ = -1;

      /**
       * @private
       * @type {number}
       */
      this.orientedRevision_ = -1;

      /**
       * @private
       * @type {Array<number>}
       */
      this.orientedFlatCoordinates_ = null;

      if (layout !== undefined && ends) {
        this.setFlatCoordinates(
          layout,
          /** @type {Array<number>} */ (coordinates)
        );
        this.ends_ = ends;
      } else {
        this.setCoordinates(
          /** @type {Array<Array<import("../coordinate.js").Coordinate>>} */ (
            coordinates
          ),
          layout
        );
      }
    }

    /**
     * Append the passed linear ring to this polygon.
     * @param {LinearRing} linearRing Linear ring.
     * @api
     */
    appendLinearRing(linearRing) {
      if (!this.flatCoordinates) {
        this.flatCoordinates = linearRing.getFlatCoordinates().slice();
      } else {
        extend$2(this.flatCoordinates, linearRing.getFlatCoordinates());
      }
      this.ends_.push(this.flatCoordinates.length);
      this.changed();
    }

    /**
     * Make a complete copy of the geometry.
     * @return {!Polygon} Clone.
     * @api
     */
    clone() {
      const polygon = new Polygon(
        this.flatCoordinates.slice(),
        this.layout,
        this.ends_.slice()
      );
      polygon.applyProperties(this);
      return polygon;
    }

    /**
     * @param {number} x X.
     * @param {number} y Y.
     * @param {import("../coordinate.js").Coordinate} closestPoint Closest point.
     * @param {number} minSquaredDistance Minimum squared distance.
     * @return {number} Minimum squared distance.
     */
    closestPointXY(x, y, closestPoint, minSquaredDistance) {
      if (minSquaredDistance < closestSquaredDistanceXY(this.getExtent(), x, y)) {
        return minSquaredDistance;
      }
      if (this.maxDeltaRevision_ != this.getRevision()) {
        this.maxDelta_ = Math.sqrt(
          arrayMaxSquaredDelta(
            this.flatCoordinates,
            0,
            this.ends_,
            this.stride,
            0
          )
        );
        this.maxDeltaRevision_ = this.getRevision();
      }
      return assignClosestArrayPoint(
        this.flatCoordinates,
        0,
        this.ends_,
        this.stride,
        this.maxDelta_,
        true,
        x,
        y,
        closestPoint,
        minSquaredDistance
      );
    }

    /**
     * @param {number} x X.
     * @param {number} y Y.
     * @return {boolean} Contains (x, y).
     */
    containsXY(x, y) {
      return linearRingsContainsXY(
        this.getOrientedFlatCoordinates(),
        0,
        this.ends_,
        this.stride,
        x,
        y
      );
    }

    /**
     * Return the area of the polygon on projected plane.
     * @return {number} Area (on projected plane).
     * @api
     */
    getArea() {
      return linearRings(
        this.getOrientedFlatCoordinates(),
        0,
        this.ends_,
        this.stride
      );
    }

    /**
     * Get the coordinate array for this geometry.  This array has the structure
     * of a GeoJSON coordinate array for polygons.
     *
     * @param {boolean} [right] Orient coordinates according to the right-hand
     *     rule (counter-clockwise for exterior and clockwise for interior rings).
     *     If `false`, coordinates will be oriented according to the left-hand rule
     *     (clockwise for exterior and counter-clockwise for interior rings).
     *     By default, coordinate orientation will depend on how the geometry was
     *     constructed.
     * @return {Array<Array<import("../coordinate.js").Coordinate>>} Coordinates.
     * @api
     */
    getCoordinates(right) {
      let flatCoordinates;
      if (right !== undefined) {
        flatCoordinates = this.getOrientedFlatCoordinates().slice();
        orientLinearRings(flatCoordinates, 0, this.ends_, this.stride, right);
      } else {
        flatCoordinates = this.flatCoordinates;
      }

      return inflateCoordinatesArray(flatCoordinates, 0, this.ends_, this.stride);
    }

    /**
     * @return {Array<number>} Ends.
     */
    getEnds() {
      return this.ends_;
    }

    /**
     * @return {Array<number>} Interior point.
     */
    getFlatInteriorPoint() {
      if (this.flatInteriorPointRevision_ != this.getRevision()) {
        const flatCenter = getCenter(this.getExtent());
        this.flatInteriorPoint_ = getInteriorPointOfArray(
          this.getOrientedFlatCoordinates(),
          0,
          this.ends_,
          this.stride,
          flatCenter,
          0
        );
        this.flatInteriorPointRevision_ = this.getRevision();
      }
      return this.flatInteriorPoint_;
    }

    /**
     * Return an interior point of the polygon.
     * @return {Point} Interior point as XYM coordinate, where M is the
     * length of the horizontal intersection that the point belongs to.
     * @api
     */
    getInteriorPoint() {
      return new Point(this.getFlatInteriorPoint(), 'XYM');
    }

    /**
     * Return the number of rings of the polygon,  this includes the exterior
     * ring and any interior rings.
     *
     * @return {number} Number of rings.
     * @api
     */
    getLinearRingCount() {
      return this.ends_.length;
    }

    /**
     * Return the Nth linear ring of the polygon geometry. Return `null` if the
     * given index is out of range.
     * The exterior linear ring is available at index `0` and the interior rings
     * at index `1` and beyond.
     *
     * @param {number} index Index.
     * @return {LinearRing|null} Linear ring.
     * @api
     */
    getLinearRing(index) {
      if (index < 0 || this.ends_.length <= index) {
        return null;
      }
      return new LinearRing(
        this.flatCoordinates.slice(
          index === 0 ? 0 : this.ends_[index - 1],
          this.ends_[index]
        ),
        this.layout
      );
    }

    /**
     * Return the linear rings of the polygon.
     * @return {Array<LinearRing>} Linear rings.
     * @api
     */
    getLinearRings() {
      const layout = this.layout;
      const flatCoordinates = this.flatCoordinates;
      const ends = this.ends_;
      const linearRings = [];
      let offset = 0;
      for (let i = 0, ii = ends.length; i < ii; ++i) {
        const end = ends[i];
        const linearRing = new LinearRing(
          flatCoordinates.slice(offset, end),
          layout
        );
        linearRings.push(linearRing);
        offset = end;
      }
      return linearRings;
    }

    /**
     * @return {Array<number>} Oriented flat coordinates.
     */
    getOrientedFlatCoordinates() {
      if (this.orientedRevision_ != this.getRevision()) {
        const flatCoordinates = this.flatCoordinates;
        if (linearRingsAreOriented(flatCoordinates, 0, this.ends_, this.stride)) {
          this.orientedFlatCoordinates_ = flatCoordinates;
        } else {
          this.orientedFlatCoordinates_ = flatCoordinates.slice();
          this.orientedFlatCoordinates_.length = orientLinearRings(
            this.orientedFlatCoordinates_,
            0,
            this.ends_,
            this.stride
          );
        }
        this.orientedRevision_ = this.getRevision();
      }
      return this.orientedFlatCoordinates_;
    }

    /**
     * @param {number} squaredTolerance Squared tolerance.
     * @return {Polygon} Simplified Polygon.
     * @protected
     */
    getSimplifiedGeometryInternal(squaredTolerance) {
      const simplifiedFlatCoordinates = [];
      const simplifiedEnds = [];
      simplifiedFlatCoordinates.length = quantizeArray(
        this.flatCoordinates,
        0,
        this.ends_,
        this.stride,
        Math.sqrt(squaredTolerance),
        simplifiedFlatCoordinates,
        0,
        simplifiedEnds
      );
      return new Polygon(simplifiedFlatCoordinates, 'XY', simplifiedEnds);
    }

    /**
     * Get the type of this geometry.
     * @return {import("./Geometry.js").Type} Geometry type.
     * @api
     */
    getType() {
      return 'Polygon';
    }

    /**
     * Test if the geometry and the passed extent intersect.
     * @param {import("../extent.js").Extent} extent Extent.
     * @return {boolean} `true` if the geometry and the extent intersect.
     * @api
     */
    intersectsExtent(extent) {
      return intersectsLinearRingArray(
        this.getOrientedFlatCoordinates(),
        0,
        this.ends_,
        this.stride,
        extent
      );
    }

    /**
     * Set the coordinates of the polygon.
     * @param {!Array<Array<import("../coordinate.js").Coordinate>>} coordinates Coordinates.
     * @param {import("./Geometry.js").GeometryLayout} [layout] Layout.
     * @api
     */
    setCoordinates(coordinates, layout) {
      this.setLayout(layout, coordinates, 2);
      if (!this.flatCoordinates) {
        this.flatCoordinates = [];
      }
      const ends = deflateCoordinatesArray(
        this.flatCoordinates,
        0,
        coordinates,
        this.stride,
        this.ends_
      );
      this.flatCoordinates.length = ends.length === 0 ? 0 : ends[ends.length - 1];
      this.changed();
    }
  }

  /**
   * Create an approximation of a circle on the surface of a sphere.
   * @param {import("../coordinate.js").Coordinate} center Center (`[lon, lat]` in degrees).
   * @param {number} radius The great-circle distance from the center to
   *     the polygon vertices in meters.
   * @param {number} [n] Optional number of vertices for the resulting
   *     polygon. Default is `32`.
   * @param {number} [sphereRadius] Optional radius for the sphere (defaults to
   *     the Earth's mean radius using the WGS84 ellipsoid).
   * @return {Polygon} The "circular" polygon.
   * @api
   */
  function circular(center, radius, n, sphereRadius) {
    n = n ? n : 32;
    /** @type {Array<number>} */
    const flatCoordinates = [];
    for (let i = 0; i < n; ++i) {
      extend$2(
        flatCoordinates,
        offset(center, radius, (2 * Math.PI * i) / n, sphereRadius)
      );
    }
    flatCoordinates.push(flatCoordinates[0], flatCoordinates[1]);
    return new Polygon(flatCoordinates, 'XY', [flatCoordinates.length]);
  }

  /**
   * Create a polygon from an extent. The layout used is `XY`.
   * @param {import("../extent.js").Extent} extent The extent.
   * @return {Polygon} The polygon.
   * @api
   */
  function fromExtent(extent) {
    const minX = extent[0];
    const minY = extent[1];
    const maxX = extent[2];
    const maxY = extent[3];
    const flatCoordinates = [
      minX,
      minY,
      minX,
      maxY,
      maxX,
      maxY,
      maxX,
      minY,
      minX,
      minY,
    ];
    return new Polygon(flatCoordinates, 'XY', [flatCoordinates.length]);
  }

  /**
   * Create a regular polygon from a circle.
   * @param {import("./Circle.js").default} circle Circle geometry.
   * @param {number} [sides] Number of sides of the polygon. Default is 32.
   * @param {number} [angle] Start angle for the first vertex of the polygon in
   *     counter-clockwise radians. 0 means East. Default is 0.
   * @return {Polygon} Polygon geometry.
   * @api
   */
  function fromCircle(circle, sides, angle) {
    sides = sides ? sides : 32;
    const stride = circle.getStride();
    const layout = circle.getLayout();
    const center = circle.getCenter();
    const arrayLength = stride * (sides + 1);
    const flatCoordinates = new Array(arrayLength);
    for (let i = 0; i < arrayLength; i += stride) {
      flatCoordinates[i] = 0;
      flatCoordinates[i + 1] = 0;
      for (let j = 2; j < stride; j++) {
        flatCoordinates[i + j] = center[j];
      }
    }
    const ends = [flatCoordinates.length];
    const polygon = new Polygon(flatCoordinates, layout, ends);
    makeRegular(polygon, center, circle.getRadius(), angle);
    return polygon;
  }

  /**
   * Modify the coordinates of a polygon to make it a regular polygon.
   * @param {Polygon} polygon Polygon geometry.
   * @param {import("../coordinate.js").Coordinate} center Center of the regular polygon.
   * @param {number} radius Radius of the regular polygon.
   * @param {number} [angle] Start angle for the first vertex of the polygon in
   *     counter-clockwise radians. 0 means East. Default is 0.
   */
  function makeRegular(polygon, center, radius, angle) {
    const flatCoordinates = polygon.getFlatCoordinates();
    const stride = polygon.getStride();
    const sides = flatCoordinates.length / stride - 1;
    const startAngle = angle ? angle : 0;
    for (let i = 0; i <= sides; ++i) {
      const offset = i * stride;
      const angle = startAngle + (modulo(i, sides) * 2 * Math.PI) / sides;
      flatCoordinates[offset] = center[0] + radius * Math.cos(angle);
      flatCoordinates[offset + 1] = center[1] + radius * Math.sin(angle);
    }
    polygon.changed();
  }

  /**
   * @module ol/Geolocation
   */

  /**
   * @enum {string}
   */
  const Property$4 = {
    ACCURACY: 'accuracy',
    ACCURACY_GEOMETRY: 'accuracyGeometry',
    ALTITUDE: 'altitude',
    ALTITUDE_ACCURACY: 'altitudeAccuracy',
    HEADING: 'heading',
    POSITION: 'position',
    PROJECTION: 'projection',
    SPEED: 'speed',
    TRACKING: 'tracking',
    TRACKING_OPTIONS: 'trackingOptions',
  };

  /**
   * @classdesc
   * Events emitted on Geolocation error.
   */
  class GeolocationError extends BaseEvent {
    /**
     * @param {GeolocationPositionError} error error object.
     */
    constructor(error) {
      super(EventType$1.ERROR);

      /**
       * @type {number}
       */
      this.code = error.code;

      /**
       * @type {string}
       */
      this.message = error.message;
    }
  }

  /**
   * @typedef {Object} Options
   * @property {boolean} [tracking=false] Start Tracking right after
   * instantiation.
   * @property {PositionOptions} [trackingOptions] Tracking options.
   * See https://www.w3.org/TR/geolocation-API/#position_options_interface.
   * @property {import("./proj.js").ProjectionLike} [projection] The projection the position
   * is reported in.
   */

  /**
   * @typedef {import("./ObjectEventType").Types|'change:accuracy'|'change:accuracyGeometry'|'change:altitude'|
   *    'change:altitudeAccuracy'|'change:heading'|'change:position'|'change:projection'|'change:speed'|'change:tracking'|
   *    'change:trackingOptions'} GeolocationObjectEventTypes
   */

  /***
   * @template Return
   * @typedef {import("./Observable").OnSignature<import("./Observable").EventTypes, import("./events/Event.js").default, Return> &
   *   import("./Observable").OnSignature<GeolocationObjectEventTypes, import("./Object").ObjectEvent, Return> &
   *   import("./Observable").OnSignature<'error', GeolocationError, Return> &
   *   import("./Observable").CombinedOnSignature<import("./Observable").EventTypes|GeolocationObjectEventTypes|
   *     'error', Return>} GeolocationOnSignature
   */

  /**
   * @classdesc
   * Helper class for providing HTML5 Geolocation capabilities.
   * The [Geolocation API](https://www.w3.org/TR/geolocation-API/)
   * is used to locate a user's position.
   *
   * To get notified of position changes, register a listener for the generic
   * `change` event on your instance of {@link module:ol/Geolocation~Geolocation}.
   *
   * Example:
   *
   *     const geolocation = new Geolocation({
   *       // take the projection to use from the map's view
   *       projection: view.getProjection()
   *     });
   *     // listen to changes in position
   *     geolocation.on('change', function(evt) {
   *       window.console.log(geolocation.getPosition());
   *     });
   *
   * @fires module:ol/events/Event~BaseEvent#event:error
   * @api
   */
  class Geolocation extends BaseObject {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      super();

      /***
       * @type {GeolocationOnSignature<import("./events").EventsKey>}
       */
      this.on;

      /***
       * @type {GeolocationOnSignature<import("./events").EventsKey>}
       */
      this.once;

      /***
       * @type {GeolocationOnSignature<void>}
       */
      this.un;

      options = options || {};

      /**
       * The unprojected (EPSG:4326) device position.
       * @private
       * @type {?import("./coordinate.js").Coordinate}
       */
      this.position_ = null;

      /**
       * @private
       * @type {import("./proj.js").TransformFunction}
       */
      this.transform_ = identityTransform;

      /**
       * @private
       * @type {number|undefined}
       */
      this.watchId_ = undefined;

      this.addChangeListener(Property$4.PROJECTION, this.handleProjectionChanged_);
      this.addChangeListener(Property$4.TRACKING, this.handleTrackingChanged_);

      if (options.projection !== undefined) {
        this.setProjection(options.projection);
      }
      if (options.trackingOptions !== undefined) {
        this.setTrackingOptions(options.trackingOptions);
      }

      this.setTracking(options.tracking !== undefined ? options.tracking : false);
    }

    /**
     * Clean up.
     */
    disposeInternal() {
      this.setTracking(false);
      super.disposeInternal();
    }

    /**
     * @private
     */
    handleProjectionChanged_() {
      const projection = this.getProjection();
      if (projection) {
        this.transform_ = getTransformFromProjections(
          get$2('EPSG:4326'),
          projection
        );
        if (this.position_) {
          this.set(Property$4.POSITION, this.transform_(this.position_));
        }
      }
    }

    /**
     * @private
     */
    handleTrackingChanged_() {
      if ('geolocation' in navigator) {
        const tracking = this.getTracking();
        if (tracking && this.watchId_ === undefined) {
          this.watchId_ = navigator.geolocation.watchPosition(
            this.positionChange_.bind(this),
            this.positionError_.bind(this),
            this.getTrackingOptions()
          );
        } else if (!tracking && this.watchId_ !== undefined) {
          navigator.geolocation.clearWatch(this.watchId_);
          this.watchId_ = undefined;
        }
      }
    }

    /**
     * @private
     * @param {GeolocationPosition} position position event.
     */
    positionChange_(position) {
      const coords = position.coords;
      this.set(Property$4.ACCURACY, coords.accuracy);
      this.set(
        Property$4.ALTITUDE,
        coords.altitude === null ? undefined : coords.altitude
      );
      this.set(
        Property$4.ALTITUDE_ACCURACY,
        coords.altitudeAccuracy === null ? undefined : coords.altitudeAccuracy
      );
      this.set(
        Property$4.HEADING,
        coords.heading === null ? undefined : toRadians(coords.heading)
      );
      if (!this.position_) {
        this.position_ = [coords.longitude, coords.latitude];
      } else {
        this.position_[0] = coords.longitude;
        this.position_[1] = coords.latitude;
      }
      const projectedPosition = this.transform_(this.position_);
      this.set(Property$4.POSITION, projectedPosition);
      this.set(Property$4.SPEED, coords.speed === null ? undefined : coords.speed);
      const geometry = circular(this.position_, coords.accuracy);
      geometry.applyTransform(this.transform_);
      this.set(Property$4.ACCURACY_GEOMETRY, geometry);
      this.changed();
    }

    /**
     * @private
     * @param {GeolocationPositionError} error error object.
     */
    positionError_(error) {
      this.dispatchEvent(new GeolocationError(error));
    }

    /**
     * Get the accuracy of the position in meters.
     * @return {number|undefined} The accuracy of the position measurement in
     *     meters.
     * @observable
     * @api
     */
    getAccuracy() {
      return /** @type {number|undefined} */ (this.get(Property$4.ACCURACY));
    }

    /**
     * Get a geometry of the position accuracy.
     * @return {?import("./geom/Polygon.js").default} A geometry of the position accuracy.
     * @observable
     * @api
     */
    getAccuracyGeometry() {
      return /** @type {?import("./geom/Polygon.js").default} */ (
        this.get(Property$4.ACCURACY_GEOMETRY) || null
      );
    }

    /**
     * Get the altitude associated with the position.
     * @return {number|undefined} The altitude of the position in meters above mean
     *     sea level.
     * @observable
     * @api
     */
    getAltitude() {
      return /** @type {number|undefined} */ (this.get(Property$4.ALTITUDE));
    }

    /**
     * Get the altitude accuracy of the position.
     * @return {number|undefined} The accuracy of the altitude measurement in
     *     meters.
     * @observable
     * @api
     */
    getAltitudeAccuracy() {
      return /** @type {number|undefined} */ (
        this.get(Property$4.ALTITUDE_ACCURACY)
      );
    }

    /**
     * Get the heading as radians clockwise from North.
     * Note: depending on the browser, the heading is only defined if the `enableHighAccuracy`
     * is set to `true` in the tracking options.
     * @return {number|undefined} The heading of the device in radians from north.
     * @observable
     * @api
     */
    getHeading() {
      return /** @type {number|undefined} */ (this.get(Property$4.HEADING));
    }

    /**
     * Get the position of the device.
     * @return {import("./coordinate.js").Coordinate|undefined} The current position of the device reported
     *     in the current projection.
     * @observable
     * @api
     */
    getPosition() {
      return /** @type {import("./coordinate.js").Coordinate|undefined} */ (
        this.get(Property$4.POSITION)
      );
    }

    /**
     * Get the projection associated with the position.
     * @return {import("./proj/Projection.js").default|undefined} The projection the position is
     *     reported in.
     * @observable
     * @api
     */
    getProjection() {
      return /** @type {import("./proj/Projection.js").default|undefined} */ (
        this.get(Property$4.PROJECTION)
      );
    }

    /**
     * Get the speed in meters per second.
     * @return {number|undefined} The instantaneous speed of the device in meters
     *     per second.
     * @observable
     * @api
     */
    getSpeed() {
      return /** @type {number|undefined} */ (this.get(Property$4.SPEED));
    }

    /**
     * Determine if the device location is being tracked.
     * @return {boolean} The device location is being tracked.
     * @observable
     * @api
     */
    getTracking() {
      return /** @type {boolean} */ (this.get(Property$4.TRACKING));
    }

    /**
     * Get the tracking options.
     * See https://www.w3.org/TR/geolocation-API/#position-options.
     * @return {PositionOptions|undefined} PositionOptions as defined by
     *     the [HTML5 Geolocation spec
     *     ](https://www.w3.org/TR/geolocation-API/#position_options_interface).
     * @observable
     * @api
     */
    getTrackingOptions() {
      return /** @type {PositionOptions|undefined} */ (
        this.get(Property$4.TRACKING_OPTIONS)
      );
    }

    /**
     * Set the projection to use for transforming the coordinates.
     * @param {import("./proj.js").ProjectionLike} projection The projection the position is
     *     reported in.
     * @observable
     * @api
     */
    setProjection(projection) {
      this.set(Property$4.PROJECTION, get$2(projection));
    }

    /**
     * Enable or disable tracking.
     * @param {boolean} tracking Enable tracking.
     * @observable
     * @api
     */
    setTracking(tracking) {
      this.set(Property$4.TRACKING, tracking);
    }

    /**
     * Set the tracking options.
     * See http://www.w3.org/TR/geolocation-API/#position-options.
     * @param {PositionOptions} options PositionOptions as defined by the
     *     [HTML5 Geolocation spec
     *     ](http://www.w3.org/TR/geolocation-API/#position_options_interface).
     * @observable
     * @api
     */
    setTrackingOptions(options) {
      this.set(Property$4.TRACKING_OPTIONS, options);
    }
  }

  /**
   * @module ol/ImageBase
   */

  /**
   * @abstract
   */
  class ImageBase extends Target {
    /**
     * @param {import("./extent.js").Extent} extent Extent.
     * @param {number|undefined} resolution Resolution.
     * @param {number} pixelRatio Pixel ratio.
     * @param {import("./ImageState.js").default} state State.
     */
    constructor(extent, resolution, pixelRatio, state) {
      super();

      /**
       * @protected
       * @type {import("./extent.js").Extent}
       */
      this.extent = extent;

      /**
       * @private
       * @type {number}
       */
      this.pixelRatio_ = pixelRatio;

      /**
       * @protected
       * @type {number|undefined}
       */
      this.resolution = resolution;

      /**
       * @protected
       * @type {import("./ImageState.js").default}
       */
      this.state = state;
    }

    /**
     * @protected
     */
    changed() {
      this.dispatchEvent(EventType$1.CHANGE);
    }

    /**
     * @return {import("./extent.js").Extent} Extent.
     */
    getExtent() {
      return this.extent;
    }

    /**
     * @abstract
     * @return {HTMLCanvasElement|HTMLImageElement|HTMLVideoElement} Image.
     */
    getImage() {
      return abstract();
    }

    /**
     * @return {number} PixelRatio.
     */
    getPixelRatio() {
      return this.pixelRatio_;
    }

    /**
     * @return {number} Resolution.
     */
    getResolution() {
      return /** @type {number} */ (this.resolution);
    }

    /**
     * @return {import("./ImageState.js").default} State.
     */
    getState() {
      return this.state;
    }

    /**
     * Load not yet loaded URI.
     * @abstract
     */
    load() {
      abstract();
    }
  }

  /**
   * @module ol/ImageState
   */

  /**
   * @enum {number}
   */
  var ImageState = {
    IDLE: 0,
    LOADING: 1,
    LOADED: 2,
    ERROR: 3,
    EMPTY: 4,
  };

  /**
   * @module ol/Image
   */

  /**
   * A function that takes an {@link module:ol/Image~ImageWrapper} for the image and a
   * `{string}` for the src as arguments. It is supposed to make it so the
   * underlying image {@link module:ol/Image~ImageWrapper#getImage} is assigned the
   * content specified by the src. If not specified, the default is
   *
   *     function(image, src) {
   *       image.getImage().src = src;
   *     }
   *
   * Providing a custom `imageLoadFunction` can be useful to load images with
   * post requests or - in general - through XHR requests, where the src of the
   * image element would be set to a data URI when the content is loaded.
   *
   * @typedef {function(ImageWrapper, string): void} LoadFunction
   * @api
   */

  class ImageWrapper extends ImageBase {
    /**
     * @param {import("./extent.js").Extent} extent Extent.
     * @param {number|undefined} resolution Resolution.
     * @param {number} pixelRatio Pixel ratio.
     * @param {string} src Image source URI.
     * @param {?string} crossOrigin Cross origin.
     * @param {LoadFunction} imageLoadFunction Image load function.
     */
    constructor(
      extent,
      resolution,
      pixelRatio,
      src,
      crossOrigin,
      imageLoadFunction
    ) {
      super(extent, resolution, pixelRatio, ImageState.IDLE);

      /**
       * @private
       * @type {string}
       */
      this.src_ = src;

      /**
       * @private
       * @type {HTMLCanvasElement|HTMLImageElement|HTMLVideoElement}
       */
      this.image_ = new Image();
      if (crossOrigin !== null) {
        this.image_.crossOrigin = crossOrigin;
      }

      /**
       * @private
       * @type {?function():void}
       */
      this.unlisten_ = null;

      /**
       * @protected
       * @type {import("./ImageState.js").default}
       */
      this.state = ImageState.IDLE;

      /**
       * @private
       * @type {LoadFunction}
       */
      this.imageLoadFunction_ = imageLoadFunction;
    }

    /**
     * @return {HTMLCanvasElement|HTMLImageElement|HTMLVideoElement} Image.
     * @api
     */
    getImage() {
      return this.image_;
    }

    /**
     * Tracks loading or read errors.
     *
     * @private
     */
    handleImageError_() {
      this.state = ImageState.ERROR;
      this.unlistenImage_();
      this.changed();
    }

    /**
     * Tracks successful image load.
     *
     * @private
     */
    handleImageLoad_() {
      if (this.resolution === undefined) {
        this.resolution = getHeight(this.extent) / this.image_.height;
      }
      this.state = ImageState.LOADED;
      this.unlistenImage_();
      this.changed();
    }

    /**
     * Load the image or retry if loading previously failed.
     * Loading is taken care of by the tile queue, and calling this method is
     * only needed for preloading or for reloading in case of an error.
     * @api
     */
    load() {
      if (this.state == ImageState.IDLE || this.state == ImageState.ERROR) {
        this.state = ImageState.LOADING;
        this.changed();
        this.imageLoadFunction_(this, this.src_);
        this.unlisten_ = listenImage(
          this.image_,
          this.handleImageLoad_.bind(this),
          this.handleImageError_.bind(this)
        );
      }
    }

    /**
     * @param {HTMLCanvasElement|HTMLImageElement|HTMLVideoElement} image Image.
     */
    setImage(image) {
      this.image_ = image;
      this.resolution = getHeight(this.extent) / this.image_.height;
    }

    /**
     * Discards event handlers which listen for load completion or errors.
     *
     * @private
     */
    unlistenImage_() {
      if (this.unlisten_) {
        this.unlisten_();
        this.unlisten_ = null;
      }
    }
  }

  /**
   * @param {HTMLCanvasElement|HTMLImageElement|HTMLVideoElement} image Image element.
   * @param {function():any} loadHandler Load callback function.
   * @param {function():any} errorHandler Error callback function.
   * @return {function():void} Callback to stop listening.
   */
  function listenImage(image, loadHandler, errorHandler) {
    const img = /** @type {HTMLImageElement} */ (image);
    let listening = true;
    let decoding = false;
    let loaded = false;

    const listenerKeys = [
      listenOnce(img, EventType$1.LOAD, function () {
        loaded = true;
        if (!decoding) {
          loadHandler();
        }
      }),
    ];

    if (img.src && IMAGE_DECODE) {
      decoding = true;
      img
        .decode()
        .then(function () {
          if (listening) {
            loadHandler();
          }
        })
        .catch(function (error) {
          if (listening) {
            if (loaded) {
              loadHandler();
            } else {
              errorHandler();
            }
          }
        });
    } else {
      listenerKeys.push(listenOnce(img, EventType$1.ERROR, errorHandler));
    }

    return function unlisten() {
      listening = false;
      listenerKeys.forEach(unlistenByKey);
    };
  }

  /**
   * @module ol/ImageCanvas
   */

  /**
   * A function that is called to trigger asynchronous canvas drawing.  It is
   * called with a "done" callback that should be called when drawing is done.
   * If any error occurs during drawing, the "done" callback should be called with
   * that error.
   *
   * @typedef {function(function(Error=): void): void} Loader
   */

  class ImageCanvas extends ImageBase {
    /**
     * @param {import("./extent.js").Extent} extent Extent.
     * @param {number} resolution Resolution.
     * @param {number} pixelRatio Pixel ratio.
     * @param {HTMLCanvasElement} canvas Canvas.
     * @param {Loader} [loader] Optional loader function to
     *     support asynchronous canvas drawing.
     */
    constructor(extent, resolution, pixelRatio, canvas, loader) {
      const state = loader !== undefined ? ImageState.IDLE : ImageState.LOADED;

      super(extent, resolution, pixelRatio, state);

      /**
       * Optional canvas loader function.
       * @type {?Loader}
       * @private
       */
      this.loader_ = loader !== undefined ? loader : null;

      /**
       * @private
       * @type {HTMLCanvasElement}
       */
      this.canvas_ = canvas;

      /**
       * @private
       * @type {?Error}
       */
      this.error_ = null;
    }

    /**
     * Get any error associated with asynchronous rendering.
     * @return {?Error} Any error that occurred during rendering.
     */
    getError() {
      return this.error_;
    }

    /**
     * Handle async drawing complete.
     * @param {Error} [err] Any error during drawing.
     * @private
     */
    handleLoad_(err) {
      if (err) {
        this.error_ = err;
        this.state = ImageState.ERROR;
      } else {
        this.state = ImageState.LOADED;
      }
      this.changed();
    }

    /**
     * Load not yet loaded URI.
     */
    load() {
      if (this.state == ImageState.IDLE) {
        this.state = ImageState.LOADING;
        this.changed();
        this.loader_(this.handleLoad_.bind(this));
      }
    }

    /**
     * @return {HTMLCanvasElement} Canvas element.
     */
    getImage() {
      return this.canvas_;
    }
  }

  /**
   * @module ol/dom
   */

  //FIXME Move this function to the canvas module
  /**
   * Create an html canvas element and returns its 2d context.
   * @param {number} [width] Canvas width.
   * @param {number} [height] Canvas height.
   * @param {Array<HTMLCanvasElement>} [canvasPool] Canvas pool to take existing canvas from.
   * @param {CanvasRenderingContext2DSettings} [settings] CanvasRenderingContext2DSettings
   * @return {CanvasRenderingContext2D} The context.
   */
  function createCanvasContext2D(width, height, canvasPool, settings) {
    /** @type {HTMLCanvasElement|OffscreenCanvas} */
    let canvas;
    if (canvasPool && canvasPool.length) {
      canvas = canvasPool.shift();
    } else if (WORKER_OFFSCREEN_CANVAS) {
      canvas = new OffscreenCanvas(width || 300, height || 300);
    } else {
      canvas = document.createElement('canvas');
    }
    if (width) {
      canvas.width = width;
    }
    if (height) {
      canvas.height = height;
    }
    //FIXME Allow OffscreenCanvasRenderingContext2D as return type
    return /** @type {CanvasRenderingContext2D} */ (
      canvas.getContext('2d', settings)
    );
  }

  /**
   * Releases canvas memory to avoid exceeding memory limits in Safari.
   * See https://pqina.nl/blog/total-canvas-memory-use-exceeds-the-maximum-limit/
   * @param {CanvasRenderingContext2D} context Context.
   */
  function releaseCanvas$1(context) {
    const canvas = context.canvas;
    canvas.width = 1;
    canvas.height = 1;
    context.clearRect(0, 0, 1, 1);
  }

  /**
   * Get the current computed width for the given element including margin,
   * padding and border.
   * Equivalent to jQuery's `$(el).outerWidth(true)`.
   * @param {!HTMLElement} element Element.
   * @return {number} The width.
   */
  function outerWidth(element) {
    let width = element.offsetWidth;
    const style = getComputedStyle(element);
    width += parseInt(style.marginLeft, 10) + parseInt(style.marginRight, 10);

    return width;
  }

  /**
   * Get the current computed height for the given element including margin,
   * padding and border.
   * Equivalent to jQuery's `$(el).outerHeight(true)`.
   * @param {!HTMLElement} element Element.
   * @return {number} The height.
   */
  function outerHeight(element) {
    let height = element.offsetHeight;
    const style = getComputedStyle(element);
    height += parseInt(style.marginTop, 10) + parseInt(style.marginBottom, 10);

    return height;
  }

  /**
   * @param {Node} newNode Node to replace old node
   * @param {Node} oldNode The node to be replaced
   */
  function replaceNode(newNode, oldNode) {
    const parent = oldNode.parentNode;
    if (parent) {
      parent.replaceChild(newNode, oldNode);
    }
  }

  /**
   * @param {Node} node The node to remove.
   * @return {Node|null} The node that was removed or null.
   */
  function removeNode$1(node) {
    return node && node.parentNode ? node.parentNode.removeChild(node) : null;
  }

  /**
   * @param {Node} node The node to remove the children from.
   */
  function removeChildren(node) {
    while (node.lastChild) {
      node.removeChild(node.lastChild);
    }
  }

  /**
   * Transform the children of a parent node so they match the
   * provided list of children.  This function aims to efficiently
   * remove, add, and reorder child nodes while maintaining a simple
   * implementation (it is not guaranteed to minimize DOM operations).
   * @param {Node} node The parent node whose children need reworking.
   * @param {Array<Node>} children The desired children.
   */
  function replaceChildren(node, children) {
    const oldChildren = node.childNodes;

    for (let i = 0; true; ++i) {
      const oldChild = oldChildren[i];
      const newChild = children[i];

      // check if our work is done
      if (!oldChild && !newChild) {
        break;
      }

      // check if children match
      if (oldChild === newChild) {
        continue;
      }

      // check if a new child needs to be added
      if (!oldChild) {
        node.appendChild(newChild);
        continue;
      }

      // check if an old child needs to be removed
      if (!newChild) {
        node.removeChild(oldChild);
        --i;
        continue;
      }

      // reorder
      node.insertBefore(newChild, oldChild);
    }
  }

  /**
   * @module ol/ImageTile
   */

  class ImageTile extends Tile {
    /**
     * @param {import("./tilecoord.js").TileCoord} tileCoord Tile coordinate.
     * @param {import("./TileState.js").default} state State.
     * @param {string} src Image source URI.
     * @param {?string} crossOrigin Cross origin.
     * @param {import("./Tile.js").LoadFunction} tileLoadFunction Tile load function.
     * @param {import("./Tile.js").Options} [options] Tile options.
     */
    constructor(tileCoord, state, src, crossOrigin, tileLoadFunction, options) {
      super(tileCoord, state, options);

      /**
       * @private
       * @type {?string}
       */
      this.crossOrigin_ = crossOrigin;

      /**
       * Image URI
       *
       * @private
       * @type {string}
       */
      this.src_ = src;

      this.key = src;

      /**
       * @private
       * @type {HTMLImageElement|HTMLCanvasElement}
       */
      this.image_ = new Image();
      if (crossOrigin !== null) {
        this.image_.crossOrigin = crossOrigin;
      }

      /**
       * @private
       * @type {?function():void}
       */
      this.unlisten_ = null;

      /**
       * @private
       * @type {import("./Tile.js").LoadFunction}
       */
      this.tileLoadFunction_ = tileLoadFunction;
    }

    /**
     * Get the HTML image element for this tile (may be a Canvas, Image, or Video).
     * @return {HTMLCanvasElement|HTMLImageElement|HTMLVideoElement} Image.
     * @api
     */
    getImage() {
      return this.image_;
    }

    /**
     * Sets an HTML image element for this tile (may be a Canvas or preloaded Image).
     * @param {HTMLCanvasElement|HTMLImageElement} element Element.
     */
    setImage(element) {
      this.image_ = element;
      this.state = TileState.LOADED;
      this.unlistenImage_();
      this.changed();
    }

    /**
     * Tracks loading or read errors.
     *
     * @private
     */
    handleImageError_() {
      this.state = TileState.ERROR;
      this.unlistenImage_();
      this.image_ = getBlankImage();
      this.changed();
    }

    /**
     * Tracks successful image load.
     *
     * @private
     */
    handleImageLoad_() {
      const image = /** @type {HTMLImageElement} */ (this.image_);
      if (image.naturalWidth && image.naturalHeight) {
        this.state = TileState.LOADED;
      } else {
        this.state = TileState.EMPTY;
      }
      this.unlistenImage_();
      this.changed();
    }

    /**
     * Load the image or retry if loading previously failed.
     * Loading is taken care of by the tile queue, and calling this method is
     * only needed for preloading or for reloading in case of an error.
     *
     * To retry loading tiles on failed requests, use a custom `tileLoadFunction`
     * that checks for error status codes and reloads only when the status code is
     * 408, 429, 500, 502, 503 and 504, and only when not too many retries have been
     * made already:
     *
     * ```js
     * const retryCodes = [408, 429, 500, 502, 503, 504];
     * const retries = {};
     * source.setTileLoadFunction((tile, src) => {
     *   const image = tile.getImage();
     *   fetch(src)
     *     .then((response) => {
     *       if (retryCodes.includes(response.status)) {
     *         retries[src] = (retries[src] || 0) + 1;
     *         if (retries[src] <= 3) {
     *           setTimeout(() => tile.load(), retries[src] * 1000);
     *         }
     *         return Promise.reject();
     *       }
     *       return response.blob();
     *     })
     *     .then((blob) => {
     *       const imageUrl = URL.createObjectURL(blob);
     *       image.src = imageUrl;
     *       setTimeout(() => URL.revokeObjectURL(imageUrl), 5000);
     *     })
     *     .catch(() => tile.setState(3)); // error
     * });
     * ```
     *
     * @api
     */
    load() {
      if (this.state == TileState.ERROR) {
        this.state = TileState.IDLE;
        this.image_ = new Image();
        if (this.crossOrigin_ !== null) {
          this.image_.crossOrigin = this.crossOrigin_;
        }
      }
      if (this.state == TileState.IDLE) {
        this.state = TileState.LOADING;
        this.changed();
        this.tileLoadFunction_(this, this.src_);
        this.unlisten_ = listenImage(
          this.image_,
          this.handleImageLoad_.bind(this),
          this.handleImageError_.bind(this)
        );
      }
    }

    /**
     * Discards event handlers which listen for load completion or errors.
     *
     * @private
     */
    unlistenImage_() {
      if (this.unlisten_) {
        this.unlisten_();
        this.unlisten_ = null;
      }
    }
  }

  /**
   * Get a 1-pixel blank image.
   * @return {HTMLCanvasElement} Blank image.
   */
  function getBlankImage() {
    const ctx = createCanvasContext2D(1, 1);
    ctx.fillStyle = 'rgba(0,0,0,0)';
    ctx.fillRect(0, 0, 1, 1);
    return ctx.canvas;
  }

  /**
   * @module ol/Kinetic
   */

  /**
   * @classdesc
   * Implementation of inertial deceleration for map movement.
   *
   * @api
   */
  class Kinetic {
    /**
     * @param {number} decay Rate of decay (must be negative).
     * @param {number} minVelocity Minimum velocity (pixels/millisecond).
     * @param {number} delay Delay to consider to calculate the kinetic
     *     initial values (milliseconds).
     */
    constructor(decay, minVelocity, delay) {
      /**
       * @private
       * @type {number}
       */
      this.decay_ = decay;

      /**
       * @private
       * @type {number}
       */
      this.minVelocity_ = minVelocity;

      /**
       * @private
       * @type {number}
       */
      this.delay_ = delay;

      /**
       * @private
       * @type {Array<number>}
       */
      this.points_ = [];

      /**
       * @private
       * @type {number}
       */
      this.angle_ = 0;

      /**
       * @private
       * @type {number}
       */
      this.initialVelocity_ = 0;
    }

    /**
     * FIXME empty description for jsdoc
     */
    begin() {
      this.points_.length = 0;
      this.angle_ = 0;
      this.initialVelocity_ = 0;
    }

    /**
     * @param {number} x X.
     * @param {number} y Y.
     */
    update(x, y) {
      this.points_.push(x, y, Date.now());
    }

    /**
     * @return {boolean} Whether we should do kinetic animation.
     */
    end() {
      if (this.points_.length < 6) {
        // at least 2 points are required (i.e. there must be at least 6 elements
        // in the array)
        return false;
      }
      const delay = Date.now() - this.delay_;
      const lastIndex = this.points_.length - 3;
      if (this.points_[lastIndex + 2] < delay) {
        // the last tracked point is too old, which means that the user stopped
        // panning before releasing the map
        return false;
      }

      // get the first point which still falls into the delay time
      let firstIndex = lastIndex - 3;
      while (firstIndex > 0 && this.points_[firstIndex + 2] > delay) {
        firstIndex -= 3;
      }

      const duration = this.points_[lastIndex + 2] - this.points_[firstIndex + 2];
      // we don't want a duration of 0 (divide by zero)
      // we also make sure the user panned for a duration of at least one frame
      // (1/60s) to compute sane displacement values
      if (duration < 1000 / 60) {
        return false;
      }

      const dx = this.points_[lastIndex] - this.points_[firstIndex];
      const dy = this.points_[lastIndex + 1] - this.points_[firstIndex + 1];
      this.angle_ = Math.atan2(dy, dx);
      this.initialVelocity_ = Math.sqrt(dx * dx + dy * dy) / duration;
      return this.initialVelocity_ > this.minVelocity_;
    }

    /**
     * @return {number} Total distance travelled (pixels).
     */
    getDistance() {
      return (this.minVelocity_ - this.initialVelocity_) / this.decay_;
    }

    /**
     * @return {number} Angle of the kinetic panning animation (radians).
     */
    getAngle() {
      return this.angle_;
    }
  }

  /**
   * @module ol/color
   */

  /**
   * A color represented as a short array [red, green, blue, alpha].
   * red, green, and blue should be integers in the range 0..255 inclusive.
   * alpha should be a float in the range 0..1 inclusive. If no alpha value is
   * given then `1` will be used.
   * @typedef {Array<number>} Color
   * @api
   */

  /**
   * This RegExp matches # followed by 3, 4, 6, or 8 hex digits.
   * @const
   * @type {RegExp}
   * @private
   */
  const HEX_COLOR_RE_ = /^#([a-f0-9]{3}|[a-f0-9]{4}(?:[a-f0-9]{2}){0,2})$/i;

  /**
   * Regular expression for matching potential named color style strings.
   * @const
   * @type {RegExp}
   * @private
   */
  const NAMED_COLOR_RE_ = /^([a-z]*)$|^hsla?\(.*\)$/i;

  /**
   * Return the color as an rgba string.
   * @param {Color|string} color Color.
   * @return {string} Rgba string.
   * @api
   */
  function asString(color) {
    if (typeof color === 'string') {
      return color;
    } else {
      return toString$2(color);
    }
  }

  /**
   * Return named color as an rgba string.
   * @param {string} color Named color.
   * @return {string} Rgb string.
   */
  function fromNamed(color) {
    const el = document.createElement('div');
    el.style.color = color;
    if (el.style.color !== '') {
      document.body.appendChild(el);
      const rgb = getComputedStyle(el).color;
      document.body.removeChild(el);
      return rgb;
    } else {
      return '';
    }
  }

  /**
   * @param {string} s String.
   * @return {Color} Color.
   */
  const fromString = (function () {
    // We maintain a small cache of parsed strings.  To provide cheap LRU-like
    // semantics, whenever the cache grows too large we simply delete an
    // arbitrary 25% of the entries.

    /**
     * @const
     * @type {number}
     */
    const MAX_CACHE_SIZE = 1024;

    /**
     * @type {Object<string, Color>}
     */
    const cache = {};

    /**
     * @type {number}
     */
    let cacheSize = 0;

    return (
      /**
       * @param {string} s String.
       * @return {Color} Color.
       */
      function (s) {
        let color;
        if (cache.hasOwnProperty(s)) {
          color = cache[s];
        } else {
          if (cacheSize >= MAX_CACHE_SIZE) {
            let i = 0;
            for (const key in cache) {
              if ((i++ & 3) === 0) {
                delete cache[key];
                --cacheSize;
              }
            }
          }
          color = fromStringInternal_(s);
          cache[s] = color;
          ++cacheSize;
        }
        return color;
      }
    );
  })();

  /**
   * Return the color as an array. This function maintains a cache of calculated
   * arrays which means the result should not be modified.
   * @param {Color|string} color Color.
   * @return {Color} Color.
   * @api
   */
  function asArray(color) {
    if (Array.isArray(color)) {
      return color;
    } else {
      return fromString(color);
    }
  }

  /**
   * @param {string} s String.
   * @private
   * @return {Color} Color.
   */
  function fromStringInternal_(s) {
    let r, g, b, a, color;

    if (NAMED_COLOR_RE_.exec(s)) {
      s = fromNamed(s);
    }

    if (HEX_COLOR_RE_.exec(s)) {
      // hex
      const n = s.length - 1; // number of hex digits
      let d; // number of digits per channel
      if (n <= 4) {
        d = 1;
      } else {
        d = 2;
      }
      const hasAlpha = n === 4 || n === 8;
      r = parseInt(s.substr(1 + 0 * d, d), 16);
      g = parseInt(s.substr(1 + 1 * d, d), 16);
      b = parseInt(s.substr(1 + 2 * d, d), 16);
      if (hasAlpha) {
        a = parseInt(s.substr(1 + 3 * d, d), 16);
      } else {
        a = 255;
      }
      if (d == 1) {
        r = (r << 4) + r;
        g = (g << 4) + g;
        b = (b << 4) + b;
        if (hasAlpha) {
          a = (a << 4) + a;
        }
      }
      color = [r, g, b, a / 255];
    } else if (s.startsWith('rgba(')) {
      // rgba()
      color = s.slice(5, -1).split(',').map(Number);
      normalize(color);
    } else if (s.startsWith('rgb(')) {
      // rgb()
      color = s.slice(4, -1).split(',').map(Number);
      color.push(1);
      normalize(color);
    } else {
      assert(false, 14); // Invalid color
    }
    return color;
  }

  /**
   * TODO this function is only used in the test, we probably shouldn't export it
   * @param {Color} color Color.
   * @return {Color} Clamped color.
   */
  function normalize(color) {
    color[0] = clamp((color[0] + 0.5) | 0, 0, 255);
    color[1] = clamp((color[1] + 0.5) | 0, 0, 255);
    color[2] = clamp((color[2] + 0.5) | 0, 0, 255);
    color[3] = clamp(color[3], 0, 1);
    return color;
  }

  /**
   * @param {Color} color Color.
   * @return {string} String.
   */
  function toString$2(color) {
    let r = color[0];
    if (r != (r | 0)) {
      r = (r + 0.5) | 0;
    }
    let g = color[1];
    if (g != (g | 0)) {
      g = (g + 0.5) | 0;
    }
    let b = color[2];
    if (b != (b | 0)) {
      b = (b + 0.5) | 0;
    }
    const a = color[3] === undefined ? 1 : Math.round(color[3] * 100) / 100;
    return 'rgba(' + r + ',' + g + ',' + b + ',' + a + ')';
  }

  /**
   * @param {string} s String.
   * @return {boolean} Whether the string is actually a valid color
   */
  function isStringColor(s) {
    if (NAMED_COLOR_RE_.test(s)) {
      s = fromNamed(s);
    }
    return HEX_COLOR_RE_.test(s) || s.startsWith('rgba(') || s.startsWith('rgb(');
  }

  /**
   * @module ol/style/IconImageCache
   */

  /**
   * @classdesc
   * Singleton class. Available through {@link module:ol/style/IconImageCache.shared}.
   */
  class IconImageCache {
    constructor() {
      /**
       * @type {!Object<string, import("./IconImage.js").default>}
       * @private
       */
      this.cache_ = {};

      /**
       * @type {number}
       * @private
       */
      this.cacheSize_ = 0;

      /**
       * @type {number}
       * @private
       */
      this.maxCacheSize_ = 32;
    }

    /**
     * FIXME empty description for jsdoc
     */
    clear() {
      this.cache_ = {};
      this.cacheSize_ = 0;
    }

    /**
     * @return {boolean} Can expire cache.
     */
    canExpireCache() {
      return this.cacheSize_ > this.maxCacheSize_;
    }

    /**
     * FIXME empty description for jsdoc
     */
    expire() {
      if (this.canExpireCache()) {
        let i = 0;
        for (const key in this.cache_) {
          const iconImage = this.cache_[key];
          if ((i++ & 3) === 0 && !iconImage.hasListener()) {
            delete this.cache_[key];
            --this.cacheSize_;
          }
        }
      }
    }

    /**
     * @param {string} src Src.
     * @param {?string} crossOrigin Cross origin.
     * @param {import("../color.js").Color} color Color.
     * @return {import("./IconImage.js").default} Icon image.
     */
    get(src, crossOrigin, color) {
      const key = getKey$1(src, crossOrigin, color);
      return key in this.cache_ ? this.cache_[key] : null;
    }

    /**
     * @param {string} src Src.
     * @param {?string} crossOrigin Cross origin.
     * @param {import("../color.js").Color} color Color.
     * @param {import("./IconImage.js").default} iconImage Icon image.
     */
    set(src, crossOrigin, color, iconImage) {
      const key = getKey$1(src, crossOrigin, color);
      this.cache_[key] = iconImage;
      ++this.cacheSize_;
    }

    /**
     * Set the cache size of the icon cache. Default is `32`. Change this value when
     * your map uses more than 32 different icon images and you are not caching icon
     * styles on the application level.
     * @param {number} maxCacheSize Cache max size.
     * @api
     */
    setSize(maxCacheSize) {
      this.maxCacheSize_ = maxCacheSize;
      this.expire();
    }
  }

  /**
   * @param {string} src Src.
   * @param {?string} crossOrigin Cross origin.
   * @param {import("../color.js").Color} color Color.
   * @return {string} Cache key.
   */
  function getKey$1(src, crossOrigin, color) {
    const colorString = color ? asString(color) : 'null';
    return crossOrigin + ':' + src + ':' + colorString;
  }

  /**
   * The {@link module:ol/style/IconImageCache~IconImageCache} for
   * {@link module:ol/style/Icon~Icon} images.
   * @api
   */
  const shared = new IconImageCache();

  /**
   * @module ol/layer/Property
   */

  /**
   * @enum {string}
   */
  var LayerProperty = {
    OPACITY: 'opacity',
    VISIBLE: 'visible',
    EXTENT: 'extent',
    Z_INDEX: 'zIndex',
    MAX_RESOLUTION: 'maxResolution',
    MIN_RESOLUTION: 'minResolution',
    MAX_ZOOM: 'maxZoom',
    MIN_ZOOM: 'minZoom',
    SOURCE: 'source',
    MAP: 'map',
  };

  /**
   * @module ol/layer/Base
   */

  /**
   * A css color, or a function called with a view resolution returning a css color.
   *
   * @typedef {string|function(number):string} BackgroundColor
   * @api
   */

  /**
   * @typedef {import("../ObjectEventType").Types|'change:extent'|'change:maxResolution'|'change:maxZoom'|
   *    'change:minResolution'|'change:minZoom'|'change:opacity'|'change:visible'|'change:zIndex'} BaseLayerObjectEventTypes
   */

  /***
   * @template Return
   * @typedef {import("../Observable").OnSignature<import("../Observable").EventTypes, import("../events/Event.js").default, Return> &
   *   import("../Observable").OnSignature<BaseLayerObjectEventTypes, import("../Object").ObjectEvent, Return> &
   *   import("../Observable").CombinedOnSignature<import("../Observable").EventTypes|BaseLayerObjectEventTypes, Return>} BaseLayerOnSignature
   */

  /**
   * @typedef {Object} Options
   * @property {string} [className='ol-layer'] A CSS class name to set to the layer element.
   * @property {number} [opacity=1] Opacity (0, 1).
   * @property {boolean} [visible=true] Visibility.
   * @property {import("../extent.js").Extent} [extent] The bounding extent for layer rendering.  The layer will not be
   * rendered outside of this extent.
   * @property {number} [zIndex] The z-index for layer rendering.  At rendering time, the layers
   * will be ordered, first by Z-index and then by position. When `undefined`, a `zIndex` of 0 is assumed
   * for layers that are added to the map's `layers` collection, or `Infinity` when the layer's `setMap()`
   * method was used.
   * @property {number} [minResolution] The minimum resolution (inclusive) at which this layer will be
   * visible.
   * @property {number} [maxResolution] The maximum resolution (exclusive) below which this layer will
   * be visible.
   * @property {number} [minZoom] The minimum view zoom level (exclusive) above which this layer will be
   * visible.
   * @property {number} [maxZoom] The maximum view zoom level (inclusive) at which this layer will
   * be visible.
   * @property {BackgroundColor} [background] Background color for the layer. If not specified, no background
   * will be rendered.
   * @property {Object<string, *>} [properties] Arbitrary observable properties. Can be accessed with `#get()` and `#set()`.
   */

  /**
   * @classdesc
   * Abstract base class; normally only used for creating subclasses and not
   * instantiated in apps.
   * Note that with {@link module:ol/layer/Base~BaseLayer} and all its subclasses, any property set in
   * the options is set as a {@link module:ol/Object~BaseObject} property on the layer object, so
   * is observable, and has get/set accessors.
   *
   * @api
   */
  class BaseLayer extends BaseObject {
    /**
     * @param {Options} options Layer options.
     */
    constructor(options) {
      super();

      /***
       * @type {BaseLayerOnSignature<import("../events").EventsKey>}
       */
      this.on;

      /***
       * @type {BaseLayerOnSignature<import("../events").EventsKey>}
       */
      this.once;

      /***
       * @type {BaseLayerOnSignature<void>}
       */
      this.un;

      /**
       * @type {BackgroundColor|false}
       * @private
       */
      this.background_ = options.background;

      /**
       * @type {Object<string, *>}
       */
      const properties = Object.assign({}, options);
      if (typeof options.properties === 'object') {
        delete properties.properties;
        Object.assign(properties, options.properties);
      }

      properties[LayerProperty.OPACITY] =
        options.opacity !== undefined ? options.opacity : 1;
      assert(typeof properties[LayerProperty.OPACITY] === 'number', 64); // Layer opacity must be a number

      properties[LayerProperty.VISIBLE] =
        options.visible !== undefined ? options.visible : true;
      properties[LayerProperty.Z_INDEX] = options.zIndex;
      properties[LayerProperty.MAX_RESOLUTION] =
        options.maxResolution !== undefined ? options.maxResolution : Infinity;
      properties[LayerProperty.MIN_RESOLUTION] =
        options.minResolution !== undefined ? options.minResolution : 0;
      properties[LayerProperty.MIN_ZOOM] =
        options.minZoom !== undefined ? options.minZoom : -Infinity;
      properties[LayerProperty.MAX_ZOOM] =
        options.maxZoom !== undefined ? options.maxZoom : Infinity;

      /**
       * @type {string}
       * @private
       */
      this.className_ =
        properties.className !== undefined ? properties.className : 'ol-layer';
      delete properties.className;

      this.setProperties(properties);

      /**
       * @type {import("./Layer.js").State}
       * @private
       */
      this.state_ = null;
    }

    /**
     * Get the background for this layer.
     * @return {BackgroundColor|false} Layer background.
     */
    getBackground() {
      return this.background_;
    }

    /**
     * @return {string} CSS class name.
     */
    getClassName() {
      return this.className_;
    }

    /**
     * This method is not meant to be called by layers or layer renderers because the state
     * is incorrect if the layer is included in a layer group.
     *
     * @param {boolean} [managed] Layer is managed.
     * @return {import("./Layer.js").State} Layer state.
     */
    getLayerState(managed) {
      /** @type {import("./Layer.js").State} */
      const state =
        this.state_ ||
        /** @type {?} */ ({
          layer: this,
          managed: managed === undefined ? true : managed,
        });
      const zIndex = this.getZIndex();
      state.opacity = clamp(Math.round(this.getOpacity() * 100) / 100, 0, 1);
      state.visible = this.getVisible();
      state.extent = this.getExtent();
      state.zIndex = zIndex === undefined && !state.managed ? Infinity : zIndex;
      state.maxResolution = this.getMaxResolution();
      state.minResolution = Math.max(this.getMinResolution(), 0);
      state.minZoom = this.getMinZoom();
      state.maxZoom = this.getMaxZoom();
      this.state_ = state;

      return state;
    }

    /**
     * @abstract
     * @param {Array<import("./Layer.js").default>} [array] Array of layers (to be
     *     modified in place).
     * @return {Array<import("./Layer.js").default>} Array of layers.
     */
    getLayersArray(array) {
      return abstract();
    }

    /**
     * @abstract
     * @param {Array<import("./Layer.js").State>} [states] Optional list of layer
     *     states (to be modified in place).
     * @return {Array<import("./Layer.js").State>} List of layer states.
     */
    getLayerStatesArray(states) {
      return abstract();
    }

    /**
     * Return the {@link module:ol/extent~Extent extent} of the layer or `undefined` if it
     * will be visible regardless of extent.
     * @return {import("../extent.js").Extent|undefined} The layer extent.
     * @observable
     * @api
     */
    getExtent() {
      return /** @type {import("../extent.js").Extent|undefined} */ (
        this.get(LayerProperty.EXTENT)
      );
    }

    /**
     * Return the maximum resolution of the layer.
     * @return {number} The maximum resolution of the layer.
     * @observable
     * @api
     */
    getMaxResolution() {
      return /** @type {number} */ (this.get(LayerProperty.MAX_RESOLUTION));
    }

    /**
     * Return the minimum resolution of the layer.
     * @return {number} The minimum resolution of the layer.
     * @observable
     * @api
     */
    getMinResolution() {
      return /** @type {number} */ (this.get(LayerProperty.MIN_RESOLUTION));
    }

    /**
     * Return the minimum zoom level of the layer.
     * @return {number} The minimum zoom level of the layer.
     * @observable
     * @api
     */
    getMinZoom() {
      return /** @type {number} */ (this.get(LayerProperty.MIN_ZOOM));
    }

    /**
     * Return the maximum zoom level of the layer.
     * @return {number} The maximum zoom level of the layer.
     * @observable
     * @api
     */
    getMaxZoom() {
      return /** @type {number} */ (this.get(LayerProperty.MAX_ZOOM));
    }

    /**
     * Return the opacity of the layer (between 0 and 1).
     * @return {number} The opacity of the layer.
     * @observable
     * @api
     */
    getOpacity() {
      return /** @type {number} */ (this.get(LayerProperty.OPACITY));
    }

    /**
     * @abstract
     * @return {import("../source/Source.js").State} Source state.
     */
    getSourceState() {
      return abstract();
    }

    /**
     * Return the visibility of the layer (`true` or `false`).
     * @return {boolean} The visibility of the layer.
     * @observable
     * @api
     */
    getVisible() {
      return /** @type {boolean} */ (this.get(LayerProperty.VISIBLE));
    }

    /**
     * Return the Z-index of the layer, which is used to order layers before
     * rendering. The default Z-index is 0.
     * @return {number} The Z-index of the layer.
     * @observable
     * @api
     */
    getZIndex() {
      return /** @type {number} */ (this.get(LayerProperty.Z_INDEX));
    }

    /**
     * Sets the background color.
     * @param {BackgroundColor} [background] Background color.
     */
    setBackground(background) {
      this.background_ = background;
      this.changed();
    }

    /**
     * Set the extent at which the layer is visible.  If `undefined`, the layer
     * will be visible at all extents.
     * @param {import("../extent.js").Extent|undefined} extent The extent of the layer.
     * @observable
     * @api
     */
    setExtent(extent) {
      this.set(LayerProperty.EXTENT, extent);
    }

    /**
     * Set the maximum resolution at which the layer is visible.
     * @param {number} maxResolution The maximum resolution of the layer.
     * @observable
     * @api
     */
    setMaxResolution(maxResolution) {
      this.set(LayerProperty.MAX_RESOLUTION, maxResolution);
    }

    /**
     * Set the minimum resolution at which the layer is visible.
     * @param {number} minResolution The minimum resolution of the layer.
     * @observable
     * @api
     */
    setMinResolution(minResolution) {
      this.set(LayerProperty.MIN_RESOLUTION, minResolution);
    }

    /**
     * Set the maximum zoom (exclusive) at which the layer is visible.
     * Note that the zoom levels for layer visibility are based on the
     * view zoom level, which may be different from a tile source zoom level.
     * @param {number} maxZoom The maximum zoom of the layer.
     * @observable
     * @api
     */
    setMaxZoom(maxZoom) {
      this.set(LayerProperty.MAX_ZOOM, maxZoom);
    }

    /**
     * Set the minimum zoom (inclusive) at which the layer is visible.
     * Note that the zoom levels for layer visibility are based on the
     * view zoom level, which may be different from a tile source zoom level.
     * @param {number} minZoom The minimum zoom of the layer.
     * @observable
     * @api
     */
    setMinZoom(minZoom) {
      this.set(LayerProperty.MIN_ZOOM, minZoom);
    }

    /**
     * Set the opacity of the layer, allowed values range from 0 to 1.
     * @param {number} opacity The opacity of the layer.
     * @observable
     * @api
     */
    setOpacity(opacity) {
      assert(typeof opacity === 'number', 64); // Layer opacity must be a number
      this.set(LayerProperty.OPACITY, opacity);
    }

    /**
     * Set the visibility of the layer (`true` or `false`).
     * @param {boolean} visible The visibility of the layer.
     * @observable
     * @api
     */
    setVisible(visible) {
      this.set(LayerProperty.VISIBLE, visible);
    }

    /**
     * Set Z-index of the layer, which is used to order layers before rendering.
     * The default Z-index is 0.
     * @param {number} zindex The z-index of the layer.
     * @observable
     * @api
     */
    setZIndex(zindex) {
      this.set(LayerProperty.Z_INDEX, zindex);
    }

    /**
     * Clean up.
     */
    disposeInternal() {
      if (this.state_) {
        this.state_.layer = null;
        this.state_ = null;
      }
      super.disposeInternal();
    }
  }

  /**
   * @module ol/render/EventType
   */

  /**
   * @enum {string}
   */
  var EventType = {
    /**
     * Triggered before a layer is rendered.
     * @event module:ol/render/Event~RenderEvent#prerender
     * @api
     */
    PRERENDER: 'prerender',

    /**
     * Triggered after a layer is rendered.
     * @event module:ol/render/Event~RenderEvent#postrender
     * @api
     */
    POSTRENDER: 'postrender',

    /**
     * Triggered before layers are composed.  When dispatched by the map, the event object will not have
     * a `context` set.  When dispatched by a layer, the event object will have a `context` set.  Only
     * WebGL layers currently dispatch this event.
     * @event module:ol/render/Event~RenderEvent#precompose
     * @api
     */
    PRECOMPOSE: 'precompose',

    /**
     * Triggered after layers are composed.  When dispatched by the map, the event object will not have
     * a `context` set.  When dispatched by a layer, the event object will have a `context` set.  Only
     * WebGL layers currently dispatch this event.
     * @event module:ol/render/Event~RenderEvent#postcompose
     * @api
     */
    POSTCOMPOSE: 'postcompose',

    /**
     * Triggered when rendering is complete, i.e. all sources and tiles have
     * finished loading for the current viewport, and all tiles are faded in.
     * The event object will not have a `context` set.
     * @event module:ol/render/Event~RenderEvent#rendercomplete
     * @api
     */
    RENDERCOMPLETE: 'rendercomplete',
  };

  /**
   * @typedef {'postrender'|'precompose'|'postcompose'|'rendercomplete'} MapRenderEventTypes
   */

  /**
   * @typedef {'postrender'|'prerender'} LayerRenderEventTypes
   */

  /**
   * @module ol/layer/Layer
   */

  /**
   * @typedef {function(import("../Map.js").FrameState):HTMLElement} RenderFunction
   */

  /***
   * @template Return
   * @typedef {import("../Observable").OnSignature<import("../Observable").EventTypes, import("../events/Event.js").default, Return> &
   *   import("../Observable").OnSignature<import("./Base").BaseLayerObjectEventTypes|
   *     'change:source', import("../Object").ObjectEvent, Return> &
   *   import("../Observable").OnSignature<import("../render/EventType").LayerRenderEventTypes, import("../render/Event").default, Return> &
   *   import("../Observable").CombinedOnSignature<import("../Observable").EventTypes|import("./Base").BaseLayerObjectEventTypes|'change:source'|
   *     import("../render/EventType").LayerRenderEventTypes, Return>} LayerOnSignature
   */

  /**
   * @template {import("../source/Source.js").default} [SourceType=import("../source/Source.js").default]
   * @typedef {Object} Options
   * @property {string} [className='ol-layer'] A CSS class name to set to the layer element.
   * @property {number} [opacity=1] Opacity (0, 1).
   * @property {boolean} [visible=true] Visibility.
   * @property {import("../extent.js").Extent} [extent] The bounding extent for layer rendering.  The layer will not be
   * rendered outside of this extent.
   * @property {number} [zIndex] The z-index for layer rendering.  At rendering time, the layers
   * will be ordered, first by Z-index and then by position. When `undefined`, a `zIndex` of 0 is assumed
   * for layers that are added to the map's `layers` collection, or `Infinity` when the layer's `setMap()`
   * method was used.
   * @property {number} [minResolution] The minimum resolution (inclusive) at which this layer will be
   * visible.
   * @property {number} [maxResolution] The maximum resolution (exclusive) below which this layer will
   * be visible.
   * @property {number} [minZoom] The minimum view zoom level (exclusive) above which this layer will be
   * visible.
   * @property {number} [maxZoom] The maximum view zoom level (inclusive) at which this layer will
   * be visible.
   * @property {SourceType} [source] Source for this layer.  If not provided to the constructor,
   * the source can be set by calling {@link module:ol/layer/Layer~Layer#setSource layer.setSource(source)} after
   * construction.
   * @property {import("../Map.js").default|null} [map] Map.
   * @property {RenderFunction} [render] Render function. Takes the frame state as input and is expected to return an
   * HTML element. Will overwrite the default rendering for the layer.
   * @property {Object<string, *>} [properties] Arbitrary observable properties. Can be accessed with `#get()` and `#set()`.
   */

  /**
   * @typedef {Object} State
   * @property {import("./Layer.js").default} layer Layer.
   * @property {number} opacity Opacity, the value is rounded to two digits to appear after the decimal point.
   * @property {boolean} visible Visible.
   * @property {boolean} managed Managed.
   * @property {import("../extent.js").Extent} [extent] Extent.
   * @property {number} zIndex ZIndex.
   * @property {number} maxResolution Maximum resolution.
   * @property {number} minResolution Minimum resolution.
   * @property {number} minZoom Minimum zoom.
   * @property {number} maxZoom Maximum zoom.
   */

  /**
   * @classdesc
   * Base class from which all layer types are derived. This should only be instantiated
   * in the case where a custom layer is added to the map with a custom `render` function.
   * Such a function can be specified in the `options` object, and is expected to return an HTML element.
   *
   * A visual representation of raster or vector map data.
   * Layers group together those properties that pertain to how the data is to be
   * displayed, irrespective of the source of that data.
   *
   * Layers are usually added to a map with [map.addLayer()]{@link import("../Map.js").default#addLayer}.
   * Components like {@link module:ol/interaction/Draw~Draw} use unmanaged layers
   * internally. These unmanaged layers are associated with the map using
   * [layer.setMap()]{@link module:ol/layer/Layer~Layer#setMap} instead.
   *
   * A generic `change` event is fired when the state of the source changes.
   *
   * @fires import("../render/Event.js").RenderEvent#prerender
   * @fires import("../render/Event.js").RenderEvent#postrender
   *
   * @template {import("../source/Source.js").default} [SourceType=import("../source/Source.js").default]
   * @template {import("../renderer/Layer.js").default} [RendererType=import("../renderer/Layer.js").default]
   * @api
   */
  class Layer extends BaseLayer {
    /**
     * @param {Options<SourceType>} options Layer options.
     */
    constructor(options) {
      const baseOptions = Object.assign({}, options);
      delete baseOptions.source;

      super(baseOptions);

      /***
       * @type {LayerOnSignature<import("../events").EventsKey>}
       */
      this.on;

      /***
       * @type {LayerOnSignature<import("../events").EventsKey>}
       */
      this.once;

      /***
       * @type {LayerOnSignature<void>}
       */
      this.un;

      /**
       * @private
       * @type {?import("../events.js").EventsKey}
       */
      this.mapPrecomposeKey_ = null;

      /**
       * @private
       * @type {?import("../events.js").EventsKey}
       */
      this.mapRenderKey_ = null;

      /**
       * @private
       * @type {?import("../events.js").EventsKey}
       */
      this.sourceChangeKey_ = null;

      /**
       * @private
       * @type {RendererType}
       */
      this.renderer_ = null;

      /**
       * @protected
       * @type {boolean}
       */
      this.rendered = false;

      // Overwrite default render method with a custom one
      if (options.render) {
        this.render = options.render;
      }

      if (options.map) {
        this.setMap(options.map);
      }

      this.addChangeListener(
        LayerProperty.SOURCE,
        this.handleSourcePropertyChange_
      );

      const source = options.source
        ? /** @type {SourceType} */ (options.source)
        : null;
      this.setSource(source);
    }

    /**
     * @param {Array<import("./Layer.js").default>} [array] Array of layers (to be modified in place).
     * @return {Array<import("./Layer.js").default>} Array of layers.
     */
    getLayersArray(array) {
      array = array ? array : [];
      array.push(this);
      return array;
    }

    /**
     * @param {Array<import("./Layer.js").State>} [states] Optional list of layer states (to be modified in place).
     * @return {Array<import("./Layer.js").State>} List of layer states.
     */
    getLayerStatesArray(states) {
      states = states ? states : [];
      states.push(this.getLayerState());
      return states;
    }

    /**
     * Get the layer source.
     * @return {SourceType|null} The layer source (or `null` if not yet set).
     * @observable
     * @api
     */
    getSource() {
      return /** @type {SourceType} */ (this.get(LayerProperty.SOURCE)) || null;
    }

    /**
     * @return {SourceType|null} The source being rendered.
     */
    getRenderSource() {
      return this.getSource();
    }

    /**
     * @return {import("../source/Source.js").State} Source state.
     */
    getSourceState() {
      const source = this.getSource();
      return !source ? 'undefined' : source.getState();
    }

    /**
     * @private
     */
    handleSourceChange_() {
      this.changed();
    }

    /**
     * @private
     */
    handleSourcePropertyChange_() {
      if (this.sourceChangeKey_) {
        unlistenByKey(this.sourceChangeKey_);
        this.sourceChangeKey_ = null;
      }
      const source = this.getSource();
      if (source) {
        this.sourceChangeKey_ = listen(
          source,
          EventType$1.CHANGE,
          this.handleSourceChange_,
          this
        );
      }
      this.changed();
    }

    /**
     * @param {import("../pixel").Pixel} pixel Pixel.
     * @return {Promise<Array<import("../Feature").default>>} Promise that resolves with
     * an array of features.
     */
    getFeatures(pixel) {
      if (!this.renderer_) {
        return new Promise((resolve) => resolve([]));
      }
      return this.renderer_.getFeatures(pixel);
    }

    /**
     * @param {import("../pixel").Pixel} pixel Pixel.
     * @return {Uint8ClampedArray|Uint8Array|Float32Array|DataView|null} Pixel data.
     */
    getData(pixel) {
      if (!this.renderer_ || !this.rendered) {
        return null;
      }
      return this.renderer_.getData(pixel);
    }

    /**
     * In charge to manage the rendering of the layer. One layer type is
     * bounded with one layer renderer.
     * @param {?import("../Map.js").FrameState} frameState Frame state.
     * @param {HTMLElement} target Target which the renderer may (but need not) use
     * for rendering its content.
     * @return {HTMLElement} The rendered element.
     */
    render(frameState, target) {
      const layerRenderer = this.getRenderer();

      if (layerRenderer.prepareFrame(frameState)) {
        this.rendered = true;
        return layerRenderer.renderFrame(frameState, target);
      }
    }

    /**
     * Called when a layer is not visible during a map render.
     */
    unrender() {
      this.rendered = false;
    }

    /**
     * For use inside the library only.
     * @param {import("../Map.js").default|null} map Map.
     */
    setMapInternal(map) {
      if (!map) {
        this.unrender();
      }
      this.set(LayerProperty.MAP, map);
    }

    /**
     * For use inside the library only.
     * @return {import("../Map.js").default|null} Map.
     */
    getMapInternal() {
      return this.get(LayerProperty.MAP);
    }

    /**
     * Sets the layer to be rendered on top of other layers on a map. The map will
     * not manage this layer in its layers collection. This
     * is useful for temporary layers. To remove an unmanaged layer from the map,
     * use `#setMap(null)`.
     *
     * To add the layer to a map and have it managed by the map, use
     * {@link module:ol/Map~Map#addLayer} instead.
     * @param {import("../Map.js").default|null} map Map.
     * @api
     */
    setMap(map) {
      if (this.mapPrecomposeKey_) {
        unlistenByKey(this.mapPrecomposeKey_);
        this.mapPrecomposeKey_ = null;
      }
      if (!map) {
        this.changed();
      }
      if (this.mapRenderKey_) {
        unlistenByKey(this.mapRenderKey_);
        this.mapRenderKey_ = null;
      }
      if (map) {
        this.mapPrecomposeKey_ = listen(
          map,
          EventType.PRECOMPOSE,
          function (evt) {
            const renderEvent =
              /** @type {import("../render/Event.js").default} */ (evt);
            const layerStatesArray = renderEvent.frameState.layerStatesArray;
            const layerState = this.getLayerState(false);
            // A layer can only be added to the map once. Use either `layer.setMap()` or `map.addLayer()`, not both.
            assert(
              !layerStatesArray.some(function (arrayLayerState) {
                return arrayLayerState.layer === layerState.layer;
              }),
              67
            );
            layerStatesArray.push(layerState);
          },
          this
        );
        this.mapRenderKey_ = listen(this, EventType$1.CHANGE, map.render, map);
        this.changed();
      }
    }

    /**
     * Set the layer source.
     * @param {SourceType|null} source The layer source.
     * @observable
     * @api
     */
    setSource(source) {
      this.set(LayerProperty.SOURCE, source);
    }

    /**
     * Get the renderer for this layer.
     * @return {RendererType|null} The layer renderer.
     */
    getRenderer() {
      if (!this.renderer_) {
        this.renderer_ = this.createRenderer();
      }
      return this.renderer_;
    }

    /**
     * @return {boolean} The layer has a renderer.
     */
    hasRenderer() {
      return !!this.renderer_;
    }

    /**
     * Create a renderer for this layer.
     * @return {RendererType} A layer renderer.
     * @protected
     */
    createRenderer() {
      return null;
    }

    /**
     * Clean up.
     */
    disposeInternal() {
      if (this.renderer_) {
        this.renderer_.dispose();
        delete this.renderer_;
      }

      this.setSource(null);
      super.disposeInternal();
    }
  }

  /**
   * Return `true` if the layer is visible and if the provided view state
   * has resolution and zoom levels that are in range of the layer's min/max.
   * @param {State} layerState Layer state.
   * @param {import("../View.js").State} viewState View state.
   * @return {boolean} The layer is visible at the given view state.
   */
  function inView(layerState, viewState) {
    if (!layerState.visible) {
      return false;
    }
    const resolution = viewState.resolution;
    if (
      resolution < layerState.minResolution ||
      resolution >= layerState.maxResolution
    ) {
      return false;
    }
    const zoom = viewState.zoom;
    return zoom > layerState.minZoom && zoom <= layerState.maxZoom;
  }

  /**
   * @module ol/renderer/Map
   */

  /**
   * @typedef HitMatch
   * @property {import("../Feature.js").FeatureLike} feature Feature.
   * @property {import("../layer/Layer.js").default} layer Layer.
   * @property {import("../geom/SimpleGeometry.js").default} geometry Geometry.
   * @property {number} distanceSq Squared distance.
   * @property {import("./vector.js").FeatureCallback<T>} callback Callback.
   * @template T
   */

  /**
   * @abstract
   */
  class MapRenderer extends Disposable {
    /**
     * @param {import("../Map.js").default} map Map.
     */
    constructor(map) {
      super();

      /**
       * @private
       * @type {import("../Map.js").default}
       */
      this.map_ = map;
    }

    /**
     * @abstract
     * @param {import("../render/EventType.js").default} type Event type.
     * @param {import("../Map.js").FrameState} frameState Frame state.
     */
    dispatchRenderEvent(type, frameState) {
      abstract();
    }

    /**
     * @param {import("../Map.js").FrameState} frameState FrameState.
     * @protected
     */
    calculateMatrices2D(frameState) {
      const viewState = frameState.viewState;
      const coordinateToPixelTransform = frameState.coordinateToPixelTransform;
      const pixelToCoordinateTransform = frameState.pixelToCoordinateTransform;

      compose(
        coordinateToPixelTransform,
        frameState.size[0] / 2,
        frameState.size[1] / 2,
        1 / viewState.resolution,
        -1 / viewState.resolution,
        -viewState.rotation,
        -viewState.center[0],
        -viewState.center[1]
      );

      makeInverse(pixelToCoordinateTransform, coordinateToPixelTransform);
    }

    /**
     * @param {import("../coordinate.js").Coordinate} coordinate Coordinate.
     * @param {import("../Map.js").FrameState} frameState FrameState.
     * @param {number} hitTolerance Hit tolerance in pixels.
     * @param {boolean} checkWrapped Check for wrapped geometries.
     * @param {import("./vector.js").FeatureCallback<T>} callback Feature callback.
     * @param {S} thisArg Value to use as `this` when executing `callback`.
     * @param {function(this: U, import("../layer/Layer.js").default): boolean} layerFilter Layer filter
     *     function, only layers which are visible and for which this function
     *     returns `true` will be tested for features.  By default, all visible
     *     layers will be tested.
     * @param {U} thisArg2 Value to use as `this` when executing `layerFilter`.
     * @return {T|undefined} Callback result.
     * @template S,T,U
     */
    forEachFeatureAtCoordinate(
      coordinate,
      frameState,
      hitTolerance,
      checkWrapped,
      callback,
      thisArg,
      layerFilter,
      thisArg2
    ) {
      let result;
      const viewState = frameState.viewState;

      /**
       * @param {boolean} managed Managed layer.
       * @param {import("../Feature.js").FeatureLike} feature Feature.
       * @param {import("../layer/Layer.js").default} layer Layer.
       * @param {import("../geom/Geometry.js").default} geometry Geometry.
       * @return {T|undefined} Callback result.
       */
      function forEachFeatureAtCoordinate(managed, feature, layer, geometry) {
        return callback.call(thisArg, feature, managed ? layer : null, geometry);
      }

      const projection = viewState.projection;

      const translatedCoordinate = wrapX$1(coordinate.slice(), projection);
      const offsets = [[0, 0]];
      if (projection.canWrapX() && checkWrapped) {
        const projectionExtent = projection.getExtent();
        const worldWidth = getWidth(projectionExtent);
        offsets.push([-worldWidth, 0], [worldWidth, 0]);
      }

      const layerStates = frameState.layerStatesArray;
      const numLayers = layerStates.length;

      const matches = /** @type {Array<HitMatch<T>>} */ ([]);
      const tmpCoord = [];
      for (let i = 0; i < offsets.length; i++) {
        for (let j = numLayers - 1; j >= 0; --j) {
          const layerState = layerStates[j];
          const layer = layerState.layer;
          if (
            layer.hasRenderer() &&
            inView(layerState, viewState) &&
            layerFilter.call(thisArg2, layer)
          ) {
            const layerRenderer = layer.getRenderer();
            const source = layer.getSource();
            if (layerRenderer && source) {
              const coordinates = source.getWrapX()
                ? translatedCoordinate
                : coordinate;
              const callback = forEachFeatureAtCoordinate.bind(
                null,
                layerState.managed
              );
              tmpCoord[0] = coordinates[0] + offsets[i][0];
              tmpCoord[1] = coordinates[1] + offsets[i][1];
              result = layerRenderer.forEachFeatureAtCoordinate(
                tmpCoord,
                frameState,
                hitTolerance,
                callback,
                matches
              );
            }
            if (result) {
              return result;
            }
          }
        }
      }
      if (matches.length === 0) {
        return undefined;
      }
      const order = 1 / matches.length;
      matches.forEach((m, i) => (m.distanceSq += i * order));
      matches.sort((a, b) => a.distanceSq - b.distanceSq);
      matches.some((m) => {
        return (result = m.callback(m.feature, m.layer, m.geometry));
      });
      return result;
    }

    /**
     * @param {import("../coordinate.js").Coordinate} coordinate Coordinate.
     * @param {import("../Map.js").FrameState} frameState FrameState.
     * @param {number} hitTolerance Hit tolerance in pixels.
     * @param {boolean} checkWrapped Check for wrapped geometries.
     * @param {function(this: U, import("../layer/Layer.js").default): boolean} layerFilter Layer filter
     *     function, only layers which are visible and for which this function
     *     returns `true` will be tested for features.  By default, all visible
     *     layers will be tested.
     * @param {U} thisArg Value to use as `this` when executing `layerFilter`.
     * @return {boolean} Is there a feature at the given coordinate?
     * @template U
     */
    hasFeatureAtCoordinate(
      coordinate,
      frameState,
      hitTolerance,
      checkWrapped,
      layerFilter,
      thisArg
    ) {
      const hasFeature = this.forEachFeatureAtCoordinate(
        coordinate,
        frameState,
        hitTolerance,
        checkWrapped,
        TRUE,
        this,
        layerFilter,
        thisArg
      );

      return hasFeature !== undefined;
    }

    /**
     * @return {import("../Map.js").default} Map.
     */
    getMap() {
      return this.map_;
    }

    /**
     * Render.
     * @abstract
     * @param {?import("../Map.js").FrameState} frameState Frame state.
     */
    renderFrame(frameState) {
      abstract();
    }

    /**
     * @param {import("../Map.js").FrameState} frameState Frame state.
     * @protected
     */
    scheduleExpireIconCache(frameState) {
      if (shared.canExpireCache()) {
        frameState.postRenderFunctions.push(expireIconCache);
      }
    }
  }

  /**
   * @param {import("../Map.js").default} map Map.
   * @param {import("../Map.js").FrameState} frameState Frame state.
   */
  function expireIconCache(map, frameState) {
    shared.expire();
  }

  /**
   * @module ol/render/Event
   */

  class RenderEvent extends BaseEvent {
    /**
     * @param {import("./EventType.js").default} type Type.
     * @param {import("../transform.js").Transform} [inversePixelTransform] Transform for
     *     CSS pixels to rendered pixels.
     * @param {import("../Map.js").FrameState} [frameState] Frame state.
     * @param {?(CanvasRenderingContext2D|WebGLRenderingContext)} [context] Context.
     */
    constructor(type, inversePixelTransform, frameState, context) {
      super(type);

      /**
       * Transform from CSS pixels (relative to the top-left corner of the map viewport)
       * to rendered pixels on this event's `context`. Only available when a Canvas renderer is used, null otherwise.
       * @type {import("../transform.js").Transform|undefined}
       * @api
       */
      this.inversePixelTransform = inversePixelTransform;

      /**
       * An object representing the current render frame state.
       * @type {import("../Map.js").FrameState|undefined}
       * @api
       */
      this.frameState = frameState;

      /**
       * Canvas context. Not available when the event is dispatched by the map. For Canvas 2D layers,
       * the context will be the 2D rendering context.  For WebGL layers, the context will be the WebGL
       * context.
       * @type {CanvasRenderingContext2D|WebGLRenderingContext|undefined}
       * @api
       */
      this.context = context;
    }
  }

  /**
   * @module ol/css
   */

  /**
   * @typedef {Object} FontParameters
   * @property {string} style Style.
   * @property {string} variant Variant.
   * @property {string} weight Weight.
   * @property {string} size Size.
   * @property {string} lineHeight LineHeight.
   * @property {string} family Family.
   * @property {Array<string>} families Families.
   */

  /**
   * The CSS class for hidden feature.
   *
   * @const
   * @type {string}
   */
  const CLASS_HIDDEN = 'ol-hidden';

  /**
   * The CSS class that we'll give the DOM elements to have them selectable.
   *
   * @const
   * @type {string}
   */
  const CLASS_SELECTABLE = 'ol-selectable';

  /**
   * The CSS class that we'll give the DOM elements to have them unselectable.
   *
   * @const
   * @type {string}
   */
  const CLASS_UNSELECTABLE = 'ol-unselectable';

  /**
   * The CSS class for unsupported feature.
   *
   * @const
   * @type {string}
   */
  const CLASS_UNSUPPORTED = 'ol-unsupported';

  /**
   * The CSS class for controls.
   *
   * @const
   * @type {string}
   */
  const CLASS_CONTROL = 'ol-control';

  /**
   * The CSS class that we'll give the DOM elements that are collapsed, i.e.
   * to those elements which usually can be expanded.
   *
   * @const
   * @type {string}
   */
  const CLASS_COLLAPSED = 'ol-collapsed';

  /**
   * From https://stackoverflow.com/questions/10135697/regex-to-parse-any-css-font
   * @type {RegExp}
   */
  const fontRegEx = new RegExp(
    [
      '^\\s*(?=(?:(?:[-a-z]+\\s*){0,2}(italic|oblique))?)',
      '(?=(?:(?:[-a-z]+\\s*){0,2}(small-caps))?)',
      '(?=(?:(?:[-a-z]+\\s*){0,2}(bold(?:er)?|lighter|[1-9]00 ))?)',
      '(?:(?:normal|\\1|\\2|\\3)\\s*){0,3}((?:xx?-)?',
      '(?:small|large)|medium|smaller|larger|[\\.\\d]+(?:\\%|in|[cem]m|ex|p[ctx]))',
      '(?:\\s*\\/\\s*(normal|[\\.\\d]+(?:\\%|in|[cem]m|ex|p[ctx])?))',
      '?\\s*([-,\\"\\\'\\sa-z]+?)\\s*$',
    ].join(''),
    'i'
  );
  const fontRegExMatchIndex = [
    'style',
    'variant',
    'weight',
    'size',
    'lineHeight',
    'family',
  ];

  /**
   * Get the list of font families from a font spec.  Note that this doesn't work
   * for font families that have commas in them.
   * @param {string} fontSpec The CSS font property.
   * @return {FontParameters|null} The font parameters (or null if the input spec is invalid).
   */
  const getFontParameters = function (fontSpec) {
    const match = fontSpec.match(fontRegEx);
    if (!match) {
      return null;
    }
    const style = /** @type {FontParameters} */ ({
      lineHeight: 'normal',
      size: '1.2em',
      style: 'normal',
      weight: 'normal',
      variant: 'normal',
    });
    for (let i = 0, ii = fontRegExMatchIndex.length; i < ii; ++i) {
      const value = match[i + 1];
      if (value !== undefined) {
        style[fontRegExMatchIndex[i]] = value;
      }
    }
    style.families = style.family.split(/,\s?/);
    return style;
  };

  /**
   * @module ol/render/canvas
   */

  /**
   * @typedef {'Circle' | 'Image' | 'LineString' | 'Polygon' | 'Text' | 'Default'} BuilderType
   */

  /**
   * @typedef {Object} FillState
   * @property {import("../colorlike.js").ColorLike} fillStyle FillStyle.
   */

  /**
   * @typedef Label
   * @property {number} width Width.
   * @property {number} height Height.
   * @property {Array<string|number>} contextInstructions ContextInstructions.
   */

  /**
   * @typedef {Object} FillStrokeState
   * @property {import("../colorlike.js").ColorLike} [currentFillStyle] Current FillStyle.
   * @property {import("../colorlike.js").ColorLike} [currentStrokeStyle] Current StrokeStyle.
   * @property {CanvasLineCap} [currentLineCap] Current LineCap.
   * @property {Array<number>} currentLineDash Current LineDash.
   * @property {number} [currentLineDashOffset] Current LineDashOffset.
   * @property {CanvasLineJoin} [currentLineJoin] Current LineJoin.
   * @property {number} [currentLineWidth] Current LineWidth.
   * @property {number} [currentMiterLimit] Current MiterLimit.
   * @property {number} [lastStroke] Last stroke.
   * @property {import("../colorlike.js").ColorLike} [fillStyle] FillStyle.
   * @property {import("../colorlike.js").ColorLike} [strokeStyle] StrokeStyle.
   * @property {CanvasLineCap} [lineCap] LineCap.
   * @property {Array<number>} lineDash LineDash.
   * @property {number} [lineDashOffset] LineDashOffset.
   * @property {CanvasLineJoin} [lineJoin] LineJoin.
   * @property {number} [lineWidth] LineWidth.
   * @property {number} [miterLimit] MiterLimit.
   */

  /**
   * @typedef {Object} StrokeState
   * @property {CanvasLineCap} lineCap LineCap.
   * @property {Array<number>} lineDash LineDash.
   * @property {number} lineDashOffset LineDashOffset.
   * @property {CanvasLineJoin} lineJoin LineJoin.
   * @property {number} lineWidth LineWidth.
   * @property {number} miterLimit MiterLimit.
   * @property {import("../colorlike.js").ColorLike} strokeStyle StrokeStyle.
   */

  /**
   * @typedef {Object} TextState
   * @property {string} font Font.
   * @property {CanvasTextAlign} [textAlign] TextAlign.
   * @property {import("../style/Text.js").TextJustify} [justify] Justify.
   * @property {CanvasTextBaseline} textBaseline TextBaseline.
   * @property {import("../style/Text.js").TextPlacement} [placement] Placement.
   * @property {number} [maxAngle] MaxAngle.
   * @property {boolean} [overflow] Overflow.
   * @property {import("../style/Fill.js").default} [backgroundFill] BackgroundFill.
   * @property {import("../style/Stroke.js").default} [backgroundStroke] BackgroundStroke.
   * @property {import("../size.js").Size} [scale] Scale.
   * @property {Array<number>} [padding] Padding.
   */

  /**
   * @typedef {Object} SerializableInstructions
   * @property {Array<*>} instructions The rendering instructions.
   * @property {Array<*>} hitDetectionInstructions The rendering hit detection instructions.
   * @property {Array<number>} coordinates The array of all coordinates.
   * @property {!Object<string, TextState>} [textStates] The text states (decluttering).
   * @property {!Object<string, FillState>} [fillStates] The fill states (decluttering).
   * @property {!Object<string, StrokeState>} [strokeStates] The stroke states (decluttering).
   */

  /**
   * @typedef {Object<number, import("./canvas/Executor.js").ReplayImageOrLabelArgs>} DeclutterImageWithText
   */

  /**
   * @const
   * @type {string}
   */
  const defaultFont = '10px sans-serif';

  /**
   * @const
   * @type {import("../colorlike.js").ColorLike}
   */
  const defaultFillStyle = '#000';

  /**
   * @const
   * @type {CanvasLineCap}
   */
  const defaultLineCap = 'round';

  /**
   * @const
   * @type {Array<number>}
   */
  const defaultLineDash = [];

  /**
   * @const
   * @type {number}
   */
  const defaultLineDashOffset = 0;

  /**
   * @const
   * @type {CanvasLineJoin}
   */
  const defaultLineJoin = 'round';

  /**
   * @const
   * @type {number}
   */
  const defaultMiterLimit = 10;

  /**
   * @const
   * @type {import("../colorlike.js").ColorLike}
   */
  const defaultStrokeStyle = '#000';

  /**
   * @const
   * @type {CanvasTextAlign}
   */
  const defaultTextAlign = 'center';

  /**
   * @const
   * @type {CanvasTextBaseline}
   */
  const defaultTextBaseline = 'middle';

  /**
   * @const
   * @type {Array<number>}
   */
  const defaultPadding = [0, 0, 0, 0];

  /**
   * @const
   * @type {number}
   */
  const defaultLineWidth = 1;

  /**
   * @type {BaseObject}
   */
  const checkedFonts = new BaseObject();

  /**
   * @type {CanvasRenderingContext2D}
   */
  let measureContext$1 = null;

  /**
   * @type {string}
   */
  let measureFont;

  /**
   * @type {!Object<string, number>}
   */
  const textHeights = {};

  /**
   * Clears the label cache when a font becomes available.
   * @param {string} fontSpec CSS font spec.
   */
  const registerFont = (function () {
    const retries = 100;
    const size = '32px ';
    const referenceFonts = ['monospace', 'serif'];
    const len = referenceFonts.length;
    const text = 'wmytzilWMYTZIL@#/&?$%10\uF013';
    let interval, referenceWidth;

    /**
     * @param {string} fontStyle Css font-style
     * @param {string} fontWeight Css font-weight
     * @param {*} fontFamily Css font-family
     * @return {boolean} Font with style and weight is available
     */
    function isAvailable(fontStyle, fontWeight, fontFamily) {
      let available = true;
      for (let i = 0; i < len; ++i) {
        const referenceFont = referenceFonts[i];
        referenceWidth = measureTextWidth(
          fontStyle + ' ' + fontWeight + ' ' + size + referenceFont,
          text
        );
        if (fontFamily != referenceFont) {
          const width = measureTextWidth(
            fontStyle +
              ' ' +
              fontWeight +
              ' ' +
              size +
              fontFamily +
              ',' +
              referenceFont,
            text
          );
          // If width and referenceWidth are the same, then the fallback was used
          // instead of the font we wanted, so the font is not available.
          available = available && width != referenceWidth;
        }
      }
      if (available) {
        return true;
      }
      return false;
    }

    function check() {
      let done = true;
      const fonts = checkedFonts.getKeys();
      for (let i = 0, ii = fonts.length; i < ii; ++i) {
        const font = fonts[i];
        if (checkedFonts.get(font) < retries) {
          if (isAvailable.apply(this, font.split('\n'))) {
            clear$2(textHeights);
            // Make sure that loaded fonts are picked up by Safari
            measureContext$1 = null;
            measureFont = undefined;
            checkedFonts.set(font, retries);
          } else {
            checkedFonts.set(font, checkedFonts.get(font) + 1, true);
            done = false;
          }
        }
      }
      if (done) {
        clearInterval(interval);
        interval = undefined;
      }
    }

    return function (fontSpec) {
      const font = getFontParameters(fontSpec);
      if (!font) {
        return;
      }
      const families = font.families;
      for (let i = 0, ii = families.length; i < ii; ++i) {
        const family = families[i];
        const key = font.style + '\n' + font.weight + '\n' + family;
        if (checkedFonts.get(key) === undefined) {
          checkedFonts.set(key, retries, true);
          if (!isAvailable(font.style, font.weight, family)) {
            checkedFonts.set(key, 0, true);
            if (interval === undefined) {
              interval = setInterval(check, 32);
            }
          }
        }
      }
    };
  })();

  /**
   * @param {string} font Font to use for measuring.
   * @return {import("../size.js").Size} Measurement.
   */
  const measureTextHeight = (function () {
    /**
     * @type {HTMLDivElement}
     */
    let measureElement;
    return function (fontSpec) {
      let height = textHeights[fontSpec];
      if (height == undefined) {
        if (WORKER_OFFSCREEN_CANVAS) {
          const font = getFontParameters(fontSpec);
          const metrics = measureText$1(fontSpec, 'Žg');
          const lineHeight = isNaN(Number(font.lineHeight))
            ? 1.2
            : Number(font.lineHeight);
          height =
            lineHeight *
            (metrics.actualBoundingBoxAscent + metrics.actualBoundingBoxDescent);
        } else {
          if (!measureElement) {
            measureElement = document.createElement('div');
            measureElement.innerHTML = 'M';
            measureElement.style.minHeight = '0';
            measureElement.style.maxHeight = 'none';
            measureElement.style.height = 'auto';
            measureElement.style.padding = '0';
            measureElement.style.border = 'none';
            measureElement.style.position = 'absolute';
            measureElement.style.display = 'block';
            measureElement.style.left = '-99999px';
          }
          measureElement.style.font = fontSpec;
          document.body.appendChild(measureElement);
          height = measureElement.offsetHeight;
          document.body.removeChild(measureElement);
        }
        textHeights[fontSpec] = height;
      }
      return height;
    };
  })();

  /**
   * @param {string} font Font.
   * @param {string} text Text.
   * @return {TextMetrics} Text metrics.
   */
  function measureText$1(font, text) {
    if (!measureContext$1) {
      measureContext$1 = createCanvasContext2D(1, 1);
    }
    if (font != measureFont) {
      measureContext$1.font = font;
      measureFont = measureContext$1.font;
    }
    return measureContext$1.measureText(text);
  }

  /**
   * @param {string} font Font.
   * @param {string} text Text.
   * @return {number} Width.
   */
  function measureTextWidth(font, text) {
    return measureText$1(font, text).width;
  }

  /**
   * Measure text width using a cache.
   * @param {string} font The font.
   * @param {string} text The text to measure.
   * @param {Object<string, number>} cache A lookup of cached widths by text.
   * @return {number} The text width.
   */
  function measureAndCacheTextWidth(font, text, cache) {
    if (text in cache) {
      return cache[text];
    }
    const width = text
      .split('\n')
      .reduce((prev, curr) => Math.max(prev, measureTextWidth(font, curr)), 0);
    cache[text] = width;
    return width;
  }

  /**
   * @param {TextState} baseStyle Base style.
   * @param {Array<string>} chunks Text chunks to measure.
   * @return {{width: number, height: number, widths: Array<number>, heights: Array<number>, lineWidths: Array<number>}}} Text metrics.
   */
  function getTextDimensions(baseStyle, chunks) {
    const widths = [];
    const heights = [];
    const lineWidths = [];
    let width = 0;
    let lineWidth = 0;
    let height = 0;
    let lineHeight = 0;
    for (let i = 0, ii = chunks.length; i <= ii; i += 2) {
      const text = chunks[i];
      if (text === '\n' || i === ii) {
        width = Math.max(width, lineWidth);
        lineWidths.push(lineWidth);
        lineWidth = 0;
        height += lineHeight;
        continue;
      }
      const font = chunks[i + 1] || baseStyle.font;
      const currentWidth = measureTextWidth(font, text);
      widths.push(currentWidth);
      lineWidth += currentWidth;
      const currentHeight = measureTextHeight(font);
      heights.push(currentHeight);
      lineHeight = Math.max(lineHeight, currentHeight);
    }
    return {width, height, widths, heights, lineWidths};
  }

  /**
   * @param {CanvasRenderingContext2D} context Context.
   * @param {number} rotation Rotation.
   * @param {number} offsetX X offset.
   * @param {number} offsetY Y offset.
   */
  function rotateAtOffset(context, rotation, offsetX, offsetY) {
    if (rotation !== 0) {
      context.translate(offsetX, offsetY);
      context.rotate(rotation);
      context.translate(-offsetX, -offsetY);
    }
  }

  /**
   * @param {CanvasRenderingContext2D} context Context.
   * @param {import("../transform.js").Transform|null} transform Transform.
   * @param {number} opacity Opacity.
   * @param {Label|HTMLCanvasElement|HTMLImageElement|HTMLVideoElement} labelOrImage Label.
   * @param {number} originX Origin X.
   * @param {number} originY Origin Y.
   * @param {number} w Width.
   * @param {number} h Height.
   * @param {number} x X.
   * @param {number} y Y.
   * @param {import("../size.js").Size} scale Scale.
   */
  function drawImageOrLabel(
    context,
    transform,
    opacity,
    labelOrImage,
    originX,
    originY,
    w,
    h,
    x,
    y,
    scale
  ) {
    context.save();

    if (opacity !== 1) {
      context.globalAlpha *= opacity;
    }
    if (transform) {
      context.setTransform.apply(context, transform);
    }

    if (/** @type {*} */ (labelOrImage).contextInstructions) {
      // label
      context.translate(x, y);
      context.scale(scale[0], scale[1]);
      executeLabelInstructions(/** @type {Label} */ (labelOrImage), context);
    } else if (scale[0] < 0 || scale[1] < 0) {
      // flipped image
      context.translate(x, y);
      context.scale(scale[0], scale[1]);
      context.drawImage(
        /** @type {HTMLCanvasElement|HTMLImageElement|HTMLVideoElement} */ (
          labelOrImage
        ),
        originX,
        originY,
        w,
        h,
        0,
        0,
        w,
        h
      );
    } else {
      // if image not flipped translate and scale can be avoided
      context.drawImage(
        /** @type {HTMLCanvasElement|HTMLImageElement|HTMLVideoElement} */ (
          labelOrImage
        ),
        originX,
        originY,
        w,
        h,
        x,
        y,
        w * scale[0],
        h * scale[1]
      );
    }

    context.restore();
  }

  /**
   * @param {Label} label Label.
   * @param {CanvasRenderingContext2D} context Context.
   */
  function executeLabelInstructions(label, context) {
    const contextInstructions = label.contextInstructions;
    for (let i = 0, ii = contextInstructions.length; i < ii; i += 2) {
      if (Array.isArray(contextInstructions[i + 1])) {
        context[contextInstructions[i]].apply(
          context,
          contextInstructions[i + 1]
        );
      } else {
        context[contextInstructions[i]] = contextInstructions[i + 1];
      }
    }
  }

  /**
   * @module ol/renderer/Composite
   */

  /**
   * @classdesc
   * Canvas map renderer.
   * @api
   */
  class CompositeMapRenderer extends MapRenderer {
    /**
     * @param {import("../Map.js").default} map Map.
     */
    constructor(map) {
      super(map);

      /**
       * @type {import("../events.js").EventsKey}
       */
      this.fontChangeListenerKey_ = listen(
        checkedFonts,
        ObjectEventType.PROPERTYCHANGE,
        map.redrawText.bind(map)
      );

      /**
       * @private
       * @type {HTMLDivElement}
       */
      this.element_ = document.createElement('div');
      const style = this.element_.style;
      style.position = 'absolute';
      style.width = '100%';
      style.height = '100%';
      style.zIndex = '0';

      this.element_.className = CLASS_UNSELECTABLE + ' ol-layers';

      const container = map.getViewport();
      container.insertBefore(this.element_, container.firstChild || null);

      /**
       * @private
       * @type {Array<HTMLElement>}
       */
      this.children_ = [];

      /**
       * @private
       * @type {boolean}
       */
      this.renderedVisible_ = true;
    }

    /**
     * @param {import("../render/EventType.js").default} type Event type.
     * @param {import("../Map.js").FrameState} frameState Frame state.
     */
    dispatchRenderEvent(type, frameState) {
      const map = this.getMap();
      if (map.hasListener(type)) {
        const event = new RenderEvent(type, undefined, frameState);
        map.dispatchEvent(event);
      }
    }

    disposeInternal() {
      unlistenByKey(this.fontChangeListenerKey_);
      this.element_.parentNode.removeChild(this.element_);
      super.disposeInternal();
    }

    /**
     * Render.
     * @param {?import("../Map.js").FrameState} frameState Frame state.
     */
    renderFrame(frameState) {
      if (!frameState) {
        if (this.renderedVisible_) {
          this.element_.style.display = 'none';
          this.renderedVisible_ = false;
        }
        return;
      }

      this.calculateMatrices2D(frameState);
      this.dispatchRenderEvent(EventType.PRECOMPOSE, frameState);

      const layerStatesArray = frameState.layerStatesArray.sort(function (a, b) {
        return a.zIndex - b.zIndex;
      });
      const viewState = frameState.viewState;

      this.children_.length = 0;
      /**
       * @type {Array<import("../layer/BaseVector.js").default>}
       */
      const declutterLayers = [];
      let previousElement = null;
      for (let i = 0, ii = layerStatesArray.length; i < ii; ++i) {
        const layerState = layerStatesArray[i];
        frameState.layerIndex = i;

        const layer = layerState.layer;
        const sourceState = layer.getSourceState();
        if (
          !inView(layerState, viewState) ||
          (sourceState != 'ready' && sourceState != 'undefined')
        ) {
          layer.unrender();
          continue;
        }

        const element = layer.render(frameState, previousElement);
        if (!element) {
          continue;
        }
        if (element !== previousElement) {
          this.children_.push(element);
          previousElement = element;
        }
        if ('getDeclutter' in layer) {
          declutterLayers.push(
            /** @type {import("../layer/BaseVector.js").default} */ (layer)
          );
        }
      }
      for (let i = declutterLayers.length - 1; i >= 0; --i) {
        declutterLayers[i].renderDeclutter(frameState);
      }

      replaceChildren(this.element_, this.children_);

      this.dispatchRenderEvent(EventType.POSTCOMPOSE, frameState);

      if (!this.renderedVisible_) {
        this.element_.style.display = '';
        this.renderedVisible_ = true;
      }

      this.scheduleExpireIconCache(frameState);
    }
  }

  /**
   * @module ol/layer/Group
   */

  /**
   * @typedef {'addlayer'|'removelayer'} EventType
   */

  /**
   * @classdesc
   * A layer group triggers 'addlayer' and 'removelayer' events when layers are added to or removed from
   * the group or one of its child groups.  When a layer group is added to or removed from another layer group,
   * a single event will be triggered (instead of one per layer in the group added or removed).
   */
  class GroupEvent extends BaseEvent {
    /**
     * @param {EventType} type The event type.
     * @param {BaseLayer} layer The layer.
     */
    constructor(type, layer) {
      super(type);

      /**
       * The added or removed layer.
       * @type {BaseLayer}
       * @api
       */
      this.layer = layer;
    }
  }

  /***
   * @template Return
   * @typedef {import("../Observable").OnSignature<import("../Observable").EventTypes, import("../events/Event.js").default, Return> &
   *   import("../Observable").OnSignature<import("./Base").BaseLayerObjectEventTypes|
   *     'change:layers', import("../Object").ObjectEvent, Return> &
   *   import("../Observable").CombinedOnSignature<import("../Observable").EventTypes|import("./Base").BaseLayerObjectEventTypes|'change:layers', Return>} GroupOnSignature
   */

  /**
   * @typedef {Object} Options
   * @property {number} [opacity=1] Opacity (0, 1).
   * @property {boolean} [visible=true] Visibility.
   * @property {import("../extent.js").Extent} [extent] The bounding extent for layer rendering.  The layer will not be
   * rendered outside of this extent.
   * @property {number} [zIndex] The z-index for layer rendering.  At rendering time, the layers
   * will be ordered, first by Z-index and then by position. When `undefined`, a `zIndex` of 0 is assumed
   * for layers that are added to the map's `layers` collection, or `Infinity` when the layer's `setMap()`
   * method was used.
   * @property {number} [minResolution] The minimum resolution (inclusive) at which this layer will be
   * visible.
   * @property {number} [maxResolution] The maximum resolution (exclusive) below which this layer will
   * be visible.
   * @property {number} [minZoom] The minimum view zoom level (exclusive) above which this layer will be
   * visible.
   * @property {number} [maxZoom] The maximum view zoom level (inclusive) at which this layer will
   * be visible.
   * @property {Array<import("./Base.js").default>|Collection<import("./Base.js").default>} [layers] Child layers.
   * @property {Object<string, *>} [properties] Arbitrary observable properties. Can be accessed with `#get()` and `#set()`.
   */

  /**
   * @enum {string}
   * @private
   */
  const Property$3 = {
    LAYERS: 'layers',
  };

  /**
   * @classdesc
   * A {@link module:ol/Collection~Collection} of layers that are handled together.
   *
   * A generic `change` event is triggered when the group/Collection changes.
   *
   * @api
   */
  class LayerGroup extends BaseLayer {
    /**
     * @param {Options} [options] Layer options.
     */
    constructor(options) {
      options = options || {};
      const baseOptions = /** @type {Options} */ (Object.assign({}, options));
      delete baseOptions.layers;

      let layers = options.layers;

      super(baseOptions);

      /***
       * @type {GroupOnSignature<import("../events").EventsKey>}
       */
      this.on;

      /***
       * @type {GroupOnSignature<import("../events").EventsKey>}
       */
      this.once;

      /***
       * @type {GroupOnSignature<void>}
       */
      this.un;

      /**
       * @private
       * @type {Array<import("../events.js").EventsKey>}
       */
      this.layersListenerKeys_ = [];

      /**
       * @private
       * @type {Object<string, Array<import("../events.js").EventsKey>>}
       */
      this.listenerKeys_ = {};

      this.addChangeListener(Property$3.LAYERS, this.handleLayersChanged_);

      if (layers) {
        if (Array.isArray(layers)) {
          layers = new Collection(layers.slice(), {unique: true});
        } else {
          assert(typeof (/** @type {?} */ (layers).getArray) === 'function', 43); // Expected `layers` to be an array or a `Collection`
        }
      } else {
        layers = new Collection(undefined, {unique: true});
      }

      this.setLayers(layers);
    }

    /**
     * @private
     */
    handleLayerChange_() {
      this.changed();
    }

    /**
     * @private
     */
    handleLayersChanged_() {
      this.layersListenerKeys_.forEach(unlistenByKey);
      this.layersListenerKeys_.length = 0;

      const layers = this.getLayers();
      this.layersListenerKeys_.push(
        listen(layers, CollectionEventType.ADD, this.handleLayersAdd_, this),
        listen(layers, CollectionEventType.REMOVE, this.handleLayersRemove_, this)
      );

      for (const id in this.listenerKeys_) {
        this.listenerKeys_[id].forEach(unlistenByKey);
      }
      clear$2(this.listenerKeys_);

      const layersArray = layers.getArray();
      for (let i = 0, ii = layersArray.length; i < ii; i++) {
        const layer = layersArray[i];
        this.registerLayerListeners_(layer);
        this.dispatchEvent(new GroupEvent('addlayer', layer));
      }
      this.changed();
    }

    /**
     * @param {BaseLayer} layer The layer.
     */
    registerLayerListeners_(layer) {
      const listenerKeys = [
        listen(
          layer,
          ObjectEventType.PROPERTYCHANGE,
          this.handleLayerChange_,
          this
        ),
        listen(layer, EventType$1.CHANGE, this.handleLayerChange_, this),
      ];

      if (layer instanceof LayerGroup) {
        listenerKeys.push(
          listen(layer, 'addlayer', this.handleLayerGroupAdd_, this),
          listen(layer, 'removelayer', this.handleLayerGroupRemove_, this)
        );
      }

      this.listenerKeys_[getUid(layer)] = listenerKeys;
    }

    /**
     * @param {GroupEvent} event The layer group event.
     */
    handleLayerGroupAdd_(event) {
      this.dispatchEvent(new GroupEvent('addlayer', event.layer));
    }

    /**
     * @param {GroupEvent} event The layer group event.
     */
    handleLayerGroupRemove_(event) {
      this.dispatchEvent(new GroupEvent('removelayer', event.layer));
    }

    /**
     * @param {import("../Collection.js").CollectionEvent<import("./Base.js").default>} collectionEvent CollectionEvent.
     * @private
     */
    handleLayersAdd_(collectionEvent) {
      const layer = collectionEvent.element;
      this.registerLayerListeners_(layer);
      this.dispatchEvent(new GroupEvent('addlayer', layer));
      this.changed();
    }

    /**
     * @param {import("../Collection.js").CollectionEvent<import("./Base.js").default>} collectionEvent CollectionEvent.
     * @private
     */
    handleLayersRemove_(collectionEvent) {
      const layer = collectionEvent.element;
      const key = getUid(layer);
      this.listenerKeys_[key].forEach(unlistenByKey);
      delete this.listenerKeys_[key];
      this.dispatchEvent(new GroupEvent('removelayer', layer));
      this.changed();
    }

    /**
     * Returns the {@link module:ol/Collection~Collection collection} of {@link module:ol/layer/Layer~Layer layers}
     * in this group.
     * @return {!Collection<import("./Base.js").default>} Collection of
     *   {@link module:ol/layer/Base~BaseLayer layers} that are part of this group.
     * @observable
     * @api
     */
    getLayers() {
      return /** @type {!Collection<import("./Base.js").default>} */ (
        this.get(Property$3.LAYERS)
      );
    }

    /**
     * Set the {@link module:ol/Collection~Collection collection} of {@link module:ol/layer/Layer~Layer layers}
     * in this group.
     * @param {!Collection<import("./Base.js").default>} layers Collection of
     *   {@link module:ol/layer/Base~BaseLayer layers} that are part of this group.
     * @observable
     * @api
     */
    setLayers(layers) {
      const collection = this.getLayers();
      if (collection) {
        const currentLayers = collection.getArray();
        for (let i = 0, ii = currentLayers.length; i < ii; ++i) {
          this.dispatchEvent(new GroupEvent('removelayer', currentLayers[i]));
        }
      }

      this.set(Property$3.LAYERS, layers);
    }

    /**
     * @param {Array<import("./Layer.js").default>} [array] Array of layers (to be modified in place).
     * @return {Array<import("./Layer.js").default>} Array of layers.
     */
    getLayersArray(array) {
      array = array !== undefined ? array : [];
      this.getLayers().forEach(function (layer) {
        layer.getLayersArray(array);
      });
      return array;
    }

    /**
     * Get the layer states list and use this groups z-index as the default
     * for all layers in this and nested groups, if it is unset at this point.
     * If dest is not provided and this group's z-index is undefined
     * 0 is used a the default z-index.
     * @param {Array<import("./Layer.js").State>} [dest] Optional list
     * of layer states (to be modified in place).
     * @return {Array<import("./Layer.js").State>} List of layer states.
     */
    getLayerStatesArray(dest) {
      const states = dest !== undefined ? dest : [];
      const pos = states.length;

      this.getLayers().forEach(function (layer) {
        layer.getLayerStatesArray(states);
      });

      const ownLayerState = this.getLayerState();
      let defaultZIndex = ownLayerState.zIndex;
      if (!dest && ownLayerState.zIndex === undefined) {
        defaultZIndex = 0;
      }
      for (let i = pos, ii = states.length; i < ii; i++) {
        const layerState = states[i];
        layerState.opacity *= ownLayerState.opacity;
        layerState.visible = layerState.visible && ownLayerState.visible;
        layerState.maxResolution = Math.min(
          layerState.maxResolution,
          ownLayerState.maxResolution
        );
        layerState.minResolution = Math.max(
          layerState.minResolution,
          ownLayerState.minResolution
        );
        layerState.minZoom = Math.max(layerState.minZoom, ownLayerState.minZoom);
        layerState.maxZoom = Math.min(layerState.maxZoom, ownLayerState.maxZoom);
        if (ownLayerState.extent !== undefined) {
          if (layerState.extent !== undefined) {
            layerState.extent = getIntersection(
              layerState.extent,
              ownLayerState.extent
            );
          } else {
            layerState.extent = ownLayerState.extent;
          }
        }
        if (layerState.zIndex === undefined) {
          layerState.zIndex = defaultZIndex;
        }
      }

      return states;
    }

    /**
     * @return {import("../source/Source.js").State} Source state.
     */
    getSourceState() {
      return 'ready';
    }
  }

  /**
   * @module ol/MapEvent
   */

  /**
   * @classdesc
   * Events emitted as map events are instances of this type.
   * See {@link module:ol/Map~Map} for which events trigger a map event.
   */
  class MapEvent extends BaseEvent {
    /**
     * @param {string} type Event type.
     * @param {import("./Map.js").default} map Map.
     * @param {?import("./Map.js").FrameState} [frameState] Frame state.
     */
    constructor(type, map, frameState) {
      super(type);

      /**
       * The map where the event occurred.
       * @type {import("./Map.js").default}
       * @api
       */
      this.map = map;

      /**
       * The frame state at the time of the event.
       * @type {?import("./Map.js").FrameState}
       * @api
       */
      this.frameState = frameState !== undefined ? frameState : null;
    }
  }

  /**
   * @module ol/MapBrowserEvent
   */

  /**
   * @classdesc
   * Events emitted as map browser events are instances of this type.
   * See {@link module:ol/Map~Map} for which events trigger a map browser event.
   * @template {UIEvent} EVENT
   */
  class MapBrowserEvent extends MapEvent {
    /**
     * @param {string} type Event type.
     * @param {import("./Map.js").default} map Map.
     * @param {EVENT} originalEvent Original event.
     * @param {boolean} [dragging] Is the map currently being dragged?
     * @param {import("./Map.js").FrameState} [frameState] Frame state.
     * @param {Array<PointerEvent>} [activePointers] Active pointers.
     */
    constructor(type, map, originalEvent, dragging, frameState, activePointers) {
      super(type, map, frameState);

      /**
       * The original browser event.
       * @const
       * @type {EVENT}
       * @api
       */
      this.originalEvent = originalEvent;

      /**
       * The map pixel relative to the viewport corresponding to the original browser event.
       * @type {?import("./pixel.js").Pixel}
       */
      this.pixel_ = null;

      /**
       * The coordinate in the user projection corresponding to the original browser event.
       * @type {?import("./coordinate.js").Coordinate}
       */
      this.coordinate_ = null;

      /**
       * Indicates if the map is currently being dragged. Only set for
       * `POINTERDRAG` and `POINTERMOVE` events. Default is `false`.
       *
       * @type {boolean}
       * @api
       */
      this.dragging = dragging !== undefined ? dragging : false;

      /**
       * @type {Array<PointerEvent>|undefined}
       */
      this.activePointers = activePointers;
    }

    /**
     * The map pixel relative to the viewport corresponding to the original event.
     * @type {import("./pixel.js").Pixel}
     * @api
     */
    get pixel() {
      if (!this.pixel_) {
        this.pixel_ = this.map.getEventPixel(this.originalEvent);
      }
      return this.pixel_;
    }
    set pixel(pixel) {
      this.pixel_ = pixel;
    }

    /**
     * The coordinate corresponding to the original browser event.  This will be in the user
     * projection if one is set.  Otherwise it will be in the view projection.
     * @type {import("./coordinate.js").Coordinate}
     * @api
     */
    get coordinate() {
      if (!this.coordinate_) {
        this.coordinate_ = this.map.getCoordinateFromPixel(this.pixel);
      }
      return this.coordinate_;
    }
    set coordinate(coordinate) {
      this.coordinate_ = coordinate;
    }

    /**
     * Prevents the default browser action.
     * See https://developer.mozilla.org/en-US/docs/Web/API/event.preventDefault.
     * @api
     */
    preventDefault() {
      super.preventDefault();
      if ('preventDefault' in this.originalEvent) {
        /** @type {UIEvent} */ (this.originalEvent).preventDefault();
      }
    }

    /**
     * Prevents further propagation of the current event.
     * See https://developer.mozilla.org/en-US/docs/Web/API/event.stopPropagation.
     * @api
     */
    stopPropagation() {
      super.stopPropagation();
      if ('stopPropagation' in this.originalEvent) {
        /** @type {UIEvent} */ (this.originalEvent).stopPropagation();
      }
    }
  }

  /**
   * @module ol/MapBrowserEventType
   */

  /**
   * Constants for event names.
   * @enum {string}
   */
  var MapBrowserEventType = {
    /**
     * A true single click with no dragging and no double click. Note that this
     * event is delayed by 250 ms to ensure that it is not a double click.
     * @event module:ol/MapBrowserEvent~MapBrowserEvent#singleclick
     * @api
     */
    SINGLECLICK: 'singleclick',

    /**
     * A click with no dragging. A double click will fire two of this.
     * @event module:ol/MapBrowserEvent~MapBrowserEvent#click
     * @api
     */
    CLICK: EventType$1.CLICK,

    /**
     * A true double click, with no dragging.
     * @event module:ol/MapBrowserEvent~MapBrowserEvent#dblclick
     * @api
     */
    DBLCLICK: EventType$1.DBLCLICK,

    /**
     * Triggered when a pointer is dragged.
     * @event module:ol/MapBrowserEvent~MapBrowserEvent#pointerdrag
     * @api
     */
    POINTERDRAG: 'pointerdrag',

    /**
     * Triggered when a pointer is moved. Note that on touch devices this is
     * triggered when the map is panned, so is not the same as mousemove.
     * @event module:ol/MapBrowserEvent~MapBrowserEvent#pointermove
     * @api
     */
    POINTERMOVE: 'pointermove',

    POINTERDOWN: 'pointerdown',
    POINTERUP: 'pointerup',
    POINTEROVER: 'pointerover',
    POINTEROUT: 'pointerout',
    POINTERENTER: 'pointerenter',
    POINTERLEAVE: 'pointerleave',
    POINTERCANCEL: 'pointercancel',
  };

  /***
   * @typedef {'singleclick'|'click'|'dblclick'|'pointerdrag'|'pointermove'} Types
   */

  /**
   * @module ol/pointer/EventType
   */

  /**
   * Constants for event names.
   * @enum {string}
   */
  var PointerEventType = {
    POINTERMOVE: 'pointermove',
    POINTERDOWN: 'pointerdown',
    POINTERUP: 'pointerup',
    POINTEROVER: 'pointerover',
    POINTEROUT: 'pointerout',
    POINTERENTER: 'pointerenter',
    POINTERLEAVE: 'pointerleave',
    POINTERCANCEL: 'pointercancel',
  };

  /**
   * @module ol/MapBrowserEventHandler
   */

  class MapBrowserEventHandler extends Target {
    /**
     * @param {import("./Map.js").default} map The map with the viewport to listen to events on.
     * @param {number} [moveTolerance] The minimal distance the pointer must travel to trigger a move.
     */
    constructor(map, moveTolerance) {
      super(map);

      /**
       * This is the element that we will listen to the real events on.
       * @type {import("./Map.js").default}
       * @private
       */
      this.map_ = map;

      /**
       * @type {any}
       * @private
       */
      this.clickTimeoutId_;

      /**
       * Emulate dblclick and singleclick. Will be true when only one pointer is active.
       * @type {boolean}
       */
      this.emulateClicks_ = false;

      /**
       * @type {boolean}
       * @private
       */
      this.dragging_ = false;

      /**
       * @type {!Array<import("./events.js").EventsKey>}
       * @private
       */
      this.dragListenerKeys_ = [];

      /**
       * @type {number}
       * @private
       */
      this.moveTolerance_ = moveTolerance === undefined ? 1 : moveTolerance;

      /**
       * The most recent "down" type event (or null if none have occurred).
       * Set on pointerdown.
       * @type {PointerEvent|null}
       * @private
       */
      this.down_ = null;

      const element = this.map_.getViewport();

      /**
       * @type {Array<PointerEvent>}
       * @private
       */
      this.activePointers_ = [];

      /**
       * @type {!Object<number, Event>}
       * @private
       */
      this.trackedTouches_ = {};

      this.element_ = element;

      /**
       * @type {?import("./events.js").EventsKey}
       * @private
       */
      this.pointerdownListenerKey_ = listen(
        element,
        PointerEventType.POINTERDOWN,
        this.handlePointerDown_,
        this
      );

      /**
       * @type {PointerEvent}
       * @private
       */
      this.originalPointerMoveEvent_;

      /**
       * @type {?import("./events.js").EventsKey}
       * @private
       */
      this.relayedListenerKey_ = listen(
        element,
        PointerEventType.POINTERMOVE,
        this.relayMoveEvent_,
        this
      );

      /**
       * @private
       */
      this.boundHandleTouchMove_ = this.handleTouchMove_.bind(this);

      this.element_.addEventListener(
        EventType$1.TOUCHMOVE,
        this.boundHandleTouchMove_,
        PASSIVE_EVENT_LISTENERS ? {passive: false} : false
      );
    }

    /**
     * @param {PointerEvent} pointerEvent Pointer
     * event.
     * @private
     */
    emulateClick_(pointerEvent) {
      let newEvent = new MapBrowserEvent(
        MapBrowserEventType.CLICK,
        this.map_,
        pointerEvent
      );
      this.dispatchEvent(newEvent);
      if (this.clickTimeoutId_ !== undefined) {
        // double-click
        clearTimeout(this.clickTimeoutId_);
        this.clickTimeoutId_ = undefined;
        newEvent = new MapBrowserEvent(
          MapBrowserEventType.DBLCLICK,
          this.map_,
          pointerEvent
        );
        this.dispatchEvent(newEvent);
      } else {
        // click
        this.clickTimeoutId_ = setTimeout(
          /** @this {MapBrowserEventHandler} */
          function () {
            this.clickTimeoutId_ = undefined;
            const newEvent = new MapBrowserEvent(
              MapBrowserEventType.SINGLECLICK,
              this.map_,
              pointerEvent
            );
            this.dispatchEvent(newEvent);
          }.bind(this),
          250
        );
      }
    }

    /**
     * Keeps track on how many pointers are currently active.
     *
     * @param {PointerEvent} pointerEvent Pointer
     * event.
     * @private
     */
    updateActivePointers_(pointerEvent) {
      const event = pointerEvent;
      const id = event.pointerId;

      if (
        event.type == MapBrowserEventType.POINTERUP ||
        event.type == MapBrowserEventType.POINTERCANCEL
      ) {
        delete this.trackedTouches_[id];
        for (const pointerId in this.trackedTouches_) {
          if (this.trackedTouches_[pointerId].target !== event.target) {
            // Some platforms assign a new pointerId when the target changes.
            // If this happens, delete one tracked pointer. If there is more
            // than one tracked pointer for the old target, it will be cleared
            // by subsequent POINTERUP events from other pointers.
            delete this.trackedTouches_[pointerId];
            break;
          }
        }
      } else if (
        event.type == MapBrowserEventType.POINTERDOWN ||
        event.type == MapBrowserEventType.POINTERMOVE
      ) {
        this.trackedTouches_[id] = event;
      }
      this.activePointers_ = Object.values(this.trackedTouches_);
    }

    /**
     * @param {PointerEvent} pointerEvent Pointer
     * event.
     * @private
     */
    handlePointerUp_(pointerEvent) {
      this.updateActivePointers_(pointerEvent);
      const newEvent = new MapBrowserEvent(
        MapBrowserEventType.POINTERUP,
        this.map_,
        pointerEvent,
        undefined,
        undefined,
        this.activePointers_
      );
      this.dispatchEvent(newEvent);

      // We emulate click events on left mouse button click, touch contact, and pen
      // contact. isMouseActionButton returns true in these cases (evt.button is set
      // to 0).
      // See http://www.w3.org/TR/pointerevents/#button-states
      // We only fire click, singleclick, and doubleclick if nobody has called
      // event.preventDefault().
      if (
        this.emulateClicks_ &&
        !newEvent.defaultPrevented &&
        !this.dragging_ &&
        this.isMouseActionButton_(pointerEvent)
      ) {
        this.emulateClick_(this.down_);
      }

      if (this.activePointers_.length === 0) {
        this.dragListenerKeys_.forEach(unlistenByKey);
        this.dragListenerKeys_.length = 0;
        this.dragging_ = false;
        this.down_ = null;
      }
    }

    /**
     * @param {PointerEvent} pointerEvent Pointer
     * event.
     * @return {boolean} If the left mouse button was pressed.
     * @private
     */
    isMouseActionButton_(pointerEvent) {
      return pointerEvent.button === 0;
    }

    /**
     * @param {PointerEvent} pointerEvent Pointer
     * event.
     * @private
     */
    handlePointerDown_(pointerEvent) {
      this.emulateClicks_ = this.activePointers_.length === 0;
      this.updateActivePointers_(pointerEvent);
      const newEvent = new MapBrowserEvent(
        MapBrowserEventType.POINTERDOWN,
        this.map_,
        pointerEvent,
        undefined,
        undefined,
        this.activePointers_
      );
      this.dispatchEvent(newEvent);

      // Store a copy of the down event
      this.down_ = /** @type {PointerEvent} */ ({});
      for (const property in pointerEvent) {
        const value = pointerEvent[property];
        this.down_[property] = typeof value === 'function' ? VOID : value;
      }

      if (this.dragListenerKeys_.length === 0) {
        const doc = this.map_.getOwnerDocument();
        this.dragListenerKeys_.push(
          listen(
            doc,
            MapBrowserEventType.POINTERMOVE,
            this.handlePointerMove_,
            this
          ),
          listen(doc, MapBrowserEventType.POINTERUP, this.handlePointerUp_, this),
          /* Note that the listener for `pointercancel is set up on
           * `pointerEventHandler_` and not `documentPointerEventHandler_` like
           * the `pointerup` and `pointermove` listeners.
           *
           * The reason for this is the following: `TouchSource.vacuumTouches_()`
           * issues `pointercancel` events, when there was no `touchend` for a
           * `touchstart`. Now, let's say a first `touchstart` is registered on
           * `pointerEventHandler_`. The `documentPointerEventHandler_` is set up.
           * But `documentPointerEventHandler_` doesn't know about the first
           * `touchstart`. If there is no `touchend` for the `touchstart`, we can
           * only receive a `touchcancel` from `pointerEventHandler_`, because it is
           * only registered there.
           */
          listen(
            this.element_,
            MapBrowserEventType.POINTERCANCEL,
            this.handlePointerUp_,
            this
          )
        );
        if (this.element_.getRootNode && this.element_.getRootNode() !== doc) {
          this.dragListenerKeys_.push(
            listen(
              this.element_.getRootNode(),
              MapBrowserEventType.POINTERUP,
              this.handlePointerUp_,
              this
            )
          );
        }
      }
    }

    /**
     * @param {PointerEvent} pointerEvent Pointer
     * event.
     * @private
     */
    handlePointerMove_(pointerEvent) {
      // Between pointerdown and pointerup, pointermove events are triggered.
      // To avoid a 'false' touchmove event to be dispatched, we test if the pointer
      // moved a significant distance.
      if (this.isMoving_(pointerEvent)) {
        this.updateActivePointers_(pointerEvent);
        this.dragging_ = true;
        const newEvent = new MapBrowserEvent(
          MapBrowserEventType.POINTERDRAG,
          this.map_,
          pointerEvent,
          this.dragging_,
          undefined,
          this.activePointers_
        );
        this.dispatchEvent(newEvent);
      }
    }

    /**
     * Wrap and relay a pointermove event.
     * @param {PointerEvent} pointerEvent Pointer
     * event.
     * @private
     */
    relayMoveEvent_(pointerEvent) {
      this.originalPointerMoveEvent_ = pointerEvent;
      const dragging = !!(this.down_ && this.isMoving_(pointerEvent));
      this.dispatchEvent(
        new MapBrowserEvent(
          MapBrowserEventType.POINTERMOVE,
          this.map_,
          pointerEvent,
          dragging
        )
      );
    }

    /**
     * Flexible handling of a `touch-action: none` css equivalent: because calling
     * `preventDefault()` on a `pointermove` event does not stop native page scrolling
     * and zooming, we also listen for `touchmove` and call `preventDefault()` on it
     * when an interaction (currently `DragPan` handles the event.
     * @param {TouchEvent} event Event.
     * @private
     */
    handleTouchMove_(event) {
      // Due to https://github.com/mpizenberg/elm-pep/issues/2, `this.originalPointerMoveEvent_`
      // may not be initialized yet when we get here on a platform without native pointer events.
      const originalEvent = this.originalPointerMoveEvent_;
      if (
        (!originalEvent || originalEvent.defaultPrevented) &&
        (typeof event.cancelable !== 'boolean' || event.cancelable === true)
      ) {
        event.preventDefault();
      }
    }

    /**
     * @param {PointerEvent} pointerEvent Pointer
     * event.
     * @return {boolean} Is moving.
     * @private
     */
    isMoving_(pointerEvent) {
      return (
        this.dragging_ ||
        Math.abs(pointerEvent.clientX - this.down_.clientX) >
          this.moveTolerance_ ||
        Math.abs(pointerEvent.clientY - this.down_.clientY) > this.moveTolerance_
      );
    }

    /**
     * Clean up.
     */
    disposeInternal() {
      if (this.relayedListenerKey_) {
        unlistenByKey(this.relayedListenerKey_);
        this.relayedListenerKey_ = null;
      }
      this.element_.removeEventListener(
        EventType$1.TOUCHMOVE,
        this.boundHandleTouchMove_
      );

      if (this.pointerdownListenerKey_) {
        unlistenByKey(this.pointerdownListenerKey_);
        this.pointerdownListenerKey_ = null;
      }

      this.dragListenerKeys_.forEach(unlistenByKey);
      this.dragListenerKeys_.length = 0;

      this.element_ = null;
      super.disposeInternal();
    }
  }

  /**
   * @module ol/MapEventType
   */

  /**
   * @enum {string}
   */
  var MapEventType = {
    /**
     * Triggered after a map frame is rendered.
     * @event module:ol/MapEvent~MapEvent#postrender
     * @api
     */
    POSTRENDER: 'postrender',

    /**
     * Triggered when the map starts moving.
     * @event module:ol/MapEvent~MapEvent#movestart
     * @api
     */
    MOVESTART: 'movestart',

    /**
     * Triggered after the map is moved.
     * @event module:ol/MapEvent~MapEvent#moveend
     * @api
     */
    MOVEEND: 'moveend',

    /**
     * Triggered when loading of additional map data (tiles, images, features) starts.
     * @event module:ol/MapEvent~MapEvent#loadstart
     * @api
     */
    LOADSTART: 'loadstart',

    /**
     * Triggered when loading of additional map data has completed.
     * @event module:ol/MapEvent~MapEvent#loadend
     * @api
     */
    LOADEND: 'loadend',
  };

  /***
   * @typedef {'postrender'|'movestart'|'moveend'|'loadstart'|'loadend'} Types
   */

  /**
   * @module ol/MapProperty
   */

  /**
   * @enum {string}
   */
  var MapProperty = {
    LAYERGROUP: 'layergroup',
    SIZE: 'size',
    TARGET: 'target',
    VIEW: 'view',
  };

  /**
   * @module ol/structs/PriorityQueue
   */

  /**
   * @type {number}
   */
  const DROP = Infinity;

  /**
   * @classdesc
   * Priority queue.
   *
   * The implementation is inspired from the Closure Library's Heap class and
   * Python's heapq module.
   *
   * See https://github.com/google/closure-library/blob/master/closure/goog/structs/heap.js
   * and https://hg.python.org/cpython/file/2.7/Lib/heapq.py.
   *
   * @template T
   */
  class PriorityQueue {
    /**
     * @param {function(T): number} priorityFunction Priority function.
     * @param {function(T): string} keyFunction Key function.
     */
    constructor(priorityFunction, keyFunction) {
      /**
       * @type {function(T): number}
       * @private
       */
      this.priorityFunction_ = priorityFunction;

      /**
       * @type {function(T): string}
       * @private
       */
      this.keyFunction_ = keyFunction;

      /**
       * @type {Array<T>}
       * @private
       */
      this.elements_ = [];

      /**
       * @type {Array<number>}
       * @private
       */
      this.priorities_ = [];

      /**
       * @type {!Object<string, boolean>}
       * @private
       */
      this.queuedElements_ = {};
    }

    /**
     * FIXME empty description for jsdoc
     */
    clear() {
      this.elements_.length = 0;
      this.priorities_.length = 0;
      clear$2(this.queuedElements_);
    }

    /**
     * Remove and return the highest-priority element. O(log N).
     * @return {T} Element.
     */
    dequeue() {
      const elements = this.elements_;
      const priorities = this.priorities_;
      const element = elements[0];
      if (elements.length == 1) {
        elements.length = 0;
        priorities.length = 0;
      } else {
        elements[0] = elements.pop();
        priorities[0] = priorities.pop();
        this.siftUp_(0);
      }
      const elementKey = this.keyFunction_(element);
      delete this.queuedElements_[elementKey];
      return element;
    }

    /**
     * Enqueue an element. O(log N).
     * @param {T} element Element.
     * @return {boolean} The element was added to the queue.
     */
    enqueue(element) {
      assert(!(this.keyFunction_(element) in this.queuedElements_), 31); // Tried to enqueue an `element` that was already added to the queue
      const priority = this.priorityFunction_(element);
      if (priority != DROP) {
        this.elements_.push(element);
        this.priorities_.push(priority);
        this.queuedElements_[this.keyFunction_(element)] = true;
        this.siftDown_(0, this.elements_.length - 1);
        return true;
      }
      return false;
    }

    /**
     * @return {number} Count.
     */
    getCount() {
      return this.elements_.length;
    }

    /**
     * Gets the index of the left child of the node at the given index.
     * @param {number} index The index of the node to get the left child for.
     * @return {number} The index of the left child.
     * @private
     */
    getLeftChildIndex_(index) {
      return index * 2 + 1;
    }

    /**
     * Gets the index of the right child of the node at the given index.
     * @param {number} index The index of the node to get the right child for.
     * @return {number} The index of the right child.
     * @private
     */
    getRightChildIndex_(index) {
      return index * 2 + 2;
    }

    /**
     * Gets the index of the parent of the node at the given index.
     * @param {number} index The index of the node to get the parent for.
     * @return {number} The index of the parent.
     * @private
     */
    getParentIndex_(index) {
      return (index - 1) >> 1;
    }

    /**
     * Make this a heap. O(N).
     * @private
     */
    heapify_() {
      let i;
      for (i = (this.elements_.length >> 1) - 1; i >= 0; i--) {
        this.siftUp_(i);
      }
    }

    /**
     * @return {boolean} Is empty.
     */
    isEmpty() {
      return this.elements_.length === 0;
    }

    /**
     * @param {string} key Key.
     * @return {boolean} Is key queued.
     */
    isKeyQueued(key) {
      return key in this.queuedElements_;
    }

    /**
     * @param {T} element Element.
     * @return {boolean} Is queued.
     */
    isQueued(element) {
      return this.isKeyQueued(this.keyFunction_(element));
    }

    /**
     * @param {number} index The index of the node to move down.
     * @private
     */
    siftUp_(index) {
      const elements = this.elements_;
      const priorities = this.priorities_;
      const count = elements.length;
      const element = elements[index];
      const priority = priorities[index];
      const startIndex = index;

      while (index < count >> 1) {
        const lIndex = this.getLeftChildIndex_(index);
        const rIndex = this.getRightChildIndex_(index);

        const smallerChildIndex =
          rIndex < count && priorities[rIndex] < priorities[lIndex]
            ? rIndex
            : lIndex;

        elements[index] = elements[smallerChildIndex];
        priorities[index] = priorities[smallerChildIndex];
        index = smallerChildIndex;
      }

      elements[index] = element;
      priorities[index] = priority;
      this.siftDown_(startIndex, index);
    }

    /**
     * @param {number} startIndex The index of the root.
     * @param {number} index The index of the node to move up.
     * @private
     */
    siftDown_(startIndex, index) {
      const elements = this.elements_;
      const priorities = this.priorities_;
      const element = elements[index];
      const priority = priorities[index];

      while (index > startIndex) {
        const parentIndex = this.getParentIndex_(index);
        if (priorities[parentIndex] > priority) {
          elements[index] = elements[parentIndex];
          priorities[index] = priorities[parentIndex];
          index = parentIndex;
        } else {
          break;
        }
      }
      elements[index] = element;
      priorities[index] = priority;
    }

    /**
     * FIXME empty description for jsdoc
     */
    reprioritize() {
      const priorityFunction = this.priorityFunction_;
      const elements = this.elements_;
      const priorities = this.priorities_;
      let index = 0;
      const n = elements.length;
      let element, i, priority;
      for (i = 0; i < n; ++i) {
        element = elements[i];
        priority = priorityFunction(element);
        if (priority == DROP) {
          delete this.queuedElements_[this.keyFunction_(element)];
        } else {
          priorities[index] = priority;
          elements[index++] = element;
        }
      }
      elements.length = index;
      priorities.length = index;
      this.heapify_();
    }
  }

  /**
   * @module ol/TileQueue
   */

  /**
   * @typedef {function(import("./Tile.js").default, string, import("./coordinate.js").Coordinate, number): number} PriorityFunction
   */

  class TileQueue extends PriorityQueue {
    /**
     * @param {PriorityFunction} tilePriorityFunction Tile priority function.
     * @param {function(): ?} tileChangeCallback Function called on each tile change event.
     */
    constructor(tilePriorityFunction, tileChangeCallback) {
      super(
        /**
         * @param {Array} element Element.
         * @return {number} Priority.
         */
        function (element) {
          return tilePriorityFunction.apply(null, element);
        },
        /**
         * @param {Array} element Element.
         * @return {string} Key.
         */
        function (element) {
          return /** @type {import("./Tile.js").default} */ (element[0]).getKey();
        }
      );

      /** @private */
      this.boundHandleTileChange_ = this.handleTileChange.bind(this);

      /**
       * @private
       * @type {function(): ?}
       */
      this.tileChangeCallback_ = tileChangeCallback;

      /**
       * @private
       * @type {number}
       */
      this.tilesLoading_ = 0;

      /**
       * @private
       * @type {!Object<string,boolean>}
       */
      this.tilesLoadingKeys_ = {};
    }

    /**
     * @param {Array} element Element.
     * @return {boolean} The element was added to the queue.
     */
    enqueue(element) {
      const added = super.enqueue(element);
      if (added) {
        const tile = element[0];
        tile.addEventListener(EventType$1.CHANGE, this.boundHandleTileChange_);
      }
      return added;
    }

    /**
     * @return {number} Number of tiles loading.
     */
    getTilesLoading() {
      return this.tilesLoading_;
    }

    /**
     * @param {import("./events/Event.js").default} event Event.
     * @protected
     */
    handleTileChange(event) {
      const tile = /** @type {import("./Tile.js").default} */ (event.target);
      const state = tile.getState();
      if (
        state === TileState.LOADED ||
        state === TileState.ERROR ||
        state === TileState.EMPTY
      ) {
        if (state !== TileState.ERROR) {
          tile.removeEventListener(EventType$1.CHANGE, this.boundHandleTileChange_);
        }
        const tileKey = tile.getKey();
        if (tileKey in this.tilesLoadingKeys_) {
          delete this.tilesLoadingKeys_[tileKey];
          --this.tilesLoading_;
        }
        this.tileChangeCallback_();
      }
    }

    /**
     * @param {number} maxTotalLoading Maximum number tiles to load simultaneously.
     * @param {number} maxNewLoads Maximum number of new tiles to load.
     */
    loadMoreTiles(maxTotalLoading, maxNewLoads) {
      let newLoads = 0;
      let state, tile, tileKey;
      while (
        this.tilesLoading_ < maxTotalLoading &&
        newLoads < maxNewLoads &&
        this.getCount() > 0
      ) {
        tile = /** @type {import("./Tile.js").default} */ (this.dequeue()[0]);
        tileKey = tile.getKey();
        state = tile.getState();
        if (state === TileState.IDLE && !(tileKey in this.tilesLoadingKeys_)) {
          this.tilesLoadingKeys_[tileKey] = true;
          ++this.tilesLoading_;
          ++newLoads;
          tile.load();
        }
      }
    }
  }

  /**
   * @param {import('./Map.js').FrameState} frameState Frame state.
   * @param {import("./Tile.js").default} tile Tile.
   * @param {string} tileSourceKey Tile source key.
   * @param {import("./coordinate.js").Coordinate} tileCenter Tile center.
   * @param {number} tileResolution Tile resolution.
   * @return {number} Tile priority.
   */
  function getTilePriority(
    frameState,
    tile,
    tileSourceKey,
    tileCenter,
    tileResolution
  ) {
    // Filter out tiles at higher zoom levels than the current zoom level, or that
    // are outside the visible extent.
    if (!frameState || !(tileSourceKey in frameState.wantedTiles)) {
      return DROP;
    }
    if (!frameState.wantedTiles[tileSourceKey][tile.getKey()]) {
      return DROP;
    }
    // Prioritize the highest zoom level tiles closest to the focus.
    // Tiles at higher zoom levels are prioritized using Math.log(tileResolution).
    // Within a zoom level, tiles are prioritized by the distance in pixels between
    // the center of the tile and the center of the viewport.  The factor of 65536
    // means that the prioritization should behave as desired for tiles up to
    // 65536 * Math.log(2) = 45426 pixels from the focus.
    const center = frameState.viewState.center;
    const deltaX = tileCenter[0] - center[0];
    const deltaY = tileCenter[1] - center[1];
    return (
      65536 * Math.log(tileResolution) +
      Math.sqrt(deltaX * deltaX + deltaY * deltaY) / tileResolution
    );
  }

  /**
   * @module ol/ViewHint
   */

  /**
   * @enum {number}
   */
  var ViewHint = {
    ANIMATING: 0,
    INTERACTING: 1,
  };

  /**
   * @module ol/ViewProperty
   */

  /**
   * @enum {string}
   */
  var ViewProperty = {
    CENTER: 'center',
    RESOLUTION: 'resolution',
    ROTATION: 'rotation',
  };

  /**
   * @module ol/tilegrid/common
   */

  /**
   * Default maximum zoom for default tile grids.
   * @type {number}
   */
  const DEFAULT_MAX_ZOOM = 42;

  /**
   * Default tile size.
   * @type {number}
   */
  const DEFAULT_TILE_SIZE = 256;

  /**
   * @module ol/centerconstraint
   */

  /**
   * @typedef {function((import("./coordinate.js").Coordinate|undefined), number, import("./size.js").Size, boolean=, Array<number>=): (import("./coordinate.js").Coordinate|undefined)} Type
   */

  /**
   * @param {import("./extent.js").Extent} extent Extent.
   * @param {boolean} onlyCenter If true, the constraint will only apply to the view center.
   * @param {boolean} smooth If true, the view will be able to go slightly out of the given extent
   * (only during interaction and animation).
   * @return {Type} The constraint.
   */
  function createExtent(extent, onlyCenter, smooth) {
    return (
      /**
       * @param {import("./coordinate.js").Coordinate|undefined} center Center.
       * @param {number|undefined} resolution Resolution.
       * @param {import("./size.js").Size} size Viewport size; unused if `onlyCenter` was specified.
       * @param {boolean} [isMoving] True if an interaction or animation is in progress.
       * @param {Array<number>} [centerShift] Shift between map center and viewport center.
       * @return {import("./coordinate.js").Coordinate|undefined} Center.
       */
      function (center, resolution, size, isMoving, centerShift) {
        if (!center) {
          return undefined;
        }
        if (!resolution && !onlyCenter) {
          return center;
        }
        const viewWidth = onlyCenter ? 0 : size[0] * resolution;
        const viewHeight = onlyCenter ? 0 : size[1] * resolution;
        const shiftX = centerShift ? centerShift[0] : 0;
        const shiftY = centerShift ? centerShift[1] : 0;
        let minX = extent[0] + viewWidth / 2 + shiftX;
        let maxX = extent[2] - viewWidth / 2 + shiftX;
        let minY = extent[1] + viewHeight / 2 + shiftY;
        let maxY = extent[3] - viewHeight / 2 + shiftY;

        // note: when zooming out of bounds, min and max values for x and y may
        // end up inverted (min > max); this has to be accounted for
        if (minX > maxX) {
          minX = (maxX + minX) / 2;
          maxX = minX;
        }
        if (minY > maxY) {
          minY = (maxY + minY) / 2;
          maxY = minY;
        }

        let x = clamp(center[0], minX, maxX);
        let y = clamp(center[1], minY, maxY);

        // during an interaction, allow some overscroll
        if (isMoving && smooth && resolution) {
          const ratio = 30 * resolution;
          x +=
            -ratio * Math.log(1 + Math.max(0, minX - center[0]) / ratio) +
            ratio * Math.log(1 + Math.max(0, center[0] - maxX) / ratio);
          y +=
            -ratio * Math.log(1 + Math.max(0, minY - center[1]) / ratio) +
            ratio * Math.log(1 + Math.max(0, center[1] - maxY) / ratio);
        }

        return [x, y];
      }
    );
  }

  /**
   * @param {import("./coordinate.js").Coordinate} [center] Center.
   * @return {import("./coordinate.js").Coordinate|undefined} Center.
   */
  function none$1(center) {
    return center;
  }

  /**
   * @module ol/resolutionconstraint
   */

  /**
   * @typedef {function((number|undefined), number, import("./size.js").Size, boolean=): (number|undefined)} Type
   */

  /**
   * Returns a modified resolution taking into account the viewport size and maximum
   * allowed extent.
   * @param {number} resolution Resolution
   * @param {import("./extent.js").Extent} maxExtent Maximum allowed extent.
   * @param {import("./size.js").Size} viewportSize Viewport size.
   * @param {boolean} showFullExtent Whether to show the full extent.
   * @return {number} Capped resolution.
   */
  function getViewportClampedResolution(
    resolution,
    maxExtent,
    viewportSize,
    showFullExtent
  ) {
    const xResolution = getWidth(maxExtent) / viewportSize[0];
    const yResolution = getHeight(maxExtent) / viewportSize[1];

    if (showFullExtent) {
      return Math.min(resolution, Math.max(xResolution, yResolution));
    }
    return Math.min(resolution, Math.min(xResolution, yResolution));
  }

  /**
   * Returns a modified resolution to be between maxResolution and minResolution while
   * still allowing the value to be slightly out of bounds.
   * Note: the computation is based on the logarithm function (ln):
   *  - at 1, ln(x) is 0
   *  - above 1, ln(x) keeps increasing but at a much slower pace than x
   * The final result is clamped to prevent getting too far away from bounds.
   * @param {number} resolution Resolution.
   * @param {number} maxResolution Max resolution.
   * @param {number} minResolution Min resolution.
   * @return {number} Smoothed resolution.
   */
  function getSmoothClampedResolution(resolution, maxResolution, minResolution) {
    let result = Math.min(resolution, maxResolution);
    const ratio = 50;

    result *=
      Math.log(1 + ratio * Math.max(0, resolution / maxResolution - 1)) / ratio +
      1;
    if (minResolution) {
      result = Math.max(result, minResolution);
      result /=
        Math.log(1 + ratio * Math.max(0, minResolution / resolution - 1)) /
          ratio +
        1;
    }
    return clamp(result, minResolution / 2, maxResolution * 2);
  }

  /**
   * @param {Array<number>} resolutions Resolutions.
   * @param {boolean} [smooth] If true, the view will be able to slightly exceed resolution limits. Default: true.
   * @param {import("./extent.js").Extent} [maxExtent] Maximum allowed extent.
   * @param {boolean} [showFullExtent] If true, allows us to show the full extent. Default: false.
   * @return {Type} Zoom function.
   */
  function createSnapToResolutions(
    resolutions,
    smooth,
    maxExtent,
    showFullExtent
  ) {
    smooth = smooth !== undefined ? smooth : true;
    return (
      /**
       * @param {number|undefined} resolution Resolution.
       * @param {number} direction Direction.
       * @param {import("./size.js").Size} size Viewport size.
       * @param {boolean} [isMoving] True if an interaction or animation is in progress.
       * @return {number|undefined} Resolution.
       */
      function (resolution, direction, size, isMoving) {
        if (resolution !== undefined) {
          const maxResolution = resolutions[0];
          const minResolution = resolutions[resolutions.length - 1];
          const cappedMaxRes = maxExtent
            ? getViewportClampedResolution(
                maxResolution,
                maxExtent,
                size,
                showFullExtent
              )
            : maxResolution;

          // during interacting or animating, allow intermediary values
          if (isMoving) {
            if (!smooth) {
              return clamp(resolution, minResolution, cappedMaxRes);
            }
            return getSmoothClampedResolution(
              resolution,
              cappedMaxRes,
              minResolution
            );
          }

          const capped = Math.min(cappedMaxRes, resolution);
          const z = Math.floor(linearFindNearest(resolutions, capped, direction));
          if (resolutions[z] > cappedMaxRes && z < resolutions.length - 1) {
            return resolutions[z + 1];
          }
          return resolutions[z];
        } else {
          return undefined;
        }
      }
    );
  }

  /**
   * @param {number} power Power.
   * @param {number} maxResolution Maximum resolution.
   * @param {number} [minResolution] Minimum resolution.
   * @param {boolean} [smooth] If true, the view will be able to slightly exceed resolution limits. Default: true.
   * @param {import("./extent.js").Extent} [maxExtent] Maximum allowed extent.
   * @param {boolean} [showFullExtent] If true, allows us to show the full extent. Default: false.
   * @return {Type} Zoom function.
   */
  function createSnapToPower(
    power,
    maxResolution,
    minResolution,
    smooth,
    maxExtent,
    showFullExtent
  ) {
    smooth = smooth !== undefined ? smooth : true;
    minResolution = minResolution !== undefined ? minResolution : 0;

    return (
      /**
       * @param {number|undefined} resolution Resolution.
       * @param {number} direction Direction.
       * @param {import("./size.js").Size} size Viewport size.
       * @param {boolean} [isMoving] True if an interaction or animation is in progress.
       * @return {number|undefined} Resolution.
       */
      function (resolution, direction, size, isMoving) {
        if (resolution !== undefined) {
          const cappedMaxRes = maxExtent
            ? getViewportClampedResolution(
                maxResolution,
                maxExtent,
                size,
                showFullExtent
              )
            : maxResolution;

          // during interacting or animating, allow intermediary values
          if (isMoving) {
            if (!smooth) {
              return clamp(resolution, minResolution, cappedMaxRes);
            }
            return getSmoothClampedResolution(
              resolution,
              cappedMaxRes,
              minResolution
            );
          }

          const tolerance = 1e-9;
          const minZoomLevel = Math.ceil(
            Math.log(maxResolution / cappedMaxRes) / Math.log(power) - tolerance
          );
          const offset = -direction * (0.5 - tolerance) + 0.5;
          const capped = Math.min(cappedMaxRes, resolution);
          const cappedZoomLevel = Math.floor(
            Math.log(maxResolution / capped) / Math.log(power) + offset
          );
          const zoomLevel = Math.max(minZoomLevel, cappedZoomLevel);
          const newResolution = maxResolution / Math.pow(power, zoomLevel);
          return clamp(newResolution, minResolution, cappedMaxRes);
        } else {
          return undefined;
        }
      }
    );
  }

  /**
   * @param {number} maxResolution Max resolution.
   * @param {number} minResolution Min resolution.
   * @param {boolean} [smooth] If true, the view will be able to slightly exceed resolution limits. Default: true.
   * @param {import("./extent.js").Extent} [maxExtent] Maximum allowed extent.
   * @param {boolean} [showFullExtent] If true, allows us to show the full extent. Default: false.
   * @return {Type} Zoom function.
   */
  function createMinMaxResolution(
    maxResolution,
    minResolution,
    smooth,
    maxExtent,
    showFullExtent
  ) {
    smooth = smooth !== undefined ? smooth : true;

    return (
      /**
       * @param {number|undefined} resolution Resolution.
       * @param {number} direction Direction.
       * @param {import("./size.js").Size} size Viewport size.
       * @param {boolean} [isMoving] True if an interaction or animation is in progress.
       * @return {number|undefined} Resolution.
       */
      function (resolution, direction, size, isMoving) {
        if (resolution !== undefined) {
          const cappedMaxRes = maxExtent
            ? getViewportClampedResolution(
                maxResolution,
                maxExtent,
                size,
                showFullExtent
              )
            : maxResolution;

          if (!smooth || !isMoving) {
            return clamp(resolution, minResolution, cappedMaxRes);
          }
          return getSmoothClampedResolution(
            resolution,
            cappedMaxRes,
            minResolution
          );
        } else {
          return undefined;
        }
      }
    );
  }

  /**
   * @module ol/rotationconstraint
   */

  /**
   * @typedef {function((number|undefined), boolean=): (number|undefined)} Type
   */

  /**
   * @param {number|undefined} rotation Rotation.
   * @return {number|undefined} Rotation.
   */
  function disable(rotation) {
    if (rotation !== undefined) {
      return 0;
    } else {
      return undefined;
    }
  }

  /**
   * @param {number|undefined} rotation Rotation.
   * @return {number|undefined} Rotation.
   */
  function none(rotation) {
    if (rotation !== undefined) {
      return rotation;
    } else {
      return undefined;
    }
  }

  /**
   * @param {number} n N.
   * @return {Type} Rotation constraint.
   */
  function createSnapToN(n) {
    const theta = (2 * Math.PI) / n;
    return (
      /**
       * @param {number|undefined} rotation Rotation.
       * @param {boolean} [isMoving] True if an interaction or animation is in progress.
       * @return {number|undefined} Rotation.
       */
      function (rotation, isMoving) {
        if (isMoving) {
          return rotation;
        }

        if (rotation !== undefined) {
          rotation = Math.floor(rotation / theta + 0.5) * theta;
          return rotation;
        } else {
          return undefined;
        }
      }
    );
  }

  /**
   * @param {number} [tolerance] Tolerance.
   * @return {Type} Rotation constraint.
   */
  function createSnapToZero(tolerance) {
    tolerance = tolerance || toRadians(5);
    return (
      /**
       * @param {number|undefined} rotation Rotation.
       * @param {boolean} [isMoving] True if an interaction or animation is in progress.
       * @return {number|undefined} Rotation.
       */
      function (rotation, isMoving) {
        if (isMoving) {
          return rotation;
        }

        if (rotation !== undefined) {
          if (Math.abs(rotation) <= tolerance) {
            return 0;
          } else {
            return rotation;
          }
        } else {
          return undefined;
        }
      }
    );
  }

  /**
   * @module ol/View
   */

  /**
   * An animation configuration
   *
   * @typedef {Object} Animation
   * @property {import("./coordinate.js").Coordinate} [sourceCenter] Source center.
   * @property {import("./coordinate.js").Coordinate} [targetCenter] Target center.
   * @property {number} [sourceResolution] Source resolution.
   * @property {number} [targetResolution] Target resolution.
   * @property {number} [sourceRotation] Source rotation.
   * @property {number} [targetRotation] Target rotation.
   * @property {import("./coordinate.js").Coordinate} [anchor] Anchor.
   * @property {number} start Start.
   * @property {number} duration Duration.
   * @property {boolean} complete Complete.
   * @property {function(number):number} easing Easing.
   * @property {function(boolean):void} callback Callback.
   */

  /**
   * @typedef {Object} Constraints
   * @property {import("./centerconstraint.js").Type} center Center.
   * @property {import("./resolutionconstraint.js").Type} resolution Resolution.
   * @property {import("./rotationconstraint.js").Type} rotation Rotation.
   */

  /**
   * @typedef {Object} FitOptions
   * @property {import("./size.js").Size} [size] The size in pixels of the box to fit
   * the extent into. Default is the current size of the first map in the DOM that
   * uses this view, or `[100, 100]` if no such map is found.
   * @property {!Array<number>} [padding=[0, 0, 0, 0]] Padding (in pixels) to be
   * cleared inside the view. Values in the array are top, right, bottom and left
   * padding.
   * @property {boolean} [nearest=false] If the view `constrainResolution` option is `true`,
   * get the nearest extent instead of the closest that actually fits the view.
   * @property {number} [minResolution=0] Minimum resolution that we zoom to.
   * @property {number} [maxZoom] Maximum zoom level that we zoom to. If
   * `minResolution` is given, this property is ignored.
   * @property {number} [duration] The duration of the animation in milliseconds.
   * By default, there is no animation to the target extent.
   * @property {function(number):number} [easing] The easing function used during
   * the animation (defaults to {@link module:ol/easing.inAndOut}).
   * The function will be called for each frame with a number representing a
   * fraction of the animation's duration.  The function should return a number
   * between 0 and 1 representing the progress toward the destination state.
   * @property {function(boolean):void} [callback] Function called when the view is in
   * its final position. The callback will be called with `true` if the animation
   * series completed on its own or `false` if it was cancelled.
   */

  /**
   * @typedef {Object} ViewOptions
   * @property {import("./coordinate.js").Coordinate} [center] The initial center for
   * the view. If a user projection is not set, the coordinate system for the center is
   * specified with the `projection` option. Layer sources will not be fetched if this
   * is not set, but the center can be set later with {@link #setCenter}.
   * @property {boolean|number} [constrainRotation=true] Rotation constraint.
   * `false` means no constraint. `true` means no constraint, but snap to zero
   * near zero. A number constrains the rotation to that number of values. For
   * example, `4` will constrain the rotation to 0, 90, 180, and 270 degrees.
   * @property {boolean} [enableRotation=true] Enable rotation.
   * If `false`, a rotation constraint that always sets the rotation to zero is
   * used. The `constrainRotation` option has no effect if `enableRotation` is
   * `false`.
   * @property {import("./extent.js").Extent} [extent] The extent that constrains the
   * view, in other words, nothing outside of this extent can be visible on the map.
   * @property {boolean} [constrainOnlyCenter=false] If true, the extent
   * constraint will only apply to the view center and not the whole extent.
   * @property {boolean} [smoothExtentConstraint=true] If true, the extent
   * constraint will be applied smoothly, i.e. allow the view to go slightly outside
   * of the given `extent`.
   * @property {number} [maxResolution] The maximum resolution used to determine
   * the resolution constraint. It is used together with `minResolution` (or
   * `maxZoom`) and `zoomFactor`. If unspecified it is calculated in such a way
   * that the projection's validity extent fits in a 256x256 px tile. If the
   * projection is Spherical Mercator (the default) then `maxResolution` defaults
   * to `40075016.68557849 / 256 = 156543.03392804097`.
   * @property {number} [minResolution] The minimum resolution used to determine
   * the resolution constraint.  It is used together with `maxResolution` (or
   * `minZoom`) and `zoomFactor`.  If unspecified it is calculated assuming 29
   * zoom levels (with a factor of 2). If the projection is Spherical Mercator
   * (the default) then `minResolution` defaults to
   * `40075016.68557849 / 256 / Math.pow(2, 28) = 0.0005831682455839253`.
   * @property {number} [maxZoom=28] The maximum zoom level used to determine the
   * resolution constraint. It is used together with `minZoom` (or
   * `maxResolution`) and `zoomFactor`.  Note that if `minResolution` is also
   * provided, it is given precedence over `maxZoom`.
   * @property {number} [minZoom=0] The minimum zoom level used to determine the
   * resolution constraint. It is used together with `maxZoom` (or
   * `minResolution`) and `zoomFactor`.  Note that if `maxResolution` is also
   * provided, it is given precedence over `minZoom`.
   * @property {boolean} [multiWorld=false] If `false` the view is constrained so
   * only one world is visible, and you cannot pan off the edge.  If `true` the map
   * may show multiple worlds at low zoom levels.  Only used if the `projection` is
   * global.  Note that if `extent` is also provided it is given precedence.
   * @property {boolean} [constrainResolution=false] If true, the view will always
   * animate to the closest zoom level after an interaction; false means
   * intermediary zoom levels are allowed.
   * @property {boolean} [smoothResolutionConstraint=true] If true, the resolution
   * min/max values will be applied smoothly, i. e. allow the view to exceed slightly
   * the given resolution or zoom bounds.
   * @property {boolean} [showFullExtent=false] Allow the view to be zoomed out to
   * show the full configured extent. By default, when a view is configured with an
   * extent, users will not be able to zoom out so the viewport exceeds the extent in
   * either dimension. This means the full extent may not be visible if the viewport
   * is taller or wider than the aspect ratio of the configured extent. If
   * showFullExtent is true, the user will be able to zoom out so that the viewport
   * exceeds the height or width of the configured extent, but not both, allowing the
   * full extent to be shown.
   * @property {import("./proj.js").ProjectionLike} [projection='EPSG:3857'] The
   * projection. The default is Spherical Mercator.
   * @property {number} [resolution] The initial resolution for the view. The
   * units are `projection` units per pixel (e.g. meters per pixel). An
   * alternative to setting this is to set `zoom`. Layer sources will not be
   * fetched if neither this nor `zoom` are defined, but they can be set later
   * with {@link #setZoom} or {@link #setResolution}.
   * @property {Array<number>} [resolutions] Resolutions that determine the
   * zoom levels if specified. The index in the array corresponds to the zoom level,
   * therefore the resolution values have to be in descending order. It also constrains
   * the resolution by the minimum and maximum value. If set the `maxResolution`,
   * `minResolution`, `minZoom`, `maxZoom`, and `zoomFactor` options are ignored.
   * @property {number} [rotation=0] The initial rotation for the view in radians
   * (positive rotation clockwise, 0 means North).
   * @property {number} [zoom] Only used if `resolution` is not defined. Zoom
   * level used to calculate the initial resolution for the view.
   * @property {number} [zoomFactor=2] The zoom factor used to compute the
   * corresponding resolution.
   * @property {!Array<number>} [padding=[0, 0, 0, 0]] Padding (in css pixels).
   * If the map viewport is partially covered with other content (overlays) along
   * its edges, this setting allows to shift the center of the viewport away from
   * that content. The order of the values is top, right, bottom, left.
   */

  /**
   * @typedef {Object} AnimationOptions
   * @property {import("./coordinate.js").Coordinate} [center] The center of the view at the end of
   * the animation.
   * @property {number} [zoom] The zoom level of the view at the end of the
   * animation. This takes precedence over `resolution`.
   * @property {number} [resolution] The resolution of the view at the end
   * of the animation.  If `zoom` is also provided, this option will be ignored.
   * @property {number} [rotation] The rotation of the view at the end of
   * the animation.
   * @property {import("./coordinate.js").Coordinate} [anchor] Optional anchor to remain fixed
   * during a rotation or resolution animation.
   * @property {number} [duration=1000] The duration of the animation in milliseconds.
   * @property {function(number):number} [easing] The easing function used
   * during the animation (defaults to {@link module:ol/easing.inAndOut}).
   * The function will be called for each frame with a number representing a
   * fraction of the animation's duration.  The function should return a number
   * between 0 and 1 representing the progress toward the destination state.
   */

  /**
   * @typedef {Object} State
   * @property {import("./coordinate.js").Coordinate} center Center.
   * @property {import("./proj/Projection.js").default} projection Projection.
   * @property {number} resolution Resolution.
   * @property {import("./coordinate.js").Coordinate} [nextCenter] The next center during an animation series.
   * @property {number} [nextResolution] The next resolution during an animation series.
   * @property {number} [nextRotation] The next rotation during an animation series.
   * @property {number} rotation Rotation.
   * @property {number} zoom Zoom.
   */

  /**
   * Default min zoom level for the map view.
   * @type {number}
   */
  const DEFAULT_MIN_ZOOM = 0;

  /**
   * @typedef {import("./ObjectEventType").Types|'change:center'|'change:resolution'|'change:rotation'} ViewObjectEventTypes
   */

  /***
   * @template Return
   * @typedef {import("./Observable").OnSignature<import("./Observable").EventTypes, import("./events/Event.js").default, Return> &
   *   import("./Observable").OnSignature<ViewObjectEventTypes, import("./Object").ObjectEvent, Return> &
   *   import("./Observable").CombinedOnSignature<import("./Observable").EventTypes|ViewObjectEventTypes, Return>} ViewOnSignature
   */

  /**
   * @classdesc
   * A View object represents a simple 2D view of the map.
   *
   * This is the object to act upon to change the center, resolution,
   * and rotation of the map.
   *
   * A View has a `projection`. The projection determines the
   * coordinate system of the center, and its units determine the units of the
   * resolution (projection units per pixel). The default projection is
   * Web Mercator (EPSG:3857).
   *
   * ### The view states
   *
   * A View is determined by three states: `center`, `resolution`,
   * and `rotation`. Each state has a corresponding getter and setter, e.g.
   * `getCenter` and `setCenter` for the `center` state.
   *
   * The `zoom` state is actually not saved on the view: all computations
   * internally use the `resolution` state. Still, the `setZoom` and `getZoom`
   * methods are available, as well as `getResolutionForZoom` and
   * `getZoomForResolution` to switch from one system to the other.
   *
   * ### The constraints
   *
   * `setCenter`, `setResolution` and `setRotation` can be used to change the
   * states of the view, but any constraint defined in the constructor will
   * be applied along the way.
   *
   * A View object can have a *resolution constraint*, a *rotation constraint*
   * and a *center constraint*.
   *
   * The *resolution constraint* typically restricts min/max values and
   * snaps to specific resolutions. It is determined by the following
   * options: `resolutions`, `maxResolution`, `maxZoom` and `zoomFactor`.
   * If `resolutions` is set, the other three options are ignored. See
   * documentation for each option for more information. By default, the view
   * only has a min/max restriction and allow intermediary zoom levels when
   * pinch-zooming for example.
   *
   * The *rotation constraint* snaps to specific angles. It is determined
   * by the following options: `enableRotation` and `constrainRotation`.
   * By default rotation is allowed and its value is snapped to zero when approaching the
   * horizontal.
   *
   * The *center constraint* is determined by the `extent` option. By
   * default the view center is not constrained at all.
   *
   * ### Changing the view state
   *
   * It is important to note that `setZoom`, `setResolution`, `setCenter` and
   * `setRotation` are subject to the above mentioned constraints. As such, it
   * may sometimes not be possible to know in advance the resulting state of the
   * View. For example, calling `setResolution(10)` does not guarantee that
   * `getResolution()` will return `10`.
   *
   * A consequence of this is that, when applying a delta on the view state, one
   * should use `adjustCenter`, `adjustRotation`, `adjustZoom` and `adjustResolution`
   * rather than the corresponding setters. This will let view do its internal
   * computations. Besides, the `adjust*` methods also take an `anchor`
   * argument which allows specifying an origin for the transformation.
   *
   * ### Interacting with the view
   *
   * View constraints are usually only applied when the view is *at rest*, meaning that
   * no interaction or animation is ongoing. As such, if the user puts the view in a
   * state that is not equivalent to a constrained one (e.g. rotating the view when
   * the snap angle is 0), an animation will be triggered at the interaction end to
   * put back the view to a stable state;
   *
   * @api
   */
  class View extends BaseObject {
    /**
     * @param {ViewOptions} [options] View options.
     */
    constructor(options) {
      super();

      /***
       * @type {ViewOnSignature<import("./events").EventsKey>}
       */
      this.on;

      /***
       * @type {ViewOnSignature<import("./events").EventsKey>}
       */
      this.once;

      /***
       * @type {ViewOnSignature<void>}
       */
      this.un;

      options = Object.assign({}, options);

      /**
       * @private
       * @type {Array<number>}
       */
      this.hints_ = [0, 0];

      /**
       * @private
       * @type {Array<Array<Animation>>}
       */
      this.animations_ = [];

      /**
       * @private
       * @type {number|undefined}
       */
      this.updateAnimationKey_;

      /**
       * @private
       * @const
       * @type {import("./proj/Projection.js").default}
       */
      this.projection_ = createProjection(options.projection, 'EPSG:3857');

      /**
       * @private
       * @type {import("./size.js").Size}
       */
      this.viewportSize_ = [100, 100];

      /**
       * @private
       * @type {import("./coordinate.js").Coordinate|undefined}
       */
      this.targetCenter_ = null;

      /**
       * @private
       * @type {number|undefined}
       */
      this.targetResolution_;

      /**
       * @private
       * @type {number|undefined}
       */
      this.targetRotation_;

      /**
       * @private
       * @type {import("./coordinate.js").Coordinate}
       */
      this.nextCenter_ = null;

      /**
       * @private
       * @type {number}
       */
      this.nextResolution_;

      /**
       * @private
       * @type {number}
       */
      this.nextRotation_;

      /**
       * @private
       * @type {import("./coordinate.js").Coordinate|undefined}
       */
      this.cancelAnchor_ = undefined;

      if (options.projection) {
        disableCoordinateWarning();
      }
      if (options.center) {
        options.center = fromUserCoordinate(options.center, this.projection_);
      }
      if (options.extent) {
        options.extent = fromUserExtent(options.extent, this.projection_);
      }

      this.applyOptions_(options);
    }

    /**
     * Set up the view with the given options.
     * @param {ViewOptions} options View options.
     */
    applyOptions_(options) {
      const properties = Object.assign({}, options);
      for (const key in ViewProperty) {
        delete properties[key];
      }
      this.setProperties(properties, true);

      const resolutionConstraintInfo = createResolutionConstraint(options);

      /**
       * @private
       * @type {number}
       */
      this.maxResolution_ = resolutionConstraintInfo.maxResolution;

      /**
       * @private
       * @type {number}
       */
      this.minResolution_ = resolutionConstraintInfo.minResolution;

      /**
       * @private
       * @type {number}
       */
      this.zoomFactor_ = resolutionConstraintInfo.zoomFactor;

      /**
       * @private
       * @type {Array<number>|undefined}
       */
      this.resolutions_ = options.resolutions;

      /**
       * @type {Array<number>|undefined}
       * @private
       */
      this.padding_ = options.padding;

      /**
       * @private
       * @type {number}
       */
      this.minZoom_ = resolutionConstraintInfo.minZoom;

      const centerConstraint = createCenterConstraint(options);
      const resolutionConstraint = resolutionConstraintInfo.constraint;
      const rotationConstraint = createRotationConstraint(options);

      /**
       * @private
       * @type {Constraints}
       */
      this.constraints_ = {
        center: centerConstraint,
        resolution: resolutionConstraint,
        rotation: rotationConstraint,
      };

      this.setRotation(options.rotation !== undefined ? options.rotation : 0);
      this.setCenterInternal(
        options.center !== undefined ? options.center : null
      );
      if (options.resolution !== undefined) {
        this.setResolution(options.resolution);
      } else if (options.zoom !== undefined) {
        this.setZoom(options.zoom);
      }
    }

    /**
     * Padding (in css pixels).
     * If the map viewport is partially covered with other content (overlays) along
     * its edges, this setting allows to shift the center of the viewport away from that
     * content. The order of the values in the array is top, right, bottom, left.
     * The default is no padding, which is equivalent to `[0, 0, 0, 0]`.
     * @type {Array<number>|undefined}
     * @api
     */
    get padding() {
      return this.padding_;
    }
    set padding(padding) {
      let oldPadding = this.padding_;
      this.padding_ = padding;
      const center = this.getCenter();
      if (center) {
        const newPadding = padding || [0, 0, 0, 0];
        oldPadding = oldPadding || [0, 0, 0, 0];
        const resolution = this.getResolution();
        const offsetX =
          (resolution / 2) *
          (newPadding[3] - oldPadding[3] + oldPadding[1] - newPadding[1]);
        const offsetY =
          (resolution / 2) *
          (newPadding[0] - oldPadding[0] + oldPadding[2] - newPadding[2]);
        this.setCenterInternal([center[0] + offsetX, center[1] - offsetY]);
      }
    }

    /**
     * Get an updated version of the view options used to construct the view.  The
     * current resolution (or zoom), center, and rotation are applied to any stored
     * options.  The provided options can be used to apply new min/max zoom or
     * resolution limits.
     * @param {ViewOptions} newOptions New options to be applied.
     * @return {ViewOptions} New options updated with the current view state.
     */
    getUpdatedOptions_(newOptions) {
      const options = this.getProperties();

      // preserve resolution (or zoom)
      if (options.resolution !== undefined) {
        options.resolution = this.getResolution();
      } else {
        options.zoom = this.getZoom();
      }

      // preserve center
      options.center = this.getCenterInternal();

      // preserve rotation
      options.rotation = this.getRotation();

      return Object.assign({}, options, newOptions);
    }

    /**
     * Animate the view.  The view's center, zoom (or resolution), and rotation
     * can be animated for smooth transitions between view states.  For example,
     * to animate the view to a new zoom level:
     *
     *     view.animate({zoom: view.getZoom() + 1});
     *
     * By default, the animation lasts one second and uses in-and-out easing.  You
     * can customize this behavior by including `duration` (in milliseconds) and
     * `easing` options (see {@link module:ol/easing}).
     *
     * To chain together multiple animations, call the method with multiple
     * animation objects.  For example, to first zoom and then pan:
     *
     *     view.animate({zoom: 10}, {center: [0, 0]});
     *
     * If you provide a function as the last argument to the animate method, it
     * will get called at the end of an animation series.  The callback will be
     * called with `true` if the animation series completed on its own or `false`
     * if it was cancelled.
     *
     * Animations are cancelled by user interactions (e.g. dragging the map) or by
     * calling `view.setCenter()`, `view.setResolution()`, or `view.setRotation()`
     * (or another method that calls one of these).
     *
     * @param {...(AnimationOptions|function(boolean): void)} var_args Animation
     *     options.  Multiple animations can be run in series by passing multiple
     *     options objects.  To run multiple animations in parallel, call the method
     *     multiple times.  An optional callback can be provided as a final
     *     argument.  The callback will be called with a boolean indicating whether
     *     the animation completed without being cancelled.
     * @api
     */
    animate(var_args) {
      if (this.isDef() && !this.getAnimating()) {
        this.resolveConstraints(0);
      }
      const args = new Array(arguments.length);
      for (let i = 0; i < args.length; ++i) {
        let options = arguments[i];
        if (options.center) {
          options = Object.assign({}, options);
          options.center = fromUserCoordinate(
            options.center,
            this.getProjection()
          );
        }
        if (options.anchor) {
          options = Object.assign({}, options);
          options.anchor = fromUserCoordinate(
            options.anchor,
            this.getProjection()
          );
        }
        args[i] = options;
      }
      this.animateInternal.apply(this, args);
    }

    /**
     * @param {...(AnimationOptions|function(boolean): void)} var_args Animation options.
     */
    animateInternal(var_args) {
      let animationCount = arguments.length;
      let callback;
      if (
        animationCount > 1 &&
        typeof arguments[animationCount - 1] === 'function'
      ) {
        callback = arguments[animationCount - 1];
        --animationCount;
      }

      let i = 0;
      for (; i < animationCount && !this.isDef(); ++i) {
        // if view properties are not yet set, shortcut to the final state
        const state = arguments[i];
        if (state.center) {
          this.setCenterInternal(state.center);
        }
        if (state.zoom !== undefined) {
          this.setZoom(state.zoom);
        } else if (state.resolution) {
          this.setResolution(state.resolution);
        }
        if (state.rotation !== undefined) {
          this.setRotation(state.rotation);
        }
      }
      if (i === animationCount) {
        if (callback) {
          animationCallback(callback, true);
        }
        return;
      }

      let start = Date.now();
      let center = this.targetCenter_.slice();
      let resolution = this.targetResolution_;
      let rotation = this.targetRotation_;
      const series = [];
      for (; i < animationCount; ++i) {
        const options = /** @type {AnimationOptions} */ (arguments[i]);

        const animation = {
          start: start,
          complete: false,
          anchor: options.anchor,
          duration: options.duration !== undefined ? options.duration : 1000,
          easing: options.easing || inAndOut,
          callback: callback,
        };

        if (options.center) {
          animation.sourceCenter = center;
          animation.targetCenter = options.center.slice();
          center = animation.targetCenter;
        }

        if (options.zoom !== undefined) {
          animation.sourceResolution = resolution;
          animation.targetResolution = this.getResolutionForZoom(options.zoom);
          resolution = animation.targetResolution;
        } else if (options.resolution) {
          animation.sourceResolution = resolution;
          animation.targetResolution = options.resolution;
          resolution = animation.targetResolution;
        }

        if (options.rotation !== undefined) {
          animation.sourceRotation = rotation;
          const delta =
            modulo(options.rotation - rotation + Math.PI, 2 * Math.PI) - Math.PI;
          animation.targetRotation = rotation + delta;
          rotation = animation.targetRotation;
        }

        // check if animation is a no-op
        if (isNoopAnimation(animation)) {
          animation.complete = true;
          // we still push it onto the series for callback handling
        } else {
          start += animation.duration;
        }
        series.push(animation);
      }
      this.animations_.push(series);
      this.setHint(ViewHint.ANIMATING, 1);
      this.updateAnimations_();
    }

    /**
     * Determine if the view is being animated.
     * @return {boolean} The view is being animated.
     * @api
     */
    getAnimating() {
      return this.hints_[ViewHint.ANIMATING] > 0;
    }

    /**
     * Determine if the user is interacting with the view, such as panning or zooming.
     * @return {boolean} The view is being interacted with.
     * @api
     */
    getInteracting() {
      return this.hints_[ViewHint.INTERACTING] > 0;
    }

    /**
     * Cancel any ongoing animations.
     * @api
     */
    cancelAnimations() {
      this.setHint(ViewHint.ANIMATING, -this.hints_[ViewHint.ANIMATING]);
      let anchor;
      for (let i = 0, ii = this.animations_.length; i < ii; ++i) {
        const series = this.animations_[i];
        if (series[0].callback) {
          animationCallback(series[0].callback, false);
        }
        if (!anchor) {
          for (let j = 0, jj = series.length; j < jj; ++j) {
            const animation = series[j];
            if (!animation.complete) {
              anchor = animation.anchor;
              break;
            }
          }
        }
      }
      this.animations_.length = 0;
      this.cancelAnchor_ = anchor;
      this.nextCenter_ = null;
      this.nextResolution_ = NaN;
      this.nextRotation_ = NaN;
    }

    /**
     * Update all animations.
     */
    updateAnimations_() {
      if (this.updateAnimationKey_ !== undefined) {
        cancelAnimationFrame(this.updateAnimationKey_);
        this.updateAnimationKey_ = undefined;
      }
      if (!this.getAnimating()) {
        return;
      }
      const now = Date.now();
      let more = false;
      for (let i = this.animations_.length - 1; i >= 0; --i) {
        const series = this.animations_[i];
        let seriesComplete = true;
        for (let j = 0, jj = series.length; j < jj; ++j) {
          const animation = series[j];
          if (animation.complete) {
            continue;
          }
          const elapsed = now - animation.start;
          let fraction =
            animation.duration > 0 ? elapsed / animation.duration : 1;
          if (fraction >= 1) {
            animation.complete = true;
            fraction = 1;
          } else {
            seriesComplete = false;
          }
          const progress = animation.easing(fraction);
          if (animation.sourceCenter) {
            const x0 = animation.sourceCenter[0];
            const y0 = animation.sourceCenter[1];
            const x1 = animation.targetCenter[0];
            const y1 = animation.targetCenter[1];
            this.nextCenter_ = animation.targetCenter;
            const x = x0 + progress * (x1 - x0);
            const y = y0 + progress * (y1 - y0);
            this.targetCenter_ = [x, y];
          }
          if (animation.sourceResolution && animation.targetResolution) {
            const resolution =
              progress === 1
                ? animation.targetResolution
                : animation.sourceResolution +
                  progress *
                    (animation.targetResolution - animation.sourceResolution);
            if (animation.anchor) {
              const size = this.getViewportSize_(this.getRotation());
              const constrainedResolution = this.constraints_.resolution(
                resolution,
                0,
                size,
                true
              );
              this.targetCenter_ = this.calculateCenterZoom(
                constrainedResolution,
                animation.anchor
              );
            }
            this.nextResolution_ = animation.targetResolution;
            this.targetResolution_ = resolution;
            this.applyTargetState_(true);
          }
          if (
            animation.sourceRotation !== undefined &&
            animation.targetRotation !== undefined
          ) {
            const rotation =
              progress === 1
                ? modulo(animation.targetRotation + Math.PI, 2 * Math.PI) -
                  Math.PI
                : animation.sourceRotation +
                  progress *
                    (animation.targetRotation - animation.sourceRotation);
            if (animation.anchor) {
              const constrainedRotation = this.constraints_.rotation(
                rotation,
                true
              );
              this.targetCenter_ = this.calculateCenterRotate(
                constrainedRotation,
                animation.anchor
              );
            }
            this.nextRotation_ = animation.targetRotation;
            this.targetRotation_ = rotation;
          }
          this.applyTargetState_(true);
          more = true;
          if (!animation.complete) {
            break;
          }
        }
        if (seriesComplete) {
          this.animations_[i] = null;
          this.setHint(ViewHint.ANIMATING, -1);
          this.nextCenter_ = null;
          this.nextResolution_ = NaN;
          this.nextRotation_ = NaN;
          const callback = series[0].callback;
          if (callback) {
            animationCallback(callback, true);
          }
        }
      }
      // prune completed series
      this.animations_ = this.animations_.filter(Boolean);
      if (more && this.updateAnimationKey_ === undefined) {
        this.updateAnimationKey_ = requestAnimationFrame(
          this.updateAnimations_.bind(this)
        );
      }
    }

    /**
     * @param {number} rotation Target rotation.
     * @param {import("./coordinate.js").Coordinate} anchor Rotation anchor.
     * @return {import("./coordinate.js").Coordinate|undefined} Center for rotation and anchor.
     */
    calculateCenterRotate(rotation, anchor) {
      let center;
      const currentCenter = this.getCenterInternal();
      if (currentCenter !== undefined) {
        center = [currentCenter[0] - anchor[0], currentCenter[1] - anchor[1]];
        rotate$1(center, rotation - this.getRotation());
        add(center, anchor);
      }
      return center;
    }

    /**
     * @param {number} resolution Target resolution.
     * @param {import("./coordinate.js").Coordinate} anchor Zoom anchor.
     * @return {import("./coordinate.js").Coordinate|undefined} Center for resolution and anchor.
     */
    calculateCenterZoom(resolution, anchor) {
      let center;
      const currentCenter = this.getCenterInternal();
      const currentResolution = this.getResolution();
      if (currentCenter !== undefined && currentResolution !== undefined) {
        const x =
          anchor[0] -
          (resolution * (anchor[0] - currentCenter[0])) / currentResolution;
        const y =
          anchor[1] -
          (resolution * (anchor[1] - currentCenter[1])) / currentResolution;
        center = [x, y];
      }
      return center;
    }

    /**
     * Returns the current viewport size.
     * @private
     * @param {number} [rotation] Take into account the rotation of the viewport when giving the size
     * @return {import("./size.js").Size} Viewport size or `[100, 100]` when no viewport is found.
     */
    getViewportSize_(rotation) {
      const size = this.viewportSize_;
      if (rotation) {
        const w = size[0];
        const h = size[1];
        return [
          Math.abs(w * Math.cos(rotation)) + Math.abs(h * Math.sin(rotation)),
          Math.abs(w * Math.sin(rotation)) + Math.abs(h * Math.cos(rotation)),
        ];
      } else {
        return size;
      }
    }

    /**
     * Stores the viewport size on the view. The viewport size is not read every time from the DOM
     * to avoid performance hit and layout reflow.
     * This should be done on map size change.
     * Note: the constraints are not resolved during an animation to avoid stopping it
     * @param {import("./size.js").Size} [size] Viewport size; if undefined, [100, 100] is assumed
     */
    setViewportSize(size) {
      this.viewportSize_ = Array.isArray(size) ? size.slice() : [100, 100];
      if (!this.getAnimating()) {
        this.resolveConstraints(0);
      }
    }

    /**
     * Get the view center.
     * @return {import("./coordinate.js").Coordinate|undefined} The center of the view.
     * @observable
     * @api
     */
    getCenter() {
      const center = this.getCenterInternal();
      if (!center) {
        return center;
      }
      return toUserCoordinate(center, this.getProjection());
    }

    /**
     * Get the view center without transforming to user projection.
     * @return {import("./coordinate.js").Coordinate|undefined} The center of the view.
     */
    getCenterInternal() {
      return /** @type {import("./coordinate.js").Coordinate|undefined} */ (
        this.get(ViewProperty.CENTER)
      );
    }

    /**
     * @return {Constraints} Constraints.
     */
    getConstraints() {
      return this.constraints_;
    }

    /**
     * @return {boolean} Resolution constraint is set
     */
    getConstrainResolution() {
      return this.get('constrainResolution');
    }

    /**
     * @param {Array<number>} [hints] Destination array.
     * @return {Array<number>} Hint.
     */
    getHints(hints) {
      if (hints !== undefined) {
        hints[0] = this.hints_[0];
        hints[1] = this.hints_[1];
        return hints;
      } else {
        return this.hints_.slice();
      }
    }

    /**
     * Calculate the extent for the current view state and the passed size.
     * The size is the pixel dimensions of the box into which the calculated extent
     * should fit. In most cases you want to get the extent of the entire map,
     * that is `map.getSize()`.
     * @param {import("./size.js").Size} [size] Box pixel size. If not provided, the size
     * of the map that uses this view will be used.
     * @return {import("./extent.js").Extent} Extent.
     * @api
     */
    calculateExtent(size) {
      const extent = this.calculateExtentInternal(size);
      return toUserExtent(extent, this.getProjection());
    }

    /**
     * @param {import("./size.js").Size} [size] Box pixel size. If not provided,
     * the map's last known viewport size will be used.
     * @return {import("./extent.js").Extent} Extent.
     */
    calculateExtentInternal(size) {
      size = size || this.getViewportSizeMinusPadding_();
      const center = /** @type {!import("./coordinate.js").Coordinate} */ (
        this.getCenterInternal()
      );
      assert(center, 1); // The view center is not defined
      const resolution = /** @type {!number} */ (this.getResolution());
      assert(resolution !== undefined, 2); // The view resolution is not defined
      const rotation = /** @type {!number} */ (this.getRotation());
      assert(rotation !== undefined, 3); // The view rotation is not defined

      return getForViewAndSize(center, resolution, rotation, size);
    }

    /**
     * Get the maximum resolution of the view.
     * @return {number} The maximum resolution of the view.
     * @api
     */
    getMaxResolution() {
      return this.maxResolution_;
    }

    /**
     * Get the minimum resolution of the view.
     * @return {number} The minimum resolution of the view.
     * @api
     */
    getMinResolution() {
      return this.minResolution_;
    }

    /**
     * Get the maximum zoom level for the view.
     * @return {number} The maximum zoom level.
     * @api
     */
    getMaxZoom() {
      return /** @type {number} */ (
        this.getZoomForResolution(this.minResolution_)
      );
    }

    /**
     * Set a new maximum zoom level for the view.
     * @param {number} zoom The maximum zoom level.
     * @api
     */
    setMaxZoom(zoom) {
      this.applyOptions_(this.getUpdatedOptions_({maxZoom: zoom}));
    }

    /**
     * Get the minimum zoom level for the view.
     * @return {number} The minimum zoom level.
     * @api
     */
    getMinZoom() {
      return /** @type {number} */ (
        this.getZoomForResolution(this.maxResolution_)
      );
    }

    /**
     * Set a new minimum zoom level for the view.
     * @param {number} zoom The minimum zoom level.
     * @api
     */
    setMinZoom(zoom) {
      this.applyOptions_(this.getUpdatedOptions_({minZoom: zoom}));
    }

    /**
     * Set whether the view should allow intermediary zoom levels.
     * @param {boolean} enabled Whether the resolution is constrained.
     * @api
     */
    setConstrainResolution(enabled) {
      this.applyOptions_(this.getUpdatedOptions_({constrainResolution: enabled}));
    }

    /**
     * Get the view projection.
     * @return {import("./proj/Projection.js").default} The projection of the view.
     * @api
     */
    getProjection() {
      return this.projection_;
    }

    /**
     * Get the view resolution.
     * @return {number|undefined} The resolution of the view.
     * @observable
     * @api
     */
    getResolution() {
      return /** @type {number|undefined} */ (this.get(ViewProperty.RESOLUTION));
    }

    /**
     * Get the resolutions for the view. This returns the array of resolutions
     * passed to the constructor of the View, or undefined if none were given.
     * @return {Array<number>|undefined} The resolutions of the view.
     * @api
     */
    getResolutions() {
      return this.resolutions_;
    }

    /**
     * Get the resolution for a provided extent (in map units) and size (in pixels).
     * @param {import("./extent.js").Extent} extent Extent.
     * @param {import("./size.js").Size} [size] Box pixel size.
     * @return {number} The resolution at which the provided extent will render at
     *     the given size.
     * @api
     */
    getResolutionForExtent(extent, size) {
      return this.getResolutionForExtentInternal(
        fromUserExtent(extent, this.getProjection()),
        size
      );
    }

    /**
     * Get the resolution for a provided extent (in map units) and size (in pixels).
     * @param {import("./extent.js").Extent} extent Extent.
     * @param {import("./size.js").Size} [size] Box pixel size.
     * @return {number} The resolution at which the provided extent will render at
     *     the given size.
     */
    getResolutionForExtentInternal(extent, size) {
      size = size || this.getViewportSizeMinusPadding_();
      const xResolution = getWidth(extent) / size[0];
      const yResolution = getHeight(extent) / size[1];
      return Math.max(xResolution, yResolution);
    }

    /**
     * Return a function that returns a value between 0 and 1 for a
     * resolution. Exponential scaling is assumed.
     * @param {number} [power] Power.
     * @return {function(number): number} Resolution for value function.
     */
    getResolutionForValueFunction(power) {
      power = power || 2;
      const maxResolution = this.getConstrainedResolution(this.maxResolution_);
      const minResolution = this.minResolution_;
      const max = Math.log(maxResolution / minResolution) / Math.log(power);
      return (
        /**
         * @param {number} value Value.
         * @return {number} Resolution.
         */
        function (value) {
          const resolution = maxResolution / Math.pow(power, value * max);
          return resolution;
        }
      );
    }

    /**
     * Get the view rotation.
     * @return {number} The rotation of the view in radians.
     * @observable
     * @api
     */
    getRotation() {
      return /** @type {number} */ (this.get(ViewProperty.ROTATION));
    }

    /**
     * Return a function that returns a resolution for a value between
     * 0 and 1. Exponential scaling is assumed.
     * @param {number} [power] Power.
     * @return {function(number): number} Value for resolution function.
     */
    getValueForResolutionFunction(power) {
      const logPower = Math.log(power || 2);
      const maxResolution = this.getConstrainedResolution(this.maxResolution_);
      const minResolution = this.minResolution_;
      const max = Math.log(maxResolution / minResolution) / logPower;
      return (
        /**
         * @param {number} resolution Resolution.
         * @return {number} Value.
         */
        function (resolution) {
          const value = Math.log(maxResolution / resolution) / logPower / max;
          return value;
        }
      );
    }

    /**
     * Returns the size of the viewport minus padding.
     * @private
     * @param {number} [rotation] Take into account the rotation of the viewport when giving the size
     * @return {import("./size.js").Size} Viewport size reduced by the padding.
     */
    getViewportSizeMinusPadding_(rotation) {
      let size = this.getViewportSize_(rotation);
      const padding = this.padding_;
      if (padding) {
        size = [
          size[0] - padding[1] - padding[3],
          size[1] - padding[0] - padding[2],
        ];
      }
      return size;
    }

    /**
     * @return {State} View state.
     */
    getState() {
      const projection = this.getProjection();
      const resolution = this.getResolution();
      const rotation = this.getRotation();
      let center = /** @type {import("./coordinate.js").Coordinate} */ (
        this.getCenterInternal()
      );
      const padding = this.padding_;
      if (padding) {
        const reducedSize = this.getViewportSizeMinusPadding_();
        center = calculateCenterOn(
          center,
          this.getViewportSize_(),
          [reducedSize[0] / 2 + padding[3], reducedSize[1] / 2 + padding[0]],
          resolution,
          rotation
        );
      }
      return {
        center: center.slice(0),
        projection: projection !== undefined ? projection : null,
        resolution: resolution,
        nextCenter: this.nextCenter_,
        nextResolution: this.nextResolution_,
        nextRotation: this.nextRotation_,
        rotation: rotation,
        zoom: this.getZoom(),
      };
    }

    /**
     * Get the current zoom level. This method may return non-integer zoom levels
     * if the view does not constrain the resolution, or if an interaction or
     * animation is underway.
     * @return {number|undefined} Zoom.
     * @api
     */
    getZoom() {
      let zoom;
      const resolution = this.getResolution();
      if (resolution !== undefined) {
        zoom = this.getZoomForResolution(resolution);
      }
      return zoom;
    }

    /**
     * Get the zoom level for a resolution.
     * @param {number} resolution The resolution.
     * @return {number|undefined} The zoom level for the provided resolution.
     * @api
     */
    getZoomForResolution(resolution) {
      let offset = this.minZoom_ || 0;
      let max, zoomFactor;
      if (this.resolutions_) {
        const nearest = linearFindNearest(this.resolutions_, resolution, 1);
        offset = nearest;
        max = this.resolutions_[nearest];
        if (nearest == this.resolutions_.length - 1) {
          zoomFactor = 2;
        } else {
          zoomFactor = max / this.resolutions_[nearest + 1];
        }
      } else {
        max = this.maxResolution_;
        zoomFactor = this.zoomFactor_;
      }
      return offset + Math.log(max / resolution) / Math.log(zoomFactor);
    }

    /**
     * Get the resolution for a zoom level.
     * @param {number} zoom Zoom level.
     * @return {number} The view resolution for the provided zoom level.
     * @api
     */
    getResolutionForZoom(zoom) {
      if (this.resolutions_) {
        if (this.resolutions_.length <= 1) {
          return 0;
        }
        const baseLevel = clamp(
          Math.floor(zoom),
          0,
          this.resolutions_.length - 2
        );
        const zoomFactor =
          this.resolutions_[baseLevel] / this.resolutions_[baseLevel + 1];
        return (
          this.resolutions_[baseLevel] /
          Math.pow(zoomFactor, clamp(zoom - baseLevel, 0, 1))
        );
      } else {
        return (
          this.maxResolution_ / Math.pow(this.zoomFactor_, zoom - this.minZoom_)
        );
      }
    }

    /**
     * Fit the given geometry or extent based on the given map size and border.
     * The size is pixel dimensions of the box to fit the extent into.
     * In most cases you will want to use the map size, that is `map.getSize()`.
     * Takes care of the map angle.
     * @param {import("./geom/SimpleGeometry.js").default|import("./extent.js").Extent} geometryOrExtent The geometry or
     *     extent to fit the view to.
     * @param {FitOptions} [options] Options.
     * @api
     */
    fit(geometryOrExtent, options) {
      /** @type {import("./geom/SimpleGeometry.js").default} */
      let geometry;
      assert(
        Array.isArray(geometryOrExtent) ||
          typeof (/** @type {?} */ (geometryOrExtent).getSimplifiedGeometry) ===
            'function',
        24
      ); // Invalid extent or geometry provided as `geometry`
      if (Array.isArray(geometryOrExtent)) {
        assert(!isEmpty(geometryOrExtent), 25); // Cannot fit empty extent provided as `geometry`
        const extent = fromUserExtent(geometryOrExtent, this.getProjection());
        geometry = fromExtent(extent);
      } else if (geometryOrExtent.getType() === 'Circle') {
        const extent = fromUserExtent(
          geometryOrExtent.getExtent(),
          this.getProjection()
        );
        geometry = fromExtent(extent);
        geometry.rotate(this.getRotation(), getCenter(extent));
      } else {
        const userProjection = getUserProjection();
        if (userProjection) {
          geometry = /** @type {import("./geom/SimpleGeometry.js").default} */ (
            geometryOrExtent
              .clone()
              .transform(userProjection, this.getProjection())
          );
        } else {
          geometry = geometryOrExtent;
        }
      }

      this.fitInternal(geometry, options);
    }

    /**
     * Calculate rotated extent
     * @param {import("./geom/SimpleGeometry.js").default} geometry The geometry.
     * @return {import("./extent").Extent} The rotated extent for the geometry.
     */
    rotatedExtentForGeometry(geometry) {
      const rotation = this.getRotation();
      const cosAngle = Math.cos(rotation);
      const sinAngle = Math.sin(-rotation);
      const coords = geometry.getFlatCoordinates();
      const stride = geometry.getStride();
      let minRotX = +Infinity;
      let minRotY = +Infinity;
      let maxRotX = -Infinity;
      let maxRotY = -Infinity;
      for (let i = 0, ii = coords.length; i < ii; i += stride) {
        const rotX = coords[i] * cosAngle - coords[i + 1] * sinAngle;
        const rotY = coords[i] * sinAngle + coords[i + 1] * cosAngle;
        minRotX = Math.min(minRotX, rotX);
        minRotY = Math.min(minRotY, rotY);
        maxRotX = Math.max(maxRotX, rotX);
        maxRotY = Math.max(maxRotY, rotY);
      }
      return [minRotX, minRotY, maxRotX, maxRotY];
    }

    /**
     * @param {import("./geom/SimpleGeometry.js").default} geometry The geometry.
     * @param {FitOptions} [options] Options.
     */
    fitInternal(geometry, options) {
      options = options || {};
      let size = options.size;
      if (!size) {
        size = this.getViewportSizeMinusPadding_();
      }
      const padding =
        options.padding !== undefined ? options.padding : [0, 0, 0, 0];
      const nearest = options.nearest !== undefined ? options.nearest : false;
      let minResolution;
      if (options.minResolution !== undefined) {
        minResolution = options.minResolution;
      } else if (options.maxZoom !== undefined) {
        minResolution = this.getResolutionForZoom(options.maxZoom);
      } else {
        minResolution = 0;
      }

      const rotatedExtent = this.rotatedExtentForGeometry(geometry);

      // calculate resolution
      let resolution = this.getResolutionForExtentInternal(rotatedExtent, [
        size[0] - padding[1] - padding[3],
        size[1] - padding[0] - padding[2],
      ]);
      resolution = isNaN(resolution)
        ? minResolution
        : Math.max(resolution, minResolution);
      resolution = this.getConstrainedResolution(resolution, nearest ? 0 : 1);

      // calculate center
      const rotation = this.getRotation();
      const sinAngle = Math.sin(rotation);
      const cosAngle = Math.cos(rotation);
      const centerRot = getCenter(rotatedExtent);
      centerRot[0] += ((padding[1] - padding[3]) / 2) * resolution;
      centerRot[1] += ((padding[0] - padding[2]) / 2) * resolution;
      const centerX = centerRot[0] * cosAngle - centerRot[1] * sinAngle;
      const centerY = centerRot[1] * cosAngle + centerRot[0] * sinAngle;
      const center = this.getConstrainedCenter([centerX, centerY], resolution);
      const callback = options.callback ? options.callback : VOID;

      if (options.duration !== undefined) {
        this.animateInternal(
          {
            resolution: resolution,
            center: center,
            duration: options.duration,
            easing: options.easing,
          },
          callback
        );
      } else {
        this.targetResolution_ = resolution;
        this.targetCenter_ = center;
        this.applyTargetState_(false, true);
        animationCallback(callback, true);
      }
    }

    /**
     * Center on coordinate and view position.
     * @param {import("./coordinate.js").Coordinate} coordinate Coordinate.
     * @param {import("./size.js").Size} size Box pixel size.
     * @param {import("./pixel.js").Pixel} position Position on the view to center on.
     * @api
     */
    centerOn(coordinate, size, position) {
      this.centerOnInternal(
        fromUserCoordinate(coordinate, this.getProjection()),
        size,
        position
      );
    }

    /**
     * @param {import("./coordinate.js").Coordinate} coordinate Coordinate.
     * @param {import("./size.js").Size} size Box pixel size.
     * @param {import("./pixel.js").Pixel} position Position on the view to center on.
     */
    centerOnInternal(coordinate, size, position) {
      this.setCenterInternal(
        calculateCenterOn(
          coordinate,
          size,
          position,
          this.getResolution(),
          this.getRotation()
        )
      );
    }

    /**
     * Calculates the shift between map and viewport center.
     * @param {import("./coordinate.js").Coordinate} center Center.
     * @param {number} resolution Resolution.
     * @param {number} rotation Rotation.
     * @param {import("./size.js").Size} size Size.
     * @return {Array<number>|undefined} Center shift.
     */
    calculateCenterShift(center, resolution, rotation, size) {
      let centerShift;
      const padding = this.padding_;
      if (padding && center) {
        const reducedSize = this.getViewportSizeMinusPadding_(-rotation);
        const shiftedCenter = calculateCenterOn(
          center,
          size,
          [reducedSize[0] / 2 + padding[3], reducedSize[1] / 2 + padding[0]],
          resolution,
          rotation
        );
        centerShift = [
          center[0] - shiftedCenter[0],
          center[1] - shiftedCenter[1],
        ];
      }
      return centerShift;
    }

    /**
     * @return {boolean} Is defined.
     */
    isDef() {
      return !!this.getCenterInternal() && this.getResolution() !== undefined;
    }

    /**
     * Adds relative coordinates to the center of the view. Any extent constraint will apply.
     * @param {import("./coordinate.js").Coordinate} deltaCoordinates Relative value to add.
     * @api
     */
    adjustCenter(deltaCoordinates) {
      const center = toUserCoordinate(this.targetCenter_, this.getProjection());
      this.setCenter([
        center[0] + deltaCoordinates[0],
        center[1] + deltaCoordinates[1],
      ]);
    }

    /**
     * Adds relative coordinates to the center of the view. Any extent constraint will apply.
     * @param {import("./coordinate.js").Coordinate} deltaCoordinates Relative value to add.
     */
    adjustCenterInternal(deltaCoordinates) {
      const center = this.targetCenter_;
      this.setCenterInternal([
        center[0] + deltaCoordinates[0],
        center[1] + deltaCoordinates[1],
      ]);
    }

    /**
     * Multiply the view resolution by a ratio, optionally using an anchor. Any resolution
     * constraint will apply.
     * @param {number} ratio The ratio to apply on the view resolution.
     * @param {import("./coordinate.js").Coordinate} [anchor] The origin of the transformation.
     * @api
     */
    adjustResolution(ratio, anchor) {
      anchor = anchor && fromUserCoordinate(anchor, this.getProjection());
      this.adjustResolutionInternal(ratio, anchor);
    }

    /**
     * Multiply the view resolution by a ratio, optionally using an anchor. Any resolution
     * constraint will apply.
     * @param {number} ratio The ratio to apply on the view resolution.
     * @param {import("./coordinate.js").Coordinate} [anchor] The origin of the transformation.
     */
    adjustResolutionInternal(ratio, anchor) {
      const isMoving = this.getAnimating() || this.getInteracting();
      const size = this.getViewportSize_(this.getRotation());
      const newResolution = this.constraints_.resolution(
        this.targetResolution_ * ratio,
        0,
        size,
        isMoving
      );

      if (anchor) {
        this.targetCenter_ = this.calculateCenterZoom(newResolution, anchor);
      }

      this.targetResolution_ *= ratio;
      this.applyTargetState_();
    }

    /**
     * Adds a value to the view zoom level, optionally using an anchor. Any resolution
     * constraint will apply.
     * @param {number} delta Relative value to add to the zoom level.
     * @param {import("./coordinate.js").Coordinate} [anchor] The origin of the transformation.
     * @api
     */
    adjustZoom(delta, anchor) {
      this.adjustResolution(Math.pow(this.zoomFactor_, -delta), anchor);
    }

    /**
     * Adds a value to the view rotation, optionally using an anchor. Any rotation
     * constraint will apply.
     * @param {number} delta Relative value to add to the zoom rotation, in radians.
     * @param {import("./coordinate.js").Coordinate} [anchor] The rotation center.
     * @api
     */
    adjustRotation(delta, anchor) {
      if (anchor) {
        anchor = fromUserCoordinate(anchor, this.getProjection());
      }
      this.adjustRotationInternal(delta, anchor);
    }

    /**
     * @param {number} delta Relative value to add to the zoom rotation, in radians.
     * @param {import("./coordinate.js").Coordinate} [anchor] The rotation center.
     */
    adjustRotationInternal(delta, anchor) {
      const isMoving = this.getAnimating() || this.getInteracting();
      const newRotation = this.constraints_.rotation(
        this.targetRotation_ + delta,
        isMoving
      );
      if (anchor) {
        this.targetCenter_ = this.calculateCenterRotate(newRotation, anchor);
      }
      this.targetRotation_ += delta;
      this.applyTargetState_();
    }

    /**
     * Set the center of the current view. Any extent constraint will apply.
     * @param {import("./coordinate.js").Coordinate|undefined} center The center of the view.
     * @observable
     * @api
     */
    setCenter(center) {
      this.setCenterInternal(
        center ? fromUserCoordinate(center, this.getProjection()) : center
      );
    }

    /**
     * Set the center using the view projection (not the user projection).
     * @param {import("./coordinate.js").Coordinate|undefined} center The center of the view.
     */
    setCenterInternal(center) {
      this.targetCenter_ = center;
      this.applyTargetState_();
    }

    /**
     * @param {import("./ViewHint.js").default} hint Hint.
     * @param {number} delta Delta.
     * @return {number} New value.
     */
    setHint(hint, delta) {
      this.hints_[hint] += delta;
      this.changed();
      return this.hints_[hint];
    }

    /**
     * Set the resolution for this view. Any resolution constraint will apply.
     * @param {number|undefined} resolution The resolution of the view.
     * @observable
     * @api
     */
    setResolution(resolution) {
      this.targetResolution_ = resolution;
      this.applyTargetState_();
    }

    /**
     * Set the rotation for this view. Any rotation constraint will apply.
     * @param {number} rotation The rotation of the view in radians.
     * @observable
     * @api
     */
    setRotation(rotation) {
      this.targetRotation_ = rotation;
      this.applyTargetState_();
    }

    /**
     * Zoom to a specific zoom level. Any resolution constrain will apply.
     * @param {number} zoom Zoom level.
     * @api
     */
    setZoom(zoom) {
      this.setResolution(this.getResolutionForZoom(zoom));
    }

    /**
     * Recompute rotation/resolution/center based on target values.
     * Note: we have to compute rotation first, then resolution and center considering that
     * parameters can influence one another in case a view extent constraint is present.
     * @param {boolean} [doNotCancelAnims] Do not cancel animations.
     * @param {boolean} [forceMoving] Apply constraints as if the view is moving.
     * @private
     */
    applyTargetState_(doNotCancelAnims, forceMoving) {
      const isMoving =
        this.getAnimating() || this.getInteracting() || forceMoving;

      // compute rotation
      const newRotation = this.constraints_.rotation(
        this.targetRotation_,
        isMoving
      );
      const size = this.getViewportSize_(newRotation);
      const newResolution = this.constraints_.resolution(
        this.targetResolution_,
        0,
        size,
        isMoving
      );
      const newCenter = this.constraints_.center(
        this.targetCenter_,
        newResolution,
        size,
        isMoving,
        this.calculateCenterShift(
          this.targetCenter_,
          newResolution,
          newRotation,
          size
        )
      );

      if (this.get(ViewProperty.ROTATION) !== newRotation) {
        this.set(ViewProperty.ROTATION, newRotation);
      }
      if (this.get(ViewProperty.RESOLUTION) !== newResolution) {
        this.set(ViewProperty.RESOLUTION, newResolution);
        this.set('zoom', this.getZoom(), true);
      }
      if (
        !newCenter ||
        !this.get(ViewProperty.CENTER) ||
        !equals$1(this.get(ViewProperty.CENTER), newCenter)
      ) {
        this.set(ViewProperty.CENTER, newCenter);
      }

      if (this.getAnimating() && !doNotCancelAnims) {
        this.cancelAnimations();
      }
      this.cancelAnchor_ = undefined;
    }

    /**
     * If any constraints need to be applied, an animation will be triggered.
     * This is typically done on interaction end.
     * Note: calling this with a duration of 0 will apply the constrained values straight away,
     * without animation.
     * @param {number} [duration] The animation duration in ms.
     * @param {number} [resolutionDirection] Which direction to zoom.
     * @param {import("./coordinate.js").Coordinate} [anchor] The origin of the transformation.
     */
    resolveConstraints(duration, resolutionDirection, anchor) {
      duration = duration !== undefined ? duration : 200;
      const direction = resolutionDirection || 0;

      const newRotation = this.constraints_.rotation(this.targetRotation_);
      const size = this.getViewportSize_(newRotation);
      const newResolution = this.constraints_.resolution(
        this.targetResolution_,
        direction,
        size
      );
      const newCenter = this.constraints_.center(
        this.targetCenter_,
        newResolution,
        size,
        false,
        this.calculateCenterShift(
          this.targetCenter_,
          newResolution,
          newRotation,
          size
        )
      );

      if (duration === 0 && !this.cancelAnchor_) {
        this.targetResolution_ = newResolution;
        this.targetRotation_ = newRotation;
        this.targetCenter_ = newCenter;
        this.applyTargetState_();
        return;
      }

      anchor = anchor || (duration === 0 ? this.cancelAnchor_ : undefined);
      this.cancelAnchor_ = undefined;

      if (
        this.getResolution() !== newResolution ||
        this.getRotation() !== newRotation ||
        !this.getCenterInternal() ||
        !equals$1(this.getCenterInternal(), newCenter)
      ) {
        if (this.getAnimating()) {
          this.cancelAnimations();
        }

        this.animateInternal({
          rotation: newRotation,
          center: newCenter,
          resolution: newResolution,
          duration: duration,
          easing: easeOut,
          anchor: anchor,
        });
      }
    }

    /**
     * Notify the View that an interaction has started.
     * The view state will be resolved to a stable one if needed
     * (depending on its constraints).
     * @api
     */
    beginInteraction() {
      this.resolveConstraints(0);

      this.setHint(ViewHint.INTERACTING, 1);
    }

    /**
     * Notify the View that an interaction has ended. The view state will be resolved
     * to a stable one if needed (depending on its constraints).
     * @param {number} [duration] Animation duration in ms.
     * @param {number} [resolutionDirection] Which direction to zoom.
     * @param {import("./coordinate.js").Coordinate} [anchor] The origin of the transformation.
     * @api
     */
    endInteraction(duration, resolutionDirection, anchor) {
      anchor = anchor && fromUserCoordinate(anchor, this.getProjection());
      this.endInteractionInternal(duration, resolutionDirection, anchor);
    }

    /**
     * Notify the View that an interaction has ended. The view state will be resolved
     * to a stable one if needed (depending on its constraints).
     * @param {number} [duration] Animation duration in ms.
     * @param {number} [resolutionDirection] Which direction to zoom.
     * @param {import("./coordinate.js").Coordinate} [anchor] The origin of the transformation.
     */
    endInteractionInternal(duration, resolutionDirection, anchor) {
      this.setHint(ViewHint.INTERACTING, -1);

      this.resolveConstraints(duration, resolutionDirection, anchor);
    }

    /**
     * Get a valid position for the view center according to the current constraints.
     * @param {import("./coordinate.js").Coordinate|undefined} targetCenter Target center position.
     * @param {number} [targetResolution] Target resolution. If not supplied, the current one will be used.
     * This is useful to guess a valid center position at a different zoom level.
     * @return {import("./coordinate.js").Coordinate|undefined} Valid center position.
     */
    getConstrainedCenter(targetCenter, targetResolution) {
      const size = this.getViewportSize_(this.getRotation());
      return this.constraints_.center(
        targetCenter,
        targetResolution || this.getResolution(),
        size
      );
    }

    /**
     * Get a valid zoom level according to the current view constraints.
     * @param {number|undefined} targetZoom Target zoom.
     * @param {number} [direction=0] Indicate which resolution should be used
     * by a renderer if the view resolution does not match any resolution of the tile source.
     * If 0, the nearest resolution will be used. If 1, the nearest lower resolution
     * will be used. If -1, the nearest higher resolution will be used.
     * @return {number|undefined} Valid zoom level.
     */
    getConstrainedZoom(targetZoom, direction) {
      const targetRes = this.getResolutionForZoom(targetZoom);
      return this.getZoomForResolution(
        this.getConstrainedResolution(targetRes, direction)
      );
    }

    /**
     * Get a valid resolution according to the current view constraints.
     * @param {number|undefined} targetResolution Target resolution.
     * @param {number} [direction=0] Indicate which resolution should be used
     * by a renderer if the view resolution does not match any resolution of the tile source.
     * If 0, the nearest resolution will be used. If 1, the nearest lower resolution
     * will be used. If -1, the nearest higher resolution will be used.
     * @return {number|undefined} Valid resolution.
     */
    getConstrainedResolution(targetResolution, direction) {
      direction = direction || 0;
      const size = this.getViewportSize_(this.getRotation());

      return this.constraints_.resolution(targetResolution, direction, size);
    }
  }

  /**
   * @param {Function} callback Callback.
   * @param {*} returnValue Return value.
   */
  function animationCallback(callback, returnValue) {
    setTimeout(function () {
      callback(returnValue);
    }, 0);
  }

  /**
   * @param {ViewOptions} options View options.
   * @return {import("./centerconstraint.js").Type} The constraint.
   */
  function createCenterConstraint(options) {
    if (options.extent !== undefined) {
      const smooth =
        options.smoothExtentConstraint !== undefined
          ? options.smoothExtentConstraint
          : true;
      return createExtent(options.extent, options.constrainOnlyCenter, smooth);
    }

    const projection = createProjection(options.projection, 'EPSG:3857');
    if (options.multiWorld !== true && projection.isGlobal()) {
      const extent = projection.getExtent().slice();
      extent[0] = -Infinity;
      extent[2] = Infinity;
      return createExtent(extent, false, false);
    }

    return none$1;
  }

  /**
   * @param {ViewOptions} options View options.
   * @return {{constraint: import("./resolutionconstraint.js").Type, maxResolution: number,
   *     minResolution: number, minZoom: number, zoomFactor: number}} The constraint.
   */
  function createResolutionConstraint(options) {
    let resolutionConstraint;
    let maxResolution;
    let minResolution;

    // TODO: move these to be ol constants
    // see https://github.com/openlayers/openlayers/issues/2076
    const defaultMaxZoom = 28;
    const defaultZoomFactor = 2;

    let minZoom =
      options.minZoom !== undefined ? options.minZoom : DEFAULT_MIN_ZOOM;

    let maxZoom =
      options.maxZoom !== undefined ? options.maxZoom : defaultMaxZoom;

    const zoomFactor =
      options.zoomFactor !== undefined ? options.zoomFactor : defaultZoomFactor;

    const multiWorld =
      options.multiWorld !== undefined ? options.multiWorld : false;

    const smooth =
      options.smoothResolutionConstraint !== undefined
        ? options.smoothResolutionConstraint
        : true;

    const showFullExtent =
      options.showFullExtent !== undefined ? options.showFullExtent : false;

    const projection = createProjection(options.projection, 'EPSG:3857');
    const projExtent = projection.getExtent();
    let constrainOnlyCenter = options.constrainOnlyCenter;
    let extent = options.extent;
    if (!multiWorld && !extent && projection.isGlobal()) {
      constrainOnlyCenter = false;
      extent = projExtent;
    }

    if (options.resolutions !== undefined) {
      const resolutions = options.resolutions;
      maxResolution = resolutions[minZoom];
      minResolution =
        resolutions[maxZoom] !== undefined
          ? resolutions[maxZoom]
          : resolutions[resolutions.length - 1];

      if (options.constrainResolution) {
        resolutionConstraint = createSnapToResolutions(
          resolutions,
          smooth,
          !constrainOnlyCenter && extent,
          showFullExtent
        );
      } else {
        resolutionConstraint = createMinMaxResolution(
          maxResolution,
          minResolution,
          smooth,
          !constrainOnlyCenter && extent,
          showFullExtent
        );
      }
    } else {
      // calculate the default min and max resolution
      const size = !projExtent
        ? // use an extent that can fit the whole world if need be
          (360 * METERS_PER_UNIT$1.degrees) / projection.getMetersPerUnit()
        : Math.max(getWidth(projExtent), getHeight(projExtent));

      const defaultMaxResolution =
        size / DEFAULT_TILE_SIZE / Math.pow(defaultZoomFactor, DEFAULT_MIN_ZOOM);

      const defaultMinResolution =
        defaultMaxResolution /
        Math.pow(defaultZoomFactor, defaultMaxZoom - DEFAULT_MIN_ZOOM);

      // user provided maxResolution takes precedence
      maxResolution = options.maxResolution;
      if (maxResolution !== undefined) {
        minZoom = 0;
      } else {
        maxResolution = defaultMaxResolution / Math.pow(zoomFactor, minZoom);
      }

      // user provided minResolution takes precedence
      minResolution = options.minResolution;
      if (minResolution === undefined) {
        if (options.maxZoom !== undefined) {
          if (options.maxResolution !== undefined) {
            minResolution = maxResolution / Math.pow(zoomFactor, maxZoom);
          } else {
            minResolution = defaultMaxResolution / Math.pow(zoomFactor, maxZoom);
          }
        } else {
          minResolution = defaultMinResolution;
        }
      }

      // given discrete zoom levels, minResolution may be different than provided
      maxZoom =
        minZoom +
        Math.floor(
          Math.log(maxResolution / minResolution) / Math.log(zoomFactor)
        );
      minResolution = maxResolution / Math.pow(zoomFactor, maxZoom - minZoom);

      if (options.constrainResolution) {
        resolutionConstraint = createSnapToPower(
          zoomFactor,
          maxResolution,
          minResolution,
          smooth,
          !constrainOnlyCenter && extent,
          showFullExtent
        );
      } else {
        resolutionConstraint = createMinMaxResolution(
          maxResolution,
          minResolution,
          smooth,
          !constrainOnlyCenter && extent,
          showFullExtent
        );
      }
    }
    return {
      constraint: resolutionConstraint,
      maxResolution: maxResolution,
      minResolution: minResolution,
      minZoom: minZoom,
      zoomFactor: zoomFactor,
    };
  }

  /**
   * @param {ViewOptions} options View options.
   * @return {import("./rotationconstraint.js").Type} Rotation constraint.
   */
  function createRotationConstraint(options) {
    const enableRotation =
      options.enableRotation !== undefined ? options.enableRotation : true;
    if (enableRotation) {
      const constrainRotation = options.constrainRotation;
      if (constrainRotation === undefined || constrainRotation === true) {
        return createSnapToZero();
      } else if (constrainRotation === false) {
        return none;
      } else if (typeof constrainRotation === 'number') {
        return createSnapToN(constrainRotation);
      } else {
        return none;
      }
    } else {
      return disable;
    }
  }

  /**
   * Determine if an animation involves no view change.
   * @param {Animation} animation The animation.
   * @return {boolean} The animation involves no view change.
   */
  function isNoopAnimation(animation) {
    if (animation.sourceCenter && animation.targetCenter) {
      if (!equals$1(animation.sourceCenter, animation.targetCenter)) {
        return false;
      }
    }
    if (animation.sourceResolution !== animation.targetResolution) {
      return false;
    }
    if (animation.sourceRotation !== animation.targetRotation) {
      return false;
    }
    return true;
  }

  /**
   * @param {import("./coordinate.js").Coordinate} coordinate Coordinate.
   * @param {import("./size.js").Size} size Box pixel size.
   * @param {import("./pixel.js").Pixel} position Position on the view to center on.
   * @param {number} resolution Resolution.
   * @param {number} rotation Rotation.
   * @return {import("./coordinate.js").Coordinate} Shifted center.
   */
  function calculateCenterOn(coordinate, size, position, resolution, rotation) {
    // calculate rotated position
    const cosAngle = Math.cos(-rotation);
    let sinAngle = Math.sin(-rotation);
    let rotX = coordinate[0] * cosAngle - coordinate[1] * sinAngle;
    let rotY = coordinate[1] * cosAngle + coordinate[0] * sinAngle;
    rotX += (size[0] / 2 - position[0]) * resolution;
    rotY += (position[1] - size[1] / 2) * resolution;

    // go back to original angle
    sinAngle = -sinAngle; // go back to original rotation
    const centerX = rotX * cosAngle - rotY * sinAngle;
    const centerY = rotY * cosAngle + rotX * sinAngle;

    return [centerX, centerY];
  }

  /**
   * @module ol/control/Control
   */

  /**
   * @typedef {Object} Options
   * @property {HTMLElement} [element] The element is the control's
   * container element. This only needs to be specified if you're developing
   * a custom control.
   * @property {function(import("../MapEvent.js").default):void} [render] Function called when
   * the control should be re-rendered. This is called in a `requestAnimationFrame`
   * callback.
   * @property {HTMLElement|string} [target] Specify a target if you want
   * the control to be rendered outside of the map's viewport.
   */

  /**
   * @classdesc
   * A control is a visible widget with a DOM element in a fixed position on the
   * screen. They can involve user input (buttons), or be informational only;
   * the position is determined using CSS. By default these are placed in the
   * container with CSS class name `ol-overlaycontainer-stopevent`, but can use
   * any outside DOM element.
   *
   * This is the base class for controls. You can use it for simple custom
   * controls by creating the element with listeners, creating an instance:
   * ```js
   * const myControl = new Control({element: myElement});
   * ```
   * and then adding this to the map.
   *
   * The main advantage of having this as a control rather than a simple separate
   * DOM element is that preventing propagation is handled for you. Controls
   * will also be objects in a {@link module:ol/Collection~Collection}, so you can use their methods.
   *
   * You can also extend this base for your own control class. See
   * examples/custom-controls for an example of how to do this.
   *
   * @api
   */
  class Control extends BaseObject {
    /**
     * @param {Options} options Control options.
     */
    constructor(options) {
      super();

      const element = options.element;
      if (element && !options.target && !element.style.pointerEvents) {
        element.style.pointerEvents = 'auto';
      }

      /**
       * @protected
       * @type {HTMLElement}
       */
      this.element = element ? element : null;

      /**
       * @private
       * @type {HTMLElement}
       */
      this.target_ = null;

      /**
       * @private
       * @type {import("../Map.js").default|null}
       */
      this.map_ = null;

      /**
       * @protected
       * @type {!Array<import("../events.js").EventsKey>}
       */
      this.listenerKeys = [];

      if (options.render) {
        this.render = options.render;
      }

      if (options.target) {
        this.setTarget(options.target);
      }
    }

    /**
     * Clean up.
     */
    disposeInternal() {
      removeNode$1(this.element);
      super.disposeInternal();
    }

    /**
     * Get the map associated with this control.
     * @return {import("../Map.js").default|null} Map.
     * @api
     */
    getMap() {
      return this.map_;
    }

    /**
     * Remove the control from its current map and attach it to the new map.
     * Pass `null` to just remove the control from the current map.
     * Subclasses may set up event handlers to get notified about changes to
     * the map here.
     * @param {import("../Map.js").default|null} map Map.
     * @api
     */
    setMap(map) {
      if (this.map_) {
        removeNode$1(this.element);
      }
      for (let i = 0, ii = this.listenerKeys.length; i < ii; ++i) {
        unlistenByKey(this.listenerKeys[i]);
      }
      this.listenerKeys.length = 0;
      this.map_ = map;
      if (map) {
        const target = this.target_
          ? this.target_
          : map.getOverlayContainerStopEvent();
        target.appendChild(this.element);
        if (this.render !== VOID) {
          this.listenerKeys.push(
            listen(map, MapEventType.POSTRENDER, this.render, this)
          );
        }
        map.render();
      }
    }

    /**
     * Renders the control.
     * @param {import("../MapEvent.js").default} mapEvent Map event.
     * @api
     */
    render(mapEvent) {}

    /**
     * This function is used to set a target element for the control. It has no
     * effect if it is called after the control has been added to the map (i.e.
     * after `setMap` is called on the control). If no `target` is set in the
     * options passed to the control constructor and if `setTarget` is not called
     * then the control is added to the map's overlay container.
     * @param {HTMLElement|string} target Target.
     * @api
     */
    setTarget(target) {
      this.target_ =
        typeof target === 'string' ? document.getElementById(target) : target;
    }
  }

  /**
   * @module ol/control/Attribution
   */

  /**
   * @typedef {Object} Options
   * @property {string} [className='ol-attribution'] CSS class name.
   * @property {HTMLElement|string} [target] Specify a target if you
   * want the control to be rendered outside of the map's
   * viewport.
   * @property {boolean} [collapsible] Specify if attributions can
   * be collapsed. If not specified, sources control this behavior with their
   * `attributionsCollapsible` setting.
   * @property {boolean} [collapsed=true] Specify if attributions should
   * be collapsed at startup.
   * @property {string} [tipLabel='Attributions'] Text label to use for the button tip.
   * @property {string|HTMLElement} [label='i'] Text label to use for the
   * collapsed attributions button.
   * Instead of text, also an element (e.g. a `span` element) can be used.
   * @property {string} [expandClassName=className + '-expand'] CSS class name for the
   * collapsed attributions button.
   * @property {string|HTMLElement} [collapseLabel='›'] Text label to use
   * for the expanded attributions button.
   * Instead of text, also an element (e.g. a `span` element) can be used.
   * @property {string} [collapseClassName=className + '-collapse'] CSS class name for the
   * expanded attributions button.
   * @property {function(import("../MapEvent.js").default):void} [render] Function called when
   * the control should be re-rendered. This is called in a `requestAnimationFrame`
   * callback.
   */

  /**
   * @classdesc
   * Control to show all the attributions associated with the layer sources
   * in the map. This control is one of the default controls included in maps.
   * By default it will show in the bottom right portion of the map, but this can
   * be changed by using a css selector for `.ol-attribution`.
   *
   * @api
   */
  class Attribution extends Control {
    /**
     * @param {Options} [options] Attribution options.
     */
    constructor(options) {
      options = options ? options : {};

      super({
        element: document.createElement('div'),
        render: options.render,
        target: options.target,
      });

      /**
       * @private
       * @type {HTMLElement}
       */
      this.ulElement_ = document.createElement('ul');

      /**
       * @private
       * @type {boolean}
       */
      this.collapsed_ =
        options.collapsed !== undefined ? options.collapsed : true;

      /**
       * @private
       * @type {boolean}
       */
      this.userCollapsed_ = this.collapsed_;

      /**
       * @private
       * @type {boolean}
       */
      this.overrideCollapsible_ = options.collapsible !== undefined;

      /**
       * @private
       * @type {boolean}
       */
      this.collapsible_ =
        options.collapsible !== undefined ? options.collapsible : true;

      if (!this.collapsible_) {
        this.collapsed_ = false;
      }

      const className =
        options.className !== undefined ? options.className : 'ol-attribution';

      const tipLabel =
        options.tipLabel !== undefined ? options.tipLabel : 'Attributions';

      const expandClassName =
        options.expandClassName !== undefined
          ? options.expandClassName
          : className + '-expand';

      const collapseLabel =
        options.collapseLabel !== undefined ? options.collapseLabel : '\u203A';

      const collapseClassName =
        options.collapseClassName !== undefined
          ? options.collapseClassName
          : className + '-collapse';

      if (typeof collapseLabel === 'string') {
        /**
         * @private
         * @type {HTMLElement}
         */
        this.collapseLabel_ = document.createElement('span');
        this.collapseLabel_.textContent = collapseLabel;
        this.collapseLabel_.className = collapseClassName;
      } else {
        this.collapseLabel_ = collapseLabel;
      }

      const label = options.label !== undefined ? options.label : 'i';

      if (typeof label === 'string') {
        /**
         * @private
         * @type {HTMLElement}
         */
        this.label_ = document.createElement('span');
        this.label_.textContent = label;
        this.label_.className = expandClassName;
      } else {
        this.label_ = label;
      }

      const activeLabel =
        this.collapsible_ && !this.collapsed_ ? this.collapseLabel_ : this.label_;

      /**
       * @private
       * @type {HTMLElement}
       */
      this.toggleButton_ = document.createElement('button');
      this.toggleButton_.setAttribute('type', 'button');
      this.toggleButton_.setAttribute('aria-expanded', String(!this.collapsed_));
      this.toggleButton_.title = tipLabel;
      this.toggleButton_.appendChild(activeLabel);

      this.toggleButton_.addEventListener(
        EventType$1.CLICK,
        this.handleClick_.bind(this),
        false
      );

      const cssClasses =
        className +
        ' ' +
        CLASS_UNSELECTABLE +
        ' ' +
        CLASS_CONTROL +
        (this.collapsed_ && this.collapsible_ ? ' ' + CLASS_COLLAPSED : '') +
        (this.collapsible_ ? '' : ' ol-uncollapsible');
      const element = this.element;
      element.className = cssClasses;
      element.appendChild(this.toggleButton_);
      element.appendChild(this.ulElement_);

      /**
       * A list of currently rendered resolutions.
       * @type {Array<string>}
       * @private
       */
      this.renderedAttributions_ = [];

      /**
       * @private
       * @type {boolean}
       */
      this.renderedVisible_ = true;
    }

    /**
     * Collect a list of visible attributions and set the collapsible state.
     * @param {import("../Map.js").FrameState} frameState Frame state.
     * @return {Array<string>} Attributions.
     * @private
     */
    collectSourceAttributions_(frameState) {
      /**
       * Used to determine if an attribution already exists.
       * @type {!Object<string, boolean>}
       */
      const lookup = {};

      /**
       * A list of visible attributions.
       * @type {Array<string>}
       */
      const visibleAttributions = [];

      let collapsible = true;
      const layerStatesArray = frameState.layerStatesArray;
      for (let i = 0, ii = layerStatesArray.length; i < ii; ++i) {
        const layerState = layerStatesArray[i];
        if (!inView(layerState, frameState.viewState)) {
          continue;
        }

        const source = /** @type {import("../layer/Layer.js").default} */ (
          layerState.layer
        ).getSource();
        if (!source) {
          continue;
        }

        const attributionGetter = source.getAttributions();
        if (!attributionGetter) {
          continue;
        }

        const attributions = attributionGetter(frameState);
        if (!attributions) {
          continue;
        }

        collapsible =
          collapsible && source.getAttributionsCollapsible() !== false;

        if (Array.isArray(attributions)) {
          for (let j = 0, jj = attributions.length; j < jj; ++j) {
            if (!(attributions[j] in lookup)) {
              visibleAttributions.push(attributions[j]);
              lookup[attributions[j]] = true;
            }
          }
        } else {
          if (!(attributions in lookup)) {
            visibleAttributions.push(attributions);
            lookup[attributions] = true;
          }
        }
      }
      if (!this.overrideCollapsible_) {
        this.setCollapsible(collapsible);
      }
      return visibleAttributions;
    }

    /**
     * @private
     * @param {?import("../Map.js").FrameState} frameState Frame state.
     */
    updateElement_(frameState) {
      if (!frameState) {
        if (this.renderedVisible_) {
          this.element.style.display = 'none';
          this.renderedVisible_ = false;
        }
        return;
      }

      const attributions = this.collectSourceAttributions_(frameState);

      const visible = attributions.length > 0;
      if (this.renderedVisible_ != visible) {
        this.element.style.display = visible ? '' : 'none';
        this.renderedVisible_ = visible;
      }

      if (equals$3(attributions, this.renderedAttributions_)) {
        return;
      }

      removeChildren(this.ulElement_);

      // append the attributions
      for (let i = 0, ii = attributions.length; i < ii; ++i) {
        const element = document.createElement('li');
        element.innerHTML = attributions[i];
        this.ulElement_.appendChild(element);
      }

      this.renderedAttributions_ = attributions;
    }

    /**
     * @param {MouseEvent} event The event to handle
     * @private
     */
    handleClick_(event) {
      event.preventDefault();
      this.handleToggle_();
      this.userCollapsed_ = this.collapsed_;
    }

    /**
     * @private
     */
    handleToggle_() {
      this.element.classList.toggle(CLASS_COLLAPSED);
      if (this.collapsed_) {
        replaceNode(this.collapseLabel_, this.label_);
      } else {
        replaceNode(this.label_, this.collapseLabel_);
      }
      this.collapsed_ = !this.collapsed_;
      this.toggleButton_.setAttribute('aria-expanded', String(!this.collapsed_));
    }

    /**
     * Return `true` if the attribution is collapsible, `false` otherwise.
     * @return {boolean} True if the widget is collapsible.
     * @api
     */
    getCollapsible() {
      return this.collapsible_;
    }

    /**
     * Set whether the attribution should be collapsible.
     * @param {boolean} collapsible True if the widget is collapsible.
     * @api
     */
    setCollapsible(collapsible) {
      if (this.collapsible_ === collapsible) {
        return;
      }
      this.collapsible_ = collapsible;
      this.element.classList.toggle('ol-uncollapsible');
      if (this.userCollapsed_) {
        this.handleToggle_();
      }
    }

    /**
     * Collapse or expand the attribution according to the passed parameter. Will
     * not do anything if the attribution isn't collapsible or if the current
     * collapsed state is already the one requested.
     * @param {boolean} collapsed True if the widget is collapsed.
     * @api
     */
    setCollapsed(collapsed) {
      this.userCollapsed_ = collapsed;
      if (!this.collapsible_ || this.collapsed_ === collapsed) {
        return;
      }
      this.handleToggle_();
    }

    /**
     * Return `true` when the attribution is currently collapsed or `false`
     * otherwise.
     * @return {boolean} True if the widget is collapsed.
     * @api
     */
    getCollapsed() {
      return this.collapsed_;
    }

    /**
     * Update the attribution element.
     * @param {import("../MapEvent.js").default} mapEvent Map event.
     * @override
     */
    render(mapEvent) {
      this.updateElement_(mapEvent.frameState);
    }
  }

  /**
   * @module ol/control/Rotate
   */

  /**
   * @typedef {Object} Options
   * @property {string} [className='ol-rotate'] CSS class name.
   * @property {string|HTMLElement} [label='⇧'] Text label to use for the rotate button.
   * Instead of text, also an element (e.g. a `span` element) can be used.
   * @property {string} [tipLabel='Reset rotation'] Text label to use for the rotate tip.
   * @property {string} [compassClassName='ol-compass'] CSS class name for the compass.
   * @property {number} [duration=250] Animation duration in milliseconds.
   * @property {boolean} [autoHide=true] Hide the control when rotation is 0.
   * @property {function(import("../MapEvent.js").default):void} [render] Function called when the control should
   * be re-rendered. This is called in a `requestAnimationFrame` callback.
   * @property {function():void} [resetNorth] Function called when the control is clicked.
   * This will override the default `resetNorth`.
   * @property {HTMLElement|string} [target] Specify a target if you want the control to be
   * rendered outside of the map's viewport.
   */

  /**
   * @classdesc
   * A button control to reset rotation to 0.
   * To style this control use css selector `.ol-rotate`. A `.ol-hidden` css
   * selector is added to the button when the rotation is 0.
   *
   * @api
   */
  class Rotate extends Control {
    /**
     * @param {Options} [options] Rotate options.
     */
    constructor(options) {
      options = options ? options : {};

      super({
        element: document.createElement('div'),
        render: options.render,
        target: options.target,
      });

      const className =
        options.className !== undefined ? options.className : 'ol-rotate';

      const label = options.label !== undefined ? options.label : '\u21E7';

      const compassClassName =
        options.compassClassName !== undefined
          ? options.compassClassName
          : 'ol-compass';

      /**
       * @type {HTMLElement}
       * @private
       */
      this.label_ = null;

      if (typeof label === 'string') {
        this.label_ = document.createElement('span');
        this.label_.className = compassClassName;
        this.label_.textContent = label;
      } else {
        this.label_ = label;
        this.label_.classList.add(compassClassName);
      }

      const tipLabel = options.tipLabel ? options.tipLabel : 'Reset rotation';

      const button = document.createElement('button');
      button.className = className + '-reset';
      button.setAttribute('type', 'button');
      button.title = tipLabel;
      button.appendChild(this.label_);

      button.addEventListener(
        EventType$1.CLICK,
        this.handleClick_.bind(this),
        false
      );

      const cssClasses =
        className + ' ' + CLASS_UNSELECTABLE + ' ' + CLASS_CONTROL;
      const element = this.element;
      element.className = cssClasses;
      element.appendChild(button);

      this.callResetNorth_ = options.resetNorth ? options.resetNorth : undefined;

      /**
       * @type {number}
       * @private
       */
      this.duration_ = options.duration !== undefined ? options.duration : 250;

      /**
       * @type {boolean}
       * @private
       */
      this.autoHide_ = options.autoHide !== undefined ? options.autoHide : true;

      /**
       * @private
       * @type {number|undefined}
       */
      this.rotation_ = undefined;

      if (this.autoHide_) {
        this.element.classList.add(CLASS_HIDDEN);
      }
    }

    /**
     * @param {MouseEvent} event The event to handle
     * @private
     */
    handleClick_(event) {
      event.preventDefault();
      if (this.callResetNorth_ !== undefined) {
        this.callResetNorth_();
      } else {
        this.resetNorth_();
      }
    }

    /**
     * @private
     */
    resetNorth_() {
      const map = this.getMap();
      const view = map.getView();
      if (!view) {
        // the map does not have a view, so we can't act
        // upon it
        return;
      }
      const rotation = view.getRotation();
      if (rotation !== undefined) {
        if (this.duration_ > 0 && rotation % (2 * Math.PI) !== 0) {
          view.animate({
            rotation: 0,
            duration: this.duration_,
            easing: easeOut,
          });
        } else {
          view.setRotation(0);
        }
      }
    }

    /**
     * Update the rotate control element.
     * @param {import("../MapEvent.js").default} mapEvent Map event.
     * @override
     */
    render(mapEvent) {
      const frameState = mapEvent.frameState;
      if (!frameState) {
        return;
      }
      const rotation = frameState.viewState.rotation;
      if (rotation != this.rotation_) {
        const transform = 'rotate(' + rotation + 'rad)';
        if (this.autoHide_) {
          const contains = this.element.classList.contains(CLASS_HIDDEN);
          if (!contains && rotation === 0) {
            this.element.classList.add(CLASS_HIDDEN);
          } else if (contains && rotation !== 0) {
            this.element.classList.remove(CLASS_HIDDEN);
          }
        }
        this.label_.style.transform = transform;
      }
      this.rotation_ = rotation;
    }
  }

  /**
   * @module ol/control/Zoom
   */

  /**
   * @typedef {Object} Options
   * @property {number} [duration=250] Animation duration in milliseconds.
   * @property {string} [className='ol-zoom'] CSS class name.
   * @property {string} [zoomInClassName=className + '-in'] CSS class name for the zoom-in button.
   * @property {string} [zoomOutClassName=className + '-out'] CSS class name for the zoom-out button.
   * @property {string|HTMLElement} [zoomInLabel='+'] Text label to use for the zoom-in
   * button. Instead of text, also an element (e.g. a `span` element) can be used.
   * @property {string|HTMLElement} [zoomOutLabel='–'] Text label to use for the zoom-out button.
   * Instead of text, also an element (e.g. a `span` element) can be used.
   * @property {string} [zoomInTipLabel='Zoom in'] Text label to use for the button tip.
   * @property {string} [zoomOutTipLabel='Zoom out'] Text label to use for the button tip.
   * @property {number} [delta=1] The zoom delta applied on each click.
   * @property {HTMLElement|string} [target] Specify a target if you want the control to be
   * rendered outside of the map's viewport.
   */

  /**
   * @classdesc
   * A control with 2 buttons, one for zoom in and one for zoom out.
   * This control is one of the default controls of a map. To style this control
   * use css selectors `.ol-zoom-in` and `.ol-zoom-out`.
   *
   * @api
   */
  class Zoom extends Control {
    /**
     * @param {Options} [options] Zoom options.
     */
    constructor(options) {
      options = options ? options : {};

      super({
        element: document.createElement('div'),
        target: options.target,
      });

      const className =
        options.className !== undefined ? options.className : 'ol-zoom';

      const delta = options.delta !== undefined ? options.delta : 1;

      const zoomInClassName =
        options.zoomInClassName !== undefined
          ? options.zoomInClassName
          : className + '-in';

      const zoomOutClassName =
        options.zoomOutClassName !== undefined
          ? options.zoomOutClassName
          : className + '-out';

      const zoomInLabel =
        options.zoomInLabel !== undefined ? options.zoomInLabel : '+';
      const zoomOutLabel =
        options.zoomOutLabel !== undefined ? options.zoomOutLabel : '\u2013';

      const zoomInTipLabel =
        options.zoomInTipLabel !== undefined ? options.zoomInTipLabel : 'Zoom in';
      const zoomOutTipLabel =
        options.zoomOutTipLabel !== undefined
          ? options.zoomOutTipLabel
          : 'Zoom out';

      const inElement = document.createElement('button');
      inElement.className = zoomInClassName;
      inElement.setAttribute('type', 'button');
      inElement.title = zoomInTipLabel;
      inElement.appendChild(
        typeof zoomInLabel === 'string'
          ? document.createTextNode(zoomInLabel)
          : zoomInLabel
      );

      inElement.addEventListener(
        EventType$1.CLICK,
        this.handleClick_.bind(this, delta),
        false
      );

      const outElement = document.createElement('button');
      outElement.className = zoomOutClassName;
      outElement.setAttribute('type', 'button');
      outElement.title = zoomOutTipLabel;
      outElement.appendChild(
        typeof zoomOutLabel === 'string'
          ? document.createTextNode(zoomOutLabel)
          : zoomOutLabel
      );

      outElement.addEventListener(
        EventType$1.CLICK,
        this.handleClick_.bind(this, -delta),
        false
      );

      const cssClasses =
        className + ' ' + CLASS_UNSELECTABLE + ' ' + CLASS_CONTROL;
      const element = this.element;
      element.className = cssClasses;
      element.appendChild(inElement);
      element.appendChild(outElement);

      /**
       * @type {number}
       * @private
       */
      this.duration_ = options.duration !== undefined ? options.duration : 250;
    }

    /**
     * @param {number} delta Zoom delta.
     * @param {MouseEvent} event The event to handle
     * @private
     */
    handleClick_(delta, event) {
      event.preventDefault();
      this.zoomByDelta_(delta);
    }

    /**
     * @param {number} delta Zoom delta.
     * @private
     */
    zoomByDelta_(delta) {
      const map = this.getMap();
      const view = map.getView();
      if (!view) {
        // the map does not have a view, so we can't act
        // upon it
        return;
      }
      const currentZoom = view.getZoom();
      if (currentZoom !== undefined) {
        const newZoom = view.getConstrainedZoom(currentZoom + delta);
        if (this.duration_ > 0) {
          if (view.getAnimating()) {
            view.cancelAnimations();
          }
          view.animate({
            zoom: newZoom,
            duration: this.duration_,
            easing: easeOut,
          });
        } else {
          view.setZoom(newZoom);
        }
      }
    }
  }

  /**
   * @module ol/control/defaults
   */

  /**
   * @typedef {Object} DefaultsOptions
   * @property {boolean} [attribution=true] Include
   * {@link module:ol/control/Attribution~Attribution}.
   * @property {import("./Attribution.js").Options} [attributionOptions]
   * Options for {@link module:ol/control/Attribution~Attribution}.
   * @property {boolean} [rotate=true] Include
   * {@link module:ol/control/Rotate~Rotate}.
   * @property {import("./Rotate.js").Options} [rotateOptions] Options
   * for {@link module:ol/control/Rotate~Rotate}.
   * @property {boolean} [zoom] Include {@link module:ol/control/Zoom~Zoom}.
   * @property {import("./Zoom.js").Options} [zoomOptions] Options for
   * {@link module:ol/control/Zoom~Zoom}.
   */

  /**
   * Set of controls included in maps by default. Unless configured otherwise,
   * this returns a collection containing an instance of each of the following
   * controls:
   * * {@link module:ol/control/Zoom~Zoom}
   * * {@link module:ol/control/Rotate~Rotate}
   * * {@link module:ol/control/Attribution~Attribution}
   *
   * @param {DefaultsOptions} [options] Options for the default controls.
   * @return {Collection<import("./Control.js").default>} A collection of controls
   * to be used with the {@link module:ol/Map~Map} constructor's `controls` option.
   * @api
   */
  function defaults$1(options) {
    options = options ? options : {};

    /** @type {Collection<import("./Control.js").default>} */
    const controls = new Collection();

    const zoomControl = options.zoom !== undefined ? options.zoom : true;
    if (zoomControl) {
      controls.push(new Zoom(options.zoomOptions));
    }

    const rotateControl = options.rotate !== undefined ? options.rotate : true;
    if (rotateControl) {
      controls.push(new Rotate(options.rotateOptions));
    }

    const attributionControl =
      options.attribution !== undefined ? options.attribution : true;
    if (attributionControl) {
      controls.push(new Attribution(options.attributionOptions));
    }

    return controls;
  }

  /**
   * @module ol/interaction/Property
   */

  /**
   * @enum {string}
   */
  var InteractionProperty = {
    ACTIVE: 'active',
  };

  /**
   * @module ol/interaction/Interaction
   */

  /***
   * @template Return
   * @typedef {import("../Observable").OnSignature<import("../Observable").EventTypes, import("../events/Event.js").default, Return> &
   *   import("../Observable").OnSignature<import("../ObjectEventType").Types|
   *     'change:active', import("../Object").ObjectEvent, Return> &
   *   import("../Observable").CombinedOnSignature<import("../Observable").EventTypes|import("../ObjectEventType").Types|
   *     'change:active', Return>} InteractionOnSignature
   */

  /**
   * Object literal with config options for interactions.
   * @typedef {Object} InteractionOptions
   * @property {function(import("../MapBrowserEvent.js").default):boolean} handleEvent
   * Method called by the map to notify the interaction that a browser event was
   * dispatched to the map. If the function returns a falsy value, propagation of
   * the event to other interactions in the map's interactions chain will be
   * prevented (this includes functions with no explicit return). The interactions
   * are traversed in reverse order of the interactions collection of the map.
   */

  /**
   * @classdesc
   * Abstract base class; normally only used for creating subclasses and not
   * instantiated in apps.
   * User actions that change the state of the map. Some are similar to controls,
   * but are not associated with a DOM element.
   * For example, {@link module:ol/interaction/KeyboardZoom~KeyboardZoom} is
   * functionally the same as {@link module:ol/control/Zoom~Zoom}, but triggered
   * by a keyboard event not a button element event.
   * Although interactions do not have a DOM element, some of them do render
   * vectors and so are visible on the screen.
   * @api
   */
  class Interaction extends BaseObject {
    /**
     * @param {InteractionOptions} [options] Options.
     */
    constructor(options) {
      super();

      /***
       * @type {InteractionOnSignature<import("../events").EventsKey>}
       */
      this.on;

      /***
       * @type {InteractionOnSignature<import("../events").EventsKey>}
       */
      this.once;

      /***
       * @type {InteractionOnSignature<void>}
       */
      this.un;

      if (options && options.handleEvent) {
        this.handleEvent = options.handleEvent;
      }

      /**
       * @private
       * @type {import("../Map.js").default|null}
       */
      this.map_ = null;

      this.setActive(true);
    }

    /**
     * Return whether the interaction is currently active.
     * @return {boolean} `true` if the interaction is active, `false` otherwise.
     * @observable
     * @api
     */
    getActive() {
      return /** @type {boolean} */ (this.get(InteractionProperty.ACTIVE));
    }

    /**
     * Get the map associated with this interaction.
     * @return {import("../Map.js").default|null} Map.
     * @api
     */
    getMap() {
      return this.map_;
    }

    /**
     * Handles the {@link module:ol/MapBrowserEvent~MapBrowserEvent map browser event}.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Map browser event.
     * @return {boolean} `false` to stop event propagation.
     * @api
     */
    handleEvent(mapBrowserEvent) {
      return true;
    }

    /**
     * Activate or deactivate the interaction.
     * @param {boolean} active Active.
     * @observable
     * @api
     */
    setActive(active) {
      this.set(InteractionProperty.ACTIVE, active);
    }

    /**
     * Remove the interaction from its current map and attach it to the new map.
     * Subclasses may set up event handlers to get notified about changes to
     * the map here.
     * @param {import("../Map.js").default|null} map Map.
     */
    setMap(map) {
      this.map_ = map;
    }
  }

  /**
   * @param {import("../View.js").default} view View.
   * @param {import("../coordinate.js").Coordinate} delta Delta.
   * @param {number} [duration] Duration.
   */
  function pan(view, delta, duration) {
    const currentCenter = view.getCenterInternal();
    if (currentCenter) {
      const center = [currentCenter[0] + delta[0], currentCenter[1] + delta[1]];
      view.animateInternal({
        duration: duration !== undefined ? duration : 250,
        easing: linear,
        center: view.getConstrainedCenter(center),
      });
    }
  }

  /**
   * @param {import("../View.js").default} view View.
   * @param {number} delta Delta from previous zoom level.
   * @param {import("../coordinate.js").Coordinate} [anchor] Anchor coordinate in the user projection.
   * @param {number} [duration] Duration.
   */
  function zoomByDelta(view, delta, anchor, duration) {
    const currentZoom = view.getZoom();

    if (currentZoom === undefined) {
      return;
    }

    const newZoom = view.getConstrainedZoom(currentZoom + delta);
    const newResolution = view.getResolutionForZoom(newZoom);

    if (view.getAnimating()) {
      view.cancelAnimations();
    }
    view.animate({
      resolution: newResolution,
      anchor: anchor,
      duration: duration !== undefined ? duration : 250,
      easing: easeOut,
    });
  }

  /**
   * @module ol/interaction/DoubleClickZoom
   */

  /**
   * @typedef {Object} Options
   * @property {number} [duration=250] Animation duration in milliseconds.
   * @property {number} [delta=1] The zoom delta applied on each double click.
   */

  /**
   * @classdesc
   * Allows the user to zoom by double-clicking on the map.
   * @api
   */
  class DoubleClickZoom extends Interaction {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      super();

      options = options ? options : {};

      /**
       * @private
       * @type {number}
       */
      this.delta_ = options.delta ? options.delta : 1;

      /**
       * @private
       * @type {number}
       */
      this.duration_ = options.duration !== undefined ? options.duration : 250;
    }

    /**
     * Handles the {@link module:ol/MapBrowserEvent~MapBrowserEvent map browser event} (if it was a
     * doubleclick) and eventually zooms the map.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Map browser event.
     * @return {boolean} `false` to stop event propagation.
     */
    handleEvent(mapBrowserEvent) {
      let stopEvent = false;
      if (mapBrowserEvent.type == MapBrowserEventType.DBLCLICK) {
        const browserEvent = /** @type {MouseEvent} */ (
          mapBrowserEvent.originalEvent
        );
        const map = mapBrowserEvent.map;
        const anchor = mapBrowserEvent.coordinate;
        const delta = browserEvent.shiftKey ? -this.delta_ : this.delta_;
        const view = map.getView();
        zoomByDelta(view, delta, anchor, this.duration_);
        browserEvent.preventDefault();
        stopEvent = true;
      }
      return !stopEvent;
    }
  }

  /**
   * @module ol/interaction/Pointer
   */

  /**
   * @typedef {Object} Options
   * @property {function(import("../MapBrowserEvent.js").default):boolean} [handleDownEvent]
   * Function handling "down" events. If the function returns `true` then a drag
   * sequence is started.
   * @property {function(import("../MapBrowserEvent.js").default):void} [handleDragEvent]
   * Function handling "drag" events. This function is called on "move" events
   * during a drag sequence.
   * @property {function(import("../MapBrowserEvent.js").default):boolean} [handleEvent]
   * Method called by the map to notify the interaction that a browser event was
   * dispatched to the map. The function may return `false` to prevent the
   * propagation of the event to other interactions in the map's interactions
   * chain.
   * @property {function(import("../MapBrowserEvent.js").default):void} [handleMoveEvent]
   * Function handling "move" events. This function is called on "move" events.
   * This functions is also called during a drag sequence, so during a drag
   * sequence both the `handleDragEvent` function and this function are called.
   * If `handleDownEvent` is defined and it returns true this function will not
   * be called during a drag sequence.
   * @property {function(import("../MapBrowserEvent.js").default):boolean} [handleUpEvent]
   *  Function handling "up" events. If the function returns `false` then the
   * current drag sequence is stopped.
   * @property {function(boolean):boolean} [stopDown]
   * Should the down event be propagated to other interactions, or should be
   * stopped?
   */

  /**
   * @classdesc
   * Base class that calls user-defined functions on `down`, `move` and `up`
   * events. This class also manages "drag sequences".
   *
   * When the `handleDownEvent` user function returns `true` a drag sequence is
   * started. During a drag sequence the `handleDragEvent` user function is
   * called on `move` events. The drag sequence ends when the `handleUpEvent`
   * user function is called and returns `false`.
   * @api
   */
  class PointerInteraction extends Interaction {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      options = options ? options : {};

      super(
        /** @type {import("./Interaction.js").InteractionOptions} */ (options)
      );

      if (options.handleDownEvent) {
        this.handleDownEvent = options.handleDownEvent;
      }

      if (options.handleDragEvent) {
        this.handleDragEvent = options.handleDragEvent;
      }

      if (options.handleMoveEvent) {
        this.handleMoveEvent = options.handleMoveEvent;
      }

      if (options.handleUpEvent) {
        this.handleUpEvent = options.handleUpEvent;
      }

      if (options.stopDown) {
        this.stopDown = options.stopDown;
      }

      /**
       * @type {boolean}
       * @protected
       */
      this.handlingDownUpSequence = false;

      /**
       * @type {Array<PointerEvent>}
       * @protected
       */
      this.targetPointers = [];
    }

    /**
     * Returns the current number of pointers involved in the interaction,
     * e.g. `2` when two fingers are used.
     * @return {number} The number of pointers.
     * @api
     */
    getPointerCount() {
      return this.targetPointers.length;
    }

    /**
     * Handle pointer down events.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Event.
     * @return {boolean} If the event was consumed.
     * @protected
     */
    handleDownEvent(mapBrowserEvent) {
      return false;
    }

    /**
     * Handle pointer drag events.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Event.
     * @protected
     */
    handleDragEvent(mapBrowserEvent) {}

    /**
     * Handles the {@link module:ol/MapBrowserEvent~MapBrowserEvent map browser event} and may call into
     * other functions, if event sequences like e.g. 'drag' or 'down-up' etc. are
     * detected.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Map browser event.
     * @return {boolean} `false` to stop event propagation.
     * @api
     */
    handleEvent(mapBrowserEvent) {
      if (!mapBrowserEvent.originalEvent) {
        return true;
      }

      let stopEvent = false;
      this.updateTrackedPointers_(mapBrowserEvent);
      if (this.handlingDownUpSequence) {
        if (mapBrowserEvent.type == MapBrowserEventType.POINTERDRAG) {
          this.handleDragEvent(mapBrowserEvent);
          // prevent page scrolling during dragging
          mapBrowserEvent.originalEvent.preventDefault();
        } else if (mapBrowserEvent.type == MapBrowserEventType.POINTERUP) {
          const handledUp = this.handleUpEvent(mapBrowserEvent);
          this.handlingDownUpSequence =
            handledUp && this.targetPointers.length > 0;
        }
      } else {
        if (mapBrowserEvent.type == MapBrowserEventType.POINTERDOWN) {
          const handled = this.handleDownEvent(mapBrowserEvent);
          this.handlingDownUpSequence = handled;
          stopEvent = this.stopDown(handled);
        } else if (mapBrowserEvent.type == MapBrowserEventType.POINTERMOVE) {
          this.handleMoveEvent(mapBrowserEvent);
        }
      }
      return !stopEvent;
    }

    /**
     * Handle pointer move events.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Event.
     * @protected
     */
    handleMoveEvent(mapBrowserEvent) {}

    /**
     * Handle pointer up events.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Event.
     * @return {boolean} If the event was consumed.
     * @protected
     */
    handleUpEvent(mapBrowserEvent) {
      return false;
    }

    /**
     * This function is used to determine if "down" events should be propagated
     * to other interactions or should be stopped.
     * @param {boolean} handled Was the event handled by the interaction?
     * @return {boolean} Should the `down` event be stopped?
     */
    stopDown(handled) {
      return handled;
    }

    /**
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Event.
     * @private
     */
    updateTrackedPointers_(mapBrowserEvent) {
      if (mapBrowserEvent.activePointers) {
        this.targetPointers = mapBrowserEvent.activePointers;
      }
    }
  }

  /**
   * @param {Array<PointerEvent>} pointerEvents List of events.
   * @return {import("../pixel.js").Pixel} Centroid pixel.
   */
  function centroid(pointerEvents) {
    const length = pointerEvents.length;
    let clientX = 0;
    let clientY = 0;
    for (let i = 0; i < length; i++) {
      clientX += pointerEvents[i].clientX;
      clientY += pointerEvents[i].clientY;
    }
    return [clientX / length, clientY / length];
  }

  /**
   * @module ol/events/condition
   */

  /**
   * A function that takes an {@link module:ol/MapBrowserEvent~MapBrowserEvent} and returns a
   * `{boolean}`. If the condition is met, true should be returned.
   *
   * @typedef {function(this: ?, import("../MapBrowserEvent.js").default): boolean} Condition
   */

  /**
   * Creates a condition function that passes when all provided conditions pass.
   * @param {...Condition} var_args Conditions to check.
   * @return {Condition} Condition function.
   */
  function all$1(var_args) {
    const conditions = arguments;
    /**
     * @param {import("../MapBrowserEvent.js").default} event Event.
     * @return {boolean} All conditions passed.
     */
    return function (event) {
      let pass = true;
      for (let i = 0, ii = conditions.length; i < ii; ++i) {
        pass = pass && conditions[i](event);
        if (!pass) {
          break;
        }
      }
      return pass;
    };
  }

  /**
   * Return `true` if only the alt-key is pressed, `false` otherwise (e.g. when
   * additionally the shift-key is pressed).
   *
   * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Map browser event.
   * @return {boolean} True if only the alt key is pressed.
   * @api
   */
  const altKeyOnly = function (mapBrowserEvent) {
    const originalEvent = /** @type {KeyboardEvent|MouseEvent|TouchEvent} */ (
      mapBrowserEvent.originalEvent
    );
    return (
      originalEvent.altKey &&
      !(originalEvent.metaKey || originalEvent.ctrlKey) &&
      !originalEvent.shiftKey
    );
  };

  /**
   * Return `true` if only the alt-key and shift-key is pressed, `false` otherwise
   * (e.g. when additionally the platform-modifier-key is pressed).
   *
   * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Map browser event.
   * @return {boolean} True if only the alt and shift keys are pressed.
   * @api
   */
  const altShiftKeysOnly = function (mapBrowserEvent) {
    const originalEvent = /** @type {KeyboardEvent|MouseEvent|TouchEvent} */ (
      mapBrowserEvent.originalEvent
    );
    return (
      originalEvent.altKey &&
      !(originalEvent.metaKey || originalEvent.ctrlKey) &&
      originalEvent.shiftKey
    );
  };

  /**
   * Return `true` if the map has the focus. This condition requires a map target
   * element with a `tabindex` attribute, e.g. `<div id="map" tabindex="1">`.
   *
   * @param {import("../MapBrowserEvent.js").default} event Map browser event.
   * @return {boolean} The map has the focus.
   * @api
   */
  const focus = function (event) {
    const targetElement = event.map.getTargetElement();
    const activeElement = event.map.getOwnerDocument().activeElement;
    return targetElement.contains(activeElement);
  };

  /**
   * Return `true` if the map has the focus or no 'tabindex' attribute set.
   *
   * @param {import("../MapBrowserEvent.js").default} event Map browser event.
   * @return {boolean} The map container has the focus or no 'tabindex' attribute.
   */
  const focusWithTabindex = function (event) {
    return event.map.getTargetElement().hasAttribute('tabindex')
      ? focus(event)
      : true;
  };

  /**
   * Return always true.
   *
   * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Map browser event.
   * @return {boolean} True.
   * @api
   */
  const always = TRUE;

  /**
   * Return `true` if the event is a `click` event, `false` otherwise.
   *
   * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Map browser event.
   * @return {boolean} True if the event is a map `click` event.
   * @api
   */
  const click = function (mapBrowserEvent) {
    return mapBrowserEvent.type == MapBrowserEventType.CLICK;
  };

  /**
   * Return `true` if the event has an "action"-producing mouse button.
   *
   * By definition, this includes left-click on windows/linux, and left-click
   * without the ctrl key on Macs.
   *
   * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Map browser event.
   * @return {boolean} The result.
   */
  const mouseActionButton = function (mapBrowserEvent) {
    const originalEvent = /** @type {MouseEvent} */ (
      mapBrowserEvent.originalEvent
    );
    return originalEvent.button == 0 && !(WEBKIT && MAC && originalEvent.ctrlKey);
  };

  /**
   * Return always false.
   *
   * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Map browser event.
   * @return {boolean} False.
   * @api
   */
  const never = FALSE;

  /**
   * Return `true` if the browser event is a `pointermove` event, `false`
   * otherwise.
   *
   * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Map browser event.
   * @return {boolean} True if the browser event is a `pointermove` event.
   * @api
   */
  const pointerMove = function (mapBrowserEvent) {
    return mapBrowserEvent.type == 'pointermove';
  };

  /**
   * Return `true` if the event is a map `singleclick` event, `false` otherwise.
   *
   * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Map browser event.
   * @return {boolean} True if the event is a map `singleclick` event.
   * @api
   */
  const singleClick = function (mapBrowserEvent) {
    return mapBrowserEvent.type == MapBrowserEventType.SINGLECLICK;
  };

  /**
   * Return `true` if the event is a map `dblclick` event, `false` otherwise.
   *
   * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Map browser event.
   * @return {boolean} True if the event is a map `dblclick` event.
   * @api
   */
  const doubleClick = function (mapBrowserEvent) {
    return mapBrowserEvent.type == MapBrowserEventType.DBLCLICK;
  };

  /**
   * Return `true` if no modifier key (alt-, shift- or platform-modifier-key) is
   * pressed.
   *
   * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Map browser event.
   * @return {boolean} True only if there no modifier keys are pressed.
   * @api
   */
  const noModifierKeys = function (mapBrowserEvent) {
    const originalEvent = /** @type {KeyboardEvent|MouseEvent|TouchEvent} */ (
      mapBrowserEvent.originalEvent
    );
    return (
      !originalEvent.altKey &&
      !(originalEvent.metaKey || originalEvent.ctrlKey) &&
      !originalEvent.shiftKey
    );
  };

  /**
   * Return `true` if only the platform-modifier-key (the meta-key on Mac,
   * ctrl-key otherwise) is pressed, `false` otherwise (e.g. when additionally
   * the shift-key is pressed).
   *
   * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Map browser event.
   * @return {boolean} True if only the platform modifier key is pressed.
   * @api
   */
  const platformModifierKeyOnly = function (mapBrowserEvent) {
    const originalEvent = /** @type {KeyboardEvent|MouseEvent|TouchEvent} */ (
      mapBrowserEvent.originalEvent
    );
    return (
      !originalEvent.altKey &&
      (MAC ? originalEvent.metaKey : originalEvent.ctrlKey) &&
      !originalEvent.shiftKey
    );
  };

  /**
   * Return `true` if only the shift-key is pressed, `false` otherwise (e.g. when
   * additionally the alt-key is pressed).
   *
   * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Map browser event.
   * @return {boolean} True if only the shift key is pressed.
   * @api
   */
  const shiftKeyOnly = function (mapBrowserEvent) {
    const originalEvent = /** @type {KeyboardEvent|MouseEvent|TouchEvent} */ (
      mapBrowserEvent.originalEvent
    );
    return (
      !originalEvent.altKey &&
      !(originalEvent.metaKey || originalEvent.ctrlKey) &&
      originalEvent.shiftKey
    );
  };

  /**
   * Return `true` if the target element is not editable, i.e. not an `input`,
   * `select`, or `textarea` element and no `contenteditable` attribute is
   * set or inherited, `false` otherwise.
   *
   * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Map browser event.
   * @return {boolean} True only if the target element is not editable.
   * @api
   */
  const targetNotEditable = function (mapBrowserEvent) {
    const originalEvent = /** @type {KeyboardEvent|MouseEvent|TouchEvent} */ (
      mapBrowserEvent.originalEvent
    );
    const tagName = /** @type {Element} */ (originalEvent.target).tagName;
    return (
      tagName !== 'INPUT' &&
      tagName !== 'SELECT' &&
      tagName !== 'TEXTAREA' &&
      // `isContentEditable` is only available on `HTMLElement`, but it may also be a
      // different type like `SVGElement`.
      // @ts-ignore
      !originalEvent.target.isContentEditable
    );
  };

  /**
   * Return `true` if the event originates from a mouse device.
   *
   * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Map browser event.
   * @return {boolean} True if the event originates from a mouse device.
   * @api
   */
  const mouseOnly = function (mapBrowserEvent) {
    const pointerEvent = /** @type {import("../MapBrowserEvent").default} */ (
      mapBrowserEvent
    ).originalEvent;
    assert(pointerEvent !== undefined, 56); // mapBrowserEvent must originate from a pointer event
    // see https://www.w3.org/TR/pointerevents/#widl-PointerEvent-pointerType
    return pointerEvent.pointerType == 'mouse';
  };

  /**
   * Return `true` if the event originates from a touchable device.
   *
   * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Map browser event.
   * @return {boolean} True if the event originates from a touchable device.
   * @api
   */
  const touchOnly = function (mapBrowserEvent) {
    const pointerEvt = /** @type {import("../MapBrowserEvent").default} */ (
      mapBrowserEvent
    ).originalEvent;
    assert(pointerEvt !== undefined, 56); // mapBrowserEvent must originate from a pointer event
    // see https://www.w3.org/TR/pointerevents/#widl-PointerEvent-pointerType
    return pointerEvt.pointerType === 'touch';
  };

  /**
   * Return `true` if the event originates from a digital pen.
   *
   * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Map browser event.
   * @return {boolean} True if the event originates from a digital pen.
   * @api
   */
  const penOnly = function (mapBrowserEvent) {
    const pointerEvt = /** @type {import("../MapBrowserEvent").default} */ (
      mapBrowserEvent
    ).originalEvent;
    assert(pointerEvt !== undefined, 56); // mapBrowserEvent must originate from a pointer event
    // see https://www.w3.org/TR/pointerevents/#widl-PointerEvent-pointerType
    return pointerEvt.pointerType === 'pen';
  };

  /**
   * Return `true` if the event originates from a primary pointer in
   * contact with the surface or if the left mouse button is pressed.
   * See https://www.w3.org/TR/pointerevents/#button-states.
   *
   * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Map browser event.
   * @return {boolean} True if the event originates from a primary pointer.
   * @api
   */
  const primaryAction = function (mapBrowserEvent) {
    const pointerEvent = /** @type {import("../MapBrowserEvent").default} */ (
      mapBrowserEvent
    ).originalEvent;
    assert(pointerEvent !== undefined, 56); // mapBrowserEvent must originate from a pointer event
    return pointerEvent.isPrimary && pointerEvent.button === 0;
  };

  /**
   * @module ol/interaction/DragPan
   */

  /**
   * @typedef {Object} Options
   * @property {import("../events/condition.js").Condition} [condition] A function that takes an {@link module:ol/MapBrowserEvent~MapBrowserEvent} and returns a boolean
   * to indicate whether that event should be handled.
   * Default is {@link module:ol/events/condition.noModifierKeys} and {@link module:ol/events/condition.primaryAction}.
   * @property {boolean} [onFocusOnly=false] When the map's target has a `tabindex` attribute set,
   * the interaction will only handle events when the map has the focus.
   * @property {import("../Kinetic.js").default} [kinetic] Kinetic inertia to apply to the pan.
   */

  /**
   * @classdesc
   * Allows the user to pan the map by dragging the map.
   * @api
   */
  class DragPan extends PointerInteraction {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      super({
        stopDown: FALSE,
      });

      options = options ? options : {};

      /**
       * @private
       * @type {import("../Kinetic.js").default|undefined}
       */
      this.kinetic_ = options.kinetic;

      /**
       * @type {import("../pixel.js").Pixel}
       */
      this.lastCentroid = null;

      /**
       * @type {number}
       */
      this.lastPointersCount_;

      /**
       * @type {boolean}
       */
      this.panning_ = false;

      const condition = options.condition
        ? options.condition
        : all$1(noModifierKeys, primaryAction);

      /**
       * @private
       * @type {import("../events/condition.js").Condition}
       */
      this.condition_ = options.onFocusOnly
        ? all$1(focusWithTabindex, condition)
        : condition;

      /**
       * @private
       * @type {boolean}
       */
      this.noKinetic_ = false;
    }

    /**
     * Handle pointer drag events.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Event.
     */
    handleDragEvent(mapBrowserEvent) {
      if (!this.panning_) {
        this.panning_ = true;
        this.getMap().getView().beginInteraction();
      }
      const targetPointers = this.targetPointers;
      const centroid$1 = centroid(targetPointers);
      if (targetPointers.length == this.lastPointersCount_) {
        if (this.kinetic_) {
          this.kinetic_.update(centroid$1[0], centroid$1[1]);
        }
        if (this.lastCentroid) {
          const delta = [
            this.lastCentroid[0] - centroid$1[0],
            centroid$1[1] - this.lastCentroid[1],
          ];
          const map = mapBrowserEvent.map;
          const view = map.getView();
          scale$2(delta, view.getResolution());
          rotate$1(delta, view.getRotation());
          view.adjustCenterInternal(delta);
        }
      } else if (this.kinetic_) {
        // reset so we don't overestimate the kinetic energy after
        // after one finger down, tiny drag, second finger down
        this.kinetic_.begin();
      }
      this.lastCentroid = centroid$1;
      this.lastPointersCount_ = targetPointers.length;
      mapBrowserEvent.originalEvent.preventDefault();
    }

    /**
     * Handle pointer up events.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Event.
     * @return {boolean} If the event was consumed.
     */
    handleUpEvent(mapBrowserEvent) {
      const map = mapBrowserEvent.map;
      const view = map.getView();
      if (this.targetPointers.length === 0) {
        if (!this.noKinetic_ && this.kinetic_ && this.kinetic_.end()) {
          const distance = this.kinetic_.getDistance();
          const angle = this.kinetic_.getAngle();
          const center = view.getCenterInternal();
          const centerpx = map.getPixelFromCoordinateInternal(center);
          const dest = map.getCoordinateFromPixelInternal([
            centerpx[0] - distance * Math.cos(angle),
            centerpx[1] - distance * Math.sin(angle),
          ]);
          view.animateInternal({
            center: view.getConstrainedCenter(dest),
            duration: 500,
            easing: easeOut,
          });
        }
        if (this.panning_) {
          this.panning_ = false;
          view.endInteraction();
        }
        return false;
      } else {
        if (this.kinetic_) {
          // reset so we don't overestimate the kinetic energy after
          // after one finger up, tiny drag, second finger up
          this.kinetic_.begin();
        }
        this.lastCentroid = null;
        return true;
      }
    }

    /**
     * Handle pointer down events.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Event.
     * @return {boolean} If the event was consumed.
     */
    handleDownEvent(mapBrowserEvent) {
      if (this.targetPointers.length > 0 && this.condition_(mapBrowserEvent)) {
        const map = mapBrowserEvent.map;
        const view = map.getView();
        this.lastCentroid = null;
        // stop any current animation
        if (view.getAnimating()) {
          view.cancelAnimations();
        }
        if (this.kinetic_) {
          this.kinetic_.begin();
        }
        // No kinetic as soon as more than one pointer on the screen is
        // detected. This is to prevent nasty pans after pinch.
        this.noKinetic_ = this.targetPointers.length > 1;
        return true;
      } else {
        return false;
      }
    }
  }

  /**
   * @module ol/interaction/DragRotate
   */

  /**
   * @typedef {Object} Options
   * @property {import("../events/condition.js").Condition} [condition] A function that takes an
   * {@link module:ol/MapBrowserEvent~MapBrowserEvent} and returns a boolean
   * to indicate whether that event should be handled.
   * Default is {@link module:ol/events/condition.altShiftKeysOnly}.
   * @property {number} [duration=250] Animation duration in milliseconds.
   */

  /**
   * @classdesc
   * Allows the user to rotate the map by clicking and dragging on the map,
   * normally combined with an {@link module:ol/events/condition} that limits
   * it to when the alt and shift keys are held down.
   *
   * This interaction is only supported for mouse devices.
   * @api
   */
  class DragRotate extends PointerInteraction {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      options = options ? options : {};

      super({
        stopDown: FALSE,
      });

      /**
       * @private
       * @type {import("../events/condition.js").Condition}
       */
      this.condition_ = options.condition ? options.condition : altShiftKeysOnly;

      /**
       * @private
       * @type {number|undefined}
       */
      this.lastAngle_ = undefined;

      /**
       * @private
       * @type {number}
       */
      this.duration_ = options.duration !== undefined ? options.duration : 250;
    }

    /**
     * Handle pointer drag events.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Event.
     */
    handleDragEvent(mapBrowserEvent) {
      if (!mouseOnly(mapBrowserEvent)) {
        return;
      }

      const map = mapBrowserEvent.map;
      const view = map.getView();
      if (view.getConstraints().rotation === disable) {
        return;
      }
      const size = map.getSize();
      const offset = mapBrowserEvent.pixel;
      const theta = Math.atan2(size[1] / 2 - offset[1], offset[0] - size[0] / 2);
      if (this.lastAngle_ !== undefined) {
        const delta = theta - this.lastAngle_;
        view.adjustRotationInternal(-delta);
      }
      this.lastAngle_ = theta;
    }

    /**
     * Handle pointer up events.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Event.
     * @return {boolean} If the event was consumed.
     */
    handleUpEvent(mapBrowserEvent) {
      if (!mouseOnly(mapBrowserEvent)) {
        return true;
      }

      const map = mapBrowserEvent.map;
      const view = map.getView();
      view.endInteraction(this.duration_);
      return false;
    }

    /**
     * Handle pointer down events.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Event.
     * @return {boolean} If the event was consumed.
     */
    handleDownEvent(mapBrowserEvent) {
      if (!mouseOnly(mapBrowserEvent)) {
        return false;
      }

      if (
        mouseActionButton(mapBrowserEvent) &&
        this.condition_(mapBrowserEvent)
      ) {
        const map = mapBrowserEvent.map;
        map.getView().beginInteraction();
        this.lastAngle_ = undefined;
        return true;
      } else {
        return false;
      }
    }
  }

  /**
   * @module ol/render/Box
   */

  class RenderBox extends Disposable {
    /**
     * @param {string} className CSS class name.
     */
    constructor(className) {
      super();

      /**
       * @type {import("../geom/Polygon.js").default}
       * @private
       */
      this.geometry_ = null;

      /**
       * @type {HTMLDivElement}
       * @private
       */
      this.element_ = document.createElement('div');
      this.element_.style.position = 'absolute';
      this.element_.style.pointerEvents = 'auto';
      this.element_.className = 'ol-box ' + className;

      /**
       * @private
       * @type {import("../Map.js").default|null}
       */
      this.map_ = null;

      /**
       * @private
       * @type {import("../pixel.js").Pixel}
       */
      this.startPixel_ = null;

      /**
       * @private
       * @type {import("../pixel.js").Pixel}
       */
      this.endPixel_ = null;
    }

    /**
     * Clean up.
     */
    disposeInternal() {
      this.setMap(null);
    }

    /**
     * @private
     */
    render_() {
      const startPixel = this.startPixel_;
      const endPixel = this.endPixel_;
      const px = 'px';
      const style = this.element_.style;
      style.left = Math.min(startPixel[0], endPixel[0]) + px;
      style.top = Math.min(startPixel[1], endPixel[1]) + px;
      style.width = Math.abs(endPixel[0] - startPixel[0]) + px;
      style.height = Math.abs(endPixel[1] - startPixel[1]) + px;
    }

    /**
     * @param {import("../Map.js").default|null} map Map.
     */
    setMap(map) {
      if (this.map_) {
        this.map_.getOverlayContainer().removeChild(this.element_);
        const style = this.element_.style;
        style.left = 'inherit';
        style.top = 'inherit';
        style.width = 'inherit';
        style.height = 'inherit';
      }
      this.map_ = map;
      if (this.map_) {
        this.map_.getOverlayContainer().appendChild(this.element_);
      }
    }

    /**
     * @param {import("../pixel.js").Pixel} startPixel Start pixel.
     * @param {import("../pixel.js").Pixel} endPixel End pixel.
     */
    setPixels(startPixel, endPixel) {
      this.startPixel_ = startPixel;
      this.endPixel_ = endPixel;
      this.createOrUpdateGeometry();
      this.render_();
    }

    /**
     * Creates or updates the cached geometry.
     */
    createOrUpdateGeometry() {
      const startPixel = this.startPixel_;
      const endPixel = this.endPixel_;
      const pixels = [
        startPixel,
        [startPixel[0], endPixel[1]],
        endPixel,
        [endPixel[0], startPixel[1]],
      ];
      const coordinates = pixels.map(
        this.map_.getCoordinateFromPixelInternal,
        this.map_
      );
      // close the polygon
      coordinates[4] = coordinates[0].slice();
      if (!this.geometry_) {
        this.geometry_ = new Polygon([coordinates]);
      } else {
        this.geometry_.setCoordinates([coordinates]);
      }
    }

    /**
     * @return {import("../geom/Polygon.js").default} Geometry.
     */
    getGeometry() {
      return this.geometry_;
    }
  }

  /**
   * @module ol/interaction/DragBox
   */

  /**
   * A function that takes a {@link module:ol/MapBrowserEvent~MapBrowserEvent} and two
   * {@link module:ol/pixel~Pixel}s and returns a `{boolean}`. If the condition is met,
   * true should be returned.
   * @typedef {function(this: ?, import("../MapBrowserEvent.js").default, import("../pixel.js").Pixel, import("../pixel.js").Pixel):boolean} EndCondition
   */

  /**
   * @typedef {Object} Options
   * @property {string} [className='ol-dragbox'] CSS class name for styling the box.
   * @property {import("../events/condition.js").Condition} [condition] A function that takes an {@link module:ol/MapBrowserEvent~MapBrowserEvent} and returns a boolean
   * to indicate whether that event should be handled.
   * Default is {@link ol/events/condition~mouseActionButton}.
   * @property {number} [minArea=64] The minimum area of the box in pixel, this value is used by the default
   * `boxEndCondition` function.
   * @property {EndCondition} [boxEndCondition] A function that takes a {@link module:ol/MapBrowserEvent~MapBrowserEvent} and two
   * {@link module:ol/pixel~Pixel}s to indicate whether a `boxend` event should be fired.
   * Default is `true` if the area of the box is bigger than the `minArea` option.
   * @property {function(this:DragBox, import("../MapBrowserEvent.js").default):void} [onBoxEnd] Code to execute just
   * before `boxend` is fired.
   */

  /**
   * @enum {string}
   */
  const DragBoxEventType = {
    /**
     * Triggered upon drag box start.
     * @event DragBoxEvent#boxstart
     * @api
     */
    BOXSTART: 'boxstart',

    /**
     * Triggered on drag when box is active.
     * @event DragBoxEvent#boxdrag
     * @api
     */
    BOXDRAG: 'boxdrag',

    /**
     * Triggered upon drag box end.
     * @event DragBoxEvent#boxend
     * @api
     */
    BOXEND: 'boxend',

    /**
     * Triggered upon drag box canceled.
     * @event DragBoxEvent#boxcancel
     * @api
     */
    BOXCANCEL: 'boxcancel',
  };

  /**
   * @classdesc
   * Events emitted by {@link module:ol/interaction/DragBox~DragBox} instances are instances of
   * this type.
   */
  class DragBoxEvent extends BaseEvent {
    /**
     * @param {string} type The event type.
     * @param {import("../coordinate.js").Coordinate} coordinate The event coordinate.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Originating event.
     */
    constructor(type, coordinate, mapBrowserEvent) {
      super(type);

      /**
       * The coordinate of the drag event.
       * @const
       * @type {import("../coordinate.js").Coordinate}
       * @api
       */
      this.coordinate = coordinate;

      /**
       * @const
       * @type {import("../MapBrowserEvent.js").default}
       * @api
       */
      this.mapBrowserEvent = mapBrowserEvent;
    }
  }

  /***
   * @template Return
   * @typedef {import("../Observable").OnSignature<import("../Observable").EventTypes, import("../events/Event.js").default, Return> &
   *   import("../Observable").OnSignature<import("../ObjectEventType").Types|
   *     'change:active', import("../Object").ObjectEvent, Return> &
   *   import("../Observable").OnSignature<'boxcancel'|'boxdrag'|'boxend'|'boxstart', DragBoxEvent, Return> &
   *   import("../Observable").CombinedOnSignature<import("../Observable").EventTypes|import("../ObjectEventType").Types|
   *     'change:active'|'boxcancel'|'boxdrag'|'boxend', Return>} DragBoxOnSignature
   */

  /**
   * @classdesc
   * Allows the user to draw a vector box by clicking and dragging on the map,
   * normally combined with an {@link module:ol/events/condition} that limits
   * it to when the shift or other key is held down. This is used, for example,
   * for zooming to a specific area of the map
   * (see {@link module:ol/interaction/DragZoom~DragZoom} and
   * {@link module:ol/interaction/DragRotateAndZoom~DragRotateAndZoom}).
   *
   * @fires DragBoxEvent
   * @api
   */
  class DragBox extends PointerInteraction {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      super();

      /***
       * @type {DragBoxOnSignature<import("../events").EventsKey>}
       */
      this.on;

      /***
       * @type {DragBoxOnSignature<import("../events").EventsKey>}
       */
      this.once;

      /***
       * @type {DragBoxOnSignature<void>}
       */
      this.un;

      options = options ? options : {};

      /**
       * @type {import("../render/Box.js").default}
       * @private
       */
      this.box_ = new RenderBox(options.className || 'ol-dragbox');

      /**
       * @type {number}
       * @private
       */
      this.minArea_ = options.minArea !== undefined ? options.minArea : 64;

      if (options.onBoxEnd) {
        this.onBoxEnd = options.onBoxEnd;
      }

      /**
       * @type {import("../pixel.js").Pixel}
       * @private
       */
      this.startPixel_ = null;

      /**
       * @private
       * @type {import("../events/condition.js").Condition}
       */
      this.condition_ = options.condition ? options.condition : mouseActionButton;

      /**
       * @private
       * @type {EndCondition}
       */
      this.boxEndCondition_ = options.boxEndCondition
        ? options.boxEndCondition
        : this.defaultBoxEndCondition;
    }

    /**
     * The default condition for determining whether the boxend event
     * should fire.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent The originating MapBrowserEvent
     *     leading to the box end.
     * @param {import("../pixel.js").Pixel} startPixel The starting pixel of the box.
     * @param {import("../pixel.js").Pixel} endPixel The end pixel of the box.
     * @return {boolean} Whether or not the boxend condition should be fired.
     */
    defaultBoxEndCondition(mapBrowserEvent, startPixel, endPixel) {
      const width = endPixel[0] - startPixel[0];
      const height = endPixel[1] - startPixel[1];
      return width * width + height * height >= this.minArea_;
    }

    /**
     * Returns geometry of last drawn box.
     * @return {import("../geom/Polygon.js").default} Geometry.
     * @api
     */
    getGeometry() {
      return this.box_.getGeometry();
    }

    /**
     * Handle pointer drag events.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Event.
     */
    handleDragEvent(mapBrowserEvent) {
      this.box_.setPixels(this.startPixel_, mapBrowserEvent.pixel);

      this.dispatchEvent(
        new DragBoxEvent(
          DragBoxEventType.BOXDRAG,
          mapBrowserEvent.coordinate,
          mapBrowserEvent
        )
      );
    }

    /**
     * Handle pointer up events.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Event.
     * @return {boolean} If the event was consumed.
     */
    handleUpEvent(mapBrowserEvent) {
      this.box_.setMap(null);

      const completeBox = this.boxEndCondition_(
        mapBrowserEvent,
        this.startPixel_,
        mapBrowserEvent.pixel
      );
      if (completeBox) {
        this.onBoxEnd(mapBrowserEvent);
      }
      this.dispatchEvent(
        new DragBoxEvent(
          completeBox ? DragBoxEventType.BOXEND : DragBoxEventType.BOXCANCEL,
          mapBrowserEvent.coordinate,
          mapBrowserEvent
        )
      );
      return false;
    }

    /**
     * Handle pointer down events.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Event.
     * @return {boolean} If the event was consumed.
     */
    handleDownEvent(mapBrowserEvent) {
      if (this.condition_(mapBrowserEvent)) {
        this.startPixel_ = mapBrowserEvent.pixel;
        this.box_.setMap(mapBrowserEvent.map);
        this.box_.setPixels(this.startPixel_, this.startPixel_);
        this.dispatchEvent(
          new DragBoxEvent(
            DragBoxEventType.BOXSTART,
            mapBrowserEvent.coordinate,
            mapBrowserEvent
          )
        );
        return true;
      } else {
        return false;
      }
    }

    /**
     * Function to execute just before `onboxend` is fired
     * @param {import("../MapBrowserEvent.js").default} event Event.
     */
    onBoxEnd(event) {}
  }

  /**
   * @module ol/interaction/DragZoom
   */

  /**
   * @typedef {Object} Options
   * @property {string} [className='ol-dragzoom'] CSS class name for styling the
   * box.
   * @property {import("../events/condition.js").Condition} [condition] A function that
   * takes an {@link module:ol/MapBrowserEvent~MapBrowserEvent} and returns a
   * boolean to indicate whether that event should be handled.
   * Default is {@link module:ol/events/condition.shiftKeyOnly}.
   * @property {number} [duration=200] Animation duration in milliseconds.
   * @property {boolean} [out=false] Use interaction for zooming out.
   * @property {number} [minArea=64] The minimum area of the box in pixel, this value is used by the parent default
   * `boxEndCondition` function.
   */

  /**
   * @classdesc
   * Allows the user to zoom the map by clicking and dragging on the map,
   * normally combined with an {@link module:ol/events/condition} that limits
   * it to when a key, shift by default, is held down.
   *
   * To change the style of the box, use CSS and the `.ol-dragzoom` selector, or
   * your custom one configured with `className`.
   * @api
   */
  class DragZoom extends DragBox {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      options = options ? options : {};

      const condition = options.condition ? options.condition : shiftKeyOnly;

      super({
        condition: condition,
        className: options.className || 'ol-dragzoom',
        minArea: options.minArea,
      });

      /**
       * @private
       * @type {number}
       */
      this.duration_ = options.duration !== undefined ? options.duration : 200;

      /**
       * @private
       * @type {boolean}
       */
      this.out_ = options.out !== undefined ? options.out : false;
    }

    /**
     * Function to execute just before `onboxend` is fired
     * @param {import("../MapBrowserEvent.js").default} event Event.
     */
    onBoxEnd(event) {
      const map = this.getMap();
      const view = /** @type {!import("../View.js").default} */ (map.getView());
      let geometry = this.getGeometry();

      if (this.out_) {
        const rotatedExtent = view.rotatedExtentForGeometry(geometry);
        const resolution = view.getResolutionForExtentInternal(rotatedExtent);
        const factor = view.getResolution() / resolution;
        geometry = geometry.clone();
        geometry.scale(factor * factor);
      }

      view.fitInternal(geometry, {
        duration: this.duration_,
        easing: easeOut,
      });
    }
  }

  /**
   * @module ol/events/KeyCode
   */

  /**
   * @enum {number}
   * @const
   */
  var KeyCode = {
    LEFT: 37,
    UP: 38,
    RIGHT: 39,
    DOWN: 40,
  };

  /**
   * @module ol/interaction/KeyboardPan
   */

  /**
   * @typedef {Object} Options
   * @property {import("../events/condition.js").Condition} [condition] A function that
   * takes an {@link module:ol/MapBrowserEvent~MapBrowserEvent} and returns a
   * boolean to indicate whether that event should be handled. Default is
   * {@link module:ol/events/condition.noModifierKeys} and
   * {@link module:ol/events/condition.targetNotEditable}.
   * @property {number} [duration=100] Animation duration in milliseconds.
   * @property {number} [pixelDelta=128] The amount of pixels to pan on each key
   * press.
   */

  /**
   * @classdesc
   * Allows the user to pan the map using keyboard arrows.
   * Note that, although this interaction is by default included in maps,
   * the keys can only be used when browser focus is on the element to which
   * the keyboard events are attached. By default, this is the map div,
   * though you can change this with the `keyboardEventTarget` in
   * {@link module:ol/Map~Map}. `document` never loses focus but, for any other
   * element, focus will have to be on, and returned to, this element if the keys
   * are to function.
   * See also {@link module:ol/interaction/KeyboardZoom~KeyboardZoom}.
   * @api
   */
  class KeyboardPan extends Interaction {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      super();

      options = options || {};

      /**
       * @private
       * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Browser event.
       * @return {boolean} Combined condition result.
       */
      this.defaultCondition_ = function (mapBrowserEvent) {
        return (
          noModifierKeys(mapBrowserEvent) && targetNotEditable(mapBrowserEvent)
        );
      };

      /**
       * @private
       * @type {import("../events/condition.js").Condition}
       */
      this.condition_ =
        options.condition !== undefined
          ? options.condition
          : this.defaultCondition_;

      /**
       * @private
       * @type {number}
       */
      this.duration_ = options.duration !== undefined ? options.duration : 100;

      /**
       * @private
       * @type {number}
       */
      this.pixelDelta_ =
        options.pixelDelta !== undefined ? options.pixelDelta : 128;
    }

    /**
     * Handles the {@link module:ol/MapBrowserEvent~MapBrowserEvent map browser event} if it was a
     * `KeyEvent`, and decides the direction to pan to (if an arrow key was
     * pressed).
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Map browser event.
     * @return {boolean} `false` to stop event propagation.
     * @this {KeyboardPan}
     */
    handleEvent(mapBrowserEvent) {
      let stopEvent = false;
      if (mapBrowserEvent.type == EventType$1.KEYDOWN) {
        const keyEvent = /** @type {KeyboardEvent} */ (
          mapBrowserEvent.originalEvent
        );
        const keyCode = keyEvent.keyCode;
        if (
          this.condition_(mapBrowserEvent) &&
          (keyCode == KeyCode.DOWN ||
            keyCode == KeyCode.LEFT ||
            keyCode == KeyCode.RIGHT ||
            keyCode == KeyCode.UP)
        ) {
          const map = mapBrowserEvent.map;
          const view = map.getView();
          const mapUnitsDelta = view.getResolution() * this.pixelDelta_;
          let deltaX = 0,
            deltaY = 0;
          if (keyCode == KeyCode.DOWN) {
            deltaY = -mapUnitsDelta;
          } else if (keyCode == KeyCode.LEFT) {
            deltaX = -mapUnitsDelta;
          } else if (keyCode == KeyCode.RIGHT) {
            deltaX = mapUnitsDelta;
          } else {
            deltaY = mapUnitsDelta;
          }
          const delta = [deltaX, deltaY];
          rotate$1(delta, view.getRotation());
          pan(view, delta, this.duration_);
          keyEvent.preventDefault();
          stopEvent = true;
        }
      }
      return !stopEvent;
    }
  }

  /**
   * @module ol/interaction/KeyboardZoom
   */

  /**
   * @typedef {Object} Options
   * @property {number} [duration=100] Animation duration in milliseconds.
   * @property {import("../events/condition.js").Condition} [condition] A function that
   * takes an {@link module:ol/MapBrowserEvent~MapBrowserEvent} and returns a
   * boolean to indicate whether that event should be handled. Default is
   * {@link module:ol/events/condition.targetNotEditable}.
   * @property {number} [delta=1] The zoom level delta on each key press.
   */

  /**
   * @classdesc
   * Allows the user to zoom the map using keyboard + and -.
   * Note that, although this interaction is by default included in maps,
   * the keys can only be used when browser focus is on the element to which
   * the keyboard events are attached. By default, this is the map div,
   * though you can change this with the `keyboardEventTarget` in
   * {@link module:ol/Map~Map}. `document` never loses focus but, for any other
   * element, focus will have to be on, and returned to, this element if the keys
   * are to function.
   * See also {@link module:ol/interaction/KeyboardPan~KeyboardPan}.
   * @api
   */
  class KeyboardZoom extends Interaction {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      super();

      options = options ? options : {};

      /**
       * @private
       * @type {import("../events/condition.js").Condition}
       */
      this.condition_ = options.condition ? options.condition : targetNotEditable;

      /**
       * @private
       * @type {number}
       */
      this.delta_ = options.delta ? options.delta : 1;

      /**
       * @private
       * @type {number}
       */
      this.duration_ = options.duration !== undefined ? options.duration : 100;
    }

    /**
     * Handles the {@link module:ol/MapBrowserEvent~MapBrowserEvent map browser event} if it was a
     * `KeyEvent`, and decides whether to zoom in or out (depending on whether the
     * key pressed was '+' or '-').
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Map browser event.
     * @return {boolean} `false` to stop event propagation.
     * @this {KeyboardZoom}
     */
    handleEvent(mapBrowserEvent) {
      let stopEvent = false;
      if (
        mapBrowserEvent.type == EventType$1.KEYDOWN ||
        mapBrowserEvent.type == EventType$1.KEYPRESS
      ) {
        const keyEvent = /** @type {KeyboardEvent} */ (
          mapBrowserEvent.originalEvent
        );
        const charCode = keyEvent.charCode;
        if (
          this.condition_(mapBrowserEvent) &&
          (charCode == '+'.charCodeAt(0) || charCode == '-'.charCodeAt(0))
        ) {
          const map = mapBrowserEvent.map;
          const delta =
            charCode == '+'.charCodeAt(0) ? this.delta_ : -this.delta_;
          const view = map.getView();
          zoomByDelta(view, delta, undefined, this.duration_);
          keyEvent.preventDefault();
          stopEvent = true;
        }
      }
      return !stopEvent;
    }
  }

  /**
   * @module ol/interaction/MouseWheelZoom
   */

  /**
   * @typedef {'trackpad' | 'wheel'} Mode
   */

  /**
   * @typedef {Object} Options
   * @property {import("../events/condition.js").Condition} [condition] A function that
   * takes an {@link module:ol/MapBrowserEvent~MapBrowserEvent} and returns a
   * boolean to indicate whether that event should be handled. Default is
   * {@link module:ol/events/condition.always}.
   * @property {boolean} [onFocusOnly=false] When the map's target has a `tabindex` attribute set,
   * the interaction will only handle events when the map has the focus.
   * @property {number} [maxDelta=1] Maximum mouse wheel delta.
   * @property {number} [duration=250] Animation duration in milliseconds.
   * @property {number} [timeout=80] Mouse wheel timeout duration in milliseconds.
   * @property {boolean} [useAnchor=true] Enable zooming using the mouse's
   * location as the anchor. When set to `false`, zooming in and out will zoom to
   * the center of the screen instead of zooming on the mouse's location.
   * @property {boolean} [constrainResolution=false] If true, the mouse wheel zoom
   * event will always animate to the closest zoom level after an interaction;
   * false means intermediary zoom levels are allowed.
   */

  /**
   * @classdesc
   * Allows the user to zoom the map by scrolling the mouse wheel.
   * @api
   */
  class MouseWheelZoom extends Interaction {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      options = options ? options : {};

      super(
        /** @type {import("./Interaction.js").InteractionOptions} */ (options)
      );

      /**
       * @private
       * @type {number}
       */
      this.totalDelta_ = 0;

      /**
       * @private
       * @type {number}
       */
      this.lastDelta_ = 0;

      /**
       * @private
       * @type {number}
       */
      this.maxDelta_ = options.maxDelta !== undefined ? options.maxDelta : 1;

      /**
       * @private
       * @type {number}
       */
      this.duration_ = options.duration !== undefined ? options.duration : 250;

      /**
       * @private
       * @type {number}
       */
      this.timeout_ = options.timeout !== undefined ? options.timeout : 80;

      /**
       * @private
       * @type {boolean}
       */
      this.useAnchor_ =
        options.useAnchor !== undefined ? options.useAnchor : true;

      /**
       * @private
       * @type {boolean}
       */
      this.constrainResolution_ =
        options.constrainResolution !== undefined
          ? options.constrainResolution
          : false;

      const condition = options.condition ? options.condition : always;

      /**
       * @private
       * @type {import("../events/condition.js").Condition}
       */
      this.condition_ = options.onFocusOnly
        ? all$1(focusWithTabindex, condition)
        : condition;

      /**
       * @private
       * @type {?import("../coordinate.js").Coordinate}
       */
      this.lastAnchor_ = null;

      /**
       * @private
       * @type {number|undefined}
       */
      this.startTime_ = undefined;

      /**
       * @private
       * @type {?}
       */
      this.timeoutId_;

      /**
       * @private
       * @type {Mode|undefined}
       */
      this.mode_ = undefined;

      /**
       * Trackpad events separated by this delay will be considered separate
       * interactions.
       * @type {number}
       */
      this.trackpadEventGap_ = 400;

      /**
       * @type {?}
       */
      this.trackpadTimeoutId_;

      /**
       * The number of delta values per zoom level
       * @private
       * @type {number}
       */
      this.deltaPerZoom_ = 300;
    }

    /**
     * @private
     */
    endInteraction_() {
      this.trackpadTimeoutId_ = undefined;
      const map = this.getMap();
      if (!map) {
        return;
      }
      const view = map.getView();
      view.endInteraction(
        undefined,
        this.lastDelta_ ? (this.lastDelta_ > 0 ? 1 : -1) : 0,
        this.lastAnchor_
      );
    }

    /**
     * Handles the {@link module:ol/MapBrowserEvent~MapBrowserEvent map browser event} (if it was a mousewheel-event) and eventually
     * zooms the map.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Map browser event.
     * @return {boolean} `false` to stop event propagation.
     */
    handleEvent(mapBrowserEvent) {
      if (!this.condition_(mapBrowserEvent)) {
        return true;
      }
      const type = mapBrowserEvent.type;
      if (type !== EventType$1.WHEEL) {
        return true;
      }

      const map = mapBrowserEvent.map;
      const wheelEvent = /** @type {WheelEvent} */ (
        mapBrowserEvent.originalEvent
      );
      wheelEvent.preventDefault();

      if (this.useAnchor_) {
        this.lastAnchor_ = mapBrowserEvent.coordinate;
      }

      // Delta normalisation inspired by
      // https://github.com/mapbox/mapbox-gl-js/blob/001c7b9/js/ui/handler/scroll_zoom.js
      let delta;
      if (mapBrowserEvent.type == EventType$1.WHEEL) {
        delta = wheelEvent.deltaY;
        if (FIREFOX && wheelEvent.deltaMode === WheelEvent.DOM_DELTA_PIXEL) {
          delta /= DEVICE_PIXEL_RATIO;
        }
        if (wheelEvent.deltaMode === WheelEvent.DOM_DELTA_LINE) {
          delta *= 40;
        }
      }

      if (delta === 0) {
        return false;
      } else {
        this.lastDelta_ = delta;
      }

      const now = Date.now();

      if (this.startTime_ === undefined) {
        this.startTime_ = now;
      }

      if (!this.mode_ || now - this.startTime_ > this.trackpadEventGap_) {
        this.mode_ = Math.abs(delta) < 4 ? 'trackpad' : 'wheel';
      }

      const view = map.getView();
      if (
        this.mode_ === 'trackpad' &&
        !(view.getConstrainResolution() || this.constrainResolution_)
      ) {
        if (this.trackpadTimeoutId_) {
          clearTimeout(this.trackpadTimeoutId_);
        } else {
          if (view.getAnimating()) {
            view.cancelAnimations();
          }
          view.beginInteraction();
        }
        this.trackpadTimeoutId_ = setTimeout(
          this.endInteraction_.bind(this),
          this.timeout_
        );
        view.adjustZoom(-delta / this.deltaPerZoom_, this.lastAnchor_);
        this.startTime_ = now;
        return false;
      }

      this.totalDelta_ += delta;

      const timeLeft = Math.max(this.timeout_ - (now - this.startTime_), 0);

      clearTimeout(this.timeoutId_);
      this.timeoutId_ = setTimeout(
        this.handleWheelZoom_.bind(this, map),
        timeLeft
      );

      return false;
    }

    /**
     * @private
     * @param {import("../Map.js").default} map Map.
     */
    handleWheelZoom_(map) {
      const view = map.getView();
      if (view.getAnimating()) {
        view.cancelAnimations();
      }
      let delta =
        -clamp(
          this.totalDelta_,
          -this.maxDelta_ * this.deltaPerZoom_,
          this.maxDelta_ * this.deltaPerZoom_
        ) / this.deltaPerZoom_;
      if (view.getConstrainResolution() || this.constrainResolution_) {
        // view has a zoom constraint, zoom by 1
        delta = delta ? (delta > 0 ? 1 : -1) : 0;
      }
      zoomByDelta(view, delta, this.lastAnchor_, this.duration_);

      this.mode_ = undefined;
      this.totalDelta_ = 0;
      this.lastAnchor_ = null;
      this.startTime_ = undefined;
      this.timeoutId_ = undefined;
    }

    /**
     * Enable or disable using the mouse's location as an anchor when zooming
     * @param {boolean} useAnchor true to zoom to the mouse's location, false
     * to zoom to the center of the map
     * @api
     */
    setMouseAnchor(useAnchor) {
      this.useAnchor_ = useAnchor;
      if (!useAnchor) {
        this.lastAnchor_ = null;
      }
    }
  }

  /**
   * @module ol/interaction/PinchRotate
   */

  /**
   * @typedef {Object} Options
   * @property {number} [duration=250] The duration of the animation in
   * milliseconds.
   * @property {number} [threshold=0.3] Minimal angle in radians to start a rotation.
   */

  /**
   * @classdesc
   * Allows the user to rotate the map by twisting with two fingers
   * on a touch screen.
   * @api
   */
  class PinchRotate extends PointerInteraction {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      options = options ? options : {};

      const pointerOptions = /** @type {import("./Pointer.js").Options} */ (
        options
      );

      if (!pointerOptions.stopDown) {
        pointerOptions.stopDown = FALSE;
      }

      super(pointerOptions);

      /**
       * @private
       * @type {import("../coordinate.js").Coordinate}
       */
      this.anchor_ = null;

      /**
       * @private
       * @type {number|undefined}
       */
      this.lastAngle_ = undefined;

      /**
       * @private
       * @type {boolean}
       */
      this.rotating_ = false;

      /**
       * @private
       * @type {number}
       */
      this.rotationDelta_ = 0.0;

      /**
       * @private
       * @type {number}
       */
      this.threshold_ = options.threshold !== undefined ? options.threshold : 0.3;

      /**
       * @private
       * @type {number}
       */
      this.duration_ = options.duration !== undefined ? options.duration : 250;
    }

    /**
     * Handle pointer drag events.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Event.
     */
    handleDragEvent(mapBrowserEvent) {
      let rotationDelta = 0.0;

      const touch0 = this.targetPointers[0];
      const touch1 = this.targetPointers[1];

      // angle between touches
      const angle = Math.atan2(
        touch1.clientY - touch0.clientY,
        touch1.clientX - touch0.clientX
      );

      if (this.lastAngle_ !== undefined) {
        const delta = angle - this.lastAngle_;
        this.rotationDelta_ += delta;
        if (!this.rotating_ && Math.abs(this.rotationDelta_) > this.threshold_) {
          this.rotating_ = true;
        }
        rotationDelta = delta;
      }
      this.lastAngle_ = angle;

      const map = mapBrowserEvent.map;
      const view = map.getView();
      if (view.getConstraints().rotation === disable) {
        return;
      }

      // rotate anchor point.
      // FIXME: should be the intersection point between the lines:
      //     touch0,touch1 and previousTouch0,previousTouch1
      const viewportPosition = map.getViewport().getBoundingClientRect();
      const centroid$1 = centroid(this.targetPointers);
      centroid$1[0] -= viewportPosition.left;
      centroid$1[1] -= viewportPosition.top;
      this.anchor_ = map.getCoordinateFromPixelInternal(centroid$1);

      // rotate
      if (this.rotating_) {
        map.render();
        view.adjustRotationInternal(rotationDelta, this.anchor_);
      }
    }

    /**
     * Handle pointer up events.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Event.
     * @return {boolean} If the event was consumed.
     */
    handleUpEvent(mapBrowserEvent) {
      if (this.targetPointers.length < 2) {
        const map = mapBrowserEvent.map;
        const view = map.getView();
        view.endInteraction(this.duration_);
        return false;
      } else {
        return true;
      }
    }

    /**
     * Handle pointer down events.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Event.
     * @return {boolean} If the event was consumed.
     */
    handleDownEvent(mapBrowserEvent) {
      if (this.targetPointers.length >= 2) {
        const map = mapBrowserEvent.map;
        this.anchor_ = null;
        this.lastAngle_ = undefined;
        this.rotating_ = false;
        this.rotationDelta_ = 0.0;
        if (!this.handlingDownUpSequence) {
          map.getView().beginInteraction();
        }
        return true;
      } else {
        return false;
      }
    }
  }

  /**
   * @module ol/interaction/PinchZoom
   */

  /**
   * @typedef {Object} Options
   * @property {number} [duration=400] Animation duration in milliseconds.
   */

  /**
   * @classdesc
   * Allows the user to zoom the map by pinching with two fingers
   * on a touch screen.
   * @api
   */
  class PinchZoom extends PointerInteraction {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      options = options ? options : {};

      const pointerOptions = /** @type {import("./Pointer.js").Options} */ (
        options
      );

      if (!pointerOptions.stopDown) {
        pointerOptions.stopDown = FALSE;
      }

      super(pointerOptions);

      /**
       * @private
       * @type {import("../coordinate.js").Coordinate}
       */
      this.anchor_ = null;

      /**
       * @private
       * @type {number}
       */
      this.duration_ = options.duration !== undefined ? options.duration : 400;

      /**
       * @private
       * @type {number|undefined}
       */
      this.lastDistance_ = undefined;

      /**
       * @private
       * @type {number}
       */
      this.lastScaleDelta_ = 1;
    }

    /**
     * Handle pointer drag events.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Event.
     */
    handleDragEvent(mapBrowserEvent) {
      let scaleDelta = 1.0;

      const touch0 = this.targetPointers[0];
      const touch1 = this.targetPointers[1];
      const dx = touch0.clientX - touch1.clientX;
      const dy = touch0.clientY - touch1.clientY;

      // distance between touches
      const distance = Math.sqrt(dx * dx + dy * dy);

      if (this.lastDistance_ !== undefined) {
        scaleDelta = this.lastDistance_ / distance;
      }
      this.lastDistance_ = distance;

      const map = mapBrowserEvent.map;
      const view = map.getView();

      if (scaleDelta != 1.0) {
        this.lastScaleDelta_ = scaleDelta;
      }

      // scale anchor point.
      const viewportPosition = map.getViewport().getBoundingClientRect();
      const centroid$1 = centroid(this.targetPointers);
      centroid$1[0] -= viewportPosition.left;
      centroid$1[1] -= viewportPosition.top;
      this.anchor_ = map.getCoordinateFromPixelInternal(centroid$1);

      // scale, bypass the resolution constraint
      map.render();
      view.adjustResolutionInternal(scaleDelta, this.anchor_);
    }

    /**
     * Handle pointer up events.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Event.
     * @return {boolean} If the event was consumed.
     */
    handleUpEvent(mapBrowserEvent) {
      if (this.targetPointers.length < 2) {
        const map = mapBrowserEvent.map;
        const view = map.getView();
        const direction = this.lastScaleDelta_ > 1 ? 1 : -1;
        view.endInteraction(this.duration_, direction);
        return false;
      } else {
        return true;
      }
    }

    /**
     * Handle pointer down events.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Event.
     * @return {boolean} If the event was consumed.
     */
    handleDownEvent(mapBrowserEvent) {
      if (this.targetPointers.length >= 2) {
        const map = mapBrowserEvent.map;
        this.anchor_ = null;
        this.lastDistance_ = undefined;
        this.lastScaleDelta_ = 1;
        if (!this.handlingDownUpSequence) {
          map.getView().beginInteraction();
        }
        return true;
      } else {
        return false;
      }
    }
  }

  /**
   * @module ol/interaction/defaults
   */

  /**
   * @typedef {Object} DefaultsOptions
   * @property {boolean} [altShiftDragRotate=true] Whether Alt-Shift-drag rotate is
   * desired.
   * @property {boolean} [onFocusOnly=false] Interact only when the map has the
   * focus. This affects the `MouseWheelZoom` and `DragPan` interactions and is
   * useful when page scroll is desired for maps that do not have the browser's
   * focus.
   * @property {boolean} [doubleClickZoom=true] Whether double click zoom is
   * desired.
   * @property {boolean} [keyboard=true] Whether keyboard interaction is desired.
   * @property {boolean} [mouseWheelZoom=true] Whether mousewheel zoom is desired.
   * @property {boolean} [shiftDragZoom=true] Whether Shift-drag zoom is desired.
   * @property {boolean} [dragPan=true] Whether drag pan is desired.
   * @property {boolean} [pinchRotate=true] Whether pinch rotate is desired.
   * @property {boolean} [pinchZoom=true] Whether pinch zoom is desired.
   * @property {number} [zoomDelta] Zoom level delta when using keyboard or double click zoom.
   * @property {number} [zoomDuration] Duration of the zoom animation in
   * milliseconds.
   */

  /**
   * Set of interactions included in maps by default. Specific interactions can be
   * excluded by setting the appropriate option to false in the constructor
   * options, but the order of the interactions is fixed.  If you want to specify
   * a different order for interactions, you will need to create your own
   * {@link module:ol/interaction/Interaction~Interaction} instances and insert
   * them into a {@link module:ol/Collection~Collection} in the order you want
   * before creating your {@link module:ol/Map~Map} instance. Changing the order can
   * be of interest if the event propagation needs to be stopped at a point.
   * The default set of interactions, in sequence, is:
   * * {@link module:ol/interaction/DragRotate~DragRotate}
   * * {@link module:ol/interaction/DoubleClickZoom~DoubleClickZoom}
   * * {@link module:ol/interaction/DragPan~DragPan}
   * * {@link module:ol/interaction/PinchRotate~PinchRotate}
   * * {@link module:ol/interaction/PinchZoom~PinchZoom}
   * * {@link module:ol/interaction/KeyboardPan~KeyboardPan}
   * * {@link module:ol/interaction/KeyboardZoom~KeyboardZoom}
   * * {@link module:ol/interaction/MouseWheelZoom~MouseWheelZoom}
   * * {@link module:ol/interaction/DragZoom~DragZoom}
   *
   * @param {DefaultsOptions} [options] Defaults options.
   * @return {Collection<import("./Interaction.js").default>}
   * A collection of interactions to be used with the {@link module:ol/Map~Map}
   * constructor's `interactions` option.
   * @api
   */
  function defaults(options) {
    options = options ? options : {};

    /** @type {Collection<import("./Interaction.js").default>} */
    const interactions = new Collection();

    const kinetic = new Kinetic(-0.005, 0.05, 100);

    const altShiftDragRotate =
      options.altShiftDragRotate !== undefined
        ? options.altShiftDragRotate
        : true;
    if (altShiftDragRotate) {
      interactions.push(new DragRotate());
    }

    const doubleClickZoom =
      options.doubleClickZoom !== undefined ? options.doubleClickZoom : true;
    if (doubleClickZoom) {
      interactions.push(
        new DoubleClickZoom({
          delta: options.zoomDelta,
          duration: options.zoomDuration,
        })
      );
    }

    const dragPan = options.dragPan !== undefined ? options.dragPan : true;
    if (dragPan) {
      interactions.push(
        new DragPan({
          onFocusOnly: options.onFocusOnly,
          kinetic: kinetic,
        })
      );
    }

    const pinchRotate =
      options.pinchRotate !== undefined ? options.pinchRotate : true;
    if (pinchRotate) {
      interactions.push(new PinchRotate());
    }

    const pinchZoom = options.pinchZoom !== undefined ? options.pinchZoom : true;
    if (pinchZoom) {
      interactions.push(
        new PinchZoom({
          duration: options.zoomDuration,
        })
      );
    }

    const keyboard = options.keyboard !== undefined ? options.keyboard : true;
    if (keyboard) {
      interactions.push(new KeyboardPan());
      interactions.push(
        new KeyboardZoom({
          delta: options.zoomDelta,
          duration: options.zoomDuration,
        })
      );
    }

    const mouseWheelZoom =
      options.mouseWheelZoom !== undefined ? options.mouseWheelZoom : true;
    if (mouseWheelZoom) {
      interactions.push(
        new MouseWheelZoom({
          onFocusOnly: options.onFocusOnly,
          duration: options.zoomDuration,
        })
      );
    }

    const shiftDragZoom =
      options.shiftDragZoom !== undefined ? options.shiftDragZoom : true;
    if (shiftDragZoom) {
      interactions.push(
        new DragZoom({
          duration: options.zoomDuration,
        })
      );
    }

    return interactions;
  }

  /**
   * @module ol/size
   */

  /**
   * An array of numbers representing a size: `[width, height]`.
   * @typedef {Array<number>} Size
   * @api
   */

  /**
   * Returns a buffered size.
   * @param {Size} size Size.
   * @param {number} num The amount by which to buffer.
   * @param {Size} [dest] Optional reusable size array.
   * @return {Size} The buffered size.
   */
  function buffer(size, num, dest) {
    if (dest === undefined) {
      dest = [0, 0];
    }
    dest[0] = size[0] + 2 * num;
    dest[1] = size[1] + 2 * num;
    return dest;
  }

  /**
   * Determines if a size has a positive area.
   * @param {Size} size The size to test.
   * @return {boolean} The size has a positive area.
   */
  function hasArea(size) {
    return size[0] > 0 && size[1] > 0;
  }

  /**
   * Returns a size scaled by a ratio. The result will be an array of integers.
   * @param {Size} size Size.
   * @param {number} ratio Ratio.
   * @param {Size} [dest] Optional reusable size array.
   * @return {Size} The scaled size.
   */
  function scale(size, ratio, dest) {
    if (dest === undefined) {
      dest = [0, 0];
    }
    dest[0] = (size[0] * ratio + 0.5) | 0;
    dest[1] = (size[1] * ratio + 0.5) | 0;
    return dest;
  }

  /**
   * Returns an `Size` array for the passed in number (meaning: square) or
   * `Size` array.
   * (meaning: non-square),
   * @param {number|Size} size Width and height.
   * @param {Size} [dest] Optional reusable size array.
   * @return {Size} Size.
   * @api
   */
  function toSize(size, dest) {
    if (Array.isArray(size)) {
      return size;
    } else {
      if (dest === undefined) {
        dest = [size, size];
      } else {
        dest[0] = size;
        dest[1] = size;
      }
      return dest;
    }
  }

  /**
   * @module ol/Map
   */

  /**
   * State of the current frame. Only `pixelRatio`, `time` and `viewState` should
   * be used in applications.
   * @typedef {Object} FrameState
   * @property {number} pixelRatio The pixel ratio of the frame.
   * @property {number} time The time when rendering of the frame was requested.
   * @property {import("./View.js").State} viewState The state of the current view.
   * @property {boolean} animate Animate.
   * @property {import("./transform.js").Transform} coordinateToPixelTransform CoordinateToPixelTransform.
   * @property {import("rbush").default} declutterTree DeclutterTree.
   * @property {null|import("./extent.js").Extent} extent Extent.
   * @property {import("./extent.js").Extent} [nextExtent] Next extent during an animation series.
   * @property {number} index Index.
   * @property {Array<import("./layer/Layer.js").State>} layerStatesArray LayerStatesArray.
   * @property {number} layerIndex LayerIndex.
   * @property {import("./transform.js").Transform} pixelToCoordinateTransform PixelToCoordinateTransform.
   * @property {Array<PostRenderFunction>} postRenderFunctions PostRenderFunctions.
   * @property {import("./size.js").Size} size Size.
   * @property {TileQueue} tileQueue TileQueue.
   * @property {!Object<string, Object<string, boolean>>} usedTiles UsedTiles.
   * @property {Array<number>} viewHints ViewHints.
   * @property {!Object<string, Object<string, boolean>>} wantedTiles WantedTiles.
   * @property {string} mapId The id of the map.
   * @property {Object<string, boolean>} renderTargets Identifiers of previously rendered elements.
   */

  /**
   * @typedef {function(Map, ?FrameState): any} PostRenderFunction
   */

  /**
   * @typedef {Object} AtPixelOptions
   * @property {undefined|function(import("./layer/Layer.js").default<import("./source/Source").default>): boolean} [layerFilter] Layer filter
   * function. The filter function will receive one argument, the
   * {@link module:ol/layer/Layer~Layer layer-candidate} and it should return a boolean value.
   * Only layers which are visible and for which this function returns `true`
   * will be tested for features. By default, all visible layers will be tested.
   * @property {number} [hitTolerance=0] Hit-detection tolerance in css pixels. Pixels
   * inside the radius around the given position will be checked for features.
   * @property {boolean} [checkWrapped=true] Check-Wrapped Will check for wrapped geometries inside the range of
   *   +/- 1 world width. Works only if a projection is used that can be wrapped.
   */

  /**
   * @typedef {Object} MapOptionsInternal
   * @property {Collection<import("./control/Control.js").default>} [controls] Controls.
   * @property {Collection<import("./interaction/Interaction.js").default>} [interactions] Interactions.
   * @property {HTMLElement|Document} keyboardEventTarget KeyboardEventTarget.
   * @property {Collection<import("./Overlay.js").default>} overlays Overlays.
   * @property {Object<string, *>} values Values.
   */

  /**
   * @typedef {import("./ObjectEventType").Types|'change:layergroup'|'change:size'|'change:target'|'change:view'} MapObjectEventTypes
   */

  /***
   * @template Return
   * @typedef {import("./Observable").OnSignature<import("./Observable").EventTypes, import("./events/Event.js").default, Return> &
   *    import("./Observable").OnSignature<MapObjectEventTypes, import("./Object").ObjectEvent, Return> &
   *    import("./Observable").OnSignature<import("./MapBrowserEventType").Types, import("./MapBrowserEvent").default, Return> &
   *    import("./Observable").OnSignature<import("./MapEventType").Types, import("./MapEvent").default, Return> &
   *    import("./Observable").OnSignature<import("./render/EventType").MapRenderEventTypes, import("./render/Event").default, Return> &
   *    import("./Observable").CombinedOnSignature<import("./Observable").EventTypes|MapObjectEventTypes|
   *      import("./MapBrowserEventType").Types|import("./MapEventType").Types|
   *      import("./render/EventType").MapRenderEventTypes, Return>} MapEventHandler
   */

  /**
   * Object literal with config options for the map.
   * @typedef {Object} MapOptions
   * @property {Collection<import("./control/Control.js").default>|Array<import("./control/Control.js").default>} [controls]
   * Controls initially added to the map. If not specified,
   * {@link module:ol/control/defaults.defaults} is used.
   * @property {number} [pixelRatio=window.devicePixelRatio] The ratio between
   * physical pixels and device-independent pixels (dips) on the device.
   * @property {Collection<import("./interaction/Interaction.js").default>|Array<import("./interaction/Interaction.js").default>} [interactions]
   * Interactions that are initially added to the map. If not specified,
   * {@link module:ol/interaction/defaults.defaults} is used.
   * @property {HTMLElement|Document|string} [keyboardEventTarget] The element to
   * listen to keyboard events on. This determines when the `KeyboardPan` and
   * `KeyboardZoom` interactions trigger. For example, if this option is set to
   * `document` the keyboard interactions will always trigger. If this option is
   * not specified, the element the library listens to keyboard events on is the
   * map target (i.e. the user-provided div for the map). If this is not
   * `document`, the target element needs to be focused for key events to be
   * emitted, requiring that the target element has a `tabindex` attribute.
   * @property {Array<import("./layer/Base.js").default>|Collection<import("./layer/Base.js").default>|LayerGroup} [layers]
   * Layers. If this is not defined, a map with no layers will be rendered. Note
   * that layers are rendered in the order supplied, so if you want, for example,
   * a vector layer to appear on top of a tile layer, it must come after the tile
   * layer.
   * @property {number} [maxTilesLoading=16] Maximum number tiles to load
   * simultaneously.
   * @property {number} [moveTolerance=1] The minimum distance in pixels the
   * cursor must move to be detected as a map move event instead of a click.
   * Increasing this value can make it easier to click on the map.
   * @property {Collection<import("./Overlay.js").default>|Array<import("./Overlay.js").default>} [overlays]
   * Overlays initially added to the map. By default, no overlays are added.
   * @property {HTMLElement|string} [target] The container for the map, either the
   * element itself or the `id` of the element. If not specified at construction
   * time, {@link module:ol/Map~Map#setTarget} must be called for the map to be
   * rendered. If passed by element, the container can be in a secondary document.
   * @property {View|Promise<import("./View.js").ViewOptions>} [view] The map's view.  No layer sources will be
   * fetched unless this is specified at construction time or through
   * {@link module:ol/Map~Map#setView}.
   */

  /**
   * @param {import("./layer/Base.js").default} layer Layer.
   */
  function removeLayerMapProperty(layer) {
    if (layer instanceof Layer) {
      layer.setMapInternal(null);
      return;
    }
    if (layer instanceof LayerGroup) {
      layer.getLayers().forEach(removeLayerMapProperty);
    }
  }

  /**
   * @param {import("./layer/Base.js").default} layer Layer.
   * @param {Map} map Map.
   */
  function setLayerMapProperty(layer, map) {
    if (layer instanceof Layer) {
      layer.setMapInternal(map);
      return;
    }
    if (layer instanceof LayerGroup) {
      const layers = layer.getLayers().getArray();
      for (let i = 0, ii = layers.length; i < ii; ++i) {
        setLayerMapProperty(layers[i], map);
      }
    }
  }

  /**
   * @classdesc
   * The map is the core component of OpenLayers. For a map to render, a view,
   * one or more layers, and a target container are needed:
   *
   *     import Map from 'ol/Map';
   *     import View from 'ol/View';
   *     import TileLayer from 'ol/layer/Tile';
   *     import OSM from 'ol/source/OSM';
   *
   *     const map = new Map({
   *       view: new View({
   *         center: [0, 0],
   *         zoom: 1,
   *       }),
   *       layers: [
   *         new TileLayer({
   *           source: new OSM(),
   *         }),
   *       ],
   *       target: 'map',
   *     });
   *
   * The above snippet creates a map using a {@link module:ol/layer/Tile~TileLayer} to
   * display {@link module:ol/source/OSM~OSM} OSM data and render it to a DOM
   * element with the id `map`.
   *
   * The constructor places a viewport container (with CSS class name
   * `ol-viewport`) in the target element (see `getViewport()`), and then two
   * further elements within the viewport: one with CSS class name
   * `ol-overlaycontainer-stopevent` for controls and some overlays, and one with
   * CSS class name `ol-overlaycontainer` for other overlays (see the `stopEvent`
   * option of {@link module:ol/Overlay~Overlay} for the difference). The map
   * itself is placed in a further element within the viewport.
   *
   * Layers are stored as a {@link module:ol/Collection~Collection} in
   * layerGroups. A top-level group is provided by the library. This is what is
   * accessed by `getLayerGroup` and `setLayerGroup`. Layers entered in the
   * options are added to this group, and `addLayer` and `removeLayer` change the
   * layer collection in the group. `getLayers` is a convenience function for
   * `getLayerGroup().getLayers()`. Note that {@link module:ol/layer/Group~LayerGroup}
   * is a subclass of {@link module:ol/layer/Base~BaseLayer}, so layers entered in the
   * options or added with `addLayer` can be groups, which can contain further
   * groups, and so on.
   *
   * @fires import("./MapBrowserEvent.js").MapBrowserEvent
   * @fires import("./MapEvent.js").MapEvent
   * @fires import("./render/Event.js").default#precompose
   * @fires import("./render/Event.js").default#postcompose
   * @fires import("./render/Event.js").default#rendercomplete
   * @api
   */
  class Map extends BaseObject {
    /**
     * @param {MapOptions} [options] Map options.
     */
    constructor(options) {
      super();

      options = options || {};

      /***
       * @type {MapEventHandler<import("./events").EventsKey>}
       */
      this.on;

      /***
       * @type {MapEventHandler<import("./events").EventsKey>}
       */
      this.once;

      /***
       * @type {MapEventHandler<void>}
       */
      this.un;

      const optionsInternal = createOptionsInternal(options);

      /**
       * @private
       * @type {boolean|undefined}
       */
      this.renderComplete_;

      /**
       * @private
       * @type {boolean}
       */
      this.loaded_ = true;

      /** @private */
      this.boundHandleBrowserEvent_ = this.handleBrowserEvent.bind(this);

      /**
       * @type {number}
       * @private
       */
      this.maxTilesLoading_ =
        options.maxTilesLoading !== undefined ? options.maxTilesLoading : 16;

      /**
       * @private
       * @type {number}
       */
      this.pixelRatio_ =
        options.pixelRatio !== undefined
          ? options.pixelRatio
          : DEVICE_PIXEL_RATIO;

      /**
       * @private
       * @type {*}
       */
      this.postRenderTimeoutHandle_;

      /**
       * @private
       * @type {number|undefined}
       */
      this.animationDelayKey_;

      /**
       * @private
       */
      this.animationDelay_ = this.animationDelay_.bind(this);

      /**
       * @private
       * @type {import("./transform.js").Transform}
       */
      this.coordinateToPixelTransform_ = create$2();

      /**
       * @private
       * @type {import("./transform.js").Transform}
       */
      this.pixelToCoordinateTransform_ = create$2();

      /**
       * @private
       * @type {number}
       */
      this.frameIndex_ = 0;

      /**
       * @private
       * @type {?FrameState}
       */
      this.frameState_ = null;

      /**
       * The extent at the previous 'moveend' event.
       * @private
       * @type {import("./extent.js").Extent}
       */
      this.previousExtent_ = null;

      /**
       * @private
       * @type {?import("./events.js").EventsKey}
       */
      this.viewPropertyListenerKey_ = null;

      /**
       * @private
       * @type {?import("./events.js").EventsKey}
       */
      this.viewChangeListenerKey_ = null;

      /**
       * @private
       * @type {?Array<import("./events.js").EventsKey>}
       */
      this.layerGroupPropertyListenerKeys_ = null;

      /**
       * @private
       * @type {!HTMLElement}
       */
      this.viewport_ = document.createElement('div');
      this.viewport_.className =
        'ol-viewport' + ('ontouchstart' in window ? ' ol-touch' : '');
      this.viewport_.style.position = 'relative';
      this.viewport_.style.overflow = 'hidden';
      this.viewport_.style.width = '100%';
      this.viewport_.style.height = '100%';

      /**
       * @private
       * @type {!HTMLElement}
       */
      this.overlayContainer_ = document.createElement('div');
      this.overlayContainer_.style.position = 'absolute';
      this.overlayContainer_.style.zIndex = '0';
      this.overlayContainer_.style.width = '100%';
      this.overlayContainer_.style.height = '100%';
      this.overlayContainer_.style.pointerEvents = 'none';
      this.overlayContainer_.className = 'ol-overlaycontainer';
      this.viewport_.appendChild(this.overlayContainer_);

      /**
       * @private
       * @type {!HTMLElement}
       */
      this.overlayContainerStopEvent_ = document.createElement('div');
      this.overlayContainerStopEvent_.style.position = 'absolute';
      this.overlayContainerStopEvent_.style.zIndex = '0';
      this.overlayContainerStopEvent_.style.width = '100%';
      this.overlayContainerStopEvent_.style.height = '100%';
      this.overlayContainerStopEvent_.style.pointerEvents = 'none';
      this.overlayContainerStopEvent_.className = 'ol-overlaycontainer-stopevent';
      this.viewport_.appendChild(this.overlayContainerStopEvent_);

      /**
       * @private
       * @type {MapBrowserEventHandler}
       */
      this.mapBrowserEventHandler_ = null;

      /**
       * @private
       * @type {number}
       */
      this.moveTolerance_ = options.moveTolerance;

      /**
       * @private
       * @type {HTMLElement|Document}
       */
      this.keyboardEventTarget_ = optionsInternal.keyboardEventTarget;

      /**
       * @private
       * @type {?Array<import("./events.js").EventsKey>}
       */
      this.targetChangeHandlerKeys_ = null;

      /**
       * @type {Collection<import("./control/Control.js").default>}
       * @protected
       */
      this.controls = optionsInternal.controls || defaults$1();

      /**
       * @type {Collection<import("./interaction/Interaction.js").default>}
       * @protected
       */
      this.interactions =
        optionsInternal.interactions ||
        defaults({
          onFocusOnly: true,
        });

      /**
       * @type {Collection<import("./Overlay.js").default>}
       * @private
       */
      this.overlays_ = optionsInternal.overlays;

      /**
       * A lookup of overlays by id.
       * @private
       * @type {Object<string, import("./Overlay.js").default>}
       */
      this.overlayIdIndex_ = {};

      /**
       * @type {import("./renderer/Map.js").default|null}
       * @private
       */
      this.renderer_ = null;

      /**
       * @private
       * @type {!Array<PostRenderFunction>}
       */
      this.postRenderFunctions_ = [];

      /**
       * @private
       * @type {TileQueue}
       */
      this.tileQueue_ = new TileQueue(
        this.getTilePriority.bind(this),
        this.handleTileChange_.bind(this)
      );

      this.addChangeListener(
        MapProperty.LAYERGROUP,
        this.handleLayerGroupChanged_
      );
      this.addChangeListener(MapProperty.VIEW, this.handleViewChanged_);
      this.addChangeListener(MapProperty.SIZE, this.handleSizeChanged_);
      this.addChangeListener(MapProperty.TARGET, this.handleTargetChanged_);

      // setProperties will trigger the rendering of the map if the map
      // is "defined" already.
      this.setProperties(optionsInternal.values);

      const map = this;
      if (options.view && !(options.view instanceof View)) {
        options.view.then(function (viewOptions) {
          map.setView(new View(viewOptions));
        });
      }

      this.controls.addEventListener(
        CollectionEventType.ADD,
        /**
         * @param {import("./Collection.js").CollectionEvent<import("./control/Control.js").default>} event CollectionEvent
         */
        function (event) {
          event.element.setMap(this);
        }.bind(this)
      );

      this.controls.addEventListener(
        CollectionEventType.REMOVE,
        /**
         * @param {import("./Collection.js").CollectionEvent<import("./control/Control.js").default>} event CollectionEvent.
         */
        function (event) {
          event.element.setMap(null);
        }.bind(this)
      );

      this.interactions.addEventListener(
        CollectionEventType.ADD,
        /**
         * @param {import("./Collection.js").CollectionEvent<import("./interaction/Interaction.js").default>} event CollectionEvent.
         */
        function (event) {
          event.element.setMap(this);
        }.bind(this)
      );

      this.interactions.addEventListener(
        CollectionEventType.REMOVE,
        /**
         * @param {import("./Collection.js").CollectionEvent<import("./interaction/Interaction.js").default>} event CollectionEvent.
         */
        function (event) {
          event.element.setMap(null);
        }.bind(this)
      );

      this.overlays_.addEventListener(
        CollectionEventType.ADD,
        /**
         * @param {import("./Collection.js").CollectionEvent<import("./Overlay.js").default>} event CollectionEvent.
         */
        function (event) {
          this.addOverlayInternal_(event.element);
        }.bind(this)
      );

      this.overlays_.addEventListener(
        CollectionEventType.REMOVE,
        /**
         * @param {import("./Collection.js").CollectionEvent<import("./Overlay.js").default>} event CollectionEvent.
         */
        function (event) {
          const id = event.element.getId();
          if (id !== undefined) {
            delete this.overlayIdIndex_[id.toString()];
          }
          event.element.setMap(null);
        }.bind(this)
      );

      this.controls.forEach(
        /**
         * @param {import("./control/Control.js").default} control Control.
         * @this {Map}
         */
        function (control) {
          control.setMap(this);
        }.bind(this)
      );

      this.interactions.forEach(
        /**
         * @param {import("./interaction/Interaction.js").default} interaction Interaction.
         * @this {Map}
         */
        function (interaction) {
          interaction.setMap(this);
        }.bind(this)
      );

      this.overlays_.forEach(this.addOverlayInternal_.bind(this));
    }

    /**
     * Add the given control to the map.
     * @param {import("./control/Control.js").default} control Control.
     * @api
     */
    addControl(control) {
      this.getControls().push(control);
    }

    /**
     * Add the given interaction to the map. If you want to add an interaction
     * at another point of the collection use `getInteractions()` and the methods
     * available on {@link module:ol/Collection~Collection}. This can be used to
     * stop the event propagation from the handleEvent function. The interactions
     * get to handle the events in the reverse order of this collection.
     * @param {import("./interaction/Interaction.js").default} interaction Interaction to add.
     * @api
     */
    addInteraction(interaction) {
      this.getInteractions().push(interaction);
    }

    /**
     * Adds the given layer to the top of this map. If you want to add a layer
     * elsewhere in the stack, use `getLayers()` and the methods available on
     * {@link module:ol/Collection~Collection}.
     * @param {import("./layer/Base.js").default} layer Layer.
     * @api
     */
    addLayer(layer) {
      const layers = this.getLayerGroup().getLayers();
      layers.push(layer);
    }

    /**
     * @param {import("./layer/Group.js").GroupEvent} event The layer add event.
     * @private
     */
    handleLayerAdd_(event) {
      setLayerMapProperty(event.layer, this);
    }

    /**
     * Add the given overlay to the map.
     * @param {import("./Overlay.js").default} overlay Overlay.
     * @api
     */
    addOverlay(overlay) {
      this.getOverlays().push(overlay);
    }

    /**
     * This deals with map's overlay collection changes.
     * @param {import("./Overlay.js").default} overlay Overlay.
     * @private
     */
    addOverlayInternal_(overlay) {
      const id = overlay.getId();
      if (id !== undefined) {
        this.overlayIdIndex_[id.toString()] = overlay;
      }
      overlay.setMap(this);
    }

    /**
     *
     * Clean up.
     */
    disposeInternal() {
      this.controls.clear();
      this.interactions.clear();
      this.overlays_.clear();
      this.setTarget(null);
      super.disposeInternal();
    }

    /**
     * Detect features that intersect a pixel on the viewport, and execute a
     * callback with each intersecting feature. Layers included in the detection can
     * be configured through the `layerFilter` option in `options`.
     * @param {import("./pixel.js").Pixel} pixel Pixel.
     * @param {function(import("./Feature.js").FeatureLike, import("./layer/Layer.js").default<import("./source/Source").default>, import("./geom/SimpleGeometry.js").default): T} callback Feature callback. The callback will be
     *     called with two arguments. The first argument is one
     *     {@link module:ol/Feature~Feature feature} or
     *     {@link module:ol/render/Feature~RenderFeature render feature} at the pixel, the second is
     *     the {@link module:ol/layer/Layer~Layer layer} of the feature and will be null for
     *     unmanaged layers. To stop detection, callback functions can return a
     *     truthy value.
     * @param {AtPixelOptions} [options] Optional options.
     * @return {T|undefined} Callback result, i.e. the return value of last
     * callback execution, or the first truthy callback return value.
     * @template T
     * @api
     */
    forEachFeatureAtPixel(pixel, callback, options) {
      if (!this.frameState_ || !this.renderer_) {
        return;
      }
      const coordinate = this.getCoordinateFromPixelInternal(pixel);
      options = options !== undefined ? options : {};
      const hitTolerance =
        options.hitTolerance !== undefined ? options.hitTolerance : 0;
      const layerFilter =
        options.layerFilter !== undefined ? options.layerFilter : TRUE;
      const checkWrapped = options.checkWrapped !== false;
      return this.renderer_.forEachFeatureAtCoordinate(
        coordinate,
        this.frameState_,
        hitTolerance,
        checkWrapped,
        callback,
        null,
        layerFilter,
        null
      );
    }

    /**
     * Get all features that intersect a pixel on the viewport.
     * @param {import("./pixel.js").Pixel} pixel Pixel.
     * @param {AtPixelOptions} [options] Optional options.
     * @return {Array<import("./Feature.js").FeatureLike>} The detected features or
     * an empty array if none were found.
     * @api
     */
    getFeaturesAtPixel(pixel, options) {
      const features = [];
      this.forEachFeatureAtPixel(
        pixel,
        function (feature) {
          features.push(feature);
        },
        options
      );
      return features;
    }

    /**
     * Get all layers from all layer groups.
     * @return {Array<import("./layer/Layer.js").default>} Layers.
     * @api
     */
    getAllLayers() {
      const layers = [];
      function addLayersFrom(layerGroup) {
        layerGroup.forEach(function (layer) {
          if (layer instanceof LayerGroup) {
            addLayersFrom(layer.getLayers());
          } else {
            layers.push(layer);
          }
        });
      }
      addLayersFrom(this.getLayers());
      return layers;
    }

    /**
     * Detect if features intersect a pixel on the viewport. Layers included in the
     * detection can be configured through the `layerFilter` option.
     * @param {import("./pixel.js").Pixel} pixel Pixel.
     * @param {AtPixelOptions} [options] Optional options.
     * @return {boolean} Is there a feature at the given pixel?
     * @api
     */
    hasFeatureAtPixel(pixel, options) {
      if (!this.frameState_ || !this.renderer_) {
        return false;
      }
      const coordinate = this.getCoordinateFromPixelInternal(pixel);
      options = options !== undefined ? options : {};
      const layerFilter =
        options.layerFilter !== undefined ? options.layerFilter : TRUE;
      const hitTolerance =
        options.hitTolerance !== undefined ? options.hitTolerance : 0;
      const checkWrapped = options.checkWrapped !== false;
      return this.renderer_.hasFeatureAtCoordinate(
        coordinate,
        this.frameState_,
        hitTolerance,
        checkWrapped,
        layerFilter,
        null
      );
    }

    /**
     * Returns the coordinate in user projection for a browser event.
     * @param {MouseEvent} event Event.
     * @return {import("./coordinate.js").Coordinate} Coordinate.
     * @api
     */
    getEventCoordinate(event) {
      return this.getCoordinateFromPixel(this.getEventPixel(event));
    }

    /**
     * Returns the coordinate in view projection for a browser event.
     * @param {MouseEvent} event Event.
     * @return {import("./coordinate.js").Coordinate} Coordinate.
     */
    getEventCoordinateInternal(event) {
      return this.getCoordinateFromPixelInternal(this.getEventPixel(event));
    }

    /**
     * Returns the map pixel position for a browser event relative to the viewport.
     * @param {UIEvent} event Event.
     * @return {import("./pixel.js").Pixel} Pixel.
     * @api
     */
    getEventPixel(event) {
      const viewportPosition = this.viewport_.getBoundingClientRect();
      const eventPosition =
        //FIXME Are we really calling this with a TouchEvent anywhere?
        'changedTouches' in event
          ? /** @type {TouchEvent} */ (event).changedTouches[0]
          : /** @type {MouseEvent} */ (event);

      return [
        eventPosition.clientX - viewportPosition.left,
        eventPosition.clientY - viewportPosition.top,
      ];
    }

    /**
     * Get the target in which this map is rendered.
     * Note that this returns what is entered as an option or in setTarget:
     * if that was an element, it returns an element; if a string, it returns that.
     * @return {HTMLElement|string|undefined} The Element or id of the Element that the
     *     map is rendered in.
     * @observable
     * @api
     */
    getTarget() {
      return /** @type {HTMLElement|string|undefined} */ (
        this.get(MapProperty.TARGET)
      );
    }

    /**
     * Get the DOM element into which this map is rendered. In contrast to
     * `getTarget` this method always return an `Element`, or `null` if the
     * map has no target.
     * @return {HTMLElement} The element that the map is rendered in.
     * @api
     */
    getTargetElement() {
      const target = this.getTarget();
      if (target !== undefined) {
        return typeof target === 'string'
          ? document.getElementById(target)
          : target;
      } else {
        return null;
      }
    }

    /**
     * Get the coordinate for a given pixel.  This returns a coordinate in the
     * user projection.
     * @param {import("./pixel.js").Pixel} pixel Pixel position in the map viewport.
     * @return {import("./coordinate.js").Coordinate} The coordinate for the pixel position.
     * @api
     */
    getCoordinateFromPixel(pixel) {
      return toUserCoordinate(
        this.getCoordinateFromPixelInternal(pixel),
        this.getView().getProjection()
      );
    }

    /**
     * Get the coordinate for a given pixel.  This returns a coordinate in the
     * map view projection.
     * @param {import("./pixel.js").Pixel} pixel Pixel position in the map viewport.
     * @return {import("./coordinate.js").Coordinate} The coordinate for the pixel position.
     */
    getCoordinateFromPixelInternal(pixel) {
      const frameState = this.frameState_;
      if (!frameState) {
        return null;
      } else {
        return apply(
          frameState.pixelToCoordinateTransform,
          pixel.slice()
        );
      }
    }

    /**
     * Get the map controls. Modifying this collection changes the controls
     * associated with the map.
     * @return {Collection<import("./control/Control.js").default>} Controls.
     * @api
     */
    getControls() {
      return this.controls;
    }

    /**
     * Get the map overlays. Modifying this collection changes the overlays
     * associated with the map.
     * @return {Collection<import("./Overlay.js").default>} Overlays.
     * @api
     */
    getOverlays() {
      return this.overlays_;
    }

    /**
     * Get an overlay by its identifier (the value returned by overlay.getId()).
     * Note that the index treats string and numeric identifiers as the same. So
     * `map.getOverlayById(2)` will return an overlay with id `'2'` or `2`.
     * @param {string|number} id Overlay identifier.
     * @return {import("./Overlay.js").default} Overlay.
     * @api
     */
    getOverlayById(id) {
      const overlay = this.overlayIdIndex_[id.toString()];
      return overlay !== undefined ? overlay : null;
    }

    /**
     * Get the map interactions. Modifying this collection changes the interactions
     * associated with the map.
     *
     * Interactions are used for e.g. pan, zoom and rotate.
     * @return {Collection<import("./interaction/Interaction.js").default>} Interactions.
     * @api
     */
    getInteractions() {
      return this.interactions;
    }

    /**
     * Get the layergroup associated with this map.
     * @return {LayerGroup} A layer group containing the layers in this map.
     * @observable
     * @api
     */
    getLayerGroup() {
      return /** @type {LayerGroup} */ (this.get(MapProperty.LAYERGROUP));
    }

    /**
     * Clear any existing layers and add layers to the map.
     * @param {Array<import("./layer/Base.js").default>|Collection<import("./layer/Base.js").default>} layers The layers to be added to the map.
     * @api
     */
    setLayers(layers) {
      const group = this.getLayerGroup();
      if (layers instanceof Collection) {
        group.setLayers(layers);
        return;
      }

      const collection = group.getLayers();
      collection.clear();
      collection.extend(layers);
    }

    /**
     * Get the collection of layers associated with this map.
     * @return {!Collection<import("./layer/Base.js").default>} Layers.
     * @api
     */
    getLayers() {
      const layers = this.getLayerGroup().getLayers();
      return layers;
    }

    /**
     * @return {boolean} Layers have sources that are still loading.
     */
    getLoadingOrNotReady() {
      const layerStatesArray = this.getLayerGroup().getLayerStatesArray();
      for (let i = 0, ii = layerStatesArray.length; i < ii; ++i) {
        const state = layerStatesArray[i];
        if (!state.visible) {
          continue;
        }
        const renderer = state.layer.getRenderer();
        if (renderer && !renderer.ready) {
          return true;
        }
        const source = state.layer.getSource();
        if (source && source.loading) {
          return true;
        }
      }
      return false;
    }

    /**
     * Get the pixel for a coordinate.  This takes a coordinate in the user
     * projection and returns the corresponding pixel.
     * @param {import("./coordinate.js").Coordinate} coordinate A map coordinate.
     * @return {import("./pixel.js").Pixel} A pixel position in the map viewport.
     * @api
     */
    getPixelFromCoordinate(coordinate) {
      const viewCoordinate = fromUserCoordinate(
        coordinate,
        this.getView().getProjection()
      );
      return this.getPixelFromCoordinateInternal(viewCoordinate);
    }

    /**
     * Get the pixel for a coordinate.  This takes a coordinate in the map view
     * projection and returns the corresponding pixel.
     * @param {import("./coordinate.js").Coordinate} coordinate A map coordinate.
     * @return {import("./pixel.js").Pixel} A pixel position in the map viewport.
     */
    getPixelFromCoordinateInternal(coordinate) {
      const frameState = this.frameState_;
      if (!frameState) {
        return null;
      } else {
        return apply(
          frameState.coordinateToPixelTransform,
          coordinate.slice(0, 2)
        );
      }
    }

    /**
     * Get the map renderer.
     * @return {import("./renderer/Map.js").default|null} Renderer
     */
    getRenderer() {
      return this.renderer_;
    }

    /**
     * Get the size of this map.
     * @return {import("./size.js").Size|undefined} The size in pixels of the map in the DOM.
     * @observable
     * @api
     */
    getSize() {
      return /** @type {import("./size.js").Size|undefined} */ (
        this.get(MapProperty.SIZE)
      );
    }

    /**
     * Get the view associated with this map. A view manages properties such as
     * center and resolution.
     * @return {View} The view that controls this map.
     * @observable
     * @api
     */
    getView() {
      return /** @type {View} */ (this.get(MapProperty.VIEW));
    }

    /**
     * Get the element that serves as the map viewport.
     * @return {HTMLElement} Viewport.
     * @api
     */
    getViewport() {
      return this.viewport_;
    }

    /**
     * Get the element that serves as the container for overlays.  Elements added to
     * this container will let mousedown and touchstart events through to the map,
     * so clicks and gestures on an overlay will trigger {@link module:ol/MapBrowserEvent~MapBrowserEvent}
     * events.
     * @return {!HTMLElement} The map's overlay container.
     */
    getOverlayContainer() {
      return this.overlayContainer_;
    }

    /**
     * Get the element that serves as a container for overlays that don't allow
     * event propagation. Elements added to this container won't let mousedown and
     * touchstart events through to the map, so clicks and gestures on an overlay
     * don't trigger any {@link module:ol/MapBrowserEvent~MapBrowserEvent}.
     * @return {!HTMLElement} The map's overlay container that stops events.
     */
    getOverlayContainerStopEvent() {
      return this.overlayContainerStopEvent_;
    }

    /**
     * @return {!Document} The document where the map is displayed.
     */
    getOwnerDocument() {
      const targetElement = this.getTargetElement();
      return targetElement ? targetElement.ownerDocument : document;
    }

    /**
     * @param {import("./Tile.js").default} tile Tile.
     * @param {string} tileSourceKey Tile source key.
     * @param {import("./coordinate.js").Coordinate} tileCenter Tile center.
     * @param {number} tileResolution Tile resolution.
     * @return {number} Tile priority.
     */
    getTilePriority(tile, tileSourceKey, tileCenter, tileResolution) {
      return getTilePriority(
        this.frameState_,
        tile,
        tileSourceKey,
        tileCenter,
        tileResolution
      );
    }

    /**
     * @param {UIEvent} browserEvent Browser event.
     * @param {string} [type] Type.
     */
    handleBrowserEvent(browserEvent, type) {
      type = type || browserEvent.type;
      const mapBrowserEvent = new MapBrowserEvent(type, this, browserEvent);
      this.handleMapBrowserEvent(mapBrowserEvent);
    }

    /**
     * @param {MapBrowserEvent} mapBrowserEvent The event to handle.
     */
    handleMapBrowserEvent(mapBrowserEvent) {
      if (!this.frameState_) {
        // With no view defined, we cannot translate pixels into geographical
        // coordinates so interactions cannot be used.
        return;
      }
      const originalEvent = /** @type {PointerEvent} */ (
        mapBrowserEvent.originalEvent
      );
      const eventType = originalEvent.type;
      if (
        eventType === PointerEventType.POINTERDOWN ||
        eventType === EventType$1.WHEEL ||
        eventType === EventType$1.KEYDOWN
      ) {
        const doc = this.getOwnerDocument();
        const rootNode = this.viewport_.getRootNode
          ? this.viewport_.getRootNode()
          : doc;
        const target = /** @type {Node} */ (originalEvent.target);
        if (
          // Abort if the target is a child of the container for elements whose events are not meant
          // to be handled by map interactions.
          this.overlayContainerStopEvent_.contains(target) ||
          // Abort if the event target is a child of the container that is no longer in the page.
          // It's possible for the target to no longer be in the page if it has been removed in an
          // event listener, this might happen in a Control that recreates it's content based on
          // user interaction either manually or via a render in something like https://reactjs.org/
          !(rootNode === doc ? doc.documentElement : rootNode).contains(target)
        ) {
          return;
        }
      }
      mapBrowserEvent.frameState = this.frameState_;
      if (this.dispatchEvent(mapBrowserEvent) !== false) {
        const interactionsArray = this.getInteractions().getArray().slice();
        for (let i = interactionsArray.length - 1; i >= 0; i--) {
          const interaction = interactionsArray[i];
          if (
            interaction.getMap() !== this ||
            !interaction.getActive() ||
            !this.getTargetElement()
          ) {
            continue;
          }
          const cont = interaction.handleEvent(mapBrowserEvent);
          if (!cont || mapBrowserEvent.propagationStopped) {
            break;
          }
        }
      }
    }

    /**
     * @protected
     */
    handlePostRender() {
      const frameState = this.frameState_;

      // Manage the tile queue
      // Image loads are expensive and a limited resource, so try to use them
      // efficiently:
      // * When the view is static we allow a large number of parallel tile loads
      //   to complete the frame as quickly as possible.
      // * When animating or interacting, image loads can cause janks, so we reduce
      //   the maximum number of loads per frame and limit the number of parallel
      //   tile loads to remain reactive to view changes and to reduce the chance of
      //   loading tiles that will quickly disappear from view.
      const tileQueue = this.tileQueue_;
      if (!tileQueue.isEmpty()) {
        let maxTotalLoading = this.maxTilesLoading_;
        let maxNewLoads = maxTotalLoading;
        if (frameState) {
          const hints = frameState.viewHints;
          if (hints[ViewHint.ANIMATING] || hints[ViewHint.INTERACTING]) {
            const lowOnFrameBudget = Date.now() - frameState.time > 8;
            maxTotalLoading = lowOnFrameBudget ? 0 : 8;
            maxNewLoads = lowOnFrameBudget ? 0 : 2;
          }
        }
        if (tileQueue.getTilesLoading() < maxTotalLoading) {
          tileQueue.reprioritize(); // FIXME only call if view has changed
          tileQueue.loadMoreTiles(maxTotalLoading, maxNewLoads);
        }
      }

      if (frameState && this.renderer_ && !frameState.animate) {
        if (this.renderComplete_ === true) {
          if (this.hasListener(EventType.RENDERCOMPLETE)) {
            this.renderer_.dispatchRenderEvent(
              EventType.RENDERCOMPLETE,
              frameState
            );
          }
          if (this.loaded_ === false) {
            this.loaded_ = true;
            this.dispatchEvent(
              new MapEvent(MapEventType.LOADEND, this, frameState)
            );
          }
        } else if (this.loaded_ === true) {
          this.loaded_ = false;
          this.dispatchEvent(
            new MapEvent(MapEventType.LOADSTART, this, frameState)
          );
        }
      }

      const postRenderFunctions = this.postRenderFunctions_;
      for (let i = 0, ii = postRenderFunctions.length; i < ii; ++i) {
        postRenderFunctions[i](this, frameState);
      }
      postRenderFunctions.length = 0;
    }

    /**
     * @private
     */
    handleSizeChanged_() {
      if (this.getView() && !this.getView().getAnimating()) {
        this.getView().resolveConstraints(0);
      }

      this.render();
    }

    /**
     * @private
     */
    handleTargetChanged_() {
      if (this.mapBrowserEventHandler_) {
        for (let i = 0, ii = this.targetChangeHandlerKeys_.length; i < ii; ++i) {
          unlistenByKey(this.targetChangeHandlerKeys_[i]);
        }
        this.targetChangeHandlerKeys_ = null;
        this.viewport_.removeEventListener(
          EventType$1.CONTEXTMENU,
          this.boundHandleBrowserEvent_
        );
        this.viewport_.removeEventListener(
          EventType$1.WHEEL,
          this.boundHandleBrowserEvent_
        );
        this.mapBrowserEventHandler_.dispose();
        this.mapBrowserEventHandler_ = null;
        removeNode$1(this.viewport_);
      }

      // target may be undefined, null, a string or an Element.
      // If it's a string we convert it to an Element before proceeding.
      // If it's not now an Element we remove the viewport from the DOM.
      // If it's an Element we append the viewport element to it.

      const targetElement = this.getTargetElement();
      if (!targetElement) {
        if (this.renderer_) {
          clearTimeout(this.postRenderTimeoutHandle_);
          this.postRenderTimeoutHandle_ = undefined;
          this.postRenderFunctions_.length = 0;
          this.renderer_.dispose();
          this.renderer_ = null;
        }
        if (this.animationDelayKey_) {
          cancelAnimationFrame(this.animationDelayKey_);
          this.animationDelayKey_ = undefined;
        }
      } else {
        targetElement.appendChild(this.viewport_);
        if (!this.renderer_) {
          this.renderer_ = new CompositeMapRenderer(this);
        }

        this.mapBrowserEventHandler_ = new MapBrowserEventHandler(
          this,
          this.moveTolerance_
        );
        for (const key in MapBrowserEventType) {
          this.mapBrowserEventHandler_.addEventListener(
            MapBrowserEventType[key],
            this.handleMapBrowserEvent.bind(this)
          );
        }
        this.viewport_.addEventListener(
          EventType$1.CONTEXTMENU,
          this.boundHandleBrowserEvent_,
          false
        );
        this.viewport_.addEventListener(
          EventType$1.WHEEL,
          this.boundHandleBrowserEvent_,
          PASSIVE_EVENT_LISTENERS ? {passive: false} : false
        );

        const defaultView = this.getOwnerDocument().defaultView;
        const keyboardEventTarget = !this.keyboardEventTarget_
          ? targetElement
          : this.keyboardEventTarget_;
        this.targetChangeHandlerKeys_ = [
          listen(
            keyboardEventTarget,
            EventType$1.KEYDOWN,
            this.handleBrowserEvent,
            this
          ),
          listen(
            keyboardEventTarget,
            EventType$1.KEYPRESS,
            this.handleBrowserEvent,
            this
          ),
          listen(defaultView, EventType$1.RESIZE, this.updateSize, this),
        ];
      }

      this.updateSize();
      // updateSize calls setSize, so no need to call this.render
      // ourselves here.
    }

    /**
     * @private
     */
    handleTileChange_() {
      this.render();
    }

    /**
     * @private
     */
    handleViewPropertyChanged_() {
      this.render();
    }

    /**
     * @private
     */
    handleViewChanged_() {
      if (this.viewPropertyListenerKey_) {
        unlistenByKey(this.viewPropertyListenerKey_);
        this.viewPropertyListenerKey_ = null;
      }
      if (this.viewChangeListenerKey_) {
        unlistenByKey(this.viewChangeListenerKey_);
        this.viewChangeListenerKey_ = null;
      }
      const view = this.getView();
      if (view) {
        this.updateViewportSize_();

        this.viewPropertyListenerKey_ = listen(
          view,
          ObjectEventType.PROPERTYCHANGE,
          this.handleViewPropertyChanged_,
          this
        );
        this.viewChangeListenerKey_ = listen(
          view,
          EventType$1.CHANGE,
          this.handleViewPropertyChanged_,
          this
        );

        view.resolveConstraints(0);
      }
      this.render();
    }

    /**
     * @private
     */
    handleLayerGroupChanged_() {
      if (this.layerGroupPropertyListenerKeys_) {
        this.layerGroupPropertyListenerKeys_.forEach(unlistenByKey);
        this.layerGroupPropertyListenerKeys_ = null;
      }
      const layerGroup = this.getLayerGroup();
      if (layerGroup) {
        this.handleLayerAdd_(new GroupEvent('addlayer', layerGroup));
        this.layerGroupPropertyListenerKeys_ = [
          listen(layerGroup, ObjectEventType.PROPERTYCHANGE, this.render, this),
          listen(layerGroup, EventType$1.CHANGE, this.render, this),
          listen(layerGroup, 'addlayer', this.handleLayerAdd_, this),
          listen(layerGroup, 'removelayer', this.handleLayerRemove_, this),
        ];
      }
      this.render();
    }

    /**
     * @return {boolean} Is rendered.
     */
    isRendered() {
      return !!this.frameState_;
    }

    /**
     * @private
     */
    animationDelay_() {
      this.animationDelayKey_ = undefined;
      this.renderFrame_(Date.now());
    }

    /**
     * Requests an immediate render in a synchronous manner.
     * @api
     */
    renderSync() {
      if (this.animationDelayKey_) {
        cancelAnimationFrame(this.animationDelayKey_);
      }
      this.animationDelay_();
    }

    /**
     * Redraws all text after new fonts have loaded
     */
    redrawText() {
      const layerStates = this.getLayerGroup().getLayerStatesArray();
      for (let i = 0, ii = layerStates.length; i < ii; ++i) {
        const layer = layerStates[i].layer;
        if (layer.hasRenderer()) {
          layer.getRenderer().handleFontsChanged();
        }
      }
    }

    /**
     * Request a map rendering (at the next animation frame).
     * @api
     */
    render() {
      if (this.renderer_ && this.animationDelayKey_ === undefined) {
        this.animationDelayKey_ = requestAnimationFrame(this.animationDelay_);
      }
    }

    /**
     * Remove the given control from the map.
     * @param {import("./control/Control.js").default} control Control.
     * @return {import("./control/Control.js").default|undefined} The removed control (or undefined
     *     if the control was not found).
     * @api
     */
    removeControl(control) {
      return this.getControls().remove(control);
    }

    /**
     * Remove the given interaction from the map.
     * @param {import("./interaction/Interaction.js").default} interaction Interaction to remove.
     * @return {import("./interaction/Interaction.js").default|undefined} The removed interaction (or
     *     undefined if the interaction was not found).
     * @api
     */
    removeInteraction(interaction) {
      return this.getInteractions().remove(interaction);
    }

    /**
     * Removes the given layer from the map.
     * @param {import("./layer/Base.js").default} layer Layer.
     * @return {import("./layer/Base.js").default|undefined} The removed layer (or undefined if the
     *     layer was not found).
     * @api
     */
    removeLayer(layer) {
      const layers = this.getLayerGroup().getLayers();
      return layers.remove(layer);
    }

    /**
     * @param {import("./layer/Group.js").GroupEvent} event The layer remove event.
     * @private
     */
    handleLayerRemove_(event) {
      removeLayerMapProperty(event.layer);
    }

    /**
     * Remove the given overlay from the map.
     * @param {import("./Overlay.js").default} overlay Overlay.
     * @return {import("./Overlay.js").default|undefined} The removed overlay (or undefined
     *     if the overlay was not found).
     * @api
     */
    removeOverlay(overlay) {
      return this.getOverlays().remove(overlay);
    }

    /**
     * @param {number} time Time.
     * @private
     */
    renderFrame_(time) {
      const size = this.getSize();
      const view = this.getView();
      const previousFrameState = this.frameState_;
      /** @type {?FrameState} */
      let frameState = null;
      if (size !== undefined && hasArea(size) && view && view.isDef()) {
        const viewHints = view.getHints(
          this.frameState_ ? this.frameState_.viewHints : undefined
        );
        const viewState = view.getState();
        frameState = {
          animate: false,
          coordinateToPixelTransform: this.coordinateToPixelTransform_,
          declutterTree: null,
          extent: getForViewAndSize(
            viewState.center,
            viewState.resolution,
            viewState.rotation,
            size
          ),
          index: this.frameIndex_++,
          layerIndex: 0,
          layerStatesArray: this.getLayerGroup().getLayerStatesArray(),
          pixelRatio: this.pixelRatio_,
          pixelToCoordinateTransform: this.pixelToCoordinateTransform_,
          postRenderFunctions: [],
          size: size,
          tileQueue: this.tileQueue_,
          time: time,
          usedTiles: {},
          viewState: viewState,
          viewHints: viewHints,
          wantedTiles: {},
          mapId: getUid(this),
          renderTargets: {},
        };
        if (viewState.nextCenter && viewState.nextResolution) {
          const rotation = isNaN(viewState.nextRotation)
            ? viewState.rotation
            : viewState.nextRotation;

          frameState.nextExtent = getForViewAndSize(
            viewState.nextCenter,
            viewState.nextResolution,
            rotation,
            size
          );
        }
      }

      this.frameState_ = frameState;
      this.renderer_.renderFrame(frameState);

      if (frameState) {
        if (frameState.animate) {
          this.render();
        }
        Array.prototype.push.apply(
          this.postRenderFunctions_,
          frameState.postRenderFunctions
        );

        if (previousFrameState) {
          const moveStart =
            !this.previousExtent_ ||
            (!isEmpty(this.previousExtent_) &&
              !equals$2(frameState.extent, this.previousExtent_));
          if (moveStart) {
            this.dispatchEvent(
              new MapEvent(MapEventType.MOVESTART, this, previousFrameState)
            );
            this.previousExtent_ = createOrUpdateEmpty(this.previousExtent_);
          }
        }

        const idle =
          this.previousExtent_ &&
          !frameState.viewHints[ViewHint.ANIMATING] &&
          !frameState.viewHints[ViewHint.INTERACTING] &&
          !equals$2(frameState.extent, this.previousExtent_);

        if (idle) {
          this.dispatchEvent(
            new MapEvent(MapEventType.MOVEEND, this, frameState)
          );
          clone(frameState.extent, this.previousExtent_);
        }
      }

      this.dispatchEvent(new MapEvent(MapEventType.POSTRENDER, this, frameState));

      this.renderComplete_ =
        this.hasListener(MapEventType.LOADSTART) ||
        this.hasListener(MapEventType.LOADEND) ||
        this.hasListener(EventType.RENDERCOMPLETE)
          ? !this.tileQueue_.getTilesLoading() &&
            !this.tileQueue_.getCount() &&
            !this.getLoadingOrNotReady()
          : undefined;

      if (!this.postRenderTimeoutHandle_) {
        this.postRenderTimeoutHandle_ = setTimeout(() => {
          this.postRenderTimeoutHandle_ = undefined;
          this.handlePostRender();
        }, 0);
      }
    }

    /**
     * Sets the layergroup of this map.
     * @param {LayerGroup} layerGroup A layer group containing the layers in this map.
     * @observable
     * @api
     */
    setLayerGroup(layerGroup) {
      const oldLayerGroup = this.getLayerGroup();
      if (oldLayerGroup) {
        this.handleLayerRemove_(new GroupEvent('removelayer', oldLayerGroup));
      }
      this.set(MapProperty.LAYERGROUP, layerGroup);
    }

    /**
     * Set the size of this map.
     * @param {import("./size.js").Size|undefined} size The size in pixels of the map in the DOM.
     * @observable
     * @api
     */
    setSize(size) {
      this.set(MapProperty.SIZE, size);
    }

    /**
     * Set the target element to render this map into.
     * @param {HTMLElement|string} [target] The Element or id of the Element
     *     that the map is rendered in.
     * @observable
     * @api
     */
    setTarget(target) {
      this.set(MapProperty.TARGET, target);
    }

    /**
     * Set the view for this map.
     * @param {View|Promise<import("./View.js").ViewOptions>} view The view that controls this map.
     * It is also possible to pass a promise that resolves to options for constructing a view.  This
     * alternative allows view properties to be resolved by sources or other components that load
     * view-related metadata.
     * @observable
     * @api
     */
    setView(view) {
      if (!view || view instanceof View) {
        this.set(MapProperty.VIEW, view);
        return;
      }
      this.set(MapProperty.VIEW, new View());

      const map = this;
      view.then(function (viewOptions) {
        map.setView(new View(viewOptions));
      });
    }

    /**
     * Force a recalculation of the map viewport size.  This should be called when
     * third-party code changes the size of the map viewport.
     * @api
     */
    updateSize() {
      const targetElement = this.getTargetElement();

      let size = undefined;
      if (targetElement) {
        const computedStyle = getComputedStyle(targetElement);
        const width =
          targetElement.offsetWidth -
          parseFloat(computedStyle['borderLeftWidth']) -
          parseFloat(computedStyle['paddingLeft']) -
          parseFloat(computedStyle['paddingRight']) -
          parseFloat(computedStyle['borderRightWidth']);
        const height =
          targetElement.offsetHeight -
          parseFloat(computedStyle['borderTopWidth']) -
          parseFloat(computedStyle['paddingTop']) -
          parseFloat(computedStyle['paddingBottom']) -
          parseFloat(computedStyle['borderBottomWidth']);
        if (!isNaN(width) && !isNaN(height)) {
          size = [width, height];
          if (
            !hasArea(size) &&
            !!(
              targetElement.offsetWidth ||
              targetElement.offsetHeight ||
              targetElement.getClientRects().length
            )
          ) {
            // eslint-disable-next-line
            console.warn(
              "No map visible because the map container's width or height are 0."
            );
          }
        }
      }

      this.setSize(size);
      this.updateViewportSize_();
    }

    /**
     * Recomputes the viewport size and save it on the view object (if any)
     * @private
     */
    updateViewportSize_() {
      const view = this.getView();
      if (view) {
        let size = undefined;
        const computedStyle = getComputedStyle(this.viewport_);
        if (computedStyle.width && computedStyle.height) {
          size = [
            parseInt(computedStyle.width, 10),
            parseInt(computedStyle.height, 10),
          ];
        }
        view.setViewportSize(size);
      }
    }
  }

  /**
   * @param {MapOptions} options Map options.
   * @return {MapOptionsInternal} Internal map options.
   */
  function createOptionsInternal(options) {
    /**
     * @type {HTMLElement|Document}
     */
    let keyboardEventTarget = null;
    if (options.keyboardEventTarget !== undefined) {
      keyboardEventTarget =
        typeof options.keyboardEventTarget === 'string'
          ? document.getElementById(options.keyboardEventTarget)
          : options.keyboardEventTarget;
    }

    /**
     * @type {Object<string, *>}
     */
    const values = {};

    const layerGroup =
      options.layers &&
      typeof (/** @type {?} */ (options.layers).getLayers) === 'function'
        ? /** @type {LayerGroup} */ (options.layers)
        : new LayerGroup({
            layers:
              /** @type {Collection<import("./layer/Base.js").default>|Array<import("./layer/Base.js").default>} */ (
                options.layers
              ),
          });
    values[MapProperty.LAYERGROUP] = layerGroup;

    values[MapProperty.TARGET] = options.target;

    values[MapProperty.VIEW] =
      options.view instanceof View ? options.view : new View();

    /** @type {Collection<import("./control/Control.js").default>} */
    let controls;
    if (options.controls !== undefined) {
      if (Array.isArray(options.controls)) {
        controls = new Collection(options.controls.slice());
      } else {
        assert(
          typeof (/** @type {?} */ (options.controls).getArray) === 'function',
          47
        ); // Expected `controls` to be an array or an `import("./Collection.js").Collection`
        controls = options.controls;
      }
    }

    /** @type {Collection<import("./interaction/Interaction").default>} */
    let interactions;
    if (options.interactions !== undefined) {
      if (Array.isArray(options.interactions)) {
        interactions = new Collection(options.interactions.slice());
      } else {
        assert(
          typeof (/** @type {?} */ (options.interactions).getArray) ===
            'function',
          48
        ); // Expected `interactions` to be an array or an `import("./Collection.js").Collection`
        interactions = options.interactions;
      }
    }

    /** @type {Collection<import("./Overlay.js").default>} */
    let overlays;
    if (options.overlays !== undefined) {
      if (Array.isArray(options.overlays)) {
        overlays = new Collection(options.overlays.slice());
      } else {
        assert(
          typeof (/** @type {?} */ (options.overlays).getArray) === 'function',
          49
        ); // Expected `overlays` to be an array or an `import("./Collection.js").Collection`
        overlays = options.overlays;
      }
    } else {
      overlays = new Collection();
    }

    return {
      controls: controls,
      interactions: interactions,
      keyboardEventTarget: keyboardEventTarget,
      overlays: overlays,
      values: values,
    };
  }

  /**
   * @module ol/Overlay
   */

  /**
   * @typedef {'bottom-left' | 'bottom-center' | 'bottom-right' | 'center-left' | 'center-center' | 'center-right' | 'top-left' | 'top-center' | 'top-right'} Positioning
   * The overlay position: `'bottom-left'`, `'bottom-center'`,  `'bottom-right'`,
   * `'center-left'`, `'center-center'`, `'center-right'`, `'top-left'`,
   * `'top-center'`, or `'top-right'`.
   */

  /**
   * @typedef {Object} Options
   * @property {number|string} [id] Set the overlay id. The overlay id can be used
   * with the {@link module:ol/Map~Map#getOverlayById} method.
   * @property {HTMLElement} [element] The overlay element.
   * @property {Array<number>} [offset=[0, 0]] Offsets in pixels used when positioning
   * the overlay. The first element in the
   * array is the horizontal offset. A positive value shifts the overlay right.
   * The second element in the array is the vertical offset. A positive value
   * shifts the overlay down.
   * @property {import("./coordinate.js").Coordinate} [position] The overlay position
   * in map projection.
   * @property {Positioning} [positioning='top-left'] Defines how
   * the overlay is actually positioned with respect to its `position` property.
   * Possible values are `'bottom-left'`, `'bottom-center'`, `'bottom-right'`,
   * `'center-left'`, `'center-center'`, `'center-right'`, `'top-left'`,
   * `'top-center'`, and `'top-right'`.
   * @property {boolean} [stopEvent=true] Whether event propagation to the map
   * viewport should be stopped. If `true` the overlay is placed in the same
   * container as that of the controls (CSS class name
   * `ol-overlaycontainer-stopevent`); if `false` it is placed in the container
   * with CSS class name specified by the `className` property.
   * @property {boolean} [insertFirst=true] Whether the overlay is inserted first
   * in the overlay container, or appended. If the overlay is placed in the same
   * container as that of the controls (see the `stopEvent` option) you will
   * probably set `insertFirst` to `true` so the overlay is displayed below the
   * controls.
   * @property {PanIntoViewOptions|boolean} [autoPan=false] Pan the map when calling
   * `setPosition`, so that the overlay is entirely visible in the current viewport.
   * @property {string} [className='ol-overlay-container ol-selectable'] CSS class
   * name.
   */

  /**
   * @typedef {Object} PanOptions
   * @property {number} [duration=1000] The duration of the animation in
   * milliseconds.
   * @property {function(number):number} [easing] The easing function to use. Can
   * be one from {@link module:ol/easing} or a custom function.
   * Default is {@link module:ol/easing.inAndOut}.
   */

  /**
   * @typedef {Object} PanIntoViewOptions
   * @property {PanOptions} [animation={}] The animation parameters for the pan
   * @property {number} [margin=20] The margin (in pixels) between the
   * overlay and the borders of the map when panning into view.
   */

  /**
   * @enum {string}
   * @protected
   */
  const Property$2 = {
    ELEMENT: 'element',
    MAP: 'map',
    OFFSET: 'offset',
    POSITION: 'position',
    POSITIONING: 'positioning',
  };

  /**
   * @typedef {import("./ObjectEventType").Types|'change:element'|'change:map'|'change:offset'|'change:position'|
   *   'change:positioning'} OverlayObjectEventTypes
   */

  /***
   * @template Return
   * @typedef {import("./Observable").OnSignature<import("./Observable").EventTypes, import("./events/Event.js").default, Return> &
   *   import("./Observable").OnSignature<OverlayObjectEventTypes, import("./Object").ObjectEvent, Return> &
   *   import("./Observable").CombinedOnSignature<import("./Observable").EventTypes|OverlayObjectEventTypes, Return>} OverlayOnSignature
   */

  /**
   * @classdesc
   * An element to be displayed over the map and attached to a single map
   * location.  Like {@link module:ol/control/Control~Control}, Overlays are
   * visible widgets. Unlike Controls, they are not in a fixed position on the
   * screen, but are tied to a geographical coordinate, so panning the map will
   * move an Overlay but not a Control.
   *
   * Example:
   *
   *     import Overlay from 'ol/Overlay';
   *
   *     // ...
   *     const popup = new Overlay({
   *       element: document.getElementById('popup'),
   *     });
   *     popup.setPosition(coordinate);
   *     map.addOverlay(popup);
   *
   * @api
   */
  class Overlay extends BaseObject {
    /**
     * @param {Options} options Overlay options.
     */
    constructor(options) {
      super();

      /***
       * @type {OverlayOnSignature<import("./events").EventsKey>}
       */
      this.on;

      /***
       * @type {OverlayOnSignature<import("./events").EventsKey>}
       */
      this.once;

      /***
       * @type {OverlayOnSignature<void>}
       */
      this.un;

      /**
       * @protected
       * @type {Options}
       */
      this.options = options;

      /**
       * @protected
       * @type {number|string|undefined}
       */
      this.id = options.id;

      /**
       * @protected
       * @type {boolean}
       */
      this.insertFirst =
        options.insertFirst !== undefined ? options.insertFirst : true;

      /**
       * @protected
       * @type {boolean}
       */
      this.stopEvent = options.stopEvent !== undefined ? options.stopEvent : true;

      /**
       * @protected
       * @type {HTMLElement}
       */
      this.element = document.createElement('div');
      this.element.className =
        options.className !== undefined
          ? options.className
          : 'ol-overlay-container ' + CLASS_SELECTABLE;
      this.element.style.position = 'absolute';
      this.element.style.pointerEvents = 'auto';

      /**
       * @protected
       * @type {PanIntoViewOptions|undefined}
       */
      this.autoPan = options.autoPan === true ? {} : options.autoPan || undefined;

      /**
       * @protected
       * @type {{transform_: string,
       *         visible: boolean}}
       */
      this.rendered = {
        transform_: '',
        visible: true,
      };

      /**
       * @protected
       * @type {?import("./events.js").EventsKey}
       */
      this.mapPostrenderListenerKey = null;

      this.addChangeListener(Property$2.ELEMENT, this.handleElementChanged);
      this.addChangeListener(Property$2.MAP, this.handleMapChanged);
      this.addChangeListener(Property$2.OFFSET, this.handleOffsetChanged);
      this.addChangeListener(Property$2.POSITION, this.handlePositionChanged);
      this.addChangeListener(Property$2.POSITIONING, this.handlePositioningChanged);

      if (options.element !== undefined) {
        this.setElement(options.element);
      }

      this.setOffset(options.offset !== undefined ? options.offset : [0, 0]);

      this.setPositioning(options.positioning || 'top-left');

      if (options.position !== undefined) {
        this.setPosition(options.position);
      }
    }

    /**
     * Get the DOM element of this overlay.
     * @return {HTMLElement|undefined} The Element containing the overlay.
     * @observable
     * @api
     */
    getElement() {
      return /** @type {HTMLElement|undefined} */ (this.get(Property$2.ELEMENT));
    }

    /**
     * Get the overlay identifier which is set on constructor.
     * @return {number|string|undefined} Id.
     * @api
     */
    getId() {
      return this.id;
    }

    /**
     * Get the map associated with this overlay.
     * @return {import("./Map.js").default|null} The map that the
     * overlay is part of.
     * @observable
     * @api
     */
    getMap() {
      return /** @type {import("./Map.js").default|null} */ (
        this.get(Property$2.MAP) || null
      );
    }

    /**
     * Get the offset of this overlay.
     * @return {Array<number>} The offset.
     * @observable
     * @api
     */
    getOffset() {
      return /** @type {Array<number>} */ (this.get(Property$2.OFFSET));
    }

    /**
     * Get the current position of this overlay.
     * @return {import("./coordinate.js").Coordinate|undefined} The spatial point that the overlay is
     *     anchored at.
     * @observable
     * @api
     */
    getPosition() {
      return /** @type {import("./coordinate.js").Coordinate|undefined} */ (
        this.get(Property$2.POSITION)
      );
    }

    /**
     * Get the current positioning of this overlay.
     * @return {Positioning} How the overlay is positioned
     *     relative to its point on the map.
     * @observable
     * @api
     */
    getPositioning() {
      return /** @type {Positioning} */ (this.get(Property$2.POSITIONING));
    }

    /**
     * @protected
     */
    handleElementChanged() {
      removeChildren(this.element);
      const element = this.getElement();
      if (element) {
        this.element.appendChild(element);
      }
    }

    /**
     * @protected
     */
    handleMapChanged() {
      if (this.mapPostrenderListenerKey) {
        removeNode$1(this.element);
        unlistenByKey(this.mapPostrenderListenerKey);
        this.mapPostrenderListenerKey = null;
      }
      const map = this.getMap();
      if (map) {
        this.mapPostrenderListenerKey = listen(
          map,
          MapEventType.POSTRENDER,
          this.render,
          this
        );
        this.updatePixelPosition();
        const container = this.stopEvent
          ? map.getOverlayContainerStopEvent()
          : map.getOverlayContainer();
        if (this.insertFirst) {
          container.insertBefore(this.element, container.childNodes[0] || null);
        } else {
          container.appendChild(this.element);
        }
        this.performAutoPan();
      }
    }

    /**
     * @protected
     */
    render() {
      this.updatePixelPosition();
    }

    /**
     * @protected
     */
    handleOffsetChanged() {
      this.updatePixelPosition();
    }

    /**
     * @protected
     */
    handlePositionChanged() {
      this.updatePixelPosition();
      this.performAutoPan();
    }

    /**
     * @protected
     */
    handlePositioningChanged() {
      this.updatePixelPosition();
    }

    /**
     * Set the DOM element to be associated with this overlay.
     * @param {HTMLElement|undefined} element The Element containing the overlay.
     * @observable
     * @api
     */
    setElement(element) {
      this.set(Property$2.ELEMENT, element);
    }

    /**
     * Set the map to be associated with this overlay.
     * @param {import("./Map.js").default|null} map The map that the
     * overlay is part of. Pass `null` to just remove the overlay from the current map.
     * @observable
     * @api
     */
    setMap(map) {
      this.set(Property$2.MAP, map);
    }

    /**
     * Set the offset for this overlay.
     * @param {Array<number>} offset Offset.
     * @observable
     * @api
     */
    setOffset(offset) {
      this.set(Property$2.OFFSET, offset);
    }

    /**
     * Set the position for this overlay. If the position is `undefined` the
     * overlay is hidden.
     * @param {import("./coordinate.js").Coordinate|undefined} position The spatial point that the overlay
     *     is anchored at.
     * @observable
     * @api
     */
    setPosition(position) {
      this.set(Property$2.POSITION, position);
    }

    /**
     * Pan the map so that the overlay is entirely visible in the current viewport
     * (if necessary) using the configured autoPan parameters
     * @protected
     */
    performAutoPan() {
      if (this.autoPan) {
        this.panIntoView(this.autoPan);
      }
    }

    /**
     * Pan the map so that the overlay is entirely visible in the current viewport
     * (if necessary).
     * @param {PanIntoViewOptions} [panIntoViewOptions] Options for the pan action
     * @api
     */
    panIntoView(panIntoViewOptions) {
      const map = this.getMap();

      if (!map || !map.getTargetElement() || !this.get(Property$2.POSITION)) {
        return;
      }

      const mapRect = this.getRect(map.getTargetElement(), map.getSize());
      const element = this.getElement();
      const overlayRect = this.getRect(element, [
        outerWidth(element),
        outerHeight(element),
      ]);

      panIntoViewOptions = panIntoViewOptions || {};

      const myMargin =
        panIntoViewOptions.margin === undefined ? 20 : panIntoViewOptions.margin;
      if (!containsExtent(mapRect, overlayRect)) {
        // the overlay is not completely inside the viewport, so pan the map
        const offsetLeft = overlayRect[0] - mapRect[0];
        const offsetRight = mapRect[2] - overlayRect[2];
        const offsetTop = overlayRect[1] - mapRect[1];
        const offsetBottom = mapRect[3] - overlayRect[3];

        const delta = [0, 0];
        if (offsetLeft < 0) {
          // move map to the left
          delta[0] = offsetLeft - myMargin;
        } else if (offsetRight < 0) {
          // move map to the right
          delta[0] = Math.abs(offsetRight) + myMargin;
        }
        if (offsetTop < 0) {
          // move map up
          delta[1] = offsetTop - myMargin;
        } else if (offsetBottom < 0) {
          // move map down
          delta[1] = Math.abs(offsetBottom) + myMargin;
        }

        if (delta[0] !== 0 || delta[1] !== 0) {
          const center = /** @type {import("./coordinate.js").Coordinate} */ (
            map.getView().getCenterInternal()
          );
          const centerPx = map.getPixelFromCoordinateInternal(center);
          if (!centerPx) {
            return;
          }
          const newCenterPx = [centerPx[0] + delta[0], centerPx[1] + delta[1]];

          const panOptions = panIntoViewOptions.animation || {};
          map.getView().animateInternal({
            center: map.getCoordinateFromPixelInternal(newCenterPx),
            duration: panOptions.duration,
            easing: panOptions.easing,
          });
        }
      }
    }

    /**
     * Get the extent of an element relative to the document
     * @param {HTMLElement} element The element.
     * @param {import("./size.js").Size} size The size of the element.
     * @return {import("./extent.js").Extent} The extent.
     * @protected
     */
    getRect(element, size) {
      const box = element.getBoundingClientRect();
      const offsetX = box.left + window.pageXOffset;
      const offsetY = box.top + window.pageYOffset;
      return [offsetX, offsetY, offsetX + size[0], offsetY + size[1]];
    }

    /**
     * Set the positioning for this overlay.
     * @param {Positioning} positioning how the overlay is
     *     positioned relative to its point on the map.
     * @observable
     * @api
     */
    setPositioning(positioning) {
      this.set(Property$2.POSITIONING, positioning);
    }

    /**
     * Modify the visibility of the element.
     * @param {boolean} visible Element visibility.
     * @protected
     */
    setVisible(visible) {
      if (this.rendered.visible !== visible) {
        this.element.style.display = visible ? '' : 'none';
        this.rendered.visible = visible;
      }
    }

    /**
     * Update pixel position.
     * @protected
     */
    updatePixelPosition() {
      const map = this.getMap();
      const position = this.getPosition();
      if (!map || !map.isRendered() || !position) {
        this.setVisible(false);
        return;
      }

      const pixel = map.getPixelFromCoordinate(position);
      const mapSize = map.getSize();
      this.updateRenderedPosition(pixel, mapSize);
    }

    /**
     * @param {import("./pixel.js").Pixel} pixel The pixel location.
     * @param {import("./size.js").Size|undefined} mapSize The map size.
     * @protected
     */
    updateRenderedPosition(pixel, mapSize) {
      const style = this.element.style;
      const offset = this.getOffset();

      const positioning = this.getPositioning();

      this.setVisible(true);

      const x = Math.round(pixel[0] + offset[0]) + 'px';
      const y = Math.round(pixel[1] + offset[1]) + 'px';
      let posX = '0%';
      let posY = '0%';
      if (
        positioning == 'bottom-right' ||
        positioning == 'center-right' ||
        positioning == 'top-right'
      ) {
        posX = '-100%';
      } else if (
        positioning == 'bottom-center' ||
        positioning == 'center-center' ||
        positioning == 'top-center'
      ) {
        posX = '-50%';
      }
      if (
        positioning == 'bottom-left' ||
        positioning == 'bottom-center' ||
        positioning == 'bottom-right'
      ) {
        posY = '-100%';
      } else if (
        positioning == 'center-left' ||
        positioning == 'center-center' ||
        positioning == 'center-right'
      ) {
        posY = '-50%';
      }
      const transform = `translate(${posX}, ${posY}) translate(${x}, ${y})`;
      if (this.rendered.transform_ != transform) {
        this.rendered.transform_ = transform;
        style.transform = transform;
      }
    }

    /**
     * returns the options this Overlay has been created with
     * @return {Options} overlay options
     */
    getOptions() {
      return this.options;
    }
  }

  /**
   * @module ol/structs/LRUCache
   */

  /**
   * @typedef {Object} Entry
   * @property {string} key_ Key.
   * @property {Object} newer Newer.
   * @property {Object} older Older.
   * @property {*} value_ Value.
   */

  /**
   * @classdesc
   * Implements a Least-Recently-Used cache where the keys do not conflict with
   * Object's properties (e.g. 'hasOwnProperty' is not allowed as a key). Expiring
   * items from the cache is the responsibility of the user.
   *
   * @fires import("../events/Event.js").default
   * @template T
   */
  class LRUCache {
    /**
     * @param {number} [highWaterMark] High water mark.
     */
    constructor(highWaterMark) {
      /**
       * Desired max cache size after expireCache(). If set to 0, no cache entries
       * will be pruned at all.
       * @type {number}
       */
      this.highWaterMark = highWaterMark !== undefined ? highWaterMark : 2048;

      /**
       * @private
       * @type {number}
       */
      this.count_ = 0;

      /**
       * @private
       * @type {!Object<string, Entry>}
       */
      this.entries_ = {};

      /**
       * @private
       * @type {?Entry}
       */
      this.oldest_ = null;

      /**
       * @private
       * @type {?Entry}
       */
      this.newest_ = null;
    }

    /**
     * @return {boolean} Can expire cache.
     */
    canExpireCache() {
      return this.highWaterMark > 0 && this.getCount() > this.highWaterMark;
    }

    /**
     * Expire the cache.
     * @param {!Object<string, boolean>} [keep] Keys to keep. To be implemented by subclasses.
     */
    expireCache(keep) {
      while (this.canExpireCache()) {
        this.pop();
      }
    }

    /**
     * FIXME empty description for jsdoc
     */
    clear() {
      this.count_ = 0;
      this.entries_ = {};
      this.oldest_ = null;
      this.newest_ = null;
    }

    /**
     * @param {string} key Key.
     * @return {boolean} Contains key.
     */
    containsKey(key) {
      return this.entries_.hasOwnProperty(key);
    }

    /**
     * @param {function(T, string, LRUCache<T>): ?} f The function
     *     to call for every entry from the oldest to the newer. This function takes
     *     3 arguments (the entry value, the entry key and the LRUCache object).
     *     The return value is ignored.
     */
    forEach(f) {
      let entry = this.oldest_;
      while (entry) {
        f(entry.value_, entry.key_, this);
        entry = entry.newer;
      }
    }

    /**
     * @param {string} key Key.
     * @param {*} [options] Options (reserved for subclasses).
     * @return {T} Value.
     */
    get(key, options) {
      const entry = this.entries_[key];
      assert(entry !== undefined, 15); // Tried to get a value for a key that does not exist in the cache
      if (entry === this.newest_) {
        return entry.value_;
      } else if (entry === this.oldest_) {
        this.oldest_ = /** @type {Entry} */ (this.oldest_.newer);
        this.oldest_.older = null;
      } else {
        entry.newer.older = entry.older;
        entry.older.newer = entry.newer;
      }
      entry.newer = null;
      entry.older = this.newest_;
      this.newest_.newer = entry;
      this.newest_ = entry;
      return entry.value_;
    }

    /**
     * Remove an entry from the cache.
     * @param {string} key The entry key.
     * @return {T} The removed entry.
     */
    remove(key) {
      const entry = this.entries_[key];
      assert(entry !== undefined, 15); // Tried to get a value for a key that does not exist in the cache
      if (entry === this.newest_) {
        this.newest_ = /** @type {Entry} */ (entry.older);
        if (this.newest_) {
          this.newest_.newer = null;
        }
      } else if (entry === this.oldest_) {
        this.oldest_ = /** @type {Entry} */ (entry.newer);
        if (this.oldest_) {
          this.oldest_.older = null;
        }
      } else {
        entry.newer.older = entry.older;
        entry.older.newer = entry.newer;
      }
      delete this.entries_[key];
      --this.count_;
      return entry.value_;
    }

    /**
     * @return {number} Count.
     */
    getCount() {
      return this.count_;
    }

    /**
     * @return {Array<string>} Keys.
     */
    getKeys() {
      const keys = new Array(this.count_);
      let i = 0;
      let entry;
      for (entry = this.newest_; entry; entry = entry.older) {
        keys[i++] = entry.key_;
      }
      return keys;
    }

    /**
     * @return {Array<T>} Values.
     */
    getValues() {
      const values = new Array(this.count_);
      let i = 0;
      let entry;
      for (entry = this.newest_; entry; entry = entry.older) {
        values[i++] = entry.value_;
      }
      return values;
    }

    /**
     * @return {T} Last value.
     */
    peekLast() {
      return this.oldest_.value_;
    }

    /**
     * @return {string} Last key.
     */
    peekLastKey() {
      return this.oldest_.key_;
    }

    /**
     * Get the key of the newest item in the cache.  Throws if the cache is empty.
     * @return {string} The newest key.
     */
    peekFirstKey() {
      return this.newest_.key_;
    }

    /**
     * Return an entry without updating least recently used time.
     * @param {string} key Key.
     * @return {T} Value.
     */
    peek(key) {
      if (!this.containsKey(key)) {
        return undefined;
      }
      return this.entries_[key].value_;
    }

    /**
     * @return {T} value Value.
     */
    pop() {
      const entry = this.oldest_;
      delete this.entries_[entry.key_];
      if (entry.newer) {
        entry.newer.older = null;
      }
      this.oldest_ = /** @type {Entry} */ (entry.newer);
      if (!this.oldest_) {
        this.newest_ = null;
      }
      --this.count_;
      return entry.value_;
    }

    /**
     * @param {string} key Key.
     * @param {T} value Value.
     */
    replace(key, value) {
      this.get(key); // update `newest_`
      this.entries_[key].value_ = value;
    }

    /**
     * @param {string} key Key.
     * @param {T} value Value.
     */
    set(key, value) {
      assert(!(key in this.entries_), 16); // Tried to set a value for a key that is used already
      const entry = {
        key_: key,
        newer: null,
        older: this.newest_,
        value_: value,
      };
      if (!this.newest_) {
        this.oldest_ = entry;
      } else {
        this.newest_.newer = entry;
      }
      this.newest_ = entry;
      this.entries_[key] = entry;
      ++this.count_;
    }

    /**
     * Set a maximum number of entries for the cache.
     * @param {number} size Cache size.
     * @api
     */
    setSize(size) {
      this.highWaterMark = size;
    }
  }

  /**
   * @module ol/tilecoord
   */

  /**
   * An array of three numbers representing the location of a tile in a tile
   * grid. The order is `z` (zoom level), `x` (column), and `y` (row).
   * @typedef {Array<number>} TileCoord
   * @api
   */

  /**
   * @param {number} z Z.
   * @param {number} x X.
   * @param {number} y Y.
   * @param {TileCoord} [tileCoord] Tile coordinate.
   * @return {TileCoord} Tile coordinate.
   */
  function createOrUpdate$1(z, x, y, tileCoord) {
    if (tileCoord !== undefined) {
      tileCoord[0] = z;
      tileCoord[1] = x;
      tileCoord[2] = y;
      return tileCoord;
    } else {
      return [z, x, y];
    }
  }

  /**
   * @param {number} z Z.
   * @param {number} x X.
   * @param {number} y Y.
   * @return {string} Key.
   */
  function getKeyZXY(z, x, y) {
    return z + '/' + x + '/' + y;
  }

  /**
   * Get the key for a tile coord.
   * @param {TileCoord} tileCoord The tile coord.
   * @return {string} Key.
   */
  function getKey(tileCoord) {
    return getKeyZXY(tileCoord[0], tileCoord[1], tileCoord[2]);
  }

  /**
   * Get the tile cache key for a tile key obtained through `tile.getKey()`.
   * @param {string} tileKey The tile key.
   * @return {string} The cache key.
   */
  function getCacheKeyForTileKey(tileKey) {
    const [z, x, y] = tileKey
      .substring(tileKey.lastIndexOf('/') + 1, tileKey.length)
      .split(',')
      .map(Number);
    return getKeyZXY(z, x, y);
  }

  /**
   * Get a tile coord given a key.
   * @param {string} key The tile coord key.
   * @return {TileCoord} The tile coord.
   */
  function fromKey(key) {
    return key.split('/').map(Number);
  }

  /**
   * @param {TileCoord} tileCoord Tile coord.
   * @return {number} Hash.
   */
  function hash(tileCoord) {
    return (tileCoord[1] << tileCoord[0]) + tileCoord[2];
  }

  /**
   * @param {TileCoord} tileCoord Tile coordinate.
   * @param {!import("./tilegrid/TileGrid.js").default} tileGrid Tile grid.
   * @return {boolean} Tile coordinate is within extent and zoom level range.
   */
  function withinExtentAndZ(tileCoord, tileGrid) {
    const z = tileCoord[0];
    const x = tileCoord[1];
    const y = tileCoord[2];

    if (tileGrid.getMinZoom() > z || z > tileGrid.getMaxZoom()) {
      return false;
    }
    const tileRange = tileGrid.getFullTileRange(z);
    if (!tileRange) {
      return true;
    } else {
      return tileRange.containsXY(x, y);
    }
  }

  /**
   * @module ol/TileCache
   */

  class TileCache extends LRUCache {
    clear() {
      while (this.getCount() > 0) {
        this.pop().release();
      }
      super.clear();
    }

    /**
     * @param {!Object<string, boolean>} usedTiles Used tiles.
     */
    expireCache(usedTiles) {
      while (this.canExpireCache()) {
        const tile = this.peekLast();
        if (tile.getKey() in usedTiles) {
          break;
        } else {
          this.pop().release();
        }
      }
    }

    /**
     * Prune all tiles from the cache that don't have the same z as the newest tile.
     */
    pruneExceptNewestZ() {
      if (this.getCount() === 0) {
        return;
      }
      const key = this.peekFirstKey();
      const tileCoord = fromKey(key);
      const z = tileCoord[0];
      this.forEach(
        function (tile) {
          if (tile.tileCoord[0] !== z) {
            this.remove(getKey(tile.tileCoord));
            tile.release();
          }
        }.bind(this)
      );
    }
  }

  /**
   * @module ol/TileRange
   */

  /**
   * A representation of a contiguous block of tiles.  A tile range is specified
   * by its min/max tile coordinates and is inclusive of coordinates.
   */
  class TileRange {
    /**
     * @param {number} minX Minimum X.
     * @param {number} maxX Maximum X.
     * @param {number} minY Minimum Y.
     * @param {number} maxY Maximum Y.
     */
    constructor(minX, maxX, minY, maxY) {
      /**
       * @type {number}
       */
      this.minX = minX;

      /**
       * @type {number}
       */
      this.maxX = maxX;

      /**
       * @type {number}
       */
      this.minY = minY;

      /**
       * @type {number}
       */
      this.maxY = maxY;
    }

    /**
     * @param {import("./tilecoord.js").TileCoord} tileCoord Tile coordinate.
     * @return {boolean} Contains tile coordinate.
     */
    contains(tileCoord) {
      return this.containsXY(tileCoord[1], tileCoord[2]);
    }

    /**
     * @param {TileRange} tileRange Tile range.
     * @return {boolean} Contains.
     */
    containsTileRange(tileRange) {
      return (
        this.minX <= tileRange.minX &&
        tileRange.maxX <= this.maxX &&
        this.minY <= tileRange.minY &&
        tileRange.maxY <= this.maxY
      );
    }

    /**
     * @param {number} x Tile coordinate x.
     * @param {number} y Tile coordinate y.
     * @return {boolean} Contains coordinate.
     */
    containsXY(x, y) {
      return this.minX <= x && x <= this.maxX && this.minY <= y && y <= this.maxY;
    }

    /**
     * @param {TileRange} tileRange Tile range.
     * @return {boolean} Equals.
     */
    equals(tileRange) {
      return (
        this.minX == tileRange.minX &&
        this.minY == tileRange.minY &&
        this.maxX == tileRange.maxX &&
        this.maxY == tileRange.maxY
      );
    }

    /**
     * @param {TileRange} tileRange Tile range.
     */
    extend(tileRange) {
      if (tileRange.minX < this.minX) {
        this.minX = tileRange.minX;
      }
      if (tileRange.maxX > this.maxX) {
        this.maxX = tileRange.maxX;
      }
      if (tileRange.minY < this.minY) {
        this.minY = tileRange.minY;
      }
      if (tileRange.maxY > this.maxY) {
        this.maxY = tileRange.maxY;
      }
    }

    /**
     * @return {number} Height.
     */
    getHeight() {
      return this.maxY - this.minY + 1;
    }

    /**
     * @return {import("./size.js").Size} Size.
     */
    getSize() {
      return [this.getWidth(), this.getHeight()];
    }

    /**
     * @return {number} Width.
     */
    getWidth() {
      return this.maxX - this.minX + 1;
    }

    /**
     * @param {TileRange} tileRange Tile range.
     * @return {boolean} Intersects.
     */
    intersects(tileRange) {
      return (
        this.minX <= tileRange.maxX &&
        this.maxX >= tileRange.minX &&
        this.minY <= tileRange.maxY &&
        this.maxY >= tileRange.minY
      );
    }
  }

  /**
   * @param {number} minX Minimum X.
   * @param {number} maxX Maximum X.
   * @param {number} minY Minimum Y.
   * @param {number} maxY Maximum Y.
   * @param {TileRange} [tileRange] TileRange.
   * @return {TileRange} Tile range.
   */
  function createOrUpdate(minX, maxX, minY, maxY, tileRange) {
    if (tileRange !== undefined) {
      tileRange.minX = minX;
      tileRange.maxX = maxX;
      tileRange.minY = minY;
      tileRange.maxY = maxY;
      return tileRange;
    } else {
      return new TileRange(minX, maxX, minY, maxY);
    }
  }

  /**
   * @module ol/VectorRenderTile
   */

  /**
   * @typedef {Object} ReplayState
   * @property {boolean} dirty Dirty.
   * @property {null|import("./render.js").OrderFunction} renderedRenderOrder RenderedRenderOrder.
   * @property {number} renderedTileRevision RenderedTileRevision.
   * @property {number} renderedResolution RenderedResolution.
   * @property {number} renderedRevision RenderedRevision.
   * @property {number} renderedTileResolution RenderedTileResolution.
   * @property {number} renderedTileZ RenderedTileZ.
   */

  /**
   * @type {Array<HTMLCanvasElement>}
   */
  const canvasPool$2 = [];

  class VectorRenderTile extends Tile {
    /**
     * @param {import("./tilecoord.js").TileCoord} tileCoord Tile coordinate.
     * @param {import("./TileState.js").default} state State.
     * @param {import("./tilecoord.js").TileCoord} urlTileCoord Wrapped tile coordinate for source urls.
     * @param {function(VectorRenderTile):Array<import("./VectorTile").default>} getSourceTiles Function
     * to get source tiles for this tile.
     */
    constructor(tileCoord, state, urlTileCoord, getSourceTiles) {
      super(tileCoord, state, {transition: 0});

      /**
       * @private
       * @type {!Object<string, CanvasRenderingContext2D>}
       */
      this.context_ = {};

      /**
       * Executor groups by layer uid. Entries are read/written by the renderer.
       * @type {Object<string, Array<import("./render/canvas/ExecutorGroup.js").default>>}
       */
      this.executorGroups = {};

      /**
       * Executor groups for decluttering, by layer uid. Entries are read/written by the renderer.
       * @type {Object<string, Array<import("./render/canvas/ExecutorGroup.js").default>>}
       */
      this.declutterExecutorGroups = {};

      /**
       * Number of loading source tiles. Read/written by the source.
       * @type {number}
       */
      this.loadingSourceTiles = 0;

      /**
       * @type {Object<number, ImageData>}
       */
      this.hitDetectionImageData = {};

      /**
       * @private
       * @type {!Object<string, ReplayState>}
       */
      this.replayState_ = {};

      /**
       * @type {Array<import("./VectorTile.js").default>}
       */
      this.sourceTiles = [];

      /**
       * @type {Object<string, boolean>}
       */
      this.errorTileKeys = {};

      /**
       * @type {number}
       */
      this.wantedResolution;

      /**
       * @type {!function():Array<import("./VectorTile.js").default>}
       */
      this.getSourceTiles = getSourceTiles.bind(undefined, this);

      /**
       * @type {import("./tilecoord.js").TileCoord}
       */
      this.wrappedTileCoord = urlTileCoord;
    }

    /**
     * @param {import("./layer/Layer.js").default} layer Layer.
     * @return {CanvasRenderingContext2D} The rendering context.
     */
    getContext(layer) {
      const key = getUid(layer);
      if (!(key in this.context_)) {
        this.context_[key] = createCanvasContext2D(1, 1, canvasPool$2);
      }
      return this.context_[key];
    }

    /**
     * @param {import("./layer/Layer.js").default} layer Layer.
     * @return {boolean} Tile has a rendering context for the given layer.
     */
    hasContext(layer) {
      return getUid(layer) in this.context_;
    }

    /**
     * Get the Canvas for this tile.
     * @param {import("./layer/Layer.js").default} layer Layer.
     * @return {HTMLCanvasElement} Canvas.
     */
    getImage(layer) {
      return this.hasContext(layer) ? this.getContext(layer).canvas : null;
    }

    /**
     * @param {import("./layer/Layer.js").default} layer Layer.
     * @return {ReplayState} The replay state.
     */
    getReplayState(layer) {
      const key = getUid(layer);
      if (!(key in this.replayState_)) {
        this.replayState_[key] = {
          dirty: false,
          renderedRenderOrder: null,
          renderedResolution: NaN,
          renderedRevision: -1,
          renderedTileResolution: NaN,
          renderedTileRevision: -1,
          renderedTileZ: -1,
        };
      }
      return this.replayState_[key];
    }

    /**
     * Load the tile.
     */
    load() {
      this.getSourceTiles();
    }

    /**
     * Remove from the cache due to expiry
     */
    release() {
      for (const key in this.context_) {
        const context = this.context_[key];
        releaseCanvas$1(context);
        canvasPool$2.push(context.canvas);
        delete this.context_[key];
      }
      super.release();
    }
  }

  /**
   * @module ol/VectorTile
   */

  class VectorTile$1 extends Tile {
    /**
     * @param {import("./tilecoord.js").TileCoord} tileCoord Tile coordinate.
     * @param {import("./TileState.js").default} state State.
     * @param {string} src Data source url.
     * @param {import("./format/Feature.js").default} format Feature format.
     * @param {import("./Tile.js").LoadFunction} tileLoadFunction Tile load function.
     * @param {import("./Tile.js").Options} [options] Tile options.
     */
    constructor(tileCoord, state, src, format, tileLoadFunction, options) {
      super(tileCoord, state, options);

      /**
       * Extent of this tile; set by the source.
       * @type {import("./extent.js").Extent}
       */
      this.extent = null;

      /**
       * @private
       * @type {import("./format/Feature.js").default}
       */
      this.format_ = format;

      /**
       * @private
       * @type {Array<import("./Feature.js").default>}
       */
      this.features_ = null;

      /**
       * @private
       * @type {import("./featureloader.js").FeatureLoader}
       */
      this.loader_;

      /**
       * Feature projection of this tile; set by the source.
       * @type {import("./proj/Projection.js").default}
       */
      this.projection = null;

      /**
       * Resolution of this tile; set by the source.
       * @type {number}
       */
      this.resolution;

      /**
       * @private
       * @type {import("./Tile.js").LoadFunction}
       */
      this.tileLoadFunction_ = tileLoadFunction;

      /**
       * @private
       * @type {string}
       */
      this.url_ = src;

      this.key = src;
    }

    /**
     * Get the feature format assigned for reading this tile's features.
     * @return {import("./format/Feature.js").default} Feature format.
     * @api
     */
    getFormat() {
      return this.format_;
    }

    /**
     * Get the features for this tile. Geometries will be in the view projection.
     * @return {Array<import("./Feature.js").FeatureLike>} Features.
     * @api
     */
    getFeatures() {
      return this.features_;
    }

    /**
     * Load not yet loaded URI.
     */
    load() {
      if (this.state == TileState.IDLE) {
        this.setState(TileState.LOADING);
        this.tileLoadFunction_(this, this.url_);
        if (this.loader_) {
          this.loader_(this.extent, this.resolution, this.projection);
        }
      }
    }

    /**
     * Handler for successful tile load.
     * @param {Array<import("./Feature.js").default>} features The loaded features.
     * @param {import("./proj/Projection.js").default} dataProjection Data projection.
     */
    onLoad(features, dataProjection) {
      this.setFeatures(features);
    }

    /**
     * Handler for tile load errors.
     */
    onError() {
      this.setState(TileState.ERROR);
    }

    /**
     * Function for use in an {@link module:ol/source/VectorTile~VectorTile}'s `tileLoadFunction`.
     * Sets the features for the tile.
     * @param {Array<import("./Feature.js").default>} features Features.
     * @api
     */
    setFeatures(features) {
      this.features_ = features;
      this.setState(TileState.LOADED);
    }

    /**
     * Set the feature loader for reading this tile's features.
     * @param {import("./featureloader.js").FeatureLoader} loader Feature loader.
     * @api
     */
    setLoader(loader) {
      this.loader_ = loader;
    }
  }

  /**
   * @module ol/colorlike
   */

  /**
   * A type accepted by CanvasRenderingContext2D.fillStyle
   * or CanvasRenderingContext2D.strokeStyle.
   * Represents a color, pattern, or gradient. The origin for patterns and
   * gradients as fill style is an increment of 512 css pixels from map coordinate
   * `[0, 0]`. For seamless repeat patterns, width and height of the pattern image
   * must be a factor of two (2, 4, 8, ..., 512).
   *
   * @typedef {string|CanvasPattern|CanvasGradient} ColorLike
   * @api
   */

  /**
   * @param {import("./color.js").Color|ColorLike} color Color.
   * @return {ColorLike} The color as an {@link ol/colorlike~ColorLike}.
   * @api
   */
  function asColorLike(color) {
    if (Array.isArray(color)) {
      return toString$2(color);
    } else {
      return color;
    }
  }

  /**
   * @module ol/featureloader
   */

  /**
   *
   * @type {boolean}
   * @private
   */
  let withCredentials = false;

  /**
   * {@link module:ol/source/Vector~VectorSource} sources use a function of this type to
   * load features.
   *
   * This function takes up to 5 arguments. These are an {@link module:ol/extent~Extent} representing
   * the area to be loaded, a `{number}` representing the resolution (map units per pixel), an
   * {@link module:ol/proj/Projection~Projection} for the projection, an optional success callback that should get
   * the loaded features passed as an argument and an optional failure callback with no arguments. If
   * the callbacks are not used, the corresponding vector source will not fire `'featuresloadend'` and
   * `'featuresloaderror'` events. `this` within the function is bound to the
   * {@link module:ol/source/Vector~VectorSource} it's called from.
   *
   * The function is responsible for loading the features and adding them to the
   * source.
   * @typedef {function(this:(import("./source/Vector").default|import("./VectorTile.js").default),
   *           import("./extent.js").Extent,
   *           number,
   *           import("./proj/Projection.js").default,
   *           function(Array<import("./Feature.js").default>): void=,
   *           function(): void=): void} FeatureLoader
   * @api
   */

  /**
   * {@link module:ol/source/Vector~VectorSource} sources use a function of this type to
   * get the url to load features from.
   *
   * This function takes an {@link module:ol/extent~Extent} representing the area
   * to be loaded, a `{number}` representing the resolution (map units per pixel)
   * and an {@link module:ol/proj/Projection~Projection} for the projection  as
   * arguments and returns a `{string}` representing the URL.
   * @typedef {function(import("./extent.js").Extent, number, import("./proj/Projection.js").default): string} FeatureUrlFunction
   * @api
   */

  /**
   * @param {string|FeatureUrlFunction} url Feature URL service.
   * @param {import("./format/Feature.js").default} format Feature format.
   * @param {import("./extent.js").Extent} extent Extent.
   * @param {number} resolution Resolution.
   * @param {import("./proj/Projection.js").default} projection Projection.
   * @param {function(Array<import("./Feature.js").default>, import("./proj/Projection.js").default): void} success Success
   *      Function called with the loaded features and optionally with the data projection.
   * @param {function(): void} failure Failure
   *      Function called when loading failed.
   */
  function loadFeaturesXhr(
    url,
    format,
    extent,
    resolution,
    projection,
    success,
    failure
  ) {
    const xhr = new XMLHttpRequest();
    xhr.open(
      'GET',
      typeof url === 'function' ? url(extent, resolution, projection) : url,
      true
    );
    if (format.getType() == 'arraybuffer') {
      xhr.responseType = 'arraybuffer';
    }
    xhr.withCredentials = withCredentials;
    /**
     * @param {Event} event Event.
     * @private
     */
    xhr.onload = function (event) {
      // status will be 0 for file:// urls
      if (!xhr.status || (xhr.status >= 200 && xhr.status < 300)) {
        const type = format.getType();
        /** @type {Document|Node|Object|string|undefined} */
        let source;
        if (type == 'json' || type == 'text') {
          source = xhr.responseText;
        } else if (type == 'xml') {
          source = xhr.responseXML;
          if (!source) {
            source = new DOMParser().parseFromString(
              xhr.responseText,
              'application/xml'
            );
          }
        } else if (type == 'arraybuffer') {
          source = /** @type {ArrayBuffer} */ (xhr.response);
        }
        if (source) {
          success(
            /** @type {Array<import("./Feature.js").default>} */
            (
              format.readFeatures(source, {
                extent: extent,
                featureProjection: projection,
              })
            ),
            format.readProjection(source)
          );
        } else {
          failure();
        }
      } else {
        failure();
      }
    };
    /**
     * @private
     */
    xhr.onerror = failure;
    xhr.send();
  }

  /**
   * Create an XHR feature loader for a `url` and `format`. The feature loader
   * loads features (with XHR), parses the features, and adds them to the
   * vector source.
   * @param {string|FeatureUrlFunction} url Feature URL service.
   * @param {import("./format/Feature.js").default} format Feature format.
   * @return {FeatureLoader} The feature loader.
   * @api
   */
  function xhr(url, format) {
    /**
     * @param {import("./extent.js").Extent} extent Extent.
     * @param {number} resolution Resolution.
     * @param {import("./proj/Projection.js").default} projection Projection.
     * @param {function(Array<import("./Feature.js").default>): void} [success] Success
     *      Function called when loading succeeded.
     * @param {function(): void} [failure] Failure
     *      Function called when loading failed.
     * @this {import("./source/Vector").default}
     */
    return function (extent, resolution, projection, success, failure) {
      const source = /** @type {import("./source/Vector").default} */ (this);
      loadFeaturesXhr(
        url,
        format,
        extent,
        resolution,
        projection,
        /**
         * @param {Array<import("./Feature.js").default>} features The loaded features.
         * @param {import("./proj/Projection.js").default} dataProjection Data
         * projection.
         */
        function (features, dataProjection) {
          source.addFeatures(features);
          if (success !== undefined) {
            success(features);
          }
        },
        /* FIXME handle error */ failure ? failure : VOID
      );
    };
  }

  /**
   * Setter for the withCredentials configuration for the XHR.
   *
   * @param {boolean} xhrWithCredentials The value of withCredentials to set.
   * Compare https://developer.mozilla.org/en-US/docs/Web/API/XMLHttpRequest/
   * @api
   */
  function setWithCredentials(xhrWithCredentials) {
    withCredentials = xhrWithCredentials;
  }

  /**
   * @module ol/loadingstrategy
   */

  /**
   * Strategy function for loading all features with a single request.
   * @param {import("./extent.js").Extent} extent Extent.
   * @param {number} resolution Resolution.
   * @return {Array<import("./extent.js").Extent>} Extents.
   * @api
   */
  function all(extent, resolution) {
    return [[-Infinity, -Infinity, Infinity, Infinity]];
  }

  /**
   * Strategy function for loading features based on the view's extent and
   * resolution.
   * @param {import("./extent.js").Extent} extent Extent.
   * @param {number} resolution Resolution.
   * @return {Array<import("./extent.js").Extent>} Extents.
   * @api
   */
  function bbox$1(extent, resolution) {
    return [extent];
  }

  /**
   * Creates a strategy function for loading features based on a tile grid.
   * @param {import("./tilegrid/TileGrid.js").default} tileGrid Tile grid.
   * @return {function(import("./extent.js").Extent, number, import("./proj.js").Projection): Array<import("./extent.js").Extent>} Loading strategy.
   * @api
   */
  function tile(tileGrid) {
    return (
      /**
       * @param {import("./extent.js").Extent} extent Extent.
       * @param {number} resolution Resolution.
       * @param {import("./proj.js").Projection} projection Projection.
       * @return {Array<import("./extent.js").Extent>} Extents.
       */
      function (extent, resolution, projection) {
        const z = tileGrid.getZForResolution(
          fromUserResolution(resolution, projection)
        );
        const tileRange = tileGrid.getTileRangeForExtentAndZ(
          fromUserExtent(extent, projection),
          z
        );
        /** @type {Array<import("./extent.js").Extent>} */
        const extents = [];
        /** @type {import("./tilecoord.js").TileCoord} */
        const tileCoord = [z, 0, 0];
        for (
          tileCoord[1] = tileRange.minX;
          tileCoord[1] <= tileRange.maxX;
          ++tileCoord[1]
        ) {
          for (
            tileCoord[2] = tileRange.minY;
            tileCoord[2] <= tileRange.maxY;
            ++tileCoord[2]
          ) {
            extents.push(
              toUserExtent(tileGrid.getTileCoordExtent(tileCoord), projection)
            );
          }
        }
        return extents;
      }
    );
  }

  /**
   * @module ol/net
   */

  /**
   * Simple JSONP helper. Supports error callbacks and a custom callback param.
   * The error callback will be called when no JSONP is executed after 10 seconds.
   *
   * @param {string} url Request url. A 'callback' query parameter will be
   *     appended.
   * @param {Function} callback Callback on success.
   * @param {Function} [errback] Callback on error.
   * @param {string} [callbackParam] Custom query parameter for the JSONP
   *     callback. Default is 'callback'.
   */
  function jsonp(url, callback, errback, callbackParam) {
    const script = document.createElement('script');
    const key = 'olc_' + getUid(callback);
    function cleanup() {
      delete window[key];
      script.parentNode.removeChild(script);
    }
    script.async = true;
    script.src =
      url +
      (url.includes('?') ? '&' : '?') +
      (callbackParam || 'callback') +
      '=' +
      key;
    const timer = setTimeout(function () {
      cleanup();
      if (errback) {
        errback();
      }
    }, 10000);
    window[key] = function (data) {
      clearTimeout(timer);
      cleanup();
      callback(data);
    };
    document.head.appendChild(script);
  }

  class ResponseError extends Error {
    /**
     * @param {XMLHttpRequest} response The XHR object.
     */
    constructor(response) {
      const message = 'Unexpected response status: ' + response.status;
      super(message);

      /**
       * @type {string}
       */
      this.name = 'ResponseError';

      /**
       * @type {XMLHttpRequest}
       */
      this.response = response;
    }
  }

  class ClientError extends Error {
    /**
     * @param {XMLHttpRequest} client The XHR object.
     */
    constructor(client) {
      super('Failed to issue request');

      /**
       * @type {string}
       */
      this.name = 'ClientError';

      /**
       * @type {XMLHttpRequest}
       */
      this.client = client;
    }
  }

  /**
   * @param {string} url The URL.
   * @return {Promise<Object>} A promise that resolves to the JSON response.
   */
  function getJSON(url) {
    return new Promise(function (resolve, reject) {
      /**
       * @param {ProgressEvent<XMLHttpRequest>} event The load event.
       */
      function onLoad(event) {
        const client = event.target;
        // status will be 0 for file:// urls
        if (!client.status || (client.status >= 200 && client.status < 300)) {
          let data;
          try {
            data = JSON.parse(client.responseText);
          } catch (err) {
            const message = 'Error parsing response text as JSON: ' + err.message;
            reject(new Error(message));
            return;
          }
          resolve(data);
          return;
        }

        reject(new ResponseError(client));
      }

      /**
       * @param {ProgressEvent<XMLHttpRequest>} event The error event.
       */
      function onError(event) {
        reject(new ClientError(event.target));
      }

      const client = new XMLHttpRequest();
      client.addEventListener('load', onLoad);
      client.addEventListener('error', onError);
      client.open('GET', url);
      client.setRequestHeader('Accept', 'application/json');
      client.send();
    });
  }

  /**
   * @param {string} base The base URL.
   * @param {string} url The potentially relative URL.
   * @return {string} The full URL.
   */
  function resolveUrl(base, url) {
    if (url.includes('://')) {
      return url;
    }
    return new URL(url, base).href;
  }

  let originalXHR;
  function overrideXHR(xhr) {
    if (typeof XMLHttpRequest !== 'undefined') {
      originalXHR = XMLHttpRequest;
    }
    global.XMLHttpRequest = xhr;
  }

  function restoreXHR() {
    global.XMLHttpRequest = originalXHR;
  }

  /**
   * @module ol/render/VectorContext
   */

  /**
   * @classdesc
   * Context for drawing geometries.  A vector context is available on render
   * events and does not need to be constructed directly.
   * @api
   */
  class VectorContext {
    /**
     * Render a geometry with a custom renderer.
     *
     * @param {import("../geom/SimpleGeometry.js").default} geometry Geometry.
     * @param {import("../Feature.js").FeatureLike} feature Feature.
     * @param {Function} renderer Renderer.
     * @param {Function} hitDetectionRenderer Renderer.
     */
    drawCustom(geometry, feature, renderer, hitDetectionRenderer) {}

    /**
     * Render a geometry.
     *
     * @param {import("../geom/Geometry.js").default} geometry The geometry to render.
     */
    drawGeometry(geometry) {}

    /**
     * Set the rendering style.
     *
     * @param {import("../style/Style.js").default} style The rendering style.
     */
    setStyle(style) {}

    /**
     * @param {import("../geom/Circle.js").default} circleGeometry Circle geometry.
     * @param {import("../Feature.js").default} feature Feature.
     */
    drawCircle(circleGeometry, feature) {}

    /**
     * @param {import("../Feature.js").default} feature Feature.
     * @param {import("../style/Style.js").default} style Style.
     */
    drawFeature(feature, style) {}

    /**
     * @param {import("../geom/GeometryCollection.js").default} geometryCollectionGeometry Geometry collection.
     * @param {import("../Feature.js").default} feature Feature.
     */
    drawGeometryCollection(geometryCollectionGeometry, feature) {}

    /**
     * @param {import("../geom/LineString.js").default|import("./Feature.js").default} lineStringGeometry Line string geometry.
     * @param {import("../Feature.js").FeatureLike} feature Feature.
     */
    drawLineString(lineStringGeometry, feature) {}

    /**
     * @param {import("../geom/MultiLineString.js").default|import("./Feature.js").default} multiLineStringGeometry MultiLineString geometry.
     * @param {import("../Feature.js").FeatureLike} feature Feature.
     */
    drawMultiLineString(multiLineStringGeometry, feature) {}

    /**
     * @param {import("../geom/MultiPoint.js").default|import("./Feature.js").default} multiPointGeometry MultiPoint geometry.
     * @param {import("../Feature.js").FeatureLike} feature Feature.
     */
    drawMultiPoint(multiPointGeometry, feature) {}

    /**
     * @param {import("../geom/MultiPolygon.js").default} multiPolygonGeometry MultiPolygon geometry.
     * @param {import("../Feature.js").FeatureLike} feature Feature.
     */
    drawMultiPolygon(multiPolygonGeometry, feature) {}

    /**
     * @param {import("../geom/Point.js").default|import("./Feature.js").default} pointGeometry Point geometry.
     * @param {import("../Feature.js").FeatureLike} feature Feature.
     */
    drawPoint(pointGeometry, feature) {}

    /**
     * @param {import("../geom/Polygon.js").default|import("./Feature.js").default} polygonGeometry Polygon geometry.
     * @param {import("../Feature.js").FeatureLike} feature Feature.
     */
    drawPolygon(polygonGeometry, feature) {}

    /**
     * @param {import("../geom/SimpleGeometry.js").default|import("./Feature.js").default} geometry Geometry.
     * @param {import("../Feature.js").FeatureLike} feature Feature.
     */
    drawText(geometry, feature) {}

    /**
     * @param {import("../style/Fill.js").default} fillStyle Fill style.
     * @param {import("../style/Stroke.js").default} strokeStyle Stroke style.
     */
    setFillStrokeStyle(fillStyle, strokeStyle) {}

    /**
     * @param {import("../style/Image.js").default} imageStyle Image style.
     * @param {import("../render/canvas.js").DeclutterImageWithText} [declutterImageWithText] Shared data for combined decluttering with a text style.
     */
    setImageStyle(imageStyle, declutterImageWithText) {}

    /**
     * @param {import("../style/Text.js").default} textStyle Text style.
     * @param {import("../render/canvas.js").DeclutterImageWithText} [declutterImageWithText] Shared data for combined decluttering with an image style.
     */
    setTextStyle(textStyle, declutterImageWithText) {}
  }

  /**
   * @module ol/render/canvas/Immediate
   */

  /**
   * @classdesc
   * A concrete subclass of {@link module:ol/render/VectorContext~VectorContext} that implements
   * direct rendering of features and geometries to an HTML5 Canvas context.
   * Instances of this class are created internally by the library and
   * provided to application code as vectorContext member of the
   * {@link module:ol/render/Event~RenderEvent} object associated with postcompose, precompose and
   * render events emitted by layers and maps.
   */
  class CanvasImmediateRenderer extends VectorContext {
    /**
     * @param {CanvasRenderingContext2D} context Context.
     * @param {number} pixelRatio Pixel ratio.
     * @param {import("../../extent.js").Extent} extent Extent.
     * @param {import("../../transform.js").Transform} transform Transform.
     * @param {number} viewRotation View rotation.
     * @param {number} [squaredTolerance] Optional squared tolerance for simplification.
     * @param {import("../../proj.js").TransformFunction} [userTransform] Transform from user to view projection.
     */
    constructor(
      context,
      pixelRatio,
      extent,
      transform,
      viewRotation,
      squaredTolerance,
      userTransform
    ) {
      super();

      /**
       * @private
       * @type {CanvasRenderingContext2D}
       */
      this.context_ = context;

      /**
       * @private
       * @type {number}
       */
      this.pixelRatio_ = pixelRatio;

      /**
       * @private
       * @type {import("../../extent.js").Extent}
       */
      this.extent_ = extent;

      /**
       * @private
       * @type {import("../../transform.js").Transform}
       */
      this.transform_ = transform;

      /**
       * @private
       * @type {number}
       */
      this.viewRotation_ = viewRotation;

      /**
       * @private
       * @type {number}
       */
      this.squaredTolerance_ = squaredTolerance;

      /**
       * @private
       * @type {import("../../proj.js").TransformFunction}
       */
      this.userTransform_ = userTransform;

      /**
       * @private
       * @type {?import("../canvas.js").FillState}
       */
      this.contextFillState_ = null;

      /**
       * @private
       * @type {?import("../canvas.js").StrokeState}
       */
      this.contextStrokeState_ = null;

      /**
       * @private
       * @type {?import("../canvas.js").TextState}
       */
      this.contextTextState_ = null;

      /**
       * @private
       * @type {?import("../canvas.js").FillState}
       */
      this.fillState_ = null;

      /**
       * @private
       * @type {?import("../canvas.js").StrokeState}
       */
      this.strokeState_ = null;

      /**
       * @private
       * @type {HTMLCanvasElement|HTMLVideoElement|HTMLImageElement}
       */
      this.image_ = null;

      /**
       * @private
       * @type {number}
       */
      this.imageAnchorX_ = 0;

      /**
       * @private
       * @type {number}
       */
      this.imageAnchorY_ = 0;

      /**
       * @private
       * @type {number}
       */
      this.imageHeight_ = 0;

      /**
       * @private
       * @type {number}
       */
      this.imageOpacity_ = 0;

      /**
       * @private
       * @type {number}
       */
      this.imageOriginX_ = 0;

      /**
       * @private
       * @type {number}
       */
      this.imageOriginY_ = 0;

      /**
       * @private
       * @type {boolean}
       */
      this.imageRotateWithView_ = false;

      /**
       * @private
       * @type {number}
       */
      this.imageRotation_ = 0;

      /**
       * @private
       * @type {import("../../size.js").Size}
       */
      this.imageScale_ = [0, 0];

      /**
       * @private
       * @type {number}
       */
      this.imageWidth_ = 0;

      /**
       * @private
       * @type {string}
       */
      this.text_ = '';

      /**
       * @private
       * @type {number}
       */
      this.textOffsetX_ = 0;

      /**
       * @private
       * @type {number}
       */
      this.textOffsetY_ = 0;

      /**
       * @private
       * @type {boolean}
       */
      this.textRotateWithView_ = false;

      /**
       * @private
       * @type {number}
       */
      this.textRotation_ = 0;

      /**
       * @private
       * @type {import("../../size.js").Size}
       */
      this.textScale_ = [0, 0];

      /**
       * @private
       * @type {?import("../canvas.js").FillState}
       */
      this.textFillState_ = null;

      /**
       * @private
       * @type {?import("../canvas.js").StrokeState}
       */
      this.textStrokeState_ = null;

      /**
       * @private
       * @type {?import("../canvas.js").TextState}
       */
      this.textState_ = null;

      /**
       * @private
       * @type {Array<number>}
       */
      this.pixelCoordinates_ = [];

      /**
       * @private
       * @type {import("../../transform.js").Transform}
       */
      this.tmpLocalTransform_ = create$2();
    }

    /**
     * @param {Array<number>} flatCoordinates Flat coordinates.
     * @param {number} offset Offset.
     * @param {number} end End.
     * @param {number} stride Stride.
     * @private
     */
    drawImages_(flatCoordinates, offset, end, stride) {
      if (!this.image_) {
        return;
      }
      const pixelCoordinates = transform2D(
        flatCoordinates,
        offset,
        end,
        stride,
        this.transform_,
        this.pixelCoordinates_
      );
      const context = this.context_;
      const localTransform = this.tmpLocalTransform_;
      const alpha = context.globalAlpha;
      if (this.imageOpacity_ != 1) {
        context.globalAlpha = alpha * this.imageOpacity_;
      }
      let rotation = this.imageRotation_;
      if (this.imageRotateWithView_) {
        rotation += this.viewRotation_;
      }
      for (let i = 0, ii = pixelCoordinates.length; i < ii; i += 2) {
        const x = pixelCoordinates[i] - this.imageAnchorX_;
        const y = pixelCoordinates[i + 1] - this.imageAnchorY_;
        if (
          rotation !== 0 ||
          this.imageScale_[0] != 1 ||
          this.imageScale_[1] != 1
        ) {
          const centerX = x + this.imageAnchorX_;
          const centerY = y + this.imageAnchorY_;
          compose(
            localTransform,
            centerX,
            centerY,
            1,
            1,
            rotation,
            -centerX,
            -centerY
          );
          context.setTransform.apply(context, localTransform);
          context.translate(centerX, centerY);
          context.scale(this.imageScale_[0], this.imageScale_[1]);
          context.drawImage(
            this.image_,
            this.imageOriginX_,
            this.imageOriginY_,
            this.imageWidth_,
            this.imageHeight_,
            -this.imageAnchorX_,
            -this.imageAnchorY_,
            this.imageWidth_,
            this.imageHeight_
          );
          context.setTransform(1, 0, 0, 1, 0, 0);
        } else {
          context.drawImage(
            this.image_,
            this.imageOriginX_,
            this.imageOriginY_,
            this.imageWidth_,
            this.imageHeight_,
            x,
            y,
            this.imageWidth_,
            this.imageHeight_
          );
        }
      }
      if (this.imageOpacity_ != 1) {
        context.globalAlpha = alpha;
      }
    }

    /**
     * @param {Array<number>} flatCoordinates Flat coordinates.
     * @param {number} offset Offset.
     * @param {number} end End.
     * @param {number} stride Stride.
     * @private
     */
    drawText_(flatCoordinates, offset, end, stride) {
      if (!this.textState_ || this.text_ === '') {
        return;
      }
      if (this.textFillState_) {
        this.setContextFillState_(this.textFillState_);
      }
      if (this.textStrokeState_) {
        this.setContextStrokeState_(this.textStrokeState_);
      }
      this.setContextTextState_(this.textState_);
      const pixelCoordinates = transform2D(
        flatCoordinates,
        offset,
        end,
        stride,
        this.transform_,
        this.pixelCoordinates_
      );
      const context = this.context_;
      let rotation = this.textRotation_;
      if (this.textRotateWithView_) {
        rotation += this.viewRotation_;
      }
      for (; offset < end; offset += stride) {
        const x = pixelCoordinates[offset] + this.textOffsetX_;
        const y = pixelCoordinates[offset + 1] + this.textOffsetY_;
        if (
          rotation !== 0 ||
          this.textScale_[0] != 1 ||
          this.textScale_[1] != 1
        ) {
          context.translate(x - this.textOffsetX_, y - this.textOffsetY_);
          context.rotate(rotation);
          context.translate(this.textOffsetX_, this.textOffsetY_);
          context.scale(this.textScale_[0], this.textScale_[1]);
          if (this.textStrokeState_) {
            context.strokeText(this.text_, 0, 0);
          }
          if (this.textFillState_) {
            context.fillText(this.text_, 0, 0);
          }
          context.setTransform(1, 0, 0, 1, 0, 0);
        } else {
          if (this.textStrokeState_) {
            context.strokeText(this.text_, x, y);
          }
          if (this.textFillState_) {
            context.fillText(this.text_, x, y);
          }
        }
      }
    }

    /**
     * @param {Array<number>} flatCoordinates Flat coordinates.
     * @param {number} offset Offset.
     * @param {number} end End.
     * @param {number} stride Stride.
     * @param {boolean} close Close.
     * @private
     * @return {number} end End.
     */
    moveToLineTo_(flatCoordinates, offset, end, stride, close) {
      const context = this.context_;
      const pixelCoordinates = transform2D(
        flatCoordinates,
        offset,
        end,
        stride,
        this.transform_,
        this.pixelCoordinates_
      );
      context.moveTo(pixelCoordinates[0], pixelCoordinates[1]);
      let length = pixelCoordinates.length;
      if (close) {
        length -= 2;
      }
      for (let i = 2; i < length; i += 2) {
        context.lineTo(pixelCoordinates[i], pixelCoordinates[i + 1]);
      }
      if (close) {
        context.closePath();
      }
      return end;
    }

    /**
     * @param {Array<number>} flatCoordinates Flat coordinates.
     * @param {number} offset Offset.
     * @param {Array<number>} ends Ends.
     * @param {number} stride Stride.
     * @private
     * @return {number} End.
     */
    drawRings_(flatCoordinates, offset, ends, stride) {
      for (let i = 0, ii = ends.length; i < ii; ++i) {
        offset = this.moveToLineTo_(
          flatCoordinates,
          offset,
          ends[i],
          stride,
          true
        );
      }
      return offset;
    }

    /**
     * Render a circle geometry into the canvas.  Rendering is immediate and uses
     * the current fill and stroke styles.
     *
     * @param {import("../../geom/Circle.js").default} geometry Circle geometry.
     * @api
     */
    drawCircle(geometry) {
      if (!intersects$3(this.extent_, geometry.getExtent())) {
        return;
      }
      if (this.fillState_ || this.strokeState_) {
        if (this.fillState_) {
          this.setContextFillState_(this.fillState_);
        }
        if (this.strokeState_) {
          this.setContextStrokeState_(this.strokeState_);
        }
        const pixelCoordinates = transformGeom2D(
          geometry,
          this.transform_,
          this.pixelCoordinates_
        );
        const dx = pixelCoordinates[2] - pixelCoordinates[0];
        const dy = pixelCoordinates[3] - pixelCoordinates[1];
        const radius = Math.sqrt(dx * dx + dy * dy);
        const context = this.context_;
        context.beginPath();
        context.arc(
          pixelCoordinates[0],
          pixelCoordinates[1],
          radius,
          0,
          2 * Math.PI
        );
        if (this.fillState_) {
          context.fill();
        }
        if (this.strokeState_) {
          context.stroke();
        }
      }
      if (this.text_ !== '') {
        this.drawText_(geometry.getCenter(), 0, 2, 2);
      }
    }

    /**
     * Set the rendering style.  Note that since this is an immediate rendering API,
     * any `zIndex` on the provided style will be ignored.
     *
     * @param {import("../../style/Style.js").default} style The rendering style.
     * @api
     */
    setStyle(style) {
      this.setFillStrokeStyle(style.getFill(), style.getStroke());
      this.setImageStyle(style.getImage());
      this.setTextStyle(style.getText());
    }

    /**
     * @param {import("../../transform.js").Transform} transform Transform.
     */
    setTransform(transform) {
      this.transform_ = transform;
    }

    /**
     * Render a geometry into the canvas.  Call
     * {@link module:ol/render/canvas/Immediate~CanvasImmediateRenderer#setStyle renderer.setStyle()} first to set the rendering style.
     *
     * @param {import("../../geom/Geometry.js").default|import("../Feature.js").default} geometry The geometry to render.
     * @api
     */
    drawGeometry(geometry) {
      const type = geometry.getType();
      switch (type) {
        case 'Point':
          this.drawPoint(
            /** @type {import("../../geom/Point.js").default} */ (geometry)
          );
          break;
        case 'LineString':
          this.drawLineString(
            /** @type {import("../../geom/LineString.js").default} */ (geometry)
          );
          break;
        case 'Polygon':
          this.drawPolygon(
            /** @type {import("../../geom/Polygon.js").default} */ (geometry)
          );
          break;
        case 'MultiPoint':
          this.drawMultiPoint(
            /** @type {import("../../geom/MultiPoint.js").default} */ (geometry)
          );
          break;
        case 'MultiLineString':
          this.drawMultiLineString(
            /** @type {import("../../geom/MultiLineString.js").default} */ (
              geometry
            )
          );
          break;
        case 'MultiPolygon':
          this.drawMultiPolygon(
            /** @type {import("../../geom/MultiPolygon.js").default} */ (geometry)
          );
          break;
        case 'GeometryCollection':
          this.drawGeometryCollection(
            /** @type {import("../../geom/GeometryCollection.js").default} */ (
              geometry
            )
          );
          break;
        case 'Circle':
          this.drawCircle(
            /** @type {import("../../geom/Circle.js").default} */ (geometry)
          );
          break;
      }
    }

    /**
     * Render a feature into the canvas.  Note that any `zIndex` on the provided
     * style will be ignored - features are rendered immediately in the order that
     * this method is called.  If you need `zIndex` support, you should be using an
     * {@link module:ol/layer/Vector~VectorLayer} instead.
     *
     * @param {import("../../Feature.js").default} feature Feature.
     * @param {import("../../style/Style.js").default} style Style.
     * @api
     */
    drawFeature(feature, style) {
      const geometry = style.getGeometryFunction()(feature);
      if (!geometry || !intersects$3(this.extent_, geometry.getExtent())) {
        return;
      }
      this.setStyle(style);
      this.drawGeometry(geometry);
    }

    /**
     * Render a GeometryCollection to the canvas.  Rendering is immediate and
     * uses the current styles appropriate for each geometry in the collection.
     *
     * @param {import("../../geom/GeometryCollection.js").default} geometry Geometry collection.
     */
    drawGeometryCollection(geometry) {
      const geometries = geometry.getGeometriesArray();
      for (let i = 0, ii = geometries.length; i < ii; ++i) {
        this.drawGeometry(geometries[i]);
      }
    }

    /**
     * Render a Point geometry into the canvas.  Rendering is immediate and uses
     * the current style.
     *
     * @param {import("../../geom/Point.js").default|import("../Feature.js").default} geometry Point geometry.
     */
    drawPoint(geometry) {
      if (this.squaredTolerance_) {
        geometry = /** @type {import("../../geom/Point.js").default} */ (
          geometry.simplifyTransformed(
            this.squaredTolerance_,
            this.userTransform_
          )
        );
      }
      const flatCoordinates = geometry.getFlatCoordinates();
      const stride = geometry.getStride();
      if (this.image_) {
        this.drawImages_(flatCoordinates, 0, flatCoordinates.length, stride);
      }
      if (this.text_ !== '') {
        this.drawText_(flatCoordinates, 0, flatCoordinates.length, stride);
      }
    }

    /**
     * Render a MultiPoint geometry  into the canvas.  Rendering is immediate and
     * uses the current style.
     *
     * @param {import("../../geom/MultiPoint.js").default|import("../Feature.js").default} geometry MultiPoint geometry.
     */
    drawMultiPoint(geometry) {
      if (this.squaredTolerance_) {
        geometry = /** @type {import("../../geom/MultiPoint.js").default} */ (
          geometry.simplifyTransformed(
            this.squaredTolerance_,
            this.userTransform_
          )
        );
      }
      const flatCoordinates = geometry.getFlatCoordinates();
      const stride = geometry.getStride();
      if (this.image_) {
        this.drawImages_(flatCoordinates, 0, flatCoordinates.length, stride);
      }
      if (this.text_ !== '') {
        this.drawText_(flatCoordinates, 0, flatCoordinates.length, stride);
      }
    }

    /**
     * Render a LineString into the canvas.  Rendering is immediate and uses
     * the current style.
     *
     * @param {import("../../geom/LineString.js").default|import("../Feature.js").default} geometry LineString geometry.
     */
    drawLineString(geometry) {
      if (this.squaredTolerance_) {
        geometry = /** @type {import("../../geom/LineString.js").default} */ (
          geometry.simplifyTransformed(
            this.squaredTolerance_,
            this.userTransform_
          )
        );
      }
      if (!intersects$3(this.extent_, geometry.getExtent())) {
        return;
      }
      if (this.strokeState_) {
        this.setContextStrokeState_(this.strokeState_);
        const context = this.context_;
        const flatCoordinates = geometry.getFlatCoordinates();
        context.beginPath();
        this.moveToLineTo_(
          flatCoordinates,
          0,
          flatCoordinates.length,
          geometry.getStride(),
          false
        );
        context.stroke();
      }
      if (this.text_ !== '') {
        const flatMidpoint = geometry.getFlatMidpoint();
        this.drawText_(flatMidpoint, 0, 2, 2);
      }
    }

    /**
     * Render a MultiLineString geometry into the canvas.  Rendering is immediate
     * and uses the current style.
     *
     * @param {import("../../geom/MultiLineString.js").default|import("../Feature.js").default} geometry MultiLineString geometry.
     */
    drawMultiLineString(geometry) {
      if (this.squaredTolerance_) {
        geometry =
          /** @type {import("../../geom/MultiLineString.js").default} */ (
            geometry.simplifyTransformed(
              this.squaredTolerance_,
              this.userTransform_
            )
          );
      }
      const geometryExtent = geometry.getExtent();
      if (!intersects$3(this.extent_, geometryExtent)) {
        return;
      }
      if (this.strokeState_) {
        this.setContextStrokeState_(this.strokeState_);
        const context = this.context_;
        const flatCoordinates = geometry.getFlatCoordinates();
        let offset = 0;
        const ends = /** @type {Array<number>} */ (geometry.getEnds());
        const stride = geometry.getStride();
        context.beginPath();
        for (let i = 0, ii = ends.length; i < ii; ++i) {
          offset = this.moveToLineTo_(
            flatCoordinates,
            offset,
            ends[i],
            stride,
            false
          );
        }
        context.stroke();
      }
      if (this.text_ !== '') {
        const flatMidpoints = geometry.getFlatMidpoints();
        this.drawText_(flatMidpoints, 0, flatMidpoints.length, 2);
      }
    }

    /**
     * Render a Polygon geometry into the canvas.  Rendering is immediate and uses
     * the current style.
     *
     * @param {import("../../geom/Polygon.js").default|import("../Feature.js").default} geometry Polygon geometry.
     */
    drawPolygon(geometry) {
      if (this.squaredTolerance_) {
        geometry = /** @type {import("../../geom/Polygon.js").default} */ (
          geometry.simplifyTransformed(
            this.squaredTolerance_,
            this.userTransform_
          )
        );
      }
      if (!intersects$3(this.extent_, geometry.getExtent())) {
        return;
      }
      if (this.strokeState_ || this.fillState_) {
        if (this.fillState_) {
          this.setContextFillState_(this.fillState_);
        }
        if (this.strokeState_) {
          this.setContextStrokeState_(this.strokeState_);
        }
        const context = this.context_;
        context.beginPath();
        this.drawRings_(
          geometry.getOrientedFlatCoordinates(),
          0,
          /** @type {Array<number>} */ (geometry.getEnds()),
          geometry.getStride()
        );
        if (this.fillState_) {
          context.fill();
        }
        if (this.strokeState_) {
          context.stroke();
        }
      }
      if (this.text_ !== '') {
        const flatInteriorPoint = geometry.getFlatInteriorPoint();
        this.drawText_(flatInteriorPoint, 0, 2, 2);
      }
    }

    /**
     * Render MultiPolygon geometry into the canvas.  Rendering is immediate and
     * uses the current style.
     * @param {import("../../geom/MultiPolygon.js").default} geometry MultiPolygon geometry.
     */
    drawMultiPolygon(geometry) {
      if (this.squaredTolerance_) {
        geometry = /** @type {import("../../geom/MultiPolygon.js").default} */ (
          geometry.simplifyTransformed(
            this.squaredTolerance_,
            this.userTransform_
          )
        );
      }
      if (!intersects$3(this.extent_, geometry.getExtent())) {
        return;
      }
      if (this.strokeState_ || this.fillState_) {
        if (this.fillState_) {
          this.setContextFillState_(this.fillState_);
        }
        if (this.strokeState_) {
          this.setContextStrokeState_(this.strokeState_);
        }
        const context = this.context_;
        const flatCoordinates = geometry.getOrientedFlatCoordinates();
        let offset = 0;
        const endss = geometry.getEndss();
        const stride = geometry.getStride();
        context.beginPath();
        for (let i = 0, ii = endss.length; i < ii; ++i) {
          const ends = endss[i];
          offset = this.drawRings_(flatCoordinates, offset, ends, stride);
        }
        if (this.fillState_) {
          context.fill();
        }
        if (this.strokeState_) {
          context.stroke();
        }
      }
      if (this.text_ !== '') {
        const flatInteriorPoints = geometry.getFlatInteriorPoints();
        this.drawText_(flatInteriorPoints, 0, flatInteriorPoints.length, 2);
      }
    }

    /**
     * @param {import("../canvas.js").FillState} fillState Fill state.
     * @private
     */
    setContextFillState_(fillState) {
      const context = this.context_;
      const contextFillState = this.contextFillState_;
      if (!contextFillState) {
        context.fillStyle = fillState.fillStyle;
        this.contextFillState_ = {
          fillStyle: fillState.fillStyle,
        };
      } else {
        if (contextFillState.fillStyle != fillState.fillStyle) {
          contextFillState.fillStyle = fillState.fillStyle;
          context.fillStyle = fillState.fillStyle;
        }
      }
    }

    /**
     * @param {import("../canvas.js").StrokeState} strokeState Stroke state.
     * @private
     */
    setContextStrokeState_(strokeState) {
      const context = this.context_;
      const contextStrokeState = this.contextStrokeState_;
      if (!contextStrokeState) {
        context.lineCap = strokeState.lineCap;
        context.setLineDash(strokeState.lineDash);
        context.lineDashOffset = strokeState.lineDashOffset;
        context.lineJoin = strokeState.lineJoin;
        context.lineWidth = strokeState.lineWidth;
        context.miterLimit = strokeState.miterLimit;
        context.strokeStyle = strokeState.strokeStyle;
        this.contextStrokeState_ = {
          lineCap: strokeState.lineCap,
          lineDash: strokeState.lineDash,
          lineDashOffset: strokeState.lineDashOffset,
          lineJoin: strokeState.lineJoin,
          lineWidth: strokeState.lineWidth,
          miterLimit: strokeState.miterLimit,
          strokeStyle: strokeState.strokeStyle,
        };
      } else {
        if (contextStrokeState.lineCap != strokeState.lineCap) {
          contextStrokeState.lineCap = strokeState.lineCap;
          context.lineCap = strokeState.lineCap;
        }
        if (!equals$3(contextStrokeState.lineDash, strokeState.lineDash)) {
          context.setLineDash(
            (contextStrokeState.lineDash = strokeState.lineDash)
          );
        }
        if (contextStrokeState.lineDashOffset != strokeState.lineDashOffset) {
          contextStrokeState.lineDashOffset = strokeState.lineDashOffset;
          context.lineDashOffset = strokeState.lineDashOffset;
        }
        if (contextStrokeState.lineJoin != strokeState.lineJoin) {
          contextStrokeState.lineJoin = strokeState.lineJoin;
          context.lineJoin = strokeState.lineJoin;
        }
        if (contextStrokeState.lineWidth != strokeState.lineWidth) {
          contextStrokeState.lineWidth = strokeState.lineWidth;
          context.lineWidth = strokeState.lineWidth;
        }
        if (contextStrokeState.miterLimit != strokeState.miterLimit) {
          contextStrokeState.miterLimit = strokeState.miterLimit;
          context.miterLimit = strokeState.miterLimit;
        }
        if (contextStrokeState.strokeStyle != strokeState.strokeStyle) {
          contextStrokeState.strokeStyle = strokeState.strokeStyle;
          context.strokeStyle = strokeState.strokeStyle;
        }
      }
    }

    /**
     * @param {import("../canvas.js").TextState} textState Text state.
     * @private
     */
    setContextTextState_(textState) {
      const context = this.context_;
      const contextTextState = this.contextTextState_;
      const textAlign = textState.textAlign
        ? textState.textAlign
        : defaultTextAlign;
      if (!contextTextState) {
        context.font = textState.font;
        context.textAlign = textAlign;
        context.textBaseline = textState.textBaseline;
        this.contextTextState_ = {
          font: textState.font,
          textAlign: textAlign,
          textBaseline: textState.textBaseline,
        };
      } else {
        if (contextTextState.font != textState.font) {
          contextTextState.font = textState.font;
          context.font = textState.font;
        }
        if (contextTextState.textAlign != textAlign) {
          contextTextState.textAlign = textAlign;
          context.textAlign = textAlign;
        }
        if (contextTextState.textBaseline != textState.textBaseline) {
          contextTextState.textBaseline = textState.textBaseline;
          context.textBaseline = textState.textBaseline;
        }
      }
    }

    /**
     * Set the fill and stroke style for subsequent draw operations.  To clear
     * either fill or stroke styles, pass null for the appropriate parameter.
     *
     * @param {import("../../style/Fill.js").default} fillStyle Fill style.
     * @param {import("../../style/Stroke.js").default} strokeStyle Stroke style.
     */
    setFillStrokeStyle(fillStyle, strokeStyle) {
      if (!fillStyle) {
        this.fillState_ = null;
      } else {
        const fillStyleColor = fillStyle.getColor();
        this.fillState_ = {
          fillStyle: asColorLike(
            fillStyleColor ? fillStyleColor : defaultFillStyle
          ),
        };
      }
      if (!strokeStyle) {
        this.strokeState_ = null;
      } else {
        const strokeStyleColor = strokeStyle.getColor();
        const strokeStyleLineCap = strokeStyle.getLineCap();
        const strokeStyleLineDash = strokeStyle.getLineDash();
        const strokeStyleLineDashOffset = strokeStyle.getLineDashOffset();
        const strokeStyleLineJoin = strokeStyle.getLineJoin();
        const strokeStyleWidth = strokeStyle.getWidth();
        const strokeStyleMiterLimit = strokeStyle.getMiterLimit();
        const lineDash = strokeStyleLineDash
          ? strokeStyleLineDash
          : defaultLineDash;
        this.strokeState_ = {
          lineCap:
            strokeStyleLineCap !== undefined
              ? strokeStyleLineCap
              : defaultLineCap,
          lineDash:
            this.pixelRatio_ === 1
              ? lineDash
              : lineDash.map((n) => n * this.pixelRatio_),
          lineDashOffset:
            (strokeStyleLineDashOffset
              ? strokeStyleLineDashOffset
              : defaultLineDashOffset) * this.pixelRatio_,
          lineJoin:
            strokeStyleLineJoin !== undefined
              ? strokeStyleLineJoin
              : defaultLineJoin,
          lineWidth:
            (strokeStyleWidth !== undefined
              ? strokeStyleWidth
              : defaultLineWidth) * this.pixelRatio_,
          miterLimit:
            strokeStyleMiterLimit !== undefined
              ? strokeStyleMiterLimit
              : defaultMiterLimit,
          strokeStyle: asColorLike(
            strokeStyleColor ? strokeStyleColor : defaultStrokeStyle
          ),
        };
      }
    }

    /**
     * Set the image style for subsequent draw operations.  Pass null to remove
     * the image style.
     *
     * @param {import("../../style/Image.js").default} imageStyle Image style.
     */
    setImageStyle(imageStyle) {
      let imageSize;
      if (!imageStyle || !(imageSize = imageStyle.getSize())) {
        this.image_ = null;
        return;
      }
      const imagePixelRatio = imageStyle.getPixelRatio(this.pixelRatio_);
      const imageAnchor = imageStyle.getAnchor();
      const imageOrigin = imageStyle.getOrigin();
      this.image_ = imageStyle.getImage(this.pixelRatio_);
      this.imageAnchorX_ = imageAnchor[0] * imagePixelRatio;
      this.imageAnchorY_ = imageAnchor[1] * imagePixelRatio;
      this.imageHeight_ = imageSize[1] * imagePixelRatio;
      this.imageOpacity_ = imageStyle.getOpacity();
      this.imageOriginX_ = imageOrigin[0];
      this.imageOriginY_ = imageOrigin[1];
      this.imageRotateWithView_ = imageStyle.getRotateWithView();
      this.imageRotation_ = imageStyle.getRotation();
      const imageScale = imageStyle.getScaleArray();
      this.imageScale_ = [
        (imageScale[0] * this.pixelRatio_) / imagePixelRatio,
        (imageScale[1] * this.pixelRatio_) / imagePixelRatio,
      ];
      this.imageWidth_ = imageSize[0] * imagePixelRatio;
    }

    /**
     * Set the text style for subsequent draw operations.  Pass null to
     * remove the text style.
     *
     * @param {import("../../style/Text.js").default} textStyle Text style.
     */
    setTextStyle(textStyle) {
      if (!textStyle) {
        this.text_ = '';
      } else {
        const textFillStyle = textStyle.getFill();
        if (!textFillStyle) {
          this.textFillState_ = null;
        } else {
          const textFillStyleColor = textFillStyle.getColor();
          this.textFillState_ = {
            fillStyle: asColorLike(
              textFillStyleColor ? textFillStyleColor : defaultFillStyle
            ),
          };
        }
        const textStrokeStyle = textStyle.getStroke();
        if (!textStrokeStyle) {
          this.textStrokeState_ = null;
        } else {
          const textStrokeStyleColor = textStrokeStyle.getColor();
          const textStrokeStyleLineCap = textStrokeStyle.getLineCap();
          const textStrokeStyleLineDash = textStrokeStyle.getLineDash();
          const textStrokeStyleLineDashOffset =
            textStrokeStyle.getLineDashOffset();
          const textStrokeStyleLineJoin = textStrokeStyle.getLineJoin();
          const textStrokeStyleWidth = textStrokeStyle.getWidth();
          const textStrokeStyleMiterLimit = textStrokeStyle.getMiterLimit();
          this.textStrokeState_ = {
            lineCap:
              textStrokeStyleLineCap !== undefined
                ? textStrokeStyleLineCap
                : defaultLineCap,
            lineDash: textStrokeStyleLineDash
              ? textStrokeStyleLineDash
              : defaultLineDash,
            lineDashOffset: textStrokeStyleLineDashOffset
              ? textStrokeStyleLineDashOffset
              : defaultLineDashOffset,
            lineJoin:
              textStrokeStyleLineJoin !== undefined
                ? textStrokeStyleLineJoin
                : defaultLineJoin,
            lineWidth:
              textStrokeStyleWidth !== undefined
                ? textStrokeStyleWidth
                : defaultLineWidth,
            miterLimit:
              textStrokeStyleMiterLimit !== undefined
                ? textStrokeStyleMiterLimit
                : defaultMiterLimit,
            strokeStyle: asColorLike(
              textStrokeStyleColor ? textStrokeStyleColor : defaultStrokeStyle
            ),
          };
        }
        const textFont = textStyle.getFont();
        const textOffsetX = textStyle.getOffsetX();
        const textOffsetY = textStyle.getOffsetY();
        const textRotateWithView = textStyle.getRotateWithView();
        const textRotation = textStyle.getRotation();
        const textScale = textStyle.getScaleArray();
        const textText = textStyle.getText();
        const textTextAlign = textStyle.getTextAlign();
        const textTextBaseline = textStyle.getTextBaseline();
        this.textState_ = {
          font: textFont !== undefined ? textFont : defaultFont,
          textAlign:
            textTextAlign !== undefined ? textTextAlign : defaultTextAlign,
          textBaseline:
            textTextBaseline !== undefined
              ? textTextBaseline
              : defaultTextBaseline,
        };
        this.text_ =
          textText !== undefined
            ? Array.isArray(textText)
              ? textText.reduce((acc, t, i) => (acc += i % 2 ? ' ' : t), '')
              : textText
            : '';
        this.textOffsetX_ =
          textOffsetX !== undefined ? this.pixelRatio_ * textOffsetX : 0;
        this.textOffsetY_ =
          textOffsetY !== undefined ? this.pixelRatio_ * textOffsetY : 0;
        this.textRotateWithView_ =
          textRotateWithView !== undefined ? textRotateWithView : false;
        this.textRotation_ = textRotation !== undefined ? textRotation : 0;
        this.textScale_ = [
          this.pixelRatio_ * textScale[0],
          this.pixelRatio_ * textScale[1],
        ];
      }
    }
  }

  /**
   * @module ol/renderer/vector
   */

  /**
   * Feature callback. The callback will be called with three arguments. The first
   * argument is one {@link module:ol/Feature~Feature feature} or {@link module:ol/render/Feature~RenderFeature render feature}
   * at the pixel, the second is the {@link module:ol/layer/Layer~Layer layer} of the feature and will be null for
   * unmanaged layers. The third is the {@link module:ol/geom/SimpleGeometry~SimpleGeometry} of the feature. For features
   * with a GeometryCollection geometry, it will be the first detected geometry from the collection.
   * @template T
   * @typedef {function(import("../Feature.js").FeatureLike, import("../layer/Layer.js").default<import("../source/Source").default>, import("../geom/SimpleGeometry.js").default): T} FeatureCallback
   */

  /**
   * Tolerance for geometry simplification in device pixels.
   * @type {number}
   */
  const SIMPLIFY_TOLERANCE = 0.5;

  /**
   * @const
   * @type {Object<import("../geom/Geometry.js").Type,
   *                function(import("../render/canvas/BuilderGroup.js").default, import("../geom/Geometry.js").default,
   *                         import("../style/Style.js").default, Object): void>}
   */
  const GEOMETRY_RENDERERS = {
    'Point': renderPointGeometry,
    'LineString': renderLineStringGeometry,
    'Polygon': renderPolygonGeometry,
    'MultiPoint': renderMultiPointGeometry,
    'MultiLineString': renderMultiLineStringGeometry,
    'MultiPolygon': renderMultiPolygonGeometry,
    'GeometryCollection': renderGeometryCollectionGeometry,
    'Circle': renderCircleGeometry,
  };

  /**
   * @param {import("../Feature.js").FeatureLike} feature1 Feature 1.
   * @param {import("../Feature.js").FeatureLike} feature2 Feature 2.
   * @return {number} Order.
   */
  function defaultOrder(feature1, feature2) {
    return parseInt(getUid(feature1), 10) - parseInt(getUid(feature2), 10);
  }

  /**
   * @param {number} resolution Resolution.
   * @param {number} pixelRatio Pixel ratio.
   * @return {number} Squared pixel tolerance.
   */
  function getSquaredTolerance(resolution, pixelRatio) {
    const tolerance = getTolerance(resolution, pixelRatio);
    return tolerance * tolerance;
  }

  /**
   * @param {number} resolution Resolution.
   * @param {number} pixelRatio Pixel ratio.
   * @return {number} Pixel tolerance.
   */
  function getTolerance(resolution, pixelRatio) {
    return (SIMPLIFY_TOLERANCE * resolution) / pixelRatio;
  }

  /**
   * @param {import("../render/canvas/BuilderGroup.js").default} builderGroup Builder group.
   * @param {import("../geom/Circle.js").default} geometry Geometry.
   * @param {import("../style/Style.js").default} style Style.
   * @param {import("../Feature.js").default} feature Feature.
   * @param {import("../render/canvas/BuilderGroup.js").default} [declutterBuilderGroup] Builder for decluttering.
   */
  function renderCircleGeometry(
    builderGroup,
    geometry,
    style,
    feature,
    declutterBuilderGroup
  ) {
    const fillStyle = style.getFill();
    const strokeStyle = style.getStroke();
    if (fillStyle || strokeStyle) {
      const circleReplay = builderGroup.getBuilder(style.getZIndex(), 'Circle');
      circleReplay.setFillStrokeStyle(fillStyle, strokeStyle);
      circleReplay.drawCircle(geometry, feature);
    }
    const textStyle = style.getText();
    if (textStyle && textStyle.getText()) {
      const textReplay = (declutterBuilderGroup || builderGroup).getBuilder(
        style.getZIndex(),
        'Text'
      );
      textReplay.setTextStyle(textStyle);
      textReplay.drawText(geometry, feature);
    }
  }

  /**
   * @param {import("../render/canvas/BuilderGroup.js").default} replayGroup Replay group.
   * @param {import("../Feature.js").FeatureLike} feature Feature.
   * @param {import("../style/Style.js").default} style Style.
   * @param {number} squaredTolerance Squared tolerance.
   * @param {function(import("../events/Event.js").default): void} listener Listener function.
   * @param {import("../proj.js").TransformFunction} [transform] Transform from user to view projection.
   * @param {import("../render/canvas/BuilderGroup.js").default} [declutterBuilderGroup] Builder for decluttering.
   * @return {boolean} `true` if style is loading.
   */
  function renderFeature$1(
    replayGroup,
    feature,
    style,
    squaredTolerance,
    listener,
    transform,
    declutterBuilderGroup
  ) {
    let loading = false;
    const imageStyle = style.getImage();
    if (imageStyle) {
      const imageState = imageStyle.getImageState();
      if (imageState == ImageState.LOADED || imageState == ImageState.ERROR) {
        imageStyle.unlistenImageChange(listener);
      } else {
        if (imageState == ImageState.IDLE) {
          imageStyle.load();
        }
        imageStyle.listenImageChange(listener);
        loading = true;
      }
    }
    renderFeatureInternal(
      replayGroup,
      feature,
      style,
      squaredTolerance,
      transform,
      declutterBuilderGroup
    );

    return loading;
  }

  /**
   * @param {import("../render/canvas/BuilderGroup.js").default} replayGroup Replay group.
   * @param {import("../Feature.js").FeatureLike} feature Feature.
   * @param {import("../style/Style.js").default} style Style.
   * @param {number} squaredTolerance Squared tolerance.
   * @param {import("../proj.js").TransformFunction} [transform] Optional transform function.
   * @param {import("../render/canvas/BuilderGroup.js").default} [declutterBuilderGroup] Builder for decluttering.
   */
  function renderFeatureInternal(
    replayGroup,
    feature,
    style,
    squaredTolerance,
    transform,
    declutterBuilderGroup
  ) {
    const geometry = style.getGeometryFunction()(feature);
    if (!geometry) {
      return;
    }
    const simplifiedGeometry = geometry.simplifyTransformed(
      squaredTolerance,
      transform
    );
    const renderer = style.getRenderer();
    if (renderer) {
      renderGeometry(replayGroup, simplifiedGeometry, style, feature);
    } else {
      const geometryRenderer = GEOMETRY_RENDERERS[simplifiedGeometry.getType()];
      geometryRenderer(
        replayGroup,
        simplifiedGeometry,
        style,
        feature,
        declutterBuilderGroup
      );
    }
  }

  /**
   * @param {import("../render/canvas/BuilderGroup.js").default} replayGroup Replay group.
   * @param {import("../geom/Geometry.js").default|import("../render/Feature.js").default} geometry Geometry.
   * @param {import("../style/Style.js").default} style Style.
   * @param {import("../Feature.js").FeatureLike} feature Feature.
   */
  function renderGeometry(replayGroup, geometry, style, feature) {
    if (geometry.getType() == 'GeometryCollection') {
      const geometries =
        /** @type {import("../geom/GeometryCollection.js").default} */ (
          geometry
        ).getGeometries();
      for (let i = 0, ii = geometries.length; i < ii; ++i) {
        renderGeometry(replayGroup, geometries[i], style, feature);
      }
      return;
    }
    const replay = replayGroup.getBuilder(style.getZIndex(), 'Default');
    replay.drawCustom(
      /** @type {import("../geom/SimpleGeometry.js").default} */ (geometry),
      feature,
      style.getRenderer(),
      style.getHitDetectionRenderer()
    );
  }

  /**
   * @param {import("../render/canvas/BuilderGroup.js").default} replayGroup Replay group.
   * @param {import("../geom/GeometryCollection.js").default} geometry Geometry.
   * @param {import("../style/Style.js").default} style Style.
   * @param {import("../Feature.js").default} feature Feature.
   * @param {import("../render/canvas/BuilderGroup.js").default} [declutterBuilderGroup] Builder for decluttering.
   */
  function renderGeometryCollectionGeometry(
    replayGroup,
    geometry,
    style,
    feature,
    declutterBuilderGroup
  ) {
    const geometries = geometry.getGeometriesArray();
    let i, ii;
    for (i = 0, ii = geometries.length; i < ii; ++i) {
      const geometryRenderer = GEOMETRY_RENDERERS[geometries[i].getType()];
      geometryRenderer(
        replayGroup,
        geometries[i],
        style,
        feature,
        declutterBuilderGroup
      );
    }
  }

  /**
   * @param {import("../render/canvas/BuilderGroup.js").default} builderGroup Replay group.
   * @param {import("../geom/LineString.js").default|import("../render/Feature.js").default} geometry Geometry.
   * @param {import("../style/Style.js").default} style Style.
   * @param {import("../Feature.js").FeatureLike} feature Feature.
   * @param {import("../render/canvas/BuilderGroup.js").default} [declutterBuilderGroup] Builder for decluttering.
   */
  function renderLineStringGeometry(
    builderGroup,
    geometry,
    style,
    feature,
    declutterBuilderGroup
  ) {
    const strokeStyle = style.getStroke();
    if (strokeStyle) {
      const lineStringReplay = builderGroup.getBuilder(
        style.getZIndex(),
        'LineString'
      );
      lineStringReplay.setFillStrokeStyle(null, strokeStyle);
      lineStringReplay.drawLineString(geometry, feature);
    }
    const textStyle = style.getText();
    if (textStyle && textStyle.getText()) {
      const textReplay = (declutterBuilderGroup || builderGroup).getBuilder(
        style.getZIndex(),
        'Text'
      );
      textReplay.setTextStyle(textStyle);
      textReplay.drawText(geometry, feature);
    }
  }

  /**
   * @param {import("../render/canvas/BuilderGroup.js").default} builderGroup Replay group.
   * @param {import("../geom/MultiLineString.js").default|import("../render/Feature.js").default} geometry Geometry.
   * @param {import("../style/Style.js").default} style Style.
   * @param {import("../Feature.js").FeatureLike} feature Feature.
   * @param {import("../render/canvas/BuilderGroup.js").default} [declutterBuilderGroup] Builder for decluttering.
   */
  function renderMultiLineStringGeometry(
    builderGroup,
    geometry,
    style,
    feature,
    declutterBuilderGroup
  ) {
    const strokeStyle = style.getStroke();
    if (strokeStyle) {
      const lineStringReplay = builderGroup.getBuilder(
        style.getZIndex(),
        'LineString'
      );
      lineStringReplay.setFillStrokeStyle(null, strokeStyle);
      lineStringReplay.drawMultiLineString(geometry, feature);
    }
    const textStyle = style.getText();
    if (textStyle && textStyle.getText()) {
      const textReplay = (declutterBuilderGroup || builderGroup).getBuilder(
        style.getZIndex(),
        'Text'
      );
      textReplay.setTextStyle(textStyle);
      textReplay.drawText(geometry, feature);
    }
  }

  /**
   * @param {import("../render/canvas/BuilderGroup.js").default} builderGroup Replay group.
   * @param {import("../geom/MultiPolygon.js").default} geometry Geometry.
   * @param {import("../style/Style.js").default} style Style.
   * @param {import("../Feature.js").default} feature Feature.
   * @param {import("../render/canvas/BuilderGroup.js").default} [declutterBuilderGroup] Builder for decluttering.
   */
  function renderMultiPolygonGeometry(
    builderGroup,
    geometry,
    style,
    feature,
    declutterBuilderGroup
  ) {
    const fillStyle = style.getFill();
    const strokeStyle = style.getStroke();
    if (strokeStyle || fillStyle) {
      const polygonReplay = builderGroup.getBuilder(style.getZIndex(), 'Polygon');
      polygonReplay.setFillStrokeStyle(fillStyle, strokeStyle);
      polygonReplay.drawMultiPolygon(geometry, feature);
    }
    const textStyle = style.getText();
    if (textStyle && textStyle.getText()) {
      const textReplay = (declutterBuilderGroup || builderGroup).getBuilder(
        style.getZIndex(),
        'Text'
      );
      textReplay.setTextStyle(textStyle);
      textReplay.drawText(geometry, feature);
    }
  }

  /**
   * @param {import("../render/canvas/BuilderGroup.js").default} builderGroup Replay group.
   * @param {import("../geom/Point.js").default|import("../render/Feature.js").default} geometry Geometry.
   * @param {import("../style/Style.js").default} style Style.
   * @param {import("../Feature.js").FeatureLike} feature Feature.
   * @param {import("../render/canvas/BuilderGroup.js").default} [declutterBuilderGroup] Builder for decluttering.
   */
  function renderPointGeometry(
    builderGroup,
    geometry,
    style,
    feature,
    declutterBuilderGroup
  ) {
    const imageStyle = style.getImage();
    const textStyle = style.getText();
    /** @type {import("../render/canvas.js").DeclutterImageWithText} */
    let declutterImageWithText;
    if (imageStyle) {
      if (imageStyle.getImageState() != ImageState.LOADED) {
        return;
      }
      let imageBuilderGroup = builderGroup;
      if (declutterBuilderGroup) {
        const declutterMode = imageStyle.getDeclutterMode();
        if (declutterMode !== 'none') {
          imageBuilderGroup = declutterBuilderGroup;
          if (declutterMode === 'obstacle') {
            // draw in non-declutter group:
            const imageReplay = builderGroup.getBuilder(
              style.getZIndex(),
              'Image'
            );
            imageReplay.setImageStyle(imageStyle, declutterImageWithText);
            imageReplay.drawPoint(geometry, feature);
          } else if (textStyle && textStyle.getText()) {
            declutterImageWithText = {};
          }
        }
      }
      const imageReplay = imageBuilderGroup.getBuilder(
        style.getZIndex(),
        'Image'
      );
      imageReplay.setImageStyle(imageStyle, declutterImageWithText);
      imageReplay.drawPoint(geometry, feature);
    }
    if (textStyle && textStyle.getText()) {
      let textBuilderGroup = builderGroup;
      if (declutterBuilderGroup) {
        textBuilderGroup = declutterBuilderGroup;
      }
      const textReplay = textBuilderGroup.getBuilder(style.getZIndex(), 'Text');
      textReplay.setTextStyle(textStyle, declutterImageWithText);
      textReplay.drawText(geometry, feature);
    }
  }

  /**
   * @param {import("../render/canvas/BuilderGroup.js").default} builderGroup Replay group.
   * @param {import("../geom/MultiPoint.js").default|import("../render/Feature.js").default} geometry Geometry.
   * @param {import("../style/Style.js").default} style Style.
   * @param {import("../Feature.js").FeatureLike} feature Feature.
   * @param {import("../render/canvas/BuilderGroup.js").default} [declutterBuilderGroup] Builder for decluttering.
   */
  function renderMultiPointGeometry(
    builderGroup,
    geometry,
    style,
    feature,
    declutterBuilderGroup
  ) {
    const imageStyle = style.getImage();
    const textStyle = style.getText();
    /** @type {import("../render/canvas.js").DeclutterImageWithText} */
    let declutterImageWithText;
    if (imageStyle) {
      if (imageStyle.getImageState() != ImageState.LOADED) {
        return;
      }
      let imageBuilderGroup = builderGroup;
      if (declutterBuilderGroup) {
        const declutterMode = imageStyle.getDeclutterMode();
        if (declutterMode !== 'none') {
          imageBuilderGroup = declutterBuilderGroup;
          if (declutterMode === 'obstacle') {
            // draw in non-declutter group:
            const imageReplay = builderGroup.getBuilder(
              style.getZIndex(),
              'Image'
            );
            imageReplay.setImageStyle(imageStyle, declutterImageWithText);
            imageReplay.drawMultiPoint(geometry, feature);
          } else if (textStyle && textStyle.getText()) {
            declutterImageWithText = {};
          }
        }
      }
      const imageReplay = imageBuilderGroup.getBuilder(
        style.getZIndex(),
        'Image'
      );
      imageReplay.setImageStyle(imageStyle, declutterImageWithText);
      imageReplay.drawMultiPoint(geometry, feature);
    }
    if (textStyle && textStyle.getText()) {
      let textBuilderGroup = builderGroup;
      if (declutterBuilderGroup) {
        textBuilderGroup = declutterBuilderGroup;
      }
      const textReplay = textBuilderGroup.getBuilder(style.getZIndex(), 'Text');
      textReplay.setTextStyle(textStyle, declutterImageWithText);
      textReplay.drawText(geometry, feature);
    }
  }

  /**
   * @param {import("../render/canvas/BuilderGroup.js").default} builderGroup Replay group.
   * @param {import("../geom/Polygon.js").default|import("../render/Feature.js").default} geometry Geometry.
   * @param {import("../style/Style.js").default} style Style.
   * @param {import("../Feature.js").FeatureLike} feature Feature.
   * @param {import("../render/canvas/BuilderGroup.js").default} [declutterBuilderGroup] Builder for decluttering.
   */
  function renderPolygonGeometry(
    builderGroup,
    geometry,
    style,
    feature,
    declutterBuilderGroup
  ) {
    const fillStyle = style.getFill();
    const strokeStyle = style.getStroke();
    if (fillStyle || strokeStyle) {
      const polygonReplay = builderGroup.getBuilder(style.getZIndex(), 'Polygon');
      polygonReplay.setFillStrokeStyle(fillStyle, strokeStyle);
      polygonReplay.drawPolygon(geometry, feature);
    }
    const textStyle = style.getText();
    if (textStyle && textStyle.getText()) {
      const textReplay = (declutterBuilderGroup || builderGroup).getBuilder(
        style.getZIndex(),
        'Text'
      );
      textReplay.setTextStyle(textStyle);
      textReplay.drawText(geometry, feature);
    }
  }

  /**
   * @module ol/render
   */

  /**
   * @typedef {Object} State
   * @property {CanvasRenderingContext2D} context Canvas context that the layer is being rendered to.
   * @property {import("./Feature.js").FeatureLike} feature Feature.
   * @property {import("./geom/SimpleGeometry.js").default} geometry Geometry.
   * @property {number} pixelRatio Pixel ratio used by the layer renderer.
   * @property {number} resolution Resolution that the render batch was created and optimized for.
   * This is not the view's resolution that is being rendered.
   * @property {number} rotation Rotation of the rendered layer in radians.
   */

  /**
   * A function to be used when sorting features before rendering.
   * It takes two instances of {@link module:ol/Feature~Feature} or
   * {@link module:ol/render/Feature~RenderFeature} and returns a `{number}`.
   *
   * @typedef {function(import("./Feature.js").FeatureLike, import("./Feature.js").FeatureLike):number} OrderFunction
   */

  /**
   * @typedef {Object} ToContextOptions
   * @property {import("./size.js").Size} [size] Desired size of the canvas in css
   * pixels. When provided, both canvas and css size will be set according to the
   * `pixelRatio`. If not provided, the current canvas and css sizes will not be
   * altered.
   * @property {number} [pixelRatio=window.devicePixelRatio] Pixel ratio (canvas
   * pixel to css pixel ratio) for the canvas.
   */

  /**
   * Binds a Canvas Immediate API to a canvas context, to allow drawing geometries
   * to the context's canvas.
   *
   * The units for geometry coordinates are css pixels relative to the top left
   * corner of the canvas element.
   * ```js
   * import {toContext} from 'ol/render';
   * import Fill from 'ol/style/Fill';
   * import Polygon from 'ol/geom/Polygon';
   *
   * const canvas = document.createElement('canvas');
   * const render = toContext(
   *     canvas.getContext('2d'),
   *     {size: [100, 100]}
   * );
   * render.setFillStrokeStyle(new Fill({ color: blue }));
   * render.drawPolygon(
   *     new Polygon([[[0, 0], [100, 100], [100, 0], [0, 0]]])
   * );
   * ```
   *
   * @param {CanvasRenderingContext2D} context Canvas context.
   * @param {ToContextOptions} [options] Options.
   * @return {CanvasImmediateRenderer} Canvas Immediate.
   * @api
   */
  function toContext(context, options) {
    const canvas = context.canvas;
    options = options ? options : {};
    const pixelRatio = options.pixelRatio || DEVICE_PIXEL_RATIO;
    const size = options.size;
    if (size) {
      canvas.width = size[0] * pixelRatio;
      canvas.height = size[1] * pixelRatio;
      canvas.style.width = size[0] + 'px';
      canvas.style.height = size[1] + 'px';
    }
    const extent = [0, 0, canvas.width, canvas.height];
    const transform = scale$3(create$2(), pixelRatio, pixelRatio);
    return new CanvasImmediateRenderer(context, pixelRatio, extent, transform, 0);
  }

  /**
   * Gets a vector context for drawing to the event's canvas.
   * @param {import("./render/Event.js").default} event Render event.
   * @return {CanvasImmediateRenderer} Vector context.
   * @api
   */
  function getVectorContext(event) {
    if (!(event.context instanceof CanvasRenderingContext2D)) {
      throw new Error('Only works for render events from Canvas 2D layers');
    }

    // canvas may be at a different pixel ratio than frameState.pixelRatio
    const canvasPixelRatio = event.inversePixelTransform[0];
    const frameState = event.frameState;
    const transform = multiply(
      event.inversePixelTransform.slice(),
      frameState.coordinateToPixelTransform
    );
    const squaredTolerance = getSquaredTolerance(
      frameState.viewState.resolution,
      canvasPixelRatio
    );
    let userTransform;
    const userProjection = getUserProjection();
    if (userProjection) {
      userTransform = getTransformFromProjections(
        userProjection,
        frameState.viewState.projection
      );
    }

    return new CanvasImmediateRenderer(
      event.context,
      canvasPixelRatio,
      frameState.extent,
      transform,
      frameState.viewState.rotation,
      squaredTolerance,
      userTransform
    );
  }

  /**
   * Gets the pixel of the event's canvas context from the map viewport's CSS pixel.
   * @param {import("./render/Event.js").default} event Render event.
   * @param {import("./pixel.js").Pixel} pixel CSS pixel relative to the top-left
   * corner of the map viewport.
   * @return {import("./pixel.js").Pixel} Pixel on the event's canvas context.
   * @api
   */
  function getRenderPixel(event, pixel) {
    return apply(event.inversePixelTransform, pixel.slice(0));
  }

  /**
   * @module ol/reproj
   */

  let brokenDiagonalRendering_;

  /**
   * @type {Array<HTMLCanvasElement>}
   */
  const canvasPool$1 = [];

  /**
   * This draws a small triangle into a canvas by setting the triangle as the clip region
   * and then drawing a (too large) rectangle
   *
   * @param {CanvasRenderingContext2D} ctx The context in which to draw the triangle
   * @param {number} u1 The x-coordinate of the second point. The first point is 0,0.
   * @param {number} v1 The y-coordinate of the second point.
   * @param {number} u2 The x-coordinate of the third point.
   * @param {number} v2 The y-coordinate of the third point.
   */
  function drawTestTriangle(ctx, u1, v1, u2, v2) {
    ctx.beginPath();
    ctx.moveTo(0, 0);
    ctx.lineTo(u1, v1);
    ctx.lineTo(u2, v2);
    ctx.closePath();
    ctx.save();
    ctx.clip();
    ctx.fillRect(0, 0, Math.max(u1, u2) + 1, Math.max(v1, v2));
    ctx.restore();
  }

  /**
   * Given the data from getImageData, see if the right values appear at the provided offset.
   * Returns true if either the color or transparency is off
   *
   * @param {Uint8ClampedArray} data The data returned from getImageData
   * @param {number} offset The pixel offset from the start of data.
   * @return {boolean} true if the diagonal rendering is broken
   */
  function verifyBrokenDiagonalRendering(data, offset) {
    // the values ought to be close to the rgba(210, 0, 0, 0.75)
    return (
      Math.abs(data[offset * 4] - 210) > 2 ||
      Math.abs(data[offset * 4 + 3] - 0.75 * 255) > 2
    );
  }

  /**
   * Determines if the current browser configuration can render triangular clip regions correctly.
   * This value is cached so the function is only expensive the first time called.
   * Firefox on Windows (as of now) does not if HWA is enabled. See https://bugzilla.mozilla.org/show_bug.cgi?id=1606976
   * Chrome works, and everything seems to work on OSX and Android. This function caches the
   * result. I suppose that it is conceivably possible that a browser might flip modes while the app is
   * running, but lets hope not.
   *
   * @return {boolean} true if the Diagonal Rendering is broken.
   */
  function isBrokenDiagonalRendering() {
    if (brokenDiagonalRendering_ === undefined) {
      const ctx = document.createElement('canvas').getContext('2d');
      ctx.globalCompositeOperation = 'lighter';
      ctx.fillStyle = 'rgba(210, 0, 0, 0.75)';
      drawTestTriangle(ctx, 4, 5, 4, 0);
      drawTestTriangle(ctx, 4, 5, 0, 5);
      const data = ctx.getImageData(0, 0, 3, 3).data;
      brokenDiagonalRendering_ =
        verifyBrokenDiagonalRendering(data, 0) ||
        verifyBrokenDiagonalRendering(data, 4) ||
        verifyBrokenDiagonalRendering(data, 8);
    }

    return brokenDiagonalRendering_;
  }

  /**
   * Calculates ideal resolution to use from the source in order to achieve
   * pixel mapping as close as possible to 1:1 during reprojection.
   * The resolution is calculated regardless of what resolutions
   * are actually available in the dataset (TileGrid, Image, ...).
   *
   * @param {import("./proj/Projection.js").default} sourceProj Source projection.
   * @param {import("./proj/Projection.js").default} targetProj Target projection.
   * @param {import("./coordinate.js").Coordinate} targetCenter Target center.
   * @param {number} targetResolution Target resolution.
   * @return {number} The best resolution to use. Can be +-Infinity, NaN or 0.
   */
  function calculateSourceResolution(
    sourceProj,
    targetProj,
    targetCenter,
    targetResolution
  ) {
    const sourceCenter = transform(targetCenter, targetProj, sourceProj);

    // calculate the ideal resolution of the source data
    let sourceResolution = getPointResolution(
      targetProj,
      targetResolution,
      targetCenter
    );

    const targetMetersPerUnit = targetProj.getMetersPerUnit();
    if (targetMetersPerUnit !== undefined) {
      sourceResolution *= targetMetersPerUnit;
    }
    const sourceMetersPerUnit = sourceProj.getMetersPerUnit();
    if (sourceMetersPerUnit !== undefined) {
      sourceResolution /= sourceMetersPerUnit;
    }

    // Based on the projection properties, the point resolution at the specified
    // coordinates may be slightly different. We need to reverse-compensate this
    // in order to achieve optimal results.

    const sourceExtent = sourceProj.getExtent();
    if (!sourceExtent || containsCoordinate(sourceExtent, sourceCenter)) {
      const compensationFactor =
        getPointResolution(sourceProj, sourceResolution, sourceCenter) /
        sourceResolution;
      if (isFinite(compensationFactor) && compensationFactor > 0) {
        sourceResolution /= compensationFactor;
      }
    }

    return sourceResolution;
  }

  /**
   * Calculates ideal resolution to use from the source in order to achieve
   * pixel mapping as close as possible to 1:1 during reprojection.
   * The resolution is calculated regardless of what resolutions
   * are actually available in the dataset (TileGrid, Image, ...).
   *
   * @param {import("./proj/Projection.js").default} sourceProj Source projection.
   * @param {import("./proj/Projection.js").default} targetProj Target projection.
   * @param {import("./extent.js").Extent} targetExtent Target extent
   * @param {number} targetResolution Target resolution.
   * @return {number} The best resolution to use. Can be +-Infinity, NaN or 0.
   */
  function calculateSourceExtentResolution(
    sourceProj,
    targetProj,
    targetExtent,
    targetResolution
  ) {
    const targetCenter = getCenter(targetExtent);
    let sourceResolution = calculateSourceResolution(
      sourceProj,
      targetProj,
      targetCenter,
      targetResolution
    );

    if (!isFinite(sourceResolution) || sourceResolution <= 0) {
      forEachCorner(targetExtent, function (corner) {
        sourceResolution = calculateSourceResolution(
          sourceProj,
          targetProj,
          corner,
          targetResolution
        );
        return isFinite(sourceResolution) && sourceResolution > 0;
      });
    }

    return sourceResolution;
  }

  /**
   * @typedef {Object} ImageExtent
   * @property {import("./extent.js").Extent} extent Extent.
   * @property {HTMLCanvasElement|HTMLImageElement|HTMLVideoElement} image Image.
   */

  /**
   * Renders the source data into new canvas based on the triangulation.
   *
   * @param {number} width Width of the canvas.
   * @param {number} height Height of the canvas.
   * @param {number} pixelRatio Pixel ratio.
   * @param {number} sourceResolution Source resolution.
   * @param {import("./extent.js").Extent} sourceExtent Extent of the data source.
   * @param {number} targetResolution Target resolution.
   * @param {import("./extent.js").Extent} targetExtent Target extent.
   * @param {import("./reproj/Triangulation.js").default} triangulation Calculated triangulation.
   * @param {Array<ImageExtent>} sources Array of sources.
   * @param {number} gutter Gutter of the sources.
   * @param {boolean} [renderEdges] Render reprojection edges.
   * @param {boolean} [interpolate] Use linear interpolation when resampling.
   * @return {HTMLCanvasElement} Canvas with reprojected data.
   */
  function render(
    width,
    height,
    pixelRatio,
    sourceResolution,
    sourceExtent,
    targetResolution,
    targetExtent,
    triangulation,
    sources,
    gutter,
    renderEdges,
    interpolate
  ) {
    const context = createCanvasContext2D(
      Math.round(pixelRatio * width),
      Math.round(pixelRatio * height),
      canvasPool$1
    );

    if (!interpolate) {
      context.imageSmoothingEnabled = false;
    }

    if (sources.length === 0) {
      return context.canvas;
    }

    context.scale(pixelRatio, pixelRatio);

    function pixelRound(value) {
      return Math.round(value * pixelRatio) / pixelRatio;
    }

    context.globalCompositeOperation = 'lighter';

    const sourceDataExtent = createEmpty();
    sources.forEach(function (src, i, arr) {
      extend$1(sourceDataExtent, src.extent);
    });

    const canvasWidthInUnits = getWidth(sourceDataExtent);
    const canvasHeightInUnits = getHeight(sourceDataExtent);
    const stitchContext = createCanvasContext2D(
      Math.round((pixelRatio * canvasWidthInUnits) / sourceResolution),
      Math.round((pixelRatio * canvasHeightInUnits) / sourceResolution)
    );

    if (!interpolate) {
      stitchContext.imageSmoothingEnabled = false;
    }

    const stitchScale = pixelRatio / sourceResolution;

    sources.forEach(function (src, i, arr) {
      const xPos = src.extent[0] - sourceDataExtent[0];
      const yPos = -(src.extent[3] - sourceDataExtent[3]);
      const srcWidth = getWidth(src.extent);
      const srcHeight = getHeight(src.extent);

      // This test should never fail -- but it does. Need to find a fix the upstream condition
      if (src.image.width > 0 && src.image.height > 0) {
        stitchContext.drawImage(
          src.image,
          gutter,
          gutter,
          src.image.width - 2 * gutter,
          src.image.height - 2 * gutter,
          xPos * stitchScale,
          yPos * stitchScale,
          srcWidth * stitchScale,
          srcHeight * stitchScale
        );
      }
    });

    const targetTopLeft = getTopLeft(targetExtent);

    triangulation.getTriangles().forEach(function (triangle, i, arr) {
      /* Calculate affine transform (src -> dst)
       * Resulting matrix can be used to transform coordinate
       * from `sourceProjection` to destination pixels.
       *
       * To optimize number of context calls and increase numerical stability,
       * we also do the following operations:
       * trans(-topLeftExtentCorner), scale(1 / targetResolution), scale(1, -1)
       * here before solving the linear system so [ui, vi] are pixel coordinates.
       *
       * Src points: xi, yi
       * Dst points: ui, vi
       * Affine coefficients: aij
       *
       * | x0 y0 1  0  0 0 |   |a00|   |u0|
       * | x1 y1 1  0  0 0 |   |a01|   |u1|
       * | x2 y2 1  0  0 0 | x |a02| = |u2|
       * |  0  0 0 x0 y0 1 |   |a10|   |v0|
       * |  0  0 0 x1 y1 1 |   |a11|   |v1|
       * |  0  0 0 x2 y2 1 |   |a12|   |v2|
       */
      const source = triangle.source;
      const target = triangle.target;
      let x0 = source[0][0],
        y0 = source[0][1];
      let x1 = source[1][0],
        y1 = source[1][1];
      let x2 = source[2][0],
        y2 = source[2][1];
      // Make sure that everything is on pixel boundaries
      const u0 = pixelRound((target[0][0] - targetTopLeft[0]) / targetResolution);
      const v0 = pixelRound(
        -(target[0][1] - targetTopLeft[1]) / targetResolution
      );
      const u1 = pixelRound((target[1][0] - targetTopLeft[0]) / targetResolution);
      const v1 = pixelRound(
        -(target[1][1] - targetTopLeft[1]) / targetResolution
      );
      const u2 = pixelRound((target[2][0] - targetTopLeft[0]) / targetResolution);
      const v2 = pixelRound(
        -(target[2][1] - targetTopLeft[1]) / targetResolution
      );

      // Shift all the source points to improve numerical stability
      // of all the subsequent calculations. The [x0, y0] is used here.
      // This is also used to simplify the linear system.
      const sourceNumericalShiftX = x0;
      const sourceNumericalShiftY = y0;
      x0 = 0;
      y0 = 0;
      x1 -= sourceNumericalShiftX;
      y1 -= sourceNumericalShiftY;
      x2 -= sourceNumericalShiftX;
      y2 -= sourceNumericalShiftY;

      const augmentedMatrix = [
        [x1, y1, 0, 0, u1 - u0],
        [x2, y2, 0, 0, u2 - u0],
        [0, 0, x1, y1, v1 - v0],
        [0, 0, x2, y2, v2 - v0],
      ];
      const affineCoefs = solveLinearSystem(augmentedMatrix);
      if (!affineCoefs) {
        return;
      }

      context.save();
      context.beginPath();

      if (isBrokenDiagonalRendering() || !interpolate) {
        // Make sure that all lines are horizontal or vertical
        context.moveTo(u1, v1);
        // This is the diagonal line. Do it in 4 steps
        const steps = 4;
        const ud = u0 - u1;
        const vd = v0 - v1;
        for (let step = 0; step < steps; step++) {
          // Go horizontally
          context.lineTo(
            u1 + pixelRound(((step + 1) * ud) / steps),
            v1 + pixelRound((step * vd) / (steps - 1))
          );
          // Go vertically
          if (step != steps - 1) {
            context.lineTo(
              u1 + pixelRound(((step + 1) * ud) / steps),
              v1 + pixelRound(((step + 1) * vd) / (steps - 1))
            );
          }
        }
        // We are almost at u0r, v0r
        context.lineTo(u2, v2);
      } else {
        context.moveTo(u1, v1);
        context.lineTo(u0, v0);
        context.lineTo(u2, v2);
      }

      context.clip();

      context.transform(
        affineCoefs[0],
        affineCoefs[2],
        affineCoefs[1],
        affineCoefs[3],
        u0,
        v0
      );

      context.translate(
        sourceDataExtent[0] - sourceNumericalShiftX,
        sourceDataExtent[3] - sourceNumericalShiftY
      );

      context.scale(
        sourceResolution / pixelRatio,
        -sourceResolution / pixelRatio
      );

      context.drawImage(stitchContext.canvas, 0, 0);
      context.restore();
    });

    if (renderEdges) {
      context.save();

      context.globalCompositeOperation = 'source-over';
      context.strokeStyle = 'black';
      context.lineWidth = 1;

      triangulation.getTriangles().forEach(function (triangle, i, arr) {
        const target = triangle.target;
        const u0 = (target[0][0] - targetTopLeft[0]) / targetResolution;
        const v0 = -(target[0][1] - targetTopLeft[1]) / targetResolution;
        const u1 = (target[1][0] - targetTopLeft[0]) / targetResolution;
        const v1 = -(target[1][1] - targetTopLeft[1]) / targetResolution;
        const u2 = (target[2][0] - targetTopLeft[0]) / targetResolution;
        const v2 = -(target[2][1] - targetTopLeft[1]) / targetResolution;

        context.beginPath();
        context.moveTo(u1, v1);
        context.lineTo(u0, v0);
        context.lineTo(u2, v2);
        context.closePath();
        context.stroke();
      });

      context.restore();
    }
    return context.canvas;
  }

  /**
   * @module ol/reproj/common
   */

  /**
   * Default maximum allowed threshold  (in pixels) for reprojection
   * triangulation.
   * @type {number}
   */
  const ERROR_THRESHOLD = 0.5;

  /**
   * @module ol/reproj/Triangulation
   */

  /**
   * Single triangle; consists of 3 source points and 3 target points.
   * @typedef {Object} Triangle
   * @property {Array<import("../coordinate.js").Coordinate>} source Source.
   * @property {Array<import("../coordinate.js").Coordinate>} target Target.
   */

  /**
   * Maximum number of subdivision steps during raster reprojection triangulation.
   * Prevents high memory usage and large number of proj4 calls (for certain
   * transformations and areas). At most `2*(2^this)` triangles are created for
   * each triangulated extent (tile/image).
   * @type {number}
   */
  const MAX_SUBDIVISION = 10;

  /**
   * Maximum allowed size of triangle relative to world width. When transforming
   * corners of world extent between certain projections, the resulting
   * triangulation seems to have zero error and no subdivision is performed. If
   * the triangle width is more than this (relative to world width; 0-1),
   * subdivison is forced (up to `MAX_SUBDIVISION`). Default is `0.25`.
   * @type {number}
   */
  const MAX_TRIANGLE_WIDTH = 0.25;

  /**
   * @classdesc
   * Class containing triangulation of the given target extent.
   * Used for determining source data and the reprojection itself.
   */
  class Triangulation {
    /**
     * @param {import("../proj/Projection.js").default} sourceProj Source projection.
     * @param {import("../proj/Projection.js").default} targetProj Target projection.
     * @param {import("../extent.js").Extent} targetExtent Target extent to triangulate.
     * @param {import("../extent.js").Extent} maxSourceExtent Maximal source extent that can be used.
     * @param {number} errorThreshold Acceptable error (in source units).
     * @param {?number} destinationResolution The (optional) resolution of the destination.
     */
    constructor(
      sourceProj,
      targetProj,
      targetExtent,
      maxSourceExtent,
      errorThreshold,
      destinationResolution
    ) {
      /**
       * @type {import("../proj/Projection.js").default}
       * @private
       */
      this.sourceProj_ = sourceProj;

      /**
       * @type {import("../proj/Projection.js").default}
       * @private
       */
      this.targetProj_ = targetProj;

      /** @type {!Object<string, import("../coordinate.js").Coordinate>} */
      let transformInvCache = {};
      const transformInv = getTransform(this.targetProj_, this.sourceProj_);

      /**
       * @param {import("../coordinate.js").Coordinate} c A coordinate.
       * @return {import("../coordinate.js").Coordinate} Transformed coordinate.
       * @private
       */
      this.transformInv_ = function (c) {
        const key = c[0] + '/' + c[1];
        if (!transformInvCache[key]) {
          transformInvCache[key] = transformInv(c);
        }
        return transformInvCache[key];
      };

      /**
       * @type {import("../extent.js").Extent}
       * @private
       */
      this.maxSourceExtent_ = maxSourceExtent;

      /**
       * @type {number}
       * @private
       */
      this.errorThresholdSquared_ = errorThreshold * errorThreshold;

      /**
       * @type {Array<Triangle>}
       * @private
       */
      this.triangles_ = [];

      /**
       * Indicates that the triangulation crosses edge of the source projection.
       * @type {boolean}
       * @private
       */
      this.wrapsXInSource_ = false;

      /**
       * @type {boolean}
       * @private
       */
      this.canWrapXInSource_ =
        this.sourceProj_.canWrapX() &&
        !!maxSourceExtent &&
        !!this.sourceProj_.getExtent() &&
        getWidth(maxSourceExtent) == getWidth(this.sourceProj_.getExtent());

      /**
       * @type {?number}
       * @private
       */
      this.sourceWorldWidth_ = this.sourceProj_.getExtent()
        ? getWidth(this.sourceProj_.getExtent())
        : null;

      /**
       * @type {?number}
       * @private
       */
      this.targetWorldWidth_ = this.targetProj_.getExtent()
        ? getWidth(this.targetProj_.getExtent())
        : null;

      const destinationTopLeft = getTopLeft(targetExtent);
      const destinationTopRight = getTopRight(targetExtent);
      const destinationBottomRight = getBottomRight(targetExtent);
      const destinationBottomLeft = getBottomLeft(targetExtent);
      const sourceTopLeft = this.transformInv_(destinationTopLeft);
      const sourceTopRight = this.transformInv_(destinationTopRight);
      const sourceBottomRight = this.transformInv_(destinationBottomRight);
      const sourceBottomLeft = this.transformInv_(destinationBottomLeft);

      /*
       * The maxSubdivision controls how many splittings of the target area can
       * be done. The idea here is to do a linear mapping of the target areas
       * but the actual overal reprojection (can be) extremely non-linear. The
       * default value of MAX_SUBDIVISION was chosen based on mapping a 256x256
       * tile size. However this function is also called to remap canvas rendered
       * layers which can be much larger. This calculation increases the maxSubdivision
       * value by the right factor so that each 256x256 pixel area has
       * MAX_SUBDIVISION divisions.
       */
      const maxSubdivision =
        MAX_SUBDIVISION +
        (destinationResolution
          ? Math.max(
              0,
              Math.ceil(
                Math.log2(
                  getArea$1(targetExtent) /
                    (destinationResolution * destinationResolution * 256 * 256)
                )
              )
            )
          : 0);

      this.addQuad_(
        destinationTopLeft,
        destinationTopRight,
        destinationBottomRight,
        destinationBottomLeft,
        sourceTopLeft,
        sourceTopRight,
        sourceBottomRight,
        sourceBottomLeft,
        maxSubdivision
      );

      if (this.wrapsXInSource_) {
        let leftBound = Infinity;
        this.triangles_.forEach(function (triangle, i, arr) {
          leftBound = Math.min(
            leftBound,
            triangle.source[0][0],
            triangle.source[1][0],
            triangle.source[2][0]
          );
        });

        // Shift triangles to be as close to `leftBound` as possible
        // (if the distance is more than `worldWidth / 2` it can be closer.
        this.triangles_.forEach(
          function (triangle) {
            if (
              Math.max(
                triangle.source[0][0],
                triangle.source[1][0],
                triangle.source[2][0]
              ) -
                leftBound >
              this.sourceWorldWidth_ / 2
            ) {
              const newTriangle = [
                [triangle.source[0][0], triangle.source[0][1]],
                [triangle.source[1][0], triangle.source[1][1]],
                [triangle.source[2][0], triangle.source[2][1]],
              ];
              if (newTriangle[0][0] - leftBound > this.sourceWorldWidth_ / 2) {
                newTriangle[0][0] -= this.sourceWorldWidth_;
              }
              if (newTriangle[1][0] - leftBound > this.sourceWorldWidth_ / 2) {
                newTriangle[1][0] -= this.sourceWorldWidth_;
              }
              if (newTriangle[2][0] - leftBound > this.sourceWorldWidth_ / 2) {
                newTriangle[2][0] -= this.sourceWorldWidth_;
              }

              // Rarely (if the extent contains both the dateline and prime meridian)
              // the shift can in turn break some triangles.
              // Detect this here and don't shift in such cases.
              const minX = Math.min(
                newTriangle[0][0],
                newTriangle[1][0],
                newTriangle[2][0]
              );
              const maxX = Math.max(
                newTriangle[0][0],
                newTriangle[1][0],
                newTriangle[2][0]
              );
              if (maxX - minX < this.sourceWorldWidth_ / 2) {
                triangle.source = newTriangle;
              }
            }
          }.bind(this)
        );
      }

      transformInvCache = {};
    }

    /**
     * Adds triangle to the triangulation.
     * @param {import("../coordinate.js").Coordinate} a The target a coordinate.
     * @param {import("../coordinate.js").Coordinate} b The target b coordinate.
     * @param {import("../coordinate.js").Coordinate} c The target c coordinate.
     * @param {import("../coordinate.js").Coordinate} aSrc The source a coordinate.
     * @param {import("../coordinate.js").Coordinate} bSrc The source b coordinate.
     * @param {import("../coordinate.js").Coordinate} cSrc The source c coordinate.
     * @private
     */
    addTriangle_(a, b, c, aSrc, bSrc, cSrc) {
      this.triangles_.push({
        source: [aSrc, bSrc, cSrc],
        target: [a, b, c],
      });
    }

    /**
     * Adds quad (points in clock-wise order) to the triangulation
     * (and reprojects the vertices) if valid.
     * Performs quad subdivision if needed to increase precision.
     *
     * @param {import("../coordinate.js").Coordinate} a The target a coordinate.
     * @param {import("../coordinate.js").Coordinate} b The target b coordinate.
     * @param {import("../coordinate.js").Coordinate} c The target c coordinate.
     * @param {import("../coordinate.js").Coordinate} d The target d coordinate.
     * @param {import("../coordinate.js").Coordinate} aSrc The source a coordinate.
     * @param {import("../coordinate.js").Coordinate} bSrc The source b coordinate.
     * @param {import("../coordinate.js").Coordinate} cSrc The source c coordinate.
     * @param {import("../coordinate.js").Coordinate} dSrc The source d coordinate.
     * @param {number} maxSubdivision Maximal allowed subdivision of the quad.
     * @private
     */
    addQuad_(a, b, c, d, aSrc, bSrc, cSrc, dSrc, maxSubdivision) {
      const sourceQuadExtent = boundingExtent([aSrc, bSrc, cSrc, dSrc]);
      const sourceCoverageX = this.sourceWorldWidth_
        ? getWidth(sourceQuadExtent) / this.sourceWorldWidth_
        : null;
      const sourceWorldWidth = /** @type {number} */ (this.sourceWorldWidth_);

      // when the quad is wrapped in the source projection
      // it covers most of the projection extent, but not fully
      const wrapsX =
        this.sourceProj_.canWrapX() &&
        sourceCoverageX > 0.5 &&
        sourceCoverageX < 1;

      let needsSubdivision = false;

      if (maxSubdivision > 0) {
        if (this.targetProj_.isGlobal() && this.targetWorldWidth_) {
          const targetQuadExtent = boundingExtent([a, b, c, d]);
          const targetCoverageX =
            getWidth(targetQuadExtent) / this.targetWorldWidth_;
          needsSubdivision =
            targetCoverageX > MAX_TRIANGLE_WIDTH || needsSubdivision;
        }
        if (!wrapsX && this.sourceProj_.isGlobal() && sourceCoverageX) {
          needsSubdivision =
            sourceCoverageX > MAX_TRIANGLE_WIDTH || needsSubdivision;
        }
      }

      if (!needsSubdivision && this.maxSourceExtent_) {
        if (
          isFinite(sourceQuadExtent[0]) &&
          isFinite(sourceQuadExtent[1]) &&
          isFinite(sourceQuadExtent[2]) &&
          isFinite(sourceQuadExtent[3])
        ) {
          if (!intersects$3(sourceQuadExtent, this.maxSourceExtent_)) {
            // whole quad outside source projection extent -> ignore
            return;
          }
        }
      }

      let isNotFinite = 0;

      if (!needsSubdivision) {
        if (
          !isFinite(aSrc[0]) ||
          !isFinite(aSrc[1]) ||
          !isFinite(bSrc[0]) ||
          !isFinite(bSrc[1]) ||
          !isFinite(cSrc[0]) ||
          !isFinite(cSrc[1]) ||
          !isFinite(dSrc[0]) ||
          !isFinite(dSrc[1])
        ) {
          if (maxSubdivision > 0) {
            needsSubdivision = true;
          } else {
            // It might be the case that only 1 of the points is infinite. In this case
            // we can draw a single triangle with the other three points
            isNotFinite =
              (!isFinite(aSrc[0]) || !isFinite(aSrc[1]) ? 8 : 0) +
              (!isFinite(bSrc[0]) || !isFinite(bSrc[1]) ? 4 : 0) +
              (!isFinite(cSrc[0]) || !isFinite(cSrc[1]) ? 2 : 0) +
              (!isFinite(dSrc[0]) || !isFinite(dSrc[1]) ? 1 : 0);
            if (
              isNotFinite != 1 &&
              isNotFinite != 2 &&
              isNotFinite != 4 &&
              isNotFinite != 8
            ) {
              return;
            }
          }
        }
      }

      if (maxSubdivision > 0) {
        if (!needsSubdivision) {
          const center = [(a[0] + c[0]) / 2, (a[1] + c[1]) / 2];
          const centerSrc = this.transformInv_(center);

          let dx;
          if (wrapsX) {
            const centerSrcEstimX =
              (modulo(aSrc[0], sourceWorldWidth) +
                modulo(cSrc[0], sourceWorldWidth)) /
              2;
            dx = centerSrcEstimX - modulo(centerSrc[0], sourceWorldWidth);
          } else {
            dx = (aSrc[0] + cSrc[0]) / 2 - centerSrc[0];
          }
          const dy = (aSrc[1] + cSrc[1]) / 2 - centerSrc[1];
          const centerSrcErrorSquared = dx * dx + dy * dy;
          needsSubdivision = centerSrcErrorSquared > this.errorThresholdSquared_;
        }
        if (needsSubdivision) {
          if (Math.abs(a[0] - c[0]) <= Math.abs(a[1] - c[1])) {
            // split horizontally (top & bottom)
            const bc = [(b[0] + c[0]) / 2, (b[1] + c[1]) / 2];
            const bcSrc = this.transformInv_(bc);
            const da = [(d[0] + a[0]) / 2, (d[1] + a[1]) / 2];
            const daSrc = this.transformInv_(da);

            this.addQuad_(
              a,
              b,
              bc,
              da,
              aSrc,
              bSrc,
              bcSrc,
              daSrc,
              maxSubdivision - 1
            );
            this.addQuad_(
              da,
              bc,
              c,
              d,
              daSrc,
              bcSrc,
              cSrc,
              dSrc,
              maxSubdivision - 1
            );
          } else {
            // split vertically (left & right)
            const ab = [(a[0] + b[0]) / 2, (a[1] + b[1]) / 2];
            const abSrc = this.transformInv_(ab);
            const cd = [(c[0] + d[0]) / 2, (c[1] + d[1]) / 2];
            const cdSrc = this.transformInv_(cd);

            this.addQuad_(
              a,
              ab,
              cd,
              d,
              aSrc,
              abSrc,
              cdSrc,
              dSrc,
              maxSubdivision - 1
            );
            this.addQuad_(
              ab,
              b,
              c,
              cd,
              abSrc,
              bSrc,
              cSrc,
              cdSrc,
              maxSubdivision - 1
            );
          }
          return;
        }
      }

      if (wrapsX) {
        if (!this.canWrapXInSource_) {
          return;
        }
        this.wrapsXInSource_ = true;
      }

      // Exactly zero or one of *Src is not finite
      // The triangles must have the diagonal line as the first side
      // This is to allow easy code in reproj.s to make it straight for broken
      // browsers that can't handle diagonal clipping
      if ((isNotFinite & 0xb) == 0) {
        this.addTriangle_(a, c, d, aSrc, cSrc, dSrc);
      }
      if ((isNotFinite & 0xe) == 0) {
        this.addTriangle_(a, c, b, aSrc, cSrc, bSrc);
      }
      if (isNotFinite) {
        // Try the other two triangles
        if ((isNotFinite & 0xd) == 0) {
          this.addTriangle_(b, d, a, bSrc, dSrc, aSrc);
        }
        if ((isNotFinite & 0x7) == 0) {
          this.addTriangle_(b, d, c, bSrc, dSrc, cSrc);
        }
      }
    }

    /**
     * Calculates extent of the `source` coordinates from all the triangles.
     *
     * @return {import("../extent.js").Extent} Calculated extent.
     */
    calculateSourceExtent() {
      const extent = createEmpty();

      this.triangles_.forEach(function (triangle, i, arr) {
        const src = triangle.source;
        extendCoordinate(extent, src[0]);
        extendCoordinate(extent, src[1]);
        extendCoordinate(extent, src[2]);
      });

      return extent;
    }

    /**
     * @return {Array<Triangle>} Array of the calculated triangles.
     */
    getTriangles() {
      return this.triangles_;
    }
  }

  /**
   * @module ol/reproj/Tile
   */

  /**
   * @typedef {function(number, number, number, number) : import("../Tile.js").default} FunctionType
   */

  /**
   * @classdesc
   * Class encapsulating single reprojected tile.
   * See {@link module:ol/source/TileImage~TileImage}.
   *
   */
  class ReprojTile extends Tile {
    /**
     * @param {import("../proj/Projection.js").default} sourceProj Source projection.
     * @param {import("../tilegrid/TileGrid.js").default} sourceTileGrid Source tile grid.
     * @param {import("../proj/Projection.js").default} targetProj Target projection.
     * @param {import("../tilegrid/TileGrid.js").default} targetTileGrid Target tile grid.
     * @param {import("../tilecoord.js").TileCoord} tileCoord Coordinate of the tile.
     * @param {import("../tilecoord.js").TileCoord} wrappedTileCoord Coordinate of the tile wrapped in X.
     * @param {number} pixelRatio Pixel ratio.
     * @param {number} gutter Gutter of the source tiles.
     * @param {FunctionType} getTileFunction
     *     Function returning source tiles (z, x, y, pixelRatio).
     * @param {number} [errorThreshold] Acceptable reprojection error (in px).
     * @param {boolean} [renderEdges] Render reprojection edges.
     * @param {boolean} [interpolate] Use linear interpolation when resampling.
     */
    constructor(
      sourceProj,
      sourceTileGrid,
      targetProj,
      targetTileGrid,
      tileCoord,
      wrappedTileCoord,
      pixelRatio,
      gutter,
      getTileFunction,
      errorThreshold,
      renderEdges,
      interpolate
    ) {
      super(tileCoord, TileState.IDLE, {interpolate: !!interpolate});

      /**
       * @private
       * @type {boolean}
       */
      this.renderEdges_ = renderEdges !== undefined ? renderEdges : false;

      /**
       * @private
       * @type {number}
       */
      this.pixelRatio_ = pixelRatio;

      /**
       * @private
       * @type {number}
       */
      this.gutter_ = gutter;

      /**
       * @private
       * @type {HTMLCanvasElement}
       */
      this.canvas_ = null;

      /**
       * @private
       * @type {import("../tilegrid/TileGrid.js").default}
       */
      this.sourceTileGrid_ = sourceTileGrid;

      /**
       * @private
       * @type {import("../tilegrid/TileGrid.js").default}
       */
      this.targetTileGrid_ = targetTileGrid;

      /**
       * @private
       * @type {import("../tilecoord.js").TileCoord}
       */
      this.wrappedTileCoord_ = wrappedTileCoord ? wrappedTileCoord : tileCoord;

      /**
       * @private
       * @type {!Array<import("../Tile.js").default>}
       */
      this.sourceTiles_ = [];

      /**
       * @private
       * @type {?Array<import("../events.js").EventsKey>}
       */
      this.sourcesListenerKeys_ = null;

      /**
       * @private
       * @type {number}
       */
      this.sourceZ_ = 0;

      const targetExtent = targetTileGrid.getTileCoordExtent(
        this.wrappedTileCoord_
      );
      const maxTargetExtent = this.targetTileGrid_.getExtent();
      let maxSourceExtent = this.sourceTileGrid_.getExtent();

      const limitedTargetExtent = maxTargetExtent
        ? getIntersection(targetExtent, maxTargetExtent)
        : targetExtent;

      if (getArea$1(limitedTargetExtent) === 0) {
        // Tile is completely outside range -> EMPTY
        // TODO: is it actually correct that the source even creates the tile ?
        this.state = TileState.EMPTY;
        return;
      }

      const sourceProjExtent = sourceProj.getExtent();
      if (sourceProjExtent) {
        if (!maxSourceExtent) {
          maxSourceExtent = sourceProjExtent;
        } else {
          maxSourceExtent = getIntersection(maxSourceExtent, sourceProjExtent);
        }
      }

      const targetResolution = targetTileGrid.getResolution(
        this.wrappedTileCoord_[0]
      );

      const sourceResolution = calculateSourceExtentResolution(
        sourceProj,
        targetProj,
        limitedTargetExtent,
        targetResolution
      );

      if (!isFinite(sourceResolution) || sourceResolution <= 0) {
        // invalid sourceResolution -> EMPTY
        // probably edges of the projections when no extent is defined
        this.state = TileState.EMPTY;
        return;
      }

      const errorThresholdInPixels =
        errorThreshold !== undefined ? errorThreshold : ERROR_THRESHOLD;

      /**
       * @private
       * @type {!import("./Triangulation.js").default}
       */
      this.triangulation_ = new Triangulation(
        sourceProj,
        targetProj,
        limitedTargetExtent,
        maxSourceExtent,
        sourceResolution * errorThresholdInPixels,
        targetResolution
      );

      if (this.triangulation_.getTriangles().length === 0) {
        // no valid triangles -> EMPTY
        this.state = TileState.EMPTY;
        return;
      }

      this.sourceZ_ = sourceTileGrid.getZForResolution(sourceResolution);
      let sourceExtent = this.triangulation_.calculateSourceExtent();

      if (maxSourceExtent) {
        if (sourceProj.canWrapX()) {
          sourceExtent[1] = clamp(
            sourceExtent[1],
            maxSourceExtent[1],
            maxSourceExtent[3]
          );
          sourceExtent[3] = clamp(
            sourceExtent[3],
            maxSourceExtent[1],
            maxSourceExtent[3]
          );
        } else {
          sourceExtent = getIntersection(sourceExtent, maxSourceExtent);
        }
      }

      if (!getArea$1(sourceExtent)) {
        this.state = TileState.EMPTY;
      } else {
        const sourceRange = sourceTileGrid.getTileRangeForExtentAndZ(
          sourceExtent,
          this.sourceZ_
        );

        for (let srcX = sourceRange.minX; srcX <= sourceRange.maxX; srcX++) {
          for (let srcY = sourceRange.minY; srcY <= sourceRange.maxY; srcY++) {
            const tile = getTileFunction(this.sourceZ_, srcX, srcY, pixelRatio);
            if (tile) {
              this.sourceTiles_.push(tile);
            }
          }
        }

        if (this.sourceTiles_.length === 0) {
          this.state = TileState.EMPTY;
        }
      }
    }

    /**
     * Get the HTML Canvas element for this tile.
     * @return {HTMLCanvasElement} Canvas.
     */
    getImage() {
      return this.canvas_;
    }

    /**
     * @private
     */
    reproject_() {
      const sources = [];
      this.sourceTiles_.forEach(
        function (tile, i, arr) {
          if (tile && tile.getState() == TileState.LOADED) {
            sources.push({
              extent: this.sourceTileGrid_.getTileCoordExtent(tile.tileCoord),
              image: tile.getImage(),
            });
          }
        }.bind(this)
      );
      this.sourceTiles_.length = 0;

      if (sources.length === 0) {
        this.state = TileState.ERROR;
      } else {
        const z = this.wrappedTileCoord_[0];
        const size = this.targetTileGrid_.getTileSize(z);
        const width = typeof size === 'number' ? size : size[0];
        const height = typeof size === 'number' ? size : size[1];
        const targetResolution = this.targetTileGrid_.getResolution(z);
        const sourceResolution = this.sourceTileGrid_.getResolution(
          this.sourceZ_
        );

        const targetExtent = this.targetTileGrid_.getTileCoordExtent(
          this.wrappedTileCoord_
        );

        this.canvas_ = render(
          width,
          height,
          this.pixelRatio_,
          sourceResolution,
          this.sourceTileGrid_.getExtent(),
          targetResolution,
          targetExtent,
          this.triangulation_,
          sources,
          this.gutter_,
          this.renderEdges_,
          this.interpolate
        );

        this.state = TileState.LOADED;
      }
      this.changed();
    }

    /**
     * Load not yet loaded URI.
     */
    load() {
      if (this.state == TileState.IDLE) {
        this.state = TileState.LOADING;
        this.changed();

        let leftToLoad = 0;

        this.sourcesListenerKeys_ = [];
        this.sourceTiles_.forEach(
          function (tile, i, arr) {
            const state = tile.getState();
            if (state == TileState.IDLE || state == TileState.LOADING) {
              leftToLoad++;

              const sourceListenKey = listen(
                tile,
                EventType$1.CHANGE,
                function (e) {
                  const state = tile.getState();
                  if (
                    state == TileState.LOADED ||
                    state == TileState.ERROR ||
                    state == TileState.EMPTY
                  ) {
                    unlistenByKey(sourceListenKey);
                    leftToLoad--;
                    if (leftToLoad === 0) {
                      this.unlistenSources_();
                      this.reproject_();
                    }
                  }
                },
                this
              );
              this.sourcesListenerKeys_.push(sourceListenKey);
            }
          }.bind(this)
        );

        if (leftToLoad === 0) {
          setTimeout(this.reproject_.bind(this), 0);
        } else {
          this.sourceTiles_.forEach(function (tile, i, arr) {
            const state = tile.getState();
            if (state == TileState.IDLE) {
              tile.load();
            }
          });
        }
      }
    }

    /**
     * @private
     */
    unlistenSources_() {
      this.sourcesListenerKeys_.forEach(unlistenByKey);
      this.sourcesListenerKeys_ = null;
    }

    /**
     * Remove from the cache due to expiry
     */
    release() {
      if (this.canvas_) {
        releaseCanvas$1(this.canvas_.getContext('2d'));
        canvasPool$1.push(this.canvas_);
        this.canvas_ = null;
      }
      super.release();
    }
  }

  /**
   * @module ol/source/TileEventType
   */

  /**
   * @enum {string}
   */
  var TileEventType = {
    /**
     * Triggered when a tile starts loading.
     * @event module:ol/source/Tile.TileSourceEvent#tileloadstart
     * @api
     */
    TILELOADSTART: 'tileloadstart',

    /**
     * Triggered when a tile finishes loading, either when its data is loaded,
     * or when loading was aborted because the tile is no longer needed.
     * @event module:ol/source/Tile.TileSourceEvent#tileloadend
     * @api
     */
    TILELOADEND: 'tileloadend',

    /**
     * Triggered if tile loading results in an error. Note that this is not the
     * right place to re-fetch tiles. See {@link module:ol/ImageTile~ImageTile#load}
     * for details.
     * @event module:ol/source/Tile.TileSourceEvent#tileloaderror
     * @api
     */
    TILELOADERROR: 'tileloaderror',
  };

  /**
   * @typedef {'tileloadstart'|'tileloadend'|'tileloaderror'} TileSourceEventTypes
   */

  /**
   * @module ol/source/Source
   */

  /**
   * @typedef {'undefined' | 'loading' | 'ready' | 'error'} State
   * State of the source, one of 'undefined', 'loading', 'ready' or 'error'.
   */

  /**
   * A function that takes a {@link module:ol/Map~FrameState} and returns a string or
   * an array of strings representing source attributions.
   *
   * @typedef {function(import("../Map.js").FrameState): (string|Array<string>)} Attribution
   */

  /**
   * A type that can be used to provide attribution information for data sources.
   *
   * It represents either
   * * a simple string (e.g. `'© Acme Inc.'`)
   * * an array of simple strings (e.g. `['© Acme Inc.', '© Bacme Inc.']`)
   * * a function that returns a string or array of strings ({@link module:ol/source/Source~Attribution})
   *
   * @typedef {string|Array<string>|Attribution} AttributionLike
   */

  /**
   * @typedef {Object} Options
   * @property {AttributionLike} [attributions] Attributions.
   * @property {boolean} [attributionsCollapsible=true] Attributions are collapsible.
   * @property {import("../proj.js").ProjectionLike} [projection] Projection. Default is the view projection.
   * @property {import("./Source.js").State} [state='ready'] State.
   * @property {boolean} [wrapX=false] WrapX.
   * @property {boolean} [interpolate=false] Use interpolated values when resampling.  By default,
   * the nearest neighbor is used when resampling.
   */

  /**
   * @classdesc
   * Abstract base class; normally only used for creating subclasses and not
   * instantiated in apps.
   * Base class for {@link module:ol/layer/Layer~Layer} sources.
   *
   * A generic `change` event is triggered when the state of the source changes.
   * @abstract
   * @api
   */
  class Source extends BaseObject {
    /**
     * @param {Options} options Source options.
     */
    constructor(options) {
      super();

      /**
       * @protected
       * @type {import("../proj/Projection.js").default|null}
       */
      this.projection = get$2(options.projection);

      /**
       * @private
       * @type {?Attribution}
       */
      this.attributions_ = adaptAttributions(options.attributions);

      /**
       * @private
       * @type {boolean}
       */
      this.attributionsCollapsible_ =
        options.attributionsCollapsible !== undefined
          ? options.attributionsCollapsible
          : true;

      /**
       * This source is currently loading data. Sources that defer loading to the
       * map's tile queue never set this to `true`.
       * @type {boolean}
       */
      this.loading = false;

      /**
       * @private
       * @type {import("./Source.js").State}
       */
      this.state_ = options.state !== undefined ? options.state : 'ready';

      /**
       * @private
       * @type {boolean}
       */
      this.wrapX_ = options.wrapX !== undefined ? options.wrapX : false;

      /**
       * @private
       * @type {boolean}
       */
      this.interpolate_ = !!options.interpolate;

      /**
       * @protected
       * @type {function(import("../View.js").ViewOptions):void}
       */
      this.viewResolver = null;

      /**
       * @protected
       * @type {function(Error):void}
       */
      this.viewRejector = null;

      const self = this;
      /**
       * @private
       * @type {Promise<import("../View.js").ViewOptions>}
       */
      this.viewPromise_ = new Promise(function (resolve, reject) {
        self.viewResolver = resolve;
        self.viewRejector = reject;
      });
    }

    /**
     * Get the attribution function for the source.
     * @return {?Attribution} Attribution function.
     * @api
     */
    getAttributions() {
      return this.attributions_;
    }

    /**
     * @return {boolean} Attributions are collapsible.
     * @api
     */
    getAttributionsCollapsible() {
      return this.attributionsCollapsible_;
    }

    /**
     * Get the projection of the source.
     * @return {import("../proj/Projection.js").default|null} Projection.
     * @api
     */
    getProjection() {
      return this.projection;
    }

    /**
     * @abstract
     * @return {Array<number>|null} Resolutions.
     */
    getResolutions() {
      return abstract();
    }

    /**
     * @return {Promise<import("../View.js").ViewOptions>} A promise for view-related properties.
     */
    getView() {
      return this.viewPromise_;
    }

    /**
     * Get the state of the source, see {@link import("./Source.js").State} for possible states.
     * @return {import("./Source.js").State} State.
     * @api
     */
    getState() {
      return this.state_;
    }

    /**
     * @return {boolean|undefined} Wrap X.
     */
    getWrapX() {
      return this.wrapX_;
    }

    /**
     * @return {boolean} Use linear interpolation when resampling.
     */
    getInterpolate() {
      return this.interpolate_;
    }

    /**
     * Refreshes the source. The source will be cleared, and data from the server will be reloaded.
     * @api
     */
    refresh() {
      this.changed();
    }

    /**
     * Set the attributions of the source.
     * @param {AttributionLike|undefined} attributions Attributions.
     *     Can be passed as `string`, `Array<string>`, {@link module:ol/source/Source~Attribution},
     *     or `undefined`.
     * @api
     */
    setAttributions(attributions) {
      this.attributions_ = adaptAttributions(attributions);
      this.changed();
    }

    /**
     * Set the state of the source.
     * @param {import("./Source.js").State} state State.
     */
    setState(state) {
      this.state_ = state;
      this.changed();
    }
  }

  /**
   * Turns the attributions option into an attributions function.
   * @param {AttributionLike|undefined} attributionLike The attribution option.
   * @return {Attribution|null} An attribution function (or null).
   */
  function adaptAttributions(attributionLike) {
    if (!attributionLike) {
      return null;
    }
    if (Array.isArray(attributionLike)) {
      return function (frameState) {
        return attributionLike;
      };
    }

    if (typeof attributionLike === 'function') {
      return attributionLike;
    }

    return function (frameState) {
      return [attributionLike];
    };
  }

  /**
   * @module ol/tilegrid/TileGrid
   */

  /**
   * @private
   * @type {import("../tilecoord.js").TileCoord}
   */
  const tmpTileCoord = [0, 0, 0];

  /**
   * Number of decimal digits to consider in integer values when rounding.
   * @type {number}
   */
  const DECIMALS$1 = 5;

  /**
   * @typedef {Object} Options
   * @property {import("../extent.js").Extent} [extent] Extent for the tile grid. No tiles outside this
   * extent will be requested by {@link module:ol/source/Tile~TileSource} sources. When no `origin` or
   * `origins` are configured, the `origin` will be set to the top-left corner of the extent.
   * @property {number} [minZoom=0] Minimum zoom.
   * @property {import("../coordinate.js").Coordinate} [origin] The tile grid origin, i.e. where the `x`
   * and `y` axes meet (`[z, 0, 0]`). Tile coordinates increase left to right and downwards. If not
   * specified, `extent` or `origins` must be provided.
   * @property {Array<import("../coordinate.js").Coordinate>} [origins] Tile grid origins, i.e. where
   * the `x` and `y` axes meet (`[z, 0, 0]`), for each zoom level. If given, the array length
   * should match the length of the `resolutions` array, i.e. each resolution can have a different
   * origin. Tile coordinates increase left to right and downwards. If not specified, `extent` or
   * `origin` must be provided.
   * @property {!Array<number>} resolutions Resolutions. The array index of each resolution needs
   * to match the zoom level. This means that even if a `minZoom` is configured, the resolutions
   * array will have a length of `maxZoom + 1`.
   * @property {Array<import("../size.js").Size>} [sizes] Number of tile rows and columns
   * of the grid for each zoom level. If specified the values
   * define each zoom level's extent together with the `origin` or `origins`.
   * A grid `extent` can be configured in addition, and will further limit the extent
   * for which tile requests are made by sources. If the bottom-left corner of
   * an extent is used as `origin` or `origins`, then the `y` value must be
   * negative because OpenLayers tile coordinates use the top left as the origin.
   * @property {number|import("../size.js").Size} [tileSize] Tile size.
   * Default is `[256, 256]`.
   * @property {Array<number|import("../size.js").Size>} [tileSizes] Tile sizes. If given, the array length
   * should match the length of the `resolutions` array, i.e. each resolution can have a different
   * tile size.
   */

  /**
   * @classdesc
   * Base class for setting the grid pattern for sources accessing tiled-image
   * servers.
   * @api
   */
  class TileGrid {
    /**
     * @param {Options} options Tile grid options.
     */
    constructor(options) {
      /**
       * @protected
       * @type {number}
       */
      this.minZoom = options.minZoom !== undefined ? options.minZoom : 0;

      /**
       * @private
       * @type {!Array<number>}
       */
      this.resolutions_ = options.resolutions;
      assert(
        isSorted(
          this.resolutions_,
          function (a, b) {
            return b - a;
          },
          true
        ),
        17
      ); // `resolutions` must be sorted in descending order

      // check if we've got a consistent zoom factor and origin
      let zoomFactor;
      if (!options.origins) {
        for (let i = 0, ii = this.resolutions_.length - 1; i < ii; ++i) {
          if (!zoomFactor) {
            zoomFactor = this.resolutions_[i] / this.resolutions_[i + 1];
          } else {
            if (this.resolutions_[i] / this.resolutions_[i + 1] !== zoomFactor) {
              zoomFactor = undefined;
              break;
            }
          }
        }
      }

      /**
       * @private
       * @type {number|undefined}
       */
      this.zoomFactor_ = zoomFactor;

      /**
       * @protected
       * @type {number}
       */
      this.maxZoom = this.resolutions_.length - 1;

      /**
       * @private
       * @type {import("../coordinate.js").Coordinate|null}
       */
      this.origin_ = options.origin !== undefined ? options.origin : null;

      /**
       * @private
       * @type {Array<import("../coordinate.js").Coordinate>}
       */
      this.origins_ = null;
      if (options.origins !== undefined) {
        this.origins_ = options.origins;
        assert(this.origins_.length == this.resolutions_.length, 20); // Number of `origins` and `resolutions` must be equal
      }

      const extent = options.extent;

      if (extent !== undefined && !this.origin_ && !this.origins_) {
        this.origin_ = getTopLeft(extent);
      }

      assert(
        (!this.origin_ && this.origins_) || (this.origin_ && !this.origins_),
        18
      ); // Either `origin` or `origins` must be configured, never both

      /**
       * @private
       * @type {Array<number|import("../size.js").Size>}
       */
      this.tileSizes_ = null;
      if (options.tileSizes !== undefined) {
        this.tileSizes_ = options.tileSizes;
        assert(this.tileSizes_.length == this.resolutions_.length, 19); // Number of `tileSizes` and `resolutions` must be equal
      }

      /**
       * @private
       * @type {number|import("../size.js").Size}
       */
      this.tileSize_ =
        options.tileSize !== undefined
          ? options.tileSize
          : !this.tileSizes_
          ? DEFAULT_TILE_SIZE
          : null;
      assert(
        (!this.tileSize_ && this.tileSizes_) ||
          (this.tileSize_ && !this.tileSizes_),
        22
      ); // Either `tileSize` or `tileSizes` must be configured, never both

      /**
       * @private
       * @type {import("../extent.js").Extent}
       */
      this.extent_ = extent !== undefined ? extent : null;

      /**
       * @private
       * @type {Array<import("../TileRange.js").default>}
       */
      this.fullTileRanges_ = null;

      /**
       * @private
       * @type {import("../size.js").Size}
       */
      this.tmpSize_ = [0, 0];

      /**
       * @private
       * @type {import("../extent.js").Extent}
       */
      this.tmpExtent_ = [0, 0, 0, 0];

      if (options.sizes !== undefined) {
        this.fullTileRanges_ = options.sizes.map(function (size, z) {
          const tileRange = new TileRange(
            Math.min(0, size[0]),
            Math.max(size[0] - 1, -1),
            Math.min(0, size[1]),
            Math.max(size[1] - 1, -1)
          );
          if (extent) {
            const restrictedTileRange = this.getTileRangeForExtentAndZ(extent, z);
            tileRange.minX = Math.max(restrictedTileRange.minX, tileRange.minX);
            tileRange.maxX = Math.min(restrictedTileRange.maxX, tileRange.maxX);
            tileRange.minY = Math.max(restrictedTileRange.minY, tileRange.minY);
            tileRange.maxY = Math.min(restrictedTileRange.maxY, tileRange.maxY);
          }
          return tileRange;
        }, this);
      } else if (extent) {
        this.calculateTileRanges_(extent);
      }
    }

    /**
     * Call a function with each tile coordinate for a given extent and zoom level.
     *
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {number} zoom Integer zoom level.
     * @param {function(import("../tilecoord.js").TileCoord): void} callback Function called with each tile coordinate.
     * @api
     */
    forEachTileCoord(extent, zoom, callback) {
      const tileRange = this.getTileRangeForExtentAndZ(extent, zoom);
      for (let i = tileRange.minX, ii = tileRange.maxX; i <= ii; ++i) {
        for (let j = tileRange.minY, jj = tileRange.maxY; j <= jj; ++j) {
          callback([zoom, i, j]);
        }
      }
    }

    /**
     * @param {import("../tilecoord.js").TileCoord} tileCoord Tile coordinate.
     * @param {function(number, import("../TileRange.js").default): boolean} callback Callback.
     * @param {import("../TileRange.js").default} [tempTileRange] Temporary import("../TileRange.js").default object.
     * @param {import("../extent.js").Extent} [tempExtent] Temporary import("../extent.js").Extent object.
     * @return {boolean} Callback succeeded.
     */
    forEachTileCoordParentTileRange(
      tileCoord,
      callback,
      tempTileRange,
      tempExtent
    ) {
      let tileRange, x, y;
      let tileCoordExtent = null;
      let z = tileCoord[0] - 1;
      if (this.zoomFactor_ === 2) {
        x = tileCoord[1];
        y = tileCoord[2];
      } else {
        tileCoordExtent = this.getTileCoordExtent(tileCoord, tempExtent);
      }
      while (z >= this.minZoom) {
        if (this.zoomFactor_ === 2) {
          x = Math.floor(x / 2);
          y = Math.floor(y / 2);
          tileRange = createOrUpdate(x, x, y, y, tempTileRange);
        } else {
          tileRange = this.getTileRangeForExtentAndZ(
            tileCoordExtent,
            z,
            tempTileRange
          );
        }
        if (callback(z, tileRange)) {
          return true;
        }
        --z;
      }
      return false;
    }

    /**
     * Get the extent for this tile grid, if it was configured.
     * @return {import("../extent.js").Extent} Extent.
     * @api
     */
    getExtent() {
      return this.extent_;
    }

    /**
     * Get the maximum zoom level for the grid.
     * @return {number} Max zoom.
     * @api
     */
    getMaxZoom() {
      return this.maxZoom;
    }

    /**
     * Get the minimum zoom level for the grid.
     * @return {number} Min zoom.
     * @api
     */
    getMinZoom() {
      return this.minZoom;
    }

    /**
     * Get the origin for the grid at the given zoom level.
     * @param {number} z Integer zoom level.
     * @return {import("../coordinate.js").Coordinate} Origin.
     * @api
     */
    getOrigin(z) {
      if (this.origin_) {
        return this.origin_;
      } else {
        return this.origins_[z];
      }
    }

    /**
     * Get the resolution for the given zoom level.
     * @param {number} z Integer zoom level.
     * @return {number} Resolution.
     * @api
     */
    getResolution(z) {
      return this.resolutions_[z];
    }

    /**
     * Get the list of resolutions for the tile grid.
     * @return {Array<number>} Resolutions.
     * @api
     */
    getResolutions() {
      return this.resolutions_;
    }

    /**
     * @param {import("../tilecoord.js").TileCoord} tileCoord Tile coordinate.
     * @param {import("../TileRange.js").default} [tempTileRange] Temporary import("../TileRange.js").default object.
     * @param {import("../extent.js").Extent} [tempExtent] Temporary import("../extent.js").Extent object.
     * @return {import("../TileRange.js").default|null} Tile range.
     */
    getTileCoordChildTileRange(tileCoord, tempTileRange, tempExtent) {
      if (tileCoord[0] < this.maxZoom) {
        if (this.zoomFactor_ === 2) {
          const minX = tileCoord[1] * 2;
          const minY = tileCoord[2] * 2;
          return createOrUpdate(
            minX,
            minX + 1,
            minY,
            minY + 1,
            tempTileRange
          );
        }
        const tileCoordExtent = this.getTileCoordExtent(
          tileCoord,
          tempExtent || this.tmpExtent_
        );
        return this.getTileRangeForExtentAndZ(
          tileCoordExtent,
          tileCoord[0] + 1,
          tempTileRange
        );
      }
      return null;
    }

    /**
     * @param {import("../tilecoord.js").TileCoord} tileCoord Tile coordinate.
     * @param {number} z Integer zoom level.
     * @param {import("../TileRange.js").default} [tempTileRange] Temporary import("../TileRange.js").default object.
     * @return {import("../TileRange.js").default|null} Tile range.
     */
    getTileRangeForTileCoordAndZ(tileCoord, z, tempTileRange) {
      if (z > this.maxZoom || z < this.minZoom) {
        return null;
      }

      const tileCoordZ = tileCoord[0];
      const tileCoordX = tileCoord[1];
      const tileCoordY = tileCoord[2];

      if (z === tileCoordZ) {
        return createOrUpdate(
          tileCoordX,
          tileCoordY,
          tileCoordX,
          tileCoordY,
          tempTileRange
        );
      }

      if (this.zoomFactor_) {
        const factor = Math.pow(this.zoomFactor_, z - tileCoordZ);
        const minX = Math.floor(tileCoordX * factor);
        const minY = Math.floor(tileCoordY * factor);
        if (z < tileCoordZ) {
          return createOrUpdate(minX, minX, minY, minY, tempTileRange);
        }

        const maxX = Math.floor(factor * (tileCoordX + 1)) - 1;
        const maxY = Math.floor(factor * (tileCoordY + 1)) - 1;
        return createOrUpdate(minX, maxX, minY, maxY, tempTileRange);
      }

      const tileCoordExtent = this.getTileCoordExtent(tileCoord, this.tmpExtent_);
      return this.getTileRangeForExtentAndZ(tileCoordExtent, z, tempTileRange);
    }

    /**
     * Get the extent for a tile range.
     * @param {number} z Integer zoom level.
     * @param {import("../TileRange.js").default} tileRange Tile range.
     * @param {import("../extent.js").Extent} [tempExtent] Temporary import("../extent.js").Extent object.
     * @return {import("../extent.js").Extent} Extent.
     */
    getTileRangeExtent(z, tileRange, tempExtent) {
      const origin = this.getOrigin(z);
      const resolution = this.getResolution(z);
      const tileSize = toSize(this.getTileSize(z), this.tmpSize_);
      const minX = origin[0] + tileRange.minX * tileSize[0] * resolution;
      const maxX = origin[0] + (tileRange.maxX + 1) * tileSize[0] * resolution;
      const minY = origin[1] + tileRange.minY * tileSize[1] * resolution;
      const maxY = origin[1] + (tileRange.maxY + 1) * tileSize[1] * resolution;
      return createOrUpdate$2(minX, minY, maxX, maxY, tempExtent);
    }

    /**
     * Get a tile range for the given extent and integer zoom level.
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {number} z Integer zoom level.
     * @param {import("../TileRange.js").default} [tempTileRange] Temporary tile range object.
     * @return {import("../TileRange.js").default} Tile range.
     */
    getTileRangeForExtentAndZ(extent, z, tempTileRange) {
      const tileCoord = tmpTileCoord;
      this.getTileCoordForXYAndZ_(extent[0], extent[3], z, false, tileCoord);
      const minX = tileCoord[1];
      const minY = tileCoord[2];
      this.getTileCoordForXYAndZ_(extent[2], extent[1], z, true, tileCoord);
      return createOrUpdate(
        minX,
        tileCoord[1],
        minY,
        tileCoord[2],
        tempTileRange
      );
    }

    /**
     * @param {import("../tilecoord.js").TileCoord} tileCoord Tile coordinate.
     * @return {import("../coordinate.js").Coordinate} Tile center.
     */
    getTileCoordCenter(tileCoord) {
      const origin = this.getOrigin(tileCoord[0]);
      const resolution = this.getResolution(tileCoord[0]);
      const tileSize = toSize(this.getTileSize(tileCoord[0]), this.tmpSize_);
      return [
        origin[0] + (tileCoord[1] + 0.5) * tileSize[0] * resolution,
        origin[1] - (tileCoord[2] + 0.5) * tileSize[1] * resolution,
      ];
    }

    /**
     * Get the extent of a tile coordinate.
     *
     * @param {import("../tilecoord.js").TileCoord} tileCoord Tile coordinate.
     * @param {import("../extent.js").Extent} [tempExtent] Temporary extent object.
     * @return {import("../extent.js").Extent} Extent.
     * @api
     */
    getTileCoordExtent(tileCoord, tempExtent) {
      const origin = this.getOrigin(tileCoord[0]);
      const resolution = this.getResolution(tileCoord[0]);
      const tileSize = toSize(this.getTileSize(tileCoord[0]), this.tmpSize_);
      const minX = origin[0] + tileCoord[1] * tileSize[0] * resolution;
      const minY = origin[1] - (tileCoord[2] + 1) * tileSize[1] * resolution;
      const maxX = minX + tileSize[0] * resolution;
      const maxY = minY + tileSize[1] * resolution;
      return createOrUpdate$2(minX, minY, maxX, maxY, tempExtent);
    }

    /**
     * Get the tile coordinate for the given map coordinate and resolution.  This
     * method considers that coordinates that intersect tile boundaries should be
     * assigned the higher tile coordinate.
     *
     * @param {import("../coordinate.js").Coordinate} coordinate Coordinate.
     * @param {number} resolution Resolution.
     * @param {import("../tilecoord.js").TileCoord} [opt_tileCoord] Destination import("../tilecoord.js").TileCoord object.
     * @return {import("../tilecoord.js").TileCoord} Tile coordinate.
     * @api
     */
    getTileCoordForCoordAndResolution(coordinate, resolution, opt_tileCoord) {
      return this.getTileCoordForXYAndResolution_(
        coordinate[0],
        coordinate[1],
        resolution,
        false,
        opt_tileCoord
      );
    }

    /**
     * Note that this method should not be called for resolutions that correspond
     * to an integer zoom level.  Instead call the `getTileCoordForXYAndZ_` method.
     * @param {number} x X.
     * @param {number} y Y.
     * @param {number} resolution Resolution (for a non-integer zoom level).
     * @param {boolean} reverseIntersectionPolicy Instead of letting edge
     *     intersections go to the higher tile coordinate, let edge intersections
     *     go to the lower tile coordinate.
     * @param {import("../tilecoord.js").TileCoord} [opt_tileCoord] Temporary import("../tilecoord.js").TileCoord object.
     * @return {import("../tilecoord.js").TileCoord} Tile coordinate.
     * @private
     */
    getTileCoordForXYAndResolution_(
      x,
      y,
      resolution,
      reverseIntersectionPolicy,
      opt_tileCoord
    ) {
      const z = this.getZForResolution(resolution);
      const scale = resolution / this.getResolution(z);
      const origin = this.getOrigin(z);
      const tileSize = toSize(this.getTileSize(z), this.tmpSize_);

      let tileCoordX = (scale * (x - origin[0])) / resolution / tileSize[0];
      let tileCoordY = (scale * (origin[1] - y)) / resolution / tileSize[1];

      if (reverseIntersectionPolicy) {
        tileCoordX = ceil(tileCoordX, DECIMALS$1) - 1;
        tileCoordY = ceil(tileCoordY, DECIMALS$1) - 1;
      } else {
        tileCoordX = floor(tileCoordX, DECIMALS$1);
        tileCoordY = floor(tileCoordY, DECIMALS$1);
      }

      return createOrUpdate$1(z, tileCoordX, tileCoordY, opt_tileCoord);
    }

    /**
     * Although there is repetition between this method and `getTileCoordForXYAndResolution_`,
     * they should have separate implementations.  This method is for integer zoom
     * levels.  The other method should only be called for resolutions corresponding
     * to non-integer zoom levels.
     * @param {number} x Map x coordinate.
     * @param {number} y Map y coordinate.
     * @param {number} z Integer zoom level.
     * @param {boolean} reverseIntersectionPolicy Instead of letting edge
     *     intersections go to the higher tile coordinate, let edge intersections
     *     go to the lower tile coordinate.
     * @param {import("../tilecoord.js").TileCoord} [opt_tileCoord] Temporary import("../tilecoord.js").TileCoord object.
     * @return {import("../tilecoord.js").TileCoord} Tile coordinate.
     * @private
     */
    getTileCoordForXYAndZ_(x, y, z, reverseIntersectionPolicy, opt_tileCoord) {
      const origin = this.getOrigin(z);
      const resolution = this.getResolution(z);
      const tileSize = toSize(this.getTileSize(z), this.tmpSize_);

      let tileCoordX = (x - origin[0]) / resolution / tileSize[0];
      let tileCoordY = (origin[1] - y) / resolution / tileSize[1];

      if (reverseIntersectionPolicy) {
        tileCoordX = ceil(tileCoordX, DECIMALS$1) - 1;
        tileCoordY = ceil(tileCoordY, DECIMALS$1) - 1;
      } else {
        tileCoordX = floor(tileCoordX, DECIMALS$1);
        tileCoordY = floor(tileCoordY, DECIMALS$1);
      }

      return createOrUpdate$1(z, tileCoordX, tileCoordY, opt_tileCoord);
    }

    /**
     * Get a tile coordinate given a map coordinate and zoom level.
     * @param {import("../coordinate.js").Coordinate} coordinate Coordinate.
     * @param {number} z Zoom level.
     * @param {import("../tilecoord.js").TileCoord} [opt_tileCoord] Destination import("../tilecoord.js").TileCoord object.
     * @return {import("../tilecoord.js").TileCoord} Tile coordinate.
     * @api
     */
    getTileCoordForCoordAndZ(coordinate, z, opt_tileCoord) {
      return this.getTileCoordForXYAndZ_(
        coordinate[0],
        coordinate[1],
        z,
        false,
        opt_tileCoord
      );
    }

    /**
     * @param {import("../tilecoord.js").TileCoord} tileCoord Tile coordinate.
     * @return {number} Tile resolution.
     */
    getTileCoordResolution(tileCoord) {
      return this.resolutions_[tileCoord[0]];
    }

    /**
     * Get the tile size for a zoom level. The type of the return value matches the
     * `tileSize` or `tileSizes` that the tile grid was configured with. To always
     * get an {@link import("../size.js").Size}, run the result through {@link module:ol/size.toSize}.
     * @param {number} z Z.
     * @return {number|import("../size.js").Size} Tile size.
     * @api
     */
    getTileSize(z) {
      if (this.tileSize_) {
        return this.tileSize_;
      } else {
        return this.tileSizes_[z];
      }
    }

    /**
     * @param {number} z Zoom level.
     * @return {import("../TileRange.js").default} Extent tile range for the specified zoom level.
     */
    getFullTileRange(z) {
      if (!this.fullTileRanges_) {
        return this.extent_
          ? this.getTileRangeForExtentAndZ(this.extent_, z)
          : null;
      } else {
        return this.fullTileRanges_[z];
      }
    }

    /**
     * @param {number} resolution Resolution.
     * @param {number|import("../array.js").NearestDirectionFunction} [opt_direction]
     *     If 0, the nearest resolution will be used.
     *     If 1, the nearest higher resolution (lower Z) will be used. If -1, the
     *     nearest lower resolution (higher Z) will be used. Default is 0.
     *     Use a {@link module:ol/array~NearestDirectionFunction} for more precise control.
     *
     * For example to change tile Z at the midpoint of zoom levels
     * ```js
     * function(value, high, low) {
     *   return value - low * Math.sqrt(high / low);
     * }
     * ```
     * @return {number} Z.
     * @api
     */
    getZForResolution(resolution, opt_direction) {
      const z = linearFindNearest(
        this.resolutions_,
        resolution,
        opt_direction || 0
      );
      return clamp(z, this.minZoom, this.maxZoom);
    }

    /**
     * The tile with the provided tile coordinate intersects the given viewport.
     * @param {import('../tilecoord.js').TileCoord} tileCoord Tile coordinate.
     * @param {Array<number>} viewport Viewport as returned from {@link module:ol/extent.getRotatedViewport}.
     * @return {boolean} The tile with the provided tile coordinate intersects the given viewport.
     */
    tileCoordIntersectsViewport(tileCoord, viewport) {
      return intersectsLinearRing(
        viewport,
        0,
        viewport.length,
        2,
        this.getTileCoordExtent(tileCoord)
      );
    }

    /**
     * @param {!import("../extent.js").Extent} extent Extent for this tile grid.
     * @private
     */
    calculateTileRanges_(extent) {
      const length = this.resolutions_.length;
      const fullTileRanges = new Array(length);
      for (let z = this.minZoom; z < length; ++z) {
        fullTileRanges[z] = this.getTileRangeForExtentAndZ(extent, z);
      }
      this.fullTileRanges_ = fullTileRanges;
    }
  }

  /**
   * @module ol/tilegrid
   */

  /**
   * @param {import("./proj/Projection.js").default} projection Projection.
   * @return {!TileGrid} Default tile grid for the
   * passed projection.
   */
  function getForProjection(projection) {
    let tileGrid = projection.getDefaultTileGrid();
    if (!tileGrid) {
      tileGrid = createForProjection(projection);
      projection.setDefaultTileGrid(tileGrid);
    }
    return tileGrid;
  }

  /**
   * @param {TileGrid} tileGrid Tile grid.
   * @param {import("./tilecoord.js").TileCoord} tileCoord Tile coordinate.
   * @param {import("./proj/Projection.js").default} projection Projection.
   * @return {import("./tilecoord.js").TileCoord} Tile coordinate.
   */
  function wrapX(tileGrid, tileCoord, projection) {
    const z = tileCoord[0];
    const center = tileGrid.getTileCoordCenter(tileCoord);
    const projectionExtent = extentFromProjection(projection);
    if (!containsCoordinate(projectionExtent, center)) {
      const worldWidth = getWidth(projectionExtent);
      const worldsAway = Math.ceil(
        (projectionExtent[0] - center[0]) / worldWidth
      );
      center[0] += worldWidth * worldsAway;
      return tileGrid.getTileCoordForCoordAndZ(center, z);
    } else {
      return tileCoord;
    }
  }

  /**
   * @param {import("./extent.js").Extent} extent Extent.
   * @param {number} [maxZoom] Maximum zoom level (default is
   *     DEFAULT_MAX_ZOOM).
   * @param {number|import("./size.js").Size} [tileSize] Tile size (default uses
   *     DEFAULT_TILE_SIZE).
   * @param {import("./extent.js").Corner} [corner] Extent corner (default is `'top-left'`).
   * @return {!TileGrid} TileGrid instance.
   */
  function createForExtent(extent, maxZoom, tileSize, corner) {
    corner = corner !== undefined ? corner : 'top-left';

    const resolutions = resolutionsFromExtent(extent, maxZoom, tileSize);

    return new TileGrid({
      extent: extent,
      origin: getCorner(extent, corner),
      resolutions: resolutions,
      tileSize: tileSize,
    });
  }

  /**
   * @typedef {Object} XYZOptions
   * @property {import("./extent.js").Extent} [extent] Extent for the tile grid. The origin for an XYZ tile grid is the
   * top-left corner of the extent. If `maxResolution` is not provided the zero level of the grid is defined by the resolution
   * at which one tile fits in the provided extent. If not provided, the extent of the EPSG:3857 projection is used.
   * @property {number} [maxResolution] Resolution at level zero.
   * @property {number} [maxZoom] Maximum zoom. The default is `42`. This determines the number of levels
   * in the grid set. For example, a `maxZoom` of 21 means there are 22 levels in the grid set.
   * @property {number} [minZoom=0] Minimum zoom.
   * @property {number|import("./size.js").Size} [tileSize=[256, 256]] Tile size in pixels.
   */

  /**
   * Creates a tile grid with a standard XYZ tiling scheme.
   * @param {XYZOptions} [options] Tile grid options.
   * @return {!TileGrid} Tile grid instance.
   * @api
   */
  function createXYZ(options) {
    const xyzOptions = options || {};

    const extent = xyzOptions.extent || get$2('EPSG:3857').getExtent();

    const gridOptions = {
      extent: extent,
      minZoom: xyzOptions.minZoom,
      tileSize: xyzOptions.tileSize,
      resolutions: resolutionsFromExtent(
        extent,
        xyzOptions.maxZoom,
        xyzOptions.tileSize,
        xyzOptions.maxResolution
      ),
    };
    return new TileGrid(gridOptions);
  }

  /**
   * Create a resolutions array from an extent.  A zoom factor of 2 is assumed.
   * @param {import("./extent.js").Extent} extent Extent.
   * @param {number} [maxZoom] Maximum zoom level (default is
   *     DEFAULT_MAX_ZOOM).
   * @param {number|import("./size.js").Size} [tileSize] Tile size (default uses
   *     DEFAULT_TILE_SIZE).
   * @param {number} [maxResolution] Resolution at level zero.
   * @return {!Array<number>} Resolutions array.
   */
  function resolutionsFromExtent(extent, maxZoom, tileSize, maxResolution) {
    maxZoom = maxZoom !== undefined ? maxZoom : DEFAULT_MAX_ZOOM;
    tileSize = toSize(tileSize !== undefined ? tileSize : DEFAULT_TILE_SIZE);

    const height = getHeight(extent);
    const width = getWidth(extent);

    maxResolution =
      maxResolution > 0
        ? maxResolution
        : Math.max(width / tileSize[0], height / tileSize[1]);

    const length = maxZoom + 1;
    const resolutions = new Array(length);
    for (let z = 0; z < length; ++z) {
      resolutions[z] = maxResolution / Math.pow(2, z);
    }
    return resolutions;
  }

  /**
   * @param {import("./proj.js").ProjectionLike} projection Projection.
   * @param {number} [maxZoom] Maximum zoom level (default is
   *     DEFAULT_MAX_ZOOM).
   * @param {number|import("./size.js").Size} [tileSize] Tile size (default uses
   *     DEFAULT_TILE_SIZE).
   * @param {import("./extent.js").Corner} [corner] Extent corner (default is `'top-left'`).
   * @return {!TileGrid} TileGrid instance.
   */
  function createForProjection(projection, maxZoom, tileSize, corner) {
    const extent = extentFromProjection(projection);
    return createForExtent(extent, maxZoom, tileSize, corner);
  }

  /**
   * Generate a tile grid extent from a projection.  If the projection has an
   * extent, it is used.  If not, a global extent is assumed.
   * @param {import("./proj.js").ProjectionLike} projection Projection.
   * @return {import("./extent.js").Extent} Extent.
   */
  function extentFromProjection(projection) {
    projection = get$2(projection);
    let extent = projection.getExtent();
    if (!extent) {
      const half =
        (180 * METERS_PER_UNIT$1.degrees) / projection.getMetersPerUnit();
      extent = createOrUpdate$2(-half, -half, half, half);
    }
    return extent;
  }

  /**
   * @module ol/source/Tile
   */

  /***
   * @template Return
   * @typedef {import("../Observable").OnSignature<import("../Observable").EventTypes, import("../events/Event.js").default, Return> &
   *   import("../Observable").OnSignature<import("../ObjectEventType").Types, import("../Object").ObjectEvent, Return> &
   *   import("../Observable").OnSignature<import("./TileEventType").TileSourceEventTypes, TileSourceEvent, Return> &
   *   import("../Observable").CombinedOnSignature<import("../Observable").EventTypes|import("../ObjectEventType").Types|
   *     import("./TileEventType").TileSourceEventTypes, Return>} TileSourceOnSignature
   */

  /**
   * @typedef {Object} Options
   * @property {import("./Source.js").AttributionLike} [attributions] Attributions.
   * @property {boolean} [attributionsCollapsible=true] Attributions are collapsible.
   * @property {number} [cacheSize] CacheSize.
   * @property {boolean} [opaque=false] Whether the layer is opaque.
   * @property {number} [tilePixelRatio] TilePixelRatio.
   * @property {import("../proj.js").ProjectionLike} [projection] Projection.
   * @property {import("./Source.js").State} [state] State.
   * @property {import("../tilegrid/TileGrid.js").default} [tileGrid] TileGrid.
   * @property {boolean} [wrapX=false] WrapX.
   * @property {number} [transition] Transition.
   * @property {string} [key] Key.
   * @property {number|import("../array.js").NearestDirectionFunction} [zDirection=0] ZDirection.
   * @property {boolean} [interpolate=false] Use interpolated values when resampling.  By default,
   * the nearest neighbor is used when resampling.
   */

  /**
   * @classdesc
   * Abstract base class; normally only used for creating subclasses and not
   * instantiated in apps.
   * Base class for sources providing images divided into a tile grid.
   * @abstract
   * @api
   */
  class TileSource extends Source {
    /**
     * @param {Options} options SourceTile source options.
     */
    constructor(options) {
      super({
        attributions: options.attributions,
        attributionsCollapsible: options.attributionsCollapsible,
        projection: options.projection,
        state: options.state,
        wrapX: options.wrapX,
        interpolate: options.interpolate,
      });

      /***
       * @type {TileSourceOnSignature<import("../events").EventsKey>}
       */
      this.on;

      /***
       * @type {TileSourceOnSignature<import("../events").EventsKey>}
       */
      this.once;

      /***
       * @type {TileSourceOnSignature<void>}
       */
      this.un;

      /**
       * @private
       * @type {boolean}
       */
      this.opaque_ = options.opaque !== undefined ? options.opaque : false;

      /**
       * @private
       * @type {number}
       */
      this.tilePixelRatio_ =
        options.tilePixelRatio !== undefined ? options.tilePixelRatio : 1;

      /**
       * @type {import("../tilegrid/TileGrid.js").default|null}
       */
      this.tileGrid = options.tileGrid !== undefined ? options.tileGrid : null;

      const tileSize = [256, 256];
      if (this.tileGrid) {
        toSize(this.tileGrid.getTileSize(this.tileGrid.getMinZoom()), tileSize);
      }

      /**
       * @protected
       * @type {import("../TileCache.js").default}
       */
      this.tileCache = new TileCache(options.cacheSize || 0);

      /**
       * @protected
       * @type {import("../size.js").Size}
       */
      this.tmpSize = [0, 0];

      /**
       * @private
       * @type {string}
       */
      this.key_ = options.key || '';

      /**
       * @protected
       * @type {import("../Tile.js").Options}
       */
      this.tileOptions = {
        transition: options.transition,
        interpolate: options.interpolate,
      };

      /**
       * zDirection hint, read by the renderer. Indicates which resolution should be used
       * by a renderer if the views resolution does not match any resolution of the tile source.
       * If 0, the nearest resolution will be used. If 1, the nearest lower resolution
       * will be used. If -1, the nearest higher resolution will be used.
       * @type {number|import("../array.js").NearestDirectionFunction}
       */
      this.zDirection = options.zDirection ? options.zDirection : 0;
    }

    /**
     * @return {boolean} Can expire cache.
     */
    canExpireCache() {
      return this.tileCache.canExpireCache();
    }

    /**
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @param {!Object<string, boolean>} usedTiles Used tiles.
     */
    expireCache(projection, usedTiles) {
      const tileCache = this.getTileCacheForProjection(projection);
      if (tileCache) {
        tileCache.expireCache(usedTiles);
      }
    }

    /**
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @param {number} z Zoom level.
     * @param {import("../TileRange.js").default} tileRange Tile range.
     * @param {function(import("../Tile.js").default):(boolean|void)} callback Called with each
     *     loaded tile.  If the callback returns `false`, the tile will not be
     *     considered loaded.
     * @return {boolean} The tile range is fully covered with loaded tiles.
     */
    forEachLoadedTile(projection, z, tileRange, callback) {
      const tileCache = this.getTileCacheForProjection(projection);
      if (!tileCache) {
        return false;
      }

      let covered = true;
      let tile, tileCoordKey, loaded;
      for (let x = tileRange.minX; x <= tileRange.maxX; ++x) {
        for (let y = tileRange.minY; y <= tileRange.maxY; ++y) {
          tileCoordKey = getKeyZXY(z, x, y);
          loaded = false;
          if (tileCache.containsKey(tileCoordKey)) {
            tile = /** @type {!import("../Tile.js").default} */ (
              tileCache.get(tileCoordKey)
            );
            loaded = tile.getState() === TileState.LOADED;
            if (loaded) {
              loaded = callback(tile) !== false;
            }
          }
          if (!loaded) {
            covered = false;
          }
        }
      }
      return covered;
    }

    /**
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @return {number} Gutter.
     */
    getGutterForProjection(projection) {
      return 0;
    }

    /**
     * Return the key to be used for all tiles in the source.
     * @return {string} The key for all tiles.
     */
    getKey() {
      return this.key_;
    }

    /**
     * Set the value to be used as the key for all tiles in the source.
     * @param {string} key The key for tiles.
     * @protected
     */
    setKey(key) {
      if (this.key_ !== key) {
        this.key_ = key;
        this.changed();
      }
    }

    /**
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @return {boolean} Opaque.
     */
    getOpaque(projection) {
      return this.opaque_;
    }

    /**
     * @return {Array<number>|null} Resolutions.
     */
    getResolutions() {
      if (!this.tileGrid) {
        return null;
      }
      return this.tileGrid.getResolutions();
    }

    /**
     * @abstract
     * @param {number} z Tile coordinate z.
     * @param {number} x Tile coordinate x.
     * @param {number} y Tile coordinate y.
     * @param {number} pixelRatio Pixel ratio.
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @return {!import("../Tile.js").default} Tile.
     */
    getTile(z, x, y, pixelRatio, projection) {
      return abstract();
    }

    /**
     * Return the tile grid of the tile source.
     * @return {import("../tilegrid/TileGrid.js").default|null} Tile grid.
     * @api
     */
    getTileGrid() {
      return this.tileGrid;
    }

    /**
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @return {!import("../tilegrid/TileGrid.js").default} Tile grid.
     */
    getTileGridForProjection(projection) {
      if (!this.tileGrid) {
        return getForProjection(projection);
      } else {
        return this.tileGrid;
      }
    }

    /**
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @return {import("../TileCache.js").default} Tile cache.
     * @protected
     */
    getTileCacheForProjection(projection) {
      const sourceProjection = this.getProjection();
      assert(
        sourceProjection === null || equivalent(sourceProjection, projection),
        68 // A VectorTile source can only be rendered if it has a projection compatible with the view projection.
      );
      return this.tileCache;
    }

    /**
     * Get the tile pixel ratio for this source. Subclasses may override this
     * method, which is meant to return a supported pixel ratio that matches the
     * provided `pixelRatio` as close as possible.
     * @param {number} pixelRatio Pixel ratio.
     * @return {number} Tile pixel ratio.
     */
    getTilePixelRatio(pixelRatio) {
      return this.tilePixelRatio_;
    }

    /**
     * @param {number} z Z.
     * @param {number} pixelRatio Pixel ratio.
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @return {import("../size.js").Size} Tile size.
     */
    getTilePixelSize(z, pixelRatio, projection) {
      const tileGrid = this.getTileGridForProjection(projection);
      const tilePixelRatio = this.getTilePixelRatio(pixelRatio);
      const tileSize = toSize(tileGrid.getTileSize(z), this.tmpSize);
      if (tilePixelRatio == 1) {
        return tileSize;
      } else {
        return scale(tileSize, tilePixelRatio, this.tmpSize);
      }
    }

    /**
     * Returns a tile coordinate wrapped around the x-axis. When the tile coordinate
     * is outside the resolution and extent range of the tile grid, `null` will be
     * returned.
     * @param {import("../tilecoord.js").TileCoord} tileCoord Tile coordinate.
     * @param {import("../proj/Projection.js").default} [projection] Projection.
     * @return {import("../tilecoord.js").TileCoord} Tile coordinate to be passed to the tileUrlFunction or
     *     null if no tile URL should be created for the passed `tileCoord`.
     */
    getTileCoordForTileUrlFunction(tileCoord, projection) {
      projection = projection !== undefined ? projection : this.getProjection();
      const tileGrid = this.getTileGridForProjection(projection);
      if (this.getWrapX() && projection.isGlobal()) {
        tileCoord = wrapX(tileGrid, tileCoord, projection);
      }
      return withinExtentAndZ(tileCoord, tileGrid) ? tileCoord : null;
    }

    /**
     * Remove all cached tiles from the source. The next render cycle will fetch new tiles.
     * @api
     */
    clear() {
      this.tileCache.clear();
    }

    refresh() {
      this.clear();
      super.refresh();
    }

    /**
     * Increases the cache size if needed
     * @param {number} tileCount Minimum number of tiles needed.
     * @param {import("../proj/Projection.js").default} projection Projection.
     */
    updateCacheSize(tileCount, projection) {
      const tileCache = this.getTileCacheForProjection(projection);
      if (tileCount > tileCache.highWaterMark) {
        tileCache.highWaterMark = tileCount;
      }
    }

    /**
     * Marks a tile coord as being used, without triggering a load.
     * @abstract
     * @param {number} z Tile coordinate z.
     * @param {number} x Tile coordinate x.
     * @param {number} y Tile coordinate y.
     * @param {import("../proj/Projection.js").default} projection Projection.
     */
    useTile(z, x, y, projection) {}
  }

  /**
   * @classdesc
   * Events emitted by {@link module:ol/source/Tile~TileSource} instances are instances of this
   * type.
   */
  class TileSourceEvent extends BaseEvent {
    /**
     * @param {string} type Type.
     * @param {import("../Tile.js").default} tile The tile.
     */
    constructor(type, tile) {
      super(type);

      /**
       * The tile related to the event.
       * @type {import("../Tile.js").default}
       * @api
       */
      this.tile = tile;
    }
  }

  /**
   * @module ol/tileurlfunction
   */

  /**
   * @param {string} template Template.
   * @param {import("./tilegrid/TileGrid.js").default} tileGrid Tile grid.
   * @return {import("./Tile.js").UrlFunction} Tile URL function.
   */
  function createFromTemplate(template, tileGrid) {
    const zRegEx = /\{z\}/g;
    const xRegEx = /\{x\}/g;
    const yRegEx = /\{y\}/g;
    const dashYRegEx = /\{-y\}/g;
    return (
      /**
       * @param {import("./tilecoord.js").TileCoord} tileCoord Tile Coordinate.
       * @param {number} pixelRatio Pixel ratio.
       * @param {import("./proj/Projection.js").default} projection Projection.
       * @return {string|undefined} Tile URL.
       */
      function (tileCoord, pixelRatio, projection) {
        if (!tileCoord) {
          return undefined;
        } else {
          return template
            .replace(zRegEx, tileCoord[0].toString())
            .replace(xRegEx, tileCoord[1].toString())
            .replace(yRegEx, tileCoord[2].toString())
            .replace(dashYRegEx, function () {
              const z = tileCoord[0];
              const range = tileGrid.getFullTileRange(z);
              assert(range, 55); // The {-y} placeholder requires a tile grid with extent
              const y = range.getHeight() - tileCoord[2] - 1;
              return y.toString();
            });
        }
      }
    );
  }

  /**
   * @param {Array<string>} templates Templates.
   * @param {import("./tilegrid/TileGrid.js").default} tileGrid Tile grid.
   * @return {import("./Tile.js").UrlFunction} Tile URL function.
   */
  function createFromTemplates(templates, tileGrid) {
    const len = templates.length;
    const tileUrlFunctions = new Array(len);
    for (let i = 0; i < len; ++i) {
      tileUrlFunctions[i] = createFromTemplate(templates[i], tileGrid);
    }
    return createFromTileUrlFunctions(tileUrlFunctions);
  }

  /**
   * @param {Array<import("./Tile.js").UrlFunction>} tileUrlFunctions Tile URL Functions.
   * @return {import("./Tile.js").UrlFunction} Tile URL function.
   */
  function createFromTileUrlFunctions(tileUrlFunctions) {
    if (tileUrlFunctions.length === 1) {
      return tileUrlFunctions[0];
    }
    return (
      /**
       * @param {import("./tilecoord.js").TileCoord} tileCoord Tile Coordinate.
       * @param {number} pixelRatio Pixel ratio.
       * @param {import("./proj/Projection.js").default} projection Projection.
       * @return {string|undefined} Tile URL.
       */
      function (tileCoord, pixelRatio, projection) {
        if (!tileCoord) {
          return undefined;
        } else {
          const h = hash(tileCoord);
          const index = modulo(h, tileUrlFunctions.length);
          return tileUrlFunctions[index](tileCoord, pixelRatio, projection);
        }
      }
    );
  }

  /**
   * @param {import("./tilecoord.js").TileCoord} tileCoord Tile coordinate.
   * @param {number} pixelRatio Pixel ratio.
   * @param {import("./proj/Projection.js").default} projection Projection.
   * @return {string|undefined} Tile URL.
   */
  function nullTileUrlFunction(tileCoord, pixelRatio, projection) {
    return undefined;
  }

  /**
   * @param {string} url URL.
   * @return {Array<string>} Array of urls.
   */
  function expandUrl(url) {
    const urls = [];
    let match = /\{([a-z])-([a-z])\}/.exec(url);
    if (match) {
      // char range
      const startCharCode = match[1].charCodeAt(0);
      const stopCharCode = match[2].charCodeAt(0);
      let charCode;
      for (charCode = startCharCode; charCode <= stopCharCode; ++charCode) {
        urls.push(url.replace(match[0], String.fromCharCode(charCode)));
      }
      return urls;
    }
    match = /\{(\d+)-(\d+)\}/.exec(url);
    if (match) {
      // number range
      const stop = parseInt(match[2], 10);
      for (let i = parseInt(match[1], 10); i <= stop; i++) {
        urls.push(url.replace(match[0], i.toString()));
      }
      return urls;
    }
    urls.push(url);
    return urls;
  }

  /**
   * @module ol/source/UrlTile
   */

  /**
   * @typedef {Object} Options
   * @property {import("./Source.js").AttributionLike} [attributions] Attributions.
   * @property {boolean} [attributionsCollapsible=true] Attributions are collapsible.
   * @property {number} [cacheSize] Cache size.
   * @property {boolean} [opaque=false] Whether the layer is opaque.
   * @property {import("../proj.js").ProjectionLike} [projection] Projection.
   * @property {import("./Source.js").State} [state] State.
   * @property {import("../tilegrid/TileGrid.js").default} [tileGrid] TileGrid.
   * @property {import("../Tile.js").LoadFunction} tileLoadFunction TileLoadFunction.
   * @property {number} [tilePixelRatio] TilePixelRatio.
   * @property {import("../Tile.js").UrlFunction} [tileUrlFunction] TileUrlFunction.
   * @property {string} [url] Url.
   * @property {Array<string>} [urls] Urls.
   * @property {boolean} [wrapX=true] WrapX.
   * @property {number} [transition] Transition.
   * @property {string} [key] Key.
   * @property {number|import("../array.js").NearestDirectionFunction} [zDirection=0] ZDirection.
   * @property {boolean} [interpolate=false] Use interpolated values when resampling.  By default,
   * the nearest neighbor is used when resampling.
   */

  /**
   * @classdesc
   * Base class for sources providing tiles divided into a tile grid over http.
   *
   * @fires import("./Tile.js").TileSourceEvent
   */
  class UrlTile extends TileSource {
    /**
     * @param {Options} options Image tile options.
     */
    constructor(options) {
      super({
        attributions: options.attributions,
        cacheSize: options.cacheSize,
        opaque: options.opaque,
        projection: options.projection,
        state: options.state,
        tileGrid: options.tileGrid,
        tilePixelRatio: options.tilePixelRatio,
        wrapX: options.wrapX,
        transition: options.transition,
        interpolate: options.interpolate,
        key: options.key,
        attributionsCollapsible: options.attributionsCollapsible,
        zDirection: options.zDirection,
      });

      /**
       * @private
       * @type {boolean}
       */
      this.generateTileUrlFunction_ =
        this.tileUrlFunction === UrlTile.prototype.tileUrlFunction;

      /**
       * @protected
       * @type {import("../Tile.js").LoadFunction}
       */
      this.tileLoadFunction = options.tileLoadFunction;

      if (options.tileUrlFunction) {
        this.tileUrlFunction = options.tileUrlFunction;
      }

      /**
       * @protected
       * @type {!Array<string>|null}
       */
      this.urls = null;

      if (options.urls) {
        this.setUrls(options.urls);
      } else if (options.url) {
        this.setUrl(options.url);
      }

      /**
       * @private
       * @type {!Object<string, boolean>}
       */
      this.tileLoadingKeys_ = {};
    }

    /**
     * Return the tile load function of the source.
     * @return {import("../Tile.js").LoadFunction} TileLoadFunction
     * @api
     */
    getTileLoadFunction() {
      return this.tileLoadFunction;
    }

    /**
     * Return the tile URL function of the source.
     * @return {import("../Tile.js").UrlFunction} TileUrlFunction
     * @api
     */
    getTileUrlFunction() {
      return Object.getPrototypeOf(this).tileUrlFunction === this.tileUrlFunction
        ? this.tileUrlFunction.bind(this)
        : this.tileUrlFunction;
    }

    /**
     * Return the URLs used for this source.
     * When a tileUrlFunction is used instead of url or urls,
     * null will be returned.
     * @return {!Array<string>|null} URLs.
     * @api
     */
    getUrls() {
      return this.urls;
    }

    /**
     * Handle tile change events.
     * @param {import("../events/Event.js").default} event Event.
     * @protected
     */
    handleTileChange(event) {
      const tile = /** @type {import("../Tile.js").default} */ (event.target);
      const uid = getUid(tile);
      const tileState = tile.getState();
      let type;
      if (tileState == TileState.LOADING) {
        this.tileLoadingKeys_[uid] = true;
        type = TileEventType.TILELOADSTART;
      } else if (uid in this.tileLoadingKeys_) {
        delete this.tileLoadingKeys_[uid];
        type =
          tileState == TileState.ERROR
            ? TileEventType.TILELOADERROR
            : tileState == TileState.LOADED
            ? TileEventType.TILELOADEND
            : undefined;
      }
      if (type != undefined) {
        this.dispatchEvent(new TileSourceEvent(type, tile));
      }
    }

    /**
     * Set the tile load function of the source.
     * @param {import("../Tile.js").LoadFunction} tileLoadFunction Tile load function.
     * @api
     */
    setTileLoadFunction(tileLoadFunction) {
      this.tileCache.clear();
      this.tileLoadFunction = tileLoadFunction;
      this.changed();
    }

    /**
     * Set the tile URL function of the source.
     * @param {import("../Tile.js").UrlFunction} tileUrlFunction Tile URL function.
     * @param {string} [key] Optional new tile key for the source.
     * @api
     */
    setTileUrlFunction(tileUrlFunction, key) {
      this.tileUrlFunction = tileUrlFunction;
      this.tileCache.pruneExceptNewestZ();
      if (typeof key !== 'undefined') {
        this.setKey(key);
      } else {
        this.changed();
      }
    }

    /**
     * Set the URL to use for requests.
     * @param {string} url URL.
     * @api
     */
    setUrl(url) {
      const urls = expandUrl(url);
      this.urls = urls;
      this.setUrls(urls);
    }

    /**
     * Set the URLs to use for requests.
     * @param {Array<string>} urls URLs.
     * @api
     */
    setUrls(urls) {
      this.urls = urls;
      const key = urls.join('\n');
      if (this.generateTileUrlFunction_) {
        this.setTileUrlFunction(createFromTemplates(urls, this.tileGrid), key);
      } else {
        this.setKey(key);
      }
    }

    /**
     * @param {import("../tilecoord.js").TileCoord} tileCoord Tile coordinate.
     * @param {number} pixelRatio Pixel ratio.
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @return {string|undefined} Tile URL.
     */
    tileUrlFunction(tileCoord, pixelRatio, projection) {
      return undefined;
    }

    /**
     * Marks a tile coord as being used, without triggering a load.
     * @param {number} z Tile coordinate z.
     * @param {number} x Tile coordinate x.
     * @param {number} y Tile coordinate y.
     */
    useTile(z, x, y) {
      const tileCoordKey = getKeyZXY(z, x, y);
      if (this.tileCache.containsKey(tileCoordKey)) {
        this.tileCache.get(tileCoordKey);
      }
    }
  }

  /**
   * @module ol/source/TileImage
   */

  /**
   * @typedef {Object} Options
   * @property {import("./Source.js").AttributionLike} [attributions] Attributions.
   * @property {boolean} [attributionsCollapsible=true] Attributions are collapsible.
   * @property {number} [cacheSize] Initial tile cache size. Will auto-grow to hold at least the number of tiles in the viewport.
   * @property {null|string} [crossOrigin] The `crossOrigin` attribute for loaded images.  Note that
   * you must provide a `crossOrigin` value if you want to access pixel data with the Canvas renderer.
   * See https://developer.mozilla.org/en-US/docs/Web/HTML/CORS_enabled_image for more detail.
   * @property {boolean} [interpolate=true] Use interpolated values when resampling.  By default,
   * linear interpolation is used when resampling.  Set to false to use the nearest neighbor instead.
   * @property {boolean} [opaque=false] Whether the layer is opaque.
   * @property {import("../proj.js").ProjectionLike} [projection] Projection. Default is the view projection.
   * @property {number} [reprojectionErrorThreshold=0.5] Maximum allowed reprojection error (in pixels).
   * Higher values can increase reprojection performance, but decrease precision.
   * @property {import("./Source.js").State} [state] Source state.
   * @property {typeof import("../ImageTile.js").default} [tileClass] Class used to instantiate image tiles.
   * Default is {@link module:ol/ImageTile~ImageTile}.
   * @property {import("../tilegrid/TileGrid.js").default} [tileGrid] Tile grid.
   * @property {import("../Tile.js").LoadFunction} [tileLoadFunction] Optional function to load a tile given a URL. The default is
   * ```js
   * function(imageTile, src) {
   *   imageTile.getImage().src = src;
   * };
   * ```
   * @property {number} [tilePixelRatio=1] The pixel ratio used by the tile service. For example, if the tile
   * service advertizes 256px by 256px tiles but actually sends 512px
   * by 512px images (for retina/hidpi devices) then `tilePixelRatio`
   * should be set to `2`.
   * @property {import("../Tile.js").UrlFunction} [tileUrlFunction] Optional function to get tile URL given a tile coordinate and the projection.
   * @property {string} [url] URL template. Must include `{x}`, `{y}` or `{-y}`, and `{z}` placeholders.
   * A `{?-?}` template pattern, for example `subdomain{a-f}.domain.com`, may be
   * used instead of defining each one separately in the `urls` option.
   * @property {Array<string>} [urls] An array of URL templates.
   * @property {boolean} [wrapX] Whether to wrap the world horizontally. The default, is to
   * request out-of-bounds tiles from the server. When set to `false`, only one
   * world will be rendered. When set to `true`, tiles will be requested for one
   * world only, but they will be wrapped horizontally to render multiple worlds.
   * @property {number} [transition] Duration of the opacity transition for rendering.
   * To disable the opacity transition, pass `transition: 0`.
   * @property {string} [key] Optional tile key for proper cache fetching
   * @property {number|import("../array.js").NearestDirectionFunction} [zDirection=0]
   * Choose whether to use tiles with a higher or lower zoom level when between integer
   * zoom levels. See {@link module:ol/tilegrid/TileGrid~TileGrid#getZForResolution}.
   */

  /**
   * @classdesc
   * Base class for sources providing images divided into a tile grid.
   *
   * @fires import("./Tile.js").TileSourceEvent
   * @api
   */
  class TileImage extends UrlTile {
    /**
     * @param {!Options} options Image tile options.
     */
    constructor(options) {
      super({
        attributions: options.attributions,
        cacheSize: options.cacheSize,
        opaque: options.opaque,
        projection: options.projection,
        state: options.state,
        tileGrid: options.tileGrid,
        tileLoadFunction: options.tileLoadFunction
          ? options.tileLoadFunction
          : defaultTileLoadFunction,
        tilePixelRatio: options.tilePixelRatio,
        tileUrlFunction: options.tileUrlFunction,
        url: options.url,
        urls: options.urls,
        wrapX: options.wrapX,
        transition: options.transition,
        interpolate:
          options.interpolate !== undefined ? options.interpolate : true,
        key: options.key,
        attributionsCollapsible: options.attributionsCollapsible,
        zDirection: options.zDirection,
      });

      /**
       * @protected
       * @type {?string}
       */
      this.crossOrigin =
        options.crossOrigin !== undefined ? options.crossOrigin : null;

      /**
       * @protected
       * @type {typeof ImageTile}
       */
      this.tileClass =
        options.tileClass !== undefined ? options.tileClass : ImageTile;

      /**
       * @protected
       * @type {!Object<string, TileCache>}
       */
      this.tileCacheForProjection = {};

      /**
       * @protected
       * @type {!Object<string, import("../tilegrid/TileGrid.js").default>}
       */
      this.tileGridForProjection = {};

      /**
       * @private
       * @type {number|undefined}
       */
      this.reprojectionErrorThreshold_ = options.reprojectionErrorThreshold;

      /**
       * @private
       * @type {boolean}
       */
      this.renderReprojectionEdges_ = false;
    }

    /**
     * @return {boolean} Can expire cache.
     */
    canExpireCache() {
      if (this.tileCache.canExpireCache()) {
        return true;
      } else {
        for (const key in this.tileCacheForProjection) {
          if (this.tileCacheForProjection[key].canExpireCache()) {
            return true;
          }
        }
      }
      return false;
    }

    /**
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @param {!Object<string, boolean>} usedTiles Used tiles.
     */
    expireCache(projection, usedTiles) {
      const usedTileCache = this.getTileCacheForProjection(projection);

      this.tileCache.expireCache(
        this.tileCache == usedTileCache ? usedTiles : {}
      );
      for (const id in this.tileCacheForProjection) {
        const tileCache = this.tileCacheForProjection[id];
        tileCache.expireCache(tileCache == usedTileCache ? usedTiles : {});
      }
    }

    /**
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @return {number} Gutter.
     */
    getGutterForProjection(projection) {
      if (
        this.getProjection() &&
        projection &&
        !equivalent(this.getProjection(), projection)
      ) {
        return 0;
      } else {
        return this.getGutter();
      }
    }

    /**
     * @return {number} Gutter.
     */
    getGutter() {
      return 0;
    }

    /**
     * Return the key to be used for all tiles in the source.
     * @return {string} The key for all tiles.
     */
    getKey() {
      let key = super.getKey();
      if (!this.getInterpolate()) {
        key += ':disable-interpolation';
      }
      return key;
    }

    /**
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @return {boolean} Opaque.
     */
    getOpaque(projection) {
      if (
        this.getProjection() &&
        projection &&
        !equivalent(this.getProjection(), projection)
      ) {
        return false;
      } else {
        return super.getOpaque(projection);
      }
    }

    /**
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @return {!import("../tilegrid/TileGrid.js").default} Tile grid.
     */
    getTileGridForProjection(projection) {
      const thisProj = this.getProjection();
      if (this.tileGrid && (!thisProj || equivalent(thisProj, projection))) {
        return this.tileGrid;
      } else {
        const projKey = getUid(projection);
        if (!(projKey in this.tileGridForProjection)) {
          this.tileGridForProjection[projKey] =
            getForProjection(projection);
        }
        return this.tileGridForProjection[projKey];
      }
    }

    /**
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @return {import("../TileCache.js").default} Tile cache.
     */
    getTileCacheForProjection(projection) {
      const thisProj = this.getProjection();
      if (!thisProj || equivalent(thisProj, projection)) {
        return this.tileCache;
      } else {
        const projKey = getUid(projection);
        if (!(projKey in this.tileCacheForProjection)) {
          this.tileCacheForProjection[projKey] = new TileCache(
            this.tileCache.highWaterMark
          );
        }
        return this.tileCacheForProjection[projKey];
      }
    }

    /**
     * @param {number} z Tile coordinate z.
     * @param {number} x Tile coordinate x.
     * @param {number} y Tile coordinate y.
     * @param {number} pixelRatio Pixel ratio.
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @param {string} key The key set on the tile.
     * @return {!ImageTile} Tile.
     * @private
     */
    createTile_(z, x, y, pixelRatio, projection, key) {
      const tileCoord = [z, x, y];
      const urlTileCoord = this.getTileCoordForTileUrlFunction(
        tileCoord,
        projection
      );
      const tileUrl = urlTileCoord
        ? this.tileUrlFunction(urlTileCoord, pixelRatio, projection)
        : undefined;
      const tile = new this.tileClass(
        tileCoord,
        tileUrl !== undefined ? TileState.IDLE : TileState.EMPTY,
        tileUrl !== undefined ? tileUrl : '',
        this.crossOrigin,
        this.tileLoadFunction,
        this.tileOptions
      );
      tile.key = key;
      tile.addEventListener(EventType$1.CHANGE, this.handleTileChange.bind(this));
      return tile;
    }

    /**
     * @param {number} z Tile coordinate z.
     * @param {number} x Tile coordinate x.
     * @param {number} y Tile coordinate y.
     * @param {number} pixelRatio Pixel ratio.
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @return {!(ImageTile|ReprojTile)} Tile.
     */
    getTile(z, x, y, pixelRatio, projection) {
      const sourceProjection = this.getProjection();
      if (
        !sourceProjection ||
        !projection ||
        equivalent(sourceProjection, projection)
      ) {
        return this.getTileInternal(
          z,
          x,
          y,
          pixelRatio,
          sourceProjection || projection
        );
      } else {
        const cache = this.getTileCacheForProjection(projection);
        const tileCoord = [z, x, y];
        let tile;
        const tileCoordKey = getKey(tileCoord);
        if (cache.containsKey(tileCoordKey)) {
          tile = cache.get(tileCoordKey);
        }
        const key = this.getKey();
        if (tile && tile.key == key) {
          return tile;
        } else {
          const sourceTileGrid = this.getTileGridForProjection(sourceProjection);
          const targetTileGrid = this.getTileGridForProjection(projection);
          const wrappedTileCoord = this.getTileCoordForTileUrlFunction(
            tileCoord,
            projection
          );
          const newTile = new ReprojTile(
            sourceProjection,
            sourceTileGrid,
            projection,
            targetTileGrid,
            tileCoord,
            wrappedTileCoord,
            this.getTilePixelRatio(pixelRatio),
            this.getGutter(),
            function (z, x, y, pixelRatio) {
              return this.getTileInternal(z, x, y, pixelRatio, sourceProjection);
            }.bind(this),
            this.reprojectionErrorThreshold_,
            this.renderReprojectionEdges_,
            this.getInterpolate()
          );
          newTile.key = key;

          if (tile) {
            newTile.interimTile = tile;
            newTile.refreshInterimChain();
            cache.replace(tileCoordKey, newTile);
          } else {
            cache.set(tileCoordKey, newTile);
          }
          return newTile;
        }
      }
    }

    /**
     * @param {number} z Tile coordinate z.
     * @param {number} x Tile coordinate x.
     * @param {number} y Tile coordinate y.
     * @param {number} pixelRatio Pixel ratio.
     * @param {!import("../proj/Projection.js").default} projection Projection.
     * @return {!(ImageTile|ReprojTile)} Tile.
     * @protected
     */
    getTileInternal(z, x, y, pixelRatio, projection) {
      let tile = null;
      const tileCoordKey = getKeyZXY(z, x, y);
      const key = this.getKey();
      if (!this.tileCache.containsKey(tileCoordKey)) {
        tile = this.createTile_(z, x, y, pixelRatio, projection, key);
        this.tileCache.set(tileCoordKey, tile);
      } else {
        tile = this.tileCache.get(tileCoordKey);
        if (tile.key != key) {
          // The source's params changed. If the tile has an interim tile and if we
          // can use it then we use it. Otherwise we create a new tile.  In both
          // cases we attempt to assign an interim tile to the new tile.
          const interimTile = tile;
          tile = this.createTile_(z, x, y, pixelRatio, projection, key);

          //make the new tile the head of the list,
          if (interimTile.getState() == TileState.IDLE) {
            //the old tile hasn't begun loading yet, and is now outdated, so we can simply discard it
            tile.interimTile = interimTile.interimTile;
          } else {
            tile.interimTile = interimTile;
          }
          tile.refreshInterimChain();
          this.tileCache.replace(tileCoordKey, tile);
        }
      }
      return tile;
    }

    /**
     * Sets whether to render reprojection edges or not (usually for debugging).
     * @param {boolean} render Render the edges.
     * @api
     */
    setRenderReprojectionEdges(render) {
      if (this.renderReprojectionEdges_ == render) {
        return;
      }
      this.renderReprojectionEdges_ = render;
      for (const id in this.tileCacheForProjection) {
        this.tileCacheForProjection[id].clear();
      }
      this.changed();
    }

    /**
     * Sets the tile grid to use when reprojecting the tiles to the given
     * projection instead of the default tile grid for the projection.
     *
     * This can be useful when the default tile grid cannot be created
     * (e.g. projection has no extent defined) or
     * for optimization reasons (custom tile size, resolutions, ...).
     *
     * @param {import("../proj.js").ProjectionLike} projection Projection.
     * @param {import("../tilegrid/TileGrid.js").default} tilegrid Tile grid to use for the projection.
     * @api
     */
    setTileGridForProjection(projection, tilegrid) {
      const proj = get$2(projection);
      if (proj) {
        const projKey = getUid(proj);
        if (!(projKey in this.tileGridForProjection)) {
          this.tileGridForProjection[projKey] = tilegrid;
        }
      }
    }

    clear() {
      super.clear();
      for (const id in this.tileCacheForProjection) {
        this.tileCacheForProjection[id].clear();
      }
    }
  }

  /**
   * @param {ImageTile} imageTile Image tile.
   * @param {string} src Source.
   */
  function defaultTileLoadFunction(imageTile, src) {
    /** @type {HTMLImageElement|HTMLVideoElement} */ (imageTile.getImage()).src =
      src;
  }

  /**
   * @module ol/source/BingMaps
   */

  /**
   * @param {import('../tilecoord.js').TileCoord} tileCoord Tile coord.
   * @return {string} Quad key.
   */
  function quadKey(tileCoord) {
    const z = tileCoord[0];
    const digits = new Array(z);
    let mask = 1 << (z - 1);
    let i, charCode;
    for (i = 0; i < z; ++i) {
      // 48 is charCode for 0 - '0'.charCodeAt(0)
      charCode = 48;
      if (tileCoord[1] & mask) {
        charCode += 1;
      }
      if (tileCoord[2] & mask) {
        charCode += 2;
      }
      digits[i] = String.fromCharCode(charCode);
      mask >>= 1;
    }
    return digits.join('');
  }

  /**
   * The attribution containing a link to the Microsoft® Bing™ Maps Platform APIs’
   * Terms Of Use.
   * @const
   * @type {string}
   */
  const TOS_ATTRIBUTION =
    '<a class="ol-attribution-bing-tos" ' +
    'href="https://www.microsoft.com/maps/product/terms.html" target="_blank">' +
    'Terms of Use</a>';

  /**
   * @typedef {Object} Options
   * @property {number} [cacheSize] Initial tile cache size. Will auto-grow to hold at least the number of tiles in the viewport.
   * @property {boolean} [hidpi=false] If `true` hidpi tiles will be requested.
   * @property {string} [culture='en-us'] Culture code.
   * @property {string} key Bing Maps API key. Get yours at https://www.bingmapsportal.com/.
   * @property {string} imagerySet Type of imagery.
   * @property {boolean} [interpolate=true] Use interpolated values when resampling.  By default,
   * linear interpolation is used when resampling.  Set to false to use the nearest neighbor instead.
   * @property {number} [maxZoom=21] Max zoom. Default is what's advertized by the BingMaps service.
   * @property {number} [reprojectionErrorThreshold=0.5] Maximum allowed reprojection error (in pixels).
   * Higher values can increase reprojection performance, but decrease precision.
   * @property {import("../Tile.js").LoadFunction} [tileLoadFunction] Optional function to load a tile given a URL. The default is
   * ```js
   * function(imageTile, src) {
   *   imageTile.getImage().src = src;
   * };
   * ```
   * @property {boolean} [wrapX=true] Whether to wrap the world horizontally.
   * @property {number} [transition] Duration of the opacity transition for rendering.
   * To disable the opacity transition, pass `transition: 0`.
   * @property {number|import("../array.js").NearestDirectionFunction} [zDirection=0]
   * Choose whether to use tiles with a higher or lower zoom level when between integer
   * zoom levels. See {@link module:ol/tilegrid/TileGrid~TileGrid#getZForResolution}.
   */

  /**
   * @typedef {Object} BingMapsImageryMetadataResponse
   * @property {number} statusCode The response status code
   * @property {string} statusDescription The response status description
   * @property {string} authenticationResultCode The authentication result code
   * @property {Array<ResourceSet>} resourceSets The array of resource sets
   */

  /**
   * @typedef {Object} ResourceSet
   * @property {Array<Resource>} resources Resources.
   */

  /**
   * @typedef {Object} Resource
   * @property {number} imageHeight The image height
   * @property {number} imageWidth The image width
   * @property {number} zoomMin The minimum zoom level
   * @property {number} zoomMax The maximum zoom level
   * @property {string} imageUrl The image URL
   * @property {Array<string>} imageUrlSubdomains The image URL subdomains for rotation
   * @property {Array<ImageryProvider>} [imageryProviders] The array of ImageryProviders
   */

  /**
   * @typedef {Object} ImageryProvider
   * @property {Array<CoverageArea>} coverageAreas The coverage areas
   * @property {string} [attribution] The attribution
   */

  /**
   * @typedef {Object} CoverageArea
   * @property {number} zoomMin The minimum zoom
   * @property {number} zoomMax The maximum zoom
   * @property {Array<number>} bbox The coverage bounding box
   */

  /**
   * @classdesc
   * Layer source for Bing Maps tile data.
   * @api
   */
  class BingMaps extends TileImage {
    /**
     * @param {Options} options Bing Maps options.
     */
    constructor(options) {
      const hidpi = options.hidpi !== undefined ? options.hidpi : false;

      super({
        cacheSize: options.cacheSize,
        crossOrigin: 'anonymous',
        interpolate: options.interpolate,
        opaque: true,
        projection: get$2('EPSG:3857'),
        reprojectionErrorThreshold: options.reprojectionErrorThreshold,
        state: 'loading',
        tileLoadFunction: options.tileLoadFunction,
        tilePixelRatio: hidpi ? 2 : 1,
        wrapX: options.wrapX !== undefined ? options.wrapX : true,
        transition: options.transition,
        zDirection: options.zDirection,
      });

      /**
       * @private
       * @type {boolean}
       */
      this.hidpi_ = hidpi;

      /**
       * @private
       * @type {string}
       */
      this.culture_ = options.culture !== undefined ? options.culture : 'en-us';

      /**
       * @private
       * @type {number}
       */
      this.maxZoom_ = options.maxZoom !== undefined ? options.maxZoom : -1;

      /**
       * @private
       * @type {string}
       */
      this.apiKey_ = options.key;

      /**
       * @private
       * @type {string}
       */
      this.imagerySet_ = options.imagerySet;

      const url =
        'https://dev.virtualearth.net/REST/v1/Imagery/Metadata/' +
        this.imagerySet_ +
        '?uriScheme=https&include=ImageryProviders&key=' +
        this.apiKey_ +
        '&c=' +
        this.culture_;

      jsonp(
        url,
        this.handleImageryMetadataResponse.bind(this),
        undefined,
        'jsonp'
      );
    }

    /**
     * Get the api key used for this source.
     *
     * @return {string} The api key.
     * @api
     */
    getApiKey() {
      return this.apiKey_;
    }

    /**
     * Get the imagery set associated with this source.
     *
     * @return {string} The imagery set.
     * @api
     */
    getImagerySet() {
      return this.imagerySet_;
    }

    /**
     * @param {BingMapsImageryMetadataResponse} response Response.
     */
    handleImageryMetadataResponse(response) {
      if (
        response.statusCode != 200 ||
        response.statusDescription != 'OK' ||
        response.authenticationResultCode != 'ValidCredentials' ||
        response.resourceSets.length != 1 ||
        response.resourceSets[0].resources.length != 1
      ) {
        this.setState('error');
        return;
      }

      const resource = response.resourceSets[0].resources[0];
      const maxZoom = this.maxZoom_ == -1 ? resource.zoomMax : this.maxZoom_;

      const sourceProjection = this.getProjection();
      const extent = extentFromProjection(sourceProjection);
      const scale = this.hidpi_ ? 2 : 1;
      const tileSize =
        resource.imageWidth == resource.imageHeight
          ? resource.imageWidth / scale
          : [resource.imageWidth / scale, resource.imageHeight / scale];

      const tileGrid = createXYZ({
        extent: extent,
        minZoom: resource.zoomMin,
        maxZoom: maxZoom,
        tileSize: tileSize,
      });
      this.tileGrid = tileGrid;

      const culture = this.culture_;
      const hidpi = this.hidpi_;
      this.tileUrlFunction = createFromTileUrlFunctions(
        resource.imageUrlSubdomains.map(function (subdomain) {
          /** @type {import('../tilecoord.js').TileCoord} */
          const quadKeyTileCoord = [0, 0, 0];
          const imageUrl = resource.imageUrl
            .replace('{subdomain}', subdomain)
            .replace('{culture}', culture);
          return (
            /**
             * @param {import("../tilecoord.js").TileCoord} tileCoord Tile coordinate.
             * @param {number} pixelRatio Pixel ratio.
             * @param {import("../proj/Projection.js").default} projection Projection.
             * @return {string|undefined} Tile URL.
             */
            function (tileCoord, pixelRatio, projection) {
              if (!tileCoord) {
                return undefined;
              } else {
                createOrUpdate$1(
                  tileCoord[0],
                  tileCoord[1],
                  tileCoord[2],
                  quadKeyTileCoord
                );
                let url = imageUrl;
                if (hidpi) {
                  url += '&dpi=d1&device=mobile';
                }
                return url.replace('{quadkey}', quadKey(quadKeyTileCoord));
              }
            }
          );
        })
      );

      if (resource.imageryProviders) {
        const transform = getTransformFromProjections(
          get$2('EPSG:4326'),
          this.getProjection()
        );

        this.setAttributions(
          function (frameState) {
            const attributions = [];
            const viewState = frameState.viewState;
            const tileGrid = this.getTileGrid();
            const z = tileGrid.getZForResolution(
              viewState.resolution,
              this.zDirection
            );
            const tileCoord = tileGrid.getTileCoordForCoordAndZ(
              viewState.center,
              z
            );
            const zoom = tileCoord[0];
            resource.imageryProviders.map(function (imageryProvider) {
              let intersecting = false;
              const coverageAreas = imageryProvider.coverageAreas;
              for (let i = 0, ii = coverageAreas.length; i < ii; ++i) {
                const coverageArea = coverageAreas[i];
                if (
                  zoom >= coverageArea.zoomMin &&
                  zoom <= coverageArea.zoomMax
                ) {
                  const bbox = coverageArea.bbox;
                  const epsg4326Extent = [bbox[1], bbox[0], bbox[3], bbox[2]];
                  const extent = applyTransform(epsg4326Extent, transform);
                  if (intersects$3(extent, frameState.extent)) {
                    intersecting = true;
                    break;
                  }
                }
              }
              if (intersecting) {
                attributions.push(imageryProvider.attribution);
              }
            });

            attributions.push(TOS_ATTRIBUTION);
            return attributions;
          }.bind(this)
        );
      }

      this.setState('ready');
    }
  }

  /**
   * @module ol/source/XYZ
   */

  /**
   * @typedef {Object} Options
   * @property {import("./Source.js").AttributionLike} [attributions] Attributions.
   * @property {boolean} [attributionsCollapsible=true] Attributions are collapsible.
   * @property {number} [cacheSize] Initial tile cache size. Will auto-grow to hold at least the number of tiles in the viewport.
   * @property {null|string} [crossOrigin] The `crossOrigin` attribute for loaded images.  Note that
   * you must provide a `crossOrigin` value if you want to access pixel data with the Canvas renderer.
   * See https://developer.mozilla.org/en-US/docs/Web/HTML/CORS_enabled_image for more detail.
   * @property {boolean} [interpolate=true] Use interpolated values when resampling.  By default,
   * linear interpolation is used when resampling.  Set to false to use the nearest neighbor instead.
   * @property {boolean} [opaque=false] Whether the layer is opaque.
   * @property {import("../proj.js").ProjectionLike} [projection='EPSG:3857'] Projection.
   * @property {number} [reprojectionErrorThreshold=0.5] Maximum allowed reprojection error (in pixels).
   * Higher values can increase reprojection performance, but decrease precision.
   * @property {number} [maxZoom=42] Optional max zoom level. Not used if `tileGrid` is provided.
   * @property {number} [minZoom=0] Optional min zoom level. Not used if `tileGrid` is provided.
   * @property {number} [maxResolution] Optional tile grid resolution at level zero. Not used if `tileGrid` is provided.
   * @property {import("../tilegrid/TileGrid.js").default} [tileGrid] Tile grid.
   * @property {import("../Tile.js").LoadFunction} [tileLoadFunction] Optional function to load a tile given a URL. The default is
   * ```js
   * function(imageTile, src) {
   *   imageTile.getImage().src = src;
   * };
   * ```
   * @property {number} [tilePixelRatio=1] The pixel ratio used by the tile service.
   * For example, if the tile service advertizes 256px by 256px tiles but actually sends 512px
   * by 512px images (for retina/hidpi devices) then `tilePixelRatio`
   * should be set to `2`.
   * @property {number|import("../size.js").Size} [tileSize=[256, 256]] The tile size used by the tile service.
   * Not used if `tileGrid` is provided.
   * @property {number} [gutter=0] The size in pixels of the gutter around image tiles to ignore.
   * This allows artifacts of rendering at tile edges to be ignored.
   * Supported images should be wider and taller than the tile size by a value of `2 x gutter`.
   * @property {import("../Tile.js").UrlFunction} [tileUrlFunction] Optional function to get
   * tile URL given a tile coordinate and the projection.
   * Required if `url` or `urls` are not provided.
   * @property {string} [url] URL template. Must include `{x}`, `{y}` or `{-y}`,
   * and `{z}` placeholders. A `{?-?}` template pattern, for example `subdomain{a-f}.domain.com`,
   * may be used instead of defining each one separately in the `urls` option.
   * @property {Array<string>} [urls] An array of URL templates.
   * @property {boolean} [wrapX=true] Whether to wrap the world horizontally.
   * @property {number} [transition=250] Duration of the opacity transition for rendering.
   * To disable the opacity transition, pass `transition: 0`.
   * @property {number|import("../array.js").NearestDirectionFunction} [zDirection=0]
   * Choose whether to use tiles with a higher or lower zoom level when between integer
   * zoom levels. See {@link module:ol/tilegrid/TileGrid~TileGrid#getZForResolution}.
   */

  /**
   * @classdesc
   * Layer source for tile data with URLs in a set XYZ format that are
   * defined in a URL template. By default, this follows the widely-used
   * Google grid where `x` 0 and `y` 0 are in the top left. Grids like
   * TMS where `x` 0 and `y` 0 are in the bottom left can be used by
   * using the `{-y}` placeholder in the URL template, so long as the
   * source does not have a custom tile grid. In this case
   * a `tileUrlFunction` can be used, such as:
   * ```js
   *  tileUrlFunction: function(coordinate) {
   *    return 'http://mapserver.com/' + coordinate[0] + '/' +
   *      coordinate[1] + '/' + (-coordinate[2] - 1) + '.png';
   *  }
   * ```
   * @api
   */
  class XYZ extends TileImage {
    /**
     * @param {Options} [options] XYZ options.
     */
    constructor(options) {
      options = options || {};

      const projection =
        options.projection !== undefined ? options.projection : 'EPSG:3857';

      const tileGrid =
        options.tileGrid !== undefined
          ? options.tileGrid
          : createXYZ({
              extent: extentFromProjection(projection),
              maxResolution: options.maxResolution,
              maxZoom: options.maxZoom,
              minZoom: options.minZoom,
              tileSize: options.tileSize,
            });

      super({
        attributions: options.attributions,
        cacheSize: options.cacheSize,
        crossOrigin: options.crossOrigin,
        interpolate: options.interpolate,
        opaque: options.opaque,
        projection: projection,
        reprojectionErrorThreshold: options.reprojectionErrorThreshold,
        tileGrid: tileGrid,
        tileLoadFunction: options.tileLoadFunction,
        tilePixelRatio: options.tilePixelRatio,
        tileUrlFunction: options.tileUrlFunction,
        url: options.url,
        urls: options.urls,
        wrapX: options.wrapX !== undefined ? options.wrapX : true,
        transition: options.transition,
        attributionsCollapsible: options.attributionsCollapsible,
        zDirection: options.zDirection,
      });

      /**
       * @private
       * @type {number}
       */
      this.gutter_ = options.gutter !== undefined ? options.gutter : 0;
    }

    /**
     * @return {number} Gutter.
     */
    getGutter() {
      return this.gutter_;
    }
  }

  /**
   * @module ol/source/CartoDB
   */

  /**
   * @typedef {Object} Options
   * @property {import("./Source.js").AttributionLike} [attributions] Attributions.
   * @property {number} [cacheSize] Initial tile cache size. Will auto-grow to hold at least the number of tiles in the viewport.
   * @property {null|string} [crossOrigin] The `crossOrigin` attribute for loaded images.  Note that
   * you must provide a `crossOrigin` value if you want to access pixel data with the Canvas renderer.
   * See https://developer.mozilla.org/en-US/docs/Web/HTML/CORS_enabled_image for more detail.
   * @property {import("../proj.js").ProjectionLike} [projection='EPSG:3857'] Projection.
   * @property {number} [maxZoom=18] Max zoom.
   * @property {number} [minZoom] Minimum zoom.
   * @property {boolean} [wrapX=true] Whether to wrap the world horizontally.
   * @property {Object} [config] If using anonymous maps, the CartoDB config to use. See
   * https://carto.com/developers/maps-api/guides/anonymous-maps/
   * for more detail.
   * If using named maps, a key-value lookup with the template parameters.
   * See https://carto.com/developers/maps-api/guides/named-maps/
   * for more detail.
   * @property {string} [map] If using named maps, this will be the name of the template to load.
   * See https://carto.com/developers/maps-api/guides/named-maps/
   * for more detail.
   * @property {string} [account] Username as used to access public Carto dashboard at https://{username}.carto.com/.
   * @property {number} [transition=250] Duration of the opacity transition for rendering.
   * To disable the opacity transition, pass `transition: 0`.
   * @property {number|import("../array.js").NearestDirectionFunction} [zDirection=0]
   * Choose whether to use tiles with a higher or lower zoom level when between integer
   * zoom levels. See {@link module:ol/tilegrid/TileGrid~TileGrid#getZForResolution}.
   */

  /**
   * @typedef {Object} CartoDBLayerInfo
   * @property {string} layergroupid The layer group ID
   * @property {{https: string}} cdn_url The CDN URL
   */

  /**
   * @classdesc
   * Layer source for the CartoDB Maps API.
   * @api
   */
  class CartoDB extends XYZ {
    /**
     * @param {Options} options CartoDB options.
     */
    constructor(options) {
      super({
        attributions: options.attributions,
        cacheSize: options.cacheSize,
        crossOrigin: options.crossOrigin,
        maxZoom: options.maxZoom !== undefined ? options.maxZoom : 18,
        minZoom: options.minZoom,
        projection: options.projection,
        transition: options.transition,
        wrapX: options.wrapX,
        zDirection: options.zDirection,
      });

      /**
       * @type {string}
       * @private
       */
      this.account_ = options.account;

      /**
       * @type {string}
       * @private
       */
      this.mapId_ = options.map || '';

      /**
       * @type {!Object}
       * @private
       */
      this.config_ = options.config || {};

      /**
       * @type {!Object<string, CartoDBLayerInfo>}
       * @private
       */
      this.templateCache_ = {};

      this.initializeMap_();
    }

    /**
     * Returns the current config.
     * @return {!Object} The current configuration.
     * @api
     */
    getConfig() {
      return this.config_;
    }

    /**
     * Updates the carto db config.
     * @param {Object} config a key-value lookup. Values will replace current values
     *     in the config.
     * @api
     */
    updateConfig(config) {
      Object.assign(this.config_, config);
      this.initializeMap_();
    }

    /**
     * Sets the CartoDB config
     * @param {Object} config In the case of anonymous maps, a CartoDB configuration
     *     object.
     * If using named maps, a key-value lookup with the template parameters.
     * @api
     */
    setConfig(config) {
      this.config_ = config || {};
      this.initializeMap_();
    }

    /**
     * Issue a request to initialize the CartoDB map.
     * @private
     */
    initializeMap_() {
      const paramHash = JSON.stringify(this.config_);
      if (this.templateCache_[paramHash]) {
        this.applyTemplate_(this.templateCache_[paramHash]);
        return;
      }
      let mapUrl = 'https://' + this.account_ + '.carto.com/api/v1/map';

      if (this.mapId_) {
        mapUrl += '/named/' + this.mapId_;
      }

      const client = new XMLHttpRequest();
      client.addEventListener(
        'load',
        this.handleInitResponse_.bind(this, paramHash)
      );
      client.addEventListener('error', this.handleInitError_.bind(this));
      client.open('POST', mapUrl);
      client.setRequestHeader('Content-type', 'application/json');
      client.send(JSON.stringify(this.config_));
    }

    /**
     * Handle map initialization response.
     * @param {string} paramHash a hash representing the parameter set that was used
     *     for the request
     * @param {Event} event Event.
     * @private
     */
    handleInitResponse_(paramHash, event) {
      const client = /** @type {XMLHttpRequest} */ (event.target);
      // status will be 0 for file:// urls
      if (!client.status || (client.status >= 200 && client.status < 300)) {
        let response;
        try {
          response = /** @type {CartoDBLayerInfo} */ (
            JSON.parse(client.responseText)
          );
        } catch (err) {
          this.setState('error');
          return;
        }
        this.applyTemplate_(response);
        this.templateCache_[paramHash] = response;
        this.setState('ready');
      } else {
        this.setState('error');
      }
    }

    /**
     * @private
     * @param {Event} event Event.
     */
    handleInitError_(event) {
      this.setState('error');
    }

    /**
     * Apply the new tile urls returned by carto db
     * @param {CartoDBLayerInfo} data Result of carto db call.
     * @private
     */
    applyTemplate_(data) {
      const tilesUrl =
        'https://' +
        data.cdn_url.https +
        '/' +
        this.account_ +
        '/api/v1/map/' +
        data.layergroupid +
        '/{z}/{x}/{y}.png';
      this.setUrl(tilesUrl);
    }
  }

  function quickselect(arr, k, left, right, compare) {
      quickselectStep(arr, k, left || 0, right || (arr.length - 1), compare || defaultCompare);
  }

  function quickselectStep(arr, k, left, right, compare) {

      while (right > left) {
          if (right - left > 600) {
              var n = right - left + 1;
              var m = k - left + 1;
              var z = Math.log(n);
              var s = 0.5 * Math.exp(2 * z / 3);
              var sd = 0.5 * Math.sqrt(z * s * (n - s) / n) * (m - n / 2 < 0 ? -1 : 1);
              var newLeft = Math.max(left, Math.floor(k - m * s / n + sd));
              var newRight = Math.min(right, Math.floor(k + (n - m) * s / n + sd));
              quickselectStep(arr, k, newLeft, newRight, compare);
          }

          var t = arr[k];
          var i = left;
          var j = right;

          swap(arr, left, k);
          if (compare(arr[right], t) > 0) swap(arr, left, right);

          while (i < j) {
              swap(arr, i, j);
              i++;
              j--;
              while (compare(arr[i], t) < 0) i++;
              while (compare(arr[j], t) > 0) j--;
          }

          if (compare(arr[left], t) === 0) swap(arr, left, j);
          else {
              j++;
              swap(arr, j, right);
          }

          if (j <= k) left = j + 1;
          if (k <= j) right = j - 1;
      }
  }

  function swap(arr, i, j) {
      var tmp = arr[i];
      arr[i] = arr[j];
      arr[j] = tmp;
  }

  function defaultCompare(a, b) {
      return a < b ? -1 : a > b ? 1 : 0;
  }

  class RBush$1 {
      constructor(maxEntries = 9) {
          // max entries in a node is 9 by default; min node fill is 40% for best performance
          this._maxEntries = Math.max(4, maxEntries);
          this._minEntries = Math.max(2, Math.ceil(this._maxEntries * 0.4));
          this.clear();
      }

      all() {
          return this._all(this.data, []);
      }

      search(bbox) {
          let node = this.data;
          const result = [];

          if (!intersects$2(bbox, node)) return result;

          const toBBox = this.toBBox;
          const nodesToSearch = [];

          while (node) {
              for (let i = 0; i < node.children.length; i++) {
                  const child = node.children[i];
                  const childBBox = node.leaf ? toBBox(child) : child;

                  if (intersects$2(bbox, childBBox)) {
                      if (node.leaf) result.push(child);
                      else if (contains$1(bbox, childBBox)) this._all(child, result);
                      else nodesToSearch.push(child);
                  }
              }
              node = nodesToSearch.pop();
          }

          return result;
      }

      collides(bbox) {
          let node = this.data;

          if (!intersects$2(bbox, node)) return false;

          const nodesToSearch = [];
          while (node) {
              for (let i = 0; i < node.children.length; i++) {
                  const child = node.children[i];
                  const childBBox = node.leaf ? this.toBBox(child) : child;

                  if (intersects$2(bbox, childBBox)) {
                      if (node.leaf || contains$1(bbox, childBBox)) return true;
                      nodesToSearch.push(child);
                  }
              }
              node = nodesToSearch.pop();
          }

          return false;
      }

      load(data) {
          if (!(data && data.length)) return this;

          if (data.length < this._minEntries) {
              for (let i = 0; i < data.length; i++) {
                  this.insert(data[i]);
              }
              return this;
          }

          // recursively build the tree with the given data from scratch using OMT algorithm
          let node = this._build(data.slice(), 0, data.length - 1, 0);

          if (!this.data.children.length) {
              // save as is if tree is empty
              this.data = node;

          } else if (this.data.height === node.height) {
              // split root if trees have the same height
              this._splitRoot(this.data, node);

          } else {
              if (this.data.height < node.height) {
                  // swap trees if inserted one is bigger
                  const tmpNode = this.data;
                  this.data = node;
                  node = tmpNode;
              }

              // insert the small tree into the large tree at appropriate level
              this._insert(node, this.data.height - node.height - 1, true);
          }

          return this;
      }

      insert(item) {
          if (item) this._insert(item, this.data.height - 1);
          return this;
      }

      clear() {
          this.data = createNode([]);
          return this;
      }

      remove(item, equalsFn) {
          if (!item) return this;

          let node = this.data;
          const bbox = this.toBBox(item);
          const path = [];
          const indexes = [];
          let i, parent, goingUp;

          // depth-first iterative tree traversal
          while (node || path.length) {

              if (!node) { // go up
                  node = path.pop();
                  parent = path[path.length - 1];
                  i = indexes.pop();
                  goingUp = true;
              }

              if (node.leaf) { // check current node
                  const index = findItem(item, node.children, equalsFn);

                  if (index !== -1) {
                      // item found, remove the item and condense tree upwards
                      node.children.splice(index, 1);
                      path.push(node);
                      this._condense(path);
                      return this;
                  }
              }

              if (!goingUp && !node.leaf && contains$1(node, bbox)) { // go down
                  path.push(node);
                  indexes.push(i);
                  i = 0;
                  parent = node;
                  node = node.children[0];

              } else if (parent) { // go right
                  i++;
                  node = parent.children[i];
                  goingUp = false;

              } else node = null; // nothing found
          }

          return this;
      }

      toBBox(item) { return item; }

      compareMinX(a, b) { return a.minX - b.minX; }
      compareMinY(a, b) { return a.minY - b.minY; }

      toJSON() { return this.data; }

      fromJSON(data) {
          this.data = data;
          return this;
      }

      _all(node, result) {
          const nodesToSearch = [];
          while (node) {
              if (node.leaf) result.push(...node.children);
              else nodesToSearch.push(...node.children);

              node = nodesToSearch.pop();
          }
          return result;
      }

      _build(items, left, right, height) {

          const N = right - left + 1;
          let M = this._maxEntries;
          let node;

          if (N <= M) {
              // reached leaf level; return leaf
              node = createNode(items.slice(left, right + 1));
              calcBBox(node, this.toBBox);
              return node;
          }

          if (!height) {
              // target height of the bulk-loaded tree
              height = Math.ceil(Math.log(N) / Math.log(M));

              // target number of root entries to maximize storage utilization
              M = Math.ceil(N / Math.pow(M, height - 1));
          }

          node = createNode([]);
          node.leaf = false;
          node.height = height;

          // split the items into M mostly square tiles

          const N2 = Math.ceil(N / M);
          const N1 = N2 * Math.ceil(Math.sqrt(M));

          multiSelect(items, left, right, N1, this.compareMinX);

          for (let i = left; i <= right; i += N1) {

              const right2 = Math.min(i + N1 - 1, right);

              multiSelect(items, i, right2, N2, this.compareMinY);

              for (let j = i; j <= right2; j += N2) {

                  const right3 = Math.min(j + N2 - 1, right2);

                  // pack each entry recursively
                  node.children.push(this._build(items, j, right3, height - 1));
              }
          }

          calcBBox(node, this.toBBox);

          return node;
      }

      _chooseSubtree(bbox, node, level, path) {
          while (true) {
              path.push(node);

              if (node.leaf || path.length - 1 === level) break;

              let minArea = Infinity;
              let minEnlargement = Infinity;
              let targetNode;

              for (let i = 0; i < node.children.length; i++) {
                  const child = node.children[i];
                  const area = bboxArea(child);
                  const enlargement = enlargedArea(bbox, child) - area;

                  // choose entry with the least area enlargement
                  if (enlargement < minEnlargement) {
                      minEnlargement = enlargement;
                      minArea = area < minArea ? area : minArea;
                      targetNode = child;

                  } else if (enlargement === minEnlargement) {
                      // otherwise choose one with the smallest area
                      if (area < minArea) {
                          minArea = area;
                          targetNode = child;
                      }
                  }
              }

              node = targetNode || node.children[0];
          }

          return node;
      }

      _insert(item, level, isNode) {
          const bbox = isNode ? item : this.toBBox(item);
          const insertPath = [];

          // find the best node for accommodating the item, saving all nodes along the path too
          const node = this._chooseSubtree(bbox, this.data, level, insertPath);

          // put the item into the node
          node.children.push(item);
          extend(node, bbox);

          // split on node overflow; propagate upwards if necessary
          while (level >= 0) {
              if (insertPath[level].children.length > this._maxEntries) {
                  this._split(insertPath, level);
                  level--;
              } else break;
          }

          // adjust bboxes along the insertion path
          this._adjustParentBBoxes(bbox, insertPath, level);
      }

      // split overflowed node into two
      _split(insertPath, level) {
          const node = insertPath[level];
          const M = node.children.length;
          const m = this._minEntries;

          this._chooseSplitAxis(node, m, M);

          const splitIndex = this._chooseSplitIndex(node, m, M);

          const newNode = createNode(node.children.splice(splitIndex, node.children.length - splitIndex));
          newNode.height = node.height;
          newNode.leaf = node.leaf;

          calcBBox(node, this.toBBox);
          calcBBox(newNode, this.toBBox);

          if (level) insertPath[level - 1].children.push(newNode);
          else this._splitRoot(node, newNode);
      }

      _splitRoot(node, newNode) {
          // split root node
          this.data = createNode([node, newNode]);
          this.data.height = node.height + 1;
          this.data.leaf = false;
          calcBBox(this.data, this.toBBox);
      }

      _chooseSplitIndex(node, m, M) {
          let index;
          let minOverlap = Infinity;
          let minArea = Infinity;

          for (let i = m; i <= M - m; i++) {
              const bbox1 = distBBox(node, 0, i, this.toBBox);
              const bbox2 = distBBox(node, i, M, this.toBBox);

              const overlap = intersectionArea(bbox1, bbox2);
              const area = bboxArea(bbox1) + bboxArea(bbox2);

              // choose distribution with minimum overlap
              if (overlap < minOverlap) {
                  minOverlap = overlap;
                  index = i;

                  minArea = area < minArea ? area : minArea;

              } else if (overlap === minOverlap) {
                  // otherwise choose distribution with minimum area
                  if (area < minArea) {
                      minArea = area;
                      index = i;
                  }
              }
          }

          return index || M - m;
      }

      // sorts node children by the best axis for split
      _chooseSplitAxis(node, m, M) {
          const compareMinX = node.leaf ? this.compareMinX : compareNodeMinX;
          const compareMinY = node.leaf ? this.compareMinY : compareNodeMinY;
          const xMargin = this._allDistMargin(node, m, M, compareMinX);
          const yMargin = this._allDistMargin(node, m, M, compareMinY);

          // if total distributions margin value is minimal for x, sort by minX,
          // otherwise it's already sorted by minY
          if (xMargin < yMargin) node.children.sort(compareMinX);
      }

      // total margin of all possible split distributions where each node is at least m full
      _allDistMargin(node, m, M, compare) {
          node.children.sort(compare);

          const toBBox = this.toBBox;
          const leftBBox = distBBox(node, 0, m, toBBox);
          const rightBBox = distBBox(node, M - m, M, toBBox);
          let margin = bboxMargin(leftBBox) + bboxMargin(rightBBox);

          for (let i = m; i < M - m; i++) {
              const child = node.children[i];
              extend(leftBBox, node.leaf ? toBBox(child) : child);
              margin += bboxMargin(leftBBox);
          }

          for (let i = M - m - 1; i >= m; i--) {
              const child = node.children[i];
              extend(rightBBox, node.leaf ? toBBox(child) : child);
              margin += bboxMargin(rightBBox);
          }

          return margin;
      }

      _adjustParentBBoxes(bbox, path, level) {
          // adjust bboxes along the given tree path
          for (let i = level; i >= 0; i--) {
              extend(path[i], bbox);
          }
      }

      _condense(path) {
          // go through the path, removing empty nodes and updating bboxes
          for (let i = path.length - 1, siblings; i >= 0; i--) {
              if (path[i].children.length === 0) {
                  if (i > 0) {
                      siblings = path[i - 1].children;
                      siblings.splice(siblings.indexOf(path[i]), 1);

                  } else this.clear();

              } else calcBBox(path[i], this.toBBox);
          }
      }
  }

  function findItem(item, items, equalsFn) {
      if (!equalsFn) return items.indexOf(item);

      for (let i = 0; i < items.length; i++) {
          if (equalsFn(item, items[i])) return i;
      }
      return -1;
  }

  // calculate node's bbox from bboxes of its children
  function calcBBox(node, toBBox) {
      distBBox(node, 0, node.children.length, toBBox, node);
  }

  // min bounding rectangle of node children from k to p-1
  function distBBox(node, k, p, toBBox, destNode) {
      if (!destNode) destNode = createNode(null);
      destNode.minX = Infinity;
      destNode.minY = Infinity;
      destNode.maxX = -Infinity;
      destNode.maxY = -Infinity;

      for (let i = k; i < p; i++) {
          const child = node.children[i];
          extend(destNode, node.leaf ? toBBox(child) : child);
      }

      return destNode;
  }

  function extend(a, b) {
      a.minX = Math.min(a.minX, b.minX);
      a.minY = Math.min(a.minY, b.minY);
      a.maxX = Math.max(a.maxX, b.maxX);
      a.maxY = Math.max(a.maxY, b.maxY);
      return a;
  }

  function compareNodeMinX(a, b) { return a.minX - b.minX; }
  function compareNodeMinY(a, b) { return a.minY - b.minY; }

  function bboxArea(a)   { return (a.maxX - a.minX) * (a.maxY - a.minY); }
  function bboxMargin(a) { return (a.maxX - a.minX) + (a.maxY - a.minY); }

  function enlargedArea(a, b) {
      return (Math.max(b.maxX, a.maxX) - Math.min(b.minX, a.minX)) *
             (Math.max(b.maxY, a.maxY) - Math.min(b.minY, a.minY));
  }

  function intersectionArea(a, b) {
      const minX = Math.max(a.minX, b.minX);
      const minY = Math.max(a.minY, b.minY);
      const maxX = Math.min(a.maxX, b.maxX);
      const maxY = Math.min(a.maxY, b.maxY);

      return Math.max(0, maxX - minX) *
             Math.max(0, maxY - minY);
  }

  function contains$1(a, b) {
      return a.minX <= b.minX &&
             a.minY <= b.minY &&
             b.maxX <= a.maxX &&
             b.maxY <= a.maxY;
  }

  function intersects$2(a, b) {
      return b.minX <= a.maxX &&
             b.minY <= a.maxY &&
             b.maxX >= a.minX &&
             b.maxY >= a.minY;
  }

  function createNode(children) {
      return {
          children,
          height: 1,
          leaf: true,
          minX: Infinity,
          minY: Infinity,
          maxX: -Infinity,
          maxY: -Infinity
      };
  }

  // sort an array so that items come in groups of n unsorted items, with groups sorted between each other;
  // combines selection algorithm with binary divide & conquer approach

  function multiSelect(arr, left, right, n, compare) {
      const stack = [left, right];

      while (stack.length) {
          right = stack.pop();
          left = stack.pop();

          if (right - left <= n) continue;

          const mid = left + Math.ceil((right - left) / n / 2) * n;
          quickselect(arr, mid, left, right, compare);

          stack.push(left, mid, mid, right);
      }
  }

  /**
   * @module ol/structs/RBush
   */

  /**
   * @typedef {Object} Entry
   * @property {number} minX MinX.
   * @property {number} minY MinY.
   * @property {number} maxX MaxX.
   * @property {number} maxY MaxY.
   * @property {Object} [value] Value.
   */

  /**
   * @classdesc
   * Wrapper around the RBush by Vladimir Agafonkin.
   * See https://github.com/mourner/rbush.
   *
   * @template T
   */
  class RBush {
    /**
     * @param {number} [maxEntries] Max entries.
     */
    constructor(maxEntries) {
      /**
       * @private
       */
      this.rbush_ = new RBush$1(maxEntries);

      /**
       * A mapping between the objects added to this rbush wrapper
       * and the objects that are actually added to the internal rbush.
       * @private
       * @type {Object<string, Entry>}
       */
      this.items_ = {};
    }

    /**
     * Insert a value into the RBush.
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {T} value Value.
     */
    insert(extent, value) {
      /** @type {Entry} */
      const item = {
        minX: extent[0],
        minY: extent[1],
        maxX: extent[2],
        maxY: extent[3],
        value: value,
      };

      this.rbush_.insert(item);
      this.items_[getUid(value)] = item;
    }

    /**
     * Bulk-insert values into the RBush.
     * @param {Array<import("../extent.js").Extent>} extents Extents.
     * @param {Array<T>} values Values.
     */
    load(extents, values) {
      const items = new Array(values.length);
      for (let i = 0, l = values.length; i < l; i++) {
        const extent = extents[i];
        const value = values[i];

        /** @type {Entry} */
        const item = {
          minX: extent[0],
          minY: extent[1],
          maxX: extent[2],
          maxY: extent[3],
          value: value,
        };
        items[i] = item;
        this.items_[getUid(value)] = item;
      }
      this.rbush_.load(items);
    }

    /**
     * Remove a value from the RBush.
     * @param {T} value Value.
     * @return {boolean} Removed.
     */
    remove(value) {
      const uid = getUid(value);

      // get the object in which the value was wrapped when adding to the
      // internal rbush. then use that object to do the removal.
      const item = this.items_[uid];
      delete this.items_[uid];
      return this.rbush_.remove(item) !== null;
    }

    /**
     * Update the extent of a value in the RBush.
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {T} value Value.
     */
    update(extent, value) {
      const item = this.items_[getUid(value)];
      const bbox = [item.minX, item.minY, item.maxX, item.maxY];
      if (!equals$2(bbox, extent)) {
        this.remove(value);
        this.insert(extent, value);
      }
    }

    /**
     * Return all values in the RBush.
     * @return {Array<T>} All.
     */
    getAll() {
      const items = this.rbush_.all();
      return items.map(function (item) {
        return item.value;
      });
    }

    /**
     * Return all values in the given extent.
     * @param {import("../extent.js").Extent} extent Extent.
     * @return {Array<T>} All in extent.
     */
    getInExtent(extent) {
      /** @type {Entry} */
      const bbox = {
        minX: extent[0],
        minY: extent[1],
        maxX: extent[2],
        maxY: extent[3],
      };
      const items = this.rbush_.search(bbox);
      return items.map(function (item) {
        return item.value;
      });
    }

    /**
     * Calls a callback function with each value in the tree.
     * If the callback returns a truthy value, this value is returned without
     * checking the rest of the tree.
     * @param {function(T): *} callback Callback.
     * @return {*} Callback return value.
     */
    forEach(callback) {
      return this.forEach_(this.getAll(), callback);
    }

    /**
     * Calls a callback function with each value in the provided extent.
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {function(T): *} callback Callback.
     * @return {*} Callback return value.
     */
    forEachInExtent(extent, callback) {
      return this.forEach_(this.getInExtent(extent), callback);
    }

    /**
     * @param {Array<T>} values Values.
     * @param {function(T): *} callback Callback.
     * @private
     * @return {*} Callback return value.
     */
    forEach_(values, callback) {
      let result;
      for (let i = 0, l = values.length; i < l; i++) {
        result = callback(values[i]);
        if (result) {
          return result;
        }
      }
      return result;
    }

    /**
     * @return {boolean} Is empty.
     */
    isEmpty() {
      return isEmpty$1(this.items_);
    }

    /**
     * Remove all values from the RBush.
     */
    clear() {
      this.rbush_.clear();
      this.items_ = {};
    }

    /**
     * @param {import("../extent.js").Extent} [extent] Extent.
     * @return {import("../extent.js").Extent} Extent.
     */
    getExtent(extent) {
      const data = this.rbush_.toJSON();
      return createOrUpdate$2(data.minX, data.minY, data.maxX, data.maxY, extent);
    }

    /**
     * @param {RBush} rbush R-Tree.
     */
    concat(rbush) {
      this.rbush_.load(rbush.rbush_.all());
      for (const i in rbush.items_) {
        this.items_[i] = rbush.items_[i];
      }
    }
  }

  /**
   * @module ol/source/VectorEventType
   */

  /**
   * @enum {string}
   */
  var VectorEventType = {
    /**
     * Triggered when a feature is added to the source.
     * @event module:ol/source/Vector.VectorSourceEvent#addfeature
     * @api
     */
    ADDFEATURE: 'addfeature',

    /**
     * Triggered when a feature is updated.
     * @event module:ol/source/Vector.VectorSourceEvent#changefeature
     * @api
     */
    CHANGEFEATURE: 'changefeature',

    /**
     * Triggered when the clear method is called on the source.
     * @event module:ol/source/Vector.VectorSourceEvent#clear
     * @api
     */
    CLEAR: 'clear',

    /**
     * Triggered when a feature is removed from the source.
     * See {@link module:ol/source/Vector~VectorSource#clear source.clear()} for exceptions.
     * @event module:ol/source/Vector.VectorSourceEvent#removefeature
     * @api
     */
    REMOVEFEATURE: 'removefeature',

    /**
     * Triggered when features starts loading.
     * @event module:ol/source/Vector.VectorSourceEvent#featuresloadstart
     * @api
     */
    FEATURESLOADSTART: 'featuresloadstart',

    /**
     * Triggered when features finishes loading.
     * @event module:ol/source/Vector.VectorSourceEvent#featuresloadend
     * @api
     */
    FEATURESLOADEND: 'featuresloadend',

    /**
     * Triggered if feature loading results in an error.
     * @event module:ol/source/Vector.VectorSourceEvent#featuresloaderror
     * @api
     */
    FEATURESLOADERROR: 'featuresloaderror',
  };

  /**
   * @typedef {'addfeature'|'changefeature'|'clear'|'removefeature'|'featuresloadstart'|'featuresloadend'|'featuresloaderror'} VectorSourceEventTypes
   */

  /**
   * @module ol/source/Vector
   */

  /**
   * A function that takes an {@link module:ol/extent~Extent} and a resolution as arguments, and
   * returns an array of {@link module:ol/extent~Extent} with the extents to load. Usually this
   * is one of the standard {@link module:ol/loadingstrategy} strategies.
   *
   * @typedef {function(import("../extent.js").Extent, number, import("../proj/Projection.js").default): Array<import("../extent.js").Extent>} LoadingStrategy
   * @api
   */

  /**
   * @classdesc
   * Events emitted by {@link module:ol/source/Vector~VectorSource} instances are instances of this
   * type.
   * @template {import("../geom/Geometry.js").default} [Geometry=import("../geom/Geometry.js").default]
   */
  class VectorSourceEvent extends BaseEvent {
    /**
     * @param {string} type Type.
     * @param {import("../Feature.js").default<Geometry>} [feature] Feature.
     * @param {Array<import("../Feature.js").default<Geometry>>} [features] Features.
     */
    constructor(type, feature, features) {
      super(type);

      /**
       * The added or removed feature for the `ADDFEATURE` and `REMOVEFEATURE` events, `undefined` otherwise.
       * @type {import("../Feature.js").default<Geometry>|undefined}
       * @api
       */
      this.feature = feature;

      /**
       * The loaded features for the `FEATURESLOADED` event, `undefined` otherwise.
       * @type {Array<import("../Feature.js").default<Geometry>>|undefined}
       * @api
       */
      this.features = features;
    }
  }

  /***
   * @template Return
   * @typedef {import("../Observable").OnSignature<import("../Observable").EventTypes, import("../events/Event.js").default, Return> &
   *   import("../Observable").OnSignature<import("../ObjectEventType").Types, import("../Object").ObjectEvent, Return> &
   *   import("../Observable").OnSignature<import("./VectorEventType").VectorSourceEventTypes, VectorSourceEvent, Return> &
   *   import("../Observable").CombinedOnSignature<import("../Observable").EventTypes|import("../ObjectEventType").Types|
   *     import("./VectorEventType").VectorSourceEventTypes, Return>} VectorSourceOnSignature
   */

  /**
   * @typedef {Object} Options
   * @property {import("./Source.js").AttributionLike} [attributions] Attributions.
   * @property {Array<import("../Feature.js").default<Geometry>>|Collection<import("../Feature.js").default<Geometry>>} [features]
   * Features. If provided as {@link module:ol/Collection~Collection}, the features in the source
   * and the collection will stay in sync.
   * @property {import("../format/Feature.js").default} [format] The feature format used by the XHR
   * feature loader when `url` is set. Required if `url` is set, otherwise ignored.
   * @property {import("../featureloader.js").FeatureLoader} [loader]
   * The loader function used to load features, from a remote source for example.
   * If this is not set and `url` is set, the source will create and use an XHR
   * feature loader. The `'featuresloadend'` and `'featuresloaderror'` events
   * will only fire if the `success` and `failure` callbacks are used.
   *
   * Example:
   *
   * ```js
   * import {Vector} from 'ol/source';
   * import {GeoJSON} from 'ol/format';
   * import {bbox} from 'ol/loadingstrategy';
   *
   * const vectorSource = new Vector({
   *   format: new GeoJSON(),
   *   loader: function(extent, resolution, projection, success, failure) {
   *      const proj = projection.getCode();
   *      const url = 'https://ahocevar.com/geoserver/wfs?service=WFS&' +
   *          'version=1.1.0&request=GetFeature&typename=osm:water_areas&' +
   *          'outputFormat=application/json&srsname=' + proj + '&' +
   *          'bbox=' + extent.join(',') + ',' + proj;
   *      const xhr = new XMLHttpRequest();
   *      xhr.open('GET', url);
   *      const onError = function() {
   *        vectorSource.removeLoadedExtent(extent);
   *        failure();
   *      }
   *      xhr.onerror = onError;
   *      xhr.onload = function() {
   *        if (xhr.status == 200) {
   *          const features = vectorSource.getFormat().readFeatures(xhr.responseText);
   *          vectorSource.addFeatures(features);
   *          success(features);
   *        } else {
   *          onError();
   *        }
   *      }
   *      xhr.send();
   *    },
   *    strategy: bbox,
   *  });
   * ```
   * @property {boolean} [overlaps=true] This source may have overlapping geometries.
   * Setting this to `false` (e.g. for sources with polygons that represent administrative
   * boundaries or TopoJSON sources) allows the renderer to optimise fill and
   * stroke operations.
   * @property {LoadingStrategy} [strategy] The loading strategy to use.
   * By default an {@link module:ol/loadingstrategy.all}
   * strategy is used, a one-off strategy which loads all features at once.
   * @property {string|import("../featureloader.js").FeatureUrlFunction} [url]
   * Setting this option instructs the source to load features using an XHR loader
   * (see {@link module:ol/featureloader.xhr}). Use a `string` and an
   * {@link module:ol/loadingstrategy.all} for a one-off download of all features from
   * the given URL. Use a {@link module:ol/featureloader~FeatureUrlFunction} to generate the url with
   * other loading strategies.
   * Requires `format` to be set as well.
   * When default XHR feature loader is provided, the features will
   * be transformed from the data projection to the view projection
   * during parsing. If your remote data source does not advertise its projection
   * properly, this transformation will be incorrect. For some formats, the
   * default projection (usually EPSG:4326) can be overridden by setting the
   * dataProjection constructor option on the format.
   * Note that if a source contains non-feature data, such as a GeoJSON geometry
   * or a KML NetworkLink, these will be ignored. Use a custom loader to load these.
   * @property {boolean} [useSpatialIndex=true]
   * By default, an RTree is used as spatial index. When features are removed and
   * added frequently, and the total number of features is low, setting this to
   * `false` may improve performance.
   *
   * Note that
   * {@link module:ol/source/Vector~VectorSource#getFeaturesInExtent},
   * {@link module:ol/source/Vector~VectorSource#getClosestFeatureToCoordinate} and
   * {@link module:ol/source/Vector~VectorSource#getExtent} cannot be used when `useSpatialIndex` is
   * set to `false`, and {@link module:ol/source/Vector~VectorSource#forEachFeatureInExtent} will loop
   * through all features.
   *
   * When set to `false`, the features will be maintained in an
   * {@link module:ol/Collection~Collection}, which can be retrieved through
   * {@link module:ol/source/Vector~VectorSource#getFeaturesCollection}.
   * @property {boolean} [wrapX=true] Wrap the world horizontally. For vector editing across the
   * -180° and 180° meridians to work properly, this should be set to `false`. The
   * resulting geometry coordinates will then exceed the world bounds.
   * @template {import("../geom/Geometry.js").default} [Geometry=import("../geom/Geometry.js").default]
   */

  /**
   * @classdesc
   * Provides a source of features for vector layers. Vector features provided
   * by this source are suitable for editing. See {@link module:ol/source/VectorTile~VectorTile} for
   * vector data that is optimized for rendering.
   *
   * @fires VectorSourceEvent
   * @api
   * @template {import("../geom/Geometry.js").default} [Geometry=import("../geom/Geometry.js").default]
   */
  class VectorSource extends Source {
    /**
     * @param {Options<Geometry>} [options] Vector source options.
     */
    constructor(options) {
      options = options || {};

      super({
        attributions: options.attributions,
        interpolate: true,
        projection: undefined,
        state: 'ready',
        wrapX: options.wrapX !== undefined ? options.wrapX : true,
      });

      /***
       * @type {VectorSourceOnSignature<import("../events").EventsKey>}
       */
      this.on;

      /***
       * @type {VectorSourceOnSignature<import("../events").EventsKey>}
       */
      this.once;

      /***
       * @type {VectorSourceOnSignature<void>}
       */
      this.un;

      /**
       * @private
       * @type {import("../featureloader.js").FeatureLoader}
       */
      this.loader_ = VOID;

      /**
       * @private
       * @type {import("../format/Feature.js").default|undefined}
       */
      this.format_ = options.format;

      /**
       * @private
       * @type {boolean}
       */
      this.overlaps_ = options.overlaps === undefined ? true : options.overlaps;

      /**
       * @private
       * @type {string|import("../featureloader.js").FeatureUrlFunction|undefined}
       */
      this.url_ = options.url;

      if (options.loader !== undefined) {
        this.loader_ = options.loader;
      } else if (this.url_ !== undefined) {
        assert(this.format_, 7); // `format` must be set when `url` is set
        // create a XHR feature loader for "url" and "format"
        this.loader_ = xhr(
          this.url_,
          /** @type {import("../format/Feature.js").default} */ (this.format_)
        );
      }

      /**
       * @private
       * @type {LoadingStrategy}
       */
      this.strategy_ =
        options.strategy !== undefined ? options.strategy : all;

      const useSpatialIndex =
        options.useSpatialIndex !== undefined ? options.useSpatialIndex : true;

      /**
       * @private
       * @type {RBush<import("../Feature.js").default<Geometry>>}
       */
      this.featuresRtree_ = useSpatialIndex ? new RBush() : null;

      /**
       * @private
       * @type {RBush<{extent: import("../extent.js").Extent}>}
       */
      this.loadedExtentsRtree_ = new RBush();

      /**
       * @type {number}
       * @private
       */
      this.loadingExtentsCount_ = 0;

      /**
       * @private
       * @type {!Object<string, import("../Feature.js").default<Geometry>>}
       */
      this.nullGeometryFeatures_ = {};

      /**
       * A lookup of features by id (the return from feature.getId()).
       * @private
       * @type {!Object<string, import("../Feature.js").default<Geometry>>}
       */
      this.idIndex_ = {};

      /**
       * A lookup of features by uid (using getUid(feature)).
       * @private
       * @type {!Object<string, import("../Feature.js").default<Geometry>>}
       */
      this.uidIndex_ = {};

      /**
       * @private
       * @type {Object<string, Array<import("../events.js").EventsKey>>}
       */
      this.featureChangeKeys_ = {};

      /**
       * @private
       * @type {Collection<import("../Feature.js").default<Geometry>>|null}
       */
      this.featuresCollection_ = null;

      /** @type {Collection<import("../Feature.js").default<Geometry>>} */
      let collection;
      /** @type {Array<import("../Feature.js").default<Geometry>>} */
      let features;
      if (Array.isArray(options.features)) {
        features = options.features;
      } else if (options.features) {
        collection = options.features;
        features = collection.getArray();
      }
      if (!useSpatialIndex && collection === undefined) {
        collection = new Collection(features);
      }
      if (features !== undefined) {
        this.addFeaturesInternal(features);
      }
      if (collection !== undefined) {
        this.bindFeaturesCollection_(collection);
      }
    }

    /**
     * Add a single feature to the source.  If you want to add a batch of features
     * at once, call {@link module:ol/source/Vector~VectorSource#addFeatures #addFeatures()}
     * instead. A feature will not be added to the source if feature with
     * the same id is already there. The reason for this behavior is to avoid
     * feature duplication when using bbox or tile loading strategies.
     * Note: this also applies if an {@link module:ol/Collection~Collection} is used for features,
     * meaning that if a feature with a duplicate id is added in the collection, it will
     * be removed from it right away.
     * @param {import("../Feature.js").default<Geometry>} feature Feature to add.
     * @api
     */
    addFeature(feature) {
      this.addFeatureInternal(feature);
      this.changed();
    }

    /**
     * Add a feature without firing a `change` event.
     * @param {import("../Feature.js").default<Geometry>} feature Feature.
     * @protected
     */
    addFeatureInternal(feature) {
      const featureKey = getUid(feature);

      if (!this.addToIndex_(featureKey, feature)) {
        if (this.featuresCollection_) {
          this.featuresCollection_.remove(feature);
        }
        return;
      }

      this.setupChangeEvents_(featureKey, feature);

      const geometry = feature.getGeometry();
      if (geometry) {
        const extent = geometry.getExtent();
        if (this.featuresRtree_) {
          this.featuresRtree_.insert(extent, feature);
        }
      } else {
        this.nullGeometryFeatures_[featureKey] = feature;
      }

      this.dispatchEvent(
        new VectorSourceEvent(VectorEventType.ADDFEATURE, feature)
      );
    }

    /**
     * @param {string} featureKey Unique identifier for the feature.
     * @param {import("../Feature.js").default<Geometry>} feature The feature.
     * @private
     */
    setupChangeEvents_(featureKey, feature) {
      this.featureChangeKeys_[featureKey] = [
        listen(feature, EventType$1.CHANGE, this.handleFeatureChange_, this),
        listen(
          feature,
          ObjectEventType.PROPERTYCHANGE,
          this.handleFeatureChange_,
          this
        ),
      ];
    }

    /**
     * @param {string} featureKey Unique identifier for the feature.
     * @param {import("../Feature.js").default<Geometry>} feature The feature.
     * @return {boolean} The feature is "valid", in the sense that it is also a
     *     candidate for insertion into the Rtree.
     * @private
     */
    addToIndex_(featureKey, feature) {
      let valid = true;
      const id = feature.getId();
      if (id !== undefined) {
        if (!(id.toString() in this.idIndex_)) {
          this.idIndex_[id.toString()] = feature;
        } else {
          valid = false;
        }
      }
      if (valid) {
        assert(!(featureKey in this.uidIndex_), 30); // The passed `feature` was already added to the source
        this.uidIndex_[featureKey] = feature;
      }
      return valid;
    }

    /**
     * Add a batch of features to the source.
     * @param {Array<import("../Feature.js").default<Geometry>>} features Features to add.
     * @api
     */
    addFeatures(features) {
      this.addFeaturesInternal(features);
      this.changed();
    }

    /**
     * Add features without firing a `change` event.
     * @param {Array<import("../Feature.js").default<Geometry>>} features Features.
     * @protected
     */
    addFeaturesInternal(features) {
      const extents = [];
      const newFeatures = [];
      const geometryFeatures = [];

      for (let i = 0, length = features.length; i < length; i++) {
        const feature = features[i];
        const featureKey = getUid(feature);
        if (this.addToIndex_(featureKey, feature)) {
          newFeatures.push(feature);
        }
      }

      for (let i = 0, length = newFeatures.length; i < length; i++) {
        const feature = newFeatures[i];
        const featureKey = getUid(feature);
        this.setupChangeEvents_(featureKey, feature);

        const geometry = feature.getGeometry();
        if (geometry) {
          const extent = geometry.getExtent();
          extents.push(extent);
          geometryFeatures.push(feature);
        } else {
          this.nullGeometryFeatures_[featureKey] = feature;
        }
      }
      if (this.featuresRtree_) {
        this.featuresRtree_.load(extents, geometryFeatures);
      }

      if (this.hasListener(VectorEventType.ADDFEATURE)) {
        for (let i = 0, length = newFeatures.length; i < length; i++) {
          this.dispatchEvent(
            new VectorSourceEvent(VectorEventType.ADDFEATURE, newFeatures[i])
          );
        }
      }
    }

    /**
     * @param {!Collection<import("../Feature.js").default<Geometry>>} collection Collection.
     * @private
     */
    bindFeaturesCollection_(collection) {
      let modifyingCollection = false;
      this.addEventListener(
        VectorEventType.ADDFEATURE,
        /**
         * @param {VectorSourceEvent<Geometry>} evt The vector source event
         */
        function (evt) {
          if (!modifyingCollection) {
            modifyingCollection = true;
            collection.push(evt.feature);
            modifyingCollection = false;
          }
        }
      );
      this.addEventListener(
        VectorEventType.REMOVEFEATURE,
        /**
         * @param {VectorSourceEvent<Geometry>} evt The vector source event
         */
        function (evt) {
          if (!modifyingCollection) {
            modifyingCollection = true;
            collection.remove(evt.feature);
            modifyingCollection = false;
          }
        }
      );
      collection.addEventListener(
        CollectionEventType.ADD,
        /**
         * @param {import("../Collection.js").CollectionEvent<import("../Feature.js").default<Geometry>>} evt The collection event
         */
        function (evt) {
          if (!modifyingCollection) {
            modifyingCollection = true;
            this.addFeature(evt.element);
            modifyingCollection = false;
          }
        }.bind(this)
      );
      collection.addEventListener(
        CollectionEventType.REMOVE,
        /**
         * @param {import("../Collection.js").CollectionEvent<import("../Feature.js").default<Geometry>>} evt The collection event
         */
        function (evt) {
          if (!modifyingCollection) {
            modifyingCollection = true;
            this.removeFeature(evt.element);
            modifyingCollection = false;
          }
        }.bind(this)
      );
      this.featuresCollection_ = collection;
    }

    /**
     * Remove all features from the source.
     * @param {boolean} [fast] Skip dispatching of {@link module:ol/source/Vector.VectorSourceEvent#event:removefeature} events.
     * @api
     */
    clear(fast) {
      if (fast) {
        for (const featureId in this.featureChangeKeys_) {
          const keys = this.featureChangeKeys_[featureId];
          keys.forEach(unlistenByKey);
        }
        if (!this.featuresCollection_) {
          this.featureChangeKeys_ = {};
          this.idIndex_ = {};
          this.uidIndex_ = {};
        }
      } else {
        if (this.featuresRtree_) {
          const removeAndIgnoreReturn = function (feature) {
            this.removeFeatureInternal(feature);
          }.bind(this);
          this.featuresRtree_.forEach(removeAndIgnoreReturn);
          for (const id in this.nullGeometryFeatures_) {
            this.removeFeatureInternal(this.nullGeometryFeatures_[id]);
          }
        }
      }
      if (this.featuresCollection_) {
        this.featuresCollection_.clear();
      }

      if (this.featuresRtree_) {
        this.featuresRtree_.clear();
      }
      this.nullGeometryFeatures_ = {};

      const clearEvent = new VectorSourceEvent(VectorEventType.CLEAR);
      this.dispatchEvent(clearEvent);
      this.changed();
    }

    /**
     * Iterate through all features on the source, calling the provided callback
     * with each one.  If the callback returns any "truthy" value, iteration will
     * stop and the function will return the same value.
     * Note: this function only iterate through the feature that have a defined geometry.
     *
     * @param {function(import("../Feature.js").default<Geometry>): T} callback Called with each feature
     *     on the source.  Return a truthy value to stop iteration.
     * @return {T|undefined} The return value from the last call to the callback.
     * @template T
     * @api
     */
    forEachFeature(callback) {
      if (this.featuresRtree_) {
        return this.featuresRtree_.forEach(callback);
      } else if (this.featuresCollection_) {
        this.featuresCollection_.forEach(callback);
      }
    }

    /**
     * Iterate through all features whose geometries contain the provided
     * coordinate, calling the callback with each feature.  If the callback returns
     * a "truthy" value, iteration will stop and the function will return the same
     * value.
     *
     * @param {import("../coordinate.js").Coordinate} coordinate Coordinate.
     * @param {function(import("../Feature.js").default<Geometry>): T} callback Called with each feature
     *     whose goemetry contains the provided coordinate.
     * @return {T|undefined} The return value from the last call to the callback.
     * @template T
     */
    forEachFeatureAtCoordinateDirect(coordinate, callback) {
      const extent = [coordinate[0], coordinate[1], coordinate[0], coordinate[1]];
      return this.forEachFeatureInExtent(extent, function (feature) {
        const geometry = feature.getGeometry();
        if (geometry.intersectsCoordinate(coordinate)) {
          return callback(feature);
        } else {
          return undefined;
        }
      });
    }

    /**
     * Iterate through all features whose bounding box intersects the provided
     * extent (note that the feature's geometry may not intersect the extent),
     * calling the callback with each feature.  If the callback returns a "truthy"
     * value, iteration will stop and the function will return the same value.
     *
     * If you are interested in features whose geometry intersects an extent, call
     * the {@link module:ol/source/Vector~VectorSource#forEachFeatureIntersectingExtent #forEachFeatureIntersectingExtent()} method instead.
     *
     * When `useSpatialIndex` is set to false, this method will loop through all
     * features, equivalent to {@link module:ol/source/Vector~VectorSource#forEachFeature #forEachFeature()}.
     *
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {function(import("../Feature.js").default<Geometry>): T} callback Called with each feature
     *     whose bounding box intersects the provided extent.
     * @return {T|undefined} The return value from the last call to the callback.
     * @template T
     * @api
     */
    forEachFeatureInExtent(extent, callback) {
      if (this.featuresRtree_) {
        return this.featuresRtree_.forEachInExtent(extent, callback);
      } else if (this.featuresCollection_) {
        this.featuresCollection_.forEach(callback);
      }
    }

    /**
     * Iterate through all features whose geometry intersects the provided extent,
     * calling the callback with each feature.  If the callback returns a "truthy"
     * value, iteration will stop and the function will return the same value.
     *
     * If you only want to test for bounding box intersection, call the
     * {@link module:ol/source/Vector~VectorSource#forEachFeatureInExtent #forEachFeatureInExtent()} method instead.
     *
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {function(import("../Feature.js").default<Geometry>): T} callback Called with each feature
     *     whose geometry intersects the provided extent.
     * @return {T|undefined} The return value from the last call to the callback.
     * @template T
     * @api
     */
    forEachFeatureIntersectingExtent(extent, callback) {
      return this.forEachFeatureInExtent(
        extent,
        /**
         * @param {import("../Feature.js").default<Geometry>} feature Feature.
         * @return {T|undefined} The return value from the last call to the callback.
         */
        function (feature) {
          const geometry = feature.getGeometry();
          if (geometry.intersectsExtent(extent)) {
            const result = callback(feature);
            if (result) {
              return result;
            }
          }
        }
      );
    }

    /**
     * Get the features collection associated with this source. Will be `null`
     * unless the source was configured with `useSpatialIndex` set to `false`, or
     * with an {@link module:ol/Collection~Collection} as `features`.
     * @return {Collection<import("../Feature.js").default<Geometry>>|null} The collection of features.
     * @api
     */
    getFeaturesCollection() {
      return this.featuresCollection_;
    }

    /**
     * Get a snapshot of the features currently on the source in random order. The returned array
     * is a copy, the features are references to the features in the source.
     * @return {Array<import("../Feature.js").default<Geometry>>} Features.
     * @api
     */
    getFeatures() {
      let features;
      if (this.featuresCollection_) {
        features = this.featuresCollection_.getArray().slice(0);
      } else if (this.featuresRtree_) {
        features = this.featuresRtree_.getAll();
        if (!isEmpty$1(this.nullGeometryFeatures_)) {
          extend$2(features, Object.values(this.nullGeometryFeatures_));
        }
      }
      return /** @type {Array<import("../Feature.js").default<Geometry>>} */ (
        features
      );
    }

    /**
     * Get all features whose geometry intersects the provided coordinate.
     * @param {import("../coordinate.js").Coordinate} coordinate Coordinate.
     * @return {Array<import("../Feature.js").default<Geometry>>} Features.
     * @api
     */
    getFeaturesAtCoordinate(coordinate) {
      const features = [];
      this.forEachFeatureAtCoordinateDirect(coordinate, function (feature) {
        features.push(feature);
      });
      return features;
    }

    /**
     * Get all features whose bounding box intersects the provided extent.  Note that this returns an array of
     * all features intersecting the given extent in random order (so it may include
     * features whose geometries do not intersect the extent).
     *
     * When `useSpatialIndex` is set to false, this method will return all
     * features.
     *
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {import("../proj/Projection.js").default} [projection] Include features
     * where `extent` exceeds the x-axis bounds of `projection` and wraps around the world.
     * @return {Array<import("../Feature.js").default<Geometry>>} Features.
     * @api
     */
    getFeaturesInExtent(extent, projection) {
      if (this.featuresRtree_) {
        const multiWorld = projection && projection.canWrapX() && this.getWrapX();

        if (!multiWorld) {
          return this.featuresRtree_.getInExtent(extent);
        }

        const extents = wrapAndSliceX(extent, projection);

        return [].concat(
          ...extents.map((anExtent) => this.featuresRtree_.getInExtent(anExtent))
        );
      } else if (this.featuresCollection_) {
        return this.featuresCollection_.getArray().slice(0);
      } else {
        return [];
      }
    }

    /**
     * Get the closest feature to the provided coordinate.
     *
     * This method is not available when the source is configured with
     * `useSpatialIndex` set to `false`.
     * @param {import("../coordinate.js").Coordinate} coordinate Coordinate.
     * @param {function(import("../Feature.js").default<Geometry>):boolean} [filter] Feature filter function.
     *     The filter function will receive one argument, the {@link module:ol/Feature~Feature feature}
     *     and it should return a boolean value. By default, no filtering is made.
     * @return {import("../Feature.js").default<Geometry>} Closest feature.
     * @api
     */
    getClosestFeatureToCoordinate(coordinate, filter) {
      // Find the closest feature using branch and bound.  We start searching an
      // infinite extent, and find the distance from the first feature found.  This
      // becomes the closest feature.  We then compute a smaller extent which any
      // closer feature must intersect.  We continue searching with this smaller
      // extent, trying to find a closer feature.  Every time we find a closer
      // feature, we update the extent being searched so that any even closer
      // feature must intersect it.  We continue until we run out of features.
      const x = coordinate[0];
      const y = coordinate[1];
      let closestFeature = null;
      const closestPoint = [NaN, NaN];
      let minSquaredDistance = Infinity;
      const extent = [-Infinity, -Infinity, Infinity, Infinity];
      filter = filter ? filter : TRUE;
      this.featuresRtree_.forEachInExtent(
        extent,
        /**
         * @param {import("../Feature.js").default<Geometry>} feature Feature.
         */
        function (feature) {
          if (filter(feature)) {
            const geometry = feature.getGeometry();
            const previousMinSquaredDistance = minSquaredDistance;
            minSquaredDistance = geometry.closestPointXY(
              x,
              y,
              closestPoint,
              minSquaredDistance
            );
            if (minSquaredDistance < previousMinSquaredDistance) {
              closestFeature = feature;
              // This is sneaky.  Reduce the extent that it is currently being
              // searched while the R-Tree traversal using this same extent object
              // is still in progress.  This is safe because the new extent is
              // strictly contained by the old extent.
              const minDistance = Math.sqrt(minSquaredDistance);
              extent[0] = x - minDistance;
              extent[1] = y - minDistance;
              extent[2] = x + minDistance;
              extent[3] = y + minDistance;
            }
          }
        }
      );
      return closestFeature;
    }

    /**
     * Get the extent of the features currently in the source.
     *
     * This method is not available when the source is configured with
     * `useSpatialIndex` set to `false`.
     * @param {import("../extent.js").Extent} [extent] Destination extent. If provided, no new extent
     *     will be created. Instead, that extent's coordinates will be overwritten.
     * @return {import("../extent.js").Extent} Extent.
     * @api
     */
    getExtent(extent) {
      return this.featuresRtree_.getExtent(extent);
    }

    /**
     * Get a feature by its identifier (the value returned by feature.getId()).
     * Note that the index treats string and numeric identifiers as the same.  So
     * `source.getFeatureById(2)` will return a feature with id `'2'` or `2`.
     *
     * @param {string|number} id Feature identifier.
     * @return {import("../Feature.js").default<Geometry>|null} The feature (or `null` if not found).
     * @api
     */
    getFeatureById(id) {
      const feature = this.idIndex_[id.toString()];
      return feature !== undefined ? feature : null;
    }

    /**
     * Get a feature by its internal unique identifier (using `getUid`).
     *
     * @param {string} uid Feature identifier.
     * @return {import("../Feature.js").default<Geometry>|null} The feature (or `null` if not found).
     */
    getFeatureByUid(uid) {
      const feature = this.uidIndex_[uid];
      return feature !== undefined ? feature : null;
    }

    /**
     * Get the format associated with this source.
     *
     * @return {import("../format/Feature.js").default|undefined} The feature format.
     * @api
     */
    getFormat() {
      return this.format_;
    }

    /**
     * @return {boolean} The source can have overlapping geometries.
     */
    getOverlaps() {
      return this.overlaps_;
    }

    /**
     * Get the url associated with this source.
     *
     * @return {string|import("../featureloader.js").FeatureUrlFunction|undefined} The url.
     * @api
     */
    getUrl() {
      return this.url_;
    }

    /**
     * @param {Event} event Event.
     * @private
     */
    handleFeatureChange_(event) {
      const feature = /** @type {import("../Feature.js").default<Geometry>} */ (
        event.target
      );
      const featureKey = getUid(feature);
      const geometry = feature.getGeometry();
      if (!geometry) {
        if (!(featureKey in this.nullGeometryFeatures_)) {
          if (this.featuresRtree_) {
            this.featuresRtree_.remove(feature);
          }
          this.nullGeometryFeatures_[featureKey] = feature;
        }
      } else {
        const extent = geometry.getExtent();
        if (featureKey in this.nullGeometryFeatures_) {
          delete this.nullGeometryFeatures_[featureKey];
          if (this.featuresRtree_) {
            this.featuresRtree_.insert(extent, feature);
          }
        } else {
          if (this.featuresRtree_) {
            this.featuresRtree_.update(extent, feature);
          }
        }
      }
      const id = feature.getId();
      if (id !== undefined) {
        const sid = id.toString();
        if (this.idIndex_[sid] !== feature) {
          this.removeFromIdIndex_(feature);
          this.idIndex_[sid] = feature;
        }
      } else {
        this.removeFromIdIndex_(feature);
        this.uidIndex_[featureKey] = feature;
      }
      this.changed();
      this.dispatchEvent(
        new VectorSourceEvent(VectorEventType.CHANGEFEATURE, feature)
      );
    }

    /**
     * Returns true if the feature is contained within the source.
     * @param {import("../Feature.js").default<Geometry>} feature Feature.
     * @return {boolean} Has feature.
     * @api
     */
    hasFeature(feature) {
      const id = feature.getId();
      if (id !== undefined) {
        return id in this.idIndex_;
      } else {
        return getUid(feature) in this.uidIndex_;
      }
    }

    /**
     * @return {boolean} Is empty.
     */
    isEmpty() {
      if (this.featuresRtree_) {
        return (
          this.featuresRtree_.isEmpty() && isEmpty$1(this.nullGeometryFeatures_)
        );
      }
      if (this.featuresCollection_) {
        return this.featuresCollection_.getLength() === 0;
      }
      return true;
    }

    /**
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {number} resolution Resolution.
     * @param {import("../proj/Projection.js").default} projection Projection.
     */
    loadFeatures(extent, resolution, projection) {
      const loadedExtentsRtree = this.loadedExtentsRtree_;
      const extentsToLoad = this.strategy_(extent, resolution, projection);
      for (let i = 0, ii = extentsToLoad.length; i < ii; ++i) {
        const extentToLoad = extentsToLoad[i];
        const alreadyLoaded = loadedExtentsRtree.forEachInExtent(
          extentToLoad,
          /**
           * @param {{extent: import("../extent.js").Extent}} object Object.
           * @return {boolean} Contains.
           */
          function (object) {
            return containsExtent(object.extent, extentToLoad);
          }
        );
        if (!alreadyLoaded) {
          ++this.loadingExtentsCount_;
          this.dispatchEvent(
            new VectorSourceEvent(VectorEventType.FEATURESLOADSTART)
          );
          this.loader_.call(
            this,
            extentToLoad,
            resolution,
            projection,
            function (features) {
              --this.loadingExtentsCount_;
              this.dispatchEvent(
                new VectorSourceEvent(
                  VectorEventType.FEATURESLOADEND,
                  undefined,
                  features
                )
              );
            }.bind(this),
            function () {
              --this.loadingExtentsCount_;
              this.dispatchEvent(
                new VectorSourceEvent(VectorEventType.FEATURESLOADERROR)
              );
            }.bind(this)
          );
          loadedExtentsRtree.insert(extentToLoad, {extent: extentToLoad.slice()});
        }
      }
      this.loading =
        this.loader_.length < 4 ? false : this.loadingExtentsCount_ > 0;
    }

    refresh() {
      this.clear(true);
      this.loadedExtentsRtree_.clear();
      super.refresh();
    }

    /**
     * Remove an extent from the list of loaded extents.
     * @param {import("../extent.js").Extent} extent Extent.
     * @api
     */
    removeLoadedExtent(extent) {
      const loadedExtentsRtree = this.loadedExtentsRtree_;
      let obj;
      loadedExtentsRtree.forEachInExtent(extent, function (object) {
        if (equals$2(object.extent, extent)) {
          obj = object;
          return true;
        }
      });
      if (obj) {
        loadedExtentsRtree.remove(obj);
      }
    }

    /**
     * Remove a single feature from the source.  If you want to remove all features
     * at once, use the {@link module:ol/source/Vector~VectorSource#clear #clear()} method
     * instead.
     * @param {import("../Feature.js").default<Geometry>} feature Feature to remove.
     * @api
     */
    removeFeature(feature) {
      if (!feature) {
        return;
      }
      const featureKey = getUid(feature);
      if (featureKey in this.nullGeometryFeatures_) {
        delete this.nullGeometryFeatures_[featureKey];
      } else {
        if (this.featuresRtree_) {
          this.featuresRtree_.remove(feature);
        }
      }
      const result = this.removeFeatureInternal(feature);
      if (result) {
        this.changed();
      }
    }

    /**
     * Remove feature without firing a `change` event.
     * @param {import("../Feature.js").default<Geometry>} feature Feature.
     * @return {import("../Feature.js").default<Geometry>|undefined} The removed feature
     *     (or undefined if the feature was not found).
     * @protected
     */
    removeFeatureInternal(feature) {
      const featureKey = getUid(feature);
      const featureChangeKeys = this.featureChangeKeys_[featureKey];
      if (!featureChangeKeys) {
        return;
      }
      featureChangeKeys.forEach(unlistenByKey);
      delete this.featureChangeKeys_[featureKey];
      const id = feature.getId();
      if (id !== undefined) {
        delete this.idIndex_[id.toString()];
      }
      delete this.uidIndex_[featureKey];
      this.dispatchEvent(
        new VectorSourceEvent(VectorEventType.REMOVEFEATURE, feature)
      );
      return feature;
    }

    /**
     * Remove a feature from the id index.  Called internally when the feature id
     * may have changed.
     * @param {import("../Feature.js").default<Geometry>} feature The feature.
     * @return {boolean} Removed the feature from the index.
     * @private
     */
    removeFromIdIndex_(feature) {
      let removed = false;
      for (const id in this.idIndex_) {
        if (this.idIndex_[id] === feature) {
          delete this.idIndex_[id];
          removed = true;
          break;
        }
      }
      return removed;
    }

    /**
     * Set the new loader of the source. The next render cycle will use the
     * new loader.
     * @param {import("../featureloader.js").FeatureLoader} loader The loader to set.
     * @api
     */
    setLoader(loader) {
      this.loader_ = loader;
    }

    /**
     * Points the source to a new url. The next render cycle will use the new url.
     * @param {string|import("../featureloader.js").FeatureUrlFunction} url Url.
     * @api
     */
    setUrl(url) {
      assert(this.format_, 7); // `format` must be set when `url` is set
      this.url_ = url;
      this.setLoader(xhr(url, this.format_));
    }
  }

  /**
   * @module ol/source/Cluster
   */

  /**
   * @typedef {Object} Options
   * @property {import("./Source.js").AttributionLike} [attributions] Attributions.
   * @property {number} [distance=20] Distance in pixels within which features will
   * be clustered together.
   * @property {number} [minDistance=0] Minimum distance in pixels between clusters.
   * Will be capped at the configured distance.
   * By default no minimum distance is guaranteed. This config can be used to avoid
   * overlapping icons. As a tradoff, the cluster feature's position will no longer be
   * the center of all its features.
   * @property {function(Feature):Point} [geometryFunction]
   * Function that takes an {@link module:ol/Feature~Feature} as argument and returns an
   * {@link module:ol/geom/Point~Point} as cluster calculation point for the feature. When a
   * feature should not be considered for clustering, the function should return
   * `null`. The default, which works when the underlying source contains point
   * features only, is
   * ```js
   * function(feature) {
   *   return feature.getGeometry();
   * }
   * ```
   * See {@link module:ol/geom/Polygon~Polygon#getInteriorPoint} for a way to get a cluster
   * calculation point for polygons.
   * @property {function(Point, Array<Feature>):Feature} [createCluster]
   * Function that takes the cluster's center {@link module:ol/geom/Point~Point} and an array
   * of {@link module:ol/Feature~Feature} included in this cluster. Must return a
   * {@link module:ol/Feature~Feature} that will be used to render. Default implementation is:
   * ```js
   * function(point, features) {
   *   return new Feature({
   *     geometry: point,
   *     features: features
   *   });
   * }
   * ```
   * @property {VectorSource} [source=null] Source.
   * @property {boolean} [wrapX=true] Whether to wrap the world horizontally.
   */

  /**
   * @classdesc
   * Layer source to cluster vector data. Works out of the box with point
   * geometries. For other geometry types, or if not all geometries should be
   * considered for clustering, a custom `geometryFunction` can be defined.
   *
   * If the instance is disposed without also disposing the underlying
   * source `setSource(null)` has to be called to remove the listener reference
   * from the wrapped source.
   * @api
   */
  class Cluster extends VectorSource {
    /**
     * @param {Options} options Cluster options.
     */
    constructor(options) {
      super({
        attributions: options.attributions,
        wrapX: options.wrapX,
      });

      /**
       * @type {number|undefined}
       * @protected
       */
      this.resolution = undefined;

      /**
       * @type {number}
       * @protected
       */
      this.distance = options.distance !== undefined ? options.distance : 20;

      /**
       * @type {number}
       * @protected
       */
      this.minDistance = options.minDistance || 0;

      /**
       * @type {number}
       * @protected
       */
      this.interpolationRatio = 0;

      /**
       * @type {Array<Feature>}
       * @protected
       */
      this.features = [];

      /**
       * @param {Feature} feature Feature.
       * @return {Point} Cluster calculation point.
       * @protected
       */
      this.geometryFunction =
        options.geometryFunction ||
        function (feature) {
          const geometry = /** @type {Point} */ (feature.getGeometry());
          assert(geometry.getType() == 'Point', 10); // The default `geometryFunction` can only handle `Point` geometries
          return geometry;
        };

      /**
       * @type {function(Point, Array<Feature>):Feature}
       * @private
       */
      this.createCustomCluster_ = options.createCluster;

      /**
       * @type {VectorSource|null}
       * @protected
       */
      this.source = null;

      /**
       * @private
       */
      this.boundRefresh_ = this.refresh.bind(this);

      this.updateDistance(this.distance, this.minDistance);
      this.setSource(options.source || null);
    }

    /**
     * Remove all features from the source.
     * @param {boolean} [fast] Skip dispatching of {@link module:ol/source/VectorEventType~VectorEventType#removefeature} events.
     * @api
     */
    clear(fast) {
      this.features.length = 0;
      super.clear(fast);
    }

    /**
     * Get the distance in pixels between clusters.
     * @return {number} Distance.
     * @api
     */
    getDistance() {
      return this.distance;
    }

    /**
     * Get a reference to the wrapped source.
     * @return {VectorSource|null} Source.
     * @api
     */
    getSource() {
      return this.source;
    }

    /**
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {number} resolution Resolution.
     * @param {import("../proj/Projection.js").default} projection Projection.
     */
    loadFeatures(extent, resolution, projection) {
      this.source.loadFeatures(extent, resolution, projection);
      if (resolution !== this.resolution) {
        this.resolution = resolution;
        this.refresh();
      }
    }

    /**
     * Set the distance within which features will be clusterd together.
     * @param {number} distance The distance in pixels.
     * @api
     */
    setDistance(distance) {
      this.updateDistance(distance, this.minDistance);
    }

    /**
     * Set the minimum distance between clusters. Will be capped at the
     * configured distance.
     * @param {number} minDistance The minimum distance in pixels.
     * @api
     */
    setMinDistance(minDistance) {
      this.updateDistance(this.distance, minDistance);
    }

    /**
     * The configured minimum distance between clusters.
     * @return {number} The minimum distance in pixels.
     * @api
     */
    getMinDistance() {
      return this.minDistance;
    }

    /**
     * Replace the wrapped source.
     * @param {VectorSource|null} source The new source for this instance.
     * @api
     */
    setSource(source) {
      if (this.source) {
        this.source.removeEventListener(EventType$1.CHANGE, this.boundRefresh_);
      }
      this.source = source;
      if (source) {
        source.addEventListener(EventType$1.CHANGE, this.boundRefresh_);
      }
      this.refresh();
    }

    /**
     * Handle the source changing.
     */
    refresh() {
      this.clear();
      this.cluster();
      this.addFeatures(this.features);
    }

    /**
     * Update the distances and refresh the source if necessary.
     * @param {number} distance The new distance.
     * @param {number} minDistance The new minimum distance.
     */
    updateDistance(distance, minDistance) {
      const ratio =
        distance === 0 ? 0 : Math.min(minDistance, distance) / distance;
      const changed =
        distance !== this.distance || this.interpolationRatio !== ratio;
      this.distance = distance;
      this.minDistance = minDistance;
      this.interpolationRatio = ratio;
      if (changed) {
        this.refresh();
      }
    }

    /**
     * @protected
     */
    cluster() {
      if (this.resolution === undefined || !this.source) {
        return;
      }
      const extent = createEmpty();
      const mapDistance = this.distance * this.resolution;
      const features = this.source.getFeatures();

      /** @type {Object<string, true>} */
      const clustered = {};

      for (let i = 0, ii = features.length; i < ii; i++) {
        const feature = features[i];
        if (!(getUid(feature) in clustered)) {
          const geometry = this.geometryFunction(feature);
          if (geometry) {
            const coordinates = geometry.getCoordinates();
            createOrUpdateFromCoordinate(coordinates, extent);
            buffer$1(extent, mapDistance, extent);

            const neighbors = this.source
              .getFeaturesInExtent(extent)
              .filter(function (neighbor) {
                const uid = getUid(neighbor);
                if (uid in clustered) {
                  return false;
                }
                clustered[uid] = true;
                return true;
              });
            this.features.push(this.createCluster(neighbors, extent));
          }
        }
      }
    }

    /**
     * @param {Array<Feature>} features Features
     * @param {import("../extent.js").Extent} extent The searched extent for these features.
     * @return {Feature} The cluster feature.
     * @protected
     */
    createCluster(features, extent) {
      const centroid = [0, 0];
      for (let i = features.length - 1; i >= 0; --i) {
        const geometry = this.geometryFunction(features[i]);
        if (geometry) {
          add(centroid, geometry.getCoordinates());
        } else {
          features.splice(i, 1);
        }
      }
      scale$2(centroid, 1 / features.length);
      const searchCenter = getCenter(extent);
      const ratio = this.interpolationRatio;
      const geometry = new Point([
        centroid[0] * (1 - ratio) + searchCenter[0] * ratio,
        centroid[1] * (1 - ratio) + searchCenter[1] * ratio,
      ]);
      if (this.createCustomCluster_) {
        return this.createCustomCluster_(geometry, features);
      } else {
        return new Feature({
          geometry,
          features,
        });
      }
    }
  }

  /**
   * @module ol/source/DataTile
   */

  /**
   * Data tile loading function.  The function is called with z, x, and y tile coordinates and
   * returns {@link import("../DataTile.js").Data data} for a tile or a promise for the same.
   * @typedef {function(number, number, number) : (import("../DataTile.js").Data|Promise<import("../DataTile.js").Data>)} Loader
   */

  /**
   * @typedef {Object} Options
   * @property {Loader} [loader] Data loader.  Called with z, x, and y tile coordinates.
   * Returns {@link import("../DataTile.js").Data data} for a tile or a promise for the same.
   * @property {import("./Source.js").AttributionLike} [attributions] Attributions.
   * @property {boolean} [attributionsCollapsible=true] Attributions are collapsible.
   * @property {number} [maxZoom=42] Optional max zoom level. Not used if `tileGrid` is provided.
   * @property {number} [minZoom=0] Optional min zoom level. Not used if `tileGrid` is provided.
   * @property {number|import("../size.js").Size} [tileSize=[256, 256]] The pixel width and height of the source tiles.
   * This may be different than the rendered pixel size if a `tileGrid` is provided.
   * @property {number} [gutter=0] The size in pixels of the gutter around data tiles to ignore.
   * This allows artifacts of rendering at tile edges to be ignored.
   * Supported data should be wider and taller than the tile size by a value of `2 x gutter`.
   * @property {number} [maxResolution] Optional tile grid resolution at level zero. Not used if `tileGrid` is provided.
   * @property {import("../proj.js").ProjectionLike} [projection='EPSG:3857'] Tile projection.
   * @property {import("../tilegrid/TileGrid.js").default} [tileGrid] Tile grid.
   * @property {boolean} [opaque=false] Whether the layer is opaque.
   * @property {import("./Source.js").State} [state] The source state.
   * @property {boolean} [wrapX=false] Render tiles beyond the antimeridian.
   * @property {number} [transition] Transition time when fading in new tiles (in miliseconds).
   * @property {number} [bandCount=4] Number of bands represented in the data.
   * @property {boolean} [interpolate=false] Use interpolated values when resampling.  By default,
   * the nearest neighbor is used when resampling.
   */

  /**
   * @classdesc
   * A source for typed array data tiles.
   *
   * @fires import("./Tile.js").TileSourceEvent
   * @api
   */
  class DataTileSource extends TileSource {
    /**
     * @param {Options} options Image tile options.
     */
    constructor(options) {
      const projection =
        options.projection === undefined ? 'EPSG:3857' : options.projection;

      let tileGrid = options.tileGrid;
      if (tileGrid === undefined && projection) {
        tileGrid = createXYZ({
          extent: extentFromProjection(projection),
          maxResolution: options.maxResolution,
          maxZoom: options.maxZoom,
          minZoom: options.minZoom,
          tileSize: options.tileSize,
        });
      }

      super({
        cacheSize: 0.1, // don't cache on the source
        attributions: options.attributions,
        attributionsCollapsible: options.attributionsCollapsible,
        projection: projection,
        tileGrid: tileGrid,
        opaque: options.opaque,
        state: options.state,
        wrapX: options.wrapX,
        transition: options.transition,
        interpolate: options.interpolate,
      });

      /**
       * @private
       * @type {number}
       */
      this.gutter_ = options.gutter !== undefined ? options.gutter : 0;

      /**
       * @private
       * @type {import('../size.js').Size|null}
       */
      this.tileSize_ = options.tileSize ? toSize(options.tileSize) : null;

      /**
       * @private
       * @type {Array<import('../size.js').Size>|null}
       */
      this.tileSizes_ = null;

      /**
       * @private
       * @type {!Object<string, boolean>}
       */
      this.tileLoadingKeys_ = {};

      /**
       * @private
       */
      this.loader_ = options.loader;

      this.handleTileChange_ = this.handleTileChange_.bind(this);

      /**
       * @type {number}
       */
      this.bandCount = options.bandCount === undefined ? 4 : options.bandCount; // assume RGBA if undefined
    }

    /**
     * Set the source tile sizes.  The length of the array is expected to match the number of
     * levels in the tile grid.
     * @protected
     * @param {Array<import('../size.js').Size>} tileSizes An array of tile sizes.
     */
    setTileSizes(tileSizes) {
      this.tileSizes_ = tileSizes;
    }

    /**
     * Get the source tile size at the given zoom level.  This may be different than the rendered tile
     * size.
     * @protected
     * @param {number} z Tile zoom level.
     * @return {import('../size.js').Size} The source tile size.
     */
    getTileSize(z) {
      if (this.tileSizes_) {
        return this.tileSizes_[z];
      }
      if (this.tileSize_) {
        return this.tileSize_;
      }
      const tileGrid = this.getTileGrid();
      return tileGrid ? toSize(tileGrid.getTileSize(z)) : [256, 256];
    }

    /**
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @return {number} Gutter.
     */
    getGutterForProjection(projection) {
      return this.gutter_;
    }

    /**
     * @param {Loader} loader The data loader.
     * @protected
     */
    setLoader(loader) {
      this.loader_ = loader;
    }

    /**
     * @param {number} z Tile coordinate z.
     * @param {number} x Tile coordinate x.
     * @param {number} y Tile coordinate y.
     * @param {number} pixelRatio Pixel ratio.
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @return {!DataTile} Tile.
     */
    getTile(z, x, y, pixelRatio, projection) {
      const size = this.getTileSize(z);
      const tileCoordKey = getKeyZXY(z, x, y);
      if (this.tileCache.containsKey(tileCoordKey)) {
        return this.tileCache.get(tileCoordKey);
      }

      const sourceLoader = this.loader_;

      function loader() {
        return toPromise(function () {
          return sourceLoader(z, x, y);
        });
      }

      const options = Object.assign(
        {
          tileCoord: [z, x, y],
          loader: loader,
          size: size,
        },
        this.tileOptions
      );

      const tile = new DataTile(options);
      tile.key = this.getKey();
      tile.addEventListener(EventType$1.CHANGE, this.handleTileChange_);

      this.tileCache.set(tileCoordKey, tile);
      return tile;
    }

    /**
     * Handle tile change events.
     * @param {import("../events/Event.js").default} event Event.
     */
    handleTileChange_(event) {
      const tile = /** @type {import("../Tile.js").default} */ (event.target);
      const uid = getUid(tile);
      const tileState = tile.getState();
      let type;
      if (tileState == TileState.LOADING) {
        this.tileLoadingKeys_[uid] = true;
        type = TileEventType.TILELOADSTART;
      } else if (uid in this.tileLoadingKeys_) {
        delete this.tileLoadingKeys_[uid];
        type =
          tileState == TileState.ERROR
            ? TileEventType.TILELOADERROR
            : tileState == TileState.LOADED
            ? TileEventType.TILELOADEND
            : undefined;
      }
      if (type) {
        this.dispatchEvent(new TileSourceEvent(type, tile));
      }
    }
  }

  /**
   * @module ol/source/GeoTIFF
   */

  /**
   * Determine if an image type is a mask.
   * See https://www.awaresystems.be/imaging/tiff/tifftags/newsubfiletype.html
   * @param {GeoTIFFImage} image The image.
   * @return {boolean} The image is a mask.
   */
  function isMask(image) {
    const fileDirectory = image.fileDirectory;
    const type = fileDirectory.NewSubfileType || 0;
    return (type & 4) === 4;
  }

  /**
   * @typedef {Object} SourceInfo
   * @property {string} [url] URL for the source GeoTIFF.
   * @property {Array<string>} [overviews] List of any overview URLs, only applies if the url parameter is given.
   * @property {Blob} [blob] Blob containing the source GeoTIFF. `blob` and `url` are mutually exclusive.
   * @property {number} [min=0] The minimum source data value.  Rendered values are scaled from 0 to 1 based on
   * the configured min and max.  If not provided and raster statistics are available, those will be used instead.
   * If neither are available, the minimum for the data type will be used.  To disable this behavior, set
   * the `normalize` option to `false` in the constructor.
   * @property {number} [max] The maximum source data value.  Rendered values are scaled from 0 to 1 based on
   * the configured min and max.  If not provided and raster statistics are available, those will be used instead.
   * If neither are available, the maximum for the data type will be used.  To disable this behavior, set
   * the `normalize` option to `false` in the constructor.
   * @property {number} [nodata] Values to discard (overriding any nodata values in the metadata).
   * When provided, an additional alpha band will be added to the data.  Often the GeoTIFF metadata
   * will include information about nodata values, so you should only need to set this property if
   * you find that it is not already extracted from the metadata.
   * @property {Array<number>} [bands] Band numbers to be read from (where the first band is `1`). If not provided, all bands will
   * be read. For example, if a GeoTIFF has blue (1), green (2), red (3), and near-infrared (4) bands, and you only need the
   * near-infrared band, configure `bands: [4]`.
   */

  /**
   * @typedef {Object} GeoKeys
   * @property {number} GTModelTypeGeoKey Model type.
   * @property {number} GTRasterTypeGeoKey Raster type.
   * @property {number} GeogAngularUnitsGeoKey Angular units.
   * @property {number} GeogInvFlatteningGeoKey Inverse flattening.
   * @property {number} GeogSemiMajorAxisGeoKey Semi-major axis.
   * @property {number} GeographicTypeGeoKey Geographic coordinate system code.
   * @property {number} ProjLinearUnitsGeoKey Projected linear unit code.
   * @property {number} ProjectedCSTypeGeoKey Projected coordinate system code.
   */

  /**
   * @typedef {import("geotiff").GeoTIFF} GeoTIFF
   */

  /**
   * @typedef {import("geotiff").MultiGeoTIFF} MultiGeoTIFF
   */

  /**
   * @typedef {Object} GDALMetadata
   * @property {string} STATISTICS_MINIMUM The minimum value (as a string).
   * @property {string} STATISTICS_MAXIMUM The maximum value (as a string).
   */

  const STATISTICS_MAXIMUM = 'STATISTICS_MAXIMUM';
  const STATISTICS_MINIMUM = 'STATISTICS_MINIMUM';

  /**
   * @typedef {import("geotiff").GeoTIFFImage} GeoTIFFImage
   */

  let workerPool;
  function getWorkerPool() {
    if (!workerPool) {
      workerPool = new geotiff.Pool();
    }
    return workerPool;
  }

  /**
   * Get the bounding box of an image.  If the image does not have an affine transform,
   * the pixel bounds are returned.
   * @param {GeoTIFFImage} image The image.
   * @return {Array<number>} The image bounding box.
   */
  function getBoundingBox(image) {
    try {
      return image.getBoundingBox();
    } catch (_) {
      const fileDirectory = image.fileDirectory;
      return [0, 0, fileDirectory.ImageWidth, fileDirectory.ImageLength];
    }
  }

  /**
   * Get the origin of an image.  If the image does not have an affine transform,
   * the top-left corner of the pixel bounds is returned.
   * @param {GeoTIFFImage} image The image.
   * @return {Array<number>} The image origin.
   */
  function getOrigin(image) {
    try {
      return image.getOrigin().slice(0, 2);
    } catch (_) {
      return [0, image.fileDirectory.ImageLength];
    }
  }

  /**
   * Get the resolution of an image.  If the image does not have an affine transform,
   * the width of the image is compared with the reference image.
   * @param {GeoTIFFImage} image The image.
   * @param {GeoTIFFImage} referenceImage The reference image.
   * @return {Array<number>} The map x and y units per pixel.
   */
  function getResolutions(image, referenceImage) {
    try {
      return image.getResolution(referenceImage);
    } catch (_) {
      return [
        referenceImage.fileDirectory.ImageWidth / image.fileDirectory.ImageWidth,
        referenceImage.fileDirectory.ImageHeight /
          image.fileDirectory.ImageHeight,
      ];
    }
  }

  /**
   * @param {GeoTIFFImage} image A GeoTIFF.
   * @return {import("../proj/Projection.js").default} The image projection.
   */
  function getProjection(image) {
    const geoKeys = image.geoKeys;
    if (!geoKeys) {
      return null;
    }

    if (geoKeys.ProjectedCSTypeGeoKey) {
      const code = 'EPSG:' + geoKeys.ProjectedCSTypeGeoKey;
      let projection = get$2(code);
      if (!projection) {
        const units = fromCode(geoKeys.ProjLinearUnitsGeoKey);
        if (units) {
          projection = new Projection({
            code: code,
            units: units,
          });
        }
      }
      return projection;
    }

    if (geoKeys.GeographicTypeGeoKey) {
      const code = 'EPSG:' + geoKeys.GeographicTypeGeoKey;
      let projection = get$2(code);
      if (!projection) {
        const units = fromCode(geoKeys.GeogAngularUnitsGeoKey);
        if (units) {
          projection = new Projection({
            code: code,
            units: units,
          });
        }
      }
      return projection;
    }

    return null;
  }

  /**
   * @param {GeoTIFF|MultiGeoTIFF} tiff A GeoTIFF.
   * @return {Promise<Array<GeoTIFFImage>>} Resolves to a list of images.
   */
  function getImagesForTIFF(tiff) {
    return tiff.getImageCount().then(function (count) {
      const requests = new Array(count);
      for (let i = 0; i < count; ++i) {
        requests[i] = tiff.getImage(i);
      }
      return Promise.all(requests);
    });
  }

  /**
   * @param {SourceInfo} source The GeoTIFF source.
   * @param {Object} options Options for the GeoTIFF source.
   * @return {Promise<Array<GeoTIFFImage>>} Resolves to a list of images.
   */
  function getImagesForSource(source, options) {
    let request;
    if (source.blob) {
      request = geotiff.fromBlob(source.blob);
    } else if (source.overviews) {
      request = geotiff.fromUrls(source.url, source.overviews, options);
    } else {
      request = geotiff.fromUrl(source.url, options);
    }
    return request.then(getImagesForTIFF);
  }

  /**
   * @param {number|Array<number>|Array<Array<number>>} expected Expected value.
   * @param {number|Array<number>|Array<Array<number>>} got Actual value.
   * @param {number} tolerance Accepted tolerance in fraction of expected between expected and got.
   * @param {string} message The error message.
   * @param {function(Error):void} rejector A function to be called with any error.
   */
  function assertEqual(expected, got, tolerance, message, rejector) {
    if (Array.isArray(expected)) {
      const length = expected.length;
      if (!Array.isArray(got) || length != got.length) {
        const error = new Error(message);
        rejector(error);
        throw error;
      }
      for (let i = 0; i < length; ++i) {
        assertEqual(expected[i], got[i], tolerance, message, rejector);
      }
      return;
    }

    got = /** @type {number} */ (got);
    if (Math.abs(expected - got) > tolerance * expected) {
      throw new Error(message);
    }
  }

  /**
   * @param {Array} array The data array.
   * @return {number} The minimum value.
   */
  function getMinForDataType(array) {
    if (array instanceof Int8Array) {
      return -128;
    }
    if (array instanceof Int16Array) {
      return -32768;
    }
    if (array instanceof Int32Array) {
      return -2147483648;
    }
    if (array instanceof Float32Array) {
      return 1.2e-38;
    }
    return 0;
  }

  /**
   * @param {Array} array The data array.
   * @return {number} The maximum value.
   */
  function getMaxForDataType(array) {
    if (array instanceof Int8Array) {
      return 127;
    }
    if (array instanceof Uint8Array) {
      return 255;
    }
    if (array instanceof Uint8ClampedArray) {
      return 255;
    }
    if (array instanceof Int16Array) {
      return 32767;
    }
    if (array instanceof Uint16Array) {
      return 65535;
    }
    if (array instanceof Int32Array) {
      return 2147483647;
    }
    if (array instanceof Uint32Array) {
      return 4294967295;
    }
    if (array instanceof Float32Array) {
      return 3.4e38;
    }
    return 255;
  }

  /**
   * @typedef {Object} GeoTIFFSourceOptions
   * @property {boolean} [forceXHR=false] Whether to force the usage of the browsers XMLHttpRequest API.
   * @property {Object<string, string>} [headers] additional key-value pairs of headers to be passed with each request. Key is the header name, value the header value.
   * @property {string} [credentials] How credentials shall be handled. See
   * https://developer.mozilla.org/en-US/docs/Web/API/fetch for reference and possible values
   * @property {number} [maxRanges] The maximum amount of ranges to request in a single multi-range request.
   * By default only a single range is used.
   * @property {boolean} [allowFullFile=false] Whether or not a full file is accepted when only a portion is
   * requested. Only use this when you know the source image to be small enough to fit in memory.
   * @property {number} [blockSize=65536] The block size to use.
   * @property {number} [cacheSize=100] The number of blocks that shall be held in a LRU cache.
   */

  /**
   * @typedef {Object} Options
   * @property {Array<SourceInfo>} sources List of information about GeoTIFF sources.
   * Multiple sources can be combined when their resolution sets are equal after applying a scale.
   * The list of sources defines a mapping between input bands as they are read from each GeoTIFF and
   * the output bands that are provided by data tiles. To control which bands to read from each GeoTIFF,
   * use the {@link import("./GeoTIFF.js").SourceInfo bands} property. If, for example, you specify two
   * sources, one with 3 bands and {@link import("./GeoTIFF.js").SourceInfo nodata} configured, and
   * another with 1 band, the resulting data tiles will have 5 bands: 3 from the first source, 1 alpha
   * band from the first source, and 1 band from the second source.
   * @property {GeoTIFFSourceOptions} [sourceOptions] Additional options to be passed to [geotiff.js](https://geotiffjs.github.io/geotiff.js/module-geotiff.html)'s `fromUrl` or `fromUrls` methods.
   * @property {boolean} [convertToRGB = false] By default, bands from the sources are read as-is. When
   * reading GeoTIFFs with the purpose of displaying them as RGB images, setting this to `true` will
   * convert other color spaces (YCbCr, CMYK) to RGB.
   * @property {boolean} [normalize=true] By default, the source data is normalized to values between
   * 0 and 1 with scaling factors based on the raster statistics or `min` and `max` properties of each source.
   * If instead you want to work with the raw values in a style expression, set this to `false`.  Setting this option
   * to `false` will make it so any `min` and `max` properties on sources are ignored.
   * @property {boolean} [opaque=false] Whether the layer is opaque.
   * @property {number} [transition=250] Duration of the opacity transition for rendering.
   * To disable the opacity transition, pass `transition: 0`.
   * @property {boolean} [wrapX=false] Render tiles beyond the tile grid extent.
   * @property {boolean} [interpolate=true] Use interpolated values when resampling.  By default,
   * the linear interpolation is used to resample the data.  If false, nearest neighbor is used.
   */

  /**
   * @classdesc
   * A source for working with GeoTIFF data.
   * @api
   */
  class GeoTIFFSource extends DataTileSource {
    /**
     * @param {Options} options Data tile options.
     */
    constructor(options) {
      super({
        state: 'loading',
        tileGrid: null,
        projection: null,
        opaque: options.opaque,
        transition: options.transition,
        interpolate: options.interpolate !== false,
        wrapX: options.wrapX,
      });

      /**
       * @type {Array<SourceInfo>}
       * @private
       */
      this.sourceInfo_ = options.sources;

      const numSources = this.sourceInfo_.length;

      /**
       * @type {Object}
       * @private
       */
      this.sourceOptions_ = options.sourceOptions;

      /**
       * @type {Array<Array<GeoTIFFImage>>}
       * @private
       */
      this.sourceImagery_ = new Array(numSources);

      /**
       * @type {Array<Array<GeoTIFFImage>>}
       * @private
       */
      this.sourceMasks_ = new Array(numSources);

      /**
       * @type {Array<number>}
       * @private
       */
      this.resolutionFactors_ = new Array(numSources);

      /**
       * @type {Array<number>}
       * @private
       */
      this.samplesPerPixel_;

      /**
       * @type {Array<Array<number>>}
       * @private
       */
      this.nodataValues_;

      /**
       * @type {Array<Array<GDALMetadata>>}
       * @private
       */
      this.metadata_;

      /**
       * @type {boolean}
       * @private
       */
      this.normalize_ = options.normalize !== false;

      /**
       * @type {boolean}
       * @private
       */
      this.addAlpha_ = false;

      /**
       * @type {Error}
       * @private
       */
      this.error_ = null;

      /**
       * @type {'readRasters' | 'readRGB'}
       */
      this.readMethod_ = options.convertToRGB ? 'readRGB' : 'readRasters';

      this.setKey(this.sourceInfo_.map((source) => source.url).join(','));

      const self = this;
      const requests = new Array(numSources);
      for (let i = 0; i < numSources; ++i) {
        requests[i] = getImagesForSource(
          this.sourceInfo_[i],
          this.sourceOptions_
        );
      }
      Promise.all(requests)
        .then(function (sources) {
          self.configure_(sources);
        })
        .catch(function (error) {
          console.error(error); // eslint-disable-line no-console
          self.error_ = error;
          self.setState('error');
        });
    }

    /**
     * @return {Error} A source loading error. When the source state is `error`, use this function
     * to get more information about the error. To debug a faulty configuration, you may want to use
     * a listener like
     * ```js
     * geotiffSource.on('change', () => {
     *   if (geotiffSource.getState() === 'error') {
     *     console.error(geotiffSource.getError());
     *   }
     * });
     * ```
     */
    getError() {
      return this.error_;
    }

    /**
     * Configure the tile grid based on images within the source GeoTIFFs.  Each GeoTIFF
     * must have the same internal tiled structure.
     * @param {Array<Array<GeoTIFFImage>>} sources Each source is a list of images
     * from a single GeoTIFF.
     * @private
     */
    configure_(sources) {
      let extent;
      let origin;
      let commonRenderTileSizes;
      let commonSourceTileSizes;
      let resolutions;
      const samplesPerPixel = new Array(sources.length);
      const nodataValues = new Array(sources.length);
      const metadata = new Array(sources.length);
      let minZoom = 0;

      const sourceCount = sources.length;
      for (let sourceIndex = 0; sourceIndex < sourceCount; ++sourceIndex) {
        const images = [];
        const masks = [];
        sources[sourceIndex].forEach((item) => {
          if (isMask(item)) {
            masks.push(item);
          } else {
            images.push(item);
          }
        });

        const imageCount = images.length;
        if (masks.length > 0 && masks.length !== imageCount) {
          throw new Error(
            `Expected one mask per image found ${masks.length} masks and ${imageCount} images`
          );
        }

        let sourceExtent;
        let sourceOrigin;
        const sourceTileSizes = new Array(imageCount);
        const renderTileSizes = new Array(imageCount);
        const sourceResolutions = new Array(imageCount);

        nodataValues[sourceIndex] = new Array(imageCount);
        metadata[sourceIndex] = new Array(imageCount);

        for (let imageIndex = 0; imageIndex < imageCount; ++imageIndex) {
          const image = images[imageIndex];
          const nodataValue = image.getGDALNoData();
          metadata[sourceIndex][imageIndex] = image.getGDALMetadata(0);
          nodataValues[sourceIndex][imageIndex] = nodataValue;

          const wantedSamples = this.sourceInfo_[sourceIndex].bands;
          samplesPerPixel[sourceIndex] = wantedSamples
            ? wantedSamples.length
            : image.getSamplesPerPixel();
          const level = imageCount - (imageIndex + 1);

          if (!sourceExtent) {
            sourceExtent = getBoundingBox(image);
          }

          if (!sourceOrigin) {
            sourceOrigin = getOrigin(image);
          }

          const imageResolutions = getResolutions(image, images[0]);
          sourceResolutions[level] = imageResolutions[0];

          const sourceTileSize = [image.getTileWidth(), image.getTileHeight()];
          sourceTileSizes[level] = sourceTileSize;

          const aspectRatio = imageResolutions[0] / Math.abs(imageResolutions[1]);
          renderTileSizes[level] = [
            sourceTileSize[0],
            sourceTileSize[1] / aspectRatio,
          ];
        }

        if (!extent) {
          extent = sourceExtent;
        } else {
          getIntersection(extent, sourceExtent, extent);
        }

        if (!origin) {
          origin = sourceOrigin;
        } else {
          const message = `Origin mismatch for source ${sourceIndex}, got [${sourceOrigin}] but expected [${origin}]`;
          assertEqual(origin, sourceOrigin, 0, message, this.viewRejector);
        }

        if (!resolutions) {
          resolutions = sourceResolutions;
          this.resolutionFactors_[sourceIndex] = 1;
        } else {
          if (resolutions.length - minZoom > sourceResolutions.length) {
            minZoom = resolutions.length - sourceResolutions.length;
          }
          const resolutionFactor =
            resolutions[resolutions.length - 1] /
            sourceResolutions[sourceResolutions.length - 1];
          this.resolutionFactors_[sourceIndex] = resolutionFactor;
          const scaledSourceResolutions = sourceResolutions.map(
            (resolution) => (resolution *= resolutionFactor)
          );
          const message = `Resolution mismatch for source ${sourceIndex}, got [${scaledSourceResolutions}] but expected [${resolutions}]`;
          assertEqual(
            resolutions.slice(minZoom, resolutions.length),
            scaledSourceResolutions,
            0.02,
            message,
            this.viewRejector
          );
        }

        if (!commonRenderTileSizes) {
          commonRenderTileSizes = renderTileSizes;
        } else {
          assertEqual(
            commonRenderTileSizes.slice(minZoom, commonRenderTileSizes.length),
            renderTileSizes,
            0.01,
            `Tile size mismatch for source ${sourceIndex}`,
            this.viewRejector
          );
        }

        if (!commonSourceTileSizes) {
          commonSourceTileSizes = sourceTileSizes;
        } else {
          assertEqual(
            commonSourceTileSizes.slice(minZoom, commonSourceTileSizes.length),
            sourceTileSizes,
            0,
            `Tile size mismatch for source ${sourceIndex}`,
            this.viewRejector
          );
        }

        this.sourceImagery_[sourceIndex] = images.reverse();
        this.sourceMasks_[sourceIndex] = masks.reverse();
      }

      for (let i = 0, ii = this.sourceImagery_.length; i < ii; ++i) {
        const sourceImagery = this.sourceImagery_[i];
        while (sourceImagery.length < resolutions.length) {
          sourceImagery.unshift(undefined);
        }
      }

      if (!this.getProjection()) {
        const firstSource = sources[0];
        for (let i = firstSource.length - 1; i >= 0; --i) {
          const image = firstSource[i];
          const projection = getProjection(image);
          if (projection) {
            this.projection = projection;
            break;
          }
        }
      }

      this.samplesPerPixel_ = samplesPerPixel;
      this.nodataValues_ = nodataValues;
      this.metadata_ = metadata;

      // decide if we need to add an alpha band to handle nodata
      outer: for (let sourceIndex = 0; sourceIndex < sourceCount; ++sourceIndex) {
        // option 1: source is configured with a nodata value
        if (this.sourceInfo_[sourceIndex].nodata !== undefined) {
          this.addAlpha_ = true;
          break;
        }
        if (this.sourceMasks_[sourceIndex].length) {
          this.addAlpha_ = true;
          break;
        }

        const values = nodataValues[sourceIndex];

        // option 2: check image metadata for limited bands
        const bands = this.sourceInfo_[sourceIndex].bands;
        if (bands) {
          for (let i = 0; i < bands.length; ++i) {
            if (values[bands[i] - 1] !== null) {
              this.addAlpha_ = true;
              break outer;
            }
          }
          continue;
        }

        // option 3: check image metadata for all bands
        for (let imageIndex = 0; imageIndex < values.length; ++imageIndex) {
          if (values[imageIndex] !== null) {
            this.addAlpha_ = true;
            break outer;
          }
        }
      }

      const additionalBands = this.addAlpha_ ? 1 : 0;
      this.bandCount =
        samplesPerPixel.reduce((accumulator, value) => {
          accumulator += value;
          return accumulator;
        }, 0) + additionalBands;

      const tileGrid = new TileGrid({
        extent: extent,
        minZoom: minZoom,
        origin: origin,
        resolutions: resolutions,
        tileSizes: commonRenderTileSizes,
      });

      this.tileGrid = tileGrid;
      this.setTileSizes(commonSourceTileSizes);

      this.setLoader(this.loadTile_.bind(this));
      this.setState('ready');
      this.viewResolver({
        projection: this.projection,
        resolutions: resolutions,
        center: toUserCoordinate(getCenter(extent), this.projection),
        extent: toUserExtent(extent, this.projection),
        zoom: 0,
      });
    }

    /**
     * @param {number} z The z tile index.
     * @param {number} x The x tile index.
     * @param {number} y The y tile index.
     * @return {Promise} The composed tile data.
     * @private
     */
    loadTile_(z, x, y) {
      const sourceTileSize = this.getTileSize(z);
      const sourceCount = this.sourceImagery_.length;
      const requests = new Array(sourceCount * 2);
      const nodataValues = this.nodataValues_;
      const sourceInfo = this.sourceInfo_;
      const pool = getWorkerPool();
      for (let sourceIndex = 0; sourceIndex < sourceCount; ++sourceIndex) {
        const source = sourceInfo[sourceIndex];
        const resolutionFactor = this.resolutionFactors_[sourceIndex];
        const pixelBounds = [
          Math.round(x * (sourceTileSize[0] * resolutionFactor)),
          Math.round(y * (sourceTileSize[1] * resolutionFactor)),
          Math.round((x + 1) * (sourceTileSize[0] * resolutionFactor)),
          Math.round((y + 1) * (sourceTileSize[1] * resolutionFactor)),
        ];
        const image = this.sourceImagery_[sourceIndex][z];
        let samples;
        if (source.bands) {
          samples = source.bands.map(function (bandNumber) {
            return bandNumber - 1;
          });
        }

        /** @type {number|Array<number>} */
        let fillValue;
        if ('nodata' in source && source.nodata !== null) {
          fillValue = source.nodata;
        } else {
          if (!samples) {
            fillValue = nodataValues[sourceIndex];
          } else {
            fillValue = samples.map(function (sampleIndex) {
              return nodataValues[sourceIndex][sampleIndex];
            });
          }
        }

        requests[sourceIndex] = image[this.readMethod_]({
          window: pixelBounds,
          width: sourceTileSize[0],
          height: sourceTileSize[1],
          samples: samples,
          fillValue: fillValue,
          pool: pool,
          interleave: false,
        });

        // requests after `sourceCount` are for mask data (if any)
        const maskIndex = sourceCount + sourceIndex;
        const mask = this.sourceMasks_[sourceIndex][z];
        if (!mask) {
          requests[maskIndex] = Promise.resolve(null);
          continue;
        }

        requests[maskIndex] = mask.readRasters({
          window: pixelBounds,
          width: sourceTileSize[0],
          height: sourceTileSize[1],
          samples: [0],
          pool: pool,
          interleave: false,
        });
      }

      return Promise.all(requests)
        .then(this.composeTile_.bind(this, sourceTileSize))
        .catch(function (error) {
          console.error(error); // eslint-disable-line no-console
          throw error;
        });
    }

    /**
     * @param {import("../size.js").Size} sourceTileSize The source tile size.
     * @param {Array} sourceSamples The source samples.
     * @return {import("../DataTile.js").Data} The composed tile data.
     * @private
     */
    composeTile_(sourceTileSize, sourceSamples) {
      const metadata = this.metadata_;
      const sourceInfo = this.sourceInfo_;
      const sourceCount = this.sourceImagery_.length;
      const bandCount = this.bandCount;
      const samplesPerPixel = this.samplesPerPixel_;
      const nodataValues = this.nodataValues_;
      const normalize = this.normalize_;
      const addAlpha = this.addAlpha_;

      const pixelCount = sourceTileSize[0] * sourceTileSize[1];
      const dataLength = pixelCount * bandCount;

      /** @type {Uint8Array|Float32Array} */
      let data;
      if (normalize) {
        data = new Uint8Array(dataLength);
      } else {
        data = new Float32Array(dataLength);
      }

      let dataIndex = 0;
      for (let pixelIndex = 0; pixelIndex < pixelCount; ++pixelIndex) {
        let transparent = addAlpha;
        for (let sourceIndex = 0; sourceIndex < sourceCount; ++sourceIndex) {
          const source = sourceInfo[sourceIndex];

          let min = source.min;
          let max = source.max;
          let gain, bias;
          if (normalize) {
            const stats = metadata[sourceIndex][0];
            if (min === undefined) {
              if (stats && STATISTICS_MINIMUM in stats) {
                min = parseFloat(stats[STATISTICS_MINIMUM]);
              } else {
                min = getMinForDataType(sourceSamples[sourceIndex][0]);
              }
            }
            if (max === undefined) {
              if (stats && STATISTICS_MAXIMUM in stats) {
                max = parseFloat(stats[STATISTICS_MAXIMUM]);
              } else {
                max = getMaxForDataType(sourceSamples[sourceIndex][0]);
              }
            }

            gain = 255 / (max - min);
            bias = -min * gain;
          }

          for (
            let sampleIndex = 0;
            sampleIndex < samplesPerPixel[sourceIndex];
            ++sampleIndex
          ) {
            const sourceValue =
              sourceSamples[sourceIndex][sampleIndex][pixelIndex];

            let value;
            if (normalize) {
              value = clamp(gain * sourceValue + bias, 0, 255);
            } else {
              value = sourceValue;
            }

            if (!addAlpha) {
              data[dataIndex] = value;
            } else {
              let nodata = source.nodata;
              if (nodata === undefined) {
                let bandIndex;
                if (source.bands) {
                  bandIndex = source.bands[sampleIndex] - 1;
                } else {
                  bandIndex = sampleIndex;
                }
                nodata = nodataValues[sourceIndex][bandIndex];
              }

              const nodataIsNaN = isNaN(nodata);
              if (
                (!nodataIsNaN && sourceValue !== nodata) ||
                (nodataIsNaN && !isNaN(sourceValue))
              ) {
                transparent = false;
                data[dataIndex] = value;
              }
            }
            dataIndex++;
          }
          if (!transparent) {
            const maskIndex = sourceCount + sourceIndex;
            const mask = sourceSamples[maskIndex];
            if (mask && !mask[0][pixelIndex]) {
              transparent = true;
            }
          }
        }
        if (addAlpha) {
          if (!transparent) {
            data[dataIndex] = 255;
          }
          dataIndex++;
        }
      }

      return data;
    }
  }

  /**
   * Get a promise for view properties based on the source.  Use the result of this function
   * as the `view` option in a map constructor.
   *
   *     const source = new GeoTIFF(options);
   *
   *     const map = new Map({
   *       target: 'map',
   *       layers: [
   *         new TileLayer({
   *           source: source,
   *         }),
   *       ],
   *       view: source.getView(),
   *     });
   *
   * @function
   * @return {Promise<import("../View.js").ViewOptions>} A promise for view-related properties.
   * @api
   *
   */
  GeoTIFFSource.prototype.getView;

  /**
   * @module ol/source/Zoomify
   */

  /**
   * @typedef {'default' | 'truncated'} TierSizeCalculation
   */

  class CustomTile$1 extends ImageTile {
    /**
     * @param {import("../size.js").Size} tileSize Full tile size.
     * @param {import("../tilecoord.js").TileCoord} tileCoord Tile coordinate.
     * @param {import("../TileState.js").default} state State.
     * @param {string} src Image source URI.
     * @param {?string} crossOrigin Cross origin.
     * @param {import("../Tile.js").LoadFunction} tileLoadFunction Tile load function.
     * @param {import("../Tile.js").Options} [options] Tile options.
     */
    constructor(
      tileSize,
      tileCoord,
      state,
      src,
      crossOrigin,
      tileLoadFunction,
      options
    ) {
      super(tileCoord, state, src, crossOrigin, tileLoadFunction, options);

      /**
       * @private
       * @type {HTMLCanvasElement|HTMLImageElement|HTMLVideoElement}
       */
      this.zoomifyImage_ = null;

      /**
       * @type {import("../size.js").Size}
       */
      this.tileSize_ = tileSize;
    }

    /**
     * Get the image element for this tile.
     * @return {HTMLCanvasElement|HTMLImageElement|HTMLVideoElement} Image.
     */
    getImage() {
      if (this.zoomifyImage_) {
        return this.zoomifyImage_;
      }
      const image = super.getImage();
      if (this.state == TileState.LOADED) {
        const tileSize = this.tileSize_;
        if (image.width == tileSize[0] && image.height == tileSize[1]) {
          this.zoomifyImage_ = image;
          return image;
        } else {
          const context = createCanvasContext2D(tileSize[0], tileSize[1]);
          context.drawImage(image, 0, 0);
          this.zoomifyImage_ = context.canvas;
          return context.canvas;
        }
      } else {
        return image;
      }
    }
  }

  /**
   * @typedef {Object} Options
   * @property {import("./Source.js").AttributionLike} [attributions] Attributions.
   * @property {number} [cacheSize] Initial tile cache size. Will auto-grow to hold at least the number of tiles in the viewport.
   * @property {null|string} [crossOrigin] The `crossOrigin` attribute for loaded images.  Note that
   * you must provide a `crossOrigin` value  you want to access pixel data with the Canvas renderer.
   * See https://developer.mozilla.org/en-US/docs/Web/HTML/CORS_enabled_image for more detail.
   * @property {boolean} [interpolate=true] Use interpolated values when resampling.  By default,
   * linear interpolation is used when resampling.  Set to false to use the nearest neighbor instead.
   * @property {import("../proj.js").ProjectionLike} [projection] Projection.
   * @property {number} [tilePixelRatio] The pixel ratio used by the tile service. For example, if the tile service advertizes 256px by 256px tiles but actually sends 512px by 512px images (for retina/hidpi devices) then `tilePixelRatio` should be set to `2`
   * @property {number} [reprojectionErrorThreshold=0.5] Maximum allowed reprojection error (in pixels).
   * Higher values can increase reprojection performance, but decrease precision.
   * @property {string} url URL template or base URL of the Zoomify service.
   * A base URL is the fixed part
   * of the URL, excluding the tile group, z, x, and y folder structure, e.g.
   * `http://my.zoomify.info/IMAGE.TIF/`. A URL template must include
   * `{TileGroup}`, `{x}`, `{y}`, and `{z}` placeholders, e.g.
   * `http://my.zoomify.info/IMAGE.TIF/{TileGroup}/{z}-{x}-{y}.jpg`.
   * Internet Imaging Protocol (IIP) with JTL extension can be also used with
   * `{tileIndex}` and `{z}` placeholders, e.g.
   * `http://my.zoomify.info?FIF=IMAGE.TIF&JTL={z},{tileIndex}`.
   * A `{?-?}` template pattern, for example `subdomain{a-f}.domain.com`, may be
   * used instead of defining each one separately in the `urls` option.
   * @property {TierSizeCalculation} [tierSizeCalculation] Tier size calculation method: `default` or `truncated`.
   * @property {import("../size.js").Size} size Size.
   * @property {import("../extent.js").Extent} [extent] Extent for the TileGrid that is created.
   * Default sets the TileGrid in the
   * fourth quadrant, meaning extent is `[0, -height, width, 0]`. To change the
   * extent to the first quadrant (the default for OpenLayers 2) set the extent
   * as `[0, 0, width, height]`.
   * @property {number} [transition] Duration of the opacity transition for rendering.
   * To disable the opacity transition, pass `transition: 0`.
   * @property {number} [tileSize=256] Tile size. Same tile size is used for all zoom levels.
   * @property {number|import("../array.js").NearestDirectionFunction} [zDirection=0]
   * Choose whether to use tiles with a higher or lower zoom level when between integer
   * zoom levels. See {@link module:ol/tilegrid/TileGrid~TileGrid#getZForResolution}.
   */

  /**
   * @classdesc
   * Layer source for tile data in Zoomify format (both Zoomify and Internet
   * Imaging Protocol are supported).
   * @api
   */
  class Zoomify extends TileImage {
    /**
     * @param {Options} options Options.
     */
    constructor(options) {
      const size = options.size;
      const tierSizeCalculation =
        options.tierSizeCalculation !== undefined
          ? options.tierSizeCalculation
          : 'default';

      const tilePixelRatio = options.tilePixelRatio || 1;
      const imageWidth = size[0];
      const imageHeight = size[1];
      const tierSizeInTiles = [];
      const tileSize = options.tileSize || DEFAULT_TILE_SIZE;
      let tileSizeForTierSizeCalculation = tileSize * tilePixelRatio;

      switch (tierSizeCalculation) {
        case 'default':
          while (
            imageWidth > tileSizeForTierSizeCalculation ||
            imageHeight > tileSizeForTierSizeCalculation
          ) {
            tierSizeInTiles.push([
              Math.ceil(imageWidth / tileSizeForTierSizeCalculation),
              Math.ceil(imageHeight / tileSizeForTierSizeCalculation),
            ]);
            tileSizeForTierSizeCalculation += tileSizeForTierSizeCalculation;
          }
          break;
        case 'truncated':
          let width = imageWidth;
          let height = imageHeight;
          while (
            width > tileSizeForTierSizeCalculation ||
            height > tileSizeForTierSizeCalculation
          ) {
            tierSizeInTiles.push([
              Math.ceil(width / tileSizeForTierSizeCalculation),
              Math.ceil(height / tileSizeForTierSizeCalculation),
            ]);
            width >>= 1;
            height >>= 1;
          }
          break;
        default: // Unknown `tierSizeCalculation` configured
          assert(false, 53);
          break;
      }

      tierSizeInTiles.push([1, 1]);
      tierSizeInTiles.reverse();

      const resolutions = [tilePixelRatio];
      const tileCountUpToTier = [0];
      for (let i = 1, ii = tierSizeInTiles.length; i < ii; i++) {
        resolutions.push(tilePixelRatio << i);
        tileCountUpToTier.push(
          tierSizeInTiles[i - 1][0] * tierSizeInTiles[i - 1][1] +
            tileCountUpToTier[i - 1]
        );
      }
      resolutions.reverse();

      const tileGrid = new TileGrid({
        tileSize: tileSize,
        extent: options.extent || [0, -imageHeight, imageWidth, 0],
        resolutions: resolutions,
      });

      let url = options.url;
      if (url && !url.includes('{TileGroup}') && !url.includes('{tileIndex}')) {
        url += '{TileGroup}/{z}-{x}-{y}.jpg';
      }
      const urls = expandUrl(url);

      let tileWidth = tileSize * tilePixelRatio;

      /**
       * @param {string} template Template.
       * @return {import("../Tile.js").UrlFunction} Tile URL function.
       */
      function createFromTemplate(template) {
        return (
          /**
           * @param {import("../tilecoord.js").TileCoord} tileCoord Tile Coordinate.
           * @param {number} pixelRatio Pixel ratio.
           * @param {import("../proj/Projection.js").default} projection Projection.
           * @return {string|undefined} Tile URL.
           */
          function (tileCoord, pixelRatio, projection) {
            if (!tileCoord) {
              return undefined;
            } else {
              const tileCoordZ = tileCoord[0];
              const tileCoordX = tileCoord[1];
              const tileCoordY = tileCoord[2];
              const tileIndex =
                tileCoordX + tileCoordY * tierSizeInTiles[tileCoordZ][0];
              const tileGroup =
                ((tileIndex + tileCountUpToTier[tileCoordZ]) / tileWidth) | 0;
              const localContext = {
                'z': tileCoordZ,
                'x': tileCoordX,
                'y': tileCoordY,
                'tileIndex': tileIndex,
                'TileGroup': 'TileGroup' + tileGroup,
              };
              return template.replace(/\{(\w+?)\}/g, function (m, p) {
                return localContext[p];
              });
            }
          }
        );
      }

      const tileUrlFunction = createFromTileUrlFunctions(
        urls.map(createFromTemplate)
      );

      const ZoomifyTileClass = CustomTile$1.bind(
        null,
        toSize(tileSize * tilePixelRatio)
      );

      super({
        attributions: options.attributions,
        cacheSize: options.cacheSize,
        crossOrigin: options.crossOrigin,
        interpolate: options.interpolate,
        projection: options.projection,
        tilePixelRatio: tilePixelRatio,
        reprojectionErrorThreshold: options.reprojectionErrorThreshold,
        tileClass: ZoomifyTileClass,
        tileGrid: tileGrid,
        tileUrlFunction: tileUrlFunction,
        transition: options.transition,
      });

      /**
       * @type {number|import("../array.js").NearestDirectionFunction}
       */
      this.zDirection = options.zDirection;

      // Server retina tile detection (non-standard):
      // Try loading the center tile for the highest resolution. If it is not
      // available, we are dealing with retina tiles, and need to adjust the
      // tile url calculation.
      const tileUrl = tileGrid.getTileCoordForCoordAndResolution(
        getCenter(tileGrid.getExtent()),
        resolutions[resolutions.length - 1]
      );
      const testTileUrl = tileUrlFunction(tileUrl, 1, null);
      const image = new Image();
      image.addEventListener(
        'error',
        function () {
          tileWidth = tileSize;
          this.changed();
        }.bind(this)
      );
      image.src = testTileUrl;
    }
  }

  /**
   * @module ol/format/IIIFInfo
   */

  /**
   * @typedef {Object} PreferredOptions
   * @property {string} [format] Preferred image format. Will be used if the image information
   * indicates support for that format.
   * @property {string} [quality] IIIF image qualitiy.  Will be used if the image information
   * indicates support for that quality.
   */

  /**
   * @typedef {Object} SupportedFeatures
   * @property {Array<string>} [supports] Supported IIIF image size and region
   * calculation features.
   * @property {Array<string>} [formats] Supported image formats.
   * @property {Array<string>} [qualities] Supported IIIF image qualities.
   */

  /**
   * @typedef {Object} TileInfo
   * @property {Array<number>} scaleFactors Supported resolution scaling factors.
   * @property {number} width Tile width in pixels.
   * @property {number} [height] Tile height in pixels. Same as tile width if height is
   * not given.
   */

  /**
   * @typedef {Object} IiifProfile
   * @property {Array<string>} [formats] Supported image formats for the image service.
   * @property {Array<string>} [qualities] Supported IIIF image qualities.
   * @property {Array<string>} [supports] Supported features.
   * @property {number} [maxArea] Maximum area (pixels) available for this image service.
   * @property {number} [maxHeight] Maximum height.
   * @property {number} [maxWidth] Maximum width.
   */

  /**
   * @typedef {Object<string,string|number|Array<number|string|IiifProfile|Object<string, number>|TileInfo>>}
   *    ImageInformationResponse
   */

  /**
   * Enum representing the major IIIF Image API versions
   * @enum {string}
   */
  const Versions = {
    VERSION1: 'version1',
    VERSION2: 'version2',
    VERSION3: 'version3',
  };

  /**
   * Supported image formats, qualities and supported region / size calculation features
   * for different image API versions and compliance levels
   * @const
   * @type {Object<string, Object<string, SupportedFeatures>>}
   */
  const IIIF_PROFILE_VALUES = {};
  IIIF_PROFILE_VALUES[Versions.VERSION1] = {
    'level0': {
      supports: [],
      formats: [],
      qualities: ['native'],
    },
    'level1': {
      supports: ['regionByPx', 'sizeByW', 'sizeByH', 'sizeByPct'],
      formats: ['jpg'],
      qualities: ['native'],
    },
    'level2': {
      supports: [
        'regionByPx',
        'regionByPct',
        'sizeByW',
        'sizeByH',
        'sizeByPct',
        'sizeByConfinedWh',
        'sizeByWh',
      ],
      formats: ['jpg', 'png'],
      qualities: ['native', 'color', 'grey', 'bitonal'],
    },
  };
  IIIF_PROFILE_VALUES[Versions.VERSION2] = {
    'level0': {
      supports: [],
      formats: ['jpg'],
      qualities: ['default'],
    },
    'level1': {
      supports: ['regionByPx', 'sizeByW', 'sizeByH', 'sizeByPct'],
      formats: ['jpg'],
      qualities: ['default'],
    },
    'level2': {
      supports: [
        'regionByPx',
        'regionByPct',
        'sizeByW',
        'sizeByH',
        'sizeByPct',
        'sizeByConfinedWh',
        'sizeByDistortedWh',
        'sizeByWh',
      ],
      formats: ['jpg', 'png'],
      qualities: ['default', 'bitonal'],
    },
  };
  IIIF_PROFILE_VALUES[Versions.VERSION3] = {
    'level0': {
      supports: [],
      formats: ['jpg'],
      qualities: ['default'],
    },
    'level1': {
      supports: ['regionByPx', 'regionSquare', 'sizeByW', 'sizeByH', 'sizeByWh'],
      formats: ['jpg'],
      qualities: ['default'],
    },
    'level2': {
      supports: [
        'regionByPx',
        'regionSquare',
        'regionByPct',
        'sizeByW',
        'sizeByH',
        'sizeByPct',
        'sizeByConfinedWh',
        'sizeByWh',
      ],
      formats: ['jpg', 'png'],
      qualities: ['default'],
    },
  };
  IIIF_PROFILE_VALUES['none'] = {
    'none': {
      supports: [],
      formats: [],
      qualities: [],
    },
  };

  const COMPLIANCE_VERSION1 =
    /^https?:\/\/library\.stanford\.edu\/iiif\/image-api\/(?:1\.1\/)?compliance\.html#level[0-2]$/;
  const COMPLIANCE_VERSION2 =
    /^https?:\/\/iiif\.io\/api\/image\/2\/level[0-2](?:\.json)?$/;
  const COMPLIANCE_VERSION3 =
    /(^https?:\/\/iiif\.io\/api\/image\/3\/level[0-2](?:\.json)?$)|(^level[0-2]$)/;

  function generateVersion1Options(iiifInfo) {
    let levelProfile = iiifInfo.getComplianceLevelSupportedFeatures();
    // Version 1.0 and 1.1 do not require a profile.
    if (levelProfile === undefined) {
      levelProfile = IIIF_PROFILE_VALUES[Versions.VERSION1]['level0'];
    }
    return {
      url:
        iiifInfo.imageInfo['@id'] === undefined
          ? undefined
          : iiifInfo.imageInfo['@id'].replace(/\/?(?:info\.json)?$/g, ''),
      supports: levelProfile.supports,
      formats: [
        ...levelProfile.formats,
        iiifInfo.imageInfo.formats === undefined
          ? []
          : iiifInfo.imageInfo.formats,
      ],
      qualities: [
        ...levelProfile.qualities,
        iiifInfo.imageInfo.qualities === undefined
          ? []
          : iiifInfo.imageInfo.qualities,
      ],
      resolutions: iiifInfo.imageInfo.scale_factors,
      tileSize:
        iiifInfo.imageInfo.tile_width !== undefined
          ? iiifInfo.imageInfo.tile_height !== undefined
            ? [iiifInfo.imageInfo.tile_width, iiifInfo.imageInfo.tile_height]
            : [iiifInfo.imageInfo.tile_width, iiifInfo.imageInfo.tile_width]
          : iiifInfo.imageInfo.tile_height != undefined
          ? [iiifInfo.imageInfo.tile_height, iiifInfo.imageInfo.tile_height]
          : undefined,
    };
  }

  function generateVersion2Options(iiifInfo) {
    const levelProfile = iiifInfo.getComplianceLevelSupportedFeatures(),
      additionalProfile =
        Array.isArray(iiifInfo.imageInfo.profile) &&
        iiifInfo.imageInfo.profile.length > 1,
      profileSupports =
        additionalProfile && iiifInfo.imageInfo.profile[1].supports
          ? iiifInfo.imageInfo.profile[1].supports
          : [],
      profileFormats =
        additionalProfile && iiifInfo.imageInfo.profile[1].formats
          ? iiifInfo.imageInfo.profile[1].formats
          : [],
      profileQualities =
        additionalProfile && iiifInfo.imageInfo.profile[1].qualities
          ? iiifInfo.imageInfo.profile[1].qualities
          : [];
    return {
      url: iiifInfo.imageInfo['@id'].replace(/\/?(?:info\.json)?$/g, ''),
      sizes:
        iiifInfo.imageInfo.sizes === undefined
          ? undefined
          : iiifInfo.imageInfo.sizes.map(function (size) {
              return [size.width, size.height];
            }),
      tileSize:
        iiifInfo.imageInfo.tiles === undefined
          ? undefined
          : [
              iiifInfo.imageInfo.tiles.map(function (tile) {
                return tile.width;
              })[0],
              iiifInfo.imageInfo.tiles.map(function (tile) {
                return tile.height === undefined ? tile.width : tile.height;
              })[0],
            ],
      resolutions:
        iiifInfo.imageInfo.tiles === undefined
          ? undefined
          : iiifInfo.imageInfo.tiles.map(function (tile) {
              return tile.scaleFactors;
            })[0],
      supports: [...levelProfile.supports, ...profileSupports],
      formats: [...levelProfile.formats, ...profileFormats],
      qualities: [...levelProfile.qualities, ...profileQualities],
    };
  }

  function generateVersion3Options(iiifInfo) {
    const levelProfile = iiifInfo.getComplianceLevelSupportedFeatures(),
      formats =
        iiifInfo.imageInfo.extraFormats === undefined
          ? levelProfile.formats
          : [...levelProfile.formats, ...iiifInfo.imageInfo.extraFormats],
      preferredFormat =
        iiifInfo.imageInfo.preferredFormats !== undefined &&
        Array.isArray(iiifInfo.imageInfo.preferredFormats) &&
        iiifInfo.imageInfo.preferredFormats.length > 0
          ? iiifInfo.imageInfo.preferredFormats
              .filter(function (format) {
                return ['jpg', 'png', 'gif'].includes(format);
              })
              .reduce(function (acc, format) {
                return acc === undefined && formats.includes(format)
                  ? format
                  : acc;
              }, undefined)
          : undefined;
    return {
      url: iiifInfo.imageInfo['id'],
      sizes:
        iiifInfo.imageInfo.sizes === undefined
          ? undefined
          : iiifInfo.imageInfo.sizes.map(function (size) {
              return [size.width, size.height];
            }),
      tileSize:
        iiifInfo.imageInfo.tiles === undefined
          ? undefined
          : [
              iiifInfo.imageInfo.tiles.map(function (tile) {
                return tile.width;
              })[0],
              iiifInfo.imageInfo.tiles.map(function (tile) {
                return tile.height;
              })[0],
            ],
      resolutions:
        iiifInfo.imageInfo.tiles === undefined
          ? undefined
          : iiifInfo.imageInfo.tiles.map(function (tile) {
              return tile.scaleFactors;
            })[0],
      supports:
        iiifInfo.imageInfo.extraFeatures === undefined
          ? levelProfile.supports
          : [...levelProfile.supports, ...iiifInfo.imageInfo.extraFeatures],
      formats: formats,
      qualities:
        iiifInfo.imageInfo.extraQualities === undefined
          ? levelProfile.qualities
          : [...levelProfile.qualities, ...iiifInfo.imageInfo.extraQualities],
      preferredFormat: preferredFormat,
    };
  }

  const versionFunctions = {};
  versionFunctions[Versions.VERSION1] = generateVersion1Options;
  versionFunctions[Versions.VERSION2] = generateVersion2Options;
  versionFunctions[Versions.VERSION3] = generateVersion3Options;

  /**
   * @classdesc
   * Format for transforming IIIF Image API image information responses into
   * IIIF tile source ready options
   *
   * @api
   */
  class IIIFInfo {
    /**
     * @param {string|ImageInformationResponse} imageInfo
     * Deserialized image information JSON response object or JSON response as string
     */
    constructor(imageInfo) {
      this.setImageInfo(imageInfo);
    }

    /**
     * @param {string|ImageInformationResponse} imageInfo
     * Deserialized image information JSON response object or JSON response as string
     * @api
     */
    setImageInfo(imageInfo) {
      if (typeof imageInfo == 'string') {
        this.imageInfo = JSON.parse(imageInfo);
      } else {
        this.imageInfo = imageInfo;
      }
    }

    /**
     * @return {Versions} Major IIIF version.
     * @api
     */
    getImageApiVersion() {
      if (this.imageInfo === undefined) {
        return;
      }
      let context = this.imageInfo['@context'] || 'ol-no-context';
      if (typeof context == 'string') {
        context = [context];
      }
      for (let i = 0; i < context.length; i++) {
        switch (context[i]) {
          case 'http://library.stanford.edu/iiif/image-api/1.1/context.json':
          case 'http://iiif.io/api/image/1/context.json':
            return Versions.VERSION1;
          case 'http://iiif.io/api/image/2/context.json':
            return Versions.VERSION2;
          case 'http://iiif.io/api/image/3/context.json':
            return Versions.VERSION3;
          case 'ol-no-context':
            // Image API 1.0 has no '@context'
            if (
              this.getComplianceLevelEntryFromProfile(Versions.VERSION1) &&
              this.imageInfo.identifier
            ) {
              return Versions.VERSION1;
            }
            break;
        }
      }
      assert(false, 61);
    }

    /**
     * @param {Versions} version Optional IIIF image API version
     * @return {string} Compliance level as it appears in the IIIF image information
     * response.
     */
    getComplianceLevelEntryFromProfile(version) {
      if (this.imageInfo === undefined || this.imageInfo.profile === undefined) {
        return;
      }
      if (version === undefined) {
        version = this.getImageApiVersion();
      }
      switch (version) {
        case Versions.VERSION1:
          if (COMPLIANCE_VERSION1.test(this.imageInfo.profile)) {
            return this.imageInfo.profile;
          }
          break;
        case Versions.VERSION3:
          if (COMPLIANCE_VERSION3.test(this.imageInfo.profile)) {
            return this.imageInfo.profile;
          }
          break;
        case Versions.VERSION2:
          if (
            typeof this.imageInfo.profile === 'string' &&
            COMPLIANCE_VERSION2.test(this.imageInfo.profile)
          ) {
            return this.imageInfo.profile;
          }
          if (
            Array.isArray(this.imageInfo.profile) &&
            this.imageInfo.profile.length > 0 &&
            typeof this.imageInfo.profile[0] === 'string' &&
            COMPLIANCE_VERSION2.test(this.imageInfo.profile[0])
          ) {
            return this.imageInfo.profile[0];
          }
          break;
      }
    }

    /**
     * @param {Versions} version Optional IIIF image API version
     * @return {string} Compliance level, on of 'level0', 'level1' or 'level2' or undefined
     */
    getComplianceLevelFromProfile(version) {
      const complianceLevel = this.getComplianceLevelEntryFromProfile(version);
      if (complianceLevel === undefined) {
        return undefined;
      }
      const level = complianceLevel.match(/level[0-2](?:\.json)?$/g);
      return Array.isArray(level) ? level[0].replace('.json', '') : undefined;
    }

    /**
     * @return {SupportedFeatures} Image formats, qualities and region / size calculation
     * methods that are supported by the IIIF service.
     */
    getComplianceLevelSupportedFeatures() {
      if (this.imageInfo === undefined) {
        return;
      }
      const version = this.getImageApiVersion();
      const level = this.getComplianceLevelFromProfile(version);
      if (level === undefined) {
        return IIIF_PROFILE_VALUES['none']['none'];
      }
      return IIIF_PROFILE_VALUES[version][level];
    }

    /**
     * @param {PreferredOptions} [preferredOptions] Optional options for preferred format and quality.
     * @return {import("../source/IIIF.js").Options} IIIF tile source ready constructor options.
     * @api
     */
    getTileSourceOptions(preferredOptions) {
      const options = preferredOptions || {},
        version = this.getImageApiVersion();
      if (version === undefined) {
        return;
      }
      const imageOptions =
        version === undefined ? undefined : versionFunctions[version](this);
      if (imageOptions === undefined) {
        return;
      }
      return {
        url: imageOptions.url,
        version: version,
        size: [this.imageInfo.width, this.imageInfo.height],
        sizes: imageOptions.sizes,
        format:
          options.format !== undefined &&
          imageOptions.formats.includes(options.format)
            ? options.format
            : imageOptions.preferredFormat !== undefined
            ? imageOptions.preferredFormat
            : 'jpg',
        supports: imageOptions.supports,
        quality:
          options.quality && imageOptions.qualities.includes(options.quality)
            ? options.quality
            : imageOptions.qualities.includes('native')
            ? 'native'
            : 'default',
        resolutions: Array.isArray(imageOptions.resolutions)
          ? imageOptions.resolutions.sort(function (a, b) {
              return b - a;
            })
          : undefined,
        tileSize: imageOptions.tileSize,
      };
    }
  }

  /**
   * @module ol/source/IIIF
   */

  /**
   * @typedef {Object} Options
   * @property {import("./Source.js").AttributionLike} [attributions] Attributions.
   * @property {boolean} [attributionsCollapsible=true] Attributions are collapsible.
   * @property {number} [cacheSize] Size of the cache.
   * @property {null|string} [crossOrigin] The value for the crossOrigin option of the request.
   * @property {import("../extent.js").Extent} [extent=[0, -height, width, 0]] The extent.
   * @property {string} [format='jpg'] Requested image format.
   * @property {boolean} [interpolate=true] Use interpolated values when resampling.  By default,
   * linear interpolation is used when resampling.  Set to false to use the nearest neighbor instead.
   * @property {import("../proj.js").ProjectionLike} [projection] Projection.
   * @property {string} [quality] Requested IIIF image quality. Default is 'native'
   * for version 1, 'default' for versions 2 and 3.
   * @property {number} [reprojectionErrorThreshold=0.5] Maximum allowed reprojection error (in pixels).
   * Higher values can increase reprojection performance, but decrease precision.
   * @property {Array<number>} [resolutions] Supported resolutions as given in IIIF 'scaleFactors'
   * @property {import("../size.js").Size} size Size of the image [width, height].
   * @property {Array<import("../size.js").Size>} [sizes] Supported scaled image sizes.
   * Content of the IIIF info.json 'sizes' property, but as array of Size objects.
   * @property {import("./Source.js").State} [state] Source state.
   * @property {Array<string>} [supports=[]] Supported IIIF region and size calculation
   * features.
   * @property {number} [tilePixelRatio] Tile pixel ratio.
   * @property {number|import("../size.js").Size} [tileSize] Tile size.
   * Same tile size is used for all zoom levels. If tile size is a number,
   * a square tile is assumed. If the IIIF image service supports arbitrary
   * tiling (sizeByH, sizeByW, sizeByWh or sizeByPct as well as regionByPx or regionByPct
   * are supported), the default tilesize is 256.
   * @property {number} [transition] Transition.
   * @property {string} [url] Base URL of the IIIF Image service.
   * This should be the same as the IIIF Image ID.
   * @property {import("../format/IIIFInfo.js").Versions} [version=Versions.VERSION2] Service's IIIF Image API version.
   * @property {number|import("../array.js").NearestDirectionFunction} [zDirection=0]
   * Choose whether to use tiles with a higher or lower zoom level when between integer
   * zoom levels. See {@link module:ol/tilegrid/TileGrid~TileGrid#getZForResolution}.
   */

  function formatPercentage(percentage) {
    return percentage.toLocaleString('en', {maximumFractionDigits: 10});
  }

  /**
   * @classdesc
   * Layer source for IIIF Image API services.
   * @api
   */
  class IIIF extends TileImage {
    /**
     * @param {Options} [options] Tile source options. Use {@link import("../format/IIIFInfo.js").IIIFInfo}
     * to parse Image API service information responses into constructor options.
     * @api
     */
    constructor(options) {
      /**
       * @type {Partial<Options>}
       */
      const partialOptions = options || {};

      let baseUrl = partialOptions.url || '';
      baseUrl =
        baseUrl +
        (baseUrl.lastIndexOf('/') === baseUrl.length - 1 || baseUrl === ''
          ? ''
          : '/');
      const version = partialOptions.version || Versions.VERSION2;
      const sizes = partialOptions.sizes || [];
      const size = partialOptions.size;
      assert(
        size != undefined &&
          Array.isArray(size) &&
          size.length == 2 &&
          !isNaN(size[0]) &&
          size[0] > 0 &&
          !isNaN(size[1]) &&
          size[1] > 0,
        60
      );
      const width = size[0];
      const height = size[1];
      const tileSize = partialOptions.tileSize;
      const tilePixelRatio = partialOptions.tilePixelRatio || 1;
      const format = partialOptions.format || 'jpg';
      const quality =
        partialOptions.quality ||
        (partialOptions.version == Versions.VERSION1 ? 'native' : 'default');
      let resolutions = partialOptions.resolutions || [];
      const supports = partialOptions.supports || [];
      const extent = partialOptions.extent || [0, -height, width, 0];

      const supportsListedSizes =
        sizes != undefined && Array.isArray(sizes) && sizes.length > 0;
      const supportsListedTiles =
        tileSize !== undefined &&
        ((typeof tileSize === 'number' &&
          Number.isInteger(tileSize) &&
          tileSize > 0) ||
          (Array.isArray(tileSize) && tileSize.length > 0));
      const supportsArbitraryTiling =
        supports != undefined &&
        Array.isArray(supports) &&
        (supports.includes('regionByPx') || supports.includes('regionByPct')) &&
        (supports.includes('sizeByWh') ||
          supports.includes('sizeByH') ||
          supports.includes('sizeByW') ||
          supports.includes('sizeByPct'));

      let tileWidth, tileHeight, maxZoom;

      resolutions.sort(function (a, b) {
        return b - a;
      });

      if (supportsListedTiles || supportsArbitraryTiling) {
        if (tileSize != undefined) {
          if (
            typeof tileSize === 'number' &&
            Number.isInteger(tileSize) &&
            tileSize > 0
          ) {
            tileWidth = tileSize;
            tileHeight = tileSize;
          } else if (Array.isArray(tileSize) && tileSize.length > 0) {
            if (
              tileSize.length == 1 ||
              (tileSize[1] == undefined && Number.isInteger(tileSize[0]))
            ) {
              tileWidth = tileSize[0];
              tileHeight = tileSize[0];
            }
            if (tileSize.length == 2) {
              if (
                Number.isInteger(tileSize[0]) &&
                Number.isInteger(tileSize[1])
              ) {
                tileWidth = tileSize[0];
                tileHeight = tileSize[1];
              } else if (
                tileSize[0] == undefined &&
                Number.isInteger(tileSize[1])
              ) {
                tileWidth = tileSize[1];
                tileHeight = tileSize[1];
              }
            }
          }
        }
        if (tileWidth === undefined || tileHeight === undefined) {
          tileWidth = DEFAULT_TILE_SIZE;
          tileHeight = DEFAULT_TILE_SIZE;
        }
        if (resolutions.length == 0) {
          maxZoom = Math.max(
            Math.ceil(Math.log(width / tileWidth) / Math.LN2),
            Math.ceil(Math.log(height / tileHeight) / Math.LN2)
          );
          for (let i = maxZoom; i >= 0; i--) {
            resolutions.push(Math.pow(2, i));
          }
        } else {
          const maxScaleFactor = Math.max(...resolutions);
          // TODO maxScaleFactor might not be a power to 2
          maxZoom = Math.round(Math.log(maxScaleFactor) / Math.LN2);
        }
      } else {
        // No tile support.
        tileWidth = width;
        tileHeight = height;
        resolutions = [];
        if (supportsListedSizes) {
          /*
           * 'sizes' provided. Use full region in different resolutions. Every
           * resolution has only one tile.
           */
          sizes.sort(function (a, b) {
            return a[0] - b[0];
          });
          maxZoom = -1;
          const ignoredSizesIndex = [];
          for (let i = 0; i < sizes.length; i++) {
            const resolution = width / sizes[i][0];
            if (
              resolutions.length > 0 &&
              resolutions[resolutions.length - 1] == resolution
            ) {
              ignoredSizesIndex.push(i);
              continue;
            }
            resolutions.push(resolution);
            maxZoom++;
          }
          if (ignoredSizesIndex.length > 0) {
            for (let i = 0; i < ignoredSizesIndex.length; i++) {
              sizes.splice(ignoredSizesIndex[i] - i, 1);
            }
          }
        } else {
          // No useful image information at all. Try pseudo tile with full image.
          resolutions.push(1);
          sizes.push([width, height]);
          maxZoom = 0;
        }
      }

      const tileGrid = new TileGrid({
        tileSize: [tileWidth, tileHeight],
        extent: extent,
        origin: getTopLeft(extent),
        resolutions: resolutions,
      });

      const tileUrlFunction = function (tileCoord, pixelRatio, projection) {
        let regionParam, sizeParam;
        const zoom = tileCoord[0];
        if (zoom > maxZoom) {
          return;
        }
        const tileX = tileCoord[1],
          tileY = tileCoord[2],
          scale = resolutions[zoom];
        if (
          tileX === undefined ||
          tileY === undefined ||
          scale === undefined ||
          tileX < 0 ||
          Math.ceil(width / scale / tileWidth) <= tileX ||
          tileY < 0 ||
          Math.ceil(height / scale / tileHeight) <= tileY
        ) {
          return;
        }
        if (supportsArbitraryTiling || supportsListedTiles) {
          const regionX = tileX * tileWidth * scale,
            regionY = tileY * tileHeight * scale;
          let regionW = tileWidth * scale,
            regionH = tileHeight * scale,
            sizeW = tileWidth,
            sizeH = tileHeight;
          if (regionX + regionW > width) {
            regionW = width - regionX;
          }
          if (regionY + regionH > height) {
            regionH = height - regionY;
          }
          if (regionX + tileWidth * scale > width) {
            sizeW = Math.floor((width - regionX + scale - 1) / scale);
          }
          if (regionY + tileHeight * scale > height) {
            sizeH = Math.floor((height - regionY + scale - 1) / scale);
          }
          if (
            regionX == 0 &&
            regionW == width &&
            regionY == 0 &&
            regionH == height
          ) {
            // canonical full image region parameter is 'full', not 'x,y,w,h'
            regionParam = 'full';
          } else if (
            !supportsArbitraryTiling ||
            supports.includes('regionByPx')
          ) {
            regionParam = regionX + ',' + regionY + ',' + regionW + ',' + regionH;
          } else if (supports.includes('regionByPct')) {
            const pctX = formatPercentage((regionX / width) * 100),
              pctY = formatPercentage((regionY / height) * 100),
              pctW = formatPercentage((regionW / width) * 100),
              pctH = formatPercentage((regionH / height) * 100);
            regionParam = 'pct:' + pctX + ',' + pctY + ',' + pctW + ',' + pctH;
          }
          if (
            version == Versions.VERSION3 &&
            (!supportsArbitraryTiling || supports.includes('sizeByWh'))
          ) {
            sizeParam = sizeW + ',' + sizeH;
          } else if (!supportsArbitraryTiling || supports.includes('sizeByW')) {
            sizeParam = sizeW + ',';
          } else if (supports.includes('sizeByH')) {
            sizeParam = ',' + sizeH;
          } else if (supports.includes('sizeByWh')) {
            sizeParam = sizeW + ',' + sizeH;
          } else if (supports.includes('sizeByPct')) {
            sizeParam = 'pct:' + formatPercentage(100 / scale);
          }
        } else {
          regionParam = 'full';
          if (supportsListedSizes) {
            const regionWidth = sizes[zoom][0],
              regionHeight = sizes[zoom][1];
            if (version == Versions.VERSION3) {
              if (regionWidth == width && regionHeight == height) {
                sizeParam = 'max';
              } else {
                sizeParam = regionWidth + ',' + regionHeight;
              }
            } else {
              if (regionWidth == width) {
                sizeParam = 'full';
              } else {
                sizeParam = regionWidth + ',';
              }
            }
          } else {
            sizeParam = version == Versions.VERSION3 ? 'max' : 'full';
          }
        }
        return (
          baseUrl + regionParam + '/' + sizeParam + '/0/' + quality + '.' + format
        );
      };

      const IiifTileClass = CustomTile$1.bind(
        null,
        toSize(tileSize || 256).map(function (size) {
          return size * tilePixelRatio;
        })
      );

      super({
        attributions: partialOptions.attributions,
        attributionsCollapsible: partialOptions.attributionsCollapsible,
        cacheSize: partialOptions.cacheSize,
        crossOrigin: partialOptions.crossOrigin,
        interpolate: partialOptions.interpolate,
        projection: partialOptions.projection,
        reprojectionErrorThreshold: partialOptions.reprojectionErrorThreshold,
        state: partialOptions.state,
        tileClass: IiifTileClass,
        tileGrid: tileGrid,
        tilePixelRatio: partialOptions.tilePixelRatio,
        tileUrlFunction: tileUrlFunction,
        transition: partialOptions.transition,
      });

      /**
       * @type {number|import("../array.js").NearestDirectionFunction}
       */
      this.zDirection = partialOptions.zDirection;
    }
  }

  /**
   * @module ol/reproj/Image
   */

  /**
   * @typedef {function(import("../extent.js").Extent, number, number) : import("../ImageBase.js").default} FunctionType
   */

  /**
   * @classdesc
   * Class encapsulating single reprojected image.
   * See {@link module:ol/source/Image~ImageSource}.
   */
  class ReprojImage extends ImageBase {
    /**
     * @param {import("../proj/Projection.js").default} sourceProj Source projection (of the data).
     * @param {import("../proj/Projection.js").default} targetProj Target projection.
     * @param {import("../extent.js").Extent} targetExtent Target extent.
     * @param {number} targetResolution Target resolution.
     * @param {number} pixelRatio Pixel ratio.
     * @param {FunctionType} getImageFunction
     *     Function returning source images (extent, resolution, pixelRatio).
     * @param {boolean} interpolate Use linear interpolation when resampling.
     */
    constructor(
      sourceProj,
      targetProj,
      targetExtent,
      targetResolution,
      pixelRatio,
      getImageFunction,
      interpolate
    ) {
      const maxSourceExtent = sourceProj.getExtent();
      const maxTargetExtent = targetProj.getExtent();

      const limitedTargetExtent = maxTargetExtent
        ? getIntersection(targetExtent, maxTargetExtent)
        : targetExtent;

      const targetCenter = getCenter(limitedTargetExtent);
      const sourceResolution = calculateSourceResolution(
        sourceProj,
        targetProj,
        targetCenter,
        targetResolution
      );

      const errorThresholdInPixels = ERROR_THRESHOLD;

      const triangulation = new Triangulation(
        sourceProj,
        targetProj,
        limitedTargetExtent,
        maxSourceExtent,
        sourceResolution * errorThresholdInPixels,
        targetResolution
      );

      const sourceExtent = triangulation.calculateSourceExtent();
      const sourceImage = getImageFunction(
        sourceExtent,
        sourceResolution,
        pixelRatio
      );
      const state = sourceImage ? ImageState.IDLE : ImageState.EMPTY;
      const sourcePixelRatio = sourceImage ? sourceImage.getPixelRatio() : 1;

      super(targetExtent, targetResolution, sourcePixelRatio, state);

      /**
       * @private
       * @type {import("../proj/Projection.js").default}
       */
      this.targetProj_ = targetProj;

      /**
       * @private
       * @type {import("../extent.js").Extent}
       */
      this.maxSourceExtent_ = maxSourceExtent;

      /**
       * @private
       * @type {!import("./Triangulation.js").default}
       */
      this.triangulation_ = triangulation;

      /**
       * @private
       * @type {number}
       */
      this.targetResolution_ = targetResolution;

      /**
       * @private
       * @type {import("../extent.js").Extent}
       */
      this.targetExtent_ = targetExtent;

      /**
       * @private
       * @type {import("../ImageBase.js").default}
       */
      this.sourceImage_ = sourceImage;

      /**
       * @private
       * @type {number}
       */
      this.sourcePixelRatio_ = sourcePixelRatio;

      /**
       * @private
       * @type {boolean}
       */
      this.interpolate_ = interpolate;

      /**
       * @private
       * @type {HTMLCanvasElement}
       */
      this.canvas_ = null;

      /**
       * @private
       * @type {?import("../events.js").EventsKey}
       */
      this.sourceListenerKey_ = null;
    }

    /**
     * Clean up.
     */
    disposeInternal() {
      if (this.state == ImageState.LOADING) {
        this.unlistenSource_();
      }
      super.disposeInternal();
    }

    /**
     * @return {HTMLCanvasElement} Image.
     */
    getImage() {
      return this.canvas_;
    }

    /**
     * @return {import("../proj/Projection.js").default} Projection.
     */
    getProjection() {
      return this.targetProj_;
    }

    /**
     * @private
     */
    reproject_() {
      const sourceState = this.sourceImage_.getState();
      if (sourceState == ImageState.LOADED) {
        const width = getWidth(this.targetExtent_) / this.targetResolution_;
        const height = getHeight(this.targetExtent_) / this.targetResolution_;

        this.canvas_ = render(
          width,
          height,
          this.sourcePixelRatio_,
          this.sourceImage_.getResolution(),
          this.maxSourceExtent_,
          this.targetResolution_,
          this.targetExtent_,
          this.triangulation_,
          [
            {
              extent: this.sourceImage_.getExtent(),
              image: this.sourceImage_.getImage(),
            },
          ],
          0,
          undefined,
          this.interpolate_
        );
      }
      this.state = sourceState;
      this.changed();
    }

    /**
     * Load not yet loaded URI.
     */
    load() {
      if (this.state == ImageState.IDLE) {
        this.state = ImageState.LOADING;
        this.changed();

        const sourceState = this.sourceImage_.getState();
        if (sourceState == ImageState.LOADED || sourceState == ImageState.ERROR) {
          this.reproject_();
        } else {
          this.sourceListenerKey_ = listen(
            this.sourceImage_,
            EventType$1.CHANGE,
            function (e) {
              const sourceState = this.sourceImage_.getState();
              if (
                sourceState == ImageState.LOADED ||
                sourceState == ImageState.ERROR
              ) {
                this.unlistenSource_();
                this.reproject_();
              }
            },
            this
          );
          this.sourceImage_.load();
        }
      }
    }

    /**
     * @private
     */
    unlistenSource_() {
      unlistenByKey(
        /** @type {!import("../events.js").EventsKey} */ (this.sourceListenerKey_)
      );
      this.sourceListenerKey_ = null;
    }
  }

  /**
   * @module ol/source/Image
   */

  /**
   * @enum {string}
   */
  const ImageSourceEventType = {
    /**
     * Triggered when an image starts loading.
     * @event module:ol/source/Image.ImageSourceEvent#imageloadstart
     * @api
     */
    IMAGELOADSTART: 'imageloadstart',

    /**
     * Triggered when an image finishes loading.
     * @event module:ol/source/Image.ImageSourceEvent#imageloadend
     * @api
     */
    IMAGELOADEND: 'imageloadend',

    /**
     * Triggered if image loading results in an error.
     * @event module:ol/source/Image.ImageSourceEvent#imageloaderror
     * @api
     */
    IMAGELOADERROR: 'imageloaderror',
  };

  /**
   * @typedef {'imageloadend'|'imageloaderror'|'imageloadstart'} ImageSourceEventTypes
   */

  /**
   * @classdesc
   * Events emitted by {@link module:ol/source/Image~ImageSource} instances are instances of this
   * type.
   */
  class ImageSourceEvent extends BaseEvent {
    /**
     * @param {string} type Type.
     * @param {import("../Image.js").default} image The image.
     */
    constructor(type, image) {
      super(type);

      /**
       * The image related to the event.
       * @type {import("../Image.js").default}
       * @api
       */
      this.image = image;
    }
  }

  /***
   * @template Return
   * @typedef {import("../Observable").OnSignature<import("../Observable").EventTypes, import("../events/Event.js").default, Return> &
   *   import("../Observable").OnSignature<import("../ObjectEventType").Types, import("../Object").ObjectEvent, Return> &
   *   import("../Observable").OnSignature<ImageSourceEventTypes, ImageSourceEvent, Return> &
   *   import("../Observable").CombinedOnSignature<import("../Observable").EventTypes|import("../ObjectEventType").Types
   *     |ImageSourceEventTypes, Return>} ImageSourceOnSignature
   */

  /**
   * @typedef {Object} Options
   * @property {import("./Source.js").AttributionLike} [attributions] Attributions.
   * @property {boolean} [interpolate=true] Use interpolated values when resampling.  By default,
   * linear interpolation is used when resampling.  Set to false to use the nearest neighbor instead.
   * @property {import("../proj.js").ProjectionLike} [projection] Projection.
   * @property {Array<number>} [resolutions] Resolutions.
   * @property {import("./Source.js").State} [state] State.
   */

  /**
   * @classdesc
   * Abstract base class; normally only used for creating subclasses and not
   * instantiated in apps.
   * Base class for sources providing a single image.
   * @abstract
   * @fires module:ol/source/Image.ImageSourceEvent
   * @api
   */
  class ImageSource extends Source {
    /**
     * @param {Options} options Single image source options.
     */
    constructor(options) {
      super({
        attributions: options.attributions,
        projection: options.projection,
        state: options.state,
        interpolate:
          options.interpolate !== undefined ? options.interpolate : true,
      });

      /***
       * @type {ImageSourceOnSignature<import("../events").EventsKey>}
       */
      this.on;

      /***
       * @type {ImageSourceOnSignature<import("../events").EventsKey>}
       */
      this.once;

      /***
       * @type {ImageSourceOnSignature<void>}
       */
      this.un;

      /**
       * @private
       * @type {Array<number>|null}
       */
      this.resolutions_ =
        options.resolutions !== undefined ? options.resolutions : null;

      /**
       * @private
       * @type {import("../reproj/Image.js").default}
       */
      this.reprojectedImage_ = null;

      /**
       * @private
       * @type {number}
       */
      this.reprojectedRevision_ = 0;
    }

    /**
     * @return {Array<number>|null} Resolutions.
     */
    getResolutions() {
      return this.resolutions_;
    }

    /**
     * @protected
     * @param {number} resolution Resolution.
     * @return {number} Resolution.
     */
    findNearestResolution(resolution) {
      if (this.resolutions_) {
        const idx = linearFindNearest(this.resolutions_, resolution, 0);
        resolution = this.resolutions_[idx];
      }
      return resolution;
    }

    /**
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {number} resolution Resolution.
     * @param {number} pixelRatio Pixel ratio.
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @return {import("../ImageBase.js").default} Single image.
     */
    getImage(extent, resolution, pixelRatio, projection) {
      const sourceProjection = this.getProjection();
      if (
        !sourceProjection ||
        !projection ||
        equivalent(sourceProjection, projection)
      ) {
        if (sourceProjection) {
          projection = sourceProjection;
        }
        return this.getImageInternal(extent, resolution, pixelRatio, projection);
      } else {
        if (this.reprojectedImage_) {
          if (
            this.reprojectedRevision_ == this.getRevision() &&
            equivalent(this.reprojectedImage_.getProjection(), projection) &&
            this.reprojectedImage_.getResolution() == resolution &&
            equals$2(this.reprojectedImage_.getExtent(), extent)
          ) {
            return this.reprojectedImage_;
          }
          this.reprojectedImage_.dispose();
          this.reprojectedImage_ = null;
        }

        this.reprojectedImage_ = new ReprojImage(
          sourceProjection,
          projection,
          extent,
          resolution,
          pixelRatio,
          function (extent, resolution, pixelRatio) {
            return this.getImageInternal(
              extent,
              resolution,
              pixelRatio,
              sourceProjection
            );
          }.bind(this),
          this.getInterpolate()
        );
        this.reprojectedRevision_ = this.getRevision();

        return this.reprojectedImage_;
      }
    }

    /**
     * @abstract
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {number} resolution Resolution.
     * @param {number} pixelRatio Pixel ratio.
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @return {import("../ImageBase.js").default} Single image.
     * @protected
     */
    getImageInternal(extent, resolution, pixelRatio, projection) {
      return abstract();
    }

    /**
     * Handle image change events.
     * @param {import("../events/Event.js").default} event Event.
     * @protected
     */
    handleImageChange(event) {
      const image = /** @type {import("../Image.js").default} */ (event.target);
      let type;
      switch (image.getState()) {
        case ImageState.LOADING:
          this.loading = true;
          type = ImageSourceEventType.IMAGELOADSTART;
          break;
        case ImageState.LOADED:
          this.loading = false;
          type = ImageSourceEventType.IMAGELOADEND;
          break;
        case ImageState.ERROR:
          this.loading = false;
          type = ImageSourceEventType.IMAGELOADERROR;
          break;
        default:
          return;
      }
      if (this.hasListener(type)) {
        this.dispatchEvent(new ImageSourceEvent(type, image));
      }
    }
  }

  /**
   * Default image load function for image sources that use import("../Image.js").Image image
   * instances.
   * @param {import("../Image.js").default} image Image.
   * @param {string} src Source.
   */
  function defaultImageLoadFunction(image, src) {
    /** @type {HTMLImageElement|HTMLVideoElement} */ (image.getImage()).src = src;
  }

  /**
   * @module ol/uri
   */

  /**
   * Appends query parameters to a URI.
   *
   * @param {string} uri The original URI, which may already have query data.
   * @param {!Object} params An object where keys are URI-encoded parameter keys,
   *     and the values are arbitrary types or arrays.
   * @return {string} The new URI.
   */
  function appendParams(uri, params) {
    const keyParams = [];
    // Skip any null or undefined parameter values
    Object.keys(params).forEach(function (k) {
      if (params[k] !== null && params[k] !== undefined) {
        keyParams.push(k + '=' + encodeURIComponent(params[k]));
      }
    });
    const qs = keyParams.join('&');
    // remove any trailing ? or &
    uri = uri.replace(/[?&]$/, '');
    // append ? or & depending on whether uri has existing parameters
    uri += uri.includes('?') ? '&' : '?';
    return uri + qs;
  }

  /**
   * @module ol/source/ImageArcGISRest
   */

  /**
   * @typedef {Object} Options
   * @property {import("./Source.js").AttributionLike} [attributions] Attributions.
   * @property {null|string} [crossOrigin] The `crossOrigin` attribute for loaded images.  Note that
   * you must provide a `crossOrigin` value if you want to access pixel data with the Canvas renderer.
   * See https://developer.mozilla.org/en-US/docs/Web/HTML/CORS_enabled_image for more detail.
   * @property {boolean} [hidpi=true] Use the `ol/Map#pixelRatio` value when requesting the image from
   * the remote server.
   * @property {import("../Image.js").LoadFunction} [imageLoadFunction] Optional function to load an image given
   * a URL.
   * @property {boolean} [interpolate=true] Use interpolated values when resampling.  By default,
   * linear interpolation is used when resampling.  Set to false to use the nearest neighbor instead.
   * @property {Object<string,*>} [params] ArcGIS Rest parameters. This field is optional. Service
   * defaults will be used for any fields not specified. `FORMAT` is `PNG32` by default. `F` is
   * `IMAGE` by default. `TRANSPARENT` is `true` by default.  `BBOX`, `SIZE`, `BBOXSR`, and `IMAGESR`
   * will be set dynamically. Set `LAYERS` to override the default service layer visibility. See
   * https://developers.arcgis.com/rest/services-reference/export-map.htm
   * for further reference.
   * @property {import("../proj.js").ProjectionLike} [projection] Projection. Default is the view projection.
   * The projection code must contain a numeric end portion separated by :
   * or the entire code must form a valid ArcGIS SpatialReference definition.
   * @property {number} [ratio=1.5] Ratio. `1` means image requests are the size of the map viewport,
   * `2` means twice the size of the map viewport, and so on.
   * @property {Array<number>} [resolutions] Resolutions. If specified, requests will be made for
   * these resolutions only.
   * @property {string} [url] ArcGIS Rest service URL for a Map Service or Image Service. The url
   * should include /MapServer or /ImageServer.
   */

  /**
   * @classdesc
   * Source for data from ArcGIS Rest services providing single, untiled images.
   * Useful when underlying map service has labels.
   *
   * If underlying map service is not using labels,
   * take advantage of ol image caching and use
   * {@link module:ol/source/TileArcGISRest~TileArcGISRest} data source.
   *
   * @fires module:ol/source/Image.ImageSourceEvent
   * @api
   */
  class ImageArcGISRest extends ImageSource {
    /**
     * @param {Options} [options] Image ArcGIS Rest Options.
     */
    constructor(options) {
      options = options ? options : {};

      super({
        attributions: options.attributions,
        interpolate: options.interpolate,
        projection: options.projection,
        resolutions: options.resolutions,
      });

      /**
       * @private
       * @type {?string}
       */
      this.crossOrigin_ =
        options.crossOrigin !== undefined ? options.crossOrigin : null;

      /**
       * @private
       * @type {boolean}
       */
      this.hidpi_ = options.hidpi !== undefined ? options.hidpi : true;

      /**
       * @private
       * @type {string|undefined}
       */
      this.url_ = options.url;

      /**
       * @private
       * @type {import("../Image.js").LoadFunction}
       */
      this.imageLoadFunction_ =
        options.imageLoadFunction !== undefined
          ? options.imageLoadFunction
          : defaultImageLoadFunction;

      /**
       * @private
       * @type {!Object}
       */
      this.params_ = options.params || {};

      /**
       * @private
       * @type {import("../Image.js").default}
       */
      this.image_ = null;

      /**
       * @private
       * @type {import("../size.js").Size}
       */
      this.imageSize_ = [0, 0];

      /**
       * @private
       * @type {number}
       */
      this.renderedRevision_ = 0;

      /**
       * @private
       * @type {number}
       */
      this.ratio_ = options.ratio !== undefined ? options.ratio : 1.5;
    }

    /**
     * Get the user-provided params, i.e. those passed to the constructor through
     * the "params" option, and possibly updated using the updateParams method.
     * @return {Object} Params.
     * @api
     */
    getParams() {
      return this.params_;
    }

    /**
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {number} resolution Resolution.
     * @param {number} pixelRatio Pixel ratio.
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @return {import("../Image.js").default} Single image.
     */
    getImageInternal(extent, resolution, pixelRatio, projection) {
      if (this.url_ === undefined) {
        return null;
      }

      resolution = this.findNearestResolution(resolution);
      pixelRatio = this.hidpi_ ? pixelRatio : 1;

      const image = this.image_;
      if (
        image &&
        this.renderedRevision_ == this.getRevision() &&
        image.getResolution() == resolution &&
        image.getPixelRatio() == pixelRatio &&
        containsExtent(image.getExtent(), extent)
      ) {
        return image;
      }

      const params = {
        'F': 'image',
        'FORMAT': 'PNG32',
        'TRANSPARENT': true,
      };
      Object.assign(params, this.params_);

      extent = extent.slice();
      const centerX = (extent[0] + extent[2]) / 2;
      const centerY = (extent[1] + extent[3]) / 2;
      if (this.ratio_ != 1) {
        const halfWidth = (this.ratio_ * getWidth(extent)) / 2;
        const halfHeight = (this.ratio_ * getHeight(extent)) / 2;
        extent[0] = centerX - halfWidth;
        extent[1] = centerY - halfHeight;
        extent[2] = centerX + halfWidth;
        extent[3] = centerY + halfHeight;
      }

      const imageResolution = resolution / pixelRatio;

      // Compute an integer width and height.
      const width = Math.ceil(getWidth(extent) / imageResolution);
      const height = Math.ceil(getHeight(extent) / imageResolution);

      // Modify the extent to match the integer width and height.
      extent[0] = centerX - (imageResolution * width) / 2;
      extent[2] = centerX + (imageResolution * width) / 2;
      extent[1] = centerY - (imageResolution * height) / 2;
      extent[3] = centerY + (imageResolution * height) / 2;

      this.imageSize_[0] = width;
      this.imageSize_[1] = height;

      const url = this.getRequestUrl_(
        extent,
        this.imageSize_,
        pixelRatio,
        projection,
        params
      );

      this.image_ = new ImageWrapper(
        extent,
        resolution,
        pixelRatio,
        url,
        this.crossOrigin_,
        this.imageLoadFunction_
      );

      this.renderedRevision_ = this.getRevision();

      this.image_.addEventListener(
        EventType$1.CHANGE,
        this.handleImageChange.bind(this)
      );

      return this.image_;
    }

    /**
     * Return the image load function of the source.
     * @return {import("../Image.js").LoadFunction} The image load function.
     * @api
     */
    getImageLoadFunction() {
      return this.imageLoadFunction_;
    }

    /**
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {import("../size.js").Size} size Size.
     * @param {number} pixelRatio Pixel ratio.
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @param {Object} params Params.
     * @return {string} Request URL.
     * @private
     */
    getRequestUrl_(extent, size, pixelRatio, projection, params) {
      // ArcGIS Server only wants the numeric portion of the projection ID.
      // (if there is no numeric portion the entire projection code must
      // form a valid ArcGIS SpatialReference definition).
      const srid = projection
        .getCode()
        .split(/:(?=\d+$)/)
        .pop();

      params['SIZE'] = size[0] + ',' + size[1];
      params['BBOX'] = extent.join(',');
      params['BBOXSR'] = srid;
      params['IMAGESR'] = srid;
      params['DPI'] = Math.round(90 * pixelRatio);

      const url = this.url_;

      const modifiedUrl = url
        .replace(/MapServer\/?$/, 'MapServer/export')
        .replace(/ImageServer\/?$/, 'ImageServer/exportImage');
      if (modifiedUrl == url) {
        assert(false, 50); // `options.featureTypes` should be an Array
      }
      return appendParams(modifiedUrl, params);
    }

    /**
     * Return the URL used for this ArcGIS source.
     * @return {string|undefined} URL.
     * @api
     */
    getUrl() {
      return this.url_;
    }

    /**
     * Set the image load function of the source.
     * @param {import("../Image.js").LoadFunction} imageLoadFunction Image load function.
     * @api
     */
    setImageLoadFunction(imageLoadFunction) {
      this.image_ = null;
      this.imageLoadFunction_ = imageLoadFunction;
      this.changed();
    }

    /**
     * Set the URL to use for requests.
     * @param {string|undefined} url URL.
     * @api
     */
    setUrl(url) {
      if (url != this.url_) {
        this.url_ = url;
        this.image_ = null;
        this.changed();
      }
    }

    /**
     * Update the user-provided params.
     * @param {Object} params Params.
     * @api
     */
    updateParams(params) {
      Object.assign(this.params_, params);
      this.image_ = null;
      this.changed();
    }
  }

  /**
   * @module ol/source/ImageCanvas
   */

  /**
   * A function returning the canvas element (`{HTMLCanvasElement}`)
   * used by the source as an image. The arguments passed to the function are:
   * {@link module:ol/extent~Extent} the image extent, `{number}` the image resolution,
   * `{number}` the pixel ratio of the map, {@link module:ol/size~Size} the image size,
   * and {@link module:ol/proj/Projection~Projection} the image projection. The canvas returned by
   * this function is cached by the source. The this keyword inside the function
   * references the {@link module:ol/source/ImageCanvas~ImageCanvasSource}.
   *
   * @typedef {function(this:import("../ImageCanvas.js").default, import("../extent.js").Extent, number,
   *     number, import("../size.js").Size, import("../proj/Projection.js").default): HTMLCanvasElement} FunctionType
   */

  /**
   * @typedef {Object} Options
   * @property {import("./Source.js").AttributionLike} [attributions] Attributions.
   * @property {FunctionType} [canvasFunction] Canvas function.
   * The function returning the canvas element used by the source
   * as an image. The arguments passed to the function are: {@link import("../extent.js").Extent} the
   * image extent, `{number}` the image resolution, `{number}` the pixel ratio of the map,
   * {@link import("../size.js").Size} the image size, and {@link import("../proj/Projection.js").default} the image
   * projection. The canvas returned by this function is cached by the source. If
   * the value returned by the function is later changed then
   * `changed` should be called on the source for the source to
   * invalidate the current cached image. See: {@link module:ol/Observable~Observable#changed}
   * @property {boolean} [interpolate=true] Use interpolated values when resampling.  By default,
   * linear interpolation is used when resampling.  Set to false to use the nearest neighbor instead.
   * @property {import("../proj.js").ProjectionLike} [projection] Projection. Default is the view projection.
   * @property {number} [ratio=1.5] Ratio. 1 means canvases are the size of the map viewport, 2 means twice the
   * width and height of the map viewport, and so on. Must be `1` or higher.
   * @property {Array<number>} [resolutions] Resolutions.
   * If specified, new canvases will be created for these resolutions
   * @property {import("./Source.js").State} [state] Source state.
   */

  /**
   * @classdesc
   * Base class for image sources where a canvas element is the image.
   * @api
   */
  class ImageCanvasSource extends ImageSource {
    /**
     * @param {Options} [options] ImageCanvas options.
     */
    constructor(options) {
      options = options ? options : {};

      super({
        attributions: options.attributions,
        interpolate: options.interpolate,
        projection: options.projection,
        resolutions: options.resolutions,
        state: options.state,
      });

      /**
       * @private
       * @type {FunctionType}
       */
      this.canvasFunction_ = options.canvasFunction;

      /**
       * @private
       * @type {import("../ImageCanvas.js").default}
       */
      this.canvas_ = null;

      /**
       * @private
       * @type {number}
       */
      this.renderedRevision_ = 0;

      /**
       * @private
       * @type {number}
       */
      this.ratio_ = options.ratio !== undefined ? options.ratio : 1.5;
    }

    /**
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {number} resolution Resolution.
     * @param {number} pixelRatio Pixel ratio.
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @return {import("../ImageCanvas.js").default} Single image.
     */
    getImageInternal(extent, resolution, pixelRatio, projection) {
      resolution = this.findNearestResolution(resolution);

      let canvas = this.canvas_;
      if (
        canvas &&
        this.renderedRevision_ == this.getRevision() &&
        canvas.getResolution() == resolution &&
        canvas.getPixelRatio() == pixelRatio &&
        containsExtent(canvas.getExtent(), extent)
      ) {
        return canvas;
      }

      extent = extent.slice();
      scaleFromCenter(extent, this.ratio_);
      const width = getWidth(extent) / resolution;
      const height = getHeight(extent) / resolution;
      const size = [width * pixelRatio, height * pixelRatio];

      const canvasElement = this.canvasFunction_.call(
        this,
        extent,
        resolution,
        pixelRatio,
        size,
        projection
      );
      if (canvasElement) {
        canvas = new ImageCanvas(extent, resolution, pixelRatio, canvasElement);
      }
      this.canvas_ = canvas;
      this.renderedRevision_ = this.getRevision();

      return canvas;
    }
  }

  /**
   * @module ol/source/ImageMapGuide
   */

  /**
   * @typedef {Object} Options
   * @property {string} [url] The mapagent url.
   * @property {null|string} [crossOrigin] The `crossOrigin` attribute for loaded images.  Note that
   * you must provide a `crossOrigin` value if you want to access pixel data with the Canvas renderer.
   * See https://developer.mozilla.org/en-US/docs/Web/HTML/CORS_enabled_image for more detail.
   * @property {number} [displayDpi=96] The display resolution.
   * @property {number} [metersPerUnit=1] The meters-per-unit value.
   * @property {boolean} [hidpi=true] Use the `ol/Map#pixelRatio` value when requesting
   * the image from the remote server.
   * @property {boolean} [useOverlay] If `true`, will use `GETDYNAMICMAPOVERLAYIMAGE`.
   * @property {import("../proj.js").ProjectionLike} [projection] Projection. Default is the view projection.
   * @property {number} [ratio=1] Ratio. `1` means image requests are the size of the map viewport, `2` means
   * twice the width and height of the map viewport, and so on. Must be `1` or higher.
   * @property {Array<number>} [resolutions] Resolutions.
   * If specified, requests will be made for these resolutions only.
   * @property {import("../Image.js").LoadFunction} [imageLoadFunction] Optional function to load an image given a URL.
   * @property {boolean} [interpolate=true] Use interpolated values when resampling.  By default,
   * linear interpolation is used when resampling.  Set to false to use the nearest neighbor instead.
   * @property {Object} [params] Additional parameters.
   */

  /**
   * @classdesc
   * Source for images from Mapguide servers
   *
   * @fires module:ol/source/Image.ImageSourceEvent
   * @api
   */
  class ImageMapGuide extends ImageSource {
    /**
     * @param {Options} options ImageMapGuide options.
     */
    constructor(options) {
      super({
        interpolate: options.interpolate,
        projection: options.projection,
        resolutions: options.resolutions,
      });

      /**
       * @private
       * @type {?string}
       */
      this.crossOrigin_ =
        options.crossOrigin !== undefined ? options.crossOrigin : null;

      /**
       * @private
       * @type {number}
       */
      this.displayDpi_ =
        options.displayDpi !== undefined ? options.displayDpi : 96;

      /**
       * @private
       * @type {!Object}
       */
      this.params_ = options.params || {};

      /**
       * @private
       * @type {string|undefined}
       */
      this.url_ = options.url;

      /**
       * @private
       * @type {import("../Image.js").LoadFunction}
       */
      this.imageLoadFunction_ =
        options.imageLoadFunction !== undefined
          ? options.imageLoadFunction
          : defaultImageLoadFunction;

      /**
       * @private
       * @type {boolean}
       */
      this.hidpi_ = options.hidpi !== undefined ? options.hidpi : true;

      /**
       * @private
       * @type {number}
       */
      this.metersPerUnit_ =
        options.metersPerUnit !== undefined ? options.metersPerUnit : 1;

      /**
       * @private
       * @type {number}
       */
      this.ratio_ = options.ratio !== undefined ? options.ratio : 1;

      /**
       * @private
       * @type {boolean}
       */
      this.useOverlay_ =
        options.useOverlay !== undefined ? options.useOverlay : false;

      /**
       * @private
       * @type {import("../Image.js").default}
       */
      this.image_ = null;

      /**
       * @private
       * @type {number}
       */
      this.renderedRevision_ = 0;
    }

    /**
     * Get the user-provided params, i.e. those passed to the constructor through
     * the "params" option, and possibly updated using the updateParams method.
     * @return {Object} Params.
     * @api
     */
    getParams() {
      return this.params_;
    }

    /**
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {number} resolution Resolution.
     * @param {number} pixelRatio Pixel ratio.
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @return {import("../Image.js").default} Single image.
     */
    getImageInternal(extent, resolution, pixelRatio, projection) {
      resolution = this.findNearestResolution(resolution);
      pixelRatio = this.hidpi_ ? pixelRatio : 1;

      let image = this.image_;
      if (
        image &&
        this.renderedRevision_ == this.getRevision() &&
        image.getResolution() == resolution &&
        image.getPixelRatio() == pixelRatio &&
        containsExtent(image.getExtent(), extent)
      ) {
        return image;
      }

      if (this.ratio_ != 1) {
        extent = extent.slice();
        scaleFromCenter(extent, this.ratio_);
      }
      const width = getWidth(extent) / resolution;
      const height = getHeight(extent) / resolution;
      const size = [width * pixelRatio, height * pixelRatio];

      if (this.url_ !== undefined) {
        const imageUrl = this.getUrl(
          this.url_,
          this.params_,
          extent,
          size,
          projection
        );
        image = new ImageWrapper(
          extent,
          resolution,
          pixelRatio,
          imageUrl,
          this.crossOrigin_,
          this.imageLoadFunction_
        );
        image.addEventListener(
          EventType$1.CHANGE,
          this.handleImageChange.bind(this)
        );
      } else {
        image = null;
      }
      this.image_ = image;
      this.renderedRevision_ = this.getRevision();

      return image;
    }

    /**
     * Return the image load function of the source.
     * @return {import("../Image.js").LoadFunction} The image load function.
     * @api
     */
    getImageLoadFunction() {
      return this.imageLoadFunction_;
    }

    /**
     * Update the user-provided params.
     * @param {Object} params Params.
     * @api
     */
    updateParams(params) {
      Object.assign(this.params_, params);
      this.changed();
    }

    /**
     * @param {string} baseUrl The mapagent url.
     * @param {Object<string, string|number>} params Request parameters.
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {import("../size.js").Size} size Size.
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @return {string} The mapagent map image request URL.
     */
    getUrl(baseUrl, params, extent, size, projection) {
      const scale = getScale(extent, size, this.metersPerUnit_, this.displayDpi_);
      const center = getCenter(extent);
      const baseParams = {
        'OPERATION': this.useOverlay_
          ? 'GETDYNAMICMAPOVERLAYIMAGE'
          : 'GETMAPIMAGE',
        'VERSION': '2.0.0',
        'LOCALE': 'en',
        'CLIENTAGENT': 'ol/source/ImageMapGuide source',
        'CLIP': '1',
        'SETDISPLAYDPI': this.displayDpi_,
        'SETDISPLAYWIDTH': Math.round(size[0]),
        'SETDISPLAYHEIGHT': Math.round(size[1]),
        'SETVIEWSCALE': scale,
        'SETVIEWCENTERX': center[0],
        'SETVIEWCENTERY': center[1],
      };
      Object.assign(baseParams, params);
      return appendParams(baseUrl, baseParams);
    }

    /**
     * Set the image load function of the MapGuide source.
     * @param {import("../Image.js").LoadFunction} imageLoadFunction Image load function.
     * @api
     */
    setImageLoadFunction(imageLoadFunction) {
      this.image_ = null;
      this.imageLoadFunction_ = imageLoadFunction;
      this.changed();
    }
  }

  /**
   * @param {import("../extent.js").Extent} extent The map extents.
   * @param {import("../size.js").Size} size The viewport size.
   * @param {number} metersPerUnit The meters-per-unit value.
   * @param {number} dpi The display resolution.
   * @return {number} The computed map scale.
   */
  function getScale(extent, size, metersPerUnit, dpi) {
    const mcsW = getWidth(extent);
    const mcsH = getHeight(extent);
    const devW = size[0];
    const devH = size[1];
    const mpp = 0.0254 / dpi;
    if (devH * mcsW > devW * mcsH) {
      return (mcsW * metersPerUnit) / (devW * mpp); // width limited
    } else {
      return (mcsH * metersPerUnit) / (devH * mpp); // height limited
    }
  }

  /**
   * @module ol/source/ImageStatic
   */

  /**
   * @typedef {Object} Options
   * @property {import("./Source.js").AttributionLike} [attributions] Attributions.
   * @property {null|string} [crossOrigin] The `crossOrigin` attribute for loaded images.  Note that
   * you must provide a `crossOrigin` value if you want to access pixel data with the Canvas renderer.
   * See https://developer.mozilla.org/en-US/docs/Web/HTML/CORS_enabled_image for more detail.
   * @property {import("../extent.js").Extent} [imageExtent] Extent of the image in map coordinates.
   * This is the [left, bottom, right, top] map coordinates of your image.
   * @property {import("../Image.js").LoadFunction} [imageLoadFunction] Optional function to load an image given a URL.
   * @property {boolean} [interpolate=true] Use interpolated values when resampling.  By default,
   * linear interpolation is used when resampling.  Set to false to use the nearest neighbor instead.
   * @property {import("../proj.js").ProjectionLike} [projection] Projection. Default is the view projection.
   * @property {import("../size.js").Size} [imageSize] Size of the image in pixels. Usually the image size is auto-detected, so this
   * only needs to be set if auto-detection fails for some reason.
   * @property {string} url Image URL.
   */

  /**
   * @classdesc
   * A layer source for displaying a single, static image.
   * @api
   */
  class Static extends ImageSource {
    /**
     * @param {Options} options ImageStatic options.
     */
    constructor(options) {
      const crossOrigin =
        options.crossOrigin !== undefined ? options.crossOrigin : null;

      const /** @type {import("../Image.js").LoadFunction} */ imageLoadFunction =
          options.imageLoadFunction !== undefined
            ? options.imageLoadFunction
            : defaultImageLoadFunction;

      super({
        attributions: options.attributions,
        interpolate: options.interpolate,
        projection: get$2(options.projection),
      });

      /**
       * @private
       * @type {string}
       */
      this.url_ = options.url;

      /**
       * @private
       * @type {import("../extent.js").Extent}
       */
      this.imageExtent_ = options.imageExtent;

      /**
       * @private
       * @type {import("../Image.js").default}
       */
      this.image_ = new ImageWrapper(
        this.imageExtent_,
        undefined,
        1,
        this.url_,
        crossOrigin,
        imageLoadFunction
      );

      /**
       * @private
       * @type {import("../size.js").Size|null}
       */
      this.imageSize_ = options.imageSize ? options.imageSize : null;

      this.image_.addEventListener(
        EventType$1.CHANGE,
        this.handleImageChange.bind(this)
      );
    }

    /**
     * Returns the image extent
     * @return {import("../extent.js").Extent} image extent.
     * @api
     */
    getImageExtent() {
      return this.imageExtent_;
    }

    /**
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {number} resolution Resolution.
     * @param {number} pixelRatio Pixel ratio.
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @return {import("../Image.js").default} Single image.
     */
    getImageInternal(extent, resolution, pixelRatio, projection) {
      if (intersects$3(extent, this.image_.getExtent())) {
        return this.image_;
      }
      return null;
    }

    /**
     * Return the URL used for this image source.
     * @return {string} URL.
     * @api
     */
    getUrl() {
      return this.url_;
    }

    /**
     * @param {import("../events/Event.js").default} evt Event.
     */
    handleImageChange(evt) {
      if (this.image_.getState() == ImageState.LOADED) {
        const imageExtent = this.image_.getExtent();
        const image = this.image_.getImage();
        let imageWidth, imageHeight;
        if (this.imageSize_) {
          imageWidth = this.imageSize_[0];
          imageHeight = this.imageSize_[1];
        } else {
          imageWidth = image.width;
          imageHeight = image.height;
        }
        const extentWidth = getWidth(imageExtent);
        const extentHeight = getHeight(imageExtent);
        const xResolution = extentWidth / imageWidth;
        const yResolution = extentHeight / imageHeight;
        let targetWidth = imageWidth;
        let targetHeight = imageHeight;
        if (xResolution > yResolution) {
          targetWidth = Math.round(extentWidth / yResolution);
        } else {
          targetHeight = Math.round(extentHeight / xResolution);
        }
        if (targetWidth !== imageWidth || targetHeight !== imageHeight) {
          const context = createCanvasContext2D(targetWidth, targetHeight);
          if (!this.getInterpolate()) {
            context.imageSmoothingEnabled = false;
          }
          const canvas = context.canvas;
          context.drawImage(
            image,
            0,
            0,
            imageWidth,
            imageHeight,
            0,
            0,
            canvas.width,
            canvas.height
          );
          this.image_.setImage(canvas);
        }
      }
      super.handleImageChange(evt);
    }
  }

  /**
   * @module ol/source/wms
   */

  /**
   * Default WMS version.
   * @type {string}
   */
  const DEFAULT_VERSION$1 = '1.3.0';

  /**
   * @api
   * @typedef {'carmentaserver' | 'geoserver' | 'mapserver' | 'qgis'} ServerType
   * Set the server type to use implementation-specific parameters beyond the WMS specification.
   *  - `'carmentaserver'`: HiDPI support for [Carmenta Server](https://www.carmenta.com/en/products/carmenta-server)
   *  - `'geoserver'`: HiDPI support for [GeoServer](https://geoserver.org/)
   *  - `'mapserver'`: HiDPI support for [MapServer](https://mapserver.org/)
   *  - `'qgis'`: HiDPI support for [QGIS](https://qgis.org/)
   */

  /**
   * @module ol/source/ImageWMS
   */

  /**
   * Number of decimal digits to consider in integer values when rounding.
   * @type {number}
   */
  const DECIMALS = 4;

  /**
   * @const
   * @type {import("../size.js").Size}
   */
  const GETFEATUREINFO_IMAGE_SIZE = [101, 101];

  /**
   * @typedef {Object} Options
   * @property {import("./Source.js").AttributionLike} [attributions] Attributions.
   * @property {null|string} [crossOrigin] The `crossOrigin` attribute for loaded images.  Note that
   * you must provide a `crossOrigin` value if you want to access pixel data with the Canvas renderer.
   * See https://developer.mozilla.org/en-US/docs/Web/HTML/CORS_enabled_image for more detail.
   * @property {boolean} [hidpi=true] Use the `ol/Map#pixelRatio` value when requesting
   * the image from the remote server.
   * @property {import("./wms.js").ServerType} [serverType] The type of
   * the remote WMS server: `mapserver`, `geoserver`, `carmentaserver`, or `qgis`.
   * Only needed if `hidpi` is `true`.
   * @property {import("../Image.js").LoadFunction} [imageLoadFunction] Optional function to load an image given a URL.
   * @property {boolean} [interpolate=true] Use interpolated values when resampling.  By default,
   * linear interpolation is used when resampling.  Set to false to use the nearest neighbor instead.
   * @property {Object<string,*>} [params] WMS request parameters.
   * At least a `LAYERS` param is required. `STYLES` is
   * `''` by default. `VERSION` is `1.3.0` by default. `WIDTH`, `HEIGHT`, `BBOX`
   * and `CRS` (`SRS` for WMS version < 1.3.0) will be set dynamically.
   * @property {import("../proj.js").ProjectionLike} [projection] Projection. Default is the view projection.
   * @property {number} [ratio=1.5] Ratio. `1` means image requests are the size of the map viewport, `2` means
   * twice the width and height of the map viewport, and so on. Must be `1` or
   * higher.
   * @property {Array<number>} [resolutions] Resolutions.
   * If specified, requests will be made for these resolutions only.
   * @property {string} [url] WMS service URL.
   */

  /**
   * @classdesc
   * Source for WMS servers providing single, untiled images.
   *
   * @fires module:ol/source/Image.ImageSourceEvent
   * @api
   */
  class ImageWMS extends ImageSource {
    /**
     * @param {Options} [options] ImageWMS options.
     */
    constructor(options) {
      options = options ? options : {};

      super({
        attributions: options.attributions,
        interpolate: options.interpolate,
        projection: options.projection,
        resolutions: options.resolutions,
      });

      /**
       * @private
       * @type {?string}
       */
      this.crossOrigin_ =
        options.crossOrigin !== undefined ? options.crossOrigin : null;

      /**
       * @private
       * @type {string|undefined}
       */
      this.url_ = options.url;

      /**
       * @private
       * @type {import("../Image.js").LoadFunction}
       */
      this.imageLoadFunction_ =
        options.imageLoadFunction !== undefined
          ? options.imageLoadFunction
          : defaultImageLoadFunction;

      /**
       * @private
       * @type {!Object}
       */
      this.params_ = Object.assign({}, options.params);

      /**
       * @private
       * @type {boolean}
       */
      this.v13_ = true;
      this.updateV13_();

      /**
       * @private
       * @type {import("./wms.js").ServerType}
       */
      this.serverType_ = options.serverType;

      /**
       * @private
       * @type {boolean}
       */
      this.hidpi_ = options.hidpi !== undefined ? options.hidpi : true;

      /**
       * @private
       * @type {import("../Image.js").default}
       */
      this.image_ = null;

      /**
       * @private
       * @type {import("../size.js").Size}
       */
      this.imageSize_ = [0, 0];

      /**
       * @private
       * @type {number}
       */
      this.renderedRevision_ = 0;

      /**
       * @private
       * @type {number}
       */
      this.ratio_ = options.ratio !== undefined ? options.ratio : 1.5;
    }

    /**
     * Return the GetFeatureInfo URL for the passed coordinate, resolution, and
     * projection. Return `undefined` if the GetFeatureInfo URL cannot be
     * constructed.
     * @param {import("../coordinate.js").Coordinate} coordinate Coordinate.
     * @param {number} resolution Resolution.
     * @param {import("../proj.js").ProjectionLike} projection Projection.
     * @param {!Object} params GetFeatureInfo params. `INFO_FORMAT` at least should
     *     be provided. If `QUERY_LAYERS` is not provided then the layers specified
     *     in the `LAYERS` parameter will be used. `VERSION` should not be
     *     specified here.
     * @return {string|undefined} GetFeatureInfo URL.
     * @api
     */
    getFeatureInfoUrl(coordinate, resolution, projection, params) {
      if (this.url_ === undefined) {
        return undefined;
      }
      const projectionObj = get$2(projection);
      const sourceProjectionObj = this.getProjection();

      if (sourceProjectionObj && sourceProjectionObj !== projectionObj) {
        resolution = calculateSourceResolution(
          sourceProjectionObj,
          projectionObj,
          coordinate,
          resolution
        );
        coordinate = transform(coordinate, projectionObj, sourceProjectionObj);
      }

      const extent = getForViewAndSize(
        coordinate,
        resolution,
        0,
        GETFEATUREINFO_IMAGE_SIZE
      );

      const baseParams = {
        'SERVICE': 'WMS',
        'VERSION': DEFAULT_VERSION$1,
        'REQUEST': 'GetFeatureInfo',
        'FORMAT': 'image/png',
        'TRANSPARENT': true,
        'QUERY_LAYERS': this.params_['LAYERS'],
      };
      Object.assign(baseParams, this.params_, params);

      const x = floor((coordinate[0] - extent[0]) / resolution, DECIMALS);
      const y = floor((extent[3] - coordinate[1]) / resolution, DECIMALS);
      baseParams[this.v13_ ? 'I' : 'X'] = x;
      baseParams[this.v13_ ? 'J' : 'Y'] = y;

      return this.getRequestUrl_(
        extent,
        GETFEATUREINFO_IMAGE_SIZE,
        1,
        sourceProjectionObj || projectionObj,
        baseParams
      );
    }

    /**
     * Return the GetLegendGraphic URL, optionally optimized for the passed
     * resolution and possibly including any passed specific parameters. Returns
     * `undefined` if the GetLegendGraphic URL cannot be constructed.
     *
     * @param {number} [resolution] Resolution. If set to undefined, `SCALE`
     *     will not be calculated and included in URL.
     * @param {Object} [params] GetLegendGraphic params. If `LAYER` is set, the
     *     request is generated for this wms layer, else it will try to use the
     *     configured wms layer. Default `FORMAT` is `image/png`.
     *     `VERSION` should not be specified here.
     * @return {string|undefined} GetLegendGraphic URL.
     * @api
     */
    getLegendUrl(resolution, params) {
      if (this.url_ === undefined) {
        return undefined;
      }

      const baseParams = {
        'SERVICE': 'WMS',
        'VERSION': DEFAULT_VERSION$1,
        'REQUEST': 'GetLegendGraphic',
        'FORMAT': 'image/png',
      };

      if (params === undefined || params['LAYER'] === undefined) {
        const layers = this.params_.LAYERS;
        const isSingleLayer = !Array.isArray(layers) || layers.length === 1;
        if (!isSingleLayer) {
          return undefined;
        }
        baseParams['LAYER'] = layers;
      }

      if (resolution !== undefined) {
        const mpu = this.getProjection()
          ? this.getProjection().getMetersPerUnit()
          : 1;
        const pixelSize = 0.00028;
        baseParams['SCALE'] = (resolution * mpu) / pixelSize;
      }

      Object.assign(baseParams, params);

      return appendParams(/** @type {string} */ (this.url_), baseParams);
    }

    /**
     * Get the user-provided params, i.e. those passed to the constructor through
     * the "params" option, and possibly updated using the updateParams method.
     * @return {Object} Params.
     * @api
     */
    getParams() {
      return this.params_;
    }

    /**
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {number} resolution Resolution.
     * @param {number} pixelRatio Pixel ratio.
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @return {import("../Image.js").default} Single image.
     */
    getImageInternal(extent, resolution, pixelRatio, projection) {
      if (this.url_ === undefined) {
        return null;
      }

      resolution = this.findNearestResolution(resolution);

      if (pixelRatio != 1 && (!this.hidpi_ || this.serverType_ === undefined)) {
        pixelRatio = 1;
      }

      const imageResolution = resolution / pixelRatio;

      const center = getCenter(extent);
      const viewWidth = ceil(getWidth(extent) / imageResolution, DECIMALS);
      const viewHeight = ceil(getHeight(extent) / imageResolution, DECIMALS);
      const viewExtent = getForViewAndSize(center, imageResolution, 0, [
        viewWidth,
        viewHeight,
      ]);
      const requestWidth = ceil(
        (this.ratio_ * getWidth(extent)) / imageResolution,
        DECIMALS
      );
      const requestHeight = ceil(
        (this.ratio_ * getHeight(extent)) / imageResolution,
        DECIMALS
      );
      const requestExtent = getForViewAndSize(center, imageResolution, 0, [
        requestWidth,
        requestHeight,
      ]);

      const image = this.image_;
      if (
        image &&
        this.renderedRevision_ == this.getRevision() &&
        image.getResolution() == resolution &&
        image.getPixelRatio() == pixelRatio &&
        containsExtent(image.getExtent(), viewExtent)
      ) {
        return image;
      }

      const params = {
        'SERVICE': 'WMS',
        'VERSION': DEFAULT_VERSION$1,
        'REQUEST': 'GetMap',
        'FORMAT': 'image/png',
        'TRANSPARENT': true,
      };
      Object.assign(params, this.params_);

      this.imageSize_[0] = round(
        getWidth(requestExtent) / imageResolution,
        DECIMALS
      );
      this.imageSize_[1] = round(
        getHeight(requestExtent) / imageResolution,
        DECIMALS
      );

      const url = this.getRequestUrl_(
        requestExtent,
        this.imageSize_,
        pixelRatio,
        projection,
        params
      );

      this.image_ = new ImageWrapper(
        requestExtent,
        resolution,
        pixelRatio,
        url,
        this.crossOrigin_,
        this.imageLoadFunction_
      );

      this.renderedRevision_ = this.getRevision();

      this.image_.addEventListener(
        EventType$1.CHANGE,
        this.handleImageChange.bind(this)
      );

      return this.image_;
    }

    /**
     * Return the image load function of the source.
     * @return {import("../Image.js").LoadFunction} The image load function.
     * @api
     */
    getImageLoadFunction() {
      return this.imageLoadFunction_;
    }

    /**
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {import("../size.js").Size} size Size.
     * @param {number} pixelRatio Pixel ratio.
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @param {Object} params Params.
     * @return {string} Request URL.
     * @private
     */
    getRequestUrl_(extent, size, pixelRatio, projection, params) {
      assert(this.url_ !== undefined, 9); // `url` must be configured or set using `#setUrl()`

      params[this.v13_ ? 'CRS' : 'SRS'] = projection.getCode();

      if (!('STYLES' in this.params_)) {
        params['STYLES'] = '';
      }

      if (pixelRatio != 1) {
        switch (this.serverType_) {
          case 'geoserver':
            const dpi = (90 * pixelRatio + 0.5) | 0;
            if ('FORMAT_OPTIONS' in params) {
              params['FORMAT_OPTIONS'] += ';dpi:' + dpi;
            } else {
              params['FORMAT_OPTIONS'] = 'dpi:' + dpi;
            }
            break;
          case 'mapserver':
            params['MAP_RESOLUTION'] = 90 * pixelRatio;
            break;
          case 'carmentaserver':
          case 'qgis':
            params['DPI'] = 90 * pixelRatio;
            break;
          default: // Unknown `serverType` configured
            assert(false, 8);
            break;
        }
      }

      params['WIDTH'] = size[0];
      params['HEIGHT'] = size[1];

      const axisOrientation = projection.getAxisOrientation();
      let bbox;
      if (this.v13_ && axisOrientation.substr(0, 2) == 'ne') {
        bbox = [extent[1], extent[0], extent[3], extent[2]];
      } else {
        bbox = extent;
      }
      params['BBOX'] = bbox.join(',');

      return appendParams(/** @type {string} */ (this.url_), params);
    }

    /**
     * Return the URL used for this WMS source.
     * @return {string|undefined} URL.
     * @api
     */
    getUrl() {
      return this.url_;
    }

    /**
     * Set the image load function of the source.
     * @param {import("../Image.js").LoadFunction} imageLoadFunction Image load function.
     * @api
     */
    setImageLoadFunction(imageLoadFunction) {
      this.image_ = null;
      this.imageLoadFunction_ = imageLoadFunction;
      this.changed();
    }

    /**
     * Set the URL to use for requests.
     * @param {string|undefined} url URL.
     * @api
     */
    setUrl(url) {
      if (url != this.url_) {
        this.url_ = url;
        this.image_ = null;
        this.changed();
      }
    }

    /**
     * Update the user-provided params.
     * @param {Object} params Params.
     * @api
     */
    updateParams(params) {
      Object.assign(this.params_, params);
      this.updateV13_();
      this.image_ = null;
      this.changed();
    }

    /**
     * @private
     */
    updateV13_() {
      const version = this.params_['VERSION'] || DEFAULT_VERSION$1;
      this.v13_ = compareVersions(version, '1.3') >= 0;
    }
  }

  /**
   * @module ol/source/OSM
   */

  /**
   * The attribution containing a link to the OpenStreetMap Copyright and License
   * page.
   * @const
   * @type {string}
   * @api
   */
  const ATTRIBUTION =
    '&#169; ' +
    '<a href="https://www.openstreetmap.org/copyright" target="_blank">OpenStreetMap</a> ' +
    'contributors.';

  /**
   * @typedef {Object} Options
   * @property {import("./Source.js").AttributionLike} [attributions] Attributions.
   * @property {number} [cacheSize] Initial tile cache size. Will auto-grow to hold at least the number of tiles in the viewport.
   * @property {null|string} [crossOrigin='anonymous'] The `crossOrigin` attribute for loaded images.  Note that
   * you must provide a `crossOrigin` value if you want to access pixel data with the Canvas renderer.
   * See https://developer.mozilla.org/en-US/docs/Web/HTML/CORS_enabled_image for more detail.
   * @property {boolean} [interpolate=true] Use interpolated values when resampling.  By default,
   * linear interpolation is used when resampling.  Set to false to use the nearest neighbor instead.
   * @property {number} [maxZoom=19] Max zoom.
   * @property {boolean} [opaque=true] Whether the layer is opaque.
   * @property {number} [reprojectionErrorThreshold=0.5] Maximum allowed reprojection error (in pixels).
   * Higher values can increase reprojection performance, but decrease precision.
   * @property {import("../Tile.js").LoadFunction} [tileLoadFunction] Optional function to load a tile given a URL. The default is
   * ```js
   * function(imageTile, src) {
   *   imageTile.getImage().src = src;
   * };
   * ```
   * @property {number} [transition=250] Duration of the opacity transition for rendering.
   * To disable the opacity transition, pass `transition: 0`.
   * @property {string} [url='https://tile.openstreetmap.org/{z}/{x}/{y}.png'] URL template.
   * Must include `{x}`, `{y}` or `{-y}`, and `{z}` placeholders.
   * @property {boolean} [wrapX=true] Whether to wrap the world horizontally.
   * @property {number|import("../array.js").NearestDirectionFunction} [zDirection=0]
   * Choose whether to use tiles with a higher or lower zoom level when between integer
   * zoom levels. See {@link module:ol/tilegrid/TileGrid~TileGrid#getZForResolution}.
   */

  /**
   * @classdesc
   * Layer source for the OpenStreetMap tile server.
   * @api
   */
  class OSM extends XYZ {
    /**
     * @param {Options} [options] Open Street Map options.
     */
    constructor(options) {
      options = options || {};

      let attributions;
      if (options.attributions !== undefined) {
        attributions = options.attributions;
      } else {
        attributions = [ATTRIBUTION];
      }

      const crossOrigin =
        options.crossOrigin !== undefined ? options.crossOrigin : 'anonymous';

      const url =
        options.url !== undefined
          ? options.url
          : 'https://tile.openstreetmap.org/{z}/{x}/{y}.png';

      super({
        attributions: attributions,
        attributionsCollapsible: false,
        cacheSize: options.cacheSize,
        crossOrigin: crossOrigin,
        interpolate: options.interpolate,
        maxZoom: options.maxZoom !== undefined ? options.maxZoom : 19,
        opaque: options.opaque !== undefined ? options.opaque : true,
        reprojectionErrorThreshold: options.reprojectionErrorThreshold,
        tileLoadFunction: options.tileLoadFunction,
        transition: options.transition,
        url: url,
        wrapX: options.wrapX,
        zDirection: options.zDirection,
      });
    }
  }

  /**
   * @module ol/layer/BaseImage
   */

  /**
   * @template {import("../source/Image.js").default} ImageSourceType
   * @typedef {Object} Options
   * @property {string} [className='ol-layer'] A CSS class name to set to the layer element.
   * @property {number} [opacity=1] Opacity (0, 1).
   * @property {boolean} [visible=true] Visibility.
   * @property {import("../extent.js").Extent} [extent] The bounding extent for layer rendering.  The layer will not be
   * rendered outside of this extent.
   * @property {number} [zIndex] The z-index for layer rendering.  At rendering time, the layers
   * will be ordered, first by Z-index and then by position. When `undefined`, a `zIndex` of 0 is assumed
   * for layers that are added to the map's `layers` collection, or `Infinity` when the layer's `setMap()`
   * method was used.
   * @property {number} [minResolution] The minimum resolution (inclusive) at which this layer will be
   * visible.
   * @property {number} [maxResolution] The maximum resolution (exclusive) below which this layer will
   * be visible.
   * @property {number} [minZoom] The minimum view zoom level (exclusive) above which this layer will be
   * visible.
   * @property {number} [maxZoom] The maximum view zoom level (inclusive) at which this layer will
   * be visible.
   * @property {import("../Map.js").default} [map] Sets the layer as overlay on a map. The map will not manage
   * this layer in its layers collection, and the layer will be rendered on top. This is useful for
   * temporary layers. The standard way to add a layer to a map and have it managed by the map is to
   * use {@link import("../Map.js").default#addLayer map.addLayer()}.
   * @property {ImageSourceType} [source] Source for this layer.
   * @property {Object<string, *>} [properties] Arbitrary observable properties. Can be accessed with `#get()` and `#set()`.
   */

  /**
   * @classdesc
   * Server-rendered images that are available for arbitrary extents and
   * resolutions.
   * Note that any property set in the options is set as a {@link module:ol/Object~BaseObject}
   * property on the layer object; for example, setting `title: 'My Title'` in the
   * options means that `title` is observable, and has get/set accessors.
   *
   * @template {import("../source/Image.js").default} ImageSourceType
   * @template {import("../renderer/Layer.js").default} RendererType
   * @extends {Layer<ImageSourceType, RendererType>}
   * @api
   */
  class BaseImageLayer extends Layer {
    /**
     * @param {Options<ImageSourceType>} [options] Layer options.
     */
    constructor(options) {
      options = options ? options : {};
      super(options);
    }
  }

  /**
   * @module ol/renderer/Layer
   */

  /**
   * @template {import("../layer/Layer.js").default} LayerType
   */
  class LayerRenderer extends Observable {
    /**
     * @param {LayerType} layer Layer.
     */
    constructor(layer) {
      super();

      /**
       * The renderer is initialized and ready to render.
       * @type {boolean}
       */
      this.ready = true;

      /** @private */
      this.boundHandleImageChange_ = this.handleImageChange_.bind(this);

      /**
       * @protected
       * @type {LayerType}
       */
      this.layer_ = layer;

      /**
       * @type {import("../render/canvas/ExecutorGroup").default}
       */
      this.declutterExecutorGroup = null;
    }

    /**
     * Asynchronous layer level hit detection.
     * @param {import("../pixel.js").Pixel} pixel Pixel.
     * @return {Promise<Array<import("../Feature").default>>} Promise that resolves with
     * an array of features.
     */
    getFeatures(pixel) {
      return abstract();
    }

    /**
     * @param {import("../pixel.js").Pixel} pixel Pixel.
     * @return {Uint8ClampedArray|Uint8Array|Float32Array|DataView|null} Pixel data.
     */
    getData(pixel) {
      return null;
    }

    /**
     * Determine whether render should be called.
     * @abstract
     * @param {import("../Map.js").FrameState} frameState Frame state.
     * @return {boolean} Layer is ready to be rendered.
     */
    prepareFrame(frameState) {
      return abstract();
    }

    /**
     * Render the layer.
     * @abstract
     * @param {import("../Map.js").FrameState} frameState Frame state.
     * @param {HTMLElement} target Target that may be used to render content to.
     * @return {HTMLElement} The rendered element.
     */
    renderFrame(frameState, target) {
      return abstract();
    }

    /**
     * @param {Object<number, Object<string, import("../Tile.js").default>>} tiles Lookup of loaded tiles by zoom level.
     * @param {number} zoom Zoom level.
     * @param {import("../Tile.js").default} tile Tile.
     * @return {boolean|void} If `false`, the tile will not be considered loaded.
     */
    loadedTileCallback(tiles, zoom, tile) {
      if (!tiles[zoom]) {
        tiles[zoom] = {};
      }
      tiles[zoom][tile.tileCoord.toString()] = tile;
      return undefined;
    }

    /**
     * Create a function that adds loaded tiles to the tile lookup.
     * @param {import("../source/Tile.js").default} source Tile source.
     * @param {import("../proj/Projection.js").default} projection Projection of the tiles.
     * @param {Object<number, Object<string, import("../Tile.js").default>>} tiles Lookup of loaded tiles by zoom level.
     * @return {function(number, import("../TileRange.js").default):boolean} A function that can be
     *     called with a zoom level and a tile range to add loaded tiles to the lookup.
     * @protected
     */
    createLoadedTileFinder(source, projection, tiles) {
      return (
        /**
         * @param {number} zoom Zoom level.
         * @param {import("../TileRange.js").default} tileRange Tile range.
         * @return {boolean} The tile range is fully loaded.
         * @this {LayerRenderer}
         */
        function (zoom, tileRange) {
          const callback = this.loadedTileCallback.bind(this, tiles, zoom);
          return source.forEachLoadedTile(projection, zoom, tileRange, callback);
        }.bind(this)
      );
    }
    /**
     * @abstract
     * @param {import("../coordinate.js").Coordinate} coordinate Coordinate.
     * @param {import("../Map.js").FrameState} frameState Frame state.
     * @param {number} hitTolerance Hit tolerance in pixels.
     * @param {import("./vector.js").FeatureCallback<T>} callback Feature callback.
     * @param {Array<import("./Map.js").HitMatch<T>>} matches The hit detected matches with tolerance.
     * @return {T|undefined} Callback result.
     * @template T
     */
    forEachFeatureAtCoordinate(
      coordinate,
      frameState,
      hitTolerance,
      callback,
      matches
    ) {
      return undefined;
    }

    /**
     * @return {LayerType} Layer.
     */
    getLayer() {
      return this.layer_;
    }

    /**
     * Perform action necessary to get the layer rendered after new fonts have loaded
     * @abstract
     */
    handleFontsChanged() {}

    /**
     * Handle changes in image state.
     * @param {import("../events/Event.js").default} event Image change event.
     * @private
     */
    handleImageChange_(event) {
      const image = /** @type {import("../Image.js").default} */ (event.target);
      if (image.getState() === ImageState.LOADED) {
        this.renderIfReadyAndVisible();
      }
    }

    /**
     * Load the image if not already loaded, and register the image change
     * listener if needed.
     * @param {import("../ImageBase.js").default} image Image.
     * @return {boolean} `true` if the image is already loaded, `false` otherwise.
     * @protected
     */
    loadImage(image) {
      let imageState = image.getState();
      if (imageState != ImageState.LOADED && imageState != ImageState.ERROR) {
        image.addEventListener(EventType$1.CHANGE, this.boundHandleImageChange_);
      }
      if (imageState == ImageState.IDLE) {
        image.load();
        imageState = image.getState();
      }
      return imageState == ImageState.LOADED;
    }

    /**
     * @protected
     */
    renderIfReadyAndVisible() {
      const layer = this.getLayer();
      if (layer && layer.getVisible() && layer.getSourceState() === 'ready') {
        layer.changed();
      }
    }

    /**
     * Clean up.
     */
    disposeInternal() {
      delete this.layer_;
      super.disposeInternal();
    }
  }

  /**
   * @module ol/renderer/canvas/Layer
   */

  /**
   * @type {Array<HTMLCanvasElement>}
   */
  const canvasPool = [];

  /**
   * @type {CanvasRenderingContext2D}
   */
  let pixelContext$1 = null;

  function createPixelContext$1() {
    const canvas = document.createElement('canvas');
    canvas.width = 1;
    canvas.height = 1;
    pixelContext$1 = canvas.getContext('2d');
  }

  /**
   * @abstract
   * @template {import("../../layer/Layer.js").default} LayerType
   * @extends {LayerRenderer<LayerType>}
   */
  class CanvasLayerRenderer extends LayerRenderer {
    /**
     * @param {LayerType} layer Layer.
     */
    constructor(layer) {
      super(layer);

      /**
       * @protected
       * @type {HTMLElement}
       */
      this.container = null;

      /**
       * @protected
       * @type {number}
       */
      this.renderedResolution;

      /**
       * A temporary transform.  The values in this transform should only be used in a
       * function that sets the values.
       * @protected
       * @type {import("../../transform.js").Transform}
       */
      this.tempTransform = create$2();

      /**
       * The transform for rendered pixels to viewport CSS pixels.  This transform must
       * be set when rendering a frame and may be used by other functions after rendering.
       * @protected
       * @type {import("../../transform.js").Transform}
       */
      this.pixelTransform = create$2();

      /**
       * The transform for viewport CSS pixels to rendered pixels.  This transform must
       * be set when rendering a frame and may be used by other functions after rendering.
       * @protected
       * @type {import("../../transform.js").Transform}
       */
      this.inversePixelTransform = create$2();

      /**
       * @type {CanvasRenderingContext2D}
       */
      this.context = null;

      /**
       * @type {boolean}
       */
      this.containerReused = false;

      /**
       * @private
       * @type {CanvasRenderingContext2D}
       */
      this.pixelContext_ = null;

      /**
       * @protected
       * @type {import("../../Map.js").FrameState|null}
       */
      this.frameState = null;
    }

    /**
     * @param {HTMLCanvasElement|HTMLImageElement|HTMLVideoElement} image Image.
     * @param {number} col The column index.
     * @param {number} row The row index.
     * @return {Uint8ClampedArray|null} The image data.
     */
    getImageData(image, col, row) {
      if (!pixelContext$1) {
        createPixelContext$1();
      }
      pixelContext$1.clearRect(0, 0, 1, 1);

      let data;
      try {
        pixelContext$1.drawImage(image, col, row, 1, 1, 0, 0, 1, 1);
        data = pixelContext$1.getImageData(0, 0, 1, 1).data;
      } catch (err) {
        pixelContext$1 = null;
        return null;
      }
      return data;
    }

    /**
     * @param {import('../../Map.js').FrameState} frameState Frame state.
     * @return {string} Background color.
     */
    getBackground(frameState) {
      const layer = this.getLayer();
      let background = layer.getBackground();
      if (typeof background === 'function') {
        background = background(frameState.viewState.resolution);
      }
      return background || undefined;
    }

    /**
     * Get a rendering container from an existing target, if compatible.
     * @param {HTMLElement} target Potential render target.
     * @param {string} transform CSS Transform.
     * @param {string} [backgroundColor] Background color.
     */
    useContainer(target, transform, backgroundColor) {
      const layerClassName = this.getLayer().getClassName();
      let container, context;
      if (
        target &&
        target.className === layerClassName &&
        (!backgroundColor ||
          (target &&
            target.style.backgroundColor &&
            equals$3(
              asArray(target.style.backgroundColor),
              asArray(backgroundColor)
            )))
      ) {
        const canvas = target.firstElementChild;
        if (canvas instanceof HTMLCanvasElement) {
          context = canvas.getContext('2d');
        }
      }
      if (context && context.canvas.style.transform === transform) {
        // Container of the previous layer renderer can be used.
        this.container = target;
        this.context = context;
        this.containerReused = true;
      } else if (this.containerReused) {
        // Previously reused container cannot be used any more.
        this.container = null;
        this.context = null;
        this.containerReused = false;
      }
      if (!this.container) {
        container = document.createElement('div');
        container.className = layerClassName;
        let style = container.style;
        style.position = 'absolute';
        style.width = '100%';
        style.height = '100%';
        context = createCanvasContext2D();
        const canvas = context.canvas;
        container.appendChild(canvas);
        style = canvas.style;
        style.position = 'absolute';
        style.left = '0';
        style.transformOrigin = 'top left';
        this.container = container;
        this.context = context;
      }
      if (
        !this.containerReused &&
        backgroundColor &&
        !this.container.style.backgroundColor
      ) {
        this.container.style.backgroundColor = backgroundColor;
      }
    }

    /**
     * @param {CanvasRenderingContext2D} context Context.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @param {import("../../extent.js").Extent} extent Clip extent.
     * @protected
     */
    clipUnrotated(context, frameState, extent) {
      const topLeft = getTopLeft(extent);
      const topRight = getTopRight(extent);
      const bottomRight = getBottomRight(extent);
      const bottomLeft = getBottomLeft(extent);

      apply(frameState.coordinateToPixelTransform, topLeft);
      apply(frameState.coordinateToPixelTransform, topRight);
      apply(frameState.coordinateToPixelTransform, bottomRight);
      apply(frameState.coordinateToPixelTransform, bottomLeft);

      const inverted = this.inversePixelTransform;
      apply(inverted, topLeft);
      apply(inverted, topRight);
      apply(inverted, bottomRight);
      apply(inverted, bottomLeft);

      context.save();
      context.beginPath();
      context.moveTo(Math.round(topLeft[0]), Math.round(topLeft[1]));
      context.lineTo(Math.round(topRight[0]), Math.round(topRight[1]));
      context.lineTo(Math.round(bottomRight[0]), Math.round(bottomRight[1]));
      context.lineTo(Math.round(bottomLeft[0]), Math.round(bottomLeft[1]));
      context.clip();
    }

    /**
     * @param {import("../../render/EventType.js").default} type Event type.
     * @param {CanvasRenderingContext2D} context Context.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @private
     */
    dispatchRenderEvent_(type, context, frameState) {
      const layer = this.getLayer();
      if (layer.hasListener(type)) {
        const event = new RenderEvent(
          type,
          this.inversePixelTransform,
          frameState,
          context
        );
        layer.dispatchEvent(event);
      }
    }

    /**
     * @param {CanvasRenderingContext2D} context Context.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @protected
     */
    preRender(context, frameState) {
      this.frameState = frameState;
      this.dispatchRenderEvent_(EventType.PRERENDER, context, frameState);
    }

    /**
     * @param {CanvasRenderingContext2D} context Context.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @protected
     */
    postRender(context, frameState) {
      this.dispatchRenderEvent_(EventType.POSTRENDER, context, frameState);
    }

    /**
     * Creates a transform for rendering to an element that will be rotated after rendering.
     * @param {import("../../coordinate.js").Coordinate} center Center.
     * @param {number} resolution Resolution.
     * @param {number} rotation Rotation.
     * @param {number} pixelRatio Pixel ratio.
     * @param {number} width Width of the rendered element (in pixels).
     * @param {number} height Height of the rendered element (in pixels).
     * @param {number} offsetX Offset on the x-axis in view coordinates.
     * @protected
     * @return {!import("../../transform.js").Transform} Transform.
     */
    getRenderTransform(
      center,
      resolution,
      rotation,
      pixelRatio,
      width,
      height,
      offsetX
    ) {
      const dx1 = width / 2;
      const dy1 = height / 2;
      const sx = pixelRatio / resolution;
      const sy = -sx;
      const dx2 = -center[0] + offsetX;
      const dy2 = -center[1];
      return compose(
        this.tempTransform,
        dx1,
        dy1,
        sx,
        sy,
        -rotation,
        dx2,
        dy2
      );
    }

    /**
     * Clean up.
     */
    disposeInternal() {
      delete this.frameState;
      super.disposeInternal();
    }
  }

  /**
   * @module ol/renderer/canvas/ImageLayer
   */

  /**
   * @classdesc
   * Canvas renderer for image layers.
   * @api
   */
  class CanvasImageLayerRenderer extends CanvasLayerRenderer {
    /**
     * @param {import("../../layer/Image.js").default} imageLayer Image layer.
     */
    constructor(imageLayer) {
      super(imageLayer);

      /**
       * @protected
       * @type {?import("../../ImageBase.js").default}
       */
      this.image_ = null;
    }

    /**
     * @return {HTMLCanvasElement|HTMLImageElement|HTMLVideoElement} Image.
     */
    getImage() {
      return !this.image_ ? null : this.image_.getImage();
    }

    /**
     * Determine whether render should be called.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @return {boolean} Layer is ready to be rendered.
     */
    prepareFrame(frameState) {
      const layerState = frameState.layerStatesArray[frameState.layerIndex];
      const pixelRatio = frameState.pixelRatio;
      const viewState = frameState.viewState;
      const viewResolution = viewState.resolution;

      const imageSource = this.getLayer().getSource();

      const hints = frameState.viewHints;

      let renderedExtent = frameState.extent;
      if (layerState.extent !== undefined) {
        renderedExtent = getIntersection(
          renderedExtent,
          fromUserExtent(layerState.extent, viewState.projection)
        );
      }

      if (
        !hints[ViewHint.ANIMATING] &&
        !hints[ViewHint.INTERACTING] &&
        !isEmpty(renderedExtent)
      ) {
        if (imageSource) {
          const projection = viewState.projection;
          const image = imageSource.getImage(
            renderedExtent,
            viewResolution,
            pixelRatio,
            projection
          );
          if (image) {
            if (this.loadImage(image)) {
              this.image_ = image;
            } else if (image.getState() === ImageState.EMPTY) {
              this.image_ = null;
            }
          }
        } else {
          this.image_ = null;
        }
      }

      return !!this.image_;
    }

    /**
     * @param {import("../../pixel.js").Pixel} pixel Pixel.
     * @return {Uint8ClampedArray} Data at the pixel location.
     */
    getData(pixel) {
      const frameState = this.frameState;
      if (!frameState) {
        return null;
      }

      const layer = this.getLayer();
      const coordinate = apply(
        frameState.pixelToCoordinateTransform,
        pixel.slice()
      );

      const layerExtent = layer.getExtent();
      if (layerExtent) {
        if (!containsCoordinate(layerExtent, coordinate)) {
          return null;
        }
      }

      const imageExtent = this.image_.getExtent();
      const img = this.image_.getImage();

      const imageMapWidth = getWidth(imageExtent);
      const col = Math.floor(
        img.width * ((coordinate[0] - imageExtent[0]) / imageMapWidth)
      );
      if (col < 0 || col >= img.width) {
        return null;
      }

      const imageMapHeight = getHeight(imageExtent);
      const row = Math.floor(
        img.height * ((imageExtent[3] - coordinate[1]) / imageMapHeight)
      );
      if (row < 0 || row >= img.height) {
        return null;
      }

      return this.getImageData(img, col, row);
    }

    /**
     * Render the layer.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @param {HTMLElement} target Target that may be used to render content to.
     * @return {HTMLElement} The rendered element.
     */
    renderFrame(frameState, target) {
      const image = this.image_;
      const imageExtent = image.getExtent();
      const imageResolution = image.getResolution();
      const imagePixelRatio = image.getPixelRatio();
      const layerState = frameState.layerStatesArray[frameState.layerIndex];
      const pixelRatio = frameState.pixelRatio;
      const viewState = frameState.viewState;
      const viewCenter = viewState.center;
      const viewResolution = viewState.resolution;
      const scale =
        (pixelRatio * imageResolution) / (viewResolution * imagePixelRatio);

      const extent = frameState.extent;
      const resolution = viewState.resolution;
      const rotation = viewState.rotation;
      // desired dimensions of the canvas in pixels
      const width = Math.round((getWidth(extent) / resolution) * pixelRatio);
      const height = Math.round((getHeight(extent) / resolution) * pixelRatio);

      // set forward and inverse pixel transforms
      compose(
        this.pixelTransform,
        frameState.size[0] / 2,
        frameState.size[1] / 2,
        1 / pixelRatio,
        1 / pixelRatio,
        rotation,
        -width / 2,
        -height / 2
      );
      makeInverse(this.inversePixelTransform, this.pixelTransform);

      const canvasTransform = toString$3(this.pixelTransform);

      this.useContainer(target, canvasTransform, this.getBackground(frameState));

      const context = this.context;
      const canvas = context.canvas;

      if (canvas.width != width || canvas.height != height) {
        canvas.width = width;
        canvas.height = height;
      } else if (!this.containerReused) {
        context.clearRect(0, 0, width, height);
      }

      // clipped rendering if layer extent is set
      let clipped = false;
      let render = true;
      if (layerState.extent) {
        const layerExtent = fromUserExtent(
          layerState.extent,
          viewState.projection
        );
        render = intersects$3(layerExtent, frameState.extent);
        clipped = render && !containsExtent(layerExtent, frameState.extent);
        if (clipped) {
          this.clipUnrotated(context, frameState, layerExtent);
        }
      }

      const img = image.getImage();

      const transform = compose(
        this.tempTransform,
        width / 2,
        height / 2,
        scale,
        scale,
        0,
        (imagePixelRatio * (imageExtent[0] - viewCenter[0])) / imageResolution,
        (imagePixelRatio * (viewCenter[1] - imageExtent[3])) / imageResolution
      );

      this.renderedResolution = (imageResolution * pixelRatio) / imagePixelRatio;

      const dw = img.width * transform[0];
      const dh = img.height * transform[3];

      if (!this.getLayer().getSource().getInterpolate()) {
        context.imageSmoothingEnabled = false;
      }

      this.preRender(context, frameState);
      if (render && dw >= 0.5 && dh >= 0.5) {
        const dx = transform[4];
        const dy = transform[5];
        const opacity = layerState.opacity;
        let previousAlpha;
        if (opacity !== 1) {
          previousAlpha = context.globalAlpha;
          context.globalAlpha = opacity;
        }
        context.drawImage(img, 0, 0, +img.width, +img.height, dx, dy, dw, dh);
        if (opacity !== 1) {
          context.globalAlpha = previousAlpha;
        }
      }
      this.postRender(context, frameState);

      if (clipped) {
        context.restore();
      }
      context.imageSmoothingEnabled = true;

      if (canvasTransform !== canvas.style.transform) {
        canvas.style.transform = canvasTransform;
      }

      return this.container;
    }
  }

  /**
   * @module ol/layer/Image
   */

  /**
   * @classdesc
   * Server-rendered images that are available for arbitrary extents and
   * resolutions.
   * Note that any property set in the options is set as a {@link module:ol/Object~BaseObject}
   * property on the layer object; for example, setting `title: 'My Title'` in the
   * options means that `title` is observable, and has get/set accessors.
   *
   * @template {import("../source/Image.js").default} ImageSourceType
   * @extends {BaseImageLayer<ImageSourceType, CanvasImageLayerRenderer>}
   * @api
   */
  class ImageLayer extends BaseImageLayer {
    /**
     * @param {import("./BaseImage.js").Options<ImageSourceType>} [options] Layer options.
     */
    constructor(options) {
      super(options);
    }

    createRenderer() {
      return new CanvasImageLayerRenderer(this);
    }

    /**
     * Get data for a pixel location.  A four element RGBA array will be returned.  For requests outside the
     * layer extent, `null` will be returned.  Data for an image can only be retrieved if the
     * source's `crossOrigin` property is set.
     *
     * ```js
     * // display layer data on every pointer move
     * map.on('pointermove', (event) => {
     *   console.log(layer.getData(event.pixel));
     * });
     * ```
     * @param {import("../pixel").Pixel} pixel Pixel.
     * @return {Uint8ClampedArray|Uint8Array|Float32Array|DataView|null} Pixel data.
     * @api
     */
    getData(pixel) {
      return super.getData(pixel);
    }
  }

  /**
   * @module ol/layer/TileProperty
   */

  /**
   * @enum {string}
   */
  var TileProperty = {
    PRELOAD: 'preload',
    USE_INTERIM_TILES_ON_ERROR: 'useInterimTilesOnError',
  };

  /**
   * @module ol/layer/BaseTile
   */

  /***
   * @template Return
   * @typedef {import("../Observable").OnSignature<import("../Observable").EventTypes, import("../events/Event.js").default, Return> &
   *   import("../Observable").OnSignature<import("./Base").BaseLayerObjectEventTypes|
   *     'change:source'|'change:preload'|'change:useInterimTilesOnError', import("../Object").ObjectEvent, Return> &
   *   import("../Observable").OnSignature<import("../render/EventType").LayerRenderEventTypes, import("../render/Event").default, Return> &
   *   import("../Observable").CombinedOnSignature<import("../Observable").EventTypes|import("./Base").BaseLayerObjectEventTypes|
   *   'change:source'|'change:preload'|'change:useInterimTilesOnError'|import("../render/EventType").LayerRenderEventTypes, Return>} BaseTileLayerOnSignature
   */

  /**
   * @template {import("../source/Tile.js").default} TileSourceType
   * @typedef {Object} Options
   * @property {string} [className='ol-layer'] A CSS class name to set to the layer element.
   * @property {number} [opacity=1] Opacity (0, 1).
   * @property {boolean} [visible=true] Visibility.
   * @property {import("../extent.js").Extent} [extent] The bounding extent for layer rendering.  The layer will not be
   * rendered outside of this extent.
   * @property {number} [zIndex] The z-index for layer rendering.  At rendering time, the layers
   * will be ordered, first by Z-index and then by position. When `undefined`, a `zIndex` of 0 is assumed
   * for layers that are added to the map's `layers` collection, or `Infinity` when the layer's `setMap()`
   * method was used.
   * @property {number} [minResolution] The minimum resolution (inclusive) at which this layer will be
   * visible.
   * @property {number} [maxResolution] The maximum resolution (exclusive) below which this layer will
   * be visible.
   * @property {number} [minZoom] The minimum view zoom level (exclusive) above which this layer will be
   * visible.
   * @property {number} [maxZoom] The maximum view zoom level (inclusive) at which this layer will
   * be visible.
   * @property {number} [preload=0] Preload. Load low-resolution tiles up to `preload` levels. `0`
   * means no preloading.
   * @property {TileSourceType} [source] Source for this layer.
   * @property {import("../Map.js").default} [map] Sets the layer as overlay on a map. The map will not manage
   * this layer in its layers collection, and the layer will be rendered on top. This is useful for
   * temporary layers. The standard way to add a layer to a map and have it managed by the map is to
   * use {@link import("../Map.js").default#addLayer map.addLayer()}.
   * @property {boolean} [useInterimTilesOnError=true] Use interim tiles on error.
   * @property {Object<string, *>} [properties] Arbitrary observable properties. Can be accessed with `#get()` and `#set()`.
   */

  /**
   * @classdesc
   * For layer sources that provide pre-rendered, tiled images in grids that are
   * organized by zoom levels for specific resolutions.
   * Note that any property set in the options is set as a {@link module:ol/Object~BaseObject}
   * property on the layer object; for example, setting `title: 'My Title'` in the
   * options means that `title` is observable, and has get/set accessors.
   *
   * @template {import("../source/Tile.js").default} TileSourceType
   * @template {import("../renderer/Layer.js").default} RendererType
   * @extends {Layer<TileSourceType, RendererType>}
   * @api
   */
  class BaseTileLayer extends Layer {
    /**
     * @param {Options<TileSourceType>} [options] Tile layer options.
     */
    constructor(options) {
      options = options ? options : {};

      const baseOptions = Object.assign({}, options);

      delete baseOptions.preload;
      delete baseOptions.useInterimTilesOnError;
      super(baseOptions);

      /***
       * @type {BaseTileLayerOnSignature<import("../events").EventsKey>}
       */
      this.on;

      /***
       * @type {BaseTileLayerOnSignature<import("../events").EventsKey>}
       */
      this.once;

      /***
       * @type {BaseTileLayerOnSignature<void>}
       */
      this.un;

      this.setPreload(options.preload !== undefined ? options.preload : 0);
      this.setUseInterimTilesOnError(
        options.useInterimTilesOnError !== undefined
          ? options.useInterimTilesOnError
          : true
      );
    }

    /**
     * Return the level as number to which we will preload tiles up to.
     * @return {number} The level to preload tiles up to.
     * @observable
     * @api
     */
    getPreload() {
      return /** @type {number} */ (this.get(TileProperty.PRELOAD));
    }

    /**
     * Set the level as number to which we will preload tiles up to.
     * @param {number} preload The level to preload tiles up to.
     * @observable
     * @api
     */
    setPreload(preload) {
      this.set(TileProperty.PRELOAD, preload);
    }

    /**
     * Whether we use interim tiles on error.
     * @return {boolean} Use interim tiles on error.
     * @observable
     * @api
     */
    getUseInterimTilesOnError() {
      return /** @type {boolean} */ (
        this.get(TileProperty.USE_INTERIM_TILES_ON_ERROR)
      );
    }

    /**
     * Set whether we use interim tiles on error.
     * @param {boolean} useInterimTilesOnError Use interim tiles on error.
     * @observable
     * @api
     */
    setUseInterimTilesOnError(useInterimTilesOnError) {
      this.set(TileProperty.USE_INTERIM_TILES_ON_ERROR, useInterimTilesOnError);
    }

    /**
     * Get data for a pixel location.  The return type depends on the source data.  For image tiles,
     * a four element RGBA array will be returned.  For data tiles, the array length will match the
     * number of bands in the dataset.  For requests outside the layer extent, `null` will be returned.
     * Data for a image tiles can only be retrieved if the source's `crossOrigin` property is set.
     *
     * ```js
     * // display layer data on every pointer move
     * map.on('pointermove', (event) => {
     *   console.log(layer.getData(event.pixel));
     * });
     * ```
     * @param {import("../pixel").Pixel} pixel Pixel.
     * @return {Uint8ClampedArray|Uint8Array|Float32Array|DataView|null} Pixel data.
     * @api
     */
    getData(pixel) {
      return super.getData(pixel);
    }
  }

  /**
   * @module ol/renderer/canvas/TileLayer
   */

  /**
   * @classdesc
   * Canvas renderer for tile layers.
   * @api
   * @template {import("../../layer/Tile.js").default<import("../../source/Tile.js").default>|import("../../layer/VectorTile.js").default} [LayerType=import("../../layer/Tile.js").default<import("../../source/Tile.js").default>|import("../../layer/VectorTile.js").default]
   * @extends {CanvasLayerRenderer<LayerType>}
   */
  class CanvasTileLayerRenderer extends CanvasLayerRenderer {
    /**
     * @param {LayerType} tileLayer Tile layer.
     */
    constructor(tileLayer) {
      super(tileLayer);

      /**
       * Rendered extent has changed since the previous `renderFrame()` call
       * @type {boolean}
       */
      this.extentChanged = true;

      /**
       * @private
       * @type {?import("../../extent.js").Extent}
       */
      this.renderedExtent_ = null;

      /**
       * @protected
       * @type {number}
       */
      this.renderedPixelRatio;

      /**
       * @protected
       * @type {import("../../proj/Projection.js").default}
       */
      this.renderedProjection = null;

      /**
       * @protected
       * @type {number}
       */
      this.renderedRevision;

      /**
       * @protected
       * @type {!Array<import("../../Tile.js").default>}
       */
      this.renderedTiles = [];

      /**
       * @private
       * @type {boolean}
       */
      this.newTiles_ = false;

      /**
       * @protected
       * @type {import("../../extent.js").Extent}
       */
      this.tmpExtent = createEmpty();

      /**
       * @private
       * @type {import("../../TileRange.js").default}
       */
      this.tmpTileRange_ = new TileRange(0, 0, 0, 0);
    }

    /**
     * @protected
     * @param {import("../../Tile.js").default} tile Tile.
     * @return {boolean} Tile is drawable.
     */
    isDrawableTile(tile) {
      const tileLayer = this.getLayer();
      const tileState = tile.getState();
      const useInterimTilesOnError = tileLayer.getUseInterimTilesOnError();
      return (
        tileState == TileState.LOADED ||
        tileState == TileState.EMPTY ||
        (tileState == TileState.ERROR && !useInterimTilesOnError)
      );
    }

    /**
     * @param {number} z Tile coordinate z.
     * @param {number} x Tile coordinate x.
     * @param {number} y Tile coordinate y.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @return {!import("../../Tile.js").default} Tile.
     */
    getTile(z, x, y, frameState) {
      const pixelRatio = frameState.pixelRatio;
      const projection = frameState.viewState.projection;
      const tileLayer = this.getLayer();
      const tileSource = tileLayer.getSource();
      let tile = tileSource.getTile(z, x, y, pixelRatio, projection);
      if (tile.getState() == TileState.ERROR) {
        if (tileLayer.getUseInterimTilesOnError() && tileLayer.getPreload() > 0) {
          // Preloaded tiles for lower resolutions might have finished loading.
          this.newTiles_ = true;
        }
      }
      if (!this.isDrawableTile(tile)) {
        tile = tile.getInterimTile();
      }
      return tile;
    }

    /**
     * @param {import("../../pixel.js").Pixel} pixel Pixel.
     * @return {Uint8ClampedArray} Data at the pixel location.
     */
    getData(pixel) {
      const frameState = this.frameState;
      if (!frameState) {
        return null;
      }

      const layer = this.getLayer();
      const coordinate = apply(
        frameState.pixelToCoordinateTransform,
        pixel.slice()
      );

      const layerExtent = layer.getExtent();
      if (layerExtent) {
        if (!containsCoordinate(layerExtent, coordinate)) {
          return null;
        }
      }

      const pixelRatio = frameState.pixelRatio;
      const projection = frameState.viewState.projection;
      const viewState = frameState.viewState;
      const source = layer.getRenderSource();
      const tileGrid = source.getTileGridForProjection(viewState.projection);
      const tilePixelRatio = source.getTilePixelRatio(frameState.pixelRatio);

      for (
        let z = tileGrid.getZForResolution(viewState.resolution);
        z >= tileGrid.getMinZoom();
        --z
      ) {
        const tileCoord = tileGrid.getTileCoordForCoordAndZ(coordinate, z);
        const tile = source.getTile(
          z,
          tileCoord[1],
          tileCoord[2],
          pixelRatio,
          projection
        );
        if (
          !(tile instanceof ImageTile || tile instanceof ReprojTile) ||
          (tile instanceof ReprojTile && tile.getState() === TileState.EMPTY)
        ) {
          return null;
        }

        if (tile.getState() !== TileState.LOADED) {
          continue;
        }

        const tileOrigin = tileGrid.getOrigin(z);
        const tileSize = toSize(tileGrid.getTileSize(z));
        const tileResolution = tileGrid.getResolution(z);

        const col = Math.floor(
          tilePixelRatio *
            ((coordinate[0] - tileOrigin[0]) / tileResolution -
              tileCoord[1] * tileSize[0])
        );

        const row = Math.floor(
          tilePixelRatio *
            ((tileOrigin[1] - coordinate[1]) / tileResolution -
              tileCoord[2] * tileSize[1])
        );

        const gutter = Math.round(
          tilePixelRatio * source.getGutterForProjection(viewState.projection)
        );

        return this.getImageData(tile.getImage(), col + gutter, row + gutter);
      }

      return null;
    }

    /**
     * @param {Object<number, Object<string, import("../../Tile.js").default>>} tiles Lookup of loaded tiles by zoom level.
     * @param {number} zoom Zoom level.
     * @param {import("../../Tile.js").default} tile Tile.
     * @return {boolean|void} If `false`, the tile will not be considered loaded.
     */
    loadedTileCallback(tiles, zoom, tile) {
      if (this.isDrawableTile(tile)) {
        return super.loadedTileCallback(tiles, zoom, tile);
      }
      return false;
    }

    /**
     * Determine whether render should be called.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @return {boolean} Layer is ready to be rendered.
     */
    prepareFrame(frameState) {
      return !!this.getLayer().getSource();
    }

    /**
     * Render the layer.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @param {HTMLElement} target Target that may be used to render content to.
     * @return {HTMLElement} The rendered element.
     */
    renderFrame(frameState, target) {
      const layerState = frameState.layerStatesArray[frameState.layerIndex];
      const viewState = frameState.viewState;
      const projection = viewState.projection;
      const viewResolution = viewState.resolution;
      const viewCenter = viewState.center;
      const rotation = viewState.rotation;
      const pixelRatio = frameState.pixelRatio;

      const tileLayer = this.getLayer();
      const tileSource = tileLayer.getSource();
      const sourceRevision = tileSource.getRevision();
      const tileGrid = tileSource.getTileGridForProjection(projection);
      const z = tileGrid.getZForResolution(viewResolution, tileSource.zDirection);
      const tileResolution = tileGrid.getResolution(z);

      let extent = frameState.extent;
      const resolution = frameState.viewState.resolution;
      const tilePixelRatio = tileSource.getTilePixelRatio(pixelRatio);
      // desired dimensions of the canvas in pixels
      const width = Math.round((getWidth(extent) / resolution) * pixelRatio);
      const height = Math.round((getHeight(extent) / resolution) * pixelRatio);

      const layerExtent =
        layerState.extent && fromUserExtent(layerState.extent, projection);
      if (layerExtent) {
        extent = getIntersection(
          extent,
          fromUserExtent(layerState.extent, projection)
        );
      }

      const dx = (tileResolution * width) / 2 / tilePixelRatio;
      const dy = (tileResolution * height) / 2 / tilePixelRatio;
      const canvasExtent = [
        viewCenter[0] - dx,
        viewCenter[1] - dy,
        viewCenter[0] + dx,
        viewCenter[1] + dy,
      ];

      const tileRange = tileGrid.getTileRangeForExtentAndZ(extent, z);

      /**
       * @type {Object<number, Object<string, import("../../Tile.js").default>>}
       */
      const tilesToDrawByZ = {};
      tilesToDrawByZ[z] = {};

      const findLoadedTiles = this.createLoadedTileFinder(
        tileSource,
        projection,
        tilesToDrawByZ
      );

      const tmpExtent = this.tmpExtent;
      const tmpTileRange = this.tmpTileRange_;
      this.newTiles_ = false;
      const viewport = rotation
        ? getRotatedViewport(
            viewState.center,
            resolution,
            rotation,
            frameState.size
          )
        : undefined;
      for (let x = tileRange.minX; x <= tileRange.maxX; ++x) {
        for (let y = tileRange.minY; y <= tileRange.maxY; ++y) {
          if (
            rotation &&
            !tileGrid.tileCoordIntersectsViewport([z, x, y], viewport)
          ) {
            continue;
          }
          const tile = this.getTile(z, x, y, frameState);
          if (this.isDrawableTile(tile)) {
            const uid = getUid(this);
            if (tile.getState() == TileState.LOADED) {
              tilesToDrawByZ[z][tile.tileCoord.toString()] = tile;
              let inTransition = tile.inTransition(uid);
              if (inTransition && layerState.opacity !== 1) {
                // Skipping transition when layer is not fully opaque avoids visual artifacts.
                tile.endTransition(uid);
                inTransition = false;
              }
              if (
                !this.newTiles_ &&
                (inTransition || !this.renderedTiles.includes(tile))
              ) {
                this.newTiles_ = true;
              }
            }
            if (tile.getAlpha(uid, frameState.time) === 1) {
              // don't look for alt tiles if alpha is 1
              continue;
            }
          }

          const childTileRange = tileGrid.getTileCoordChildTileRange(
            tile.tileCoord,
            tmpTileRange,
            tmpExtent
          );

          let covered = false;
          if (childTileRange) {
            covered = findLoadedTiles(z + 1, childTileRange);
          }
          if (!covered) {
            tileGrid.forEachTileCoordParentTileRange(
              tile.tileCoord,
              findLoadedTiles,
              tmpTileRange,
              tmpExtent
            );
          }
        }
      }

      const canvasScale =
        ((tileResolution / viewResolution) * pixelRatio) / tilePixelRatio;

      // set forward and inverse pixel transforms
      compose(
        this.pixelTransform,
        frameState.size[0] / 2,
        frameState.size[1] / 2,
        1 / pixelRatio,
        1 / pixelRatio,
        rotation,
        -width / 2,
        -height / 2
      );

      const canvasTransform = toString$3(this.pixelTransform);

      this.useContainer(target, canvasTransform, this.getBackground(frameState));
      const context = this.context;
      const canvas = context.canvas;

      makeInverse(this.inversePixelTransform, this.pixelTransform);

      // set scale transform for calculating tile positions on the canvas
      compose(
        this.tempTransform,
        width / 2,
        height / 2,
        canvasScale,
        canvasScale,
        0,
        -width / 2,
        -height / 2
      );

      if (canvas.width != width || canvas.height != height) {
        canvas.width = width;
        canvas.height = height;
      } else if (!this.containerReused) {
        context.clearRect(0, 0, width, height);
      }

      if (layerExtent) {
        this.clipUnrotated(context, frameState, layerExtent);
      }

      if (!tileSource.getInterpolate()) {
        context.imageSmoothingEnabled = false;
      }

      this.preRender(context, frameState);

      this.renderedTiles.length = 0;
      /** @type {Array<number>} */
      let zs = Object.keys(tilesToDrawByZ).map(Number);
      zs.sort(numberSafeCompareFunction);

      let clips, clipZs, currentClip;
      if (
        layerState.opacity === 1 &&
        (!this.containerReused ||
          tileSource.getOpaque(frameState.viewState.projection))
      ) {
        zs = zs.reverse();
      } else {
        clips = [];
        clipZs = [];
      }
      for (let i = zs.length - 1; i >= 0; --i) {
        const currentZ = zs[i];
        const currentTilePixelSize = tileSource.getTilePixelSize(
          currentZ,
          pixelRatio,
          projection
        );
        const currentResolution = tileGrid.getResolution(currentZ);
        const currentScale = currentResolution / tileResolution;
        const dx = currentTilePixelSize[0] * currentScale * canvasScale;
        const dy = currentTilePixelSize[1] * currentScale * canvasScale;
        const originTileCoord = tileGrid.getTileCoordForCoordAndZ(
          getTopLeft(canvasExtent),
          currentZ
        );
        const originTileExtent = tileGrid.getTileCoordExtent(originTileCoord);
        const origin = apply(this.tempTransform, [
          (tilePixelRatio * (originTileExtent[0] - canvasExtent[0])) /
            tileResolution,
          (tilePixelRatio * (canvasExtent[3] - originTileExtent[3])) /
            tileResolution,
        ]);
        const tileGutter =
          tilePixelRatio * tileSource.getGutterForProjection(projection);
        const tilesToDraw = tilesToDrawByZ[currentZ];
        for (const tileCoordKey in tilesToDraw) {
          const tile = /** @type {import("../../ImageTile.js").default} */ (
            tilesToDraw[tileCoordKey]
          );
          const tileCoord = tile.tileCoord;

          // Calculate integer positions and sizes so that tiles align
          const xIndex = originTileCoord[1] - tileCoord[1];
          const nextX = Math.round(origin[0] - (xIndex - 1) * dx);
          const yIndex = originTileCoord[2] - tileCoord[2];
          const nextY = Math.round(origin[1] - (yIndex - 1) * dy);
          const x = Math.round(origin[0] - xIndex * dx);
          const y = Math.round(origin[1] - yIndex * dy);
          const w = nextX - x;
          const h = nextY - y;
          const transition = z === currentZ;

          const inTransition =
            transition && tile.getAlpha(getUid(this), frameState.time) !== 1;
          let contextSaved = false;
          if (!inTransition) {
            if (clips) {
              // Clip mask for regions in this tile that already filled by a higher z tile
              currentClip = [x, y, x + w, y, x + w, y + h, x, y + h];
              for (let i = 0, ii = clips.length; i < ii; ++i) {
                if (z !== currentZ && currentZ < clipZs[i]) {
                  const clip = clips[i];
                  if (
                    intersects$3(
                      [x, y, x + w, y + h],
                      [clip[0], clip[3], clip[4], clip[7]]
                    )
                  ) {
                    if (!contextSaved) {
                      context.save();
                      contextSaved = true;
                    }
                    context.beginPath();
                    // counter-clockwise (outer ring) for current tile
                    context.moveTo(currentClip[0], currentClip[1]);
                    context.lineTo(currentClip[2], currentClip[3]);
                    context.lineTo(currentClip[4], currentClip[5]);
                    context.lineTo(currentClip[6], currentClip[7]);
                    // clockwise (inner ring) for higher z tile
                    context.moveTo(clip[6], clip[7]);
                    context.lineTo(clip[4], clip[5]);
                    context.lineTo(clip[2], clip[3]);
                    context.lineTo(clip[0], clip[1]);
                    context.clip();
                  }
                }
              }
              clips.push(currentClip);
              clipZs.push(currentZ);
            } else {
              context.clearRect(x, y, w, h);
            }
          }
          this.drawTileImage(
            tile,
            frameState,
            x,
            y,
            w,
            h,
            tileGutter,
            transition
          );
          if (clips && !inTransition) {
            if (contextSaved) {
              context.restore();
            }
            this.renderedTiles.unshift(tile);
          } else {
            this.renderedTiles.push(tile);
          }
          this.updateUsedTiles(frameState.usedTiles, tileSource, tile);
        }
      }

      this.renderedRevision = sourceRevision;
      this.renderedResolution = tileResolution;
      this.extentChanged =
        !this.renderedExtent_ || !equals$2(this.renderedExtent_, canvasExtent);
      this.renderedExtent_ = canvasExtent;
      this.renderedPixelRatio = pixelRatio;
      this.renderedProjection = projection;

      this.manageTilePyramid(
        frameState,
        tileSource,
        tileGrid,
        pixelRatio,
        projection,
        extent,
        z,
        tileLayer.getPreload()
      );
      this.scheduleExpireCache(frameState, tileSource);

      this.postRender(context, frameState);

      if (layerState.extent) {
        context.restore();
      }
      context.imageSmoothingEnabled = true;

      if (canvasTransform !== canvas.style.transform) {
        canvas.style.transform = canvasTransform;
      }

      return this.container;
    }

    /**
     * @param {import("../../ImageTile.js").default} tile Tile.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @param {number} x Left of the tile.
     * @param {number} y Top of the tile.
     * @param {number} w Width of the tile.
     * @param {number} h Height of the tile.
     * @param {number} gutter Tile gutter.
     * @param {boolean} transition Apply an alpha transition.
     */
    drawTileImage(tile, frameState, x, y, w, h, gutter, transition) {
      const image = this.getTileImage(tile);
      if (!image) {
        return;
      }
      const uid = getUid(this);
      const layerState = frameState.layerStatesArray[frameState.layerIndex];
      const alpha =
        layerState.opacity *
        (transition ? tile.getAlpha(uid, frameState.time) : 1);
      const alphaChanged = alpha !== this.context.globalAlpha;
      if (alphaChanged) {
        this.context.save();
        this.context.globalAlpha = alpha;
      }
      this.context.drawImage(
        image,
        gutter,
        gutter,
        image.width - 2 * gutter,
        image.height - 2 * gutter,
        x,
        y,
        w,
        h
      );

      if (alphaChanged) {
        this.context.restore();
      }
      if (alpha !== layerState.opacity) {
        frameState.animate = true;
      } else if (transition) {
        tile.endTransition(uid);
      }
    }

    /**
     * @return {HTMLCanvasElement} Image
     */
    getImage() {
      const context = this.context;
      return context ? context.canvas : null;
    }

    /**
     * Get the image from a tile.
     * @param {import("../../ImageTile.js").default} tile Tile.
     * @return {HTMLCanvasElement|HTMLImageElement|HTMLVideoElement} Image.
     * @protected
     */
    getTileImage(tile) {
      return tile.getImage();
    }

    /**
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @param {import("../../source/Tile.js").default} tileSource Tile source.
     * @protected
     */
    scheduleExpireCache(frameState, tileSource) {
      if (tileSource.canExpireCache()) {
        /**
         * @param {import("../../source/Tile.js").default} tileSource Tile source.
         * @param {import("../../Map.js").default} map Map.
         * @param {import("../../Map.js").FrameState} frameState Frame state.
         */
        const postRenderFunction = function (tileSource, map, frameState) {
          const tileSourceKey = getUid(tileSource);
          if (tileSourceKey in frameState.usedTiles) {
            tileSource.expireCache(
              frameState.viewState.projection,
              frameState.usedTiles[tileSourceKey]
            );
          }
        }.bind(null, tileSource);

        frameState.postRenderFunctions.push(
          /** @type {import("../../Map.js").PostRenderFunction} */ (
            postRenderFunction
          )
        );
      }
    }

    /**
     * @param {!Object<string, !Object<string, boolean>>} usedTiles Used tiles.
     * @param {import("../../source/Tile.js").default} tileSource Tile source.
     * @param {import('../../Tile.js').default} tile Tile.
     * @protected
     */
    updateUsedTiles(usedTiles, tileSource, tile) {
      // FIXME should we use tilesToDrawByZ instead?
      const tileSourceKey = getUid(tileSource);
      if (!(tileSourceKey in usedTiles)) {
        usedTiles[tileSourceKey] = {};
      }
      usedTiles[tileSourceKey][tile.getKey()] = true;
    }

    /**
     * Manage tile pyramid.
     * This function performs a number of functions related to the tiles at the
     * current zoom and lower zoom levels:
     * - registers idle tiles in frameState.wantedTiles so that they are not
     *   discarded by the tile queue
     * - enqueues missing tiles
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @param {import("../../source/Tile.js").default} tileSource Tile source.
     * @param {import("../../tilegrid/TileGrid.js").default} tileGrid Tile grid.
     * @param {number} pixelRatio Pixel ratio.
     * @param {import("../../proj/Projection.js").default} projection Projection.
     * @param {import("../../extent.js").Extent} extent Extent.
     * @param {number} currentZ Current Z.
     * @param {number} preload Load low resolution tiles up to `preload` levels.
     * @param {function(import("../../Tile.js").default):void} [tileCallback] Tile callback.
     * @protected
     */
    manageTilePyramid(
      frameState,
      tileSource,
      tileGrid,
      pixelRatio,
      projection,
      extent,
      currentZ,
      preload,
      tileCallback
    ) {
      const tileSourceKey = getUid(tileSource);
      if (!(tileSourceKey in frameState.wantedTiles)) {
        frameState.wantedTiles[tileSourceKey] = {};
      }
      const wantedTiles = frameState.wantedTiles[tileSourceKey];
      const tileQueue = frameState.tileQueue;
      const minZoom = tileGrid.getMinZoom();
      const rotation = frameState.viewState.rotation;
      const viewport = rotation
        ? getRotatedViewport(
            frameState.viewState.center,
            frameState.viewState.resolution,
            rotation,
            frameState.size
          )
        : undefined;
      let tileCount = 0;
      let tile, tileRange, tileResolution, x, y, z;
      for (z = minZoom; z <= currentZ; ++z) {
        tileRange = tileGrid.getTileRangeForExtentAndZ(extent, z, tileRange);
        tileResolution = tileGrid.getResolution(z);
        for (x = tileRange.minX; x <= tileRange.maxX; ++x) {
          for (y = tileRange.minY; y <= tileRange.maxY; ++y) {
            if (
              rotation &&
              !tileGrid.tileCoordIntersectsViewport([z, x, y], viewport)
            ) {
              continue;
            }
            if (currentZ - z <= preload) {
              ++tileCount;
              tile = tileSource.getTile(z, x, y, pixelRatio, projection);
              if (tile.getState() == TileState.IDLE) {
                wantedTiles[tile.getKey()] = true;
                if (!tileQueue.isKeyQueued(tile.getKey())) {
                  tileQueue.enqueue([
                    tile,
                    tileSourceKey,
                    tileGrid.getTileCoordCenter(tile.tileCoord),
                    tileResolution,
                  ]);
                }
              }
              if (tileCallback !== undefined) {
                tileCallback(tile);
              }
            } else {
              tileSource.useTile(z, x, y, projection);
            }
          }
        }
      }
      tileSource.updateCacheSize(tileCount, projection);
    }
  }

  /**
   * @module ol/layer/Tile
   */

  /**
   * @classdesc
   * For layer sources that provide pre-rendered, tiled images in grids that are
   * organized by zoom levels for specific resolutions.
   * Note that any property set in the options is set as a {@link module:ol/Object~BaseObject}
   * property on the layer object; for example, setting `title: 'My Title'` in the
   * options means that `title` is observable, and has get/set accessors.
   *
   * @template {import("../source/Tile.js").default} TileSourceType
   * @extends BaseTileLayer<TileSourceType, CanvasTileLayerRenderer>
   * @api
   */
  class TileLayer extends BaseTileLayer {
    /**
     * @param {import("./BaseTile.js").Options<TileSourceType>} [options] Tile layer options.
     */
    constructor(options) {
      super(options);
    }

    createRenderer() {
      return new CanvasTileLayerRenderer(this);
    }
  }

  /**
   * @module ol/source/Raster
   */

  let hasImageData = true;
  try {
    new ImageData(10, 10);
  } catch (_) {
    hasImageData = false;
  }

  /** @type {CanvasRenderingContext2D} */
  let context;

  /**
   * @param {Uint8ClampedArray} data Image data.
   * @param {number} width Number of columns.
   * @param {number} height Number of rows.
   * @return {ImageData} Image data.
   */
  function newImageData(data, width, height) {
    if (hasImageData) {
      return new ImageData(data, width, height);
    }

    if (!context) {
      context = document.createElement('canvas').getContext('2d');
    }
    const imageData = context.createImageData(width, height);
    imageData.data.set(data);
    return imageData;
  }

  /**
   * @typedef {Object} MinionData
   * @property {Array<ArrayBuffer>} buffers Array of buffers.
   * @property {Object} meta Operation metadata.
   * @property {boolean} imageOps The operation is an image operation.
   * @property {number} width The width of the image.
   * @property {number} height The height of the image.
   */

  /* istanbul ignore next */
  /**
   * Create a function for running operations.  This function is serialized for
   * use in a worker.
   * @param {function(Array, Object):*} operation The operation.
   * @return {function(MinionData):ArrayBuffer} A function that takes an object with
   * buffers, meta, imageOps, width, and height properties and returns an array
   * buffer.
   */
  function createMinion(operation) {
    let workerHasImageData = true;
    try {
      new ImageData(10, 10);
    } catch (_) {
      workerHasImageData = false;
    }

    function newWorkerImageData(data, width, height) {
      if (workerHasImageData) {
        return new ImageData(data, width, height);
      } else {
        return {data: data, width: width, height: height};
      }
    }

    return function (data) {
      // bracket notation for minification support
      const buffers = data['buffers'];
      const meta = data['meta'];
      const imageOps = data['imageOps'];
      const width = data['width'];
      const height = data['height'];

      const numBuffers = buffers.length;
      const numBytes = buffers[0].byteLength;

      if (imageOps) {
        const images = new Array(numBuffers);
        for (let b = 0; b < numBuffers; ++b) {
          images[b] = newWorkerImageData(
            new Uint8ClampedArray(buffers[b]),
            width,
            height
          );
        }
        const output = operation(images, meta).data;
        return output.buffer;
      }

      const output = new Uint8ClampedArray(numBytes);
      const arrays = new Array(numBuffers);
      const pixels = new Array(numBuffers);
      for (let b = 0; b < numBuffers; ++b) {
        arrays[b] = new Uint8ClampedArray(buffers[b]);
        pixels[b] = [0, 0, 0, 0];
      }
      for (let i = 0; i < numBytes; i += 4) {
        for (let j = 0; j < numBuffers; ++j) {
          const array = arrays[j];
          pixels[j][0] = array[i];
          pixels[j][1] = array[i + 1];
          pixels[j][2] = array[i + 2];
          pixels[j][3] = array[i + 3];
        }
        const pixel = operation(pixels, meta);
        output[i] = pixel[0];
        output[i + 1] = pixel[1];
        output[i + 2] = pixel[2];
        output[i + 3] = pixel[3];
      }
      return output.buffer;
    };
  }

  /**
   * Create a worker for running operations.
   * @param {ProcessorOptions} config Processor options.
   * @param {function(MessageEvent): void} onMessage Called with a message event.
   * @return {Worker} The worker.
   */
  function createWorker(config, onMessage) {
    const lib = Object.keys(config.lib || {}).map(function (name) {
      return 'const ' + name + ' = ' + config.lib[name].toString() + ';';
    });

    const lines = lib.concat([
      'const __minion__ = (' + createMinion.toString() + ')(',
      config.operation.toString(),
      ');',
      'self.addEventListener("message", function(event) {',
      '  const buffer = __minion__(event.data);',
      '  self.postMessage({buffer: buffer, meta: event.data.meta}, [buffer]);',
      '});',
    ]);

    const worker = new Worker(
      typeof Blob === 'undefined'
        ? 'data:text/javascript;base64,' +
          Buffer.from(lines.join('\n'), 'binary').toString('base64')
        : URL.createObjectURL(new Blob(lines, {type: 'text/javascript'}))
    );
    worker.addEventListener('message', onMessage);
    return worker;
  }

  /**
   * @typedef {Object} FauxMessageEvent
   * @property {Object} data Message data.
   */

  /**
   * Create a faux worker for running operations.
   * @param {ProcessorOptions} config Configuration.
   * @param {function(FauxMessageEvent): void} onMessage Called with a message event.
   * @return {Object} The faux worker.
   */
  function createFauxWorker(config, onMessage) {
    const minion = createMinion(config.operation);
    let terminated = false;
    return {
      postMessage: function (data) {
        setTimeout(function () {
          if (terminated) {
            return;
          }
          onMessage({data: {buffer: minion(data), meta: data['meta']}});
        }, 0);
      },
      terminate: function () {
        terminated = true;
      },
    };
  }

  /**
   * @typedef {function(Error, ImageData, (Object|Array<Object>)): void} JobCallback
   */

  /**
   * @typedef {Object} Job
   * @property {Object} meta Job metadata.
   * @property {Array<ImageData>} inputs Array of input data.
   * @property {JobCallback} callback Called when the job is complete.
   */

  /**
   * @typedef {Object} ProcessorOptions
   * @property {number} threads Number of workers to spawn.
   * @property {Operation} operation The operation.
   * @property {Object<string, Function>} [lib] Functions that will be made available to operations run in a worker.
   * @property {number} queue The number of queued jobs to allow.
   * @property {boolean} [imageOps=false] Pass all the image data to the operation instead of a single pixel.
   */

  /**
   * @classdesc
   * A processor runs pixel or image operations in workers.
   */
  class Processor extends Disposable {
    /**
     * @param {ProcessorOptions} config Configuration.
     */
    constructor(config) {
      super();

      this._imageOps = !!config.imageOps;
      let threads;
      if (config.threads === 0) {
        threads = 0;
      } else if (this._imageOps) {
        threads = 1;
      } else {
        threads = config.threads || 1;
      }

      /**
       * @type {Array<Worker>}
       */
      const workers = new Array(threads);
      if (threads) {
        for (let i = 0; i < threads; ++i) {
          workers[i] = createWorker(config, this._onWorkerMessage.bind(this, i));
        }
      } else {
        workers[0] = createFauxWorker(
          config,
          this._onWorkerMessage.bind(this, 0)
        );
      }
      this._workers = workers;

      /**
       * @type {Array<Job>}
       * @private
       */
      this._queue = [];

      this._maxQueueLength = config.queue || Infinity;
      this._running = 0;

      /**
       * @type {Object<number, any>}
       * @private
       */
      this._dataLookup = {};

      /**
       * @type {Job}
       * @private
       */
      this._job = null;
    }

    /**
     * Run operation on input data.
     * @param {Array<ImageData>} inputs Array of image data.
     * @param {Object} meta A user data object.  This is passed to all operations
     *     and must be serializable.
     * @param {function(Error, ImageData, Object): void} callback Called when work
     *     completes.  The first argument is any error.  The second is the ImageData
     *     generated by operations.  The third is the user data object.
     */
    process(inputs, meta, callback) {
      this._enqueue({
        inputs: inputs,
        meta: meta,
        callback: callback,
      });
      this._dispatch();
    }

    /**
     * Add a job to the queue.
     * @param {Job} job The job.
     */
    _enqueue(job) {
      this._queue.push(job);
      while (this._queue.length > this._maxQueueLength) {
        this._queue.shift().callback(null, null);
      }
    }

    /**
     * Dispatch a job.
     */
    _dispatch() {
      if (this._running || this._queue.length === 0) {
        return;
      }

      const job = this._queue.shift();
      this._job = job;
      const width = job.inputs[0].width;
      const height = job.inputs[0].height;
      const buffers = job.inputs.map(function (input) {
        return input.data.buffer;
      });
      const threads = this._workers.length;
      this._running = threads;
      if (threads === 1) {
        this._workers[0].postMessage(
          {
            buffers: buffers,
            meta: job.meta,
            imageOps: this._imageOps,
            width: width,
            height: height,
          },
          buffers
        );
        return;
      }

      const length = job.inputs[0].data.length;
      const segmentLength = 4 * Math.ceil(length / 4 / threads);
      for (let i = 0; i < threads; ++i) {
        const offset = i * segmentLength;
        const slices = [];
        for (let j = 0, jj = buffers.length; j < jj; ++j) {
          slices.push(buffers[j].slice(offset, offset + segmentLength));
        }
        this._workers[i].postMessage(
          {
            buffers: slices,
            meta: job.meta,
            imageOps: this._imageOps,
            width: width,
            height: height,
          },
          slices
        );
      }
    }

    /**
     * Handle messages from the worker.
     * @param {number} index The worker index.
     * @param {MessageEvent} event The message event.
     */
    _onWorkerMessage(index, event) {
      if (this.disposed) {
        return;
      }
      this._dataLookup[index] = event.data;
      --this._running;
      if (this._running === 0) {
        this._resolveJob();
      }
    }

    /**
     * Resolve a job.  If there are no more worker threads, the processor callback
     * will be called.
     */
    _resolveJob() {
      const job = this._job;
      const threads = this._workers.length;
      let data, meta;
      if (threads === 1) {
        data = new Uint8ClampedArray(this._dataLookup[0]['buffer']);
        meta = this._dataLookup[0]['meta'];
      } else {
        const length = job.inputs[0].data.length;
        data = new Uint8ClampedArray(length);
        meta = new Array(threads);
        const segmentLength = 4 * Math.ceil(length / 4 / threads);
        for (let i = 0; i < threads; ++i) {
          const buffer = this._dataLookup[i]['buffer'];
          const offset = i * segmentLength;
          data.set(new Uint8ClampedArray(buffer), offset);
          meta[i] = this._dataLookup[i]['meta'];
        }
      }
      this._job = null;
      this._dataLookup = {};
      job.callback(
        null,
        newImageData(data, job.inputs[0].width, job.inputs[0].height),
        meta
      );
      this._dispatch();
    }

    /**
     * Terminate all workers associated with the processor.
     */
    disposeInternal() {
      for (let i = 0; i < this._workers.length; ++i) {
        this._workers[i].terminate();
      }
      this._workers.length = 0;
    }
  }

  /**
   * A function that takes an array of input data, performs some operation, and
   * returns an array of output data.
   * For `pixel` type operations, the function will be called with an array of
   * pixels, where each pixel is an array of four numbers (`[r, g, b, a]`) in the
   * range of 0 - 255. It should return a single pixel array.
   * For `'image'` type operations, functions will be called with an array of
   * [ImageData](https://developer.mozilla.org/en-US/docs/Web/API/ImageData)
   * and should return a single
   * [ImageData](https://developer.mozilla.org/en-US/docs/Web/API/ImageData).
   * The operations
   * are called with a second "data" argument, which can be used for storage.  The
   * data object is accessible from raster events, where it can be initialized in
   * "beforeoperations" and accessed again in "afteroperations".
   *
   * @typedef {function((Array<Array<number>>|Array<ImageData>), Object):
   *     (Array<number>|ImageData)} Operation
   */

  /**
   * @enum {string}
   */
  const RasterEventType = {
    /**
     * Triggered before operations are run.  Listeners will receive an event object with
     * a `data` property that can be used to make data available to operations.
     * @event module:ol/source/Raster.RasterSourceEvent#beforeoperations
     * @api
     */
    BEFOREOPERATIONS: 'beforeoperations',

    /**
     * Triggered after operations are run.  Listeners will receive an event object with
     * a `data` property.  If more than one thread is used, `data` will be an array of
     * objects.  If a single thread is used, `data` will be a single object.
     * @event module:ol/source/Raster.RasterSourceEvent#afteroperations
     * @api
     */
    AFTEROPERATIONS: 'afteroperations',
  };

  /**
   * @typedef {'pixel' | 'image'} RasterOperationType
   * Raster operation type. Supported values are `'pixel'` and `'image'`.
   */

  /**
   * @typedef {import("./Image.js").ImageSourceEventTypes|'beforeoperations'|'afteroperations'} RasterSourceEventTypes
   */

  /**
   * @classdesc
   * Events emitted by {@link module:ol/source/Raster~RasterSource} instances are instances of this
   * type.
   */
  class RasterSourceEvent extends BaseEvent {
    /**
     * @param {string} type Type.
     * @param {import("../Map.js").FrameState} frameState The frame state.
     * @param {Object|Array<Object>} data An object made available to operations.  For "afteroperations" evenets
     * this will be an array of objects if more than one thread is used.
     */
    constructor(type, frameState, data) {
      super(type);

      /**
       * The raster extent.
       * @type {import("../extent.js").Extent}
       * @api
       */
      this.extent = frameState.extent;

      /**
       * The pixel resolution (map units per pixel).
       * @type {number}
       * @api
       */
      this.resolution = frameState.viewState.resolution / frameState.pixelRatio;

      /**
       * An object made available to all operations.  This can be used by operations
       * as a storage object (e.g. for calculating statistics).
       * @type {Object}
       * @api
       */
      this.data = data;
    }
  }

  /**
   * @typedef {Object} Options
   * @property {Array<import("./Source.js").default|import("../layer/Layer.js").default>} sources Input
   * sources or layers.  For vector data, use an VectorImage layer.
   * @property {Operation} [operation] Raster operation.
   * The operation will be called with data from input sources
   * and the output will be assigned to the raster source.
   * @property {Object} [lib] Functions that will be made available to operations run in a worker.
   * @property {number} [threads] By default, operations will be run in a single worker thread.
   * To avoid using workers altogether, set `threads: 0`.  For pixel operations, operations can
   * be run in multiple worker threads.  Note that there is additional overhead in
   * transferring data to multiple workers, and that depending on the user's
   * system, it may not be possible to parallelize the work.
   * @property {RasterOperationType} [operationType='pixel'] Operation type.
   * Supported values are `'pixel'` and `'image'`.  By default,
   * `'pixel'` operations are assumed, and operations will be called with an
   * array of pixels from input sources.  If set to `'image'`, operations will
   * be called with an array of ImageData objects from input sources.
   */

  /***
   * @template Return
   * @typedef {import("../Observable").OnSignature<import("../Observable").EventTypes, import("../events/Event.js").default, Return> &
   *   import("../Observable").OnSignature<import("../ObjectEventType").Types, import("../Object").ObjectEvent, Return> &
   *   import("../Observable").OnSignature<import("./Image.js").ImageSourceEventTypes, import("./Image.js").ImageSourceEvent, Return> &
   *   import("../Observable").OnSignature<RasterSourceEventTypes, RasterSourceEvent, Return> &
   *   import("../Observable").CombinedOnSignature<import("../Observable").EventTypes|import("../ObjectEventType").Types
   *     |RasterSourceEventTypes, Return>} RasterSourceOnSignature
   */

  /**
   * @classdesc
   * A source that transforms data from any number of input sources using an
   * {@link module:ol/source/Raster~Operation} function to transform input pixel values into
   * output pixel values.
   *
   * @fires module:ol/source/Raster.RasterSourceEvent
   * @api
   */
  class RasterSource extends ImageSource {
    /**
     * @param {Options} options Options.
     */
    constructor(options) {
      super({
        projection: null,
      });

      /***
       * @type {RasterSourceOnSignature<import("../events").EventsKey>}
       */
      this.on;

      /***
       * @type {RasterSourceOnSignature<import("../events").EventsKey>}
       */
      this.once;

      /***
       * @type {RasterSourceOnSignature<void>}
       */
      this.un;

      /**
       * @private
       * @type {Processor}
       */
      this.processor_ = null;

      /**
       * @private
       * @type {RasterOperationType}
       */
      this.operationType_ =
        options.operationType !== undefined ? options.operationType : 'pixel';

      /**
       * @private
       * @type {number}
       */
      this.threads_ = options.threads !== undefined ? options.threads : 1;

      /**
       * @private
       * @type {Array<import("../layer/Layer.js").default>}
       */
      this.layers_ = createLayers(options.sources);

      const changed = this.changed.bind(this);
      for (let i = 0, ii = this.layers_.length; i < ii; ++i) {
        this.layers_[i].addEventListener(EventType$1.CHANGE, changed);
      }

      /**
       * @private
       * @type {import("../TileQueue.js").default}
       */
      this.tileQueue_ = new TileQueue(function () {
        return 1;
      }, this.changed.bind(this));

      /**
       * The most recently requested frame state.
       * @type {import("../Map.js").FrameState}
       * @private
       */
      this.requestedFrameState_;

      /**
       * The most recently rendered image canvas.
       * @type {import("../ImageCanvas.js").default}
       * @private
       */
      this.renderedImageCanvas_ = null;

      /**
       * The most recently rendered revision.
       * @type {number}
       */
      this.renderedRevision_;

      /**
       * @private
       * @type {import("../Map.js").FrameState}
       */
      this.frameState_ = {
        animate: false,
        coordinateToPixelTransform: create$2(),
        declutterTree: null,
        extent: null,
        index: 0,
        layerIndex: 0,
        layerStatesArray: getLayerStatesArray(this.layers_),
        pixelRatio: 1,
        pixelToCoordinateTransform: create$2(),
        postRenderFunctions: [],
        size: [0, 0],
        tileQueue: this.tileQueue_,
        time: Date.now(),
        usedTiles: {},
        viewState: /** @type {import("../View.js").State} */ ({
          rotation: 0,
        }),
        viewHints: [],
        wantedTiles: {},
        mapId: getUid(this),
        renderTargets: {},
      };

      this.setAttributions(function (frameState) {
        const attributions = [];
        for (
          let index = 0, iMax = options.sources.length;
          index < iMax;
          ++index
        ) {
          const sourceOrLayer = options.sources[index];
          const source =
            sourceOrLayer instanceof Source
              ? sourceOrLayer
              : sourceOrLayer.getSource();
          const attributionGetter = source.getAttributions();
          if (typeof attributionGetter === 'function') {
            const sourceAttribution = attributionGetter(frameState);
            attributions.push.apply(attributions, sourceAttribution);
          }
        }
        return attributions.length !== 0 ? attributions : null;
      });

      if (options.operation !== undefined) {
        this.setOperation(options.operation, options.lib);
      }
    }

    /**
     * Set the operation.
     * @param {Operation} operation New operation.
     * @param {Object} [lib] Functions that will be available to operations run
     *     in a worker.
     * @api
     */
    setOperation(operation, lib) {
      if (this.processor_) {
        this.processor_.dispose();
      }

      this.processor_ = new Processor({
        operation: operation,
        imageOps: this.operationType_ === 'image',
        queue: 1,
        lib: lib,
        threads: this.threads_,
      });
      this.changed();
    }

    /**
     * Update the stored frame state.
     * @param {import("../extent.js").Extent} extent The view extent (in map units).
     * @param {number} resolution The view resolution.
     * @param {import("../proj/Projection.js").default} projection The view projection.
     * @return {import("../Map.js").FrameState} The updated frame state.
     * @private
     */
    updateFrameState_(extent, resolution, projection) {
      const frameState = /** @type {import("../Map.js").FrameState} */ (
        Object.assign({}, this.frameState_)
      );

      frameState.viewState = /** @type {import("../View.js").State} */ (
        Object.assign({}, frameState.viewState)
      );

      const center = getCenter(extent);

      frameState.extent = extent.slice();
      frameState.size[0] = Math.round(getWidth(extent) / resolution);
      frameState.size[1] = Math.round(getHeight(extent) / resolution);
      frameState.time = Date.now();

      const viewState = frameState.viewState;
      viewState.center = center;
      viewState.projection = projection;
      viewState.resolution = resolution;
      return frameState;
    }

    /**
     * Determine if all sources are ready.
     * @return {boolean} All sources are ready.
     * @private
     */
    allSourcesReady_() {
      let ready = true;
      let source;
      for (let i = 0, ii = this.layers_.length; i < ii; ++i) {
        source = this.layers_[i].getSource();
        if (source.getState() !== 'ready') {
          ready = false;
          break;
        }
      }
      return ready;
    }

    /**
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {number} resolution Resolution.
     * @param {number} pixelRatio Pixel ratio.
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @return {import("../ImageCanvas.js").default} Single image.
     */
    getImage(extent, resolution, pixelRatio, projection) {
      if (!this.allSourcesReady_()) {
        return null;
      }

      const frameState = this.updateFrameState_(extent, resolution, projection);
      this.requestedFrameState_ = frameState;

      // check if we can't reuse the existing ol/ImageCanvas
      if (this.renderedImageCanvas_) {
        const renderedResolution = this.renderedImageCanvas_.getResolution();
        const renderedExtent = this.renderedImageCanvas_.getExtent();
        if (
          resolution !== renderedResolution ||
          !equals$2(extent, renderedExtent)
        ) {
          this.renderedImageCanvas_ = null;
        }
      }

      if (
        !this.renderedImageCanvas_ ||
        this.getRevision() !== this.renderedRevision_
      ) {
        this.processSources_();
      }

      frameState.tileQueue.loadMoreTiles(16, 16);

      if (frameState.animate) {
        requestAnimationFrame(this.changed.bind(this));
      }

      return this.renderedImageCanvas_;
    }

    /**
     * Start processing source data.
     * @private
     */
    processSources_() {
      const frameState = this.requestedFrameState_;
      const len = this.layers_.length;
      const imageDatas = new Array(len);
      for (let i = 0; i < len; ++i) {
        frameState.layerIndex = i;
        const imageData = getImageData(this.layers_[i], frameState);
        if (imageData) {
          imageDatas[i] = imageData;
        } else {
          return;
        }
      }

      const data = {};
      this.dispatchEvent(
        new RasterSourceEvent(RasterEventType.BEFOREOPERATIONS, frameState, data)
      );
      this.processor_.process(
        imageDatas,
        data,
        this.onWorkerComplete_.bind(this, frameState)
      );
    }

    /**
     * Called when pixel processing is complete.
     * @param {import("../Map.js").FrameState} frameState The frame state.
     * @param {Error} err Any error during processing.
     * @param {ImageData} output The output image data.
     * @param {Object|Array<Object>} data The user data (or an array if more than one thread).
     * @private
     */
    onWorkerComplete_(frameState, err, output, data) {
      if (err || !output) {
        return;
      }

      // do nothing if extent or resolution changed
      const extent = frameState.extent;
      const resolution = frameState.viewState.resolution;
      if (
        resolution !== this.requestedFrameState_.viewState.resolution ||
        !equals$2(extent, this.requestedFrameState_.extent)
      ) {
        return;
      }

      let context;
      if (this.renderedImageCanvas_) {
        context = this.renderedImageCanvas_.getImage().getContext('2d');
      } else {
        const width = Math.round(getWidth(extent) / resolution);
        const height = Math.round(getHeight(extent) / resolution);
        context = createCanvasContext2D(width, height);
        this.renderedImageCanvas_ = new ImageCanvas(
          extent,
          resolution,
          1,
          context.canvas
        );
      }
      context.putImageData(output, 0, 0);

      this.changed();
      this.renderedRevision_ = this.getRevision();

      this.dispatchEvent(
        new RasterSourceEvent(RasterEventType.AFTEROPERATIONS, frameState, data)
      );
      if (frameState.animate) {
        requestAnimationFrame(this.changed.bind(this));
      }
    }

    disposeInternal() {
      if (this.processor_) {
        this.processor_.dispose();
      }
      super.disposeInternal();
    }
  }

  /**
   * Clean up and unregister the worker.
   * @function
   * @api
   */
  RasterSource.prototype.dispose;

  /**
   * A reusable canvas context.
   * @type {CanvasRenderingContext2D}
   * @private
   */
  let sharedContext = null;

  /**
   * Get image data from a layer.
   * @param {import("../layer/Layer.js").default} layer Layer to render.
   * @param {import("../Map.js").FrameState} frameState The frame state.
   * @return {ImageData} The image data.
   */
  function getImageData(layer, frameState) {
    const renderer = layer.getRenderer();
    if (!renderer) {
      throw new Error('Unsupported layer type: ' + layer);
    }

    if (!renderer.prepareFrame(frameState)) {
      return null;
    }
    const width = frameState.size[0];
    const height = frameState.size[1];
    if (width === 0 || height === 0) {
      return null;
    }
    const container = renderer.renderFrame(frameState, null);
    let element;
    if (container instanceof HTMLCanvasElement) {
      element = container;
    } else {
      if (container) {
        element = container.firstElementChild;
      }
      if (!(element instanceof HTMLCanvasElement)) {
        throw new Error('Unsupported rendered element: ' + element);
      }
      if (element.width === width && element.height === height) {
        const context = element.getContext('2d');
        return context.getImageData(0, 0, width, height);
      }
    }

    if (!sharedContext) {
      sharedContext = createCanvasContext2D(width, height);
    } else {
      const canvas = sharedContext.canvas;
      if (canvas.width !== width || canvas.height !== height) {
        sharedContext = createCanvasContext2D(width, height);
      } else {
        sharedContext.clearRect(0, 0, width, height);
      }
    }
    sharedContext.drawImage(element, 0, 0, width, height);
    return sharedContext.getImageData(0, 0, width, height);
  }

  /**
   * Get a list of layer states from a list of layers.
   * @param {Array<import("../layer/Layer.js").default>} layers Layers.
   * @return {Array<import("../layer/Layer.js").State>} The layer states.
   */
  function getLayerStatesArray(layers) {
    return layers.map(function (layer) {
      return layer.getLayerState();
    });
  }

  /**
   * Create layers for all sources.
   * @param {Array<import("./Source.js").default|import("../layer/Layer.js").default>} sources The sources.
   * @return {Array<import("../layer/Layer.js").default>} Array of layers.
   */
  function createLayers(sources) {
    const len = sources.length;
    const layers = new Array(len);
    for (let i = 0; i < len; ++i) {
      layers[i] = createLayer(sources[i]);
    }
    return layers;
  }

  /**
   * Create a layer for the provided source.
   * @param {import("./Source.js").default|import("../layer/Layer.js").default} layerOrSource The layer or source.
   * @return {import("../layer/Layer.js").default} The layer.
   */
  function createLayer(layerOrSource) {
    // @type {import("../layer/Layer.js").default}
    let layer;
    if (layerOrSource instanceof Source) {
      if (layerOrSource instanceof TileSource) {
        layer = new TileLayer({source: layerOrSource});
      } else if (layerOrSource instanceof ImageSource) {
        layer = new ImageLayer({source: layerOrSource});
      }
    } else {
      layer = layerOrSource;
    }
    return layer;
  }

  /**
   * @module ol/source/Stamen
   */

  /**
   * @const
   * @type {Array<string>}
   */
  const ATTRIBUTIONS = [
    'Map tiles by <a href="https://stamen.com/" target="_blank">Stamen Design</a>, ' +
      'under <a href="https://creativecommons.org/licenses/by/3.0/" target="_blank">CC BY' +
      ' 3.0</a>.',
    ATTRIBUTION,
  ];

  /**
   * @type {Object<string, {extension: string, opaque: boolean}>}
   */
  const LayerConfig = {
    'terrain': {
      extension: 'jpg',
      opaque: true,
    },
    'terrain-background': {
      extension: 'jpg',
      opaque: true,
    },
    'terrain-labels': {
      extension: 'png',
      opaque: false,
    },
    'terrain-lines': {
      extension: 'png',
      opaque: false,
    },
    'toner-background': {
      extension: 'png',
      opaque: true,
    },
    'toner': {
      extension: 'png',
      opaque: true,
    },
    'toner-hybrid': {
      extension: 'png',
      opaque: false,
    },
    'toner-labels': {
      extension: 'png',
      opaque: false,
    },
    'toner-lines': {
      extension: 'png',
      opaque: false,
    },
    'toner-lite': {
      extension: 'png',
      opaque: true,
    },
    'watercolor': {
      extension: 'jpg',
      opaque: true,
    },
  };

  /**
   * @type {Object<string, {minZoom: number, maxZoom: number}>}
   */
  const ProviderConfig = {
    'terrain': {
      minZoom: 0,
      maxZoom: 18,
    },
    'toner': {
      minZoom: 0,
      maxZoom: 20,
    },
    'watercolor': {
      minZoom: 0,
      maxZoom: 18,
    },
  };

  /**
   * @typedef {Object} Options
   * @property {number} [cacheSize] Initial tile cache size. Will auto-grow to hold at least the number of tiles in the viewport.
   * @property {boolean} [interpolate=true] Use interpolated values when resampling.  By default,
   * linear interpolation is used when resampling.  Set to false to use the nearest neighbor instead.
   * @property {string} layer Layer name.
   * @property {number} [minZoom] Minimum zoom.
   * @property {number} [maxZoom] Maximum zoom.
   * @property {number} [reprojectionErrorThreshold=0.5] Maximum allowed reprojection error (in pixels).
   * Higher values can increase reprojection performance, but decrease precision.
   * @property {import("../Tile.js").LoadFunction} [tileLoadFunction]
   * Optional function to load a tile given a URL. The default is
   * ```js
   * function(imageTile, src) {
   *   imageTile.getImage().src = src;
   * };
   * ```
   * @property {number} [transition=250] Duration of the opacity transition for rendering.
   * To disable the opacity transition, pass `transition: 0`.
   * @property {string} [url] URL template. Must include `{x}`, `{y}` or `{-y}`, and `{z}` placeholders.
   * @property {boolean} [wrapX=true] Whether to wrap the world horizontally.
   * @property {number|import("../array.js").NearestDirectionFunction} [zDirection=0]
   * Choose whether to use tiles with a higher or lower zoom level when between integer
   * zoom levels. See {@link module:ol/tilegrid/TileGrid~TileGrid#getZForResolution}.
   */

  /**
   * @classdesc
   * Layer source for the Stamen tile server.
   * @api
   */
  class Stamen extends XYZ {
    /**
     * @param {Options} options Stamen options.
     */
    constructor(options) {
      const i = options.layer.indexOf('-');
      const provider = i == -1 ? options.layer : options.layer.slice(0, i);
      const providerConfig = ProviderConfig[provider];

      const layerConfig = LayerConfig[options.layer];

      const url =
        options.url !== undefined
          ? options.url
          : 'https://stamen-tiles-{a-d}.a.ssl.fastly.net/' +
            options.layer +
            '/{z}/{x}/{y}.' +
            layerConfig.extension;

      super({
        attributions: ATTRIBUTIONS,
        cacheSize: options.cacheSize,
        crossOrigin: 'anonymous',
        interpolate: options.interpolate,
        maxZoom:
          options.maxZoom != undefined ? options.maxZoom : providerConfig.maxZoom,
        minZoom:
          options.minZoom != undefined ? options.minZoom : providerConfig.minZoom,
        opaque: layerConfig.opaque,
        reprojectionErrorThreshold: options.reprojectionErrorThreshold,
        tileLoadFunction: options.tileLoadFunction,
        transition: options.transition,
        url: url,
        wrapX: options.wrapX,
        zDirection: options.zDirection,
      });
    }
  }

  /**
   * @module ol/source/TileArcGISRest
   */

  /**
   * @typedef {Object} Options
   * @property {import("./Source.js").AttributionLike} [attributions] Attributions.
   * @property {number} [cacheSize] Initial tile cache size. Will auto-grow to hold at least the number of tiles in the viewport.
   * @property {null|string} [crossOrigin] The `crossOrigin` attribute for loaded images.  Note that
   * you must provide a `crossOrigin` value if you want to access pixel data with the Canvas renderer.
   * See https://developer.mozilla.org/en-US/docs/Web/HTML/CORS_enabled_image for more detail.
   * @property {boolean} [interpolate=true] Use interpolated values when resampling.  By default,
   * linear interpolation is used when resampling.  Set to false to use the nearest neighbor instead.
   * @property {Object<string,*>} [params] ArcGIS Rest parameters. This field is optional. Service defaults will be
   * used for any fields not specified. `FORMAT` is `PNG32` by default. `F` is `IMAGE` by
   * default. `TRANSPARENT` is `true` by default.  `BBOX`, `SIZE`, `BBOXSR`,
   * and `IMAGESR` will be set dynamically. Set `LAYERS` to
   * override the default service layer visibility. See
   * https://developers.arcgis.com/rest/services-reference/export-map.htm
   * for further reference.
   * @property {boolean} [hidpi=true] Use the `ol/Map#pixelRatio` value when requesting
   * the image from the remote server.
   * @property {import("../tilegrid/TileGrid.js").default} [tileGrid] Tile grid. Base this on the resolutions,
   * tilesize and extent supported by the server.
   * If this is not defined, a default grid will be used: if there is a projection
   * extent, the grid will be based on that; if not, a grid based on a global
   * extent with origin at 0,0 will be used.
   * @property {import("../proj.js").ProjectionLike} [projection] Projection. Default is the view projection.
   * The projection code must contain a numeric end portion separated by :
   * or the entire code must form a valid ArcGIS SpatialReference definition.
   * @property {number} [reprojectionErrorThreshold=0.5] Maximum allowed reprojection error (in pixels).
   * Higher values can increase reprojection performance, but decrease precision.
   * @property {import("../Tile.js").LoadFunction} [tileLoadFunction] Optional function to load a tile given a URL.
   * The default is
   * ```js
   * function(imageTile, src) {
   *   imageTile.getImage().src = src;
   * };
   * ```
   * @property {string} [url] ArcGIS Rest service URL for a Map Service or Image Service. The
   * url should include /MapServer or /ImageServer.
   * @property {boolean} [wrapX=true] Whether to wrap the world horizontally.
   * @property {number} [transition] Duration of the opacity transition for rendering.  To disable the opacity
   * transition, pass `transition: 0`.
   * @property {Array<string>} [urls] ArcGIS Rest service urls. Use this instead of `url` when the ArcGIS
   * Service supports multiple urls for export requests.
   * @property {number|import("../array.js").NearestDirectionFunction} [zDirection=0]
   * Choose whether to use tiles with a higher or lower zoom level when between integer
   * zoom levels. See {@link module:ol/tilegrid/TileGrid~TileGrid#getZForResolution}.
   */

  /**
   * @classdesc
   * Layer source for tile data from ArcGIS Rest services. Map and Image
   * Services are supported.
   *
   * For cached ArcGIS services, better performance is available using the
   * {@link module:ol/source/XYZ~XYZ} data source.
   * @api
   */
  class TileArcGISRest extends TileImage {
    /**
     * @param {Options} [options] Tile ArcGIS Rest options.
     */
    constructor(options) {
      options = options ? options : {};

      super({
        attributions: options.attributions,
        cacheSize: options.cacheSize,
        crossOrigin: options.crossOrigin,
        interpolate: options.interpolate,
        projection: options.projection,
        reprojectionErrorThreshold: options.reprojectionErrorThreshold,
        tileGrid: options.tileGrid,
        tileLoadFunction: options.tileLoadFunction,
        url: options.url,
        urls: options.urls,
        wrapX: options.wrapX !== undefined ? options.wrapX : true,
        transition: options.transition,
        zDirection: options.zDirection,
      });

      /**
       * @private
       * @type {!Object}
       */
      this.params_ = options.params || {};

      /**
       * @private
       * @type {boolean}
       */
      this.hidpi_ = options.hidpi !== undefined ? options.hidpi : true;

      /**
       * @private
       * @type {import("../extent.js").Extent}
       */
      this.tmpExtent_ = createEmpty();

      this.setKey(this.getKeyForParams_());
    }

    /**
     * @private
     * @return {string} The key for the current params.
     */
    getKeyForParams_() {
      let i = 0;
      const res = [];
      for (const key in this.params_) {
        res[i++] = key + '-' + this.params_[key];
      }
      return res.join('/');
    }

    /**
     * Get the user-provided params, i.e. those passed to the constructor through
     * the "params" option, and possibly updated using the updateParams method.
     * @return {Object} Params.
     * @api
     */
    getParams() {
      return this.params_;
    }

    /**
     * @param {import("../tilecoord.js").TileCoord} tileCoord Tile coordinate.
     * @param {import("../size.js").Size} tileSize Tile size.
     * @param {import("../extent.js").Extent} tileExtent Tile extent.
     * @param {number} pixelRatio Pixel ratio.
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @param {Object} params Params.
     * @return {string|undefined} Request URL.
     * @private
     */
    getRequestUrl_(
      tileCoord,
      tileSize,
      tileExtent,
      pixelRatio,
      projection,
      params
    ) {
      const urls = this.urls;
      if (!urls) {
        return undefined;
      }

      // ArcGIS Server only wants the numeric portion of the projection ID.
      // (if there is no numeric portion the entire projection code must
      // form a valid ArcGIS SpatialReference definition).
      const srid = projection
        .getCode()
        .split(/:(?=\d+$)/)
        .pop();

      params['SIZE'] = tileSize[0] + ',' + tileSize[1];
      params['BBOX'] = tileExtent.join(',');
      params['BBOXSR'] = srid;
      params['IMAGESR'] = srid;
      params['DPI'] = Math.round(
        params['DPI'] ? params['DPI'] * pixelRatio : 90 * pixelRatio
      );

      let url;
      if (urls.length == 1) {
        url = urls[0];
      } else {
        const index = modulo(hash(tileCoord), urls.length);
        url = urls[index];
      }

      const modifiedUrl = url
        .replace(/MapServer\/?$/, 'MapServer/export')
        .replace(/ImageServer\/?$/, 'ImageServer/exportImage');
      return appendParams(modifiedUrl, params);
    }

    /**
     * Get the tile pixel ratio for this source.
     * @param {number} pixelRatio Pixel ratio.
     * @return {number} Tile pixel ratio.
     */
    getTilePixelRatio(pixelRatio) {
      return this.hidpi_ ? pixelRatio : 1;
    }

    /**
     * Update the user-provided params.
     * @param {Object} params Params.
     * @api
     */
    updateParams(params) {
      Object.assign(this.params_, params);
      this.setKey(this.getKeyForParams_());
    }

    /**
     * @param {import("../tilecoord.js").TileCoord} tileCoord The tile coordinate
     * @param {number} pixelRatio The pixel ratio
     * @param {import("../proj/Projection.js").default} projection The projection
     * @return {string|undefined} The tile URL
     * @override
     */
    tileUrlFunction(tileCoord, pixelRatio, projection) {
      let tileGrid = this.getTileGrid();
      if (!tileGrid) {
        tileGrid = this.getTileGridForProjection(projection);
      }

      if (tileGrid.getResolutions().length <= tileCoord[0]) {
        return undefined;
      }

      if (pixelRatio != 1 && !this.hidpi_) {
        pixelRatio = 1;
      }

      const tileExtent = tileGrid.getTileCoordExtent(tileCoord, this.tmpExtent_);
      let tileSize = toSize(tileGrid.getTileSize(tileCoord[0]), this.tmpSize);

      if (pixelRatio != 1) {
        tileSize = scale(tileSize, pixelRatio, this.tmpSize);
      }

      // Apply default params and override with user specified values.
      const baseParams = {
        'F': 'image',
        'FORMAT': 'PNG32',
        'TRANSPARENT': true,
      };
      Object.assign(baseParams, this.params_);

      return this.getRequestUrl_(
        tileCoord,
        tileSize,
        tileExtent,
        pixelRatio,
        projection,
        baseParams
      );
    }
  }

  /**
   * @module ol/source/TileDebug
   */

  /**
   * @typedef {Object} Options
   * @property {import("../proj.js").ProjectionLike} [projection='EPSG:3857'] Optional projection.
   * @property {import("../tilegrid/TileGrid.js").default} [tileGrid] Tile grid.
   * @property {boolean} [wrapX=true] Whether to wrap the world horizontally.
   * @property {number|import("../array.js").NearestDirectionFunction} [zDirection=0]
   * Set to `1` when debugging `VectorTile` sources with a default configuration.
   * Choose whether to use tiles with a higher or lower zoom level when between integer
   * zoom levels. See {@link module:ol/tilegrid/TileGrid~TileGrid#getZForResolution}.
   * @property {string} [template='z:{z} x:{x} y:{y}'] Template for labeling the tiles.
   * Should include `{x}`, `{y}` or `{-y}`, and `{z}` placeholders.
   */

  /**
   * @classdesc
   * A pseudo tile source, which does not fetch tiles from a server, but renders
   * a grid outline for the tile grid/projection along with the coordinates for
   * each tile. See examples/canvas-tiles for an example.
   * @api
   */
  class TileDebug extends XYZ {
    /**
     * @param {Options} [options] Debug tile options.
     */
    constructor(options) {
      /**
       * @type {Options}
       */
      options = options || {};

      super({
        opaque: false,
        projection: options.projection,
        tileGrid: options.tileGrid,
        wrapX: options.wrapX !== undefined ? options.wrapX : true,
        zDirection: options.zDirection,
        url: options.template || 'z:{z} x:{x} y:{y}',
        tileLoadFunction: (tile, text) => {
          const z = tile.getTileCoord()[0];
          const tileSize = toSize(this.tileGrid.getTileSize(z));
          const context = createCanvasContext2D(tileSize[0], tileSize[1]);

          context.strokeStyle = 'grey';
          context.strokeRect(0.5, 0.5, tileSize[0] + 0.5, tileSize[1] + 0.5);

          context.fillStyle = 'grey';
          context.strokeStyle = 'white';
          context.textAlign = 'center';
          context.textBaseline = 'middle';
          context.font = '24px sans-serif';
          context.lineWidth = 4;
          context.strokeText(text, tileSize[0] / 2, tileSize[1] / 2, tileSize[0]);
          context.fillText(text, tileSize[0] / 2, tileSize[1] / 2, tileSize[0]);

          /** @type {import("../ImageTile.js").default} */ (tile).setImage(
            context.canvas
          );
        },
      });
    }
  }

  /**
   * @module ol/source/TileJSON
   */

  /**
   * @typedef {Object} Config
   * @property {string} [name] The name.
   * @property {string} [description] The description.
   * @property {string} [version] The version.
   * @property {string} [attribution] The attribution.
   * @property {string} [template] The template.
   * @property {string} [legend] The legend.
   * @property {string} [scheme] The scheme.
   * @property {Array<string>} tiles The tile URL templates.
   * @property {Array<string>} [grids] Optional grids.
   * @property {number} [minzoom] Minimum zoom level.
   * @property {number} [maxzoom] Maximum zoom level.
   * @property {Array<number>} [bounds] Optional bounds.
   * @property {Array<number>} [center] Optional center.
   */

  /**
   * @typedef {Object} Options
   * @property {import("./Source.js").AttributionLike} [attributions] Attributions.
   * @property {number} [cacheSize] Initial tile cache size. Will auto-grow to hold at least the number of tiles in the viewport.
   * @property {null|string} [crossOrigin] The `crossOrigin` attribute for loaded images.  Note that
   * you must provide a `crossOrigin` value if you want to access pixel data with the Canvas renderer.
   * See https://developer.mozilla.org/en-US/docs/Web/HTML/CORS_enabled_image for more detail.
   * @property {boolean} [interpolate=true] Use interpolated values when resampling.  By default,
   * linear interpolation is used when resampling.  Set to false to use the nearest neighbor instead.
   * @property {boolean} [jsonp=false] Use JSONP with callback to load the TileJSON.
   * Useful when the server does not support CORS..
   * @property {number} [reprojectionErrorThreshold=0.5] Maximum allowed reprojection error (in pixels).
   * Higher values can increase reprojection performance, but decrease precision.
   * @property {Config} [tileJSON] TileJSON configuration for this source.
   * If not provided, `url` must be configured.
   * @property {import("../Tile.js").LoadFunction} [tileLoadFunction] Optional function to load a tile given a URL. The default is
   * ```js
   * function(imageTile, src) {
   *   imageTile.getImage().src = src;
   * };
   * ```
   * @property {number|import("../size.js").Size} [tileSize=[256, 256]] The tile size used by the tile service.
   * Note: `tileSize` and other non-standard TileJSON properties are currently ignored.
   * @property {string} [url] URL to the TileJSON file. If not provided, `tileJSON` must be configured.
   * @property {boolean} [wrapX=true] Whether to wrap the world horizontally.
   * @property {number} [transition] Duration of the opacity transition for rendering.
   * To disable the opacity transition, pass `transition: 0`.
   * @property {number|import("../array.js").NearestDirectionFunction} [zDirection=0]
   * Choose whether to use tiles with a higher or lower zoom level when between integer
   * zoom levels. See {@link module:ol/tilegrid/TileGrid~TileGrid#getZForResolution}.
   */

  /**
   * @classdesc
   * Layer source for tile data in TileJSON format.
   * @api
   */
  class TileJSON extends TileImage {
    /**
     * @param {Options} options TileJSON options.
     */
    constructor(options) {
      super({
        attributions: options.attributions,
        cacheSize: options.cacheSize,
        crossOrigin: options.crossOrigin,
        interpolate: options.interpolate,
        projection: get$2('EPSG:3857'),
        reprojectionErrorThreshold: options.reprojectionErrorThreshold,
        state: 'loading',
        tileLoadFunction: options.tileLoadFunction,
        wrapX: options.wrapX !== undefined ? options.wrapX : true,
        transition: options.transition,
        zDirection: options.zDirection,
      });

      /**
       * @type {Config}
       * @private
       */
      this.tileJSON_ = null;

      /**
       * @type {number|import("../size.js").Size}
       * @private
       */
      this.tileSize_ = options.tileSize;

      if (options.url) {
        if (options.jsonp) {
          jsonp(
            options.url,
            this.handleTileJSONResponse.bind(this),
            this.handleTileJSONError.bind(this)
          );
        } else {
          const client = new XMLHttpRequest();
          client.addEventListener('load', this.onXHRLoad_.bind(this));
          client.addEventListener('error', this.onXHRError_.bind(this));
          client.open('GET', options.url);
          client.send();
        }
      } else if (options.tileJSON) {
        this.handleTileJSONResponse(options.tileJSON);
      } else {
        assert(false, 51); // Either `url` or `tileJSON` options must be provided
      }
    }

    /**
     * @private
     * @param {Event} event The load event.
     */
    onXHRLoad_(event) {
      const client = /** @type {XMLHttpRequest} */ (event.target);
      // status will be 0 for file:// urls
      if (!client.status || (client.status >= 200 && client.status < 300)) {
        let response;
        try {
          response = /** @type {Config} */ (JSON.parse(client.responseText));
        } catch (err) {
          this.handleTileJSONError();
          return;
        }
        this.handleTileJSONResponse(response);
      } else {
        this.handleTileJSONError();
      }
    }

    /**
     * @private
     * @param {Event} event The error event.
     */
    onXHRError_(event) {
      this.handleTileJSONError();
    }

    /**
     * @return {Config} The tilejson object.
     * @api
     */
    getTileJSON() {
      return this.tileJSON_;
    }

    /**
     * @protected
     * @param {Config} tileJSON Tile JSON.
     */
    handleTileJSONResponse(tileJSON) {
      const epsg4326Projection = get$2('EPSG:4326');

      const sourceProjection = this.getProjection();
      let extent;
      if (tileJSON['bounds'] !== undefined) {
        const transform = getTransformFromProjections(
          epsg4326Projection,
          sourceProjection
        );
        extent = applyTransform(tileJSON['bounds'], transform);
      }

      const gridExtent = extentFromProjection(sourceProjection);
      const minZoom = tileJSON['minzoom'] || 0;
      const maxZoom = tileJSON['maxzoom'] || 22;
      const tileGrid = createXYZ({
        extent: gridExtent,
        maxZoom: maxZoom,
        minZoom: minZoom,
        tileSize: this.tileSize_,
      });
      this.tileGrid = tileGrid;

      this.tileUrlFunction = createFromTemplates(tileJSON['tiles'], tileGrid);

      if (tileJSON['attribution'] && !this.getAttributions()) {
        const attributionExtent = extent !== undefined ? extent : gridExtent;
        this.setAttributions(function (frameState) {
          if (intersects$3(attributionExtent, frameState.extent)) {
            return [tileJSON['attribution']];
          }
          return null;
        });
      }
      this.tileJSON_ = tileJSON;
      this.setState('ready');
    }

    /**
     * @protected
     */
    handleTileJSONError() {
      this.setState('error');
    }
  }

  /**
   * @module ol/source/TileWMS
   */

  /**
   * @typedef {Object} Options
   * @property {import("./Source.js").AttributionLike} [attributions] Attributions.
   * @property {boolean} [attributionsCollapsible=true] Attributions are collapsible.
   * @property {number} [cacheSize] Initial tile cache size. Will auto-grow to hold at least the number of tiles in the viewport.
   * @property {null|string} [crossOrigin] The `crossOrigin` attribute for loaded images.  Note that
   * you must provide a `crossOrigin` value if you want to access pixel data with the Canvas renderer.
   * See https://developer.mozilla.org/en-US/docs/Web/HTML/CORS_enabled_image for more detail.
   * @property {boolean} [interpolate=true] Use interpolated values when resampling.  By default,
   * linear interpolation is used when resampling.  Set to false to use the nearest neighbor instead.
   * @property {Object<string,*>} params WMS request parameters.
   * At least a `LAYERS` param is required. `STYLES` is
   * `''` by default. `VERSION` is `1.3.0` by default. `WIDTH`, `HEIGHT`, `BBOX`
   * and `CRS` (`SRS` for WMS version < 1.3.0) will be set dynamically.
   * @property {number} [gutter=0]
   * The size in pixels of the gutter around image tiles to ignore. By setting
   * this property to a non-zero value, images will be requested that are wider
   * and taller than the tile size by a value of `2 x gutter`.
   * Using a non-zero value allows artifacts of rendering at tile edges to be
   * ignored. If you control the WMS service it is recommended to address
   * "artifacts at tile edges" issues by properly configuring the WMS service. For
   * example, MapServer has a `tile_map_edge_buffer` configuration parameter for
   * this. See https://mapserver.org/output/tile_mode.html.
   * @property {boolean} [hidpi=true] Use the `ol/Map#pixelRatio` value when requesting
   * the image from the remote server.
   * @property {import("../proj.js").ProjectionLike} [projection] Projection. Default is the view projection.
   * @property {number} [reprojectionErrorThreshold=0.5] Maximum allowed reprojection error (in pixels).
   * Higher values can increase reprojection performance, but decrease precision.
   * @property {typeof import("../ImageTile.js").default} [tileClass] Class used to instantiate image tiles.
   * Default is {@link module:ol/ImageTile~ImageTile}.
   * @property {import("../tilegrid/TileGrid.js").default} [tileGrid] Tile grid. Base this on the resolutions,
   * tilesize and extent supported by the server.
   * If this is not defined, a default grid will be used: if there is a projection
   * extent, the grid will be based on that; if not, a grid based on a global
   * extent with origin at 0,0 will be used.
   * @property {import("./wms.js").ServerType} [serverType] The type of
   * the remote WMS server: `mapserver`, `geoserver`, `carmentaserver`, or `qgis`.
   * Only needed if `hidpi` is `true`.
   * @property {import("../Tile.js").LoadFunction} [tileLoadFunction] Optional function to load a tile given a URL. The default is
   * ```js
   * function(imageTile, src) {
   *   imageTile.getImage().src = src;
   * };
   * ```
   * @property {string} [url] WMS service URL.
   * @property {Array<string>} [urls] WMS service urls.
   * Use this instead of `url` when the WMS supports multiple urls for GetMap requests.
   * @property {boolean} [wrapX=true] Whether to wrap the world horizontally.
   * When set to `false`, only one world
   * will be rendered. When `true`, tiles will be requested for one world only,
   * but they will be wrapped horizontally to render multiple worlds.
   * @property {number} [transition] Duration of the opacity transition for rendering.
   * To disable the opacity transition, pass `transition: 0`.
   * @property {number|import("../array.js").NearestDirectionFunction} [zDirection=0]
   * Choose whether to use tiles with a higher or lower zoom level when between integer
   * zoom levels. See {@link module:ol/tilegrid/TileGrid~TileGrid#getZForResolution}.
   */

  /**
   * @classdesc
   * Layer source for tile data from WMS servers.
   * @api
   */
  class TileWMS extends TileImage {
    /**
     * @param {Options} [options] Tile WMS options.
     */
    constructor(options) {
      options = options ? options : /** @type {Options} */ ({});

      const params = Object.assign({}, options.params);

      const transparent = 'TRANSPARENT' in params ? params['TRANSPARENT'] : true;

      super({
        attributions: options.attributions,
        attributionsCollapsible: options.attributionsCollapsible,
        cacheSize: options.cacheSize,
        crossOrigin: options.crossOrigin,
        interpolate: options.interpolate,
        opaque: !transparent,
        projection: options.projection,
        reprojectionErrorThreshold: options.reprojectionErrorThreshold,
        tileClass: options.tileClass,
        tileGrid: options.tileGrid,
        tileLoadFunction: options.tileLoadFunction,
        url: options.url,
        urls: options.urls,
        wrapX: options.wrapX !== undefined ? options.wrapX : true,
        transition: options.transition,
        zDirection: options.zDirection,
      });

      /**
       * @private
       * @type {number}
       */
      this.gutter_ = options.gutter !== undefined ? options.gutter : 0;

      /**
       * @private
       * @type {!Object}
       */
      this.params_ = params;

      /**
       * @private
       * @type {boolean}
       */
      this.v13_ = true;

      /**
       * @private
       * @type {import("./wms.js").ServerType}
       */
      this.serverType_ = options.serverType;

      /**
       * @private
       * @type {boolean}
       */
      this.hidpi_ = options.hidpi !== undefined ? options.hidpi : true;

      /**
       * @private
       * @type {import("../extent.js").Extent}
       */
      this.tmpExtent_ = createEmpty();

      this.updateV13_();
      this.setKey(this.getKeyForParams_());
    }

    /**
     * Return the GetFeatureInfo URL for the passed coordinate, resolution, and
     * projection. Return `undefined` if the GetFeatureInfo URL cannot be
     * constructed.
     * @param {import("../coordinate.js").Coordinate} coordinate Coordinate.
     * @param {number} resolution Resolution.
     * @param {import("../proj.js").ProjectionLike} projection Projection.
     * @param {!Object} params GetFeatureInfo params. `INFO_FORMAT` at least should
     *     be provided. If `QUERY_LAYERS` is not provided then the layers specified
     *     in the `LAYERS` parameter will be used. `VERSION` should not be
     *     specified here.
     * @return {string|undefined} GetFeatureInfo URL.
     * @api
     */
    getFeatureInfoUrl(coordinate, resolution, projection, params) {
      const projectionObj = get$2(projection);
      const sourceProjectionObj = this.getProjection();

      let tileGrid = this.getTileGrid();
      if (!tileGrid) {
        tileGrid = this.getTileGridForProjection(projectionObj);
      }

      const z = tileGrid.getZForResolution(resolution, this.zDirection);
      const tileCoord = tileGrid.getTileCoordForCoordAndZ(coordinate, z);

      if (tileGrid.getResolutions().length <= tileCoord[0]) {
        return undefined;
      }

      let tileResolution = tileGrid.getResolution(tileCoord[0]);
      let tileExtent = tileGrid.getTileCoordExtent(tileCoord, this.tmpExtent_);
      let tileSize = toSize(tileGrid.getTileSize(tileCoord[0]), this.tmpSize);

      const gutter = this.gutter_;
      if (gutter !== 0) {
        tileSize = buffer(tileSize, gutter, this.tmpSize);
        tileExtent = buffer$1(tileExtent, tileResolution * gutter, tileExtent);
      }

      if (sourceProjectionObj && sourceProjectionObj !== projectionObj) {
        tileResolution = calculateSourceResolution(
          sourceProjectionObj,
          projectionObj,
          coordinate,
          tileResolution
        );
        tileExtent = transformExtent(
          tileExtent,
          projectionObj,
          sourceProjectionObj
        );
        coordinate = transform(coordinate, projectionObj, sourceProjectionObj);
      }

      const baseParams = {
        'SERVICE': 'WMS',
        'VERSION': DEFAULT_VERSION$1,
        'REQUEST': 'GetFeatureInfo',
        'FORMAT': 'image/png',
        'TRANSPARENT': true,
        'QUERY_LAYERS': this.params_['LAYERS'],
      };
      Object.assign(baseParams, this.params_, params);

      const x = Math.floor((coordinate[0] - tileExtent[0]) / tileResolution);
      const y = Math.floor((tileExtent[3] - coordinate[1]) / tileResolution);

      baseParams[this.v13_ ? 'I' : 'X'] = x;
      baseParams[this.v13_ ? 'J' : 'Y'] = y;

      return this.getRequestUrl_(
        tileCoord,
        tileSize,
        tileExtent,
        1,
        sourceProjectionObj || projectionObj,
        baseParams
      );
    }

    /**
     * Return the GetLegendGraphic URL, optionally optimized for the passed
     * resolution and possibly including any passed specific parameters. Returns
     * `undefined` if the GetLegendGraphic URL cannot be constructed.
     *
     * @param {number} [resolution] Resolution. If set to undefined, `SCALE`
     *     will not be calculated and included in URL.
     * @param {Object} [params] GetLegendGraphic params. If `LAYER` is set, the
     *     request is generated for this wms layer, else it will try to use the
     *     configured wms layer. Default `FORMAT` is `image/png`.
     *     `VERSION` should not be specified here.
     * @return {string|undefined} GetLegendGraphic URL.
     * @api
     */
    getLegendUrl(resolution, params) {
      if (this.urls[0] === undefined) {
        return undefined;
      }

      const baseParams = {
        'SERVICE': 'WMS',
        'VERSION': DEFAULT_VERSION$1,
        'REQUEST': 'GetLegendGraphic',
        'FORMAT': 'image/png',
      };

      if (params === undefined || params['LAYER'] === undefined) {
        const layers = this.params_.LAYERS;
        const isSingleLayer = !Array.isArray(layers) || layers.length === 1;
        if (!isSingleLayer) {
          return undefined;
        }
        baseParams['LAYER'] = layers;
      }

      if (resolution !== undefined) {
        const mpu = this.getProjection()
          ? this.getProjection().getMetersPerUnit()
          : 1;
        const pixelSize = 0.00028;
        baseParams['SCALE'] = (resolution * mpu) / pixelSize;
      }

      Object.assign(baseParams, params);

      return appendParams(/** @type {string} */ (this.urls[0]), baseParams);
    }

    /**
     * @return {number} Gutter.
     */
    getGutter() {
      return this.gutter_;
    }

    /**
     * Get the user-provided params, i.e. those passed to the constructor through
     * the "params" option, and possibly updated using the updateParams method.
     * @return {Object} Params.
     * @api
     */
    getParams() {
      return this.params_;
    }

    /**
     * @param {import("../tilecoord.js").TileCoord} tileCoord Tile coordinate.
     * @param {import("../size.js").Size} tileSize Tile size.
     * @param {import("../extent.js").Extent} tileExtent Tile extent.
     * @param {number} pixelRatio Pixel ratio.
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @param {Object} params Params.
     * @return {string|undefined} Request URL.
     * @private
     */
    getRequestUrl_(
      tileCoord,
      tileSize,
      tileExtent,
      pixelRatio,
      projection,
      params
    ) {
      const urls = this.urls;
      if (!urls) {
        return undefined;
      }

      params['WIDTH'] = tileSize[0];
      params['HEIGHT'] = tileSize[1];

      params[this.v13_ ? 'CRS' : 'SRS'] = projection.getCode();

      if (!('STYLES' in this.params_)) {
        params['STYLES'] = '';
      }

      if (pixelRatio != 1) {
        switch (this.serverType_) {
          case 'geoserver':
            const dpi = (90 * pixelRatio + 0.5) | 0;
            if ('FORMAT_OPTIONS' in params) {
              params['FORMAT_OPTIONS'] += ';dpi:' + dpi;
            } else {
              params['FORMAT_OPTIONS'] = 'dpi:' + dpi;
            }
            break;
          case 'mapserver':
            params['MAP_RESOLUTION'] = 90 * pixelRatio;
            break;
          case 'carmentaserver':
          case 'qgis':
            params['DPI'] = 90 * pixelRatio;
            break;
          default: // Unknown `serverType` configured
            assert(false, 52);
            break;
        }
      }

      const axisOrientation = projection.getAxisOrientation();
      const bbox = tileExtent;
      if (this.v13_ && axisOrientation.substr(0, 2) == 'ne') {
        let tmp;
        tmp = tileExtent[0];
        bbox[0] = tileExtent[1];
        bbox[1] = tmp;
        tmp = tileExtent[2];
        bbox[2] = tileExtent[3];
        bbox[3] = tmp;
      }
      params['BBOX'] = bbox.join(',');

      let url;
      if (urls.length == 1) {
        url = urls[0];
      } else {
        const index = modulo(hash(tileCoord), urls.length);
        url = urls[index];
      }
      return appendParams(url, params);
    }

    /**
     * Get the tile pixel ratio for this source.
     * @param {number} pixelRatio Pixel ratio.
     * @return {number} Tile pixel ratio.
     */
    getTilePixelRatio(pixelRatio) {
      return !this.hidpi_ || this.serverType_ === undefined ? 1 : pixelRatio;
    }

    /**
     * @private
     * @return {string} The key for the current params.
     */
    getKeyForParams_() {
      let i = 0;
      const res = [];
      for (const key in this.params_) {
        res[i++] = key + '-' + this.params_[key];
      }
      return res.join('/');
    }

    /**
     * Update the user-provided params.
     * @param {Object} params Params.
     * @api
     */
    updateParams(params) {
      Object.assign(this.params_, params);
      this.updateV13_();
      this.setKey(this.getKeyForParams_());
    }

    /**
     * @private
     */
    updateV13_() {
      const version = this.params_['VERSION'] || DEFAULT_VERSION$1;
      this.v13_ = compareVersions(version, '1.3') >= 0;
    }

    /**
     * @param {import("../tilecoord.js").TileCoord} tileCoord The tile coordinate
     * @param {number} pixelRatio The pixel ratio
     * @param {import("../proj/Projection.js").default} projection The projection
     * @return {string|undefined} The tile URL
     * @override
     */
    tileUrlFunction(tileCoord, pixelRatio, projection) {
      let tileGrid = this.getTileGrid();
      if (!tileGrid) {
        tileGrid = this.getTileGridForProjection(projection);
      }

      if (tileGrid.getResolutions().length <= tileCoord[0]) {
        return undefined;
      }

      if (pixelRatio != 1 && (!this.hidpi_ || this.serverType_ === undefined)) {
        pixelRatio = 1;
      }

      const tileResolution = tileGrid.getResolution(tileCoord[0]);
      let tileExtent = tileGrid.getTileCoordExtent(tileCoord, this.tmpExtent_);
      let tileSize = toSize(tileGrid.getTileSize(tileCoord[0]), this.tmpSize);

      const gutter = this.gutter_;
      if (gutter !== 0) {
        tileSize = buffer(tileSize, gutter, this.tmpSize);
        tileExtent = buffer$1(tileExtent, tileResolution * gutter, tileExtent);
      }

      if (pixelRatio != 1) {
        tileSize = scale(tileSize, pixelRatio, this.tmpSize);
      }

      const baseParams = {
        'SERVICE': 'WMS',
        'VERSION': DEFAULT_VERSION$1,
        'REQUEST': 'GetMap',
        'FORMAT': 'image/png',
        'TRANSPARENT': true,
      };
      Object.assign(baseParams, this.params_);

      return this.getRequestUrl_(
        tileCoord,
        tileSize,
        tileExtent,
        pixelRatio,
        projection,
        baseParams
      );
    }
  }

  /**
   * @module ol/source/UTFGrid
   */

  /**
   * @typedef {Object} UTFGridJSON
   * @property {Array<string>} grid The grid.
   * @property {Array<string>} keys The keys.
   * @property {Object<string, Object>} [data] Optional data.
   */

  class CustomTile extends Tile {
    /**
     * @param {import("../tilecoord.js").TileCoord} tileCoord Tile coordinate.
     * @param {import("../TileState.js").default} state State.
     * @param {string} src Image source URI.
     * @param {import("../extent.js").Extent} extent Extent of the tile.
     * @param {boolean} preemptive Load the tile when visible (before it's needed).
     * @param {boolean} jsonp Load the tile as a script.
     */
    constructor(tileCoord, state, src, extent, preemptive, jsonp) {
      super(tileCoord, state);

      /**
       * @private
       * @type {string}
       */
      this.src_ = src;

      /**
       * @private
       * @type {import("../extent.js").Extent}
       */
      this.extent_ = extent;

      /**
       * @private
       * @type {boolean}
       */
      this.preemptive_ = preemptive;

      /**
       * @private
       * @type {Array<string>}
       */
      this.grid_ = null;

      /**
       * @private
       * @type {Array<string>}
       */
      this.keys_ = null;

      /**
       * @private
       * @type {Object<string, Object>|undefined}
       */
      this.data_ = null;

      /**
       * @private
       * @type {boolean}
       */
      this.jsonp_ = jsonp;
    }

    /**
     * Get the image element for this tile.
     * @return {HTMLImageElement} Image.
     */
    getImage() {
      return null;
    }

    /**
     * Synchronously returns data at given coordinate (if available).
     * @param {import("../coordinate.js").Coordinate} coordinate Coordinate.
     * @return {*} The data.
     */
    getData(coordinate) {
      if (!this.grid_ || !this.keys_) {
        return null;
      }
      const xRelative =
        (coordinate[0] - this.extent_[0]) / (this.extent_[2] - this.extent_[0]);
      const yRelative =
        (coordinate[1] - this.extent_[1]) / (this.extent_[3] - this.extent_[1]);

      const row = this.grid_[Math.floor((1 - yRelative) * this.grid_.length)];

      if (typeof row !== 'string') {
        return null;
      }

      let code = row.charCodeAt(Math.floor(xRelative * row.length));
      if (code >= 93) {
        code--;
      }
      if (code >= 35) {
        code--;
      }
      code -= 32;

      let data = null;
      if (code in this.keys_) {
        const id = this.keys_[code];
        if (this.data_ && id in this.data_) {
          data = this.data_[id];
        } else {
          data = id;
        }
      }
      return data;
    }

    /**
     * Calls the callback (synchronously by default) with the available data
     * for given coordinate (or `null` if not yet loaded).
     * @param {import("../coordinate.js").Coordinate} coordinate Coordinate.
     * @param {function(*): void} callback Callback.
     * @param {boolean} [request] If `true` the callback is always async.
     *                               The tile data is requested if not yet loaded.
     */
    forDataAtCoordinate(coordinate, callback, request) {
      if (this.state == TileState.EMPTY && request === true) {
        this.state = TileState.IDLE;
        listenOnce(
          this,
          EventType$1.CHANGE,
          function (e) {
            callback(this.getData(coordinate));
          },
          this
        );
        this.loadInternal_();
      } else {
        if (request === true) {
          setTimeout(
            function () {
              callback(this.getData(coordinate));
            }.bind(this),
            0
          );
        } else {
          callback(this.getData(coordinate));
        }
      }
    }

    /**
     * Return the key to be used for all tiles in the source.
     * @return {string} The key for all tiles.
     */
    getKey() {
      return this.src_;
    }

    /**
     * @private
     */
    handleError_() {
      this.state = TileState.ERROR;
      this.changed();
    }

    /**
     * @param {!UTFGridJSON} json UTFGrid data.
     * @private
     */
    handleLoad_(json) {
      this.grid_ = json['grid'];
      this.keys_ = json['keys'];
      this.data_ = json['data'];

      this.state = TileState.LOADED;
      this.changed();
    }

    /**
     * @private
     */
    loadInternal_() {
      if (this.state == TileState.IDLE) {
        this.state = TileState.LOADING;
        if (this.jsonp_) {
          jsonp(
            this.src_,
            this.handleLoad_.bind(this),
            this.handleError_.bind(this)
          );
        } else {
          const client = new XMLHttpRequest();
          client.addEventListener('load', this.onXHRLoad_.bind(this));
          client.addEventListener('error', this.onXHRError_.bind(this));
          client.open('GET', this.src_);
          client.send();
        }
      }
    }

    /**
     * @private
     * @param {Event} event The load event.
     */
    onXHRLoad_(event) {
      const client = /** @type {XMLHttpRequest} */ (event.target);
      // status will be 0 for file:// urls
      if (!client.status || (client.status >= 200 && client.status < 300)) {
        let response;
        try {
          response = /** @type {!UTFGridJSON} */ (
            JSON.parse(client.responseText)
          );
        } catch (err) {
          this.handleError_();
          return;
        }
        this.handleLoad_(response);
      } else {
        this.handleError_();
      }
    }

    /**
     * @private
     * @param {Event} event The error event.
     */
    onXHRError_(event) {
      this.handleError_();
    }

    /**
     */
    load() {
      if (this.preemptive_) {
        this.loadInternal_();
      } else {
        this.setState(TileState.EMPTY);
      }
    }
  }

  /**
   * @typedef {Object} Options
   * @property {boolean} [preemptive=true]
   * If `true` the UTFGrid source loads the tiles based on their "visibility".
   * This improves the speed of response, but increases traffic.
   * Note that if set to `false` (lazy loading), you need to pass `true` as
   * `request` to the `forDataAtCoordinateAndResolution` method otherwise no
   * data will ever be loaded.
   * @property {boolean} [jsonp=false] Use JSONP with callback to load the TileJSON.
   * Useful when the server does not support CORS..
   * @property {import("./TileJSON.js").Config} [tileJSON] TileJSON configuration for this source.
   * If not provided, `url` must be configured.
   * @property {string} [url] TileJSON endpoint that provides the configuration for this source.
   * Request will be made through JSONP. If not provided, `tileJSON` must be configured.
   * @property {number|import("../array.js").NearestDirectionFunction} [zDirection=0]
   * Choose whether to use tiles with a higher or lower zoom level when between integer
   * zoom levels. See {@link module:ol/tilegrid/TileGrid~TileGrid#getZForResolution}.
   */

  /**
   * @classdesc
   * Layer source for UTFGrid interaction data loaded from TileJSON format.
   * @api
   */
  class UTFGrid extends TileSource {
    /**
     * @param {Options} options Source options.
     */
    constructor(options) {
      super({
        projection: get$2('EPSG:3857'),
        state: 'loading',
        zDirection: options.zDirection,
      });

      /**
       * @private
       * @type {boolean}
       */
      this.preemptive_ =
        options.preemptive !== undefined ? options.preemptive : true;

      /**
       * @private
       * @type {!import("../Tile.js").UrlFunction}
       */
      this.tileUrlFunction_ = nullTileUrlFunction;

      /**
       * @private
       * @type {string|undefined}
       */
      this.template_ = undefined;

      /**
       * @private
       * @type {boolean}
       */
      this.jsonp_ = options.jsonp || false;

      if (options.url) {
        if (this.jsonp_) {
          jsonp(
            options.url,
            this.handleTileJSONResponse.bind(this),
            this.handleTileJSONError.bind(this)
          );
        } else {
          const client = new XMLHttpRequest();
          client.addEventListener('load', this.onXHRLoad_.bind(this));
          client.addEventListener('error', this.onXHRError_.bind(this));
          client.open('GET', options.url);
          client.send();
        }
      } else if (options.tileJSON) {
        this.handleTileJSONResponse(options.tileJSON);
      } else {
        assert(false, 51); // Either `url` or `tileJSON` options must be provided
      }
    }

    /**
     * @private
     * @param {Event} event The load event.
     */
    onXHRLoad_(event) {
      const client = /** @type {XMLHttpRequest} */ (event.target);
      // status will be 0 for file:// urls
      if (!client.status || (client.status >= 200 && client.status < 300)) {
        let response;
        try {
          response = /** @type {import("./TileJSON.js").Config} */ (
            JSON.parse(client.responseText)
          );
        } catch (err) {
          this.handleTileJSONError();
          return;
        }
        this.handleTileJSONResponse(response);
      } else {
        this.handleTileJSONError();
      }
    }

    /**
     * @private
     * @param {Event} event The error event.
     */
    onXHRError_(event) {
      this.handleTileJSONError();
    }

    /**
     * Return the template from TileJSON.
     * @return {string|undefined} The template from TileJSON.
     * @api
     */
    getTemplate() {
      return this.template_;
    }

    /**
     * Calls the callback (synchronously by default) with the available data
     * for given coordinate and resolution (or `null` if not yet loaded or
     * in case of an error).
     * @param {import("../coordinate.js").Coordinate} coordinate Coordinate.
     * @param {number} resolution Resolution.
     * @param {function(*): void} callback Callback.
     * @param {boolean} [request] If `true` the callback is always async.
     *                               The tile data is requested if not yet loaded.
     * @api
     */
    forDataAtCoordinateAndResolution(coordinate, resolution, callback, request) {
      if (this.tileGrid) {
        const z = this.tileGrid.getZForResolution(resolution, this.zDirection);
        const tileCoord = this.tileGrid.getTileCoordForCoordAndZ(coordinate, z);
        const tile = /** @type {!CustomTile} */ (
          this.getTile(
            tileCoord[0],
            tileCoord[1],
            tileCoord[2],
            1,
            this.getProjection()
          )
        );
        tile.forDataAtCoordinate(coordinate, callback, request);
      } else {
        if (request === true) {
          setTimeout(function () {
            callback(null);
          }, 0);
        } else {
          callback(null);
        }
      }
    }

    /**
     * @protected
     */
    handleTileJSONError() {
      this.setState('error');
    }

    /**
     * TODO: very similar to ol/source/TileJSON#handleTileJSONResponse
     * @protected
     * @param {import("./TileJSON.js").Config} tileJSON Tile JSON.
     */
    handleTileJSONResponse(tileJSON) {
      const epsg4326Projection = get$2('EPSG:4326');

      const sourceProjection = this.getProjection();
      let extent;
      if (tileJSON['bounds'] !== undefined) {
        const transform = getTransformFromProjections(
          epsg4326Projection,
          sourceProjection
        );
        extent = applyTransform(tileJSON['bounds'], transform);
      }

      const gridExtent = extentFromProjection(sourceProjection);
      const minZoom = tileJSON['minzoom'] || 0;
      const maxZoom = tileJSON['maxzoom'] || 22;
      const tileGrid = createXYZ({
        extent: gridExtent,
        maxZoom: maxZoom,
        minZoom: minZoom,
      });
      this.tileGrid = tileGrid;

      this.template_ = tileJSON['template'];

      const grids = tileJSON['grids'];
      if (!grids) {
        this.setState('error');
        return;
      }

      this.tileUrlFunction_ = createFromTemplates(grids, tileGrid);

      if (tileJSON['attribution'] !== undefined) {
        const attributionExtent = extent !== undefined ? extent : gridExtent;
        this.setAttributions(function (frameState) {
          if (intersects$3(attributionExtent, frameState.extent)) {
            return [tileJSON['attribution']];
          }
          return null;
        });
      }

      this.setState('ready');
    }

    /**
     * @param {number} z Tile coordinate z.
     * @param {number} x Tile coordinate x.
     * @param {number} y Tile coordinate y.
     * @param {number} pixelRatio Pixel ratio.
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @return {!CustomTile} Tile.
     */
    getTile(z, x, y, pixelRatio, projection) {
      const tileCoordKey = getKeyZXY(z, x, y);
      if (this.tileCache.containsKey(tileCoordKey)) {
        return this.tileCache.get(tileCoordKey);
      } else {
        const tileCoord = [z, x, y];
        const urlTileCoord = this.getTileCoordForTileUrlFunction(
          tileCoord,
          projection
        );
        const tileUrl = this.tileUrlFunction_(
          urlTileCoord,
          pixelRatio,
          projection
        );
        const tile = new CustomTile(
          tileCoord,
          tileUrl !== undefined ? TileState.IDLE : TileState.EMPTY,
          tileUrl !== undefined ? tileUrl : '',
          this.tileGrid.getTileCoordExtent(tileCoord),
          this.preemptive_,
          this.jsonp_
        );
        this.tileCache.set(tileCoordKey, tile);
        return tile;
      }
    }

    /**
     * Marks a tile coord as being used, without triggering a load.
     * @param {number} z Tile coordinate z.
     * @param {number} x Tile coordinate x.
     * @param {number} y Tile coordinate y.
     */
    useTile(z, x, y) {
      const tileCoordKey = getKeyZXY(z, x, y);
      if (this.tileCache.containsKey(tileCoordKey)) {
        this.tileCache.get(tileCoordKey);
      }
    }
  }

  /**
   * @module ol/source/VectorTile
   */

  /**
   * @typedef {Object} Options
   * @property {import("./Source.js").AttributionLike} [attributions] Attributions.
   * @property {boolean} [attributionsCollapsible=true] Attributions are collapsible.
   * @property {number} [cacheSize] Initial tile cache size. Will auto-grow to hold at least twice the number of tiles in the viewport.
   * @property {import("../extent.js").Extent} [extent] Extent.
   * @property {import("../format/Feature.js").default} [format] Feature format for tiles. Used and required by the default.
   * @property {boolean} [overlaps=true] This source may have overlapping geometries. Setting this
   * to `false` (e.g. for sources with polygons that represent administrative
   * boundaries or TopoJSON sources) allows the renderer to optimise fill and
   * stroke operations.
   * @property {import("../proj.js").ProjectionLike} [projection='EPSG:3857'] Projection of the tile grid.
   * @property {import("./Source.js").State} [state] Source state.
   * @property {typeof import("../VectorTile.js").default} [tileClass] Class used to instantiate image tiles.
   * Default is {@link module:ol/VectorTile~VectorTile}.
   * @property {number} [maxZoom=22] Optional max zoom level. Not used if `tileGrid` is provided.
   * @property {number} [minZoom] Optional min zoom level. Not used if `tileGrid` is provided.
   * @property {number|import("../size.js").Size} [tileSize=512] Optional tile size. Not used if `tileGrid` is provided.
   * @property {number} [maxResolution] Optional tile grid resolution at level zero. Not used if `tileGrid` is provided.
   * @property {import("../tilegrid/TileGrid.js").default} [tileGrid] Tile grid.
   * @property {import("../Tile.js").LoadFunction} [tileLoadFunction]
   * Optional function to load a tile given a URL. Could look like this for pbf tiles:
   * ```js
   * function(tile, url) {
   *   tile.setLoader(function(extent, resolution, projection) {
   *     fetch(url).then(function(response) {
   *       response.arrayBuffer().then(function(data) {
   *         const format = tile.getFormat() // ol/format/MVT configured as source format
   *         const features = format.readFeatures(data, {
   *           extent: extent,
   *           featureProjection: projection
   *         });
   *         tile.setFeatures(features);
   *       });
   *     });
   *   });
   * }
   * ```
   * If you do not need extent, resolution and projection to get the features for a tile (e.g.
   * for GeoJSON tiles), your `tileLoadFunction` does not need a `setLoader()` call. Only make sure
   * to call `setFeatures()` on the tile:
   * ```js
   * const format = new GeoJSON({featureProjection: map.getView().getProjection()});
   * async function tileLoadFunction(tile, url) {
   *   const response = await fetch(url);
   *   const data = await response.json();
   *   tile.setFeatures(format.readFeatures(data));
   * }
   * ```
   * @property {import("../Tile.js").UrlFunction} [tileUrlFunction] Optional function to get tile URL given a tile coordinate and the projection.
   * @property {string} [url] URL template. Must include `{x}`, `{y}` or `{-y}`, and `{z}` placeholders.
   * A `{?-?}` template pattern, for example `subdomain{a-f}.domain.com`, may be
   * used instead of defining each one separately in the `urls` option.
   * @property {number} [transition] A duration for tile opacity
   * transitions in milliseconds. A duration of 0 disables the opacity transition.
   * @property {Array<string>} [urls] An array of URL templates.
   * @property {boolean} [wrapX=true] Whether to wrap the world horizontally.
   * When set to `false`, only one world
   * will be rendered. When set to `true`, tiles will be wrapped horizontally to
   * render multiple worlds.
   * @property {number|import("../array.js").NearestDirectionFunction} [zDirection=1]
   * Choose whether to use tiles with a higher or lower zoom level when between integer
   * zoom levels. See {@link module:ol/tilegrid/TileGrid~TileGrid#getZForResolution}.
   */

  /**
   * @classdesc
   * Class for layer sources providing vector data divided into a tile grid, to be
   * used with {@link module:ol/layer/VectorTile~VectorTileLayer}. Although this source receives tiles
   * with vector features from the server, it is not meant for feature editing.
   * Features are optimized for rendering, their geometries are clipped at or near
   * tile boundaries and simplified for a view resolution. See
   * {@link module:ol/source/Vector~VectorSource} for vector sources that are suitable for feature
   * editing.
   *
   * @fires import("./Tile.js").TileSourceEvent
   * @api
   */
  class VectorTile extends UrlTile {
    /**
     * @param {!Options} options Vector tile options.
     */
    constructor(options) {
      const projection = options.projection || 'EPSG:3857';

      const extent = options.extent || extentFromProjection(projection);

      const tileGrid =
        options.tileGrid ||
        createXYZ({
          extent: extent,
          maxResolution: options.maxResolution,
          maxZoom: options.maxZoom !== undefined ? options.maxZoom : 22,
          minZoom: options.minZoom,
          tileSize: options.tileSize || 512,
        });

      super({
        attributions: options.attributions,
        attributionsCollapsible: options.attributionsCollapsible,
        cacheSize: options.cacheSize,
        interpolate: true,
        opaque: false,
        projection: projection,
        state: options.state,
        tileGrid: tileGrid,
        tileLoadFunction: options.tileLoadFunction
          ? options.tileLoadFunction
          : defaultLoadFunction,
        tileUrlFunction: options.tileUrlFunction,
        url: options.url,
        urls: options.urls,
        wrapX: options.wrapX === undefined ? true : options.wrapX,
        transition: options.transition,
        zDirection: options.zDirection === undefined ? 1 : options.zDirection,
      });

      /**
       * @private
       * @type {import("../format/Feature.js").default|null}
       */
      this.format_ = options.format ? options.format : null;

      /**
       * @private
       * @type {TileCache}
       */
      this.sourceTileCache = new TileCache(this.tileCache.highWaterMark);

      /**
       * @private
       * @type {boolean}
       */
      this.overlaps_ = options.overlaps == undefined ? true : options.overlaps;

      /**
       * @protected
       * @type {typeof import("../VectorTile.js").default}
       */
      this.tileClass = options.tileClass ? options.tileClass : VectorTile$1;

      /**
       * @private
       * @type {Object<string, import("../tilegrid/TileGrid.js").default>}
       */
      this.tileGrids_ = {};
    }

    /**
     * Get features whose bounding box intersects the provided extent. Only features for cached
     * tiles for the last rendered zoom level are available in the source. So this method is only
     * suitable for requesting tiles for extents that are currently rendered.
     *
     * Features are returned in random tile order and as they are included in the tiles. This means
     * they can be clipped, duplicated across tiles, and simplified to the render resolution.
     *
     * @param {import("../extent.js").Extent} extent Extent.
     * @return {Array<import("../Feature.js").FeatureLike>} Features.
     * @api
     */
    getFeaturesInExtent(extent) {
      const features = [];
      const tileCache = this.tileCache;
      if (tileCache.getCount() === 0) {
        return features;
      }
      const z = fromKey(tileCache.peekFirstKey())[0];
      const tileGrid = this.tileGrid;
      tileCache.forEach(function (tile) {
        if (tile.tileCoord[0] !== z || tile.getState() !== TileState.LOADED) {
          return;
        }
        const sourceTiles = tile.getSourceTiles();
        for (let i = 0, ii = sourceTiles.length; i < ii; ++i) {
          const sourceTile = sourceTiles[i];
          const tileCoord = sourceTile.tileCoord;
          if (intersects$3(extent, tileGrid.getTileCoordExtent(tileCoord))) {
            const tileFeatures = sourceTile.getFeatures();
            if (tileFeatures) {
              for (let j = 0, jj = tileFeatures.length; j < jj; ++j) {
                const candidate = tileFeatures[j];
                const geometry = candidate.getGeometry();
                if (intersects$3(extent, geometry.getExtent())) {
                  features.push(candidate);
                }
              }
            }
          }
        }
      });
      return features;
    }

    /**
     * @return {boolean} The source can have overlapping geometries.
     */
    getOverlaps() {
      return this.overlaps_;
    }

    /**
     * clear {@link module:ol/TileCache~TileCache} and delete all source tiles
     * @api
     */
    clear() {
      this.tileCache.clear();
      this.sourceTileCache.clear();
    }

    /**
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @param {!Object<string, boolean>} usedTiles Used tiles.
     */
    expireCache(projection, usedTiles) {
      const tileCache = this.getTileCacheForProjection(projection);
      const usedSourceTiles = Object.keys(usedTiles).reduce((acc, key) => {
        const cacheKey = getCacheKeyForTileKey(key);
        const tile = tileCache.peek(cacheKey);
        if (tile) {
          const sourceTiles = tile.sourceTiles;
          for (let i = 0, ii = sourceTiles.length; i < ii; ++i) {
            acc[sourceTiles[i].getKey()] = true;
          }
        }
        return acc;
      }, {});
      super.expireCache(projection, usedTiles);
      this.sourceTileCache.expireCache(usedSourceTiles);
    }

    /**
     * @param {number} pixelRatio Pixel ratio.
     * @param {import("../proj/Projection").default} projection Projection.
     * @param {VectorRenderTile} tile Vector image tile.
     * @return {Array<import("../VectorTile").default>} Tile keys.
     */
    getSourceTiles(pixelRatio, projection, tile) {
      if (tile.getState() === TileState.IDLE) {
        tile.setState(TileState.LOADING);
        const urlTileCoord = tile.wrappedTileCoord;
        const tileGrid = this.getTileGridForProjection(projection);
        const extent = tileGrid.getTileCoordExtent(urlTileCoord);
        const z = urlTileCoord[0];
        const resolution = tileGrid.getResolution(z);
        // make extent 1 pixel smaller so we don't load tiles for < 0.5 pixel render space
        buffer$1(extent, -resolution, extent);
        const sourceTileGrid = this.tileGrid;
        const sourceExtent = sourceTileGrid.getExtent();
        if (sourceExtent) {
          getIntersection(extent, sourceExtent, extent);
        }
        const sourceZ = sourceTileGrid.getZForResolution(
          resolution,
          this.zDirection
        );

        sourceTileGrid.forEachTileCoord(extent, sourceZ, (sourceTileCoord) => {
          const tileUrl = this.tileUrlFunction(
            sourceTileCoord,
            pixelRatio,
            projection
          );
          const sourceTile = this.sourceTileCache.containsKey(tileUrl)
            ? this.sourceTileCache.get(tileUrl)
            : new this.tileClass(
                sourceTileCoord,
                tileUrl ? TileState.IDLE : TileState.EMPTY,
                tileUrl,
                this.format_,
                this.tileLoadFunction
              );
          tile.sourceTiles.push(sourceTile);
          const sourceTileState = sourceTile.getState();
          if (sourceTileState < TileState.LOADED) {
            const listenChange = (event) => {
              this.handleTileChange(event);
              const state = sourceTile.getState();
              if (state === TileState.LOADED || state === TileState.ERROR) {
                const sourceTileKey = sourceTile.getKey();
                if (sourceTileKey in tile.errorTileKeys) {
                  if (sourceTile.getState() === TileState.LOADED) {
                    delete tile.errorTileKeys[sourceTileKey];
                  }
                } else {
                  tile.loadingSourceTiles--;
                }
                if (state === TileState.ERROR) {
                  tile.errorTileKeys[sourceTileKey] = true;
                } else {
                  sourceTile.removeEventListener(EventType$1.CHANGE, listenChange);
                }
                if (tile.loadingSourceTiles === 0) {
                  tile.setState(
                    isEmpty$1(tile.errorTileKeys)
                      ? TileState.LOADED
                      : TileState.ERROR
                  );
                }
              }
            };
            sourceTile.addEventListener(EventType$1.CHANGE, listenChange);
            tile.loadingSourceTiles++;
          }
          if (sourceTileState === TileState.IDLE) {
            sourceTile.extent =
              sourceTileGrid.getTileCoordExtent(sourceTileCoord);
            sourceTile.projection = projection;
            sourceTile.resolution = sourceTileGrid.getResolution(
              sourceTileCoord[0]
            );
            this.sourceTileCache.set(tileUrl, sourceTile);
            sourceTile.load();
          }
        });
        if (!tile.loadingSourceTiles) {
          tile.setState(
            tile.sourceTiles.some(
              (sourceTile) => sourceTile.getState() === TileState.ERROR
            )
              ? TileState.ERROR
              : TileState.LOADED
          );
        }
      }

      return tile.sourceTiles;
    }

    /**
     * @param {number} z Tile coordinate z.
     * @param {number} x Tile coordinate x.
     * @param {number} y Tile coordinate y.
     * @param {number} pixelRatio Pixel ratio.
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @return {!VectorRenderTile} Tile.
     */
    getTile(z, x, y, pixelRatio, projection) {
      const coordKey = getKeyZXY(z, x, y);
      const key = this.getKey();
      let tile;
      if (this.tileCache.containsKey(coordKey)) {
        tile = this.tileCache.get(coordKey);
        if (tile.key === key) {
          return tile;
        }
      }
      const tileCoord = [z, x, y];
      let urlTileCoord = this.getTileCoordForTileUrlFunction(
        tileCoord,
        projection
      );
      const sourceExtent = this.getTileGrid().getExtent();
      const tileGrid = this.getTileGridForProjection(projection);
      if (urlTileCoord && sourceExtent) {
        const tileExtent = tileGrid.getTileCoordExtent(urlTileCoord);
        // make extent 1 pixel smaller so we don't load tiles for < 0.5 pixel render space
        buffer$1(tileExtent, -tileGrid.getResolution(z), tileExtent);
        if (!intersects$3(sourceExtent, tileExtent)) {
          urlTileCoord = null;
        }
      }
      let empty = true;
      if (urlTileCoord !== null) {
        const sourceTileGrid = this.tileGrid;
        const resolution = tileGrid.getResolution(z);
        const sourceZ = sourceTileGrid.getZForResolution(resolution, 1);
        // make extent 1 pixel smaller so we don't load tiles for < 0.5 pixel render space
        const extent = tileGrid.getTileCoordExtent(urlTileCoord);
        buffer$1(extent, -resolution, extent);
        sourceTileGrid.forEachTileCoord(
          extent,
          sourceZ,
          function (sourceTileCoord) {
            empty =
              empty &&
              !this.tileUrlFunction(sourceTileCoord, pixelRatio, projection);
          }.bind(this)
        );
      }
      const newTile = new VectorRenderTile(
        tileCoord,
        empty ? TileState.EMPTY : TileState.IDLE,
        urlTileCoord,
        this.getSourceTiles.bind(this, pixelRatio, projection)
      );

      newTile.key = key;
      if (tile) {
        newTile.interimTile = tile;
        newTile.refreshInterimChain();
        this.tileCache.replace(coordKey, newTile);
      } else {
        this.tileCache.set(coordKey, newTile);
      }
      return newTile;
    }

    /**
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @return {!import("../tilegrid/TileGrid.js").default} Tile grid.
     */
    getTileGridForProjection(projection) {
      const code = projection.getCode();
      let tileGrid = this.tileGrids_[code];
      if (!tileGrid) {
        // A tile grid that matches the tile size of the source tile grid is more
        // likely to have 1:1 relationships between source tiles and rendered tiles.
        const sourceTileGrid = this.tileGrid;
        const resolutions = sourceTileGrid.getResolutions().slice();
        const origins = resolutions.map(function (resolution, z) {
          return sourceTileGrid.getOrigin(z);
        });
        const tileSizes = resolutions.map(function (resolution, z) {
          return sourceTileGrid.getTileSize(z);
        });
        const length = DEFAULT_MAX_ZOOM + 1;
        for (let z = resolutions.length; z < length; ++z) {
          resolutions.push(resolutions[z - 1] / 2);
          origins.push(origins[z - 1]);
          tileSizes.push(tileSizes[z - 1]);
        }
        tileGrid = new TileGrid({
          extent: sourceTileGrid.getExtent(),
          origins: origins,
          resolutions: resolutions,
          tileSizes: tileSizes,
        });
        this.tileGrids_[code] = tileGrid;
      }
      return tileGrid;
    }

    /**
     * Get the tile pixel ratio for this source.
     * @param {number} pixelRatio Pixel ratio.
     * @return {number} Tile pixel ratio.
     */
    getTilePixelRatio(pixelRatio) {
      return pixelRatio;
    }

    /**
     * @param {number} z Z.
     * @param {number} pixelRatio Pixel ratio.
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @return {import("../size.js").Size} Tile size.
     */
    getTilePixelSize(z, pixelRatio, projection) {
      const tileGrid = this.getTileGridForProjection(projection);
      const tileSize = toSize(tileGrid.getTileSize(z), this.tmpSize);
      return [
        Math.round(tileSize[0] * pixelRatio),
        Math.round(tileSize[1] * pixelRatio),
      ];
    }

    /**
     * Increases the cache size if needed
     * @param {number} tileCount Minimum number of tiles needed.
     * @param {import("../proj/Projection.js").default} projection Projection.
     */
    updateCacheSize(tileCount, projection) {
      super.updateCacheSize(tileCount * 2, projection);
      this.sourceTileCache.highWaterMark =
        this.getTileCacheForProjection(projection).highWaterMark;
    }
  }

  /**
   * Sets the loader for a tile.
   * @param {import("../VectorTile.js").default} tile Vector tile.
   * @param {string} url URL.
   */
  function defaultLoadFunction(tile, url) {
    tile.setLoader(
      /**
       * @param {import("../extent.js").Extent} extent Extent.
       * @param {number} resolution Resolution.
       * @param {import("../proj/Projection.js").default} projection Projection.
       */
      function (extent, resolution, projection) {
        loadFeaturesXhr(
          url,
          tile.getFormat(),
          extent,
          resolution,
          projection,
          tile.onLoad.bind(tile),
          tile.onError.bind(tile)
        );
      }
    );
  }

  /**
   * @module ol/tilegrid/WMTS
   */

  /**
   * @typedef {Object} Options
   * @property {import("../extent.js").Extent} [extent] Extent for the tile grid. No tiles
   * outside this extent will be requested by {@link module:ol/source/Tile~TileSource} sources.
   * When no `origin` or `origins` are configured, the `origin` will be set to the
   * top-left corner of the extent.
   * @property {import("../coordinate.js").Coordinate} [origin] The tile grid origin, i.e.
   * where the `x` and `y` axes meet (`[z, 0, 0]`). Tile coordinates increase left
   * to right and downwards. If not specified, `extent` or `origins` must be provided.
   * @property {Array<import("../coordinate.js").Coordinate>} [origins] Tile grid origins,
   * i.e. where the `x` and `y` axes meet (`[z, 0, 0]`), for each zoom level. If
   * given, the array length should match the length of the `resolutions` array, i.e.
   * each resolution can have a different origin. Tile coordinates increase left to
   * right and downwards. If not specified, `extent` or `origin` must be provided.
   * @property {!Array<number>} resolutions Resolutions. The array index of each
   * resolution needs to match the zoom level. This means that even if a `minZoom`
   * is configured, the resolutions array will have a length of `maxZoom + 1`
   * @property {!Array<string>} matrixIds matrix IDs. The length of this array needs
   * to match the length of the `resolutions` array.
   * @property {Array<import("../size.js").Size>} [sizes] Number of tile rows and columns
   * of the grid for each zoom level. The values here are the `TileMatrixWidth` and
   * `TileMatrixHeight` advertised in the GetCapabilities response of the WMTS, and
   * define each zoom level's extent together with the `origin` or `origins`.
   * A grid `extent` can be configured in addition, and will further limit the extent for
   * which tile requests are made by sources. If the bottom-left corner of
   * an extent is used as `origin` or `origins`, then the `y` value must be
   * negative because OpenLayers tile coordinates use the top left as the origin.
   * @property {number|import("../size.js").Size} [tileSize] Tile size.
   * @property {Array<number|import("../size.js").Size>} [tileSizes] Tile sizes. The length of
   * this array needs to match the length of the `resolutions` array.
   */

  /**
   * @classdesc
   * Set the grid pattern for sources accessing WMTS tiled-image servers.
   * @api
   */
  class WMTSTileGrid extends TileGrid {
    /**
     * @param {Options} options WMTS options.
     */
    constructor(options) {
      super({
        extent: options.extent,
        origin: options.origin,
        origins: options.origins,
        resolutions: options.resolutions,
        tileSize: options.tileSize,
        tileSizes: options.tileSizes,
        sizes: options.sizes,
      });

      /**
       * @private
       * @type {!Array<string>}
       */
      this.matrixIds_ = options.matrixIds;
    }

    /**
     * @param {number} z Z.
     * @return {string} MatrixId..
     */
    getMatrixId(z) {
      return this.matrixIds_[z];
    }

    /**
     * Get the list of matrix identifiers.
     * @return {Array<string>} MatrixIds.
     * @api
     */
    getMatrixIds() {
      return this.matrixIds_;
    }
  }

  /**
   * Create a tile grid from a WMTS capabilities matrix set and an
   * optional TileMatrixSetLimits.
   * @param {Object} matrixSet An object representing a matrixSet in the
   *     capabilities document.
   * @param {import("../extent.js").Extent} [extent] An optional extent to restrict the tile
   *     ranges the server provides.
   * @param {Array<Object>} [matrixLimits] An optional object representing
   *     the available matrices for tileGrid.
   * @return {WMTSTileGrid} WMTS tileGrid instance.
   * @api
   */
  function createFromCapabilitiesMatrixSet(
    matrixSet,
    extent,
    matrixLimits
  ) {
    /** @type {!Array<number>} */
    const resolutions = [];
    /** @type {!Array<string>} */
    const matrixIds = [];
    /** @type {!Array<import("../coordinate.js").Coordinate>} */
    const origins = [];
    /** @type {!Array<number|import("../size.js").Size>} */
    const tileSizes = [];
    /** @type {!Array<import("../size.js").Size>} */
    const sizes = [];

    matrixLimits = matrixLimits !== undefined ? matrixLimits : [];

    const supportedCRSPropName = 'SupportedCRS';
    const matrixIdsPropName = 'TileMatrix';
    const identifierPropName = 'Identifier';
    const scaleDenominatorPropName = 'ScaleDenominator';
    const topLeftCornerPropName = 'TopLeftCorner';
    const tileWidthPropName = 'TileWidth';
    const tileHeightPropName = 'TileHeight';

    const code = matrixSet[supportedCRSPropName];
    const projection = get$2(code);
    const metersPerUnit = projection.getMetersPerUnit();
    // swap origin x and y coordinates if axis orientation is lat/long
    const switchOriginXY = projection.getAxisOrientation().substr(0, 2) == 'ne';

    matrixSet[matrixIdsPropName].sort(function (a, b) {
      return b[scaleDenominatorPropName] - a[scaleDenominatorPropName];
    });

    matrixSet[matrixIdsPropName].forEach(function (elt) {
      let matrixAvailable;
      // use of matrixLimits to filter TileMatrices from GetCapabilities
      // TileMatrixSet from unavailable matrix levels.
      if (matrixLimits.length > 0) {
        matrixAvailable = matrixLimits.find(function (elt_ml) {
          if (elt[identifierPropName] == elt_ml[matrixIdsPropName]) {
            return true;
          }
          // Fallback for tileMatrix identifiers that don't get prefixed
          // by their tileMatrixSet identifiers.
          if (!elt[identifierPropName].includes(':')) {
            return (
              matrixSet[identifierPropName] + ':' + elt[identifierPropName] ===
              elt_ml[matrixIdsPropName]
            );
          }
          return false;
        });
      } else {
        matrixAvailable = true;
      }

      if (matrixAvailable) {
        matrixIds.push(elt[identifierPropName]);
        const resolution =
          (elt[scaleDenominatorPropName] * 0.28e-3) / metersPerUnit;
        const tileWidth = elt[tileWidthPropName];
        const tileHeight = elt[tileHeightPropName];
        if (switchOriginXY) {
          origins.push([
            elt[topLeftCornerPropName][1],
            elt[topLeftCornerPropName][0],
          ]);
        } else {
          origins.push(elt[topLeftCornerPropName]);
        }
        resolutions.push(resolution);
        tileSizes.push(
          tileWidth == tileHeight ? tileWidth : [tileWidth, tileHeight]
        );
        sizes.push([elt['MatrixWidth'], elt['MatrixHeight']]);
      }
    });

    return new WMTSTileGrid({
      extent: extent,
      origins: origins,
      resolutions: resolutions,
      matrixIds: matrixIds,
      tileSizes: tileSizes,
      sizes: sizes,
    });
  }

  /**
   * @module ol/source/WMTS
   */

  /**
   * Request encoding. One of 'KVP', 'REST'.
   * @typedef {'KVP' | 'REST'} RequestEncoding
   */

  /**
   * @typedef {Object} Options
   * @property {import("./Source.js").AttributionLike} [attributions] Attributions.
   * @property {boolean} [attributionsCollapsible=true] Attributions are collapsible.
   * @property {number} [cacheSize] Initial tile cache size. Will auto-grow to hold at least the number of tiles in the viewport.
   * @property {null|string} [crossOrigin] The `crossOrigin` attribute for loaded images.  Note that
   * you must provide a `crossOrigin` value if you want to access pixel data with the Canvas renderer.
   * See https://developer.mozilla.org/en-US/docs/Web/HTML/CORS_enabled_image for more detail.
   * @property {boolean} [interpolate=true] Use interpolated values when resampling.  By default,
   * linear interpolation is used when resampling.  Set to false to use the nearest neighbor instead.
   * @property {import("../tilegrid/WMTS.js").default} tileGrid Tile grid.
   * @property {import("../proj.js").ProjectionLike} [projection] Projection. Default is the view projection.
   * @property {number} [reprojectionErrorThreshold=0.5] Maximum allowed reprojection error (in pixels).
   * Higher values can increase reprojection performance, but decrease precision.
   * @property {RequestEncoding} [requestEncoding='KVP'] Request encoding.
   * @property {string} layer Layer name as advertised in the WMTS capabilities.
   * @property {string} style Style name as advertised in the WMTS capabilities.
   * @property {typeof import("../ImageTile.js").default} [tileClass]  Class used to instantiate image tiles. Default is {@link module:ol/ImageTile~ImageTile}.
   * @property {number} [tilePixelRatio=1] The pixel ratio used by the tile service.
   * For example, if the tile service advertizes 256px by 256px tiles but actually sends 512px
   * by 512px images (for retina/hidpi devices) then `tilePixelRatio`
   * should be set to `2`.
   * @property {string} [format='image/jpeg'] Image format. Only used when `requestEncoding` is `'KVP'`.
   * @property {string} [version='1.0.0'] WMTS version.
   * @property {string} matrixSet Matrix set.
   * @property {!Object} [dimensions] Additional "dimensions" for tile requests.
   * This is an object with properties named like the advertised WMTS dimensions.
   * @property {string} [url]  A URL for the service.
   * For the RESTful request encoding, this is a URL
   * template.  For KVP encoding, it is normal URL. A `{?-?}` template pattern,
   * for example `subdomain{a-f}.domain.com`, may be used instead of defining
   * each one separately in the `urls` option.
   * @property {import("../Tile.js").LoadFunction} [tileLoadFunction] Optional function to load a tile given a URL. The default is
   * ```js
   * function(imageTile, src) {
   *   imageTile.getImage().src = src;
   * };
   * ```
   * @property {Array<string>} [urls] An array of URLs.
   * Requests will be distributed among the URLs in this array.
   * @property {boolean} [wrapX=false] Whether to wrap the world horizontally.
   * @property {number} [transition] Duration of the opacity transition for rendering.
   * To disable the opacity transition, pass `transition: 0`.
   * @property {number|import("../array.js").NearestDirectionFunction} [zDirection=0]
   * Choose whether to use tiles with a higher or lower zoom level when between integer
   * zoom levels. See {@link module:ol/tilegrid/TileGrid~TileGrid#getZForResolution}.
   */

  /**
   * @classdesc
   * Layer source for tile data from WMTS servers.
   * @api
   */
  class WMTS extends TileImage {
    /**
     * @param {Options} options WMTS options.
     */
    constructor(options) {
      // TODO: add support for TileMatrixLimits

      const requestEncoding =
        options.requestEncoding !== undefined ? options.requestEncoding : 'KVP';

      // FIXME: should we create a default tileGrid?
      // we could issue a getCapabilities xhr to retrieve missing configuration
      const tileGrid = options.tileGrid;

      let urls = options.urls;
      if (urls === undefined && options.url !== undefined) {
        urls = expandUrl(options.url);
      }

      super({
        attributions: options.attributions,
        attributionsCollapsible: options.attributionsCollapsible,
        cacheSize: options.cacheSize,
        crossOrigin: options.crossOrigin,
        interpolate: options.interpolate,
        projection: options.projection,
        reprojectionErrorThreshold: options.reprojectionErrorThreshold,
        tileClass: options.tileClass,
        tileGrid: tileGrid,
        tileLoadFunction: options.tileLoadFunction,
        tilePixelRatio: options.tilePixelRatio,
        urls: urls,
        wrapX: options.wrapX !== undefined ? options.wrapX : false,
        transition: options.transition,
        zDirection: options.zDirection,
      });

      /**
       * @private
       * @type {string}
       */
      this.version_ = options.version !== undefined ? options.version : '1.0.0';

      /**
       * @private
       * @type {string}
       */
      this.format_ = options.format !== undefined ? options.format : 'image/jpeg';

      /**
       * @private
       * @type {!Object}
       */
      this.dimensions_ =
        options.dimensions !== undefined ? options.dimensions : {};

      /**
       * @private
       * @type {string}
       */
      this.layer_ = options.layer;

      /**
       * @private
       * @type {string}
       */
      this.matrixSet_ = options.matrixSet;

      /**
       * @private
       * @type {string}
       */
      this.style_ = options.style;

      // FIXME: should we guess this requestEncoding from options.url(s)
      //        structure? that would mean KVP only if a template is not provided.

      /**
       * @private
       * @type {RequestEncoding}
       */
      this.requestEncoding_ = requestEncoding;

      this.setKey(this.getKeyForDimensions_());

      if (urls && urls.length > 0) {
        this.tileUrlFunction = createFromTileUrlFunctions(
          urls.map(this.createFromWMTSTemplate.bind(this))
        );
      }
    }

    /**
     * Set the URLs to use for requests.
     * URLs may contain OGC conform URL Template Variables: {TileMatrix}, {TileRow}, {TileCol}.
     * @param {Array<string>} urls URLs.
     */
    setUrls(urls) {
      this.urls = urls;
      const key = urls.join('\n');
      this.setTileUrlFunction(
        createFromTileUrlFunctions(
          urls.map(this.createFromWMTSTemplate.bind(this))
        ),
        key
      );
    }

    /**
     * Get the dimensions, i.e. those passed to the constructor through the
     * "dimensions" option, and possibly updated using the updateDimensions
     * method.
     * @return {!Object} Dimensions.
     * @api
     */
    getDimensions() {
      return this.dimensions_;
    }

    /**
     * Return the image format of the WMTS source.
     * @return {string} Format.
     * @api
     */
    getFormat() {
      return this.format_;
    }

    /**
     * Return the layer of the WMTS source.
     * @return {string} Layer.
     * @api
     */
    getLayer() {
      return this.layer_;
    }

    /**
     * Return the matrix set of the WMTS source.
     * @return {string} MatrixSet.
     * @api
     */
    getMatrixSet() {
      return this.matrixSet_;
    }

    /**
     * Return the request encoding, either "KVP" or "REST".
     * @return {RequestEncoding} Request encoding.
     * @api
     */
    getRequestEncoding() {
      return this.requestEncoding_;
    }

    /**
     * Return the style of the WMTS source.
     * @return {string} Style.
     * @api
     */
    getStyle() {
      return this.style_;
    }

    /**
     * Return the version of the WMTS source.
     * @return {string} Version.
     * @api
     */
    getVersion() {
      return this.version_;
    }

    /**
     * @private
     * @return {string} The key for the current dimensions.
     */
    getKeyForDimensions_() {
      let i = 0;
      const res = [];
      for (const key in this.dimensions_) {
        res[i++] = key + '-' + this.dimensions_[key];
      }
      return res.join('/');
    }

    /**
     * Update the dimensions.
     * @param {Object} dimensions Dimensions.
     * @api
     */
    updateDimensions(dimensions) {
      Object.assign(this.dimensions_, dimensions);
      this.setKey(this.getKeyForDimensions_());
    }

    /**
     * @param {string} template Template.
     * @return {import("../Tile.js").UrlFunction} Tile URL function.
     */
    createFromWMTSTemplate(template) {
      const requestEncoding = this.requestEncoding_;

      // context property names are lower case to allow for a case insensitive
      // replacement as some services use different naming conventions
      const context = {
        'layer': this.layer_,
        'style': this.style_,
        'tilematrixset': this.matrixSet_,
      };

      if (requestEncoding == 'KVP') {
        Object.assign(context, {
          'Service': 'WMTS',
          'Request': 'GetTile',
          'Version': this.version_,
          'Format': this.format_,
        });
      }

      // TODO: we may want to create our own appendParams function so that params
      // order conforms to wmts spec guidance, and so that we can avoid to escape
      // special template params

      template =
        requestEncoding == 'KVP'
          ? appendParams(template, context)
          : template.replace(/\{(\w+?)\}/g, function (m, p) {
              return p.toLowerCase() in context ? context[p.toLowerCase()] : m;
            });

      const tileGrid = /** @type {import("../tilegrid/WMTS.js").default} */ (
        this.tileGrid
      );
      const dimensions = this.dimensions_;

      return (
        /**
         * @param {import("../tilecoord.js").TileCoord} tileCoord Tile coordinate.
         * @param {number} pixelRatio Pixel ratio.
         * @param {import("../proj/Projection.js").default} projection Projection.
         * @return {string|undefined} Tile URL.
         */
        function (tileCoord, pixelRatio, projection) {
          if (!tileCoord) {
            return undefined;
          } else {
            const localContext = {
              'TileMatrix': tileGrid.getMatrixId(tileCoord[0]),
              'TileCol': tileCoord[1],
              'TileRow': tileCoord[2],
            };
            Object.assign(localContext, dimensions);
            let url = template;
            if (requestEncoding == 'KVP') {
              url = appendParams(url, localContext);
            } else {
              url = url.replace(/\{(\w+?)\}/g, function (m, p) {
                return localContext[p];
              });
            }
            return url;
          }
        }
      );
    }
  }

  /**
   * Generate source options from a capabilities object.
   * @param {Object} wmtsCap An object representing the capabilities document.
   * @param {!Object} config Configuration properties for the layer.  Defaults for
   *                  the layer will apply if not provided.
   *
   * Required config properties:
   *  - layer - {string} The layer identifier.
   *
   * Optional config properties:
   *  - matrixSet - {string} The matrix set identifier, required if there is
   *       more than one matrix set in the layer capabilities.
   *  - projection - {string} The desired CRS when no matrixSet is specified.
   *       eg: "EPSG:3857". If the desired projection is not available,
   *       an error is thrown.
   *  - requestEncoding - {string} url encoding format for the layer. Default is
   *       the first tile url format found in the GetCapabilities response.
   *  - style - {string} The name of the style
   *  - format - {string} Image format for the layer. Default is the first
   *       format returned in the GetCapabilities response.
   *  - crossOrigin - {string|null|undefined} Cross origin. Default is `undefined`.
   * @return {Options|null} WMTS source options object or `null` if the layer was not found.
   * @api
   */
  function optionsFromCapabilities(wmtsCap, config) {
    const layers = wmtsCap['Contents']['Layer'];
    const l = layers.find(function (elt) {
      return elt['Identifier'] == config['layer'];
    });
    if (!l) {
      return null;
    }
    const tileMatrixSets = wmtsCap['Contents']['TileMatrixSet'];
    let idx;
    if (l['TileMatrixSetLink'].length > 1) {
      if ('projection' in config) {
        idx = l['TileMatrixSetLink'].findIndex(function (elt) {
          const tileMatrixSet = tileMatrixSets.find(function (el) {
            return el['Identifier'] == elt['TileMatrixSet'];
          });
          const supportedCRS = tileMatrixSet['SupportedCRS'];
          const proj1 = get$2(supportedCRS);
          const proj2 = get$2(config['projection']);
          if (proj1 && proj2) {
            return equivalent(proj1, proj2);
          } else {
            return supportedCRS == config['projection'];
          }
        });
      } else {
        idx = l['TileMatrixSetLink'].findIndex(function (elt) {
          return elt['TileMatrixSet'] == config['matrixSet'];
        });
      }
    } else {
      idx = 0;
    }
    if (idx < 0) {
      idx = 0;
    }
    const matrixSet =
      /** @type {string} */
      (l['TileMatrixSetLink'][idx]['TileMatrixSet']);
    const matrixLimits =
      /** @type {Array<Object>} */
      (l['TileMatrixSetLink'][idx]['TileMatrixSetLimits']);

    let format = /** @type {string} */ (l['Format'][0]);
    if ('format' in config) {
      format = config['format'];
    }
    idx = l['Style'].findIndex(function (elt) {
      if ('style' in config) {
        return elt['Title'] == config['style'];
      } else {
        return elt['isDefault'];
      }
    });
    if (idx < 0) {
      idx = 0;
    }
    const style = /** @type {string} */ (l['Style'][idx]['Identifier']);

    const dimensions = {};
    if ('Dimension' in l) {
      l['Dimension'].forEach(function (elt, index, array) {
        const key = elt['Identifier'];
        let value = elt['Default'];
        if (value === undefined) {
          value = elt['Value'][0];
        }
        dimensions[key] = value;
      });
    }

    const matrixSets = wmtsCap['Contents']['TileMatrixSet'];
    const matrixSetObj = matrixSets.find(function (elt) {
      return elt['Identifier'] == matrixSet;
    });

    let projection;
    const code = matrixSetObj['SupportedCRS'];
    if (code) {
      projection = get$2(code);
    }
    if ('projection' in config) {
      const projConfig = get$2(config['projection']);
      if (projConfig) {
        if (!projection || equivalent(projConfig, projection)) {
          projection = projConfig;
        }
      }
    }

    let wrapX = false;
    const switchXY = projection.getAxisOrientation().substr(0, 2) == 'ne';

    let matrix = matrixSetObj.TileMatrix[0];

    // create default matrixLimit
    let selectedMatrixLimit = {
      MinTileCol: 0,
      MinTileRow: 0,
      // subtract one to end up at tile top left
      MaxTileCol: matrix.MatrixWidth - 1,
      MaxTileRow: matrix.MatrixHeight - 1,
    };

    //in case of matrix limits, use matrix limits to calculate extent
    if (matrixLimits) {
      selectedMatrixLimit = matrixLimits[matrixLimits.length - 1];
      const m = matrixSetObj.TileMatrix.find(
        (tileMatrixValue) =>
          tileMatrixValue.Identifier === selectedMatrixLimit.TileMatrix ||
          matrixSetObj.Identifier + ':' + tileMatrixValue.Identifier ===
            selectedMatrixLimit.TileMatrix
      );
      if (m) {
        matrix = m;
      }
    }

    const resolution =
      (matrix.ScaleDenominator * 0.00028) / projection.getMetersPerUnit(); // WMTS 1.0.0: standardized rendering pixel size
    const origin = switchXY
      ? [matrix.TopLeftCorner[1], matrix.TopLeftCorner[0]]
      : matrix.TopLeftCorner;
    const tileSpanX = matrix.TileWidth * resolution;
    const tileSpanY = matrix.TileHeight * resolution;
    let matrixSetExtent = matrixSetObj['BoundingBox'];
    if (matrixSetExtent && switchXY) {
      matrixSetExtent = [
        matrixSetExtent[1],
        matrixSetExtent[0],
        matrixSetExtent[3],
        matrixSetExtent[2],
      ];
    }
    let extent = [
      origin[0] + tileSpanX * selectedMatrixLimit.MinTileCol,
      // add one to get proper bottom/right coordinate
      origin[1] - tileSpanY * (1 + selectedMatrixLimit.MaxTileRow),
      origin[0] + tileSpanX * (1 + selectedMatrixLimit.MaxTileCol),
      origin[1] - tileSpanY * selectedMatrixLimit.MinTileRow,
    ];

    if (
      matrixSetExtent !== undefined &&
      !containsExtent(matrixSetExtent, extent)
    ) {
      const wgs84BoundingBox = l['WGS84BoundingBox'];
      const wgs84ProjectionExtent = get$2('EPSG:4326').getExtent();
      extent = matrixSetExtent;
      if (wgs84BoundingBox) {
        wrapX =
          wgs84BoundingBox[0] === wgs84ProjectionExtent[0] &&
          wgs84BoundingBox[2] === wgs84ProjectionExtent[2];
      } else {
        const wgs84MatrixSetExtent = transformExtent(
          matrixSetExtent,
          matrixSetObj['SupportedCRS'],
          'EPSG:4326'
        );
        // Ignore slight deviation from the correct x limits
        wrapX =
          wgs84MatrixSetExtent[0] - 1e-10 <= wgs84ProjectionExtent[0] &&
          wgs84MatrixSetExtent[2] + 1e-10 >= wgs84ProjectionExtent[2];
      }
    }

    const tileGrid = createFromCapabilitiesMatrixSet(
      matrixSetObj,
      extent,
      matrixLimits
    );

    /** @type {!Array<string>} */
    const urls = [];
    let requestEncoding = config['requestEncoding'];
    requestEncoding = requestEncoding !== undefined ? requestEncoding : '';

    if (
      'OperationsMetadata' in wmtsCap &&
      'GetTile' in wmtsCap['OperationsMetadata']
    ) {
      const gets = wmtsCap['OperationsMetadata']['GetTile']['DCP']['HTTP']['Get'];

      for (let i = 0, ii = gets.length; i < ii; ++i) {
        if (gets[i]['Constraint']) {
          const constraint = gets[i]['Constraint'].find(function (element) {
            return element['name'] == 'GetEncoding';
          });
          const encodings = constraint['AllowedValues']['Value'];

          if (requestEncoding === '') {
            // requestEncoding not provided, use the first encoding from the list
            requestEncoding = encodings[0];
          }
          if (requestEncoding === 'KVP') {
            if (encodings.includes('KVP')) {
              urls.push(/** @type {string} */ (gets[i]['href']));
            }
          } else {
            break;
          }
        } else if (gets[i]['href']) {
          requestEncoding = 'KVP';
          urls.push(/** @type {string} */ (gets[i]['href']));
        }
      }
    }
    if (urls.length === 0) {
      requestEncoding = 'REST';
      l['ResourceURL'].forEach(function (element) {
        if (element['resourceType'] === 'tile') {
          format = element['format'];
          urls.push(/** @type {string} */ (element['template']));
        }
      });
    }

    return {
      urls: urls,
      layer: config['layer'],
      matrixSet: matrixSet,
      format: format,
      projection: projection,
      requestEncoding: requestEncoding,
      tileGrid: tileGrid,
      style: style,
      dimensions: dimensions,
      wrapX: wrapX,
      crossOrigin: config['crossOrigin'],
    };
  }

  /**
   * @module ol/source
   */

  /**
   * Creates a sources function from a tile grid. This function can be used as value for the
   * `sources` property of the {@link module:ol/layer/Layer~Layer} subclasses that support it.
   * @param {import("./tilegrid/TileGrid.js").default} tileGrid Tile grid.
   * @param {function(import("./tilecoord.js").TileCoord): import("./source/Source.js").default} factory Source factory.
   * This function takes a {@link module:ol/tilecoord~TileCoord} as argument and is expected to return a
   * {@link module:ol/source/Source~Source}. **Note**: The returned sources should have a tile grid with
   * a limited set of resolutions, matching the resolution range of a single zoom level of the pyramid
   * `tileGrid` that `sourcesFromTileGrid` was called with.
   * @return {function(import("./extent.js").Extent, number): Array<import("./source/Source.js").default>} Sources function.
   * @api
   */
  function sourcesFromTileGrid(tileGrid, factory) {
    const sourceCache = new LRUCache(32);
    const tileGridExtent = tileGrid.getExtent();
    return function (extent, resolution) {
      sourceCache.expireCache();
      if (tileGridExtent) {
        extent = getIntersection(tileGridExtent, extent);
      }
      const z = tileGrid.getZForResolution(resolution);
      const wantedSources = [];
      tileGrid.forEachTileCoord(extent, z, (tileCoord) => {
        const key = tileCoord.toString();
        if (!sourceCache.containsKey(key)) {
          const source = factory(tileCoord);
          sourceCache.set(key, source);
        }
        wantedSources.push(sourceCache.get(key));
      });
      return wantedSources;
    };
  }

  /**
   * @module ol/webgl
   */

  /**
   * Constants taken from goog.webgl
   */

  /**
   * Used by {@link module:ol/webgl/Helper~WebGLHelper} for buffers containing vertices data, such as
   * position, color, texture coordinate, etc. These vertices are then referenced by an index buffer
   * to be drawn on screen (see {@link module:ol/webgl.ELEMENT_ARRAY_BUFFER}).
   * @const
   * @type {number}
   * @api
   */
  const ARRAY_BUFFER = 0x8892;

  /**
   * Used by {@link module:ol/webgl/Helper~WebGLHelper} for buffers containing indices data.
   * Index buffers are essentially lists of references to vertices defined in a vertex buffer
   * (see {@link module:ol/webgl.ARRAY_BUFFER}), and define the primitives (triangles) to be drawn.
   * @const
   * @type {number}
   * @api
   */
  const ELEMENT_ARRAY_BUFFER = 0x8893;

  /**
   * Used by {link module:ol/webgl/Buffer~WebGLArrayBuffer}.
   * @const
   * @type {number}
   * @api
   */
  const STREAM_DRAW = 0x88e0;

  /**
   * Used by {link module:ol/webgl/Buffer~WebGLArrayBuffer}.
   * @const
   * @type {number}
   * @api
   */
  const STATIC_DRAW = 0x88e4;

  /**
   * Used by {link module:ol/webgl/Buffer~WebGLArrayBuffer}.
   * @const
   * @type {number}
   * @api
   */
  const DYNAMIC_DRAW = 0x88e8;

  /**
   * @const
   * @type {number}
   */
  const UNSIGNED_BYTE = 0x1401;

  /**
   * @const
   * @type {number}
   */
  const UNSIGNED_SHORT = 0x1403;

  /**
   * @const
   * @type {number}
   */
  const UNSIGNED_INT = 0x1405;

  /**
   * @const
   * @type {number}
   */
  const FLOAT = 0x1406;

  /** end of goog.webgl constants
   */

  /**
   * @const
   * @type {Array<string>}
   */
  const CONTEXT_IDS = ['experimental-webgl', 'webgl', 'webkit-3d', 'moz-webgl'];

  /**
   * @param {HTMLCanvasElement} canvas Canvas.
   * @param {Object} [attributes] Attributes.
   * @return {WebGLRenderingContext} WebGL rendering context.
   */
  function getContext(canvas, attributes) {
    attributes = Object.assign(
      {
        preserveDrawingBuffer: true,
        antialias: SAFARI_BUG_237906 ? false : true, // https://bugs.webkit.org/show_bug.cgi?id=237906
      },
      attributes
    );
    const ii = CONTEXT_IDS.length;
    for (let i = 0; i < ii; ++i) {
      try {
        const context = canvas.getContext(CONTEXT_IDS[i], attributes);
        if (context) {
          return /** @type {!WebGLRenderingContext} */ (context);
        }
      } catch (e) {
        // pass
      }
    }
    return null;
  }

  /**
   * @type {Array<string>}
   */
  let supportedExtensions;

  /**
   * @return {Array<string>} List of supported WebGL extensions.
   */
  function getSupportedExtensions() {
    if (!supportedExtensions) {
      const canvas = document.createElement('canvas');
      const gl = getContext(canvas);
      if (gl) {
        supportedExtensions = gl.getSupportedExtensions();
      }
    }
    return supportedExtensions;
  }

  /**
   * @module ol/xml
   */

  /**
   * When using {@link module:ol/xml.makeChildAppender} or
   * {@link module:ol/xml.makeSimpleNodeFactory}, the top `objectStack` item needs
   * to have this structure.
   * @typedef {Object} NodeStackItem
   * @property {Element} node Node.
   */

  /**
   * @typedef {function(Element, Array<*>): void} Parser
   */

  /**
   * @typedef {function(Element, *, Array<*>): void} Serializer
   */

  /**
   * @type {string}
   */
  const XML_SCHEMA_INSTANCE_URI =
    'http://www.w3.org/2001/XMLSchema-instance';

  /**
   * @param {string} namespaceURI Namespace URI.
   * @param {string} qualifiedName Qualified name.
   * @return {Element} Node.
   */
  function createElementNS(namespaceURI, qualifiedName) {
    return getDocument().createElementNS(namespaceURI, qualifiedName);
  }

  /**
   * Recursively grab all text content of child nodes into a single string.
   * @param {Node} node Node.
   * @param {boolean} normalizeWhitespace Normalize whitespace: remove all line
   * breaks.
   * @return {string} All text content.
   * @api
   */
  function getAllTextContent(node, normalizeWhitespace) {
    return getAllTextContent_(node, normalizeWhitespace, []).join('');
  }

  /**
   * Recursively grab all text content of child nodes into a single string.
   * @param {Node} node Node.
   * @param {boolean} normalizeWhitespace Normalize whitespace: remove all line
   * breaks.
   * @param {Array<string>} accumulator Accumulator.
   * @private
   * @return {Array<string>} Accumulator.
   */
  function getAllTextContent_(node, normalizeWhitespace, accumulator) {
    if (
      node.nodeType == Node.CDATA_SECTION_NODE ||
      node.nodeType == Node.TEXT_NODE
    ) {
      if (normalizeWhitespace) {
        accumulator.push(String(node.nodeValue).replace(/(\r\n|\r|\n)/g, ''));
      } else {
        accumulator.push(node.nodeValue);
      }
    } else {
      let n;
      for (n = node.firstChild; n; n = n.nextSibling) {
        getAllTextContent_(n, normalizeWhitespace, accumulator);
      }
    }
    return accumulator;
  }

  /**
   * @param {Object} object Object.
   * @return {boolean} Is a document.
   */
  function isDocument(object) {
    return 'documentElement' in object;
  }

  /**
   * @param {Element} node Node.
   * @param {?string} namespaceURI Namespace URI.
   * @param {string} name Attribute name.
   * @return {string} Value
   */
  function getAttributeNS(node, namespaceURI, name) {
    return node.getAttributeNS(namespaceURI, name) || '';
  }

  /**
   * Parse an XML string to an XML Document.
   * @param {string} xml XML.
   * @return {Document} Document.
   * @api
   */
  function parse(xml) {
    return new DOMParser().parseFromString(xml, 'application/xml');
  }

  /**
   * Make an array extender function for extending the array at the top of the
   * object stack.
   * @param {function(this: T, Node, Array<*>): (Array<*>|undefined)} valueReader Value reader.
   * @param {T} [thisArg] The object to use as `this` in `valueReader`.
   * @return {Parser} Parser.
   * @template T
   */
  function makeArrayExtender(valueReader, thisArg) {
    return (
      /**
       * @param {Node} node Node.
       * @param {Array<*>} objectStack Object stack.
       */
      function (node, objectStack) {
        const value = valueReader.call(
          thisArg !== undefined ? thisArg : this,
          node,
          objectStack
        );
        if (value !== undefined) {
          const array = /** @type {Array<*>} */ (
            objectStack[objectStack.length - 1]
          );
          extend$2(array, value);
        }
      }
    );
  }

  /**
   * Make an array pusher function for pushing to the array at the top of the
   * object stack.
   * @param {function(this: T, Element, Array<*>): *} valueReader Value reader.
   * @param {T} [thisArg] The object to use as `this` in `valueReader`.
   * @return {Parser} Parser.
   * @template T
   */
  function makeArrayPusher(valueReader, thisArg) {
    return (
      /**
       * @param {Element} node Node.
       * @param {Array<*>} objectStack Object stack.
       */
      function (node, objectStack) {
        const value = valueReader.call(
          thisArg !== undefined ? thisArg : this,
          node,
          objectStack
        );
        if (value !== undefined) {
          const array = /** @type {Array<*>} */ (
            objectStack[objectStack.length - 1]
          );
          array.push(value);
        }
      }
    );
  }

  /**
   * Make an object stack replacer function for replacing the object at the
   * top of the stack.
   * @param {function(this: T, Node, Array<*>): *} valueReader Value reader.
   * @param {T} [thisArg] The object to use as `this` in `valueReader`.
   * @return {Parser} Parser.
   * @template T
   */
  function makeReplacer(valueReader, thisArg) {
    return (
      /**
       * @param {Node} node Node.
       * @param {Array<*>} objectStack Object stack.
       */
      function (node, objectStack) {
        const value = valueReader.call(
          thisArg !== undefined ? thisArg : this,
          node,
          objectStack
        );
        if (value !== undefined) {
          objectStack[objectStack.length - 1] = value;
        }
      }
    );
  }

  /**
   * Make an object property pusher function for adding a property to the
   * object at the top of the stack.
   * @param {function(this: T, Element, Array<*>): *} valueReader Value reader.
   * @param {string} [property] Property.
   * @param {T} [thisArg] The object to use as `this` in `valueReader`.
   * @return {Parser} Parser.
   * @template T
   */
  function makeObjectPropertyPusher(valueReader, property, thisArg) {
    return (
      /**
       * @param {Element} node Node.
       * @param {Array<*>} objectStack Object stack.
       */
      function (node, objectStack) {
        const value = valueReader.call(
          thisArg !== undefined ? thisArg : this,
          node,
          objectStack
        );
        if (value !== undefined) {
          const object = /** @type {!Object} */ (
            objectStack[objectStack.length - 1]
          );
          const name = property !== undefined ? property : node.localName;
          let array;
          if (name in object) {
            array = object[name];
          } else {
            array = [];
            object[name] = array;
          }
          array.push(value);
        }
      }
    );
  }

  /**
   * Make an object property setter function.
   * @param {function(this: T, Element, Array<*>): *} valueReader Value reader.
   * @param {string} [property] Property.
   * @param {T} [thisArg] The object to use as `this` in `valueReader`.
   * @return {Parser} Parser.
   * @template T
   */
  function makeObjectPropertySetter(valueReader, property, thisArg) {
    return (
      /**
       * @param {Element} node Node.
       * @param {Array<*>} objectStack Object stack.
       */
      function (node, objectStack) {
        const value = valueReader.call(
          thisArg !== undefined ? thisArg : this,
          node,
          objectStack
        );
        if (value !== undefined) {
          const object = /** @type {!Object} */ (
            objectStack[objectStack.length - 1]
          );
          const name = property !== undefined ? property : node.localName;
          object[name] = value;
        }
      }
    );
  }

  /**
   * Create a serializer that appends nodes written by its `nodeWriter` to its
   * designated parent. The parent is the `node` of the
   * {@link module:ol/xml~NodeStackItem} at the top of the `objectStack`.
   * @param {function(this: T, Node, V, Array<*>): void} nodeWriter Node writer.
   * @param {T} [thisArg] The object to use as `this` in `nodeWriter`.
   * @return {Serializer} Serializer.
   * @template T, V
   */
  function makeChildAppender(nodeWriter, thisArg) {
    return function (node, value, objectStack) {
      nodeWriter.call(
        thisArg !== undefined ? thisArg : this,
        node,
        value,
        objectStack
      );
      const parent = /** @type {NodeStackItem} */ (
        objectStack[objectStack.length - 1]
      );
      const parentNode = parent.node;
      parentNode.appendChild(node);
    };
  }

  /**
   * Create a serializer that calls the provided `nodeWriter` from
   * {@link module:ol/xml.serialize}. This can be used by the parent writer to have the
   * `nodeWriter` called with an array of values when the `nodeWriter` was
   * designed to serialize a single item. An example would be a LineString
   * geometry writer, which could be reused for writing MultiLineString
   * geometries.
   * @param {function(this: T, Element, V, Array<*>): void} nodeWriter Node writer.
   * @param {T} [thisArg] The object to use as `this` in `nodeWriter`.
   * @return {Serializer} Serializer.
   * @template T, V
   */
  function makeArraySerializer(nodeWriter, thisArg) {
    let serializersNS, nodeFactory;
    return function (node, value, objectStack) {
      if (serializersNS === undefined) {
        serializersNS = {};
        const serializers = {};
        serializers[node.localName] = nodeWriter;
        serializersNS[node.namespaceURI] = serializers;
        nodeFactory = makeSimpleNodeFactory(node.localName);
      }
      serialize(serializersNS, nodeFactory, value, objectStack);
    };
  }

  /**
   * Create a node factory which can use the `keys` passed to
   * {@link module:ol/xml.serialize} or {@link module:ol/xml.pushSerializeAndPop} as node names,
   * or a fixed node name. The namespace of the created nodes can either be fixed,
   * or the parent namespace will be used.
   * @param {string} [fixedNodeName] Fixed node name which will be used for all
   *     created nodes. If not provided, the 3rd argument to the resulting node
   *     factory needs to be provided and will be the nodeName.
   * @param {string} [fixedNamespaceURI] Fixed namespace URI which will be used for
   *     all created nodes. If not provided, the namespace of the parent node will
   *     be used.
   * @return {function(*, Array<*>, string=): (Node|undefined)} Node factory.
   */
  function makeSimpleNodeFactory(fixedNodeName, fixedNamespaceURI) {
    return (
      /**
       * @param {*} value Value.
       * @param {Array<*>} objectStack Object stack.
       * @param {string} [newNodeName] Node name.
       * @return {Node} Node.
       */
      function (value, objectStack, newNodeName) {
        const context = /** @type {NodeStackItem} */ (
          objectStack[objectStack.length - 1]
        );
        const node = context.node;
        let nodeName = fixedNodeName;
        if (nodeName === undefined) {
          nodeName = newNodeName;
        }

        const namespaceURI =
          fixedNamespaceURI !== undefined ? fixedNamespaceURI : node.namespaceURI;
        return createElementNS(namespaceURI, /** @type {string} */ (nodeName));
      }
    );
  }

  /**
   * A node factory that creates a node using the parent's `namespaceURI` and the
   * `nodeName` passed by {@link module:ol/xml.serialize} or
   * {@link module:ol/xml.pushSerializeAndPop} to the node factory.
   * @const
   * @type {function(*, Array<*>, string=): (Node|undefined)}
   */
  const OBJECT_PROPERTY_NODE_FACTORY = makeSimpleNodeFactory();

  /**
   * Create an array of `values` to be used with {@link module:ol/xml.serialize} or
   * {@link module:ol/xml.pushSerializeAndPop}, where `orderedKeys` has to be provided as
   * `key` argument.
   * @param {Object<string, *>} object Key-value pairs for the sequence. Keys can
   *     be a subset of the `orderedKeys`.
   * @param {Array<string>} orderedKeys Keys in the order of the sequence.
   * @return {Array<*>} Values in the order of the sequence. The resulting array
   *     has the same length as the `orderedKeys` array. Values that are not
   *     present in `object` will be `undefined` in the resulting array.
   */
  function makeSequence(object, orderedKeys) {
    const length = orderedKeys.length;
    const sequence = new Array(length);
    for (let i = 0; i < length; ++i) {
      sequence[i] = object[orderedKeys[i]];
    }
    return sequence;
  }

  /**
   * Create a namespaced structure, using the same values for each namespace.
   * This can be used as a starting point for versioned parsers, when only a few
   * values are version specific.
   * @param {Array<string>} namespaceURIs Namespace URIs.
   * @param {T} structure Structure.
   * @param {Object<string, T>} [structureNS] Namespaced structure to add to.
   * @return {Object<string, T>} Namespaced structure.
   * @template T
   */
  function makeStructureNS(namespaceURIs, structure, structureNS) {
    structureNS = structureNS !== undefined ? structureNS : {};
    let i, ii;
    for (i = 0, ii = namespaceURIs.length; i < ii; ++i) {
      structureNS[namespaceURIs[i]] = structure;
    }
    return structureNS;
  }

  /**
   * Parse a node using the parsers and object stack.
   * @param {Object<string, Object<string, Parser>>} parsersNS
   *     Parsers by namespace.
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @param {*} [thisArg] The object to use as `this`.
   */
  function parseNode(parsersNS, node, objectStack, thisArg) {
    let n;
    for (n = node.firstElementChild; n; n = n.nextElementSibling) {
      const parsers = parsersNS[n.namespaceURI];
      if (parsers !== undefined) {
        const parser = parsers[n.localName];
        if (parser !== undefined) {
          parser.call(thisArg, n, objectStack);
        }
      }
    }
  }

  /**
   * Push an object on top of the stack, parse and return the popped object.
   * @param {T} object Object.
   * @param {Object<string, Object<string, Parser>>} parsersNS
   *     Parsers by namespace.
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @param {*} [thisArg] The object to use as `this`.
   * @return {T} Object.
   * @template T
   */
  function pushParseAndPop(object, parsersNS, node, objectStack, thisArg) {
    objectStack.push(object);
    parseNode(parsersNS, node, objectStack, thisArg);
    return /** @type {T} */ (objectStack.pop());
  }

  /**
   * Walk through an array of `values` and call a serializer for each value.
   * @param {Object<string, Object<string, Serializer>>} serializersNS
   *     Namespaced serializers.
   * @param {function(this: T, *, Array<*>, (string|undefined)): (Node|undefined)} nodeFactory
   *     Node factory. The `nodeFactory` creates the node whose namespace and name
   *     will be used to choose a node writer from `serializersNS`. This
   *     separation allows us to decide what kind of node to create, depending on
   *     the value we want to serialize. An example for this would be different
   *     geometry writers based on the geometry type.
   * @param {Array<*>} values Values to serialize. An example would be an array
   *     of {@link module:ol/Feature~Feature} instances.
   * @param {Array<*>} objectStack Node stack.
   * @param {Array<string>} [keys] Keys of the `values`. Will be passed to the
   *     `nodeFactory`. This is used for serializing object literals where the
   *     node name relates to the property key. The array length of `keys` has
   *     to match the length of `values`. For serializing a sequence, `keys`
   *     determines the order of the sequence.
   * @param {T} [thisArg] The object to use as `this` for the node factory and
   *     serializers.
   * @template T
   */
  function serialize(
    serializersNS,
    nodeFactory,
    values,
    objectStack,
    keys,
    thisArg
  ) {
    const length = (keys !== undefined ? keys : values).length;
    let value, node;
    for (let i = 0; i < length; ++i) {
      value = values[i];
      if (value !== undefined) {
        node = nodeFactory.call(
          thisArg !== undefined ? thisArg : this,
          value,
          objectStack,
          keys !== undefined ? keys[i] : undefined
        );
        if (node !== undefined) {
          serializersNS[node.namespaceURI][node.localName].call(
            thisArg,
            node,
            value,
            objectStack
          );
        }
      }
    }
  }

  /**
   * @param {O} object Object.
   * @param {Object<string, Object<string, Serializer>>} serializersNS
   *     Namespaced serializers.
   * @param {function(this: T, *, Array<*>, (string|undefined)): (Node|undefined)} nodeFactory
   *     Node factory. The `nodeFactory` creates the node whose namespace and name
   *     will be used to choose a node writer from `serializersNS`. This
   *     separation allows us to decide what kind of node to create, depending on
   *     the value we want to serialize. An example for this would be different
   *     geometry writers based on the geometry type.
   * @param {Array<*>} values Values to serialize. An example would be an array
   *     of {@link module:ol/Feature~Feature} instances.
   * @param {Array<*>} objectStack Node stack.
   * @param {Array<string>} [keys] Keys of the `values`. Will be passed to the
   *     `nodeFactory`. This is used for serializing object literals where the
   *     node name relates to the property key. The array length of `keys` has
   *     to match the length of `values`. For serializing a sequence, `keys`
   *     determines the order of the sequence.
   * @param {T} [thisArg] The object to use as `this` for the node factory and
   *     serializers.
   * @return {O|undefined} Object.
   * @template O, T
   */
  function pushSerializeAndPop(
    object,
    serializersNS,
    nodeFactory,
    values,
    objectStack,
    keys,
    thisArg
  ) {
    objectStack.push(object);
    serialize(serializersNS, nodeFactory, values, objectStack, keys, thisArg);
    return /** @type {O|undefined} */ (objectStack.pop());
  }

  let xmlSerializer_ = undefined;

  /**
   * Register a XMLSerializer. Can be used  to inject a XMLSerializer
   * where there is no globally available implementation.
   *
   * @param {XMLSerializer} xmlSerializer A XMLSerializer.
   * @api
   */
  function registerXMLSerializer(xmlSerializer) {
    xmlSerializer_ = xmlSerializer;
  }

  /**
   * @return {XMLSerializer} The XMLSerializer.
   */
  function getXMLSerializer() {
    if (xmlSerializer_ === undefined && typeof XMLSerializer !== 'undefined') {
      xmlSerializer_ = new XMLSerializer();
    }
    return xmlSerializer_;
  }

  let document_ = undefined;

  /**
   * Register a Document to use when creating nodes for XML serializations. Can be used
   * to inject a Document where there is no globally available implementation.
   *
   * @param {Document} document A Document.
   * @api
   */
  function registerDocument(document) {
    document_ = document;
  }

  /**
   * Get a document that should be used when creating nodes for XML serializations.
   * @return {Document} The document.
   */
  function getDocument() {
    if (document_ === undefined && typeof document !== 'undefined') {
      document_ = document.implementation.createDocument('', '', null);
    }
    return document_;
  }

  /**
   * @module ol/webgl/Buffer
   */

  /**
   * Used to describe the intended usage for the data: `STATIC_DRAW`, `STREAM_DRAW`
   * or `DYNAMIC_DRAW`.
   * @enum {number}
   */
  const BufferUsage = {
    STATIC_DRAW: STATIC_DRAW,
    STREAM_DRAW: STREAM_DRAW,
    DYNAMIC_DRAW: DYNAMIC_DRAW,
  };

  /**
   * @classdesc
   * Object used to store an array of data as well as usage information for that data.
   * Stores typed arrays internally, either Float32Array or Uint16/32Array depending on
   * the buffer type (ARRAY_BUFFER or ELEMENT_ARRAY_BUFFER) and available extensions.
   *
   * To populate the array, you can either use:
   * * A size using `#ofSize(buffer)`
   * * An `ArrayBuffer` object using `#fromArrayBuffer(buffer)`
   * * A plain array using `#fromArray(array)`
   *
   * Note:
   * See the documentation of [WebGLRenderingContext.bufferData](https://developer.mozilla.org/en-US/docs/Web/API/WebGLRenderingContext/bufferData)
   * for more info on buffer usage.
   * @api
   */
  class WebGLArrayBuffer {
    /**
     * @param {number} type Buffer type, either ARRAY_BUFFER or ELEMENT_ARRAY_BUFFER.
     * @param {number} [usage] Intended usage, either `STATIC_DRAW`, `STREAM_DRAW` or `DYNAMIC_DRAW`.
     * Default is `DYNAMIC_DRAW`.
     */
    constructor(type, usage) {
      /**
       * @private
       * @type {Float32Array|Uint32Array}
       */
      this.array = null;

      /**
       * @private
       * @type {number}
       */
      this.type = type;

      assert(type === ARRAY_BUFFER || type === ELEMENT_ARRAY_BUFFER, 62);

      /**
       * @private
       * @type {number}
       */
      this.usage = usage !== undefined ? usage : BufferUsage.STATIC_DRAW;
    }

    /**
     * Populates the buffer with an array of the given size (all values will be zeroes).
     * @param {number} size Array size
     */
    ofSize(size) {
      this.array = new (getArrayClassForType(this.type))(size);
    }

    /**
     * Populates the buffer with an array of the given size.
     * @param {Array<number>} array Numerical array
     */
    fromArray(array) {
      this.array = getArrayClassForType(this.type).from(array);
    }

    /**
     * Populates the buffer with a raw binary array buffer.
     * @param {ArrayBuffer} buffer Raw binary buffer to populate the array with. Note that this buffer must have been
     * initialized for the same typed array class.
     */
    fromArrayBuffer(buffer) {
      this.array = new (getArrayClassForType(this.type))(buffer);
    }

    /**
     * @return {number} Buffer type.
     */
    getType() {
      return this.type;
    }

    /**
     * Will return null if the buffer was not initialized
     * @return {Float32Array|Uint32Array} Array.
     */
    getArray() {
      return this.array;
    }

    /**
     * @return {number} Usage.
     */
    getUsage() {
      return this.usage;
    }

    /**
     * Will return 0 if the buffer is not initialized
     * @return {number} Array size
     */
    getSize() {
      return this.array ? this.array.length : 0;
    }
  }

  /**
   * Returns a typed array constructor based on the given buffer type
   * @param {number} type Buffer type, either ARRAY_BUFFER or ELEMENT_ARRAY_BUFFER.
   * @return {Float32ArrayConstructor|Uint32ArrayConstructor} The typed array class to use for this buffer.
   */
  function getArrayClassForType(type) {
    switch (type) {
      case ARRAY_BUFFER:
        return Float32Array;
      case ELEMENT_ARRAY_BUFFER:
        return Uint32Array;
      default:
        return Float32Array;
    }
  }

  /**
   * @module ol/webgl/ContextEventType
   */

  /**
   * @enum {string}
   */
  var ContextEventType = {
    LOST: 'webglcontextlost',
    RESTORED: 'webglcontextrestored',
  };

  /**
   * @module ol/webgl/PostProcessingPass
   */

  const DEFAULT_VERTEX_SHADER = `
  precision mediump float;
  
  attribute vec2 a_position;
  varying vec2 v_texCoord;
  varying vec2 v_screenCoord;
  
  uniform vec2 u_screenSize;
   
  void main() {
    v_texCoord = a_position * 0.5 + 0.5;
    v_screenCoord = v_texCoord * u_screenSize;
    gl_Position = vec4(a_position, 0.0, 1.0);
  }
`;

  const DEFAULT_FRAGMENT_SHADER = `
  precision mediump float;
   
  uniform sampler2D u_image;
  uniform float u_opacity;
   
  varying vec2 v_texCoord;
   
  void main() {
    gl_FragColor = texture2D(u_image, v_texCoord) * u_opacity;
  }
`;

  /**
   * @typedef {Object} Options
   * @property {WebGLRenderingContext} webGlContext WebGL context; mandatory.
   * @property {number} [scaleRatio] Scale ratio; if < 1, the post process will render to a texture smaller than
   * the main canvas that will then be sampled up (useful for saving resource on blur steps).
   * @property {string} [vertexShader] Vertex shader source
   * @property {string} [fragmentShader] Fragment shader source
   * @property {Object<string,import("./Helper").UniformValue>} [uniforms] Uniform definitions for the post process step
   */

  /**
   * @typedef {Object} UniformInternalDescription
   * @property {import("./Helper").UniformValue} value Value
   * @property {number} location Location
   * @property {WebGLTexture} [texture] Texture
   * @private
   */

  /**
   * @classdesc
   * This class is used to define Post Processing passes with custom shaders and uniforms.
   * This is used internally by {@link module:ol/webgl/Helper~WebGLHelper}.
   *
   * Please note that the final output on the DOM canvas is expected to have premultiplied alpha, which means that
   * a pixel which is 100% red with an opacity of 50% must have a color of (r=0.5, g=0, b=0, a=0.5).
   * Failing to provide pixel colors with premultiplied alpha will result in render anomalies.
   *
   * The default post-processing pass does *not* multiply color values with alpha value, it expects color values to be
   * premultiplied.
   *
   * Default shaders are shown hereafter:
   *
   * * Vertex shader:
   *
   *   ```
   *   precision mediump float;
   *
   *   attribute vec2 a_position;
   *   varying vec2 v_texCoord;
   *   varying vec2 v_screenCoord;
   *
   *   uniform vec2 u_screenSize;
   *
   *   void main() {
   *     v_texCoord = a_position * 0.5 + 0.5;
   *     v_screenCoord = v_texCoord * u_screenSize;
   *     gl_Position = vec4(a_position, 0.0, 1.0);
   *   }
   *   ```
   *
   * * Fragment shader:
   *
   *   ```
   *   precision mediump float;
   *
   *   uniform sampler2D u_image;
   *   uniform float u_opacity;
   *
   *   varying vec2 v_texCoord;
   *
   *   void main() {
   *     gl_FragColor = texture2D(u_image, v_texCoord) * u_opacity;
   *   }
   *   ```
   *
   * @api
   */
  class WebGLPostProcessingPass {
    /**
     * @param {Options} options Options.
     */
    constructor(options) {
      this.gl_ = options.webGlContext;
      const gl = this.gl_;

      this.scaleRatio_ = options.scaleRatio || 1;

      this.renderTargetTexture_ = gl.createTexture();
      this.renderTargetTextureSize_ = null;

      this.frameBuffer_ = gl.createFramebuffer();

      // compile the program for the frame buffer
      // TODO: make compilation errors show up
      const vertexShader = gl.createShader(gl.VERTEX_SHADER);
      gl.shaderSource(
        vertexShader,
        options.vertexShader || DEFAULT_VERTEX_SHADER
      );
      gl.compileShader(vertexShader);
      const fragmentShader = gl.createShader(gl.FRAGMENT_SHADER);
      gl.shaderSource(
        fragmentShader,
        options.fragmentShader || DEFAULT_FRAGMENT_SHADER
      );
      gl.compileShader(fragmentShader);
      this.renderTargetProgram_ = gl.createProgram();
      gl.attachShader(this.renderTargetProgram_, vertexShader);
      gl.attachShader(this.renderTargetProgram_, fragmentShader);
      gl.linkProgram(this.renderTargetProgram_);

      // bind the vertices buffer for the frame buffer
      this.renderTargetVerticesBuffer_ = gl.createBuffer();
      const verticesArray = [-1, -1, 1, -1, -1, 1, 1, -1, 1, 1, -1, 1];
      gl.bindBuffer(gl.ARRAY_BUFFER, this.renderTargetVerticesBuffer_);
      gl.bufferData(
        gl.ARRAY_BUFFER,
        new Float32Array(verticesArray),
        gl.STATIC_DRAW
      );

      this.renderTargetAttribLocation_ = gl.getAttribLocation(
        this.renderTargetProgram_,
        'a_position'
      );
      this.renderTargetUniformLocation_ = gl.getUniformLocation(
        this.renderTargetProgram_,
        'u_screenSize'
      );
      this.renderTargetOpacityLocation_ = gl.getUniformLocation(
        this.renderTargetProgram_,
        'u_opacity'
      );
      this.renderTargetTextureLocation_ = gl.getUniformLocation(
        this.renderTargetProgram_,
        'u_image'
      );

      /**
       * Holds info about custom uniforms used in the post processing pass
       * @type {Array<UniformInternalDescription>}
       * @private
       */
      this.uniforms_ = [];
      options.uniforms &&
        Object.keys(options.uniforms).forEach(
          function (name) {
            this.uniforms_.push({
              value: options.uniforms[name],
              location: gl.getUniformLocation(this.renderTargetProgram_, name),
            });
          }.bind(this)
        );
    }

    /**
     * Get the WebGL rendering context
     * @return {WebGLRenderingContext} The rendering context.
     * @api
     */
    getGL() {
      return this.gl_;
    }

    /**
     * Initialize the render target texture of the post process, make sure it is at the
     * right size and bind it as a render target for the next draw calls.
     * The last step to be initialized will be the one where the primitives are rendered.
     * @param {import("../Map.js").FrameState} frameState current frame state
     * @api
     */
    init(frameState) {
      const gl = this.getGL();
      const textureSize = [
        gl.drawingBufferWidth * this.scaleRatio_,
        gl.drawingBufferHeight * this.scaleRatio_,
      ];

      // rendering goes to my buffer
      gl.bindFramebuffer(gl.FRAMEBUFFER, this.getFrameBuffer());
      gl.viewport(0, 0, textureSize[0], textureSize[1]);

      // if size has changed: adjust canvas & render target texture
      if (
        !this.renderTargetTextureSize_ ||
        this.renderTargetTextureSize_[0] !== textureSize[0] ||
        this.renderTargetTextureSize_[1] !== textureSize[1]
      ) {
        this.renderTargetTextureSize_ = textureSize;

        // create a new texture
        const level = 0;
        const internalFormat = gl.RGBA;
        const border = 0;
        const format = gl.RGBA;
        const type = gl.UNSIGNED_BYTE;
        const data = null;
        gl.bindTexture(gl.TEXTURE_2D, this.renderTargetTexture_);
        gl.texImage2D(
          gl.TEXTURE_2D,
          level,
          internalFormat,
          textureSize[0],
          textureSize[1],
          border,
          format,
          type,
          data
        );

        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.CLAMP_TO_EDGE);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.CLAMP_TO_EDGE);

        // bind the texture to the framebuffer
        gl.framebufferTexture2D(
          gl.FRAMEBUFFER,
          gl.COLOR_ATTACHMENT0,
          gl.TEXTURE_2D,
          this.renderTargetTexture_,
          0
        );
      }
    }

    /**
     * Render to the next postprocessing pass (or to the canvas if final pass).
     * @param {import("../Map.js").FrameState} frameState current frame state
     * @param {WebGLPostProcessingPass} [nextPass] Next pass, optional
     * @param {function(WebGLRenderingContext, import("../Map.js").FrameState):void} [preCompose] Called before composing.
     * @param {function(WebGLRenderingContext, import("../Map.js").FrameState):void} [postCompose] Called before composing.
     * @api
     */
    apply(frameState, nextPass, preCompose, postCompose) {
      const gl = this.getGL();
      const size = frameState.size;

      gl.bindFramebuffer(
        gl.FRAMEBUFFER,
        nextPass ? nextPass.getFrameBuffer() : null
      );
      gl.activeTexture(gl.TEXTURE0);
      gl.bindTexture(gl.TEXTURE_2D, this.renderTargetTexture_);

      if (!nextPass) {
        // clear the canvas if we are the first to render to it
        // and preserveDrawingBuffer is true
        const canvasId = getUid(gl.canvas);
        if (!frameState.renderTargets[canvasId]) {
          const attributes = gl.getContextAttributes();
          if (attributes && attributes.preserveDrawingBuffer) {
            gl.clearColor(0.0, 0.0, 0.0, 0.0);
            gl.clear(gl.COLOR_BUFFER_BIT);
          }

          frameState.renderTargets[canvasId] = true;
        }
      }

      gl.enable(gl.BLEND);
      gl.blendFunc(gl.ONE, gl.ONE_MINUS_SRC_ALPHA);
      gl.viewport(0, 0, gl.drawingBufferWidth, gl.drawingBufferHeight);

      gl.bindBuffer(gl.ARRAY_BUFFER, this.renderTargetVerticesBuffer_);

      gl.useProgram(this.renderTargetProgram_);
      gl.enableVertexAttribArray(this.renderTargetAttribLocation_);
      gl.vertexAttribPointer(
        this.renderTargetAttribLocation_,
        2,
        gl.FLOAT,
        false,
        0,
        0
      );
      gl.uniform2f(this.renderTargetUniformLocation_, size[0], size[1]);
      gl.uniform1i(this.renderTargetTextureLocation_, 0);

      const opacity = frameState.layerStatesArray[frameState.layerIndex].opacity;
      gl.uniform1f(this.renderTargetOpacityLocation_, opacity);

      this.applyUniforms(frameState);

      if (preCompose) {
        preCompose(gl, frameState);
      }
      gl.drawArrays(gl.TRIANGLES, 0, 6);
      if (postCompose) {
        postCompose(gl, frameState);
      }
    }

    /**
     * @return {WebGLFramebuffer} Frame buffer
     * @api
     */
    getFrameBuffer() {
      return this.frameBuffer_;
    }

    /**
     * Sets the custom uniforms based on what was given in the constructor.
     * @param {import("../Map.js").FrameState} frameState Frame state.
     * @private
     */
    applyUniforms(frameState) {
      const gl = this.getGL();

      let value;
      let textureSlot = 1;
      this.uniforms_.forEach(function (uniform) {
        value =
          typeof uniform.value === 'function'
            ? uniform.value(frameState)
            : uniform.value;

        // apply value based on type
        if (value instanceof HTMLCanvasElement || value instanceof ImageData) {
          // create a texture & put data
          if (!uniform.texture) {
            uniform.texture = gl.createTexture();
          }
          gl.activeTexture(gl[`TEXTURE${textureSlot}`]);
          gl.bindTexture(gl.TEXTURE_2D, uniform.texture);
          gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR);
          gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.CLAMP_TO_EDGE);
          gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.CLAMP_TO_EDGE);

          if (value instanceof ImageData) {
            gl.texImage2D(
              gl.TEXTURE_2D,
              0,
              gl.RGBA,
              gl.RGBA,
              value.width,
              value.height,
              0,
              gl.UNSIGNED_BYTE,
              new Uint8Array(value.data)
            );
          } else {
            gl.texImage2D(
              gl.TEXTURE_2D,
              0,
              gl.RGBA,
              gl.RGBA,
              gl.UNSIGNED_BYTE,
              value
            );
          }

          // fill texture slots
          gl.uniform1i(uniform.location, textureSlot++);
        } else if (Array.isArray(value)) {
          switch (value.length) {
            case 2:
              gl.uniform2f(uniform.location, value[0], value[1]);
              return;
            case 3:
              gl.uniform3f(uniform.location, value[0], value[1], value[2]);
              return;
            case 4:
              gl.uniform4f(
                uniform.location,
                value[0],
                value[1],
                value[2],
                value[3]
              );
              return;
            default:
              return;
          }
        } else if (typeof value === 'number') {
          gl.uniform1f(uniform.location, value);
        }
      });
    }
  }

  /**
   * @module ol/vec/mat4
   */

  /**
   * @return {Array<number>} "4x4 matrix representing a 3D identity transform."
   */
  function create$1() {
    return [1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 0, 1];
  }

  /**
   * @param {Array<number>} mat4 Flattened 4x4 matrix receiving the result.
   * @param {import("../transform.js").Transform} transform Transformation matrix.
   * @return {Array<number>} "2D transformation matrix as flattened 4x4 matrix."
   */
  function fromTransform(mat4, transform) {
    mat4[0] = transform[0];
    mat4[1] = transform[1];
    mat4[4] = transform[2];
    mat4[5] = transform[3];
    mat4[12] = transform[4];
    mat4[13] = transform[5];
    return mat4;
  }

  /**
   * @module ol/webgl/Helper
   */

  /**
   * Names of uniforms made available to all shaders.
   * Please note: changing these *will* break custom shaders!
   * @enum {string}
   */
  const DefaultUniform = {
    PROJECTION_MATRIX: 'u_projectionMatrix',
    OFFSET_SCALE_MATRIX: 'u_offsetScaleMatrix',
    OFFSET_ROTATION_MATRIX: 'u_offsetRotateMatrix',
    TIME: 'u_time',
    ZOOM: 'u_zoom',
    RESOLUTION: 'u_resolution',
    SIZE_PX: 'u_sizePx',
    PIXEL_RATIO: 'u_pixelRatio',
  };

  /**
   * Attribute types, either `UNSIGNED_BYTE`, `UNSIGNED_SHORT`, `UNSIGNED_INT` or `FLOAT`
   * Note: an attribute stored in a `Float32Array` should be of type `FLOAT`.
   * @enum {number}
   */
  const AttributeType = {
    UNSIGNED_BYTE: UNSIGNED_BYTE,
    UNSIGNED_SHORT: UNSIGNED_SHORT,
    UNSIGNED_INT: UNSIGNED_INT,
    FLOAT: FLOAT,
  };

  /**
   * Description of an attribute in a buffer
   * @typedef {Object} AttributeDescription
   * @property {string} name Attribute name to use in shaders
   * @property {number} size Number of components per attributes
   * @property {AttributeType} [type] Attribute type, i.e. number of bytes used to store the value. This is
   * determined by the class of typed array which the buffer uses (eg. `Float32Array` for a `FLOAT` attribute).
   * Default is `FLOAT`.
   */

  /**
   * @typedef {number|Array<number>|HTMLCanvasElement|HTMLImageElement|ImageData|import("../transform").Transform} UniformLiteralValue
   */

  /**
   * Uniform value can be a number, array of numbers (2 to 4), canvas element or a callback returning
   * one of the previous types.
   * @typedef {UniformLiteralValue|function(import("../Map.js").FrameState):UniformLiteralValue} UniformValue
   */

  /**
   * @typedef {Object} PostProcessesOptions
   * @property {number} [scaleRatio] Scale ratio; if < 1, the post process will render to a texture smaller than
   * the main canvas which will then be sampled up (useful for saving resource on blur steps).
   * @property {string} [vertexShader] Vertex shader source
   * @property {string} [fragmentShader] Fragment shader source
   * @property {Object<string,UniformValue>} [uniforms] Uniform definitions for the post process step
   */

  /**
   * @typedef {Object} Options
   * @property {Object<string,UniformValue>} [uniforms] Uniform definitions; property names must match the uniform
   * names in the provided or default shaders.
   * @property {Array<PostProcessesOptions>} [postProcesses] Post-processes definitions
   * @property {string} [canvasCacheKey] The cache key for the canvas.
   */

  /**
   * @typedef {Object} UniformInternalDescription
   * @property {string} name Name
   * @property {UniformValue} [value] Value
   * @property {WebGLTexture} [texture] Texture
   * @private
   */

  /**
   * @typedef {Object} CanvasCacheItem
   * @property {HTMLCanvasElement} canvas Canvas element.
   * @property {number} users The count of users of this canvas.
   */

  /**
   * @type {Object<string,CanvasCacheItem>}
   */
  const canvasCache = {};

  /**
   * @param {string} key The cache key for the canvas.
   * @return {string} The shared cache key.
   */
  function getSharedCanvasCacheKey(key) {
    return 'shared/' + key;
  }

  let uniqueCanvasCacheKeyCount = 0;

  /**
   * @return {string} The unique cache key.
   */
  function getUniqueCanvasCacheKey() {
    const key = 'unique/' + uniqueCanvasCacheKeyCount;
    uniqueCanvasCacheKeyCount += 1;
    return key;
  }

  /**
   * @param {string} key The cache key for the canvas.
   * @return {HTMLCanvasElement} The canvas.
   */
  function getCanvas(key) {
    let cacheItem = canvasCache[key];
    if (!cacheItem) {
      const canvas = document.createElement('canvas');
      canvas.style.position = 'absolute';
      canvas.style.left = '0';
      cacheItem = {users: 0, canvas};
      canvasCache[key] = cacheItem;
    }

    cacheItem.users += 1;
    return cacheItem.canvas;
  }

  /**
   * @param {string} key The cache key for the canvas.
   */
  function releaseCanvas(key) {
    const cacheItem = canvasCache[key];
    if (!cacheItem) {
      return;
    }

    cacheItem.users -= 1;
    if (cacheItem.users > 0) {
      return;
    }

    const canvas = cacheItem.canvas;
    const gl = getContext(canvas);
    const extension = gl.getExtension('WEBGL_lose_context');
    if (extension) {
      extension.loseContext();
    }

    delete canvasCache[key];
  }

  /**
   * @classdesc
   * This class is intended to provide low-level functions related to WebGL rendering, so that accessing
   * directly the WebGL API should not be required anymore.
   *
   * Several operations are handled by the `WebGLHelper` class:
   *
   * ### Define custom shaders and uniforms
   *
   *   *Shaders* are low-level programs executed on the GPU and written in GLSL. There are two types of shaders:
   *
   *   Vertex shaders are used to manipulate the position and attribute of *vertices* of rendered primitives (ie. corners of a square).
   *   Outputs are:
   *
   *   * `gl_Position`: position of the vertex in screen space
   *
   *   * Varyings usually prefixed with `v_` are passed on to the fragment shader
   *
   *   Fragment shaders are used to control the actual color of the pixels drawn on screen. Their only output is `gl_FragColor`.
   *
   *   Both shaders can take *uniforms* or *attributes* as input. Attributes are explained later. Uniforms are common, read-only values that
   *   can be changed at every frame and can be of type float, arrays of float or images.
   *
   *   Shaders must be compiled and assembled into a program like so:
   *   ```js
   *   // here we simply create two shaders and assemble them in a program which is then used
   *   // for subsequent rendering calls; note how a frameState is required to set up a program,
   *   // as several default uniforms are computed from it (projection matrix, zoom level, etc.)
   *   const vertexShader = new WebGLVertex(VERTEX_SHADER);
   *   const fragmentShader = new WebGLFragment(FRAGMENT_SHADER);
   *   const program = this.context.getProgram(fragmentShader, vertexShader);
   *   helper.useProgram(this.program, frameState);
   *   ```
   *
   *   Uniforms are defined using the `uniforms` option and can either be explicit values or callbacks taking the frame state as argument.
   *   You can also change their value along the way like so:
   *   ```js
   *   helper.setUniformFloatValue('u_value', valueAsNumber);
   *   ```
   *
   * ### Defining post processing passes
   *
   *   *Post processing* describes the act of rendering primitives to a texture, and then rendering this texture to the final canvas
   *   while applying special effects in screen space.
   *   Typical uses are: blurring, color manipulation, depth of field, filtering...
   *
   *   The `WebGLHelper` class offers the possibility to define post processes at creation time using the `postProcesses` option.
   *   A post process step accepts the following options:
   *
   *   * `fragmentShader` and `vertexShader`: text literals in GLSL language that will be compiled and used in the post processing step.
   *   * `uniforms`: uniforms can be defined for the post processing steps just like for the main render.
   *   * `scaleRatio`: allows using an intermediate texture smaller or higher than the final canvas in the post processing step.
   *     This is typically used in blur steps to reduce the performance overhead by using an already downsampled texture as input.
   *
   *   The {@link module:ol/webgl/PostProcessingPass~WebGLPostProcessingPass} class is used internally, refer to its documentation for more info.
   *
   * ### Binding WebGL buffers and flushing data into them
   *
   *   Data that must be passed to the GPU has to be transferred using {@link module:ol/webgl/Buffer~WebGLArrayBuffer} objects.
   *   A buffer has to be created only once, but must be bound every time the buffer content will be used for rendering.
   *   This is done using {@link bindBuffer}.
   *   When the buffer's array content has changed, the new data has to be flushed to the GPU memory; this is done using
   *   {@link flushBufferData}. Note: this operation is expensive and should be done as infrequently as possible.
   *
   *   When binding an array buffer, a `target` parameter must be given: it should be either {@link module:ol/webgl.ARRAY_BUFFER}
   *   (if the buffer contains vertices data) or {@link module:ol/webgl.ELEMENT_ARRAY_BUFFER} (if the buffer contains indices data).
   *
   *   Examples below:
   *   ```js
   *   // at initialization phase
   *   const verticesBuffer = new WebGLArrayBuffer([], DYNAMIC_DRAW);
   *   const indicesBuffer = new WebGLArrayBuffer([], DYNAMIC_DRAW);
   *
   *   // when array values have changed
   *   helper.flushBufferData(ARRAY_BUFFER, this.verticesBuffer);
   *   helper.flushBufferData(ELEMENT_ARRAY_BUFFER, this.indicesBuffer);
   *
   *   // at rendering phase
   *   helper.bindBuffer(ARRAY_BUFFER, this.verticesBuffer);
   *   helper.bindBuffer(ELEMENT_ARRAY_BUFFER, this.indicesBuffer);
   *   ```
   *
   * ### Specifying attributes
   *
   *   The GPU only receives the data as arrays of numbers. These numbers must be handled differently depending on what it describes (position, texture coordinate...).
   *   Attributes are used to specify these uses. Specify the attribute names with
   *   {@link module:ol/webgl/Helper~WebGLHelper#enableAttributes} (see code snippet below).
   *
   *   Please note that you will have to specify the type and offset of the attributes in the data array. You can refer to the documentation of [WebGLRenderingContext.vertexAttribPointer](https://developer.mozilla.org/en-US/docs/Web/API/WebGLRenderingContext/vertexAttribPointer) for more explanation.
   *   ```js
   *   // here we indicate that the data array has the following structure:
   *   // [posX, posY, offsetX, offsetY, texCoordU, texCoordV, posX, posY, ...]
   *   helper.enableAttributes([
   *     {
   *        name: 'a_position',
   *        size: 2
   *     },
   *     {
   *       name: 'a_offset',
   *       size: 2
   *     },
   *     {
   *       name: 'a_texCoord',
   *       size: 2
   *     }
   *   ])
   *   ```
   *
   * ### Rendering primitives
   *
   *   Once all the steps above have been achieved, rendering primitives to the screen is done using {@link prepareDraw}, {@link drawElements} and {@link finalizeDraw}.
   *   ```js
   *   // frame preparation step
   *   helper.prepareDraw(frameState);
   *
   *   // call this for every data array that has to be rendered on screen
   *   helper.drawElements(0, this.indicesBuffer.getArray().length);
   *
   *   // finalize the rendering by applying post processes
   *   helper.finalizeDraw(frameState);
   *   ```
   *
   * For an example usage of this class, refer to {@link module:ol/renderer/webgl/PointsLayer~WebGLPointsLayerRenderer}.
   */
  class WebGLHelper extends Disposable {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      super();
      options = options || {};

      /** @private */
      this.boundHandleWebGLContextLost_ = this.handleWebGLContextLost.bind(this);

      /** @private */
      this.boundHandleWebGLContextRestored_ =
        this.handleWebGLContextRestored.bind(this);

      /**
       * @private
       * @type {string}
       */
      this.canvasCacheKey_ = options.canvasCacheKey
        ? getSharedCanvasCacheKey(options.canvasCacheKey)
        : getUniqueCanvasCacheKey();

      /**
       * @private
       * @type {HTMLCanvasElement}
       */
      this.canvas_ = getCanvas(this.canvasCacheKey_);

      /**
       * @private
       * @type {WebGLRenderingContext}
       */
      this.gl_ = getContext(this.canvas_);

      /**
       * @private
       * @type {!Object<string, BufferCacheEntry>}
       */
      this.bufferCache_ = {};

      /**
       * @private
       * @type {Object<string, Object>}
       */
      this.extensionCache_ = {};

      /**
       * @private
       * @type {WebGLProgram}
       */
      this.currentProgram_ = null;

      this.canvas_.addEventListener(
        ContextEventType.LOST,
        this.boundHandleWebGLContextLost_
      );
      this.canvas_.addEventListener(
        ContextEventType.RESTORED,
        this.boundHandleWebGLContextRestored_
      );

      /**
       * @private
       * @type {import("../transform.js").Transform}
       */
      this.offsetRotateMatrix_ = create$2();

      /**
       * @private
       * @type {import("../transform.js").Transform}
       */
      this.offsetScaleMatrix_ = create$2();

      /**
       * @private
       * @type {Array<number>}
       */
      this.tmpMat4_ = create$1();

      /**
       * @private
       * @type {Object<string, WebGLUniformLocation>}
       */
      this.uniformLocations_ = {};

      /**
       * @private
       * @type {Object<string, number>}
       */
      this.attribLocations_ = {};

      /**
       * Holds info about custom uniforms used in the post processing pass.
       * If the uniform is a texture, the WebGL Texture object will be stored here.
       * @type {Array<UniformInternalDescription>}
       * @private
       */
      this.uniforms_ = [];
      if (options.uniforms) {
        this.setUniforms(options.uniforms);
      }

      const gl = this.getGL();

      /**
       * An array of PostProcessingPass objects is kept in this variable, built from the steps provided in the
       * options. If no post process was given, a default one is used (so as not to have to make an exception to
       * the frame buffer logic).
       * @type {Array<WebGLPostProcessingPass>}
       * @private
       */
      this.postProcessPasses_ = options.postProcesses
        ? options.postProcesses.map(function (options) {
            return new WebGLPostProcessingPass({
              webGlContext: gl,
              scaleRatio: options.scaleRatio,
              vertexShader: options.vertexShader,
              fragmentShader: options.fragmentShader,
              uniforms: options.uniforms,
            });
          })
        : [new WebGLPostProcessingPass({webGlContext: gl})];

      /**
       * @type {string|null}
       * @private
       */
      this.shaderCompileErrors_ = null;

      /**
       * @type {number}
       * @private
       */
      this.startTime_ = Date.now();
    }

    /**
     * @param {Object<string, UniformValue>} uniforms Uniform definitions.
     */
    setUniforms(uniforms) {
      this.uniforms_ = [];
      for (const name in uniforms) {
        this.uniforms_.push({
          name: name,
          value: uniforms[name],
        });
      }
      this.uniformLocations_ = {};
    }

    /**
     * @param {string} canvasCacheKey The canvas cache key.
     * @return {boolean} The provided key matches the one this helper was constructed with.
     */
    canvasCacheKeyMatches(canvasCacheKey) {
      return this.canvasCacheKey_ === getSharedCanvasCacheKey(canvasCacheKey);
    }

    /**
     * Get a WebGL extension.  If the extension is not supported, null is returned.
     * Extensions are cached after they are enabled for the first time.
     * @param {string} name The extension name.
     * @return {Object|null} The extension or null if not supported.
     */
    getExtension(name) {
      if (name in this.extensionCache_) {
        return this.extensionCache_[name];
      }
      const extension = this.gl_.getExtension(name);
      this.extensionCache_[name] = extension;
      return extension;
    }

    /**
     * Just bind the buffer if it's in the cache. Otherwise create
     * the WebGL buffer, bind it, populate it, and add an entry to
     * the cache.
     * @param {import("./Buffer").default} buffer Buffer.
     */
    bindBuffer(buffer) {
      const gl = this.getGL();
      const bufferKey = getUid(buffer);
      let bufferCache = this.bufferCache_[bufferKey];
      if (!bufferCache) {
        const webGlBuffer = gl.createBuffer();
        bufferCache = {
          buffer: buffer,
          webGlBuffer: webGlBuffer,
        };
        this.bufferCache_[bufferKey] = bufferCache;
      }
      gl.bindBuffer(buffer.getType(), bufferCache.webGlBuffer);
    }

    /**
     * Update the data contained in the buffer array; this is required for the
     * new data to be rendered
     * @param {import("./Buffer").default} buffer Buffer.
     */
    flushBufferData(buffer) {
      const gl = this.getGL();
      this.bindBuffer(buffer);
      gl.bufferData(buffer.getType(), buffer.getArray(), buffer.getUsage());
    }

    /**
     * @param {import("./Buffer.js").default} buf Buffer.
     */
    deleteBuffer(buf) {
      const gl = this.getGL();
      const bufferKey = getUid(buf);
      const bufferCacheEntry = this.bufferCache_[bufferKey];
      if (bufferCacheEntry && !gl.isContextLost()) {
        gl.deleteBuffer(bufferCacheEntry.webGlBuffer);
      }
      delete this.bufferCache_[bufferKey];
    }

    /**
     * Clean up.
     */
    disposeInternal() {
      this.canvas_.removeEventListener(
        ContextEventType.LOST,
        this.boundHandleWebGLContextLost_
      );
      this.canvas_.removeEventListener(
        ContextEventType.RESTORED,
        this.boundHandleWebGLContextRestored_
      );

      releaseCanvas(this.canvasCacheKey_);

      delete this.gl_;
      delete this.canvas_;
    }

    /**
     * Clear the buffer & set the viewport to draw.
     * Post process passes will be initialized here, the first one being bound as a render target for
     * subsequent draw calls.
     * @param {import("../Map.js").FrameState} frameState current frame state
     * @param {boolean} [disableAlphaBlend] If true, no alpha blending will happen.
     */
    prepareDraw(frameState, disableAlphaBlend) {
      const gl = this.getGL();
      const canvas = this.getCanvas();
      const size = frameState.size;
      const pixelRatio = frameState.pixelRatio;

      canvas.width = size[0] * pixelRatio;
      canvas.height = size[1] * pixelRatio;
      canvas.style.width = size[0] + 'px';
      canvas.style.height = size[1] + 'px';

      // loop backwards in post processes list
      for (let i = this.postProcessPasses_.length - 1; i >= 0; i--) {
        this.postProcessPasses_[i].init(frameState);
      }

      gl.bindTexture(gl.TEXTURE_2D, null);

      gl.clearColor(0.0, 0.0, 0.0, 0.0);
      gl.clear(gl.COLOR_BUFFER_BIT);

      gl.enable(gl.BLEND);
      gl.blendFunc(gl.ONE, disableAlphaBlend ? gl.ZERO : gl.ONE_MINUS_SRC_ALPHA);
    }

    /**
     * Clear the render target & bind it for future draw operations.
     * This is similar to `prepareDraw`, only post processes will not be applied.
     * Note: the whole viewport will be drawn to the render target, regardless of its size.
     * @param {import("../Map.js").FrameState} frameState current frame state
     * @param {import("./RenderTarget.js").default} renderTarget Render target to draw to
     * @param {boolean} [disableAlphaBlend] If true, no alpha blending will happen.
     */
    prepareDrawToRenderTarget(frameState, renderTarget, disableAlphaBlend) {
      const gl = this.getGL();
      const size = renderTarget.getSize();

      gl.bindFramebuffer(gl.FRAMEBUFFER, renderTarget.getFramebuffer());
      gl.viewport(0, 0, size[0], size[1]);
      gl.bindTexture(gl.TEXTURE_2D, renderTarget.getTexture());
      gl.clearColor(0.0, 0.0, 0.0, 0.0);
      gl.clear(gl.COLOR_BUFFER_BIT);
      gl.enable(gl.BLEND);
      gl.blendFunc(gl.ONE, disableAlphaBlend ? gl.ZERO : gl.ONE_MINUS_SRC_ALPHA);
    }

    /**
     * Execute a draw call based on the currently bound program, texture, buffers, attributes.
     * @param {number} start Start index.
     * @param {number} end End index.
     */
    drawElements(start, end) {
      const gl = this.getGL();
      this.getExtension('OES_element_index_uint');

      const elementType = gl.UNSIGNED_INT;
      const elementSize = 4;

      const numItems = end - start;
      const offsetInBytes = start * elementSize;
      gl.drawElements(gl.TRIANGLES, numItems, elementType, offsetInBytes);
    }

    /**
     * Apply the successive post process passes which will eventually render to the actual canvas.
     * @param {import("../Map.js").FrameState} frameState current frame state
     * @param {function(WebGLRenderingContext, import("../Map.js").FrameState):void} [preCompose] Called before composing.
     * @param {function(WebGLRenderingContext, import("../Map.js").FrameState):void} [postCompose] Called before composing.
     */
    finalizeDraw(frameState, preCompose, postCompose) {
      // apply post processes using the next one as target
      for (let i = 0, ii = this.postProcessPasses_.length; i < ii; i++) {
        if (i === ii - 1) {
          this.postProcessPasses_[i].apply(
            frameState,
            null,
            preCompose,
            postCompose
          );
        } else {
          this.postProcessPasses_[i].apply(
            frameState,
            this.postProcessPasses_[i + 1]
          );
        }
      }
    }

    /**
     * @return {HTMLCanvasElement} Canvas.
     */
    getCanvas() {
      return this.canvas_;
    }

    /**
     * Get the WebGL rendering context
     * @return {WebGLRenderingContext} The rendering context.
     */
    getGL() {
      return this.gl_;
    }

    /**
     * Sets the default matrix uniforms for a given frame state. This is called internally in `prepareDraw`.
     * @param {import("../Map.js").FrameState} frameState Frame state.
     */
    applyFrameState(frameState) {
      const size = frameState.size;
      const rotation = frameState.viewState.rotation;
      const pixelRatio = frameState.pixelRatio;

      const offsetScaleMatrix = reset(this.offsetScaleMatrix_);
      scale$3(offsetScaleMatrix, 2 / size[0], 2 / size[1]);

      const offsetRotateMatrix = reset(this.offsetRotateMatrix_);
      if (rotation !== 0) {
        rotate$2(offsetRotateMatrix, -rotation);
      }

      this.setUniformMatrixValue(
        DefaultUniform.OFFSET_SCALE_MATRIX,
        fromTransform(this.tmpMat4_, offsetScaleMatrix)
      );
      this.setUniformMatrixValue(
        DefaultUniform.OFFSET_ROTATION_MATRIX,
        fromTransform(this.tmpMat4_, offsetRotateMatrix)
      );

      this.setUniformFloatValue(
        DefaultUniform.TIME,
        (Date.now() - this.startTime_) * 0.001
      );
      this.setUniformFloatValue(DefaultUniform.ZOOM, frameState.viewState.zoom);
      this.setUniformFloatValue(
        DefaultUniform.RESOLUTION,
        frameState.viewState.resolution
      );
      this.setUniformFloatValue(DefaultUniform.PIXEL_RATIO, pixelRatio);
      this.setUniformFloatVec2(DefaultUniform.SIZE_PX, [size[0], size[1]]);
    }

    /**
     * Sets the custom uniforms based on what was given in the constructor. This is called internally in `prepareDraw`.
     * @param {import("../Map.js").FrameState} frameState Frame state.
     */
    applyUniforms(frameState) {
      const gl = this.getGL();

      let value;
      let textureSlot = 0;
      this.uniforms_.forEach(
        function (uniform) {
          value =
            typeof uniform.value === 'function'
              ? uniform.value(frameState)
              : uniform.value;

          // apply value based on type
          if (
            value instanceof HTMLCanvasElement ||
            value instanceof HTMLImageElement ||
            value instanceof ImageData
          ) {
            // create a texture & put data
            if (!uniform.texture) {
              uniform.prevValue = undefined;
              uniform.texture = gl.createTexture();
            }
            gl.activeTexture(gl[`TEXTURE${textureSlot}`]);
            gl.bindTexture(gl.TEXTURE_2D, uniform.texture);
            gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR);
            gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.CLAMP_TO_EDGE);
            gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.CLAMP_TO_EDGE);

            const imageReady =
              !(value instanceof HTMLImageElement) ||
              /** @type {HTMLImageElement} */ (value).complete;
            if (imageReady && uniform.prevValue !== value) {
              uniform.prevValue = value;
              gl.texImage2D(
                gl.TEXTURE_2D,
                0,
                gl.RGBA,
                gl.RGBA,
                gl.UNSIGNED_BYTE,
                value
              );
            }

            // fill texture slots by increasing index
            gl.uniform1i(this.getUniformLocation(uniform.name), textureSlot++);
          } else if (Array.isArray(value) && value.length === 6) {
            this.setUniformMatrixValue(
              uniform.name,
              fromTransform(this.tmpMat4_, value)
            );
          } else if (Array.isArray(value) && value.length <= 4) {
            switch (value.length) {
              case 2:
                gl.uniform2f(
                  this.getUniformLocation(uniform.name),
                  value[0],
                  value[1]
                );
                return;
              case 3:
                gl.uniform3f(
                  this.getUniformLocation(uniform.name),
                  value[0],
                  value[1],
                  value[2]
                );
                return;
              case 4:
                gl.uniform4f(
                  this.getUniformLocation(uniform.name),
                  value[0],
                  value[1],
                  value[2],
                  value[3]
                );
                return;
              default:
                return;
            }
          } else if (typeof value === 'number') {
            gl.uniform1f(this.getUniformLocation(uniform.name), value);
          }
        }.bind(this)
      );
    }

    /**
     * Set up a program for use. The program will be set as the current one. Then, the uniforms used
     * in the program will be set based on the current frame state and the helper configuration.
     * @param {WebGLProgram} program Program.
     * @param {import("../Map.js").FrameState} frameState Frame state.
     */
    useProgram(program, frameState) {
      const gl = this.getGL();
      gl.useProgram(program);
      this.currentProgram_ = program;
      this.uniformLocations_ = {};
      this.attribLocations_ = {};
      this.applyFrameState(frameState);
      this.applyUniforms(frameState);
    }

    /**
     * Will attempt to compile a vertex or fragment shader based on source
     * On error, the shader will be returned but
     * `gl.getShaderParameter(shader, gl.COMPILE_STATUS)` will return `true`
     * Use `gl.getShaderInfoLog(shader)` to have details
     * @param {string} source Shader source
     * @param {ShaderType} type VERTEX_SHADER or FRAGMENT_SHADER
     * @return {WebGLShader} Shader object
     */
    compileShader(source, type) {
      const gl = this.getGL();
      const shader = gl.createShader(type);
      gl.shaderSource(shader, source);
      gl.compileShader(shader);
      return shader;
    }

    /**
     * Create a program for a vertex and fragment shader.  Throws if shader compilation fails.
     * @param {string} fragmentShaderSource Fragment shader source.
     * @param {string} vertexShaderSource Vertex shader source.
     * @return {WebGLProgram} Program
     */
    getProgram(fragmentShaderSource, vertexShaderSource) {
      const gl = this.getGL();

      const fragmentShader = this.compileShader(
        fragmentShaderSource,
        gl.FRAGMENT_SHADER
      );

      const vertexShader = this.compileShader(
        vertexShaderSource,
        gl.VERTEX_SHADER
      );

      const program = gl.createProgram();
      gl.attachShader(program, fragmentShader);
      gl.attachShader(program, vertexShader);
      gl.linkProgram(program);

      if (!gl.getShaderParameter(fragmentShader, gl.COMPILE_STATUS)) {
        const message = `Fragment shader compliation failed: ${gl.getShaderInfoLog(
        fragmentShader
      )}`;
        throw new Error(message);
      }
      gl.deleteShader(fragmentShader);

      if (!gl.getShaderParameter(vertexShader, gl.COMPILE_STATUS)) {
        const message = `Vertex shader compilation failed: ${gl.getShaderInfoLog(
        vertexShader
      )}`;
        throw new Error(message);
      }
      gl.deleteShader(vertexShader);

      if (!gl.getProgramParameter(program, gl.LINK_STATUS)) {
        const message = `GL program linking failed: ${gl.getShaderInfoLog(
        vertexShader
      )}`;
        throw new Error(message);
      }

      return program;
    }

    /**
     * Will get the location from the shader or the cache
     * @param {string} name Uniform name
     * @return {WebGLUniformLocation} uniformLocation
     */
    getUniformLocation(name) {
      if (this.uniformLocations_[name] === undefined) {
        this.uniformLocations_[name] = this.getGL().getUniformLocation(
          this.currentProgram_,
          name
        );
      }
      return this.uniformLocations_[name];
    }

    /**
     * Will get the location from the shader or the cache
     * @param {string} name Attribute name
     * @return {number} attribLocation
     */
    getAttributeLocation(name) {
      if (this.attribLocations_[name] === undefined) {
        this.attribLocations_[name] = this.getGL().getAttribLocation(
          this.currentProgram_,
          name
        );
      }
      return this.attribLocations_[name];
    }

    /**
     * Modifies the given transform to apply the rotation/translation/scaling of the given frame state.
     * The resulting transform can be used to convert world space coordinates to view coordinates.
     * @param {import("../Map.js").FrameState} frameState Frame state.
     * @param {import("../transform").Transform} transform Transform to update.
     * @return {import("../transform").Transform} The updated transform object.
     */
    makeProjectionTransform(frameState, transform) {
      const size = frameState.size;
      const rotation = frameState.viewState.rotation;
      const resolution = frameState.viewState.resolution;
      const center = frameState.viewState.center;

      reset(transform);
      compose(
        transform,
        0,
        0,
        2 / (resolution * size[0]),
        2 / (resolution * size[1]),
        -rotation,
        -center[0],
        -center[1]
      );
      return transform;
    }

    /**
     * Give a value for a standard float uniform
     * @param {string} uniform Uniform name
     * @param {number} value Value
     */
    setUniformFloatValue(uniform, value) {
      this.getGL().uniform1f(this.getUniformLocation(uniform), value);
    }

    /**
     * Give a value for a vec2 uniform
     * @param {string} uniform Uniform name
     * @param {Array<number>} value Array of length 4.
     */
    setUniformFloatVec2(uniform, value) {
      this.getGL().uniform2fv(this.getUniformLocation(uniform), value);
    }

    /**
     * Give a value for a vec4 uniform
     * @param {string} uniform Uniform name
     * @param {Array<number>} value Array of length 4.
     */
    setUniformFloatVec4(uniform, value) {
      this.getGL().uniform4fv(this.getUniformLocation(uniform), value);
    }

    /**
     * Give a value for a standard matrix4 uniform
     * @param {string} uniform Uniform name
     * @param {Array<number>} value Matrix value
     */
    setUniformMatrixValue(uniform, value) {
      this.getGL().uniformMatrix4fv(
        this.getUniformLocation(uniform),
        false,
        value
      );
    }

    /**
     * Will set the currently bound buffer to an attribute of the shader program. Used by `#enableAttributes`
     * internally.
     * @param {string} attribName Attribute name
     * @param {number} size Number of components per attributes
     * @param {number} type UNSIGNED_INT, UNSIGNED_BYTE, UNSIGNED_SHORT or FLOAT
     * @param {number} stride Stride in bytes (0 means attribs are packed)
     * @param {number} offset Offset in bytes
     * @private
     */
    enableAttributeArray_(attribName, size, type, stride, offset) {
      const location = this.getAttributeLocation(attribName);
      // the attribute has not been found in the shaders; do not enable it
      if (location < 0) {
        return;
      }
      this.getGL().enableVertexAttribArray(location);
      this.getGL().vertexAttribPointer(
        location,
        size,
        type,
        false,
        stride,
        offset
      );
    }

    /**
     * Will enable the following attributes to be read from the currently bound buffer,
     * i.e. tell the GPU where to read the different attributes in the buffer. An error in the
     * size/type/order of attributes will most likely break the rendering and throw a WebGL exception.
     * @param {Array<AttributeDescription>} attributes Ordered list of attributes to read from the buffer
     */
    enableAttributes(attributes) {
      const stride = computeAttributesStride(attributes);
      let offset = 0;
      for (let i = 0; i < attributes.length; i++) {
        const attr = attributes[i];
        this.enableAttributeArray_(
          attr.name,
          attr.size,
          attr.type || FLOAT,
          stride,
          offset
        );
        offset += attr.size * getByteSizeFromType(attr.type);
      }
    }

    /**
     * WebGL context was lost
     * @private
     */
    handleWebGLContextLost() {
      clear$2(this.bufferCache_);
      this.currentProgram_ = null;
    }

    /**
     * WebGL context was restored
     * @private
     */
    handleWebGLContextRestored() {}

    /**
     * Will create or reuse a given webgl texture and apply the given size. If no image data
     * specified, the texture will be empty, otherwise image data will be used and the `size`
     * parameter will be ignored.
     * Note: wrap parameters are set to clamp to edge, min filter is set to linear.
     * @param {Array<number>} size Expected size of the texture
     * @param {ImageData|HTMLImageElement|HTMLCanvasElement} [data] Image data/object to bind to the texture
     * @param {WebGLTexture} [texture] Existing texture to reuse
     * @return {WebGLTexture} The generated texture
     */
    createTexture(size, data, texture) {
      const gl = this.getGL();
      texture = texture || gl.createTexture();

      // set params & size
      const level = 0;
      const internalFormat = gl.RGBA;
      const border = 0;
      const format = gl.RGBA;
      const type = gl.UNSIGNED_BYTE;
      gl.bindTexture(gl.TEXTURE_2D, texture);
      if (data) {
        gl.texImage2D(gl.TEXTURE_2D, level, internalFormat, format, type, data);
      } else {
        gl.texImage2D(
          gl.TEXTURE_2D,
          level,
          internalFormat,
          size[0],
          size[1],
          border,
          format,
          type,
          null
        );
      }
      gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.LINEAR);
      gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.CLAMP_TO_EDGE);
      gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.CLAMP_TO_EDGE);

      return texture;
    }
  }

  /**
   * Compute a stride in bytes based on a list of attributes
   * @param {Array<AttributeDescription>} attributes Ordered list of attributes
   * @return {number} Stride, ie amount of values for each vertex in the vertex buffer
   */
  function computeAttributesStride(attributes) {
    let stride = 0;
    for (let i = 0; i < attributes.length; i++) {
      const attr = attributes[i];
      stride += attr.size * getByteSizeFromType(attr.type);
    }
    return stride;
  }

  /**
   * Computes the size in byte of an attribute type.
   * @param {AttributeType} type Attribute type
   * @return {number} The size in bytes
   */
  function getByteSizeFromType(type) {
    switch (type) {
      case AttributeType.UNSIGNED_BYTE:
        return Uint8Array.BYTES_PER_ELEMENT;
      case AttributeType.UNSIGNED_SHORT:
        return Uint16Array.BYTES_PER_ELEMENT;
      case AttributeType.UNSIGNED_INT:
        return Uint32Array.BYTES_PER_ELEMENT;
      case AttributeType.FLOAT:
      default:
        return Float32Array.BYTES_PER_ELEMENT;
    }
  }

  /**
   * @module ol/webgl/PaletteTexture
   */

  class PaletteTexture {
    /**
     * @param {string} name The name of the texture.
     * @param {Uint8Array} data The texture data.
     */
    constructor(name, data) {
      this.name = name;
      this.data = data;

      /**
       * @type {WebGLTexture}
       * @private
       */
      this.texture_ = null;
    }

    /**
     * @param {WebGLRenderingContext} gl Rendering context.
     * @return {WebGLTexture} The texture.
     */
    getTexture(gl) {
      if (!this.texture_) {
        const texture = gl.createTexture();
        gl.bindTexture(gl.TEXTURE_2D, texture);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.CLAMP_TO_EDGE);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.CLAMP_TO_EDGE);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.NEAREST);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.NEAREST);
        gl.texImage2D(
          gl.TEXTURE_2D,
          0,
          gl.RGBA,
          this.data.length / 4,
          1,
          0,
          gl.RGBA,
          gl.UNSIGNED_BYTE,
          this.data
        );
        this.texture_ = texture;
      }
      return this.texture_;
    }
  }

  /**
   * A wrapper class to simplify rendering to a texture instead of the final canvas
   * @module ol/webgl/RenderTarget
   */

  // for pixel color reading
  const tmpArray4 = new Uint8Array(4);

  /**
   * @classdesc
   * This class is a wrapper around the association of both a `WebGLTexture` and a `WebGLFramebuffer` instances,
   * simplifying initialization and binding for rendering.
   * @api
   */
  class WebGLRenderTarget {
    /**
     * @param {import("./Helper.js").default} helper WebGL helper; mandatory.
     * @param {Array<number>} [size] Expected size of the render target texture; note: this can be changed later on.
     */
    constructor(helper, size) {
      /**
       * @private
       * @type {import("./Helper.js").default}
       */
      this.helper_ = helper;
      const gl = helper.getGL();

      /**
       * @private
       * @type {WebGLTexture}
       */
      this.texture_ = gl.createTexture();

      /**
       * @private
       * @type {WebGLFramebuffer}
       */
      this.framebuffer_ = gl.createFramebuffer();

      /**
       * @type {Array<number>}
       * @private
       */
      this.size_ = size || [1, 1];

      /**
       * @type {Uint8Array}
       * @private
       */
      this.data_ = new Uint8Array(0);

      /**
       * @type {boolean}
       * @private
       */
      this.dataCacheDirty_ = true;

      this.updateSize_();
    }

    /**
     * Changes the size of the render target texture. Note: will do nothing if the size
     * is already the same.
     * @param {Array<number>} size Expected size of the render target texture
     * @api
     */
    setSize(size) {
      if (equals$3(size, this.size_)) {
        return;
      }
      this.size_[0] = size[0];
      this.size_[1] = size[1];
      this.updateSize_();
    }

    /**
     * Returns the size of the render target texture
     * @return {Array<number>} Size of the render target texture
     * @api
     */
    getSize() {
      return this.size_;
    }

    /**
     * This will cause following calls to `#readAll` or `#readPixel` to download the content of the
     * render target into memory, which is an expensive operation.
     * This content will be kept in cache but should be cleared after each new render.
     * @api
     */
    clearCachedData() {
      this.dataCacheDirty_ = true;
    }

    /**
     * Returns the full content of the frame buffer as a series of r, g, b, a components
     * in the 0-255 range (unsigned byte).
     * @return {Uint8Array} Integer array of color values
     * @api
     */
    readAll() {
      if (this.dataCacheDirty_) {
        const size = this.size_;
        const gl = this.helper_.getGL();

        gl.bindFramebuffer(gl.FRAMEBUFFER, this.framebuffer_);
        gl.readPixels(
          0,
          0,
          size[0],
          size[1],
          gl.RGBA,
          gl.UNSIGNED_BYTE,
          this.data_
        );
        this.dataCacheDirty_ = false;
      }
      return this.data_;
    }

    /**
     * Reads one pixel of the frame buffer as an array of r, g, b, a components
     * in the 0-255 range (unsigned byte).
     * If x and/or y are outside of existing data, an array filled with 0 is returned.
     * @param {number} x Pixel coordinate
     * @param {number} y Pixel coordinate
     * @return {Uint8Array} Integer array with one color value (4 components)
     * @api
     */
    readPixel(x, y) {
      if (x < 0 || y < 0 || x > this.size_[0] || y >= this.size_[1]) {
        tmpArray4[0] = 0;
        tmpArray4[1] = 0;
        tmpArray4[2] = 0;
        tmpArray4[3] = 0;
        return tmpArray4;
      }

      this.readAll();
      const index =
        Math.floor(x) + (this.size_[1] - Math.floor(y) - 1) * this.size_[0];
      tmpArray4[0] = this.data_[index * 4];
      tmpArray4[1] = this.data_[index * 4 + 1];
      tmpArray4[2] = this.data_[index * 4 + 2];
      tmpArray4[3] = this.data_[index * 4 + 3];
      return tmpArray4;
    }

    /**
     * @return {WebGLTexture} Texture to render to
     */
    getTexture() {
      return this.texture_;
    }

    /**
     * @return {WebGLFramebuffer} Frame buffer of the render target
     */
    getFramebuffer() {
      return this.framebuffer_;
    }

    /**
     * @private
     */
    updateSize_() {
      const size = this.size_;
      const gl = this.helper_.getGL();

      this.texture_ = this.helper_.createTexture(size, null, this.texture_);

      gl.bindFramebuffer(gl.FRAMEBUFFER, this.framebuffer_);
      gl.viewport(0, 0, size[0], size[1]);
      gl.framebufferTexture2D(
        gl.FRAMEBUFFER,
        gl.COLOR_ATTACHMENT0,
        gl.TEXTURE_2D,
        this.texture_,
        0
      );

      this.data_ = new Uint8Array(size[0] * size[1] * 4);
    }
  }

  /**
   * @module ol/webgl/TileTexture
   */

  /**
   * @param {WebGLRenderingContext} gl The WebGL context.
   * @param {WebGLTexture} texture The texture.
   * @param {boolean} interpolate Interpolate when resampling.
   */
  function bindAndConfigure(gl, texture, interpolate) {
    const resampleFilter = interpolate ? gl.LINEAR : gl.NEAREST;
    gl.bindTexture(gl.TEXTURE_2D, texture);
    gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.CLAMP_TO_EDGE);
    gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.CLAMP_TO_EDGE);
    gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, resampleFilter);
    gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, resampleFilter);
  }

  /**
   * @param {WebGLRenderingContext} gl The WebGL context.
   * @param {WebGLTexture} texture The texture.
   * @param {HTMLImageElement|HTMLCanvasElement|HTMLVideoElement} image The image.
   * @param {boolean} interpolate Interpolate when resampling.
   */
  function uploadImageTexture(gl, texture, image, interpolate) {
    bindAndConfigure(gl, texture, interpolate);

    gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, gl.RGBA, gl.UNSIGNED_BYTE, image);
  }

  /**
   * @param {import("./Helper.js").default} helper The WebGL helper.
   * @param {WebGLTexture} texture The texture.
   * @param {import("../DataTile.js").Data} data The pixel data.
   * @param {import("../size.js").Size} size The pixel size.
   * @param {number} bandCount The band count.
   * @param {boolean} interpolate Interpolate when resampling.
   */
  function uploadDataTexture(
    helper,
    texture,
    data,
    size,
    bandCount,
    interpolate
  ) {
    const gl = helper.getGL();
    let textureType;
    let canInterpolate;
    if (data instanceof Float32Array) {
      textureType = gl.FLOAT;
      helper.getExtension('OES_texture_float');
      const extension = helper.getExtension('OES_texture_float_linear');
      canInterpolate = extension !== null;
    } else {
      textureType = gl.UNSIGNED_BYTE;
      canInterpolate = true;
    }
    bindAndConfigure(gl, texture, interpolate && canInterpolate);

    const bytesPerRow = data.byteLength / size[1];
    let unpackAlignment = 1;
    if (bytesPerRow % 8 === 0) {
      unpackAlignment = 8;
    } else if (bytesPerRow % 4 === 0) {
      unpackAlignment = 4;
    } else if (bytesPerRow % 2 === 0) {
      unpackAlignment = 2;
    }

    let format;
    switch (bandCount) {
      case 1: {
        format = gl.LUMINANCE;
        break;
      }
      case 2: {
        format = gl.LUMINANCE_ALPHA;
        break;
      }
      case 3: {
        format = gl.RGB;
        break;
      }
      case 4: {
        format = gl.RGBA;
        break;
      }
      default: {
        throw new Error(`Unsupported number of bands: ${bandCount}`);
      }
    }

    const oldUnpackAlignment = gl.getParameter(gl.UNPACK_ALIGNMENT);
    gl.pixelStorei(gl.UNPACK_ALIGNMENT, unpackAlignment);
    gl.texImage2D(
      gl.TEXTURE_2D,
      0,
      format,
      size[0],
      size[1],
      0,
      format,
      textureType,
      data
    );
    gl.pixelStorei(gl.UNPACK_ALIGNMENT, oldUnpackAlignment);
  }

  /**
   * @type {CanvasRenderingContext2D}
   */
  let pixelContext = null;

  function createPixelContext() {
    const canvas = document.createElement('canvas');
    canvas.width = 1;
    canvas.height = 1;
    pixelContext = canvas.getContext('2d');
  }

  /**
   * @typedef {import("../DataTile.js").default|ImageTile|ReprojTile} TileType
   */

  /**
   * @typedef {Object} Options
   * @property {TileType} tile The tile.
   * @property {import("../tilegrid/TileGrid.js").default} grid Tile grid.
   * @property {import("../webgl/Helper.js").default} helper WebGL helper.
   * @property {number} [gutter=0] The size in pixels of the gutter around image tiles to ignore.
   */

  class TileTexture extends Target {
    /**
     * @param {Options} options The tile texture options.
     */
    constructor(options) {
      super();

      /**
       * @type {TileType}
       */
      this.tile;

      /**
       * @type {Array<WebGLTexture>}
       */
      this.textures = [];
      this.handleTileChange_ = this.handleTileChange_.bind(this);

      /**
       * @type {import("../size.js").Size}
       * @private
       */
      this.renderSize_ = toSize(
        options.grid.getTileSize(options.tile.tileCoord[0])
      );

      /**
       * @type {number}
       * @private
       */
      this.gutter_ = options.gutter || 0;

      /**
       * @type {number}
       */
      this.bandCount = NaN;

      /**
       * @type {import("../webgl/Helper.js").default}
       * @private
       */
      this.helper_ = options.helper;

      const coords = new WebGLArrayBuffer(ARRAY_BUFFER, STATIC_DRAW);
      coords.fromArray([
        0, // P0
        1,
        1, // P1
        1,
        1, // P2
        0,
        0, // P3
        0,
      ]);
      this.helper_.flushBufferData(coords);

      /**
       * @type {WebGLArrayBuffer}
       */
      this.coords = coords;

      this.setTile(options.tile);
    }

    /**
     * @param {TileType} tile Tile.
     */
    setTile(tile) {
      if (tile !== this.tile) {
        if (this.tile) {
          this.tile.removeEventListener(EventType$1.CHANGE, this.handleTileChange_);
        }
        this.tile = tile;
        this.textures.length = 0;
        this.loaded = tile.getState() === TileState.LOADED;
        if (this.loaded) {
          this.uploadTile_();
        } else {
          if (tile instanceof ImageTile) {
            const image = tile.getImage();
            if (image instanceof Image && !image.crossOrigin) {
              image.crossOrigin = 'anonymous';
            }
          }
          tile.addEventListener(EventType$1.CHANGE, this.handleTileChange_);
        }
      }
    }

    uploadTile_() {
      const helper = this.helper_;
      const gl = helper.getGL();
      const tile = this.tile;

      if (tile instanceof ImageTile || tile instanceof ReprojTile) {
        const texture = gl.createTexture();
        this.textures.push(texture);
        this.bandCount = 4;
        uploadImageTexture(gl, texture, tile.getImage(), tile.interpolate);
        return;
      }

      const sourceTileSize = tile.getSize();
      const pixelSize = [
        sourceTileSize[0] + 2 * this.gutter_,
        sourceTileSize[1] + 2 * this.gutter_,
      ];
      const data = tile.getData();
      const isFloat = data instanceof Float32Array;
      const pixelCount = pixelSize[0] * pixelSize[1];
      const DataType = isFloat ? Float32Array : Uint8Array;
      const bytesPerElement = DataType.BYTES_PER_ELEMENT;
      const bytesPerRow = data.byteLength / pixelSize[1];

      this.bandCount = Math.floor(bytesPerRow / bytesPerElement / pixelSize[0]);
      const textureCount = Math.ceil(this.bandCount / 4);

      if (textureCount === 1) {
        const texture = gl.createTexture();
        this.textures.push(texture);
        uploadDataTexture(
          helper,
          texture,
          data,
          pixelSize,
          this.bandCount,
          tile.interpolate
        );
        return;
      }

      const textureDataArrays = new Array(textureCount);
      for (let textureIndex = 0; textureIndex < textureCount; ++textureIndex) {
        const texture = gl.createTexture();
        this.textures.push(texture);

        const bandCount =
          textureIndex < textureCount - 1 ? 4 : ((this.bandCount - 1) % 4) + 1;
        textureDataArrays[textureIndex] = new DataType(pixelCount * bandCount);
      }

      let dataIndex = 0;
      let rowOffset = 0;
      const colCount = pixelSize[0] * this.bandCount;
      for (let rowIndex = 0; rowIndex < pixelSize[1]; ++rowIndex) {
        for (let colIndex = 0; colIndex < colCount; ++colIndex) {
          const dataValue = data[rowOffset + colIndex];

          const pixelIndex = Math.floor(dataIndex / this.bandCount);
          const bandIndex = colIndex % this.bandCount;
          const textureIndex = Math.floor(bandIndex / 4);
          const textureData = textureDataArrays[textureIndex];
          const bandCount = textureData.length / pixelCount;
          const textureBandIndex = bandIndex % 4;
          textureData[pixelIndex * bandCount + textureBandIndex] = dataValue;

          ++dataIndex;
        }
        rowOffset += bytesPerRow / bytesPerElement;
      }

      for (let textureIndex = 0; textureIndex < textureCount; ++textureIndex) {
        const texture = this.textures[textureIndex];
        const textureData = textureDataArrays[textureIndex];
        const bandCount = textureData.length / pixelCount;
        uploadDataTexture(
          helper,
          texture,
          textureData,
          pixelSize,
          bandCount,
          tile.interpolate
        );
      }
    }

    handleTileChange_() {
      if (this.tile.getState() === TileState.LOADED) {
        this.loaded = true;
        this.uploadTile_();
        this.dispatchEvent(EventType$1.CHANGE);
      }
    }

    disposeInternal() {
      const gl = this.helper_.getGL();
      this.helper_.deleteBuffer(this.coords);
      for (let i = 0; i < this.textures.length; ++i) {
        gl.deleteTexture(this.textures[i]);
      }
      this.tile.removeEventListener(EventType$1.CHANGE, this.handleTileChange_);
    }

    /**
     * Get data for a pixel.  If the tile is not loaded, null is returned.
     * @param {number} renderCol The column index (in rendered tile space).
     * @param {number} renderRow The row index (in rendered tile space).
     * @return {import("../DataTile.js").Data|null} The data.
     */
    getPixelData(renderCol, renderRow) {
      if (!this.loaded) {
        return null;
      }
      const renderWidth = this.renderSize_[0];
      const renderHeight = this.renderSize_[1];
      const gutter = this.gutter_;

      if (this.tile instanceof DataTile) {
        const sourceSize = this.tile.getSize();

        const sourceWidthWithoutGutter = sourceSize[0];
        const sourceHeightWithoutGutter = sourceSize[1];
        const sourceWidth = sourceWidthWithoutGutter + 2 * gutter;
        const sourceHeight = sourceHeightWithoutGutter + 2 * gutter;

        const sourceCol =
          gutter +
          Math.floor(sourceWidthWithoutGutter * (renderCol / renderWidth));

        const sourceRow =
          gutter +
          Math.floor(sourceHeightWithoutGutter * (renderRow / renderHeight));

        const data = this.tile.getData();
        if (data instanceof DataView) {
          const bytesPerPixel = data.byteLength / (sourceWidth * sourceHeight);
          const offset = bytesPerPixel * (sourceRow * sourceWidth + sourceCol);
          const buffer = data.buffer.slice(offset, offset + bytesPerPixel);
          return new DataView(buffer);
        }

        const offset = this.bandCount * (sourceRow * sourceWidth + sourceCol);
        return data.slice(offset, offset + this.bandCount);
      }

      if (!pixelContext) {
        createPixelContext();
      }
      pixelContext.clearRect(0, 0, 1, 1);

      const image = this.tile.getImage();
      const sourceWidth = image.width;
      const sourceHeight = image.height;

      const sourceWidthWithoutGutter = sourceWidth - 2 * gutter;
      const sourceHeightWithoutGutter = sourceHeight - 2 * gutter;

      const sourceCol =
        gutter + Math.floor(sourceWidthWithoutGutter * (renderCol / renderWidth));

      const sourceRow =
        gutter +
        Math.floor(sourceHeightWithoutGutter * (renderRow / renderHeight));

      let data;
      try {
        pixelContext.drawImage(image, sourceCol, sourceRow, 1, 1, 0, 0, 1, 1);
        data = pixelContext.getImageData(0, 0, 1, 1).data;
      } catch (err) {
        pixelContext = null;
        return null;
      }
      return data;
    }
  }

  /**
   * @module ol/renderer/webgl/Layer
   */

  /**
   * @typedef {Object} PostProcessesOptions
   * @property {number} [scaleRatio] Scale ratio; if < 1, the post process will render to a texture smaller than
   * the main canvas that will then be sampled up (useful for saving resource on blur steps).
   * @property {string} [vertexShader] Vertex shader source
   * @property {string} [fragmentShader] Fragment shader source
   * @property {Object<string,import("../../webgl/Helper").UniformValue>} [uniforms] Uniform definitions for the post process step
   */

  /**
   * @typedef {Object} Options
   * @property {Object<string,import("../../webgl/Helper").UniformValue>} [uniforms] Uniform definitions for the post process steps
   * @property {Array<PostProcessesOptions>} [postProcesses] Post-processes definitions
   */

  /**
   * @classdesc
   * Base WebGL renderer class.
   * Holds all logic related to data manipulation & some common rendering logic
   * @template {import("../../layer/Layer.js").default} LayerType
   * @extends {LayerRenderer<LayerType>}
   */
  class WebGLLayerRenderer extends LayerRenderer {
    /**
     * @param {LayerType} layer Layer.
     * @param {Options} [options] Options.
     */
    constructor(layer, options) {
      super(layer);

      options = options || {};

      /**
       * The transform for viewport CSS pixels to rendered pixels.  This transform is only
       * set before dispatching rendering events.
       * @private
       * @type {import("../../transform.js").Transform}
       */
      this.inversePixelTransform_ = create$2();

      /**
       * @private
       * @type {CanvasRenderingContext2D}
       */
      this.pixelContext_ = null;

      /**
       * @private
       */
      this.postProcesses_ = options.postProcesses;

      /**
       * @private
       */
      this.uniforms_ = options.uniforms;

      /**
       * @type {WebGLHelper}
       * @protected
       */
      this.helper;

      layer.addChangeListener(LayerProperty.MAP, this.removeHelper.bind(this));

      this.dispatchPreComposeEvent = this.dispatchPreComposeEvent.bind(this);
      this.dispatchPostComposeEvent = this.dispatchPostComposeEvent.bind(this);
    }

    /**
     * @param {WebGLRenderingContext} context The WebGL rendering context.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @protected
     */
    dispatchPreComposeEvent(context, frameState) {
      const layer = this.getLayer();
      if (layer.hasListener(EventType.PRECOMPOSE)) {
        const event = new RenderEvent(
          EventType.PRECOMPOSE,
          undefined,
          frameState,
          context
        );
        layer.dispatchEvent(event);
      }
    }

    /**
     * @param {WebGLRenderingContext} context The WebGL rendering context.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @protected
     */
    dispatchPostComposeEvent(context, frameState) {
      const layer = this.getLayer();
      if (layer.hasListener(EventType.POSTCOMPOSE)) {
        const event = new RenderEvent(
          EventType.POSTCOMPOSE,
          undefined,
          frameState,
          context
        );
        layer.dispatchEvent(event);
      }
    }

    /**
     * Reset options (only handles uniforms).
     * @param {Options} options Options.
     */
    reset(options) {
      this.uniforms_ = options.uniforms;
      if (this.helper) {
        this.helper.setUniforms(this.uniforms_);
      }
    }

    /**
     * @protected
     */
    removeHelper() {
      if (this.helper) {
        this.helper.dispose();
        delete this.helper;
      }
    }

    /**
     * Determine whether renderFrame should be called.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @return {boolean} Layer is ready to be rendered.
     */
    prepareFrame(frameState) {
      if (this.getLayer().getRenderSource()) {
        let incrementGroup = true;
        let groupNumber = -1;
        let className;
        for (let i = 0, ii = frameState.layerStatesArray.length; i < ii; i++) {
          const layer = frameState.layerStatesArray[i].layer;
          const renderer = layer.getRenderer();
          if (!(renderer instanceof WebGLLayerRenderer)) {
            incrementGroup = true;
            continue;
          }
          const layerClassName = layer.getClassName();
          if (incrementGroup || layerClassName !== className) {
            groupNumber += 1;
            incrementGroup = false;
          }
          className = layerClassName;
          if (renderer === this) {
            break;
          }
        }

        const canvasCacheKey =
          'map/' + frameState.mapId + '/group/' + groupNumber;

        if (!this.helper || !this.helper.canvasCacheKeyMatches(canvasCacheKey)) {
          this.removeHelper();

          this.helper = new WebGLHelper({
            postProcesses: this.postProcesses_,
            uniforms: this.uniforms_,
            canvasCacheKey: canvasCacheKey,
          });

          if (className) {
            this.helper.getCanvas().className = className;
          }

          this.afterHelperCreated();
        }
      }

      return this.prepareFrameInternal(frameState);
    }

    /**
     * @protected
     */
    afterHelperCreated() {}

    /**
     * Determine whether renderFrame should be called.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @return {boolean} Layer is ready to be rendered.
     * @protected
     */
    prepareFrameInternal(frameState) {
      return true;
    }

    /**
     * Clean up.
     */
    disposeInternal() {
      this.removeHelper();
      super.disposeInternal();
    }

    /**
     * @param {import("../../render/EventType.js").default} type Event type.
     * @param {WebGLRenderingContext} context The rendering context.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @private
     */
    dispatchRenderEvent_(type, context, frameState) {
      const layer = this.getLayer();
      if (layer.hasListener(type)) {
        compose(
          this.inversePixelTransform_,
          0,
          0,
          frameState.pixelRatio,
          -frameState.pixelRatio,
          0,
          0,
          -frameState.size[1]
        );

        const event = new RenderEvent(
          type,
          this.inversePixelTransform_,
          frameState,
          context
        );
        layer.dispatchEvent(event);
      }
    }

    /**
     * @param {WebGLRenderingContext} context The rendering context.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @protected
     */
    preRender(context, frameState) {
      this.dispatchRenderEvent_(EventType.PRERENDER, context, frameState);
    }

    /**
     * @param {WebGLRenderingContext} context The rendering context.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @protected
     */
    postRender(context, frameState) {
      this.dispatchRenderEvent_(EventType.POSTRENDER, context, frameState);
    }
  }

  /**
   * @module ol/renderer/webgl/TileLayer
   */

  const Uniforms = {
    TILE_TEXTURE_ARRAY: 'u_tileTextures',
    TILE_TRANSFORM: 'u_tileTransform',
    TRANSITION_ALPHA: 'u_transitionAlpha',
    DEPTH: 'u_depth',
    TEXTURE_PIXEL_WIDTH: 'u_texturePixelWidth',
    TEXTURE_PIXEL_HEIGHT: 'u_texturePixelHeight',
    TEXTURE_RESOLUTION: 'u_textureResolution', // map units per texture pixel
    TEXTURE_ORIGIN_X: 'u_textureOriginX', // map x coordinate of left edge of texture
    TEXTURE_ORIGIN_Y: 'u_textureOriginY', // map y coordinate of top edge of texture
    RENDER_EXTENT: 'u_renderExtent', // intersection of layer, source, and view extent
    RESOLUTION: 'u_resolution',
    ZOOM: 'u_zoom',
  };

  const Attributes$3 = {
    TEXTURE_COORD: 'a_textureCoord',
  };

  /**
   * @type {Array<import('../../webgl/Helper.js').AttributeDescription>}
   */
  const attributeDescriptions = [
    {
      name: Attributes$3.TEXTURE_COORD,
      size: 2,
      type: AttributeType.FLOAT,
    },
  ];

  /**
   * @type {Object<string, boolean>}
   */
  const empty = {};

  /**
   * Transform a zoom level into a depth value ranging from -1 to 1.
   * @param {number} z A zoom level.
   * @return {number} A depth value.
   */
  function depthForZ(z) {
    return 2 * (1 - 1 / (z + 1)) - 1;
  }

  /**
   * Add a tile texture to the lookup.
   * @param {Object<number, Array<import("../../webgl/TileTexture.js").default>>} tileTexturesByZ Lookup of
   * tile textures by zoom level.
   * @param {import("../../webgl/TileTexture.js").default} tileTexture A tile texture.
   * @param {number} z The zoom level.
   */
  function addTileTextureToLookup(tileTexturesByZ, tileTexture, z) {
    if (!(z in tileTexturesByZ)) {
      tileTexturesByZ[z] = [];
    }
    tileTexturesByZ[z].push(tileTexture);
  }

  /**
   * @param {import("../../Map.js").FrameState} frameState Frame state.
   * @param {import("../../extent.js").Extent} extent The frame extent.
   * @return {import("../../extent.js").Extent} Frame extent intersected with layer extents.
   */
  function getRenderExtent(frameState, extent) {
    const layerState = frameState.layerStatesArray[frameState.layerIndex];
    if (layerState.extent) {
      extent = getIntersection(
        extent,
        fromUserExtent(layerState.extent, frameState.viewState.projection)
      );
    }
    const source = /** @type {import("../../source/Tile.js").default} */ (
      layerState.layer.getRenderSource()
    );
    if (!source.getWrapX()) {
      const gridExtent = source
        .getTileGridForProjection(frameState.viewState.projection)
        .getExtent();
      if (gridExtent) {
        extent = getIntersection(extent, gridExtent);
      }
    }
    return extent;
  }

  function getCacheKey(source, tileCoord) {
    return `${source.getKey()},${getKey(tileCoord)}`;
  }

  /**
   * @typedef {Object} Options
   * @property {string} vertexShader Vertex shader source.
   * @property {string} fragmentShader Fragment shader source.
   * @property {Object<string, import("../../webgl/Helper").UniformValue>} [uniforms] Additional uniforms
   * made available to shaders.
   * @property {Array<import("../../webgl/PaletteTexture.js").default>} [paletteTextures] Palette textures.
   * @property {number} [cacheSize=512] The texture cache size.
   */

  /**
   * @typedef {import("../../layer/WebGLTile.js").default} LayerType
   */

  /**
   * @classdesc
   * WebGL renderer for tile layers.
   * @extends {WebGLLayerRenderer<LayerType>}
   * @api
   */
  class WebGLTileLayerRenderer extends WebGLLayerRenderer {
    /**
     * @param {LayerType} tileLayer Tile layer.
     * @param {Options} options Options.
     */
    constructor(tileLayer, options) {
      super(tileLayer, {
        uniforms: options.uniforms,
      });

      /**
       * The last call to `renderFrame` was completed with all tiles loaded
       * @type {boolean}
       */
      this.renderComplete = false;

      /**
       * This transform converts texture coordinates to screen coordinates.
       * @type {import("../../transform.js").Transform}
       * @private
       */
      this.tileTransform_ = create$2();

      /**
       * @type {Array<number>}
       * @private
       */
      this.tempMat4_ = create$1();

      /**
       * @type {import("../../TileRange.js").default}
       * @private
       */
      this.tempTileRange_ = new TileRange(0, 0, 0, 0);

      /**
       * @type {import("../../tilecoord.js").TileCoord}
       * @private
       */
      this.tempTileCoord_ = createOrUpdate$1(0, 0, 0);

      /**
       * @type {import("../../size.js").Size}
       * @private
       */
      this.tempSize_ = [0, 0];

      /**
       * @type {WebGLProgram}
       * @private
       */
      this.program_;

      /**
       * @private
       */
      this.vertexShader_ = options.vertexShader;

      /**
       * @private
       */
      this.fragmentShader_ = options.fragmentShader;

      /**
       * Tiles are rendered as a quad with the following structure:
       *
       *  [P3]---------[P2]
       *   |`           |
       *   |  `     B   |
       *   |    `       |
       *   |      `     |
       *   |   A    `   |
       *   |          ` |
       *  [P0]---------[P1]
       *
       * Triangle A: P0, P1, P3
       * Triangle B: P1, P2, P3
       *
       * @private
       */
      this.indices_ = new WebGLArrayBuffer(ELEMENT_ARRAY_BUFFER, STATIC_DRAW);
      this.indices_.fromArray([0, 1, 3, 1, 2, 3]);

      const cacheSize = options.cacheSize !== undefined ? options.cacheSize : 512;

      /**
       * @type {import("../../structs/LRUCache.js").default<import("../../webgl/TileTexture.js").default>}
       * @private
       */
      this.tileTextureCache_ = new LRUCache(cacheSize);

      /**
       * @type {Array<import("../../webgl/PaletteTexture.js").default>}
       * @private
       */
      this.paletteTextures_ = options.paletteTextures || [];

      /**
       * @private
       * @type {import("../../Map.js").FrameState|null}
       */
      this.frameState_ = null;

      /**
       * @private
       * @type {import("../../proj/Projection.js").default}
       */
      this.projection_ = undefined;
    }

    /**
     * @param {Options} options Options.
     */
    reset(options) {
      super.reset({
        uniforms: options.uniforms,
      });
      this.vertexShader_ = options.vertexShader;
      this.fragmentShader_ = options.fragmentShader;
      this.paletteTextures_ = options.paletteTextures || [];

      if (this.helper) {
        this.program_ = this.helper.getProgram(
          this.fragmentShader_,
          this.vertexShader_
        );
      }
    }

    afterHelperCreated() {
      this.program_ = this.helper.getProgram(
        this.fragmentShader_,
        this.vertexShader_
      );

      this.helper.flushBufferData(this.indices_);
    }

    /**
     * @param {import("../../webgl/TileTexture").TileType} tile Tile.
     * @return {boolean} Tile is drawable.
     * @private
     */
    isDrawableTile_(tile) {
      const tileLayer = this.getLayer();
      const tileState = tile.getState();
      const useInterimTilesOnError = tileLayer.getUseInterimTilesOnError();
      return (
        tileState == TileState.LOADED ||
        tileState == TileState.EMPTY ||
        (tileState == TileState.ERROR && !useInterimTilesOnError)
      );
    }

    /**
     * Determine whether renderFrame should be called.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @return {boolean} Layer is ready to be rendered.
     */
    prepareFrameInternal(frameState) {
      if (!this.projection_) {
        this.projection_ = frameState.viewState.projection;
      } else if (frameState.viewState.projection !== this.projection_) {
        this.clearCache();
        this.projection_ = frameState.viewState.projection;
      }

      const layer = this.getLayer();
      const source = layer.getRenderSource();
      if (!source) {
        return false;
      }

      if (isEmpty(getRenderExtent(frameState, frameState.extent))) {
        return false;
      }
      return source.getState() === 'ready';
    }

    /**
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @param {import("../../extent.js").Extent} extent The extent to be rendered.
     * @param {number} initialZ The zoom level.
     * @param {Object<number, Array<TileTexture>>} tileTexturesByZ The zoom level.
     * @param {number} preload Number of additional levels to load.
     */
    enqueueTiles(frameState, extent, initialZ, tileTexturesByZ, preload) {
      const viewState = frameState.viewState;
      const tileLayer = this.getLayer();
      const tileSource = tileLayer.getRenderSource();
      const tileGrid = tileSource.getTileGridForProjection(viewState.projection);
      const gutter = tileSource.getGutterForProjection(viewState.projection);

      const tileSourceKey = getUid(tileSource);
      if (!(tileSourceKey in frameState.wantedTiles)) {
        frameState.wantedTiles[tileSourceKey] = {};
      }

      const wantedTiles = frameState.wantedTiles[tileSourceKey];
      const tileTextureCache = this.tileTextureCache_;

      const map = tileLayer.getMapInternal();
      const minZ = Math.max(
        initialZ - preload,
        tileGrid.getMinZoom(),
        tileGrid.getZForResolution(
          Math.min(
            tileLayer.getMaxResolution(),
            map
              ? map
                  .getView()
                  .getResolutionForZoom(Math.max(tileLayer.getMinZoom(), 0))
              : tileGrid.getResolution(0)
          ),
          tileSource.zDirection
        )
      );
      for (let z = initialZ; z >= minZ; --z) {
        const tileRange = tileGrid.getTileRangeForExtentAndZ(
          extent,
          z,
          this.tempTileRange_
        );

        const tileResolution = tileGrid.getResolution(z);

        for (let x = tileRange.minX; x <= tileRange.maxX; ++x) {
          for (let y = tileRange.minY; y <= tileRange.maxY; ++y) {
            const tileCoord = createOrUpdate$1(z, x, y, this.tempTileCoord_);
            const cacheKey = getCacheKey(tileSource, tileCoord);

            /** @type {TileTexture} */
            let tileTexture;

            /** @type {import("../../webgl/TileTexture").TileType} */
            let tile;

            if (tileTextureCache.containsKey(cacheKey)) {
              tileTexture = tileTextureCache.get(cacheKey);
              tile = tileTexture.tile;
            }
            if (!tileTexture || tileTexture.tile.key !== tileSource.getKey()) {
              tile = tileSource.getTile(
                z,
                x,
                y,
                frameState.pixelRatio,
                viewState.projection
              );
              if (!tileTexture) {
                tileTexture = new TileTexture({
                  tile: tile,
                  grid: tileGrid,
                  helper: this.helper,
                  gutter: gutter,
                });
                tileTextureCache.set(cacheKey, tileTexture);
              } else {
                if (this.isDrawableTile_(tile)) {
                  tileTexture.setTile(tile);
                } else {
                  const interimTile =
                    /** @type {import("../../webgl/TileTexture").TileType} */ (
                      tile.getInterimTile()
                    );
                  tileTexture.setTile(interimTile);
                }
              }
            }

            addTileTextureToLookup(tileTexturesByZ, tileTexture, z);

            const tileQueueKey = tile.getKey();
            wantedTiles[tileQueueKey] = true;

            if (tile.getState() === TileState.IDLE) {
              if (!frameState.tileQueue.isKeyQueued(tileQueueKey)) {
                frameState.tileQueue.enqueue([
                  tile,
                  tileSourceKey,
                  tileGrid.getTileCoordCenter(tileCoord),
                  tileResolution,
                ]);
              }
            }
          }
        }
      }
    }

    /**
     * Render the layer.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @return {HTMLElement} The rendered element.
     */
    renderFrame(frameState) {
      this.frameState_ = frameState;
      this.renderComplete = true;
      const gl = this.helper.getGL();
      this.preRender(gl, frameState);

      const viewState = frameState.viewState;
      const tileLayer = this.getLayer();
      const tileSource = tileLayer.getRenderSource();
      const tileGrid = tileSource.getTileGridForProjection(viewState.projection);
      const gutter = tileSource.getGutterForProjection(viewState.projection);
      const extent = getRenderExtent(frameState, frameState.extent);
      const z = tileGrid.getZForResolution(
        viewState.resolution,
        tileSource.zDirection
      );

      /**
       * @type {Object<number, Array<import("../../webgl/TileTexture.js").default>>}
       */
      const tileTexturesByZ = {};

      const preload = tileLayer.getPreload();
      if (frameState.nextExtent) {
        const targetZ = tileGrid.getZForResolution(
          viewState.nextResolution,
          tileSource.zDirection
        );
        const nextExtent = getRenderExtent(frameState, frameState.nextExtent);
        this.enqueueTiles(
          frameState,
          nextExtent,
          targetZ,
          tileTexturesByZ,
          preload
        );
      }

      this.enqueueTiles(frameState, extent, z, tileTexturesByZ, 0);
      if (preload > 0) {
        setTimeout(() => {
          this.enqueueTiles(
            frameState,
            extent,
            z - 1,
            tileTexturesByZ,
            preload - 1
          );
        }, 0);
      }

      /**
       * A lookup of alpha values for tiles at the target rendering resolution
       * for tiles that are in transition.  If a tile coord key is absent from
       * this lookup, the tile should be rendered at alpha 1.
       * @type {Object<string, number>}
       */
      const alphaLookup = {};

      const uid = getUid(this);
      const time = frameState.time;
      let blend = false;

      // look for cached tiles to use if a target tile is not ready
      const tileTextures = tileTexturesByZ[z];
      for (let i = 0, ii = tileTextures.length; i < ii; ++i) {
        const tileTexture = tileTextures[i];
        const tile = tileTexture.tile;
        if (tile instanceof ReprojTile && tile.getState() === TileState.EMPTY) {
          continue;
        }
        const tileCoord = tile.tileCoord;

        if (tileTexture.loaded) {
          const alpha = tile.getAlpha(uid, time);
          if (alpha === 1) {
            // no need to look for alt tiles
            tile.endTransition(uid);
            continue;
          }
          blend = true;
          const tileCoordKey = getKey(tileCoord);
          alphaLookup[tileCoordKey] = alpha;
        }
        this.renderComplete = false;

        // first look for child tiles (at z + 1)
        const coveredByChildren = this.findAltTiles_(
          tileGrid,
          tileCoord,
          z + 1,
          tileTexturesByZ
        );

        if (coveredByChildren) {
          continue;
        }

        // next look for parent tiles
        const minZoom = tileGrid.getMinZoom();
        for (let parentZ = z - 1; parentZ >= minZoom; --parentZ) {
          const coveredByParent = this.findAltTiles_(
            tileGrid,
            tileCoord,
            parentZ,
            tileTexturesByZ
          );

          if (coveredByParent) {
            break;
          }
        }
      }

      this.helper.useProgram(this.program_, frameState);
      this.helper.prepareDraw(frameState, !blend);

      const zs = Object.keys(tileTexturesByZ)
        .map(Number)
        .sort(numberSafeCompareFunction);

      const centerX = viewState.center[0];
      const centerY = viewState.center[1];

      for (let j = 0, jj = zs.length; j < jj; ++j) {
        const tileZ = zs[j];
        const tileResolution = tileGrid.getResolution(tileZ);
        const tileSize = toSize(tileGrid.getTileSize(tileZ), this.tempSize_);
        const tileOrigin = tileGrid.getOrigin(tileZ);

        const tileWidthWithGutter = tileSize[0] + 2 * gutter;
        const tileHeightWithGutter = tileSize[1] + 2 * gutter;
        const aspectRatio = tileWidthWithGutter / tileHeightWithGutter;

        const centerI =
          (centerX - tileOrigin[0]) / (tileSize[0] * tileResolution);
        const centerJ =
          (tileOrigin[1] - centerY) / (tileSize[1] * tileResolution);

        const tileScale = viewState.resolution / tileResolution;

        const depth = depthForZ(tileZ);
        const tileTextures = tileTexturesByZ[tileZ];
        for (let i = 0, ii = tileTextures.length; i < ii; ++i) {
          const tileTexture = tileTextures[i];
          if (!tileTexture.loaded) {
            continue;
          }
          const tile = tileTexture.tile;
          const tileCoord = tile.tileCoord;
          const tileCoordKey = getKey(tileCoord);

          const tileCenterI = tileCoord[1];
          const tileCenterJ = tileCoord[2];

          reset(this.tileTransform_);
          scale$3(
            this.tileTransform_,
            2 / ((frameState.size[0] * tileScale) / tileWidthWithGutter),
            -2 / ((frameState.size[1] * tileScale) / tileWidthWithGutter)
          );
          rotate$2(this.tileTransform_, viewState.rotation);
          scale$3(this.tileTransform_, 1, 1 / aspectRatio);
          translate$1(
            this.tileTransform_,
            (tileSize[0] * (tileCenterI - centerI) - gutter) /
              tileWidthWithGutter,
            (tileSize[1] * (tileCenterJ - centerJ) - gutter) /
              tileHeightWithGutter
          );

          this.helper.setUniformMatrixValue(
            Uniforms.TILE_TRANSFORM,
            fromTransform(this.tempMat4_, this.tileTransform_)
          );

          this.helper.bindBuffer(tileTexture.coords);
          this.helper.bindBuffer(this.indices_);
          this.helper.enableAttributes(attributeDescriptions);

          let textureSlot = 0;
          while (textureSlot < tileTexture.textures.length) {
            const textureProperty = 'TEXTURE' + textureSlot;
            const uniformName = `${Uniforms.TILE_TEXTURE_ARRAY}[${textureSlot}]`;
            gl.activeTexture(gl[textureProperty]);
            gl.bindTexture(gl.TEXTURE_2D, tileTexture.textures[textureSlot]);
            gl.uniform1i(
              this.helper.getUniformLocation(uniformName),
              textureSlot
            );
            ++textureSlot;
          }

          for (
            let paletteIndex = 0;
            paletteIndex < this.paletteTextures_.length;
            ++paletteIndex
          ) {
            const paletteTexture = this.paletteTextures_[paletteIndex];
            gl.activeTexture(gl['TEXTURE' + textureSlot]);
            const texture = paletteTexture.getTexture(gl);
            gl.bindTexture(gl.TEXTURE_2D, texture);
            gl.uniform1i(
              this.helper.getUniformLocation(paletteTexture.name),
              textureSlot
            );
            ++textureSlot;
          }

          const alpha =
            tileCoordKey in alphaLookup ? alphaLookup[tileCoordKey] : 1;

          if (alpha < 1) {
            frameState.animate = true;
          }

          this.helper.setUniformFloatValue(Uniforms.TRANSITION_ALPHA, alpha);
          this.helper.setUniformFloatValue(Uniforms.DEPTH, depth);
          this.helper.setUniformFloatValue(
            Uniforms.TEXTURE_PIXEL_WIDTH,
            tileWidthWithGutter
          );
          this.helper.setUniformFloatValue(
            Uniforms.TEXTURE_PIXEL_HEIGHT,
            tileHeightWithGutter
          );
          this.helper.setUniformFloatValue(
            Uniforms.TEXTURE_RESOLUTION,
            tileResolution
          );
          this.helper.setUniformFloatValue(
            Uniforms.TEXTURE_ORIGIN_X,
            tileOrigin[0] +
              tileCenterI * tileSize[0] * tileResolution -
              gutter * tileResolution
          );
          this.helper.setUniformFloatValue(
            Uniforms.TEXTURE_ORIGIN_Y,
            tileOrigin[1] -
              tileCenterJ * tileSize[1] * tileResolution +
              gutter * tileResolution
          );
          let gutterExtent = extent;
          if (gutter > 0) {
            gutterExtent = tileGrid.getTileCoordExtent(tileCoord);
            getIntersection(gutterExtent, extent, gutterExtent);
          }
          this.helper.setUniformFloatVec4(Uniforms.RENDER_EXTENT, gutterExtent);
          this.helper.setUniformFloatValue(
            Uniforms.RESOLUTION,
            viewState.resolution
          );
          this.helper.setUniformFloatValue(Uniforms.ZOOM, viewState.zoom);

          this.helper.drawElements(0, this.indices_.getSize());
        }
      }

      this.helper.finalizeDraw(
        frameState,
        this.dispatchPreComposeEvent,
        this.dispatchPostComposeEvent
      );

      const canvas = this.helper.getCanvas();

      const tileTextureCache = this.tileTextureCache_;
      while (tileTextureCache.canExpireCache()) {
        const tileTexture = tileTextureCache.pop();
        tileTexture.dispose();
      }

      // TODO: let the renderers manage their own cache instead of managing the source cache
      /**
       * Here we unconditionally expire the source cache since the renderer maintains
       * its own cache.
       * @param {import("../../Map.js").default} map Map.
       * @param {import("../../Map.js").FrameState} frameState Frame state.
       */
      const postRenderFunction = function (map, frameState) {
        tileSource.updateCacheSize(0.1, frameState.viewState.projection);
        tileSource.expireCache(frameState.viewState.projection, empty);
      };

      frameState.postRenderFunctions.push(postRenderFunction);

      this.postRender(gl, frameState);
      return canvas;
    }

    /**
     * @param {import("../../pixel.js").Pixel} pixel Pixel.
     * @return {Uint8ClampedArray|Uint8Array|Float32Array|DataView} Data at the pixel location.
     */
    getData(pixel) {
      const gl = this.helper.getGL();
      if (!gl) {
        return null;
      }

      const frameState = this.frameState_;
      if (!frameState) {
        return null;
      }

      const layer = this.getLayer();
      const coordinate = apply(
        frameState.pixelToCoordinateTransform,
        pixel.slice()
      );

      const viewState = frameState.viewState;
      const layerExtent = layer.getExtent();
      if (layerExtent) {
        if (
          !containsCoordinate(
            fromUserExtent(layerExtent, viewState.projection),
            coordinate
          )
        ) {
          return null;
        }
      }

      // determine last source suitable for rendering at coordinate
      const sources = layer.getSources(
        boundingExtent([coordinate]),
        viewState.resolution
      );
      let i, source, tileGrid;
      for (i = sources.length - 1; i >= 0; --i) {
        source = sources[i];
        if (source.getState() === 'ready') {
          tileGrid = source.getTileGridForProjection(viewState.projection);
          if (source.getWrapX()) {
            break;
          }
          const gridExtent = tileGrid.getExtent();
          if (!gridExtent || containsCoordinate(gridExtent, coordinate)) {
            break;
          }
        }
      }
      if (i < 0) {
        return null;
      }

      const tileTextureCache = this.tileTextureCache_;
      for (
        let z = tileGrid.getZForResolution(viewState.resolution);
        z >= tileGrid.getMinZoom();
        --z
      ) {
        const tileCoord = tileGrid.getTileCoordForCoordAndZ(coordinate, z);
        const cacheKey = getCacheKey(source, tileCoord);
        if (!tileTextureCache.containsKey(cacheKey)) {
          continue;
        }
        const tileTexture = tileTextureCache.get(cacheKey);
        const tile = tileTexture.tile;
        if (tile instanceof ReprojTile && tile.getState() === TileState.EMPTY) {
          return null;
        }
        if (!tileTexture.loaded) {
          continue;
        }
        const tileOrigin = tileGrid.getOrigin(z);
        const tileSize = toSize(tileGrid.getTileSize(z));
        const tileResolution = tileGrid.getResolution(z);

        const col =
          (coordinate[0] - tileOrigin[0]) / tileResolution -
          tileCoord[1] * tileSize[0];

        const row =
          (tileOrigin[1] - coordinate[1]) / tileResolution -
          tileCoord[2] * tileSize[1];

        return tileTexture.getPixelData(col, row);
      }
      return null;
    }

    /**
     * Look for tiles covering the provided tile coordinate at an alternate
     * zoom level.  Loaded tiles will be added to the provided tile texture lookup.
     * @param {import("../../tilegrid/TileGrid.js").default} tileGrid The tile grid.
     * @param {import("../../tilecoord.js").TileCoord} tileCoord The target tile coordinate.
     * @param {number} altZ The alternate zoom level.
     * @param {Object<number, Array<import("../../webgl/TileTexture.js").default>>} tileTexturesByZ Lookup of
     * tile textures by zoom level.
     * @return {boolean} The tile coordinate is covered by loaded tiles at the alternate zoom level.
     * @private
     */
    findAltTiles_(tileGrid, tileCoord, altZ, tileTexturesByZ) {
      const tileRange = tileGrid.getTileRangeForTileCoordAndZ(
        tileCoord,
        altZ,
        this.tempTileRange_
      );

      if (!tileRange) {
        return false;
      }

      let covered = true;
      const tileTextureCache = this.tileTextureCache_;
      const source = this.getLayer().getRenderSource();
      for (let x = tileRange.minX; x <= tileRange.maxX; ++x) {
        for (let y = tileRange.minY; y <= tileRange.maxY; ++y) {
          const cacheKey = getCacheKey(source, [altZ, x, y]);
          let loaded = false;
          if (tileTextureCache.containsKey(cacheKey)) {
            const tileTexture = tileTextureCache.get(cacheKey);
            if (tileTexture.loaded) {
              addTileTextureToLookup(tileTexturesByZ, tileTexture, altZ);
              loaded = true;
            }
          }
          if (!loaded) {
            covered = false;
          }
        }
      }
      return covered;
    }

    clearCache() {
      const tileTextureCache = this.tileTextureCache_;
      tileTextureCache.forEach((tileTexture) => tileTexture.dispose());
      tileTextureCache.clear();
    }

    removeHelper() {
      if (this.helper) {
        this.clearCache();
      }

      super.removeHelper();
    }

    /**
     * Clean up.
     */
    disposeInternal() {
      const helper = this.helper;
      if (helper) {
        const gl = helper.getGL();
        gl.deleteProgram(this.program_);
        delete this.program_;

        helper.deleteBuffer(this.indices_);
      }

      super.disposeInternal();

      delete this.indices_;
      delete this.tileTextureCache_;
      delete this.frameState_;
    }
  }

  /**
   * Operators and utilities used for style expressions
   * @module ol/style/expressions
   */

  /**
   * Base type used for literal style parameters; can be a number literal or the output of an operator,
   * which in turns takes {@link import("./expressions.js").ExpressionValue} arguments.
   *
   * The following operators can be used:
   *
   * * Reading operators:
   *   * `['band', bandIndex, xOffset, yOffset]` For tile layers only. Fetches pixel values from band
   *     `bandIndex` of the source's data. The first `bandIndex` of the source data is `1`. Fetched values
   *     are in the 0..1 range. {@link import("../source/TileImage.js").default} sources have 4 bands: red,
   *     green, blue and alpha. {@link import("../source/DataTile.js").default} sources can have any number
   *     of bands, depending on the underlying data source and
   *     {@link import("../source/GeoTIFF.js").Options configuration}. `xOffset` and `yOffset` are optional
   *     and allow specifying pixel offsets for x and y. This is used for sampling data from neighboring pixels.
   *   * `['get', 'attributeName']` fetches a feature attribute (it will be prefixed by `a_` in the shader)
   *     Note: those will be taken from the attributes provided to the renderer
   *   * `['resolution']` returns the current resolution
   *   * `['time']` returns the time in seconds since the creation of the layer
   *   * `['var', 'varName']` fetches a value from the style variables, or 0 if undefined
   *   * `['zoom']` returns the current zoom level
   *
   * * Math operators:
   *   * `['*', value1, value2]` multiplies `value1` by `value2`
   *   * `['/', value1, value2]` divides `value1` by `value2`
   *   * `['+', value1, value2]` adds `value1` and `value2`
   *   * `['-', value1, value2]` subtracts `value2` from `value1`
   *   * `['clamp', value, low, high]` clamps `value` between `low` and `high`
   *   * `['%', value1, value2]` returns the result of `value1 % value2` (modulo)
   *   * `['^', value1, value2]` returns the value of `value1` raised to the `value2` power
   *   * `['abs', value1]` returns the absolute value of `value1`
   *   * `['floor', value1]` returns the nearest integer less than or equal to `value1`
   *   * `['round', value1]` returns the nearest integer to `value1`
   *   * `['ceil', value1]` returns the nearest integer greater than or equal to `value1`
   *   * `['sin', value1]` returns the sine of `value1`
   *   * `['cos', value1]` returns the cosine of `value1`
   *   * `['atan', value1, value2]` returns `atan2(value1, value2)`. If `value2` is not provided, returns `atan(value1)`
   *
   * * Transform operators:
   *   * `['case', condition1, output1, ...conditionN, outputN, fallback]` selects the first output whose corresponding
   *     condition evaluates to `true`. If no match is found, returns the `fallback` value.
   *     All conditions should be `boolean`, output and fallback can be any kind.
   *   * `['match', input, match1, output1, ...matchN, outputN, fallback]` compares the `input` value against all
   *     provided `matchX` values, returning the output associated with the first valid match. If no match is found,
   *     returns the `fallback` value.
   *     `input` and `matchX` values must all be of the same type, and can be `number` or `string`. `outputX` and
   *     `fallback` values must be of the same type, and can be of any kind.
   *   * `['interpolate', interpolation, input, stop1, output1, ...stopN, outputN]` returns a value by interpolating between
   *     pairs of inputs and outputs; `interpolation` can either be `['linear']` or `['exponential', base]` where `base` is
   *     the rate of increase from stop A to stop B (i.e. power to which the interpolation ratio is raised); a value
   *     of 1 is equivalent to `['linear']`.
   *     `input` and `stopX` values must all be of type `number`. `outputX` values can be `number` or `color` values.
   *     Note: `input` will be clamped between `stop1` and `stopN`, meaning that all output values will be comprised
   *     between `output1` and `outputN`.
   *
   * * Logical operators:
   *   * `['<', value1, value2]` returns `true` if `value1` is strictly lower than `value2`, or `false` otherwise.
   *   * `['<=', value1, value2]` returns `true` if `value1` is lower than or equals `value2`, or `false` otherwise.
   *   * `['>', value1, value2]` returns `true` if `value1` is strictly greater than `value2`, or `false` otherwise.
   *   * `['>=', value1, value2]` returns `true` if `value1` is greater than or equals `value2`, or `false` otherwise.
   *   * `['==', value1, value2]` returns `true` if `value1` equals `value2`, or `false` otherwise.
   *   * `['!=', value1, value2]` returns `true` if `value1` does not equal `value2`, or `false` otherwise.
   *   * `['!', value1]` returns `false` if `value1` is `true` or greater than `0`, or `true` otherwise.
   *   * `['all', value1, value2, ...]` returns `true` if all the inputs are `true`, `false` otherwise.
   *   * `['any', value1, value2, ...]` returns `true` if any of the inputs are `true`, `false` otherwise.
   *   * `['between', value1, value2, value3]` returns `true` if `value1` is contained between `value2` and `value3`
   *     (inclusively), or `false` otherwise.
   *
   * * Conversion operators:
   *   * `['array', value1, ...valueN]` creates a numerical array from `number` values; please note that the amount of
   *     values can currently only be 2, 3 or 4.
   *   * `['color', red, green, blue, alpha]` creates a `color` value from `number` values; the `alpha` parameter is
   *     optional; if not specified, it will be set to 1.
   *     Note: `red`, `green` and `blue` components must be values between 0 and 255; `alpha` between 0 and 1.
   *   * `['palette', index, colors]` picks a `color` value from an array of colors using the given index; the `index`
   *     expression must evaluate to a number; the items in the `colors` array must be strings with hex colors
   *     (e.g. `'#86A136'`), colors using the rgba[a] functional notation (e.g. `'rgb(134, 161, 54)'` or `'rgba(134, 161, 54, 1)'`),
   *     named colors (e.g. `'red'`), or array literals with 3 ([r, g, b]) or 4 ([r, g, b, a]) values (with r, g, and b
   *     in the 0-255 range and a in the 0-1 range).
   *
   * Values can either be literals or another operator, as they will be evaluated recursively.
   * Literal values can be of the following types:
   * * `boolean`
   * * `number`
   * * `string`
   * * {@link module:ol/color~Color}
   *
   * @typedef {Array<*>|import("../color.js").Color|string|number|boolean} ExpressionValue
   * @api
   */

  /**
   * Possible inferred types from a given value or expression.
   * Note: these are binary flags.
   * @enum {number}
   */
  const ValueTypes = {
    NUMBER: 0b00001,
    STRING: 0b00010,
    COLOR: 0b00100,
    BOOLEAN: 0b01000,
    NUMBER_ARRAY: 0b10000,
    ANY: 0b11111,
    NONE: 0,
  };

  /**
   * An operator declaration must contain two methods: `getReturnType` which returns a type based on
   * the operator arguments, and `toGlsl` which returns a GLSL-compatible string.
   * Note: both methods can process arguments recursively.
   * @typedef {Object} Operator
   * @property {function(Array<ExpressionValue>): ValueTypes|number} getReturnType Returns one or several types
   * @property {function(ParsingContext, Array<ExpressionValue>, ValueTypes=): string} toGlsl Returns a GLSL-compatible string
   * Note: takes in an optional type hint as 3rd parameter
   */

  /**
   * Operator declarations
   * @type {Object<string, Operator>}
   */
  const Operators = {};

  /**
   * Returns the possible types for a given value (each type being a binary flag)
   * To test a value use e.g. `getValueType(v) & ValueTypes.BOOLEAN`
   * @param {ExpressionValue} value Value
   * @return {ValueTypes|number} Type or types inferred from the value
   */
  function getValueType(value) {
    if (typeof value === 'number') {
      return ValueTypes.NUMBER;
    }
    if (typeof value === 'boolean') {
      return ValueTypes.BOOLEAN;
    }
    if (typeof value === 'string') {
      if (isStringColor(value)) {
        return ValueTypes.COLOR | ValueTypes.STRING;
      }
      return ValueTypes.STRING;
    }
    if (!Array.isArray(value)) {
      throw new Error(`Unhandled value type: ${JSON.stringify(value)}`);
    }
    const valueArr = /** @type {Array<*>} */ (value);
    const onlyNumbers = valueArr.every(function (v) {
      return typeof v === 'number';
    });
    if (onlyNumbers) {
      if (valueArr.length === 3 || valueArr.length === 4) {
        return ValueTypes.COLOR | ValueTypes.NUMBER_ARRAY;
      }
      return ValueTypes.NUMBER_ARRAY;
    }
    if (typeof valueArr[0] !== 'string') {
      throw new Error(
        `Expected an expression operator but received: ${JSON.stringify(
        valueArr
      )}`
      );
    }
    const operator = Operators[valueArr[0]];
    if (operator === undefined) {
      throw new Error(
        `Unrecognized expression operator: ${JSON.stringify(valueArr)}`
      );
    }
    return operator.getReturnType(valueArr.slice(1));
  }

  /**
   * Checks if only one value type is enabled in the input number.
   * @param {ValueTypes|number} valueType Number containing value type binary flags
   * @return {boolean} True if only one type flag is enabled, false if zero or multiple
   */
  function isTypeUnique(valueType) {
    return Math.log2(valueType) % 1 === 0;
  }

  /**
   * Context available during the parsing of an expression.
   * @typedef {Object} ParsingContext
   * @property {boolean} [inFragmentShader] If false, means the expression output should be made for a vertex shader
   * @property {Array<string>} variables List of variables used in the expression; contains **unprefixed names**
   * @property {Array<string>} attributes List of attributes used in the expression; contains **unprefixed names**
   * @property {Object<string, number>} stringLiteralsMap This object maps all encountered string values to a number
   * @property {Object<string, string>} functions Lookup of functions used by the style.
   * @property {number} [bandCount] Number of bands per pixel.
   * @property {Array<PaletteTexture>} [paletteTextures] List of palettes used by the style.
   */

  /**
   * Will return the number as a float with a dot separator, which is required by GLSL.
   * @param {number} v Numerical value.
   * @return {string} The value as string.
   */
  function numberToGlsl(v) {
    const s = v.toString();
    return s.includes('.') ? s : s + '.0';
  }

  /**
   * Will return the number array as a float with a dot separator, concatenated with ', '.
   * @param {Array<number>} array Numerical values array.
   * @return {string} The array as a vector, e. g.: `vec3(1.0, 2.0, 3.0)`.
   */
  function arrayToGlsl(array) {
    if (array.length < 2 || array.length > 4) {
      throw new Error(
        '`formatArray` can only output `vec2`, `vec3` or `vec4` arrays.'
      );
    }
    return `vec${array.length}(${array.map(numberToGlsl).join(', ')})`;
  }

  /**
   * Will normalize and converts to string a `vec4` color array compatible with GLSL.
   * @param {string|import("../color.js").Color} color Color either in string format or [r, g, b, a] array format,
   * with RGB components in the 0..255 range and the alpha component in the 0..1 range.
   * Note that the final array will always have 4 components.
   * @return {string} The color expressed in the `vec4(1.0, 1.0, 1.0, 1.0)` form.
   */
  function colorToGlsl(color) {
    const array = asArray(color).slice();
    if (array.length < 4) {
      array.push(1);
    }
    return arrayToGlsl(
      array.map(function (c, i) {
        return i < 3 ? c / 255 : c;
      })
    );
  }

  /**
   * Returns a stable equivalent number for the string literal.
   * @param {ParsingContext} context Parsing context
   * @param {string} string String literal value
   * @return {number} Number equivalent
   */
  function getStringNumberEquivalent(context, string) {
    if (context.stringLiteralsMap[string] === undefined) {
      context.stringLiteralsMap[string] = Object.keys(
        context.stringLiteralsMap
      ).length;
    }
    return context.stringLiteralsMap[string];
  }

  /**
   * Returns a stable equivalent number for the string literal, for use in shaders. This number is then
   * converted to be a GLSL-compatible string.
   * @param {ParsingContext} context Parsing context
   * @param {string} string String literal value
   * @return {string} GLSL-compatible string containing a number
   */
  function stringToGlsl(context, string) {
    return numberToGlsl(getStringNumberEquivalent(context, string));
  }

  /**
   * Recursively parses a style expression and outputs a GLSL-compatible string. Takes in a parsing context that
   * will be read and modified during the parsing operation.
   * @param {ParsingContext} context Parsing context
   * @param {ExpressionValue} value Value
   * @param {ValueTypes|number} [typeHint] Hint for the expected final type (can be several types combined)
   * @return {string} GLSL-compatible output
   */
  function expressionToGlsl(context, value, typeHint) {
    // operator
    if (Array.isArray(value) && typeof value[0] === 'string') {
      const operator = Operators[value[0]];
      if (operator === undefined) {
        throw new Error(
          `Unrecognized expression operator: ${JSON.stringify(value)}`
        );
      }
      return operator.toGlsl(context, value.slice(1), typeHint);
    }

    const valueType = getValueType(value);
    if ((valueType & ValueTypes.NUMBER) > 0) {
      return numberToGlsl(/** @type {number} */ (value));
    }

    if ((valueType & ValueTypes.BOOLEAN) > 0) {
      return value.toString();
    }

    if (
      (valueType & ValueTypes.STRING) > 0 &&
      (typeHint === undefined || typeHint == ValueTypes.STRING)
    ) {
      return stringToGlsl(context, value.toString());
    }

    if (
      (valueType & ValueTypes.COLOR) > 0 &&
      (typeHint === undefined || typeHint == ValueTypes.COLOR)
    ) {
      return colorToGlsl(/** @type {Array<number> | string} */ (value));
    }

    if ((valueType & ValueTypes.NUMBER_ARRAY) > 0) {
      return arrayToGlsl(/** @type {Array<number>} */ (value));
    }

    throw new Error(`Unexpected expression ${value} (expected type ${typeHint})`);
  }

  function assertNumber(value) {
    if (!(getValueType(value) & ValueTypes.NUMBER)) {
      throw new Error(
        `A numeric value was expected, got ${JSON.stringify(value)} instead`
      );
    }
  }
  function assertNumbers(values) {
    for (let i = 0; i < values.length; i++) {
      assertNumber(values[i]);
    }
  }
  function assertString(value) {
    if (!(getValueType(value) & ValueTypes.STRING)) {
      throw new Error(
        `A string value was expected, got ${JSON.stringify(value)} instead`
      );
    }
  }
  function assertBoolean(value) {
    if (!(getValueType(value) & ValueTypes.BOOLEAN)) {
      throw new Error(
        `A boolean value was expected, got ${JSON.stringify(value)} instead`
      );
    }
  }
  function assertArgsCount(args, count) {
    if (args.length !== count) {
      throw new Error(
        `Exactly ${count} arguments were expected, got ${args.length} instead`
      );
    }
  }
  function assertArgsMinCount(args, count) {
    if (args.length < count) {
      throw new Error(
        `At least ${count} arguments were expected, got ${args.length} instead`
      );
    }
  }
  function assertArgsMaxCount(args, count) {
    if (args.length > count) {
      throw new Error(
        `At most ${count} arguments were expected, got ${args.length} instead`
      );
    }
  }
  function assertArgsEven(args) {
    if (args.length % 2 !== 0) {
      throw new Error(
        `An even amount of arguments was expected, got ${args} instead`
      );
    }
  }
  function assertArgsOdd(args) {
    if (args.length % 2 === 0) {
      throw new Error(
        `An odd amount of arguments was expected, got ${args} instead`
      );
    }
  }
  function assertUniqueInferredType(args, types) {
    if (!isTypeUnique(types)) {
      throw new Error(
        `Could not infer only one type from the following expression: ${JSON.stringify(
        args
      )}`
      );
    }
  }

  Operators['get'] = {
    getReturnType: function (args) {
      return ValueTypes.ANY;
    },
    toGlsl: function (context, args) {
      assertArgsCount(args, 1);
      assertString(args[0]);
      const value = args[0].toString();
      if (!context.attributes.includes(value)) {
        context.attributes.push(value);
      }
      const prefix = context.inFragmentShader ? 'v_' : 'a_';
      return prefix + value;
    },
  };

  /**
   * Get the uniform name given a variable name.
   * @param {string} variableName The variable name.
   * @return {string} The uniform name.
   */
  function uniformNameForVariable(variableName) {
    return 'u_var_' + variableName;
  }

  Operators['var'] = {
    getReturnType: function (args) {
      return ValueTypes.ANY;
    },
    toGlsl: function (context, args) {
      assertArgsCount(args, 1);
      assertString(args[0]);
      const value = args[0].toString();
      if (!context.variables.includes(value)) {
        context.variables.push(value);
      }
      return uniformNameForVariable(value);
    },
  };

  const PALETTE_TEXTURE_ARRAY = 'u_paletteTextures';

  // ['palette', index, colors]
  Operators['palette'] = {
    getReturnType: function (args) {
      return ValueTypes.COLOR;
    },
    toGlsl: function (context, args) {
      assertArgsCount(args, 2);
      assertNumber(args[0]);
      const index = expressionToGlsl(context, args[0]);
      const colors = args[1];
      if (!Array.isArray(colors)) {
        throw new Error('The second argument of palette must be an array');
      }
      const numColors = colors.length;
      const palette = new Uint8Array(numColors * 4);
      for (let i = 0; i < numColors; i++) {
        const candidate = colors[i];
        /**
         * @type {import('../color.js').Color}
         */
        let color;
        if (typeof candidate === 'string') {
          color = fromString(candidate);
        } else {
          if (!Array.isArray(candidate)) {
            throw new Error(
              'The second argument of palette must be an array of strings or colors'
            );
          }
          const length = candidate.length;
          if (length === 4) {
            color = candidate;
          } else {
            if (length !== 3) {
              throw new Error(
                `Expected palette color to have 3 or 4 values, got ${length}`
              );
            }
            color = [candidate[0], candidate[1], candidate[2], 1];
          }
        }
        const offset = i * 4;
        palette[offset] = color[0];
        palette[offset + 1] = color[1];
        palette[offset + 2] = color[2];
        palette[offset + 3] = color[3] * 255;
      }
      if (!context.paletteTextures) {
        context.paletteTextures = [];
      }
      const paletteName = `${PALETTE_TEXTURE_ARRAY}[${context.paletteTextures.length}]`;
      const paletteTexture = new PaletteTexture(paletteName, palette);
      context.paletteTextures.push(paletteTexture);
      return `texture2D(${paletteName}, vec2((${index} + 0.5) / ${numColors}.0, 0.5))`;
    },
  };

  const GET_BAND_VALUE_FUNC = 'getBandValue';

  Operators['band'] = {
    getReturnType: function (args) {
      return ValueTypes.NUMBER;
    },
    toGlsl: function (context, args) {
      assertArgsMinCount(args, 1);
      assertArgsMaxCount(args, 3);
      const band = args[0];

      if (!(GET_BAND_VALUE_FUNC in context.functions)) {
        let ifBlocks = '';
        const bandCount = context.bandCount || 1;
        for (let i = 0; i < bandCount; i++) {
          const colorIndex = Math.floor(i / 4);
          let bandIndex = i % 4;
          if (i === bandCount - 1 && bandIndex === 1) {
            // LUMINANCE_ALPHA - band 1 assigned to rgb and band 2 assigned to alpha
            bandIndex = 3;
          }
          const textureName = `${Uniforms.TILE_TEXTURE_ARRAY}[${colorIndex}]`;
          ifBlocks += `
          if (band == ${i + 1}.0) {
            return texture2D(${textureName}, v_textureCoord + vec2(dx, dy))[${bandIndex}];
          }
        `;
        }

        context.functions[GET_BAND_VALUE_FUNC] = `
        float getBandValue(float band, float xOffset, float yOffset) {
          float dx = xOffset / ${Uniforms.TEXTURE_PIXEL_WIDTH};
          float dy = yOffset / ${Uniforms.TEXTURE_PIXEL_HEIGHT};
          ${ifBlocks}
        }
      `;
      }

      const bandExpression = expressionToGlsl(context, band);
      const xOffsetExpression = expressionToGlsl(context, args[1] || 0);
      const yOffsetExpression = expressionToGlsl(context, args[2] || 0);
      return `${GET_BAND_VALUE_FUNC}(${bandExpression}, ${xOffsetExpression}, ${yOffsetExpression})`;
    },
  };

  Operators['time'] = {
    getReturnType: function (args) {
      return ValueTypes.NUMBER;
    },
    toGlsl: function (context, args) {
      assertArgsCount(args, 0);
      return 'u_time';
    },
  };

  Operators['zoom'] = {
    getReturnType: function (args) {
      return ValueTypes.NUMBER;
    },
    toGlsl: function (context, args) {
      assertArgsCount(args, 0);
      return 'u_zoom';
    },
  };

  Operators['resolution'] = {
    getReturnType: function (args) {
      return ValueTypes.NUMBER;
    },
    toGlsl: function (context, args) {
      assertArgsCount(args, 0);
      return 'u_resolution';
    },
  };

  Operators['*'] = {
    getReturnType: function (args) {
      return ValueTypes.NUMBER;
    },
    toGlsl: function (context, args) {
      assertArgsCount(args, 2);
      assertNumbers(args);
      return `(${expressionToGlsl(context, args[0])} * ${expressionToGlsl(
      context,
      args[1]
    )})`;
    },
  };

  Operators['/'] = {
    getReturnType: function (args) {
      return ValueTypes.NUMBER;
    },
    toGlsl: function (context, args) {
      assertArgsCount(args, 2);
      assertNumbers(args);
      return `(${expressionToGlsl(context, args[0])} / ${expressionToGlsl(
      context,
      args[1]
    )})`;
    },
  };

  Operators['+'] = {
    getReturnType: function (args) {
      return ValueTypes.NUMBER;
    },
    toGlsl: function (context, args) {
      assertArgsCount(args, 2);
      assertNumbers(args);
      return `(${expressionToGlsl(context, args[0])} + ${expressionToGlsl(
      context,
      args[1]
    )})`;
    },
  };

  Operators['-'] = {
    getReturnType: function (args) {
      return ValueTypes.NUMBER;
    },
    toGlsl: function (context, args) {
      assertArgsCount(args, 2);
      assertNumbers(args);
      return `(${expressionToGlsl(context, args[0])} - ${expressionToGlsl(
      context,
      args[1]
    )})`;
    },
  };

  Operators['clamp'] = {
    getReturnType: function (args) {
      return ValueTypes.NUMBER;
    },
    toGlsl: function (context, args) {
      assertArgsCount(args, 3);
      assertNumbers(args);
      const min = expressionToGlsl(context, args[1]);
      const max = expressionToGlsl(context, args[2]);
      return `clamp(${expressionToGlsl(context, args[0])}, ${min}, ${max})`;
    },
  };

  Operators['%'] = {
    getReturnType: function (args) {
      return ValueTypes.NUMBER;
    },
    toGlsl: function (context, args) {
      assertArgsCount(args, 2);
      assertNumbers(args);
      return `mod(${expressionToGlsl(context, args[0])}, ${expressionToGlsl(
      context,
      args[1]
    )})`;
    },
  };

  Operators['^'] = {
    getReturnType: function (args) {
      return ValueTypes.NUMBER;
    },
    toGlsl: function (context, args) {
      assertArgsCount(args, 2);
      assertNumbers(args);
      return `pow(${expressionToGlsl(context, args[0])}, ${expressionToGlsl(
      context,
      args[1]
    )})`;
    },
  };

  Operators['abs'] = {
    getReturnType: function (args) {
      return ValueTypes.NUMBER;
    },
    toGlsl: function (context, args) {
      assertArgsCount(args, 1);
      assertNumbers(args);
      return `abs(${expressionToGlsl(context, args[0])})`;
    },
  };

  Operators['floor'] = {
    getReturnType: function (args) {
      return ValueTypes.NUMBER;
    },
    toGlsl: function (context, args) {
      assertArgsCount(args, 1);
      assertNumbers(args);
      return `floor(${expressionToGlsl(context, args[0])})`;
    },
  };

  Operators['round'] = {
    getReturnType: function (args) {
      return ValueTypes.NUMBER;
    },
    toGlsl: function (context, args) {
      assertArgsCount(args, 1);
      assertNumbers(args);
      return `floor(${expressionToGlsl(context, args[0])} + 0.5)`;
    },
  };

  Operators['ceil'] = {
    getReturnType: function (args) {
      return ValueTypes.NUMBER;
    },
    toGlsl: function (context, args) {
      assertArgsCount(args, 1);
      assertNumbers(args);
      return `ceil(${expressionToGlsl(context, args[0])})`;
    },
  };

  Operators['sin'] = {
    getReturnType: function (args) {
      return ValueTypes.NUMBER;
    },
    toGlsl: function (context, args) {
      assertArgsCount(args, 1);
      assertNumbers(args);
      return `sin(${expressionToGlsl(context, args[0])})`;
    },
  };

  Operators['cos'] = {
    getReturnType: function (args) {
      return ValueTypes.NUMBER;
    },
    toGlsl: function (context, args) {
      assertArgsCount(args, 1);
      assertNumbers(args);
      return `cos(${expressionToGlsl(context, args[0])})`;
    },
  };

  Operators['atan'] = {
    getReturnType: function (args) {
      return ValueTypes.NUMBER;
    },
    toGlsl: function (context, args) {
      assertArgsMinCount(args, 1);
      assertArgsMaxCount(args, 2);
      assertNumbers(args);
      return args.length === 2
        ? `atan(${expressionToGlsl(context, args[0])}, ${expressionToGlsl(
          context,
          args[1]
        )})`
        : `atan(${expressionToGlsl(context, args[0])})`;
    },
  };

  Operators['>'] = {
    getReturnType: function (args) {
      return ValueTypes.BOOLEAN;
    },
    toGlsl: function (context, args) {
      assertArgsCount(args, 2);
      assertNumbers(args);
      return `(${expressionToGlsl(context, args[0])} > ${expressionToGlsl(
      context,
      args[1]
    )})`;
    },
  };

  Operators['>='] = {
    getReturnType: function (args) {
      return ValueTypes.BOOLEAN;
    },
    toGlsl: function (context, args) {
      assertArgsCount(args, 2);
      assertNumbers(args);
      return `(${expressionToGlsl(context, args[0])} >= ${expressionToGlsl(
      context,
      args[1]
    )})`;
    },
  };

  Operators['<'] = {
    getReturnType: function (args) {
      return ValueTypes.BOOLEAN;
    },
    toGlsl: function (context, args) {
      assertArgsCount(args, 2);
      assertNumbers(args);
      return `(${expressionToGlsl(context, args[0])} < ${expressionToGlsl(
      context,
      args[1]
    )})`;
    },
  };

  Operators['<='] = {
    getReturnType: function (args) {
      return ValueTypes.BOOLEAN;
    },
    toGlsl: function (context, args) {
      assertArgsCount(args, 2);
      assertNumbers(args);
      return `(${expressionToGlsl(context, args[0])} <= ${expressionToGlsl(
      context,
      args[1]
    )})`;
    },
  };

  function getEqualOperator(operator) {
    return {
      getReturnType: function (args) {
        return ValueTypes.BOOLEAN;
      },
      toGlsl: function (context, args) {
        assertArgsCount(args, 2);

        // find common type
        let type = ValueTypes.ANY;
        for (let i = 0; i < args.length; i++) {
          type &= getValueType(args[i]);
        }
        if (type === ValueTypes.NONE) {
          throw new Error(
            `All arguments should be of compatible type, got ${JSON.stringify(
            args
          )} instead`
          );
        }

        // Since it's not possible to have color types here, we can leave it out
        // This fixes issues in case the value type is ambiguously detected as a color (e.g. the string 'red')
        type &= ~ValueTypes.COLOR;

        return `(${expressionToGlsl(
        context,
        args[0],
        type
      )} ${operator} ${expressionToGlsl(context, args[1], type)})`;
      },
    };
  }

  Operators['=='] = getEqualOperator('==');

  Operators['!='] = getEqualOperator('!=');

  Operators['!'] = {
    getReturnType: function (args) {
      return ValueTypes.BOOLEAN;
    },
    toGlsl: function (context, args) {
      assertArgsCount(args, 1);
      assertBoolean(args[0]);
      return `(!${expressionToGlsl(context, args[0])})`;
    },
  };

  function getDecisionOperator(operator) {
    return {
      getReturnType: function (args) {
        return ValueTypes.BOOLEAN;
      },
      toGlsl: function (context, args) {
        assertArgsMinCount(args, 2);
        for (let i = 0; i < args.length; i++) {
          assertBoolean(args[i]);
        }
        let result = '';
        result = args
          .map((arg) => expressionToGlsl(context, arg))
          .join(` ${operator} `);
        result = `(${result})`;
        return result;
      },
    };
  }

  Operators['all'] = getDecisionOperator('&&');

  Operators['any'] = getDecisionOperator('||');

  Operators['between'] = {
    getReturnType: function (args) {
      return ValueTypes.BOOLEAN;
    },
    toGlsl: function (context, args) {
      assertArgsCount(args, 3);
      assertNumbers(args);
      const min = expressionToGlsl(context, args[1]);
      const max = expressionToGlsl(context, args[2]);
      const value = expressionToGlsl(context, args[0]);
      return `(${value} >= ${min} && ${value} <= ${max})`;
    },
  };

  Operators['array'] = {
    getReturnType: function (args) {
      return ValueTypes.NUMBER_ARRAY;
    },
    toGlsl: function (context, args) {
      assertArgsMinCount(args, 2);
      assertArgsMaxCount(args, 4);
      assertNumbers(args);
      const parsedArgs = args.map(function (val) {
        return expressionToGlsl(context, val, ValueTypes.NUMBER);
      });
      return `vec${args.length}(${parsedArgs.join(', ')})`;
    },
  };

  Operators['color'] = {
    getReturnType: function (args) {
      return ValueTypes.COLOR;
    },
    toGlsl: function (context, args) {
      assertArgsMinCount(args, 3);
      assertArgsMaxCount(args, 4);
      assertNumbers(args);
      const array = /** @type {Array<number>} */ (args);
      if (args.length === 3) {
        array.push(1);
      }
      const parsedArgs = args.map(function (val, i) {
        return (
          expressionToGlsl(context, val, ValueTypes.NUMBER) +
          (i < 3 ? ' / 255.0' : '')
        );
      });
      return `vec${args.length}(${parsedArgs.join(', ')})`;
    },
  };

  Operators['interpolate'] = {
    getReturnType: function (args) {
      let type = ValueTypes.COLOR | ValueTypes.NUMBER;
      for (let i = 3; i < args.length; i += 2) {
        type = type & getValueType(args[i]);
      }
      return type;
    },
    toGlsl: function (context, args, typeHint) {
      assertArgsEven(args);
      assertArgsMinCount(args, 6);

      // validate interpolation type
      const type = args[0];
      let interpolation;
      switch (type[0]) {
        case 'linear':
          interpolation = 1;
          break;
        case 'exponential':
          interpolation = type[1];
          break;
        default:
          interpolation = null;
      }
      if (!interpolation) {
        throw new Error(
          `Invalid interpolation type for "interpolate" operator, received: ${JSON.stringify(
          type
        )}`
        );
      }

      // compute input/output types
      typeHint = typeHint !== undefined ? typeHint : ValueTypes.ANY;
      const outputType = Operators['interpolate'].getReturnType(args) & typeHint;
      assertUniqueInferredType(args, outputType);

      const input = expressionToGlsl(context, args[1]);
      const exponent = numberToGlsl(interpolation);

      let result = '';
      for (let i = 2; i < args.length - 2; i += 2) {
        const stop1 = expressionToGlsl(context, args[i]);
        const output1 =
          result || expressionToGlsl(context, args[i + 1], outputType);
        const stop2 = expressionToGlsl(context, args[i + 2]);
        const output2 = expressionToGlsl(context, args[i + 3], outputType);
        result = `mix(${output1}, ${output2}, pow(clamp((${input} - ${stop1}) / (${stop2} - ${stop1}), 0.0, 1.0), ${exponent}))`;
      }
      return result;
    },
  };

  Operators['match'] = {
    getReturnType: function (args) {
      let type = ValueTypes.ANY;
      for (let i = 2; i < args.length; i += 2) {
        type = type & getValueType(args[i]);
      }
      type = type & getValueType(args[args.length - 1]);
      return type;
    },
    toGlsl: function (context, args, typeHint) {
      assertArgsEven(args);
      assertArgsMinCount(args, 4);

      typeHint = typeHint !== undefined ? typeHint : ValueTypes.ANY;
      const outputType = Operators['match'].getReturnType(args) & typeHint;
      assertUniqueInferredType(args, outputType);

      const input = expressionToGlsl(context, args[0]);
      const fallback = expressionToGlsl(
        context,
        args[args.length - 1],
        outputType
      );
      let result = null;
      for (let i = args.length - 3; i >= 1; i -= 2) {
        const match = expressionToGlsl(context, args[i]);
        const output = expressionToGlsl(context, args[i + 1], outputType);
        result = `(${input} == ${match} ? ${output} : ${result || fallback})`;
      }
      return result;
    },
  };

  Operators['case'] = {
    getReturnType: function (args) {
      let type = ValueTypes.ANY;
      for (let i = 1; i < args.length; i += 2) {
        type = type & getValueType(args[i]);
      }
      type = type & getValueType(args[args.length - 1]);
      return type;
    },
    toGlsl: function (context, args, typeHint) {
      assertArgsOdd(args);
      assertArgsMinCount(args, 3);

      typeHint = typeHint !== undefined ? typeHint : ValueTypes.ANY;
      const outputType = Operators['case'].getReturnType(args) & typeHint;
      assertUniqueInferredType(args, outputType);
      for (let i = 0; i < args.length - 1; i += 2) {
        assertBoolean(args[i]);
      }

      const fallback = expressionToGlsl(
        context,
        args[args.length - 1],
        outputType
      );
      let result = null;
      for (let i = args.length - 3; i >= 0; i -= 2) {
        const condition = expressionToGlsl(context, args[i]);
        const output = expressionToGlsl(context, args[i + 1], outputType);
        result = `(${condition} ? ${output} : ${result || fallback})`;
      }
      return result;
    },
  };

  /**
   * Classes and utilities for generating shaders from literal style objects
   * @module ol/webgl/ShaderBuilder
   */

  /**
   * @typedef {Object} VaryingDescription
   * @property {string} name Varying name, as will be declared in the header.
   * @property {string} type Varying type, either `float`, `vec2`, `vec4`...
   * @property {string} expression Expression which will be assigned to the varying in the vertex shader, and
   * passed on to the fragment shader.
   */

  /**
   * @classdesc
   * This class implements a classic builder pattern for generating many different types of shaders.
   * Methods can be chained, e. g.:
   *
   * ```js
   * const shader = new ShaderBuilder()
   *   .addVarying('v_width', 'float', 'a_width')
   *   .addUniform('u_time')
   *   .setColorExpression('...')
   *   .setSizeExpression('...')
   *   .outputSymbolFragmentShader();
   * ```
   */
  class ShaderBuilder {
    constructor() {
      /**
       * Uniforms; these will be declared in the header (should include the type).
       * @type {Array<string>}
       * @private
       */
      this.uniforms = [];

      /**
       * Attributes; these will be declared in the header (should include the type).
       * @type {Array<string>}
       * @private
       */
      this.attributes = [];

      /**
       * Varyings with a name, a type and an expression.
       * @type {Array<VaryingDescription>}
       * @private
       */
      this.varyings = [];

      /**
       * @type {string}
       * @private
       */
      this.sizeExpression = 'vec2(1.0)';

      /**
       * @type {string}
       * @private
       */
      this.rotationExpression = '0.0';

      /**
       * @type {string}
       * @private
       */
      this.offsetExpression = 'vec2(0.0)';

      /**
       * @type {string}
       * @private
       */
      this.colorExpression = 'vec4(1.0)';

      /**
       * @type {string}
       * @private
       */
      this.texCoordExpression = 'vec4(0.0, 0.0, 1.0, 1.0)';

      /**
       * @type {string}
       * @private
       */
      this.discardExpression = 'false';

      /**
       * @type {boolean}
       * @private
       */
      this.rotateWithView = false;
    }

    /**
     * Adds a uniform accessible in both fragment and vertex shaders.
     * The given name should include a type, such as `sampler2D u_texture`.
     * @param {string} name Uniform name
     * @return {ShaderBuilder} the builder object
     */
    addUniform(name) {
      this.uniforms.push(name);
      return this;
    }

    /**
     * Adds an attribute accessible in the vertex shader, read from the geometry buffer.
     * The given name should include a type, such as `vec2 a_position`.
     * @param {string} name Attribute name
     * @return {ShaderBuilder} the builder object
     */
    addAttribute(name) {
      this.attributes.push(name);
      return this;
    }

    /**
     * Adds a varying defined in the vertex shader and accessible from the fragment shader.
     * The type and expression of the varying have to be specified separately.
     * @param {string} name Varying name
     * @param {'float'|'vec2'|'vec3'|'vec4'} type Type
     * @param {string} expression Expression used to assign a value to the varying.
     * @return {ShaderBuilder} the builder object
     */
    addVarying(name, type, expression) {
      this.varyings.push({
        name: name,
        type: type,
        expression: expression,
      });
      return this;
    }

    /**
     * Sets an expression to compute the size of the shape.
     * This expression can use all the uniforms and attributes available
     * in the vertex shader, and should evaluate to a `vec2` value.
     * @param {string} expression Size expression
     * @return {ShaderBuilder} the builder object
     */
    setSizeExpression(expression) {
      this.sizeExpression = expression;
      return this;
    }

    /**
     * Sets an expression to compute the rotation of the shape.
     * This expression can use all the uniforms and attributes available
     * in the vertex shader, and should evaluate to a `float` value in radians.
     * @param {string} expression Size expression
     * @return {ShaderBuilder} the builder object
     */
    setRotationExpression(expression) {
      this.rotationExpression = expression;
      return this;
    }

    /**
     * Sets an expression to compute the offset of the symbol from the point center.
     * This expression can use all the uniforms and attributes available
     * in the vertex shader, and should evaluate to a `vec2` value.
     * Note: will only be used for point geometry shaders.
     * @param {string} expression Offset expression
     * @return {ShaderBuilder} the builder object
     */
    setSymbolOffsetExpression(expression) {
      this.offsetExpression = expression;
      return this;
    }

    /**
     * Sets an expression to compute the color of the shape.
     * This expression can use all the uniforms, varyings and attributes available
     * in the fragment shader, and should evaluate to a `vec4` value.
     * @param {string} expression Color expression
     * @return {ShaderBuilder} the builder object
     */
    setColorExpression(expression) {
      this.colorExpression = expression;
      return this;
    }

    /**
     * Sets an expression to compute the texture coordinates of the vertices.
     * This expression can use all the uniforms and attributes available
     * in the vertex shader, and should evaluate to a `vec4` value.
     * @param {string} expression Texture coordinate expression
     * @return {ShaderBuilder} the builder object
     */
    setTextureCoordinateExpression(expression) {
      this.texCoordExpression = expression;
      return this;
    }

    /**
     * Sets an expression to determine whether a fragment (pixel) should be discarded,
     * i.e. not drawn at all.
     * This expression can use all the uniforms, varyings and attributes available
     * in the fragment shader, and should evaluate to a `bool` value (it will be
     * used in an `if` statement)
     * @param {string} expression Fragment discard expression
     * @return {ShaderBuilder} the builder object
     */
    setFragmentDiscardExpression(expression) {
      this.discardExpression = expression;
      return this;
    }

    /**
     * Sets whether the symbols should rotate with the view or stay aligned with the map.
     * Note: will only be used for point geometry shaders.
     * @param {boolean} rotateWithView Rotate with view
     * @return {ShaderBuilder} the builder object
     */
    setSymbolRotateWithView(rotateWithView) {
      this.rotateWithView = rotateWithView;
      return this;
    }

    /**
     * @return {string} Previously set size expression
     */
    getSizeExpression() {
      return this.sizeExpression;
    }

    /**
     * @return {string} Previously set symbol offset expression
     */
    getOffsetExpression() {
      return this.offsetExpression;
    }

    /**
     * @return {string} Previously set color expression
     */
    getColorExpression() {
      return this.colorExpression;
    }

    /**
     * @return {string} Previously set texture coordinate expression
     */
    getTextureCoordinateExpression() {
      return this.texCoordExpression;
    }

    /**
     * @return {string} Previously set fragment discard expression
     */
    getFragmentDiscardExpression() {
      return this.discardExpression;
    }

    /**
     * Generates a symbol vertex shader from the builder parameters,
     * intended to be used on point geometries.
     *
     * Three uniforms are hardcoded in all shaders: `u_projectionMatrix`, `u_offsetScaleMatrix`,
     * `u_offsetRotateMatrix`, `u_time`.
     *
     * The following attributes are hardcoded and expected to be present in the vertex buffers:
     * `vec2 a_position`, `float a_index` (being the index of the vertex in the quad, 0 to 3).
     *
     * The following varyings are hardcoded and gives the coordinate of the pixel both in the quad and on the texture:
     * `vec2 v_quadCoord`, `vec2 v_texCoord`
     *
     * @param {boolean} [forHitDetection] If true, the shader will be modified to include hit detection variables
     * (namely, hit color with encoded feature id).
     * @return {string} The full shader as a string.
     */
    getSymbolVertexShader(forHitDetection) {
      const offsetMatrix = this.rotateWithView
        ? 'u_offsetScaleMatrix * u_offsetRotateMatrix'
        : 'u_offsetScaleMatrix';

      let attributes = this.attributes;
      let varyings = this.varyings;

      if (forHitDetection) {
        attributes = attributes.concat('vec4 a_hitColor');
        varyings = varyings.concat({
          name: 'v_hitColor',
          type: 'vec4',
          expression: 'a_hitColor',
        });
      }

      return `precision mediump float;
uniform mat4 u_projectionMatrix;
uniform mat4 u_offsetScaleMatrix;
uniform mat4 u_offsetRotateMatrix;
uniform float u_time;
uniform float u_zoom;
uniform float u_resolution;
${this.uniforms
  .map(function (uniform) {
    return 'uniform ' + uniform + ';';
  })
  .join('\n')}
attribute vec2 a_position;
attribute float a_index;
${attributes
  .map(function (attribute) {
    return 'attribute ' + attribute + ';';
  })
  .join('\n')}
varying vec2 v_texCoord;
varying vec2 v_quadCoord;
${varyings
  .map(function (varying) {
    return 'varying ' + varying.type + ' ' + varying.name + ';';
  })
  .join('\n')}
void main(void) {
  mat4 offsetMatrix = ${offsetMatrix};
  vec2 halfSize = ${this.sizeExpression} * 0.5;
  vec2 offset = ${this.offsetExpression};
  float angle = ${this.rotationExpression};
  float offsetX;
  float offsetY;
  if (a_index == 0.0) {
    offsetX = (offset.x - halfSize.x) * cos(angle) + (offset.y - halfSize.y) * sin(angle);
    offsetY = (offset.y - halfSize.y) * cos(angle) - (offset.x - halfSize.x) * sin(angle);
  } else if (a_index == 1.0) {
    offsetX = (offset.x + halfSize.x) * cos(angle) + (offset.y - halfSize.y) * sin(angle);
    offsetY = (offset.y - halfSize.y) * cos(angle) - (offset.x + halfSize.x) * sin(angle);
  } else if (a_index == 2.0) {
    offsetX = (offset.x + halfSize.x) * cos(angle) + (offset.y + halfSize.y) * sin(angle);
    offsetY = (offset.y + halfSize.y) * cos(angle) - (offset.x + halfSize.x) * sin(angle);
  } else {
    offsetX = (offset.x - halfSize.x) * cos(angle) + (offset.y + halfSize.y) * sin(angle);
    offsetY = (offset.y + halfSize.y) * cos(angle) - (offset.x - halfSize.x) * sin(angle);
  }
  vec4 offsets = offsetMatrix * vec4(offsetX, offsetY, 0.0, 0.0);
  gl_Position = u_projectionMatrix * vec4(a_position, 0.0, 1.0) + offsets;
  vec4 texCoord = ${this.texCoordExpression};
  float u = a_index == 0.0 || a_index == 3.0 ? texCoord.s : texCoord.p;
  float v = a_index == 2.0 || a_index == 3.0 ? texCoord.t : texCoord.q;
  v_texCoord = vec2(u, v);
  u = a_index == 0.0 || a_index == 3.0 ? 0.0 : 1.0;
  v = a_index == 2.0 || a_index == 3.0 ? 0.0 : 1.0;
  v_quadCoord = vec2(u, v);
${varyings
  .map(function (varying) {
    return '  ' + varying.name + ' = ' + varying.expression + ';';
  })
  .join('\n')}
}`;
    }

    /**
     * Generates a symbol fragment shader from the builder parameters,
     * intended to be used on point geometries.
     *
     * Expects the following varyings to be transmitted by the vertex shader:
     * `vec2 v_quadCoord`, `vec2 v_texCoord`
     *
     * @param {boolean} [forHitDetection] If true, the shader will be modified to include hit detection variables
     * (namely, hit color with encoded feature id).
     * @return {string} The full shader as a string.
     */
    getSymbolFragmentShader(forHitDetection) {
      const hitDetectionBypass = forHitDetection
        ? '  if (gl_FragColor.a < 0.1) { discard; } gl_FragColor = v_hitColor;'
        : '';

      let varyings = this.varyings;

      if (forHitDetection) {
        varyings = varyings.concat({
          name: 'v_hitColor',
          type: 'vec4',
          expression: 'a_hitColor',
        });
      }

      return `precision mediump float;
uniform float u_time;
uniform float u_zoom;
uniform float u_resolution;
${this.uniforms
  .map(function (uniform) {
    return 'uniform ' + uniform + ';';
  })
  .join('\n')}
varying vec2 v_texCoord;
varying vec2 v_quadCoord;
${varyings
  .map(function (varying) {
    return 'varying ' + varying.type + ' ' + varying.name + ';';
  })
  .join('\n')}
void main(void) {
  if (${this.discardExpression}) { discard; }
  gl_FragColor = ${this.colorExpression};
  gl_FragColor.rgb *= gl_FragColor.a;
${hitDetectionBypass}
}`;
    }
  }

  /**
   * @typedef {Object} StyleParseResult
   * @property {ShaderBuilder} builder Shader builder pre-configured according to a given style
   * @property {Object<string,import("./Helper").UniformValue>} uniforms Uniform definitions.
   * @property {Array<import("../renderer/webgl/PointsLayer").CustomAttribute>} attributes Attribute descriptions.
   */

  /**
   * Parses a {@link import("../style/literal").LiteralStyle} object and returns a {@link ShaderBuilder}
   * object that has been configured according to the given style, as well as `attributes` and `uniforms`
   * arrays to be fed to the `WebGLPointsRenderer` class.
   *
   * Also returns `uniforms` and `attributes` properties as expected by the
   * {@link module:ol/renderer/webgl/PointsLayer~WebGLPointsLayerRenderer}.
   *
   * @param {import("../style/literal").LiteralStyle} style Literal style.
   * @return {StyleParseResult} Result containing shader params, attributes and uniforms.
   */
  function parseLiteralStyle(style) {
    const symbStyle = style.symbol;
    const size = symbStyle.size !== undefined ? symbStyle.size : 1;
    const color = symbStyle.color || 'white';
    const texCoord = symbStyle.textureCoord || [0, 0, 1, 1];
    const offset = symbStyle.offset || [0, 0];
    const opacity = symbStyle.opacity !== undefined ? symbStyle.opacity : 1;
    const rotation = symbStyle.rotation !== undefined ? symbStyle.rotation : 0;

    /**
     * @type {import("../style/expressions.js").ParsingContext}
     */
    const vertContext = {
      inFragmentShader: false,
      variables: [],
      attributes: [],
      stringLiteralsMap: {},
      functions: {},
    };
    const parsedSize = expressionToGlsl(
      vertContext,
      size,
      ValueTypes.NUMBER_ARRAY | ValueTypes.NUMBER
    );
    const parsedOffset = expressionToGlsl(
      vertContext,
      offset,
      ValueTypes.NUMBER_ARRAY
    );
    const parsedTexCoord = expressionToGlsl(
      vertContext,
      texCoord,
      ValueTypes.NUMBER_ARRAY
    );
    const parsedRotation = expressionToGlsl(
      vertContext,
      rotation,
      ValueTypes.NUMBER
    );

    /**
     * @type {import("../style/expressions.js").ParsingContext}
     */
    const fragContext = {
      inFragmentShader: true,
      variables: vertContext.variables,
      attributes: [],
      stringLiteralsMap: vertContext.stringLiteralsMap,
      functions: {},
    };
    const parsedColor = expressionToGlsl(fragContext, color, ValueTypes.COLOR);
    const parsedOpacity = expressionToGlsl(
      fragContext,
      opacity,
      ValueTypes.NUMBER
    );

    let opacityFilter = '1.0';
    const visibleSize = `vec2(${expressionToGlsl(
    fragContext,
    size,
    ValueTypes.NUMBER_ARRAY | ValueTypes.NUMBER
  )}).x`;
    switch (symbStyle.symbolType) {
      case 'square':
        break;
      case 'image':
        break;
      // taken from https://thebookofshaders.com/07/
      case 'circle':
        opacityFilter = `(1.0-smoothstep(1.-4./${visibleSize},1.,dot(v_quadCoord-.5,v_quadCoord-.5)*4.))`;
        break;
      case 'triangle':
        const st = '(v_quadCoord*2.-1.)';
        const a = `(atan(${st}.x,${st}.y))`;
        opacityFilter = `(1.0-smoothstep(.5-3./${visibleSize},.5,cos(floor(.5+${a}/2.094395102)*2.094395102-${a})*length(${st})))`;
        break;

      default:
        throw new Error('Unexpected symbol type: ' + symbStyle.symbolType);
    }

    const builder = new ShaderBuilder()
      .setSizeExpression(`vec2(${parsedSize})`)
      .setRotationExpression(parsedRotation)
      .setSymbolOffsetExpression(parsedOffset)
      .setTextureCoordinateExpression(parsedTexCoord)
      .setSymbolRotateWithView(!!symbStyle.rotateWithView)
      .setColorExpression(
        `vec4(${parsedColor}.rgb, ${parsedColor}.a * ${parsedOpacity} * ${opacityFilter})`
      );

    if (style.filter) {
      const parsedFilter = expressionToGlsl(
        fragContext,
        style.filter,
        ValueTypes.BOOLEAN
      );
      builder.setFragmentDiscardExpression(`!${parsedFilter}`);
    }

    /** @type {Object<string,import("../webgl/Helper").UniformValue>} */
    const uniforms = {};

    // define one uniform per variable
    fragContext.variables.forEach(function (varName) {
      const uniformName = uniformNameForVariable(varName);
      builder.addUniform(`float ${uniformName}`);
      uniforms[uniformName] = function () {
        if (!style.variables || style.variables[varName] === undefined) {
          throw new Error(
            `The following variable is missing from the style: ${varName}`
          );
        }
        let value = style.variables[varName];
        if (typeof value === 'string') {
          value = getStringNumberEquivalent(vertContext, value);
        }
        return value !== undefined ? value : -9999999; // to avoid matching with the first string literal
      };
    });

    if (symbStyle.symbolType === 'image' && symbStyle.src) {
      const texture = new Image();
      texture.crossOrigin =
        symbStyle.crossOrigin === undefined ? 'anonymous' : symbStyle.crossOrigin;
      texture.src = symbStyle.src;
      builder
        .addUniform('sampler2D u_texture')
        .setColorExpression(
          builder.getColorExpression() + ' * texture2D(u_texture, v_texCoord)'
        );
      uniforms['u_texture'] = texture;
    }

    // for each feature attribute used in the fragment shader, define a varying that will be used to pass data
    // from the vertex to the fragment shader, as well as an attribute in the vertex shader (if not already present)
    fragContext.attributes.forEach(function (attrName) {
      if (!vertContext.attributes.includes(attrName)) {
        vertContext.attributes.push(attrName);
      }
      builder.addVarying(`v_${attrName}`, 'float', `a_${attrName}`);
    });

    // for each feature attribute used in the vertex shader, define an attribute in the vertex shader.
    vertContext.attributes.forEach(function (attrName) {
      builder.addAttribute(`float a_${attrName}`);
    });

    return {
      builder: builder,
      attributes: vertContext.attributes.map(function (attributeName) {
        return {
          name: attributeName,
          callback: function (feature, props) {
            let value = props[attributeName];
            if (typeof value === 'string') {
              value = getStringNumberEquivalent(vertContext, value);
            }
            return value !== undefined ? value : -9999999; // to avoid matching with the first string literal
          },
        };
      }),
      uniforms: uniforms,
    };
  }

  /**
   * @module ol/style/Image
   */

  /**
   * @typedef {Object} Options
   * @property {number} opacity Opacity.
   * @property {boolean} rotateWithView If the image should get rotated with the view.
   * @property {number} rotation Rotation.
   * @property {number|import("../size.js").Size} scale Scale.
   * @property {Array<number>} displacement Displacement.
   * @property {"declutter"|"obstacle"|"none"|undefined} declutterMode Declutter mode: `declutter`, `obstacle`, 'none */

  /**
   * @classdesc
   * A base class used for creating subclasses and not instantiated in
   * apps. Base class for {@link module:ol/style/Icon~Icon}, {@link module:ol/style/Circle~CircleStyle} and
   * {@link module:ol/style/RegularShape~RegularShape}.
   * @abstract
   * @api
   */
  class ImageStyle {
    /**
     * @param {Options} options Options.
     */
    constructor(options) {
      /**
       * @private
       * @type {number}
       */
      this.opacity_ = options.opacity;

      /**
       * @private
       * @type {boolean}
       */
      this.rotateWithView_ = options.rotateWithView;

      /**
       * @private
       * @type {number}
       */
      this.rotation_ = options.rotation;

      /**
       * @private
       * @type {number|import("../size.js").Size}
       */
      this.scale_ = options.scale;

      /**
       * @private
       * @type {import("../size.js").Size}
       */
      this.scaleArray_ = toSize(options.scale);

      /**
       * @private
       * @type {Array<number>}
       */
      this.displacement_ = options.displacement;

      /**
       * @private
       * @type {"declutter"|"obstacle"|"none"|undefined}
       */
      this.declutterMode_ = options.declutterMode;
    }

    /**
     * Clones the style.
     * @return {ImageStyle} The cloned style.
     * @api
     */
    clone() {
      const scale = this.getScale();
      return new ImageStyle({
        opacity: this.getOpacity(),
        scale: Array.isArray(scale) ? scale.slice() : scale,
        rotation: this.getRotation(),
        rotateWithView: this.getRotateWithView(),
        displacement: this.getDisplacement().slice(),
        declutterMode: this.getDeclutterMode(),
      });
    }

    /**
     * Get the symbolizer opacity.
     * @return {number} Opacity.
     * @api
     */
    getOpacity() {
      return this.opacity_;
    }

    /**
     * Determine whether the symbolizer rotates with the map.
     * @return {boolean} Rotate with map.
     * @api
     */
    getRotateWithView() {
      return this.rotateWithView_;
    }

    /**
     * Get the symoblizer rotation.
     * @return {number} Rotation.
     * @api
     */
    getRotation() {
      return this.rotation_;
    }

    /**
     * Get the symbolizer scale.
     * @return {number|import("../size.js").Size} Scale.
     * @api
     */
    getScale() {
      return this.scale_;
    }

    /**
     * Get the symbolizer scale array.
     * @return {import("../size.js").Size} Scale array.
     */
    getScaleArray() {
      return this.scaleArray_;
    }

    /**
     * Get the displacement of the shape
     * @return {Array<number>} Shape's center displacement
     * @api
     */
    getDisplacement() {
      return this.displacement_;
    }

    /**
     * Get the declutter mode of the shape
     * @return {"declutter"|"obstacle"|"none"|undefined} Shape's declutter mode
     * @api
     */
    getDeclutterMode() {
      return this.declutterMode_;
    }

    /**
     * Get the anchor point in pixels. The anchor determines the center point for the
     * symbolizer.
     * @abstract
     * @return {Array<number>} Anchor.
     */
    getAnchor() {
      return abstract();
    }

    /**
     * Get the image element for the symbolizer.
     * @abstract
     * @param {number} pixelRatio Pixel ratio.
     * @return {HTMLCanvasElement|HTMLVideoElement|HTMLImageElement} Image element.
     */
    getImage(pixelRatio) {
      return abstract();
    }

    /**
     * @abstract
     * @return {HTMLCanvasElement|HTMLVideoElement|HTMLImageElement} Image element.
     */
    getHitDetectionImage() {
      return abstract();
    }

    /**
     * Get the image pixel ratio.
     * @param {number} pixelRatio Pixel ratio.
     * @return {number} Pixel ratio.
     */
    getPixelRatio(pixelRatio) {
      return 1;
    }

    /**
     * @abstract
     * @return {import("../ImageState.js").default} Image state.
     */
    getImageState() {
      return abstract();
    }

    /**
     * @abstract
     * @return {import("../size.js").Size} Image size.
     */
    getImageSize() {
      return abstract();
    }

    /**
     * Get the origin of the symbolizer.
     * @abstract
     * @return {Array<number>} Origin.
     */
    getOrigin() {
      return abstract();
    }

    /**
     * Get the size of the symbolizer (in pixels).
     * @abstract
     * @return {import("../size.js").Size} Size.
     */
    getSize() {
      return abstract();
    }

    /**
     * Set the displacement.
     *
     * @param {Array<number>} displacement Displacement.
     * @api
     */
    setDisplacement(displacement) {
      this.displacement_ = displacement;
    }

    /**
     * Set the opacity.
     *
     * @param {number} opacity Opacity.
     * @api
     */
    setOpacity(opacity) {
      this.opacity_ = opacity;
    }

    /**
     * Set whether to rotate the style with the view.
     *
     * @param {boolean} rotateWithView Rotate with map.
     * @api
     */
    setRotateWithView(rotateWithView) {
      this.rotateWithView_ = rotateWithView;
    }

    /**
     * Set the rotation.
     *
     * @param {number} rotation Rotation.
     * @api
     */
    setRotation(rotation) {
      this.rotation_ = rotation;
    }
    /**
     * Set the scale.
     *
     * @param {number|import("../size.js").Size} scale Scale.
     * @api
     */
    setScale(scale) {
      this.scale_ = scale;
      this.scaleArray_ = toSize(scale);
    }

    /**
     * @abstract
     * @param {function(import("../events/Event.js").default): void} listener Listener function.
     */
    listenImageChange(listener) {
      abstract();
    }

    /**
     * Load not yet loaded URI.
     * @abstract
     */
    load() {
      abstract();
    }

    /**
     * @abstract
     * @param {function(import("../events/Event.js").default): void} listener Listener function.
     */
    unlistenImageChange(listener) {
      abstract();
    }
  }

  /**
   * @module ol/style/RegularShape
   */

  /**
   * Specify radius for regular polygons, or radius1 and radius2 for stars.
   * @typedef {Object} Options
   * @property {import("./Fill.js").default} [fill] Fill style.
   * @property {number} points Number of points for stars and regular polygons. In case of a polygon, the number of points
   * is the number of sides.
   * @property {number} [radius] Radius of a regular polygon.
   * @property {number} [radius1] First radius of a star. Ignored if radius is set.
   * @property {number} [radius2] Second radius of a star.
   * @property {number} [angle=0] Shape's angle in radians. A value of 0 will have one of the shape's points facing up.
   * @property {Array<number>} [displacement=[0, 0]] Displacement of the shape in pixels.
   * Positive values will shift the shape right and up.
   * @property {import("./Stroke.js").default} [stroke] Stroke style.
   * @property {number} [rotation=0] Rotation in radians (positive rotation clockwise).
   * @property {boolean} [rotateWithView=false] Whether to rotate the shape with the view.
   * @property {number|import("../size.js").Size} [scale=1] Scale. Unless two dimensional scaling is required a better
   * result may be obtained with appropriate settings for `radius`, `radius1` and `radius2`.
   * @property {"declutter"|"obstacle"|"none"|undefined} [declutterMode] Declutter mode.
   */

  /**
   * @typedef {Object} RenderOptions
   * @property {import("../colorlike.js").ColorLike} [strokeStyle] StrokeStyle.
   * @property {number} strokeWidth StrokeWidth.
   * @property {number} size Size.
   * @property {Array<number>|null} lineDash LineDash.
   * @property {number} lineDashOffset LineDashOffset.
   * @property {CanvasLineJoin} lineJoin LineJoin.
   * @property {number} miterLimit MiterLimit.
   */

  /**
   * @classdesc
   * Set regular shape style for vector features. The resulting shape will be
   * a regular polygon when `radius` is provided, or a star when `radius1` and
   * `radius2` are provided.
   * @api
   */
  class RegularShape extends ImageStyle {
    /**
     * @param {Options} options Options.
     */
    constructor(options) {
      /**
       * @type {boolean}
       */
      const rotateWithView =
        options.rotateWithView !== undefined ? options.rotateWithView : false;

      super({
        opacity: 1,
        rotateWithView: rotateWithView,
        rotation: options.rotation !== undefined ? options.rotation : 0,
        scale: options.scale !== undefined ? options.scale : 1,
        displacement:
          options.displacement !== undefined ? options.displacement : [0, 0],
        declutterMode: options.declutterMode,
      });

      /**
       * @private
       * @type {Object<number, HTMLCanvasElement>}
       */
      this.canvas_ = undefined;

      /**
       * @private
       * @type {HTMLCanvasElement}
       */
      this.hitDetectionCanvas_ = null;

      /**
       * @private
       * @type {import("./Fill.js").default}
       */
      this.fill_ = options.fill !== undefined ? options.fill : null;

      /**
       * @private
       * @type {Array<number>}
       */
      this.origin_ = [0, 0];

      /**
       * @private
       * @type {number}
       */
      this.points_ = options.points;

      /**
       * @protected
       * @type {number}
       */
      this.radius_ =
        options.radius !== undefined ? options.radius : options.radius1;

      /**
       * @private
       * @type {number|undefined}
       */
      this.radius2_ = options.radius2;

      /**
       * @private
       * @type {number}
       */
      this.angle_ = options.angle !== undefined ? options.angle : 0;

      /**
       * @private
       * @type {import("./Stroke.js").default}
       */
      this.stroke_ = options.stroke !== undefined ? options.stroke : null;

      /**
       * @private
       * @type {import("../size.js").Size}
       */
      this.size_ = null;

      /**
       * @private
       * @type {RenderOptions}
       */
      this.renderOptions_ = null;

      this.render();
    }

    /**
     * Clones the style.
     * @return {RegularShape} The cloned style.
     * @api
     */
    clone() {
      const scale = this.getScale();
      const style = new RegularShape({
        fill: this.getFill() ? this.getFill().clone() : undefined,
        points: this.getPoints(),
        radius: this.getRadius(),
        radius2: this.getRadius2(),
        angle: this.getAngle(),
        stroke: this.getStroke() ? this.getStroke().clone() : undefined,
        rotation: this.getRotation(),
        rotateWithView: this.getRotateWithView(),
        scale: Array.isArray(scale) ? scale.slice() : scale,
        displacement: this.getDisplacement().slice(),
        declutterMode: this.getDeclutterMode(),
      });
      style.setOpacity(this.getOpacity());
      return style;
    }

    /**
     * Get the anchor point in pixels. The anchor determines the center point for the
     * symbolizer.
     * @return {Array<number>} Anchor.
     * @api
     */
    getAnchor() {
      const size = this.size_;
      if (!size) {
        return null;
      }
      const displacement = this.getDisplacement();
      const scale = this.getScaleArray();
      // anchor is scaled by renderer but displacement should not be scaled
      // so divide by scale here
      return [
        size[0] / 2 - displacement[0] / scale[0],
        size[1] / 2 + displacement[1] / scale[1],
      ];
    }

    /**
     * Get the angle used in generating the shape.
     * @return {number} Shape's rotation in radians.
     * @api
     */
    getAngle() {
      return this.angle_;
    }

    /**
     * Get the fill style for the shape.
     * @return {import("./Fill.js").default} Fill style.
     * @api
     */
    getFill() {
      return this.fill_;
    }

    /**
     * Set the fill style.
     * @param {import("./Fill.js").default} fill Fill style.
     * @api
     */
    setFill(fill) {
      this.fill_ = fill;
      this.render();
    }

    /**
     * @return {HTMLCanvasElement} Image element.
     */
    getHitDetectionImage() {
      if (!this.hitDetectionCanvas_) {
        this.createHitDetectionCanvas_(this.renderOptions_);
      }
      return this.hitDetectionCanvas_;
    }

    /**
     * Get the image icon.
     * @param {number} pixelRatio Pixel ratio.
     * @return {HTMLCanvasElement} Image or Canvas element.
     * @api
     */
    getImage(pixelRatio) {
      let image = this.canvas_[pixelRatio];
      if (!image) {
        const renderOptions = this.renderOptions_;
        const context = createCanvasContext2D(
          renderOptions.size * pixelRatio,
          renderOptions.size * pixelRatio
        );
        this.draw_(renderOptions, context, pixelRatio);

        image = context.canvas;
        this.canvas_[pixelRatio] = image;
      }
      return image;
    }

    /**
     * Get the image pixel ratio.
     * @param {number} pixelRatio Pixel ratio.
     * @return {number} Pixel ratio.
     */
    getPixelRatio(pixelRatio) {
      return pixelRatio;
    }

    /**
     * @return {import("../size.js").Size} Image size.
     */
    getImageSize() {
      return this.size_;
    }

    /**
     * @return {import("../ImageState.js").default} Image state.
     */
    getImageState() {
      return ImageState.LOADED;
    }

    /**
     * Get the origin of the symbolizer.
     * @return {Array<number>} Origin.
     * @api
     */
    getOrigin() {
      return this.origin_;
    }

    /**
     * Get the number of points for generating the shape.
     * @return {number} Number of points for stars and regular polygons.
     * @api
     */
    getPoints() {
      return this.points_;
    }

    /**
     * Get the (primary) radius for the shape.
     * @return {number} Radius.
     * @api
     */
    getRadius() {
      return this.radius_;
    }

    /**
     * Get the secondary radius for the shape.
     * @return {number|undefined} Radius2.
     * @api
     */
    getRadius2() {
      return this.radius2_;
    }

    /**
     * Get the size of the symbolizer (in pixels).
     * @return {import("../size.js").Size} Size.
     * @api
     */
    getSize() {
      return this.size_;
    }

    /**
     * Get the stroke style for the shape.
     * @return {import("./Stroke.js").default} Stroke style.
     * @api
     */
    getStroke() {
      return this.stroke_;
    }

    /**
     * Set the stroke style.
     * @param {import("./Stroke.js").default} stroke Stroke style.
     * @api
     */
    setStroke(stroke) {
      this.stroke_ = stroke;
      this.render();
    }

    /**
     * @param {function(import("../events/Event.js").default): void} listener Listener function.
     */
    listenImageChange(listener) {}

    /**
     * Load not yet loaded URI.
     */
    load() {}

    /**
     * @param {function(import("../events/Event.js").default): void} listener Listener function.
     */
    unlistenImageChange(listener) {}

    /**
     * Calculate additional canvas size needed for the miter.
     * @param {string} lineJoin Line join
     * @param {number} strokeWidth Stroke width
     * @param {number} miterLimit Miter limit
     * @return {number} Additional canvas size needed
     * @private
     */
    calculateLineJoinSize_(lineJoin, strokeWidth, miterLimit) {
      if (
        strokeWidth === 0 ||
        this.points_ === Infinity ||
        (lineJoin !== 'bevel' && lineJoin !== 'miter')
      ) {
        return strokeWidth;
      }
      // m  | ^
      // i  | |\                  .
      // t >|  #\
      // e  | |\ \              .
      // r      \s\
      //      |  \t\          .                 .
      //          \r\                      .   .
      //      |    \o\      .          .  . . .
      //          e \k\            .  .    . .
      //      |      \e\  .    .  .       . .
      //       d      \ \  .  .          . .
      //      | _ _a_ _\#  .            . .
      //   r1          / `             . .
      //      |                       . .
      //       b     /               . .
      //      |                     . .
      //           / r2            . .
      //      |                        .   .
      //         /                           .   .
      //      |α                                   .   .
      //       /                                         .   .
      //      ° center
      let r1 = this.radius_;
      let r2 = this.radius2_ === undefined ? r1 : this.radius2_;
      if (r1 < r2) {
        const tmp = r1;
        r1 = r2;
        r2 = tmp;
      }
      const points =
        this.radius2_ === undefined ? this.points_ : this.points_ * 2;
      const alpha = (2 * Math.PI) / points;
      const a = r2 * Math.sin(alpha);
      const b = Math.sqrt(r2 * r2 - a * a);
      const d = r1 - b;
      const e = Math.sqrt(a * a + d * d);
      const miterRatio = e / a;
      if (lineJoin === 'miter' && miterRatio <= miterLimit) {
        return miterRatio * strokeWidth;
      }
      // Calculate the distnce from center to the stroke corner where
      // it was cut short because of the miter limit.
      //              l
      //        ----+---- <= distance from center to here is maxr
      //       /####|k ##\
      //      /#####^#####\
      //     /#### /+\# s #\
      //    /### h/+++\# t #\
      //   /### t/+++++\# r #\
      //  /### a/+++++++\# o #\
      // /### p/++ fill +\# k #\
      ///#### /+++++^+++++\# e #\
      //#####/+++++/+\+++++\#####\
      const k = strokeWidth / 2 / miterRatio;
      const l = (strokeWidth / 2) * (d / e);
      const maxr = Math.sqrt((r1 + k) * (r1 + k) + l * l);
      const bevelAdd = maxr - r1;
      if (this.radius2_ === undefined || lineJoin === 'bevel') {
        return bevelAdd * 2;
      }
      // If outer miter is over the miter limit the inner miter may reach through the
      // center and be longer than the bevel, same calculation as above but swap r1 / r2.
      const aa = r1 * Math.sin(alpha);
      const bb = Math.sqrt(r1 * r1 - aa * aa);
      const dd = r2 - bb;
      const ee = Math.sqrt(aa * aa + dd * dd);
      const innerMiterRatio = ee / aa;
      if (innerMiterRatio <= miterLimit) {
        const innerLength = (innerMiterRatio * strokeWidth) / 2 - r2 - r1;
        return 2 * Math.max(bevelAdd, innerLength);
      }
      return bevelAdd * 2;
    }

    /**
     * @return {RenderOptions}  The render options
     * @protected
     */
    createRenderOptions() {
      let lineJoin = defaultLineJoin;
      let miterLimit = 0;
      let lineDash = null;
      let lineDashOffset = 0;
      let strokeStyle;
      let strokeWidth = 0;

      if (this.stroke_) {
        strokeStyle = this.stroke_.getColor();
        if (strokeStyle === null) {
          strokeStyle = defaultStrokeStyle;
        }
        strokeStyle = asColorLike(strokeStyle);
        strokeWidth = this.stroke_.getWidth();
        if (strokeWidth === undefined) {
          strokeWidth = defaultLineWidth;
        }
        lineDash = this.stroke_.getLineDash();
        lineDashOffset = this.stroke_.getLineDashOffset();
        lineJoin = this.stroke_.getLineJoin();
        if (lineJoin === undefined) {
          lineJoin = defaultLineJoin;
        }
        miterLimit = this.stroke_.getMiterLimit();
        if (miterLimit === undefined) {
          miterLimit = defaultMiterLimit;
        }
      }

      const add = this.calculateLineJoinSize_(lineJoin, strokeWidth, miterLimit);
      const maxRadius = Math.max(this.radius_, this.radius2_ || 0);
      const size = Math.ceil(2 * maxRadius + add);

      return {
        strokeStyle: strokeStyle,
        strokeWidth: strokeWidth,
        size: size,
        lineDash: lineDash,
        lineDashOffset: lineDashOffset,
        lineJoin: lineJoin,
        miterLimit: miterLimit,
      };
    }

    /**
     * @protected
     */
    render() {
      this.renderOptions_ = this.createRenderOptions();
      const size = this.renderOptions_.size;
      this.canvas_ = {};
      this.size_ = [size, size];
    }

    /**
     * @private
     * @param {RenderOptions} renderOptions Render options.
     * @param {CanvasRenderingContext2D} context The rendering context.
     * @param {number} pixelRatio The pixel ratio.
     */
    draw_(renderOptions, context, pixelRatio) {
      context.scale(pixelRatio, pixelRatio);
      // set origin to canvas center
      context.translate(renderOptions.size / 2, renderOptions.size / 2);

      this.createPath_(context);

      if (this.fill_) {
        let color = this.fill_.getColor();
        if (color === null) {
          color = defaultFillStyle;
        }
        context.fillStyle = asColorLike(color);
        context.fill();
      }
      if (this.stroke_) {
        context.strokeStyle = renderOptions.strokeStyle;
        context.lineWidth = renderOptions.strokeWidth;
        if (renderOptions.lineDash) {
          context.setLineDash(renderOptions.lineDash);
          context.lineDashOffset = renderOptions.lineDashOffset;
        }
        context.lineJoin = renderOptions.lineJoin;
        context.miterLimit = renderOptions.miterLimit;
        context.stroke();
      }
    }

    /**
     * @private
     * @param {RenderOptions} renderOptions Render options.
     */
    createHitDetectionCanvas_(renderOptions) {
      if (this.fill_) {
        let color = this.fill_.getColor();

        // determine if fill is transparent (or pattern or gradient)
        let opacity = 0;
        if (typeof color === 'string') {
          color = asArray(color);
        }
        if (color === null) {
          opacity = 1;
        } else if (Array.isArray(color)) {
          opacity = color.length === 4 ? color[3] : 1;
        }
        if (opacity === 0) {
          // if a transparent fill style is set, create an extra hit-detection image
          // with a default fill style
          const context = createCanvasContext2D(
            renderOptions.size,
            renderOptions.size
          );
          this.hitDetectionCanvas_ = context.canvas;

          this.drawHitDetectionCanvas_(renderOptions, context);
        }
      }
      if (!this.hitDetectionCanvas_) {
        this.hitDetectionCanvas_ = this.getImage(1);
      }
    }

    /**
     * @private
     * @param {CanvasRenderingContext2D} context The context to draw in.
     */
    createPath_(context) {
      let points = this.points_;
      const radius = this.radius_;
      if (points === Infinity) {
        context.arc(0, 0, radius, 0, 2 * Math.PI);
      } else {
        const radius2 = this.radius2_ === undefined ? radius : this.radius2_;
        if (this.radius2_ !== undefined) {
          points *= 2;
        }
        const startAngle = this.angle_ - Math.PI / 2;
        const step = (2 * Math.PI) / points;
        for (let i = 0; i < points; i++) {
          const angle0 = startAngle + i * step;
          const radiusC = i % 2 === 0 ? radius : radius2;
          context.lineTo(radiusC * Math.cos(angle0), radiusC * Math.sin(angle0));
        }
        context.closePath();
      }
    }

    /**
     * @private
     * @param {RenderOptions} renderOptions Render options.
     * @param {CanvasRenderingContext2D} context The context.
     */
    drawHitDetectionCanvas_(renderOptions, context) {
      // set origin to canvas center
      context.translate(renderOptions.size / 2, renderOptions.size / 2);

      this.createPath_(context);

      context.fillStyle = defaultFillStyle;
      context.fill();
      if (this.stroke_) {
        context.strokeStyle = renderOptions.strokeStyle;
        context.lineWidth = renderOptions.strokeWidth;
        if (renderOptions.lineDash) {
          context.setLineDash(renderOptions.lineDash);
          context.lineDashOffset = renderOptions.lineDashOffset;
        }
        context.lineJoin = renderOptions.lineJoin;
        context.miterLimit = renderOptions.miterLimit;
        context.stroke();
      }
    }
  }

  /**
   * @module ol/style/Circle
   */

  /**
   * @typedef {Object} Options
   * @property {import("./Fill.js").default} [fill] Fill style.
   * @property {number} radius Circle radius.
   * @property {import("./Stroke.js").default} [stroke] Stroke style.
   * @property {Array<number>} [displacement=[0,0]] displacement
   * @property {number|import("../size.js").Size} [scale=1] Scale. A two dimensional scale will produce an ellipse.
   * Unless two dimensional scaling is required a better result may be obtained with an appropriate setting for `radius`.
   * @property {number} [rotation=0] Rotation in radians
   * (positive rotation clockwise, meaningful only when used in conjunction with a two dimensional scale).
   * @property {boolean} [rotateWithView=false] Whether to rotate the shape with the view
   * (meaningful only when used in conjunction with a two dimensional scale).
   * @property {"declutter"|"obstacle"|"none"|undefined} [declutterMode] Declutter mode
   */

  /**
   * @classdesc
   * Set circle style for vector features.
   * @api
   */
  class CircleStyle extends RegularShape {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      options = options ? options : {radius: 5};

      super({
        points: Infinity,
        fill: options.fill,
        radius: options.radius,
        stroke: options.stroke,
        scale: options.scale !== undefined ? options.scale : 1,
        rotation: options.rotation !== undefined ? options.rotation : 0,
        rotateWithView:
          options.rotateWithView !== undefined ? options.rotateWithView : false,
        displacement:
          options.displacement !== undefined ? options.displacement : [0, 0],
        declutterMode: options.declutterMode,
      });
    }

    /**
     * Clones the style.
     * @return {CircleStyle} The cloned style.
     * @api
     */
    clone() {
      const scale = this.getScale();
      const style = new CircleStyle({
        fill: this.getFill() ? this.getFill().clone() : undefined,
        stroke: this.getStroke() ? this.getStroke().clone() : undefined,
        radius: this.getRadius(),
        scale: Array.isArray(scale) ? scale.slice() : scale,
        rotation: this.getRotation(),
        rotateWithView: this.getRotateWithView(),
        displacement: this.getDisplacement().slice(),
        declutterMode: this.getDeclutterMode(),
      });
      style.setOpacity(this.getOpacity());
      return style;
    }

    /**
     * Set the circle radius.
     *
     * @param {number} radius Circle radius.
     * @api
     */
    setRadius(radius) {
      this.radius_ = radius;
      this.render();
    }
  }

  /**
   * @module ol/style/Fill
   */

  /**
   * @typedef {Object} Options
   * @property {import("../color.js").Color|import("../colorlike.js").ColorLike|null} [color=null] A color, gradient or pattern.
   * See {@link module:ol/color~Color} and {@link module:ol/colorlike~ColorLike} for possible formats.
   * Default null; if null, the Canvas/renderer default black will be used.
   */

  /**
   * @classdesc
   * Set fill style for vector features.
   * @api
   */
  class Fill {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      options = options || {};

      /**
       * @private
       * @type {import("../color.js").Color|import("../colorlike.js").ColorLike|null}
       */
      this.color_ = options.color !== undefined ? options.color : null;
    }

    /**
     * Clones the style. The color is not cloned if it is an {@link module:ol/colorlike~ColorLike}.
     * @return {Fill} The cloned style.
     * @api
     */
    clone() {
      const color = this.getColor();
      return new Fill({
        color: Array.isArray(color) ? color.slice() : color || undefined,
      });
    }

    /**
     * Get the fill color.
     * @return {import("../color.js").Color|import("../colorlike.js").ColorLike|null} Color.
     * @api
     */
    getColor() {
      return this.color_;
    }

    /**
     * Set the color.
     *
     * @param {import("../color.js").Color|import("../colorlike.js").ColorLike|null} color Color.
     * @api
     */
    setColor(color) {
      this.color_ = color;
    }
  }

  /**
   * @module ol/style/IconImage
   */

  /**
   * @type {CanvasRenderingContext2D}
   */
  let taintedTestContext = null;

  class IconImage extends Target {
    /**
     * @param {HTMLImageElement|HTMLCanvasElement} image Image.
     * @param {string|undefined} src Src.
     * @param {import("../size.js").Size} size Size.
     * @param {?string} crossOrigin Cross origin.
     * @param {import("../ImageState.js").default} imageState Image state.
     * @param {import("../color.js").Color} color Color.
     */
    constructor(image, src, size, crossOrigin, imageState, color) {
      super();

      /**
       * @private
       * @type {HTMLImageElement|HTMLCanvasElement}
       */
      this.hitDetectionImage_ = null;

      /**
       * @private
       * @type {HTMLImageElement|HTMLCanvasElement}
       */
      this.image_ = image;

      /**
       * @private
       * @type {string|null}
       */
      this.crossOrigin_ = crossOrigin;

      /**
       * @private
       * @type {Object<number, HTMLCanvasElement>}
       */
      this.canvas_ = {};

      /**
       * @private
       * @type {import("../color.js").Color}
       */
      this.color_ = color;

      /**
       * @private
       * @type {?function():void}
       */
      this.unlisten_ = null;

      /**
       * @private
       * @type {import("../ImageState.js").default}
       */
      this.imageState_ = imageState;

      /**
       * @private
       * @type {import("../size.js").Size}
       */
      this.size_ = size;

      /**
       * @private
       * @type {string|undefined}
       */
      this.src_ = src;

      /**
       * @private
       */
      this.tainted_;
    }

    /**
     * @private
     */
    initializeImage_() {
      this.image_ = new Image();
      if (this.crossOrigin_ !== null) {
        this.image_.crossOrigin = this.crossOrigin_;
      }
    }

    /**
     * @private
     * @return {boolean} The image canvas is tainted.
     */
    isTainted_() {
      if (this.tainted_ === undefined && this.imageState_ === ImageState.LOADED) {
        if (!taintedTestContext) {
          taintedTestContext = createCanvasContext2D(1, 1);
        }
        taintedTestContext.drawImage(this.image_, 0, 0);
        try {
          taintedTestContext.getImageData(0, 0, 1, 1);
          this.tainted_ = false;
        } catch (e) {
          taintedTestContext = null;
          this.tainted_ = true;
        }
      }
      return this.tainted_ === true;
    }

    /**
     * @private
     */
    dispatchChangeEvent_() {
      this.dispatchEvent(EventType$1.CHANGE);
    }

    /**
     * @private
     */
    handleImageError_() {
      this.imageState_ = ImageState.ERROR;
      this.unlistenImage_();
      this.dispatchChangeEvent_();
    }

    /**
     * @private
     */
    handleImageLoad_() {
      this.imageState_ = ImageState.LOADED;
      if (this.size_) {
        this.image_.width = this.size_[0];
        this.image_.height = this.size_[1];
      } else {
        this.size_ = [this.image_.width, this.image_.height];
      }
      this.unlistenImage_();
      this.dispatchChangeEvent_();
    }

    /**
     * @param {number} pixelRatio Pixel ratio.
     * @return {HTMLImageElement|HTMLCanvasElement} Image or Canvas element.
     */
    getImage(pixelRatio) {
      if (!this.image_) {
        this.initializeImage_();
      }
      this.replaceColor_(pixelRatio);
      return this.canvas_[pixelRatio] ? this.canvas_[pixelRatio] : this.image_;
    }

    /**
     * @param {number} pixelRatio Pixel ratio.
     * @return {number} Image or Canvas element.
     */
    getPixelRatio(pixelRatio) {
      this.replaceColor_(pixelRatio);
      return this.canvas_[pixelRatio] ? pixelRatio : 1;
    }

    /**
     * @return {import("../ImageState.js").default} Image state.
     */
    getImageState() {
      return this.imageState_;
    }

    /**
     * @return {HTMLImageElement|HTMLCanvasElement} Image element.
     */
    getHitDetectionImage() {
      if (!this.image_) {
        this.initializeImage_();
      }
      if (!this.hitDetectionImage_) {
        if (this.isTainted_()) {
          const width = this.size_[0];
          const height = this.size_[1];
          const context = createCanvasContext2D(width, height);
          context.fillRect(0, 0, width, height);
          this.hitDetectionImage_ = context.canvas;
        } else {
          this.hitDetectionImage_ = this.image_;
        }
      }
      return this.hitDetectionImage_;
    }

    /**
     * Get the size of the icon (in pixels).
     * @return {import("../size.js").Size} Image size.
     */
    getSize() {
      return this.size_;
    }

    /**
     * @return {string|undefined} Image src.
     */
    getSrc() {
      return this.src_;
    }

    /**
     * Load not yet loaded URI.
     */
    load() {
      if (this.imageState_ !== ImageState.IDLE) {
        return;
      }
      if (!this.image_) {
        this.initializeImage_();
      }

      this.imageState_ = ImageState.LOADING;
      try {
        /** @type {HTMLImageElement} */ (this.image_).src = this.src_;
      } catch (e) {
        this.handleImageError_();
      }
      this.unlisten_ = listenImage(
        this.image_,
        this.handleImageLoad_.bind(this),
        this.handleImageError_.bind(this)
      );
    }

    /**
     * @param {number} pixelRatio Pixel ratio.
     * @private
     */
    replaceColor_(pixelRatio) {
      if (
        !this.color_ ||
        this.canvas_[pixelRatio] ||
        this.imageState_ !== ImageState.LOADED
      ) {
        return;
      }

      const image = this.image_;
      const canvas = document.createElement('canvas');
      canvas.width = Math.ceil(image.width * pixelRatio);
      canvas.height = Math.ceil(image.height * pixelRatio);

      const ctx = canvas.getContext('2d');
      ctx.scale(pixelRatio, pixelRatio);
      ctx.drawImage(image, 0, 0);

      ctx.globalCompositeOperation = 'multiply';
      ctx.fillStyle = asString(this.color_);
      ctx.fillRect(0, 0, canvas.width / pixelRatio, canvas.height / pixelRatio);

      ctx.globalCompositeOperation = 'destination-in';
      ctx.drawImage(image, 0, 0);

      this.canvas_[pixelRatio] = canvas;
    }

    /**
     * Discards event handlers which listen for load completion or errors.
     *
     * @private
     */
    unlistenImage_() {
      if (this.unlisten_) {
        this.unlisten_();
        this.unlisten_ = null;
      }
    }
  }

  /**
   * @param {HTMLImageElement|HTMLCanvasElement} image Image.
   * @param {string} src Src.
   * @param {import("../size.js").Size} size Size.
   * @param {?string} crossOrigin Cross origin.
   * @param {import("../ImageState.js").default} imageState Image state.
   * @param {import("../color.js").Color} color Color.
   * @return {IconImage} Icon image.
   */
  function get$1(image, src, size, crossOrigin, imageState, color) {
    let iconImage = shared.get(src, crossOrigin, color);
    if (!iconImage) {
      iconImage = new IconImage(image, src, size, crossOrigin, imageState, color);
      shared.set(src, crossOrigin, color, iconImage);
    }
    return iconImage;
  }

  /**
   * @module ol/style/Icon
   */

  /**
   * @typedef {'fraction' | 'pixels'} IconAnchorUnits
   * Anchor unit can be either a fraction of the icon size or in pixels.
   */

  /**
   * @typedef {'bottom-left' | 'bottom-right' | 'top-left' | 'top-right'} IconOrigin
   * Icon origin. One of 'bottom-left', 'bottom-right', 'top-left', 'top-right'.
   */

  /**
   * @typedef {Object} Options
   * @property {Array<number>} [anchor=[0.5, 0.5]] Anchor. Default value is the icon center.
   * @property {IconOrigin} [anchorOrigin='top-left'] Origin of the anchor: `bottom-left`, `bottom-right`,
   * `top-left` or `top-right`.
   * @property {IconAnchorUnits} [anchorXUnits='fraction'] Units in which the anchor x value is
   * specified. A value of `'fraction'` indicates the x value is a fraction of the icon. A value of `'pixels'` indicates
   * the x value in pixels.
   * @property {IconAnchorUnits} [anchorYUnits='fraction'] Units in which the anchor y value is
   * specified. A value of `'fraction'` indicates the y value is a fraction of the icon. A value of `'pixels'` indicates
   * the y value in pixels.
   * @property {import("../color.js").Color|string} [color] Color to tint the icon. If not specified,
   * the icon will be left as is.
   * @property {null|string} [crossOrigin] The `crossOrigin` attribute for loaded images. Note that you must provide a
   * `crossOrigin` value if you want to access pixel data with the Canvas renderer.
   * See https://developer.mozilla.org/en-US/docs/Web/HTML/CORS_enabled_image for more detail.
   * @property {HTMLImageElement|HTMLCanvasElement} [img] Image object for the icon. If the `src` option is not provided then the
   * provided image must already be loaded. And in that case, it is required
   * to provide the size of the image, with the `imgSize` option.
   * @property {import("../size.js").Size} [imgSize] Image size in pixels. Only required if `img` is set and `src` is not.
   * The provided `imgSize` needs to match the actual size of the image.
   * @property {Array<number>} [displacement=[0, 0]] Displacement of the icon in pixels.
   * Positive values will shift the icon right and up.
   * @property {number} [opacity=1] Opacity of the icon.
   * @property {number|import("../size.js").Size} [scale=1] Scale.
   * @property {boolean} [rotateWithView=false] Whether to rotate the icon with the view.
   * @property {number} [rotation=0] Rotation in radians (positive rotation clockwise).
   * @property {Array<number>} [offset=[0, 0]] Offset which, together with `size` and `offsetOrigin`, defines the
   * sub-rectangle to use from the original (sprite) image.
   * @property {IconOrigin} [offsetOrigin='top-left'] Origin of the offset: `bottom-left`, `bottom-right`,
   * `top-left` or `top-right`.
   * @property {import("../size.js").Size} [size] Icon size in pixels. Used together with `offset` to define the
   * sub-rectangle to use from the original (sprite) image.
   * @property {string} [src] Image source URI.
   * @property {"declutter"|"obstacle"|"none"|undefined} [declutterMode] Declutter mode.
   */

  /**
   * @classdesc
   * Set icon style for vector features.
   * @api
   */
  class Icon extends ImageStyle {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      options = options || {};

      /**
       * @type {number}
       */
      const opacity = options.opacity !== undefined ? options.opacity : 1;

      /**
       * @type {number}
       */
      const rotation = options.rotation !== undefined ? options.rotation : 0;

      /**
       * @type {number|import("../size.js").Size}
       */
      const scale = options.scale !== undefined ? options.scale : 1;

      /**
       * @type {boolean}
       */
      const rotateWithView =
        options.rotateWithView !== undefined ? options.rotateWithView : false;

      super({
        opacity: opacity,
        rotation: rotation,
        scale: scale,
        displacement:
          options.displacement !== undefined ? options.displacement : [0, 0],
        rotateWithView: rotateWithView,
        declutterMode: options.declutterMode,
      });

      /**
       * @private
       * @type {Array<number>}
       */
      this.anchor_ = options.anchor !== undefined ? options.anchor : [0.5, 0.5];

      /**
       * @private
       * @type {Array<number>}
       */
      this.normalizedAnchor_ = null;

      /**
       * @private
       * @type {IconOrigin}
       */
      this.anchorOrigin_ =
        options.anchorOrigin !== undefined ? options.anchorOrigin : 'top-left';

      /**
       * @private
       * @type {IconAnchorUnits}
       */
      this.anchorXUnits_ =
        options.anchorXUnits !== undefined ? options.anchorXUnits : 'fraction';

      /**
       * @private
       * @type {IconAnchorUnits}
       */
      this.anchorYUnits_ =
        options.anchorYUnits !== undefined ? options.anchorYUnits : 'fraction';

      /**
       * @private
       * @type {?string}
       */
      this.crossOrigin_ =
        options.crossOrigin !== undefined ? options.crossOrigin : null;

      /**
       * @type {HTMLImageElement|HTMLCanvasElement}
       */
      const image = options.img !== undefined ? options.img : null;

      /**
       * @private
       * @type {import("../size.js").Size|undefined}
       */
      this.imgSize_ = options.imgSize;

      /**
       * @type {string|undefined}
       */
      let src = options.src;

      assert(!(src !== undefined && image), 4); // `image` and `src` cannot be provided at the same time
      assert(!image || (image && this.imgSize_), 5); // `imgSize` must be set when `image` is provided

      if ((src === undefined || src.length === 0) && image) {
        src = /** @type {HTMLImageElement} */ (image).src || getUid(image);
      }
      assert(src !== undefined && src.length > 0, 6); // A defined and non-empty `src` or `image` must be provided

      /**
       * @type {import("../ImageState.js").default}
       */
      const imageState =
        options.src !== undefined ? ImageState.IDLE : ImageState.LOADED;

      /**
       * @private
       * @type {import("../color.js").Color}
       */
      this.color_ = options.color !== undefined ? asArray(options.color) : null;

      /**
       * @private
       * @type {import("./IconImage.js").default}
       */
      this.iconImage_ = get$1(
        image,
        /** @type {string} */ (src),
        this.imgSize_ !== undefined ? this.imgSize_ : null,
        this.crossOrigin_,
        imageState,
        this.color_
      );

      /**
       * @private
       * @type {Array<number>}
       */
      this.offset_ = options.offset !== undefined ? options.offset : [0, 0];
      /**
       * @private
       * @type {IconOrigin}
       */
      this.offsetOrigin_ =
        options.offsetOrigin !== undefined ? options.offsetOrigin : 'top-left';

      /**
       * @private
       * @type {Array<number>}
       */
      this.origin_ = null;

      /**
       * @private
       * @type {import("../size.js").Size}
       */
      this.size_ = options.size !== undefined ? options.size : null;
    }

    /**
     * Clones the style. The underlying Image/HTMLCanvasElement is not cloned.
     * @return {Icon} The cloned style.
     * @api
     */
    clone() {
      const scale = this.getScale();
      return new Icon({
        anchor: this.anchor_.slice(),
        anchorOrigin: this.anchorOrigin_,
        anchorXUnits: this.anchorXUnits_,
        anchorYUnits: this.anchorYUnits_,
        color:
          this.color_ && this.color_.slice
            ? this.color_.slice()
            : this.color_ || undefined,
        crossOrigin: this.crossOrigin_,
        imgSize: this.imgSize_,
        offset: this.offset_.slice(),
        offsetOrigin: this.offsetOrigin_,
        opacity: this.getOpacity(),
        rotateWithView: this.getRotateWithView(),
        rotation: this.getRotation(),
        scale: Array.isArray(scale) ? scale.slice() : scale,
        size: this.size_ !== null ? this.size_.slice() : undefined,
        src: this.getSrc(),
        displacement: this.getDisplacement().slice(),
        declutterMode: this.getDeclutterMode(),
      });
    }

    /**
     * Get the anchor point in pixels. The anchor determines the center point for the
     * symbolizer.
     * @return {Array<number>} Anchor.
     * @api
     */
    getAnchor() {
      let anchor = this.normalizedAnchor_;
      if (!anchor) {
        anchor = this.anchor_;
        const size = this.getSize();
        if (
          this.anchorXUnits_ == 'fraction' ||
          this.anchorYUnits_ == 'fraction'
        ) {
          if (!size) {
            return null;
          }
          anchor = this.anchor_.slice();
          if (this.anchorXUnits_ == 'fraction') {
            anchor[0] *= size[0];
          }
          if (this.anchorYUnits_ == 'fraction') {
            anchor[1] *= size[1];
          }
        }

        if (this.anchorOrigin_ != 'top-left') {
          if (!size) {
            return null;
          }
          if (anchor === this.anchor_) {
            anchor = this.anchor_.slice();
          }
          if (
            this.anchorOrigin_ == 'top-right' ||
            this.anchorOrigin_ == 'bottom-right'
          ) {
            anchor[0] = -anchor[0] + size[0];
          }
          if (
            this.anchorOrigin_ == 'bottom-left' ||
            this.anchorOrigin_ == 'bottom-right'
          ) {
            anchor[1] = -anchor[1] + size[1];
          }
        }
        this.normalizedAnchor_ = anchor;
      }
      const displacement = this.getDisplacement();
      const scale = this.getScaleArray();
      // anchor is scaled by renderer but displacement should not be scaled
      // so divide by scale here
      return [
        anchor[0] - displacement[0] / scale[0],
        anchor[1] + displacement[1] / scale[1],
      ];
    }

    /**
     * Set the anchor point. The anchor determines the center point for the
     * symbolizer.
     *
     * @param {Array<number>} anchor Anchor.
     * @api
     */
    setAnchor(anchor) {
      this.anchor_ = anchor;
      this.normalizedAnchor_ = null;
    }

    /**
     * Get the icon color.
     * @return {import("../color.js").Color} Color.
     * @api
     */
    getColor() {
      return this.color_;
    }

    /**
     * Get the image icon.
     * @param {number} pixelRatio Pixel ratio.
     * @return {HTMLImageElement|HTMLCanvasElement} Image or Canvas element.
     * @api
     */
    getImage(pixelRatio) {
      return this.iconImage_.getImage(pixelRatio);
    }

    /**
     * Get the pixel ratio.
     * @param {number} pixelRatio Pixel ratio.
     * @return {number} The pixel ratio of the image.
     * @api
     */
    getPixelRatio(pixelRatio) {
      return this.iconImage_.getPixelRatio(pixelRatio);
    }

    /**
     * @return {import("../size.js").Size} Image size.
     */
    getImageSize() {
      return this.iconImage_.getSize();
    }

    /**
     * @return {import("../ImageState.js").default} Image state.
     */
    getImageState() {
      return this.iconImage_.getImageState();
    }

    /**
     * @return {HTMLImageElement|HTMLCanvasElement} Image element.
     */
    getHitDetectionImage() {
      return this.iconImage_.getHitDetectionImage();
    }

    /**
     * Get the origin of the symbolizer.
     * @return {Array<number>} Origin.
     * @api
     */
    getOrigin() {
      if (this.origin_) {
        return this.origin_;
      }
      let offset = this.offset_;

      if (this.offsetOrigin_ != 'top-left') {
        const size = this.getSize();
        const iconImageSize = this.iconImage_.getSize();
        if (!size || !iconImageSize) {
          return null;
        }
        offset = offset.slice();
        if (
          this.offsetOrigin_ == 'top-right' ||
          this.offsetOrigin_ == 'bottom-right'
        ) {
          offset[0] = iconImageSize[0] - size[0] - offset[0];
        }
        if (
          this.offsetOrigin_ == 'bottom-left' ||
          this.offsetOrigin_ == 'bottom-right'
        ) {
          offset[1] = iconImageSize[1] - size[1] - offset[1];
        }
      }
      this.origin_ = offset;
      return this.origin_;
    }

    /**
     * Get the image URL.
     * @return {string|undefined} Image src.
     * @api
     */
    getSrc() {
      return this.iconImage_.getSrc();
    }

    /**
     * Get the size of the icon (in pixels).
     * @return {import("../size.js").Size} Image size.
     * @api
     */
    getSize() {
      return !this.size_ ? this.iconImage_.getSize() : this.size_;
    }

    /**
     * @param {function(import("../events/Event.js").default): void} listener Listener function.
     */
    listenImageChange(listener) {
      this.iconImage_.addEventListener(EventType$1.CHANGE, listener);
    }

    /**
     * Load not yet loaded URI.
     * When rendering a feature with an icon style, the vector renderer will
     * automatically call this method. However, you might want to call this
     * method yourself for preloading or other purposes.
     * @api
     */
    load() {
      this.iconImage_.load();
    }

    /**
     * @param {function(import("../events/Event.js").default): void} listener Listener function.
     */
    unlistenImageChange(listener) {
      this.iconImage_.removeEventListener(EventType$1.CHANGE, listener);
    }
  }

  /**
   * @module ol/style/Stroke
   */

  /**
   * @typedef {Object} Options
   * @property {import("../color.js").Color|import("../colorlike.js").ColorLike} [color] A color, gradient or pattern.
   * See {@link module:ol/color~Color} and {@link module:ol/colorlike~ColorLike} for possible formats.
   * Default null; if null, the Canvas/renderer default black will be used.
   * @property {CanvasLineCap} [lineCap='round'] Line cap style: `butt`, `round`, or `square`.
   * @property {CanvasLineJoin} [lineJoin='round'] Line join style: `bevel`, `round`, or `miter`.
   * @property {Array<number>} [lineDash] Line dash pattern. Default is `null` (no dash).
   * @property {number} [lineDashOffset=0] Line dash offset.
   * @property {number} [miterLimit=10] Miter limit.
   * @property {number} [width] Width.
   */

  /**
   * @classdesc
   * Set stroke style for vector features.
   * Note that the defaults given are the Canvas defaults, which will be used if
   * option is not defined. The `get` functions return whatever was entered in
   * the options; they will not return the default.
   * @api
   */
  class Stroke {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      options = options || {};

      /**
       * @private
       * @type {import("../color.js").Color|import("../colorlike.js").ColorLike}
       */
      this.color_ = options.color !== undefined ? options.color : null;

      /**
       * @private
       * @type {CanvasLineCap|undefined}
       */
      this.lineCap_ = options.lineCap;

      /**
       * @private
       * @type {Array<number>|null}
       */
      this.lineDash_ = options.lineDash !== undefined ? options.lineDash : null;

      /**
       * @private
       * @type {number|undefined}
       */
      this.lineDashOffset_ = options.lineDashOffset;

      /**
       * @private
       * @type {CanvasLineJoin|undefined}
       */
      this.lineJoin_ = options.lineJoin;

      /**
       * @private
       * @type {number|undefined}
       */
      this.miterLimit_ = options.miterLimit;

      /**
       * @private
       * @type {number|undefined}
       */
      this.width_ = options.width;
    }

    /**
     * Clones the style.
     * @return {Stroke} The cloned style.
     * @api
     */
    clone() {
      const color = this.getColor();
      return new Stroke({
        color: Array.isArray(color) ? color.slice() : color || undefined,
        lineCap: this.getLineCap(),
        lineDash: this.getLineDash() ? this.getLineDash().slice() : undefined,
        lineDashOffset: this.getLineDashOffset(),
        lineJoin: this.getLineJoin(),
        miterLimit: this.getMiterLimit(),
        width: this.getWidth(),
      });
    }

    /**
     * Get the stroke color.
     * @return {import("../color.js").Color|import("../colorlike.js").ColorLike} Color.
     * @api
     */
    getColor() {
      return this.color_;
    }

    /**
     * Get the line cap type for the stroke.
     * @return {CanvasLineCap|undefined} Line cap.
     * @api
     */
    getLineCap() {
      return this.lineCap_;
    }

    /**
     * Get the line dash style for the stroke.
     * @return {Array<number>|null} Line dash.
     * @api
     */
    getLineDash() {
      return this.lineDash_;
    }

    /**
     * Get the line dash offset for the stroke.
     * @return {number|undefined} Line dash offset.
     * @api
     */
    getLineDashOffset() {
      return this.lineDashOffset_;
    }

    /**
     * Get the line join type for the stroke.
     * @return {CanvasLineJoin|undefined} Line join.
     * @api
     */
    getLineJoin() {
      return this.lineJoin_;
    }

    /**
     * Get the miter limit for the stroke.
     * @return {number|undefined} Miter limit.
     * @api
     */
    getMiterLimit() {
      return this.miterLimit_;
    }

    /**
     * Get the stroke width.
     * @return {number|undefined} Width.
     * @api
     */
    getWidth() {
      return this.width_;
    }

    /**
     * Set the color.
     *
     * @param {import("../color.js").Color|import("../colorlike.js").ColorLike} color Color.
     * @api
     */
    setColor(color) {
      this.color_ = color;
    }

    /**
     * Set the line cap.
     *
     * @param {CanvasLineCap|undefined} lineCap Line cap.
     * @api
     */
    setLineCap(lineCap) {
      this.lineCap_ = lineCap;
    }

    /**
     * Set the line dash.
     *
     * @param {Array<number>|null} lineDash Line dash.
     * @api
     */
    setLineDash(lineDash) {
      this.lineDash_ = lineDash;
    }

    /**
     * Set the line dash offset.
     *
     * @param {number|undefined} lineDashOffset Line dash offset.
     * @api
     */
    setLineDashOffset(lineDashOffset) {
      this.lineDashOffset_ = lineDashOffset;
    }

    /**
     * Set the line join.
     *
     * @param {CanvasLineJoin|undefined} lineJoin Line join.
     * @api
     */
    setLineJoin(lineJoin) {
      this.lineJoin_ = lineJoin;
    }

    /**
     * Set the miter limit.
     *
     * @param {number|undefined} miterLimit Miter limit.
     * @api
     */
    setMiterLimit(miterLimit) {
      this.miterLimit_ = miterLimit;
    }

    /**
     * Set the width.
     *
     * @param {number|undefined} width Width.
     * @api
     */
    setWidth(width) {
      this.width_ = width;
    }
  }

  /**
   * @module ol/style/Style
   */

  /**
   * A function that takes an {@link module:ol/Feature~Feature} and a `{number}`
   * representing the view's resolution. The function should return a
   * {@link module:ol/style/Style~Style} or an array of them. This way e.g. a
   * vector layer can be styled. If the function returns `undefined`, the
   * feature will not be rendered.
   *
   * @typedef {function(import("../Feature.js").FeatureLike, number):(Style|Array<Style>|void)} StyleFunction
   */

  /**
   * A {@link Style}, an array of {@link Style}, or a {@link StyleFunction}.
   * @typedef {Style|Array<Style>|StyleFunction} StyleLike
   */

  /**
   * A function that takes an {@link module:ol/Feature~Feature} as argument and returns an
   * {@link module:ol/geom/Geometry~Geometry} that will be rendered and styled for the feature.
   *
   * @typedef {function(import("../Feature.js").FeatureLike):
   *     (import("../geom/Geometry.js").default|import("../render/Feature.js").default|undefined)} GeometryFunction
   */

  /**
   * Custom renderer function. Takes two arguments:
   *
   * 1. The pixel coordinates of the geometry in GeoJSON notation.
   * 2. The {@link module:ol/render~State} of the layer renderer.
   *
   * @typedef {function((import("../coordinate.js").Coordinate|Array<import("../coordinate.js").Coordinate>|Array<Array<import("../coordinate.js").Coordinate>>),import("../render.js").State): void} RenderFunction
   */

  /**
   * @typedef {Object} Options
   * @property {string|import("../geom/Geometry.js").default|GeometryFunction} [geometry] Feature property or geometry
   * or function returning a geometry to render for this style.
   * @property {import("./Fill.js").default} [fill] Fill style.
   * @property {import("./Image.js").default} [image] Image style.
   * @property {RenderFunction} [renderer] Custom renderer. When configured, `fill`, `stroke` and `image` will be
   * ignored, and the provided function will be called with each render frame for each geometry.
   * @property {RenderFunction} [hitDetectionRenderer] Custom renderer for hit detection. If provided will be used
   * in hit detection rendering.
   * @property {import("./Stroke.js").default} [stroke] Stroke style.
   * @property {import("./Text.js").default} [text] Text style.
   * @property {number} [zIndex] Z index.
   */

  /**
   * @classdesc
   * Container for vector feature rendering styles. Any changes made to the style
   * or its children through `set*()` methods will not take effect until the
   * feature or layer that uses the style is re-rendered.
   *
   * ## Feature styles
   *
   * If no style is defined, the following default style is used:
   * ```js
   *  import {Circle, Fill, Stroke, Style} from 'ol/style';
   *
   *  const fill = new Fill({
   *    color: 'rgba(255,255,255,0.4)',
   *  });
   *  const stroke = new Stroke({
   *    color: '#3399CC',
   *    width: 1.25,
   *  });
   *  const styles = [
   *    new Style({
   *      image: new Circle({
   *        fill: fill,
   *        stroke: stroke,
   *        radius: 5,
   *      }),
   *      fill: fill,
   *      stroke: stroke,
   *    }),
   *  ];
   * ```
   *
   * A separate editing style has the following defaults:
   * ```js
   *  import {Circle, Fill, Stroke, Style} from 'ol/style';
   *
   *  const styles = {};
   *  const white = [255, 255, 255, 1];
   *  const blue = [0, 153, 255, 1];
   *  const width = 3;
   *  styles['Polygon'] = [
   *    new Style({
   *      fill: new Fill({
   *        color: [255, 255, 255, 0.5],
   *      }),
   *    }),
   *  ];
   *  styles['MultiPolygon'] =
   *      styles['Polygon'];
   *  styles['LineString'] = [
   *    new Style({
   *      stroke: new Stroke({
   *        color: white,
   *        width: width + 2,
   *      }),
   *    }),
   *    new Style({
   *      stroke: new Stroke({
   *        color: blue,
   *        width: width,
   *      }),
   *    }),
   *  ];
   *  styles['MultiLineString'] = styles['LineString'];
   *
   *  styles['Circle'] = styles['Polygon'].concat(
   *    styles['LineString']
   *  );
   *
   *  styles['Point'] = [
   *    new Style({
   *      image: new Circle({
   *        radius: width * 2,
   *        fill: new Fill({
   *          color: blue,
   *        }),
   *        stroke: new Stroke({
   *          color: white,
   *          width: width / 2,
   *        }),
   *      }),
   *      zIndex: Infinity,
   *    }),
   *  ];
   *  styles['MultiPoint'] =
   *      styles['Point'];
   *  styles['GeometryCollection'] =
   *      styles['Polygon'].concat(
   *          styles['LineString'],
   *          styles['Point']
   *      );
   * ```
   *
   * @api
   */
  class Style {
    /**
     * @param {Options} [options] Style options.
     */
    constructor(options) {
      options = options || {};

      /**
       * @private
       * @type {string|import("../geom/Geometry.js").default|GeometryFunction}
       */
      this.geometry_ = null;

      /**
       * @private
       * @type {!GeometryFunction}
       */
      this.geometryFunction_ = defaultGeometryFunction;

      if (options.geometry !== undefined) {
        this.setGeometry(options.geometry);
      }

      /**
       * @private
       * @type {import("./Fill.js").default}
       */
      this.fill_ = options.fill !== undefined ? options.fill : null;

      /**
       * @private
       * @type {import("./Image.js").default}
       */
      this.image_ = options.image !== undefined ? options.image : null;

      /**
       * @private
       * @type {RenderFunction|null}
       */
      this.renderer_ = options.renderer !== undefined ? options.renderer : null;

      /**
       * @private
       * @type {RenderFunction|null}
       */
      this.hitDetectionRenderer_ =
        options.hitDetectionRenderer !== undefined
          ? options.hitDetectionRenderer
          : null;

      /**
       * @private
       * @type {import("./Stroke.js").default}
       */
      this.stroke_ = options.stroke !== undefined ? options.stroke : null;

      /**
       * @private
       * @type {import("./Text.js").default}
       */
      this.text_ = options.text !== undefined ? options.text : null;

      /**
       * @private
       * @type {number|undefined}
       */
      this.zIndex_ = options.zIndex;
    }

    /**
     * Clones the style.
     * @return {Style} The cloned style.
     * @api
     */
    clone() {
      let geometry = this.getGeometry();
      if (geometry && typeof geometry === 'object') {
        geometry = /** @type {import("../geom/Geometry.js").default} */ (
          geometry
        ).clone();
      }
      return new Style({
        geometry: geometry,
        fill: this.getFill() ? this.getFill().clone() : undefined,
        image: this.getImage() ? this.getImage().clone() : undefined,
        renderer: this.getRenderer(),
        stroke: this.getStroke() ? this.getStroke().clone() : undefined,
        text: this.getText() ? this.getText().clone() : undefined,
        zIndex: this.getZIndex(),
      });
    }

    /**
     * Get the custom renderer function that was configured with
     * {@link #setRenderer} or the `renderer` constructor option.
     * @return {RenderFunction|null} Custom renderer function.
     * @api
     */
    getRenderer() {
      return this.renderer_;
    }

    /**
     * Sets a custom renderer function for this style. When set, `fill`, `stroke`
     * and `image` options of the style will be ignored.
     * @param {RenderFunction|null} renderer Custom renderer function.
     * @api
     */
    setRenderer(renderer) {
      this.renderer_ = renderer;
    }

    /**
     * Sets a custom renderer function for this style used
     * in hit detection.
     * @param {RenderFunction|null} renderer Custom renderer function.
     * @api
     */
    setHitDetectionRenderer(renderer) {
      this.hitDetectionRenderer_ = renderer;
    }

    /**
     * Get the custom renderer function that was configured with
     * {@link #setHitDetectionRenderer} or the `hitDetectionRenderer` constructor option.
     * @return {RenderFunction|null} Custom renderer function.
     * @api
     */
    getHitDetectionRenderer() {
      return this.hitDetectionRenderer_;
    }

    /**
     * Get the geometry to be rendered.
     * @return {string|import("../geom/Geometry.js").default|GeometryFunction}
     * Feature property or geometry or function that returns the geometry that will
     * be rendered with this style.
     * @api
     */
    getGeometry() {
      return this.geometry_;
    }

    /**
     * Get the function used to generate a geometry for rendering.
     * @return {!GeometryFunction} Function that is called with a feature
     * and returns the geometry to render instead of the feature's geometry.
     * @api
     */
    getGeometryFunction() {
      return this.geometryFunction_;
    }

    /**
     * Get the fill style.
     * @return {import("./Fill.js").default} Fill style.
     * @api
     */
    getFill() {
      return this.fill_;
    }

    /**
     * Set the fill style.
     * @param {import("./Fill.js").default} fill Fill style.
     * @api
     */
    setFill(fill) {
      this.fill_ = fill;
    }

    /**
     * Get the image style.
     * @return {import("./Image.js").default} Image style.
     * @api
     */
    getImage() {
      return this.image_;
    }

    /**
     * Set the image style.
     * @param {import("./Image.js").default} image Image style.
     * @api
     */
    setImage(image) {
      this.image_ = image;
    }

    /**
     * Get the stroke style.
     * @return {import("./Stroke.js").default} Stroke style.
     * @api
     */
    getStroke() {
      return this.stroke_;
    }

    /**
     * Set the stroke style.
     * @param {import("./Stroke.js").default} stroke Stroke style.
     * @api
     */
    setStroke(stroke) {
      this.stroke_ = stroke;
    }

    /**
     * Get the text style.
     * @return {import("./Text.js").default} Text style.
     * @api
     */
    getText() {
      return this.text_;
    }

    /**
     * Set the text style.
     * @param {import("./Text.js").default} text Text style.
     * @api
     */
    setText(text) {
      this.text_ = text;
    }

    /**
     * Get the z-index for the style.
     * @return {number|undefined} ZIndex.
     * @api
     */
    getZIndex() {
      return this.zIndex_;
    }

    /**
     * Set a geometry that is rendered instead of the feature's geometry.
     *
     * @param {string|import("../geom/Geometry.js").default|GeometryFunction} geometry
     *     Feature property or geometry or function returning a geometry to render
     *     for this style.
     * @api
     */
    setGeometry(geometry) {
      if (typeof geometry === 'function') {
        this.geometryFunction_ = geometry;
      } else if (typeof geometry === 'string') {
        this.geometryFunction_ = function (feature) {
          return /** @type {import("../geom/Geometry.js").default} */ (
            feature.get(geometry)
          );
        };
      } else if (!geometry) {
        this.geometryFunction_ = defaultGeometryFunction;
      } else if (geometry !== undefined) {
        this.geometryFunction_ = function () {
          return /** @type {import("../geom/Geometry.js").default} */ (geometry);
        };
      }
      this.geometry_ = geometry;
    }

    /**
     * Set the z-index.
     *
     * @param {number|undefined} zIndex ZIndex.
     * @api
     */
    setZIndex(zIndex) {
      this.zIndex_ = zIndex;
    }
  }

  /**
   * Convert the provided object into a style function.  Functions passed through
   * unchanged.  Arrays of Style or single style objects wrapped in a
   * new style function.
   * @param {StyleFunction|Array<Style>|Style} obj
   *     A style function, a single style, or an array of styles.
   * @return {StyleFunction} A style function.
   */
  function toFunction(obj) {
    let styleFunction;

    if (typeof obj === 'function') {
      styleFunction = obj;
    } else {
      /**
       * @type {Array<Style>}
       */
      let styles;
      if (Array.isArray(obj)) {
        styles = obj;
      } else {
        assert(typeof (/** @type {?} */ (obj).getZIndex) === 'function', 41); // Expected an `Style` or an array of `Style`
        const style = /** @type {Style} */ (obj);
        styles = [style];
      }
      styleFunction = function () {
        return styles;
      };
    }
    return styleFunction;
  }

  /**
   * @type {Array<Style>|null}
   */
  let defaultStyles = null;

  /**
   * @param {import("../Feature.js").FeatureLike} feature Feature.
   * @param {number} resolution Resolution.
   * @return {Array<Style>} Style.
   */
  function createDefaultStyle(feature, resolution) {
    // We don't use an immediately-invoked function
    // and a closure so we don't get an error at script evaluation time in
    // browsers that do not support Canvas. (import("./Circle.js").CircleStyle does
    // canvas.getContext('2d') at construction time, which will cause an.error
    // in such browsers.)
    if (!defaultStyles) {
      const fill = new Fill({
        color: 'rgba(255,255,255,0.4)',
      });
      const stroke = new Stroke({
        color: '#3399CC',
        width: 1.25,
      });
      defaultStyles = [
        new Style({
          image: new CircleStyle({
            fill: fill,
            stroke: stroke,
            radius: 5,
          }),
          fill: fill,
          stroke: stroke,
        }),
      ];
    }
    return defaultStyles;
  }

  /**
   * Default styles for editing features.
   * @return {Object<import("../geom/Geometry.js").Type, Array<Style>>} Styles
   */
  function createEditingStyle() {
    /** @type {Object<import("../geom/Geometry.js").Type, Array<Style>>} */
    const styles = {};
    const white = [255, 255, 255, 1];
    const blue = [0, 153, 255, 1];
    const width = 3;
    styles['Polygon'] = [
      new Style({
        fill: new Fill({
          color: [255, 255, 255, 0.5],
        }),
      }),
    ];
    styles['MultiPolygon'] = styles['Polygon'];

    styles['LineString'] = [
      new Style({
        stroke: new Stroke({
          color: white,
          width: width + 2,
        }),
      }),
      new Style({
        stroke: new Stroke({
          color: blue,
          width: width,
        }),
      }),
    ];
    styles['MultiLineString'] = styles['LineString'];

    styles['Circle'] = styles['Polygon'].concat(styles['LineString']);

    styles['Point'] = [
      new Style({
        image: new CircleStyle({
          radius: width * 2,
          fill: new Fill({
            color: blue,
          }),
          stroke: new Stroke({
            color: white,
            width: width / 2,
          }),
        }),
        zIndex: Infinity,
      }),
    ];
    styles['MultiPoint'] = styles['Point'];

    styles['GeometryCollection'] = styles['Polygon'].concat(
      styles['LineString'],
      styles['Point']
    );

    return styles;
  }

  /**
   * Function that is called with a feature and returns its default geometry.
   * @param {import("../Feature.js").FeatureLike} feature Feature to get the geometry for.
   * @return {import("../geom/Geometry.js").default|import("../render/Feature.js").default|undefined} Geometry to render.
   */
  function defaultGeometryFunction(feature) {
    return feature.getGeometry();
  }

  /**
   * @module ol/style/Text
   */

  /**
   * @typedef {'point' | 'line'} TextPlacement
   * Default text placement is `'point'`. Note that
   * `'line'` requires the underlying geometry to be a {@link module:ol/geom/LineString~LineString},
   * {@link module:ol/geom/Polygon~Polygon}, {@link module:ol/geom/MultiLineString~MultiLineString} or
   * {@link module:ol/geom/MultiPolygon~MultiPolygon}.
   */

  /**
   * @typedef {'left' | 'center' | 'right'} TextJustify
   */

  /**
   * The default fill color to use if no fill was set at construction time; a
   * blackish `#333`.
   *
   * @const {string}
   */
  const DEFAULT_FILL_COLOR = '#333';

  /**
   * @typedef {Object} Options
   * @property {string} [font] Font style as CSS `font` value, see:
   * https://developer.mozilla.org/en-US/docs/Web/API/CanvasRenderingContext2D/font. Default is `'10px sans-serif'`
   * @property {number} [maxAngle=Math.PI/4] When `placement` is set to `'line'`, allow a maximum angle between adjacent characters.
   * The expected value is in radians, and the default is 45° (`Math.PI / 4`).
   * @property {number} [offsetX=0] Horizontal text offset in pixels. A positive will shift the text right.
   * @property {number} [offsetY=0] Vertical text offset in pixels. A positive will shift the text down.
   * @property {boolean} [overflow=false] For polygon labels or when `placement` is set to `'line'`, allow text to exceed
   * the width of the polygon at the label position or the length of the path that it follows.
   * @property {TextPlacement} [placement='point'] Text placement.
   * @property {number|import("../size.js").Size} [scale] Scale.
   * @property {boolean} [rotateWithView=false] Whether to rotate the text with the view.
   * @property {number} [rotation=0] Rotation in radians (positive rotation clockwise).
   * @property {string|Array<string>} [text] Text content or rich text content. For plain text provide a string, which can
   * contain line breaks (`\n`). For rich text provide an array of text/font tuples. A tuple consists of the text to
   * render and the font to use (or `''` to use the text style's font). A line break has to be a separate tuple (i.e. `'\n', ''`).
   * **Example:** `['foo', 'bold 10px sans-serif', ' bar', 'italic 10px sans-serif', ' baz', '']` will yield "**foo** *bar* baz".
   * **Note:** Rich text is not supported for the immediate rendering API.
   * @property {CanvasTextAlign} [textAlign] Text alignment. Possible values: `'left'`, `'right'`, `'center'`, `'end'` or `'start'`.
   * Default is `'center'` for `placement: 'point'`. For `placement: 'line'`, the default is to let the renderer choose a
   * placement where `maxAngle` is not exceeded.
   * @property {TextJustify} [justify] Text justification within the text box.
   * If not set, text is justified towards the `textAlign` anchor.
   * Otherwise, use options `'left'`, `'center'`, or `'right'` to justify the text within the text box.
   * **Note:** `justify` is ignored for immediate rendering and also for `placement: 'line'`.
   * @property {CanvasTextBaseline} [textBaseline='middle'] Text base line. Possible values: `'bottom'`, `'top'`, `'middle'`, `'alphabetic'`,
   * `'hanging'`, `'ideographic'`.
   * @property {import("./Fill.js").default} [fill] Fill style. If none is provided, we'll use a dark fill-style (#333).
   * @property {import("./Stroke.js").default} [stroke] Stroke style.
   * @property {import("./Fill.js").default} [backgroundFill] Fill style for the text background when `placement` is
   * `'point'`. Default is no fill.
   * @property {import("./Stroke.js").default} [backgroundStroke] Stroke style for the text background  when `placement`
   * is `'point'`. Default is no stroke.
   * @property {Array<number>} [padding=[0, 0, 0, 0]] Padding in pixels around the text for decluttering and background. The order of
   * values in the array is `[top, right, bottom, left]`.
   */

  /**
   * @classdesc
   * Set text style for vector features.
   * @api
   */
  class Text {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      options = options || {};

      /**
       * @private
       * @type {string|undefined}
       */
      this.font_ = options.font;

      /**
       * @private
       * @type {number|undefined}
       */
      this.rotation_ = options.rotation;

      /**
       * @private
       * @type {boolean|undefined}
       */
      this.rotateWithView_ = options.rotateWithView;

      /**
       * @private
       * @type {number|import("../size.js").Size|undefined}
       */
      this.scale_ = options.scale;

      /**
       * @private
       * @type {import("../size.js").Size}
       */
      this.scaleArray_ = toSize(options.scale !== undefined ? options.scale : 1);

      /**
       * @private
       * @type {string|Array<string>|undefined}
       */
      this.text_ = options.text;

      /**
       * @private
       * @type {CanvasTextAlign|undefined}
       */
      this.textAlign_ = options.textAlign;

      /**
       * @private
       * @type {TextJustify|undefined}
       */
      this.justify_ = options.justify;

      /**
       * @private
       * @type {CanvasTextBaseline|undefined}
       */
      this.textBaseline_ = options.textBaseline;

      /**
       * @private
       * @type {import("./Fill.js").default}
       */
      this.fill_ =
        options.fill !== undefined
          ? options.fill
          : new Fill({color: DEFAULT_FILL_COLOR});

      /**
       * @private
       * @type {number}
       */
      this.maxAngle_ =
        options.maxAngle !== undefined ? options.maxAngle : Math.PI / 4;

      /**
       * @private
       * @type {TextPlacement}
       */
      this.placement_ =
        options.placement !== undefined ? options.placement : 'point';

      /**
       * @private
       * @type {boolean}
       */
      this.overflow_ = !!options.overflow;

      /**
       * @private
       * @type {import("./Stroke.js").default}
       */
      this.stroke_ = options.stroke !== undefined ? options.stroke : null;

      /**
       * @private
       * @type {number}
       */
      this.offsetX_ = options.offsetX !== undefined ? options.offsetX : 0;

      /**
       * @private
       * @type {number}
       */
      this.offsetY_ = options.offsetY !== undefined ? options.offsetY : 0;

      /**
       * @private
       * @type {import("./Fill.js").default}
       */
      this.backgroundFill_ = options.backgroundFill
        ? options.backgroundFill
        : null;

      /**
       * @private
       * @type {import("./Stroke.js").default}
       */
      this.backgroundStroke_ = options.backgroundStroke
        ? options.backgroundStroke
        : null;

      /**
       * @private
       * @type {Array<number>|null}
       */
      this.padding_ = options.padding === undefined ? null : options.padding;
    }

    /**
     * Clones the style.
     * @return {Text} The cloned style.
     * @api
     */
    clone() {
      const scale = this.getScale();
      return new Text({
        font: this.getFont(),
        placement: this.getPlacement(),
        maxAngle: this.getMaxAngle(),
        overflow: this.getOverflow(),
        rotation: this.getRotation(),
        rotateWithView: this.getRotateWithView(),
        scale: Array.isArray(scale) ? scale.slice() : scale,
        text: this.getText(),
        textAlign: this.getTextAlign(),
        justify: this.getJustify(),
        textBaseline: this.getTextBaseline(),
        fill: this.getFill() ? this.getFill().clone() : undefined,
        stroke: this.getStroke() ? this.getStroke().clone() : undefined,
        offsetX: this.getOffsetX(),
        offsetY: this.getOffsetY(),
        backgroundFill: this.getBackgroundFill()
          ? this.getBackgroundFill().clone()
          : undefined,
        backgroundStroke: this.getBackgroundStroke()
          ? this.getBackgroundStroke().clone()
          : undefined,
        padding: this.getPadding() || undefined,
      });
    }

    /**
     * Get the `overflow` configuration.
     * @return {boolean} Let text overflow the length of the path they follow.
     * @api
     */
    getOverflow() {
      return this.overflow_;
    }

    /**
     * Get the font name.
     * @return {string|undefined} Font.
     * @api
     */
    getFont() {
      return this.font_;
    }

    /**
     * Get the maximum angle between adjacent characters.
     * @return {number} Angle in radians.
     * @api
     */
    getMaxAngle() {
      return this.maxAngle_;
    }

    /**
     * Get the label placement.
     * @return {TextPlacement} Text placement.
     * @api
     */
    getPlacement() {
      return this.placement_;
    }

    /**
     * Get the x-offset for the text.
     * @return {number} Horizontal text offset.
     * @api
     */
    getOffsetX() {
      return this.offsetX_;
    }

    /**
     * Get the y-offset for the text.
     * @return {number} Vertical text offset.
     * @api
     */
    getOffsetY() {
      return this.offsetY_;
    }

    /**
     * Get the fill style for the text.
     * @return {import("./Fill.js").default} Fill style.
     * @api
     */
    getFill() {
      return this.fill_;
    }

    /**
     * Determine whether the text rotates with the map.
     * @return {boolean|undefined} Rotate with map.
     * @api
     */
    getRotateWithView() {
      return this.rotateWithView_;
    }

    /**
     * Get the text rotation.
     * @return {number|undefined} Rotation.
     * @api
     */
    getRotation() {
      return this.rotation_;
    }

    /**
     * Get the text scale.
     * @return {number|import("../size.js").Size|undefined} Scale.
     * @api
     */
    getScale() {
      return this.scale_;
    }

    /**
     * Get the symbolizer scale array.
     * @return {import("../size.js").Size} Scale array.
     */
    getScaleArray() {
      return this.scaleArray_;
    }

    /**
     * Get the stroke style for the text.
     * @return {import("./Stroke.js").default} Stroke style.
     * @api
     */
    getStroke() {
      return this.stroke_;
    }

    /**
     * Get the text to be rendered.
     * @return {string|Array<string>|undefined} Text.
     * @api
     */
    getText() {
      return this.text_;
    }

    /**
     * Get the text alignment.
     * @return {CanvasTextAlign|undefined} Text align.
     * @api
     */
    getTextAlign() {
      return this.textAlign_;
    }

    /**
     * Get the justification.
     * @return {TextJustify|undefined} Justification.
     * @api
     */
    getJustify() {
      return this.justify_;
    }

    /**
     * Get the text baseline.
     * @return {CanvasTextBaseline|undefined} Text baseline.
     * @api
     */
    getTextBaseline() {
      return this.textBaseline_;
    }

    /**
     * Get the background fill style for the text.
     * @return {import("./Fill.js").default} Fill style.
     * @api
     */
    getBackgroundFill() {
      return this.backgroundFill_;
    }

    /**
     * Get the background stroke style for the text.
     * @return {import("./Stroke.js").default} Stroke style.
     * @api
     */
    getBackgroundStroke() {
      return this.backgroundStroke_;
    }

    /**
     * Get the padding for the text.
     * @return {Array<number>|null} Padding.
     * @api
     */
    getPadding() {
      return this.padding_;
    }

    /**
     * Set the `overflow` property.
     *
     * @param {boolean} overflow Let text overflow the path that it follows.
     * @api
     */
    setOverflow(overflow) {
      this.overflow_ = overflow;
    }

    /**
     * Set the font.
     *
     * @param {string|undefined} font Font.
     * @api
     */
    setFont(font) {
      this.font_ = font;
    }

    /**
     * Set the maximum angle between adjacent characters.
     *
     * @param {number} maxAngle Angle in radians.
     * @api
     */
    setMaxAngle(maxAngle) {
      this.maxAngle_ = maxAngle;
    }

    /**
     * Set the x offset.
     *
     * @param {number} offsetX Horizontal text offset.
     * @api
     */
    setOffsetX(offsetX) {
      this.offsetX_ = offsetX;
    }

    /**
     * Set the y offset.
     *
     * @param {number} offsetY Vertical text offset.
     * @api
     */
    setOffsetY(offsetY) {
      this.offsetY_ = offsetY;
    }

    /**
     * Set the text placement.
     *
     * @param {TextPlacement} placement Placement.
     * @api
     */
    setPlacement(placement) {
      this.placement_ = placement;
    }

    /**
     * Set whether to rotate the text with the view.
     *
     * @param {boolean} rotateWithView Rotate with map.
     * @api
     */
    setRotateWithView(rotateWithView) {
      this.rotateWithView_ = rotateWithView;
    }

    /**
     * Set the fill.
     *
     * @param {import("./Fill.js").default} fill Fill style.
     * @api
     */
    setFill(fill) {
      this.fill_ = fill;
    }

    /**
     * Set the rotation.
     *
     * @param {number|undefined} rotation Rotation.
     * @api
     */
    setRotation(rotation) {
      this.rotation_ = rotation;
    }

    /**
     * Set the scale.
     *
     * @param {number|import("../size.js").Size|undefined} scale Scale.
     * @api
     */
    setScale(scale) {
      this.scale_ = scale;
      this.scaleArray_ = toSize(scale !== undefined ? scale : 1);
    }

    /**
     * Set the stroke.
     *
     * @param {import("./Stroke.js").default} stroke Stroke style.
     * @api
     */
    setStroke(stroke) {
      this.stroke_ = stroke;
    }

    /**
     * Set the text.
     *
     * @param {string|Array<string>|undefined} text Text.
     * @api
     */
    setText(text) {
      this.text_ = text;
    }

    /**
     * Set the text alignment.
     *
     * @param {CanvasTextAlign|undefined} textAlign Text align.
     * @api
     */
    setTextAlign(textAlign) {
      this.textAlign_ = textAlign;
    }

    /**
     * Set the justification.
     *
     * @param {TextJustify|undefined} justify Justification.
     * @api
     */
    setJustify(justify) {
      this.justify_ = justify;
    }

    /**
     * Set the text baseline.
     *
     * @param {CanvasTextBaseline|undefined} textBaseline Text baseline.
     * @api
     */
    setTextBaseline(textBaseline) {
      this.textBaseline_ = textBaseline;
    }

    /**
     * Set the background fill.
     *
     * @param {import("./Fill.js").default} fill Fill style.
     * @api
     */
    setBackgroundFill(fill) {
      this.backgroundFill_ = fill;
    }

    /**
     * Set the background stroke.
     *
     * @param {import("./Stroke.js").default} stroke Stroke style.
     * @api
     */
    setBackgroundStroke(stroke) {
      this.backgroundStroke_ = stroke;
    }

    /**
     * Set the padding (`[top, right, bottom, left]`).
     *
     * @param {Array<number>|null} padding Padding.
     * @api
     */
    setPadding(padding) {
      this.padding_ = padding;
    }
  }

  /**
   * @module ol/style/flat
   */

  /**
   * For static styling, the [layer.setStyle()]{@link module:ol/layer/Vector~VectorLayer#setStyle} method
   * can be called with an object literal that has fill, stroke, text, icon, regular shape, and/or circle properties.
   * @api
   *
   * @typedef {FlatFill & FlatStroke & FlatText & FlatIcon & FlatShape & FlatCircle} FlatStyle
   */

  /**
   * A flat style literal or an array of the same.
   *
   * @typedef {FlatStyle|Array<FlatStyle>} FlatStyleLike
   */

  /**
   * Fill style properties applied to polygon features.
   *
   * @typedef {Object} FlatFill
   * @property {import("../color.js").Color|import("../colorlike.js").ColorLike} [fill-color] The fill color.
   */

  /**
   * Stroke style properties applied to line strings and polygon boundaries.  To apply a stroke, at least one of
   * `stroke-color` or `stroke-width` must be provided.
   *
   * @typedef {Object} FlatStroke
   * @property {import("../color.js").Color|import("../colorlike.js").ColorLike} [stroke-color] The stroke color.
   * @property {number} [stroke-width] Stroke pixel width.
   * @property {CanvasLineCap} [stroke-line-cap='round'] Line cap style: `butt`, `round`, or `square`.
   * @property {CanvasLineJoin} [stroke-line-join='round'] Line join style: `bevel`, `round`, or `miter`.
   * @property {Array<number>} [stroke-line-dash] Line dash pattern.
   * @property {number} [stroke-line-dash-offset=0] Line dash offset.
   * @property {number} [stroke-miter-limit=10] Miter limit.
   */

  /**
   * Label style properties applied to all features.  At a minimum, a `text-value` must be provided.
   *
   * @typedef {Object} FlatText
   * @property {string|Array<string>} [text-value] Text content or rich text content. For plain text provide a string, which can
   * contain line breaks (`\n`). For rich text provide an array of text/font tuples. A tuple consists of the text to
   * render and the font to use (or `''` to use the text style's font). A line break has to be a separate tuple (i.e. `'\n', ''`).
   * **Example:** `['foo', 'bold 10px sans-serif', ' bar', 'italic 10px sans-serif', ' baz', '']` will yield "**foo** *bar* baz".
   * **Note:** Rich text is not supported for the immediate rendering API.
   * @property {string} [text-font] Font style as CSS `font` value, see:
   * https://developer.mozilla.org/en-US/docs/Web/API/CanvasRenderingContext2D/font. Default is `'10px sans-serif'`
   * @property {number} [text-max-angle=Math.PI/4] When `text-placement` is set to `'line'`, allow a maximum angle between adjacent characters.
   * The expected value is in radians, and the default is 45° (`Math.PI / 4`).
   * @property {number} [text-offset-x=0] Horizontal text offset in pixels. A positive will shift the text right.
   * @property {number} [text-offset-y=0] Vertical text offset in pixels. A positive will shift the text down.
   * @property {boolean} [text-overflow=false] For polygon labels or when `placement` is set to `'line'`, allow text to exceed
   * the width of the polygon at the label position or the length of the path that it follows.
   * @property {import("./Text.js").TextPlacement} [text-placement='point'] Text placement.
   * @property {number|import("../size.js").Size} [text-scale] Scale.
   * @property {boolean} [text-rotate-with-view=false] Whether to rotate the text with the view.
   * @property {number} [text-rotation=0] Rotation in radians (positive rotation clockwise).
   * @property {CanvasTextAlign} [text-align] Text alignment. Possible values: `'left'`, `'right'`, `'center'`, `'end'` or `'start'`.
   * Default is `'center'` for `text-placement: 'point'`. For `text-placement: 'line'`, the default is to let the renderer choose a
   * placement where `text-max-angle` is not exceeded.
   * @property {import('./Text.js').TextJustify} [text-justify] Text justification within the text box.
   * If not set, text is justified towards the `textAlign` anchor.
   * Otherwise, use options `'left'`, `'center'`, or `'right'` to justify the text within the text box.
   * **Note:** `text-justify` is ignored for immediate rendering and also for `text-placement: 'line'`.
   * @property {CanvasTextBaseline} [text-baseline='middle'] Text base line. Possible values: `'bottom'`, `'top'`, `'middle'`, `'alphabetic'`,
   * `'hanging'`, `'ideographic'`.
   * @property {Array<number>} [text-padding=[0, 0, 0, 0]] Padding in pixels around the text for decluttering and background. The order of
   * values in the array is `[top, right, bottom, left]`.
   * @property {import("../color.js").Color|import("../colorlike.js").ColorLike} [text-fill-color] The fill color.
   * @property {import("../color.js").Color|import("../colorlike.js").ColorLike} [text-background-fill-color] The fill color.
   * @property {import("../color.js").Color|import("../colorlike.js").ColorLike} [text-stroke-color] The stroke color.
   * @property {CanvasLineCap} [text-stroke-line-cap='round'] Line cap style: `butt`, `round`, or `square`.
   * @property {CanvasLineJoin} [text-stroke-line-join='round'] Line join style: `bevel`, `round`, or `miter`.
   * @property {Array<number>} [text-stroke-line-dash] Line dash pattern.
   * @property {number} [text-stroke-line-dash-offset=0] Line dash offset.
   * @property {number} [text-stroke-miter-limit=10] Miter limit.
   * @property {number} [text-stroke-width] Stroke pixel width.
   * @property {import("../color.js").Color|import("../colorlike.js").ColorLike} [text-background-stroke-color] The stroke color.
   * @property {CanvasLineCap} [text-background-stroke-line-cap='round'] Line cap style: `butt`, `round`, or `square`.
   * @property {CanvasLineJoin} [text-background-stroke-line-join='round'] Line join style: `bevel`, `round`, or `miter`.
   * @property {Array<number>} [text-background-stroke-line-dash] Line dash pattern.
   * @property {number} [text-background-stroke-line-dash-offset=0] Line dash offset.
   * @property {number} [text-background-stroke-miter-limit=10] Miter limit.
   * @property {number} [text-background-stroke-width] Stroke pixel width.
   */

  /**
   * Icon style properties applied to point features.  One of `icon-src` or `icon-img` must be provided to render
   * points with an icon.
   *
   * @typedef {Object} FlatIcon
   * @property {string} [icon-src] Image source URI.
   * @property {HTMLImageElement|HTMLCanvasElement} [icon-img] Image object for the icon. If the `icon-src` option is not provided then the
   * provided image must already be loaded. And in that case, it is required
   * to provide the size of the image, with the `icon-img-size` option.
   * @property {import("../size.js").Size} [icon-img-size] Image size in pixels. Only required if `icon-img` is set and `icon-src` is not.
   * The provided size needs to match the actual size of the image.
   * @property {Array<number>} [icon-anchor=[0.5, 0.5]] Anchor. Default value is the icon center.
   * @property {import("./Icon.js").IconOrigin} [icon-anchor-origin='top-left'] Origin of the anchor: `bottom-left`, `bottom-right`,
   * `top-left` or `top-right`.
   * @property {import("./Icon.js").IconAnchorUnits} [icon-anchor-x-units='fraction'] Units in which the anchor x value is
   * specified. A value of `'fraction'` indicates the x value is a fraction of the icon. A value of `'pixels'` indicates
   * the x value in pixels.
   * @property {import("./Icon.js").IconAnchorUnits} [icon-anchor-y-units='fraction'] Units in which the anchor y value is
   * specified. A value of `'fraction'` indicates the y value is a fraction of the icon. A value of `'pixels'` indicates
   * the y value in pixels.
   * @property {import("../color.js").Color|string} [icon-color] Color to tint the icon. If not specified,
   * the icon will be left as is.
   * @property {null|string} [icon-cross-origin] The `crossOrigin` attribute for loaded images. Note that you must provide a
   * `icon-cross-origin` value if you want to access pixel data with the Canvas renderer.
   * See https://developer.mozilla.org/en-US/docs/Web/HTML/CORS_enabled_image for more detail.
   * @property {Array<number>} [icon-offset=[0, 0]] Offset, which, together with the size and the offset origin, define the
   * sub-rectangle to use from the original icon image.
   * @property {Array<number>} [icon-displacement=[0,0]] Displacement of the icon.
   * @property {import("./Icon.js").IconOrigin} [icon-offset-origin='top-left'] Origin of the offset: `bottom-left`, `bottom-right`,
   * `top-left` or `top-right`.
   * @property {number} [icon-opacity=1] Opacity of the icon.
   * @property {number|import("../size.js").Size} [icon-scale=1] Scale.
   * @property {number} [icon-rotation=0] Rotation in radians (positive rotation clockwise).
   * @property {boolean} [icon-rotate-with-view=false] Whether to rotate the icon with the view.
   * @property {import("../size.js").Size} [icon-size] Icon size in pixel. Can be used together with `icon-offset` to define the
   * sub-rectangle to use from the origin (sprite) icon image.
   * @property {"declutter"|"obstacle"|"none"|undefined} [icon-declutter-mode] Declutter mode
   */

  /**
   * Regular shape style properties for rendering point features.  At least `shape-points` must be provided.
   *
   * @typedef {Object} FlatShape
   * @property {number} [shape-points] Number of points for stars and regular polygons. In case of a polygon, the number of points
   * is the number of sides.
   * @property {import("../color.js").Color|import("../colorlike.js").ColorLike} [shape-fill-color] The fill color.
   * @property {import("../color.js").Color|import("../colorlike.js").ColorLike} [shape-stroke-color] The stroke color.
   * @property {number} [shape-stroke-width] Stroke pixel width.
   * @property {CanvasLineCap} [shape-stroke-line-cap='round'] Line cap style: `butt`, `round`, or `square`.
   * @property {CanvasLineJoin} [shape-stroke-line-join='round'] Line join style: `bevel`, `round`, or `miter`.
   * @property {Array<number>} [shape-stroke-line-dash] Line dash pattern.
   * @property {number} [shape-stroke-line-dash-offset=0] Line dash offset.
   * @property {number} [shape-stroke-miter-limit=10] Miter limit.
   * @property {number} [shape-radius] Radius of a regular polygon.
   * @property {number} [shape-radius1] First radius of a star. Ignored if radius is set.
   * @property {number} [shape-radius2] Second radius of a star.
   * @property {number} [shape-angle=0] Shape's angle in radians. A value of 0 will have one of the shape's point facing up.
   * @property {Array<number>} [shape-displacement=[0,0]] Displacement of the shape
   * @property {number} [shape-rotation=0] Rotation in radians (positive rotation clockwise).
   * @property {boolean} [shape-rotate-with-view=false] Whether to rotate the shape with the view.
   * @property {number|import("../size.js").Size} [shape-scale=1] Scale. Unless two dimensional scaling is required a better
   * result may be obtained with appropriate settings for `shape-radius`, `shape-radius1` and `shape-radius2`.
   * @property {"declutter"|"obstacle"|"none"|undefined} [shape-declutter-mode] Declutter mode.
   */

  /**
   * Circle style properties for rendering point features.  At least `circle-radius` must be provided.
   *
   * @typedef {Object} FlatCircle
   * @property {number} [circle-radius] Circle radius.
   * @property {import("../color.js").Color|import("../colorlike.js").ColorLike} [circle-fill-color] The fill color.
   * @property {import("../color.js").Color|import("../colorlike.js").ColorLike} [circle-stroke-color] The stroke color.
   * @property {number} [circle-stroke-width] Stroke pixel width.
   * @property {CanvasLineCap} [circle-stroke-line-cap='round'] Line cap style: `butt`, `round`, or `square`.
   * @property {CanvasLineJoin} [circle-stroke-line-join='round'] Line join style: `bevel`, `round`, or `miter`.
   * @property {Array<number>} [circle-stroke-line-dash] Line dash pattern.
   * @property {number} [circle-stroke-line-dash-offset=0] Line dash offset.
   * @property {number} [circle-stroke-miter-limit=10] Miter limit.
   * @property {Array<number>} [circle-displacement=[0,0]] displacement
   * @property {number|import("../size.js").Size} [circle-scale=1] Scale. A two dimensional scale will produce an ellipse.
   * Unless two dimensional scaling is required a better result may be obtained with an appropriate setting for `circle-radius`.
   * @property {number} [circle-rotation=0] Rotation in radians
   * (positive rotation clockwise, meaningful only when used in conjunction with a two dimensional scale).
   * @property {boolean} [circle-rotate-with-view=false] Whether to rotate the shape with the view
   * (meaningful only when used in conjunction with a two dimensional scale).
   * @property {"declutter"|"obstacle"|"none"|undefined} [circle-declutter-mode] Declutter mode
   */

  /**
   * @param {FlatStyle} flatStyle A flat style literal.
   * @return {import("./Style.js").default} A style instance.
   */
  function toStyle(flatStyle) {
    const style = new Style({
      fill: getFill(flatStyle, ''),
      stroke: getStroke(flatStyle, ''),
      text: getText$1(flatStyle),
      image: getImage(flatStyle),
    });

    return style;
  }

  /**
   * @param {FlatStyle} flatStyle The flat style.
   * @param {string} prefix The property prefix.
   * @return {Fill|undefined} The fill (if any).
   */
  function getFill(flatStyle, prefix) {
    const color = flatStyle[prefix + 'fill-color'];
    if (!color) {
      return;
    }

    return new Fill({color: color});
  }

  /**
   * @param {FlatStyle} flatStyle The flat style.
   * @param {string} prefix The property prefix.
   * @return {Stroke|undefined} The stroke (if any).
   */
  function getStroke(flatStyle, prefix) {
    const width = flatStyle[prefix + 'stroke-width'];
    const color = flatStyle[prefix + 'stroke-color'];
    if (!width && !color) {
      return;
    }

    return new Stroke({
      width: width,
      color: color,
      lineCap: flatStyle[prefix + 'stroke-line-cap'],
      lineJoin: flatStyle[prefix + 'stroke-line-join'],
      lineDash: flatStyle[prefix + 'stroke-line-dash'],
      lineDashOffset: flatStyle[prefix + 'stroke-line-dash-offset'],
      miterLimit: flatStyle[prefix + 'stroke-miter-limit'],
    });
  }

  /**
   * @param {FlatStyle} flatStyle The flat style.
   * @return {Text|undefined} The text (if any).
   */
  function getText$1(flatStyle) {
    const value = flatStyle['text-value'];
    if (!value) {
      return;
    }

    const text = new Text({
      text: value,
      font: flatStyle['text-font'],
      maxAngle: flatStyle['text-max-angle'],
      offsetX: flatStyle['text-offset-x'],
      offsetY: flatStyle['text-offset-y'],
      overflow: flatStyle['text-overflow'],
      placement: flatStyle['text-placement'],
      scale: flatStyle['text-scale'],
      rotateWithView: flatStyle['text-rotate-with-view'],
      rotation: flatStyle['text-rotation'],
      textAlign: flatStyle['text-align'],
      justify: flatStyle['text-justify'],
      textBaseline: flatStyle['text-baseline'],
      padding: flatStyle['text-padding'],
      fill: getFill(flatStyle, 'text-'),
      backgroundFill: getFill(flatStyle, 'text-background-'),
      stroke: getStroke(flatStyle, 'text-'),
      backgroundStroke: getStroke(flatStyle, 'text-background-'),
    });

    return text;
  }

  /**
   * @param {FlatStyle} flatStyle The flat style.
   * @return {import("./Image.js").default|undefined} The image (if any).
   */
  function getImage(flatStyle) {
    const iconSrc = flatStyle['icon-src'];
    const iconImg = flatStyle['icon-img'];
    if (iconSrc || iconImg) {
      const icon = new Icon({
        src: iconSrc,
        img: iconImg,
        imgSize: flatStyle['icon-img-size'],
        anchor: flatStyle['icon-anchor'],
        anchorOrigin: flatStyle['icon-anchor-origin'],
        anchorXUnits: flatStyle['icon-anchor-x-units'],
        anchorYUnits: flatStyle['icon-anchor-y-units'],
        color: flatStyle['icon-color'],
        crossOrigin: flatStyle['icon-cross-origin'],
        offset: flatStyle['icon-offset'],
        displacement: flatStyle['icon-displacement'],
        opacity: flatStyle['icon-opacity'],
        scale: flatStyle['icon-scale'],
        rotation: flatStyle['icon-rotation'],
        rotateWithView: flatStyle['icon-rotate-with-view'],
        size: flatStyle['icon-size'],
        declutterMode: flatStyle['icon-declutter-mode'],
      });
      return icon;
    }

    const shapePoints = flatStyle['shape-points'];
    if (shapePoints) {
      const prefix = 'shape-';
      const shape = new RegularShape({
        points: shapePoints,
        fill: getFill(flatStyle, prefix),
        stroke: getStroke(flatStyle, prefix),
        radius: flatStyle['shape-radius'],
        radius1: flatStyle['shape-radius1'],
        radius2: flatStyle['shape-radius2'],
        angle: flatStyle['shape-angle'],
        displacement: flatStyle['shape-displacement'],
        rotation: flatStyle['shape-rotation'],
        rotateWithView: flatStyle['shape-rotate-with-view'],
        scale: flatStyle['shape-scale'],
        declutterMode: flatStyle['shape-declutter-mode'],
      });

      return shape;
    }

    const circleRadius = flatStyle['circle-radius'];
    if (circleRadius) {
      const prefix = 'circle-';
      const circle = new CircleStyle({
        radius: circleRadius,
        fill: getFill(flatStyle, prefix),
        stroke: getStroke(flatStyle, prefix),
        displacement: flatStyle['circle-displacement'],
        scale: flatStyle['circle-scale'],
        rotation: flatStyle['circle-rotation'],
        rotateWithView: flatStyle['circle-rotate-with-view'],
        declutterMode: flatStyle['circle-declutter-mode'],
      });

      return circle;
    }

    return;
  }

  /**
   * @module ol/structs/LinkedList
   */

  /**
   * @typedef {Object} Item
   * @property {Item} [prev] Previous.
   * @property {Item} [next] Next.
   * @property {?} data Data.
   */

  /**
   * @classdesc
   * Creates an empty linked list structure.
   */
  class LinkedList {
    /**
     * @param {boolean} [circular] The last item is connected to the first one,
     * and the first item to the last one. Default is true.
     */
    constructor(circular) {
      /**
       * @private
       * @type {Item|undefined}
       */
      this.first_;

      /**
       * @private
       * @type {Item|undefined}
       */
      this.last_;

      /**
       * @private
       * @type {Item|undefined}
       */
      this.head_;

      /**
       * @private
       * @type {boolean}
       */
      this.circular_ = circular === undefined ? true : circular;

      /**
       * @private
       * @type {number}
       */
      this.length_ = 0;
    }

    /**
     * Inserts an item into the linked list right after the current one.
     *
     * @param {?} data Item data.
     */
    insertItem(data) {
      /** @type {Item} */
      const item = {
        prev: undefined,
        next: undefined,
        data: data,
      };

      const head = this.head_;

      //Initialize the list.
      if (!head) {
        this.first_ = item;
        this.last_ = item;
        if (this.circular_) {
          item.next = item;
          item.prev = item;
        }
      } else {
        //Link the new item to the adjacent ones.
        const next = head.next;
        item.prev = head;
        item.next = next;
        head.next = item;
        if (next) {
          next.prev = item;
        }

        if (head === this.last_) {
          this.last_ = item;
        }
      }
      this.head_ = item;
      this.length_++;
    }

    /**
     * Removes the current item from the list. Sets the cursor to the next item,
     * if possible.
     */
    removeItem() {
      const head = this.head_;
      if (head) {
        const next = head.next;
        const prev = head.prev;
        if (next) {
          next.prev = prev;
        }
        if (prev) {
          prev.next = next;
        }
        this.head_ = next || prev;

        if (this.first_ === this.last_) {
          this.head_ = undefined;
          this.first_ = undefined;
          this.last_ = undefined;
        } else if (this.first_ === head) {
          this.first_ = this.head_;
        } else if (this.last_ === head) {
          this.last_ = prev ? this.head_.prev : this.head_;
        }
        this.length_--;
      }
    }

    /**
     * Sets the cursor to the first item, and returns the associated data.
     *
     * @return {?} Item data.
     */
    firstItem() {
      this.head_ = this.first_;
      if (this.head_) {
        return this.head_.data;
      }
      return undefined;
    }

    /**
     * Sets the cursor to the last item, and returns the associated data.
     *
     * @return {?} Item data.
     */
    lastItem() {
      this.head_ = this.last_;
      if (this.head_) {
        return this.head_.data;
      }
      return undefined;
    }

    /**
     * Sets the cursor to the next item, and returns the associated data.
     *
     * @return {?} Item data.
     */
    nextItem() {
      if (this.head_ && this.head_.next) {
        this.head_ = this.head_.next;
        return this.head_.data;
      }
      return undefined;
    }

    /**
     * Returns the next item's data without moving the cursor.
     *
     * @return {?} Item data.
     */
    getNextItem() {
      if (this.head_ && this.head_.next) {
        return this.head_.next.data;
      }
      return undefined;
    }

    /**
     * Sets the cursor to the previous item, and returns the associated data.
     *
     * @return {?} Item data.
     */
    prevItem() {
      if (this.head_ && this.head_.prev) {
        this.head_ = this.head_.prev;
        return this.head_.data;
      }
      return undefined;
    }

    /**
     * Returns the previous item's data without moving the cursor.
     *
     * @return {?} Item data.
     */
    getPrevItem() {
      if (this.head_ && this.head_.prev) {
        return this.head_.prev.data;
      }
      return undefined;
    }

    /**
     * Returns the current item's data.
     *
     * @return {?} Item data.
     */
    getCurrItem() {
      if (this.head_) {
        return this.head_.data;
      }
      return undefined;
    }

    /**
     * Sets the first item of the list. This only works for circular lists, and sets
     * the last item accordingly.
     */
    setFirstItem() {
      if (this.circular_ && this.head_) {
        this.first_ = this.head_;
        this.last_ = this.head_.prev;
      }
    }

    /**
     * Concatenates two lists.
     * @param {LinkedList} list List to merge into the current list.
     */
    concat(list) {
      if (list.head_) {
        if (this.head_) {
          const end = this.head_.next;
          this.head_.next = list.first_;
          list.first_.prev = this.head_;
          end.prev = list.last_;
          list.last_.next = end;
          this.length_ += list.length_;
        } else {
          this.head_ = list.head_;
          this.first_ = list.first_;
          this.last_ = list.last_;
          this.length_ = list.length_;
        }
        list.head_ = undefined;
        list.first_ = undefined;
        list.last_ = undefined;
        list.length_ = 0;
      }
    }

    /**
     * Returns the current length of the list.
     *
     * @return {number} Length.
     */
    getLength() {
      return this.length_;
    }
  }

  /**
   * @module ol/source/ogcTileUtil
   */

  /**
   * See https://ogcapi.ogc.org/tiles/.
   */

  /**
   * @typedef {'map' | 'vector'} TileType
   */

  /**
   * @typedef {'topLeft' | 'bottomLeft'} CornerOfOrigin
   */

  /**
   * @typedef {Object} TileSet
   * @property {TileType} dataType Type of data represented in the tileset.
   * @property {string} [tileMatrixSetDefinition] Reference to a tile matrix set definition.
   * @property {TileMatrixSet} [tileMatrixSet] Tile matrix set definition.
   * @property {Array<TileMatrixSetLimit>} [tileMatrixSetLimits] Tile matrix set limits.
   * @property {Array<Link>} links Tileset links.
   */

  /**
   * @typedef {Object} Link
   * @property {string} rel The link rel attribute.
   * @property {string} href The link URL.
   * @property {string} type The link type.
   */

  /**
   * @typedef {Object} TileMatrixSetLimit
   * @property {string} tileMatrix The tile matrix id.
   * @property {number} minTileRow The minimum tile row.
   * @property {number} maxTileRow The maximum tile row.
   * @property {number} minTileCol The minimum tile column.
   * @property {number} maxTileCol The maximum tile column.
   */

  /**
   * @typedef {Object} TileMatrixSet
   * @property {string} id The tile matrix set identifier.
   * @property {string} crs The coordinate reference system.
   * @property {Array<TileMatrix>} tileMatrices Array of tile matrices.
   */

  /**
   * @typedef {Object} TileMatrix
   * @property {string} id The tile matrix identifier.
   * @property {number} cellSize The pixel resolution (map units per pixel).
   * @property {Array<number>} pointOfOrigin The map location of the matrix origin.
   * @property {CornerOfOrigin} [cornerOfOrigin='topLeft'] The corner of the matrix that represents the origin ('topLeft' or 'bottomLeft').
   * @property {number} matrixWidth The number of columns.
   * @property {number} matrixHeight The number of rows.
   * @property {number} tileWidth The pixel width of a tile.
   * @property {number} tileHeight The pixel height of a tile.
   */

  /**
   * @type {Object<string, boolean>}
   */
  const knownMapMediaTypes = {
    'image/png': true,
    'image/jpeg': true,
    'image/gif': true,
    'image/webp': true,
  };

  /**
   * @type {Object<string, boolean>}
   */
  const knownVectorMediaTypes = {
    'application/vnd.mapbox-vector-tile': true,
    'application/geo+json': true,
  };

  /**
   * @typedef {Object} TileSetInfo
   * @property {string} urlTemplate The tile URL template.
   * @property {import("../tilegrid/TileGrid.js").default} grid The tile grid.
   * @property {import("../Tile.js").UrlFunction} urlFunction The tile URL function.
   */

  /**
   * @typedef {Object} SourceInfo
   * @property {string} url The tile set URL.
   * @property {string} mediaType The preferred tile media type.
   * @property {Array<string>} [supportedMediaTypes] The supported media types.
   * @property {import("../proj/Projection.js").default} projection The source projection.
   * @property {Object} [context] Optional context for constructing the URL.
   */

  /**
   * @param {Array<Link>} links Tileset links.
   * @param {string} [mediaType] The preferred media type.
   * @return {string} The tile URL template.
   */
  function getMapTileUrlTemplate(links, mediaType) {
    let tileUrlTemplate;
    let fallbackUrlTemplate;
    for (let i = 0; i < links.length; ++i) {
      const link = links[i];
      if (link.rel === 'item') {
        if (link.type === mediaType) {
          tileUrlTemplate = link.href;
          break;
        }
        if (knownMapMediaTypes[link.type]) {
          fallbackUrlTemplate = link.href;
        } else if (!fallbackUrlTemplate && link.type.startsWith('image/')) {
          fallbackUrlTemplate = link.href;
        }
      }
    }

    if (!tileUrlTemplate) {
      if (fallbackUrlTemplate) {
        tileUrlTemplate = fallbackUrlTemplate;
      } else {
        throw new Error('Could not find "item" link');
      }
    }

    return tileUrlTemplate;
  }

  /**
   * @param {Array<Link>} links Tileset links.
   * @param {string} [mediaType] The preferred media type.
   * @param {Array<string>} [supportedMediaTypes] The media types supported by the parser.
   * @return {string} The tile URL template.
   */
  function getVectorTileUrlTemplate(
    links,
    mediaType,
    supportedMediaTypes
  ) {
    let tileUrlTemplate;
    let fallbackUrlTemplate;

    /**
     * Lookup of URL by media type.
     * @type {Object<string, string>}
     */
    const hrefLookup = {};

    for (let i = 0; i < links.length; ++i) {
      const link = links[i];
      hrefLookup[link.type] = link.href;
      if (link.rel === 'item') {
        if (link.type === mediaType) {
          tileUrlTemplate = link.href;
          break;
        }
        if (knownVectorMediaTypes[link.type]) {
          fallbackUrlTemplate = link.href;
        }
      }
    }

    if (!tileUrlTemplate && supportedMediaTypes) {
      for (let i = 0; i < supportedMediaTypes.length; ++i) {
        const supportedMediaType = supportedMediaTypes[i];
        if (hrefLookup[supportedMediaType]) {
          tileUrlTemplate = hrefLookup[supportedMediaType];
          break;
        }
      }
    }

    if (!tileUrlTemplate) {
      if (fallbackUrlTemplate) {
        tileUrlTemplate = fallbackUrlTemplate;
      } else {
        throw new Error('Could not find "item" link');
      }
    }

    return tileUrlTemplate;
  }

  /**
   * @param {SourceInfo} sourceInfo The source info.
   * @param {TileMatrixSet} tileMatrixSet Tile matrix set.
   * @param {string} tileUrlTemplate Tile URL template.
   * @param {Array<TileMatrixSetLimit>} [tileMatrixSetLimits] Tile matrix set limits.
   * @return {TileSetInfo} Tile set info.
   */
  function parseTileMatrixSet(
    sourceInfo,
    tileMatrixSet,
    tileUrlTemplate,
    tileMatrixSetLimits
  ) {
    let projection = sourceInfo.projection;
    if (!projection) {
      projection = get$2(tileMatrixSet.crs);
      if (!projection) {
        throw new Error(`Unsupported CRS: ${tileMatrixSet.crs}`);
      }
    }
    const backwards = projection.getAxisOrientation().substr(0, 2) !== 'en';

    const matrices = tileMatrixSet.tileMatrices;

    /**
     * @type {Object<string, TileMatrix>}
     */
    const matrixLookup = {};
    for (let i = 0; i < matrices.length; ++i) {
      const matrix = matrices[i];
      matrixLookup[matrix.id] = matrix;
    }

    /**
     * @type {Object<string, TileMatrixSetLimit>}
     */
    const limitLookup = {};

    /**
     * @type {Array<string>}
     */
    const matrixIds = [];

    if (tileMatrixSetLimits) {
      for (let i = 0; i < tileMatrixSetLimits.length; ++i) {
        const limit = tileMatrixSetLimits[i];
        const id = limit.tileMatrix;
        matrixIds.push(id);
        limitLookup[id] = limit;
      }
    } else {
      for (let i = 0; i < matrices.length; ++i) {
        const id = matrices[i].id;
        matrixIds.push(id);
      }
    }

    const length = matrixIds.length;
    const origins = new Array(length);
    const resolutions = new Array(length);
    const sizes = new Array(length);
    const tileSizes = new Array(length);
    const extent = [-Infinity, -Infinity, Infinity, Infinity];

    for (let i = 0; i < length; ++i) {
      const id = matrixIds[i];
      const matrix = matrixLookup[id];
      const origin = matrix.pointOfOrigin;
      if (backwards) {
        origins[i] = [origin[1], origin[0]];
      } else {
        origins[i] = origin;
      }
      resolutions[i] = matrix.cellSize;
      sizes[i] = [matrix.matrixWidth, matrix.matrixHeight];
      tileSizes[i] = [matrix.tileWidth, matrix.tileHeight];
      const limit = limitLookup[id];
      if (limit) {
        const tileMapWidth = matrix.cellSize * matrix.tileWidth;
        const minX = origins[i][0] + limit.minTileCol * tileMapWidth;
        const maxX = origins[i][0] + (limit.maxTileCol + 1) * tileMapWidth;

        const tileMapHeight = matrix.cellSize * matrix.tileHeight;
        const upsideDown = matrix.cornerOfOrigin === 'bottomLeft';

        let minY;
        let maxY;
        if (upsideDown) {
          minY = origins[i][1] + limit.minTileRow * tileMapHeight;
          maxY = origins[i][1] + (limit.maxTileRow + 1) * tileMapHeight;
        } else {
          minY = origins[i][1] - (limit.maxTileRow + 1) * tileMapHeight;
          maxY = origins[i][1] - limit.minTileRow * tileMapHeight;
        }

        getIntersection(extent, [minX, minY, maxX, maxY], extent);
      }
    }

    const tileGrid = new TileGrid({
      origins: origins,
      resolutions: resolutions,
      sizes: sizes,
      tileSizes: tileSizes,
      extent: tileMatrixSetLimits ? extent : undefined,
    });

    const context = sourceInfo.context;
    const base = sourceInfo.url;

    function tileUrlFunction(tileCoord, pixelRatio, projection) {
      if (!tileCoord) {
        return undefined;
      }

      const id = matrixIds[tileCoord[0]];
      const matrix = matrixLookup[id];
      const upsideDown = matrix.cornerOfOrigin === 'bottomLeft';

      const localContext = {
        tileMatrix: id,
        tileCol: tileCoord[1],
        tileRow: upsideDown ? -tileCoord[2] - 1 : tileCoord[2],
      };

      if (tileMatrixSetLimits) {
        const limit = limitLookup[matrix.id];
        if (
          localContext.tileCol < limit.minTileCol ||
          localContext.tileCol > limit.maxTileCol ||
          localContext.tileRow < limit.minTileRow ||
          localContext.tileRow > limit.maxTileRow
        ) {
          return undefined;
        }
      }

      Object.assign(localContext, context);

      const url = tileUrlTemplate.replace(/\{(\w+?)\}/g, function (m, p) {
        return localContext[p];
      });

      return resolveUrl(base, url);
    }

    return {
      grid: tileGrid,
      urlTemplate: tileUrlTemplate,
      urlFunction: tileUrlFunction,
    };
  }

  /**
   * @param {SourceInfo} sourceInfo The source info.
   * @param {TileSet} tileSet Tile set.
   * @return {TileSetInfo|Promise<TileSetInfo>} Tile set info.
   */
  function parseTileSetMetadata(sourceInfo, tileSet) {
    const tileMatrixSetLimits = tileSet.tileMatrixSetLimits;
    let tileUrlTemplate;

    if (tileSet.dataType === 'map') {
      tileUrlTemplate = getMapTileUrlTemplate(
        tileSet.links,
        sourceInfo.mediaType
      );
    } else if (tileSet.dataType === 'vector') {
      tileUrlTemplate = getVectorTileUrlTemplate(
        tileSet.links,
        sourceInfo.mediaType,
        sourceInfo.supportedMediaTypes
      );
    } else {
      throw new Error('Expected tileset data type to be "map" or "vector"');
    }

    if (tileSet.tileMatrixSet) {
      return parseTileMatrixSet(
        sourceInfo,
        tileSet.tileMatrixSet,
        tileUrlTemplate,
        tileMatrixSetLimits
      );
    }

    const tileMatrixSetLink = tileSet.links.find(
      (link) =>
        link.rel === 'http://www.opengis.net/def/rel/ogc/1.0/tiling-scheme'
    );
    if (!tileMatrixSetLink) {
      throw new Error(
        'Expected http://www.opengis.net/def/rel/ogc/1.0/tiling-scheme link or tileMatrixSet'
      );
    }
    const tileMatrixSetDefinition = tileMatrixSetLink.href;

    const url = resolveUrl(sourceInfo.url, tileMatrixSetDefinition);
    return getJSON(url).then(function (tileMatrixSet) {
      return parseTileMatrixSet(
        sourceInfo,
        tileMatrixSet,
        tileUrlTemplate,
        tileMatrixSetLimits
      );
    });
  }

  /**
   * @param {SourceInfo} sourceInfo Source info.
   * @return {Promise<TileSetInfo>} Tile set info.
   */
  function getTileSetInfo(sourceInfo) {
    return getJSON(sourceInfo.url).then(function (tileSet) {
      return parseTileSetMetadata(sourceInfo, tileSet);
    });
  }

  /**
   * @module ol/source/OGCMapTile
   */

  /**
   * @typedef {Object} Options
   * @property {string} url URL to the OGC Map Tileset endpoint.
   * @property {Object} [context] A lookup of values to use in the tile URL template.  The `{tileMatrix}`
   * (zoom level), `{tileRow}`, and `{tileCol}` variables in the URL will always be provided by the source.
   * @property {string} [mediaType] The content type for the tiles (e.g. "image/png").  If not provided,
   * the source will try to find a link with rel="item" that uses a supported image type.
   * @property {import("../proj.js").ProjectionLike} [projection] Projection. By default, the projection
   * will be derived from the `crs` of the `tileMatrixSet`.  You can override this by supplying
   * a projection to the constructor.
   * @property {import("./Source.js").AttributionLike} [attributions] Attributions.
   * @property {number} [cacheSize] Tile cache size. The default depends on the screen size. Will be ignored if too small.
   * @property {null|string} [crossOrigin] The `crossOrigin` attribute for loaded images.  Note that
   * you must provide a `crossOrigin` value if you want to access pixel data with the Canvas renderer.
   * See https://developer.mozilla.org/en-US/docs/Web/HTML/CORS_enabled_image for more detail.
   * @property {boolean} [interpolate=true] Use interpolated values when resampling.  By default,
   * linear interpolation is used when resampling.  Set to false to use the nearest neighbor instead.
   * @property {number} [reprojectionErrorThreshold=0.5] Maximum allowed reprojection error (in pixels).
   * Higher values can increase reprojection performance, but decrease precision.
   * @property {import("../Tile.js").LoadFunction} [tileLoadFunction] Optional function to load a tile given a URL. The default is
   * ```js
   * function(tile, src) {
   *   tile.getImage().src = src;
   * };
   * ```
   * @property {boolean} [wrapX=true] Whether to wrap the world horizontally.
   * @property {number} [transition] Duration of the opacity transition for rendering.
   * To disable the opacity transition, pass `transition: 0`.
   */

  /**
   * @classdesc
   * Layer source for map tiles from an [OGC API - Tiles](https://ogcapi.ogc.org/tiles/) service that provides "map" type tiles.
   * The service must conform to at least the core (http://www.opengis.net/spec/ogcapi-tiles-1/1.0/conf/core)
   * and tileset (http://www.opengis.net/spec/ogcapi-tiles-1/1.0/conf/tileset) conformance classes.
   */
  class OGCMapTile extends TileImage {
    /**
     * @param {Options} options OGC map tile options.
     */
    constructor(options) {
      super({
        attributions: options.attributions,
        cacheSize: options.cacheSize,
        crossOrigin: options.crossOrigin,
        interpolate: options.interpolate,
        projection: options.projection,
        reprojectionErrorThreshold: options.reprojectionErrorThreshold,
        state: 'loading',
        tileLoadFunction: options.tileLoadFunction,
        wrapX: options.wrapX !== undefined ? options.wrapX : true,
        transition: options.transition,
      });

      const sourceInfo = {
        url: options.url,
        projection: this.getProjection(),
        mediaType: options.mediaType,
        context: options.context || null,
      };

      getTileSetInfo(sourceInfo)
        .then(this.handleTileSetInfo_.bind(this))
        .catch(this.handleError_.bind(this));
    }

    /**
     * @param {import("./ogcTileUtil.js").TileSetInfo} tileSetInfo Tile set info.
     * @private
     */
    handleTileSetInfo_(tileSetInfo) {
      this.tileGrid = tileSetInfo.grid;
      this.setTileUrlFunction(tileSetInfo.urlFunction, tileSetInfo.urlTemplate);
      this.setState('ready');
    }

    /**
     * @private
     * @param {Error} error The error.
     */
    handleError_(error) {
      console.error(error); // eslint-disable-line no-console
      this.setState('error');
    }
  }

  /**
   * @module ol/source/OGCVectorTile
   */

  /**
   * @typedef {Object} Options
   * @property {string} url URL to the OGC Vector Tileset endpoint.
   * @property {Object} [context] A lookup of values to use in the tile URL template.  The `{tileMatrix}`
   * (zoom level), `{tileRow}`, and `{tileCol}` variables in the URL will always be provided by the source.
   * @property {import("../format/Feature.js").default} format Feature parser for tiles.
   * @property {string} [mediaType] The content type for the tiles (e.g. "application/vnd.mapbox-vector-tile").  If not provided,
   * the source will try to find a link with rel="item" that uses a vector type supported by the configured format.
   * @property {import("./Source.js").AttributionLike} [attributions] Attributions.
   * @property {boolean} [attributionsCollapsible=true] Attributions are collapsible.
   * @property {number} [cacheSize] Initial tile cache size. Will auto-grow to hold at least twice the number of tiles in the viewport.
   * @property {boolean} [overlaps=true] This source may have overlapping geometries. Setting this
   * to `false` (e.g. for sources with polygons that represent administrative
   * boundaries or TopoJSON sources) allows the renderer to optimise fill and
   * stroke operations.
   * @property {import("../proj.js").ProjectionLike} [projection='EPSG:3857'] Projection of the tile grid.
   * @property {typeof import("../VectorTile.js").default} [tileClass] Class used to instantiate image tiles.
   * Default is {@link module:ol/VectorTile~VectorTile}.
   * @property {number} [transition] A duration for tile opacity
   * transitions in milliseconds. A duration of 0 disables the opacity transition.
   * @property {boolean} [wrapX=true] Whether to wrap the world horizontally.
   * When set to `false`, only one world
   * will be rendered. When set to `true`, tiles will be wrapped horizontally to
   * render multiple worlds.
   * @property {number|import("../array.js").NearestDirectionFunction} [zDirection=1]
   * Choose whether to use tiles with a higher or lower zoom level when between integer
   * zoom levels. See {@link module:ol/tilegrid/TileGrid~TileGrid#getZForResolution}.
   */

  /**
   * @classdesc
   * Layer source for map tiles from an [OGC API - Tiles](https://ogcapi.ogc.org/tiles/) service that provides "vector" type tiles.
   * The service must conform to at least the core (http://www.opengis.net/spec/ogcapi-tiles-1/1.0/conf/core)
   * and tileset (http://www.opengis.net/spec/ogcapi-tiles-1/1.0/conf/tileset) conformance classes.
   *
   * Vector tile sets may come in a variety of formats (e.g. GeoJSON, MVT).  The `format` option is used to determine
   * which of the advertised media types is used.  If you need to force the use of a particular media type, you can
   * provide the `mediaType` option.
   */
  class OGCVectorTile extends VectorTile {
    /**
     * @param {Options} options OGC vector tile options.
     */
    constructor(options) {
      super({
        attributions: options.attributions,
        attributionsCollapsible: options.attributionsCollapsible,
        cacheSize: options.cacheSize,
        format: options.format,
        overlaps: options.overlaps,
        projection: options.projection,
        tileClass: options.tileClass,
        transition: options.transition,
        wrapX: options.wrapX,
        zDirection: options.zDirection,
        state: 'loading',
      });

      const sourceInfo = {
        url: options.url,
        projection: this.getProjection(),
        mediaType: options.mediaType,
        supportedMediaTypes: options.format.supportedMediaTypes,
        context: options.context || null,
      };

      getTileSetInfo(sourceInfo)
        .then(this.handleTileSetInfo_.bind(this))
        .catch(this.handleError_.bind(this));
    }

    /**
     * @param {import("./ogcTileUtil.js").TileSetInfo} tileSetInfo Tile set info.
     * @private
     */
    handleTileSetInfo_(tileSetInfo) {
      this.tileGrid = tileSetInfo.grid;
      this.setTileUrlFunction(tileSetInfo.urlFunction, tileSetInfo.urlTemplate);
      this.setState('ready');
    }

    /**
     * @private
     * @param {Error} error The error.
     */
    handleError_(error) {
      console.error(error); // eslint-disable-line no-console
      this.setState('error');
    }
  }

  /**
   * @module ol/source/common
   */

  /**
   * Default WMS version.
   * @type {string}
   */
  const DEFAULT_WMS_VERSION = '1.3.0';

  /**
   * @module ol/layer/BaseVector
   */

  /**
   * @template {import("../source/Vector.js").default|import("../source/VectorTile.js").default} VectorSourceType
   * @typedef {Object} Options
   * @property {string} [className='ol-layer'] A CSS class name to set to the layer element.
   * @property {number} [opacity=1] Opacity (0, 1).
   * @property {boolean} [visible=true] Visibility.
   * @property {import("../extent.js").Extent} [extent] The bounding extent for layer rendering.  The layer will not be
   * rendered outside of this extent.
   * @property {number} [zIndex] The z-index for layer rendering.  At rendering time, the layers
   * will be ordered, first by Z-index and then by position. When `undefined`, a `zIndex` of 0 is assumed
   * for layers that are added to the map's `layers` collection, or `Infinity` when the layer's `setMap()`
   * method was used.
   * @property {number} [minResolution] The minimum resolution (inclusive) at which this layer will be
   * visible.
   * @property {number} [maxResolution] The maximum resolution (exclusive) below which this layer will
   * be visible.
   * @property {number} [minZoom] The minimum view zoom level (exclusive) above which this layer will be
   * visible.
   * @property {number} [maxZoom] The maximum view zoom level (inclusive) at which this layer will
   * be visible.
   * @property {import("../render.js").OrderFunction} [renderOrder] Render order. Function to be used when sorting
   * features before rendering. By default features are drawn in the order that they are created. Use
   * `null` to avoid the sort, but get an undefined draw order.
   * @property {number} [renderBuffer=100] The buffer in pixels around the viewport extent used by the
   * renderer when getting features from the vector source for the rendering or hit-detection.
   * Recommended value: the size of the largest symbol, line width or label.
   * @property {VectorSourceType} [source] Source.
   * @property {import("../Map.js").default} [map] Sets the layer as overlay on a map. The map will not manage
   * this layer in its layers collection, and the layer will be rendered on top. This is useful for
   * temporary layers. The standard way to add a layer to a map and have it managed by the map is to
   * use [map.addLayer()]{@link import("../Map.js").default#addLayer}.
   * @property {boolean} [declutter=false] Declutter images and text. Decluttering is applied to all
   * image and text styles of all Vector and VectorTile layers that have set this to `true`. The priority
   * is defined by the z-index of the layer, the `zIndex` of the style and the render order of features.
   * Higher z-index means higher priority. Within the same z-index, a feature rendered before another has
   * higher priority.
   *
   * As an optimization decluttered features from layers with the same `className` are rendered above
   * the fill and stroke styles of all of those layers regardless of z-index.  To opt out of this
   * behavior and place declutterd features with their own layer configure the layer with a `className`
   * other than `ol-layer`.
   * @property {import("../style/Style.js").StyleLike|import("../style/flat.js").FlatStyleLike|null} [style] Layer style. When set to `null`, only
   * features that have their own style will be rendered. See {@link module:ol/style/Style~Style} for the default style
   * which will be used if this is not set.
   * @property {import("./Base.js").BackgroundColor} [background] Background color for the layer. If not specified, no background
   * will be rendered.
   * @property {boolean} [updateWhileAnimating=false] When set to `true`, feature batches will
   * be recreated during animations. This means that no vectors will be shown clipped, but the
   * setting will have a performance impact for large amounts of vector data. When set to `false`,
   * batches will be recreated when no animation is active.
   * @property {boolean} [updateWhileInteracting=false] When set to `true`, feature batches will
   * be recreated during interactions. See also `updateWhileAnimating`.
   * @property {Object<string, *>} [properties] Arbitrary observable properties. Can be accessed with `#get()` and `#set()`.
   */

  /**
   * @enum {string}
   * @private
   */
  const Property$1 = {
    RENDER_ORDER: 'renderOrder',
  };

  /**
   * @classdesc
   * Vector data that is rendered client-side.
   * Note that any property set in the options is set as a {@link module:ol/Object~BaseObject}
   * property on the layer object; for example, setting `title: 'My Title'` in the
   * options means that `title` is observable, and has get/set accessors.
   *
   * @template {import("../source/Vector.js").default|import("../source/VectorTile.js").default} VectorSourceType
   * @template {import("../renderer/canvas/VectorLayer.js").default|import("../renderer/canvas/VectorTileLayer.js").default|import("../renderer/canvas/VectorImageLayer.js").default|import("../renderer/webgl/PointsLayer.js").default} RendererType
   * @extends {Layer<VectorSourceType, RendererType>}
   * @api
   */
  class BaseVectorLayer extends Layer {
    /**
     * @param {Options<VectorSourceType>} [options] Options.
     */
    constructor(options) {
      options = options ? options : {};

      const baseOptions = Object.assign({}, options);

      delete baseOptions.style;
      delete baseOptions.renderBuffer;
      delete baseOptions.updateWhileAnimating;
      delete baseOptions.updateWhileInteracting;
      super(baseOptions);

      /**
       * @private
       * @type {boolean}
       */
      this.declutter_ =
        options.declutter !== undefined ? options.declutter : false;

      /**
       * @type {number}
       * @private
       */
      this.renderBuffer_ =
        options.renderBuffer !== undefined ? options.renderBuffer : 100;

      /**
       * User provided style.
       * @type {import("../style/Style.js").StyleLike}
       * @private
       */
      this.style_ = null;

      /**
       * Style function for use within the library.
       * @type {import("../style/Style.js").StyleFunction|undefined}
       * @private
       */
      this.styleFunction_ = undefined;

      this.setStyle(options.style);

      /**
       * @type {boolean}
       * @private
       */
      this.updateWhileAnimating_ =
        options.updateWhileAnimating !== undefined
          ? options.updateWhileAnimating
          : false;

      /**
       * @type {boolean}
       * @private
       */
      this.updateWhileInteracting_ =
        options.updateWhileInteracting !== undefined
          ? options.updateWhileInteracting
          : false;
    }

    /**
     * @return {boolean} Declutter.
     */
    getDeclutter() {
      return this.declutter_;
    }

    /**
     * Get the topmost feature that intersects the given pixel on the viewport. Returns a promise
     * that resolves with an array of features. The array will either contain the topmost feature
     * when a hit was detected, or it will be empty.
     *
     * The hit detection algorithm used for this method is optimized for performance, but is less
     * accurate than the one used in [map.getFeaturesAtPixel()]{@link import("../Map.js").default#getFeaturesAtPixel}.
     * Text is not considered, and icons are only represented by their bounding box instead of the exact
     * image.
     *
     * @param {import("../pixel.js").Pixel} pixel Pixel.
     * @return {Promise<Array<import("../Feature").default>>} Promise that resolves with an array of features.
     * @api
     */
    getFeatures(pixel) {
      return super.getFeatures(pixel);
    }

    /**
     * @return {number|undefined} Render buffer.
     */
    getRenderBuffer() {
      return this.renderBuffer_;
    }

    /**
     * @return {function(import("../Feature.js").default, import("../Feature.js").default): number|null|undefined} Render
     *     order.
     */
    getRenderOrder() {
      return /** @type {import("../render.js").OrderFunction|null|undefined} */ (
        this.get(Property$1.RENDER_ORDER)
      );
    }

    /**
     * Get the style for features.  This returns whatever was passed to the `style`
     * option at construction or to the `setStyle` method.
     * @return {import("../style/Style.js").StyleLike|null|undefined} Layer style.
     * @api
     */
    getStyle() {
      return this.style_;
    }

    /**
     * Get the style function.
     * @return {import("../style/Style.js").StyleFunction|undefined} Layer style function.
     * @api
     */
    getStyleFunction() {
      return this.styleFunction_;
    }

    /**
     * @return {boolean} Whether the rendered layer should be updated while
     *     animating.
     */
    getUpdateWhileAnimating() {
      return this.updateWhileAnimating_;
    }

    /**
     * @return {boolean} Whether the rendered layer should be updated while
     *     interacting.
     */
    getUpdateWhileInteracting() {
      return this.updateWhileInteracting_;
    }

    /**
     * Render declutter items for this layer
     * @param {import("../Map.js").FrameState} frameState Frame state.
     */
    renderDeclutter(frameState) {
      if (!frameState.declutterTree) {
        frameState.declutterTree = new RBush$1(9);
      }
      /** @type {*} */ (this.getRenderer()).renderDeclutter(frameState);
    }

    /**
     * @param {import("../render.js").OrderFunction|null|undefined} renderOrder
     *     Render order.
     */
    setRenderOrder(renderOrder) {
      this.set(Property$1.RENDER_ORDER, renderOrder);
    }

    /**
     * Set the style for features.  This can be a single style object, an array
     * of styles, or a function that takes a feature and resolution and returns
     * an array of styles. If set to `null`, the layer has no style (a `null` style),
     * so only features that have their own styles will be rendered in the layer. Call
     * `setStyle()` without arguments to reset to the default style. See
     * [the ol/style/Style module]{@link module:ol/style/Style~Style} for information on the default style.
     *
     * If your layer has a static style, you can use "flat" style object literals instead of
     * using the `Style` and symbolizer constructors (`Fill`, `Stroke`, etc.).  See the documentation
     * for the [flat style types]{@link module:ol/style/flat~FlatStyle} to see what properties are supported.
     *
     * @param {import("../style/Style.js").StyleLike|import("../style/flat.js").FlatStyleLike|null} [style] Layer style.
     * @api
     */
    setStyle(style) {
      /**
       * @type {import("../style/Style.js").StyleLike|null}
       */
      let styleLike;

      if (style === undefined) {
        styleLike = createDefaultStyle;
      } else if (style === null) {
        styleLike = null;
      } else if (typeof style === 'function') {
        styleLike = style;
      } else if (style instanceof Style) {
        styleLike = style;
      } else if (Array.isArray(style)) {
        const len = style.length;

        /**
         * @type {Array<Style>}
         */
        const styles = new Array(len);

        for (let i = 0; i < len; ++i) {
          const s = style[i];
          if (s instanceof Style) {
            styles[i] = s;
          } else {
            styles[i] = toStyle(s);
          }
        }
        styleLike = styles;
      } else {
        styleLike = toStyle(style);
      }

      this.style_ = styleLike;
      this.styleFunction_ =
        style === null ? undefined : toFunction(this.style_);
      this.changed();
    }
  }

  /**
   * @module ol/render/webgl/constants
   */

  /**
   * @enum {string}
   */
  const WebGLWorkerMessageType = {
    GENERATE_POLYGON_BUFFERS: 'GENERATE_POLYGON_BUFFERS',
    GENERATE_POINT_BUFFERS: 'GENERATE_POINT_BUFFERS',
    GENERATE_LINE_STRING_BUFFERS: 'GENERATE_LINE_STRING_BUFFERS',
  };

  /**
   * @typedef {Object} WebGLWorkerGenerateBuffersMessage
   * This message will trigger the generation of a vertex and an index buffer based on the given render instructions.
   * When the buffers are generated, the worked will send a message of the same type to the main thread, with
   * the generated buffers in it.
   * Note that any addition properties present in the message *will* be sent back to the main thread.
   * @property {number} id Message id; will be used both in request and response as a means of identification
   * @property {WebGLWorkerMessageType} type Message type
   * @property {ArrayBuffer} renderInstructions Polygon render instructions raw binary buffer.
   * @property {number} [customAttributesCount] Amount of custom attributes count in the polygon render instructions.
   * @property {ArrayBuffer} [vertexBuffer] Vertices array raw binary buffer (sent by the worker).
   * @property {ArrayBuffer} [indexBuffer] Indices array raw binary buffer (sent by the worker).
   * @property {import("../../transform").Transform} [renderInstructionsTransform] Transformation matrix used to project the instructions coordinates
   */

  var earcut$1 = {exports: {}};

  earcut$1.exports = earcut;
  earcut$1.exports.default = earcut;

  function earcut(data, holeIndices, dim) {

      dim = dim || 2;

      var hasHoles = holeIndices && holeIndices.length,
          outerLen = hasHoles ? holeIndices[0] * dim : data.length,
          outerNode = linkedList(data, 0, outerLen, dim, true),
          triangles = [];

      if (!outerNode || outerNode.next === outerNode.prev) return triangles;

      var minX, minY, maxX, maxY, x, y, invSize;

      if (hasHoles) outerNode = eliminateHoles(data, holeIndices, outerNode, dim);

      // if the shape is not too simple, we'll use z-order curve hash later; calculate polygon bbox
      if (data.length > 80 * dim) {
          minX = maxX = data[0];
          minY = maxY = data[1];

          for (var i = dim; i < outerLen; i += dim) {
              x = data[i];
              y = data[i + 1];
              if (x < minX) minX = x;
              if (y < minY) minY = y;
              if (x > maxX) maxX = x;
              if (y > maxY) maxY = y;
          }

          // minX, minY and invSize are later used to transform coords into integers for z-order calculation
          invSize = Math.max(maxX - minX, maxY - minY);
          invSize = invSize !== 0 ? 1 / invSize : 0;
      }

      earcutLinked(outerNode, triangles, dim, minX, minY, invSize);

      return triangles;
  }

  // create a circular doubly linked list from polygon points in the specified winding order
  function linkedList(data, start, end, dim, clockwise) {
      var i, last;

      if (clockwise === (signedArea(data, start, end, dim) > 0)) {
          for (i = start; i < end; i += dim) last = insertNode(i, data[i], data[i + 1], last);
      } else {
          for (i = end - dim; i >= start; i -= dim) last = insertNode(i, data[i], data[i + 1], last);
      }

      if (last && equals(last, last.next)) {
          removeNode(last);
          last = last.next;
      }

      return last;
  }

  // eliminate colinear or duplicate points
  function filterPoints(start, end) {
      if (!start) return start;
      if (!end) end = start;

      var p = start,
          again;
      do {
          again = false;

          if (!p.steiner && (equals(p, p.next) || area(p.prev, p, p.next) === 0)) {
              removeNode(p);
              p = end = p.prev;
              if (p === p.next) break;
              again = true;

          } else {
              p = p.next;
          }
      } while (again || p !== end);

      return end;
  }

  // main ear slicing loop which triangulates a polygon (given as a linked list)
  function earcutLinked(ear, triangles, dim, minX, minY, invSize, pass) {
      if (!ear) return;

      // interlink polygon nodes in z-order
      if (!pass && invSize) indexCurve(ear, minX, minY, invSize);

      var stop = ear,
          prev, next;

      // iterate through ears, slicing them one by one
      while (ear.prev !== ear.next) {
          prev = ear.prev;
          next = ear.next;

          if (invSize ? isEarHashed(ear, minX, minY, invSize) : isEar(ear)) {
              // cut off the triangle
              triangles.push(prev.i / dim);
              triangles.push(ear.i / dim);
              triangles.push(next.i / dim);

              removeNode(ear);

              // skipping the next vertex leads to less sliver triangles
              ear = next.next;
              stop = next.next;

              continue;
          }

          ear = next;

          // if we looped through the whole remaining polygon and can't find any more ears
          if (ear === stop) {
              // try filtering points and slicing again
              if (!pass) {
                  earcutLinked(filterPoints(ear), triangles, dim, minX, minY, invSize, 1);

              // if this didn't work, try curing all small self-intersections locally
              } else if (pass === 1) {
                  ear = cureLocalIntersections(filterPoints(ear), triangles, dim);
                  earcutLinked(ear, triangles, dim, minX, minY, invSize, 2);

              // as a last resort, try splitting the remaining polygon into two
              } else if (pass === 2) {
                  splitEarcut(ear, triangles, dim, minX, minY, invSize);
              }

              break;
          }
      }
  }

  // check whether a polygon node forms a valid ear with adjacent nodes
  function isEar(ear) {
      var a = ear.prev,
          b = ear,
          c = ear.next;

      if (area(a, b, c) >= 0) return false; // reflex, can't be an ear

      // now make sure we don't have other points inside the potential ear
      var p = ear.next.next;

      while (p !== ear.prev) {
          if (pointInTriangle(a.x, a.y, b.x, b.y, c.x, c.y, p.x, p.y) &&
              area(p.prev, p, p.next) >= 0) return false;
          p = p.next;
      }

      return true;
  }

  function isEarHashed(ear, minX, minY, invSize) {
      var a = ear.prev,
          b = ear,
          c = ear.next;

      if (area(a, b, c) >= 0) return false; // reflex, can't be an ear

      // triangle bbox; min & max are calculated like this for speed
      var minTX = a.x < b.x ? (a.x < c.x ? a.x : c.x) : (b.x < c.x ? b.x : c.x),
          minTY = a.y < b.y ? (a.y < c.y ? a.y : c.y) : (b.y < c.y ? b.y : c.y),
          maxTX = a.x > b.x ? (a.x > c.x ? a.x : c.x) : (b.x > c.x ? b.x : c.x),
          maxTY = a.y > b.y ? (a.y > c.y ? a.y : c.y) : (b.y > c.y ? b.y : c.y);

      // z-order range for the current triangle bbox;
      var minZ = zOrder(minTX, minTY, minX, minY, invSize),
          maxZ = zOrder(maxTX, maxTY, minX, minY, invSize);

      var p = ear.prevZ,
          n = ear.nextZ;

      // look for points inside the triangle in both directions
      while (p && p.z >= minZ && n && n.z <= maxZ) {
          if (p !== ear.prev && p !== ear.next &&
              pointInTriangle(a.x, a.y, b.x, b.y, c.x, c.y, p.x, p.y) &&
              area(p.prev, p, p.next) >= 0) return false;
          p = p.prevZ;

          if (n !== ear.prev && n !== ear.next &&
              pointInTriangle(a.x, a.y, b.x, b.y, c.x, c.y, n.x, n.y) &&
              area(n.prev, n, n.next) >= 0) return false;
          n = n.nextZ;
      }

      // look for remaining points in decreasing z-order
      while (p && p.z >= minZ) {
          if (p !== ear.prev && p !== ear.next &&
              pointInTriangle(a.x, a.y, b.x, b.y, c.x, c.y, p.x, p.y) &&
              area(p.prev, p, p.next) >= 0) return false;
          p = p.prevZ;
      }

      // look for remaining points in increasing z-order
      while (n && n.z <= maxZ) {
          if (n !== ear.prev && n !== ear.next &&
              pointInTriangle(a.x, a.y, b.x, b.y, c.x, c.y, n.x, n.y) &&
              area(n.prev, n, n.next) >= 0) return false;
          n = n.nextZ;
      }

      return true;
  }

  // go through all polygon nodes and cure small local self-intersections
  function cureLocalIntersections(start, triangles, dim) {
      var p = start;
      do {
          var a = p.prev,
              b = p.next.next;

          if (!equals(a, b) && intersects$1(a, p, p.next, b) && locallyInside(a, b) && locallyInside(b, a)) {

              triangles.push(a.i / dim);
              triangles.push(p.i / dim);
              triangles.push(b.i / dim);

              // remove two nodes involved
              removeNode(p);
              removeNode(p.next);

              p = start = b;
          }
          p = p.next;
      } while (p !== start);

      return filterPoints(p);
  }

  // try splitting polygon into two and triangulate them independently
  function splitEarcut(start, triangles, dim, minX, minY, invSize) {
      // look for a valid diagonal that divides the polygon into two
      var a = start;
      do {
          var b = a.next.next;
          while (b !== a.prev) {
              if (a.i !== b.i && isValidDiagonal(a, b)) {
                  // split the polygon in two by the diagonal
                  var c = splitPolygon(a, b);

                  // filter colinear points around the cuts
                  a = filterPoints(a, a.next);
                  c = filterPoints(c, c.next);

                  // run earcut on each half
                  earcutLinked(a, triangles, dim, minX, minY, invSize);
                  earcutLinked(c, triangles, dim, minX, minY, invSize);
                  return;
              }
              b = b.next;
          }
          a = a.next;
      } while (a !== start);
  }

  // link every hole into the outer loop, producing a single-ring polygon without holes
  function eliminateHoles(data, holeIndices, outerNode, dim) {
      var queue = [],
          i, len, start, end, list;

      for (i = 0, len = holeIndices.length; i < len; i++) {
          start = holeIndices[i] * dim;
          end = i < len - 1 ? holeIndices[i + 1] * dim : data.length;
          list = linkedList(data, start, end, dim, false);
          if (list === list.next) list.steiner = true;
          queue.push(getLeftmost(list));
      }

      queue.sort(compareX);

      // process holes from left to right
      for (i = 0; i < queue.length; i++) {
          outerNode = eliminateHole(queue[i], outerNode);
          outerNode = filterPoints(outerNode, outerNode.next);
      }

      return outerNode;
  }

  function compareX(a, b) {
      return a.x - b.x;
  }

  // find a bridge between vertices that connects hole with an outer ring and and link it
  function eliminateHole(hole, outerNode) {
      var bridge = findHoleBridge(hole, outerNode);
      if (!bridge) {
          return outerNode;
      }

      var bridgeReverse = splitPolygon(bridge, hole);

      // filter collinear points around the cuts
      var filteredBridge = filterPoints(bridge, bridge.next);
      filterPoints(bridgeReverse, bridgeReverse.next);

      // Check if input node was removed by the filtering
      return outerNode === bridge ? filteredBridge : outerNode;
  }

  // David Eberly's algorithm for finding a bridge between hole and outer polygon
  function findHoleBridge(hole, outerNode) {
      var p = outerNode,
          hx = hole.x,
          hy = hole.y,
          qx = -Infinity,
          m;

      // find a segment intersected by a ray from the hole's leftmost point to the left;
      // segment's endpoint with lesser x will be potential connection point
      do {
          if (hy <= p.y && hy >= p.next.y && p.next.y !== p.y) {
              var x = p.x + (hy - p.y) * (p.next.x - p.x) / (p.next.y - p.y);
              if (x <= hx && x > qx) {
                  qx = x;
                  if (x === hx) {
                      if (hy === p.y) return p;
                      if (hy === p.next.y) return p.next;
                  }
                  m = p.x < p.next.x ? p : p.next;
              }
          }
          p = p.next;
      } while (p !== outerNode);

      if (!m) return null;

      if (hx === qx) return m; // hole touches outer segment; pick leftmost endpoint

      // look for points inside the triangle of hole point, segment intersection and endpoint;
      // if there are no points found, we have a valid connection;
      // otherwise choose the point of the minimum angle with the ray as connection point

      var stop = m,
          mx = m.x,
          my = m.y,
          tanMin = Infinity,
          tan;

      p = m;

      do {
          if (hx >= p.x && p.x >= mx && hx !== p.x &&
                  pointInTriangle(hy < my ? hx : qx, hy, mx, my, hy < my ? qx : hx, hy, p.x, p.y)) {

              tan = Math.abs(hy - p.y) / (hx - p.x); // tangential

              if (locallyInside(p, hole) &&
                  (tan < tanMin || (tan === tanMin && (p.x > m.x || (p.x === m.x && sectorContainsSector(m, p)))))) {
                  m = p;
                  tanMin = tan;
              }
          }

          p = p.next;
      } while (p !== stop);

      return m;
  }

  // whether sector in vertex m contains sector in vertex p in the same coordinates
  function sectorContainsSector(m, p) {
      return area(m.prev, m, p.prev) < 0 && area(p.next, m, m.next) < 0;
  }

  // interlink polygon nodes in z-order
  function indexCurve(start, minX, minY, invSize) {
      var p = start;
      do {
          if (p.z === null) p.z = zOrder(p.x, p.y, minX, minY, invSize);
          p.prevZ = p.prev;
          p.nextZ = p.next;
          p = p.next;
      } while (p !== start);

      p.prevZ.nextZ = null;
      p.prevZ = null;

      sortLinked(p);
  }

  // Simon Tatham's linked list merge sort algorithm
  // http://www.chiark.greenend.org.uk/~sgtatham/algorithms/listsort.html
  function sortLinked(list) {
      var i, p, q, e, tail, numMerges, pSize, qSize,
          inSize = 1;

      do {
          p = list;
          list = null;
          tail = null;
          numMerges = 0;

          while (p) {
              numMerges++;
              q = p;
              pSize = 0;
              for (i = 0; i < inSize; i++) {
                  pSize++;
                  q = q.nextZ;
                  if (!q) break;
              }
              qSize = inSize;

              while (pSize > 0 || (qSize > 0 && q)) {

                  if (pSize !== 0 && (qSize === 0 || !q || p.z <= q.z)) {
                      e = p;
                      p = p.nextZ;
                      pSize--;
                  } else {
                      e = q;
                      q = q.nextZ;
                      qSize--;
                  }

                  if (tail) tail.nextZ = e;
                  else list = e;

                  e.prevZ = tail;
                  tail = e;
              }

              p = q;
          }

          tail.nextZ = null;
          inSize *= 2;

      } while (numMerges > 1);

      return list;
  }

  // z-order of a point given coords and inverse of the longer side of data bbox
  function zOrder(x, y, minX, minY, invSize) {
      // coords are transformed into non-negative 15-bit integer range
      x = 32767 * (x - minX) * invSize;
      y = 32767 * (y - minY) * invSize;

      x = (x | (x << 8)) & 0x00FF00FF;
      x = (x | (x << 4)) & 0x0F0F0F0F;
      x = (x | (x << 2)) & 0x33333333;
      x = (x | (x << 1)) & 0x55555555;

      y = (y | (y << 8)) & 0x00FF00FF;
      y = (y | (y << 4)) & 0x0F0F0F0F;
      y = (y | (y << 2)) & 0x33333333;
      y = (y | (y << 1)) & 0x55555555;

      return x | (y << 1);
  }

  // find the leftmost node of a polygon ring
  function getLeftmost(start) {
      var p = start,
          leftmost = start;
      do {
          if (p.x < leftmost.x || (p.x === leftmost.x && p.y < leftmost.y)) leftmost = p;
          p = p.next;
      } while (p !== start);

      return leftmost;
  }

  // check if a point lies within a convex triangle
  function pointInTriangle(ax, ay, bx, by, cx, cy, px, py) {
      return (cx - px) * (ay - py) - (ax - px) * (cy - py) >= 0 &&
             (ax - px) * (by - py) - (bx - px) * (ay - py) >= 0 &&
             (bx - px) * (cy - py) - (cx - px) * (by - py) >= 0;
  }

  // check if a diagonal between two polygon nodes is valid (lies in polygon interior)
  function isValidDiagonal(a, b) {
      return a.next.i !== b.i && a.prev.i !== b.i && !intersectsPolygon(a, b) && // dones't intersect other edges
             (locallyInside(a, b) && locallyInside(b, a) && middleInside(a, b) && // locally visible
              (area(a.prev, a, b.prev) || area(a, b.prev, b)) || // does not create opposite-facing sectors
              equals(a, b) && area(a.prev, a, a.next) > 0 && area(b.prev, b, b.next) > 0); // special zero-length case
  }

  // signed area of a triangle
  function area(p, q, r) {
      return (q.y - p.y) * (r.x - q.x) - (q.x - p.x) * (r.y - q.y);
  }

  // check if two points are equal
  function equals(p1, p2) {
      return p1.x === p2.x && p1.y === p2.y;
  }

  // check if two segments intersect
  function intersects$1(p1, q1, p2, q2) {
      var o1 = sign(area(p1, q1, p2));
      var o2 = sign(area(p1, q1, q2));
      var o3 = sign(area(p2, q2, p1));
      var o4 = sign(area(p2, q2, q1));

      if (o1 !== o2 && o3 !== o4) return true; // general case

      if (o1 === 0 && onSegment(p1, p2, q1)) return true; // p1, q1 and p2 are collinear and p2 lies on p1q1
      if (o2 === 0 && onSegment(p1, q2, q1)) return true; // p1, q1 and q2 are collinear and q2 lies on p1q1
      if (o3 === 0 && onSegment(p2, p1, q2)) return true; // p2, q2 and p1 are collinear and p1 lies on p2q2
      if (o4 === 0 && onSegment(p2, q1, q2)) return true; // p2, q2 and q1 are collinear and q1 lies on p2q2

      return false;
  }

  // for collinear points p, q, r, check if point q lies on segment pr
  function onSegment(p, q, r) {
      return q.x <= Math.max(p.x, r.x) && q.x >= Math.min(p.x, r.x) && q.y <= Math.max(p.y, r.y) && q.y >= Math.min(p.y, r.y);
  }

  function sign(num) {
      return num > 0 ? 1 : num < 0 ? -1 : 0;
  }

  // check if a polygon diagonal intersects any polygon segments
  function intersectsPolygon(a, b) {
      var p = a;
      do {
          if (p.i !== a.i && p.next.i !== a.i && p.i !== b.i && p.next.i !== b.i &&
                  intersects$1(p, p.next, a, b)) return true;
          p = p.next;
      } while (p !== a);

      return false;
  }

  // check if a polygon diagonal is locally inside the polygon
  function locallyInside(a, b) {
      return area(a.prev, a, a.next) < 0 ?
          area(a, b, a.next) >= 0 && area(a, a.prev, b) >= 0 :
          area(a, b, a.prev) < 0 || area(a, a.next, b) < 0;
  }

  // check if the middle point of a polygon diagonal is inside the polygon
  function middleInside(a, b) {
      var p = a,
          inside = false,
          px = (a.x + b.x) / 2,
          py = (a.y + b.y) / 2;
      do {
          if (((p.y > py) !== (p.next.y > py)) && p.next.y !== p.y &&
                  (px < (p.next.x - p.x) * (py - p.y) / (p.next.y - p.y) + p.x))
              inside = !inside;
          p = p.next;
      } while (p !== a);

      return inside;
  }

  // link two polygon vertices with a bridge; if the vertices belong to the same ring, it splits polygon into two;
  // if one belongs to the outer ring and another to a hole, it merges it into a single ring
  function splitPolygon(a, b) {
      var a2 = new Node$1(a.i, a.x, a.y),
          b2 = new Node$1(b.i, b.x, b.y),
          an = a.next,
          bp = b.prev;

      a.next = b;
      b.prev = a;

      a2.next = an;
      an.prev = a2;

      b2.next = a2;
      a2.prev = b2;

      bp.next = b2;
      b2.prev = bp;

      return b2;
  }

  // create a node and optionally link it with previous one (in a circular doubly linked list)
  function insertNode(i, x, y, last) {
      var p = new Node$1(i, x, y);

      if (!last) {
          p.prev = p;
          p.next = p;

      } else {
          p.next = last.next;
          p.prev = last;
          last.next.prev = p;
          last.next = p;
      }
      return p;
  }

  function removeNode(p) {
      p.next.prev = p.prev;
      p.prev.next = p.next;

      if (p.prevZ) p.prevZ.nextZ = p.nextZ;
      if (p.nextZ) p.nextZ.prevZ = p.prevZ;
  }

  function Node$1(i, x, y) {
      // vertex index in coordinates array
      this.i = i;

      // vertex coordinates
      this.x = x;
      this.y = y;

      // previous and next vertex nodes in a polygon ring
      this.prev = null;
      this.next = null;

      // z-order curve value
      this.z = null;

      // previous and next nodes in z-order
      this.prevZ = null;
      this.nextZ = null;

      // indicates whether this is a steiner point
      this.steiner = false;
  }

  // return a percentage difference between the polygon area and its triangulation area;
  // used to verify correctness of triangulation
  earcut.deviation = function (data, holeIndices, dim, triangles) {
      var hasHoles = holeIndices && holeIndices.length;
      var outerLen = hasHoles ? holeIndices[0] * dim : data.length;

      var polygonArea = Math.abs(signedArea(data, 0, outerLen, dim));
      if (hasHoles) {
          for (var i = 0, len = holeIndices.length; i < len; i++) {
              var start = holeIndices[i] * dim;
              var end = i < len - 1 ? holeIndices[i + 1] * dim : data.length;
              polygonArea -= Math.abs(signedArea(data, start, end, dim));
          }
      }

      var trianglesArea = 0;
      for (i = 0; i < triangles.length; i += 3) {
          var a = triangles[i] * dim;
          var b = triangles[i + 1] * dim;
          var c = triangles[i + 2] * dim;
          trianglesArea += Math.abs(
              (data[a] - data[c]) * (data[b + 1] - data[a + 1]) -
              (data[a] - data[b]) * (data[c + 1] - data[a + 1]));
      }

      return polygonArea === 0 && trianglesArea === 0 ? 0 :
          Math.abs((trianglesArea - polygonArea) / polygonArea);
  };

  function signedArea(data, start, end, dim) {
      var sum = 0;
      for (var i = start, j = end - dim; i < end; i += dim) {
          sum += (data[j] - data[i]) * (data[i + 1] + data[j + 1]);
          j = i;
      }
      return sum;
  }

  // turn a polygon in a multi-dimensional array form (e.g. as in GeoJSON) into a form Earcut accepts
  earcut.flatten = function (data) {
      var dim = data[0][0].length,
          result = {vertices: [], holes: [], dimensions: dim},
          holeIndex = 0;

      for (var i = 0; i < data.length; i++) {
          for (var j = 0; j < data[i].length; j++) {
              for (var d = 0; d < dim; d++) result.vertices.push(data[i][j][d]);
          }
          if (i > 0) {
              holeIndex += data[i - 1].length;
              result.holes.push(holeIndex);
          }
      }
      return result;
  };

  /**
   * @module ol/render/webgl/utils
   */

  const tmpArray_ = [];

  /**
   * An object holding positions both in an index and a vertex buffer.
   * @typedef {Object} BufferPositions
   * @property {number} vertexPosition Position in the vertex buffer
   * @property {number} indexPosition Position in the index buffer
   */
  const bufferPositions_ = {vertexPosition: 0, indexPosition: 0};

  function writePointVertex(buffer, pos, x, y, index) {
    buffer[pos + 0] = x;
    buffer[pos + 1] = y;
    buffer[pos + 2] = index;
  }

  /**
   * Pushes a quad (two triangles) based on a point geometry
   * @param {Float32Array} instructions Array of render instructions for points.
   * @param {number} elementIndex Index from which render instructions will be read.
   * @param {Float32Array} vertexBuffer Buffer in the form of a typed array.
   * @param {Uint32Array} indexBuffer Buffer in the form of a typed array.
   * @param {number} customAttributesCount Amount of custom attributes for each element.
   * @param {BufferPositions} [bufferPositions] Buffer write positions; if not specified, positions will be set at 0.
   * @return {BufferPositions} New buffer positions where to write next
   * @property {number} vertexPosition New position in the vertex buffer where future writes should start.
   * @property {number} indexPosition New position in the index buffer where future writes should start.
   * @private
   */
  function writePointFeatureToBuffers(
    instructions,
    elementIndex,
    vertexBuffer,
    indexBuffer,
    customAttributesCount,
    bufferPositions
  ) {
    // This is for x, y and index
    const baseVertexAttrsCount = 3;
    const baseInstructionsCount = 2;
    const stride = baseVertexAttrsCount + customAttributesCount;

    const x = instructions[elementIndex + 0];
    const y = instructions[elementIndex + 1];

    // read custom numerical attributes on the feature
    const customAttrs = tmpArray_;
    customAttrs.length = customAttributesCount;
    for (let i = 0; i < customAttrs.length; i++) {
      customAttrs[i] = instructions[elementIndex + baseInstructionsCount + i];
    }

    let vPos = bufferPositions ? bufferPositions.vertexPosition : 0;
    let iPos = bufferPositions ? bufferPositions.indexPosition : 0;
    const baseIndex = vPos / stride;

    // push vertices for each of the four quad corners (first standard then custom attributes)
    writePointVertex(vertexBuffer, vPos, x, y, 0);
    customAttrs.length &&
      vertexBuffer.set(customAttrs, vPos + baseVertexAttrsCount);
    vPos += stride;

    writePointVertex(vertexBuffer, vPos, x, y, 1);
    customAttrs.length &&
      vertexBuffer.set(customAttrs, vPos + baseVertexAttrsCount);
    vPos += stride;

    writePointVertex(vertexBuffer, vPos, x, y, 2);
    customAttrs.length &&
      vertexBuffer.set(customAttrs, vPos + baseVertexAttrsCount);
    vPos += stride;

    writePointVertex(vertexBuffer, vPos, x, y, 3);
    customAttrs.length &&
      vertexBuffer.set(customAttrs, vPos + baseVertexAttrsCount);
    vPos += stride;

    indexBuffer[iPos++] = baseIndex;
    indexBuffer[iPos++] = baseIndex + 1;
    indexBuffer[iPos++] = baseIndex + 3;
    indexBuffer[iPos++] = baseIndex + 1;
    indexBuffer[iPos++] = baseIndex + 2;
    indexBuffer[iPos++] = baseIndex + 3;

    bufferPositions_.vertexPosition = vPos;
    bufferPositions_.indexPosition = iPos;

    return bufferPositions_;
  }

  /**
   * Pushes a single quad to form a line segment; also includes a computation for the join angles with previous and next
   * segment, in order to be able to offset the vertices correctly in the shader
   * @param {Float32Array} instructions Array of render instructions for lines.
   * @param {number} segmentStartIndex Index of the segment start point from which render instructions will be read.
   * @param {number} segmentEndIndex Index of the segment start point from which render instructions will be read.
   * @param {number|null} beforeSegmentIndex Index of the point right before the segment (null if none, e.g this is a line start)
   * @param {number|null} afterSegmentIndex Index of the point right after the segment (null if none, e.g this is a line end)
   * @param {Array<number>} vertexArray Array containing vertices.
   * @param {Array<number>} indexArray Array containing indices.
   * @param {Array<number>} customAttributes Array of custom attributes value
   * @param {import('../../transform.js').Transform} instructionsTransform Transform matrix used to project coordinates in instructions
   * @param {import('../../transform.js').Transform} invertInstructionsTransform Transform matrix used to project coordinates in instructions
   * @private
   */
  function writeLineSegmentToBuffers(
    instructions,
    segmentStartIndex,
    segmentEndIndex,
    beforeSegmentIndex,
    afterSegmentIndex,
    vertexArray,
    indexArray,
    customAttributes,
    instructionsTransform,
    invertInstructionsTransform
  ) {
    // compute the stride to determine how many vertices were already pushed
    const baseVertexAttrsCount = 5; // base attributes: x0, y0, x1, y1, params (vertex number [0-3], join angle 1, join angle 2)
    const stride = baseVertexAttrsCount + customAttributes.length;
    const baseIndex = vertexArray.length / stride;

    // The segment is composed of two positions called P0[x0, y0] and P1[x1, y1]
    // Depending on whether there are points before and after the segment, its final shape
    // will be different
    const p0 = [
      instructions[segmentStartIndex + 0],
      instructions[segmentStartIndex + 1],
    ];
    const p1 = [instructions[segmentEndIndex], instructions[segmentEndIndex + 1]];

    // to compute offsets from the line center we need to reproject
    // coordinates back in world units and compute the length of the segment
    const p0world = apply(invertInstructionsTransform, [...p0]);
    const p1world = apply(invertInstructionsTransform, [...p1]);

    function computeVertexParameters(vertexNumber, joinAngle1, joinAngle2) {
      const shift = 10000;
      const anglePrecision = 1500;
      return (
        Math.round(joinAngle1 * anglePrecision) +
        Math.round(joinAngle2 * anglePrecision) * shift +
        vertexNumber * shift * shift
      );
    }

    // compute the angle between p0pA and p0pB
    // returns a value in [0, 2PI]
    function angleBetween(p0, pA, pB) {
      const lenA = Math.sqrt(
        (pA[0] - p0[0]) * (pA[0] - p0[0]) + (pA[1] - p0[1]) * (pA[1] - p0[1])
      );
      const tangentA = [(pA[0] - p0[0]) / lenA, (pA[1] - p0[1]) / lenA];
      const orthoA = [-tangentA[1], tangentA[0]];
      const lenB = Math.sqrt(
        (pB[0] - p0[0]) * (pB[0] - p0[0]) + (pB[1] - p0[1]) * (pB[1] - p0[1])
      );
      const tangentB = [(pB[0] - p0[0]) / lenB, (pB[1] - p0[1]) / lenB];

      // this angle can be clockwise or anticlockwise; hence the computation afterwards
      const angle =
        lenA === 0 || lenB === 0
          ? 0
          : Math.acos(
              clamp(tangentB[0] * tangentA[0] + tangentB[1] * tangentA[1], -1, 1)
            );
      const isClockwise = tangentB[0] * orthoA[0] + tangentB[1] * orthoA[1] > 0;
      return !isClockwise ? Math.PI * 2 - angle : angle;
    }

    const joinBefore = beforeSegmentIndex !== null;
    const joinAfter = afterSegmentIndex !== null;

    let angle0 = 0;
    let angle1 = 0;

    // add vertices and adapt offsets for P0 in case of join
    if (joinBefore) {
      // B for before
      const pB = [
        instructions[beforeSegmentIndex],
        instructions[beforeSegmentIndex + 1],
      ];
      const pBworld = apply(invertInstructionsTransform, [...pB]);
      angle0 = angleBetween(p0world, p1world, pBworld);
    }
    // adapt offsets for P1 in case of join
    if (joinAfter) {
      // A for after
      const pA = [
        instructions[afterSegmentIndex],
        instructions[afterSegmentIndex + 1],
      ];
      const pAworld = apply(invertInstructionsTransform, [...pA]);
      angle1 = angleBetween(p1world, p0world, pAworld);
    }

    // add main segment triangles
    vertexArray.push(
      p0[0],
      p0[1],
      p1[0],
      p1[1],
      computeVertexParameters(0, angle0, angle1)
    );
    vertexArray.push(...customAttributes);

    vertexArray.push(
      p0[0],
      p0[1],
      p1[0],
      p1[1],
      computeVertexParameters(1, angle0, angle1)
    );
    vertexArray.push(...customAttributes);

    vertexArray.push(
      p0[0],
      p0[1],
      p1[0],
      p1[1],
      computeVertexParameters(2, angle0, angle1)
    );
    vertexArray.push(...customAttributes);

    vertexArray.push(
      p0[0],
      p0[1],
      p1[0],
      p1[1],
      computeVertexParameters(3, angle0, angle1)
    );
    vertexArray.push(...customAttributes);

    indexArray.push(
      baseIndex,
      baseIndex + 1,
      baseIndex + 2,
      baseIndex + 1,
      baseIndex + 3,
      baseIndex + 2
    );
  }

  /**
   * Pushes several triangles to form a polygon, including holes
   * @param {Float32Array} instructions Array of render instructions for lines.
   * @param {number} polygonStartIndex Index of the polygon start point from which render instructions will be read.
   * @param {Array<number>} vertexArray Array containing vertices.
   * @param {Array<number>} indexArray Array containing indices.
   * @param {number} customAttributesCount Amount of custom attributes for each element.
   * @return {number} Next polygon instructions index
   * @private
   */
  function writePolygonTrianglesToBuffers(
    instructions,
    polygonStartIndex,
    vertexArray,
    indexArray,
    customAttributesCount
  ) {
    const instructionsPerVertex = 2; // x, y
    const attributesPerVertex = 2 + customAttributesCount;
    let instructionsIndex = polygonStartIndex;
    const customAttributes = instructions.slice(
      instructionsIndex,
      instructionsIndex + customAttributesCount
    );
    instructionsIndex += customAttributesCount;
    const ringsCount = instructions[instructionsIndex++];
    let verticesCount = 0;
    const holes = new Array(ringsCount - 1);
    for (let i = 0; i < ringsCount; i++) {
      verticesCount += instructions[instructionsIndex++];
      if (i < ringsCount - 1) {
        holes[i] = verticesCount;
      }
    }
    const flatCoords = instructions.slice(
      instructionsIndex,
      instructionsIndex + verticesCount * instructionsPerVertex
    );

    // pushing to vertices and indices!! this is where the magic happens
    const result = earcut$1.exports(flatCoords, holes, instructionsPerVertex);
    for (let i = 0; i < result.length; i++) {
      indexArray.push(result[i] + vertexArray.length / attributesPerVertex);
    }
    for (let i = 0; i < flatCoords.length; i += 2) {
      vertexArray.push(flatCoords[i], flatCoords[i + 1], ...customAttributes);
    }

    return instructionsIndex + verticesCount * instructionsPerVertex;
  }

  /**
   * Returns a texture of 1x1 pixel, white
   * @private
   * @return {ImageData} Image data.
   */
  function getBlankImageData() {
    const canvas = document.createElement('canvas');
    const image = canvas.getContext('2d').createImageData(1, 1);
    image.data[0] = 255;
    image.data[1] = 255;
    image.data[2] = 255;
    image.data[3] = 255;
    return image;
  }

  /**
   * Generates a color array based on a numerical id
   * Note: the range for each component is 0 to 1 with 256 steps
   * @param {number} id Id
   * @param {Array<number>} [array] Reusable array
   * @return {Array<number>} Color array containing the encoded id
   */
  function colorEncodeId(id, array) {
    array = array || [];
    const radix = 256;
    const divide = radix - 1;
    array[0] = Math.floor(id / radix / radix / radix) / divide;
    array[1] = (Math.floor(id / radix / radix) % radix) / divide;
    array[2] = (Math.floor(id / radix) % radix) / divide;
    array[3] = (id % radix) / divide;
    return array;
  }

  /**
   * Reads an id from a color-encoded array
   * Note: the expected range for each component is 0 to 1 with 256 steps.
   * @param {Array<number>} color Color array containing the encoded id
   * @return {number} Decoded id
   */
  function colorDecodeId(color) {
    let id = 0;
    const radix = 256;
    const mult = radix - 1;
    id += Math.round(color[0] * radix * radix * radix * mult);
    id += Math.round(color[1] * radix * radix * mult);
    id += Math.round(color[2] * radix * mult);
    id += Math.round(color[3] * mult);
    return id;
  }

  function create() {
            const source = "const e=\"GENERATE_POLYGON_BUFFERS\",t=\"GENERATE_POINT_BUFFERS\",n=\"GENERATE_LINE_STRING_BUFFERS\",r=\"undefined\"!=typeof navigator&&void 0!==navigator.userAgent?navigator.userAgent.toLowerCase():\"\";r.includes(\"firefox\");r.includes(\"safari\")&&!r.includes(\"chrom\")&&(r.includes(\"version/15.4\")||/cpu (os|iphone os) 15_4 like mac os x/.test(r)),r.includes(\"webkit\")&&r.includes(\"edge\"),r.includes(\"macintosh\"),\"undefined\"!=typeof WorkerGlobalScope&&\"undefined\"!=typeof OffscreenCanvas&&(self,WorkerGlobalScope),function(){let e=!1;try{const t=Object.defineProperty({},\"passive\",{get:function(){e=!0}});window.addEventListener(\"_\",null,t),window.removeEventListener(\"_\",null,t)}catch(e){}}();const o={1:\"The view center is not defined\",2:\"The view resolution is not defined\",3:\"The view rotation is not defined\",4:\"`image` and `src` cannot be provided at the same time\",5:\"`imgSize` must be set when `image` is provided\",7:\"`format` must be set when `url` is set\",8:\"Unknown `serverType` configured\",9:\"`url` must be configured or set using `#setUrl()`\",10:\"The default `geometryFunction` can only handle `Point` geometries\",11:\"`options.featureTypes` must be an Array\",12:\"`options.geometryName` must also be provided when `options.bbox` is set\",13:\"Invalid corner\",14:\"Invalid color\",15:\"Tried to get a value for a key that does not exist in the cache\",16:\"Tried to set a value for a key that is used already\",17:\"`resolutions` must be sorted in descending order\",18:\"Either `origin` or `origins` must be configured, never both\",19:\"Number of `tileSizes` and `resolutions` must be equal\",20:\"Number of `origins` and `resolutions` must be equal\",22:\"Either `tileSize` or `tileSizes` must be configured, never both\",24:\"Invalid extent or geometry provided as `geometry`\",25:\"Cannot fit empty extent provided as `geometry`\",26:\"Features must have an id set\",27:\"Features must have an id set\",28:'`renderMode` must be `\"hybrid\"` or `\"vector\"`',30:\"The passed `feature` was already added to the source\",31:\"Tried to enqueue an `element` that was already added to the queue\",32:\"Transformation matrix cannot be inverted\",33:\"Invalid units\",34:\"Invalid geometry layout\",36:\"Unknown SRS type\",37:\"Unknown geometry type found\",38:\"`styleMapValue` has an unknown type\",39:\"Unknown geometry type\",40:\"Expected `feature` to have a geometry\",41:\"Expected an `ol/style/Style` or an array of `ol/style/Style.js`\",42:\"Question unknown, the answer is 42\",43:\"Expected `layers` to be an array or a `Collection`\",47:\"Expected `controls` to be an array or an `ol/Collection`\",48:\"Expected `interactions` to be an array or an `ol/Collection`\",49:\"Expected `overlays` to be an array or an `ol/Collection`\",50:\"`options.featureTypes` should be an Array\",51:\"Either `url` or `tileJSON` options must be provided\",52:\"Unknown `serverType` configured\",53:\"Unknown `tierSizeCalculation` configured\",55:\"The {-y} placeholder requires a tile grid with extent\",56:\"mapBrowserEvent must originate from a pointer event\",57:\"At least 2 conditions are required\",59:\"Invalid command found in the PBF\",60:\"Missing or invalid `size`\",61:\"Cannot determine IIIF Image API version from provided image information JSON\",62:\"A `WebGLArrayBuffer` must either be of type `ELEMENT_ARRAY_BUFFER` or `ARRAY_BUFFER`\",64:\"Layer opacity must be a number\",66:\"`forEachFeatureAtCoordinate` cannot be used on a WebGL layer if the hit detection logic has not been enabled. This is done by providing adequate shaders using the `hitVertexShader` and `hitFragmentShader` properties of `WebGLPointsLayerRenderer`\",67:\"A layer can only be added to the map once. Use either `layer.setMap()` or `map.addLayer()`, not both\",68:\"A VectorTile source can only be rendered if it has a projection compatible with the view projection\"};class i extends Error{constructor(e){const t=o[e];super(t),this.code=e,this.name=\"AssertionError\",this.message=t}}function s(e,t){const n=t[0],r=t[1];return t[0]=e[0]*n+e[2]*r+e[4],t[1]=e[1]*n+e[3]*r+e[5],t}function a(e,t){const n=(r=t)[0]*r[3]-r[1]*r[2];var r;!function(e,t){if(!e)throw new i(t)}(0!==n,32);const o=t[0],s=t[1],a=t[2],u=t[3],x=t[4],f=t[5];return e[0]=u/n,e[1]=-s/n,e[2]=-a/n,e[3]=o/n,e[4]=(a*f-u*x)/n,e[5]=-(o*f-s*x)/n,e}new Array(6);var u={exports:{}};function x(e,t,n){n=n||2;var r,o,i,s,a,u,x,y=t&&t.length,d=y?t[0]*n:e.length,h=f(e,0,d,n,!0),v=[];if(!h||h.next===h.prev)return v;if(y&&(h=function(e,t,n,r){var o,i,s,a=[];for(o=0,i=t.length;o<i;o++)(s=f(e,t[o]*r,o<i-1?t[o+1]*r:e.length,r,!1))===s.next&&(s.steiner=!0),a.push(w(s));for(a.sort(p),o=0;o<a.length;o++)n=l(n=g(a[o],n),n.next);return n}(e,t,h,n)),e.length>80*n){r=i=e[0],o=s=e[1];for(var b=n;b<d;b+=n)(a=e[b])<r&&(r=a),(u=e[b+1])<o&&(o=u),a>i&&(i=a),u>s&&(s=u);x=0!==(x=Math.max(i-r,s-o))?1/x:0}return c(h,v,n,r,o,x),v}function f(e,t,n,r,o){var i,s;if(o===B(e,t,n,r)>0)for(i=t;i<n;i+=r)s=U(i,e[i],e[i+1],s);else for(i=n-r;i>=t;i-=r)s=U(i,e[i],e[i+1],s);return s&&M(s,s.next)&&(z(s),s=s.next),s}function l(e,t){if(!e)return e;t||(t=e);var n,r=e;do{if(n=!1,r.steiner||!M(r,r.next)&&0!==Z(r.prev,r,r.next))r=r.next;else{if(z(r),(r=t=r.prev)===r.next)break;n=!0}}while(n||r!==t);return t}function c(e,t,n,r,o,i,s){if(e){!s&&i&&function(e,t,n,r){var o=e;do{null===o.z&&(o.z=m(o.x,o.y,t,n,r)),o.prevZ=o.prev,o.nextZ=o.next,o=o.next}while(o!==e);o.prevZ.nextZ=null,o.prevZ=null,function(e){var t,n,r,o,i,s,a,u,x=1;do{for(n=e,e=null,i=null,s=0;n;){for(s++,r=n,a=0,t=0;t<x&&(a++,r=r.nextZ);t++);for(u=x;a>0||u>0&&r;)0!==a&&(0===u||!r||n.z<=r.z)?(o=n,n=n.nextZ,a--):(o=r,r=r.nextZ,u--),i?i.nextZ=o:e=o,o.prevZ=i,i=o;n=r}i.nextZ=null,x*=2}while(s>1)}(o)}(e,r,o,i);for(var a,u,x=e;e.prev!==e.next;)if(a=e.prev,u=e.next,i?d(e,r,o,i):y(e))t.push(a.i/n),t.push(e.i/n),t.push(u.i/n),z(e),e=u.next,x=u.next;else if((e=u)===x){s?1===s?c(e=h(l(e),t,n),t,n,r,o,i,2):2===s&&v(e,t,n,r,o,i):c(l(e),t,n,r,o,i,1);break}}}function y(e){var t=e.prev,n=e,r=e.next;if(Z(t,n,r)>=0)return!1;for(var o=e.next.next;o!==e.prev;){if(A(t.x,t.y,n.x,n.y,r.x,r.y,o.x,o.y)&&Z(o.prev,o,o.next)>=0)return!1;o=o.next}return!0}function d(e,t,n,r){var o=e.prev,i=e,s=e.next;if(Z(o,i,s)>=0)return!1;for(var a=o.x<i.x?o.x<s.x?o.x:s.x:i.x<s.x?i.x:s.x,u=o.y<i.y?o.y<s.y?o.y:s.y:i.y<s.y?i.y:s.y,x=o.x>i.x?o.x>s.x?o.x:s.x:i.x>s.x?i.x:s.x,f=o.y>i.y?o.y>s.y?o.y:s.y:i.y>s.y?i.y:s.y,l=m(a,u,t,n,r),c=m(x,f,t,n,r),y=e.prevZ,d=e.nextZ;y&&y.z>=l&&d&&d.z<=c;){if(y!==e.prev&&y!==e.next&&A(o.x,o.y,i.x,i.y,s.x,s.y,y.x,y.y)&&Z(y.prev,y,y.next)>=0)return!1;if(y=y.prevZ,d!==e.prev&&d!==e.next&&A(o.x,o.y,i.x,i.y,s.x,s.y,d.x,d.y)&&Z(d.prev,d,d.next)>=0)return!1;d=d.nextZ}for(;y&&y.z>=l;){if(y!==e.prev&&y!==e.next&&A(o.x,o.y,i.x,i.y,s.x,s.y,y.x,y.y)&&Z(y.prev,y,y.next)>=0)return!1;y=y.prevZ}for(;d&&d.z<=c;){if(d!==e.prev&&d!==e.next&&A(o.x,o.y,i.x,i.y,s.x,s.y,d.x,d.y)&&Z(d.prev,d,d.next)>=0)return!1;d=d.nextZ}return!0}function h(e,t,n){var r=e;do{var o=r.prev,i=r.next.next;!M(o,i)&&F(o,r,r.next,i)&&k(o,i)&&k(i,o)&&(t.push(o.i/n),t.push(r.i/n),t.push(i.i/n),z(r),z(r.next),r=e=i),r=r.next}while(r!==e);return l(r)}function v(e,t,n,r,o,i){var s=e;do{for(var a=s.next.next;a!==s.prev;){if(s.i!==a.i&&E(s,a)){var u=S(s,a);return s=l(s,s.next),u=l(u,u.next),c(s,t,n,r,o,i),void c(u,t,n,r,o,i)}a=a.next}s=s.next}while(s!==e)}function p(e,t){return e.x-t.x}function g(e,t){var n=function(e,t){var n,r=t,o=e.x,i=e.y,s=-1/0;do{if(i<=r.y&&i>=r.next.y&&r.next.y!==r.y){var a=r.x+(i-r.y)*(r.next.x-r.x)/(r.next.y-r.y);if(a<=o&&a>s){if(s=a,a===o){if(i===r.y)return r;if(i===r.next.y)return r.next}n=r.x<r.next.x?r:r.next}}r=r.next}while(r!==t);if(!n)return null;if(o===s)return n;var u,x=n,f=n.x,l=n.y,c=1/0;r=n;do{o>=r.x&&r.x>=f&&o!==r.x&&A(i<l?o:s,i,f,l,i<l?s:o,i,r.x,r.y)&&(u=Math.abs(i-r.y)/(o-r.x),k(r,e)&&(u<c||u===c&&(r.x>n.x||r.x===n.x&&b(n,r)))&&(n=r,c=u)),r=r.next}while(r!==x);return n}(e,t);if(!n)return t;var r=S(n,e),o=l(n,n.next);return l(r,r.next),t===n?o:t}function b(e,t){return Z(e.prev,e,t.prev)<0&&Z(t.next,e,e.next)<0}function m(e,t,n,r,o){return(e=1431655765&((e=858993459&((e=252645135&((e=16711935&((e=32767*(e-n)*o)|e<<8))|e<<4))|e<<2))|e<<1))|(t=1431655765&((t=858993459&((t=252645135&((t=16711935&((t=32767*(t-r)*o)|t<<8))|t<<4))|t<<2))|t<<1))<<1}function w(e){var t=e,n=e;do{(t.x<n.x||t.x===n.x&&t.y<n.y)&&(n=t),t=t.next}while(t!==e);return n}function A(e,t,n,r,o,i,s,a){return(o-s)*(t-a)-(e-s)*(i-a)>=0&&(e-s)*(r-a)-(n-s)*(t-a)>=0&&(n-s)*(i-a)-(o-s)*(r-a)>=0}function E(e,t){return e.next.i!==t.i&&e.prev.i!==t.i&&!function(e,t){var n=e;do{if(n.i!==e.i&&n.next.i!==e.i&&n.i!==t.i&&n.next.i!==t.i&&F(n,n.next,e,t))return!0;n=n.next}while(n!==e);return!1}(e,t)&&(k(e,t)&&k(t,e)&&function(e,t){var n=e,r=!1,o=(e.x+t.x)/2,i=(e.y+t.y)/2;do{n.y>i!=n.next.y>i&&n.next.y!==n.y&&o<(n.next.x-n.x)*(i-n.y)/(n.next.y-n.y)+n.x&&(r=!r),n=n.next}while(n!==e);return r}(e,t)&&(Z(e.prev,e,t.prev)||Z(e,t.prev,t))||M(e,t)&&Z(e.prev,e,e.next)>0&&Z(t.prev,t,t.next)>0)}function Z(e,t,n){return(t.y-e.y)*(n.x-t.x)-(t.x-e.x)*(n.y-t.y)}function M(e,t){return e.x===t.x&&e.y===t.y}function F(e,t,n,r){var o=I(Z(e,t,n)),i=I(Z(e,t,r)),s=I(Z(n,r,e)),a=I(Z(n,r,t));return o!==i&&s!==a||(!(0!==o||!T(e,n,t))||(!(0!==i||!T(e,r,t))||(!(0!==s||!T(n,e,r))||!(0!==a||!T(n,t,r)))))}function T(e,t,n){return t.x<=Math.max(e.x,n.x)&&t.x>=Math.min(e.x,n.x)&&t.y<=Math.max(e.y,n.y)&&t.y>=Math.min(e.y,n.y)}function I(e){return e>0?1:e<0?-1:0}function k(e,t){return Z(e.prev,e,e.next)<0?Z(e,t,e.next)>=0&&Z(e,e.prev,t)>=0:Z(e,t,e.prev)<0||Z(e,e.next,t)<0}function S(e,t){var n=new R(e.i,e.x,e.y),r=new R(t.i,t.x,t.y),o=e.next,i=t.prev;return e.next=t,t.prev=e,n.next=o,o.prev=n,r.next=n,n.prev=r,i.next=r,r.prev=i,r}function U(e,t,n,r){var o=new R(e,t,n);return r?(o.next=r.next,o.prev=r,r.next.prev=o,r.next=o):(o.prev=o,o.next=o),o}function z(e){e.next.prev=e.prev,e.prev.next=e.next,e.prevZ&&(e.prevZ.nextZ=e.nextZ),e.nextZ&&(e.nextZ.prevZ=e.prevZ)}function R(e,t,n){this.i=e,this.x=t,this.y=n,this.prev=null,this.next=null,this.z=null,this.prevZ=null,this.nextZ=null,this.steiner=!1}function B(e,t,n,r){for(var o=0,i=t,s=n-r;i<n;i+=r)o+=(e[s]-e[i])*(e[i+1]+e[s+1]),s=i;return o}u.exports=x,u.exports.default=x,x.deviation=function(e,t,n,r){var o=t&&t.length,i=o?t[0]*n:e.length,s=Math.abs(B(e,0,i,n));if(o)for(var a=0,u=t.length;a<u;a++){var x=t[a]*n,f=a<u-1?t[a+1]*n:e.length;s-=Math.abs(B(e,x,f,n))}var l=0;for(a=0;a<r.length;a+=3){var c=r[a]*n,y=r[a+1]*n,d=r[a+2]*n;l+=Math.abs((e[c]-e[d])*(e[y+1]-e[c+1])-(e[c]-e[y])*(e[d+1]-e[c+1]))}return 0===s&&0===l?0:Math.abs((l-s)/s)},x.flatten=function(e){for(var t=e[0][0].length,n={vertices:[],holes:[],dimensions:t},r=0,o=0;o<e.length;o++){for(var i=0;i<e[o].length;i++)for(var s=0;s<t;s++)n.vertices.push(e[o][i][s]);o>0&&(r+=e[o-1].length,n.holes.push(r))}return n};const P=[],C={vertexPosition:0,indexPosition:0};function N(e,t,n,r,o){e[t+0]=n,e[t+1]=r,e[t+2]=o}function _(e,t,n,r,o,i){const s=3+o,a=e[t+0],u=e[t+1],x=P;x.length=o;for(let n=0;n<x.length;n++)x[n]=e[t+2+n];let f=i?i.vertexPosition:0,l=i?i.indexPosition:0;const c=f/s;return N(n,f,a,u,0),x.length&&n.set(x,f+3),f+=s,N(n,f,a,u,1),x.length&&n.set(x,f+3),f+=s,N(n,f,a,u,2),x.length&&n.set(x,f+3),f+=s,N(n,f,a,u,3),x.length&&n.set(x,f+3),f+=s,r[l++]=c,r[l++]=c+1,r[l++]=c+3,r[l++]=c+1,r[l++]=c+2,r[l++]=c+3,C.vertexPosition=f,C.indexPosition=l,C}function L(e,t,n,r,o,i,a,u,x,f){const l=5+u.length,c=i.length/l,y=[e[t+0],e[t+1]],d=[e[n],e[n+1]],h=s(f,[...y]),v=s(f,[...d]);function p(e,t,n){const r=1e4;return Math.round(1500*t)+Math.round(1500*n)*r+e*r*r}function g(e,t,n){const r=Math.sqrt((t[0]-e[0])*(t[0]-e[0])+(t[1]-e[1])*(t[1]-e[1])),o=[(t[0]-e[0])/r,(t[1]-e[1])/r],i=[-o[1],o[0]],s=Math.sqrt((n[0]-e[0])*(n[0]-e[0])+(n[1]-e[1])*(n[1]-e[1])),a=[(n[0]-e[0])/s,(n[1]-e[1])/s],u=0===r||0===s?0:Math.acos((x=a[0]*o[0]+a[1]*o[1],f=-1,l=1,Math.min(Math.max(x,f),l)));var x,f,l;return a[0]*i[0]+a[1]*i[1]>0?u:2*Math.PI-u}const b=null!==o;let m=0,w=0;if(null!==r){m=g(h,v,s(f,[...[e[r],e[r+1]]]))}if(b){w=g(v,h,s(f,[...[e[o],e[o+1]]]))}i.push(y[0],y[1],d[0],d[1],p(0,m,w)),i.push(...u),i.push(y[0],y[1],d[0],d[1],p(1,m,w)),i.push(...u),i.push(y[0],y[1],d[0],d[1],p(2,m,w)),i.push(...u),i.push(y[0],y[1],d[0],d[1],p(3,m,w)),i.push(...u),a.push(c,c+1,c+2,c+1,c+3,c+2)}function G(e,t,n,r,o){const i=2+o;let s=t;const a=e.slice(s,s+o);s+=o;const x=e[s++];let f=0;const l=new Array(x-1);for(let t=0;t<x;t++)f+=e[s++],t<x-1&&(l[t]=f);const c=e.slice(s,s+2*f),y=u.exports(c,l,2);for(let e=0;e<y.length;e++)r.push(y[e]+n.length/i);for(let e=0;e<c.length;e+=2)n.push(c[e],c[e+1],...a);return s+2*f}const O=self;O.onmessage=r=>{const o=r.data;switch(o.type){case t:{const e=3,t=2,n=o.customAttributesCount,r=t+n,i=new Float32Array(o.renderInstructions),s=i.length/r,a=4*s*(n+e),u=new Uint32Array(6*s),x=new Float32Array(a);let f;for(let e=0;e<i.length;e+=r)f=_(i,e,x,u,n,f);const l=Object.assign({vertexBuffer:x.buffer,indexBuffer:u.buffer,renderInstructions:i.buffer},o);O.postMessage(l,[x.buffer,u.buffer,i.buffer]);break}case n:{const e=[],t=[],n=o.customAttributesCount,r=2,i=new Float32Array(o.renderInstructions);let s=0;const u=o.renderInstructionsTransform,x=[1,0,0,1,0,0];let f,l;for(a(x,u);s<i.length;){l=Array.from(i.slice(s,s+n)),s+=n,f=i[s++];for(let n=0;n<f-1;n++)L(i,s+n*r,s+(n+1)*r,n>0?s+(n-1)*r:null,n<f-2?s+(n+2)*r:null,e,t,l,0,x);s+=f*r}const c=Uint32Array.from(t),y=Float32Array.from(e),d=Object.assign({vertexBuffer:y.buffer,indexBuffer:c.buffer,renderInstructions:i.buffer},o);O.postMessage(d,[y.buffer,c.buffer,i.buffer]);break}case e:{const e=[],t=[],n=o.customAttributesCount,r=new Float32Array(o.renderInstructions);let i=0;for(;i<r.length;)i=G(r,i,e,t,n);const s=Uint32Array.from(t),a=Float32Array.from(e),u=Object.assign({vertexBuffer:a.buffer,indexBuffer:s.buffer,renderInstructions:r.buffer},o);O.postMessage(u,[a.buffer,s.buffer,r.buffer]);break}}};";
            return new Worker(typeof Blob === 'undefined'
              ? 'data:application/javascript;base64,' + Buffer.from(source, 'binary').toString('base64')
              : URL.createObjectURL(new Blob([source], {type: 'application/javascript'})));
          }

  /**
   * @module ol/renderer/webgl/PointsLayer
   */

  /**
   * @typedef {Object} CustomAttribute A description of a custom attribute to be passed on to the GPU, with a value different
   * for each feature.
   * @property {string} name Attribute name.
   * @property {function(import("../../Feature").default, Object<string, *>):number} callback This callback computes the numerical value of the
   * attribute for a given feature (properties are available as 2nd arg for quicker access).
   */

  /**
   * @typedef {Object} FeatureCacheItem Object that holds a reference to a feature, its geometry and properties. Used to optimize
   * rebuildBuffers by accessing these objects quicker.
   * @property {import("../../Feature").default} feature Feature
   * @property {Object<string, *>} properties Feature properties
   * @property {import("../../geom").Geometry} geometry Feature geometry
   */

  /**
   * @typedef {Object} Options
   * @property {string} [className='ol-layer'] A CSS class name to set to the canvas element.
   * @property {Array<CustomAttribute>} [attributes] These attributes will be read from the features in the source and then
   * passed to the GPU. The `name` property of each attribute will serve as its identifier:
   *  * In the vertex shader as an `attribute` by prefixing it with `a_`
   *  * In the fragment shader as a `varying` by prefixing it with `v_`
   * Please note that these can only be numerical values.
   * @property {string} vertexShader Vertex shader source, mandatory.
   * @property {string} fragmentShader Fragment shader source, mandatory.
   * @property {string} [hitVertexShader] Vertex shader source for hit detection rendering.
   * @property {string} [hitFragmentShader] Fragment shader source for hit detection rendering.
   * @property {Object<string,import("../../webgl/Helper").UniformValue>} [uniforms] Uniform definitions for the post process steps
   * Please note that `u_texture` is reserved for the main texture slot and `u_opacity` is reserved for the layer opacity.
   * @property {Array<import("./Layer").PostProcessesOptions>} [postProcesses] Post-processes definitions
   */

  /**
   * @classdesc
   * WebGL vector renderer optimized for points.
   * All features will be rendered as quads (two triangles forming a square). New data will be flushed to the GPU
   * every time the vector source changes.
   *
   * You need to provide vertex and fragment shaders for rendering. This can be done using
   * {@link module:ol/webgl/ShaderBuilder~ShaderBuilder} utilities. These shaders shall expect a `a_position` attribute
   * containing the screen-space projected center of the quad, as well as a `a_index` attribute
   * whose value (0, 1, 2 or 3) indicates which quad vertex is currently getting processed (see structure below).
   *
   * To include variable attributes in the shaders, you need to declare them using the `attributes` property of
   * the options object like so:
   * ```js
   * new WebGLPointsLayerRenderer(layer, {
   *   attributes: [
   *     {
   *       name: 'size',
   *       callback: function(feature) {
   *         // compute something with the feature
   *       }
   *     },
   *     {
   *       name: 'weight',
   *       callback: function(feature) {
   *         // compute something with the feature
   *       }
   *     },
   *   ],
   *   vertexShader:
   *     // shader using attribute a_weight and a_size
   *   fragmentShader:
   *     // shader using varying v_weight and v_size
   * ```
   *
   * To enable hit detection, you must as well provide dedicated shaders using the `hitVertexShader`
   * and `hitFragmentShader` properties. These shall expect the `a_hitColor` attribute to contain
   * the final color that will have to be output for hit detection to work.
   *
   * The following uniform is used for the main texture: `u_texture`.
   * The following uniform is used for the layer opacity: `u_opacity`.
   *
   * Please note that the main shader output should have premultiplied alpha, otherwise visual anomalies may occur.
   *
   * Points are rendered as quads with the following structure:
   *
   * ```
   *   (u0, v1)      (u1, v1)
   *  [3]----------[2]
   *   |`           |
   *   |  `         |
   *   |    `       |
   *   |      `     |
   *   |        `   |
   *   |          ` |
   *  [0]----------[1]
   *   (u0, v0)      (u1, v0)
   *  ```
   *
   * This uses {@link module:ol/webgl/Helper~WebGLHelper} internally.
   *
   * @api
   */
  class WebGLPointsLayerRenderer extends WebGLLayerRenderer {
    /**
     * @param {import("../../layer/Layer.js").default} layer Layer.
     * @param {Options} options Options.
     */
    constructor(layer, options) {
      const uniforms = options.uniforms || {};
      const projectionMatrixTransform = create$2();
      uniforms[DefaultUniform.PROJECTION_MATRIX] = projectionMatrixTransform;

      super(layer, {
        uniforms: uniforms,
        postProcesses: options.postProcesses,
      });

      this.ready = false;

      this.sourceRevision_ = -1;

      this.verticesBuffer_ = new WebGLArrayBuffer(ARRAY_BUFFER, DYNAMIC_DRAW);
      this.hitVerticesBuffer_ = new WebGLArrayBuffer(ARRAY_BUFFER, DYNAMIC_DRAW);
      this.indicesBuffer_ = new WebGLArrayBuffer(
        ELEMENT_ARRAY_BUFFER,
        DYNAMIC_DRAW
      );

      /**
       * @private
       */
      this.vertexShader_ = options.vertexShader;

      /**
       * @private
       */
      this.fragmentShader_ = options.fragmentShader;

      /**
       * @type {WebGLProgram}
       * @private
       */
      this.program_;

      /**
       * @type {boolean}
       * @private
       */
      this.hitDetectionEnabled_ =
        options.hitFragmentShader && options.hitVertexShader ? true : false;

      /**
       * @private
       */
      this.hitVertexShader_ = options.hitVertexShader;

      /**
       * @private
       */
      this.hitFragmentShader_ = options.hitFragmentShader;

      /**
       * @type {WebGLProgram}
       * @private
       */
      this.hitProgram_;

      const customAttributes = options.attributes
        ? options.attributes.map(function (attribute) {
            return {
              name: 'a_' + attribute.name,
              size: 1,
              type: AttributeType.FLOAT,
            };
          })
        : [];

      /**
       * A list of attributes used by the renderer. By default only the position and
       * index of the vertex (0 to 3) are required.
       * @type {Array<import('../../webgl/Helper.js').AttributeDescription>}
       */
      this.attributes = [
        {
          name: 'a_position',
          size: 2,
          type: AttributeType.FLOAT,
        },
        {
          name: 'a_index',
          size: 1,
          type: AttributeType.FLOAT,
        },
      ].concat(customAttributes);

      /**
       * A list of attributes used for hit detection.
       * @type {Array<import('../../webgl/Helper.js').AttributeDescription>}
       */
      this.hitDetectionAttributes = [
        {
          name: 'a_position',
          size: 2,
          type: AttributeType.FLOAT,
        },
        {
          name: 'a_index',
          size: 1,
          type: AttributeType.FLOAT,
        },
        {
          name: 'a_hitColor',
          size: 4,
          type: AttributeType.FLOAT,
        },
        {
          name: 'a_featureUid',
          size: 1,
          type: AttributeType.FLOAT,
        },
      ].concat(customAttributes);

      this.customAttributes = options.attributes ? options.attributes : [];

      this.previousExtent_ = createEmpty();

      /**
       * This transform is updated on every frame and is the composition of:
       * - invert of the world->screen transform that was used when rebuilding buffers (see `this.renderTransform_`)
       * - current world->screen transform
       * @type {import("../../transform.js").Transform}
       * @private
       */
      this.currentTransform_ = projectionMatrixTransform;

      /**
       * This transform is updated when buffers are rebuilt and converts world space coordinates to screen space
       * @type {import("../../transform.js").Transform}
       * @private
       */
      this.renderTransform_ = create$2();

      /**
       * @type {import("../../transform.js").Transform}
       * @private
       */
      this.invertRenderTransform_ = create$2();

      /**
       * @type {Float32Array}
       * @private
       */
      this.renderInstructions_ = new Float32Array(0);

      /**
       * These instructions are used for hit detection
       * @type {Float32Array}
       * @private
       */
      this.hitRenderInstructions_ = new Float32Array(0);

      /**
       * @type {WebGLRenderTarget}
       * @private
       */
      this.hitRenderTarget_;

      /**
       * Keep track of latest message sent to worker
       * @type {number}
       * @private
       */
      this.generateBuffersRun_ = 0;

      /**
       * @private
       */
      this.worker_ = create();

      this.worker_.addEventListener(
        'message',
        /**
         * @param {*} event Event.
         * @this {WebGLPointsLayerRenderer}
         */
        function (event) {
          const received = event.data;
          if (received.type === WebGLWorkerMessageType.GENERATE_POINT_BUFFERS) {
            const projectionTransform = received.projectionTransform;
            if (received.hitDetection) {
              this.hitVerticesBuffer_.fromArrayBuffer(received.vertexBuffer);
              this.helper.flushBufferData(this.hitVerticesBuffer_);
            } else {
              this.verticesBuffer_.fromArrayBuffer(received.vertexBuffer);
              this.helper.flushBufferData(this.verticesBuffer_);
            }
            this.indicesBuffer_.fromArrayBuffer(received.indexBuffer);
            this.helper.flushBufferData(this.indicesBuffer_);

            this.renderTransform_ = projectionTransform;
            makeInverse(
              this.invertRenderTransform_,
              this.renderTransform_
            );
            if (received.hitDetection) {
              this.hitRenderInstructions_ = new Float32Array(
                event.data.renderInstructions
              );
            } else {
              this.renderInstructions_ = new Float32Array(
                event.data.renderInstructions
              );
              if (received.generateBuffersRun === this.generateBuffersRun_) {
                this.ready = true;
              }
            }

            this.getLayer().changed();
          }
        }.bind(this)
      );

      /**
       * This object will be updated when the source changes. Key is uid.
       * @type {Object<string, FeatureCacheItem>}
       * @private
       */
      this.featureCache_ = {};

      /**
       * Amount of features in the cache.
       * @type {number}
       * @private
       */
      this.featureCount_ = 0;

      const source = this.getLayer().getSource();
      this.sourceListenKeys_ = [
        listen(
          source,
          VectorEventType.ADDFEATURE,
          this.handleSourceFeatureAdded_,
          this
        ),
        listen(
          source,
          VectorEventType.CHANGEFEATURE,
          this.handleSourceFeatureChanged_,
          this
        ),
        listen(
          source,
          VectorEventType.REMOVEFEATURE,
          this.handleSourceFeatureDelete_,
          this
        ),
        listen(
          source,
          VectorEventType.CLEAR,
          this.handleSourceFeatureClear_,
          this
        ),
      ];
      source.forEachFeature(
        function (feature) {
          this.featureCache_[getUid(feature)] = {
            feature: feature,
            properties: feature.getProperties(),
            geometry: feature.getGeometry(),
          };
          this.featureCount_++;
        }.bind(this)
      );
    }

    afterHelperCreated() {
      this.program_ = this.helper.getProgram(
        this.fragmentShader_,
        this.vertexShader_
      );

      if (this.hitDetectionEnabled_) {
        this.hitProgram_ = this.helper.getProgram(
          this.hitFragmentShader_,
          this.hitVertexShader_
        );

        this.hitRenderTarget_ = new WebGLRenderTarget(this.helper);
      }
    }

    /**
     * @param {import("../../source/Vector.js").VectorSourceEvent} event Event.
     * @private
     */
    handleSourceFeatureAdded_(event) {
      const feature = event.feature;
      this.featureCache_[getUid(feature)] = {
        feature: feature,
        properties: feature.getProperties(),
        geometry: feature.getGeometry(),
      };
      this.featureCount_++;
    }

    /**
     * @param {import("../../source/Vector.js").VectorSourceEvent} event Event.
     * @private
     */
    handleSourceFeatureChanged_(event) {
      const feature = event.feature;
      this.featureCache_[getUid(feature)] = {
        feature: feature,
        properties: feature.getProperties(),
        geometry: feature.getGeometry(),
      };
    }

    /**
     * @param {import("../../source/Vector.js").VectorSourceEvent} event Event.
     * @private
     */
    handleSourceFeatureDelete_(event) {
      const feature = event.feature;
      delete this.featureCache_[getUid(feature)];
      this.featureCount_--;
    }

    /**
     * @private
     */
    handleSourceFeatureClear_() {
      this.featureCache_ = {};
      this.featureCount_ = 0;
    }

    /**
     * Render the layer.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @return {HTMLElement} The rendered element.
     */
    renderFrame(frameState) {
      const gl = this.helper.getGL();
      this.preRender(gl, frameState);

      const projection = frameState.viewState.projection;
      const layer = this.getLayer();
      const vectorSource = layer.getSource();
      // FIXME fix hit detection isn't reliable when rendering multiple worlds
      const multiWorld = vectorSource.getWrapX() && projection.canWrapX();
      const projectionExtent = projection.getExtent();

      const extent = frameState.extent;
      const worldWidth = multiWorld ? getWidth(projectionExtent) : null;
      const endWorld = multiWorld
        ? Math.ceil((extent[2] - projectionExtent[2]) / worldWidth) + 1
        : 1;

      const startWorld = multiWorld
        ? Math.floor((extent[0] - projectionExtent[0]) / worldWidth)
        : 0;

      let world = startWorld;
      const renderCount = this.indicesBuffer_.getSize();

      do {
        // apply the current projection transform with the invert of the one used to fill buffers
        this.helper.makeProjectionTransform(frameState, this.currentTransform_);
        translate$1(this.currentTransform_, world * worldWidth, 0);
        multiply(this.currentTransform_, this.invertRenderTransform_);
        this.helper.applyUniforms(frameState);

        this.helper.drawElements(0, renderCount);
      } while (++world < endWorld);

      this.helper.finalizeDraw(
        frameState,
        this.dispatchPreComposeEvent,
        this.dispatchPostComposeEvent
      );
      const canvas = this.helper.getCanvas();

      if (this.hitDetectionEnabled_) {
        this.renderHitDetection(frameState, startWorld, endWorld, worldWidth);
        this.hitRenderTarget_.clearCachedData();
      }

      this.postRender(gl, frameState);

      return canvas;
    }

    /**
     * Determine whether renderFrame should be called.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @return {boolean} Layer is ready to be rendered.
     */
    prepareFrameInternal(frameState) {
      const layer = this.getLayer();
      const vectorSource = layer.getSource();
      const viewState = frameState.viewState;
      const viewNotMoving =
        !frameState.viewHints[ViewHint.ANIMATING] &&
        !frameState.viewHints[ViewHint.INTERACTING];
      const extentChanged = !equals$2(this.previousExtent_, frameState.extent);
      const sourceChanged = this.sourceRevision_ < vectorSource.getRevision();

      if (sourceChanged) {
        this.sourceRevision_ = vectorSource.getRevision();
      }

      if (viewNotMoving && (extentChanged || sourceChanged)) {
        const projection = viewState.projection;
        const resolution = viewState.resolution;

        const renderBuffer =
          layer instanceof BaseVectorLayer ? layer.getRenderBuffer() : 0;
        const extent = buffer$1(frameState.extent, renderBuffer * resolution);
        vectorSource.loadFeatures(extent, resolution, projection);

        this.rebuildBuffers_(frameState);
        this.previousExtent_ = frameState.extent.slice();
      }

      this.helper.useProgram(this.program_, frameState);
      this.helper.prepareDraw(frameState);

      // write new data
      this.helper.bindBuffer(this.verticesBuffer_);
      this.helper.bindBuffer(this.indicesBuffer_);
      this.helper.enableAttributes(this.attributes);

      return true;
    }

    /**
     * Rebuild internal webgl buffers based on current view extent; costly, should not be called too much
     * @param {import("../../Map").FrameState} frameState Frame state.
     * @private
     */
    rebuildBuffers_(frameState) {
      // saves the projection transform for the current frame state
      const projectionTransform = create$2();
      this.helper.makeProjectionTransform(frameState, projectionTransform);

      // here we anticipate the amount of render instructions that we well generate
      // this can be done since we know that for normal render we only have x, y as base instructions,
      // and x, y, r, g, b, a and featureUid for hit render instructions
      // and we also know the amount of custom attributes to append to these
      const totalInstructionsCount =
        (2 + this.customAttributes.length) * this.featureCount_;
      if (
        !this.renderInstructions_ ||
        this.renderInstructions_.length !== totalInstructionsCount
      ) {
        this.renderInstructions_ = new Float32Array(totalInstructionsCount);
      }
      if (this.hitDetectionEnabled_) {
        const totalHitInstructionsCount =
          (7 + this.customAttributes.length) * this.featureCount_;
        if (
          !this.hitRenderInstructions_ ||
          this.hitRenderInstructions_.length !== totalHitInstructionsCount
        ) {
          this.hitRenderInstructions_ = new Float32Array(
            totalHitInstructionsCount
          );
        }
      }

      // loop on features to fill the buffer
      let featureCache, geometry;
      const tmpCoords = [];
      const tmpColor = [];
      let renderIndex = 0;
      let hitIndex = 0;
      let hitColor;
      for (const featureUid in this.featureCache_) {
        featureCache = this.featureCache_[featureUid];
        geometry = /** @type {import("../../geom").Point} */ (
          featureCache.geometry
        );
        if (!geometry || geometry.getType() !== 'Point') {
          continue;
        }

        tmpCoords[0] = geometry.getFlatCoordinates()[0];
        tmpCoords[1] = geometry.getFlatCoordinates()[1];
        apply(projectionTransform, tmpCoords);

        hitColor = colorEncodeId(hitIndex + 6, tmpColor);

        this.renderInstructions_[renderIndex++] = tmpCoords[0];
        this.renderInstructions_[renderIndex++] = tmpCoords[1];

        // for hit detection, the feature uid is saved in the opacity value
        // and the index of the opacity value is encoded in the color values
        if (this.hitDetectionEnabled_) {
          this.hitRenderInstructions_[hitIndex++] = tmpCoords[0];
          this.hitRenderInstructions_[hitIndex++] = tmpCoords[1];
          this.hitRenderInstructions_[hitIndex++] = hitColor[0];
          this.hitRenderInstructions_[hitIndex++] = hitColor[1];
          this.hitRenderInstructions_[hitIndex++] = hitColor[2];
          this.hitRenderInstructions_[hitIndex++] = hitColor[3];
          this.hitRenderInstructions_[hitIndex++] = Number(featureUid);
        }

        // pushing custom attributes
        let value;
        for (let j = 0; j < this.customAttributes.length; j++) {
          value = this.customAttributes[j].callback(
            featureCache.feature,
            featureCache.properties
          );
          this.renderInstructions_[renderIndex++] = value;
          if (this.hitDetectionEnabled_) {
            this.hitRenderInstructions_[hitIndex++] = value;
          }
        }
      }

      /** @type {import('../../render/webgl/constants.js').WebGLWorkerGenerateBuffersMessage} */
      const message = {
        id: 0,
        type: WebGLWorkerMessageType.GENERATE_POINT_BUFFERS,
        renderInstructions: this.renderInstructions_.buffer,
        customAttributesCount: this.customAttributes.length,
      };
      // additional properties will be sent back as-is by the worker
      message['projectionTransform'] = projectionTransform;
      message['generateBuffersRun'] = ++this.generateBuffersRun_;
      this.ready = false;
      this.worker_.postMessage(message, [this.renderInstructions_.buffer]);
      this.renderInstructions_ = null;

      /** @type {import('../../render/webgl/constants.js').WebGLWorkerGenerateBuffersMessage} */
      if (this.hitDetectionEnabled_) {
        const hitMessage = {
          id: 0,
          type: WebGLWorkerMessageType.GENERATE_POINT_BUFFERS,
          renderInstructions: this.hitRenderInstructions_.buffer,
          customAttributesCount: 5 + this.customAttributes.length,
        };
        hitMessage['projectionTransform'] = projectionTransform;
        hitMessage['hitDetection'] = true;
        this.worker_.postMessage(hitMessage, [
          this.hitRenderInstructions_.buffer,
        ]);
        this.hitRenderInstructions_ = null;
      }
    }

    /**
     * @param {import("../../coordinate.js").Coordinate} coordinate Coordinate.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @param {number} hitTolerance Hit tolerance in pixels.
     * @param {import("../vector.js").FeatureCallback<T>} callback Feature callback.
     * @param {Array<import("../Map.js").HitMatch<T>>} matches The hit detected matches with tolerance.
     * @return {T|undefined} Callback result.
     * @template T
     */
    forEachFeatureAtCoordinate(
      coordinate,
      frameState,
      hitTolerance,
      callback,
      matches
    ) {
      assert(this.hitDetectionEnabled_, 66);
      if (!this.hitRenderInstructions_) {
        return undefined;
      }

      const pixel = apply(
        frameState.coordinateToPixelTransform,
        coordinate.slice()
      );

      const data = this.hitRenderTarget_.readPixel(pixel[0] / 2, pixel[1] / 2);
      const color = [data[0] / 255, data[1] / 255, data[2] / 255, data[3] / 255];
      const index = colorDecodeId(color);
      const opacity = this.hitRenderInstructions_[index];
      const uid = Math.floor(opacity).toString();

      const source = this.getLayer().getSource();
      const feature = source.getFeatureByUid(uid);
      if (feature) {
        return callback(feature, this.getLayer(), null);
      }
      return undefined;
    }

    /**
     * Render the hit detection data to the corresponding render target
     * @param {import("../../Map.js").FrameState} frameState current frame state
     * @param {number} startWorld the world to render in the first iteration
     * @param {number} endWorld the last world to render
     * @param {number} worldWidth the width of the worlds being rendered
     */
    renderHitDetection(frameState, startWorld, endWorld, worldWidth) {
      // skip render entirely if vertex buffers not ready/generated yet
      if (!this.hitVerticesBuffer_.getSize()) {
        return;
      }

      let world = startWorld;

      this.hitRenderTarget_.setSize([
        Math.floor(frameState.size[0] / 2),
        Math.floor(frameState.size[1] / 2),
      ]);

      this.helper.useProgram(this.hitProgram_, frameState);
      this.helper.prepareDrawToRenderTarget(
        frameState,
        this.hitRenderTarget_,
        true
      );

      this.helper.bindBuffer(this.hitVerticesBuffer_);
      this.helper.bindBuffer(this.indicesBuffer_);
      this.helper.enableAttributes(this.hitDetectionAttributes);

      do {
        this.helper.makeProjectionTransform(frameState, this.currentTransform_);
        translate$1(this.currentTransform_, world * worldWidth, 0);
        multiply(this.currentTransform_, this.invertRenderTransform_);
        this.helper.applyUniforms(frameState);

        const renderCount = this.indicesBuffer_.getSize();
        this.helper.drawElements(0, renderCount);
      } while (++world < endWorld);
    }

    /**
     * Clean up.
     */
    disposeInternal() {
      this.worker_.terminate();
      this.layer_ = null;
      this.sourceListenKeys_.forEach(function (key) {
        unlistenByKey(key);
      });
      this.sourceListenKeys_ = null;
      super.disposeInternal();
    }
  }

  /**
   * @module ol/render/webgl/BatchRenderer
   */

  /**
   * @typedef {Object} CustomAttribute A description of a custom attribute to be passed on to the GPU, with a value different
   * for each feature.
   * @property {string} name Attribute name.
   * @property {function(import("../../Feature").default):number} callback This callback computes the numerical value of the
   * attribute for a given feature.
   */

  let workerMessageCounter = 0;

  /**
   * @classdesc Abstract class for batch renderers.
   * Batch renderers are meant to render the geometries contained in a {@link module:ol/render/webgl/GeometryBatch}
   * instance. They are responsible for generating render instructions and transforming them into WebGL buffers.
   */
  class AbstractBatchRenderer {
    /**
     * @param {import("../../webgl/Helper.js").default} helper WebGL helper instance
     * @param {Worker} worker WebGL worker instance
     * @param {string} vertexShader Vertex shader
     * @param {string} fragmentShader Fragment shader
     * @param {Array<CustomAttribute>} customAttributes List of custom attributes
     */
    constructor(helper, worker, vertexShader, fragmentShader, customAttributes) {
      /**
       * @type {import("../../webgl/Helper.js").default}
       * @private
       */
      this.helper_ = helper;

      /**
       * @type {Worker}
       * @private
       */
      this.worker_ = worker;

      /**
       * @type {WebGLProgram}
       * @private
       */
      this.program_ = this.helper_.getProgram(fragmentShader, vertexShader);

      /**
       * A list of attributes used by the renderer.
       * @type {Array<import('../../webgl/Helper.js').AttributeDescription>}
       * @protected
       */
      this.attributes = [];

      /**
       * @type {Array<CustomAttribute>}
       * @protected
       */
      this.customAttributes = customAttributes;
    }

    /**
     * Rebuild rendering instructions and webgl buffers based on the provided frame state
     * Note: this is a costly operation.
     * @param {import("./MixedGeometryBatch.js").GeometryBatch} batch Geometry batch
     * @param {import("../../Map").FrameState} frameState Frame state.
     * @param {import("../../geom/Geometry.js").Type} geometryType Geometry type
     * @param {function(): void} callback Function called once the render buffers are updated
     */
    rebuild(batch, frameState, geometryType, callback) {
      // store transform for rendering instructions
      batch.renderInstructionsTransform = this.helper_.makeProjectionTransform(
        frameState,
        create$2()
      );
      this.generateRenderInstructions(batch);
      this.generateBuffers_(batch, geometryType, callback);
    }

    /**
     * Render the geometries in the batch. This will also update the current transform used for rendering according to
     * the invert transform of the webgl buffers
     * @param {import("./MixedGeometryBatch.js").GeometryBatch} batch Geometry batch
     * @param {import("../../transform.js").Transform} currentTransform Transform
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @param {number} offsetX X offset
     */
    render(batch, currentTransform, frameState, offsetX) {
      // multiply the current projection transform with the invert of the one used to fill buffers
      this.helper_.makeProjectionTransform(frameState, currentTransform);
      translate$1(currentTransform, offsetX, 0);
      multiply(currentTransform, batch.invertVerticesBufferTransform);

      // enable program, buffers and attributes
      this.helper_.useProgram(this.program_, frameState);
      this.helper_.bindBuffer(batch.verticesBuffer);
      this.helper_.bindBuffer(batch.indicesBuffer);
      this.helper_.enableAttributes(this.attributes);

      const renderCount = batch.indicesBuffer.getSize();
      this.helper_.drawElements(0, renderCount);
    }

    /**
     * Rebuild rendering instructions based on the provided frame state
     * This is specific to the geometry type and has to be implemented by subclasses.
     * @param {import("./MixedGeometryBatch.js").GeometryBatch} batch Geometry batch
     * @protected
     */
    generateRenderInstructions(batch) {
      abstract();
    }

    /**
     * Rebuild internal webgl buffers for rendering based on the current rendering instructions;
     * This is asynchronous: webgl buffers wil _not_ be updated right away
     * @param {import("./MixedGeometryBatch.js").GeometryBatch} batch Geometry batch
     * @param {import("../../geom/Geometry.js").Type} geometryType Geometry type
     * @param {function(): void} callback Function called once the render buffers are updated
     * @private
     */
    generateBuffers_(batch, geometryType, callback) {
      const messageId = workerMessageCounter++;

      let messageType;
      switch (geometryType) {
        case 'Polygon':
          messageType = WebGLWorkerMessageType.GENERATE_POLYGON_BUFFERS;
          break;
        case 'Point':
          messageType = WebGLWorkerMessageType.GENERATE_POINT_BUFFERS;
          break;
        case 'LineString':
          messageType = WebGLWorkerMessageType.GENERATE_LINE_STRING_BUFFERS;
          break;
        // pass
      }

      /** @type {import('./constants.js').WebGLWorkerGenerateBuffersMessage} */
      const message = {
        id: messageId,
        type: messageType,
        renderInstructions: batch.renderInstructions.buffer,
        renderInstructionsTransform: batch.renderInstructionsTransform,
        customAttributesCount: this.customAttributes.length,
      };
      this.worker_.postMessage(message, [batch.renderInstructions.buffer]);

      // leave ownership of render instructions
      batch.renderInstructions = null;

      const handleMessage =
        /**
         * @param {*} event Event.
         * @this {AbstractBatchRenderer}
         */
        function (event) {
          const received = event.data;

          // this is not the response to our request: skip
          if (received.id !== messageId) {
            return;
          }

          // we've received our response: stop listening
          this.worker_.removeEventListener('message', handleMessage);

          // store transform & invert transform for webgl buffers
          batch.verticesBufferTransform = received.renderInstructionsTransform;
          makeInverse(
            batch.invertVerticesBufferTransform,
            batch.verticesBufferTransform
          );

          // copy & flush received buffers to GPU
          batch.verticesBuffer.fromArrayBuffer(received.vertexBuffer);
          this.helper_.flushBufferData(batch.verticesBuffer);
          batch.indicesBuffer.fromArrayBuffer(received.indexBuffer);
          this.helper_.flushBufferData(batch.indicesBuffer);

          // take back ownership of the render instructions for further use
          batch.renderInstructions = new Float32Array(
            received.renderInstructions
          );

          callback();
        }.bind(this);

      this.worker_.addEventListener('message', handleMessage);
    }
  }

  /**
   * @module ol/render/webgl/LineStringBatchRenderer
   */

  /**
   * Names of attributes made available to the vertex shader.
   * Please note: changing these *will* break custom shaders!
   * @enum {string}
   */
  const Attributes$2 = {
    SEGMENT_START: 'a_segmentStart',
    SEGMENT_END: 'a_segmentEnd',
    PARAMETERS: 'a_parameters',
  };

  class LineStringBatchRenderer extends AbstractBatchRenderer {
    /**
     * @param {import("../../webgl/Helper.js").default} helper WebGL helper instance
     * @param {Worker} worker WebGL worker instance
     * @param {string} vertexShader Vertex shader
     * @param {string} fragmentShader Fragment shader
     * @param {Array<import('./BatchRenderer.js').CustomAttribute>} customAttributes List of custom attributes
     */
    constructor(helper, worker, vertexShader, fragmentShader, customAttributes) {
      super(helper, worker, vertexShader, fragmentShader, customAttributes);

      // vertices for lines must hold both a position (x,y) and an offset (dx,dy)
      this.attributes = [
        {
          name: Attributes$2.SEGMENT_START,
          size: 2,
          type: AttributeType.FLOAT,
        },
        {
          name: Attributes$2.SEGMENT_END,
          size: 2,
          type: AttributeType.FLOAT,
        },
        {
          name: Attributes$2.PARAMETERS,
          size: 1,
          type: AttributeType.FLOAT,
        },
      ].concat(
        customAttributes.map(function (attribute) {
          return {
            name: 'a_' + attribute.name,
            size: 1,
            type: AttributeType.FLOAT,
          };
        })
      );
    }

    /**
     * Render instructions for lines are structured like so:
     * [ customAttr0, ... , customAttrN, numberOfVertices0, x0, y0, ... , xN, yN, numberOfVertices1, ... ]
     * @param {import("./MixedGeometryBatch.js").LineStringGeometryBatch} batch Linestring geometry batch
     * @override
     */
    generateRenderInstructions(batch) {
      // here we anticipate the amount of render instructions for lines:
      // 2 instructions per vertex for position (x and y)
      // + 1 instruction per line per custom attributes
      // + 1 instruction per line (for vertices count)
      const totalInstructionsCount =
        2 * batch.verticesCount +
        (1 + this.customAttributes.length) * batch.geometriesCount;
      if (
        !batch.renderInstructions ||
        batch.renderInstructions.length !== totalInstructionsCount
      ) {
        batch.renderInstructions = new Float32Array(totalInstructionsCount);
      }

      // loop on features to fill the render instructions
      let batchEntry;
      const flatCoords = [];
      let renderIndex = 0;
      let value;
      for (const featureUid in batch.entries) {
        batchEntry = batch.entries[featureUid];
        for (let i = 0, ii = batchEntry.flatCoordss.length; i < ii; i++) {
          flatCoords.length = batchEntry.flatCoordss[i].length;
          transform2D(
            batchEntry.flatCoordss[i],
            0,
            flatCoords.length,
            2,
            batch.renderInstructionsTransform,
            flatCoords
          );

          // custom attributes
          for (let k = 0, kk = this.customAttributes.length; k < kk; k++) {
            value = this.customAttributes[k].callback(batchEntry.feature);
            batch.renderInstructions[renderIndex++] = value;
          }

          // vertices count
          batch.renderInstructions[renderIndex++] = flatCoords.length / 2;

          // looping on points for positions
          for (let j = 0, jj = flatCoords.length; j < jj; j += 2) {
            batch.renderInstructions[renderIndex++] = flatCoords[j];
            batch.renderInstructions[renderIndex++] = flatCoords[j + 1];
          }
        }
      }
    }
  }

  /**
   * @module ol/render/webgl/MixedGeometryBatch
   */

  /**
   * @typedef {Object} GeometryBatchItem Object that holds a reference to a feature as well as the raw coordinates of its various geometries
   * @property {import("../../Feature").default} feature Feature
   * @property {Array<Array<number>>} flatCoordss Array of flat coordinates arrays, one for each geometry related to the feature
   * @property {number} [verticesCount] Only defined for linestring and polygon batches
   * @property {number} [ringsCount] Only defined for polygon batches
   * @property {Array<Array<number>>} [ringsVerticesCounts] Array of vertices counts in each ring for each geometry; only defined for polygons batches
   */

  /**
   * @typedef {PointGeometryBatch|LineStringGeometryBatch|PolygonGeometryBatch} GeometryBatch
   */

  /**
   * @typedef {Object} PolygonGeometryBatch A geometry batch specific to polygons
   * @property {Object<string, GeometryBatchItem>} entries Dictionary of all entries in the batch with associated computed values.
   * One entry corresponds to one feature. Key is feature uid.
   * @property {number} geometriesCount Amount of geometries in the batch.
   * @property {Float32Array} renderInstructions Render instructions for polygons are structured like so:
   * [ numberOfRings, numberOfVerticesInRing0, ..., numberOfVerticesInRingN, x0, y0, customAttr0, ..., xN, yN, customAttrN, numberOfRings,... ]
   * @property {WebGLArrayBuffer} verticesBuffer Vertices WebGL buffer
   * @property {WebGLArrayBuffer} indicesBuffer Indices WebGL buffer
   * @property {import("../../transform.js").Transform} renderInstructionsTransform Converts world space coordinates to screen space; applies to the rendering instructions
   * @property {import("../../transform.js").Transform} verticesBufferTransform Converts world space coordinates to screen space; applies to the webgl vertices buffer
   * @property {import("../../transform.js").Transform} invertVerticesBufferTransform Screen space to world space; applies to the webgl vertices buffer
   * @property {number} verticesCount Amount of vertices from geometries in the batch.
   * @property {number} ringsCount How many outer and inner rings in this batch.
   */

  /**
   * @typedef {Object} LineStringGeometryBatch A geometry batch specific to lines
   * @property {Object<string, GeometryBatchItem>} entries Dictionary of all entries in the batch with associated computed values.
   * One entry corresponds to one feature. Key is feature uid.
   * @property {number} geometriesCount Amount of geometries in the batch.
   * @property {Float32Array} renderInstructions Render instructions for polygons are structured like so:
   * [ numberOfRings, numberOfVerticesInRing0, ..., numberOfVerticesInRingN, x0, y0, customAttr0, ..., xN, yN, customAttrN, numberOfRings,... ]
   * @property {WebGLArrayBuffer} verticesBuffer Vertices WebGL buffer
   * @property {WebGLArrayBuffer} indicesBuffer Indices WebGL buffer
   * @property {import("../../transform.js").Transform} renderInstructionsTransform Converts world space coordinates to screen space; applies to the rendering instructions
   * @property {import("../../transform.js").Transform} verticesBufferTransform Converts world space coordinates to screen space; applies to the webgl vertices buffer
   * @property {import("../../transform.js").Transform} invertVerticesBufferTransform Screen space to world space; applies to the webgl vertices buffer
   * @property {number} verticesCount Amount of vertices from geometries in the batch.
   */

  /**
   * @typedef {Object} PointGeometryBatch A geometry batch specific to points
   * @property {Object<string, GeometryBatchItem>} entries Dictionary of all entries in the batch with associated computed values.
   * One entry corresponds to one feature. Key is feature uid.
   * @property {number} geometriesCount Amount of geometries in the batch.
   * @property {Float32Array} renderInstructions Render instructions for polygons are structured like so:
   * [ numberOfRings, numberOfVerticesInRing0, ..., numberOfVerticesInRingN, x0, y0, customAttr0, ..., xN, yN, customAttrN, numberOfRings,... ]
   * @property {WebGLArrayBuffer} verticesBuffer Vertices WebGL buffer
   * @property {WebGLArrayBuffer} indicesBuffer Indices WebGL buffer
   * @property {import("../../transform.js").Transform} renderInstructionsTransform Converts world space coordinates to screen space; applies to the rendering instructions
   * @property {import("../../transform.js").Transform} verticesBufferTransform Converts world space coordinates to screen space; applies to the webgl vertices buffer
   * @property {import("../../transform.js").Transform} invertVerticesBufferTransform Screen space to world space; applies to the webgl vertices buffer
   */

  /**
   * @classdesc This class is used to group several geometries of various types together for faster rendering.
   * Three inner batches are maintained for polygons, lines and points. Each time a feature is added, changed or removed
   * from the batch, these inner batches are modified accordingly in order to keep them up-to-date.
   *
   * A feature can be present in several inner batches, for example a polygon geometry will be present in the polygon batch
   * and its linar rings will be present in the line batch. Multi geometries are also broken down into individual geometries
   * and added to the corresponding batches in a recursive manner.
   *
   * Corresponding {@link module:ol/render/webgl/BatchRenderer} instances are then used to generate the render instructions
   * and WebGL buffers (vertices and indices) for each inner batches; render instructions are stored on the inner batches,
   * alongside the transform used to convert world coords to screen coords at the time these instructions were generated.
   * The resulting WebGL buffers are stored on the batches as well.
   *
   * An important aspect of geometry batches is that there is no guarantee that render instructions and WebGL buffers
   * are synchronized, i.e. render instructions can describe a new state while WebGL buffers might not have been written yet.
   * This is why two world-to-screen transforms are stored on each batch: one for the render instructions and one for
   * the WebGL buffers.
   */
  class MixedGeometryBatch {
    constructor() {
      /**
       * @type {PolygonGeometryBatch}
       */
      this.polygonBatch = {
        entries: {},
        geometriesCount: 0,
        verticesCount: 0,
        ringsCount: 0,
        renderInstructions: new Float32Array(0),
        verticesBuffer: new WebGLArrayBuffer(ARRAY_BUFFER, DYNAMIC_DRAW),
        indicesBuffer: new WebGLArrayBuffer(ELEMENT_ARRAY_BUFFER, DYNAMIC_DRAW),
        renderInstructionsTransform: create$2(),
        verticesBufferTransform: create$2(),
        invertVerticesBufferTransform: create$2(),
      };

      /**
       * @type {PointGeometryBatch}
       */
      this.pointBatch = {
        entries: {},
        geometriesCount: 0,
        renderInstructions: new Float32Array(0),
        verticesBuffer: new WebGLArrayBuffer(ARRAY_BUFFER, DYNAMIC_DRAW),
        indicesBuffer: new WebGLArrayBuffer(ELEMENT_ARRAY_BUFFER, DYNAMIC_DRAW),
        renderInstructionsTransform: create$2(),
        verticesBufferTransform: create$2(),
        invertVerticesBufferTransform: create$2(),
      };

      /**
       * @type {LineStringGeometryBatch}
       */
      this.lineStringBatch = {
        entries: {},
        geometriesCount: 0,
        verticesCount: 0,
        renderInstructions: new Float32Array(0),
        verticesBuffer: new WebGLArrayBuffer(ARRAY_BUFFER, DYNAMIC_DRAW),
        indicesBuffer: new WebGLArrayBuffer(ELEMENT_ARRAY_BUFFER, DYNAMIC_DRAW),
        renderInstructionsTransform: create$2(),
        verticesBufferTransform: create$2(),
        invertVerticesBufferTransform: create$2(),
      };
    }

    /**
     * @param {Array<import("../../Feature").default>} features Array of features to add to the batch
     */
    addFeatures(features) {
      for (let i = 0; i < features.length; i++) {
        this.addFeature(features[i]);
      }
    }

    /**
     * @param {import("../../Feature").default} feature Feature to add to the batch
     */
    addFeature(feature) {
      const geometry = feature.getGeometry();
      if (!geometry) {
        return;
      }
      this.addGeometry_(geometry, feature);
    }

    /**
     * @param {import("../../Feature").default} feature Feature
     * @return {GeometryBatchItem} Batch item added (or existing one)
     * @private
     */
    addFeatureEntryInPointBatch_(feature) {
      const uid = getUid(feature);
      if (!(uid in this.pointBatch.entries)) {
        this.pointBatch.entries[uid] = {
          feature: feature,
          flatCoordss: [],
        };
      }
      return this.pointBatch.entries[uid];
    }

    /**
     * @param {import("../../Feature").default} feature Feature
     * @return {GeometryBatchItem} Batch item added (or existing one)
     * @private
     */
    addFeatureEntryInLineStringBatch_(feature) {
      const uid = getUid(feature);
      if (!(uid in this.lineStringBatch.entries)) {
        this.lineStringBatch.entries[uid] = {
          feature: feature,
          flatCoordss: [],
          verticesCount: 0,
        };
      }
      return this.lineStringBatch.entries[uid];
    }

    /**
     * @param {import("../../Feature").default} feature Feature
     * @return {GeometryBatchItem} Batch item added (or existing one)
     * @private
     */
    addFeatureEntryInPolygonBatch_(feature) {
      const uid = getUid(feature);
      if (!(uid in this.polygonBatch.entries)) {
        this.polygonBatch.entries[uid] = {
          feature: feature,
          flatCoordss: [],
          verticesCount: 0,
          ringsCount: 0,
          ringsVerticesCounts: [],
        };
      }
      return this.polygonBatch.entries[uid];
    }

    /**
     * @param {import("../../Feature").default} feature Feature
     * @private
     */
    clearFeatureEntryInPointBatch_(feature) {
      const entry = this.pointBatch.entries[getUid(feature)];
      if (!entry) {
        return;
      }
      this.pointBatch.geometriesCount -= entry.flatCoordss.length;
      delete this.pointBatch.entries[getUid(feature)];
    }

    /**
     * @param {import("../../Feature").default} feature Feature
     * @private
     */
    clearFeatureEntryInLineStringBatch_(feature) {
      const entry = this.lineStringBatch.entries[getUid(feature)];
      if (!entry) {
        return;
      }
      this.lineStringBatch.verticesCount -= entry.verticesCount;
      this.lineStringBatch.geometriesCount -= entry.flatCoordss.length;
      delete this.lineStringBatch.entries[getUid(feature)];
    }

    /**
     * @param {import("../../Feature").default} feature Feature
     * @private
     */
    clearFeatureEntryInPolygonBatch_(feature) {
      const entry = this.polygonBatch.entries[getUid(feature)];
      if (!entry) {
        return;
      }
      this.polygonBatch.verticesCount -= entry.verticesCount;
      this.polygonBatch.ringsCount -= entry.ringsCount;
      this.polygonBatch.geometriesCount -= entry.flatCoordss.length;
      delete this.polygonBatch.entries[getUid(feature)];
    }

    /**
     * @param {import("../../geom").Geometry} geometry Geometry
     * @param {import("../../Feature").default} feature Feature
     * @private
     */
    addGeometry_(geometry, feature) {
      const type = geometry.getType();
      let flatCoords;
      let verticesCount;
      let batchEntry;
      switch (type) {
        case 'GeometryCollection':
          /** @type {import("../../geom").GeometryCollection} */ (geometry)
            .getGeometries()
            .map((geom) => this.addGeometry_(geom, feature));
          break;
        case 'MultiPolygon':
          /** @type {import("../../geom").MultiPolygon} */ (geometry)
            .getPolygons()
            .map((polygon) => this.addGeometry_(polygon, feature));
          break;
        case 'MultiLineString':
          /** @type {import("../../geom").MultiLineString} */ (geometry)
            .getLineStrings()
            .map((line) => this.addGeometry_(line, feature));
          break;
        case 'MultiPoint':
          /** @type {import("../../geom").MultiPoint} */ (geometry)
            .getPoints()
            .map((point) => this.addGeometry_(point, feature));
          break;
        case 'Polygon':
          const polygonGeom = /** @type {import("../../geom").Polygon} */ (
            geometry
          );
          batchEntry = this.addFeatureEntryInPolygonBatch_(feature);
          flatCoords = polygonGeom.getFlatCoordinates();
          verticesCount = flatCoords.length / 2;
          const ringsCount = polygonGeom.getLinearRingCount();
          const ringsVerticesCount = polygonGeom
            .getEnds()
            .map((end, ind, arr) =>
              ind > 0 ? (end - arr[ind - 1]) / 2 : end / 2
            );
          this.polygonBatch.verticesCount += verticesCount;
          this.polygonBatch.ringsCount += ringsCount;
          this.polygonBatch.geometriesCount++;
          batchEntry.flatCoordss.push(flatCoords);
          batchEntry.ringsVerticesCounts.push(ringsVerticesCount);
          batchEntry.verticesCount += verticesCount;
          batchEntry.ringsCount += ringsCount;
          polygonGeom
            .getLinearRings()
            .map((ring) => this.addGeometry_(ring, feature));
          break;
        case 'Point':
          const pointGeom = /** @type {import("../../geom").Point} */ (geometry);
          batchEntry = this.addFeatureEntryInPointBatch_(feature);
          flatCoords = pointGeom.getFlatCoordinates();
          this.pointBatch.geometriesCount++;
          batchEntry.flatCoordss.push(flatCoords);
          break;
        case 'LineString':
        case 'LinearRing':
          const lineGeom = /** @type {import("../../geom").LineString} */ (
            geometry
          );
          batchEntry = this.addFeatureEntryInLineStringBatch_(feature);
          flatCoords = lineGeom.getFlatCoordinates();
          verticesCount = flatCoords.length / 2;
          this.lineStringBatch.verticesCount += verticesCount;
          this.lineStringBatch.geometriesCount++;
          batchEntry.flatCoordss.push(flatCoords);
          batchEntry.verticesCount += verticesCount;
          break;
        // pass
      }
    }

    /**
     * @param {import("../../Feature").default} feature Feature
     */
    changeFeature(feature) {
      this.clearFeatureEntryInPointBatch_(feature);
      this.clearFeatureEntryInPolygonBatch_(feature);
      this.clearFeatureEntryInLineStringBatch_(feature);
      const geometry = feature.getGeometry();
      if (!geometry) {
        return;
      }
      this.addGeometry_(geometry, feature);
    }

    /**
     * @param {import("../../Feature").default} feature Feature
     */
    removeFeature(feature) {
      this.clearFeatureEntryInPointBatch_(feature);
      this.clearFeatureEntryInPolygonBatch_(feature);
      this.clearFeatureEntryInLineStringBatch_(feature);
    }

    clear() {
      this.polygonBatch.entries = {};
      this.polygonBatch.geometriesCount = 0;
      this.polygonBatch.verticesCount = 0;
      this.polygonBatch.ringsCount = 0;
      this.lineStringBatch.entries = {};
      this.lineStringBatch.geometriesCount = 0;
      this.lineStringBatch.verticesCount = 0;
      this.pointBatch.entries = {};
      this.pointBatch.geometriesCount = 0;
    }
  }

  /**
   * @module ol/render/webgl/PointBatchRenderer
   */

  /**
   * Names of attributes made available to the vertex shader.
   * Please note: changing these *will* break custom shaders!
   * @enum {string}
   */
  const Attributes$1 = {
    POSITION: 'a_position',
    INDEX: 'a_index',
  };

  class PointBatchRenderer extends AbstractBatchRenderer {
    /**
     * @param {import("../../webgl/Helper.js").default} helper WebGL helper instance
     * @param {Worker} worker WebGL worker instance
     * @param {string} vertexShader Vertex shader
     * @param {string} fragmentShader Fragment shader
     * @param {Array<import('./BatchRenderer.js').CustomAttribute>} customAttributes List of custom attributes
     */
    constructor(helper, worker, vertexShader, fragmentShader, customAttributes) {
      super(helper, worker, vertexShader, fragmentShader, customAttributes);

      // vertices for point must hold both a position (x,y) and an index (their position in the quad)
      this.attributes = [
        {
          name: Attributes$1.POSITION,
          size: 2,
          type: AttributeType.FLOAT,
        },
        {
          name: Attributes$1.INDEX,
          size: 1,
          type: AttributeType.FLOAT,
        },
      ].concat(
        customAttributes.map(function (attribute) {
          return {
            name: 'a_' + attribute.name,
            size: 1,
            type: AttributeType.FLOAT,
          };
        })
      );
    }

    /**
     * Render instructions for lines are structured like so:
     * [ x0, y0, customAttr0, ... , xN, yN, customAttrN ]
     * @param {import("./MixedGeometryBatch.js").PointGeometryBatch} batch Point geometry batch
     * @override
     */
    generateRenderInstructions(batch) {
      // here we anticipate the amount of render instructions for points:
      // 2 instructions per vertex for position (x and y)
      // + 1 instruction per vertex per custom attributes
      const totalInstructionsCount =
        (2 + this.customAttributes.length) * batch.geometriesCount;
      if (
        !batch.renderInstructions ||
        batch.renderInstructions.length !== totalInstructionsCount
      ) {
        batch.renderInstructions = new Float32Array(totalInstructionsCount);
      }

      // loop on features to fill the render instructions
      let batchEntry;
      const tmpCoords = [];
      let renderIndex = 0;
      let value;
      for (const featureUid in batch.entries) {
        batchEntry = batch.entries[featureUid];
        for (let i = 0, ii = batchEntry.flatCoordss.length; i < ii; i++) {
          tmpCoords[0] = batchEntry.flatCoordss[i][0];
          tmpCoords[1] = batchEntry.flatCoordss[i][1];
          apply(batch.renderInstructionsTransform, tmpCoords);

          batch.renderInstructions[renderIndex++] = tmpCoords[0];
          batch.renderInstructions[renderIndex++] = tmpCoords[1];

          // pushing custom attributes
          for (let j = 0, jj = this.customAttributes.length; j < jj; j++) {
            value = this.customAttributes[j].callback(batchEntry.feature);
            batch.renderInstructions[renderIndex++] = value;
          }
        }
      }
    }
  }

  /**
   * @module ol/render/webgl/PolygonBatchRenderer
   */

  /**
   * Names of attributes made available to the vertex shader.
   * Please note: changing these *will* break custom shaders!
   * @enum {string}
   */
  const Attributes = {
    POSITION: 'a_position',
  };

  class PolygonBatchRenderer extends AbstractBatchRenderer {
    /**
     * @param {import("../../webgl/Helper.js").default} helper WebGL helper instance
     * @param {Worker} worker WebGL worker instance
     * @param {string} vertexShader Vertex shader
     * @param {string} fragmentShader Fragment shader
     * @param {Array<import('./BatchRenderer.js').CustomAttribute>} customAttributes List of custom attributes
     */
    constructor(helper, worker, vertexShader, fragmentShader, customAttributes) {
      super(helper, worker, vertexShader, fragmentShader, customAttributes);

      // By default only a position attribute is required to render polygons
      this.attributes = [
        {
          name: Attributes.POSITION,
          size: 2,
          type: AttributeType.FLOAT,
        },
      ].concat(
        customAttributes.map(function (attribute) {
          return {
            name: 'a_' + attribute.name,
            size: 1,
            type: AttributeType.FLOAT,
          };
        })
      );
    }

    /**
     * Render instructions for polygons are structured like so:
     * [ customAttr0, ..., customAttrN, numberOfRings, numberOfVerticesInRing0, ..., numberOfVerticesInRingN, x0, y0, ..., xN, yN, numberOfRings,... ]
     * @param {import("./MixedGeometryBatch.js").PolygonGeometryBatch} batch Polygon geometry batch
     * @override
     */
    generateRenderInstructions(batch) {
      // here we anticipate the amount of render instructions for polygons:
      // 2 instructions per vertex for position (x and y)
      // + 1 instruction per polygon per custom attributes
      // + 1 instruction per polygon (for vertices count in polygon)
      // + 1 instruction per ring (for vertices count in ring)
      const totalInstructionsCount =
        2 * batch.verticesCount +
        (1 + this.customAttributes.length) * batch.geometriesCount +
        batch.ringsCount;
      if (
        !batch.renderInstructions ||
        batch.renderInstructions.length !== totalInstructionsCount
      ) {
        batch.renderInstructions = new Float32Array(totalInstructionsCount);
      }

      // loop on features to fill the render instructions
      let batchEntry;
      const flatCoords = [];
      let renderIndex = 0;
      let value;
      for (const featureUid in batch.entries) {
        batchEntry = batch.entries[featureUid];
        for (let i = 0, ii = batchEntry.flatCoordss.length; i < ii; i++) {
          flatCoords.length = batchEntry.flatCoordss[i].length;
          transform2D(
            batchEntry.flatCoordss[i],
            0,
            flatCoords.length,
            2,
            batch.renderInstructionsTransform,
            flatCoords
          );

          // custom attributes
          for (let k = 0, kk = this.customAttributes.length; k < kk; k++) {
            value = this.customAttributes[k].callback(batchEntry.feature);
            batch.renderInstructions[renderIndex++] = value;
          }

          // ring count
          batch.renderInstructions[renderIndex++] =
            batchEntry.ringsVerticesCounts[i].length;

          // vertices count in each ring
          for (
            let j = 0, jj = batchEntry.ringsVerticesCounts[i].length;
            j < jj;
            j++
          ) {
            batch.renderInstructions[renderIndex++] =
              batchEntry.ringsVerticesCounts[i][j];
          }

          // looping on points for positions
          for (let j = 0, jj = flatCoords.length; j < jj; j += 2) {
            batch.renderInstructions[renderIndex++] = flatCoords[j];
            batch.renderInstructions[renderIndex++] = flatCoords[j + 1];
          }
        }
      }
    }
  }

  /**
   * @module ol/renderer/webgl/shaders
   */

  /** @typedef {'color'|'opacity'|'width'} DefaultAttributes */

  /**
   * Packs red/green/blue channels of a color into a single float value; alpha is ignored.
   * This is how the color is expected to be computed.
   * @param {import("../../color.js").Color|string} color Color as array of numbers or string
   * @return {number} Float value containing the color
   */
  function packColor(color) {
    const array = asArray(color);
    const r = array[0] * 256 * 256;
    const g = array[1] * 256;
    const b = array[2];
    return r + g + b;
  }

  const DECODE_COLOR_EXPRESSION = `vec3(
  fract(floor(a_color / 256.0 / 256.0) / 256.0),
  fract(floor(a_color / 256.0) / 256.0),
  fract(a_color / 256.0)
);`;

  /**
   * Default polygon vertex shader.
   * Relies on the color and opacity attributes.
   * @type {string}
   */
  const FILL_VERTEX_SHADER = `
  precision mediump float;
  uniform mat4 u_projectionMatrix;
  attribute vec2 a_position;
  attribute float a_color;
  attribute float a_opacity;
  varying vec3 v_color;
  varying float v_opacity;

  void main(void) {
    gl_Position = u_projectionMatrix * vec4(a_position, 0.0, 1.0);
    v_color = ${DECODE_COLOR_EXPRESSION}
    v_opacity = a_opacity;
  }`;

  /**
   * Default polygon fragment shader.
   * @type {string}
   */
  const FILL_FRAGMENT_SHADER = `
  precision mediump float;
  varying vec3 v_color;
  varying float v_opacity;

  void main(void) {
    gl_FragColor = vec4(v_color, 1.0) * v_opacity;
  }`;

  /**
   * Default linestring vertex shader.
   * Relies on color, opacity and width attributes.
   * @type {string}
   */
  const STROKE_VERTEX_SHADER = `
  precision mediump float;
  uniform mat4 u_projectionMatrix;
  uniform vec2 u_sizePx;
  attribute vec2 a_segmentStart;
  attribute vec2 a_segmentEnd;
  attribute float a_parameters;
  attribute float a_color;
  attribute float a_opacity;
  attribute float a_width;
  varying vec2 v_segmentStart;
  varying vec2 v_segmentEnd;
  varying float v_angleStart;
  varying float v_angleEnd;
  varying vec3 v_color;
  varying float v_opacity;
  varying float v_width;

  vec2 worldToPx(vec2 worldPos) {
    vec4 screenPos = u_projectionMatrix * vec4(worldPos, 0.0, 1.0);
    return (0.5 * screenPos.xy + 0.5) * u_sizePx;
  }

  vec4 pxToScreen(vec2 pxPos) {
    vec2 screenPos = pxPos * 4.0 / u_sizePx;
    return vec4(screenPos.xy, 0.0, 0.0);
  }

  vec2 getOffsetDirection(vec2 normalPx, vec2 tangentPx, float joinAngle) {
    if (cos(joinAngle) > 0.93) return normalPx - tangentPx;
    float halfAngle = joinAngle / 2.0;
    vec2 angleBisectorNormal = vec2(
      sin(halfAngle) * normalPx.x + cos(halfAngle) * normalPx.y,
      -cos(halfAngle) * normalPx.x + sin(halfAngle) * normalPx.y
    );
    float length = 1.0 / sin(halfAngle);
    return angleBisectorNormal * length;
  }

  void main(void) {
    float anglePrecision = 1500.0;
    float paramShift = 10000.0;
    v_angleStart = fract(a_parameters / paramShift) * paramShift / anglePrecision;
    v_angleEnd = fract(floor(a_parameters / paramShift + 0.5) / paramShift) * paramShift / anglePrecision;
    float vertexNumber = floor(a_parameters / paramShift / paramShift + 0.0001);
    vec2 tangentPx = worldToPx(a_segmentEnd) - worldToPx(a_segmentStart);
    tangentPx = normalize(tangentPx);
    vec2 normalPx = vec2(-tangentPx.y, tangentPx.x);
    float normalDir = vertexNumber < 0.5 || (vertexNumber > 1.5 && vertexNumber < 2.5) ? 1.0 : -1.0;
    float tangentDir = vertexNumber < 1.5 ? 1.0 : -1.0;
    float angle = vertexNumber < 1.5 ? v_angleStart : v_angleEnd;
    vec2 offsetPx = getOffsetDirection(normalPx * normalDir, tangentDir * tangentPx, angle) * a_width * 0.5;
    vec2 position =  vertexNumber < 1.5 ? a_segmentStart : a_segmentEnd;
    gl_Position = u_projectionMatrix * vec4(position, 0.0, 1.0) + pxToScreen(offsetPx);
    v_segmentStart = worldToPx(a_segmentStart);
    v_segmentEnd = worldToPx(a_segmentEnd);
    v_color = ${DECODE_COLOR_EXPRESSION}
    v_opacity = a_opacity;
    v_width = a_width;
  }`;

  /**
   * Default linestring fragment shader.
   * @type {string}
   */
  const STROKE_FRAGMENT_SHADER = `
  precision mediump float;
  uniform float u_pixelRatio;
  varying vec2 v_segmentStart;
  varying vec2 v_segmentEnd;
  varying float v_angleStart;
  varying float v_angleEnd;
  varying vec3 v_color;
  varying float v_opacity;
  varying float v_width;

  float segmentDistanceField(vec2 point, vec2 start, vec2 end, float radius) {
    vec2 startToPoint = point - start;
    vec2 startToEnd = end - start;
    float ratio = clamp(dot(startToPoint, startToEnd) / dot(startToEnd, startToEnd), 0.0, 1.0);
    float dist = length(startToPoint - ratio * startToEnd);
    return 1.0 - smoothstep(radius - 1.0, radius, dist);
  }

  void main(void) {
    vec2 v_currentPoint = gl_FragCoord.xy / u_pixelRatio;
    gl_FragColor = vec4(v_color, 1.0) * v_opacity;
    gl_FragColor *= segmentDistanceField(v_currentPoint, v_segmentStart, v_segmentEnd, v_width);
  }`;

  /**
   * Default point vertex shader.
   * Relies on color and opacity attributes.
   * @type {string}
   */
  const POINT_VERTEX_SHADER = `
  precision mediump float;
  uniform mat4 u_projectionMatrix;
  uniform mat4 u_offsetScaleMatrix;
  attribute vec2 a_position;
  attribute float a_index;
  attribute float a_color;
  attribute float a_opacity;
  varying vec2 v_texCoord;
  varying vec3 v_color;
  varying float v_opacity;

  void main(void) {
    mat4 offsetMatrix = u_offsetScaleMatrix;
    float size = 6.0;
    float offsetX = a_index == 0.0 || a_index == 3.0 ? -size / 2.0 : size / 2.0;
    float offsetY = a_index == 0.0 || a_index == 1.0 ? -size / 2.0 : size / 2.0;
    vec4 offsets = offsetMatrix * vec4(offsetX, offsetY, 0.0, 0.0);
    gl_Position = u_projectionMatrix * vec4(a_position, 0.0, 1.0) + offsets;
    float u = a_index == 0.0 || a_index == 3.0 ? 0.0 : 1.0;
    float v = a_index == 0.0 || a_index == 1.0 ? 0.0 : 1.0;
    v_texCoord = vec2(u, v);
    v_color = ${DECODE_COLOR_EXPRESSION}
    v_opacity = a_opacity;
  }`;

  /**
   * Default point fragment shader.
   * @type {string}
   */
  const POINT_FRAGMENT_SHADER = `
  precision mediump float;
  varying vec3 v_color;
  varying float v_opacity;

  void main(void) {
      gl_FragColor = vec4(v_color, 1.0) * v_opacity;
  }`;

  /**
   * @module ol/renderer/webgl/VectorLayer
   */

  /**
   * @typedef {function(import("../../Feature").default, Object<string, *>):number} CustomAttributeCallback A callback computing
   * the value of a custom attribute (different for each feature) to be passed on to the GPU.
   * Properties are available as 2nd arg for quicker access.
   */

  /**
   * @typedef {Object} ShaderProgram An object containing both shaders (vertex and fragment) as well as the required attributes
   * @property {string} [vertexShader] Vertex shader source (using the default one if unspecified).
   * @property {string} [fragmentShader] Fragment shader source (using the default one if unspecified).
   * @property {Object<import("./shaders.js").DefaultAttributes,CustomAttributeCallback>} attributes Custom attributes made available in the vertex shader.
   * Keys are the names of the attributes which are then accessible in the vertex shader using the `a_` prefix, e.g.: `a_opacity`.
   * Default shaders rely on the attributes in {@link module:ol/render/webgl/shaders~DefaultAttributes}.
   */

  /**
   * @typedef {Object} Options
   * @property {string} [className='ol-layer'] A CSS class name to set to the canvas element.
   * @property {ShaderProgram} [fill] Attributes and shaders for filling polygons.
   * @property {ShaderProgram} [stroke] Attributes and shaders for line strings and polygon strokes.
   * @property {ShaderProgram} [point] Attributes and shaders for points.
   * @property {Object<string,import("../../webgl/Helper").UniformValue>} [uniforms] Uniform definitions.
   * @property {Array<import("./Layer").PostProcessesOptions>} [postProcesses] Post-processes definitions
   */

  /**
   * @param {Object<import("./shaders.js").DefaultAttributes,CustomAttributeCallback>} obj Lookup of attribute getters.
   * @return {Array<import("../../render/webgl/BatchRenderer").CustomAttribute>} An array of attribute descriptors.
   */
  function toAttributesArray(obj) {
    return Object.keys(obj).map((key) => ({name: key, callback: obj[key]}));
  }

  /**
   * @classdesc
   * Experimental WebGL vector renderer. Supports polygons, lines and points:
   *  * Polygons are broken down into triangles
   *  * Lines are rendered as strips of quads
   *  * Points are rendered as quads
   *
   * You need to provide vertex and fragment shaders as well as custom attributes for each type of geometry. All shaders
   * can access the uniforms in the {@link module:ol/webgl/Helper~DefaultUniform} enum.
   * The vertex shaders can access the following attributes depending on the geometry type:
   *  * For polygons: {@link module:ol/render/webgl/PolygonBatchRenderer~Attributes}
   *  * For line strings: {@link module:ol/render/webgl/LineStringBatchRenderer~Attributes}
   *  * For points: {@link module:ol/render/webgl/PointBatchRenderer~Attributes}
   *
   * Please note that the fragment shaders output should have premultiplied alpha, otherwise visual anomalies may occur.
   *
   * Note: this uses {@link module:ol/webgl/Helper~WebGLHelper} internally.
   */
  class WebGLVectorLayerRenderer extends WebGLLayerRenderer {
    /**
     * @param {import("../../layer/Layer.js").default} layer Layer.
     * @param {Options} options Options.
     */
    constructor(layer, options) {
      const uniforms = options.uniforms || {};
      const projectionMatrixTransform = create$2();
      uniforms[DefaultUniform.PROJECTION_MATRIX] = projectionMatrixTransform;

      super(layer, {
        uniforms: uniforms,
        postProcesses: options.postProcesses,
      });

      this.sourceRevision_ = -1;

      this.previousExtent_ = createEmpty();

      /**
       * This transform is updated on every frame and is the composition of:
       * - invert of the world->screen transform that was used when rebuilding buffers (see `this.renderTransform_`)
       * - current world->screen transform
       * @type {import("../../transform.js").Transform}
       * @private
       */
      this.currentTransform_ = projectionMatrixTransform;

      const fillAttributes = {
        color: function () {
          return packColor('#ddd');
        },
        opacity: function () {
          return 1;
        },
        ...(options.fill && options.fill.attributes),
      };

      const strokeAttributes = {
        color: function () {
          return packColor('#eee');
        },
        opacity: function () {
          return 1;
        },
        width: function () {
          return 1.5;
        },
        ...(options.stroke && options.stroke.attributes),
      };

      const pointAttributes = {
        color: function () {
          return packColor('#eee');
        },
        opacity: function () {
          return 1;
        },
        ...(options.point && options.point.attributes),
      };

      this.fillVertexShader_ =
        (options.fill && options.fill.vertexShader) || FILL_VERTEX_SHADER;
      this.fillFragmentShader_ =
        (options.fill && options.fill.fragmentShader) || FILL_FRAGMENT_SHADER;
      this.fillAttributes_ = toAttributesArray(fillAttributes);

      this.strokeVertexShader_ =
        (options.stroke && options.stroke.vertexShader) || STROKE_VERTEX_SHADER;
      this.strokeFragmentShader_ =
        (options.stroke && options.stroke.fragmentShader) ||
        STROKE_FRAGMENT_SHADER;
      this.strokeAttributes_ = toAttributesArray(strokeAttributes);

      this.pointVertexShader_ =
        (options.point && options.point.vertexShader) || POINT_VERTEX_SHADER;
      this.pointFragmentShader_ =
        (options.point && options.point.fragmentShader) || POINT_FRAGMENT_SHADER;
      this.pointAttributes_ = toAttributesArray(pointAttributes);

      /**
       * @private
       */
      this.worker_ = create();

      /**
       * @private
       */
      this.batch_ = new MixedGeometryBatch();

      const source = this.getLayer().getSource();
      this.batch_.addFeatures(source.getFeatures());
      this.sourceListenKeys_ = [
        listen(
          source,
          VectorEventType.ADDFEATURE,
          this.handleSourceFeatureAdded_,
          this
        ),
        listen(
          source,
          VectorEventType.CHANGEFEATURE,
          this.handleSourceFeatureChanged_,
          this
        ),
        listen(
          source,
          VectorEventType.REMOVEFEATURE,
          this.handleSourceFeatureDelete_,
          this
        ),
        listen(
          source,
          VectorEventType.CLEAR,
          this.handleSourceFeatureClear_,
          this
        ),
      ];
    }

    afterHelperCreated() {
      this.polygonRenderer_ = new PolygonBatchRenderer(
        this.helper,
        this.worker_,
        this.fillVertexShader_,
        this.fillFragmentShader_,
        this.fillAttributes_
      );
      this.pointRenderer_ = new PointBatchRenderer(
        this.helper,
        this.worker_,
        this.pointVertexShader_,
        this.pointFragmentShader_,
        this.pointAttributes_
      );
      this.lineStringRenderer_ = new LineStringBatchRenderer(
        this.helper,
        this.worker_,
        this.strokeVertexShader_,
        this.strokeFragmentShader_,
        this.strokeAttributes_
      );
    }

    /**
     * @param {import("../../source/Vector.js").VectorSourceEvent} event Event.
     * @private
     */
    handleSourceFeatureAdded_(event) {
      const feature = event.feature;
      this.batch_.addFeature(feature);
    }

    /**
     * @param {import("../../source/Vector.js").VectorSourceEvent} event Event.
     * @private
     */
    handleSourceFeatureChanged_(event) {
      const feature = event.feature;
      this.batch_.changeFeature(feature);
    }

    /**
     * @param {import("../../source/Vector.js").VectorSourceEvent} event Event.
     * @private
     */
    handleSourceFeatureDelete_(event) {
      const feature = event.feature;
      this.batch_.removeFeature(feature);
    }

    /**
     * @private
     */
    handleSourceFeatureClear_() {
      this.batch_.clear();
    }

    /**
     * Render the layer.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @return {HTMLElement} The rendered element.
     */
    renderFrame(frameState) {
      const gl = this.helper.getGL();
      this.preRender(gl, frameState);

      const layer = this.getLayer();
      const vectorSource = layer.getSource();
      const projection = frameState.viewState.projection;
      const multiWorld = vectorSource.getWrapX() && projection.canWrapX();
      const projectionExtent = projection.getExtent();
      const extent = frameState.extent;
      const worldWidth = multiWorld ? getWidth(projectionExtent) : null;
      const endWorld = multiWorld
        ? Math.ceil((extent[2] - projectionExtent[2]) / worldWidth) + 1
        : 1;
      let world = multiWorld
        ? Math.floor((extent[0] - projectionExtent[0]) / worldWidth)
        : 0;

      do {
        this.polygonRenderer_.render(
          this.batch_.polygonBatch,
          this.currentTransform_,
          frameState,
          world * worldWidth
        );
        this.lineStringRenderer_.render(
          this.batch_.lineStringBatch,
          this.currentTransform_,
          frameState,
          world * worldWidth
        );
        this.pointRenderer_.render(
          this.batch_.pointBatch,
          this.currentTransform_,
          frameState,
          world * worldWidth
        );
      } while (++world < endWorld);

      this.helper.finalizeDraw(frameState);

      const canvas = this.helper.getCanvas();
      const layerState = frameState.layerStatesArray[frameState.layerIndex];
      const opacity = layerState.opacity;
      if (opacity !== parseFloat(canvas.style.opacity)) {
        canvas.style.opacity = String(opacity);
      }

      this.postRender(gl, frameState);
      return canvas;
    }

    /**
     * Determine whether renderFrame should be called.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @return {boolean} Layer is ready to be rendered.
     */
    prepareFrameInternal(frameState) {
      const layer = this.getLayer();
      const vectorSource = layer.getSource();
      const viewState = frameState.viewState;
      const viewNotMoving =
        !frameState.viewHints[ViewHint.ANIMATING] &&
        !frameState.viewHints[ViewHint.INTERACTING];
      const extentChanged = !equals$2(this.previousExtent_, frameState.extent);
      const sourceChanged = this.sourceRevision_ < vectorSource.getRevision();

      if (sourceChanged) {
        this.sourceRevision_ = vectorSource.getRevision();
      }

      if (viewNotMoving && (extentChanged || sourceChanged)) {
        const projection = viewState.projection;
        const resolution = viewState.resolution;

        const renderBuffer =
          layer instanceof BaseVectorLayer ? layer.getRenderBuffer() : 0;
        const extent = buffer$1(frameState.extent, renderBuffer * resolution);
        vectorSource.loadFeatures(extent, resolution, projection);

        this.ready = false;
        let remaining = 3;
        const rebuildCb = () => {
          remaining--;
          this.ready = remaining <= 0;
          this.getLayer().changed();
        };

        this.polygonRenderer_.rebuild(
          this.batch_.polygonBatch,
          frameState,
          'Polygon',
          rebuildCb
        );
        this.lineStringRenderer_.rebuild(
          this.batch_.lineStringBatch,
          frameState,
          'LineString',
          rebuildCb
        );
        this.pointRenderer_.rebuild(
          this.batch_.pointBatch,
          frameState,
          'Point',
          rebuildCb
        );
        this.previousExtent_ = frameState.extent.slice();
      }

      this.helper.makeProjectionTransform(frameState, this.currentTransform_);
      this.helper.prepareDraw(frameState);

      return true;
    }

    /**
     * @param {import("../../coordinate.js").Coordinate} coordinate Coordinate.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @param {number} hitTolerance Hit tolerance in pixels.
     * @param {import("../vector.js").FeatureCallback<T>} callback Feature callback.
     * @param {Array<import("../Map.js").HitMatch<T>>} matches The hit detected matches with tolerance.
     * @return {T|undefined} Callback result.
     * @template T
     */
    forEachFeatureAtCoordinate(
      coordinate,
      frameState,
      hitTolerance,
      callback,
      matches
    ) {
      return undefined;
    }

    /**
     * Clean up.
     */
    disposeInternal() {
      this.worker_.terminate();
      this.layer_ = null;
      this.sourceListenKeys_.forEach(function (key) {
        unlistenByKey(key);
      });
      this.sourceListenKeys_ = null;
      super.disposeInternal();
    }
  }

  /**
   * @module ol/render/canvas/Instruction
   */

  /**
   * @enum {number}
   */
  const Instruction = {
    BEGIN_GEOMETRY: 0,
    BEGIN_PATH: 1,
    CIRCLE: 2,
    CLOSE_PATH: 3,
    CUSTOM: 4,
    DRAW_CHARS: 5,
    DRAW_IMAGE: 6,
    END_GEOMETRY: 7,
    FILL: 8,
    MOVE_TO_LINE_TO: 9,
    SET_FILL_STYLE: 10,
    SET_STROKE_STYLE: 11,
    STROKE: 12,
  };

  /**
   * @type {Array<Instruction>}
   */
  const fillInstruction = [Instruction.FILL];

  /**
   * @type {Array<Instruction>}
   */
  const strokeInstruction = [Instruction.STROKE];

  /**
   * @type {Array<Instruction>}
   */
  const beginPathInstruction = [Instruction.BEGIN_PATH];

  /**
   * @type {Array<Instruction>}
   */
  const closePathInstruction = [Instruction.CLOSE_PATH];

  /**
   * @module ol/render/canvas/Builder
   */

  class CanvasBuilder extends VectorContext {
    /**
     * @param {number} tolerance Tolerance.
     * @param {import("../../extent.js").Extent} maxExtent Maximum extent.
     * @param {number} resolution Resolution.
     * @param {number} pixelRatio Pixel ratio.
     */
    constructor(tolerance, maxExtent, resolution, pixelRatio) {
      super();

      /**
       * @protected
       * @type {number}
       */
      this.tolerance = tolerance;

      /**
       * @protected
       * @const
       * @type {import("../../extent.js").Extent}
       */
      this.maxExtent = maxExtent;

      /**
       * @protected
       * @type {number}
       */
      this.pixelRatio = pixelRatio;

      /**
       * @protected
       * @type {number}
       */
      this.maxLineWidth = 0;

      /**
       * @protected
       * @const
       * @type {number}
       */
      this.resolution = resolution;

      /**
       * @private
       * @type {Array<*>}
       */
      this.beginGeometryInstruction1_ = null;

      /**
       * @private
       * @type {Array<*>}
       */
      this.beginGeometryInstruction2_ = null;

      /**
       * @private
       * @type {import("../../extent.js").Extent}
       */
      this.bufferedMaxExtent_ = null;

      /**
       * @protected
       * @type {Array<*>}
       */
      this.instructions = [];

      /**
       * @protected
       * @type {Array<number>}
       */
      this.coordinates = [];

      /**
       * @private
       * @type {import("../../coordinate.js").Coordinate}
       */
      this.tmpCoordinate_ = [];

      /**
       * @protected
       * @type {Array<*>}
       */
      this.hitDetectionInstructions = [];

      /**
       * @protected
       * @type {import("../canvas.js").FillStrokeState}
       */
      this.state = /** @type {import("../canvas.js").FillStrokeState} */ ({});
    }

    /**
     * @protected
     * @param {Array<number>} dashArray Dash array.
     * @return {Array<number>} Dash array with pixel ratio applied
     */
    applyPixelRatio(dashArray) {
      const pixelRatio = this.pixelRatio;
      return pixelRatio == 1
        ? dashArray
        : dashArray.map(function (dash) {
            return dash * pixelRatio;
          });
    }

    /**
     * @param {Array<number>} flatCoordinates Flat coordinates.
     * @param {number} stride Stride.
     * @protected
     * @return {number} My end
     */
    appendFlatPointCoordinates(flatCoordinates, stride) {
      const extent = this.getBufferedMaxExtent();
      const tmpCoord = this.tmpCoordinate_;
      const coordinates = this.coordinates;
      let myEnd = coordinates.length;
      for (let i = 0, ii = flatCoordinates.length; i < ii; i += stride) {
        tmpCoord[0] = flatCoordinates[i];
        tmpCoord[1] = flatCoordinates[i + 1];
        if (containsCoordinate(extent, tmpCoord)) {
          coordinates[myEnd++] = tmpCoord[0];
          coordinates[myEnd++] = tmpCoord[1];
        }
      }
      return myEnd;
    }

    /**
     * @param {Array<number>} flatCoordinates Flat coordinates.
     * @param {number} offset Offset.
     * @param {number} end End.
     * @param {number} stride Stride.
     * @param {boolean} closed Last input coordinate equals first.
     * @param {boolean} skipFirst Skip first coordinate.
     * @protected
     * @return {number} My end.
     */
    appendFlatLineCoordinates(
      flatCoordinates,
      offset,
      end,
      stride,
      closed,
      skipFirst
    ) {
      const coordinates = this.coordinates;
      let myEnd = coordinates.length;
      const extent = this.getBufferedMaxExtent();
      if (skipFirst) {
        offset += stride;
      }
      let lastXCoord = flatCoordinates[offset];
      let lastYCoord = flatCoordinates[offset + 1];
      const nextCoord = this.tmpCoordinate_;
      let skipped = true;

      let i, lastRel, nextRel;
      for (i = offset + stride; i < end; i += stride) {
        nextCoord[0] = flatCoordinates[i];
        nextCoord[1] = flatCoordinates[i + 1];
        nextRel = coordinateRelationship(extent, nextCoord);
        if (nextRel !== lastRel) {
          if (skipped) {
            coordinates[myEnd++] = lastXCoord;
            coordinates[myEnd++] = lastYCoord;
            skipped = false;
          }
          coordinates[myEnd++] = nextCoord[0];
          coordinates[myEnd++] = nextCoord[1];
        } else if (nextRel === Relationship.INTERSECTING) {
          coordinates[myEnd++] = nextCoord[0];
          coordinates[myEnd++] = nextCoord[1];
          skipped = false;
        } else {
          skipped = true;
        }
        lastXCoord = nextCoord[0];
        lastYCoord = nextCoord[1];
        lastRel = nextRel;
      }

      // Last coordinate equals first or only one point to append:
      if ((closed && skipped) || i === offset + stride) {
        coordinates[myEnd++] = lastXCoord;
        coordinates[myEnd++] = lastYCoord;
      }
      return myEnd;
    }

    /**
     * @param {Array<number>} flatCoordinates Flat coordinates.
     * @param {number} offset Offset.
     * @param {Array<number>} ends Ends.
     * @param {number} stride Stride.
     * @param {Array<number>} builderEnds Builder ends.
     * @return {number} Offset.
     */
    drawCustomCoordinates_(flatCoordinates, offset, ends, stride, builderEnds) {
      for (let i = 0, ii = ends.length; i < ii; ++i) {
        const end = ends[i];
        const builderEnd = this.appendFlatLineCoordinates(
          flatCoordinates,
          offset,
          end,
          stride,
          false,
          false
        );
        builderEnds.push(builderEnd);
        offset = end;
      }
      return offset;
    }

    /**
     * @param {import("../../geom/SimpleGeometry.js").default} geometry Geometry.
     * @param {import("../../Feature.js").FeatureLike} feature Feature.
     * @param {Function} renderer Renderer.
     * @param {Function} hitDetectionRenderer Renderer.
     */
    drawCustom(geometry, feature, renderer, hitDetectionRenderer) {
      this.beginGeometry(geometry, feature);

      const type = geometry.getType();
      const stride = geometry.getStride();
      const builderBegin = this.coordinates.length;

      let flatCoordinates, builderEnd, builderEnds, builderEndss;
      let offset;

      switch (type) {
        case 'MultiPolygon':
          flatCoordinates =
            /** @type {import("../../geom/MultiPolygon.js").default} */ (
              geometry
            ).getOrientedFlatCoordinates();
          builderEndss = [];
          const endss =
            /** @type {import("../../geom/MultiPolygon.js").default} */ (
              geometry
            ).getEndss();
          offset = 0;
          for (let i = 0, ii = endss.length; i < ii; ++i) {
            const myEnds = [];
            offset = this.drawCustomCoordinates_(
              flatCoordinates,
              offset,
              endss[i],
              stride,
              myEnds
            );
            builderEndss.push(myEnds);
          }
          this.instructions.push([
            Instruction.CUSTOM,
            builderBegin,
            builderEndss,
            geometry,
            renderer,
            inflateMultiCoordinatesArray,
          ]);
          this.hitDetectionInstructions.push([
            Instruction.CUSTOM,
            builderBegin,
            builderEndss,
            geometry,
            hitDetectionRenderer || renderer,
            inflateMultiCoordinatesArray,
          ]);
          break;
        case 'Polygon':
        case 'MultiLineString':
          builderEnds = [];
          flatCoordinates =
            type == 'Polygon'
              ? /** @type {import("../../geom/Polygon.js").default} */ (
                  geometry
                ).getOrientedFlatCoordinates()
              : geometry.getFlatCoordinates();
          offset = this.drawCustomCoordinates_(
            flatCoordinates,
            0,
            /** @type {import("../../geom/Polygon.js").default|import("../../geom/MultiLineString.js").default} */ (
              geometry
            ).getEnds(),
            stride,
            builderEnds
          );
          this.instructions.push([
            Instruction.CUSTOM,
            builderBegin,
            builderEnds,
            geometry,
            renderer,
            inflateCoordinatesArray,
          ]);
          this.hitDetectionInstructions.push([
            Instruction.CUSTOM,
            builderBegin,
            builderEnds,
            geometry,
            hitDetectionRenderer || renderer,
            inflateCoordinatesArray,
          ]);
          break;
        case 'LineString':
        case 'Circle':
          flatCoordinates = geometry.getFlatCoordinates();
          builderEnd = this.appendFlatLineCoordinates(
            flatCoordinates,
            0,
            flatCoordinates.length,
            stride,
            false,
            false
          );
          this.instructions.push([
            Instruction.CUSTOM,
            builderBegin,
            builderEnd,
            geometry,
            renderer,
            inflateCoordinates,
          ]);
          this.hitDetectionInstructions.push([
            Instruction.CUSTOM,
            builderBegin,
            builderEnd,
            geometry,
            hitDetectionRenderer || renderer,
            inflateCoordinates,
          ]);
          break;
        case 'MultiPoint':
          flatCoordinates = geometry.getFlatCoordinates();
          builderEnd = this.appendFlatPointCoordinates(flatCoordinates, stride);

          if (builderEnd > builderBegin) {
            this.instructions.push([
              Instruction.CUSTOM,
              builderBegin,
              builderEnd,
              geometry,
              renderer,
              inflateCoordinates,
            ]);
            this.hitDetectionInstructions.push([
              Instruction.CUSTOM,
              builderBegin,
              builderEnd,
              geometry,
              hitDetectionRenderer || renderer,
              inflateCoordinates,
            ]);
          }
          break;
        case 'Point':
          flatCoordinates = geometry.getFlatCoordinates();
          this.coordinates.push(flatCoordinates[0], flatCoordinates[1]);
          builderEnd = this.coordinates.length;

          this.instructions.push([
            Instruction.CUSTOM,
            builderBegin,
            builderEnd,
            geometry,
            renderer,
          ]);
          this.hitDetectionInstructions.push([
            Instruction.CUSTOM,
            builderBegin,
            builderEnd,
            geometry,
            hitDetectionRenderer || renderer,
          ]);
          break;
      }
      this.endGeometry(feature);
    }

    /**
     * @protected
     * @param {import("../../geom/Geometry").default|import("../Feature.js").default} geometry The geometry.
     * @param {import("../../Feature.js").FeatureLike} feature Feature.
     */
    beginGeometry(geometry, feature) {
      this.beginGeometryInstruction1_ = [
        Instruction.BEGIN_GEOMETRY,
        feature,
        0,
        geometry,
      ];
      this.instructions.push(this.beginGeometryInstruction1_);
      this.beginGeometryInstruction2_ = [
        Instruction.BEGIN_GEOMETRY,
        feature,
        0,
        geometry,
      ];
      this.hitDetectionInstructions.push(this.beginGeometryInstruction2_);
    }

    /**
     * @return {import("../canvas.js").SerializableInstructions} the serializable instructions.
     */
    finish() {
      return {
        instructions: this.instructions,
        hitDetectionInstructions: this.hitDetectionInstructions,
        coordinates: this.coordinates,
      };
    }

    /**
     * Reverse the hit detection instructions.
     */
    reverseHitDetectionInstructions() {
      const hitDetectionInstructions = this.hitDetectionInstructions;
      // step 1 - reverse array
      hitDetectionInstructions.reverse();
      // step 2 - reverse instructions within geometry blocks
      let i;
      const n = hitDetectionInstructions.length;
      let instruction;
      let type;
      let begin = -1;
      for (i = 0; i < n; ++i) {
        instruction = hitDetectionInstructions[i];
        type = /** @type {import("./Instruction.js").default} */ (instruction[0]);
        if (type == Instruction.END_GEOMETRY) {
          begin = i;
        } else if (type == Instruction.BEGIN_GEOMETRY) {
          instruction[2] = i;
          reverseSubArray(this.hitDetectionInstructions, begin, i);
          begin = -1;
        }
      }
    }

    /**
     * @param {import("../../style/Fill.js").default} fillStyle Fill style.
     * @param {import("../../style/Stroke.js").default} strokeStyle Stroke style.
     */
    setFillStrokeStyle(fillStyle, strokeStyle) {
      const state = this.state;
      if (fillStyle) {
        const fillStyleColor = fillStyle.getColor();
        state.fillStyle = asColorLike(
          fillStyleColor ? fillStyleColor : defaultFillStyle
        );
      } else {
        state.fillStyle = undefined;
      }
      if (strokeStyle) {
        const strokeStyleColor = strokeStyle.getColor();
        state.strokeStyle = asColorLike(
          strokeStyleColor ? strokeStyleColor : defaultStrokeStyle
        );
        const strokeStyleLineCap = strokeStyle.getLineCap();
        state.lineCap =
          strokeStyleLineCap !== undefined ? strokeStyleLineCap : defaultLineCap;
        const strokeStyleLineDash = strokeStyle.getLineDash();
        state.lineDash = strokeStyleLineDash
          ? strokeStyleLineDash.slice()
          : defaultLineDash;
        const strokeStyleLineDashOffset = strokeStyle.getLineDashOffset();
        state.lineDashOffset = strokeStyleLineDashOffset
          ? strokeStyleLineDashOffset
          : defaultLineDashOffset;
        const strokeStyleLineJoin = strokeStyle.getLineJoin();
        state.lineJoin =
          strokeStyleLineJoin !== undefined
            ? strokeStyleLineJoin
            : defaultLineJoin;
        const strokeStyleWidth = strokeStyle.getWidth();
        state.lineWidth =
          strokeStyleWidth !== undefined ? strokeStyleWidth : defaultLineWidth;
        const strokeStyleMiterLimit = strokeStyle.getMiterLimit();
        state.miterLimit =
          strokeStyleMiterLimit !== undefined
            ? strokeStyleMiterLimit
            : defaultMiterLimit;

        if (state.lineWidth > this.maxLineWidth) {
          this.maxLineWidth = state.lineWidth;
          // invalidate the buffered max extent cache
          this.bufferedMaxExtent_ = null;
        }
      } else {
        state.strokeStyle = undefined;
        state.lineCap = undefined;
        state.lineDash = null;
        state.lineDashOffset = undefined;
        state.lineJoin = undefined;
        state.lineWidth = undefined;
        state.miterLimit = undefined;
      }
    }

    /**
     * @param {import("../canvas.js").FillStrokeState} state State.
     * @return {Array<*>} Fill instruction.
     */
    createFill(state) {
      const fillStyle = state.fillStyle;
      /** @type {Array<*>} */
      const fillInstruction = [Instruction.SET_FILL_STYLE, fillStyle];
      if (typeof fillStyle !== 'string') {
        // Fill is a pattern or gradient - align it!
        fillInstruction.push(true);
      }
      return fillInstruction;
    }

    /**
     * @param {import("../canvas.js").FillStrokeState} state State.
     */
    applyStroke(state) {
      this.instructions.push(this.createStroke(state));
    }

    /**
     * @param {import("../canvas.js").FillStrokeState} state State.
     * @return {Array<*>} Stroke instruction.
     */
    createStroke(state) {
      return [
        Instruction.SET_STROKE_STYLE,
        state.strokeStyle,
        state.lineWidth * this.pixelRatio,
        state.lineCap,
        state.lineJoin,
        state.miterLimit,
        this.applyPixelRatio(state.lineDash),
        state.lineDashOffset * this.pixelRatio,
      ];
    }

    /**
     * @param {import("../canvas.js").FillStrokeState} state State.
     * @param {function(this:CanvasBuilder, import("../canvas.js").FillStrokeState):Array<*>} createFill Create fill.
     */
    updateFillStyle(state, createFill) {
      const fillStyle = state.fillStyle;
      if (typeof fillStyle !== 'string' || state.currentFillStyle != fillStyle) {
        if (fillStyle !== undefined) {
          this.instructions.push(createFill.call(this, state));
        }
        state.currentFillStyle = fillStyle;
      }
    }

    /**
     * @param {import("../canvas.js").FillStrokeState} state State.
     * @param {function(this:CanvasBuilder, import("../canvas.js").FillStrokeState): void} applyStroke Apply stroke.
     */
    updateStrokeStyle(state, applyStroke) {
      const strokeStyle = state.strokeStyle;
      const lineCap = state.lineCap;
      const lineDash = state.lineDash;
      const lineDashOffset = state.lineDashOffset;
      const lineJoin = state.lineJoin;
      const lineWidth = state.lineWidth;
      const miterLimit = state.miterLimit;
      if (
        state.currentStrokeStyle != strokeStyle ||
        state.currentLineCap != lineCap ||
        (lineDash != state.currentLineDash &&
          !equals$3(state.currentLineDash, lineDash)) ||
        state.currentLineDashOffset != lineDashOffset ||
        state.currentLineJoin != lineJoin ||
        state.currentLineWidth != lineWidth ||
        state.currentMiterLimit != miterLimit
      ) {
        if (strokeStyle !== undefined) {
          applyStroke.call(this, state);
        }
        state.currentStrokeStyle = strokeStyle;
        state.currentLineCap = lineCap;
        state.currentLineDash = lineDash;
        state.currentLineDashOffset = lineDashOffset;
        state.currentLineJoin = lineJoin;
        state.currentLineWidth = lineWidth;
        state.currentMiterLimit = miterLimit;
      }
    }

    /**
     * @param {import("../../Feature.js").FeatureLike} feature Feature.
     */
    endGeometry(feature) {
      this.beginGeometryInstruction1_[2] = this.instructions.length;
      this.beginGeometryInstruction1_ = null;
      this.beginGeometryInstruction2_[2] = this.hitDetectionInstructions.length;
      this.beginGeometryInstruction2_ = null;
      const endGeometryInstruction = [Instruction.END_GEOMETRY, feature];
      this.instructions.push(endGeometryInstruction);
      this.hitDetectionInstructions.push(endGeometryInstruction);
    }

    /**
     * Get the buffered rendering extent.  Rendering will be clipped to the extent
     * provided to the constructor.  To account for symbolizers that may intersect
     * this extent, we calculate a buffered extent (e.g. based on stroke width).
     * @return {import("../../extent.js").Extent} The buffered rendering extent.
     * @protected
     */
    getBufferedMaxExtent() {
      if (!this.bufferedMaxExtent_) {
        this.bufferedMaxExtent_ = clone(this.maxExtent);
        if (this.maxLineWidth > 0) {
          const width = (this.resolution * (this.maxLineWidth + 1)) / 2;
          buffer$1(this.bufferedMaxExtent_, width, this.bufferedMaxExtent_);
        }
      }
      return this.bufferedMaxExtent_;
    }
  }

  /**
   * @module ol/render/canvas/ImageBuilder
   */

  class CanvasImageBuilder extends CanvasBuilder {
    /**
     * @param {number} tolerance Tolerance.
     * @param {import("../../extent.js").Extent} maxExtent Maximum extent.
     * @param {number} resolution Resolution.
     * @param {number} pixelRatio Pixel ratio.
     */
    constructor(tolerance, maxExtent, resolution, pixelRatio) {
      super(tolerance, maxExtent, resolution, pixelRatio);

      /**
       * @private
       * @type {HTMLCanvasElement|HTMLVideoElement|HTMLImageElement}
       */
      this.hitDetectionImage_ = null;

      /**
       * @private
       * @type {HTMLCanvasElement|HTMLVideoElement|HTMLImageElement}
       */
      this.image_ = null;

      /**
       * @private
       * @type {number|undefined}
       */
      this.imagePixelRatio_ = undefined;

      /**
       * @private
       * @type {number|undefined}
       */
      this.anchorX_ = undefined;

      /**
       * @private
       * @type {number|undefined}
       */
      this.anchorY_ = undefined;

      /**
       * @private
       * @type {number|undefined}
       */
      this.height_ = undefined;

      /**
       * @private
       * @type {number|undefined}
       */
      this.opacity_ = undefined;

      /**
       * @private
       * @type {number|undefined}
       */
      this.originX_ = undefined;

      /**
       * @private
       * @type {number|undefined}
       */
      this.originY_ = undefined;

      /**
       * @private
       * @type {boolean|undefined}
       */
      this.rotateWithView_ = undefined;

      /**
       * @private
       * @type {number|undefined}
       */
      this.rotation_ = undefined;

      /**
       * @private
       * @type {import("../../size.js").Size|undefined}
       */
      this.scale_ = undefined;

      /**
       * @private
       * @type {number|undefined}
       */
      this.width_ = undefined;

      /**
       * @private
       * @type {"declutter"|"obstacle"|"none"|undefined}
       */
      this.declutterMode_ = undefined;

      /**
       * Data shared with a text builder for combined decluttering.
       * @private
       * @type {import("../canvas.js").DeclutterImageWithText}
       */
      this.declutterImageWithText_ = undefined;
    }

    /**
     * @param {import("../../geom/Point.js").default|import("../Feature.js").default} pointGeometry Point geometry.
     * @param {import("../../Feature.js").FeatureLike} feature Feature.
     */
    drawPoint(pointGeometry, feature) {
      if (!this.image_) {
        return;
      }
      this.beginGeometry(pointGeometry, feature);
      const flatCoordinates = pointGeometry.getFlatCoordinates();
      const stride = pointGeometry.getStride();
      const myBegin = this.coordinates.length;
      const myEnd = this.appendFlatPointCoordinates(flatCoordinates, stride);
      this.instructions.push([
        Instruction.DRAW_IMAGE,
        myBegin,
        myEnd,
        this.image_,
        // Remaining arguments to DRAW_IMAGE are in alphabetical order
        this.anchorX_ * this.imagePixelRatio_,
        this.anchorY_ * this.imagePixelRatio_,
        Math.ceil(this.height_ * this.imagePixelRatio_),
        this.opacity_,
        this.originX_ * this.imagePixelRatio_,
        this.originY_ * this.imagePixelRatio_,
        this.rotateWithView_,
        this.rotation_,
        [
          (this.scale_[0] * this.pixelRatio) / this.imagePixelRatio_,
          (this.scale_[1] * this.pixelRatio) / this.imagePixelRatio_,
        ],
        Math.ceil(this.width_ * this.imagePixelRatio_),
        this.declutterMode_,
        this.declutterImageWithText_,
      ]);
      this.hitDetectionInstructions.push([
        Instruction.DRAW_IMAGE,
        myBegin,
        myEnd,
        this.hitDetectionImage_,
        // Remaining arguments to DRAW_IMAGE are in alphabetical order
        this.anchorX_,
        this.anchorY_,
        this.height_,
        this.opacity_,
        this.originX_,
        this.originY_,
        this.rotateWithView_,
        this.rotation_,
        this.scale_,
        this.width_,
        this.declutterMode_,
        this.declutterImageWithText_,
      ]);
      this.endGeometry(feature);
    }

    /**
     * @param {import("../../geom/MultiPoint.js").default|import("../Feature.js").default} multiPointGeometry MultiPoint geometry.
     * @param {import("../../Feature.js").FeatureLike} feature Feature.
     */
    drawMultiPoint(multiPointGeometry, feature) {
      if (!this.image_) {
        return;
      }
      this.beginGeometry(multiPointGeometry, feature);
      const flatCoordinates = multiPointGeometry.getFlatCoordinates();
      const stride = multiPointGeometry.getStride();
      const myBegin = this.coordinates.length;
      const myEnd = this.appendFlatPointCoordinates(flatCoordinates, stride);
      this.instructions.push([
        Instruction.DRAW_IMAGE,
        myBegin,
        myEnd,
        this.image_,
        // Remaining arguments to DRAW_IMAGE are in alphabetical order
        this.anchorX_ * this.imagePixelRatio_,
        this.anchorY_ * this.imagePixelRatio_,
        Math.ceil(this.height_ * this.imagePixelRatio_),
        this.opacity_,
        this.originX_ * this.imagePixelRatio_,
        this.originY_ * this.imagePixelRatio_,
        this.rotateWithView_,
        this.rotation_,
        [
          (this.scale_[0] * this.pixelRatio) / this.imagePixelRatio_,
          (this.scale_[1] * this.pixelRatio) / this.imagePixelRatio_,
        ],
        Math.ceil(this.width_ * this.imagePixelRatio_),
        this.declutterMode_,
        this.declutterImageWithText_,
      ]);
      this.hitDetectionInstructions.push([
        Instruction.DRAW_IMAGE,
        myBegin,
        myEnd,
        this.hitDetectionImage_,
        // Remaining arguments to DRAW_IMAGE are in alphabetical order
        this.anchorX_,
        this.anchorY_,
        this.height_,
        this.opacity_,
        this.originX_,
        this.originY_,
        this.rotateWithView_,
        this.rotation_,
        this.scale_,
        this.width_,
        this.declutterMode_,
        this.declutterImageWithText_,
      ]);
      this.endGeometry(feature);
    }

    /**
     * @return {import("../canvas.js").SerializableInstructions} the serializable instructions.
     */
    finish() {
      this.reverseHitDetectionInstructions();
      // FIXME this doesn't really protect us against further calls to draw*Geometry
      this.anchorX_ = undefined;
      this.anchorY_ = undefined;
      this.hitDetectionImage_ = null;
      this.image_ = null;
      this.imagePixelRatio_ = undefined;
      this.height_ = undefined;
      this.scale_ = undefined;
      this.opacity_ = undefined;
      this.originX_ = undefined;
      this.originY_ = undefined;
      this.rotateWithView_ = undefined;
      this.rotation_ = undefined;
      this.width_ = undefined;
      return super.finish();
    }

    /**
     * @param {import("../../style/Image.js").default} imageStyle Image style.
     * @param {Object} [sharedData] Shared data.
     */
    setImageStyle(imageStyle, sharedData) {
      const anchor = imageStyle.getAnchor();
      const size = imageStyle.getSize();
      const origin = imageStyle.getOrigin();
      this.imagePixelRatio_ = imageStyle.getPixelRatio(this.pixelRatio);
      this.anchorX_ = anchor[0];
      this.anchorY_ = anchor[1];
      this.hitDetectionImage_ = imageStyle.getHitDetectionImage();
      this.image_ = imageStyle.getImage(this.pixelRatio);
      this.height_ = size[1];
      this.opacity_ = imageStyle.getOpacity();
      this.originX_ = origin[0];
      this.originY_ = origin[1];
      this.rotateWithView_ = imageStyle.getRotateWithView();
      this.rotation_ = imageStyle.getRotation();
      this.scale_ = imageStyle.getScaleArray();
      this.width_ = size[0];
      this.declutterMode_ = imageStyle.getDeclutterMode();
      this.declutterImageWithText_ = sharedData;
    }
  }

  /**
   * @module ol/render/canvas/LineStringBuilder
   */

  class CanvasLineStringBuilder extends CanvasBuilder {
    /**
     * @param {number} tolerance Tolerance.
     * @param {import("../../extent.js").Extent} maxExtent Maximum extent.
     * @param {number} resolution Resolution.
     * @param {number} pixelRatio Pixel ratio.
     */
    constructor(tolerance, maxExtent, resolution, pixelRatio) {
      super(tolerance, maxExtent, resolution, pixelRatio);
    }

    /**
     * @param {Array<number>} flatCoordinates Flat coordinates.
     * @param {number} offset Offset.
     * @param {number} end End.
     * @param {number} stride Stride.
     * @private
     * @return {number} end.
     */
    drawFlatCoordinates_(flatCoordinates, offset, end, stride) {
      const myBegin = this.coordinates.length;
      const myEnd = this.appendFlatLineCoordinates(
        flatCoordinates,
        offset,
        end,
        stride,
        false,
        false
      );
      const moveToLineToInstruction = [
        Instruction.MOVE_TO_LINE_TO,
        myBegin,
        myEnd,
      ];
      this.instructions.push(moveToLineToInstruction);
      this.hitDetectionInstructions.push(moveToLineToInstruction);
      return end;
    }

    /**
     * @param {import("../../geom/LineString.js").default|import("../Feature.js").default} lineStringGeometry Line string geometry.
     * @param {import("../../Feature.js").FeatureLike} feature Feature.
     */
    drawLineString(lineStringGeometry, feature) {
      const state = this.state;
      const strokeStyle = state.strokeStyle;
      const lineWidth = state.lineWidth;
      if (strokeStyle === undefined || lineWidth === undefined) {
        return;
      }
      this.updateStrokeStyle(state, this.applyStroke);
      this.beginGeometry(lineStringGeometry, feature);
      this.hitDetectionInstructions.push(
        [
          Instruction.SET_STROKE_STYLE,
          state.strokeStyle,
          state.lineWidth,
          state.lineCap,
          state.lineJoin,
          state.miterLimit,
          defaultLineDash,
          defaultLineDashOffset,
        ],
        beginPathInstruction
      );
      const flatCoordinates = lineStringGeometry.getFlatCoordinates();
      const stride = lineStringGeometry.getStride();
      this.drawFlatCoordinates_(
        flatCoordinates,
        0,
        flatCoordinates.length,
        stride
      );
      this.hitDetectionInstructions.push(strokeInstruction);
      this.endGeometry(feature);
    }

    /**
     * @param {import("../../geom/MultiLineString.js").default|import("../Feature.js").default} multiLineStringGeometry MultiLineString geometry.
     * @param {import("../../Feature.js").FeatureLike} feature Feature.
     */
    drawMultiLineString(multiLineStringGeometry, feature) {
      const state = this.state;
      const strokeStyle = state.strokeStyle;
      const lineWidth = state.lineWidth;
      if (strokeStyle === undefined || lineWidth === undefined) {
        return;
      }
      this.updateStrokeStyle(state, this.applyStroke);
      this.beginGeometry(multiLineStringGeometry, feature);
      this.hitDetectionInstructions.push(
        [
          Instruction.SET_STROKE_STYLE,
          state.strokeStyle,
          state.lineWidth,
          state.lineCap,
          state.lineJoin,
          state.miterLimit,
          state.lineDash,
          state.lineDashOffset,
        ],
        beginPathInstruction
      );
      const ends = multiLineStringGeometry.getEnds();
      const flatCoordinates = multiLineStringGeometry.getFlatCoordinates();
      const stride = multiLineStringGeometry.getStride();
      let offset = 0;
      for (let i = 0, ii = ends.length; i < ii; ++i) {
        offset = this.drawFlatCoordinates_(
          flatCoordinates,
          offset,
          /** @type {number} */ (ends[i]),
          stride
        );
      }
      this.hitDetectionInstructions.push(strokeInstruction);
      this.endGeometry(feature);
    }

    /**
     * @return {import("../canvas.js").SerializableInstructions} the serializable instructions.
     */
    finish() {
      const state = this.state;
      if (
        state.lastStroke != undefined &&
        state.lastStroke != this.coordinates.length
      ) {
        this.instructions.push(strokeInstruction);
      }
      this.reverseHitDetectionInstructions();
      this.state = null;
      return super.finish();
    }

    /**
     * @param {import("../canvas.js").FillStrokeState} state State.
     */
    applyStroke(state) {
      if (
        state.lastStroke != undefined &&
        state.lastStroke != this.coordinates.length
      ) {
        this.instructions.push(strokeInstruction);
        state.lastStroke = this.coordinates.length;
      }
      state.lastStroke = 0;
      super.applyStroke(state);
      this.instructions.push(beginPathInstruction);
    }
  }

  /**
   * @module ol/render/canvas/PolygonBuilder
   */

  class CanvasPolygonBuilder extends CanvasBuilder {
    /**
     * @param {number} tolerance Tolerance.
     * @param {import("../../extent.js").Extent} maxExtent Maximum extent.
     * @param {number} resolution Resolution.
     * @param {number} pixelRatio Pixel ratio.
     */
    constructor(tolerance, maxExtent, resolution, pixelRatio) {
      super(tolerance, maxExtent, resolution, pixelRatio);
    }

    /**
     * @param {Array<number>} flatCoordinates Flat coordinates.
     * @param {number} offset Offset.
     * @param {Array<number>} ends Ends.
     * @param {number} stride Stride.
     * @private
     * @return {number} End.
     */
    drawFlatCoordinatess_(flatCoordinates, offset, ends, stride) {
      const state = this.state;
      const fill = state.fillStyle !== undefined;
      const stroke = state.strokeStyle !== undefined;
      const numEnds = ends.length;
      this.instructions.push(beginPathInstruction);
      this.hitDetectionInstructions.push(beginPathInstruction);
      for (let i = 0; i < numEnds; ++i) {
        const end = ends[i];
        const myBegin = this.coordinates.length;
        const myEnd = this.appendFlatLineCoordinates(
          flatCoordinates,
          offset,
          end,
          stride,
          true,
          !stroke
        );
        const moveToLineToInstruction = [
          Instruction.MOVE_TO_LINE_TO,
          myBegin,
          myEnd,
        ];
        this.instructions.push(moveToLineToInstruction);
        this.hitDetectionInstructions.push(moveToLineToInstruction);
        if (stroke) {
          // Performance optimization: only call closePath() when we have a stroke.
          // Otherwise the ring is closed already (see appendFlatLineCoordinates above).
          this.instructions.push(closePathInstruction);
          this.hitDetectionInstructions.push(closePathInstruction);
        }
        offset = end;
      }
      if (fill) {
        this.instructions.push(fillInstruction);
        this.hitDetectionInstructions.push(fillInstruction);
      }
      if (stroke) {
        this.instructions.push(strokeInstruction);
        this.hitDetectionInstructions.push(strokeInstruction);
      }
      return offset;
    }

    /**
     * @param {import("../../geom/Circle.js").default} circleGeometry Circle geometry.
     * @param {import("../../Feature.js").default} feature Feature.
     */
    drawCircle(circleGeometry, feature) {
      const state = this.state;
      const fillStyle = state.fillStyle;
      const strokeStyle = state.strokeStyle;
      if (fillStyle === undefined && strokeStyle === undefined) {
        return;
      }
      this.setFillStrokeStyles_();
      this.beginGeometry(circleGeometry, feature);
      if (state.fillStyle !== undefined) {
        this.hitDetectionInstructions.push([
          Instruction.SET_FILL_STYLE,
          defaultFillStyle,
        ]);
      }
      if (state.strokeStyle !== undefined) {
        this.hitDetectionInstructions.push([
          Instruction.SET_STROKE_STYLE,
          state.strokeStyle,
          state.lineWidth,
          state.lineCap,
          state.lineJoin,
          state.miterLimit,
          state.lineDash,
          state.lineDashOffset,
        ]);
      }
      const flatCoordinates = circleGeometry.getFlatCoordinates();
      const stride = circleGeometry.getStride();
      const myBegin = this.coordinates.length;
      this.appendFlatLineCoordinates(
        flatCoordinates,
        0,
        flatCoordinates.length,
        stride,
        false,
        false
      );
      const circleInstruction = [Instruction.CIRCLE, myBegin];
      this.instructions.push(beginPathInstruction, circleInstruction);
      this.hitDetectionInstructions.push(beginPathInstruction, circleInstruction);
      if (state.fillStyle !== undefined) {
        this.instructions.push(fillInstruction);
        this.hitDetectionInstructions.push(fillInstruction);
      }
      if (state.strokeStyle !== undefined) {
        this.instructions.push(strokeInstruction);
        this.hitDetectionInstructions.push(strokeInstruction);
      }
      this.endGeometry(feature);
    }

    /**
     * @param {import("../../geom/Polygon.js").default|import("../Feature.js").default} polygonGeometry Polygon geometry.
     * @param {import("../../Feature.js").FeatureLike} feature Feature.
     */
    drawPolygon(polygonGeometry, feature) {
      const state = this.state;
      const fillStyle = state.fillStyle;
      const strokeStyle = state.strokeStyle;
      if (fillStyle === undefined && strokeStyle === undefined) {
        return;
      }
      this.setFillStrokeStyles_();
      this.beginGeometry(polygonGeometry, feature);
      if (state.fillStyle !== undefined) {
        this.hitDetectionInstructions.push([
          Instruction.SET_FILL_STYLE,
          defaultFillStyle,
        ]);
      }
      if (state.strokeStyle !== undefined) {
        this.hitDetectionInstructions.push([
          Instruction.SET_STROKE_STYLE,
          state.strokeStyle,
          state.lineWidth,
          state.lineCap,
          state.lineJoin,
          state.miterLimit,
          state.lineDash,
          state.lineDashOffset,
        ]);
      }
      const ends = polygonGeometry.getEnds();
      const flatCoordinates = polygonGeometry.getOrientedFlatCoordinates();
      const stride = polygonGeometry.getStride();
      this.drawFlatCoordinatess_(
        flatCoordinates,
        0,
        /** @type {Array<number>} */ (ends),
        stride
      );
      this.endGeometry(feature);
    }

    /**
     * @param {import("../../geom/MultiPolygon.js").default} multiPolygonGeometry MultiPolygon geometry.
     * @param {import("../../Feature.js").FeatureLike} feature Feature.
     */
    drawMultiPolygon(multiPolygonGeometry, feature) {
      const state = this.state;
      const fillStyle = state.fillStyle;
      const strokeStyle = state.strokeStyle;
      if (fillStyle === undefined && strokeStyle === undefined) {
        return;
      }
      this.setFillStrokeStyles_();
      this.beginGeometry(multiPolygonGeometry, feature);
      if (state.fillStyle !== undefined) {
        this.hitDetectionInstructions.push([
          Instruction.SET_FILL_STYLE,
          defaultFillStyle,
        ]);
      }
      if (state.strokeStyle !== undefined) {
        this.hitDetectionInstructions.push([
          Instruction.SET_STROKE_STYLE,
          state.strokeStyle,
          state.lineWidth,
          state.lineCap,
          state.lineJoin,
          state.miterLimit,
          state.lineDash,
          state.lineDashOffset,
        ]);
      }
      const endss = multiPolygonGeometry.getEndss();
      const flatCoordinates = multiPolygonGeometry.getOrientedFlatCoordinates();
      const stride = multiPolygonGeometry.getStride();
      let offset = 0;
      for (let i = 0, ii = endss.length; i < ii; ++i) {
        offset = this.drawFlatCoordinatess_(
          flatCoordinates,
          offset,
          endss[i],
          stride
        );
      }
      this.endGeometry(feature);
    }

    /**
     * @return {import("../canvas.js").SerializableInstructions} the serializable instructions.
     */
    finish() {
      this.reverseHitDetectionInstructions();
      this.state = null;
      // We want to preserve topology when drawing polygons.  Polygons are
      // simplified using quantization and point elimination. However, we might
      // have received a mix of quantized and non-quantized geometries, so ensure
      // that all are quantized by quantizing all coordinates in the batch.
      const tolerance = this.tolerance;
      if (tolerance !== 0) {
        const coordinates = this.coordinates;
        for (let i = 0, ii = coordinates.length; i < ii; ++i) {
          coordinates[i] = snap(coordinates[i], tolerance);
        }
      }
      return super.finish();
    }

    /**
     * @private
     */
    setFillStrokeStyles_() {
      const state = this.state;
      const fillStyle = state.fillStyle;
      if (fillStyle !== undefined) {
        this.updateFillStyle(state, this.createFill);
      }
      if (state.strokeStyle !== undefined) {
        this.updateStrokeStyle(state, this.applyStroke);
      }
    }
  }

  /**
   * @module ol/geom/flat/straightchunk
   */

  /**
   * @param {number} maxAngle Maximum acceptable angle delta between segments.
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   * @return {Array<number>} Start and end of the first suitable chunk of the
   * given `flatCoordinates`.
   */
  function matchingChunk(maxAngle, flatCoordinates, offset, end, stride) {
    let chunkStart = offset;
    let chunkEnd = offset;
    let chunkM = 0;
    let m = 0;
    let start = offset;
    let acos, i, m12, m23, x1, y1, x12, y12, x23, y23;
    for (i = offset; i < end; i += stride) {
      const x2 = flatCoordinates[i];
      const y2 = flatCoordinates[i + 1];
      if (x1 !== undefined) {
        x23 = x2 - x1;
        y23 = y2 - y1;
        m23 = Math.sqrt(x23 * x23 + y23 * y23);
        if (x12 !== undefined) {
          m += m12;
          acos = Math.acos((x12 * x23 + y12 * y23) / (m12 * m23));
          if (acos > maxAngle) {
            if (m > chunkM) {
              chunkM = m;
              chunkStart = start;
              chunkEnd = i;
            }
            m = 0;
            start = i - stride;
          }
        }
        m12 = m23;
        x12 = x23;
        y12 = y23;
      }
      x1 = x2;
      y1 = y2;
    }
    m += m23;
    return m > chunkM ? [start, i] : [chunkStart, chunkEnd];
  }

  /**
   * @module ol/render/canvas/TextBuilder
   */
  /**
   * @const
   * @enum {number}
   */
  const TEXT_ALIGN = {
    'left': 0,
    'end': 0,
    'center': 0.5,
    'right': 1,
    'start': 1,
    'top': 0,
    'middle': 0.5,
    'hanging': 0.2,
    'alphabetic': 0.8,
    'ideographic': 0.8,
    'bottom': 1,
  };

  class CanvasTextBuilder extends CanvasBuilder {
    /**
     * @param {number} tolerance Tolerance.
     * @param {import("../../extent.js").Extent} maxExtent Maximum extent.
     * @param {number} resolution Resolution.
     * @param {number} pixelRatio Pixel ratio.
     */
    constructor(tolerance, maxExtent, resolution, pixelRatio) {
      super(tolerance, maxExtent, resolution, pixelRatio);

      /**
       * @private
       * @type {Array<HTMLCanvasElement>}
       */
      this.labels_ = null;

      /**
       * @private
       * @type {string|Array<string>}
       */
      this.text_ = '';

      /**
       * @private
       * @type {number}
       */
      this.textOffsetX_ = 0;

      /**
       * @private
       * @type {number}
       */
      this.textOffsetY_ = 0;

      /**
       * @private
       * @type {boolean|undefined}
       */
      this.textRotateWithView_ = undefined;

      /**
       * @private
       * @type {number}
       */
      this.textRotation_ = 0;

      /**
       * @private
       * @type {?import("../canvas.js").FillState}
       */
      this.textFillState_ = null;

      /**
       * @type {!Object<string, import("../canvas.js").FillState>}
       */
      this.fillStates = {};

      /**
       * @private
       * @type {?import("../canvas.js").StrokeState}
       */
      this.textStrokeState_ = null;

      /**
       * @type {!Object<string, import("../canvas.js").StrokeState>}
       */
      this.strokeStates = {};

      /**
       * @private
       * @type {import("../canvas.js").TextState}
       */
      this.textState_ = /** @type {import("../canvas.js").TextState} */ ({});

      /**
       * @type {!Object<string, import("../canvas.js").TextState>}
       */
      this.textStates = {};

      /**
       * @private
       * @type {string}
       */
      this.textKey_ = '';

      /**
       * @private
       * @type {string}
       */
      this.fillKey_ = '';

      /**
       * @private
       * @type {string}
       */
      this.strokeKey_ = '';

      /**
       * Data shared with an image builder for combined decluttering.
       * @private
       * @type {import("../canvas.js").DeclutterImageWithText}
       */
      this.declutterImageWithText_ = undefined;
    }

    /**
     * @return {import("../canvas.js").SerializableInstructions} the serializable instructions.
     */
    finish() {
      const instructions = super.finish();
      instructions.textStates = this.textStates;
      instructions.fillStates = this.fillStates;
      instructions.strokeStates = this.strokeStates;
      return instructions;
    }

    /**
     * @param {import("../../geom/SimpleGeometry.js").default|import("../Feature.js").default} geometry Geometry.
     * @param {import("../../Feature.js").FeatureLike} feature Feature.
     */
    drawText(geometry, feature) {
      const fillState = this.textFillState_;
      const strokeState = this.textStrokeState_;
      const textState = this.textState_;
      if (this.text_ === '' || !textState || (!fillState && !strokeState)) {
        return;
      }

      const coordinates = this.coordinates;
      let begin = coordinates.length;

      const geometryType = geometry.getType();
      let flatCoordinates = null;
      let stride = geometry.getStride();

      if (
        textState.placement === 'line' &&
        (geometryType == 'LineString' ||
          geometryType == 'MultiLineString' ||
          geometryType == 'Polygon' ||
          geometryType == 'MultiPolygon')
      ) {
        if (!intersects$3(this.getBufferedMaxExtent(), geometry.getExtent())) {
          return;
        }
        let ends;
        flatCoordinates = geometry.getFlatCoordinates();
        if (geometryType == 'LineString') {
          ends = [flatCoordinates.length];
        } else if (geometryType == 'MultiLineString') {
          ends = /** @type {import("../../geom/MultiLineString.js").default} */ (
            geometry
          ).getEnds();
        } else if (geometryType == 'Polygon') {
          ends = /** @type {import("../../geom/Polygon.js").default} */ (geometry)
            .getEnds()
            .slice(0, 1);
        } else if (geometryType == 'MultiPolygon') {
          const endss =
            /** @type {import("../../geom/MultiPolygon.js").default} */ (
              geometry
            ).getEndss();
          ends = [];
          for (let i = 0, ii = endss.length; i < ii; ++i) {
            ends.push(endss[i][0]);
          }
        }
        this.beginGeometry(geometry, feature);
        const textAlign = textState.textAlign;
        // No `justify` support for line placement.
        let flatOffset = 0;
        let flatEnd;
        for (let o = 0, oo = ends.length; o < oo; ++o) {
          if (textAlign == undefined) {
            const range = matchingChunk(
              textState.maxAngle,
              flatCoordinates,
              flatOffset,
              ends[o],
              stride
            );
            flatOffset = range[0];
            flatEnd = range[1];
          } else {
            flatEnd = ends[o];
          }
          for (let i = flatOffset; i < flatEnd; i += stride) {
            coordinates.push(flatCoordinates[i], flatCoordinates[i + 1]);
          }
          const end = coordinates.length;
          flatOffset = ends[o];
          this.drawChars_(begin, end);
          begin = end;
        }
        this.endGeometry(feature);
      } else {
        let geometryWidths = textState.overflow ? null : [];
        switch (geometryType) {
          case 'Point':
          case 'MultiPoint':
            flatCoordinates =
              /** @type {import("../../geom/MultiPoint.js").default} */ (
                geometry
              ).getFlatCoordinates();
            break;
          case 'LineString':
            flatCoordinates =
              /** @type {import("../../geom/LineString.js").default} */ (
                geometry
              ).getFlatMidpoint();
            break;
          case 'Circle':
            flatCoordinates =
              /** @type {import("../../geom/Circle.js").default} */ (
                geometry
              ).getCenter();
            break;
          case 'MultiLineString':
            flatCoordinates =
              /** @type {import("../../geom/MultiLineString.js").default} */ (
                geometry
              ).getFlatMidpoints();
            stride = 2;
            break;
          case 'Polygon':
            flatCoordinates =
              /** @type {import("../../geom/Polygon.js").default} */ (
                geometry
              ).getFlatInteriorPoint();
            if (!textState.overflow) {
              geometryWidths.push(flatCoordinates[2] / this.resolution);
            }
            stride = 3;
            break;
          case 'MultiPolygon':
            const interiorPoints =
              /** @type {import("../../geom/MultiPolygon.js").default} */ (
                geometry
              ).getFlatInteriorPoints();
            flatCoordinates = [];
            for (let i = 0, ii = interiorPoints.length; i < ii; i += 3) {
              if (!textState.overflow) {
                geometryWidths.push(interiorPoints[i + 2] / this.resolution);
              }
              flatCoordinates.push(interiorPoints[i], interiorPoints[i + 1]);
            }
            if (flatCoordinates.length === 0) {
              return;
            }
            stride = 2;
            break;
        }
        const end = this.appendFlatPointCoordinates(flatCoordinates, stride);
        if (end === begin) {
          return;
        }
        if (
          geometryWidths &&
          (end - begin) / 2 !== flatCoordinates.length / stride
        ) {
          let beg = begin / 2;
          geometryWidths = geometryWidths.filter((w, i) => {
            const keep =
              coordinates[(beg + i) * 2] === flatCoordinates[i * stride] &&
              coordinates[(beg + i) * 2 + 1] === flatCoordinates[i * stride + 1];
            if (!keep) {
              --beg;
            }
            return keep;
          });
        }

        this.saveTextStates_();

        if (textState.backgroundFill || textState.backgroundStroke) {
          this.setFillStrokeStyle(
            textState.backgroundFill,
            textState.backgroundStroke
          );
          if (textState.backgroundFill) {
            this.updateFillStyle(this.state, this.createFill);
            this.hitDetectionInstructions.push(this.createFill(this.state));
          }
          if (textState.backgroundStroke) {
            this.updateStrokeStyle(this.state, this.applyStroke);
            this.hitDetectionInstructions.push(this.createStroke(this.state));
          }
        }

        this.beginGeometry(geometry, feature);

        // adjust padding for negative scale
        let padding = textState.padding;
        if (
          padding != defaultPadding &&
          (textState.scale[0] < 0 || textState.scale[1] < 0)
        ) {
          let p0 = textState.padding[0];
          let p1 = textState.padding[1];
          let p2 = textState.padding[2];
          let p3 = textState.padding[3];
          if (textState.scale[0] < 0) {
            p1 = -p1;
            p3 = -p3;
          }
          if (textState.scale[1] < 0) {
            p0 = -p0;
            p2 = -p2;
          }
          padding = [p0, p1, p2, p3];
        }

        // The image is unknown at this stage so we pass null; it will be computed at render time.
        // For clarity, we pass NaN for offsetX, offsetY, width and height, which will be computed at
        // render time.
        const pixelRatio = this.pixelRatio;
        this.instructions.push([
          Instruction.DRAW_IMAGE,
          begin,
          end,
          null,
          NaN,
          NaN,
          NaN,
          1,
          0,
          0,
          this.textRotateWithView_,
          this.textRotation_,
          [1, 1],
          NaN,
          undefined,
          this.declutterImageWithText_,
          padding == defaultPadding
            ? defaultPadding
            : padding.map(function (p) {
                return p * pixelRatio;
              }),
          !!textState.backgroundFill,
          !!textState.backgroundStroke,
          this.text_,
          this.textKey_,
          this.strokeKey_,
          this.fillKey_,
          this.textOffsetX_,
          this.textOffsetY_,
          geometryWidths,
        ]);
        const scale = 1 / pixelRatio;
        this.hitDetectionInstructions.push([
          Instruction.DRAW_IMAGE,
          begin,
          end,
          null,
          NaN,
          NaN,
          NaN,
          1,
          0,
          0,
          this.textRotateWithView_,
          this.textRotation_,
          [scale, scale],
          NaN,
          undefined,
          this.declutterImageWithText_,
          padding,
          !!textState.backgroundFill,
          !!textState.backgroundStroke,
          this.text_,
          this.textKey_,
          this.strokeKey_,
          this.fillKey_,
          this.textOffsetX_,
          this.textOffsetY_,
          geometryWidths,
        ]);

        this.endGeometry(feature);
      }
    }

    /**
     * @private
     */
    saveTextStates_() {
      const strokeState = this.textStrokeState_;
      const textState = this.textState_;
      const fillState = this.textFillState_;

      const strokeKey = this.strokeKey_;
      if (strokeState) {
        if (!(strokeKey in this.strokeStates)) {
          this.strokeStates[strokeKey] = {
            strokeStyle: strokeState.strokeStyle,
            lineCap: strokeState.lineCap,
            lineDashOffset: strokeState.lineDashOffset,
            lineWidth: strokeState.lineWidth,
            lineJoin: strokeState.lineJoin,
            miterLimit: strokeState.miterLimit,
            lineDash: strokeState.lineDash,
          };
        }
      }
      const textKey = this.textKey_;
      if (!(textKey in this.textStates)) {
        this.textStates[textKey] = {
          font: textState.font,
          textAlign: textState.textAlign || defaultTextAlign,
          justify: textState.justify,
          textBaseline: textState.textBaseline || defaultTextBaseline,
          scale: textState.scale,
        };
      }
      const fillKey = this.fillKey_;
      if (fillState) {
        if (!(fillKey in this.fillStates)) {
          this.fillStates[fillKey] = {
            fillStyle: fillState.fillStyle,
          };
        }
      }
    }

    /**
     * @private
     * @param {number} begin Begin.
     * @param {number} end End.
     */
    drawChars_(begin, end) {
      const strokeState = this.textStrokeState_;
      const textState = this.textState_;

      const strokeKey = this.strokeKey_;
      const textKey = this.textKey_;
      const fillKey = this.fillKey_;
      this.saveTextStates_();

      const pixelRatio = this.pixelRatio;
      const baseline = TEXT_ALIGN[textState.textBaseline];

      const offsetY = this.textOffsetY_ * pixelRatio;
      const text = this.text_;
      const strokeWidth = strokeState
        ? (strokeState.lineWidth * Math.abs(textState.scale[0])) / 2
        : 0;

      this.instructions.push([
        Instruction.DRAW_CHARS,
        begin,
        end,
        baseline,
        textState.overflow,
        fillKey,
        textState.maxAngle,
        pixelRatio,
        offsetY,
        strokeKey,
        strokeWidth * pixelRatio,
        text,
        textKey,
        1,
      ]);
      this.hitDetectionInstructions.push([
        Instruction.DRAW_CHARS,
        begin,
        end,
        baseline,
        textState.overflow,
        fillKey,
        textState.maxAngle,
        1,
        offsetY,
        strokeKey,
        strokeWidth,
        text,
        textKey,
        1 / pixelRatio,
      ]);
    }

    /**
     * @param {import("../../style/Text.js").default} textStyle Text style.
     * @param {Object} [sharedData] Shared data.
     */
    setTextStyle(textStyle, sharedData) {
      let textState, fillState, strokeState;
      if (!textStyle) {
        this.text_ = '';
      } else {
        const textFillStyle = textStyle.getFill();
        if (!textFillStyle) {
          fillState = null;
          this.textFillState_ = fillState;
        } else {
          fillState = this.textFillState_;
          if (!fillState) {
            fillState = /** @type {import("../canvas.js").FillState} */ ({});
            this.textFillState_ = fillState;
          }
          fillState.fillStyle = asColorLike(
            textFillStyle.getColor() || defaultFillStyle
          );
        }

        const textStrokeStyle = textStyle.getStroke();
        if (!textStrokeStyle) {
          strokeState = null;
          this.textStrokeState_ = strokeState;
        } else {
          strokeState = this.textStrokeState_;
          if (!strokeState) {
            strokeState = /** @type {import("../canvas.js").StrokeState} */ ({});
            this.textStrokeState_ = strokeState;
          }
          const lineDash = textStrokeStyle.getLineDash();
          const lineDashOffset = textStrokeStyle.getLineDashOffset();
          const lineWidth = textStrokeStyle.getWidth();
          const miterLimit = textStrokeStyle.getMiterLimit();
          strokeState.lineCap = textStrokeStyle.getLineCap() || defaultLineCap;
          strokeState.lineDash = lineDash ? lineDash.slice() : defaultLineDash;
          strokeState.lineDashOffset =
            lineDashOffset === undefined ? defaultLineDashOffset : lineDashOffset;
          strokeState.lineJoin = textStrokeStyle.getLineJoin() || defaultLineJoin;
          strokeState.lineWidth =
            lineWidth === undefined ? defaultLineWidth : lineWidth;
          strokeState.miterLimit =
            miterLimit === undefined ? defaultMiterLimit : miterLimit;
          strokeState.strokeStyle = asColorLike(
            textStrokeStyle.getColor() || defaultStrokeStyle
          );
        }

        textState = this.textState_;
        const font = textStyle.getFont() || defaultFont;
        registerFont(font);
        const textScale = textStyle.getScaleArray();
        textState.overflow = textStyle.getOverflow();
        textState.font = font;
        textState.maxAngle = textStyle.getMaxAngle();
        textState.placement = textStyle.getPlacement();
        textState.textAlign = textStyle.getTextAlign();
        textState.justify = textStyle.getJustify();
        textState.textBaseline =
          textStyle.getTextBaseline() || defaultTextBaseline;
        textState.backgroundFill = textStyle.getBackgroundFill();
        textState.backgroundStroke = textStyle.getBackgroundStroke();
        textState.padding = textStyle.getPadding() || defaultPadding;
        textState.scale = textScale === undefined ? [1, 1] : textScale;

        const textOffsetX = textStyle.getOffsetX();
        const textOffsetY = textStyle.getOffsetY();
        const textRotateWithView = textStyle.getRotateWithView();
        const textRotation = textStyle.getRotation();
        this.text_ = textStyle.getText() || '';
        this.textOffsetX_ = textOffsetX === undefined ? 0 : textOffsetX;
        this.textOffsetY_ = textOffsetY === undefined ? 0 : textOffsetY;
        this.textRotateWithView_ =
          textRotateWithView === undefined ? false : textRotateWithView;
        this.textRotation_ = textRotation === undefined ? 0 : textRotation;

        this.strokeKey_ = strokeState
          ? (typeof strokeState.strokeStyle == 'string'
              ? strokeState.strokeStyle
              : getUid(strokeState.strokeStyle)) +
            strokeState.lineCap +
            strokeState.lineDashOffset +
            '|' +
            strokeState.lineWidth +
            strokeState.lineJoin +
            strokeState.miterLimit +
            '[' +
            strokeState.lineDash.join() +
            ']'
          : '';
        this.textKey_ =
          textState.font +
          textState.scale +
          (textState.textAlign || '?') +
          (textState.justify || '?') +
          (textState.textBaseline || '?');
        this.fillKey_ = fillState
          ? typeof fillState.fillStyle == 'string'
            ? fillState.fillStyle
            : '|' + getUid(fillState.fillStyle)
          : '';
      }
      this.declutterImageWithText_ = sharedData;
    }
  }

  /**
   * @module ol/render/canvas/BuilderGroup
   */

  /**
   * @type {Object<import("../canvas.js").BuilderType, typeof Builder>}
   */
  const BATCH_CONSTRUCTORS = {
    'Circle': CanvasPolygonBuilder,
    'Default': CanvasBuilder,
    'Image': CanvasImageBuilder,
    'LineString': CanvasLineStringBuilder,
    'Polygon': CanvasPolygonBuilder,
    'Text': CanvasTextBuilder,
  };

  class BuilderGroup {
    /**
     * @param {number} tolerance Tolerance.
     * @param {import("../../extent.js").Extent} maxExtent Max extent.
     * @param {number} resolution Resolution.
     * @param {number} pixelRatio Pixel ratio.
     */
    constructor(tolerance, maxExtent, resolution, pixelRatio) {
      /**
       * @private
       * @type {number}
       */
      this.tolerance_ = tolerance;

      /**
       * @private
       * @type {import("../../extent.js").Extent}
       */
      this.maxExtent_ = maxExtent;

      /**
       * @private
       * @type {number}
       */
      this.pixelRatio_ = pixelRatio;

      /**
       * @private
       * @type {number}
       */
      this.resolution_ = resolution;

      /**
       * @private
       * @type {!Object<string, !Object<import("../canvas.js").BuilderType, Builder>>}
       */
      this.buildersByZIndex_ = {};
    }

    /**
     * @return {!Object<string, !Object<import("../canvas.js").BuilderType, import("./Builder.js").SerializableInstructions>>} The serializable instructions
     */
    finish() {
      const builderInstructions = {};
      for (const zKey in this.buildersByZIndex_) {
        builderInstructions[zKey] = builderInstructions[zKey] || {};
        const builders = this.buildersByZIndex_[zKey];
        for (const builderKey in builders) {
          const builderInstruction = builders[builderKey].finish();
          builderInstructions[zKey][builderKey] = builderInstruction;
        }
      }
      return builderInstructions;
    }

    /**
     * @param {number|undefined} zIndex Z index.
     * @param {import("../canvas.js").BuilderType} builderType Replay type.
     * @return {import("../VectorContext.js").default} Replay.
     */
    getBuilder(zIndex, builderType) {
      const zIndexKey = zIndex !== undefined ? zIndex.toString() : '0';
      let replays = this.buildersByZIndex_[zIndexKey];
      if (replays === undefined) {
        replays = {};
        this.buildersByZIndex_[zIndexKey] = replays;
      }
      let replay = replays[builderType];
      if (replay === undefined) {
        const Constructor = BATCH_CONSTRUCTORS[builderType];
        replay = new Constructor(
          this.tolerance_,
          this.maxExtent_,
          this.resolution_,
          this.pixelRatio_
        );
        replays[builderType] = replay;
      }
      return replay;
    }
  }

  /**
   * @module ol/geom/flat/textpath
   */

  /**
   * @param {Array<number>} flatCoordinates Path to put text on.
   * @param {number} offset Start offset of the `flatCoordinates`.
   * @param {number} end End offset of the `flatCoordinates`.
   * @param {number} stride Stride.
   * @param {string} text Text to place on the path.
   * @param {number} startM m along the path where the text starts.
   * @param {number} maxAngle Max angle between adjacent chars in radians.
   * @param {number} scale The product of the text scale and the device pixel ratio.
   * @param {function(string, string, Object<string, number>):number} measureAndCacheTextWidth Measure and cache text width.
   * @param {string} font The font.
   * @param {Object<string, number>} cache A cache of measured widths.
   * @param {number} rotation Rotation to apply to the flatCoordinates to determine whether text needs to be reversed.
   * @return {Array<Array<*>>|null} The result array (or null if `maxAngle` was
   * exceeded). Entries of the array are x, y, anchorX, angle, chunk.
   */
  function drawTextOnPath(
    flatCoordinates,
    offset,
    end,
    stride,
    text,
    startM,
    maxAngle,
    scale,
    measureAndCacheTextWidth,
    font,
    cache,
    rotation
  ) {
    let x2 = flatCoordinates[offset];
    let y2 = flatCoordinates[offset + 1];
    let x1 = 0;
    let y1 = 0;
    let segmentLength = 0;
    let segmentM = 0;

    function advance() {
      x1 = x2;
      y1 = y2;
      offset += stride;
      x2 = flatCoordinates[offset];
      y2 = flatCoordinates[offset + 1];
      segmentM += segmentLength;
      segmentLength = Math.sqrt((x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1));
    }
    do {
      advance();
    } while (offset < end - stride && segmentM + segmentLength < startM);

    let interpolate =
      segmentLength === 0 ? 0 : (startM - segmentM) / segmentLength;
    const beginX = lerp(x1, x2, interpolate);
    const beginY = lerp(y1, y2, interpolate);

    const startOffset = offset - stride;
    const startLength = segmentM;
    const endM = startM + scale * measureAndCacheTextWidth(font, text, cache);
    while (offset < end - stride && segmentM + segmentLength < endM) {
      advance();
    }
    interpolate = segmentLength === 0 ? 0 : (endM - segmentM) / segmentLength;
    const endX = lerp(x1, x2, interpolate);
    const endY = lerp(y1, y2, interpolate);

    // Keep text upright
    let reverse;
    if (rotation) {
      const flat = [beginX, beginY, endX, endY];
      rotate(flat, 0, 4, 2, rotation, flat, flat);
      reverse = flat[0] > flat[2];
    } else {
      reverse = beginX > endX;
    }

    const PI = Math.PI;
    const result = [];
    const singleSegment = startOffset + stride === offset;

    offset = startOffset;
    segmentLength = 0;
    segmentM = startLength;
    x2 = flatCoordinates[offset];
    y2 = flatCoordinates[offset + 1];

    let previousAngle;
    // All on the same segment
    if (singleSegment) {
      advance();

      previousAngle = Math.atan2(y2 - y1, x2 - x1);
      if (reverse) {
        previousAngle += previousAngle > 0 ? -PI : PI;
      }
      const x = (endX + beginX) / 2;
      const y = (endY + beginY) / 2;
      result[0] = [x, y, (endM - startM) / 2, previousAngle, text];
      return result;
    }

    // rendering across line segments
    text = text.replace(/\n/g, ' '); // ensure rendering in single-line as all calculations below don't handle multi-lines

    for (let i = 0, ii = text.length; i < ii; ) {
      advance();
      let angle = Math.atan2(y2 - y1, x2 - x1);
      if (reverse) {
        angle += angle > 0 ? -PI : PI;
      }
      if (previousAngle !== undefined) {
        let delta = angle - previousAngle;
        delta += delta > PI ? -2 * PI : delta < -PI ? 2 * PI : 0;
        if (Math.abs(delta) > maxAngle) {
          return null;
        }
      }
      previousAngle = angle;

      const iStart = i;
      let charLength = 0;
      for (; i < ii; ++i) {
        const index = reverse ? ii - i - 1 : i;
        const len = scale * measureAndCacheTextWidth(font, text[index], cache);
        if (
          offset + stride < end &&
          segmentM + segmentLength < startM + charLength + len / 2
        ) {
          break;
        }
        charLength += len;
      }
      if (i === iStart) {
        continue;
      }
      const chars = reverse
        ? text.substring(ii - iStart, ii - i)
        : text.substring(iStart, i);
      interpolate =
        segmentLength === 0
          ? 0
          : (startM + charLength / 2 - segmentM) / segmentLength;
      const x = lerp(x1, x2, interpolate);
      const y = lerp(y1, y2, interpolate);
      result.push([x, y, charLength / 2, angle, chars]);
      startM += charLength;
    }
    return result;
  }

  /**
   * @module ol/geom/flat/length
   */

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   * @return {number} Length.
   */
  function lineStringLength(flatCoordinates, offset, end, stride) {
    let x1 = flatCoordinates[offset];
    let y1 = flatCoordinates[offset + 1];
    let length = 0;
    for (let i = offset + stride; i < end; i += stride) {
      const x2 = flatCoordinates[i];
      const y2 = flatCoordinates[i + 1];
      length += Math.sqrt((x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1));
      x1 = x2;
      y1 = y2;
    }
    return length;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   * @return {number} Perimeter.
   */
  function linearRingLength(flatCoordinates, offset, end, stride) {
    let perimeter = lineStringLength(flatCoordinates, offset, end, stride);
    const dx = flatCoordinates[end - stride] - flatCoordinates[offset];
    const dy = flatCoordinates[end - stride + 1] - flatCoordinates[offset + 1];
    perimeter += Math.sqrt(dx * dx + dy * dy);
    return perimeter;
  }

  /**
   * @module ol/render/canvas/Executor
   */

  /**
   * @typedef {Object} BBox
   * @property {number} minX Minimal x.
   * @property {number} minY Minimal y.
   * @property {number} maxX Maximal x.
   * @property {number} maxY Maximal y
   * @property {*} value Value.
   */

  /**
   * @typedef {Object} ImageOrLabelDimensions
   * @property {number} drawImageX DrawImageX.
   * @property {number} drawImageY DrawImageY.
   * @property {number} drawImageW DrawImageW.
   * @property {number} drawImageH DrawImageH.
   * @property {number} originX OriginX.
   * @property {number} originY OriginY.
   * @property {Array<number>} scale Scale.
   * @property {BBox} declutterBox DeclutterBox.
   * @property {import("../../transform.js").Transform} canvasTransform CanvasTransform.
   */

  /**
   * @typedef {{0: CanvasRenderingContext2D, 1: number, 2: import("../canvas.js").Label|HTMLImageElement|HTMLCanvasElement|HTMLVideoElement, 3: ImageOrLabelDimensions, 4: number, 5: Array<*>, 6: Array<*>}} ReplayImageOrLabelArgs
   */

  /**
   * @template T
   * @typedef {function(import("../../Feature.js").FeatureLike, import("../../geom/SimpleGeometry.js").default): T} FeatureCallback
   */

  /**
   * @type {import("../../extent.js").Extent}
   */
  const tmpExtent = createEmpty();

  /** @type {import("../../coordinate.js").Coordinate} */
  const p1 = [];
  /** @type {import("../../coordinate.js").Coordinate} */
  const p2 = [];
  /** @type {import("../../coordinate.js").Coordinate} */
  const p3 = [];
  /** @type {import("../../coordinate.js").Coordinate} */
  const p4 = [];

  /**
   * @param {ReplayImageOrLabelArgs} replayImageOrLabelArgs Arguments to replayImageOrLabel
   * @return {BBox} Declutter bbox.
   */
  function getDeclutterBox(replayImageOrLabelArgs) {
    return replayImageOrLabelArgs[3].declutterBox;
  }

  const rtlRegEx = new RegExp(
    /* eslint-disable prettier/prettier */
    '[' +
      String.fromCharCode(0x00591) + '-' + String.fromCharCode(0x008ff) +
      String.fromCharCode(0x0fb1d) + '-' + String.fromCharCode(0x0fdff) +
      String.fromCharCode(0x0fe70) + '-' + String.fromCharCode(0x0fefc) +
      String.fromCharCode(0x10800) + '-' + String.fromCharCode(0x10fff) +
      String.fromCharCode(0x1e800) + '-' + String.fromCharCode(0x1efff) +
    ']'
    /* eslint-enable prettier/prettier */
  );

  /**
   * @param {string} text Text.
   * @param {CanvasTextAlign} align Alignment.
   * @return {number} Text alignment.
   */
  function horizontalTextAlign(text, align) {
    if ((align === 'start' || align === 'end') && !rtlRegEx.test(text)) {
      align = align === 'start' ? 'left' : 'right';
    }
    return TEXT_ALIGN[align];
  }

  /**
   * @param {Array<string>} acc Accumulator.
   * @param {string} line Line of text.
   * @param {number} i Index
   * @return {Array<string>} Accumulator.
   */
  function createTextChunks(acc, line, i) {
    if (i > 0) {
      acc.push('\n', '');
    }
    acc.push(line, '');
    return acc;
  }

  class Executor {
    /**
     * @param {number} resolution Resolution.
     * @param {number} pixelRatio Pixel ratio.
     * @param {boolean} overlaps The replay can have overlapping geometries.
     * @param {import("../canvas.js").SerializableInstructions} instructions The serializable instructions
     */
    constructor(resolution, pixelRatio, overlaps, instructions) {
      /**
       * @protected
       * @type {boolean}
       */
      this.overlaps = overlaps;

      /**
       * @protected
       * @type {number}
       */
      this.pixelRatio = pixelRatio;

      /**
       * @protected
       * @const
       * @type {number}
       */
      this.resolution = resolution;

      /**
       * @private
       * @type {boolean}
       */
      this.alignFill_;

      /**
       * @protected
       * @type {Array<*>}
       */
      this.instructions = instructions.instructions;

      /**
       * @protected
       * @type {Array<number>}
       */
      this.coordinates = instructions.coordinates;

      /**
       * @private
       * @type {!Object<number,import("../../coordinate.js").Coordinate|Array<import("../../coordinate.js").Coordinate>|Array<Array<import("../../coordinate.js").Coordinate>>>}
       */
      this.coordinateCache_ = {};

      /**
       * @private
       * @type {!import("../../transform.js").Transform}
       */
      this.renderedTransform_ = create$2();

      /**
       * @protected
       * @type {Array<*>}
       */
      this.hitDetectionInstructions = instructions.hitDetectionInstructions;

      /**
       * @private
       * @type {Array<number>}
       */
      this.pixelCoordinates_ = null;

      /**
       * @private
       * @type {number}
       */
      this.viewRotation_ = 0;

      /**
       * @type {!Object<string, import("../canvas.js").FillState>}
       */
      this.fillStates = instructions.fillStates || {};

      /**
       * @type {!Object<string, import("../canvas.js").StrokeState>}
       */
      this.strokeStates = instructions.strokeStates || {};

      /**
       * @type {!Object<string, import("../canvas.js").TextState>}
       */
      this.textStates = instructions.textStates || {};

      /**
       * @private
       * @type {Object<string, Object<string, number>>}
       */
      this.widths_ = {};

      /**
       * @private
       * @type {Object<string, import("../canvas.js").Label>}
       */
      this.labels_ = {};
    }

    /**
     * @param {string|Array<string>} text Text.
     * @param {string} textKey Text style key.
     * @param {string} fillKey Fill style key.
     * @param {string} strokeKey Stroke style key.
     * @return {import("../canvas.js").Label} Label.
     */
    createLabel(text, textKey, fillKey, strokeKey) {
      const key = text + textKey + fillKey + strokeKey;
      if (this.labels_[key]) {
        return this.labels_[key];
      }
      const strokeState = strokeKey ? this.strokeStates[strokeKey] : null;
      const fillState = fillKey ? this.fillStates[fillKey] : null;
      const textState = this.textStates[textKey];
      const pixelRatio = this.pixelRatio;
      const scale = [
        textState.scale[0] * pixelRatio,
        textState.scale[1] * pixelRatio,
      ];
      const textIsArray = Array.isArray(text);
      const align = textState.justify
        ? TEXT_ALIGN[textState.justify]
        : horizontalTextAlign(
            Array.isArray(text) ? text[0] : text,
            textState.textAlign || defaultTextAlign
          );
      const strokeWidth =
        strokeKey && strokeState.lineWidth ? strokeState.lineWidth : 0;

      const chunks = textIsArray
        ? text
        : text.split('\n').reduce(createTextChunks, []);

      const {width, height, widths, heights, lineWidths} = getTextDimensions(
        textState,
        chunks
      );
      const renderWidth = width + strokeWidth;
      const contextInstructions = [];
      // make canvas 2 pixels wider to account for italic text width measurement errors
      const w = (renderWidth + 2) * scale[0];
      const h = (height + strokeWidth) * scale[1];
      /** @type {import("../canvas.js").Label} */
      const label = {
        width: w < 0 ? Math.floor(w) : Math.ceil(w),
        height: h < 0 ? Math.floor(h) : Math.ceil(h),
        contextInstructions: contextInstructions,
      };
      if (scale[0] != 1 || scale[1] != 1) {
        contextInstructions.push('scale', scale);
      }
      if (strokeKey) {
        contextInstructions.push('strokeStyle', strokeState.strokeStyle);
        contextInstructions.push('lineWidth', strokeWidth);
        contextInstructions.push('lineCap', strokeState.lineCap);
        contextInstructions.push('lineJoin', strokeState.lineJoin);
        contextInstructions.push('miterLimit', strokeState.miterLimit);
        contextInstructions.push('setLineDash', [strokeState.lineDash]);
        contextInstructions.push('lineDashOffset', strokeState.lineDashOffset);
      }
      if (fillKey) {
        contextInstructions.push('fillStyle', fillState.fillStyle);
      }
      contextInstructions.push('textBaseline', 'middle');
      contextInstructions.push('textAlign', 'center');
      const leftRight = 0.5 - align;
      let x = align * renderWidth + leftRight * strokeWidth;
      const strokeInstructions = [];
      const fillInstructions = [];
      let lineHeight = 0;
      let lineOffset = 0;
      let widthHeightIndex = 0;
      let lineWidthIndex = 0;
      let previousFont;
      for (let i = 0, ii = chunks.length; i < ii; i += 2) {
        const text = chunks[i];
        if (text === '\n') {
          lineOffset += lineHeight;
          lineHeight = 0;
          x = align * renderWidth + leftRight * strokeWidth;
          ++lineWidthIndex;
          continue;
        }
        const font = chunks[i + 1] || textState.font;
        if (font !== previousFont) {
          if (strokeKey) {
            strokeInstructions.push('font', font);
          }
          if (fillKey) {
            fillInstructions.push('font', font);
          }
          previousFont = font;
        }
        lineHeight = Math.max(lineHeight, heights[widthHeightIndex]);
        const fillStrokeArgs = [
          text,
          x +
            leftRight * widths[widthHeightIndex] +
            align * (widths[widthHeightIndex] - lineWidths[lineWidthIndex]),
          0.5 * (strokeWidth + lineHeight) + lineOffset,
        ];
        x += widths[widthHeightIndex];
        if (strokeKey) {
          strokeInstructions.push('strokeText', fillStrokeArgs);
        }
        if (fillKey) {
          fillInstructions.push('fillText', fillStrokeArgs);
        }
        ++widthHeightIndex;
      }
      Array.prototype.push.apply(contextInstructions, strokeInstructions);
      Array.prototype.push.apply(contextInstructions, fillInstructions);
      this.labels_[key] = label;
      return label;
    }

    /**
     * @param {CanvasRenderingContext2D} context Context.
     * @param {import("../../coordinate.js").Coordinate} p1 1st point of the background box.
     * @param {import("../../coordinate.js").Coordinate} p2 2nd point of the background box.
     * @param {import("../../coordinate.js").Coordinate} p3 3rd point of the background box.
     * @param {import("../../coordinate.js").Coordinate} p4 4th point of the background box.
     * @param {Array<*>} fillInstruction Fill instruction.
     * @param {Array<*>} strokeInstruction Stroke instruction.
     */
    replayTextBackground_(
      context,
      p1,
      p2,
      p3,
      p4,
      fillInstruction,
      strokeInstruction
    ) {
      context.beginPath();
      context.moveTo.apply(context, p1);
      context.lineTo.apply(context, p2);
      context.lineTo.apply(context, p3);
      context.lineTo.apply(context, p4);
      context.lineTo.apply(context, p1);
      if (fillInstruction) {
        this.alignFill_ = /** @type {boolean} */ (fillInstruction[2]);
        this.fill_(context);
      }
      if (strokeInstruction) {
        this.setStrokeStyle_(
          context,
          /** @type {Array<*>} */ (strokeInstruction)
        );
        context.stroke();
      }
    }

    /**
     * @private
     * @param {number} sheetWidth Width of the sprite sheet.
     * @param {number} sheetHeight Height of the sprite sheet.
     * @param {number} centerX X.
     * @param {number} centerY Y.
     * @param {number} width Width.
     * @param {number} height Height.
     * @param {number} anchorX Anchor X.
     * @param {number} anchorY Anchor Y.
     * @param {number} originX Origin X.
     * @param {number} originY Origin Y.
     * @param {number} rotation Rotation.
     * @param {import("../../size.js").Size} scale Scale.
     * @param {boolean} snapToPixel Snap to pixel.
     * @param {Array<number>} padding Padding.
     * @param {boolean} fillStroke Background fill or stroke.
     * @param {import("../../Feature.js").FeatureLike} feature Feature.
     * @return {ImageOrLabelDimensions} Dimensions for positioning and decluttering the image or label.
     */
    calculateImageOrLabelDimensions_(
      sheetWidth,
      sheetHeight,
      centerX,
      centerY,
      width,
      height,
      anchorX,
      anchorY,
      originX,
      originY,
      rotation,
      scale,
      snapToPixel,
      padding,
      fillStroke,
      feature
    ) {
      anchorX *= scale[0];
      anchorY *= scale[1];
      let x = centerX - anchorX;
      let y = centerY - anchorY;

      const w = width + originX > sheetWidth ? sheetWidth - originX : width;
      const h = height + originY > sheetHeight ? sheetHeight - originY : height;
      const boxW = padding[3] + w * scale[0] + padding[1];
      const boxH = padding[0] + h * scale[1] + padding[2];
      const boxX = x - padding[3];
      const boxY = y - padding[0];

      if (fillStroke || rotation !== 0) {
        p1[0] = boxX;
        p4[0] = boxX;
        p1[1] = boxY;
        p2[1] = boxY;
        p2[0] = boxX + boxW;
        p3[0] = p2[0];
        p3[1] = boxY + boxH;
        p4[1] = p3[1];
      }

      let transform;
      if (rotation !== 0) {
        transform = compose(
          create$2(),
          centerX,
          centerY,
          1,
          1,
          rotation,
          -centerX,
          -centerY
        );

        apply(transform, p1);
        apply(transform, p2);
        apply(transform, p3);
        apply(transform, p4);
        createOrUpdate$2(
          Math.min(p1[0], p2[0], p3[0], p4[0]),
          Math.min(p1[1], p2[1], p3[1], p4[1]),
          Math.max(p1[0], p2[0], p3[0], p4[0]),
          Math.max(p1[1], p2[1], p3[1], p4[1]),
          tmpExtent
        );
      } else {
        createOrUpdate$2(
          Math.min(boxX, boxX + boxW),
          Math.min(boxY, boxY + boxH),
          Math.max(boxX, boxX + boxW),
          Math.max(boxY, boxY + boxH),
          tmpExtent
        );
      }
      if (snapToPixel) {
        x = Math.round(x);
        y = Math.round(y);
      }
      return {
        drawImageX: x,
        drawImageY: y,
        drawImageW: w,
        drawImageH: h,
        originX: originX,
        originY: originY,
        declutterBox: {
          minX: tmpExtent[0],
          minY: tmpExtent[1],
          maxX: tmpExtent[2],
          maxY: tmpExtent[3],
          value: feature,
        },
        canvasTransform: transform,
        scale: scale,
      };
    }

    /**
     * @private
     * @param {CanvasRenderingContext2D} context Context.
     * @param {number} contextScale Scale of the context.
     * @param {import("../canvas.js").Label|HTMLImageElement|HTMLCanvasElement|HTMLVideoElement} imageOrLabel Image.
     * @param {ImageOrLabelDimensions} dimensions Dimensions.
     * @param {number} opacity Opacity.
     * @param {Array<*>} fillInstruction Fill instruction.
     * @param {Array<*>} strokeInstruction Stroke instruction.
     * @return {boolean} The image or label was rendered.
     */
    replayImageOrLabel_(
      context,
      contextScale,
      imageOrLabel,
      dimensions,
      opacity,
      fillInstruction,
      strokeInstruction
    ) {
      const fillStroke = !!(fillInstruction || strokeInstruction);

      const box = dimensions.declutterBox;
      const canvas = context.canvas;
      const strokePadding = strokeInstruction
        ? (strokeInstruction[2] * dimensions.scale[0]) / 2
        : 0;
      const intersects =
        box.minX - strokePadding <= canvas.width / contextScale &&
        box.maxX + strokePadding >= 0 &&
        box.minY - strokePadding <= canvas.height / contextScale &&
        box.maxY + strokePadding >= 0;

      if (intersects) {
        if (fillStroke) {
          this.replayTextBackground_(
            context,
            p1,
            p2,
            p3,
            p4,
            /** @type {Array<*>} */ (fillInstruction),
            /** @type {Array<*>} */ (strokeInstruction)
          );
        }
        drawImageOrLabel(
          context,
          dimensions.canvasTransform,
          opacity,
          imageOrLabel,
          dimensions.originX,
          dimensions.originY,
          dimensions.drawImageW,
          dimensions.drawImageH,
          dimensions.drawImageX,
          dimensions.drawImageY,
          dimensions.scale
        );
      }
      return true;
    }

    /**
     * @private
     * @param {CanvasRenderingContext2D} context Context.
     */
    fill_(context) {
      if (this.alignFill_) {
        const origin = apply(this.renderedTransform_, [0, 0]);
        const repeatSize = 512 * this.pixelRatio;
        context.save();
        context.translate(origin[0] % repeatSize, origin[1] % repeatSize);
        context.rotate(this.viewRotation_);
      }
      context.fill();
      if (this.alignFill_) {
        context.restore();
      }
    }

    /**
     * @private
     * @param {CanvasRenderingContext2D} context Context.
     * @param {Array<*>} instruction Instruction.
     */
    setStrokeStyle_(context, instruction) {
      context['strokeStyle'] =
        /** @type {import("../../colorlike.js").ColorLike} */ (instruction[1]);
      context.lineWidth = /** @type {number} */ (instruction[2]);
      context.lineCap = /** @type {CanvasLineCap} */ (instruction[3]);
      context.lineJoin = /** @type {CanvasLineJoin} */ (instruction[4]);
      context.miterLimit = /** @type {number} */ (instruction[5]);
      context.lineDashOffset = /** @type {number} */ (instruction[7]);
      context.setLineDash(/** @type {Array<number>} */ (instruction[6]));
    }

    /**
     * @private
     * @param {string|Array<string>} text The text to draw.
     * @param {string} textKey The key of the text state.
     * @param {string} strokeKey The key for the stroke state.
     * @param {string} fillKey The key for the fill state.
     * @return {{label: import("../canvas.js").Label, anchorX: number, anchorY: number}} The text image and its anchor.
     */
    drawLabelWithPointPlacement_(text, textKey, strokeKey, fillKey) {
      const textState = this.textStates[textKey];

      const label = this.createLabel(text, textKey, fillKey, strokeKey);

      const strokeState = this.strokeStates[strokeKey];
      const pixelRatio = this.pixelRatio;
      const align = horizontalTextAlign(
        Array.isArray(text) ? text[0] : text,
        textState.textAlign || defaultTextAlign
      );
      const baseline = TEXT_ALIGN[textState.textBaseline || defaultTextBaseline];
      const strokeWidth =
        strokeState && strokeState.lineWidth ? strokeState.lineWidth : 0;

      // Remove the 2 pixels we added in createLabel() for the anchor
      const width = label.width / pixelRatio - 2 * textState.scale[0];
      const anchorX = align * width + 2 * (0.5 - align) * strokeWidth;
      const anchorY =
        (baseline * label.height) / pixelRatio +
        2 * (0.5 - baseline) * strokeWidth;

      return {
        label: label,
        anchorX: anchorX,
        anchorY: anchorY,
      };
    }

    /**
     * @private
     * @param {CanvasRenderingContext2D} context Context.
     * @param {number} contextScale Scale of the context.
     * @param {import("../../transform.js").Transform} transform Transform.
     * @param {Array<*>} instructions Instructions array.
     * @param {boolean} snapToPixel Snap point symbols and text to integer pixels.
     * @param {FeatureCallback<T>} [featureCallback] Feature callback.
     * @param {import("../../extent.js").Extent} [hitExtent] Only check
     *     features that intersect this extent.
     * @param {import("rbush").default} [declutterTree] Declutter tree.
     * @return {T|undefined} Callback result.
     * @template T
     */
    execute_(
      context,
      contextScale,
      transform,
      instructions,
      snapToPixel,
      featureCallback,
      hitExtent,
      declutterTree
    ) {
      /** @type {Array<number>} */
      let pixelCoordinates;
      if (this.pixelCoordinates_ && equals$3(transform, this.renderedTransform_)) {
        pixelCoordinates = this.pixelCoordinates_;
      } else {
        if (!this.pixelCoordinates_) {
          this.pixelCoordinates_ = [];
        }
        pixelCoordinates = transform2D(
          this.coordinates,
          0,
          this.coordinates.length,
          2,
          transform,
          this.pixelCoordinates_
        );
        setFromArray(this.renderedTransform_, transform);
      }
      let i = 0; // instruction index
      const ii = instructions.length; // end of instructions
      let d = 0; // data index
      let dd; // end of per-instruction data
      let anchorX,
        anchorY,
        prevX,
        prevY,
        roundX,
        roundY,
        image,
        text,
        textKey,
        strokeKey,
        fillKey;
      let pendingFill = 0;
      let pendingStroke = 0;
      let lastFillInstruction = null;
      let lastStrokeInstruction = null;
      const coordinateCache = this.coordinateCache_;
      const viewRotation = this.viewRotation_;
      const viewRotationFromTransform =
        Math.round(Math.atan2(-transform[1], transform[0]) * 1e12) / 1e12;

      const state = /** @type {import("../../render.js").State} */ ({
        context: context,
        pixelRatio: this.pixelRatio,
        resolution: this.resolution,
        rotation: viewRotation,
      });

      // When the batch size gets too big, performance decreases. 200 is a good
      // balance between batch size and number of fill/stroke instructions.
      const batchSize =
        this.instructions != instructions || this.overlaps ? 0 : 200;
      let /** @type {import("../../Feature.js").FeatureLike} */ feature;
      let x, y, currentGeometry;
      while (i < ii) {
        const instruction = instructions[i];
        const type = /** @type {import("./Instruction.js").default} */ (
          instruction[0]
        );
        switch (type) {
          case Instruction.BEGIN_GEOMETRY:
            feature = /** @type {import("../../Feature.js").FeatureLike} */ (
              instruction[1]
            );
            currentGeometry = instruction[3];
            if (!feature.getGeometry()) {
              i = /** @type {number} */ (instruction[2]);
            } else if (
              hitExtent !== undefined &&
              !intersects$3(hitExtent, currentGeometry.getExtent())
            ) {
              i = /** @type {number} */ (instruction[2]) + 1;
            } else {
              ++i;
            }
            break;
          case Instruction.BEGIN_PATH:
            if (pendingFill > batchSize) {
              this.fill_(context);
              pendingFill = 0;
            }
            if (pendingStroke > batchSize) {
              context.stroke();
              pendingStroke = 0;
            }
            if (!pendingFill && !pendingStroke) {
              context.beginPath();
              prevX = NaN;
              prevY = NaN;
            }
            ++i;
            break;
          case Instruction.CIRCLE:
            d = /** @type {number} */ (instruction[1]);
            const x1 = pixelCoordinates[d];
            const y1 = pixelCoordinates[d + 1];
            const x2 = pixelCoordinates[d + 2];
            const y2 = pixelCoordinates[d + 3];
            const dx = x2 - x1;
            const dy = y2 - y1;
            const r = Math.sqrt(dx * dx + dy * dy);
            context.moveTo(x1 + r, y1);
            context.arc(x1, y1, r, 0, 2 * Math.PI, true);
            ++i;
            break;
          case Instruction.CLOSE_PATH:
            context.closePath();
            ++i;
            break;
          case Instruction.CUSTOM:
            d = /** @type {number} */ (instruction[1]);
            dd = instruction[2];
            const geometry =
              /** @type {import("../../geom/SimpleGeometry.js").default} */ (
                instruction[3]
              );
            const renderer = instruction[4];
            const fn = instruction.length == 6 ? instruction[5] : undefined;
            state.geometry = geometry;
            state.feature = feature;
            if (!(i in coordinateCache)) {
              coordinateCache[i] = [];
            }
            const coords = coordinateCache[i];
            if (fn) {
              fn(pixelCoordinates, d, dd, 2, coords);
            } else {
              coords[0] = pixelCoordinates[d];
              coords[1] = pixelCoordinates[d + 1];
              coords.length = 2;
            }
            renderer(coords, state);
            ++i;
            break;
          case Instruction.DRAW_IMAGE:
            d = /** @type {number} */ (instruction[1]);
            dd = /** @type {number} */ (instruction[2]);
            image =
              /** @type {HTMLCanvasElement|HTMLVideoElement|HTMLImageElement} */ (
                instruction[3]
              );

            // Remaining arguments in DRAW_IMAGE are in alphabetical order
            anchorX = /** @type {number} */ (instruction[4]);
            anchorY = /** @type {number} */ (instruction[5]);
            let height = /** @type {number} */ (instruction[6]);
            const opacity = /** @type {number} */ (instruction[7]);
            const originX = /** @type {number} */ (instruction[8]);
            const originY = /** @type {number} */ (instruction[9]);
            const rotateWithView = /** @type {boolean} */ (instruction[10]);
            let rotation = /** @type {number} */ (instruction[11]);
            const scale = /** @type {import("../../size.js").Size} */ (
              instruction[12]
            );
            let width = /** @type {number} */ (instruction[13]);
            const declutterMode =
              /** @type {"declutter"|"obstacle"|"none"|undefined} */ (
                instruction[14]
              );
            const declutterImageWithText =
              /** @type {import("../canvas.js").DeclutterImageWithText} */ (
                instruction[15]
              );

            if (!image && instruction.length >= 20) {
              // create label images
              text = /** @type {string} */ (instruction[19]);
              textKey = /** @type {string} */ (instruction[20]);
              strokeKey = /** @type {string} */ (instruction[21]);
              fillKey = /** @type {string} */ (instruction[22]);
              const labelWithAnchor = this.drawLabelWithPointPlacement_(
                text,
                textKey,
                strokeKey,
                fillKey
              );
              image = labelWithAnchor.label;
              instruction[3] = image;
              const textOffsetX = /** @type {number} */ (instruction[23]);
              anchorX = (labelWithAnchor.anchorX - textOffsetX) * this.pixelRatio;
              instruction[4] = anchorX;
              const textOffsetY = /** @type {number} */ (instruction[24]);
              anchorY = (labelWithAnchor.anchorY - textOffsetY) * this.pixelRatio;
              instruction[5] = anchorY;
              height = image.height;
              instruction[6] = height;
              width = image.width;
              instruction[13] = width;
            }

            let geometryWidths;
            if (instruction.length > 25) {
              geometryWidths = /** @type {number} */ (instruction[25]);
            }

            let padding, backgroundFill, backgroundStroke;
            if (instruction.length > 17) {
              padding = /** @type {Array<number>} */ (instruction[16]);
              backgroundFill = /** @type {boolean} */ (instruction[17]);
              backgroundStroke = /** @type {boolean} */ (instruction[18]);
            } else {
              padding = defaultPadding;
              backgroundFill = false;
              backgroundStroke = false;
            }

            if (rotateWithView && viewRotationFromTransform) {
              // Canvas is expected to be rotated to reverse view rotation.
              rotation += viewRotation;
            } else if (!rotateWithView && !viewRotationFromTransform) {
              // Canvas is not rotated, images need to be rotated back to be north-up.
              rotation -= viewRotation;
            }
            let widthIndex = 0;
            for (; d < dd; d += 2) {
              if (
                geometryWidths &&
                geometryWidths[widthIndex++] < width / this.pixelRatio
              ) {
                continue;
              }
              const dimensions = this.calculateImageOrLabelDimensions_(
                image.width,
                image.height,
                pixelCoordinates[d],
                pixelCoordinates[d + 1],
                width,
                height,
                anchorX,
                anchorY,
                originX,
                originY,
                rotation,
                scale,
                snapToPixel,
                padding,
                backgroundFill || backgroundStroke,
                feature
              );
              /** @type {ReplayImageOrLabelArgs} */
              const args = [
                context,
                contextScale,
                image,
                dimensions,
                opacity,
                backgroundFill
                  ? /** @type {Array<*>} */ (lastFillInstruction)
                  : null,
                backgroundStroke
                  ? /** @type {Array<*>} */ (lastStrokeInstruction)
                  : null,
              ];
              if (declutterTree) {
                if (declutterMode === 'none') {
                  // not rendered in declutter group
                  continue;
                } else if (declutterMode === 'obstacle') {
                  // will always be drawn, thus no collision detection, but insert as obstacle
                  declutterTree.insert(dimensions.declutterBox);
                  continue;
                } else {
                  let imageArgs;
                  let imageDeclutterBox;
                  if (declutterImageWithText) {
                    const index = dd - d;
                    if (!declutterImageWithText[index]) {
                      // We now have the image for an image+text combination.
                      declutterImageWithText[index] = args;
                      // Don't render anything for now, wait for the text.
                      continue;
                    }
                    imageArgs = declutterImageWithText[index];
                    delete declutterImageWithText[index];
                    imageDeclutterBox = getDeclutterBox(imageArgs);
                    if (declutterTree.collides(imageDeclutterBox)) {
                      continue;
                    }
                  }
                  if (declutterTree.collides(dimensions.declutterBox)) {
                    continue;
                  }
                  if (imageArgs) {
                    // We now have image and text for an image+text combination.
                    declutterTree.insert(imageDeclutterBox);
                    // Render the image before we render the text.
                    this.replayImageOrLabel_.apply(this, imageArgs);
                  }
                  declutterTree.insert(dimensions.declutterBox);
                }
              }
              this.replayImageOrLabel_.apply(this, args);
            }
            ++i;
            break;
          case Instruction.DRAW_CHARS:
            const begin = /** @type {number} */ (instruction[1]);
            const end = /** @type {number} */ (instruction[2]);
            const baseline = /** @type {number} */ (instruction[3]);
            const overflow = /** @type {number} */ (instruction[4]);
            fillKey = /** @type {string} */ (instruction[5]);
            const maxAngle = /** @type {number} */ (instruction[6]);
            const measurePixelRatio = /** @type {number} */ (instruction[7]);
            const offsetY = /** @type {number} */ (instruction[8]);
            strokeKey = /** @type {string} */ (instruction[9]);
            const strokeWidth = /** @type {number} */ (instruction[10]);
            text = /** @type {string} */ (instruction[11]);
            textKey = /** @type {string} */ (instruction[12]);
            const pixelRatioScale = [
              /** @type {number} */ (instruction[13]),
              /** @type {number} */ (instruction[13]),
            ];

            const textState = this.textStates[textKey];
            const font = textState.font;
            const textScale = [
              textState.scale[0] * measurePixelRatio,
              textState.scale[1] * measurePixelRatio,
            ];

            let cachedWidths;
            if (font in this.widths_) {
              cachedWidths = this.widths_[font];
            } else {
              cachedWidths = {};
              this.widths_[font] = cachedWidths;
            }

            const pathLength = lineStringLength(pixelCoordinates, begin, end, 2);
            const textLength =
              Math.abs(textScale[0]) *
              measureAndCacheTextWidth(font, text, cachedWidths);
            if (overflow || textLength <= pathLength) {
              const textAlign = this.textStates[textKey].textAlign;
              const startM = (pathLength - textLength) * TEXT_ALIGN[textAlign];
              const parts = drawTextOnPath(
                pixelCoordinates,
                begin,
                end,
                2,
                text,
                startM,
                maxAngle,
                Math.abs(textScale[0]),
                measureAndCacheTextWidth,
                font,
                cachedWidths,
                viewRotationFromTransform ? 0 : this.viewRotation_
              );
              drawChars: if (parts) {
                /** @type {Array<ReplayImageOrLabelArgs>} */
                const replayImageOrLabelArgs = [];
                let c, cc, chars, label, part;
                if (strokeKey) {
                  for (c = 0, cc = parts.length; c < cc; ++c) {
                    part = parts[c]; // x, y, anchorX, rotation, chunk
                    chars = /** @type {string} */ (part[4]);
                    label = this.createLabel(chars, textKey, '', strokeKey);
                    anchorX =
                      /** @type {number} */ (part[2]) +
                      (textScale[0] < 0 ? -strokeWidth : strokeWidth);
                    anchorY =
                      baseline * label.height +
                      ((0.5 - baseline) * 2 * strokeWidth * textScale[1]) /
                        textScale[0] -
                      offsetY;
                    const dimensions = this.calculateImageOrLabelDimensions_(
                      label.width,
                      label.height,
                      part[0],
                      part[1],
                      label.width,
                      label.height,
                      anchorX,
                      anchorY,
                      0,
                      0,
                      part[3],
                      pixelRatioScale,
                      false,
                      defaultPadding,
                      false,
                      feature
                    );
                    if (
                      declutterTree &&
                      declutterTree.collides(dimensions.declutterBox)
                    ) {
                      break drawChars;
                    }
                    replayImageOrLabelArgs.push([
                      context,
                      contextScale,
                      label,
                      dimensions,
                      1,
                      null,
                      null,
                    ]);
                  }
                }
                if (fillKey) {
                  for (c = 0, cc = parts.length; c < cc; ++c) {
                    part = parts[c]; // x, y, anchorX, rotation, chunk
                    chars = /** @type {string} */ (part[4]);
                    label = this.createLabel(chars, textKey, fillKey, '');
                    anchorX = /** @type {number} */ (part[2]);
                    anchorY = baseline * label.height - offsetY;
                    const dimensions = this.calculateImageOrLabelDimensions_(
                      label.width,
                      label.height,
                      part[0],
                      part[1],
                      label.width,
                      label.height,
                      anchorX,
                      anchorY,
                      0,
                      0,
                      part[3],
                      pixelRatioScale,
                      false,
                      defaultPadding,
                      false,
                      feature
                    );
                    if (
                      declutterTree &&
                      declutterTree.collides(dimensions.declutterBox)
                    ) {
                      break drawChars;
                    }
                    replayImageOrLabelArgs.push([
                      context,
                      contextScale,
                      label,
                      dimensions,
                      1,
                      null,
                      null,
                    ]);
                  }
                }
                if (declutterTree) {
                  declutterTree.load(replayImageOrLabelArgs.map(getDeclutterBox));
                }
                for (let i = 0, ii = replayImageOrLabelArgs.length; i < ii; ++i) {
                  this.replayImageOrLabel_.apply(this, replayImageOrLabelArgs[i]);
                }
              }
            }
            ++i;
            break;
          case Instruction.END_GEOMETRY:
            if (featureCallback !== undefined) {
              feature = /** @type {import("../../Feature.js").FeatureLike} */ (
                instruction[1]
              );
              const result = featureCallback(feature, currentGeometry);
              if (result) {
                return result;
              }
            }
            ++i;
            break;
          case Instruction.FILL:
            if (batchSize) {
              pendingFill++;
            } else {
              this.fill_(context);
            }
            ++i;
            break;
          case Instruction.MOVE_TO_LINE_TO:
            d = /** @type {number} */ (instruction[1]);
            dd = /** @type {number} */ (instruction[2]);
            x = pixelCoordinates[d];
            y = pixelCoordinates[d + 1];
            roundX = (x + 0.5) | 0;
            roundY = (y + 0.5) | 0;
            if (roundX !== prevX || roundY !== prevY) {
              context.moveTo(x, y);
              prevX = roundX;
              prevY = roundY;
            }
            for (d += 2; d < dd; d += 2) {
              x = pixelCoordinates[d];
              y = pixelCoordinates[d + 1];
              roundX = (x + 0.5) | 0;
              roundY = (y + 0.5) | 0;
              if (d == dd - 2 || roundX !== prevX || roundY !== prevY) {
                context.lineTo(x, y);
                prevX = roundX;
                prevY = roundY;
              }
            }
            ++i;
            break;
          case Instruction.SET_FILL_STYLE:
            lastFillInstruction = instruction;
            this.alignFill_ = instruction[2];

            if (pendingFill) {
              this.fill_(context);
              pendingFill = 0;
              if (pendingStroke) {
                context.stroke();
                pendingStroke = 0;
              }
            }

            context.fillStyle =
              /** @type {import("../../colorlike.js").ColorLike} */ (
                instruction[1]
              );
            ++i;
            break;
          case Instruction.SET_STROKE_STYLE:
            lastStrokeInstruction = instruction;
            if (pendingStroke) {
              context.stroke();
              pendingStroke = 0;
            }
            this.setStrokeStyle_(context, /** @type {Array<*>} */ (instruction));
            ++i;
            break;
          case Instruction.STROKE:
            if (batchSize) {
              pendingStroke++;
            } else {
              context.stroke();
            }
            ++i;
            break;
          default: // consume the instruction anyway, to avoid an infinite loop
            ++i;
            break;
        }
      }
      if (pendingFill) {
        this.fill_(context);
      }
      if (pendingStroke) {
        context.stroke();
      }
      return undefined;
    }

    /**
     * @param {CanvasRenderingContext2D} context Context.
     * @param {number} contextScale Scale of the context.
     * @param {import("../../transform.js").Transform} transform Transform.
     * @param {number} viewRotation View rotation.
     * @param {boolean} snapToPixel Snap point symbols and text to integer pixels.
     * @param {import("rbush").default} [declutterTree] Declutter tree.
     */
    execute(
      context,
      contextScale,
      transform,
      viewRotation,
      snapToPixel,
      declutterTree
    ) {
      this.viewRotation_ = viewRotation;
      this.execute_(
        context,
        contextScale,
        transform,
        this.instructions,
        snapToPixel,
        undefined,
        undefined,
        declutterTree
      );
    }

    /**
     * @param {CanvasRenderingContext2D} context Context.
     * @param {import("../../transform.js").Transform} transform Transform.
     * @param {number} viewRotation View rotation.
     * @param {FeatureCallback<T>} [featureCallback] Feature callback.
     * @param {import("../../extent.js").Extent} [hitExtent] Only check
     *     features that intersect this extent.
     * @return {T|undefined} Callback result.
     * @template T
     */
    executeHitDetection(
      context,
      transform,
      viewRotation,
      featureCallback,
      hitExtent
    ) {
      this.viewRotation_ = viewRotation;
      return this.execute_(
        context,
        1,
        transform,
        this.hitDetectionInstructions,
        true,
        featureCallback,
        hitExtent
      );
    }
  }

  /**
   * @module ol/render/canvas/ExecutorGroup
   */

  /**
   * @const
   * @type {Array<import("../canvas.js").BuilderType>}
   */
  const ORDER = ['Polygon', 'Circle', 'LineString', 'Image', 'Text', 'Default'];

  class ExecutorGroup {
    /**
     * @param {import("../../extent.js").Extent} maxExtent Max extent for clipping. When a
     * `maxExtent` was set on the Builder for this executor group, the same `maxExtent`
     * should be set here, unless the target context does not exceed that extent (which
     * can be the case when rendering to tiles).
     * @param {number} resolution Resolution.
     * @param {number} pixelRatio Pixel ratio.
     * @param {boolean} overlaps The executor group can have overlapping geometries.
     * @param {!Object<string, !Object<import("../canvas.js").BuilderType, import("../canvas.js").SerializableInstructions>>} allInstructions
     * The serializable instructions.
     * @param {number} [renderBuffer] Optional rendering buffer.
     */
    constructor(
      maxExtent,
      resolution,
      pixelRatio,
      overlaps,
      allInstructions,
      renderBuffer
    ) {
      /**
       * @private
       * @type {import("../../extent.js").Extent}
       */
      this.maxExtent_ = maxExtent;

      /**
       * @private
       * @type {boolean}
       */
      this.overlaps_ = overlaps;

      /**
       * @private
       * @type {number}
       */
      this.pixelRatio_ = pixelRatio;

      /**
       * @private
       * @type {number}
       */
      this.resolution_ = resolution;

      /**
       * @private
       * @type {number|undefined}
       */
      this.renderBuffer_ = renderBuffer;

      /**
       * @private
       * @type {!Object<string, !Object<import("../canvas.js").BuilderType, import("./Executor").default>>}
       */
      this.executorsByZIndex_ = {};

      /**
       * @private
       * @type {CanvasRenderingContext2D}
       */
      this.hitDetectionContext_ = null;

      /**
       * @private
       * @type {import("../../transform.js").Transform}
       */
      this.hitDetectionTransform_ = create$2();

      this.createExecutors_(allInstructions);
    }

    /**
     * @param {CanvasRenderingContext2D} context Context.
     * @param {import("../../transform.js").Transform} transform Transform.
     */
    clip(context, transform) {
      const flatClipCoords = this.getClipCoords(transform);
      context.beginPath();
      context.moveTo(flatClipCoords[0], flatClipCoords[1]);
      context.lineTo(flatClipCoords[2], flatClipCoords[3]);
      context.lineTo(flatClipCoords[4], flatClipCoords[5]);
      context.lineTo(flatClipCoords[6], flatClipCoords[7]);
      context.clip();
    }

    /**
     * Create executors and populate them using the provided instructions.
     * @private
     * @param {!Object<string, !Object<import("../canvas.js").BuilderType, import("../canvas.js").SerializableInstructions>>} allInstructions The serializable instructions
     */
    createExecutors_(allInstructions) {
      for (const zIndex in allInstructions) {
        let executors = this.executorsByZIndex_[zIndex];
        if (executors === undefined) {
          executors = {};
          this.executorsByZIndex_[zIndex] = executors;
        }
        const instructionByZindex = allInstructions[zIndex];
        for (const builderType in instructionByZindex) {
          const instructions = instructionByZindex[builderType];
          executors[builderType] = new Executor(
            this.resolution_,
            this.pixelRatio_,
            this.overlaps_,
            instructions
          );
        }
      }
    }

    /**
     * @param {Array<import("../canvas.js").BuilderType>} executors Executors.
     * @return {boolean} Has executors of the provided types.
     */
    hasExecutors(executors) {
      for (const zIndex in this.executorsByZIndex_) {
        const candidates = this.executorsByZIndex_[zIndex];
        for (let i = 0, ii = executors.length; i < ii; ++i) {
          if (executors[i] in candidates) {
            return true;
          }
        }
      }
      return false;
    }

    /**
     * @param {import("../../coordinate.js").Coordinate} coordinate Coordinate.
     * @param {number} resolution Resolution.
     * @param {number} rotation Rotation.
     * @param {number} hitTolerance Hit tolerance in pixels.
     * @param {function(import("../../Feature.js").FeatureLike, import("../../geom/SimpleGeometry.js").default, number): T} callback Feature callback.
     * @param {Array<import("../../Feature.js").FeatureLike>} declutteredFeatures Decluttered features.
     * @return {T|undefined} Callback result.
     * @template T
     */
    forEachFeatureAtCoordinate(
      coordinate,
      resolution,
      rotation,
      hitTolerance,
      callback,
      declutteredFeatures
    ) {
      hitTolerance = Math.round(hitTolerance);
      const contextSize = hitTolerance * 2 + 1;
      const transform = compose(
        this.hitDetectionTransform_,
        hitTolerance + 0.5,
        hitTolerance + 0.5,
        1 / resolution,
        -1 / resolution,
        -rotation,
        -coordinate[0],
        -coordinate[1]
      );

      const newContext = !this.hitDetectionContext_;
      if (newContext) {
        this.hitDetectionContext_ = createCanvasContext2D(
          contextSize,
          contextSize
        );
      }
      const context = this.hitDetectionContext_;

      if (
        context.canvas.width !== contextSize ||
        context.canvas.height !== contextSize
      ) {
        context.canvas.width = contextSize;
        context.canvas.height = contextSize;
      } else if (!newContext) {
        context.clearRect(0, 0, contextSize, contextSize);
      }

      /**
       * @type {import("../../extent.js").Extent}
       */
      let hitExtent;
      if (this.renderBuffer_ !== undefined) {
        hitExtent = createEmpty();
        extendCoordinate(hitExtent, coordinate);
        buffer$1(
          hitExtent,
          resolution * (this.renderBuffer_ + hitTolerance),
          hitExtent
        );
      }

      const indexes = getPixelIndexArray(hitTolerance);

      let builderType;

      /**
       * @param {import("../../Feature.js").FeatureLike} feature Feature.
       * @param {import("../../geom/SimpleGeometry.js").default} geometry Geometry.
       * @return {T|undefined} Callback result.
       */
      function featureCallback(feature, geometry) {
        const imageData = context.getImageData(
          0,
          0,
          contextSize,
          contextSize
        ).data;
        for (let i = 0, ii = indexes.length; i < ii; i++) {
          if (imageData[indexes[i]] > 0) {
            if (
              !declutteredFeatures ||
              (builderType !== 'Image' && builderType !== 'Text') ||
              declutteredFeatures.includes(feature)
            ) {
              const idx = (indexes[i] - 3) / 4;
              const x = hitTolerance - (idx % contextSize);
              const y = hitTolerance - ((idx / contextSize) | 0);
              const result = callback(feature, geometry, x * x + y * y);
              if (result) {
                return result;
              }
            }
            context.clearRect(0, 0, contextSize, contextSize);
            break;
          }
        }
        return undefined;
      }

      /** @type {Array<number>} */
      const zs = Object.keys(this.executorsByZIndex_).map(Number);
      zs.sort(numberSafeCompareFunction);

      let i, j, executors, executor, result;
      for (i = zs.length - 1; i >= 0; --i) {
        const zIndexKey = zs[i].toString();
        executors = this.executorsByZIndex_[zIndexKey];
        for (j = ORDER.length - 1; j >= 0; --j) {
          builderType = ORDER[j];
          executor = executors[builderType];
          if (executor !== undefined) {
            result = executor.executeHitDetection(
              context,
              transform,
              rotation,
              featureCallback,
              hitExtent
            );
            if (result) {
              return result;
            }
          }
        }
      }
      return undefined;
    }

    /**
     * @param {import("../../transform.js").Transform} transform Transform.
     * @return {Array<number>|null} Clip coordinates.
     */
    getClipCoords(transform) {
      const maxExtent = this.maxExtent_;
      if (!maxExtent) {
        return null;
      }
      const minX = maxExtent[0];
      const minY = maxExtent[1];
      const maxX = maxExtent[2];
      const maxY = maxExtent[3];
      const flatClipCoords = [minX, minY, minX, maxY, maxX, maxY, maxX, minY];
      transform2D(flatClipCoords, 0, 8, 2, transform, flatClipCoords);
      return flatClipCoords;
    }

    /**
     * @return {boolean} Is empty.
     */
    isEmpty() {
      return isEmpty$1(this.executorsByZIndex_);
    }

    /**
     * @param {CanvasRenderingContext2D} context Context.
     * @param {number} contextScale Scale of the context.
     * @param {import("../../transform.js").Transform} transform Transform.
     * @param {number} viewRotation View rotation.
     * @param {boolean} snapToPixel Snap point symbols and test to integer pixel.
     * @param {Array<import("../canvas.js").BuilderType>} [builderTypes] Ordered replay types to replay.
     *     Default is {@link module:ol/render/replay~ORDER}
     * @param {import("rbush").default} [declutterTree] Declutter tree.
     */
    execute(
      context,
      contextScale,
      transform,
      viewRotation,
      snapToPixel,
      builderTypes,
      declutterTree
    ) {
      /** @type {Array<number>} */
      const zs = Object.keys(this.executorsByZIndex_).map(Number);
      zs.sort(numberSafeCompareFunction);

      // setup clipping so that the parts of over-simplified geometries are not
      // visible outside the current extent when panning
      if (this.maxExtent_) {
        context.save();
        this.clip(context, transform);
      }

      builderTypes = builderTypes ? builderTypes : ORDER;
      let i, ii, j, jj, replays, replay;
      if (declutterTree) {
        zs.reverse();
      }
      for (i = 0, ii = zs.length; i < ii; ++i) {
        const zIndexKey = zs[i].toString();
        replays = this.executorsByZIndex_[zIndexKey];
        for (j = 0, jj = builderTypes.length; j < jj; ++j) {
          const builderType = builderTypes[j];
          replay = replays[builderType];
          if (replay !== undefined) {
            replay.execute(
              context,
              contextScale,
              transform,
              viewRotation,
              snapToPixel,
              declutterTree
            );
          }
        }
      }

      if (this.maxExtent_) {
        context.restore();
      }
    }
  }

  /**
   * This cache is used to store arrays of indexes for calculated pixel circles
   * to increase performance.
   * It is a static property to allow each Replaygroup to access it.
   * @type {Object<number, Array<number>>}
   */
  const circlePixelIndexArrayCache = {};

  /**
   * This methods creates an array with indexes of all pixels within a circle,
   * ordered by how close they are to the center.
   * A cache is used to increase performance.
   * @param {number} radius Radius.
   * @return {Array<number>} An array with indexes within a circle.
   */
  function getPixelIndexArray(radius) {
    if (circlePixelIndexArrayCache[radius] !== undefined) {
      return circlePixelIndexArrayCache[radius];
    }

    const size = radius * 2 + 1;
    const maxDistanceSq = radius * radius;
    const distances = new Array(maxDistanceSq + 1);
    for (let i = 0; i <= radius; ++i) {
      for (let j = 0; j <= radius; ++j) {
        const distanceSq = i * i + j * j;
        if (distanceSq > maxDistanceSq) {
          break;
        }
        let distance = distances[distanceSq];
        if (!distance) {
          distance = [];
          distances[distanceSq] = distance;
        }
        distance.push(((radius + i) * size + (radius + j)) * 4 + 3);
        if (i > 0) {
          distance.push(((radius - i) * size + (radius + j)) * 4 + 3);
        }
        if (j > 0) {
          distance.push(((radius + i) * size + (radius - j)) * 4 + 3);
          if (i > 0) {
            distance.push(((radius - i) * size + (radius - j)) * 4 + 3);
          }
        }
      }
    }

    const pixelIndex = [];
    for (let i = 0, ii = distances.length; i < ii; ++i) {
      if (distances[i]) {
        pixelIndex.push(...distances[i]);
      }
    }

    circlePixelIndexArrayCache[radius] = pixelIndex;
    return pixelIndex;
  }

  /**
   * @module ol/render/canvas/hitdetect
   */

  const HIT_DETECT_RESOLUTION = 0.5;

  /**
   * @param {import("../../size.js").Size} size Canvas size in css pixels.
   * @param {Array<import("../../transform.js").Transform>} transforms Transforms
   * for rendering features to all worlds of the viewport, from coordinates to css
   * pixels.
   * @param {Array<import("../../Feature.js").FeatureLike>} features
   * Features to consider for hit detection.
   * @param {import("../../style/Style.js").StyleFunction|undefined} styleFunction
   * Layer style function.
   * @param {import("../../extent.js").Extent} extent Extent.
   * @param {number} resolution Resolution.
   * @param {number} rotation Rotation.
   * @return {ImageData} Hit detection image data.
   */
  function createHitDetectionImageData(
    size,
    transforms,
    features,
    styleFunction,
    extent,
    resolution,
    rotation
  ) {
    const width = size[0] * HIT_DETECT_RESOLUTION;
    const height = size[1] * HIT_DETECT_RESOLUTION;
    const context = createCanvasContext2D(width, height);
    context.imageSmoothingEnabled = false;
    const canvas = context.canvas;
    const renderer = new CanvasImmediateRenderer(
      context,
      HIT_DETECT_RESOLUTION,
      extent,
      null,
      rotation
    );
    const featureCount = features.length;
    // Stretch hit detection index to use the whole available color range
    const indexFactor = Math.floor((256 * 256 * 256 - 1) / featureCount);
    const featuresByZIndex = {};
    for (let i = 1; i <= featureCount; ++i) {
      const feature = features[i - 1];
      const featureStyleFunction = feature.getStyleFunction() || styleFunction;
      if (!styleFunction) {
        continue;
      }
      let styles = featureStyleFunction(feature, resolution);
      if (!styles) {
        continue;
      }
      if (!Array.isArray(styles)) {
        styles = [styles];
      }
      const index = i * indexFactor;
      const color = '#' + ('000000' + index.toString(16)).slice(-6);
      for (let j = 0, jj = styles.length; j < jj; ++j) {
        const originalStyle = styles[j];
        const geometry = originalStyle.getGeometryFunction()(feature);
        if (!geometry || !intersects$3(extent, geometry.getExtent())) {
          continue;
        }
        const style = originalStyle.clone();
        const fill = style.getFill();
        if (fill) {
          fill.setColor(color);
        }
        const stroke = style.getStroke();
        if (stroke) {
          stroke.setColor(color);
          stroke.setLineDash(null);
        }
        style.setText(undefined);
        const image = originalStyle.getImage();
        if (image && image.getOpacity() !== 0) {
          const imgSize = image.getImageSize();
          if (!imgSize) {
            continue;
          }

          const imgContext = createCanvasContext2D(
            imgSize[0],
            imgSize[1],
            undefined,
            {alpha: false}
          );
          const img = imgContext.canvas;
          imgContext.fillStyle = color;
          imgContext.fillRect(0, 0, img.width, img.height);
          style.setImage(
            new Icon({
              img: img,
              imgSize: imgSize,
              anchor: image.getAnchor(),
              anchorXUnits: 'pixels',
              anchorYUnits: 'pixels',
              offset: image.getOrigin(),
              opacity: 1,
              size: image.getSize(),
              scale: image.getScale(),
              rotation: image.getRotation(),
              rotateWithView: image.getRotateWithView(),
            })
          );
        }
        const zIndex = style.getZIndex() || 0;
        let byGeometryType = featuresByZIndex[zIndex];
        if (!byGeometryType) {
          byGeometryType = {};
          featuresByZIndex[zIndex] = byGeometryType;
          byGeometryType['Polygon'] = [];
          byGeometryType['Circle'] = [];
          byGeometryType['LineString'] = [];
          byGeometryType['Point'] = [];
        }
        byGeometryType[geometry.getType().replace('Multi', '')].push(
          geometry,
          style
        );
      }
    }

    const zIndexKeys = Object.keys(featuresByZIndex)
      .map(Number)
      .sort(numberSafeCompareFunction);
    for (let i = 0, ii = zIndexKeys.length; i < ii; ++i) {
      const byGeometryType = featuresByZIndex[zIndexKeys[i]];
      for (const type in byGeometryType) {
        const geomAndStyle = byGeometryType[type];
        for (let j = 0, jj = geomAndStyle.length; j < jj; j += 2) {
          renderer.setStyle(geomAndStyle[j + 1]);
          for (let k = 0, kk = transforms.length; k < kk; ++k) {
            renderer.setTransform(transforms[k]);
            renderer.drawGeometry(geomAndStyle[j]);
          }
        }
      }
    }
    return context.getImageData(0, 0, canvas.width, canvas.height);
  }

  /**
   * @param {import("../../pixel").Pixel} pixel Pixel coordinate on the hit
   * detection canvas in css pixels.
   * @param {Array<import("../../Feature").FeatureLike>} features Features. Has to
   * match the `features` array that was passed to `createHitDetectionImageData()`.
   * @param {ImageData} imageData Hit detection image data generated by
   * `createHitDetectionImageData()`.
   * @return {Array<import("../../Feature").FeatureLike>} features Features.
   */
  function hitDetect(pixel, features, imageData) {
    const resultFeatures = [];
    if (imageData) {
      const x = Math.floor(Math.round(pixel[0]) * HIT_DETECT_RESOLUTION);
      const y = Math.floor(Math.round(pixel[1]) * HIT_DETECT_RESOLUTION);
      // The pixel coordinate is clamped down to the hit-detect canvas' size to account
      // for browsers returning coordinates slightly larger than the actual canvas size
      // due to a non-integer pixel ratio.
      const index =
        (clamp(x, 0, imageData.width - 1) +
          clamp(y, 0, imageData.height - 1) * imageData.width) *
        4;
      const r = imageData.data[index];
      const g = imageData.data[index + 1];
      const b = imageData.data[index + 2];
      const i = b + 256 * (g + 256 * r);
      const indexFactor = Math.floor((256 * 256 * 256 - 1) / features.length);
      if (i && i % indexFactor === 0) {
        resultFeatures.push(features[i / indexFactor - 1]);
      }
    }
    return resultFeatures;
  }

  /**
   * @module ol/renderer/canvas/VectorLayer
   */

  /**
   * @classdesc
   * Canvas renderer for vector layers.
   * @api
   */
  class CanvasVectorLayerRenderer extends CanvasLayerRenderer {
    /**
     * @param {import("../../layer/BaseVector.js").default} vectorLayer Vector layer.
     */
    constructor(vectorLayer) {
      super(vectorLayer);

      /** @private */
      this.boundHandleStyleImageChange_ = this.handleStyleImageChange_.bind(this);

      /**
       * @type {boolean}
       */
      this.animatingOrInteracting_;

      /**
       * @type {ImageData}
       */
      this.hitDetectionImageData_ = null;

      /**
       * @type {Array<import("../../Feature.js").default>}
       */
      this.renderedFeatures_ = null;

      /**
       * @private
       * @type {number}
       */
      this.renderedRevision_ = -1;

      /**
       * @private
       * @type {number}
       */
      this.renderedResolution_ = NaN;

      /**
       * @private
       * @type {import("../../extent.js").Extent}
       */
      this.renderedExtent_ = createEmpty();

      /**
       * @private
       * @type {import("../../extent.js").Extent}
       */
      this.wrappedRenderedExtent_ = createEmpty();

      /**
       * @private
       * @type {number}
       */
      this.renderedRotation_;

      /**
       * @private
       * @type {import("../../coordinate").Coordinate}
       */
      this.renderedCenter_ = null;

      /**
       * @private
       * @type {import("../../proj/Projection").default}
       */
      this.renderedProjection_ = null;

      /**
       * @private
       * @type {function(import("../../Feature.js").default, import("../../Feature.js").default): number|null}
       */
      this.renderedRenderOrder_ = null;

      /**
       * @private
       * @type {import("../../render/canvas/ExecutorGroup").default}
       */
      this.replayGroup_ = null;

      /**
       * A new replay group had to be created by `prepareFrame()`
       * @type {boolean}
       */
      this.replayGroupChanged = true;

      /**
       * @type {import("../../render/canvas/ExecutorGroup").default}
       */
      this.declutterExecutorGroup = null;

      /**
       * Clipping to be performed by `renderFrame()`
       * @type {boolean}
       */
      this.clipping = true;

      /**
       * @private
       * @type {CanvasRenderingContext2D}
       */
      this.compositionContext_ = null;

      /**
       * @private
       * @type {number}
       */
      this.opacity_ = 1;
    }

    /**
     * @param {ExecutorGroup} executorGroup Executor group.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @param {import("rbush").default} [declutterTree] Declutter tree.
     */
    renderWorlds(executorGroup, frameState, declutterTree) {
      const extent = frameState.extent;
      const viewState = frameState.viewState;
      const center = viewState.center;
      const resolution = viewState.resolution;
      const projection = viewState.projection;
      const rotation = viewState.rotation;
      const projectionExtent = projection.getExtent();
      const vectorSource = this.getLayer().getSource();
      const pixelRatio = frameState.pixelRatio;
      const viewHints = frameState.viewHints;
      const snapToPixel = !(
        viewHints[ViewHint.ANIMATING] || viewHints[ViewHint.INTERACTING]
      );
      const context = this.compositionContext_;
      const width = Math.round(frameState.size[0] * pixelRatio);
      const height = Math.round(frameState.size[1] * pixelRatio);

      const multiWorld = vectorSource.getWrapX() && projection.canWrapX();
      const worldWidth = multiWorld ? getWidth(projectionExtent) : null;
      const endWorld = multiWorld
        ? Math.ceil((extent[2] - projectionExtent[2]) / worldWidth) + 1
        : 1;
      let world = multiWorld
        ? Math.floor((extent[0] - projectionExtent[0]) / worldWidth)
        : 0;
      do {
        const transform = this.getRenderTransform(
          center,
          resolution,
          rotation,
          pixelRatio,
          width,
          height,
          world * worldWidth
        );
        executorGroup.execute(
          context,
          1,
          transform,
          rotation,
          snapToPixel,
          undefined,
          declutterTree
        );
      } while (++world < endWorld);
    }

    setupCompositionContext_() {
      if (this.opacity_ !== 1) {
        const compositionContext = createCanvasContext2D(
          this.context.canvas.width,
          this.context.canvas.height,
          canvasPool
        );
        this.compositionContext_ = compositionContext;
      } else {
        this.compositionContext_ = this.context;
      }
    }

    releaseCompositionContext_() {
      if (this.opacity_ !== 1) {
        const alpha = this.context.globalAlpha;
        this.context.globalAlpha = this.opacity_;
        this.context.drawImage(this.compositionContext_.canvas, 0, 0);
        this.context.globalAlpha = alpha;
        releaseCanvas$1(this.compositionContext_);
        canvasPool.push(this.compositionContext_.canvas);
        this.compositionContext_ = null;
      }
    }

    /**
     * Render declutter items for this layer
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     */
    renderDeclutter(frameState) {
      if (this.declutterExecutorGroup) {
        this.setupCompositionContext_();
        this.renderWorlds(
          this.declutterExecutorGroup,
          frameState,
          frameState.declutterTree
        );
        this.releaseCompositionContext_();
      }
    }

    /**
     * Render the layer.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @param {HTMLElement} target Target that may be used to render content to.
     * @return {HTMLElement} The rendered element.
     */
    renderFrame(frameState, target) {
      const pixelRatio = frameState.pixelRatio;
      const layerState = frameState.layerStatesArray[frameState.layerIndex];

      // set forward and inverse pixel transforms
      makeScale(this.pixelTransform, 1 / pixelRatio, 1 / pixelRatio);
      makeInverse(this.inversePixelTransform, this.pixelTransform);

      const canvasTransform = toString$3(this.pixelTransform);

      this.useContainer(target, canvasTransform, this.getBackground(frameState));
      const context = this.context;
      const canvas = context.canvas;

      const replayGroup = this.replayGroup_;
      const declutterExecutorGroup = this.declutterExecutorGroup;
      if (
        (!replayGroup || replayGroup.isEmpty()) &&
        (!declutterExecutorGroup || declutterExecutorGroup.isEmpty())
      ) {
        return null;
      }

      // resize and clear
      const width = Math.round(frameState.size[0] * pixelRatio);
      const height = Math.round(frameState.size[1] * pixelRatio);
      if (canvas.width != width || canvas.height != height) {
        canvas.width = width;
        canvas.height = height;
        if (canvas.style.transform !== canvasTransform) {
          canvas.style.transform = canvasTransform;
        }
      } else if (!this.containerReused) {
        context.clearRect(0, 0, width, height);
      }

      this.preRender(context, frameState);

      const viewState = frameState.viewState;
      const projection = viewState.projection;

      this.opacity_ = layerState.opacity;
      this.setupCompositionContext_();

      // clipped rendering if layer extent is set
      let clipped = false;
      let render = true;
      if (layerState.extent && this.clipping) {
        const layerExtent = fromUserExtent(layerState.extent, projection);
        render = intersects$3(layerExtent, frameState.extent);
        clipped = render && !containsExtent(layerExtent, frameState.extent);
        if (clipped) {
          this.clipUnrotated(this.compositionContext_, frameState, layerExtent);
        }
      }

      if (render) {
        this.renderWorlds(replayGroup, frameState);
      }

      if (clipped) {
        this.compositionContext_.restore();
      }

      this.releaseCompositionContext_();

      this.postRender(context, frameState);

      if (this.renderedRotation_ !== viewState.rotation) {
        this.renderedRotation_ = viewState.rotation;
        this.hitDetectionImageData_ = null;
      }
      return this.container;
    }

    /**
     * Asynchronous layer level hit detection.
     * @param {import("../../pixel.js").Pixel} pixel Pixel.
     * @return {Promise<Array<import("../../Feature").default>>} Promise that resolves with an array of features.
     */
    getFeatures(pixel) {
      return new Promise(
        /**
         * @param {function(Array<import("../../Feature").default|import("../../render/Feature").default>): void} resolve Resolver function.
         * @this {CanvasVectorLayerRenderer}
         */
        function (resolve) {
          if (!this.hitDetectionImageData_ && !this.animatingOrInteracting_) {
            const size = [this.context.canvas.width, this.context.canvas.height];
            apply(this.pixelTransform, size);
            const center = this.renderedCenter_;
            const resolution = this.renderedResolution_;
            const rotation = this.renderedRotation_;
            const projection = this.renderedProjection_;
            const extent = this.wrappedRenderedExtent_;
            const layer = this.getLayer();
            const transforms = [];
            const width = size[0] * HIT_DETECT_RESOLUTION;
            const height = size[1] * HIT_DETECT_RESOLUTION;
            transforms.push(
              this.getRenderTransform(
                center,
                resolution,
                rotation,
                HIT_DETECT_RESOLUTION,
                width,
                height,
                0
              ).slice()
            );
            const source = layer.getSource();
            const projectionExtent = projection.getExtent();
            if (
              source.getWrapX() &&
              projection.canWrapX() &&
              !containsExtent(projectionExtent, extent)
            ) {
              let startX = extent[0];
              const worldWidth = getWidth(projectionExtent);
              let world = 0;
              let offsetX;
              while (startX < projectionExtent[0]) {
                --world;
                offsetX = worldWidth * world;
                transforms.push(
                  this.getRenderTransform(
                    center,
                    resolution,
                    rotation,
                    HIT_DETECT_RESOLUTION,
                    width,
                    height,
                    offsetX
                  ).slice()
                );
                startX += worldWidth;
              }
              world = 0;
              startX = extent[2];
              while (startX > projectionExtent[2]) {
                ++world;
                offsetX = worldWidth * world;
                transforms.push(
                  this.getRenderTransform(
                    center,
                    resolution,
                    rotation,
                    HIT_DETECT_RESOLUTION,
                    width,
                    height,
                    offsetX
                  ).slice()
                );
                startX -= worldWidth;
              }
            }

            this.hitDetectionImageData_ = createHitDetectionImageData(
              size,
              transforms,
              this.renderedFeatures_,
              layer.getStyleFunction(),
              extent,
              resolution,
              rotation
            );
          }
          resolve(
            hitDetect(pixel, this.renderedFeatures_, this.hitDetectionImageData_)
          );
        }.bind(this)
      );
    }

    /**
     * @param {import("../../coordinate.js").Coordinate} coordinate Coordinate.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @param {number} hitTolerance Hit tolerance in pixels.
     * @param {import("../vector.js").FeatureCallback<T>} callback Feature callback.
     * @param {Array<import("../Map.js").HitMatch<T>>} matches The hit detected matches with tolerance.
     * @return {T|undefined} Callback result.
     * @template T
     */
    forEachFeatureAtCoordinate(
      coordinate,
      frameState,
      hitTolerance,
      callback,
      matches
    ) {
      if (!this.replayGroup_) {
        return undefined;
      }
      const resolution = frameState.viewState.resolution;
      const rotation = frameState.viewState.rotation;
      const layer = this.getLayer();

      /** @type {!Object<string, import("../Map.js").HitMatch<T>|true>} */
      const features = {};

      /**
       * @param {import("../../Feature.js").FeatureLike} feature Feature.
       * @param {import("../../geom/SimpleGeometry.js").default} geometry Geometry.
       * @param {number} distanceSq The squared distance to the click position
       * @return {T|undefined} Callback result.
       */
      const featureCallback = function (feature, geometry, distanceSq) {
        const key = getUid(feature);
        const match = features[key];
        if (!match) {
          if (distanceSq === 0) {
            features[key] = true;
            return callback(feature, layer, geometry);
          }
          matches.push(
            (features[key] = {
              feature: feature,
              layer: layer,
              geometry: geometry,
              distanceSq: distanceSq,
              callback: callback,
            })
          );
        } else if (match !== true && distanceSq < match.distanceSq) {
          if (distanceSq === 0) {
            features[key] = true;
            matches.splice(matches.lastIndexOf(match), 1);
            return callback(feature, layer, geometry);
          }
          match.geometry = geometry;
          match.distanceSq = distanceSq;
        }
        return undefined;
      };

      let result;
      const executorGroups = [this.replayGroup_];
      if (this.declutterExecutorGroup) {
        executorGroups.push(this.declutterExecutorGroup);
      }
      executorGroups.some((executorGroup) => {
        return (result = executorGroup.forEachFeatureAtCoordinate(
          coordinate,
          resolution,
          rotation,
          hitTolerance,
          featureCallback,
          executorGroup === this.declutterExecutorGroup &&
            frameState.declutterTree
            ? frameState.declutterTree.all().map((item) => item.value)
            : null
        ));
      });

      return result;
    }

    /**
     * Perform action necessary to get the layer rendered after new fonts have loaded
     */
    handleFontsChanged() {
      const layer = this.getLayer();
      if (layer.getVisible() && this.replayGroup_) {
        layer.changed();
      }
    }

    /**
     * Handle changes in image style state.
     * @param {import("../../events/Event.js").default} event Image style change event.
     * @private
     */
    handleStyleImageChange_(event) {
      this.renderIfReadyAndVisible();
    }

    /**
     * Determine whether render should be called.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @return {boolean} Layer is ready to be rendered.
     */
    prepareFrame(frameState) {
      const vectorLayer = this.getLayer();
      const vectorSource = vectorLayer.getSource();
      if (!vectorSource) {
        return false;
      }

      const animating = frameState.viewHints[ViewHint.ANIMATING];
      const interacting = frameState.viewHints[ViewHint.INTERACTING];
      const updateWhileAnimating = vectorLayer.getUpdateWhileAnimating();
      const updateWhileInteracting = vectorLayer.getUpdateWhileInteracting();

      if (
        (this.ready && !updateWhileAnimating && animating) ||
        (!updateWhileInteracting && interacting)
      ) {
        this.animatingOrInteracting_ = true;
        return true;
      }
      this.animatingOrInteracting_ = false;

      const frameStateExtent = frameState.extent;
      const viewState = frameState.viewState;
      const projection = viewState.projection;
      const resolution = viewState.resolution;
      const pixelRatio = frameState.pixelRatio;
      const vectorLayerRevision = vectorLayer.getRevision();
      const vectorLayerRenderBuffer = vectorLayer.getRenderBuffer();
      let vectorLayerRenderOrder = vectorLayer.getRenderOrder();

      if (vectorLayerRenderOrder === undefined) {
        vectorLayerRenderOrder = defaultOrder;
      }

      const center = viewState.center.slice();
      const extent = buffer$1(
        frameStateExtent,
        vectorLayerRenderBuffer * resolution
      );
      const renderedExtent = extent.slice();
      const loadExtents = [extent.slice()];
      const projectionExtent = projection.getExtent();

      if (
        vectorSource.getWrapX() &&
        projection.canWrapX() &&
        !containsExtent(projectionExtent, frameState.extent)
      ) {
        // For the replay group, we need an extent that intersects the real world
        // (-180° to +180°). To support geometries in a coordinate range from -540°
        // to +540°, we add at least 1 world width on each side of the projection
        // extent. If the viewport is wider than the world, we need to add half of
        // the viewport width to make sure we cover the whole viewport.
        const worldWidth = getWidth(projectionExtent);
        const gutter = Math.max(getWidth(extent) / 2, worldWidth);
        extent[0] = projectionExtent[0] - gutter;
        extent[2] = projectionExtent[2] + gutter;
        wrapX$1(center, projection);
        const loadExtent = wrapX$2(loadExtents[0], projection);
        // If the extent crosses the date line, we load data for both edges of the worlds
        if (
          loadExtent[0] < projectionExtent[0] &&
          loadExtent[2] < projectionExtent[2]
        ) {
          loadExtents.push([
            loadExtent[0] + worldWidth,
            loadExtent[1],
            loadExtent[2] + worldWidth,
            loadExtent[3],
          ]);
        } else if (
          loadExtent[0] > projectionExtent[0] &&
          loadExtent[2] > projectionExtent[2]
        ) {
          loadExtents.push([
            loadExtent[0] - worldWidth,
            loadExtent[1],
            loadExtent[2] - worldWidth,
            loadExtent[3],
          ]);
        }
      }

      if (
        this.ready &&
        this.renderedResolution_ == resolution &&
        this.renderedRevision_ == vectorLayerRevision &&
        this.renderedRenderOrder_ == vectorLayerRenderOrder &&
        containsExtent(this.wrappedRenderedExtent_, extent)
      ) {
        if (!equals$3(this.renderedExtent_, renderedExtent)) {
          this.hitDetectionImageData_ = null;
          this.renderedExtent_ = renderedExtent;
        }
        this.renderedCenter_ = center;
        this.replayGroupChanged = false;
        return true;
      }

      this.replayGroup_ = null;

      const replayGroup = new BuilderGroup(
        getTolerance(resolution, pixelRatio),
        extent,
        resolution,
        pixelRatio
      );

      let declutterBuilderGroup;
      if (this.getLayer().getDeclutter()) {
        declutterBuilderGroup = new BuilderGroup(
          getTolerance(resolution, pixelRatio),
          extent,
          resolution,
          pixelRatio
        );
      }

      const userProjection = getUserProjection();
      let userTransform;
      if (userProjection) {
        for (let i = 0, ii = loadExtents.length; i < ii; ++i) {
          const extent = loadExtents[i];
          const userExtent = toUserExtent(extent, projection);
          vectorSource.loadFeatures(
            userExtent,
            toUserResolution(resolution, projection),
            userProjection
          );
        }
        userTransform = getTransformFromProjections(userProjection, projection);
      } else {
        for (let i = 0, ii = loadExtents.length; i < ii; ++i) {
          vectorSource.loadFeatures(loadExtents[i], resolution, projection);
        }
      }

      const squaredTolerance = getSquaredTolerance(resolution, pixelRatio);
      let ready = true;
      const render =
        /**
         * @param {import("../../Feature.js").default} feature Feature.
         * @this {CanvasVectorLayerRenderer}
         */
        function (feature) {
          let styles;
          const styleFunction =
            feature.getStyleFunction() || vectorLayer.getStyleFunction();
          if (styleFunction) {
            styles = styleFunction(feature, resolution);
          }
          if (styles) {
            const dirty = this.renderFeature(
              feature,
              squaredTolerance,
              styles,
              replayGroup,
              userTransform,
              declutterBuilderGroup
            );
            ready = ready && !dirty;
          }
        }.bind(this);

      const userExtent = toUserExtent(extent, projection);
      /** @type {Array<import("../../Feature.js").default>} */
      const features = vectorSource.getFeaturesInExtent(userExtent);
      if (vectorLayerRenderOrder) {
        features.sort(vectorLayerRenderOrder);
      }
      for (let i = 0, ii = features.length; i < ii; ++i) {
        render(features[i]);
      }
      this.renderedFeatures_ = features;
      this.ready = ready;

      const replayGroupInstructions = replayGroup.finish();
      const executorGroup = new ExecutorGroup(
        extent,
        resolution,
        pixelRatio,
        vectorSource.getOverlaps(),
        replayGroupInstructions,
        vectorLayer.getRenderBuffer()
      );

      if (declutterBuilderGroup) {
        this.declutterExecutorGroup = new ExecutorGroup(
          extent,
          resolution,
          pixelRatio,
          vectorSource.getOverlaps(),
          declutterBuilderGroup.finish(),
          vectorLayer.getRenderBuffer()
        );
      }

      this.renderedResolution_ = resolution;
      this.renderedRevision_ = vectorLayerRevision;
      this.renderedRenderOrder_ = vectorLayerRenderOrder;
      this.renderedExtent_ = renderedExtent;
      this.wrappedRenderedExtent_ = extent;
      this.renderedCenter_ = center;
      this.renderedProjection_ = projection;
      this.replayGroup_ = executorGroup;
      this.hitDetectionImageData_ = null;

      this.replayGroupChanged = true;
      return true;
    }

    /**
     * @param {import("../../Feature.js").default} feature Feature.
     * @param {number} squaredTolerance Squared render tolerance.
     * @param {import("../../style/Style.js").default|Array<import("../../style/Style.js").default>} styles The style or array of styles.
     * @param {import("../../render/canvas/BuilderGroup.js").default} builderGroup Builder group.
     * @param {import("../../proj.js").TransformFunction} [transform] Transform from user to view projection.
     * @param {import("../../render/canvas/BuilderGroup.js").default} [declutterBuilderGroup] Builder for decluttering.
     * @return {boolean} `true` if an image is loading.
     */
    renderFeature(
      feature,
      squaredTolerance,
      styles,
      builderGroup,
      transform,
      declutterBuilderGroup
    ) {
      if (!styles) {
        return false;
      }
      let loading = false;
      if (Array.isArray(styles)) {
        for (let i = 0, ii = styles.length; i < ii; ++i) {
          loading =
            renderFeature$1(
              builderGroup,
              feature,
              styles[i],
              squaredTolerance,
              this.boundHandleStyleImageChange_,
              transform,
              declutterBuilderGroup
            ) || loading;
        }
      } else {
        loading = renderFeature$1(
          builderGroup,
          feature,
          styles,
          squaredTolerance,
          this.boundHandleStyleImageChange_,
          transform,
          declutterBuilderGroup
        );
      }
      return loading;
    }
  }

  /**
   * @module ol/renderer/canvas/VectorImageLayer
   */

  /**
   * @classdesc
   * Canvas renderer for image layers.
   * @api
   */
  class CanvasVectorImageLayerRenderer extends CanvasImageLayerRenderer {
    /**
     * @param {import("../../layer/VectorImage.js").default} layer Vector image layer.
     */
    constructor(layer) {
      super(layer);

      /**
       * @private
       * @type {import("./VectorLayer.js").default}
       */
      this.vectorRenderer_ = new CanvasVectorLayerRenderer(layer);

      /**
       * @private
       * @type {number}
       */
      this.layerImageRatio_ = layer.getImageRatio();

      /**
       * @private
       * @type {import("../../transform.js").Transform}
       */
      this.coordinateToVectorPixelTransform_ = create$2();

      /**
       * @private
       * @type {import("../../transform.js").Transform}
       */
      this.renderedPixelToCoordinateTransform_ = null;
    }

    /**
     * Clean up.
     */
    disposeInternal() {
      this.vectorRenderer_.dispose();
      super.disposeInternal();
    }

    /**
     * Asynchronous layer level hit detection.
     * @param {import("../../pixel.js").Pixel} pixel Pixel.
     * @return {Promise<Array<import("../../Feature").default>>} Promise that resolves with an array of features.
     */
    getFeatures(pixel) {
      if (!this.vectorRenderer_) {
        return new Promise((resolve) => resolve([]));
      }
      const vectorPixel = apply(
        this.coordinateToVectorPixelTransform_,
        apply(this.renderedPixelToCoordinateTransform_, pixel.slice())
      );
      return this.vectorRenderer_.getFeatures(vectorPixel);
    }

    /**
     * Perform action necessary to get the layer rendered after new fonts have loaded
     */
    handleFontsChanged() {
      this.vectorRenderer_.handleFontsChanged();
    }

    /**
     * Determine whether render should be called.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @return {boolean} Layer is ready to be rendered.
     */
    prepareFrame(frameState) {
      const pixelRatio = frameState.pixelRatio;
      const viewState = frameState.viewState;
      const viewResolution = viewState.resolution;

      const hints = frameState.viewHints;
      const vectorRenderer = this.vectorRenderer_;
      let renderedExtent = frameState.extent;
      if (this.layerImageRatio_ !== 1) {
        renderedExtent = renderedExtent.slice(0);
        scaleFromCenter(renderedExtent, this.layerImageRatio_);
      }
      const width = getWidth(renderedExtent) / viewResolution;
      const height = getHeight(renderedExtent) / viewResolution;

      if (
        !hints[ViewHint.ANIMATING] &&
        !hints[ViewHint.INTERACTING] &&
        !isEmpty(renderedExtent)
      ) {
        vectorRenderer.useContainer(null, null);
        const context = vectorRenderer.context;
        const layerState = frameState.layerStatesArray[frameState.layerIndex];
        context.globalAlpha = layerState.opacity;
        const imageLayerState = Object.assign({}, layerState, {opacity: 1});
        const imageFrameState = /** @type {import("../../Map.js").FrameState} */ (
          Object.assign({}, frameState, {
            declutterTree: new RBush$1(9),
            extent: renderedExtent,
            size: [width, height],
            viewState: /** @type {import("../../View.js").State} */ (
              Object.assign({}, frameState.viewState, {
                rotation: 0,
              })
            ),
            layerStatesArray: [imageLayerState],
            layerIndex: 0,
          })
        );
        let emptyImage = true;
        const image = new ImageCanvas(
          renderedExtent,
          viewResolution,
          pixelRatio,
          context.canvas,
          function (callback) {
            if (
              vectorRenderer.prepareFrame(imageFrameState) &&
              vectorRenderer.replayGroupChanged
            ) {
              vectorRenderer.clipping = false;
              if (vectorRenderer.renderFrame(imageFrameState, null)) {
                vectorRenderer.renderDeclutter(imageFrameState);
                emptyImage = false;
              }
              callback();
            }
          }
        );

        image.addEventListener(
          EventType$1.CHANGE,
          function () {
            if (image.getState() !== ImageState.LOADED) {
              return;
            }
            this.image_ = emptyImage ? null : image;
            const imageResolution = image.getResolution();
            const imagePixelRatio = image.getPixelRatio();
            const renderedResolution =
              (imageResolution * pixelRatio) / imagePixelRatio;
            this.renderedResolution = renderedResolution;
            this.coordinateToVectorPixelTransform_ = compose(
              this.coordinateToVectorPixelTransform_,
              width / 2,
              height / 2,
              1 / renderedResolution,
              -1 / renderedResolution,
              0,
              -viewState.center[0],
              -viewState.center[1]
            );
          }.bind(this)
        );
        image.load();
      }

      if (this.image_) {
        this.renderedPixelToCoordinateTransform_ =
          frameState.pixelToCoordinateTransform.slice();
      }

      return !!this.image_;
    }

    /**
     */
    preRender() {}

    /**
     */
    postRender() {}

    /**
     */
    renderDeclutter() {}

    /**
     * @param {import("../../coordinate.js").Coordinate} coordinate Coordinate.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @param {number} hitTolerance Hit tolerance in pixels.
     * @param {import("../vector.js").FeatureCallback<T>} callback Feature callback.
     * @param {Array<import("../Map.js").HitMatch<T>>} matches The hit detected matches with tolerance.
     * @return {T|undefined} Callback result.
     * @template T
     */
    forEachFeatureAtCoordinate(
      coordinate,
      frameState,
      hitTolerance,
      callback,
      matches
    ) {
      if (this.vectorRenderer_) {
        return this.vectorRenderer_.forEachFeatureAtCoordinate(
          coordinate,
          frameState,
          hitTolerance,
          callback,
          matches
        );
      } else {
        return super.forEachFeatureAtCoordinate(
          coordinate,
          frameState,
          hitTolerance,
          callback,
          matches
        );
      }
    }
  }

  /**
   * @module ol/renderer/canvas/VectorTileLayer
   */

  /**
   * @type {!Object<string, Array<import("../../render/canvas.js").BuilderType>>}
   */
  const IMAGE_REPLAYS = {
    'image': ['Polygon', 'Circle', 'LineString', 'Image', 'Text'],
    'hybrid': ['Polygon', 'LineString'],
    'vector': [],
  };

  /**
   * @type {!Object<string, Array<import("../../render/canvas.js").BuilderType>>}
   */
  const VECTOR_REPLAYS = {
    'hybrid': ['Image', 'Text', 'Default'],
    'vector': ['Polygon', 'Circle', 'LineString', 'Image', 'Text', 'Default'],
  };

  /**
   * @classdesc
   * Canvas renderer for vector tile layers.
   * @api
   * @extends {CanvasTileLayerRenderer<import("../../layer/VectorTile.js").default>}
   */
  class CanvasVectorTileLayerRenderer extends CanvasTileLayerRenderer {
    /**
     * @param {import("../../layer/VectorTile.js").default} layer VectorTile layer.
     */
    constructor(layer) {
      super(layer);

      /** @private */
      this.boundHandleStyleImageChange_ = this.handleStyleImageChange_.bind(this);

      /**
       * @private
       * @type {number}
       */
      this.renderedLayerRevision_;

      /**
       * @private
       * @type {import("../../transform").Transform}
       */
      this.renderedPixelToCoordinateTransform_ = null;

      /**
       * @private
       * @type {number}
       */
      this.renderedRotation_;

      /**
       * @private
       * @type {import("../../transform.js").Transform}
       */
      this.tmpTransform_ = create$2();
    }

    /**
     * @param {import("../../VectorRenderTile.js").default} tile Tile.
     * @param {number} pixelRatio Pixel ratio.
     * @param {import("../../proj/Projection").default} projection Projection.
     * @return {boolean|undefined} Tile needs to be rendered.
     */
    prepareTile(tile, pixelRatio, projection) {
      let render;
      const state = tile.getState();
      if (state === TileState.LOADED || state === TileState.ERROR) {
        this.updateExecutorGroup_(tile, pixelRatio, projection);
        if (this.tileImageNeedsRender_(tile)) {
          render = true;
        }
      }
      return render;
    }

    /**
     * @param {number} z Tile coordinate z.
     * @param {number} x Tile coordinate x.
     * @param {number} y Tile coordinate y.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @return {!import("../../Tile.js").default} Tile.
     */
    getTile(z, x, y, frameState) {
      const pixelRatio = frameState.pixelRatio;
      const viewState = frameState.viewState;
      const resolution = viewState.resolution;
      const projection = viewState.projection;
      const layer = this.getLayer();
      const tile = layer.getSource().getTile(z, x, y, pixelRatio, projection);
      const viewHints = frameState.viewHints;
      const hifi = !(
        viewHints[ViewHint.ANIMATING] || viewHints[ViewHint.INTERACTING]
      );
      if (hifi || !tile.wantedResolution) {
        tile.wantedResolution = resolution;
      }
      const render = this.prepareTile(tile, pixelRatio, projection);
      if (
        render &&
        (hifi || Date.now() - frameState.time < 8) &&
        layer.getRenderMode() !== 'vector'
      ) {
        this.renderTileImage_(tile, frameState);
      }
      return super.getTile(z, x, y, frameState);
    }

    /**
     * @param {import("../../VectorRenderTile.js").default} tile Tile.
     * @return {boolean} Tile is drawable.
     */
    isDrawableTile(tile) {
      const layer = this.getLayer();
      return (
        super.isDrawableTile(tile) &&
        (layer.getRenderMode() === 'vector'
          ? getUid(layer) in tile.executorGroups
          : tile.hasContext(layer))
      );
    }

    /**
     * @inheritDoc
     */
    getTileImage(tile) {
      return tile.getImage(this.getLayer());
    }

    /**
     * Determine whether render should be called.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @return {boolean} Layer is ready to be rendered.
     */
    prepareFrame(frameState) {
      const layerRevision = this.getLayer().getRevision();
      if (this.renderedLayerRevision_ !== layerRevision) {
        this.renderedLayerRevision_ = layerRevision;
        this.renderedTiles.length = 0;
      }
      return super.prepareFrame(frameState);
    }

    /**
     * @param {import("../../VectorRenderTile.js").default} tile Tile.
     * @param {number} pixelRatio Pixel ratio.
     * @param {import("../../proj/Projection.js").default} projection Projection.
     * @private
     */
    updateExecutorGroup_(tile, pixelRatio, projection) {
      const layer = /** @type {import("../../layer/VectorTile.js").default} */ (
        this.getLayer()
      );
      const revision = layer.getRevision();
      const renderOrder = layer.getRenderOrder() || null;

      const resolution = tile.wantedResolution;
      const builderState = tile.getReplayState(layer);
      if (
        !builderState.dirty &&
        builderState.renderedResolution === resolution &&
        builderState.renderedRevision == revision &&
        builderState.renderedRenderOrder == renderOrder
      ) {
        return;
      }

      const source = layer.getSource();
      const declutter = layer.getDeclutter();
      const sourceTileGrid = source.getTileGrid();
      const tileGrid = source.getTileGridForProjection(projection);
      const tileExtent = tileGrid.getTileCoordExtent(tile.wrappedTileCoord);

      const sourceTiles = source.getSourceTiles(pixelRatio, projection, tile);
      const layerUid = getUid(layer);
      delete tile.hitDetectionImageData[layerUid];
      tile.executorGroups[layerUid] = [];
      if (declutter) {
        tile.declutterExecutorGroups[layerUid] = [];
      }
      builderState.dirty = false;
      for (let t = 0, tt = sourceTiles.length; t < tt; ++t) {
        const sourceTile = sourceTiles[t];
        if (sourceTile.getState() != TileState.LOADED) {
          continue;
        }
        const sourceTileCoord = sourceTile.tileCoord;
        const sourceTileExtent =
          sourceTileGrid.getTileCoordExtent(sourceTileCoord);
        const sharedExtent = getIntersection(tileExtent, sourceTileExtent);
        const builderExtent = buffer$1(
          sharedExtent,
          layer.getRenderBuffer() * resolution,
          this.tmpExtent
        );
        const bufferedExtent = equals$2(sourceTileExtent, sharedExtent)
          ? null
          : builderExtent;
        const builderGroup = new BuilderGroup(
          0,
          builderExtent,
          resolution,
          pixelRatio
        );
        const declutterBuilderGroup = declutter
          ? new BuilderGroup(0, sharedExtent, resolution, pixelRatio)
          : undefined;
        const squaredTolerance = getSquaredTolerance(
          resolution,
          pixelRatio
        );

        /**
         * @param {import("../../Feature.js").FeatureLike} feature Feature.
         * @this {CanvasVectorTileLayerRenderer}
         */
        const render = function (feature) {
          let styles;
          const styleFunction =
            feature.getStyleFunction() || layer.getStyleFunction();
          if (styleFunction) {
            styles = styleFunction(feature, resolution);
          }
          if (styles) {
            const dirty = this.renderFeature(
              feature,
              squaredTolerance,
              styles,
              builderGroup,
              declutterBuilderGroup
            );
            builderState.dirty = builderState.dirty || dirty;
          }
        };

        const features = sourceTile.getFeatures();
        if (renderOrder && renderOrder !== builderState.renderedRenderOrder) {
          features.sort(renderOrder);
        }
        for (let i = 0, ii = features.length; i < ii; ++i) {
          const feature = features[i];
          if (
            !bufferedExtent ||
            intersects$3(bufferedExtent, feature.getGeometry().getExtent())
          ) {
            render.call(this, feature);
          }
        }
        const executorGroupInstructions = builderGroup.finish();
        // no need to clip when the render tile is covered by a single source tile
        const replayExtent =
          layer.getRenderMode() !== 'vector' &&
          declutter &&
          sourceTiles.length === 1
            ? null
            : sharedExtent;
        const renderingReplayGroup = new ExecutorGroup(
          replayExtent,
          resolution,
          pixelRatio,
          source.getOverlaps(),
          executorGroupInstructions,
          layer.getRenderBuffer()
        );
        tile.executorGroups[layerUid].push(renderingReplayGroup);
        if (declutterBuilderGroup) {
          const declutterExecutorGroup = new ExecutorGroup(
            null,
            resolution,
            pixelRatio,
            source.getOverlaps(),
            declutterBuilderGroup.finish(),
            layer.getRenderBuffer()
          );
          tile.declutterExecutorGroups[layerUid].push(declutterExecutorGroup);
        }
      }
      builderState.renderedRevision = revision;
      builderState.renderedRenderOrder = renderOrder;
      builderState.renderedResolution = resolution;
    }

    /**
     * @param {import("../../coordinate.js").Coordinate} coordinate Coordinate.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     * @param {number} hitTolerance Hit tolerance in pixels.
     * @param {import("../vector.js").FeatureCallback<T>} callback Feature callback.
     * @param {Array<import("../Map.js").HitMatch<T>>} matches The hit detected matches with tolerance.
     * @return {T|undefined} Callback result.
     * @template T
     */
    forEachFeatureAtCoordinate(
      coordinate,
      frameState,
      hitTolerance,
      callback,
      matches
    ) {
      const resolution = frameState.viewState.resolution;
      const rotation = frameState.viewState.rotation;
      hitTolerance = hitTolerance == undefined ? 0 : hitTolerance;
      const layer = this.getLayer();
      const source = layer.getSource();
      const tileGrid = source.getTileGridForProjection(
        frameState.viewState.projection
      );

      const hitExtent = boundingExtent([coordinate]);
      buffer$1(hitExtent, resolution * hitTolerance, hitExtent);

      /** @type {!Object<string, import("../Map.js").HitMatch<T>|true>} */
      const features = {};

      /**
       * @param {import("../../Feature.js").FeatureLike} feature Feature.
       * @param {import("../../geom/SimpleGeometry.js").default} geometry Geometry.
       * @param {number} distanceSq The squared distance to the click position.
       * @return {T|undefined} Callback result.
       */
      const featureCallback = function (feature, geometry, distanceSq) {
        let key = feature.getId();
        if (key === undefined) {
          key = getUid(feature);
        }
        const match = features[key];
        if (!match) {
          if (distanceSq === 0) {
            features[key] = true;
            return callback(feature, layer, geometry);
          }
          matches.push(
            (features[key] = {
              feature: feature,
              layer: layer,
              geometry: geometry,
              distanceSq: distanceSq,
              callback: callback,
            })
          );
        } else if (match !== true && distanceSq < match.distanceSq) {
          if (distanceSq === 0) {
            features[key] = true;
            matches.splice(matches.lastIndexOf(match), 1);
            return callback(feature, layer, geometry);
          }
          match.geometry = geometry;
          match.distanceSq = distanceSq;
        }
        return undefined;
      };

      const renderedTiles =
        /** @type {Array<import("../../VectorRenderTile.js").default>} */ (
          this.renderedTiles
        );

      let found;
      for (let i = 0, ii = renderedTiles.length; !found && i < ii; ++i) {
        const tile = renderedTiles[i];
        const tileExtent = tileGrid.getTileCoordExtent(tile.wrappedTileCoord);
        if (!intersects$3(tileExtent, hitExtent)) {
          continue;
        }

        const layerUid = getUid(layer);
        const executorGroups = [tile.executorGroups[layerUid]];
        const declutterExecutorGroups = tile.declutterExecutorGroups[layerUid];
        if (declutterExecutorGroups) {
          executorGroups.push(declutterExecutorGroups);
        }
        executorGroups.some((executorGroups) => {
          const declutteredFeatures =
            executorGroups === declutterExecutorGroups
              ? frameState.declutterTree.all().map((item) => item.value)
              : null;
          for (let t = 0, tt = executorGroups.length; t < tt; ++t) {
            const executorGroup = executorGroups[t];
            found = executorGroup.forEachFeatureAtCoordinate(
              coordinate,
              resolution,
              rotation,
              hitTolerance,
              featureCallback,
              declutteredFeatures
            );
            if (found) {
              return true;
            }
          }
        });
      }
      return found;
    }

    /**
     * Asynchronous layer level hit detection.
     * @param {import("../../pixel.js").Pixel} pixel Pixel.
     * @return {Promise<Array<import("../../Feature").default>>} Promise that resolves with an array of features.
     */
    getFeatures(pixel) {
      return new Promise(
        function (resolve, reject) {
          const layer =
            /** @type {import("../../layer/VectorTile.js").default} */ (
              this.getLayer()
            );
          const layerUid = getUid(layer);
          const source = layer.getSource();
          const projection = this.renderedProjection;
          const projectionExtent = projection.getExtent();
          const resolution = this.renderedResolution;
          const tileGrid = source.getTileGridForProjection(projection);
          const coordinate = apply(
            this.renderedPixelToCoordinateTransform_,
            pixel.slice()
          );
          const tileCoord = tileGrid.getTileCoordForCoordAndResolution(
            coordinate,
            resolution
          );
          let tile;
          for (let i = 0, ii = this.renderedTiles.length; i < ii; ++i) {
            if (
              tileCoord.toString() === this.renderedTiles[i].tileCoord.toString()
            ) {
              tile = this.renderedTiles[i];
              if (tile.getState() === TileState.LOADED) {
                const extent = tileGrid.getTileCoordExtent(tile.tileCoord);
                if (
                  source.getWrapX() &&
                  projection.canWrapX() &&
                  !containsExtent(projectionExtent, extent)
                ) {
                  wrapX$1(coordinate, projection);
                }
                break;
              }
              tile = undefined;
            }
          }
          if (!tile || tile.loadingSourceTiles > 0) {
            resolve([]);
            return;
          }
          const extent = tileGrid.getTileCoordExtent(tile.wrappedTileCoord);
          const corner = getTopLeft(extent);
          const tilePixel = [
            (coordinate[0] - corner[0]) / resolution,
            (corner[1] - coordinate[1]) / resolution,
          ];
          const features = tile
            .getSourceTiles()
            .reduce(function (accumulator, sourceTile) {
              return accumulator.concat(sourceTile.getFeatures());
            }, []);
          let hitDetectionImageData = tile.hitDetectionImageData[layerUid];
          if (!hitDetectionImageData && !this.animatingOrInteracting_) {
            const tileSize = toSize(
              tileGrid.getTileSize(
                tileGrid.getZForResolution(resolution, source.zDirection)
              )
            );
            const rotation = this.renderedRotation_;
            const transforms = [
              this.getRenderTransform(
                tileGrid.getTileCoordCenter(tile.wrappedTileCoord),
                resolution,
                0,
                HIT_DETECT_RESOLUTION,
                tileSize[0] * HIT_DETECT_RESOLUTION,
                tileSize[1] * HIT_DETECT_RESOLUTION,
                0
              ),
            ];
            hitDetectionImageData = createHitDetectionImageData(
              tileSize,
              transforms,
              features,
              layer.getStyleFunction(),
              tileGrid.getTileCoordExtent(tile.wrappedTileCoord),
              tile.getReplayState(layer).renderedResolution,
              rotation
            );
            tile.hitDetectionImageData[layerUid] = hitDetectionImageData;
          }
          resolve(hitDetect(tilePixel, features, hitDetectionImageData));
        }.bind(this)
      );
    }

    /**
     * Perform action necessary to get the layer rendered after new fonts have loaded
     */
    handleFontsChanged() {
      const layer = this.getLayer();
      if (layer.getVisible() && this.renderedLayerRevision_ !== undefined) {
        layer.changed();
      }
    }

    /**
     * Handle changes in image style state.
     * @param {import("../../events/Event.js").default} event Image style change event.
     * @private
     */
    handleStyleImageChange_(event) {
      this.renderIfReadyAndVisible();
    }

    /**
     * Render declutter items for this layer
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     */
    renderDeclutter(frameState) {
      const context = this.context;
      const alpha = context.globalAlpha;
      context.globalAlpha = this.getLayer().getOpacity();
      const viewHints = frameState.viewHints;
      const hifi = !(
        viewHints[ViewHint.ANIMATING] || viewHints[ViewHint.INTERACTING]
      );
      const tiles =
        /** @type {Array<import("../../VectorRenderTile.js").default>} */ (
          this.renderedTiles
        );
      for (let i = 0, ii = tiles.length; i < ii; ++i) {
        const tile = tiles[i];
        const declutterExecutorGroups =
          tile.declutterExecutorGroups[getUid(this.getLayer())];
        if (declutterExecutorGroups) {
          for (let j = declutterExecutorGroups.length - 1; j >= 0; --j) {
            declutterExecutorGroups[j].execute(
              this.context,
              1,
              this.getTileRenderTransform(tile, frameState),
              frameState.viewState.rotation,
              hifi,
              undefined,
              frameState.declutterTree
            );
          }
        }
      }
      context.globalAlpha = alpha;
    }

    getTileRenderTransform(tile, frameState) {
      const pixelRatio = frameState.pixelRatio;
      const viewState = frameState.viewState;
      const center = viewState.center;
      const resolution = viewState.resolution;
      const rotation = viewState.rotation;
      const size = frameState.size;
      const width = Math.round(size[0] * pixelRatio);
      const height = Math.round(size[1] * pixelRatio);

      const source = this.getLayer().getSource();
      const tileGrid = source.getTileGridForProjection(
        frameState.viewState.projection
      );
      const tileCoord = tile.tileCoord;
      const tileExtent = tileGrid.getTileCoordExtent(tile.wrappedTileCoord);
      const worldOffset =
        tileGrid.getTileCoordExtent(tileCoord, this.tmpExtent)[0] - tileExtent[0];
      const transform = multiply(
        scale$3(this.inversePixelTransform.slice(), 1 / pixelRatio, 1 / pixelRatio),
        this.getRenderTransform(
          center,
          resolution,
          rotation,
          pixelRatio,
          width,
          height,
          worldOffset
        )
      );
      return transform;
    }

    /**
     * Render the vectors for this layer.
     * @param {CanvasRenderingContext2D} context Target context.
     * @param {import("../../Map.js").FrameState} frameState Frame state.
     */
    postRender(context, frameState) {
      const viewHints = frameState.viewHints;
      const hifi = !(
        viewHints[ViewHint.ANIMATING] || viewHints[ViewHint.INTERACTING]
      );

      this.renderedPixelToCoordinateTransform_ =
        frameState.pixelToCoordinateTransform.slice();
      this.renderedRotation_ = frameState.viewState.rotation;

      const layer = /** @type {import("../../layer/VectorTile.js").default} */ (
        this.getLayer()
      );
      const renderMode = layer.getRenderMode();
      const alpha = context.globalAlpha;
      context.globalAlpha = layer.getOpacity();
      const replayTypes = VECTOR_REPLAYS[renderMode];
      const viewState = frameState.viewState;
      const rotation = viewState.rotation;
      const tileSource = layer.getSource();
      const tileGrid = tileSource.getTileGridForProjection(viewState.projection);
      const z = tileGrid.getZForResolution(
        viewState.resolution,
        tileSource.zDirection
      );

      const tiles = this.renderedTiles;
      const clips = [];
      const clipZs = [];
      let ready = true;
      for (let i = tiles.length - 1; i >= 0; --i) {
        const tile = /** @type {import("../../VectorRenderTile.js").default} */ (
          tiles[i]
        );
        ready = ready && !tile.getReplayState(layer).dirty;
        const executorGroups = tile.executorGroups[getUid(layer)].filter(
          (group) => group.hasExecutors(replayTypes)
        );
        if (executorGroups.length === 0) {
          continue;
        }
        const transform = this.getTileRenderTransform(tile, frameState);
        const currentZ = tile.tileCoord[0];
        let contextSaved = false;
        // Clip mask for regions in this tile that already filled by a higher z tile
        const currentClip = executorGroups[0].getClipCoords(transform);
        if (currentClip) {
          for (let j = 0, jj = clips.length; j < jj; ++j) {
            if (z !== currentZ && currentZ < clipZs[j]) {
              const clip = clips[j];
              if (
                intersects$3(
                  [
                    currentClip[0],
                    currentClip[3],
                    currentClip[4],
                    currentClip[7],
                  ],
                  [clip[0], clip[3], clip[4], clip[7]]
                )
              ) {
                if (!contextSaved) {
                  context.save();
                  contextSaved = true;
                }
                context.beginPath();
                // counter-clockwise (outer ring) for current tile
                context.moveTo(currentClip[0], currentClip[1]);
                context.lineTo(currentClip[2], currentClip[3]);
                context.lineTo(currentClip[4], currentClip[5]);
                context.lineTo(currentClip[6], currentClip[7]);
                // clockwise (inner ring) for higher z tile
                context.moveTo(clip[6], clip[7]);
                context.lineTo(clip[4], clip[5]);
                context.lineTo(clip[2], clip[3]);
                context.lineTo(clip[0], clip[1]);
                context.clip();
              }
            }
          }
          clips.push(currentClip);
          clipZs.push(currentZ);
        }
        for (let t = 0, tt = executorGroups.length; t < tt; ++t) {
          const executorGroup = executorGroups[t];
          executorGroup.execute(
            context,
            1,
            transform,
            rotation,
            hifi,
            replayTypes
          );
        }
        if (contextSaved) {
          context.restore();
        }
      }
      context.globalAlpha = alpha;
      this.ready = ready;

      super.postRender(context, frameState);
    }

    /**
     * @param {import("../../Feature.js").FeatureLike} feature Feature.
     * @param {number} squaredTolerance Squared tolerance.
     * @param {import("../../style/Style.js").default|Array<import("../../style/Style.js").default>} styles The style or array of styles.
     * @param {import("../../render/canvas/BuilderGroup.js").default} builderGroup Replay group.
     * @param {import("../../render/canvas/BuilderGroup.js").default} [declutterBuilderGroup] Builder group for decluttering.
     * @return {boolean} `true` if an image is loading.
     */
    renderFeature(
      feature,
      squaredTolerance,
      styles,
      builderGroup,
      declutterBuilderGroup
    ) {
      if (!styles) {
        return false;
      }
      let loading = false;
      if (Array.isArray(styles)) {
        for (let i = 0, ii = styles.length; i < ii; ++i) {
          loading =
            renderFeature$1(
              builderGroup,
              feature,
              styles[i],
              squaredTolerance,
              this.boundHandleStyleImageChange_,
              undefined,
              declutterBuilderGroup
            ) || loading;
        }
      } else {
        loading = renderFeature$1(
          builderGroup,
          feature,
          styles,
          squaredTolerance,
          this.boundHandleStyleImageChange_,
          undefined,
          declutterBuilderGroup
        );
      }
      return loading;
    }

    /**
     * @param {import("../../VectorRenderTile.js").default} tile Tile.
     * @return {boolean} A new tile image was rendered.
     * @private
     */
    tileImageNeedsRender_(tile) {
      const layer = /** @type {import("../../layer/VectorTile.js").default} */ (
        this.getLayer()
      );
      if (layer.getRenderMode() === 'vector') {
        return false;
      }
      const replayState = tile.getReplayState(layer);
      const revision = layer.getRevision();
      const resolution = tile.wantedResolution;
      return (
        replayState.renderedTileResolution !== resolution ||
        replayState.renderedTileRevision !== revision
      );
    }

    /**
     * @param {import("../../VectorRenderTile.js").default} tile Tile.
     * @param {import("../../Map").FrameState} frameState Frame state.
     * @private
     */
    renderTileImage_(tile, frameState) {
      const layer = /** @type {import("../../layer/VectorTile.js").default} */ (
        this.getLayer()
      );
      const replayState = tile.getReplayState(layer);
      const revision = layer.getRevision();
      const executorGroups = tile.executorGroups[getUid(layer)];
      replayState.renderedTileRevision = revision;

      const tileCoord = tile.wrappedTileCoord;
      const z = tileCoord[0];
      const source = layer.getSource();
      let pixelRatio = frameState.pixelRatio;
      const viewState = frameState.viewState;
      const projection = viewState.projection;
      const tileGrid = source.getTileGridForProjection(projection);
      const tileResolution = tileGrid.getResolution(tile.tileCoord[0]);
      const renderPixelRatio =
        (frameState.pixelRatio / tile.wantedResolution) * tileResolution;
      const resolution = tileGrid.getResolution(z);
      const context = tile.getContext(layer);

      // Increase tile size when overzooming for low pixel ratio, to avoid blurry tiles
      pixelRatio = Math.round(
        Math.max(pixelRatio, renderPixelRatio / pixelRatio)
      );
      const size = source.getTilePixelSize(z, pixelRatio, projection);
      context.canvas.width = size[0];
      context.canvas.height = size[1];
      const renderScale = pixelRatio / renderPixelRatio;
      if (renderScale !== 1) {
        const canvasTransform = reset(this.tmpTransform_);
        scale$3(canvasTransform, renderScale, renderScale);
        context.setTransform.apply(context, canvasTransform);
      }
      const tileExtent = tileGrid.getTileCoordExtent(tileCoord, this.tmpExtent);
      const pixelScale = renderPixelRatio / resolution;
      const transform = reset(this.tmpTransform_);
      scale$3(transform, pixelScale, -pixelScale);
      translate$1(transform, -tileExtent[0], -tileExtent[3]);
      for (let i = 0, ii = executorGroups.length; i < ii; ++i) {
        const executorGroup = executorGroups[i];
        executorGroup.execute(
          context,
          renderScale,
          transform,
          0,
          true,
          IMAGE_REPLAYS[layer.getRenderMode()]
        );
      }
      replayState.renderedTileResolution = tile.wantedResolution;
    }
  }

  /**
   * @module ol/geom/Circle
   */

  /**
   * @classdesc
   * Circle geometry.
   *
   * @api
   */
  class Circle extends SimpleGeometry {
    /**
     * @param {!import("../coordinate.js").Coordinate} center Center.
     *     For internal use, flat coordinates in combination with `layout` and no
     *     `radius` are also accepted.
     * @param {number} [radius] Radius.
     * @param {import("./Geometry.js").GeometryLayout} [layout] Layout.
     */
    constructor(center, radius, layout) {
      super();
      if (layout !== undefined && radius === undefined) {
        this.setFlatCoordinates(layout, center);
      } else {
        radius = radius ? radius : 0;
        this.setCenterAndRadius(center, radius, layout);
      }
    }

    /**
     * Make a complete copy of the geometry.
     * @return {!Circle} Clone.
     * @api
     */
    clone() {
      const circle = new Circle(
        this.flatCoordinates.slice(),
        undefined,
        this.layout
      );
      circle.applyProperties(this);
      return circle;
    }

    /**
     * @param {number} x X.
     * @param {number} y Y.
     * @param {import("../coordinate.js").Coordinate} closestPoint Closest point.
     * @param {number} minSquaredDistance Minimum squared distance.
     * @return {number} Minimum squared distance.
     */
    closestPointXY(x, y, closestPoint, minSquaredDistance) {
      const flatCoordinates = this.flatCoordinates;
      const dx = x - flatCoordinates[0];
      const dy = y - flatCoordinates[1];
      const squaredDistance = dx * dx + dy * dy;
      if (squaredDistance < minSquaredDistance) {
        if (squaredDistance === 0) {
          for (let i = 0; i < this.stride; ++i) {
            closestPoint[i] = flatCoordinates[i];
          }
        } else {
          const delta = this.getRadius() / Math.sqrt(squaredDistance);
          closestPoint[0] = flatCoordinates[0] + delta * dx;
          closestPoint[1] = flatCoordinates[1] + delta * dy;
          for (let i = 2; i < this.stride; ++i) {
            closestPoint[i] = flatCoordinates[i];
          }
        }
        closestPoint.length = this.stride;
        return squaredDistance;
      } else {
        return minSquaredDistance;
      }
    }

    /**
     * @param {number} x X.
     * @param {number} y Y.
     * @return {boolean} Contains (x, y).
     */
    containsXY(x, y) {
      const flatCoordinates = this.flatCoordinates;
      const dx = x - flatCoordinates[0];
      const dy = y - flatCoordinates[1];
      return dx * dx + dy * dy <= this.getRadiusSquared_();
    }

    /**
     * Return the center of the circle as {@link module:ol/coordinate~Coordinate coordinate}.
     * @return {import("../coordinate.js").Coordinate} Center.
     * @api
     */
    getCenter() {
      return this.flatCoordinates.slice(0, this.stride);
    }

    /**
     * @param {import("../extent.js").Extent} extent Extent.
     * @protected
     * @return {import("../extent.js").Extent} extent Extent.
     */
    computeExtent(extent) {
      const flatCoordinates = this.flatCoordinates;
      const radius = flatCoordinates[this.stride] - flatCoordinates[0];
      return createOrUpdate$2(
        flatCoordinates[0] - radius,
        flatCoordinates[1] - radius,
        flatCoordinates[0] + radius,
        flatCoordinates[1] + radius,
        extent
      );
    }

    /**
     * Return the radius of the circle.
     * @return {number} Radius.
     * @api
     */
    getRadius() {
      return Math.sqrt(this.getRadiusSquared_());
    }

    /**
     * @private
     * @return {number} Radius squared.
     */
    getRadiusSquared_() {
      const dx = this.flatCoordinates[this.stride] - this.flatCoordinates[0];
      const dy = this.flatCoordinates[this.stride + 1] - this.flatCoordinates[1];
      return dx * dx + dy * dy;
    }

    /**
     * Get the type of this geometry.
     * @return {import("./Geometry.js").Type} Geometry type.
     * @api
     */
    getType() {
      return 'Circle';
    }

    /**
     * Test if the geometry and the passed extent intersect.
     * @param {import("../extent.js").Extent} extent Extent.
     * @return {boolean} `true` if the geometry and the extent intersect.
     * @api
     */
    intersectsExtent(extent) {
      const circleExtent = this.getExtent();
      if (intersects$3(extent, circleExtent)) {
        const center = this.getCenter();

        if (extent[0] <= center[0] && extent[2] >= center[0]) {
          return true;
        }
        if (extent[1] <= center[1] && extent[3] >= center[1]) {
          return true;
        }

        return forEachCorner(extent, this.intersectsCoordinate.bind(this));
      }
      return false;
    }

    /**
     * Set the center of the circle as {@link module:ol/coordinate~Coordinate coordinate}.
     * @param {import("../coordinate.js").Coordinate} center Center.
     * @api
     */
    setCenter(center) {
      const stride = this.stride;
      const radius = this.flatCoordinates[stride] - this.flatCoordinates[0];
      const flatCoordinates = center.slice();
      flatCoordinates[stride] = flatCoordinates[0] + radius;
      for (let i = 1; i < stride; ++i) {
        flatCoordinates[stride + i] = center[i];
      }
      this.setFlatCoordinates(this.layout, flatCoordinates);
      this.changed();
    }

    /**
     * Set the center (as {@link module:ol/coordinate~Coordinate coordinate}) and the radius (as
     * number) of the circle.
     * @param {!import("../coordinate.js").Coordinate} center Center.
     * @param {number} radius Radius.
     * @param {import("./Geometry.js").GeometryLayout} [layout] Layout.
     * @api
     */
    setCenterAndRadius(center, radius, layout) {
      this.setLayout(layout, center, 0);
      if (!this.flatCoordinates) {
        this.flatCoordinates = [];
      }
      /** @type {Array<number>} */
      const flatCoordinates = this.flatCoordinates;
      let offset = deflateCoordinate(flatCoordinates, 0, center, this.stride);
      flatCoordinates[offset++] = flatCoordinates[0] + radius;
      for (let i = 1, ii = this.stride; i < ii; ++i) {
        flatCoordinates[offset++] = flatCoordinates[i];
      }
      flatCoordinates.length = offset;
      this.changed();
    }

    getCoordinates() {
      return null;
    }

    setCoordinates(coordinates, layout) {}

    /**
     * Set the radius of the circle. The radius is in the units of the projection.
     * @param {number} radius Radius.
     * @api
     */
    setRadius(radius) {
      this.flatCoordinates[this.stride] = this.flatCoordinates[0] + radius;
      this.changed();
    }

    /**
     * Rotate the geometry around a given coordinate. This modifies the geometry
     * coordinates in place.
     * @param {number} angle Rotation angle in counter-clockwise radians.
     * @param {import("../coordinate.js").Coordinate} anchor The rotation center.
     * @api
     */
    rotate(angle, anchor) {
      const center = this.getCenter();
      const stride = this.getStride();
      this.setCenter(
        rotate(center, 0, center.length, stride, angle, anchor, center)
      );
      this.changed();
    }

    /**
     * Translate the geometry.  This modifies the geometry coordinates in place.  If
     * instead you want a new geometry, first `clone()` this geometry.
     * @param {number} deltaX Delta X.
     * @param {number} deltaY Delta Y.
     * @api
     */
    translate(deltaX, deltaY) {
      const center = this.getCenter();
      const stride = this.getStride();
      this.setCenter(
        translate(center, 0, center.length, stride, deltaX, deltaY, center)
      );
      this.changed();
    }
  }

  /**
   * Transform each coordinate of the circle from one coordinate reference system
   * to another. The geometry is modified in place.
   * If you do not want the geometry modified in place, first clone() it and
   * then use this function on the clone.
   *
   * Internally a circle is currently represented by two points: the center of
   * the circle `[cx, cy]`, and the point to the right of the circle
   * `[cx + r, cy]`. This `transform` function just transforms these two points.
   * So the resulting geometry is also a circle, and that circle does not
   * correspond to the shape that would be obtained by transforming every point
   * of the original circle.
   *
   * @param {import("../proj.js").ProjectionLike} source The current projection.  Can be a
   *     string identifier or a {@link module:ol/proj/Projection~Projection} object.
   * @param {import("../proj.js").ProjectionLike} destination The desired projection.  Can be a
   *     string identifier or a {@link module:ol/proj/Projection~Projection} object.
   * @return {Circle} This geometry.  Note that original geometry is
   *     modified in place.
   * @function
   * @api
   */
  Circle.prototype.transform;

  /**
   * @module ol/geom/GeometryCollection
   */

  /**
   * @classdesc
   * An array of {@link module:ol/geom/Geometry~Geometry} objects.
   *
   * @api
   */
  class GeometryCollection extends Geometry {
    /**
     * @param {Array<Geometry>} [geometries] Geometries.
     */
    constructor(geometries) {
      super();

      /**
       * @private
       * @type {Array<Geometry>}
       */
      this.geometries_ = geometries ? geometries : null;

      /**
       * @type {Array<import("../events.js").EventsKey>}
       */
      this.changeEventsKeys_ = [];

      this.listenGeometriesChange_();
    }

    /**
     * @private
     */
    unlistenGeometriesChange_() {
      this.changeEventsKeys_.forEach(unlistenByKey);
      this.changeEventsKeys_.length = 0;
    }

    /**
     * @private
     */
    listenGeometriesChange_() {
      if (!this.geometries_) {
        return;
      }
      for (let i = 0, ii = this.geometries_.length; i < ii; ++i) {
        this.changeEventsKeys_.push(
          listen(this.geometries_[i], EventType$1.CHANGE, this.changed, this)
        );
      }
    }

    /**
     * Make a complete copy of the geometry.
     * @return {!GeometryCollection} Clone.
     * @api
     */
    clone() {
      const geometryCollection = new GeometryCollection(null);
      geometryCollection.setGeometries(this.geometries_);
      geometryCollection.applyProperties(this);
      return geometryCollection;
    }

    /**
     * @param {number} x X.
     * @param {number} y Y.
     * @param {import("../coordinate.js").Coordinate} closestPoint Closest point.
     * @param {number} minSquaredDistance Minimum squared distance.
     * @return {number} Minimum squared distance.
     */
    closestPointXY(x, y, closestPoint, minSquaredDistance) {
      if (minSquaredDistance < closestSquaredDistanceXY(this.getExtent(), x, y)) {
        return minSquaredDistance;
      }
      const geometries = this.geometries_;
      for (let i = 0, ii = geometries.length; i < ii; ++i) {
        minSquaredDistance = geometries[i].closestPointXY(
          x,
          y,
          closestPoint,
          minSquaredDistance
        );
      }
      return minSquaredDistance;
    }

    /**
     * @param {number} x X.
     * @param {number} y Y.
     * @return {boolean} Contains (x, y).
     */
    containsXY(x, y) {
      const geometries = this.geometries_;
      for (let i = 0, ii = geometries.length; i < ii; ++i) {
        if (geometries[i].containsXY(x, y)) {
          return true;
        }
      }
      return false;
    }

    /**
     * @param {import("../extent.js").Extent} extent Extent.
     * @protected
     * @return {import("../extent.js").Extent} extent Extent.
     */
    computeExtent(extent) {
      createOrUpdateEmpty(extent);
      const geometries = this.geometries_;
      for (let i = 0, ii = geometries.length; i < ii; ++i) {
        extend$1(extent, geometries[i].getExtent());
      }
      return extent;
    }

    /**
     * Return the geometries that make up this geometry collection.
     * @return {Array<Geometry>} Geometries.
     * @api
     */
    getGeometries() {
      return cloneGeometries(this.geometries_);
    }

    /**
     * @return {Array<Geometry>} Geometries.
     */
    getGeometriesArray() {
      return this.geometries_;
    }

    /**
     * @return {Array<Geometry>} Geometries.
     */
    getGeometriesArrayRecursive() {
      /** @type {Array<Geometry>} */
      let geometriesArray = [];
      const geometries = this.geometries_;
      for (let i = 0, ii = geometries.length; i < ii; ++i) {
        if (geometries[i].getType() === this.getType()) {
          geometriesArray = geometriesArray.concat(
            /** @type {GeometryCollection} */ (
              geometries[i]
            ).getGeometriesArrayRecursive()
          );
        } else {
          geometriesArray.push(geometries[i]);
        }
      }
      return geometriesArray;
    }

    /**
     * Create a simplified version of this geometry using the Douglas Peucker algorithm.
     * @param {number} squaredTolerance Squared tolerance.
     * @return {GeometryCollection} Simplified GeometryCollection.
     */
    getSimplifiedGeometry(squaredTolerance) {
      if (this.simplifiedGeometryRevision !== this.getRevision()) {
        this.simplifiedGeometryMaxMinSquaredTolerance = 0;
        this.simplifiedGeometryRevision = this.getRevision();
      }
      if (
        squaredTolerance < 0 ||
        (this.simplifiedGeometryMaxMinSquaredTolerance !== 0 &&
          squaredTolerance < this.simplifiedGeometryMaxMinSquaredTolerance)
      ) {
        return this;
      }

      const simplifiedGeometries = [];
      const geometries = this.geometries_;
      let simplified = false;
      for (let i = 0, ii = geometries.length; i < ii; ++i) {
        const geometry = geometries[i];
        const simplifiedGeometry =
          geometry.getSimplifiedGeometry(squaredTolerance);
        simplifiedGeometries.push(simplifiedGeometry);
        if (simplifiedGeometry !== geometry) {
          simplified = true;
        }
      }
      if (simplified) {
        const simplifiedGeometryCollection = new GeometryCollection(null);
        simplifiedGeometryCollection.setGeometriesArray(simplifiedGeometries);
        return simplifiedGeometryCollection;
      } else {
        this.simplifiedGeometryMaxMinSquaredTolerance = squaredTolerance;
        return this;
      }
    }

    /**
     * Get the type of this geometry.
     * @return {import("./Geometry.js").Type} Geometry type.
     * @api
     */
    getType() {
      return 'GeometryCollection';
    }

    /**
     * Test if the geometry and the passed extent intersect.
     * @param {import("../extent.js").Extent} extent Extent.
     * @return {boolean} `true` if the geometry and the extent intersect.
     * @api
     */
    intersectsExtent(extent) {
      const geometries = this.geometries_;
      for (let i = 0, ii = geometries.length; i < ii; ++i) {
        if (geometries[i].intersectsExtent(extent)) {
          return true;
        }
      }
      return false;
    }

    /**
     * @return {boolean} Is empty.
     */
    isEmpty() {
      return this.geometries_.length === 0;
    }

    /**
     * Rotate the geometry around a given coordinate. This modifies the geometry
     * coordinates in place.
     * @param {number} angle Rotation angle in radians.
     * @param {import("../coordinate.js").Coordinate} anchor The rotation center.
     * @api
     */
    rotate(angle, anchor) {
      const geometries = this.geometries_;
      for (let i = 0, ii = geometries.length; i < ii; ++i) {
        geometries[i].rotate(angle, anchor);
      }
      this.changed();
    }

    /**
     * Scale the geometry (with an optional origin).  This modifies the geometry
     * coordinates in place.
     * @abstract
     * @param {number} sx The scaling factor in the x-direction.
     * @param {number} [sy] The scaling factor in the y-direction (defaults to sx).
     * @param {import("../coordinate.js").Coordinate} [anchor] The scale origin (defaults to the center
     *     of the geometry extent).
     * @api
     */
    scale(sx, sy, anchor) {
      if (!anchor) {
        anchor = getCenter(this.getExtent());
      }
      const geometries = this.geometries_;
      for (let i = 0, ii = geometries.length; i < ii; ++i) {
        geometries[i].scale(sx, sy, anchor);
      }
      this.changed();
    }

    /**
     * Set the geometries that make up this geometry collection.
     * @param {Array<Geometry>} geometries Geometries.
     * @api
     */
    setGeometries(geometries) {
      this.setGeometriesArray(cloneGeometries(geometries));
    }

    /**
     * @param {Array<Geometry>} geometries Geometries.
     */
    setGeometriesArray(geometries) {
      this.unlistenGeometriesChange_();
      this.geometries_ = geometries;
      this.listenGeometriesChange_();
      this.changed();
    }

    /**
     * Apply a transform function to the coordinates of the geometry.
     * The geometry is modified in place.
     * If you do not want the geometry modified in place, first `clone()` it and
     * then use this function on the clone.
     * @param {import("../proj.js").TransformFunction} transformFn Transform function.
     * Called with a flat array of geometry coordinates.
     * @api
     */
    applyTransform(transformFn) {
      const geometries = this.geometries_;
      for (let i = 0, ii = geometries.length; i < ii; ++i) {
        geometries[i].applyTransform(transformFn);
      }
      this.changed();
    }

    /**
     * Translate the geometry.  This modifies the geometry coordinates in place.  If
     * instead you want a new geometry, first `clone()` this geometry.
     * @param {number} deltaX Delta X.
     * @param {number} deltaY Delta Y.
     * @api
     */
    translate(deltaX, deltaY) {
      const geometries = this.geometries_;
      for (let i = 0, ii = geometries.length; i < ii; ++i) {
        geometries[i].translate(deltaX, deltaY);
      }
      this.changed();
    }

    /**
     * Clean up.
     */
    disposeInternal() {
      this.unlistenGeometriesChange_();
      super.disposeInternal();
    }
  }

  /**
   * @param {Array<Geometry>} geometries Geometries.
   * @return {Array<Geometry>} Cloned geometries.
   */
  function cloneGeometries(geometries) {
    const clonedGeometries = [];
    for (let i = 0, ii = geometries.length; i < ii; ++i) {
      clonedGeometries.push(geometries[i].clone());
    }
    return clonedGeometries;
  }

  /**
   * @module ol/geom/flat/interpolate
   */

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   * @param {number} fraction Fraction.
   * @param {Array<number>} [dest] Destination.
   * @param {number} [dimension] Destination dimension (default is `2`)
   * @return {Array<number>} Destination.
   */
  function interpolatePoint(
    flatCoordinates,
    offset,
    end,
    stride,
    fraction,
    dest,
    dimension
  ) {
    let o, t;
    const n = (end - offset) / stride;
    if (n === 1) {
      o = offset;
    } else if (n === 2) {
      o = offset;
      t = fraction;
    } else if (n !== 0) {
      let x1 = flatCoordinates[offset];
      let y1 = flatCoordinates[offset + 1];
      let length = 0;
      const cumulativeLengths = [0];
      for (let i = offset + stride; i < end; i += stride) {
        const x2 = flatCoordinates[i];
        const y2 = flatCoordinates[i + 1];
        length += Math.sqrt((x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1));
        cumulativeLengths.push(length);
        x1 = x2;
        y1 = y2;
      }
      const target = fraction * length;
      const index = binarySearch$1(cumulativeLengths, target);
      if (index < 0) {
        t =
          (target - cumulativeLengths[-index - 2]) /
          (cumulativeLengths[-index - 1] - cumulativeLengths[-index - 2]);
        o = offset + (-index - 2) * stride;
      } else {
        o = offset + index * stride;
      }
    }
    dimension = dimension > 1 ? dimension : 2;
    dest = dest ? dest : new Array(dimension);
    for (let i = 0; i < dimension; ++i) {
      dest[i] =
        o === undefined
          ? NaN
          : t === undefined
          ? flatCoordinates[o + i]
          : lerp(flatCoordinates[o + i], flatCoordinates[o + stride + i], t);
    }
    return dest;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   * @param {number} m M.
   * @param {boolean} extrapolate Extrapolate.
   * @return {import("../../coordinate.js").Coordinate|null} Coordinate.
   */
  function lineStringCoordinateAtM(
    flatCoordinates,
    offset,
    end,
    stride,
    m,
    extrapolate
  ) {
    if (end == offset) {
      return null;
    }
    let coordinate;
    if (m < flatCoordinates[offset + stride - 1]) {
      if (extrapolate) {
        coordinate = flatCoordinates.slice(offset, offset + stride);
        coordinate[stride - 1] = m;
        return coordinate;
      } else {
        return null;
      }
    } else if (flatCoordinates[end - 1] < m) {
      if (extrapolate) {
        coordinate = flatCoordinates.slice(end - stride, end);
        coordinate[stride - 1] = m;
        return coordinate;
      } else {
        return null;
      }
    }
    // FIXME use O(1) search
    if (m == flatCoordinates[offset + stride - 1]) {
      return flatCoordinates.slice(offset, offset + stride);
    }
    let lo = offset / stride;
    let hi = end / stride;
    while (lo < hi) {
      const mid = (lo + hi) >> 1;
      if (m < flatCoordinates[(mid + 1) * stride - 1]) {
        hi = mid;
      } else {
        lo = mid + 1;
      }
    }
    const m0 = flatCoordinates[lo * stride - 1];
    if (m == m0) {
      return flatCoordinates.slice((lo - 1) * stride, (lo - 1) * stride + stride);
    }
    const m1 = flatCoordinates[(lo + 1) * stride - 1];
    const t = (m - m0) / (m1 - m0);
    coordinate = [];
    for (let i = 0; i < stride - 1; ++i) {
      coordinate.push(
        lerp(
          flatCoordinates[(lo - 1) * stride + i],
          flatCoordinates[lo * stride + i],
          t
        )
      );
    }
    coordinate.push(m);
    return coordinate;
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<number>} ends Ends.
   * @param {number} stride Stride.
   * @param {number} m M.
   * @param {boolean} extrapolate Extrapolate.
   * @param {boolean} interpolate Interpolate.
   * @return {import("../../coordinate.js").Coordinate|null} Coordinate.
   */
  function lineStringsCoordinateAtM(
    flatCoordinates,
    offset,
    ends,
    stride,
    m,
    extrapolate,
    interpolate
  ) {
    if (interpolate) {
      return lineStringCoordinateAtM(
        flatCoordinates,
        offset,
        ends[ends.length - 1],
        stride,
        m,
        extrapolate
      );
    }
    let coordinate;
    if (m < flatCoordinates[stride - 1]) {
      if (extrapolate) {
        coordinate = flatCoordinates.slice(0, stride);
        coordinate[stride - 1] = m;
        return coordinate;
      } else {
        return null;
      }
    }
    if (flatCoordinates[flatCoordinates.length - 1] < m) {
      if (extrapolate) {
        coordinate = flatCoordinates.slice(flatCoordinates.length - stride);
        coordinate[stride - 1] = m;
        return coordinate;
      } else {
        return null;
      }
    }
    for (let i = 0, ii = ends.length; i < ii; ++i) {
      const end = ends[i];
      if (offset == end) {
        continue;
      }
      if (m < flatCoordinates[offset + stride - 1]) {
        return null;
      } else if (m <= flatCoordinates[end - 1]) {
        return lineStringCoordinateAtM(
          flatCoordinates,
          offset,
          end,
          stride,
          m,
          false
        );
      }
      offset = end;
    }
    return null;
  }

  /**
   * @module ol/geom/LineString
   */

  /**
   * @classdesc
   * Linestring geometry.
   *
   * @api
   */
  class LineString extends SimpleGeometry {
    /**
     * @param {Array<import("../coordinate.js").Coordinate>|Array<number>} coordinates Coordinates.
     *     For internal use, flat coordinates in combination with `layout` are also accepted.
     * @param {import("./Geometry.js").GeometryLayout} [layout] Layout.
     */
    constructor(coordinates, layout) {
      super();

      /**
       * @private
       * @type {import("../coordinate.js").Coordinate}
       */
      this.flatMidpoint_ = null;

      /**
       * @private
       * @type {number}
       */
      this.flatMidpointRevision_ = -1;

      /**
       * @private
       * @type {number}
       */
      this.maxDelta_ = -1;

      /**
       * @private
       * @type {number}
       */
      this.maxDeltaRevision_ = -1;

      if (layout !== undefined && !Array.isArray(coordinates[0])) {
        this.setFlatCoordinates(
          layout,
          /** @type {Array<number>} */ (coordinates)
        );
      } else {
        this.setCoordinates(
          /** @type {Array<import("../coordinate.js").Coordinate>} */ (
            coordinates
          ),
          layout
        );
      }
    }

    /**
     * Append the passed coordinate to the coordinates of the linestring.
     * @param {import("../coordinate.js").Coordinate} coordinate Coordinate.
     * @api
     */
    appendCoordinate(coordinate) {
      if (!this.flatCoordinates) {
        this.flatCoordinates = coordinate.slice();
      } else {
        extend$2(this.flatCoordinates, coordinate);
      }
      this.changed();
    }

    /**
     * Make a complete copy of the geometry.
     * @return {!LineString} Clone.
     * @api
     */
    clone() {
      const lineString = new LineString(
        this.flatCoordinates.slice(),
        this.layout
      );
      lineString.applyProperties(this);
      return lineString;
    }

    /**
     * @param {number} x X.
     * @param {number} y Y.
     * @param {import("../coordinate.js").Coordinate} closestPoint Closest point.
     * @param {number} minSquaredDistance Minimum squared distance.
     * @return {number} Minimum squared distance.
     */
    closestPointXY(x, y, closestPoint, minSquaredDistance) {
      if (minSquaredDistance < closestSquaredDistanceXY(this.getExtent(), x, y)) {
        return minSquaredDistance;
      }
      if (this.maxDeltaRevision_ != this.getRevision()) {
        this.maxDelta_ = Math.sqrt(
          maxSquaredDelta(
            this.flatCoordinates,
            0,
            this.flatCoordinates.length,
            this.stride,
            0
          )
        );
        this.maxDeltaRevision_ = this.getRevision();
      }
      return assignClosestPoint(
        this.flatCoordinates,
        0,
        this.flatCoordinates.length,
        this.stride,
        this.maxDelta_,
        false,
        x,
        y,
        closestPoint,
        minSquaredDistance
      );
    }

    /**
     * Iterate over each segment, calling the provided callback.
     * If the callback returns a truthy value the function returns that
     * value immediately. Otherwise the function returns `false`.
     *
     * @param {function(this: S, import("../coordinate.js").Coordinate, import("../coordinate.js").Coordinate): T} callback Function
     *     called for each segment. The function will receive two arguments, the start and end coordinates of the segment.
     * @return {T|boolean} Value.
     * @template T,S
     * @api
     */
    forEachSegment(callback) {
      return forEach(
        this.flatCoordinates,
        0,
        this.flatCoordinates.length,
        this.stride,
        callback
      );
    }

    /**
     * Returns the coordinate at `m` using linear interpolation, or `null` if no
     * such coordinate exists.
     *
     * `extrapolate` controls extrapolation beyond the range of Ms in the
     * MultiLineString. If `extrapolate` is `true` then Ms less than the first
     * M will return the first coordinate and Ms greater than the last M will
     * return the last coordinate.
     *
     * @param {number} m M.
     * @param {boolean} [extrapolate] Extrapolate. Default is `false`.
     * @return {import("../coordinate.js").Coordinate|null} Coordinate.
     * @api
     */
    getCoordinateAtM(m, extrapolate) {
      if (this.layout != 'XYM' && this.layout != 'XYZM') {
        return null;
      }
      extrapolate = extrapolate !== undefined ? extrapolate : false;
      return lineStringCoordinateAtM(
        this.flatCoordinates,
        0,
        this.flatCoordinates.length,
        this.stride,
        m,
        extrapolate
      );
    }

    /**
     * Return the coordinates of the linestring.
     * @return {Array<import("../coordinate.js").Coordinate>} Coordinates.
     * @api
     */
    getCoordinates() {
      return inflateCoordinates(
        this.flatCoordinates,
        0,
        this.flatCoordinates.length,
        this.stride
      );
    }

    /**
     * Return the coordinate at the provided fraction along the linestring.
     * The `fraction` is a number between 0 and 1, where 0 is the start of the
     * linestring and 1 is the end.
     * @param {number} fraction Fraction.
     * @param {import("../coordinate.js").Coordinate} [dest] Optional coordinate whose values will
     *     be modified. If not provided, a new coordinate will be returned.
     * @return {import("../coordinate.js").Coordinate} Coordinate of the interpolated point.
     * @api
     */
    getCoordinateAt(fraction, dest) {
      return interpolatePoint(
        this.flatCoordinates,
        0,
        this.flatCoordinates.length,
        this.stride,
        fraction,
        dest,
        this.stride
      );
    }

    /**
     * Return the length of the linestring on projected plane.
     * @return {number} Length (on projected plane).
     * @api
     */
    getLength() {
      return lineStringLength(
        this.flatCoordinates,
        0,
        this.flatCoordinates.length,
        this.stride
      );
    }

    /**
     * @return {Array<number>} Flat midpoint.
     */
    getFlatMidpoint() {
      if (this.flatMidpointRevision_ != this.getRevision()) {
        this.flatMidpoint_ = this.getCoordinateAt(0.5, this.flatMidpoint_);
        this.flatMidpointRevision_ = this.getRevision();
      }
      return this.flatMidpoint_;
    }

    /**
     * @param {number} squaredTolerance Squared tolerance.
     * @return {LineString} Simplified LineString.
     * @protected
     */
    getSimplifiedGeometryInternal(squaredTolerance) {
      const simplifiedFlatCoordinates = [];
      simplifiedFlatCoordinates.length = douglasPeucker(
        this.flatCoordinates,
        0,
        this.flatCoordinates.length,
        this.stride,
        squaredTolerance,
        simplifiedFlatCoordinates,
        0
      );
      return new LineString(simplifiedFlatCoordinates, 'XY');
    }

    /**
     * Get the type of this geometry.
     * @return {import("./Geometry.js").Type} Geometry type.
     * @api
     */
    getType() {
      return 'LineString';
    }

    /**
     * Test if the geometry and the passed extent intersect.
     * @param {import("../extent.js").Extent} extent Extent.
     * @return {boolean} `true` if the geometry and the extent intersect.
     * @api
     */
    intersectsExtent(extent) {
      return intersectsLineString(
        this.flatCoordinates,
        0,
        this.flatCoordinates.length,
        this.stride,
        extent
      );
    }

    /**
     * Set the coordinates of the linestring.
     * @param {!Array<import("../coordinate.js").Coordinate>} coordinates Coordinates.
     * @param {import("./Geometry.js").GeometryLayout} [layout] Layout.
     * @api
     */
    setCoordinates(coordinates, layout) {
      this.setLayout(layout, coordinates, 1);
      if (!this.flatCoordinates) {
        this.flatCoordinates = [];
      }
      this.flatCoordinates.length = deflateCoordinates(
        this.flatCoordinates,
        0,
        coordinates,
        this.stride
      );
      this.changed();
    }
  }

  /**
   * @module ol/geom/MultiLineString
   */

  /**
   * @classdesc
   * Multi-linestring geometry.
   *
   * @api
   */
  class MultiLineString extends SimpleGeometry {
    /**
     * @param {Array<Array<import("../coordinate.js").Coordinate>|LineString>|Array<number>} coordinates
     *     Coordinates or LineString geometries. (For internal use, flat coordinates in
     *     combination with `layout` and `ends` are also accepted.)
     * @param {import("./Geometry.js").GeometryLayout} [layout] Layout.
     * @param {Array<number>} [ends] Flat coordinate ends for internal use.
     */
    constructor(coordinates, layout, ends) {
      super();

      /**
       * @type {Array<number>}
       * @private
       */
      this.ends_ = [];

      /**
       * @private
       * @type {number}
       */
      this.maxDelta_ = -1;

      /**
       * @private
       * @type {number}
       */
      this.maxDeltaRevision_ = -1;

      if (Array.isArray(coordinates[0])) {
        this.setCoordinates(
          /** @type {Array<Array<import("../coordinate.js").Coordinate>>} */ (
            coordinates
          ),
          layout
        );
      } else if (layout !== undefined && ends) {
        this.setFlatCoordinates(
          layout,
          /** @type {Array<number>} */ (coordinates)
        );
        this.ends_ = ends;
      } else {
        let layout = this.getLayout();
        const lineStrings = /** @type {Array<LineString>} */ (coordinates);
        const flatCoordinates = [];
        const ends = [];
        for (let i = 0, ii = lineStrings.length; i < ii; ++i) {
          const lineString = lineStrings[i];
          if (i === 0) {
            layout = lineString.getLayout();
          }
          extend$2(flatCoordinates, lineString.getFlatCoordinates());
          ends.push(flatCoordinates.length);
        }
        this.setFlatCoordinates(layout, flatCoordinates);
        this.ends_ = ends;
      }
    }

    /**
     * Append the passed linestring to the multilinestring.
     * @param {LineString} lineString LineString.
     * @api
     */
    appendLineString(lineString) {
      if (!this.flatCoordinates) {
        this.flatCoordinates = lineString.getFlatCoordinates().slice();
      } else {
        extend$2(this.flatCoordinates, lineString.getFlatCoordinates().slice());
      }
      this.ends_.push(this.flatCoordinates.length);
      this.changed();
    }

    /**
     * Make a complete copy of the geometry.
     * @return {!MultiLineString} Clone.
     * @api
     */
    clone() {
      const multiLineString = new MultiLineString(
        this.flatCoordinates.slice(),
        this.layout,
        this.ends_.slice()
      );
      multiLineString.applyProperties(this);
      return multiLineString;
    }

    /**
     * @param {number} x X.
     * @param {number} y Y.
     * @param {import("../coordinate.js").Coordinate} closestPoint Closest point.
     * @param {number} minSquaredDistance Minimum squared distance.
     * @return {number} Minimum squared distance.
     */
    closestPointXY(x, y, closestPoint, minSquaredDistance) {
      if (minSquaredDistance < closestSquaredDistanceXY(this.getExtent(), x, y)) {
        return minSquaredDistance;
      }
      if (this.maxDeltaRevision_ != this.getRevision()) {
        this.maxDelta_ = Math.sqrt(
          arrayMaxSquaredDelta(
            this.flatCoordinates,
            0,
            this.ends_,
            this.stride,
            0
          )
        );
        this.maxDeltaRevision_ = this.getRevision();
      }
      return assignClosestArrayPoint(
        this.flatCoordinates,
        0,
        this.ends_,
        this.stride,
        this.maxDelta_,
        false,
        x,
        y,
        closestPoint,
        minSquaredDistance
      );
    }

    /**
     * Returns the coordinate at `m` using linear interpolation, or `null` if no
     * such coordinate exists.
     *
     * `extrapolate` controls extrapolation beyond the range of Ms in the
     * MultiLineString. If `extrapolate` is `true` then Ms less than the first
     * M will return the first coordinate and Ms greater than the last M will
     * return the last coordinate.
     *
     * `interpolate` controls interpolation between consecutive LineStrings
     * within the MultiLineString. If `interpolate` is `true` the coordinates
     * will be linearly interpolated between the last coordinate of one LineString
     * and the first coordinate of the next LineString.  If `interpolate` is
     * `false` then the function will return `null` for Ms falling between
     * LineStrings.
     *
     * @param {number} m M.
     * @param {boolean} [extrapolate] Extrapolate. Default is `false`.
     * @param {boolean} [interpolate] Interpolate. Default is `false`.
     * @return {import("../coordinate.js").Coordinate|null} Coordinate.
     * @api
     */
    getCoordinateAtM(m, extrapolate, interpolate) {
      if (
        (this.layout != 'XYM' && this.layout != 'XYZM') ||
        this.flatCoordinates.length === 0
      ) {
        return null;
      }
      extrapolate = extrapolate !== undefined ? extrapolate : false;
      interpolate = interpolate !== undefined ? interpolate : false;
      return lineStringsCoordinateAtM(
        this.flatCoordinates,
        0,
        this.ends_,
        this.stride,
        m,
        extrapolate,
        interpolate
      );
    }

    /**
     * Return the coordinates of the multilinestring.
     * @return {Array<Array<import("../coordinate.js").Coordinate>>} Coordinates.
     * @api
     */
    getCoordinates() {
      return inflateCoordinatesArray(
        this.flatCoordinates,
        0,
        this.ends_,
        this.stride
      );
    }

    /**
     * @return {Array<number>} Ends.
     */
    getEnds() {
      return this.ends_;
    }

    /**
     * Return the linestring at the specified index.
     * @param {number} index Index.
     * @return {LineString} LineString.
     * @api
     */
    getLineString(index) {
      if (index < 0 || this.ends_.length <= index) {
        return null;
      }
      return new LineString(
        this.flatCoordinates.slice(
          index === 0 ? 0 : this.ends_[index - 1],
          this.ends_[index]
        ),
        this.layout
      );
    }

    /**
     * Return the linestrings of this multilinestring.
     * @return {Array<LineString>} LineStrings.
     * @api
     */
    getLineStrings() {
      const flatCoordinates = this.flatCoordinates;
      const ends = this.ends_;
      const layout = this.layout;
      /** @type {Array<LineString>} */
      const lineStrings = [];
      let offset = 0;
      for (let i = 0, ii = ends.length; i < ii; ++i) {
        const end = ends[i];
        const lineString = new LineString(
          flatCoordinates.slice(offset, end),
          layout
        );
        lineStrings.push(lineString);
        offset = end;
      }
      return lineStrings;
    }

    /**
     * @return {Array<number>} Flat midpoints.
     */
    getFlatMidpoints() {
      const midpoints = [];
      const flatCoordinates = this.flatCoordinates;
      let offset = 0;
      const ends = this.ends_;
      const stride = this.stride;
      for (let i = 0, ii = ends.length; i < ii; ++i) {
        const end = ends[i];
        const midpoint = interpolatePoint(
          flatCoordinates,
          offset,
          end,
          stride,
          0.5
        );
        extend$2(midpoints, midpoint);
        offset = end;
      }
      return midpoints;
    }

    /**
     * @param {number} squaredTolerance Squared tolerance.
     * @return {MultiLineString} Simplified MultiLineString.
     * @protected
     */
    getSimplifiedGeometryInternal(squaredTolerance) {
      const simplifiedFlatCoordinates = [];
      const simplifiedEnds = [];
      simplifiedFlatCoordinates.length = douglasPeuckerArray(
        this.flatCoordinates,
        0,
        this.ends_,
        this.stride,
        squaredTolerance,
        simplifiedFlatCoordinates,
        0,
        simplifiedEnds
      );
      return new MultiLineString(simplifiedFlatCoordinates, 'XY', simplifiedEnds);
    }

    /**
     * Get the type of this geometry.
     * @return {import("./Geometry.js").Type} Geometry type.
     * @api
     */
    getType() {
      return 'MultiLineString';
    }

    /**
     * Test if the geometry and the passed extent intersect.
     * @param {import("../extent.js").Extent} extent Extent.
     * @return {boolean} `true` if the geometry and the extent intersect.
     * @api
     */
    intersectsExtent(extent) {
      return intersectsLineStringArray(
        this.flatCoordinates,
        0,
        this.ends_,
        this.stride,
        extent
      );
    }

    /**
     * Set the coordinates of the multilinestring.
     * @param {!Array<Array<import("../coordinate.js").Coordinate>>} coordinates Coordinates.
     * @param {import("./Geometry.js").GeometryLayout} [layout] Layout.
     * @api
     */
    setCoordinates(coordinates, layout) {
      this.setLayout(layout, coordinates, 2);
      if (!this.flatCoordinates) {
        this.flatCoordinates = [];
      }
      const ends = deflateCoordinatesArray(
        this.flatCoordinates,
        0,
        coordinates,
        this.stride,
        this.ends_
      );
      this.flatCoordinates.length = ends.length === 0 ? 0 : ends[ends.length - 1];
      this.changed();
    }
  }

  /**
   * @module ol/geom/MultiPoint
   */

  /**
   * @classdesc
   * Multi-point geometry.
   *
   * @api
   */
  class MultiPoint extends SimpleGeometry {
    /**
     * @param {Array<import("../coordinate.js").Coordinate>|Array<number>} coordinates Coordinates.
     *     For internal use, flat coordinates in combination with `layout` are also accepted.
     * @param {import("./Geometry.js").GeometryLayout} [layout] Layout.
     */
    constructor(coordinates, layout) {
      super();
      if (layout && !Array.isArray(coordinates[0])) {
        this.setFlatCoordinates(
          layout,
          /** @type {Array<number>} */ (coordinates)
        );
      } else {
        this.setCoordinates(
          /** @type {Array<import("../coordinate.js").Coordinate>} */ (
            coordinates
          ),
          layout
        );
      }
    }

    /**
     * Append the passed point to this multipoint.
     * @param {Point} point Point.
     * @api
     */
    appendPoint(point) {
      if (!this.flatCoordinates) {
        this.flatCoordinates = point.getFlatCoordinates().slice();
      } else {
        extend$2(this.flatCoordinates, point.getFlatCoordinates());
      }
      this.changed();
    }

    /**
     * Make a complete copy of the geometry.
     * @return {!MultiPoint} Clone.
     * @api
     */
    clone() {
      const multiPoint = new MultiPoint(
        this.flatCoordinates.slice(),
        this.layout
      );
      multiPoint.applyProperties(this);
      return multiPoint;
    }

    /**
     * @param {number} x X.
     * @param {number} y Y.
     * @param {import("../coordinate.js").Coordinate} closestPoint Closest point.
     * @param {number} minSquaredDistance Minimum squared distance.
     * @return {number} Minimum squared distance.
     */
    closestPointXY(x, y, closestPoint, minSquaredDistance) {
      if (minSquaredDistance < closestSquaredDistanceXY(this.getExtent(), x, y)) {
        return minSquaredDistance;
      }
      const flatCoordinates = this.flatCoordinates;
      const stride = this.stride;
      for (let i = 0, ii = flatCoordinates.length; i < ii; i += stride) {
        const squaredDistance = squaredDistance$1(
          x,
          y,
          flatCoordinates[i],
          flatCoordinates[i + 1]
        );
        if (squaredDistance < minSquaredDistance) {
          minSquaredDistance = squaredDistance;
          for (let j = 0; j < stride; ++j) {
            closestPoint[j] = flatCoordinates[i + j];
          }
          closestPoint.length = stride;
        }
      }
      return minSquaredDistance;
    }

    /**
     * Return the coordinates of the multipoint.
     * @return {Array<import("../coordinate.js").Coordinate>} Coordinates.
     * @api
     */
    getCoordinates() {
      return inflateCoordinates(
        this.flatCoordinates,
        0,
        this.flatCoordinates.length,
        this.stride
      );
    }

    /**
     * Return the point at the specified index.
     * @param {number} index Index.
     * @return {Point} Point.
     * @api
     */
    getPoint(index) {
      const n = !this.flatCoordinates
        ? 0
        : this.flatCoordinates.length / this.stride;
      if (index < 0 || n <= index) {
        return null;
      }
      return new Point(
        this.flatCoordinates.slice(
          index * this.stride,
          (index + 1) * this.stride
        ),
        this.layout
      );
    }

    /**
     * Return the points of this multipoint.
     * @return {Array<Point>} Points.
     * @api
     */
    getPoints() {
      const flatCoordinates = this.flatCoordinates;
      const layout = this.layout;
      const stride = this.stride;
      /** @type {Array<Point>} */
      const points = [];
      for (let i = 0, ii = flatCoordinates.length; i < ii; i += stride) {
        const point = new Point(flatCoordinates.slice(i, i + stride), layout);
        points.push(point);
      }
      return points;
    }

    /**
     * Get the type of this geometry.
     * @return {import("./Geometry.js").Type} Geometry type.
     * @api
     */
    getType() {
      return 'MultiPoint';
    }

    /**
     * Test if the geometry and the passed extent intersect.
     * @param {import("../extent.js").Extent} extent Extent.
     * @return {boolean} `true` if the geometry and the extent intersect.
     * @api
     */
    intersectsExtent(extent) {
      const flatCoordinates = this.flatCoordinates;
      const stride = this.stride;
      for (let i = 0, ii = flatCoordinates.length; i < ii; i += stride) {
        const x = flatCoordinates[i];
        const y = flatCoordinates[i + 1];
        if (containsXY(extent, x, y)) {
          return true;
        }
      }
      return false;
    }

    /**
     * Set the coordinates of the multipoint.
     * @param {!Array<import("../coordinate.js").Coordinate>} coordinates Coordinates.
     * @param {import("./Geometry.js").GeometryLayout} [layout] Layout.
     * @api
     */
    setCoordinates(coordinates, layout) {
      this.setLayout(layout, coordinates, 1);
      if (!this.flatCoordinates) {
        this.flatCoordinates = [];
      }
      this.flatCoordinates.length = deflateCoordinates(
        this.flatCoordinates,
        0,
        coordinates,
        this.stride
      );
      this.changed();
    }
  }

  /**
   * @module ol/geom/flat/center
   */

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {Array<Array<number>>} endss Endss.
   * @param {number} stride Stride.
   * @return {Array<number>} Flat centers.
   */
  function linearRingss(flatCoordinates, offset, endss, stride) {
    const flatCenters = [];
    let extent = createEmpty();
    for (let i = 0, ii = endss.length; i < ii; ++i) {
      const ends = endss[i];
      extent = createOrUpdateFromFlatCoordinates(
        flatCoordinates,
        offset,
        ends[0],
        stride
      );
      flatCenters.push((extent[0] + extent[2]) / 2, (extent[1] + extent[3]) / 2);
      offset = ends[ends.length - 1];
    }
    return flatCenters;
  }

  /**
   * @module ol/geom/MultiPolygon
   */

  /**
   * @classdesc
   * Multi-polygon geometry.
   *
   * @api
   */
  class MultiPolygon extends SimpleGeometry {
    /**
     * @param {Array<Array<Array<import("../coordinate.js").Coordinate>>|Polygon>|Array<number>} coordinates Coordinates.
     *     For internal use, flat coordinates in combination with `layout` and `endss` are also accepted.
     * @param {import("./Geometry.js").GeometryLayout} [layout] Layout.
     * @param {Array<Array<number>>} [endss] Array of ends for internal use with flat coordinates.
     */
    constructor(coordinates, layout, endss) {
      super();

      /**
       * @type {Array<Array<number>>}
       * @private
       */
      this.endss_ = [];

      /**
       * @private
       * @type {number}
       */
      this.flatInteriorPointsRevision_ = -1;

      /**
       * @private
       * @type {Array<number>}
       */
      this.flatInteriorPoints_ = null;

      /**
       * @private
       * @type {number}
       */
      this.maxDelta_ = -1;

      /**
       * @private
       * @type {number}
       */
      this.maxDeltaRevision_ = -1;

      /**
       * @private
       * @type {number}
       */
      this.orientedRevision_ = -1;

      /**
       * @private
       * @type {Array<number>}
       */
      this.orientedFlatCoordinates_ = null;

      if (!endss && !Array.isArray(coordinates[0])) {
        let thisLayout = this.getLayout();
        const polygons = /** @type {Array<Polygon>} */ (coordinates);
        const flatCoordinates = [];
        const thisEndss = [];
        for (let i = 0, ii = polygons.length; i < ii; ++i) {
          const polygon = polygons[i];
          if (i === 0) {
            thisLayout = polygon.getLayout();
          }
          const offset = flatCoordinates.length;
          const ends = polygon.getEnds();
          for (let j = 0, jj = ends.length; j < jj; ++j) {
            ends[j] += offset;
          }
          extend$2(flatCoordinates, polygon.getFlatCoordinates());
          thisEndss.push(ends);
        }
        layout = thisLayout;
        coordinates = flatCoordinates;
        endss = thisEndss;
      }
      if (layout !== undefined && endss) {
        this.setFlatCoordinates(
          layout,
          /** @type {Array<number>} */ (coordinates)
        );
        this.endss_ = endss;
      } else {
        this.setCoordinates(
          /** @type {Array<Array<Array<import("../coordinate.js").Coordinate>>>} */ (
            coordinates
          ),
          layout
        );
      }
    }

    /**
     * Append the passed polygon to this multipolygon.
     * @param {Polygon} polygon Polygon.
     * @api
     */
    appendPolygon(polygon) {
      /** @type {Array<number>} */
      let ends;
      if (!this.flatCoordinates) {
        this.flatCoordinates = polygon.getFlatCoordinates().slice();
        ends = polygon.getEnds().slice();
        this.endss_.push();
      } else {
        const offset = this.flatCoordinates.length;
        extend$2(this.flatCoordinates, polygon.getFlatCoordinates());
        ends = polygon.getEnds().slice();
        for (let i = 0, ii = ends.length; i < ii; ++i) {
          ends[i] += offset;
        }
      }
      this.endss_.push(ends);
      this.changed();
    }

    /**
     * Make a complete copy of the geometry.
     * @return {!MultiPolygon} Clone.
     * @api
     */
    clone() {
      const len = this.endss_.length;
      const newEndss = new Array(len);
      for (let i = 0; i < len; ++i) {
        newEndss[i] = this.endss_[i].slice();
      }

      const multiPolygon = new MultiPolygon(
        this.flatCoordinates.slice(),
        this.layout,
        newEndss
      );
      multiPolygon.applyProperties(this);

      return multiPolygon;
    }

    /**
     * @param {number} x X.
     * @param {number} y Y.
     * @param {import("../coordinate.js").Coordinate} closestPoint Closest point.
     * @param {number} minSquaredDistance Minimum squared distance.
     * @return {number} Minimum squared distance.
     */
    closestPointXY(x, y, closestPoint, minSquaredDistance) {
      if (minSquaredDistance < closestSquaredDistanceXY(this.getExtent(), x, y)) {
        return minSquaredDistance;
      }
      if (this.maxDeltaRevision_ != this.getRevision()) {
        this.maxDelta_ = Math.sqrt(
          multiArrayMaxSquaredDelta(
            this.flatCoordinates,
            0,
            this.endss_,
            this.stride,
            0
          )
        );
        this.maxDeltaRevision_ = this.getRevision();
      }
      return assignClosestMultiArrayPoint(
        this.getOrientedFlatCoordinates(),
        0,
        this.endss_,
        this.stride,
        this.maxDelta_,
        true,
        x,
        y,
        closestPoint,
        minSquaredDistance
      );
    }

    /**
     * @param {number} x X.
     * @param {number} y Y.
     * @return {boolean} Contains (x, y).
     */
    containsXY(x, y) {
      return linearRingssContainsXY(
        this.getOrientedFlatCoordinates(),
        0,
        this.endss_,
        this.stride,
        x,
        y
      );
    }

    /**
     * Return the area of the multipolygon on projected plane.
     * @return {number} Area (on projected plane).
     * @api
     */
    getArea() {
      return linearRingss$1(
        this.getOrientedFlatCoordinates(),
        0,
        this.endss_,
        this.stride
      );
    }

    /**
     * Get the coordinate array for this geometry.  This array has the structure
     * of a GeoJSON coordinate array for multi-polygons.
     *
     * @param {boolean} [right] Orient coordinates according to the right-hand
     *     rule (counter-clockwise for exterior and clockwise for interior rings).
     *     If `false`, coordinates will be oriented according to the left-hand rule
     *     (clockwise for exterior and counter-clockwise for interior rings).
     *     By default, coordinate orientation will depend on how the geometry was
     *     constructed.
     * @return {Array<Array<Array<import("../coordinate.js").Coordinate>>>} Coordinates.
     * @api
     */
    getCoordinates(right) {
      let flatCoordinates;
      if (right !== undefined) {
        flatCoordinates = this.getOrientedFlatCoordinates().slice();
        orientLinearRingsArray(
          flatCoordinates,
          0,
          this.endss_,
          this.stride,
          right
        );
      } else {
        flatCoordinates = this.flatCoordinates;
      }

      return inflateMultiCoordinatesArray(
        flatCoordinates,
        0,
        this.endss_,
        this.stride
      );
    }

    /**
     * @return {Array<Array<number>>} Endss.
     */
    getEndss() {
      return this.endss_;
    }

    /**
     * @return {Array<number>} Flat interior points.
     */
    getFlatInteriorPoints() {
      if (this.flatInteriorPointsRevision_ != this.getRevision()) {
        const flatCenters = linearRingss(
          this.flatCoordinates,
          0,
          this.endss_,
          this.stride
        );
        this.flatInteriorPoints_ = getInteriorPointsOfMultiArray(
          this.getOrientedFlatCoordinates(),
          0,
          this.endss_,
          this.stride,
          flatCenters
        );
        this.flatInteriorPointsRevision_ = this.getRevision();
      }
      return this.flatInteriorPoints_;
    }

    /**
     * Return the interior points as {@link module:ol/geom/MultiPoint~MultiPoint multipoint}.
     * @return {MultiPoint} Interior points as XYM coordinates, where M is
     * the length of the horizontal intersection that the point belongs to.
     * @api
     */
    getInteriorPoints() {
      return new MultiPoint(this.getFlatInteriorPoints().slice(), 'XYM');
    }

    /**
     * @return {Array<number>} Oriented flat coordinates.
     */
    getOrientedFlatCoordinates() {
      if (this.orientedRevision_ != this.getRevision()) {
        const flatCoordinates = this.flatCoordinates;
        if (
          linearRingssAreOriented(flatCoordinates, 0, this.endss_, this.stride)
        ) {
          this.orientedFlatCoordinates_ = flatCoordinates;
        } else {
          this.orientedFlatCoordinates_ = flatCoordinates.slice();
          this.orientedFlatCoordinates_.length = orientLinearRingsArray(
            this.orientedFlatCoordinates_,
            0,
            this.endss_,
            this.stride
          );
        }
        this.orientedRevision_ = this.getRevision();
      }
      return this.orientedFlatCoordinates_;
    }

    /**
     * @param {number} squaredTolerance Squared tolerance.
     * @return {MultiPolygon} Simplified MultiPolygon.
     * @protected
     */
    getSimplifiedGeometryInternal(squaredTolerance) {
      const simplifiedFlatCoordinates = [];
      const simplifiedEndss = [];
      simplifiedFlatCoordinates.length = quantizeMultiArray(
        this.flatCoordinates,
        0,
        this.endss_,
        this.stride,
        Math.sqrt(squaredTolerance),
        simplifiedFlatCoordinates,
        0,
        simplifiedEndss
      );
      return new MultiPolygon(simplifiedFlatCoordinates, 'XY', simplifiedEndss);
    }

    /**
     * Return the polygon at the specified index.
     * @param {number} index Index.
     * @return {Polygon} Polygon.
     * @api
     */
    getPolygon(index) {
      if (index < 0 || this.endss_.length <= index) {
        return null;
      }
      let offset;
      if (index === 0) {
        offset = 0;
      } else {
        const prevEnds = this.endss_[index - 1];
        offset = prevEnds[prevEnds.length - 1];
      }
      const ends = this.endss_[index].slice();
      const end = ends[ends.length - 1];
      if (offset !== 0) {
        for (let i = 0, ii = ends.length; i < ii; ++i) {
          ends[i] -= offset;
        }
      }
      return new Polygon(
        this.flatCoordinates.slice(offset, end),
        this.layout,
        ends
      );
    }

    /**
     * Return the polygons of this multipolygon.
     * @return {Array<Polygon>} Polygons.
     * @api
     */
    getPolygons() {
      const layout = this.layout;
      const flatCoordinates = this.flatCoordinates;
      const endss = this.endss_;
      const polygons = [];
      let offset = 0;
      for (let i = 0, ii = endss.length; i < ii; ++i) {
        const ends = endss[i].slice();
        const end = ends[ends.length - 1];
        if (offset !== 0) {
          for (let j = 0, jj = ends.length; j < jj; ++j) {
            ends[j] -= offset;
          }
        }
        const polygon = new Polygon(
          flatCoordinates.slice(offset, end),
          layout,
          ends
        );
        polygons.push(polygon);
        offset = end;
      }
      return polygons;
    }

    /**
     * Get the type of this geometry.
     * @return {import("./Geometry.js").Type} Geometry type.
     * @api
     */
    getType() {
      return 'MultiPolygon';
    }

    /**
     * Test if the geometry and the passed extent intersect.
     * @param {import("../extent.js").Extent} extent Extent.
     * @return {boolean} `true` if the geometry and the extent intersect.
     * @api
     */
    intersectsExtent(extent) {
      return intersectsLinearRingMultiArray(
        this.getOrientedFlatCoordinates(),
        0,
        this.endss_,
        this.stride,
        extent
      );
    }

    /**
     * Set the coordinates of the multipolygon.
     * @param {!Array<Array<Array<import("../coordinate.js").Coordinate>>>} coordinates Coordinates.
     * @param {import("./Geometry.js").GeometryLayout} [layout] Layout.
     * @api
     */
    setCoordinates(coordinates, layout) {
      this.setLayout(layout, coordinates, 3);
      if (!this.flatCoordinates) {
        this.flatCoordinates = [];
      }
      const endss = deflateMultiCoordinatesArray(
        this.flatCoordinates,
        0,
        coordinates,
        this.stride,
        this.endss_
      );
      if (endss.length === 0) {
        this.flatCoordinates.length = 0;
      } else {
        const lastEnds = endss[endss.length - 1];
        this.flatCoordinates.length =
          lastEnds.length === 0 ? 0 : lastEnds[lastEnds.length - 1];
      }
      this.changed();
    }
  }

  /**
   * @module ol/render/Feature
   */

  /**
   * @type {import("../transform.js").Transform}
   */
  const tmpTransform = create$2();

  /**
   * Lightweight, read-only, {@link module:ol/Feature~Feature} and {@link module:ol/geom/Geometry~Geometry} like
   * structure, optimized for vector tile rendering and styling. Geometry access
   * through the API is limited to getting the type and extent of the geometry.
   */
  class RenderFeature {
    /**
     * @param {import("../geom/Geometry.js").Type} type Geometry type.
     * @param {Array<number>} flatCoordinates Flat coordinates. These always need
     *     to be right-handed for polygons.
     * @param {Array<number>|Array<Array<number>>} ends Ends or Endss.
     * @param {Object<string, *>} properties Properties.
     * @param {number|string|undefined} id Feature id.
     */
    constructor(type, flatCoordinates, ends, properties, id) {
      /**
       * @type {import("../style/Style.js").StyleFunction|undefined}
       */
      this.styleFunction;

      /**
       * @private
       * @type {import("../extent.js").Extent|undefined}
       */
      this.extent_;

      /**
       * @private
       * @type {number|string|undefined}
       */
      this.id_ = id;

      /**
       * @private
       * @type {import("../geom/Geometry.js").Type}
       */
      this.type_ = type;

      /**
       * @private
       * @type {Array<number>}
       */
      this.flatCoordinates_ = flatCoordinates;

      /**
       * @private
       * @type {Array<number>}
       */
      this.flatInteriorPoints_ = null;

      /**
       * @private
       * @type {Array<number>}
       */
      this.flatMidpoints_ = null;

      /**
       * @private
       * @type {Array<number>|Array<Array<number>>}
       */
      this.ends_ = ends;

      /**
       * @private
       * @type {Object<string, *>}
       */
      this.properties_ = properties;
    }

    /**
     * Get a feature property by its key.
     * @param {string} key Key
     * @return {*} Value for the requested key.
     * @api
     */
    get(key) {
      return this.properties_[key];
    }

    /**
     * Get the extent of this feature's geometry.
     * @return {import("../extent.js").Extent} Extent.
     * @api
     */
    getExtent() {
      if (!this.extent_) {
        this.extent_ =
          this.type_ === 'Point'
            ? createOrUpdateFromCoordinate(this.flatCoordinates_)
            : createOrUpdateFromFlatCoordinates(
                this.flatCoordinates_,
                0,
                this.flatCoordinates_.length,
                2
              );
      }
      return this.extent_;
    }

    /**
     * @return {Array<number>} Flat interior points.
     */
    getFlatInteriorPoint() {
      if (!this.flatInteriorPoints_) {
        const flatCenter = getCenter(this.getExtent());
        this.flatInteriorPoints_ = getInteriorPointOfArray(
          this.flatCoordinates_,
          0,
          /** @type {Array<number>} */ (this.ends_),
          2,
          flatCenter,
          0
        );
      }
      return this.flatInteriorPoints_;
    }

    /**
     * @return {Array<number>} Flat interior points.
     */
    getFlatInteriorPoints() {
      if (!this.flatInteriorPoints_) {
        const flatCenters = linearRingss(
          this.flatCoordinates_,
          0,
          /** @type {Array<Array<number>>} */ (this.ends_),
          2
        );
        this.flatInteriorPoints_ = getInteriorPointsOfMultiArray(
          this.flatCoordinates_,
          0,
          /** @type {Array<Array<number>>} */ (this.ends_),
          2,
          flatCenters
        );
      }
      return this.flatInteriorPoints_;
    }

    /**
     * @return {Array<number>} Flat midpoint.
     */
    getFlatMidpoint() {
      if (!this.flatMidpoints_) {
        this.flatMidpoints_ = interpolatePoint(
          this.flatCoordinates_,
          0,
          this.flatCoordinates_.length,
          2,
          0.5
        );
      }
      return this.flatMidpoints_;
    }

    /**
     * @return {Array<number>} Flat midpoints.
     */
    getFlatMidpoints() {
      if (!this.flatMidpoints_) {
        this.flatMidpoints_ = [];
        const flatCoordinates = this.flatCoordinates_;
        let offset = 0;
        const ends = /** @type {Array<number>} */ (this.ends_);
        for (let i = 0, ii = ends.length; i < ii; ++i) {
          const end = ends[i];
          const midpoint = interpolatePoint(flatCoordinates, offset, end, 2, 0.5);
          extend$2(this.flatMidpoints_, midpoint);
          offset = end;
        }
      }
      return this.flatMidpoints_;
    }

    /**
     * Get the feature identifier.  This is a stable identifier for the feature and
     * is set when reading data from a remote source.
     * @return {number|string|undefined} Id.
     * @api
     */
    getId() {
      return this.id_;
    }

    /**
     * @return {Array<number>} Flat coordinates.
     */
    getOrientedFlatCoordinates() {
      return this.flatCoordinates_;
    }

    /**
     * For API compatibility with {@link module:ol/Feature~Feature}, this method is useful when
     * determining the geometry type in style function (see {@link #getType}).
     * @return {RenderFeature} Feature.
     * @api
     */
    getGeometry() {
      return this;
    }

    /**
     * @param {number} squaredTolerance Squared tolerance.
     * @return {RenderFeature} Simplified geometry.
     */
    getSimplifiedGeometry(squaredTolerance) {
      return this;
    }

    /**
     * Get a transformed and simplified version of the geometry.
     * @abstract
     * @param {number} squaredTolerance Squared tolerance.
     * @param {import("../proj.js").TransformFunction} [transform] Optional transform function.
     * @return {RenderFeature} Simplified geometry.
     */
    simplifyTransformed(squaredTolerance, transform) {
      return this;
    }

    /**
     * Get the feature properties.
     * @return {Object<string, *>} Feature properties.
     * @api
     */
    getProperties() {
      return this.properties_;
    }

    /**
     * @return {number} Stride.
     */
    getStride() {
      return 2;
    }

    /**
     * @return {import('../style/Style.js').StyleFunction|undefined} Style
     */
    getStyleFunction() {
      return this.styleFunction;
    }

    /**
     * Get the type of this feature's geometry.
     * @return {import("../geom/Geometry.js").Type} Geometry type.
     * @api
     */
    getType() {
      return this.type_;
    }

    /**
     * Transform geometry coordinates from tile pixel space to projected.
     *
     * @param {import("../proj.js").ProjectionLike} projection The data projection
     */
    transform(projection) {
      projection = get$2(projection);
      const pixelExtent = projection.getExtent();
      const projectedExtent = projection.getWorldExtent();
      if (pixelExtent && projectedExtent) {
        const scale = getHeight(projectedExtent) / getHeight(pixelExtent);
        compose(
          tmpTransform,
          projectedExtent[0],
          projectedExtent[3],
          scale,
          -scale,
          0,
          0,
          0
        );
        transform2D(
          this.flatCoordinates_,
          0,
          this.flatCoordinates_.length,
          2,
          tmpTransform,
          this.flatCoordinates_
        );
      }
    }
    /**
     * @return {Array<number>|Array<Array<number>>} Ends or endss.
     */
    getEnds() {
      return this.ends_;
    }
  }

  RenderFeature.prototype.getEndss = RenderFeature.prototype.getEnds;

  /**
   * @return {Array<number>} Flat coordinates.
   */
  RenderFeature.prototype.getFlatCoordinates =
    RenderFeature.prototype.getOrientedFlatCoordinates;

  /**
   * Create a geometry from an `ol/render/Feature`
   * @param {RenderFeature} renderFeature
   * Render Feature
   * @return {Point|MultiPoint|LineString|MultiLineString|Polygon|MultiPolygon}
   * New geometry instance.
   * @api
   */
  function toGeometry(renderFeature) {
    const geometryType = renderFeature.getType();
    switch (geometryType) {
      case 'Point':
        return new Point(renderFeature.getFlatCoordinates());
      case 'MultiPoint':
        return new MultiPoint(renderFeature.getFlatCoordinates(), 'XY');
      case 'LineString':
        return new LineString(renderFeature.getFlatCoordinates(), 'XY');
      case 'MultiLineString':
        return new MultiLineString(
          renderFeature.getFlatCoordinates(),
          'XY',
          /** @type {Array<number>} */ (renderFeature.getEnds())
        );
      case 'Polygon':
        const flatCoordinates = renderFeature.getFlatCoordinates();
        const ends = /** @type {Array<number>} */ (renderFeature.getEnds());
        const endss = inflateEnds(flatCoordinates, ends);
        return endss.length > 1
          ? new MultiPolygon(flatCoordinates, 'XY', endss)
          : new Polygon(flatCoordinates, 'XY', ends);
      default:
        throw new Error('Invalid geometry type:' + geometryType);
    }
  }

  /**
   * Create an `ol/Feature` from an `ol/render/Feature`
   * @param {RenderFeature} renderFeature RenderFeature
   * @param {string} [geometryName='geometry'] Geometry name to use
   * when creating the Feature.
   * @return {Feature} Newly constructed `ol/Feature` with properties,
   * geometry, and id copied over.
   * @api
   */
  function toFeature(renderFeature, geometryName) {
    const id = renderFeature.getId();
    const geometry = toGeometry(renderFeature);
    const properties = renderFeature.getProperties();
    const feature = new Feature();
    if (geometryName !== undefined) {
      feature.setGeometryName(geometryName);
    }
    feature.setGeometry(geometry);
    if (id !== undefined) {
      feature.setId(id);
    }
    feature.setProperties(properties, true);
    return feature;
  }

  /**
   * @module ol/proj/proj4
   */

  /**
   * Make projections defined in proj4 (with `proj4.defs()`) available in
   * OpenLayers. Requires proj4 >= 2.8.0.
   *
   * This function should be called whenever changes are made to the proj4
   * registry, e.g. after calling `proj4.defs()`. Existing transforms will not be
   * modified by this function.
   *
   * @param {?} proj4 Proj4.
   * @api
   */
  function register(proj4) {
    const projCodes = Object.keys(proj4.defs);
    const len = projCodes.length;
    let i, j;
    for (i = 0; i < len; ++i) {
      const code = projCodes[i];
      if (!get$2(code)) {
        const def = proj4.defs(code);
        let units = def.units;
        if (!units && def.projName === 'longlat') {
          units = 'degrees';
        }
        addProjection(
          new Projection({
            code: code,
            axisOrientation: def.axis,
            metersPerUnit: def.to_meter,
            units,
          })
        );
      }
    }
    for (i = 0; i < len; ++i) {
      const code1 = projCodes[i];
      const proj1 = get$2(code1);
      for (j = 0; j < len; ++j) {
        const code2 = projCodes[j];
        const proj2 = get$2(code2);
        if (!get$3(code1, code2)) {
          if (proj4.defs[code1] === proj4.defs[code2]) {
            addEquivalentProjections([proj1, proj2]);
          } else {
            const transform = proj4(code1, code2);
            addCoordinateTransforms(
              proj1,
              proj2,
              createSafeCoordinateTransform(proj1, proj2, transform.forward),
              createSafeCoordinateTransform(proj2, proj1, transform.inverse)
            );
          }
        }
      }
    }
  }

  /**
   * @module ol/layer/Vector
   */

  /**
   * @classdesc
   * Vector data is rendered client-side, as vectors. This layer type provides most accurate rendering
   * even during animations. Points and labels stay upright on rotated views. For very large
   * amounts of vector data, performance may suffer during pan and zoom animations. In this case,
   * try {@link module:ol/layer/VectorImage~VectorImageLayer}.
   *
   * Note that any property set in the options is set as a {@link module:ol/Object~BaseObject}
   * property on the layer object; for example, setting `title: 'My Title'` in the
   * options means that `title` is observable, and has get/set accessors.
   *
   * @template {import("../source/Vector.js").default} VectorSourceType
   * @extends {BaseVectorLayer<VectorSourceType, CanvasVectorLayerRenderer>}
   * @api
   */
  class VectorLayer extends BaseVectorLayer {
    /**
     * @param {import("./BaseVector.js").Options<VectorSourceType>} [options] Options.
     */
    constructor(options) {
      super(options);
    }

    createRenderer() {
      return new CanvasVectorLayerRenderer(this);
    }
  }

  /**
   * @module ol/geom/flat/geodesic
   */

  /**
   * @param {function(number): import("../../coordinate.js").Coordinate} interpolate Interpolate function.
   * @param {import("../../proj.js").TransformFunction} transform Transform from longitude/latitude to
   *     projected coordinates.
   * @param {number} squaredTolerance Squared tolerance.
   * @return {Array<number>} Flat coordinates.
   */
  function line(interpolate, transform, squaredTolerance) {
    // FIXME reduce garbage generation
    // FIXME optimize stack operations

    /** @type {Array<number>} */
    const flatCoordinates = [];

    let geoA = interpolate(0);
    let geoB = interpolate(1);

    let a = transform(geoA);
    let b = transform(geoB);

    /** @type {Array<import("../../coordinate.js").Coordinate>} */
    const geoStack = [geoB, geoA];
    /** @type {Array<import("../../coordinate.js").Coordinate>} */
    const stack = [b, a];
    /** @type {Array<number>} */
    const fractionStack = [1, 0];

    /** @type {!Object<string, boolean>} */
    const fractions = {};

    let maxIterations = 1e5;
    let geoM, m, fracA, fracB, fracM, key;

    while (--maxIterations > 0 && fractionStack.length > 0) {
      // Pop the a coordinate off the stack
      fracA = fractionStack.pop();
      geoA = geoStack.pop();
      a = stack.pop();
      // Add the a coordinate if it has not been added yet
      key = fracA.toString();
      if (!(key in fractions)) {
        flatCoordinates.push(a[0], a[1]);
        fractions[key] = true;
      }
      // Pop the b coordinate off the stack
      fracB = fractionStack.pop();
      geoB = geoStack.pop();
      b = stack.pop();
      // Find the m point between the a and b coordinates
      fracM = (fracA + fracB) / 2;
      geoM = interpolate(fracM);
      m = transform(geoM);
      if (
        squaredSegmentDistance(m[0], m[1], a[0], a[1], b[0], b[1]) <
        squaredTolerance
      ) {
        // If the m point is sufficiently close to the straight line, then we
        // discard it.  Just use the b coordinate and move on to the next line
        // segment.
        flatCoordinates.push(b[0], b[1]);
        key = fracB.toString();
        fractions[key] = true;
      } else {
        // Otherwise, we need to subdivide the current line segment.  Split it
        // into two and push the two line segments onto the stack.
        fractionStack.push(fracB, fracM, fracM, fracA);
        stack.push(b, m, m, a);
        geoStack.push(geoB, geoM, geoM, geoA);
      }
    }

    return flatCoordinates;
  }

  /**
   * Generate a great-circle arcs between two lat/lon points.
   * @param {number} lon1 Longitude 1 in degrees.
   * @param {number} lat1 Latitude 1 in degrees.
   * @param {number} lon2 Longitude 2 in degrees.
   * @param {number} lat2 Latitude 2 in degrees.
   * @param {import("../../proj/Projection.js").default} projection Projection.
   * @param {number} squaredTolerance Squared tolerance.
   * @return {Array<number>} Flat coordinates.
   */
  function greatCircleArc(
    lon1,
    lat1,
    lon2,
    lat2,
    projection,
    squaredTolerance
  ) {
    const geoProjection = get$2('EPSG:4326');

    const cosLat1 = Math.cos(toRadians(lat1));
    const sinLat1 = Math.sin(toRadians(lat1));
    const cosLat2 = Math.cos(toRadians(lat2));
    const sinLat2 = Math.sin(toRadians(lat2));
    const cosDeltaLon = Math.cos(toRadians(lon2 - lon1));
    const sinDeltaLon = Math.sin(toRadians(lon2 - lon1));
    const d = sinLat1 * sinLat2 + cosLat1 * cosLat2 * cosDeltaLon;

    return line(
      /**
       * @param {number} frac Fraction.
       * @return {import("../../coordinate.js").Coordinate} Coordinate.
       */
      function (frac) {
        if (1 <= d) {
          return [lon2, lat2];
        }
        const D = frac * Math.acos(d);
        const cosD = Math.cos(D);
        const sinD = Math.sin(D);
        const y = sinDeltaLon * cosLat2;
        const x = cosLat1 * sinLat2 - sinLat1 * cosLat2 * cosDeltaLon;
        const theta = Math.atan2(y, x);
        const lat = Math.asin(sinLat1 * cosD + cosLat1 * sinD * Math.cos(theta));
        const lon =
          toRadians(lon1) +
          Math.atan2(
            Math.sin(theta) * sinD * cosLat1,
            cosD - sinLat1 * Math.sin(lat)
          );
        return [toDegrees(lon), toDegrees(lat)];
      },
      getTransform(geoProjection, projection),
      squaredTolerance
    );
  }

  /**
   * Generate a meridian (line at constant longitude).
   * @param {number} lon Longitude.
   * @param {number} lat1 Latitude 1.
   * @param {number} lat2 Latitude 2.
   * @param {import("../../proj/Projection.js").default} projection Projection.
   * @param {number} squaredTolerance Squared tolerance.
   * @return {Array<number>} Flat coordinates.
   */
  function meridian(lon, lat1, lat2, projection, squaredTolerance) {
    const epsg4326Projection = get$2('EPSG:4326');
    return line(
      /**
       * @param {number} frac Fraction.
       * @return {import("../../coordinate.js").Coordinate} Coordinate.
       */
      function (frac) {
        return [lon, lat1 + (lat2 - lat1) * frac];
      },
      getTransform(epsg4326Projection, projection),
      squaredTolerance
    );
  }

  /**
   * Generate a parallel (line at constant latitude).
   * @param {number} lat Latitude.
   * @param {number} lon1 Longitude 1.
   * @param {number} lon2 Longitude 2.
   * @param {import("../../proj/Projection.js").default} projection Projection.
   * @param {number} squaredTolerance Squared tolerance.
   * @return {Array<number>} Flat coordinates.
   */
  function parallel(lat, lon1, lon2, projection, squaredTolerance) {
    const epsg4326Projection = get$2('EPSG:4326');
    return line(
      /**
       * @param {number} frac Fraction.
       * @return {import("../../coordinate.js").Coordinate} Coordinate.
       */
      function (frac) {
        return [lon1 + (lon2 - lon1) * frac, lat];
      },
      getTransform(epsg4326Projection, projection),
      squaredTolerance
    );
  }

  /**
   * @module ol/layer/Graticule
   */

  /**
   * @type {Stroke}
   * @private
   * @const
   */
  const DEFAULT_STROKE_STYLE$1 = new Stroke({
    color: 'rgba(0,0,0,0.2)',
  });

  /**
   * @type {Array<number>}
   * @private
   */
  const INTERVALS = [
    90,
    45,
    30,
    20,
    10,
    5,
    2,
    1,
    30 / 60,
    20 / 60,
    10 / 60,
    5 / 60,
    2 / 60,
    1 / 60,
    30 / 3600,
    20 / 3600,
    10 / 3600,
    5 / 3600,
    2 / 3600,
    1 / 3600,
  ];

  /**
   * @typedef {Object} GraticuleLabelDataType
   * @property {Point} geom Geometry.
   * @property {string} text Text.
   */

  /**
   * @typedef {Object} Options
   * @property {string} [className='ol-layer'] A CSS class name to set to the layer element.
   * @property {number} [opacity=1] Opacity (0, 1).
   * @property {boolean} [visible=true] Visibility.
   * @property {import("../extent.js").Extent} [extent] The bounding extent for layer rendering.  The layer will not be
   * rendered outside of this extent.
   * @property {number} [zIndex] The z-index for layer rendering.  At rendering time, the layers
   * will be ordered, first by Z-index and then by position. When `undefined`, a `zIndex` of 0 is assumed
   * for layers that are added to the map's `layers` collection, or `Infinity` when the layer's `setMap()`
   * method was used.
   * @property {number} [minResolution] The minimum resolution (inclusive) at which this layer will be
   * visible.
   * @property {number} [maxResolution] The maximum resolution (exclusive) below which this layer will
   * be visible.
   * @property {number} [minZoom] The minimum view zoom level (exclusive) above which this layer will be
   * visible.
   * @property {number} [maxZoom] The maximum view zoom level (inclusive) at which this layer will
   * be visible.
   * @property {number} [maxLines=100] The maximum number of meridians and
   * parallels from the center of the map. The default value of 100 means that at
   * most 200 meridians and 200 parallels will be displayed. The default value is
   * appropriate for conformal projections like Spherical Mercator. If you
   * increase the value, more lines will be drawn and the drawing performance will
   * decrease.
   * @property {Stroke} [strokeStyle] The
   * stroke style to use for drawing the graticule. If not provided, the following stroke will be used:
   * ```js
   * new Stroke({
   *   color: 'rgba(0, 0, 0, 0.2)' // a not fully opaque black
   * });
   * ```
   * @property {number} [targetSize=100] The target size of the graticule cells,
   * in pixels.
   * @property {boolean} [showLabels=false] Render a label with the respective
   * latitude/longitude for each graticule line.
   * @property {function(number):string} [lonLabelFormatter] Label formatter for
   * longitudes. This function is called with the longitude as argument, and
   * should return a formatted string representing the longitude. By default,
   * labels are formatted as degrees, minutes, seconds and hemisphere.
   * @property {function(number):string} [latLabelFormatter] Label formatter for
   * latitudes. This function is called with the latitude as argument, and
   * should return a formatted string representing the latitude. By default,
   * labels are formatted as degrees, minutes, seconds and hemisphere.
   * @property {number} [lonLabelPosition=0] Longitude label position in fractions
   * (0..1) of view extent. 0 means at the bottom of the viewport, 1 means at the
   * top.
   * @property {number} [latLabelPosition=1] Latitude label position in fractions
   * (0..1) of view extent. 0 means at the left of the viewport, 1 means at the
   * right.
   * @property {Text} [lonLabelStyle] Longitude label text
   * style. If not provided, the following style will be used:
   * ```js
   * new Text({
   *   font: '12px Calibri,sans-serif',
   *   textBaseline: 'bottom',
   *   fill: new Fill({
   *     color: 'rgba(0,0,0,1)'
   *   }),
   *   stroke: new Stroke({
   *     color: 'rgba(255,255,255,1)',
   *     width: 3
   *   })
   * });
   * ```
   * Note that the default's `textBaseline` configuration will not work well for
   * `lonLabelPosition` configurations that position labels close to the top of
   * the viewport.
   * @property {Text} [latLabelStyle] Latitude label text style.
   * If not provided, the following style will be used:
   * ```js
   * new Text({
   *   font: '12px Calibri,sans-serif',
   *   textAlign: 'end',
   *   fill: new Fill({
   *     color: 'rgba(0,0,0,1)'
   *   }),
   *   stroke: Stroke({
   *     color: 'rgba(255,255,255,1)',
   *     width: 3
   *   })
   * });
   * ```
   * Note that the default's `textAlign` configuration will not work well for
   * `latLabelPosition` configurations that position labels close to the left of
   * the viewport.
   * @property {Array<number>} [intervals=[90, 45, 30, 20, 10, 5, 2, 1, 30/60, 20/60, 10/60, 5/60, 2/60, 1/60, 30/3600, 20/3600, 10/3600, 5/3600, 2/3600, 1/3600]]
   * Intervals (in degrees) for the graticule. Example to limit graticules to 30 and 10 degrees intervals:
   * ```js
   * [30, 10]
   * ```
   * @property {boolean} [wrapX=true] Whether to repeat the graticule horizontally.
   * @property {Object<string, *>} [properties] Arbitrary observable properties. Can be accessed with `#get()` and `#set()`.
   */

  /**
   * @classdesc
   * Layer that renders a grid for a coordinate system (currently only EPSG:4326 is supported).
   * Note that the view projection must define both extent and worldExtent.
   *
   * @fires import("../render/Event.js").RenderEvent
   * @extends {VectorLayer<import("../source/Vector.js").default>}
   * @api
   */
  class Graticule extends VectorLayer {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      options = options ? options : {};

      const baseOptions = Object.assign(
        {
          updateWhileAnimating: true,
          updateWhileInteracting: true,
          renderBuffer: 0,
        },
        options
      );

      delete baseOptions.maxLines;
      delete baseOptions.strokeStyle;
      delete baseOptions.targetSize;
      delete baseOptions.showLabels;
      delete baseOptions.lonLabelFormatter;
      delete baseOptions.latLabelFormatter;
      delete baseOptions.lonLabelPosition;
      delete baseOptions.latLabelPosition;
      delete baseOptions.lonLabelStyle;
      delete baseOptions.latLabelStyle;
      delete baseOptions.intervals;
      super(baseOptions);

      /**
       * @type {import("../proj/Projection.js").default}
       */
      this.projection_ = null;

      /**
       * @type {number}
       * @private
       */
      this.maxLat_ = Infinity;

      /**
       * @type {number}
       * @private
       */
      this.maxLon_ = Infinity;

      /**
       * @type {number}
       * @private
       */
      this.minLat_ = -Infinity;

      /**
       * @type {number}
       * @private
       */
      this.minLon_ = -Infinity;

      /**
       * @type {number}
       * @private
       */
      this.maxX_ = Infinity;

      /**
       * @type {number}
       * @private
       */
      this.maxY_ = Infinity;

      /**
       * @type {number}
       * @private
       */
      this.minX_ = -Infinity;

      /**
       * @type {number}
       * @private
       */
      this.minY_ = -Infinity;

      /**
       * @type {number}
       * @private
       */
      this.targetSize_ =
        options.targetSize !== undefined ? options.targetSize : 100;

      /**
       * @type {number}
       * @private
       */
      this.maxLines_ = options.maxLines !== undefined ? options.maxLines : 100;

      /**
       * @type {Array<LineString>}
       * @private
       */
      this.meridians_ = [];

      /**
       * @type {Array<LineString>}
       * @private
       */
      this.parallels_ = [];

      /**
       * @type {Stroke}
       * @private
       */
      this.strokeStyle_ =
        options.strokeStyle !== undefined
          ? options.strokeStyle
          : DEFAULT_STROKE_STYLE$1;

      /**
       * @type {import("../proj.js").TransformFunction|undefined}
       * @private
       */
      this.fromLonLatTransform_ = undefined;

      /**
       * @type {import("../proj.js").TransformFunction|undefined}
       * @private
       */
      this.toLonLatTransform_ = undefined;

      /**
       * @type {import("../coordinate.js").Coordinate}
       * @private
       */
      this.projectionCenterLonLat_ = null;

      /**
       * @type {import("../coordinate.js").Coordinate}
       * @private
       */
      this.bottomLeft_ = null;

      /**
       * @type {import("../coordinate.js").Coordinate}
       * @private
       */
      this.bottomRight_ = null;

      /**
       * @type {import("../coordinate.js").Coordinate}
       * @private
       */
      this.topLeft_ = null;

      /**
       * @type {import("../coordinate.js").Coordinate}
       * @private
       */
      this.topRight_ = null;

      /**
       * @type {Array<GraticuleLabelDataType>}
       * @private
       */
      this.meridiansLabels_ = null;

      /**
       * @type {Array<GraticuleLabelDataType>}
       * @private
       */
      this.parallelsLabels_ = null;

      if (options.showLabels) {
        /**
         * @type {null|function(number):string}
         * @private
         */
        this.lonLabelFormatter_ =
          options.lonLabelFormatter == undefined
            ? degreesToStringHDMS.bind(this, 'EW')
            : options.lonLabelFormatter;

        /**
         * @type {function(number):string}
         * @private
         */
        this.latLabelFormatter_ =
          options.latLabelFormatter == undefined
            ? degreesToStringHDMS.bind(this, 'NS')
            : options.latLabelFormatter;

        /**
         * Longitude label position in fractions (0..1) of view extent. 0 means
         * bottom, 1 means top.
         * @type {number}
         * @private
         */
        this.lonLabelPosition_ =
          options.lonLabelPosition == undefined ? 0 : options.lonLabelPosition;

        /**
         * Latitude Label position in fractions (0..1) of view extent. 0 means left, 1
         * means right.
         * @type {number}
         * @private
         */
        this.latLabelPosition_ =
          options.latLabelPosition == undefined ? 1 : options.latLabelPosition;

        /**
         * @type {Style}
         * @private
         */
        this.lonLabelStyleBase_ = new Style({
          text:
            options.lonLabelStyle !== undefined
              ? options.lonLabelStyle.clone()
              : new Text({
                  font: '12px Calibri,sans-serif',
                  textBaseline: 'bottom',
                  fill: new Fill({
                    color: 'rgba(0,0,0,1)',
                  }),
                  stroke: new Stroke({
                    color: 'rgba(255,255,255,1)',
                    width: 3,
                  }),
                }),
        });

        /**
         * @private
         * @param {import("../Feature").default} feature Feature
         * @return {Style} style
         */
        this.lonLabelStyle_ = function (feature) {
          const label = feature.get('graticule_label');
          this.lonLabelStyleBase_.getText().setText(label);
          return this.lonLabelStyleBase_;
        }.bind(this);

        /**
         * @type {Style}
         * @private
         */
        this.latLabelStyleBase_ = new Style({
          text:
            options.latLabelStyle !== undefined
              ? options.latLabelStyle.clone()
              : new Text({
                  font: '12px Calibri,sans-serif',
                  textAlign: 'right',
                  fill: new Fill({
                    color: 'rgba(0,0,0,1)',
                  }),
                  stroke: new Stroke({
                    color: 'rgba(255,255,255,1)',
                    width: 3,
                  }),
                }),
        });

        /**
         * @private
         * @param {import("../Feature").default} feature Feature
         * @return {Style} style
         */
        this.latLabelStyle_ = function (feature) {
          const label = feature.get('graticule_label');
          this.latLabelStyleBase_.getText().setText(label);
          return this.latLabelStyleBase_;
        }.bind(this);

        this.meridiansLabels_ = [];
        this.parallelsLabels_ = [];

        this.addEventListener(EventType.POSTRENDER, this.drawLabels_.bind(this));
      }

      /**
       * @type {Array<number>}
       * @private
       */
      this.intervals_ =
        options.intervals !== undefined ? options.intervals : INTERVALS;

      // use a source with a custom loader for lines & text
      this.setSource(
        new VectorSource({
          loader: this.loaderFunction.bind(this),
          strategy: this.strategyFunction.bind(this),
          features: new Collection(),
          overlaps: false,
          useSpatialIndex: false,
          wrapX: options.wrapX,
        })
      );

      /**
       * feature pool to use when updating graticule
       * @type {Array<Feature>}
       * @private
       */
      this.featurePool_ = [];

      /**
       * @type {Style}
       * @private
       */
      this.lineStyle_ = new Style({
        stroke: this.strokeStyle_,
      });

      /**
       * @type {?import("../extent.js").Extent}
       * @private
       */
      this.loadedExtent_ = null;

      /**
       * @type {?import("../extent.js").Extent}
       * @private
       */
      this.renderedExtent_ = null;

      /**
       * @type {?number}
       * @private
       */
      this.renderedResolution_ = null;

      this.setRenderOrder(null);
    }

    /**
     * Strategy function for loading features based on the view's extent and
     * resolution.
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {number} resolution Resolution.
     * @return {Array<import("../extent.js").Extent>} Extents.
     */
    strategyFunction(extent, resolution) {
      // extents may be passed in different worlds, to avoid endless loop we use only one
      let realWorldExtent = extent.slice();
      if (this.projection_ && this.getSource().getWrapX()) {
        wrapX$2(realWorldExtent, this.projection_);
      }
      if (this.loadedExtent_) {
        if (
          approximatelyEquals(this.loadedExtent_, realWorldExtent, resolution)
        ) {
          // make sure result is exactly equal to previous extent
          realWorldExtent = this.loadedExtent_.slice();
        } else {
          // we should not keep track of loaded extents
          this.getSource().removeLoadedExtent(this.loadedExtent_);
        }
      }
      return [realWorldExtent];
    }

    /**
     * Update geometries in the source based on current view
     * @param {import("../extent").Extent} extent Extent
     * @param {number} resolution Resolution
     * @param {import("../proj/Projection.js").default} projection Projection
     */
    loaderFunction(extent, resolution, projection) {
      this.loadedExtent_ = extent;
      const source = this.getSource();

      // only consider the intersection between our own extent & the requested one
      const layerExtent = this.getExtent() || [
        -Infinity,
        -Infinity,
        Infinity,
        Infinity,
      ];
      const renderExtent = getIntersection(layerExtent, extent);

      if (
        this.renderedExtent_ &&
        equals$2(this.renderedExtent_, renderExtent) &&
        this.renderedResolution_ === resolution
      ) {
        return;
      }
      this.renderedExtent_ = renderExtent;
      this.renderedResolution_ = resolution;

      // bail out if nothing to render
      if (isEmpty(renderExtent)) {
        return;
      }

      // update projection info
      const center = getCenter(renderExtent);
      const squaredTolerance = (resolution * resolution) / 4;

      const updateProjectionInfo =
        !this.projection_ || !equivalent(this.projection_, projection);

      if (updateProjectionInfo) {
        this.updateProjectionInfo_(projection);
      }

      this.createGraticule_(renderExtent, center, resolution, squaredTolerance);

      // first make sure we have enough features in the pool
      let featureCount = this.meridians_.length + this.parallels_.length;
      if (this.meridiansLabels_) {
        featureCount += this.meridians_.length;
      }
      if (this.parallelsLabels_) {
        featureCount += this.parallels_.length;
      }

      let feature;
      while (featureCount > this.featurePool_.length) {
        feature = new Feature();
        this.featurePool_.push(feature);
      }

      const featuresColl = source.getFeaturesCollection();
      featuresColl.clear();
      let poolIndex = 0;

      // add features for the lines & labels
      let i, l;
      for (i = 0, l = this.meridians_.length; i < l; ++i) {
        feature = this.featurePool_[poolIndex++];
        feature.setGeometry(this.meridians_[i]);
        feature.setStyle(this.lineStyle_);
        featuresColl.push(feature);
      }
      for (i = 0, l = this.parallels_.length; i < l; ++i) {
        feature = this.featurePool_[poolIndex++];
        feature.setGeometry(this.parallels_[i]);
        feature.setStyle(this.lineStyle_);
        featuresColl.push(feature);
      }
    }

    /**
     * @param {number} lon Longitude.
     * @param {number} minLat Minimal latitude.
     * @param {number} maxLat Maximal latitude.
     * @param {number} squaredTolerance Squared tolerance.
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {number} index Index.
     * @return {number} Index.
     * @private
     */
    addMeridian_(lon, minLat, maxLat, squaredTolerance, extent, index) {
      const lineString = this.getMeridian_(
        lon,
        minLat,
        maxLat,
        squaredTolerance,
        index
      );
      if (intersects$3(lineString.getExtent(), extent)) {
        if (this.meridiansLabels_) {
          const text = this.lonLabelFormatter_(lon);
          if (index in this.meridiansLabels_) {
            this.meridiansLabels_[index].text = text;
          } else {
            this.meridiansLabels_[index] = {
              geom: new Point([]),
              text: text,
            };
          }
        }
        this.meridians_[index++] = lineString;
      }
      return index;
    }

    /**
     * @param {number} lat Latitude.
     * @param {number} minLon Minimal longitude.
     * @param {number} maxLon Maximal longitude.
     * @param {number} squaredTolerance Squared tolerance.
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {number} index Index.
     * @return {number} Index.
     * @private
     */
    addParallel_(lat, minLon, maxLon, squaredTolerance, extent, index) {
      const lineString = this.getParallel_(
        lat,
        minLon,
        maxLon,
        squaredTolerance,
        index
      );
      if (intersects$3(lineString.getExtent(), extent)) {
        if (this.parallelsLabels_) {
          const text = this.latLabelFormatter_(lat);
          if (index in this.parallelsLabels_) {
            this.parallelsLabels_[index].text = text;
          } else {
            this.parallelsLabels_[index] = {
              geom: new Point([]),
              text: text,
            };
          }
        }
        this.parallels_[index++] = lineString;
      }
      return index;
    }

    /**
     * @param {import("../render/Event.js").default} event Render event.
     * @private
     */
    drawLabels_(event) {
      const rotation = event.frameState.viewState.rotation;
      const resolution = event.frameState.viewState.resolution;
      const size = event.frameState.size;
      const extent = event.frameState.extent;
      const rotationCenter = getCenter(extent);
      let rotationExtent = extent;
      if (rotation) {
        const unrotatedWidth = size[0] * resolution;
        const unrotatedHeight = size[1] * resolution;
        rotationExtent = [
          rotationCenter[0] - unrotatedWidth / 2,
          rotationCenter[1] - unrotatedHeight / 2,
          rotationCenter[0] + unrotatedWidth / 2,
          rotationCenter[1] + unrotatedHeight / 2,
        ];
      }

      let startWorld = 0;
      let endWorld = 0;
      let labelsAtStart = this.latLabelPosition_ < 0.5;
      const projectionExtent = this.projection_.getExtent();
      const worldWidth = getWidth(projectionExtent);
      if (
        this.getSource().getWrapX() &&
        this.projection_.canWrapX() &&
        !containsExtent(projectionExtent, extent)
      ) {
        startWorld = Math.floor((extent[0] - projectionExtent[0]) / worldWidth);
        endWorld = Math.ceil((extent[2] - projectionExtent[2]) / worldWidth);
        const inverted = Math.abs(rotation) > Math.PI / 2;
        labelsAtStart = labelsAtStart !== inverted;
      }
      const vectorContext = getVectorContext(event);

      for (let world = startWorld; world <= endWorld; ++world) {
        let poolIndex = this.meridians_.length + this.parallels_.length;
        let feature, index, l, textPoint;

        if (this.meridiansLabels_) {
          for (index = 0, l = this.meridiansLabels_.length; index < l; ++index) {
            const lineString = this.meridians_[index];
            if (!rotation && world === 0) {
              textPoint = this.getMeridianPoint_(lineString, extent, index);
            } else {
              const clone = lineString.clone();
              clone.translate(world * worldWidth, 0);
              clone.rotate(-rotation, rotationCenter);
              textPoint = this.getMeridianPoint_(clone, rotationExtent, index);
              textPoint.rotate(rotation, rotationCenter);
            }
            feature = this.featurePool_[poolIndex++];
            feature.setGeometry(textPoint);
            feature.set('graticule_label', this.meridiansLabels_[index].text);
            vectorContext.drawFeature(feature, this.lonLabelStyle_(feature));
          }
        }
        if (this.parallelsLabels_) {
          if (
            (world === startWorld && labelsAtStart) ||
            (world === endWorld && !labelsAtStart)
          ) {
            for (index = 0, l = this.parallels_.length; index < l; ++index) {
              const lineString = this.parallels_[index];
              if (!rotation && world === 0) {
                textPoint = this.getParallelPoint_(lineString, extent, index);
              } else {
                const clone = lineString.clone();
                clone.translate(world * worldWidth, 0);
                clone.rotate(-rotation, rotationCenter);
                textPoint = this.getParallelPoint_(clone, rotationExtent, index);
                textPoint.rotate(rotation, rotationCenter);
              }
              feature = this.featurePool_[poolIndex++];
              feature.setGeometry(textPoint);
              feature.set('graticule_label', this.parallelsLabels_[index].text);
              vectorContext.drawFeature(feature, this.latLabelStyle_(feature));
            }
          }
        }
      }
    }

    /**
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {import("../coordinate.js").Coordinate} center Center.
     * @param {number} resolution Resolution.
     * @param {number} squaredTolerance Squared tolerance.
     * @private
     */
    createGraticule_(extent, center, resolution, squaredTolerance) {
      const interval = this.getInterval_(resolution);
      if (interval == -1) {
        this.meridians_.length = 0;
        this.parallels_.length = 0;
        if (this.meridiansLabels_) {
          this.meridiansLabels_.length = 0;
        }
        if (this.parallelsLabels_) {
          this.parallelsLabels_.length = 0;
        }
        return;
      }

      let wrapX = false;
      const projectionExtent = this.projection_.getExtent();
      const worldWidth = getWidth(projectionExtent);
      if (
        this.getSource().getWrapX() &&
        this.projection_.canWrapX() &&
        !containsExtent(projectionExtent, extent)
      ) {
        if (getWidth(extent) >= worldWidth) {
          extent[0] = projectionExtent[0];
          extent[2] = projectionExtent[2];
        } else {
          wrapX = true;
        }
      }

      // Constrain the center to fit into the extent available to the graticule

      const validCenterP = [
        clamp(center[0], this.minX_, this.maxX_),
        clamp(center[1], this.minY_, this.maxY_),
      ];

      // Transform the center to lon lat
      // Some projections may have a void area at the poles
      // so replace any NaN latitudes with the min or max value closest to a pole

      const centerLonLat = this.toLonLatTransform_(validCenterP);
      if (isNaN(centerLonLat[1])) {
        centerLonLat[1] =
          Math.abs(this.maxLat_) >= Math.abs(this.minLat_)
            ? this.maxLat_
            : this.minLat_;
      }
      let centerLon = clamp(centerLonLat[0], this.minLon_, this.maxLon_);
      let centerLat = clamp(centerLonLat[1], this.minLat_, this.maxLat_);
      const maxLines = this.maxLines_;
      let cnt, idx, lat, lon;

      // Limit the extent to fit into the extent available to the graticule

      let validExtentP = extent;
      if (!wrapX) {
        validExtentP = [
          clamp(extent[0], this.minX_, this.maxX_),
          clamp(extent[1], this.minY_, this.maxY_),
          clamp(extent[2], this.minX_, this.maxX_),
          clamp(extent[3], this.minY_, this.maxY_),
        ];
      }

      // Transform the extent to get the lon lat ranges for the edges of the extent

      const validExtent = applyTransform(
        validExtentP,
        this.toLonLatTransform_,
        undefined,
        8
      );

      let maxLat = validExtent[3];
      let maxLon = validExtent[2];
      let minLat = validExtent[1];
      let minLon = validExtent[0];

      if (!wrapX) {
        // Check if extremities of the world extent lie inside the extent
        // (for example the pole in a polar projection)
        // and extend the extent as appropriate

        if (containsCoordinate(validExtentP, this.bottomLeft_)) {
          minLon = this.minLon_;
          minLat = this.minLat_;
        }
        if (containsCoordinate(validExtentP, this.bottomRight_)) {
          maxLon = this.maxLon_;
          minLat = this.minLat_;
        }
        if (containsCoordinate(validExtentP, this.topLeft_)) {
          minLon = this.minLon_;
          maxLat = this.maxLat_;
        }
        if (containsCoordinate(validExtentP, this.topRight_)) {
          maxLon = this.maxLon_;
          maxLat = this.maxLat_;
        }

        // The transformed center may also extend the lon lat ranges used for rendering

        maxLat = clamp(maxLat, centerLat, this.maxLat_);
        maxLon = clamp(maxLon, centerLon, this.maxLon_);
        minLat = clamp(minLat, this.minLat_, centerLat);
        minLon = clamp(minLon, this.minLon_, centerLon);
      }

      // Create meridians

      centerLon = Math.floor(centerLon / interval) * interval;
      lon = clamp(centerLon, this.minLon_, this.maxLon_);

      idx = this.addMeridian_(lon, minLat, maxLat, squaredTolerance, extent, 0);

      cnt = 0;
      if (wrapX) {
        while ((lon -= interval) >= minLon && cnt++ < maxLines) {
          idx = this.addMeridian_(
            lon,
            minLat,
            maxLat,
            squaredTolerance,
            extent,
            idx
          );
        }
      } else {
        while (lon != this.minLon_ && cnt++ < maxLines) {
          lon = Math.max(lon - interval, this.minLon_);
          idx = this.addMeridian_(
            lon,
            minLat,
            maxLat,
            squaredTolerance,
            extent,
            idx
          );
        }
      }

      lon = clamp(centerLon, this.minLon_, this.maxLon_);

      cnt = 0;
      if (wrapX) {
        while ((lon += interval) <= maxLon && cnt++ < maxLines) {
          idx = this.addMeridian_(
            lon,
            minLat,
            maxLat,
            squaredTolerance,
            extent,
            idx
          );
        }
      } else {
        while (lon != this.maxLon_ && cnt++ < maxLines) {
          lon = Math.min(lon + interval, this.maxLon_);
          idx = this.addMeridian_(
            lon,
            minLat,
            maxLat,
            squaredTolerance,
            extent,
            idx
          );
        }
      }

      this.meridians_.length = idx;
      if (this.meridiansLabels_) {
        this.meridiansLabels_.length = idx;
      }

      // Create parallels

      centerLat = Math.floor(centerLat / interval) * interval;
      lat = clamp(centerLat, this.minLat_, this.maxLat_);

      idx = this.addParallel_(lat, minLon, maxLon, squaredTolerance, extent, 0);

      cnt = 0;
      while (lat != this.minLat_ && cnt++ < maxLines) {
        lat = Math.max(lat - interval, this.minLat_);
        idx = this.addParallel_(
          lat,
          minLon,
          maxLon,
          squaredTolerance,
          extent,
          idx
        );
      }

      lat = clamp(centerLat, this.minLat_, this.maxLat_);

      cnt = 0;
      while (lat != this.maxLat_ && cnt++ < maxLines) {
        lat = Math.min(lat + interval, this.maxLat_);
        idx = this.addParallel_(
          lat,
          minLon,
          maxLon,
          squaredTolerance,
          extent,
          idx
        );
      }

      this.parallels_.length = idx;
      if (this.parallelsLabels_) {
        this.parallelsLabels_.length = idx;
      }
    }

    /**
     * @param {number} resolution Resolution.
     * @return {number} The interval in degrees.
     * @private
     */
    getInterval_(resolution) {
      const centerLon = this.projectionCenterLonLat_[0];
      const centerLat = this.projectionCenterLonLat_[1];
      let interval = -1;
      const target = Math.pow(this.targetSize_ * resolution, 2);
      /** @type {Array<number>} **/
      const p1 = [];
      /** @type {Array<number>} **/
      const p2 = [];
      for (let i = 0, ii = this.intervals_.length; i < ii; ++i) {
        const delta = clamp(this.intervals_[i] / 2, 0, 90);
        // Don't attempt to transform latitudes beyond the poles!
        const clampedLat = clamp(centerLat, -90 + delta, 90 - delta);
        p1[0] = centerLon - delta;
        p1[1] = clampedLat - delta;
        p2[0] = centerLon + delta;
        p2[1] = clampedLat + delta;
        this.fromLonLatTransform_(p1, p1);
        this.fromLonLatTransform_(p2, p2);
        const dist = Math.pow(p2[0] - p1[0], 2) + Math.pow(p2[1] - p1[1], 2);
        if (dist <= target) {
          break;
        }
        interval = this.intervals_[i];
      }
      return interval;
    }

    /**
     * @param {number} lon Longitude.
     * @param {number} minLat Minimal latitude.
     * @param {number} maxLat Maximal latitude.
     * @param {number} squaredTolerance Squared tolerance.
     * @return {LineString} The meridian line string.
     * @param {number} index Index.
     * @private
     */
    getMeridian_(lon, minLat, maxLat, squaredTolerance, index) {
      const flatCoordinates = meridian(
        lon,
        minLat,
        maxLat,
        this.projection_,
        squaredTolerance
      );
      let lineString = this.meridians_[index];
      if (!lineString) {
        lineString = new LineString(flatCoordinates, 'XY');
        this.meridians_[index] = lineString;
      } else {
        lineString.setFlatCoordinates('XY', flatCoordinates);
        lineString.changed();
      }
      return lineString;
    }

    /**
     * @param {LineString} lineString Meridian
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {number} index Index.
     * @return {Point} Meridian point.
     * @private
     */
    getMeridianPoint_(lineString, extent, index) {
      const flatCoordinates = lineString.getFlatCoordinates();
      let bottom = 1;
      let top = flatCoordinates.length - 1;
      if (flatCoordinates[bottom] > flatCoordinates[top]) {
        bottom = top;
        top = 1;
      }
      const clampedBottom = Math.max(extent[1], flatCoordinates[bottom]);
      const clampedTop = Math.min(extent[3], flatCoordinates[top]);
      const lat = clamp(
        extent[1] + Math.abs(extent[1] - extent[3]) * this.lonLabelPosition_,
        clampedBottom,
        clampedTop
      );
      const coordinate0 =
        flatCoordinates[bottom - 1] +
        ((flatCoordinates[top - 1] - flatCoordinates[bottom - 1]) *
          (lat - flatCoordinates[bottom])) /
          (flatCoordinates[top] - flatCoordinates[bottom]);
      const coordinate = [coordinate0, lat];
      const point = this.meridiansLabels_[index].geom;
      point.setCoordinates(coordinate);
      return point;
    }

    /**
     * Get the list of meridians.  Meridians are lines of equal longitude.
     * @return {Array<LineString>} The meridians.
     * @api
     */
    getMeridians() {
      return this.meridians_;
    }

    /**
     * @param {number} lat Latitude.
     * @param {number} minLon Minimal longitude.
     * @param {number} maxLon Maximal longitude.
     * @param {number} squaredTolerance Squared tolerance.
     * @return {LineString} The parallel line string.
     * @param {number} index Index.
     * @private
     */
    getParallel_(lat, minLon, maxLon, squaredTolerance, index) {
      const flatCoordinates = parallel(
        lat,
        minLon,
        maxLon,
        this.projection_,
        squaredTolerance
      );
      let lineString = this.parallels_[index];
      if (!lineString) {
        lineString = new LineString(flatCoordinates, 'XY');
      } else {
        lineString.setFlatCoordinates('XY', flatCoordinates);
        lineString.changed();
      }
      return lineString;
    }

    /**
     * @param {LineString} lineString Parallels.
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {number} index Index.
     * @return {Point} Parallel point.
     * @private
     */
    getParallelPoint_(lineString, extent, index) {
      const flatCoordinates = lineString.getFlatCoordinates();
      let left = 0;
      let right = flatCoordinates.length - 2;
      if (flatCoordinates[left] > flatCoordinates[right]) {
        left = right;
        right = 0;
      }
      const clampedLeft = Math.max(extent[0], flatCoordinates[left]);
      const clampedRight = Math.min(extent[2], flatCoordinates[right]);
      const lon = clamp(
        extent[0] + Math.abs(extent[0] - extent[2]) * this.latLabelPosition_,
        clampedLeft,
        clampedRight
      );
      const coordinate1 =
        flatCoordinates[left + 1] +
        ((flatCoordinates[right + 1] - flatCoordinates[left + 1]) *
          (lon - flatCoordinates[left])) /
          (flatCoordinates[right] - flatCoordinates[left]);
      const coordinate = [lon, coordinate1];
      const point = this.parallelsLabels_[index].geom;
      point.setCoordinates(coordinate);
      return point;
    }

    /**
     * Get the list of parallels.  Parallels are lines of equal latitude.
     * @return {Array<LineString>} The parallels.
     * @api
     */
    getParallels() {
      return this.parallels_;
    }

    /**
     * @param {import("../proj/Projection.js").default} projection Projection.
     * @private
     */
    updateProjectionInfo_(projection) {
      const epsg4326Projection = get$2('EPSG:4326');

      const worldExtent = projection.getWorldExtent();

      this.maxLat_ = worldExtent[3];
      this.maxLon_ = worldExtent[2];
      this.minLat_ = worldExtent[1];
      this.minLon_ = worldExtent[0];

      // If the world extent crosses the dateline define a custom transform to
      // return longitudes which wrap the dateline

      const toLonLatTransform = getTransform(projection, epsg4326Projection);
      if (this.minLon_ < this.maxLon_) {
        this.toLonLatTransform_ = toLonLatTransform;
      } else {
        const split = this.minLon_ + this.maxLon_ / 2;
        this.maxLon_ += 360;
        this.toLonLatTransform_ = function (coordinates, output, dimension) {
          dimension = dimension || 2;
          const lonLatCoordinates = toLonLatTransform(
            coordinates,
            output,
            dimension
          );
          for (let i = 0, l = lonLatCoordinates.length; i < l; i += dimension) {
            if (lonLatCoordinates[i] < split) {
              lonLatCoordinates[i] += 360;
            }
          }
          return lonLatCoordinates;
        };
      }

      // Transform the extent to get the limits of the view projection extent
      // which should be available to the graticule

      this.fromLonLatTransform_ = getTransform(epsg4326Projection, projection);
      const worldExtentP = applyTransform(
        [this.minLon_, this.minLat_, this.maxLon_, this.maxLat_],
        this.fromLonLatTransform_,
        undefined,
        8
      );

      this.minX_ = worldExtentP[0];
      this.maxX_ = worldExtentP[2];
      this.minY_ = worldExtentP[1];
      this.maxY_ = worldExtentP[3];

      // Determine the view projection coordinates of the extremities of the world extent
      // as these may lie inside a view extent (for example the pole in a polar projection)

      this.bottomLeft_ = this.fromLonLatTransform_([this.minLon_, this.minLat_]);
      this.bottomRight_ = this.fromLonLatTransform_([this.maxLon_, this.minLat_]);
      this.topLeft_ = this.fromLonLatTransform_([this.minLon_, this.maxLat_]);
      this.topRight_ = this.fromLonLatTransform_([this.maxLon_, this.maxLat_]);

      // Transform the projection center to lon lat
      // Some projections may have a void area at the poles
      // so replace any NaN latitudes with the min or max value closest to a pole

      this.projectionCenterLonLat_ = this.toLonLatTransform_(
        getCenter(projection.getExtent())
      );
      if (isNaN(this.projectionCenterLonLat_[1])) {
        this.projectionCenterLonLat_[1] =
          Math.abs(this.maxLat_) >= Math.abs(this.minLat_)
            ? this.maxLat_
            : this.minLat_;
      }

      this.projection_ = projection;
    }
  }

  /**
   * @module ol/layer/Heatmap
   */

  /**
   * @typedef {Object} Options
   * @property {string} [className='ol-layer'] A CSS class name to set to the layer element.
   * @property {number} [opacity=1] Opacity (0, 1).
   * @property {boolean} [visible=true] Visibility.
   * @property {import("../extent.js").Extent} [extent] The bounding extent for layer rendering.  The layer will not be
   * rendered outside of this extent.
   * @property {number} [zIndex] The z-index for layer rendering.  At rendering time, the layers
   * will be ordered, first by Z-index and then by position. When `undefined`, a `zIndex` of 0 is assumed
   * for layers that are added to the map's `layers` collection, or `Infinity` when the layer's `setMap()`
   * method was used.
   * @property {number} [minResolution] The minimum resolution (inclusive) at which this layer will be
   * visible.
   * @property {number} [maxResolution] The maximum resolution (exclusive) below which this layer will
   * be visible.
   * @property {number} [minZoom] The minimum view zoom level (exclusive) above which this layer will be
   * visible.
   * @property {number} [maxZoom] The maximum view zoom level (inclusive) at which this layer will
   * be visible.
   * @property {Array<string>} [gradient=['#00f', '#0ff', '#0f0', '#ff0', '#f00']] The color gradient
   * of the heatmap, specified as an array of CSS color strings.
   * @property {number} [radius=8] Radius size in pixels.
   * @property {number} [blur=15] Blur size in pixels.
   * @property {string|function(import("../Feature.js").default):number} [weight='weight'] The feature
   * attribute to use for the weight or a function that returns a weight from a feature. Weight values
   * should range from 0 to 1 (and values outside will be clamped to that range).
   * @property {import("../source/Vector.js").default<import("../geom/Point.js").default>} [source] Point source.
   * @property {Object<string, *>} [properties] Arbitrary observable properties. Can be accessed with `#get()` and `#set()`.
   */

  /**
   * @enum {string}
   * @private
   */
  const Property = {
    BLUR: 'blur',
    GRADIENT: 'gradient',
    RADIUS: 'radius',
  };

  /**
   * @const
   * @type {Array<string>}
   */
  const DEFAULT_GRADIENT = ['#00f', '#0ff', '#0f0', '#ff0', '#f00'];

  /**
   * @classdesc
   * Layer for rendering vector data as a heatmap.
   * Note that any property set in the options is set as a {@link module:ol/Object~BaseObject}
   * property on the layer object; for example, setting `title: 'My Title'` in the
   * options means that `title` is observable, and has get/set accessors.
   *
   * @fires import("../render/Event.js").RenderEvent
   * @extends {BaseVector<import("../source/Vector.js").default, WebGLPointsLayerRenderer>}
   * @api
   */
  class Heatmap extends BaseVectorLayer {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      options = options ? options : {};

      const baseOptions = Object.assign({}, options);

      delete baseOptions.gradient;
      delete baseOptions.radius;
      delete baseOptions.blur;
      delete baseOptions.weight;
      super(baseOptions);

      /**
       * @private
       * @type {HTMLCanvasElement}
       */
      this.gradient_ = null;

      this.addChangeListener(Property.GRADIENT, this.handleGradientChanged_);

      this.setGradient(options.gradient ? options.gradient : DEFAULT_GRADIENT);

      this.setBlur(options.blur !== undefined ? options.blur : 15);

      this.setRadius(options.radius !== undefined ? options.radius : 8);

      const weight = options.weight ? options.weight : 'weight';
      if (typeof weight === 'string') {
        this.weightFunction_ = function (feature) {
          return feature.get(weight);
        };
      } else {
        this.weightFunction_ = weight;
      }

      // For performance reasons, don't sort the features before rendering.
      // The render order is not relevant for a heatmap representation.
      this.setRenderOrder(null);
    }

    /**
     * Return the blur size in pixels.
     * @return {number} Blur size in pixels.
     * @api
     * @observable
     */
    getBlur() {
      return /** @type {number} */ (this.get(Property.BLUR));
    }

    /**
     * Return the gradient colors as array of strings.
     * @return {Array<string>} Colors.
     * @api
     * @observable
     */
    getGradient() {
      return /** @type {Array<string>} */ (this.get(Property.GRADIENT));
    }

    /**
     * Return the size of the radius in pixels.
     * @return {number} Radius size in pixel.
     * @api
     * @observable
     */
    getRadius() {
      return /** @type {number} */ (this.get(Property.RADIUS));
    }

    /**
     * @private
     */
    handleGradientChanged_() {
      this.gradient_ = createGradient(this.getGradient());
    }

    /**
     * Set the blur size in pixels.
     * @param {number} blur Blur size in pixels.
     * @api
     * @observable
     */
    setBlur(blur) {
      this.set(Property.BLUR, blur);
    }

    /**
     * Set the gradient colors as array of strings.
     * @param {Array<string>} colors Gradient.
     * @api
     * @observable
     */
    setGradient(colors) {
      this.set(Property.GRADIENT, colors);
    }

    /**
     * Set the size of the radius in pixels.
     * @param {number} radius Radius size in pixel.
     * @api
     * @observable
     */
    setRadius(radius) {
      this.set(Property.RADIUS, radius);
    }

    createRenderer() {
      return new WebGLPointsLayerRenderer(this, {
        className: this.getClassName(),
        attributes: [
          {
            name: 'weight',
            callback: function (feature) {
              const weight = this.weightFunction_(feature);
              return weight !== undefined ? clamp(weight, 0, 1) : 1;
            }.bind(this),
          },
        ],
        vertexShader: `
        precision mediump float;
        uniform mat4 u_projectionMatrix;
        uniform mat4 u_offsetScaleMatrix;
        uniform float u_size;
        attribute vec2 a_position;
        attribute float a_index;
        attribute float a_weight;

        varying vec2 v_texCoord;
        varying float v_weight;

        void main(void) {
          mat4 offsetMatrix = u_offsetScaleMatrix;
          float offsetX = a_index == 0.0 || a_index == 3.0 ? -u_size / 2.0 : u_size / 2.0;
          float offsetY = a_index == 0.0 || a_index == 1.0 ? -u_size / 2.0 : u_size / 2.0;
          vec4 offsets = offsetMatrix * vec4(offsetX, offsetY, 0.0, 0.0);
          gl_Position = u_projectionMatrix * vec4(a_position, 0.0, 1.0) + offsets;
          float u = a_index == 0.0 || a_index == 3.0 ? 0.0 : 1.0;
          float v = a_index == 0.0 || a_index == 1.0 ? 0.0 : 1.0;
          v_texCoord = vec2(u, v);
          v_weight = a_weight;
        }`,
        fragmentShader: `
        precision mediump float;
        uniform float u_blurSlope;

        varying vec2 v_texCoord;
        varying float v_weight;

        void main(void) {
          vec2 texCoord = v_texCoord * 2.0 - vec2(1.0, 1.0);
          float sqRadius = texCoord.x * texCoord.x + texCoord.y * texCoord.y;
          float value = (1.0 - sqrt(sqRadius)) * u_blurSlope;
          float alpha = smoothstep(0.0, 1.0, value) * v_weight;
          gl_FragColor = vec4(alpha, alpha, alpha, alpha);
        }`,
        hitVertexShader: `
        precision mediump float;
        uniform mat4 u_projectionMatrix;
        uniform mat4 u_offsetScaleMatrix;
        uniform float u_size;
        attribute vec2 a_position;
        attribute float a_index;
        attribute float a_weight;
        attribute vec4 a_hitColor;

        varying vec2 v_texCoord;
        varying float v_weight;
        varying vec4 v_hitColor;

        void main(void) {
          mat4 offsetMatrix = u_offsetScaleMatrix;
          float offsetX = a_index == 0.0 || a_index == 3.0 ? -u_size / 2.0 : u_size / 2.0;
          float offsetY = a_index == 0.0 || a_index == 1.0 ? -u_size / 2.0 : u_size / 2.0;
          vec4 offsets = offsetMatrix * vec4(offsetX, offsetY, 0.0, 0.0);
          gl_Position = u_projectionMatrix * vec4(a_position, 0.0, 1.0) + offsets;
          float u = a_index == 0.0 || a_index == 3.0 ? 0.0 : 1.0;
          float v = a_index == 0.0 || a_index == 1.0 ? 0.0 : 1.0;
          v_texCoord = vec2(u, v);
          v_hitColor = a_hitColor;
          v_weight = a_weight;
        }`,
        hitFragmentShader: `
        precision mediump float;
        uniform float u_blurSlope;

        varying vec2 v_texCoord;
        varying float v_weight;
        varying vec4 v_hitColor;

        void main(void) {
          vec2 texCoord = v_texCoord * 2.0 - vec2(1.0, 1.0);
          float sqRadius = texCoord.x * texCoord.x + texCoord.y * texCoord.y;
          float value = (1.0 - sqrt(sqRadius)) * u_blurSlope;
          float alpha = smoothstep(0.0, 1.0, value) * v_weight;
          if (alpha < 0.05) {
            discard;
          }

          gl_FragColor = v_hitColor;
        }`,
        uniforms: {
          u_size: function () {
            return (this.get(Property.RADIUS) + this.get(Property.BLUR)) * 2;
          }.bind(this),
          u_blurSlope: function () {
            return (
              this.get(Property.RADIUS) / Math.max(1, this.get(Property.BLUR))
            );
          }.bind(this),
        },
        postProcesses: [
          {
            fragmentShader: `
            precision mediump float;

            uniform sampler2D u_image;
            uniform sampler2D u_gradientTexture;
            uniform float u_opacity;

            varying vec2 v_texCoord;

            void main() {
              vec4 color = texture2D(u_image, v_texCoord);
              gl_FragColor.a = color.a * u_opacity;
              gl_FragColor.rgb = texture2D(u_gradientTexture, vec2(0.5, color.a)).rgb;
              gl_FragColor.rgb *= gl_FragColor.a;
            }`,
            uniforms: {
              u_gradientTexture: function () {
                return this.gradient_;
              }.bind(this),
              u_opacity: function () {
                return this.getOpacity();
              }.bind(this),
            },
          },
        ],
      });
    }

    renderDeclutter() {}
  }

  /**
   * @param {Array<string>} colors A list of colored.
   * @return {HTMLCanvasElement} canvas with gradient texture.
   */
  function createGradient(colors) {
    const width = 1;
    const height = 256;
    const context = createCanvasContext2D(width, height);

    const gradient = context.createLinearGradient(0, 0, width, height);
    const step = 1 / (colors.length - 1);
    for (let i = 0, ii = colors.length; i < ii; ++i) {
      gradient.addColorStop(i * step, colors[i]);
    }

    context.fillStyle = gradient;
    context.fillRect(0, 0, width, height);

    return context.canvas;
  }

  /**
   * @module ol/format/Feature
   */

  /**
   * @typedef {Object} ReadOptions
   * @property {import("../proj.js").ProjectionLike} [dataProjection] Projection of the data we are reading.
   * If not provided, the projection will be derived from the data (where possible) or
   * the `dataProjection` of the format is assigned (where set). If the projection
   * can not be derived from the data and if no `dataProjection` is set for a format,
   * the features will not be reprojected.
   * @property {import("../extent.js").Extent} [extent] Tile extent in map units of the tile being read.
   * This is only required when reading data with tile pixels as geometry units. When configured,
   * a `dataProjection` with `TILE_PIXELS` as `units` and the tile's pixel extent as `extent` needs to be
   * provided.
   * @property {import("../proj.js").ProjectionLike} [featureProjection] Projection of the feature geometries
   * created by the format reader. If not provided, features will be returned in the
   * `dataProjection`.
   */

  /**
   * @typedef {Object} WriteOptions
   * @property {import("../proj.js").ProjectionLike} [dataProjection] Projection of the data we are writing.
   * If not provided, the `dataProjection` of the format is assigned (where set).
   * If no `dataProjection` is set for a format, the features will be returned
   * in the `featureProjection`.
   * @property {import("../proj.js").ProjectionLike} [featureProjection] Projection of the feature geometries
   * that will be serialized by the format writer. If not provided, geometries are assumed
   * to be in the `dataProjection` if that is set; in other words, they are not transformed.
   * @property {boolean} [rightHanded] When writing geometries, follow the right-hand
   * rule for linear ring orientation.  This means that polygons will have counter-clockwise
   * exterior rings and clockwise interior rings.  By default, coordinates are serialized
   * as they are provided at construction.  If `true`, the right-hand rule will
   * be applied.  If `false`, the left-hand rule will be applied (clockwise for
   * exterior and counter-clockwise for interior rings).  Note that not all
   * formats support this.  The GeoJSON format does use this property when writing
   * geometries.
   * @property {number} [decimals] Maximum number of decimal places for coordinates.
   * Coordinates are stored internally as floats, but floating-point arithmetic can create
   * coordinates with a large number of decimal places, not generally wanted on output.
   * Set a number here to round coordinates. Can also be used to ensure that
   * coordinates read in can be written back out with the same number of decimals.
   * Default is no rounding.
   */

  /**
   * @typedef {'arraybuffer' | 'json' | 'text' | 'xml'} Type
   */

  /**
   * @classdesc
   * Abstract base class; normally only used for creating subclasses and not
   * instantiated in apps.
   * Base class for feature formats.
   * {@link module:ol/format/Feature~FeatureFormat} subclasses provide the ability to decode and encode
   * {@link module:ol/Feature~Feature} objects from a variety of commonly used geospatial
   * file formats.  See the documentation for each format for more details.
   *
   * @abstract
   * @api
   */
  class FeatureFormat {
    constructor() {
      /**
       * @protected
       * @type {import("../proj/Projection.js").default|undefined}
       */
      this.dataProjection = undefined;

      /**
       * @protected
       * @type {import("../proj/Projection.js").default|undefined}
       */
      this.defaultFeatureProjection = undefined;

      /**
       * A list media types supported by the format in descending order of preference.
       * @type {Array<string>}
       */
      this.supportedMediaTypes = null;
    }

    /**
     * Adds the data projection to the read options.
     * @param {Document|Element|Object|string} source Source.
     * @param {ReadOptions} [options] Options.
     * @return {ReadOptions|undefined} Options.
     * @protected
     */
    getReadOptions(source, options) {
      if (options) {
        let dataProjection = options.dataProjection
          ? get$2(options.dataProjection)
          : this.readProjection(source);
        if (
          options.extent &&
          dataProjection &&
          dataProjection.getUnits() === 'tile-pixels'
        ) {
          dataProjection = get$2(dataProjection);
          dataProjection.setWorldExtent(options.extent);
        }
        options = {
          dataProjection: dataProjection,
          featureProjection: options.featureProjection,
        };
      }
      return this.adaptOptions(options);
    }

    /**
     * Sets the `dataProjection` on the options, if no `dataProjection`
     * is set.
     * @param {WriteOptions|ReadOptions|undefined} options
     *     Options.
     * @protected
     * @return {WriteOptions|ReadOptions|undefined}
     *     Updated options.
     */
    adaptOptions(options) {
      return Object.assign(
        {
          dataProjection: this.dataProjection,
          featureProjection: this.defaultFeatureProjection,
        },
        options
      );
    }

    /**
     * @abstract
     * @return {Type} The format type.
     */
    getType() {
      return abstract();
    }

    /**
     * Read a single feature from a source.
     *
     * @abstract
     * @param {Document|Element|Object|string} source Source.
     * @param {ReadOptions} [options] Read options.
     * @return {import("../Feature.js").FeatureLike} Feature.
     */
    readFeature(source, options) {
      return abstract();
    }

    /**
     * Read all features from a source.
     *
     * @abstract
     * @param {Document|Element|ArrayBuffer|Object|string} source Source.
     * @param {ReadOptions} [options] Read options.
     * @return {Array<import("../Feature.js").FeatureLike>} Features.
     */
    readFeatures(source, options) {
      return abstract();
    }

    /**
     * Read a single geometry from a source.
     *
     * @abstract
     * @param {Document|Element|Object|string} source Source.
     * @param {ReadOptions} [options] Read options.
     * @return {import("../geom/Geometry.js").default} Geometry.
     */
    readGeometry(source, options) {
      return abstract();
    }

    /**
     * Read the projection from a source.
     *
     * @abstract
     * @param {Document|Element|Object|string} source Source.
     * @return {import("../proj/Projection.js").default|undefined} Projection.
     */
    readProjection(source) {
      return abstract();
    }

    /**
     * Encode a feature in this format.
     *
     * @abstract
     * @param {import("../Feature.js").default} feature Feature.
     * @param {WriteOptions} [options] Write options.
     * @return {string|ArrayBuffer} Result.
     */
    writeFeature(feature, options) {
      return abstract();
    }

    /**
     * Encode an array of features in this format.
     *
     * @abstract
     * @param {Array<import("../Feature.js").default>} features Features.
     * @param {WriteOptions} [options] Write options.
     * @return {string|ArrayBuffer} Result.
     */
    writeFeatures(features, options) {
      return abstract();
    }

    /**
     * Write a single geometry in this format.
     *
     * @abstract
     * @param {import("../geom/Geometry.js").default} geometry Geometry.
     * @param {WriteOptions} [options] Write options.
     * @return {string|ArrayBuffer} Result.
     */
    writeGeometry(geometry, options) {
      return abstract();
    }
  }

  /**
   * @param {import("../geom/Geometry.js").default} geometry Geometry.
   * @param {boolean} write Set to true for writing, false for reading.
   * @param {WriteOptions|ReadOptions} [options] Options.
   * @return {import("../geom/Geometry.js").default} Transformed geometry.
   */
  function transformGeometryWithOptions(geometry, write, options) {
    const featureProjection = options
      ? get$2(options.featureProjection)
      : null;
    const dataProjection = options ? get$2(options.dataProjection) : null;

    let transformed;
    if (
      featureProjection &&
      dataProjection &&
      !equivalent(featureProjection, dataProjection)
    ) {
      transformed = (write ? geometry.clone() : geometry).transform(
        write ? featureProjection : dataProjection,
        write ? dataProjection : featureProjection
      );
    } else {
      transformed = geometry;
    }
    if (
      write &&
      options &&
      /** @type {WriteOptions} */ (options).decimals !== undefined
    ) {
      const power = Math.pow(10, /** @type {WriteOptions} */ (options).decimals);
      // if decimals option on write, round each coordinate appropriately
      /**
       * @param {Array<number>} coordinates Coordinates.
       * @return {Array<number>} Transformed coordinates.
       */
      const transform = function (coordinates) {
        for (let i = 0, ii = coordinates.length; i < ii; ++i) {
          coordinates[i] = Math.round(coordinates[i] * power) / power;
        }
        return coordinates;
      };
      if (transformed === geometry) {
        transformed = geometry.clone();
      }
      transformed.applyTransform(transform);
    }
    return transformed;
  }

  /**
   * @param {import("../extent.js").Extent} extent Extent.
   * @param {ReadOptions} [options] Read options.
   * @return {import("../extent.js").Extent} Transformed extent.
   */
  function transformExtentWithOptions(extent, options) {
    const featureProjection = options
      ? get$2(options.featureProjection)
      : null;
    const dataProjection = options ? get$2(options.dataProjection) : null;

    if (
      featureProjection &&
      dataProjection &&
      !equivalent(featureProjection, dataProjection)
    ) {
      return transformExtent(extent, dataProjection, featureProjection);
    } else {
      return extent;
    }
  }

  var ieee754$1 = {};

  /*! ieee754. BSD-3-Clause License. Feross Aboukhadijeh <https://feross.org/opensource> */

  ieee754$1.read = function (buffer, offset, isLE, mLen, nBytes) {
    var e, m;
    var eLen = (nBytes * 8) - mLen - 1;
    var eMax = (1 << eLen) - 1;
    var eBias = eMax >> 1;
    var nBits = -7;
    var i = isLE ? (nBytes - 1) : 0;
    var d = isLE ? -1 : 1;
    var s = buffer[offset + i];

    i += d;

    e = s & ((1 << (-nBits)) - 1);
    s >>= (-nBits);
    nBits += eLen;
    for (; nBits > 0; e = (e * 256) + buffer[offset + i], i += d, nBits -= 8) {}

    m = e & ((1 << (-nBits)) - 1);
    e >>= (-nBits);
    nBits += mLen;
    for (; nBits > 0; m = (m * 256) + buffer[offset + i], i += d, nBits -= 8) {}

    if (e === 0) {
      e = 1 - eBias;
    } else if (e === eMax) {
      return m ? NaN : ((s ? -1 : 1) * Infinity)
    } else {
      m = m + Math.pow(2, mLen);
      e = e - eBias;
    }
    return (s ? -1 : 1) * m * Math.pow(2, e - mLen)
  };

  ieee754$1.write = function (buffer, value, offset, isLE, mLen, nBytes) {
    var e, m, c;
    var eLen = (nBytes * 8) - mLen - 1;
    var eMax = (1 << eLen) - 1;
    var eBias = eMax >> 1;
    var rt = (mLen === 23 ? Math.pow(2, -24) - Math.pow(2, -77) : 0);
    var i = isLE ? 0 : (nBytes - 1);
    var d = isLE ? 1 : -1;
    var s = value < 0 || (value === 0 && 1 / value < 0) ? 1 : 0;

    value = Math.abs(value);

    if (isNaN(value) || value === Infinity) {
      m = isNaN(value) ? 1 : 0;
      e = eMax;
    } else {
      e = Math.floor(Math.log(value) / Math.LN2);
      if (value * (c = Math.pow(2, -e)) < 1) {
        e--;
        c *= 2;
      }
      if (e + eBias >= 1) {
        value += rt / c;
      } else {
        value += rt * Math.pow(2, 1 - eBias);
      }
      if (value * c >= 2) {
        e++;
        c /= 2;
      }

      if (e + eBias >= eMax) {
        m = 0;
        e = eMax;
      } else if (e + eBias >= 1) {
        m = ((value * c) - 1) * Math.pow(2, mLen);
        e = e + eBias;
      } else {
        m = value * Math.pow(2, eBias - 1) * Math.pow(2, mLen);
        e = 0;
      }
    }

    for (; mLen >= 8; buffer[offset + i] = m & 0xff, i += d, m /= 256, mLen -= 8) {}

    e = (e << mLen) | m;
    eLen += mLen;
    for (; eLen > 0; buffer[offset + i] = e & 0xff, i += d, e /= 256, eLen -= 8) {}

    buffer[offset + i - d] |= s * 128;
  };

  var pbf = Pbf;

  var ieee754 = ieee754$1;

  function Pbf(buf) {
      this.buf = ArrayBuffer.isView && ArrayBuffer.isView(buf) ? buf : new Uint8Array(buf || 0);
      this.pos = 0;
      this.type = 0;
      this.length = this.buf.length;
  }

  Pbf.Varint  = 0; // varint: int32, int64, uint32, uint64, sint32, sint64, bool, enum
  Pbf.Fixed64 = 1; // 64-bit: double, fixed64, sfixed64
  Pbf.Bytes   = 2; // length-delimited: string, bytes, embedded messages, packed repeated fields
  Pbf.Fixed32 = 5; // 32-bit: float, fixed32, sfixed32

  var SHIFT_LEFT_32 = (1 << 16) * (1 << 16),
      SHIFT_RIGHT_32 = 1 / SHIFT_LEFT_32;

  // Threshold chosen based on both benchmarking and knowledge about browser string
  // data structures (which currently switch structure types at 12 bytes or more)
  var TEXT_DECODER_MIN_LENGTH = 12;
  var utf8TextDecoder = typeof TextDecoder === 'undefined' ? null : new TextDecoder('utf8');

  Pbf.prototype = {

      destroy: function() {
          this.buf = null;
      },

      // === READING =================================================================

      readFields: function(readField, result, end) {
          end = end || this.length;

          while (this.pos < end) {
              var val = this.readVarint(),
                  tag = val >> 3,
                  startPos = this.pos;

              this.type = val & 0x7;
              readField(tag, result, this);

              if (this.pos === startPos) this.skip(val);
          }
          return result;
      },

      readMessage: function(readField, result) {
          return this.readFields(readField, result, this.readVarint() + this.pos);
      },

      readFixed32: function() {
          var val = readUInt32(this.buf, this.pos);
          this.pos += 4;
          return val;
      },

      readSFixed32: function() {
          var val = readInt32(this.buf, this.pos);
          this.pos += 4;
          return val;
      },

      // 64-bit int handling is based on github.com/dpw/node-buffer-more-ints (MIT-licensed)

      readFixed64: function() {
          var val = readUInt32(this.buf, this.pos) + readUInt32(this.buf, this.pos + 4) * SHIFT_LEFT_32;
          this.pos += 8;
          return val;
      },

      readSFixed64: function() {
          var val = readUInt32(this.buf, this.pos) + readInt32(this.buf, this.pos + 4) * SHIFT_LEFT_32;
          this.pos += 8;
          return val;
      },

      readFloat: function() {
          var val = ieee754.read(this.buf, this.pos, true, 23, 4);
          this.pos += 4;
          return val;
      },

      readDouble: function() {
          var val = ieee754.read(this.buf, this.pos, true, 52, 8);
          this.pos += 8;
          return val;
      },

      readVarint: function(isSigned) {
          var buf = this.buf,
              val, b;

          b = buf[this.pos++]; val  =  b & 0x7f;        if (b < 0x80) return val;
          b = buf[this.pos++]; val |= (b & 0x7f) << 7;  if (b < 0x80) return val;
          b = buf[this.pos++]; val |= (b & 0x7f) << 14; if (b < 0x80) return val;
          b = buf[this.pos++]; val |= (b & 0x7f) << 21; if (b < 0x80) return val;
          b = buf[this.pos];   val |= (b & 0x0f) << 28;

          return readVarintRemainder(val, isSigned, this);
      },

      readVarint64: function() { // for compatibility with v2.0.1
          return this.readVarint(true);
      },

      readSVarint: function() {
          var num = this.readVarint();
          return num % 2 === 1 ? (num + 1) / -2 : num / 2; // zigzag encoding
      },

      readBoolean: function() {
          return Boolean(this.readVarint());
      },

      readString: function() {
          var end = this.readVarint() + this.pos;
          var pos = this.pos;
          this.pos = end;

          if (end - pos >= TEXT_DECODER_MIN_LENGTH && utf8TextDecoder) {
              // longer strings are fast with the built-in browser TextDecoder API
              return readUtf8TextDecoder(this.buf, pos, end);
          }
          // short strings are fast with our custom implementation
          return readUtf8(this.buf, pos, end);
      },

      readBytes: function() {
          var end = this.readVarint() + this.pos,
              buffer = this.buf.subarray(this.pos, end);
          this.pos = end;
          return buffer;
      },

      // verbose for performance reasons; doesn't affect gzipped size

      readPackedVarint: function(arr, isSigned) {
          if (this.type !== Pbf.Bytes) return arr.push(this.readVarint(isSigned));
          var end = readPackedEnd(this);
          arr = arr || [];
          while (this.pos < end) arr.push(this.readVarint(isSigned));
          return arr;
      },
      readPackedSVarint: function(arr) {
          if (this.type !== Pbf.Bytes) return arr.push(this.readSVarint());
          var end = readPackedEnd(this);
          arr = arr || [];
          while (this.pos < end) arr.push(this.readSVarint());
          return arr;
      },
      readPackedBoolean: function(arr) {
          if (this.type !== Pbf.Bytes) return arr.push(this.readBoolean());
          var end = readPackedEnd(this);
          arr = arr || [];
          while (this.pos < end) arr.push(this.readBoolean());
          return arr;
      },
      readPackedFloat: function(arr) {
          if (this.type !== Pbf.Bytes) return arr.push(this.readFloat());
          var end = readPackedEnd(this);
          arr = arr || [];
          while (this.pos < end) arr.push(this.readFloat());
          return arr;
      },
      readPackedDouble: function(arr) {
          if (this.type !== Pbf.Bytes) return arr.push(this.readDouble());
          var end = readPackedEnd(this);
          arr = arr || [];
          while (this.pos < end) arr.push(this.readDouble());
          return arr;
      },
      readPackedFixed32: function(arr) {
          if (this.type !== Pbf.Bytes) return arr.push(this.readFixed32());
          var end = readPackedEnd(this);
          arr = arr || [];
          while (this.pos < end) arr.push(this.readFixed32());
          return arr;
      },
      readPackedSFixed32: function(arr) {
          if (this.type !== Pbf.Bytes) return arr.push(this.readSFixed32());
          var end = readPackedEnd(this);
          arr = arr || [];
          while (this.pos < end) arr.push(this.readSFixed32());
          return arr;
      },
      readPackedFixed64: function(arr) {
          if (this.type !== Pbf.Bytes) return arr.push(this.readFixed64());
          var end = readPackedEnd(this);
          arr = arr || [];
          while (this.pos < end) arr.push(this.readFixed64());
          return arr;
      },
      readPackedSFixed64: function(arr) {
          if (this.type !== Pbf.Bytes) return arr.push(this.readSFixed64());
          var end = readPackedEnd(this);
          arr = arr || [];
          while (this.pos < end) arr.push(this.readSFixed64());
          return arr;
      },

      skip: function(val) {
          var type = val & 0x7;
          if (type === Pbf.Varint) while (this.buf[this.pos++] > 0x7f) {}
          else if (type === Pbf.Bytes) this.pos = this.readVarint() + this.pos;
          else if (type === Pbf.Fixed32) this.pos += 4;
          else if (type === Pbf.Fixed64) this.pos += 8;
          else throw new Error('Unimplemented type: ' + type);
      },

      // === WRITING =================================================================

      writeTag: function(tag, type) {
          this.writeVarint((tag << 3) | type);
      },

      realloc: function(min) {
          var length = this.length || 16;

          while (length < this.pos + min) length *= 2;

          if (length !== this.length) {
              var buf = new Uint8Array(length);
              buf.set(this.buf);
              this.buf = buf;
              this.length = length;
          }
      },

      finish: function() {
          this.length = this.pos;
          this.pos = 0;
          return this.buf.subarray(0, this.length);
      },

      writeFixed32: function(val) {
          this.realloc(4);
          writeInt32(this.buf, val, this.pos);
          this.pos += 4;
      },

      writeSFixed32: function(val) {
          this.realloc(4);
          writeInt32(this.buf, val, this.pos);
          this.pos += 4;
      },

      writeFixed64: function(val) {
          this.realloc(8);
          writeInt32(this.buf, val & -1, this.pos);
          writeInt32(this.buf, Math.floor(val * SHIFT_RIGHT_32), this.pos + 4);
          this.pos += 8;
      },

      writeSFixed64: function(val) {
          this.realloc(8);
          writeInt32(this.buf, val & -1, this.pos);
          writeInt32(this.buf, Math.floor(val * SHIFT_RIGHT_32), this.pos + 4);
          this.pos += 8;
      },

      writeVarint: function(val) {
          val = +val || 0;

          if (val > 0xfffffff || val < 0) {
              writeBigVarint(val, this);
              return;
          }

          this.realloc(4);

          this.buf[this.pos++] =           val & 0x7f  | (val > 0x7f ? 0x80 : 0); if (val <= 0x7f) return;
          this.buf[this.pos++] = ((val >>>= 7) & 0x7f) | (val > 0x7f ? 0x80 : 0); if (val <= 0x7f) return;
          this.buf[this.pos++] = ((val >>>= 7) & 0x7f) | (val > 0x7f ? 0x80 : 0); if (val <= 0x7f) return;
          this.buf[this.pos++] =   (val >>> 7) & 0x7f;
      },

      writeSVarint: function(val) {
          this.writeVarint(val < 0 ? -val * 2 - 1 : val * 2);
      },

      writeBoolean: function(val) {
          this.writeVarint(Boolean(val));
      },

      writeString: function(str) {
          str = String(str);
          this.realloc(str.length * 4);

          this.pos++; // reserve 1 byte for short string length

          var startPos = this.pos;
          // write the string directly to the buffer and see how much was written
          this.pos = writeUtf8(this.buf, str, this.pos);
          var len = this.pos - startPos;

          if (len >= 0x80) makeRoomForExtraLength(startPos, len, this);

          // finally, write the message length in the reserved place and restore the position
          this.pos = startPos - 1;
          this.writeVarint(len);
          this.pos += len;
      },

      writeFloat: function(val) {
          this.realloc(4);
          ieee754.write(this.buf, val, this.pos, true, 23, 4);
          this.pos += 4;
      },

      writeDouble: function(val) {
          this.realloc(8);
          ieee754.write(this.buf, val, this.pos, true, 52, 8);
          this.pos += 8;
      },

      writeBytes: function(buffer) {
          var len = buffer.length;
          this.writeVarint(len);
          this.realloc(len);
          for (var i = 0; i < len; i++) this.buf[this.pos++] = buffer[i];
      },

      writeRawMessage: function(fn, obj) {
          this.pos++; // reserve 1 byte for short message length

          // write the message directly to the buffer and see how much was written
          var startPos = this.pos;
          fn(obj, this);
          var len = this.pos - startPos;

          if (len >= 0x80) makeRoomForExtraLength(startPos, len, this);

          // finally, write the message length in the reserved place and restore the position
          this.pos = startPos - 1;
          this.writeVarint(len);
          this.pos += len;
      },

      writeMessage: function(tag, fn, obj) {
          this.writeTag(tag, Pbf.Bytes);
          this.writeRawMessage(fn, obj);
      },

      writePackedVarint:   function(tag, arr) { if (arr.length) this.writeMessage(tag, writePackedVarint, arr);   },
      writePackedSVarint:  function(tag, arr) { if (arr.length) this.writeMessage(tag, writePackedSVarint, arr);  },
      writePackedBoolean:  function(tag, arr) { if (arr.length) this.writeMessage(tag, writePackedBoolean, arr);  },
      writePackedFloat:    function(tag, arr) { if (arr.length) this.writeMessage(tag, writePackedFloat, arr);    },
      writePackedDouble:   function(tag, arr) { if (arr.length) this.writeMessage(tag, writePackedDouble, arr);   },
      writePackedFixed32:  function(tag, arr) { if (arr.length) this.writeMessage(tag, writePackedFixed32, arr);  },
      writePackedSFixed32: function(tag, arr) { if (arr.length) this.writeMessage(tag, writePackedSFixed32, arr); },
      writePackedFixed64:  function(tag, arr) { if (arr.length) this.writeMessage(tag, writePackedFixed64, arr);  },
      writePackedSFixed64: function(tag, arr) { if (arr.length) this.writeMessage(tag, writePackedSFixed64, arr); },

      writeBytesField: function(tag, buffer) {
          this.writeTag(tag, Pbf.Bytes);
          this.writeBytes(buffer);
      },
      writeFixed32Field: function(tag, val) {
          this.writeTag(tag, Pbf.Fixed32);
          this.writeFixed32(val);
      },
      writeSFixed32Field: function(tag, val) {
          this.writeTag(tag, Pbf.Fixed32);
          this.writeSFixed32(val);
      },
      writeFixed64Field: function(tag, val) {
          this.writeTag(tag, Pbf.Fixed64);
          this.writeFixed64(val);
      },
      writeSFixed64Field: function(tag, val) {
          this.writeTag(tag, Pbf.Fixed64);
          this.writeSFixed64(val);
      },
      writeVarintField: function(tag, val) {
          this.writeTag(tag, Pbf.Varint);
          this.writeVarint(val);
      },
      writeSVarintField: function(tag, val) {
          this.writeTag(tag, Pbf.Varint);
          this.writeSVarint(val);
      },
      writeStringField: function(tag, str) {
          this.writeTag(tag, Pbf.Bytes);
          this.writeString(str);
      },
      writeFloatField: function(tag, val) {
          this.writeTag(tag, Pbf.Fixed32);
          this.writeFloat(val);
      },
      writeDoubleField: function(tag, val) {
          this.writeTag(tag, Pbf.Fixed64);
          this.writeDouble(val);
      },
      writeBooleanField: function(tag, val) {
          this.writeVarintField(tag, Boolean(val));
      }
  };

  function readVarintRemainder(l, s, p) {
      var buf = p.buf,
          h, b;

      b = buf[p.pos++]; h  = (b & 0x70) >> 4;  if (b < 0x80) return toNum(l, h, s);
      b = buf[p.pos++]; h |= (b & 0x7f) << 3;  if (b < 0x80) return toNum(l, h, s);
      b = buf[p.pos++]; h |= (b & 0x7f) << 10; if (b < 0x80) return toNum(l, h, s);
      b = buf[p.pos++]; h |= (b & 0x7f) << 17; if (b < 0x80) return toNum(l, h, s);
      b = buf[p.pos++]; h |= (b & 0x7f) << 24; if (b < 0x80) return toNum(l, h, s);
      b = buf[p.pos++]; h |= (b & 0x01) << 31; if (b < 0x80) return toNum(l, h, s);

      throw new Error('Expected varint not more than 10 bytes');
  }

  function readPackedEnd(pbf) {
      return pbf.type === Pbf.Bytes ?
          pbf.readVarint() + pbf.pos : pbf.pos + 1;
  }

  function toNum(low, high, isSigned) {
      if (isSigned) {
          return high * 0x100000000 + (low >>> 0);
      }

      return ((high >>> 0) * 0x100000000) + (low >>> 0);
  }

  function writeBigVarint(val, pbf) {
      var low, high;

      if (val >= 0) {
          low  = (val % 0x100000000) | 0;
          high = (val / 0x100000000) | 0;
      } else {
          low  = ~(-val % 0x100000000);
          high = ~(-val / 0x100000000);

          if (low ^ 0xffffffff) {
              low = (low + 1) | 0;
          } else {
              low = 0;
              high = (high + 1) | 0;
          }
      }

      if (val >= 0x10000000000000000 || val < -0x10000000000000000) {
          throw new Error('Given varint doesn\'t fit into 10 bytes');
      }

      pbf.realloc(10);

      writeBigVarintLow(low, high, pbf);
      writeBigVarintHigh(high, pbf);
  }

  function writeBigVarintLow(low, high, pbf) {
      pbf.buf[pbf.pos++] = low & 0x7f | 0x80; low >>>= 7;
      pbf.buf[pbf.pos++] = low & 0x7f | 0x80; low >>>= 7;
      pbf.buf[pbf.pos++] = low & 0x7f | 0x80; low >>>= 7;
      pbf.buf[pbf.pos++] = low & 0x7f | 0x80; low >>>= 7;
      pbf.buf[pbf.pos]   = low & 0x7f;
  }

  function writeBigVarintHigh(high, pbf) {
      var lsb = (high & 0x07) << 4;

      pbf.buf[pbf.pos++] |= lsb         | ((high >>>= 3) ? 0x80 : 0); if (!high) return;
      pbf.buf[pbf.pos++]  = high & 0x7f | ((high >>>= 7) ? 0x80 : 0); if (!high) return;
      pbf.buf[pbf.pos++]  = high & 0x7f | ((high >>>= 7) ? 0x80 : 0); if (!high) return;
      pbf.buf[pbf.pos++]  = high & 0x7f | ((high >>>= 7) ? 0x80 : 0); if (!high) return;
      pbf.buf[pbf.pos++]  = high & 0x7f | ((high >>>= 7) ? 0x80 : 0); if (!high) return;
      pbf.buf[pbf.pos++]  = high & 0x7f;
  }

  function makeRoomForExtraLength(startPos, len, pbf) {
      var extraLen =
          len <= 0x3fff ? 1 :
          len <= 0x1fffff ? 2 :
          len <= 0xfffffff ? 3 : Math.floor(Math.log(len) / (Math.LN2 * 7));

      // if 1 byte isn't enough for encoding message length, shift the data to the right
      pbf.realloc(extraLen);
      for (var i = pbf.pos - 1; i >= startPos; i--) pbf.buf[i + extraLen] = pbf.buf[i];
  }

  function writePackedVarint(arr, pbf)   { for (var i = 0; i < arr.length; i++) pbf.writeVarint(arr[i]);   }
  function writePackedSVarint(arr, pbf)  { for (var i = 0; i < arr.length; i++) pbf.writeSVarint(arr[i]);  }
  function writePackedFloat(arr, pbf)    { for (var i = 0; i < arr.length; i++) pbf.writeFloat(arr[i]);    }
  function writePackedDouble(arr, pbf)   { for (var i = 0; i < arr.length; i++) pbf.writeDouble(arr[i]);   }
  function writePackedBoolean(arr, pbf)  { for (var i = 0; i < arr.length; i++) pbf.writeBoolean(arr[i]);  }
  function writePackedFixed32(arr, pbf)  { for (var i = 0; i < arr.length; i++) pbf.writeFixed32(arr[i]);  }
  function writePackedSFixed32(arr, pbf) { for (var i = 0; i < arr.length; i++) pbf.writeSFixed32(arr[i]); }
  function writePackedFixed64(arr, pbf)  { for (var i = 0; i < arr.length; i++) pbf.writeFixed64(arr[i]);  }
  function writePackedSFixed64(arr, pbf) { for (var i = 0; i < arr.length; i++) pbf.writeSFixed64(arr[i]); }

  // Buffer code below from https://github.com/feross/buffer, MIT-licensed

  function readUInt32(buf, pos) {
      return ((buf[pos]) |
          (buf[pos + 1] << 8) |
          (buf[pos + 2] << 16)) +
          (buf[pos + 3] * 0x1000000);
  }

  function writeInt32(buf, val, pos) {
      buf[pos] = val;
      buf[pos + 1] = (val >>> 8);
      buf[pos + 2] = (val >>> 16);
      buf[pos + 3] = (val >>> 24);
  }

  function readInt32(buf, pos) {
      return ((buf[pos]) |
          (buf[pos + 1] << 8) |
          (buf[pos + 2] << 16)) +
          (buf[pos + 3] << 24);
  }

  function readUtf8(buf, pos, end) {
      var str = '';
      var i = pos;

      while (i < end) {
          var b0 = buf[i];
          var c = null; // codepoint
          var bytesPerSequence =
              b0 > 0xEF ? 4 :
              b0 > 0xDF ? 3 :
              b0 > 0xBF ? 2 : 1;

          if (i + bytesPerSequence > end) break;

          var b1, b2, b3;

          if (bytesPerSequence === 1) {
              if (b0 < 0x80) {
                  c = b0;
              }
          } else if (bytesPerSequence === 2) {
              b1 = buf[i + 1];
              if ((b1 & 0xC0) === 0x80) {
                  c = (b0 & 0x1F) << 0x6 | (b1 & 0x3F);
                  if (c <= 0x7F) {
                      c = null;
                  }
              }
          } else if (bytesPerSequence === 3) {
              b1 = buf[i + 1];
              b2 = buf[i + 2];
              if ((b1 & 0xC0) === 0x80 && (b2 & 0xC0) === 0x80) {
                  c = (b0 & 0xF) << 0xC | (b1 & 0x3F) << 0x6 | (b2 & 0x3F);
                  if (c <= 0x7FF || (c >= 0xD800 && c <= 0xDFFF)) {
                      c = null;
                  }
              }
          } else if (bytesPerSequence === 4) {
              b1 = buf[i + 1];
              b2 = buf[i + 2];
              b3 = buf[i + 3];
              if ((b1 & 0xC0) === 0x80 && (b2 & 0xC0) === 0x80 && (b3 & 0xC0) === 0x80) {
                  c = (b0 & 0xF) << 0x12 | (b1 & 0x3F) << 0xC | (b2 & 0x3F) << 0x6 | (b3 & 0x3F);
                  if (c <= 0xFFFF || c >= 0x110000) {
                      c = null;
                  }
              }
          }

          if (c === null) {
              c = 0xFFFD;
              bytesPerSequence = 1;

          } else if (c > 0xFFFF) {
              c -= 0x10000;
              str += String.fromCharCode(c >>> 10 & 0x3FF | 0xD800);
              c = 0xDC00 | c & 0x3FF;
          }

          str += String.fromCharCode(c);
          i += bytesPerSequence;
      }

      return str;
  }

  function readUtf8TextDecoder(buf, pos, end) {
      return utf8TextDecoder.decode(buf.subarray(pos, end));
  }

  function writeUtf8(buf, str, pos) {
      for (var i = 0, c, lead; i < str.length; i++) {
          c = str.charCodeAt(i); // code point

          if (c > 0xD7FF && c < 0xE000) {
              if (lead) {
                  if (c < 0xDC00) {
                      buf[pos++] = 0xEF;
                      buf[pos++] = 0xBF;
                      buf[pos++] = 0xBD;
                      lead = c;
                      continue;
                  } else {
                      c = lead - 0xD800 << 10 | c - 0xDC00 | 0x10000;
                      lead = null;
                  }
              } else {
                  if (c > 0xDBFF || (i + 1 === str.length)) {
                      buf[pos++] = 0xEF;
                      buf[pos++] = 0xBF;
                      buf[pos++] = 0xBD;
                  } else {
                      lead = c;
                  }
                  continue;
              }
          } else if (lead) {
              buf[pos++] = 0xEF;
              buf[pos++] = 0xBF;
              buf[pos++] = 0xBD;
              lead = null;
          }

          if (c < 0x80) {
              buf[pos++] = c;
          } else {
              if (c < 0x800) {
                  buf[pos++] = c >> 0x6 | 0xC0;
              } else {
                  if (c < 0x10000) {
                      buf[pos++] = c >> 0xC | 0xE0;
                  } else {
                      buf[pos++] = c >> 0x12 | 0xF0;
                      buf[pos++] = c >> 0xC & 0x3F | 0x80;
                  }
                  buf[pos++] = c >> 0x6 & 0x3F | 0x80;
              }
              buf[pos++] = c & 0x3F | 0x80;
          }
      }
      return pos;
  }

  /**
   * @module ol/format/MVT
   */

  /**
   * @typedef {Object} Options
   * @property {import("../Feature.js").FeatureClass} [featureClass] Class for features returned by
   * {@link module:ol/format/MVT~MVT#readFeatures}. Set to {@link module:ol/Feature~Feature} to get full editing and geometry
   * support at the cost of decreased rendering performance. The default is
   * {@link module:ol/render/Feature~RenderFeature}, which is optimized for rendering and hit detection.
   * @property {string} [geometryName='geometry'] Geometry name to use when creating features.
   * @property {string} [layerName='layer'] Name of the feature attribute that holds the layer name.
   * @property {Array<string>} [layers] Layers to read features from. If not provided, features will be read from all
   * @property {string} [idProperty] Optional property that will be assigned as the feature id and removed from the properties.
   * layers.
   */

  /**
   * @classdesc
   * Feature format for reading data in the Mapbox MVT format.
   *
   * @param {Options} [options] Options.
   * @api
   */
  class MVT extends FeatureFormat {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      super();

      options = options ? options : {};

      /**
       * @type {Projection}
       */
      this.dataProjection = new Projection({
        code: '',
        units: 'tile-pixels',
      });

      /**
       * @private
       * @type {import("../Feature.js").FeatureClass}
       */
      this.featureClass_ = options.featureClass
        ? options.featureClass
        : RenderFeature;

      /**
       * @private
       * @type {string|undefined}
       */
      this.geometryName_ = options.geometryName;

      /**
       * @private
       * @type {string}
       */
      this.layerName_ = options.layerName ? options.layerName : 'layer';

      /**
       * @private
       * @type {Array<string>|null}
       */
      this.layers_ = options.layers ? options.layers : null;

      /**
       * @private
       * @type {string}
       */
      this.idProperty_ = options.idProperty;

      this.supportedMediaTypes = [
        'application/vnd.mapbox-vector-tile',
        'application/x-protobuf',
      ];
    }

    /**
     * Read the raw geometry from the pbf offset stored in a raw feature's geometry
     * property.
     * @param {PBF} pbf PBF.
     * @param {Object} feature Raw feature.
     * @param {Array<number>} flatCoordinates Array to store flat coordinates in.
     * @param {Array<number>} ends Array to store ends in.
     * @private
     */
    readRawGeometry_(pbf, feature, flatCoordinates, ends) {
      pbf.pos = feature.geometry;

      const end = pbf.readVarint() + pbf.pos;
      let cmd = 1;
      let length = 0;
      let x = 0;
      let y = 0;
      let coordsLen = 0;
      let currentEnd = 0;

      while (pbf.pos < end) {
        if (!length) {
          const cmdLen = pbf.readVarint();
          cmd = cmdLen & 0x7;
          length = cmdLen >> 3;
        }

        length--;

        if (cmd === 1 || cmd === 2) {
          x += pbf.readSVarint();
          y += pbf.readSVarint();

          if (cmd === 1) {
            // moveTo
            if (coordsLen > currentEnd) {
              ends.push(coordsLen);
              currentEnd = coordsLen;
            }
          }

          flatCoordinates.push(x, y);
          coordsLen += 2;
        } else if (cmd === 7) {
          if (coordsLen > currentEnd) {
            // close polygon
            flatCoordinates.push(
              flatCoordinates[currentEnd],
              flatCoordinates[currentEnd + 1]
            );
            coordsLen += 2;
          }
        } else {
          assert(false, 59); // Invalid command found in the PBF
        }
      }

      if (coordsLen > currentEnd) {
        ends.push(coordsLen);
        currentEnd = coordsLen;
      }
    }

    /**
     * @private
     * @param {PBF} pbf PBF
     * @param {Object} rawFeature Raw Mapbox feature.
     * @param {import("./Feature.js").ReadOptions} options Read options.
     * @return {import("../Feature.js").FeatureLike|null} Feature.
     */
    createFeature_(pbf, rawFeature, options) {
      const type = rawFeature.type;
      if (type === 0) {
        return null;
      }

      let feature;
      const values = rawFeature.properties;

      let id;
      if (!this.idProperty_) {
        id = rawFeature.id;
      } else {
        id = values[this.idProperty_];
        delete values[this.idProperty_];
      }

      values[this.layerName_] = rawFeature.layer.name;

      const flatCoordinates = /** @type {Array<number>} */ ([]);
      const ends = /** @type {Array<number>} */ ([]);
      this.readRawGeometry_(pbf, rawFeature, flatCoordinates, ends);

      const geometryType = getGeometryType(type, ends.length);

      if (this.featureClass_ === RenderFeature) {
        feature = new this.featureClass_(
          geometryType,
          flatCoordinates,
          ends,
          values,
          id
        );
        feature.transform(options.dataProjection);
      } else {
        let geom;
        if (geometryType == 'Polygon') {
          const endss = inflateEnds(flatCoordinates, ends);
          geom =
            endss.length > 1
              ? new MultiPolygon(flatCoordinates, 'XY', endss)
              : new Polygon(flatCoordinates, 'XY', ends);
        } else {
          geom =
            geometryType === 'Point'
              ? new Point(flatCoordinates, 'XY')
              : geometryType === 'LineString'
              ? new LineString(flatCoordinates, 'XY')
              : geometryType === 'MultiPoint'
              ? new MultiPoint(flatCoordinates, 'XY')
              : geometryType === 'MultiLineString'
              ? new MultiLineString(flatCoordinates, 'XY', ends)
              : null;
        }
        const ctor = /** @type {typeof import("../Feature.js").default} */ (
          this.featureClass_
        );
        feature = new ctor();
        if (this.geometryName_) {
          feature.setGeometryName(this.geometryName_);
        }
        const geometry = transformGeometryWithOptions(geom, false, options);
        feature.setGeometry(geometry);
        if (id !== undefined) {
          feature.setId(id);
        }
        feature.setProperties(values, true);
      }

      return feature;
    }

    /**
     * @return {import("./Feature.js").Type} Format.
     */
    getType() {
      return 'arraybuffer';
    }

    /**
     * Read all features.
     *
     * @param {ArrayBuffer} source Source.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @return {Array<import("../Feature.js").FeatureLike>} Features.
     * @api
     */
    readFeatures(source, options) {
      const layers = this.layers_;
      options = this.adaptOptions(options);
      const dataProjection = get$2(options.dataProjection);
      dataProjection.setWorldExtent(options.extent);
      options.dataProjection = dataProjection;

      const pbf$1 = new pbf(/** @type {ArrayBuffer} */ (source));
      const pbfLayers = pbf$1.readFields(layersPBFReader, {});
      const features = [];
      for (const name in pbfLayers) {
        if (layers && !layers.includes(name)) {
          continue;
        }
        const pbfLayer = pbfLayers[name];

        const extent = pbfLayer ? [0, 0, pbfLayer.extent, pbfLayer.extent] : null;
        dataProjection.setExtent(extent);

        for (let i = 0, ii = pbfLayer.length; i < ii; ++i) {
          const rawFeature = readRawFeature(pbf$1, pbfLayer, i);
          const feature = this.createFeature_(pbf$1, rawFeature, options);
          if (feature !== null) {
            features.push(feature);
          }
        }
      }

      return features;
    }

    /**
     * Read the projection from the source.
     *
     * @param {Document|Element|Object|string} source Source.
     * @return {import("../proj/Projection.js").default} Projection.
     * @api
     */
    readProjection(source) {
      return this.dataProjection;
    }

    /**
     * Sets the layers that features will be read from.
     * @param {Array<string>} layers Layers.
     * @api
     */
    setLayers(layers) {
      this.layers_ = layers;
    }
  }

  /**
   * Reader callback for parsing layers.
   * @param {number} tag The tag.
   * @param {Object} layers The layers object.
   * @param {PBF} pbf The PBF.
   */
  function layersPBFReader(tag, layers, pbf) {
    if (tag === 3) {
      const layer = {
        keys: [],
        values: [],
        features: [],
      };
      const end = pbf.readVarint() + pbf.pos;
      pbf.readFields(layerPBFReader, layer, end);
      layer.length = layer.features.length;
      if (layer.length) {
        layers[layer.name] = layer;
      }
    }
  }

  /**
   * Reader callback for parsing layer.
   * @param {number} tag The tag.
   * @param {Object} layer The layer object.
   * @param {PBF} pbf The PBF.
   */
  function layerPBFReader(tag, layer, pbf) {
    if (tag === 15) {
      layer.version = pbf.readVarint();
    } else if (tag === 1) {
      layer.name = pbf.readString();
    } else if (tag === 5) {
      layer.extent = pbf.readVarint();
    } else if (tag === 2) {
      layer.features.push(pbf.pos);
    } else if (tag === 3) {
      layer.keys.push(pbf.readString());
    } else if (tag === 4) {
      let value = null;
      const end = pbf.readVarint() + pbf.pos;
      while (pbf.pos < end) {
        tag = pbf.readVarint() >> 3;
        value =
          tag === 1
            ? pbf.readString()
            : tag === 2
            ? pbf.readFloat()
            : tag === 3
            ? pbf.readDouble()
            : tag === 4
            ? pbf.readVarint64()
            : tag === 5
            ? pbf.readVarint()
            : tag === 6
            ? pbf.readSVarint()
            : tag === 7
            ? pbf.readBoolean()
            : null;
      }
      layer.values.push(value);
    }
  }

  /**
   * Reader callback for parsing feature.
   * @param {number} tag The tag.
   * @param {Object} feature The feature object.
   * @param {PBF} pbf The PBF.
   */
  function featurePBFReader(tag, feature, pbf) {
    if (tag == 1) {
      feature.id = pbf.readVarint();
    } else if (tag == 2) {
      const end = pbf.readVarint() + pbf.pos;
      while (pbf.pos < end) {
        const key = feature.layer.keys[pbf.readVarint()];
        const value = feature.layer.values[pbf.readVarint()];
        feature.properties[key] = value;
      }
    } else if (tag == 3) {
      feature.type = pbf.readVarint();
    } else if (tag == 4) {
      feature.geometry = pbf.pos;
    }
  }

  /**
   * Read a raw feature from the pbf offset stored at index `i` in the raw layer.
   * @param {PBF} pbf PBF.
   * @param {Object} layer Raw layer.
   * @param {number} i Index of the feature in the raw layer's `features` array.
   * @return {Object} Raw feature.
   */
  function readRawFeature(pbf, layer, i) {
    pbf.pos = layer.features[i];
    const end = pbf.readVarint() + pbf.pos;

    const feature = {
      layer: layer,
      type: 0,
      properties: {},
    };
    pbf.readFields(featurePBFReader, feature, end);
    return feature;
  }

  /**
   * @param {number} type The raw feature's geometry type
   * @param {number} numEnds Number of ends of the flat coordinates of the
   * geometry.
   * @return {import("../geom/Geometry.js").Type} The geometry type.
   */
  function getGeometryType(type, numEnds) {
    /** @type {import("../geom/Geometry.js").Type} */
    let geometryType;
    if (type === 1) {
      geometryType = numEnds === 1 ? 'Point' : 'MultiPoint';
    } else if (type === 2) {
      geometryType = numEnds === 1 ? 'LineString' : 'MultiLineString';
    } else if (type === 3) {
      geometryType = 'Polygon';
      // MultiPolygon not relevant for rendering - winding order determines
      // outer rings of polygons.
    }
    return geometryType;
  }

  /**
   * @module ol/layer/VectorTile
   */

  /***
   * @template Return
   * @typedef {import("../Observable").OnSignature<import("../Observable").EventTypes, import("../events/Event.js").default, Return> &
   *   import("../Observable").OnSignature<import("./Base").BaseLayerObjectEventTypes|
   *     'change:source'|'change:preload'|'change:useInterimTilesOnError', import("../Object").ObjectEvent, Return> &
   *   import("../Observable").OnSignature<import("../render/EventType").LayerRenderEventTypes, import("../render/Event").default, Return> &
   *   import("../Observable").CombinedOnSignature<import("../Observable").EventTypes|import("./Base").BaseLayerObjectEventTypes|
   *     'change:source'|'change:preload'|'change:useInterimTilesOnError'|import("../render/EventType").LayerRenderEventTypes, Return>} VectorTileLayerOnSignature
   */

  /**
   * @typedef {'hybrid' | 'vector'} VectorTileRenderType
   */

  /**
   * @typedef {Object} Options
   * @property {string} [className='ol-layer'] A CSS class name to set to the layer element.
   * @property {number} [opacity=1] Opacity (0, 1).
   * @property {boolean} [visible=true] Visibility.
   * @property {import("../extent.js").Extent} [extent] The bounding extent for layer rendering.  The layer will not be
   * rendered outside of this extent.
   * @property {number} [zIndex] The z-index for layer rendering.  At rendering time, the layers
   * will be ordered, first by Z-index and then by position. When `undefined`, a `zIndex` of 0 is assumed
   * for layers that are added to the map's `layers` collection, or `Infinity` when the layer's `setMap()`
   * method was used.
   * @property {number} [minResolution] The minimum resolution (inclusive) at which this layer will be
   * visible.
   * @property {number} [maxResolution] The maximum resolution (exclusive) below which this layer will
   * be visible.
   * @property {number} [minZoom] The minimum view zoom level (exclusive) above which this layer will be
   * visible.
   * @property {number} [maxZoom] The maximum view zoom level (inclusive) at which this layer will
   * be visible.
   * @property {import("../render.js").OrderFunction} [renderOrder] Render order. Function to be used when sorting
   * features before rendering. By default features are drawn in the order that they are created. Use
   * `null` to avoid the sort, but get an undefined draw order.
   * @property {number} [renderBuffer=100] The buffer in pixels around the tile extent used by the
   * renderer when getting features from the vector tile for the rendering or hit-detection.
   * Recommended value: Vector tiles are usually generated with a buffer, so this value should match
   * the largest possible buffer of the used tiles. It should be at least the size of the largest
   * point symbol or line width.
   * @property {VectorTileRenderType} [renderMode='hybrid'] Render mode for vector tiles:
   *  * `'hybrid'`: Polygon and line elements are rendered as images, so pixels are scaled during zoom
   *    animations. Point symbols and texts are accurately rendered as vectors and can stay upright on
   *    rotated views.
   *  * `'vector'`: Everything is rendered as vectors. Use this mode for improved performance on vector
   *    tile layers with only a few rendered features (e.g. for highlighting a subset of features of
   *    another layer with the same source).
   * @property {import("../source/VectorTile.js").default} [source] Source.
   * @property {import("../Map.js").default} [map] Sets the layer as overlay on a map. The map will not manage
   * this layer in its layers collection, and the layer will be rendered on top. This is useful for
   * temporary layers. The standard way to add a layer to a map and have it managed by the map is to
   * use [map.addLayer()]{@link import("../Map.js").default#addLayer}.
   * @property {boolean} [declutter=false] Declutter images and text. Decluttering is applied to all
   * image and text styles of all Vector and VectorTile layers that have set this to `true`. The priority
   * is defined by the z-index of the layer, the `zIndex` of the style and the render order of features.
   * Higher z-index means higher priority. Within the same z-index, a feature rendered before another has
   * higher priority.
   *
   * As an optimization decluttered features from layers with the same `className` are rendered above
   * the fill and stroke styles of all of those layers regardless of z-index.  To opt out of this
   * behavior and place declutterd features with their own layer configure the layer with a `className`
   * other than `ol-layer`.
   * @property {import("../style/Style.js").StyleLike|null} [style] Layer style. When set to `null`, only
   * features that have their own style will be rendered. See {@link module:ol/style/Style~Style} for the default style
   * which will be used if this is not set.
   * @property {import("./Base.js").BackgroundColor|false} [background] Background color for the layer. If not specified, no
   * background will be rendered.
   * @property {boolean} [updateWhileAnimating=false] When set to `true`, feature batches will be
   * recreated during animations. This means that no vectors will be shown clipped, but the setting
   * will have a performance impact for large amounts of vector data. When set to `false`, batches
   * will be recreated when no animation is active.
   * @property {boolean} [updateWhileInteracting=false] When set to `true`, feature batches will be
   * recreated during interactions. See also `updateWhileAnimating`.
   * @property {number} [preload=0] Preload. Load low-resolution tiles up to `preload` levels. `0`
   * means no preloading.
   * @property {boolean} [useInterimTilesOnError=true] Use interim tiles on error.
   * @property {Object<string, *>} [properties] Arbitrary observable properties. Can be accessed with `#get()` and `#set()`.
   */

  /**
   * @classdesc
   * Layer for vector tile data that is rendered client-side.
   * Note that any property set in the options is set as a {@link module:ol/Object~BaseObject}
   * property on the layer object; for example, setting `title: 'My Title'` in the
   * options means that `title` is observable, and has get/set accessors.
   *
   * @param {Options} [options] Options.
   * @extends {BaseVectorLayer<import("../source/VectorTile.js").default, CanvasVectorTileLayerRenderer>}
   * @api
   */
  class VectorTileLayer extends BaseVectorLayer {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      options = options ? options : {};

      const baseOptions = /** @type {Object} */ (Object.assign({}, options));
      delete baseOptions.preload;
      delete baseOptions.useInterimTilesOnError;

      super(
        /** @type {import("./BaseVector.js").Options<import("../source/VectorTile.js").default>} */ (
          baseOptions
        )
      );

      /***
       * @type {VectorTileLayerOnSignature<import("../events").EventsKey>}
       */
      this.on;

      /***
       * @type {VectorTileLayerOnSignature<import("../events").EventsKey>}
       */
      this.once;

      /***
       * @type {VectorTileLayerOnSignature<void>}
       */
      this.un;

      const renderMode = options.renderMode || 'hybrid';
      // `renderMode` must be `'hybrid'` or `'vector'`.
      assert(renderMode == 'hybrid' || renderMode == 'vector', 28);

      /**
       * @private
       * @type {VectorTileRenderType}
       */
      this.renderMode_ = renderMode;

      this.setPreload(options.preload ? options.preload : 0);
      this.setUseInterimTilesOnError(
        options.useInterimTilesOnError !== undefined
          ? options.useInterimTilesOnError
          : true
      );

      /**
       * @return {import("./Base.js").BackgroundColor} Background color.
       * @function
       * @api
       */
      this.getBackground;

      /**
       * @param {import("./Base.js").BackgroundColor} background Background color.
       * @function
       * @api
       */
      this.setBackground;
    }

    createRenderer() {
      return new CanvasVectorTileLayerRenderer(this);
    }

    /**
     * Get the topmost feature that intersects the given pixel on the viewport. Returns a promise
     * that resolves with an array of features. The array will either contain the topmost feature
     * when a hit was detected, or it will be empty.
     *
     * The hit detection algorithm used for this method is optimized for performance, but is less
     * accurate than the one used in [map.getFeaturesAtPixel()]{@link import("../Map.js").default#getFeaturesAtPixel}.
     * Text is not considered, and icons are only represented by their bounding box instead of the exact
     * image.
     *
     * @param {import("../pixel.js").Pixel} pixel Pixel.
     * @return {Promise<Array<import("../Feature").default>>} Promise that resolves with an array of features.
     * @api
     */
    getFeatures(pixel) {
      return super.getFeatures(pixel);
    }

    /**
     * @return {VectorTileRenderType} The render mode.
     */
    getRenderMode() {
      return this.renderMode_;
    }

    /**
     * Return the level as number to which we will preload tiles up to.
     * @return {number} The level to preload tiles up to.
     * @observable
     * @api
     */
    getPreload() {
      return /** @type {number} */ (this.get(TileProperty.PRELOAD));
    }

    /**
     * Whether we use interim tiles on error.
     * @return {boolean} Use interim tiles on error.
     * @observable
     * @api
     */
    getUseInterimTilesOnError() {
      return /** @type {boolean} */ (
        this.get(TileProperty.USE_INTERIM_TILES_ON_ERROR)
      );
    }

    /**
     * Set the level as number to which we will preload tiles up to.
     * @param {number} preload The level to preload tiles up to.
     * @observable
     * @api
     */
    setPreload(preload) {
      this.set(TileProperty.PRELOAD, preload);
    }

    /**
     * Set whether we use interim tiles on error.
     * @param {boolean} useInterimTilesOnError Use interim tiles on error.
     * @observable
     * @api
     */
    setUseInterimTilesOnError(useInterimTilesOnError) {
      this.set(TileProperty.USE_INTERIM_TILES_ON_ERROR, useInterimTilesOnError);
    }
  }

  /**
   * @module ol/format/JSONFeature
   */

  /**
   * @classdesc
   * Abstract base class; normally only used for creating subclasses and not
   * instantiated in apps.
   * Base class for JSON feature formats.
   *
   * @abstract
   */
  class JSONFeature extends FeatureFormat {
    constructor() {
      super();
    }

    /**
     * @return {import("./Feature.js").Type} Format.
     */
    getType() {
      return 'json';
    }

    /**
     * Read a feature.  Only works for a single feature. Use `readFeatures` to
     * read a feature collection.
     *
     * @param {ArrayBuffer|Document|Element|Object|string} source Source.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @return {import("../Feature.js").default} Feature.
     * @api
     */
    readFeature(source, options) {
      return this.readFeatureFromObject(
        getObject(source),
        this.getReadOptions(source, options)
      );
    }

    /**
     * Read all features.  Works with both a single feature and a feature
     * collection.
     *
     * @param {ArrayBuffer|Document|Element|Object|string} source Source.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @return {Array<import("../Feature.js").default>} Features.
     * @api
     */
    readFeatures(source, options) {
      return this.readFeaturesFromObject(
        getObject(source),
        this.getReadOptions(source, options)
      );
    }

    /**
     * @abstract
     * @param {Object} object Object.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @protected
     * @return {import("../Feature.js").default} Feature.
     */
    readFeatureFromObject(object, options) {
      return abstract();
    }

    /**
     * @abstract
     * @param {Object} object Object.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @protected
     * @return {Array<import("../Feature.js").default>} Features.
     */
    readFeaturesFromObject(object, options) {
      return abstract();
    }

    /**
     * Read a geometry.
     *
     * @param {ArrayBuffer|Document|Element|Object|string} source Source.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @return {import("../geom/Geometry.js").default} Geometry.
     * @api
     */
    readGeometry(source, options) {
      return this.readGeometryFromObject(
        getObject(source),
        this.getReadOptions(source, options)
      );
    }

    /**
     * @abstract
     * @param {Object} object Object.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @protected
     * @return {import("../geom/Geometry.js").default} Geometry.
     */
    readGeometryFromObject(object, options) {
      return abstract();
    }

    /**
     * Read the projection.
     *
     * @param {ArrayBuffer|Document|Element|Object|string} source Source.
     * @return {import("../proj/Projection.js").default} Projection.
     * @api
     */
    readProjection(source) {
      return this.readProjectionFromObject(getObject(source));
    }

    /**
     * @abstract
     * @param {Object} object Object.
     * @protected
     * @return {import("../proj/Projection.js").default} Projection.
     */
    readProjectionFromObject(object) {
      return abstract();
    }

    /**
     * Encode a feature as string.
     *
     * @param {import("../Feature.js").default} feature Feature.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @return {string} Encoded feature.
     * @api
     */
    writeFeature(feature, options) {
      return JSON.stringify(this.writeFeatureObject(feature, options));
    }

    /**
     * @abstract
     * @param {import("../Feature.js").default} feature Feature.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @return {Object} Object.
     */
    writeFeatureObject(feature, options) {
      return abstract();
    }

    /**
     * Encode an array of features as string.
     *
     * @param {Array<import("../Feature.js").default>} features Features.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @return {string} Encoded features.
     * @api
     */
    writeFeatures(features, options) {
      return JSON.stringify(this.writeFeaturesObject(features, options));
    }

    /**
     * @abstract
     * @param {Array<import("../Feature.js").default>} features Features.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @return {Object} Object.
     */
    writeFeaturesObject(features, options) {
      return abstract();
    }

    /**
     * Encode a geometry as string.
     *
     * @param {import("../geom/Geometry.js").default} geometry Geometry.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @return {string} Encoded geometry.
     * @api
     */
    writeGeometry(geometry, options) {
      return JSON.stringify(this.writeGeometryObject(geometry, options));
    }

    /**
     * @abstract
     * @param {import("../geom/Geometry.js").default} geometry Geometry.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @return {Object} Object.
     */
    writeGeometryObject(geometry, options) {
      return abstract();
    }
  }

  /**
   * @param {Document|Element|Object|string} source Source.
   * @return {Object} Object.
   */
  function getObject(source) {
    if (typeof source === 'string') {
      const object = JSON.parse(source);
      return object ? /** @type {Object} */ (object) : null;
    } else if (source !== null) {
      return source;
    } else {
      return null;
    }
  }

  /**
   * @module ol/format/GeoJSON
   */

  /**
   * @typedef {import("geojson").GeoJSON} GeoJSONObject
   * @typedef {import("geojson").Feature} GeoJSONFeature
   * @typedef {import("geojson").FeatureCollection} GeoJSONFeatureCollection
   * @typedef {import("geojson").Geometry} GeoJSONGeometry
   * @typedef {import("geojson").Point} GeoJSONPoint
   * @typedef {import("geojson").LineString} GeoJSONLineString
   * @typedef {import("geojson").Polygon} GeoJSONPolygon
   * @typedef {import("geojson").MultiPoint} GeoJSONMultiPoint
   * @typedef {import("geojson").MultiLineString} GeoJSONMultiLineString
   * @typedef {import("geojson").MultiPolygon} GeoJSONMultiPolygon
   * @typedef {import("geojson").GeometryCollection} GeoJSONGeometryCollection
   */

  /**
   * @typedef {Object} Options
   * @property {import("../proj.js").ProjectionLike} [dataProjection='EPSG:4326'] Default data projection.
   * @property {import("../proj.js").ProjectionLike} [featureProjection] Projection for features read or
   * written by the format.  Options passed to read or write methods will take precedence.
   * @property {string} [geometryName] Geometry name to use when creating features.
   * @property {boolean} [extractGeometryName=false] Certain GeoJSON providers include
   * the geometry_name field in the feature GeoJSON. If set to `true` the GeoJSON reader
   * will look for that field to set the geometry name. If both this field is set to `true`
   * and a `geometryName` is provided, the `geometryName` will take precedence.
   */

  /**
   * @classdesc
   * Feature format for reading and writing data in the GeoJSON format.
   *
   * @api
   */
  class GeoJSON extends JSONFeature {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      options = options ? options : {};

      super();

      /**
       * @type {import("../proj/Projection.js").default}
       */
      this.dataProjection = get$2(
        options.dataProjection ? options.dataProjection : 'EPSG:4326'
      );

      if (options.featureProjection) {
        /**
         * @type {import("../proj/Projection.js").default}
         */
        this.defaultFeatureProjection = get$2(options.featureProjection);
      }

      /**
       * Name of the geometry attribute for features.
       * @type {string|undefined}
       * @private
       */
      this.geometryName_ = options.geometryName;

      /**
       * Look for the geometry name in the feature GeoJSON
       * @type {boolean|undefined}
       * @private
       */
      this.extractGeometryName_ = options.extractGeometryName;

      this.supportedMediaTypes = [
        'application/geo+json',
        'application/vnd.geo+json',
      ];
    }

    /**
     * @param {Object} object Object.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @protected
     * @return {import("../Feature.js").default} Feature.
     */
    readFeatureFromObject(object, options) {
      /**
       * @type {GeoJSONFeature}
       */
      let geoJSONFeature = null;
      if (object['type'] === 'Feature') {
        geoJSONFeature = /** @type {GeoJSONFeature} */ (object);
      } else {
        geoJSONFeature = {
          'type': 'Feature',
          'geometry': /** @type {GeoJSONGeometry} */ (object),
          'properties': null,
        };
      }

      const geometry = readGeometry$1(geoJSONFeature['geometry'], options);
      const feature = new Feature();
      if (this.geometryName_) {
        feature.setGeometryName(this.geometryName_);
      } else if (
        this.extractGeometryName_ &&
        'geometry_name' in geoJSONFeature !== undefined
      ) {
        feature.setGeometryName(geoJSONFeature['geometry_name']);
      }
      feature.setGeometry(geometry);

      if ('id' in geoJSONFeature) {
        feature.setId(geoJSONFeature['id']);
      }

      if (geoJSONFeature['properties']) {
        feature.setProperties(geoJSONFeature['properties'], true);
      }
      return feature;
    }

    /**
     * @param {Object} object Object.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @protected
     * @return {Array<Feature>} Features.
     */
    readFeaturesFromObject(object, options) {
      const geoJSONObject = /** @type {GeoJSONObject} */ (object);
      /** @type {Array<import("../Feature.js").default>} */
      let features = null;
      if (geoJSONObject['type'] === 'FeatureCollection') {
        const geoJSONFeatureCollection = /** @type {GeoJSONFeatureCollection} */ (
          object
        );
        features = [];
        const geoJSONFeatures = geoJSONFeatureCollection['features'];
        for (let i = 0, ii = geoJSONFeatures.length; i < ii; ++i) {
          features.push(this.readFeatureFromObject(geoJSONFeatures[i], options));
        }
      } else {
        features = [this.readFeatureFromObject(object, options)];
      }
      return features;
    }

    /**
     * @param {GeoJSONGeometry} object Object.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @protected
     * @return {import("../geom/Geometry.js").default} Geometry.
     */
    readGeometryFromObject(object, options) {
      return readGeometry$1(object, options);
    }

    /**
     * @param {Object} object Object.
     * @protected
     * @return {import("../proj/Projection.js").default} Projection.
     */
    readProjectionFromObject(object) {
      const crs = object['crs'];
      let projection;
      if (crs) {
        if (crs['type'] == 'name') {
          projection = get$2(crs['properties']['name']);
        } else if (crs['type'] === 'EPSG') {
          projection = get$2('EPSG:' + crs['properties']['code']);
        } else {
          assert(false, 36); // Unknown SRS type
        }
      } else {
        projection = this.dataProjection;
      }
      return /** @type {import("../proj/Projection.js").default} */ (projection);
    }

    /**
     * Encode a feature as a GeoJSON Feature object.
     *
     * @param {import("../Feature.js").default} feature Feature.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @return {GeoJSONFeature} Object.
     * @api
     */
    writeFeatureObject(feature, options) {
      options = this.adaptOptions(options);

      /** @type {GeoJSONFeature} */
      const object = {
        'type': 'Feature',
        geometry: null,
        properties: null,
      };

      const id = feature.getId();
      if (id !== undefined) {
        object.id = id;
      }

      if (!feature.hasProperties()) {
        return object;
      }

      const properties = feature.getProperties();
      const geometry = feature.getGeometry();
      if (geometry) {
        object.geometry = writeGeometry$1(geometry, options);

        delete properties[feature.getGeometryName()];
      }

      if (!isEmpty$1(properties)) {
        object.properties = properties;
      }

      return object;
    }

    /**
     * Encode an array of features as a GeoJSON object.
     *
     * @param {Array<import("../Feature.js").default>} features Features.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @return {GeoJSONFeatureCollection} GeoJSON Object.
     * @api
     */
    writeFeaturesObject(features, options) {
      options = this.adaptOptions(options);
      const objects = [];
      for (let i = 0, ii = features.length; i < ii; ++i) {
        objects.push(this.writeFeatureObject(features[i], options));
      }
      return {
        type: 'FeatureCollection',
        features: objects,
      };
    }

    /**
     * Encode a geometry as a GeoJSON object.
     *
     * @param {import("../geom/Geometry.js").default} geometry Geometry.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @return {GeoJSONGeometry|GeoJSONGeometryCollection} Object.
     * @api
     */
    writeGeometryObject(geometry, options) {
      return writeGeometry$1(geometry, this.adaptOptions(options));
    }
  }

  /**
   * @param {GeoJSONGeometry|GeoJSONGeometryCollection} object Object.
   * @param {import("./Feature.js").ReadOptions} [options] Read options.
   * @return {import("../geom/Geometry.js").default} Geometry.
   */
  function readGeometry$1(object, options) {
    if (!object) {
      return null;
    }

    /**
     * @type {import("../geom/Geometry.js").default}
     */
    let geometry;
    switch (object['type']) {
      case 'Point': {
        geometry = readPointGeometry$2(/** @type {GeoJSONPoint} */ (object));
        break;
      }
      case 'LineString': {
        geometry = readLineStringGeometry$2(
          /** @type {GeoJSONLineString} */ (object)
        );
        break;
      }
      case 'Polygon': {
        geometry = readPolygonGeometry$2(/** @type {GeoJSONPolygon} */ (object));
        break;
      }
      case 'MultiPoint': {
        geometry = readMultiPointGeometry$2(
          /** @type {GeoJSONMultiPoint} */ (object)
        );
        break;
      }
      case 'MultiLineString': {
        geometry = readMultiLineStringGeometry$2(
          /** @type {GeoJSONMultiLineString} */ (object)
        );
        break;
      }
      case 'MultiPolygon': {
        geometry = readMultiPolygonGeometry$2(
          /** @type {GeoJSONMultiPolygon} */ (object)
        );
        break;
      }
      case 'GeometryCollection': {
        geometry = readGeometryCollectionGeometry(
          /** @type {GeoJSONGeometryCollection} */ (object)
        );
        break;
      }
      default: {
        throw new Error('Unsupported GeoJSON type: ' + object['type']);
      }
    }
    return transformGeometryWithOptions(geometry, false, options);
  }

  /**
   * @param {GeoJSONGeometryCollection} object Object.
   * @param {import("./Feature.js").ReadOptions} [options] Read options.
   * @return {GeometryCollection} Geometry collection.
   */
  function readGeometryCollectionGeometry(object, options) {
    const geometries = object['geometries'].map(
      /**
       * @param {GeoJSONGeometry} geometry Geometry.
       * @return {import("../geom/Geometry.js").default} geometry Geometry.
       */
      function (geometry) {
        return readGeometry$1(geometry, options);
      }
    );
    return new GeometryCollection(geometries);
  }

  /**
   * @param {GeoJSONPoint} object Object.
   * @return {Point} Point.
   */
  function readPointGeometry$2(object) {
    return new Point(object['coordinates']);
  }

  /**
   * @param {GeoJSONLineString} object Object.
   * @return {LineString} LineString.
   */
  function readLineStringGeometry$2(object) {
    return new LineString(object['coordinates']);
  }

  /**
   * @param {GeoJSONMultiLineString} object Object.
   * @return {MultiLineString} MultiLineString.
   */
  function readMultiLineStringGeometry$2(object) {
    return new MultiLineString(object['coordinates']);
  }

  /**
   * @param {GeoJSONMultiPoint} object Object.
   * @return {MultiPoint} MultiPoint.
   */
  function readMultiPointGeometry$2(object) {
    return new MultiPoint(object['coordinates']);
  }

  /**
   * @param {GeoJSONMultiPolygon} object Object.
   * @return {MultiPolygon} MultiPolygon.
   */
  function readMultiPolygonGeometry$2(object) {
    return new MultiPolygon(object['coordinates']);
  }

  /**
   * @param {GeoJSONPolygon} object Object.
   * @return {Polygon} Polygon.
   */
  function readPolygonGeometry$2(object) {
    return new Polygon(object['coordinates']);
  }

  /**
   * @param {import("../geom/Geometry.js").default} geometry Geometry.
   * @param {import("./Feature.js").WriteOptions} [options] Write options.
   * @return {GeoJSONGeometry} GeoJSON geometry.
   */
  function writeGeometry$1(geometry, options) {
    geometry = transformGeometryWithOptions(geometry, true, options);
    const type = geometry.getType();

    /** @type {GeoJSONGeometry} */
    let geoJSON;
    switch (type) {
      case 'Point': {
        geoJSON = writePointGeometry$1(/** @type {Point} */ (geometry));
        break;
      }
      case 'LineString': {
        geoJSON = writeLineStringGeometry$1(
          /** @type {LineString} */ (geometry));
        break;
      }
      case 'Polygon': {
        geoJSON = writePolygonGeometry$1(
          /** @type {Polygon} */ (geometry),
          options
        );
        break;
      }
      case 'MultiPoint': {
        geoJSON = writeMultiPointGeometry$1(
          /** @type {MultiPoint} */ (geometry));
        break;
      }
      case 'MultiLineString': {
        geoJSON = writeMultiLineStringGeometry$1(
          /** @type {MultiLineString} */ (geometry));
        break;
      }
      case 'MultiPolygon': {
        geoJSON = writeMultiPolygonGeometry$1(
          /** @type {MultiPolygon} */ (geometry),
          options
        );
        break;
      }
      case 'GeometryCollection': {
        geoJSON = writeGeometryCollectionGeometry(
          /** @type {GeometryCollection} */ (geometry),
          options
        );
        break;
      }
      case 'Circle': {
        geoJSON = {
          type: 'GeometryCollection',
          geometries: [],
        };
        break;
      }
      default: {
        throw new Error('Unsupported geometry type: ' + type);
      }
    }
    return geoJSON;
  }

  /**
   * @param {GeometryCollection} geometry Geometry.
   * @param {import("./Feature.js").WriteOptions} [options] Write options.
   * @return {GeoJSONGeometryCollection} GeoJSON geometry collection.
   */
  function writeGeometryCollectionGeometry(geometry, options) {
    options = Object.assign({}, options);
    delete options.featureProjection;
    const geometries = geometry.getGeometriesArray().map(function (geometry) {
      return writeGeometry$1(geometry, options);
    });
    return {
      type: 'GeometryCollection',
      geometries: geometries,
    };
  }

  /**
   * @param {LineString} geometry Geometry.
   * @param {import("./Feature.js").WriteOptions} [options] Write options.
   * @return {GeoJSONGeometry} GeoJSON geometry.
   */
  function writeLineStringGeometry$1(geometry, options) {
    return {
      type: 'LineString',
      coordinates: geometry.getCoordinates(),
    };
  }

  /**
   * @param {MultiLineString} geometry Geometry.
   * @param {import("./Feature.js").WriteOptions} [options] Write options.
   * @return {GeoJSONGeometry} GeoJSON geometry.
   */
  function writeMultiLineStringGeometry$1(geometry, options) {
    return {
      type: 'MultiLineString',
      coordinates: geometry.getCoordinates(),
    };
  }

  /**
   * @param {MultiPoint} geometry Geometry.
   * @param {import("./Feature.js").WriteOptions} [options] Write options.
   * @return {GeoJSONGeometry} GeoJSON geometry.
   */
  function writeMultiPointGeometry$1(geometry, options) {
    return {
      type: 'MultiPoint',
      coordinates: geometry.getCoordinates(),
    };
  }

  /**
   * @param {MultiPolygon} geometry Geometry.
   * @param {import("./Feature.js").WriteOptions} [options] Write options.
   * @return {GeoJSONGeometry} GeoJSON geometry.
   */
  function writeMultiPolygonGeometry$1(geometry, options) {
    let right;
    if (options) {
      right = options.rightHanded;
    }
    return {
      type: 'MultiPolygon',
      coordinates: geometry.getCoordinates(right),
    };
  }

  /**
   * @param {Point} geometry Geometry.
   * @param {import("./Feature.js").WriteOptions} [options] Write options.
   * @return {GeoJSONGeometry} GeoJSON geometry.
   */
  function writePointGeometry$1(geometry, options) {
    return {
      type: 'Point',
      coordinates: geometry.getCoordinates(),
    };
  }

  /**
   * @param {Polygon} geometry Geometry.
   * @param {import("./Feature.js").WriteOptions} [options] Write options.
   * @return {GeoJSONGeometry} GeoJSON geometry.
   */
  function writePolygonGeometry$1(geometry, options) {
    let right;
    if (options) {
      right = options.rightHanded;
    }
    return {
      type: 'Polygon',
      coordinates: geometry.getCoordinates(right),
    };
  }

  var csscolorparser = {};

  var parseCSSColor_1;
  // (c) Dean McNamee <dean@gmail.com>, 2012.
  //
  // https://github.com/deanm/css-color-parser-js
  //
  // Permission is hereby granted, free of charge, to any person obtaining a copy
  // of this software and associated documentation files (the "Software"), to
  // deal in the Software without restriction, including without limitation the
  // rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
  // sell copies of the Software, and to permit persons to whom the Software is
  // furnished to do so, subject to the following conditions:
  //
  // The above copyright notice and this permission notice shall be included in
  // all copies or substantial portions of the Software.
  //
  // THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
  // IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
  // FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
  // AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
  // LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  // FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
  // IN THE SOFTWARE.
  // http://www.w3.org/TR/css3-color/
  var kCSSColorTable = {
      'transparent': [
          0,
          0,
          0,
          0
      ],
      'aliceblue': [
          240,
          248,
          255,
          1
      ],
      'antiquewhite': [
          250,
          235,
          215,
          1
      ],
      'aqua': [
          0,
          255,
          255,
          1
      ],
      'aquamarine': [
          127,
          255,
          212,
          1
      ],
      'azure': [
          240,
          255,
          255,
          1
      ],
      'beige': [
          245,
          245,
          220,
          1
      ],
      'bisque': [
          255,
          228,
          196,
          1
      ],
      'black': [
          0,
          0,
          0,
          1
      ],
      'blanchedalmond': [
          255,
          235,
          205,
          1
      ],
      'blue': [
          0,
          0,
          255,
          1
      ],
      'blueviolet': [
          138,
          43,
          226,
          1
      ],
      'brown': [
          165,
          42,
          42,
          1
      ],
      'burlywood': [
          222,
          184,
          135,
          1
      ],
      'cadetblue': [
          95,
          158,
          160,
          1
      ],
      'chartreuse': [
          127,
          255,
          0,
          1
      ],
      'chocolate': [
          210,
          105,
          30,
          1
      ],
      'coral': [
          255,
          127,
          80,
          1
      ],
      'cornflowerblue': [
          100,
          149,
          237,
          1
      ],
      'cornsilk': [
          255,
          248,
          220,
          1
      ],
      'crimson': [
          220,
          20,
          60,
          1
      ],
      'cyan': [
          0,
          255,
          255,
          1
      ],
      'darkblue': [
          0,
          0,
          139,
          1
      ],
      'darkcyan': [
          0,
          139,
          139,
          1
      ],
      'darkgoldenrod': [
          184,
          134,
          11,
          1
      ],
      'darkgray': [
          169,
          169,
          169,
          1
      ],
      'darkgreen': [
          0,
          100,
          0,
          1
      ],
      'darkgrey': [
          169,
          169,
          169,
          1
      ],
      'darkkhaki': [
          189,
          183,
          107,
          1
      ],
      'darkmagenta': [
          139,
          0,
          139,
          1
      ],
      'darkolivegreen': [
          85,
          107,
          47,
          1
      ],
      'darkorange': [
          255,
          140,
          0,
          1
      ],
      'darkorchid': [
          153,
          50,
          204,
          1
      ],
      'darkred': [
          139,
          0,
          0,
          1
      ],
      'darksalmon': [
          233,
          150,
          122,
          1
      ],
      'darkseagreen': [
          143,
          188,
          143,
          1
      ],
      'darkslateblue': [
          72,
          61,
          139,
          1
      ],
      'darkslategray': [
          47,
          79,
          79,
          1
      ],
      'darkslategrey': [
          47,
          79,
          79,
          1
      ],
      'darkturquoise': [
          0,
          206,
          209,
          1
      ],
      'darkviolet': [
          148,
          0,
          211,
          1
      ],
      'deeppink': [
          255,
          20,
          147,
          1
      ],
      'deepskyblue': [
          0,
          191,
          255,
          1
      ],
      'dimgray': [
          105,
          105,
          105,
          1
      ],
      'dimgrey': [
          105,
          105,
          105,
          1
      ],
      'dodgerblue': [
          30,
          144,
          255,
          1
      ],
      'firebrick': [
          178,
          34,
          34,
          1
      ],
      'floralwhite': [
          255,
          250,
          240,
          1
      ],
      'forestgreen': [
          34,
          139,
          34,
          1
      ],
      'fuchsia': [
          255,
          0,
          255,
          1
      ],
      'gainsboro': [
          220,
          220,
          220,
          1
      ],
      'ghostwhite': [
          248,
          248,
          255,
          1
      ],
      'gold': [
          255,
          215,
          0,
          1
      ],
      'goldenrod': [
          218,
          165,
          32,
          1
      ],
      'gray': [
          128,
          128,
          128,
          1
      ],
      'green': [
          0,
          128,
          0,
          1
      ],
      'greenyellow': [
          173,
          255,
          47,
          1
      ],
      'grey': [
          128,
          128,
          128,
          1
      ],
      'honeydew': [
          240,
          255,
          240,
          1
      ],
      'hotpink': [
          255,
          105,
          180,
          1
      ],
      'indianred': [
          205,
          92,
          92,
          1
      ],
      'indigo': [
          75,
          0,
          130,
          1
      ],
      'ivory': [
          255,
          255,
          240,
          1
      ],
      'khaki': [
          240,
          230,
          140,
          1
      ],
      'lavender': [
          230,
          230,
          250,
          1
      ],
      'lavenderblush': [
          255,
          240,
          245,
          1
      ],
      'lawngreen': [
          124,
          252,
          0,
          1
      ],
      'lemonchiffon': [
          255,
          250,
          205,
          1
      ],
      'lightblue': [
          173,
          216,
          230,
          1
      ],
      'lightcoral': [
          240,
          128,
          128,
          1
      ],
      'lightcyan': [
          224,
          255,
          255,
          1
      ],
      'lightgoldenrodyellow': [
          250,
          250,
          210,
          1
      ],
      'lightgray': [
          211,
          211,
          211,
          1
      ],
      'lightgreen': [
          144,
          238,
          144,
          1
      ],
      'lightgrey': [
          211,
          211,
          211,
          1
      ],
      'lightpink': [
          255,
          182,
          193,
          1
      ],
      'lightsalmon': [
          255,
          160,
          122,
          1
      ],
      'lightseagreen': [
          32,
          178,
          170,
          1
      ],
      'lightskyblue': [
          135,
          206,
          250,
          1
      ],
      'lightslategray': [
          119,
          136,
          153,
          1
      ],
      'lightslategrey': [
          119,
          136,
          153,
          1
      ],
      'lightsteelblue': [
          176,
          196,
          222,
          1
      ],
      'lightyellow': [
          255,
          255,
          224,
          1
      ],
      'lime': [
          0,
          255,
          0,
          1
      ],
      'limegreen': [
          50,
          205,
          50,
          1
      ],
      'linen': [
          250,
          240,
          230,
          1
      ],
      'magenta': [
          255,
          0,
          255,
          1
      ],
      'maroon': [
          128,
          0,
          0,
          1
      ],
      'mediumaquamarine': [
          102,
          205,
          170,
          1
      ],
      'mediumblue': [
          0,
          0,
          205,
          1
      ],
      'mediumorchid': [
          186,
          85,
          211,
          1
      ],
      'mediumpurple': [
          147,
          112,
          219,
          1
      ],
      'mediumseagreen': [
          60,
          179,
          113,
          1
      ],
      'mediumslateblue': [
          123,
          104,
          238,
          1
      ],
      'mediumspringgreen': [
          0,
          250,
          154,
          1
      ],
      'mediumturquoise': [
          72,
          209,
          204,
          1
      ],
      'mediumvioletred': [
          199,
          21,
          133,
          1
      ],
      'midnightblue': [
          25,
          25,
          112,
          1
      ],
      'mintcream': [
          245,
          255,
          250,
          1
      ],
      'mistyrose': [
          255,
          228,
          225,
          1
      ],
      'moccasin': [
          255,
          228,
          181,
          1
      ],
      'navajowhite': [
          255,
          222,
          173,
          1
      ],
      'navy': [
          0,
          0,
          128,
          1
      ],
      'oldlace': [
          253,
          245,
          230,
          1
      ],
      'olive': [
          128,
          128,
          0,
          1
      ],
      'olivedrab': [
          107,
          142,
          35,
          1
      ],
      'orange': [
          255,
          165,
          0,
          1
      ],
      'orangered': [
          255,
          69,
          0,
          1
      ],
      'orchid': [
          218,
          112,
          214,
          1
      ],
      'palegoldenrod': [
          238,
          232,
          170,
          1
      ],
      'palegreen': [
          152,
          251,
          152,
          1
      ],
      'paleturquoise': [
          175,
          238,
          238,
          1
      ],
      'palevioletred': [
          219,
          112,
          147,
          1
      ],
      'papayawhip': [
          255,
          239,
          213,
          1
      ],
      'peachpuff': [
          255,
          218,
          185,
          1
      ],
      'peru': [
          205,
          133,
          63,
          1
      ],
      'pink': [
          255,
          192,
          203,
          1
      ],
      'plum': [
          221,
          160,
          221,
          1
      ],
      'powderblue': [
          176,
          224,
          230,
          1
      ],
      'purple': [
          128,
          0,
          128,
          1
      ],
      'rebeccapurple': [
          102,
          51,
          153,
          1
      ],
      'red': [
          255,
          0,
          0,
          1
      ],
      'rosybrown': [
          188,
          143,
          143,
          1
      ],
      'royalblue': [
          65,
          105,
          225,
          1
      ],
      'saddlebrown': [
          139,
          69,
          19,
          1
      ],
      'salmon': [
          250,
          128,
          114,
          1
      ],
      'sandybrown': [
          244,
          164,
          96,
          1
      ],
      'seagreen': [
          46,
          139,
          87,
          1
      ],
      'seashell': [
          255,
          245,
          238,
          1
      ],
      'sienna': [
          160,
          82,
          45,
          1
      ],
      'silver': [
          192,
          192,
          192,
          1
      ],
      'skyblue': [
          135,
          206,
          235,
          1
      ],
      'slateblue': [
          106,
          90,
          205,
          1
      ],
      'slategray': [
          112,
          128,
          144,
          1
      ],
      'slategrey': [
          112,
          128,
          144,
          1
      ],
      'snow': [
          255,
          250,
          250,
          1
      ],
      'springgreen': [
          0,
          255,
          127,
          1
      ],
      'steelblue': [
          70,
          130,
          180,
          1
      ],
      'tan': [
          210,
          180,
          140,
          1
      ],
      'teal': [
          0,
          128,
          128,
          1
      ],
      'thistle': [
          216,
          191,
          216,
          1
      ],
      'tomato': [
          255,
          99,
          71,
          1
      ],
      'turquoise': [
          64,
          224,
          208,
          1
      ],
      'violet': [
          238,
          130,
          238,
          1
      ],
      'wheat': [
          245,
          222,
          179,
          1
      ],
      'white': [
          255,
          255,
          255,
          1
      ],
      'whitesmoke': [
          245,
          245,
          245,
          1
      ],
      'yellow': [
          255,
          255,
          0,
          1
      ],
      'yellowgreen': [
          154,
          205,
          50,
          1
      ]
  };
  function clamp_css_byte(i) {
      // Clamp to integer 0 .. 255.
      i = Math.round(i);
      // Seems to be what Chrome does (vs truncation).
      return i < 0 ? 0 : i > 255 ? 255 : i;
  }
  function clamp_css_float(f) {
      // Clamp to float 0.0 .. 1.0.
      return f < 0 ? 0 : f > 1 ? 1 : f;
  }
  function parse_css_int(str) {
      // int or percentage.
      if (str[str.length - 1] === '%')
          { return clamp_css_byte(parseFloat(str) / 100 * 255); }
      return clamp_css_byte(parseInt(str));
  }
  function parse_css_float(str) {
      // float or percentage.
      if (str[str.length - 1] === '%')
          { return clamp_css_float(parseFloat(str) / 100); }
      return clamp_css_float(parseFloat(str));
  }
  function css_hue_to_rgb(m1, m2, h) {
      if (h < 0)
          { h += 1; }
      else if (h > 1)
          { h -= 1; }
      if (h * 6 < 1)
          { return m1 + (m2 - m1) * h * 6; }
      if (h * 2 < 1)
          { return m2; }
      if (h * 3 < 2)
          { return m1 + (m2 - m1) * (2 / 3 - h) * 6; }
      return m1;
  }
  function parseCSSColor(css_str) {
      // Remove all whitespace, not compliant, but should just be more accepting.
      var str = css_str.replace(/ /g, '').toLowerCase();
      // Color keywords (and transparent) lookup.
      if (str in kCSSColorTable)
          { return kCSSColorTable[str].slice(); }
      // dup.
      // #abc and #abc123 syntax.
      if (str[0] === '#') {
          if (str.length === 4) {
              var iv = parseInt(str.substr(1), 16);
              // TODO(deanm): Stricter parsing.
              if (!(iv >= 0 && iv <= 4095))
                  { return null; }
              // Covers NaN.
              return [
                  (iv & 3840) >> 4 | (iv & 3840) >> 8,
                  iv & 240 | (iv & 240) >> 4,
                  iv & 15 | (iv & 15) << 4,
                  1
              ];
          } else if (str.length === 7) {
              var iv = parseInt(str.substr(1), 16);
              // TODO(deanm): Stricter parsing.
              if (!(iv >= 0 && iv <= 16777215))
                  { return null; }
              // Covers NaN.
              return [
                  (iv & 16711680) >> 16,
                  (iv & 65280) >> 8,
                  iv & 255,
                  1
              ];
          }
          return null;
      }
      var op = str.indexOf('('), ep = str.indexOf(')');
      if (op !== -1 && ep + 1 === str.length) {
          var fname = str.substr(0, op);
          var params = str.substr(op + 1, ep - (op + 1)).split(',');
          var alpha = 1;
          // To allow case fallthrough.
          switch (fname) {
          case 'rgba':
              if (params.length !== 4)
                  { return null; }
              alpha = parse_css_float(params.pop());
          // Fall through.
          case 'rgb':
              if (params.length !== 3)
                  { return null; }
              return [
                  parse_css_int(params[0]),
                  parse_css_int(params[1]),
                  parse_css_int(params[2]),
                  alpha
              ];
          case 'hsla':
              if (params.length !== 4)
                  { return null; }
              alpha = parse_css_float(params.pop());
          // Fall through.
          case 'hsl':
              if (params.length !== 3)
                  { return null; }
              var h = (parseFloat(params[0]) % 360 + 360) % 360 / 360;
              // 0 .. 1
              // NOTE(deanm): According to the CSS spec s/l should only be
              // percentages, but we don't bother and let float or percentage.
              var s = parse_css_float(params[1]);
              var l = parse_css_float(params[2]);
              var m2 = l <= 0.5 ? l * (s + 1) : l + s - l * s;
              var m1 = l * 2 - m2;
              return [
                  clamp_css_byte(css_hue_to_rgb(m1, m2, h + 1 / 3) * 255),
                  clamp_css_byte(css_hue_to_rgb(m1, m2, h) * 255),
                  clamp_css_byte(css_hue_to_rgb(m1, m2, h - 1 / 3) * 255),
                  alpha
              ];
          default:
              return null;
          }
      }
      return null;
  }
  try {
      parseCSSColor_1 = csscolorparser.parseCSSColor = parseCSSColor;
  } catch (e) {
  }

  //      
  /**
   * An RGBA color value. Create instances from color strings using the static
   * method `Color.parse`. The constructor accepts RGB channel values in the range
   * `[0, 1]`, premultiplied by A.
   *
   * @param {number} r The red channel.
   * @param {number} g The green channel.
   * @param {number} b The blue channel.
   * @param {number} a The alpha channel.
   * @private
   */
  var Color = function Color(r, g, b, a) {
      if ( a === void 0 ) a = 1;

      this.r = r;
      this.g = g;
      this.b = b;
      this.a = a;
  };
  /**
   * Parses valid CSS color strings and returns a `Color` instance.
   * @returns A `Color` instance, or `undefined` if the input is not a valid color string.
   */
  Color.parse = function parse (input) {
      if (!input) {
          return undefined;
      }
      if (input instanceof Color) {
          return input;
      }
      if (typeof input !== 'string') {
          return undefined;
      }
      var rgba = parseCSSColor_1(input);
      if (!rgba) {
          return undefined;
      }
      return new Color(rgba[0] / 255 * rgba[3], rgba[1] / 255 * rgba[3], rgba[2] / 255 * rgba[3], rgba[3]);
  };
  /**
   * Returns an RGBA string representing the color value.
   *
   * @returns An RGBA string.
   * @example
   * var purple = new Color.parse('purple');
   * purple.toString; // = "rgba(128,0,128,1)"
   * var translucentGreen = new Color.parse('rgba(26, 207, 26, .73)');
   * translucentGreen.toString(); // = "rgba(26,207,26,0.73)"
   */
  Color.prototype.toString = function toString () {
      var ref = this.toArray();
          var r = ref[0];
          var g = ref[1];
          var b = ref[2];
          var a = ref[3];
      return ("rgba(" + (Math.round(r)) + "," + (Math.round(g)) + "," + (Math.round(b)) + "," + a + ")");
  };
  /**
   * Returns an RGBA array of values representing the color, unpremultiplied by A.
   *
   * @returns An array of RGBA color values in the range [0, 255].
   */
  Color.prototype.toArray = function toArray () {
      var ref = this;
          var r = ref.r;
          var g = ref.g;
          var b = ref.b;
          var a = ref.a;
      return a === 0 ? [
          0,
          0,
          0,
          0
      ] : [
          r * 255 / a,
          g * 255 / a,
          b * 255 / a,
          a
      ];
  };
  /**
   * Returns a RGBA array of float values representing the color, unpremultiplied by A.
   *
   * @returns An array of RGBA color values in the range [0, 1].
   */
  Color.prototype.toArray01 = function toArray01 () {
      var ref = this;
          var r = ref.r;
          var g = ref.g;
          var b = ref.b;
          var a = ref.a;
      return a === 0 ? [
          0,
          0,
          0,
          0
      ] : [
          r / a,
          g / a,
          b / a,
          a
      ];
  };
  /**
   * Returns an RGBA array of values representing the color, premultiplied by A.
   *
   * @returns An array of RGBA color values in the range [0, 1].
   */
  Color.prototype.toArray01PremultipliedAlpha = function toArray01PremultipliedAlpha () {
      var ref = this;
          var r = ref.r;
          var g = ref.g;
          var b = ref.b;
          var a = ref.a;
      return [
          r,
          g,
          b,
          a
      ];
  };
  Color.black = new Color(0, 0, 0, 1);
  Color.white = new Color(1, 1, 1, 1);
  Color.transparent = new Color(0, 0, 0, 0);
  Color.red = new Color(1, 0, 0, 1);
  Color.blue = new Color(0, 0, 1, 1);

  function convertLiteral(value) {
      return typeof value === 'object' ? [
          'literal',
          value
      ] : value;
  }
  function convertFunction(parameters, propertySpec) {
      var stops = parameters.stops;
      if (!stops) {
          // identity function
          return convertIdentityFunction(parameters, propertySpec);
      }
      var zoomAndFeatureDependent = stops && typeof stops[0][0] === 'object';
      var featureDependent = zoomAndFeatureDependent || parameters.property !== undefined;
      var zoomDependent = zoomAndFeatureDependent || !featureDependent;
      stops = stops.map(function (stop) {
          if (!featureDependent && propertySpec.tokens && typeof stop[1] === 'string') {
              return [
                  stop[0],
                  convertTokenString(stop[1])
              ];
          }
          return [
              stop[0],
              convertLiteral(stop[1])
          ];
      });
      if (zoomAndFeatureDependent) {
          return convertZoomAndPropertyFunction(parameters, propertySpec, stops);
      } else if (zoomDependent) {
          return convertZoomFunction(parameters, propertySpec, stops);
      } else {
          return convertPropertyFunction(parameters, propertySpec, stops);
      }
  }
  function convertIdentityFunction(parameters, propertySpec) {
      var get = [
          'get',
          parameters.property
      ];
      if (parameters.default === undefined) {
          // By default, expressions for string-valued properties get coerced. To preserve
          // legacy function semantics, insert an explicit assertion instead.
          return propertySpec.type === 'string' ? [
              'string',
              get
          ] : get;
      } else if (propertySpec.type === 'enum') {
          return [
              'match',
              get,
              Object.keys(propertySpec.values),
              get,
              parameters.default
          ];
      } else {
          var expression = [
              propertySpec.type === 'color' ? 'to-color' : propertySpec.type,
              get,
              convertLiteral(parameters.default)
          ];
          if (propertySpec.type === 'array') {
              expression.splice(1, 0, propertySpec.value, propertySpec.length || null);
          }
          return expression;
      }
  }
  function getInterpolateOperator(parameters) {
      switch (parameters.colorSpace) {
      case 'hcl':
          return 'interpolate-hcl';
      case 'lab':
          return 'interpolate-lab';
      default:
          return 'interpolate';
      }
  }
  function convertZoomAndPropertyFunction(parameters, propertySpec, stops) {
      var featureFunctionParameters = {};
      var featureFunctionStops = {};
      var zoomStops = [];
      for (var s = 0; s < stops.length; s++) {
          var stop = stops[s];
          var zoom = stop[0].zoom;
          if (featureFunctionParameters[zoom] === undefined) {
              featureFunctionParameters[zoom] = {
                  zoom: zoom,
                  type: parameters.type,
                  property: parameters.property,
                  default: parameters.default
              };
              featureFunctionStops[zoom] = [];
              zoomStops.push(zoom);
          }
          featureFunctionStops[zoom].push([
              stop[0].value,
              stop[1]
          ]);
      }
      // the interpolation type for the zoom dimension of a zoom-and-property
      // function is determined directly from the style property specification
      // for which it's being used: linear for interpolatable properties, step
      // otherwise.
      var functionType = getFunctionType({}, propertySpec);
      if (functionType === 'exponential') {
          var expression = [
              getInterpolateOperator(parameters),
              ['linear'],
              ['zoom']
          ];
          for (var i = 0, list = zoomStops; i < list.length; i += 1) {
              var z = list[i];

              var output = convertPropertyFunction(featureFunctionParameters[z], propertySpec, featureFunctionStops[z]);
              appendStopPair(expression, z, output, false);
          }
          return expression;
      } else {
          var expression$1 = [
              'step',
              ['zoom']
          ];
          for (var i$1 = 0, list$1 = zoomStops; i$1 < list$1.length; i$1 += 1) {
              var z$1 = list$1[i$1];

              var output$1 = convertPropertyFunction(featureFunctionParameters[z$1], propertySpec, featureFunctionStops[z$1]);
              appendStopPair(expression$1, z$1, output$1, true);
          }
          fixupDegenerateStepCurve(expression$1);
          return expression$1;
      }
  }
  function coalesce(a, b) {
      if (a !== undefined)
          { return a; }
      if (b !== undefined)
          { return b; }
  }
  function getFallback(parameters, propertySpec) {
      var defaultValue = convertLiteral(coalesce(parameters.default, propertySpec.default));
      /*
       * Some fields with type: resolvedImage have an undefined default.
       * Because undefined is an invalid value for resolvedImage, set fallback to
       * an empty string instead of undefined to ensure output
       * passes validation.
       */
      if (defaultValue === undefined && propertySpec.type === 'resolvedImage') {
          return '';
      }
      return defaultValue;
  }
  function convertPropertyFunction(parameters, propertySpec, stops) {
      var type = getFunctionType(parameters, propertySpec);
      var get = [
          'get',
          parameters.property
      ];
      if (type === 'categorical' && typeof stops[0][0] === 'boolean') {
          var expression = ['case'];
          for (var i = 0, list = stops; i < list.length; i += 1) {
              var stop = list[i];

              expression.push([
                  '==',
                  get,
                  stop[0]
              ], stop[1]);
          }
          expression.push(getFallback(parameters, propertySpec));
          return expression;
      } else if (type === 'categorical') {
          var expression$1 = [
              'match',
              get
          ];
          for (var i$1 = 0, list$1 = stops; i$1 < list$1.length; i$1 += 1) {
              var stop$1 = list$1[i$1];

              appendStopPair(expression$1, stop$1[0], stop$1[1], false);
          }
          expression$1.push(getFallback(parameters, propertySpec));
          return expression$1;
      } else if (type === 'interval') {
          var expression$2 = [
              'step',
              [
                  'number',
                  get
              ]
          ];
          for (var i$2 = 0, list$2 = stops; i$2 < list$2.length; i$2 += 1) {
              var stop$2 = list$2[i$2];

              appendStopPair(expression$2, stop$2[0], stop$2[1], true);
          }
          fixupDegenerateStepCurve(expression$2);
          return parameters.default === undefined ? expression$2 : [
              'case',
              [
                  '==',
                  [
                      'typeof',
                      get
                  ],
                  'number'
              ],
              expression$2,
              convertLiteral(parameters.default)
          ];
      } else if (type === 'exponential') {
          var base = parameters.base !== undefined ? parameters.base : 1;
          var expression$3 = [
              getInterpolateOperator(parameters),
              base === 1 ? ['linear'] : [
                  'exponential',
                  base
              ],
              [
                  'number',
                  get
              ]
          ];
          for (var i$3 = 0, list$3 = stops; i$3 < list$3.length; i$3 += 1) {
              var stop$3 = list$3[i$3];

              appendStopPair(expression$3, stop$3[0], stop$3[1], false);
          }
          return parameters.default === undefined ? expression$3 : [
              'case',
              [
                  '==',
                  [
                      'typeof',
                      get
                  ],
                  'number'
              ],
              expression$3,
              convertLiteral(parameters.default)
          ];
      } else {
          throw new Error(("Unknown property function type " + type));
      }
  }
  function convertZoomFunction(parameters, propertySpec, stops, input) {
      if ( input === void 0 ) input = ['zoom'];

      var type = getFunctionType(parameters, propertySpec);
      var expression;
      var isStep = false;
      if (type === 'interval') {
          expression = [
              'step',
              input
          ];
          isStep = true;
      } else if (type === 'exponential') {
          var base = parameters.base !== undefined ? parameters.base : 1;
          expression = [
              getInterpolateOperator(parameters),
              base === 1 ? ['linear'] : [
                  'exponential',
                  base
              ],
              input
          ];
      } else {
          throw new Error(("Unknown zoom function type \"" + type + "\""));
      }
      for (var i = 0, list = stops; i < list.length; i += 1) {
          var stop = list[i];

          appendStopPair(expression, stop[0], stop[1], isStep);
      }
      fixupDegenerateStepCurve(expression);
      return expression;
  }
  function fixupDegenerateStepCurve(expression) {
      // degenerate step curve (i.e. a constant function): add a noop stop
      if (expression[0] === 'step' && expression.length === 3) {
          expression.push(0);
          expression.push(expression[3]);
      }
  }
  function appendStopPair(curve, input, output, isStep) {
      // Skip duplicate stop values. They were not validated for functions, but they are for expressions.
      // https://github.com/mapbox/mapbox-gl-js/issues/4107
      if (curve.length > 3 && input === curve[curve.length - 2]) {
          return;
      }
      // step curves don't get the first input value, as it is redundant.
      if (!(isStep && curve.length === 2)) {
          curve.push(input);
      }
      curve.push(output);
  }
  function getFunctionType(parameters, propertySpec) {
      if (parameters.type) {
          return parameters.type;
      } else {
          return propertySpec.expression.interpolated ? 'exponential' : 'interval';
      }
  }
  // "String with {name} token" => ["concat", "String with ", ["get", "name"], " token"]
  function convertTokenString(s) {
      var result = ['concat'];
      var re = /{([^{}]+)}/g;
      var pos = 0;
      for (var match = re.exec(s); match !== null; match = re.exec(s)) {
          var literal = s.slice(pos, re.lastIndex - match[0].length);
          pos = re.lastIndex;
          if (literal.length > 0)
              { result.push(literal); }
          result.push([
              'get',
              match[1]
          ]);
      }
      if (result.length === 1) {
          return s;
      }
      if (pos < s.length) {
          result.push(s.slice(pos));
      } else if (result.length === 2) {
          return [
              'to-string',
              result[1]
          ];
      }
      return result;
  }

  //      
  var ParsingError = /*@__PURE__*/(function (Error) {
      function ParsingError(key, message) {
          Error.call(this, message);
          this.message = message;
          this.key = key;
      }

      if ( Error ) ParsingError.__proto__ = Error;
      ParsingError.prototype = Object.create( Error && Error.prototype );
      ParsingError.prototype.constructor = ParsingError;

      return ParsingError;
  }(Error));

  //      
  /**
   * Tracks `let` bindings during expression parsing.
   * @private
   */
  var Scope = function Scope(parent, bindings) {
      if ( bindings === void 0 ) bindings = [];

      this.parent = parent;
      this.bindings = {};
      for (var i = 0, list = bindings; i < list.length; i += 1) {
          var ref = list[i];
          var name = ref[0];
          var expression = ref[1];

          this.bindings[name] = expression;
      }
  };
  Scope.prototype.concat = function concat (bindings) {
      return new Scope(this, bindings);
  };
  Scope.prototype.get = function get (name) {
      if (this.bindings[name]) {
          return this.bindings[name];
      }
      if (this.parent) {
          return this.parent.get(name);
      }
      throw new Error((name + " not found in scope."));
  };
  Scope.prototype.has = function has (name) {
      if (this.bindings[name])
          { return true; }
      return this.parent ? this.parent.has(name) : false;
  };

  //      
  var NullType = { kind: 'null' };
  var NumberType = { kind: 'number' };
  var StringType = { kind: 'string' };
  var BooleanType = { kind: 'boolean' };
  var ColorType = { kind: 'color' };
  var ObjectType = { kind: 'object' };
  var ValueType = { kind: 'value' };
  var ErrorType = { kind: 'error' };
  var CollatorType = { kind: 'collator' };
  var FormattedType = { kind: 'formatted' };
  var ResolvedImageType = { kind: 'resolvedImage' };
  function array$1(itemType, N) {
      return {
          kind: 'array',
          itemType: itemType,
          N: N
      };
  }
  function toString$1(type) {
      if (type.kind === 'array') {
          var itemType = toString$1(type.itemType);
          return typeof type.N === 'number' ? ("array<" + itemType + ", " + (type.N) + ">") : type.itemType.kind === 'value' ? 'array' : ("array<" + itemType + ">");
      } else {
          return type.kind;
      }
  }
  var valueMemberTypes = [
      NullType,
      NumberType,
      StringType,
      BooleanType,
      ColorType,
      FormattedType,
      ObjectType,
      array$1(ValueType),
      ResolvedImageType
  ];
  /**
   * Returns null if `t` is a subtype of `expected`; otherwise returns an
   * error message.
   * @private
   */
  function checkSubtype(expected, t) {
      if (t.kind === 'error') {
          // Error is a subtype of every type
          return null;
      } else if (expected.kind === 'array') {
          if (t.kind === 'array' && (t.N === 0 && t.itemType.kind === 'value' || !checkSubtype(expected.itemType, t.itemType)) && (typeof expected.N !== 'number' || expected.N === t.N)) {
              return null;
          }
      } else if (expected.kind === t.kind) {
          return null;
      } else if (expected.kind === 'value') {
          for (var i = 0, list = valueMemberTypes; i < list.length; i += 1) {
              var memberType = list[i];

              if (!checkSubtype(memberType, t)) {
                  return null;
              }
          }
      }
      return ("Expected " + (toString$1(expected)) + " but found " + (toString$1(t)) + " instead.");
  }
  function isValidType(provided, allowedTypes) {
      return allowedTypes.some(function (t) { return t.kind === provided.kind; });
  }
  function isValidNativeType(provided, allowedTypes) {
      return allowedTypes.some(function (t) {
          if (t === 'null') {
              return provided === null;
          } else if (t === 'array') {
              return Array.isArray(provided);
          } else if (t === 'object') {
              return provided && !Array.isArray(provided) && typeof provided === 'object';
          } else {
              return t === typeof provided;
          }
      });
  }

  //      
  // Flow type declarations for Intl cribbed from
  // https://github.com/facebook/flow/issues/1270
  var Collator = function Collator(caseSensitive, diacriticSensitive, locale) {
      if (caseSensitive)
          { this.sensitivity = diacriticSensitive ? 'variant' : 'case'; }
      else
          { this.sensitivity = diacriticSensitive ? 'accent' : 'base'; }
      this.locale = locale;
      this.collator = new Intl.Collator(this.locale ? this.locale : [], {
          sensitivity: this.sensitivity,
          usage: 'search'
      });
  };
  Collator.prototype.compare = function compare (lhs, rhs) {
      return this.collator.compare(lhs, rhs);
  };
  Collator.prototype.resolvedLocale = function resolvedLocale () {
      // We create a Collator without "usage: search" because we don't want
      // the search options encoded in our result (e.g. "en-u-co-search")
      return new Intl.Collator(this.locale ? this.locale : []).resolvedOptions().locale;
  };

  //      
  var FormattedSection = function FormattedSection(text, image, scale, fontStack, textColor) {
      // combine characters so that diacritic marks are not separate code points
      this.text = text.normalize ? text.normalize() : text;
      this.image = image;
      this.scale = scale;
      this.fontStack = fontStack;
      this.textColor = textColor;
  };
  var Formatted = function Formatted(sections) {
      this.sections = sections;
  };
  Formatted.fromString = function fromString (unformatted) {
      return new Formatted([new FormattedSection(unformatted, null, null, null, null)]);
  };
  Formatted.prototype.isEmpty = function isEmpty () {
      if (this.sections.length === 0)
          { return true; }
      return !this.sections.some(function (section) { return section.text.length !== 0 || section.image && section.image.name.length !== 0; });
  };
  Formatted.factory = function factory (text) {
      if (text instanceof Formatted) {
          return text;
      } else {
          return Formatted.fromString(text);
      }
  };
  Formatted.prototype.toString = function toString () {
      if (this.sections.length === 0)
          { return ''; }
      return this.sections.map(function (section) { return section.text; }).join('');
  };
  Formatted.prototype.serialize = function serialize () {
      var serialized = ['format'];
      for (var i = 0, list = this.sections; i < list.length; i += 1) {
          var section = list[i];

              if (section.image) {
              serialized.push([
                  'image',
                  section.image.name
              ]);
              continue;
          }
          serialized.push(section.text);
          var options = {};
          if (section.fontStack) {
              options['text-font'] = [
                  'literal',
                  section.fontStack.split(',')
              ];
          }
          if (section.scale) {
              options['font-scale'] = section.scale;
          }
          if (section.textColor) {
              options['text-color'] = ['rgba'].concat(section.textColor.toArray());
          }
          serialized.push(options);
      }
      return serialized;
  };

  //      
  var ResolvedImage = function ResolvedImage(options) {
      this.name = options.name;
      this.available = options.available;
  };
  ResolvedImage.prototype.toString = function toString () {
      return this.name;
  };
  ResolvedImage.fromString = function fromString (name) {
      if (!name)
          { return null; }
      // treat empty values as no image
      return new ResolvedImage({
          name: name,
          available: false
      });
  };
  ResolvedImage.prototype.serialize = function serialize () {
      return [
          'image',
          this.name
      ];
  };

  function validateRGBA(r, g, b, a) {
      if (!(typeof r === 'number' && r >= 0 && r <= 255 && typeof g === 'number' && g >= 0 && g <= 255 && typeof b === 'number' && b >= 0 && b <= 255)) {
          var value = typeof a === 'number' ? [
              r,
              g,
              b,
              a
          ] : [
              r,
              g,
              b
          ];
          return ("Invalid rgba value [" + (value.join(', ')) + "]: 'r', 'g', and 'b' must be between 0 and 255.");
      }
      if (!(typeof a === 'undefined' || typeof a === 'number' && a >= 0 && a <= 1)) {
          return ("Invalid rgba value [" + ([
              r,
              g,
              b,
              a
          ].join(', ')) + "]: 'a' must be between 0 and 1.");
      }
      return null;
  }
  function isValue(mixed) {
      if (mixed === null) {
          return true;
      } else if (typeof mixed === 'string') {
          return true;
      } else if (typeof mixed === 'boolean') {
          return true;
      } else if (typeof mixed === 'number') {
          return true;
      } else if (mixed instanceof Color) {
          return true;
      } else if (mixed instanceof Collator) {
          return true;
      } else if (mixed instanceof Formatted) {
          return true;
      } else if (mixed instanceof ResolvedImage) {
          return true;
      } else if (Array.isArray(mixed)) {
          for (var i = 0, list = mixed; i < list.length; i += 1) {
              var item = list[i];

              if (!isValue(item)) {
                  return false;
              }
          }
          return true;
      } else if (typeof mixed === 'object') {
          for (var key in mixed) {
              if (!isValue(mixed[key])) {
                  return false;
              }
          }
          return true;
      } else {
          return false;
      }
  }
  function typeOf(value) {
      if (value === null) {
          return NullType;
      } else if (typeof value === 'string') {
          return StringType;
      } else if (typeof value === 'boolean') {
          return BooleanType;
      } else if (typeof value === 'number') {
          return NumberType;
      } else if (value instanceof Color) {
          return ColorType;
      } else if (value instanceof Collator) {
          return CollatorType;
      } else if (value instanceof Formatted) {
          return FormattedType;
      } else if (value instanceof ResolvedImage) {
          return ResolvedImageType;
      } else if (Array.isArray(value)) {
          var length = value.length;
          var itemType;
          for (var i = 0, list = value; i < list.length; i += 1) {
              var item = list[i];

              var t = typeOf(item);
              if (!itemType) {
                  itemType = t;
              } else if (itemType === t) {
                  continue;
              } else {
                  itemType = ValueType;
                  break;
              }
          }
          return array$1(itemType || ValueType, length);
      } else {
          return ObjectType;
      }
  }
  function toString(value) {
      var type = typeof value;
      if (value === null) {
          return '';
      } else if (type === 'string' || type === 'number' || type === 'boolean') {
          return String(value);
      } else if (value instanceof Color || value instanceof Formatted || value instanceof ResolvedImage) {
          return value.toString();
      } else {
          return JSON.stringify(value);
      }
  }

  var Literal = function Literal(type, value) {
      this.type = type;
      this.value = value;
  };
  Literal.parse = function parse (args, context) {
      if (args.length !== 2)
          { return context.error(("'literal' expression requires exactly one argument, but found " + (args.length - 1) + " instead.")); }
      if (!isValue(args[1]))
          { return context.error("invalid value"); }
      var value = args[1];
      var type = typeOf(value);
      // special case: infer the item type if possible for zero-length arrays
      var expected = context.expectedType;
      if (type.kind === 'array' && type.N === 0 && expected && expected.kind === 'array' && (typeof expected.N !== 'number' || expected.N === 0)) {
          type = expected;
      }
      return new Literal(type, value);
  };
  Literal.prototype.evaluate = function evaluate () {
      return this.value;
  };
  Literal.prototype.eachChild = function eachChild () {
  };
  Literal.prototype.outputDefined = function outputDefined () {
      return true;
  };
  Literal.prototype.serialize = function serialize () {
      if (this.type.kind === 'array' || this.type.kind === 'object') {
          return [
              'literal',
              this.value
          ];
      } else if (this.value instanceof Color) {
          // Constant-folding can generate Literal expressions that you
          // couldn't actually generate with a "literal" expression,
          // so we have to implement an equivalent serialization here
          return ['rgba'].concat(this.value.toArray());
      } else if (this.value instanceof Formatted) {
          // Same as Color
          return this.value.serialize();
      } else {
          return this.value;
      }
  };

  //      
  var RuntimeError = function RuntimeError(message) {
      this.name = 'ExpressionEvaluationError';
      this.message = message;
  };
  RuntimeError.prototype.toJSON = function toJSON () {
      return this.message;
  };

  var types$2 = {
      string: StringType,
      number: NumberType,
      boolean: BooleanType,
      object: ObjectType
  };
  var Assertion = function Assertion(type, args) {
      this.type = type;
      this.args = args;
  };
  Assertion.parse = function parse (args, context) {
      if (args.length < 2)
          { return context.error("Expected at least one argument."); }
      var i = 1;
      var type;
      var name = args[0];
      if (name === 'array') {
          var itemType;
          if (args.length > 2) {
              var type$1 = args[1];
              if (typeof type$1 !== 'string' || !(type$1 in types$2) || type$1 === 'object')
                  { return context.error('The item type argument of "array" must be one of string, number, boolean', 1); }
              itemType = types$2[type$1];
              i++;
          } else {
              itemType = ValueType;
          }
          var N;
          if (args.length > 3) {
              if (args[2] !== null && (typeof args[2] !== 'number' || args[2] < 0 || args[2] !== Math.floor(args[2]))) {
                  return context.error('The length argument to "array" must be a positive integer literal', 2);
              }
              N = args[2];
              i++;
          }
          type = array$1(itemType, N);
      } else {
          type = types$2[name];
      }
      var parsed = [];
      for (; i < args.length; i++) {
          var input = context.parse(args[i], i, ValueType);
          if (!input)
              { return null; }
          parsed.push(input);
      }
      return new Assertion(type, parsed);
  };
  Assertion.prototype.evaluate = function evaluate (ctx) {
      for (var i = 0; i < this.args.length; i++) {
          var value = this.args[i].evaluate(ctx);
          var error = checkSubtype(this.type, typeOf(value));
          if (!error) {
              return value;
          } else if (i === this.args.length - 1) {
              throw new RuntimeError(("Expected value to be of type " + (toString$1(this.type)) + ", but found " + (toString$1(typeOf(value))) + " instead."));
          }
      }
      return null;
  };
  Assertion.prototype.eachChild = function eachChild (fn) {
      this.args.forEach(fn);
  };
  Assertion.prototype.outputDefined = function outputDefined () {
      return this.args.every(function (arg) { return arg.outputDefined(); });
  };
  Assertion.prototype.serialize = function serialize () {
      var type = this.type;
      var serialized = [type.kind];
      if (type.kind === 'array') {
          var itemType = type.itemType;
          if (itemType.kind === 'string' || itemType.kind === 'number' || itemType.kind === 'boolean') {
              serialized.push(itemType.kind);
              var N = type.N;
              if (typeof N === 'number' || this.args.length > 1) {
                  serialized.push(N);
              }
          }
      }
      return serialized.concat(this.args.map(function (arg) { return arg.serialize(); }));
  };

  //      
  var FormatExpression = function FormatExpression(sections) {
      this.type = FormattedType;
      this.sections = sections;
  };
  FormatExpression.parse = function parse (args, context) {
      if (args.length < 2) {
          return context.error("Expected at least one argument.");
      }
      var firstArg = args[1];
      if (!Array.isArray(firstArg) && typeof firstArg === 'object') {
          return context.error("First argument must be an image or text section.");
      }
      var sections = [];
      var nextTokenMayBeObject = false;
      for (var i = 1; i <= args.length - 1; ++i) {
          var arg = args[i];
          if (nextTokenMayBeObject && typeof arg === 'object' && !Array.isArray(arg)) {
              nextTokenMayBeObject = false;
              var scale = null;
              if (arg['font-scale']) {
                  scale = context.parse(arg['font-scale'], 1, NumberType);
                  if (!scale)
                      { return null; }
              }
              var font = null;
              if (arg['text-font']) {
                  font = context.parse(arg['text-font'], 1, array$1(StringType));
                  if (!font)
                      { return null; }
              }
              var textColor = null;
              if (arg['text-color']) {
                  textColor = context.parse(arg['text-color'], 1, ColorType);
                  if (!textColor)
                      { return null; }
              }
              var lastExpression = sections[sections.length - 1];
              lastExpression.scale = scale;
              lastExpression.font = font;
              lastExpression.textColor = textColor;
          } else {
              var content = context.parse(args[i], 1, ValueType);
              if (!content)
                  { return null; }
              var kind = content.type.kind;
              if (kind !== 'string' && kind !== 'value' && kind !== 'null' && kind !== 'resolvedImage')
                  { return context.error("Formatted text type must be 'string', 'value', 'image' or 'null'."); }
              nextTokenMayBeObject = true;
              sections.push({
                  content: content,
                  scale: null,
                  font: null,
                  textColor: null
              });
          }
      }
      return new FormatExpression(sections);
  };
  FormatExpression.prototype.evaluate = function evaluate (ctx) {
      var evaluateSection = function (section) {
          var evaluatedContent = section.content.evaluate(ctx);
          if (typeOf(evaluatedContent) === ResolvedImageType) {
              return new FormattedSection('', evaluatedContent, null, null, null);
          }
          return new FormattedSection(toString(evaluatedContent), null, section.scale ? section.scale.evaluate(ctx) : null, section.font ? section.font.evaluate(ctx).join(',') : null, section.textColor ? section.textColor.evaluate(ctx) : null);
      };
      return new Formatted(this.sections.map(evaluateSection));
  };
  FormatExpression.prototype.eachChild = function eachChild (fn) {
      for (var i = 0, list = this.sections; i < list.length; i += 1) {
          var section = list[i];

              fn(section.content);
          if (section.scale) {
              fn(section.scale);
          }
          if (section.font) {
              fn(section.font);
          }
          if (section.textColor) {
              fn(section.textColor);
          }
      }
  };
  FormatExpression.prototype.outputDefined = function outputDefined () {
      // Technically the combinatoric set of all children
      // Usually, this.text will be undefined anyway
      return false;
  };
  FormatExpression.prototype.serialize = function serialize () {
      var serialized = ['format'];
      for (var i = 0, list = this.sections; i < list.length; i += 1) {
          var section = list[i];

              serialized.push(section.content.serialize());
          var options = {};
          if (section.scale) {
              options['font-scale'] = section.scale.serialize();
          }
          if (section.font) {
              options['text-font'] = section.font.serialize();
          }
          if (section.textColor) {
              options['text-color'] = section.textColor.serialize();
          }
          serialized.push(options);
      }
      return serialized;
  };

  //      
  var ImageExpression = function ImageExpression(input) {
      this.type = ResolvedImageType;
      this.input = input;
  };
  ImageExpression.parse = function parse (args, context) {
      if (args.length !== 2) {
          return context.error("Expected two arguments.");
      }
      var name = context.parse(args[1], 1, StringType);
      if (!name)
          { return context.error("No image name provided."); }
      return new ImageExpression(name);
  };
  ImageExpression.prototype.evaluate = function evaluate (ctx) {
      var evaluatedImageName = this.input.evaluate(ctx);
      var value = ResolvedImage.fromString(evaluatedImageName);
      if (value && ctx.availableImages)
          { value.available = ctx.availableImages.indexOf(evaluatedImageName) > -1; }
      return value;
  };
  ImageExpression.prototype.eachChild = function eachChild (fn) {
      fn(this.input);
  };
  ImageExpression.prototype.outputDefined = function outputDefined () {
      // The output of image is determined by the list of available images in the evaluation context
      return false;
  };
  ImageExpression.prototype.serialize = function serialize () {
      return [
          'image',
          this.input.serialize()
      ];
  };

  var types$1 = {
      'to-boolean': BooleanType,
      'to-color': ColorType,
      'to-number': NumberType,
      'to-string': StringType
  };
  /**
   * Special form for error-coalescing coercion expressions "to-number",
   * "to-color".  Since these coercions can fail at runtime, they accept multiple
   * arguments, only evaluating one at a time until one succeeds.
   *
   * @private
   */
  var Coercion = function Coercion(type, args) {
      this.type = type;
      this.args = args;
  };
  Coercion.parse = function parse (args, context) {
      if (args.length < 2)
          { return context.error("Expected at least one argument."); }
      var name = args[0];
      if ((name === 'to-boolean' || name === 'to-string') && args.length !== 2)
          { return context.error("Expected one argument."); }
      var type = types$1[name];
      var parsed = [];
      for (var i = 1; i < args.length; i++) {
          var input = context.parse(args[i], i, ValueType);
          if (!input)
              { return null; }
          parsed.push(input);
      }
      return new Coercion(type, parsed);
  };
  Coercion.prototype.evaluate = function evaluate (ctx) {
      if (this.type.kind === 'boolean') {
          return Boolean(this.args[0].evaluate(ctx));
      } else if (this.type.kind === 'color') {
          var input;
          var error;
          for (var i = 0, list = this.args; i < list.length; i += 1) {
              var arg = list[i];

                  input = arg.evaluate(ctx);
              error = null;
              if (input instanceof Color) {
                  return input;
              } else if (typeof input === 'string') {
                  var c = ctx.parseColor(input);
                  if (c)
                      { return c; }
              } else if (Array.isArray(input)) {
                  if (input.length < 3 || input.length > 4) {
                      error = "Invalid rbga value " + (JSON.stringify(input)) + ": expected an array containing either three or four numeric values.";
                  } else {
                      error = validateRGBA(input[0], input[1], input[2], input[3]);
                  }
                  if (!error) {
                      return new Color(input[0] / 255, input[1] / 255, input[2] / 255, input[3]);
                  }
              }
          }
          throw new RuntimeError(error || ("Could not parse color from value '" + (typeof input === 'string' ? input : String(JSON.stringify(input))) + "'"));
      } else if (this.type.kind === 'number') {
          var value = null;
          for (var i$1 = 0, list$1 = this.args; i$1 < list$1.length; i$1 += 1) {
              var arg$1 = list$1[i$1];

                  value = arg$1.evaluate(ctx);
              if (value === null)
                  { return 0; }
              var num = Number(value);
              if (isNaN(num))
                  { continue; }
              return num;
          }
          throw new RuntimeError(("Could not convert " + (JSON.stringify(value)) + " to number."));
      } else if (this.type.kind === 'formatted') {
          // There is no explicit 'to-formatted' but this coercion can be implicitly
          // created by properties that expect the 'formatted' type.
          return Formatted.fromString(toString(this.args[0].evaluate(ctx)));
      } else if (this.type.kind === 'resolvedImage') {
          return ResolvedImage.fromString(toString(this.args[0].evaluate(ctx)));
      } else {
          return toString(this.args[0].evaluate(ctx));
      }
  };
  Coercion.prototype.eachChild = function eachChild (fn) {
      this.args.forEach(fn);
  };
  Coercion.prototype.outputDefined = function outputDefined () {
      return this.args.every(function (arg) { return arg.outputDefined(); });
  };
  Coercion.prototype.serialize = function serialize () {
      if (this.type.kind === 'formatted') {
          return new FormatExpression([{
                  content: this.args[0],
                  scale: null,
                  font: null,
                  textColor: null
              }]).serialize();
      }
      if (this.type.kind === 'resolvedImage') {
          return new ImageExpression(this.args[0]).serialize();
      }
      var serialized = [("to-" + (this.type.kind))];
      this.eachChild(function (child) {
          serialized.push(child.serialize());
      });
      return serialized;
  };

  //      
  var geometryTypes = [
      'Unknown',
      'Point',
      'LineString',
      'Polygon'
  ];
  var EvaluationContext = function EvaluationContext() {
      this.globals = null;
      this.feature = null;
      this.featureState = null;
      this.formattedSection = null;
      this._parseColorCache = {};
      this.availableImages = null;
      this.canonical = null;
      this.featureTileCoord = null;
      this.featureDistanceData = null;
  };
  EvaluationContext.prototype.id = function id () {
      return this.feature && 'id' in this.feature && this.feature.id ? this.feature.id : null;
  };
  EvaluationContext.prototype.geometryType = function geometryType () {
      return this.feature ? typeof this.feature.type === 'number' ? geometryTypes[this.feature.type] : this.feature.type : null;
  };
  EvaluationContext.prototype.geometry = function geometry () {
      return this.feature && 'geometry' in this.feature ? this.feature.geometry : null;
  };
  EvaluationContext.prototype.canonicalID = function canonicalID () {
      return this.canonical;
  };
  EvaluationContext.prototype.properties = function properties () {
      return this.feature && this.feature.properties || {};
  };
  EvaluationContext.prototype.distanceFromCenter = function distanceFromCenter () {
      if (this.featureTileCoord && this.featureDistanceData) {
          var c = this.featureDistanceData.center;
          var scale = this.featureDistanceData.scale;
          var ref = this.featureTileCoord;
              var x = ref.x;
              var y = ref.y;
          // Calculate the distance vector `d` (left handed)
          var dX = x * scale - c[0];
          var dY = y * scale - c[1];
          // The bearing vector `b` (left handed)
          var bX = this.featureDistanceData.bearing[0];
          var bY = this.featureDistanceData.bearing[1];
          // Distance is calculated as `dot(d, v)`
          var dist = bX * dX + bY * dY;
          return dist;
      }
      return 0;
  };
  EvaluationContext.prototype.parseColor = function parseColor (input) {
      var cached = this._parseColorCache[input];
      if (!cached) {
          cached = this._parseColorCache[input] = Color.parse(input);
      }
      return cached;
  };

  //      
  var CompoundExpression = function CompoundExpression(name, type, evaluate, args) {
      this.name = name;
      this.type = type;
      this._evaluate = evaluate;
      this.args = args;
  };
  CompoundExpression.prototype.evaluate = function evaluate (ctx) {
      return this._evaluate(ctx, this.args);
  };
  CompoundExpression.prototype.eachChild = function eachChild (fn) {
      this.args.forEach(fn);
  };
  CompoundExpression.prototype.outputDefined = function outputDefined () {
      return false;
  };
  CompoundExpression.prototype.serialize = function serialize () {
      return [this.name].concat(this.args.map(function (arg) { return arg.serialize(); }));
  };
  CompoundExpression.parse = function parse (args, context) {
          var ref$1;

      var op = args[0];
      var definition = CompoundExpression.definitions[op];
      if (!definition) {
          return context.error(("Unknown expression \"" + op + "\". If you wanted a literal array, use [\"literal\", [...]]."), 0);
      }
      // Now check argument types against each signature
      var type = Array.isArray(definition) ? definition[0] : definition.type;
      var availableOverloads = Array.isArray(definition) ? [[
              definition[1],
              definition[2]
          ]] : definition.overloads;
      var overloads = availableOverloads.filter(function (ref) {
              var signature = ref[0];

              return !Array.isArray(signature) || // varags
      signature.length === args.length - 1;
          }// correct param count
  );
      var signatureContext = null;
      for (var i$3 = 0, list = overloads; i$3 < list.length; i$3 += 1) {
          // Use a fresh context for each attempted signature so that, if
          // we eventually succeed, we haven't polluted `context.errors`.
          var ref = list[i$3];
              var params = ref[0];
              var evaluate = ref[1];

              signatureContext = new ParsingContext(context.registry, context.path, null, context.scope);
          // First parse all the args, potentially coercing to the
          // types expected by this overload.
          var parsedArgs = [];
          var argParseFailed = false;
          for (var i = 1; i < args.length; i++) {
              var arg = args[i];
              var expectedType = Array.isArray(params) ? params[i - 1] : params.type;
              var parsed = signatureContext.parse(arg, 1 + parsedArgs.length, expectedType);
              if (!parsed) {
                  argParseFailed = true;
                  break;
              }
              parsedArgs.push(parsed);
          }
          if (argParseFailed) {
              // Couldn't coerce args of this overload to expected type, move
              // on to next one.
              continue;
          }
          if (Array.isArray(params)) {
              if (params.length !== parsedArgs.length) {
                  signatureContext.error(("Expected " + (params.length) + " arguments, but found " + (parsedArgs.length) + " instead."));
                  continue;
              }
          }
          for (var i$1 = 0; i$1 < parsedArgs.length; i$1++) {
              var expected = Array.isArray(params) ? params[i$1] : params.type;
              var arg$1 = parsedArgs[i$1];
              signatureContext.concat(i$1 + 1).checkSubtype(expected, arg$1.type);
          }
          if (signatureContext.errors.length === 0) {
              return new CompoundExpression(op, type, evaluate, parsedArgs);
          }
      }
      if (overloads.length === 1) {
          (ref$1 = context.errors).push.apply(ref$1, signatureContext.errors);
      } else {
          var expected$1 = overloads.length ? overloads : availableOverloads;
          var signatures = expected$1.map(function (ref) {
                  var params = ref[0];

                  return stringifySignature(params);
              }).join(' | ');
          var actualTypes = [];
          // For error message, re-parse arguments without trying to
          // apply any coercions
          for (var i$2 = 1; i$2 < args.length; i$2++) {
              var parsed$1 = context.parse(args[i$2], 1 + actualTypes.length);
              if (!parsed$1)
                  { return null; }
              actualTypes.push(toString$1(parsed$1.type));
          }
          context.error(("Expected arguments of type " + signatures + ", but found (" + (actualTypes.join(', ')) + ") instead."));
      }
      return null;
  };
  CompoundExpression.register = function register (registry, definitions) {
      CompoundExpression.definitions = definitions;
      for (var name in definitions) {
          registry[name] = CompoundExpression;
      }
  };
  function stringifySignature(signature) {
      if (Array.isArray(signature)) {
          return ("(" + (signature.map(toString$1).join(', ')) + ")");
      } else {
          return ("(" + (toString$1(signature.type)) + "...)");
      }
  }

  //      
  var CollatorExpression = function CollatorExpression(caseSensitive, diacriticSensitive, locale) {
      this.type = CollatorType;
      this.locale = locale;
      this.caseSensitive = caseSensitive;
      this.diacriticSensitive = diacriticSensitive;
  };
  CollatorExpression.parse = function parse (args, context) {
      if (args.length !== 2)
          { return context.error("Expected one argument."); }
      var options = args[1];
      if (typeof options !== 'object' || Array.isArray(options))
          { return context.error("Collator options argument must be an object."); }
      var caseSensitive = context.parse(options['case-sensitive'] === undefined ? false : options['case-sensitive'], 1, BooleanType);
      if (!caseSensitive)
          { return null; }
      var diacriticSensitive = context.parse(options['diacritic-sensitive'] === undefined ? false : options['diacritic-sensitive'], 1, BooleanType);
      if (!diacriticSensitive)
          { return null; }
      var locale = null;
      if (options['locale']) {
          locale = context.parse(options['locale'], 1, StringType);
          if (!locale)
              { return null; }
      }
      return new CollatorExpression(caseSensitive, diacriticSensitive, locale);
  };
  CollatorExpression.prototype.evaluate = function evaluate (ctx) {
      return new Collator(this.caseSensitive.evaluate(ctx), this.diacriticSensitive.evaluate(ctx), this.locale ? this.locale.evaluate(ctx) : null);
  };
  CollatorExpression.prototype.eachChild = function eachChild (fn) {
      fn(this.caseSensitive);
      fn(this.diacriticSensitive);
      if (this.locale) {
          fn(this.locale);
      }
  };
  CollatorExpression.prototype.outputDefined = function outputDefined () {
      // Technically the set of possible outputs is the combinatoric set of Collators produced
      // by all possible outputs of locale/caseSensitive/diacriticSensitive
      // But for the primary use of Collators in comparison operators, we ignore the Collator's
      // possible outputs anyway, so we can get away with leaving this false for now.
      return false;
  };
  CollatorExpression.prototype.serialize = function serialize () {
      var options = {};
      options['case-sensitive'] = this.caseSensitive.serialize();
      options['diacritic-sensitive'] = this.diacriticSensitive.serialize();
      if (this.locale) {
          options['locale'] = this.locale.serialize();
      }
      return [
          'collator',
          options
      ];
  };

  //      
  // minX, minY, maxX, maxY
  var EXTENT = 8192;
  function updateBBox(bbox, coord) {
      bbox[0] = Math.min(bbox[0], coord[0]);
      bbox[1] = Math.min(bbox[1], coord[1]);
      bbox[2] = Math.max(bbox[2], coord[0]);
      bbox[3] = Math.max(bbox[3], coord[1]);
  }
  function mercatorXfromLng(lng) {
      return (180 + lng) / 360;
  }
  function mercatorYfromLat(lat) {
      return (180 - 180 / Math.PI * Math.log(Math.tan(Math.PI / 4 + lat * Math.PI / 360))) / 360;
  }
  function boxWithinBox(bbox1, bbox2) {
      if (bbox1[0] <= bbox2[0])
          { return false; }
      if (bbox1[2] >= bbox2[2])
          { return false; }
      if (bbox1[1] <= bbox2[1])
          { return false; }
      if (bbox1[3] >= bbox2[3])
          { return false; }
      return true;
  }
  function getTileCoordinates(p, canonical) {
      var x = mercatorXfromLng(p[0]);
      var y = mercatorYfromLat(p[1]);
      var tilesAtZoom = Math.pow(2, canonical.z);
      return [
          Math.round(x * tilesAtZoom * EXTENT),
          Math.round(y * tilesAtZoom * EXTENT)
      ];
  }
  function onBoundary(p, p1, p2) {
      var x1 = p[0] - p1[0];
      var y1 = p[1] - p1[1];
      var x2 = p[0] - p2[0];
      var y2 = p[1] - p2[1];
      return x1 * y2 - x2 * y1 === 0 && x1 * x2 <= 0 && y1 * y2 <= 0;
  }
  function rayIntersect(p, p1, p2) {
      return p1[1] > p[1] !== p2[1] > p[1] && p[0] < (p2[0] - p1[0]) * (p[1] - p1[1]) / (p2[1] - p1[1]) + p1[0];
  }
  // ray casting algorithm for detecting if point is in polygon
  function pointWithinPolygon(point, rings) {
      var inside = false;
      for (var i = 0, len = rings.length; i < len; i++) {
          var ring = rings[i];
          for (var j = 0, len2 = ring.length; j < len2 - 1; j++) {
              if (onBoundary(point, ring[j], ring[j + 1]))
                  { return false; }
              if (rayIntersect(point, ring[j], ring[j + 1]))
                  { inside = !inside; }
          }
      }
      return inside;
  }
  function pointWithinPolygons(point, polygons) {
      for (var i = 0; i < polygons.length; i++) {
          if (pointWithinPolygon(point, polygons[i]))
              { return true; }
      }
      return false;
  }
  function perp(v1, v2) {
      return v1[0] * v2[1] - v1[1] * v2[0];
  }
  // check if p1 and p2 are in different sides of line segment q1->q2
  function twoSided(p1, p2, q1, q2) {
      // q1->p1 (x1, y1), q1->p2 (x2, y2), q1->q2 (x3, y3)
      var x1 = p1[0] - q1[0];
      var y1 = p1[1] - q1[1];
      var x2 = p2[0] - q1[0];
      var y2 = p2[1] - q1[1];
      var x3 = q2[0] - q1[0];
      var y3 = q2[1] - q1[1];
      var det1 = x1 * y3 - x3 * y1;
      var det2 = x2 * y3 - x3 * y2;
      if (det1 > 0 && det2 < 0 || det1 < 0 && det2 > 0)
          { return true; }
      return false;
  }
  // a, b are end points for line segment1, c and d are end points for line segment2
  function lineIntersectLine(a, b, c, d) {
      // check if two segments are parallel or not
      // precondition is end point a, b is inside polygon, if line a->b is
      // parallel to polygon edge c->d, then a->b won't intersect with c->d
      var vectorP = [
          b[0] - a[0],
          b[1] - a[1]
      ];
      var vectorQ = [
          d[0] - c[0],
          d[1] - c[1]
      ];
      if (perp(vectorQ, vectorP) === 0)
          { return false; }
      // If lines are intersecting with each other, the relative location should be:
      // a and b lie in different sides of segment c->d
      // c and d lie in different sides of segment a->b
      if (twoSided(a, b, c, d) && twoSided(c, d, a, b))
          { return true; }
      return false;
  }
  function lineIntersectPolygon(p1, p2, polygon) {
      for (var i = 0, list = polygon; i < list.length; i += 1) {
          // loop through every edge of the ring
          var ring = list[i];

          for (var j = 0; j < ring.length - 1; ++j) {
              if (lineIntersectLine(p1, p2, ring[j], ring[j + 1])) {
                  return true;
              }
          }
      }
      return false;
  }
  function lineStringWithinPolygon(line, polygon) {
      // First, check if geometry points of line segments are all inside polygon
      for (var i = 0; i < line.length; ++i) {
          if (!pointWithinPolygon(line[i], polygon)) {
              return false;
          }
      }
      // Second, check if there is line segment intersecting polygon edge
      for (var i$1 = 0; i$1 < line.length - 1; ++i$1) {
          if (lineIntersectPolygon(line[i$1], line[i$1 + 1], polygon)) {
              return false;
          }
      }
      return true;
  }
  function lineStringWithinPolygons(line, polygons) {
      for (var i = 0; i < polygons.length; i++) {
          if (lineStringWithinPolygon(line, polygons[i]))
              { return true; }
      }
      return false;
  }
  function getTilePolygon(coordinates, bbox, canonical) {
      var polygon = [];
      for (var i = 0; i < coordinates.length; i++) {
          var ring = [];
          for (var j = 0; j < coordinates[i].length; j++) {
              var coord = getTileCoordinates(coordinates[i][j], canonical);
              updateBBox(bbox, coord);
              ring.push(coord);
          }
          polygon.push(ring);
      }
      return polygon;
  }
  function getTilePolygons(coordinates, bbox, canonical) {
      var polygons = [];
      for (var i = 0; i < coordinates.length; i++) {
          var polygon = getTilePolygon(coordinates[i], bbox, canonical);
          polygons.push(polygon);
      }
      return polygons;
  }
  function updatePoint(p, bbox, polyBBox, worldSize) {
      if (p[0] < polyBBox[0] || p[0] > polyBBox[2]) {
          var halfWorldSize = worldSize * 0.5;
          var shift = p[0] - polyBBox[0] > halfWorldSize ? -worldSize : polyBBox[0] - p[0] > halfWorldSize ? worldSize : 0;
          if (shift === 0) {
              shift = p[0] - polyBBox[2] > halfWorldSize ? -worldSize : polyBBox[2] - p[0] > halfWorldSize ? worldSize : 0;
          }
          p[0] += shift;
      }
      updateBBox(bbox, p);
  }
  function resetBBox(bbox) {
      bbox[0] = bbox[1] = Infinity;
      bbox[2] = bbox[3] = -Infinity;
  }
  function getTilePoints(geometry, pointBBox, polyBBox, canonical) {
      var worldSize = Math.pow(2, canonical.z) * EXTENT;
      var shifts = [
          canonical.x * EXTENT,
          canonical.y * EXTENT
      ];
      var tilePoints = [];
      if (!geometry)
          { return tilePoints; }
      for (var i$1 = 0, list$1 = geometry; i$1 < list$1.length; i$1 += 1) {
          var points = list$1[i$1];

          for (var i = 0, list = points; i < list.length; i += 1) {
              var point = list[i];

              var p = [
                  point.x + shifts[0],
                  point.y + shifts[1]
              ];
              updatePoint(p, pointBBox, polyBBox, worldSize);
              tilePoints.push(p);
          }
      }
      return tilePoints;
  }
  function getTileLines(geometry, lineBBox, polyBBox, canonical) {
      var worldSize = Math.pow(2, canonical.z) * EXTENT;
      var shifts = [
          canonical.x * EXTENT,
          canonical.y * EXTENT
      ];
      var tileLines = [];
      if (!geometry)
          { return tileLines; }
      for (var i$1 = 0, list$1 = geometry; i$1 < list$1.length; i$1 += 1) {
          var line = list$1[i$1];

          var tileLine = [];
          for (var i = 0, list = line; i < list.length; i += 1) {
              var point = list[i];

              var p = [
                  point.x + shifts[0],
                  point.y + shifts[1]
              ];
              updateBBox(lineBBox, p);
              tileLine.push(p);
          }
          tileLines.push(tileLine);
      }
      if (lineBBox[2] - lineBBox[0] <= worldSize / 2) {
          resetBBox(lineBBox);
          for (var i$3 = 0, list$3 = tileLines; i$3 < list$3.length; i$3 += 1) {
              var line$1 = list$3[i$3];

              for (var i$2 = 0, list$2 = line$1; i$2 < list$2.length; i$2 += 1) {
                  var p$1 = list$2[i$2];

                  updatePoint(p$1, lineBBox, polyBBox, worldSize);
              }
          }
      }
      return tileLines;
  }
  function pointsWithinPolygons(ctx, polygonGeometry) {
      var pointBBox = [
          Infinity,
          Infinity,
          -Infinity,
          -Infinity
      ];
      var polyBBox = [
          Infinity,
          Infinity,
          -Infinity,
          -Infinity
      ];
      var canonical = ctx.canonicalID();
      if (!canonical) {
          return false;
      }
      if (polygonGeometry.type === 'Polygon') {
          var tilePolygon = getTilePolygon(polygonGeometry.coordinates, polyBBox, canonical);
          var tilePoints = getTilePoints(ctx.geometry(), pointBBox, polyBBox, canonical);
          if (!boxWithinBox(pointBBox, polyBBox))
              { return false; }
          for (var i = 0, list = tilePoints; i < list.length; i += 1) {
              var point = list[i];

              if (!pointWithinPolygon(point, tilePolygon))
                  { return false; }
          }
      }
      if (polygonGeometry.type === 'MultiPolygon') {
          var tilePolygons = getTilePolygons(polygonGeometry.coordinates, polyBBox, canonical);
          var tilePoints$1 = getTilePoints(ctx.geometry(), pointBBox, polyBBox, canonical);
          if (!boxWithinBox(pointBBox, polyBBox))
              { return false; }
          for (var i$1 = 0, list$1 = tilePoints$1; i$1 < list$1.length; i$1 += 1) {
              var point$1 = list$1[i$1];

              if (!pointWithinPolygons(point$1, tilePolygons))
                  { return false; }
          }
      }
      return true;
  }
  function linesWithinPolygons(ctx, polygonGeometry) {
      var lineBBox = [
          Infinity,
          Infinity,
          -Infinity,
          -Infinity
      ];
      var polyBBox = [
          Infinity,
          Infinity,
          -Infinity,
          -Infinity
      ];
      var canonical = ctx.canonicalID();
      if (!canonical) {
          return false;
      }
      if (polygonGeometry.type === 'Polygon') {
          var tilePolygon = getTilePolygon(polygonGeometry.coordinates, polyBBox, canonical);
          var tileLines = getTileLines(ctx.geometry(), lineBBox, polyBBox, canonical);
          if (!boxWithinBox(lineBBox, polyBBox))
              { return false; }
          for (var i = 0, list = tileLines; i < list.length; i += 1) {
              var line = list[i];

              if (!lineStringWithinPolygon(line, tilePolygon))
                  { return false; }
          }
      }
      if (polygonGeometry.type === 'MultiPolygon') {
          var tilePolygons = getTilePolygons(polygonGeometry.coordinates, polyBBox, canonical);
          var tileLines$1 = getTileLines(ctx.geometry(), lineBBox, polyBBox, canonical);
          if (!boxWithinBox(lineBBox, polyBBox))
              { return false; }
          for (var i$1 = 0, list$1 = tileLines$1; i$1 < list$1.length; i$1 += 1) {
              var line$1 = list$1[i$1];

              if (!lineStringWithinPolygons(line$1, tilePolygons))
                  { return false; }
          }
      }
      return true;
  }
  var Within$1 = function Within(geojson, geometries) {
      this.type = BooleanType;
      this.geojson = geojson;
      this.geometries = geometries;
  };
  Within$1.parse = function parse (args, context) {
      if (args.length !== 2)
          { return context.error(("'within' expression requires exactly one argument, but found " + (args.length - 1) + " instead.")); }
      if (isValue(args[1])) {
          var geojson = args[1];
          if (geojson.type === 'FeatureCollection') {
              for (var i = 0; i < geojson.features.length; ++i) {
                  var type = geojson.features[i].geometry.type;
                  if (type === 'Polygon' || type === 'MultiPolygon') {
                      return new Within$1(geojson, geojson.features[i].geometry);
                  }
              }
          } else if (geojson.type === 'Feature') {
              var type$1 = geojson.geometry.type;
              if (type$1 === 'Polygon' || type$1 === 'MultiPolygon') {
                  return new Within$1(geojson, geojson.geometry);
              }
          } else if (geojson.type === 'Polygon' || geojson.type === 'MultiPolygon') {
              return new Within$1(geojson, geojson);
          }
      }
      return context.error("'within' expression requires valid geojson object that contains polygon geometry type.");
  };
  Within$1.prototype.evaluate = function evaluate (ctx) {
      if (ctx.geometry() != null && ctx.canonicalID() != null) {
          if (ctx.geometryType() === 'Point') {
              return pointsWithinPolygons(ctx, this.geometries);
          } else if (ctx.geometryType() === 'LineString') {
              return linesWithinPolygons(ctx, this.geometries);
          }
      }
      return false;
  };
  Within$1.prototype.eachChild = function eachChild () {
  };
  Within$1.prototype.outputDefined = function outputDefined () {
      return true;
  };
  Within$1.prototype.serialize = function serialize () {
      return [
          'within',
          this.geojson
      ];
  };

  //      
  function isFeatureConstant(e) {
      if (e instanceof CompoundExpression) {
          if (e.name === 'get' && e.args.length === 1) {
              return false;
          } else if (e.name === 'feature-state') {
              return false;
          } else if (e.name === 'has' && e.args.length === 1) {
              return false;
          } else if (e.name === 'properties' || e.name === 'geometry-type' || e.name === 'id') {
              return false;
          } else if (/^filter-/.test(e.name)) {
              return false;
          }
      }
      if (e instanceof Within$1) {
          return false;
      }
      var result = true;
      e.eachChild(function (arg) {
          if (result && !isFeatureConstant(arg)) {
              result = false;
          }
      });
      return result;
  }
  function isStateConstant(e) {
      if (e instanceof CompoundExpression) {
          if (e.name === 'feature-state') {
              return false;
          }
      }
      var result = true;
      e.eachChild(function (arg) {
          if (result && !isStateConstant(arg)) {
              result = false;
          }
      });
      return result;
  }
  function isGlobalPropertyConstant(e, properties) {
      if (e instanceof CompoundExpression && properties.indexOf(e.name) >= 0) {
          return false;
      }
      var result = true;
      e.eachChild(function (arg) {
          if (result && !isGlobalPropertyConstant(arg, properties)) {
              result = false;
          }
      });
      return result;
  }

  //      
  var Var = function Var(name, boundExpression) {
      this.type = boundExpression.type;
      this.name = name;
      this.boundExpression = boundExpression;
  };
  Var.parse = function parse (args, context) {
      if (args.length !== 2 || typeof args[1] !== 'string')
          { return context.error("'var' expression requires exactly one string literal argument."); }
      var name = args[1];
      if (!context.scope.has(name)) {
          return context.error(("Unknown variable \"" + name + "\". Make sure \"" + name + "\" has been bound in an enclosing \"let\" expression before using it."), 1);
      }
      return new Var(name, context.scope.get(name));
  };
  Var.prototype.evaluate = function evaluate (ctx) {
      return this.boundExpression.evaluate(ctx);
  };
  Var.prototype.eachChild = function eachChild () {
  };
  Var.prototype.outputDefined = function outputDefined () {
      return false;
  };
  Var.prototype.serialize = function serialize () {
      return [
          'var',
          this.name
      ];
  };

  //      
  /**
   * State associated parsing at a given point in an expression tree.
   * @private
   */
  var ParsingContext = function ParsingContext(registry, path, expectedType, scope, errors) {
      if ( path === void 0 ) path = [];
      if ( scope === void 0 ) scope = new Scope();
      if ( errors === void 0 ) errors = [];

      this.registry = registry;
      this.path = path;
      this.key = path.map(function (part) { return ("[" + part + "]"); }).join('');
      this.scope = scope;
      this.errors = errors;
      this.expectedType = expectedType;
  };
  /**
   * @param expr the JSON expression to parse
   * @param index the optional argument index if this expression is an argument of a parent expression that's being parsed
   * @param options
   * @param options.omitTypeAnnotations set true to omit inferred type annotations.  Caller beware: with this option set, the parsed expression's type will NOT satisfy `expectedType` if it would normally be wrapped in an inferred annotation.
   * @private
   */
  ParsingContext.prototype.parse = function parse (expr, index, expectedType, bindings, options) {
          if ( options === void 0 ) options = {};

      if (index) {
          return this.concat(index, expectedType, bindings)._parse(expr, options);
      }
      return this._parse(expr, options);
  };
  ParsingContext.prototype._parse = function _parse (expr, options) {
      if (expr === null || typeof expr === 'string' || typeof expr === 'boolean' || typeof expr === 'number') {
          expr = [
              'literal',
              expr
          ];
      }
      function annotate(parsed, type, typeAnnotation) {
          if (typeAnnotation === 'assert') {
              return new Assertion(type, [parsed]);
          } else if (typeAnnotation === 'coerce') {
              return new Coercion(type, [parsed]);
          } else {
              return parsed;
          }
      }
      if (Array.isArray(expr)) {
          if (expr.length === 0) {
              return this.error("Expected an array with at least one element. If you wanted a literal array, use [\"literal\", []].");
          }
          var op = expr[0];
          if (typeof op !== 'string') {
              this.error(("Expression name must be a string, but found " + (typeof op) + " instead. If you wanted a literal array, use [\"literal\", [...]]."), 0);
              return null;
          }
          var Expr = this.registry[op];
          if (Expr) {
              var parsed = Expr.parse(expr, this);
              if (!parsed)
                  { return null; }
              if (this.expectedType) {
                  var expected = this.expectedType;
                  var actual = parsed.type;
                  // When we expect a number, string, boolean, or array but have a value, wrap it in an assertion.
                  // When we expect a color or formatted string, but have a string or value, wrap it in a coercion.
                  // Otherwise, we do static type-checking.
                  //
                  // These behaviors are overridable for:
                  //   * The "coalesce" operator, which needs to omit type annotations.
                  //   * String-valued properties (e.g. `text-field`), where coercion is more convenient than assertion.
                  //
                  if ((expected.kind === 'string' || expected.kind === 'number' || expected.kind === 'boolean' || expected.kind === 'object' || expected.kind === 'array') && actual.kind === 'value') {
                      parsed = annotate(parsed, expected, options.typeAnnotation || 'assert');
                  } else if ((expected.kind === 'color' || expected.kind === 'formatted' || expected.kind === 'resolvedImage') && (actual.kind === 'value' || actual.kind === 'string')) {
                      parsed = annotate(parsed, expected, options.typeAnnotation || 'coerce');
                  } else if (this.checkSubtype(expected, actual)) {
                      return null;
                  }
              }
              // If an expression's arguments are all literals, we can evaluate
              // it immediately and replace it with a literal value in the
              // parsed/compiled result. Expressions that expect an image should
              // not be resolved here so we can later get the available images.
              if (!(parsed instanceof Literal) && parsed.type.kind !== 'resolvedImage' && isConstant(parsed)) {
                  var ec = new EvaluationContext();
                  try {
                      parsed = new Literal(parsed.type, parsed.evaluate(ec));
                  } catch (e) {
                      this.error(e.message);
                      return null;
                  }
              }
              return parsed;
          }
          return this.error(("Unknown expression \"" + op + "\". If you wanted a literal array, use [\"literal\", [...]]."), 0);
      } else if (typeof expr === 'undefined') {
          return this.error("'undefined' value invalid. Use null instead.");
      } else if (typeof expr === 'object') {
          return this.error("Bare objects invalid. Use [\"literal\", {...}] instead.");
      } else {
          return this.error(("Expected an array, but found " + (typeof expr) + " instead."));
      }
  };
  /**
   * Returns a copy of this context suitable for parsing the subexpression at
   * index `index`, optionally appending to 'let' binding map.
   *
   * Note that `errors` property, intended for collecting errors while
   * parsing, is copied by reference rather than cloned.
   * @private
   */
  ParsingContext.prototype.concat = function concat (index, expectedType, bindings) {
      var path = typeof index === 'number' ? this.path.concat(index) : this.path;
      var scope = bindings ? this.scope.concat(bindings) : this.scope;
      return new ParsingContext(this.registry, path, expectedType || null, scope, this.errors);
  };
  /**
   * Push a parsing (or type checking) error into the `this.errors`
   * @param error The message
   * @param keys Optionally specify the source of the error at a child
   * of the current expression at `this.key`.
   * @private
   */
  ParsingContext.prototype.error = function error (error$1) {
          var keys = [], len = arguments.length - 1;
          while ( len-- > 0 ) keys[ len ] = arguments[ len + 1 ];

      var key = "" + (this.key) + (keys.map(function (k) { return ("[" + k + "]"); }).join(''));
      this.errors.push(new ParsingError(key, error$1));
  };
  /**
   * Returns null if `t` is a subtype of `expected`; otherwise returns an
   * error message and also pushes it to `this.errors`.
   */
  ParsingContext.prototype.checkSubtype = function checkSubtype$1 (expected, t) {
      var error = checkSubtype(expected, t);
      if (error)
          { this.error(error); }
      return error;
  };
  function isConstant(expression) {
      if (expression instanceof Var) {
          return isConstant(expression.boundExpression);
      } else if (expression instanceof CompoundExpression && expression.name === 'error') {
          return false;
      } else if (expression instanceof CollatorExpression) {
          // Although the results of a Collator expression with fixed arguments
          // generally shouldn't change between executions, we can't serialize them
          // as constant expressions because results change based on environment.
          return false;
      } else if (expression instanceof Within$1) {
          return false;
      }
      var isTypeAnnotation = expression instanceof Coercion || expression instanceof Assertion;
      var childrenConstant = true;
      expression.eachChild(function (child) {
          // We can _almost_ assume that if `expressions` children are constant,
          // they would already have been evaluated to Literal values when they
          // were parsed.  Type annotations are the exception, because they might
          // have been inferred and added after a child was parsed.
          // So we recurse into isConstant() for the children of type annotations,
          // but otherwise simply check whether they are Literals.
          if (isTypeAnnotation) {
              childrenConstant = childrenConstant && isConstant(child);
          } else {
              childrenConstant = childrenConstant && child instanceof Literal;
          }
      });
      if (!childrenConstant) {
          return false;
      }
      return isFeatureConstant(expression) && isGlobalPropertyConstant(expression, [
          'zoom',
          'heatmap-density',
          'line-progress',
          'sky-radial-progress',
          'accumulated',
          'is-supported-script',
          'pitch',
          'distance-from-center'
      ]);
  }

  //      
  /**
   * Returns the index of the last stop <= input, or 0 if it doesn't exist.
   * @private
   */
  function findStopLessThanOrEqualTo(stops, input) {
      var lastIndex = stops.length - 1;
      var lowerIndex = 0;
      var upperIndex = lastIndex;
      var currentIndex = 0;
      var currentValue, nextValue;
      while (lowerIndex <= upperIndex) {
          currentIndex = Math.floor((lowerIndex + upperIndex) / 2);
          currentValue = stops[currentIndex];
          nextValue = stops[currentIndex + 1];
          if (currentValue <= input) {
              if (currentIndex === lastIndex || input < nextValue) {
                  // Search complete
                  return currentIndex;
              }
              lowerIndex = currentIndex + 1;
          } else if (currentValue > input) {
              upperIndex = currentIndex - 1;
          } else {
              throw new RuntimeError('Input is not a number.');
          }
      }
      return 0;
  }

  //      
  var Step = function Step(type, input, stops) {
      this.type = type;
      this.input = input;
      this.labels = [];
      this.outputs = [];
      for (var i = 0, list = stops; i < list.length; i += 1) {
          var ref = list[i];
          var label = ref[0];
          var expression = ref[1];

          this.labels.push(label);
          this.outputs.push(expression);
      }
  };
  Step.parse = function parse (args, context) {
      if (args.length - 1 < 4) {
          return context.error(("Expected at least 4 arguments, but found only " + (args.length - 1) + "."));
      }
      if ((args.length - 1) % 2 !== 0) {
          return context.error("Expected an even number of arguments.");
      }
      var input = context.parse(args[1], 1, NumberType);
      if (!input)
          { return null; }
      var stops = [];
      var outputType = null;
      if (context.expectedType && context.expectedType.kind !== 'value') {
          outputType = context.expectedType;
      }
      for (var i = 1; i < args.length; i += 2) {
          var label = i === 1 ? -Infinity : args[i];
          var value = args[i + 1];
          var labelKey = i;
          var valueKey = i + 1;
          if (typeof label !== 'number') {
              return context.error('Input/output pairs for "step" expressions must be defined using literal numeric values (not computed expressions) for the input values.', labelKey);
          }
          if (stops.length && stops[stops.length - 1][0] >= label) {
              return context.error('Input/output pairs for "step" expressions must be arranged with input values in strictly ascending order.', labelKey);
          }
          var parsed = context.parse(value, valueKey, outputType);
          if (!parsed)
              { return null; }
          outputType = outputType || parsed.type;
          stops.push([
              label,
              parsed
          ]);
      }
      return new Step(outputType, input, stops);
  };
  Step.prototype.evaluate = function evaluate (ctx) {
      var labels = this.labels;
      var outputs = this.outputs;
      if (labels.length === 1) {
          return outputs[0].evaluate(ctx);
      }
      var value = this.input.evaluate(ctx);
      if (value <= labels[0]) {
          return outputs[0].evaluate(ctx);
      }
      var stopCount = labels.length;
      if (value >= labels[stopCount - 1]) {
          return outputs[stopCount - 1].evaluate(ctx);
      }
      var index = findStopLessThanOrEqualTo(labels, value);
      return outputs[index].evaluate(ctx);
  };
  Step.prototype.eachChild = function eachChild (fn) {
      fn(this.input);
      for (var i = 0, list = this.outputs; i < list.length; i += 1) {
          var expression = list[i];

              fn(expression);
      }
  };
  Step.prototype.outputDefined = function outputDefined () {
      return this.outputs.every(function (out) { return out.outputDefined(); });
  };
  Step.prototype.serialize = function serialize () {
      var serialized = [
          'step',
          this.input.serialize()
      ];
      for (var i = 0; i < this.labels.length; i++) {
          if (i > 0) {
              serialized.push(this.labels[i]);
          }
          serialized.push(this.outputs[i].serialize());
      }
      return serialized;
  };

  /*
   * Copyright (C) 2008 Apple Inc. All Rights Reserved.
   *
   * Redistribution and use in source and binary forms, with or without
   * modification, are permitted provided that the following conditions
   * are met:
   * 1. Redistributions of source code must retain the above copyright
   *    notice, this list of conditions and the following disclaimer.
   * 2. Redistributions in binary form must reproduce the above copyright
   *    notice, this list of conditions and the following disclaimer in the
   *    documentation and/or other materials provided with the distribution.
   *
   * THIS SOFTWARE IS PROVIDED BY APPLE INC. ``AS IS'' AND ANY
   * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
   * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
   * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL APPLE INC. OR
   * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY
   * OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
   * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
   * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
   *
   * Ported from Webkit
   * http://svn.webkit.org/repository/webkit/trunk/Source/WebCore/platform/graphics/UnitBezier.h
   */

  var unitbezier = UnitBezier;
  function UnitBezier(p1x, p1y, p2x, p2y) {
      // Calculate the polynomial coefficients, implicit first and last control points are (0,0) and (1,1).
      this.cx = 3 * p1x;
      this.bx = 3 * (p2x - p1x) - this.cx;
      this.ax = 1 - this.cx - this.bx;
      this.cy = 3 * p1y;
      this.by = 3 * (p2y - p1y) - this.cy;
      this.ay = 1 - this.cy - this.by;
      this.p1x = p1x;
      this.p1y = p2y;
      this.p2x = p2x;
      this.p2y = p2y;
  }
  UnitBezier.prototype.sampleCurveX = function (t) {
      // `ax t^3 + bx t^2 + cx t' expanded using Horner's rule.
      return ((this.ax * t + this.bx) * t + this.cx) * t;
  };
  UnitBezier.prototype.sampleCurveY = function (t) {
      return ((this.ay * t + this.by) * t + this.cy) * t;
  };
  UnitBezier.prototype.sampleCurveDerivativeX = function (t) {
      return (3 * this.ax * t + 2 * this.bx) * t + this.cx;
  };
  UnitBezier.prototype.solveCurveX = function (x, epsilon) {
      if (typeof epsilon === 'undefined')
          { epsilon = 0.000001; }
      var t0, t1, t2, x2, i;
      // First try a few iterations of Newton's method -- normally very fast.
      for (t2 = x, i = 0; i < 8; i++) {
          x2 = this.sampleCurveX(t2) - x;
          if (Math.abs(x2) < epsilon)
              { return t2; }
          var d2 = this.sampleCurveDerivativeX(t2);
          if (Math.abs(d2) < 0.000001)
              { break; }
          t2 = t2 - x2 / d2;
      }
      // Fall back to the bisection method for reliability.
      t0 = 0;
      t1 = 1;
      t2 = x;
      if (t2 < t0)
          { return t0; }
      if (t2 > t1)
          { return t1; }
      while (t0 < t1) {
          x2 = this.sampleCurveX(t2);
          if (Math.abs(x2 - x) < epsilon)
              { return t2; }
          if (x > x2) {
              t0 = t2;
          } else {
              t1 = t2;
          }
          t2 = (t1 - t0) * 0.5 + t0;
      }
      // Failure.
      return t2;
  };
  UnitBezier.prototype.solve = function (x, epsilon) {
      return this.sampleCurveY(this.solveCurveX(x, epsilon));
  };

  //      
  function number(a, b, t) {
      return a * (1 - t) + b * t;
  }
  function color(from, to, t) {
      return new Color(number(from.r, to.r, t), number(from.g, to.g, t), number(from.b, to.b, t), number(from.a, to.a, t));
  }
  function array(from, to, t) {
      return from.map(function (d, i) {
          return number(d, to[i], t);
      });
  }

  var interpolate = /*#__PURE__*/Object.freeze({
    __proto__: null,
    number: number,
    color: color,
    array: array
  });

  //      
  // Constants
  var Xn = 0.95047,
      // D65 standard referent
      Yn = 1, Zn = 1.08883, t0 = 4 / 29, t1 = 6 / 29, t2 = 3 * t1 * t1, t3 = t1 * t1 * t1, deg2rad$1 = Math.PI / 180, rad2deg = 180 / Math.PI;
  // Utilities
  function xyz2lab(t) {
      return t > t3 ? Math.pow(t, 1 / 3) : t / t2 + t0;
  }
  function lab2xyz(t) {
      return t > t1 ? t * t * t : t2 * (t - t0);
  }
  function xyz2rgb(x) {
      return 255 * (x <= 0.0031308 ? 12.92 * x : 1.055 * Math.pow(x, 1 / 2.4) - 0.055);
  }
  function rgb2xyz(x) {
      x /= 255;
      return x <= 0.04045 ? x / 12.92 : Math.pow((x + 0.055) / 1.055, 2.4);
  }
  // LAB
  function rgbToLab(rgbColor) {
      var b = rgb2xyz(rgbColor.r), a = rgb2xyz(rgbColor.g), l = rgb2xyz(rgbColor.b), x = xyz2lab((0.4124564 * b + 0.3575761 * a + 0.1804375 * l) / Xn), y = xyz2lab((0.2126729 * b + 0.7151522 * a + 0.072175 * l) / Yn), z = xyz2lab((0.0193339 * b + 0.119192 * a + 0.9503041 * l) / Zn);
      return {
          l: 116 * y - 16,
          a: 500 * (x - y),
          b: 200 * (y - z),
          alpha: rgbColor.a
      };
  }
  function labToRgb(labColor) {
      var y = (labColor.l + 16) / 116, x = isNaN(labColor.a) ? y : y + labColor.a / 500, z = isNaN(labColor.b) ? y : y - labColor.b / 200;
      y = Yn * lab2xyz(y);
      x = Xn * lab2xyz(x);
      z = Zn * lab2xyz(z);
      return new Color(xyz2rgb(3.2404542 * x - 1.5371385 * y - 0.4985314 * z), // D65 -> sRGB
      xyz2rgb(-0.969266 * x + 1.8760108 * y + 0.041556 * z), xyz2rgb(0.0556434 * x - 0.2040259 * y + 1.0572252 * z), labColor.alpha);
  }
  function interpolateLab(from, to, t) {
      return {
          l: number(from.l, to.l, t),
          a: number(from.a, to.a, t),
          b: number(from.b, to.b, t),
          alpha: number(from.alpha, to.alpha, t)
      };
  }
  // HCL
  function rgbToHcl(rgbColor) {
      var ref = rgbToLab(rgbColor);
      var l = ref.l;
      var a = ref.a;
      var b = ref.b;
      var h = Math.atan2(b, a) * rad2deg;
      return {
          h: h < 0 ? h + 360 : h,
          c: Math.sqrt(a * a + b * b),
          l: l,
          alpha: rgbColor.a
      };
  }
  function hclToRgb(hclColor) {
      var h = hclColor.h * deg2rad$1, c = hclColor.c, l = hclColor.l;
      return labToRgb({
          l: l,
          a: Math.cos(h) * c,
          b: Math.sin(h) * c,
          alpha: hclColor.alpha
      });
  }
  function interpolateHue(a, b, t) {
      var d = b - a;
      return a + t * (d > 180 || d < -180 ? d - 360 * Math.round(d / 360) : d);
  }
  function interpolateHcl(from, to, t) {
      return {
          h: interpolateHue(from.h, to.h, t),
          c: number(from.c, to.c, t),
          l: number(from.l, to.l, t),
          alpha: number(from.alpha, to.alpha, t)
      };
  }
  var lab = {
      forward: rgbToLab,
      reverse: labToRgb,
      interpolate: interpolateLab
  };
  var hcl = {
      forward: rgbToHcl,
      reverse: hclToRgb,
      interpolate: interpolateHcl
  };

  //      
  var Interpolate = function Interpolate(type, operator, interpolation, input, stops) {
      this.type = type;
      this.operator = operator;
      this.interpolation = interpolation;
      this.input = input;
      this.labels = [];
      this.outputs = [];
      for (var i = 0, list = stops; i < list.length; i += 1) {
          var ref = list[i];
          var label = ref[0];
          var expression = ref[1];

          this.labels.push(label);
          this.outputs.push(expression);
      }
  };
  Interpolate.interpolationFactor = function interpolationFactor (interpolation, input, lower, upper) {
      var t = 0;
      if (interpolation.name === 'exponential') {
          t = exponentialInterpolation(input, interpolation.base, lower, upper);
      } else if (interpolation.name === 'linear') {
          t = exponentialInterpolation(input, 1, lower, upper);
      } else if (interpolation.name === 'cubic-bezier') {
          var c = interpolation.controlPoints;
          var ub = new unitbezier(c[0], c[1], c[2], c[3]);
          t = ub.solve(exponentialInterpolation(input, 1, lower, upper));
      }
      return t;
  };
  Interpolate.parse = function parse (args, context) {
      var operator = args[0];
          var interpolation = args[1];
          var input = args[2];
          var rest = args.slice(3);
      if (!Array.isArray(interpolation) || interpolation.length === 0) {
          return context.error("Expected an interpolation type expression.", 1);
      }
      if (interpolation[0] === 'linear') {
          interpolation = { name: 'linear' };
      } else if (interpolation[0] === 'exponential') {
          var base = interpolation[1];
          if (typeof base !== 'number')
              { return context.error("Exponential interpolation requires a numeric base.", 1, 1); }
          interpolation = {
              name: 'exponential',
              base: base
          };
      } else if (interpolation[0] === 'cubic-bezier') {
          var controlPoints = interpolation.slice(1);
          if (controlPoints.length !== 4 || controlPoints.some(function (t) { return typeof t !== 'number' || t < 0 || t > 1; })) {
              return context.error('Cubic bezier interpolation requires four numeric arguments with values between 0 and 1.', 1);
          }
          interpolation = {
              name: 'cubic-bezier',
              controlPoints: controlPoints
          };
      } else {
          return context.error(("Unknown interpolation type " + (String(interpolation[0]))), 1, 0);
      }
      if (args.length - 1 < 4) {
          return context.error(("Expected at least 4 arguments, but found only " + (args.length - 1) + "."));
      }
      if ((args.length - 1) % 2 !== 0) {
          return context.error("Expected an even number of arguments.");
      }
      input = context.parse(input, 2, NumberType);
      if (!input)
          { return null; }
      var stops = [];
      var outputType = null;
      if (operator === 'interpolate-hcl' || operator === 'interpolate-lab') {
          outputType = ColorType;
      } else if (context.expectedType && context.expectedType.kind !== 'value') {
          outputType = context.expectedType;
      }
      for (var i = 0; i < rest.length; i += 2) {
          var label = rest[i];
          var value = rest[i + 1];
          var labelKey = i + 3;
          var valueKey = i + 4;
          if (typeof label !== 'number') {
              return context.error('Input/output pairs for "interpolate" expressions must be defined using literal numeric values (not computed expressions) for the input values.', labelKey);
          }
          if (stops.length && stops[stops.length - 1][0] >= label) {
              return context.error('Input/output pairs for "interpolate" expressions must be arranged with input values in strictly ascending order.', labelKey);
          }
          var parsed = context.parse(value, valueKey, outputType);
          if (!parsed)
              { return null; }
          outputType = outputType || parsed.type;
          stops.push([
              label,
              parsed
          ]);
      }
      if (outputType.kind !== 'number' && outputType.kind !== 'color' && !(outputType.kind === 'array' && outputType.itemType.kind === 'number' && typeof outputType.N === 'number')) {
          return context.error(("Type " + (toString$1(outputType)) + " is not interpolatable."));
      }
      return new Interpolate(outputType, operator, interpolation, input, stops);
  };
  Interpolate.prototype.evaluate = function evaluate (ctx) {
      var labels = this.labels;
      var outputs = this.outputs;
      if (labels.length === 1) {
          return outputs[0].evaluate(ctx);
      }
      var value = this.input.evaluate(ctx);
      if (value <= labels[0]) {
          return outputs[0].evaluate(ctx);
      }
      var stopCount = labels.length;
      if (value >= labels[stopCount - 1]) {
          return outputs[stopCount - 1].evaluate(ctx);
      }
      var index = findStopLessThanOrEqualTo(labels, value);
      var lower = labels[index];
      var upper = labels[index + 1];
      var t = Interpolate.interpolationFactor(this.interpolation, value, lower, upper);
      var outputLower = outputs[index].evaluate(ctx);
      var outputUpper = outputs[index + 1].evaluate(ctx);
      if (this.operator === 'interpolate') {
          return interpolate[this.type.kind.toLowerCase()](outputLower, outputUpper, t);// eslint-disable-line import/namespace
      } else if (this.operator === 'interpolate-hcl') {
          return hcl.reverse(hcl.interpolate(hcl.forward(outputLower), hcl.forward(outputUpper), t));
      } else {
          return lab.reverse(lab.interpolate(lab.forward(outputLower), lab.forward(outputUpper), t));
      }
  };
  Interpolate.prototype.eachChild = function eachChild (fn) {
      fn(this.input);
      for (var i = 0, list = this.outputs; i < list.length; i += 1) {
          var expression = list[i];

              fn(expression);
      }
  };
  Interpolate.prototype.outputDefined = function outputDefined () {
      return this.outputs.every(function (out) { return out.outputDefined(); });
  };
  Interpolate.prototype.serialize = function serialize () {
      var interpolation;
      if (this.interpolation.name === 'linear') {
          interpolation = ['linear'];
      } else if (this.interpolation.name === 'exponential') {
          if (this.interpolation.base === 1) {
              interpolation = ['linear'];
          } else {
              interpolation = [
                  'exponential',
                  this.interpolation.base
              ];
          }
      } else {
          interpolation = ['cubic-bezier'].concat(this.interpolation.controlPoints);
      }
      var serialized = [
          this.operator,
          interpolation,
          this.input.serialize()
      ];
      for (var i = 0; i < this.labels.length; i++) {
          serialized.push(this.labels[i], this.outputs[i].serialize());
      }
      return serialized;
  };
  /**
   * Returns a ratio that can be used to interpolate between exponential function
   * stops.
   * How it works: Two consecutive stop values define a (scaled and shifted) exponential function `f(x) = a * base^x + b`, where `base` is the user-specified base,
   * and `a` and `b` are constants affording sufficient degrees of freedom to fit
   * the function to the given stops.
   *
   * Here's a bit of algebra that lets us compute `f(x)` directly from the stop
   * values without explicitly solving for `a` and `b`:
   *
   * First stop value: `f(x0) = y0 = a * base^x0 + b`
   * Second stop value: `f(x1) = y1 = a * base^x1 + b`
   * => `y1 - y0 = a(base^x1 - base^x0)`
   * => `a = (y1 - y0)/(base^x1 - base^x0)`
   *
   * Desired value: `f(x) = y = a * base^x + b`
   * => `f(x) = y0 + a * (base^x - base^x0)`
   *
   * From the above, we can replace the `a` in `a * (base^x - base^x0)` and do a
   * little algebra:
   * ```
   * a * (base^x - base^x0) = (y1 - y0)/(base^x1 - base^x0) * (base^x - base^x0)
   *                     = (y1 - y0) * (base^x - base^x0) / (base^x1 - base^x0)
   * ```
   *
   * If we let `(base^x - base^x0) / (base^x1 base^x0)`, then we have
   * `f(x) = y0 + (y1 - y0) * ratio`.  In other words, `ratio` may be treated as
   * an interpolation factor between the two stops' output values.
   *
   * (Note: a slightly different form for `ratio`,
   * `(base^(x-x0) - 1) / (base^(x1-x0) - 1) `, is equivalent, but requires fewer
   * expensive `Math.pow()` operations.)
   *
   * @private
  */
  function exponentialInterpolation(input, base, lowerValue, upperValue) {
      var difference = upperValue - lowerValue;
      var progress = input - lowerValue;
      if (difference === 0) {
          return 0;
      } else if (base === 1) {
          return progress / difference;
      } else {
          return (Math.pow(base, progress) - 1) / (Math.pow(base, difference) - 1);
      }
  }

  var Coalesce = function Coalesce(type, args) {
      this.type = type;
      this.args = args;
  };
  Coalesce.parse = function parse (args, context) {
      if (args.length < 2) {
          return context.error('Expectected at least one argument.');
      }
      var outputType = null;
      var expectedType = context.expectedType;
      if (expectedType && expectedType.kind !== 'value') {
          outputType = expectedType;
      }
      var parsedArgs = [];
      for (var i = 0, list = args.slice(1); i < list.length; i += 1) {
          var arg = list[i];

              var parsed = context.parse(arg, 1 + parsedArgs.length, outputType, undefined, { typeAnnotation: 'omit' });
          if (!parsed)
              { return null; }
          outputType = outputType || parsed.type;
          parsedArgs.push(parsed);
      }
      // Above, we parse arguments without inferred type annotation so that
      // they don't produce a runtime error for `null` input, which would
      // preempt the desired null-coalescing behavior.
      // Thus, if any of our arguments would have needed an annotation, we
      // need to wrap the enclosing coalesce expression with it instead.
      var needsAnnotation = expectedType && parsedArgs.some(function (arg) { return checkSubtype(expectedType, arg.type); });
      return needsAnnotation ? new Coalesce(ValueType, parsedArgs) : new Coalesce(outputType, parsedArgs);
  };
  Coalesce.prototype.evaluate = function evaluate (ctx) {
      var result = null;
      var argCount = 0;
      var firstImage;
      for (var i = 0, list = this.args; i < list.length; i += 1) {
          var arg = list[i];

              argCount++;
          result = arg.evaluate(ctx);
          // we need to keep track of the first requested image in a coalesce statement
          // if coalesce can't find a valid image, we return the first image so styleimagemissing can fire
          if (result && result instanceof ResolvedImage && !result.available) {
              // set to first image
              if (!firstImage) {
                  firstImage = result;
              }
              result = null;
              // if we reach the end, return the first image
              if (argCount === this.args.length) {
                  return firstImage;
              }
          }
          if (result !== null)
              { break; }
      }
      return result;
  };
  Coalesce.prototype.eachChild = function eachChild (fn) {
      this.args.forEach(fn);
  };
  Coalesce.prototype.outputDefined = function outputDefined () {
      return this.args.every(function (arg) { return arg.outputDefined(); });
  };
  Coalesce.prototype.serialize = function serialize () {
      var serialized = ['coalesce'];
      this.eachChild(function (child) {
          serialized.push(child.serialize());
      });
      return serialized;
  };

  //      
  var Let = function Let(bindings, result) {
      this.type = result.type;
      this.bindings = [].concat(bindings);
      this.result = result;
  };
  Let.prototype.evaluate = function evaluate (ctx) {
      return this.result.evaluate(ctx);
  };
  Let.prototype.eachChild = function eachChild (fn) {
      for (var i = 0, list = this.bindings; i < list.length; i += 1) {
          var binding = list[i];

              fn(binding[1]);
      }
      fn(this.result);
  };
  Let.parse = function parse (args, context) {
      if (args.length < 4)
          { return context.error(("Expected at least 3 arguments, but found " + (args.length - 1) + " instead.")); }
      var bindings = [];
      for (var i = 1; i < args.length - 1; i += 2) {
          var name = args[i];
          if (typeof name !== 'string') {
              return context.error(("Expected string, but found " + (typeof name) + " instead."), i);
          }
          if (/[^a-zA-Z0-9_]/.test(name)) {
              return context.error("Variable names must contain only alphanumeric characters or '_'.", i);
          }
          var value = context.parse(args[i + 1], i + 1);
          if (!value)
              { return null; }
          bindings.push([
              name,
              value
          ]);
      }
      var result = context.parse(args[args.length - 1], args.length - 1, context.expectedType, bindings);
      if (!result)
          { return null; }
      return new Let(bindings, result);
  };
  Let.prototype.outputDefined = function outputDefined () {
      return this.result.outputDefined();
  };
  Let.prototype.serialize = function serialize () {
      var serialized = ['let'];
      for (var i = 0, list = this.bindings; i < list.length; i += 1) {
          var ref = list[i];
              var name = ref[0];
              var expr = ref[1];

              serialized.push(name, expr.serialize());
      }
      serialized.push(this.result.serialize());
      return serialized;
  };

  //      
  var At = function At(type, index, input) {
      this.type = type;
      this.index = index;
      this.input = input;
  };
  At.parse = function parse (args, context) {
      if (args.length !== 3)
          { return context.error(("Expected 2 arguments, but found " + (args.length - 1) + " instead.")); }
      var index = context.parse(args[1], 1, NumberType);
      var input = context.parse(args[2], 2, array$1(context.expectedType || ValueType));
      if (!index || !input)
          { return null; }
      var t = input.type;
      return new At(t.itemType, index, input);
  };
  At.prototype.evaluate = function evaluate (ctx) {
      var index = this.index.evaluate(ctx);
      var array = this.input.evaluate(ctx);
      if (index < 0) {
          throw new RuntimeError(("Array index out of bounds: " + index + " < 0."));
      }
      if (index >= array.length) {
          throw new RuntimeError(("Array index out of bounds: " + index + " > " + (array.length - 1) + "."));
      }
      if (index !== Math.floor(index)) {
          throw new RuntimeError(("Array index must be an integer, but found " + index + " instead."));
      }
      return array[index];
  };
  At.prototype.eachChild = function eachChild (fn) {
      fn(this.index);
      fn(this.input);
  };
  At.prototype.outputDefined = function outputDefined () {
      return false;
  };
  At.prototype.serialize = function serialize () {
      return [
          'at',
          this.index.serialize(),
          this.input.serialize()
      ];
  };

  //      
  var In = function In(needle, haystack) {
      this.type = BooleanType;
      this.needle = needle;
      this.haystack = haystack;
  };
  In.parse = function parse (args, context) {
      if (args.length !== 3) {
          return context.error(("Expected 2 arguments, but found " + (args.length - 1) + " instead."));
      }
      var needle = context.parse(args[1], 1, ValueType);
      var haystack = context.parse(args[2], 2, ValueType);
      if (!needle || !haystack)
          { return null; }
      if (!isValidType(needle.type, [
              BooleanType,
              StringType,
              NumberType,
              NullType,
              ValueType
          ])) {
          return context.error(("Expected first argument to be of type boolean, string, number or null, but found " + (toString$1(needle.type)) + " instead"));
      }
      return new In(needle, haystack);
  };
  In.prototype.evaluate = function evaluate (ctx) {
      var needle = this.needle.evaluate(ctx);
      var haystack = this.haystack.evaluate(ctx);
      if (haystack == null)
          { return false; }
      if (!isValidNativeType(needle, [
              'boolean',
              'string',
              'number',
              'null'
          ])) {
          throw new RuntimeError(("Expected first argument to be of type boolean, string, number or null, but found " + (toString$1(typeOf(needle))) + " instead."));
      }
      if (!isValidNativeType(haystack, [
              'string',
              'array'
          ])) {
          throw new RuntimeError(("Expected second argument to be of type array or string, but found " + (toString$1(typeOf(haystack))) + " instead."));
      }
      return haystack.indexOf(needle) >= 0;
  };
  In.prototype.eachChild = function eachChild (fn) {
      fn(this.needle);
      fn(this.haystack);
  };
  In.prototype.outputDefined = function outputDefined () {
      return true;
  };
  In.prototype.serialize = function serialize () {
      return [
          'in',
          this.needle.serialize(),
          this.haystack.serialize()
      ];
  };

  //      
  var IndexOf = function IndexOf(needle, haystack, fromIndex) {
      this.type = NumberType;
      this.needle = needle;
      this.haystack = haystack;
      this.fromIndex = fromIndex;
  };
  IndexOf.parse = function parse (args, context) {
      if (args.length <= 2 || args.length >= 5) {
          return context.error(("Expected 3 or 4 arguments, but found " + (args.length - 1) + " instead."));
      }
      var needle = context.parse(args[1], 1, ValueType);
      var haystack = context.parse(args[2], 2, ValueType);
      if (!needle || !haystack)
          { return null; }
      if (!isValidType(needle.type, [
              BooleanType,
              StringType,
              NumberType,
              NullType,
              ValueType
          ])) {
          return context.error(("Expected first argument to be of type boolean, string, number or null, but found " + (toString$1(needle.type)) + " instead"));
      }
      if (args.length === 4) {
          var fromIndex = context.parse(args[3], 3, NumberType);
          if (!fromIndex)
              { return null; }
          return new IndexOf(needle, haystack, fromIndex);
      } else {
          return new IndexOf(needle, haystack);
      }
  };
  IndexOf.prototype.evaluate = function evaluate (ctx) {
      var needle = this.needle.evaluate(ctx);
      var haystack = this.haystack.evaluate(ctx);
      if (!isValidNativeType(needle, [
              'boolean',
              'string',
              'number',
              'null'
          ])) {
          throw new RuntimeError(("Expected first argument to be of type boolean, string, number or null, but found " + (toString$1(typeOf(needle))) + " instead."));
      }
      if (!isValidNativeType(haystack, [
              'string',
              'array'
          ])) {
          throw new RuntimeError(("Expected second argument to be of type array or string, but found " + (toString$1(typeOf(haystack))) + " instead."));
      }
      if (this.fromIndex) {
          var fromIndex = this.fromIndex.evaluate(ctx);
          return haystack.indexOf(needle, fromIndex);
      }
      return haystack.indexOf(needle);
  };
  IndexOf.prototype.eachChild = function eachChild (fn) {
      fn(this.needle);
      fn(this.haystack);
      if (this.fromIndex) {
          fn(this.fromIndex);
      }
  };
  IndexOf.prototype.outputDefined = function outputDefined () {
      return false;
  };
  IndexOf.prototype.serialize = function serialize () {
      if (this.fromIndex != null && this.fromIndex !== undefined) {
          var fromIndex = this.fromIndex.serialize();
          return [
              'index-of',
              this.needle.serialize(),
              this.haystack.serialize(),
              fromIndex
          ];
      }
      return [
          'index-of',
          this.needle.serialize(),
          this.haystack.serialize()
      ];
  };

  // Map input label values to output expression index
  var Match = function Match(inputType, outputType, input, cases, outputs, otherwise) {
      this.inputType = inputType;
      this.type = outputType;
      this.input = input;
      this.cases = cases;
      this.outputs = outputs;
      this.otherwise = otherwise;
  };
  Match.parse = function parse (args, context) {
      if (args.length < 5)
          { return context.error(("Expected at least 4 arguments, but found only " + (args.length - 1) + ".")); }
      if (args.length % 2 !== 1)
          { return context.error("Expected an even number of arguments."); }
      var inputType;
      var outputType;
      if (context.expectedType && context.expectedType.kind !== 'value') {
          outputType = context.expectedType;
      }
      var cases = {};
      var outputs = [];
      for (var i = 2; i < args.length - 1; i += 2) {
          var labels = args[i];
          var value = args[i + 1];
          if (!Array.isArray(labels)) {
              labels = [labels];
          }
          var labelContext = context.concat(i);
          if (labels.length === 0) {
              return labelContext.error('Expected at least one branch label.');
          }
          for (var i$1 = 0, list = labels; i$1 < list.length; i$1 += 1) {
              var label = list[i$1];

                  if (typeof label !== 'number' && typeof label !== 'string') {
                  return labelContext.error("Branch labels must be numbers or strings.");
              } else if (typeof label === 'number' && Math.abs(label) > Number.MAX_SAFE_INTEGER) {
                  return labelContext.error(("Branch labels must be integers no larger than " + (Number.MAX_SAFE_INTEGER) + "."));
              } else if (typeof label === 'number' && Math.floor(label) !== label) {
                  return labelContext.error("Numeric branch labels must be integer values.");
              } else if (!inputType) {
                  inputType = typeOf(label);
              } else if (labelContext.checkSubtype(inputType, typeOf(label))) {
                  return null;
              }
              if (typeof cases[String(label)] !== 'undefined') {
                  return labelContext.error('Branch labels must be unique.');
              }
              cases[String(label)] = outputs.length;
          }
          var result = context.parse(value, i, outputType);
          if (!result)
              { return null; }
          outputType = outputType || result.type;
          outputs.push(result);
      }
      var input = context.parse(args[1], 1, ValueType);
      if (!input)
          { return null; }
      var otherwise = context.parse(args[args.length - 1], args.length - 1, outputType);
      if (!otherwise)
          { return null; }
      if (input.type.kind !== 'value' && context.concat(1).checkSubtype(inputType, input.type)) {
          return null;
      }
      return new Match(inputType, outputType, input, cases, outputs, otherwise);
  };
  Match.prototype.evaluate = function evaluate (ctx) {
      var input = this.input.evaluate(ctx);
      var output = typeOf(input) === this.inputType && this.outputs[this.cases[input]] || this.otherwise;
      return output.evaluate(ctx);
  };
  Match.prototype.eachChild = function eachChild (fn) {
      fn(this.input);
      this.outputs.forEach(fn);
      fn(this.otherwise);
  };
  Match.prototype.outputDefined = function outputDefined () {
      return this.outputs.every(function (out) { return out.outputDefined(); }) && this.otherwise.outputDefined();
  };
  Match.prototype.serialize = function serialize () {
          var this$1$1 = this;

      var serialized = [
          'match',
          this.input.serialize()
      ];
      // Sort so serialization has an arbitrary defined order, even though
      // branch order doesn't affect evaluation
      var sortedLabels = Object.keys(this.cases).sort();
      // Group branches by unique match expression to support condensed
      // serializations of the form [case1, case2, ...] -> matchExpression
      var groupedByOutput = [];
      var outputLookup = {};
      // lookup index into groupedByOutput for a given output expression
      for (var i = 0, list = sortedLabels; i < list.length; i += 1) {
          var label = list[i];

              var outputIndex = outputLookup[this.cases[label]];
          if (outputIndex === undefined) {
              // First time seeing this output, add it to the end of the grouped list
              outputLookup[this.cases[label]] = groupedByOutput.length;
              groupedByOutput.push([
                  this.cases[label],
                  [label]
              ]);
          } else {
              // We've seen this expression before, add the label to that output's group
              groupedByOutput[outputIndex][1].push(label);
          }
      }
      var coerceLabel = function (label) { return this$1$1.inputType.kind === 'number' ? Number(label) : label; };
      for (var i$1 = 0, list$1 = groupedByOutput; i$1 < list$1.length; i$1 += 1) {
          var ref = list$1[i$1];
              var outputIndex = ref[0];
              var labels = ref[1];

              if (labels.length === 1) {
              // Only a single label matches this output expression
              serialized.push(coerceLabel(labels[0]));
          } else {
              // Array of literal labels pointing to this output expression
              serialized.push(labels.map(coerceLabel));
          }
          serialized.push(this.outputs[outputIndex$1].serialize());
      }
      serialized.push(this.otherwise.serialize());
      return serialized;
  };

  var Case = function Case(type, branches, otherwise) {
      this.type = type;
      this.branches = branches;
      this.otherwise = otherwise;
  };
  Case.parse = function parse (args, context) {
      if (args.length < 4)
          { return context.error(("Expected at least 3 arguments, but found only " + (args.length - 1) + ".")); }
      if (args.length % 2 !== 0)
          { return context.error("Expected an odd number of arguments."); }
      var outputType;
      if (context.expectedType && context.expectedType.kind !== 'value') {
          outputType = context.expectedType;
      }
      var branches = [];
      for (var i = 1; i < args.length - 1; i += 2) {
          var test = context.parse(args[i], i, BooleanType);
          if (!test)
              { return null; }
          var result = context.parse(args[i + 1], i + 1, outputType);
          if (!result)
              { return null; }
          branches.push([
              test,
              result
          ]);
          outputType = outputType || result.type;
      }
      var otherwise = context.parse(args[args.length - 1], args.length - 1, outputType);
      if (!otherwise)
          { return null; }
      return new Case(outputType, branches, otherwise);
  };
  Case.prototype.evaluate = function evaluate (ctx) {
      for (var i = 0, list = this.branches; i < list.length; i += 1) {
          var ref = list[i];
              var test = ref[0];
              var expression = ref[1];

              if (test.evaluate(ctx)) {
              return expression.evaluate(ctx);
          }
      }
      return this.otherwise.evaluate(ctx);
  };
  Case.prototype.eachChild = function eachChild (fn) {
      for (var i = 0, list = this.branches; i < list.length; i += 1) {
          var ref = list[i];
              var test = ref[0];
              var expression = ref[1];

              fn(test);
          fn(expression);
      }
      fn(this.otherwise);
  };
  Case.prototype.outputDefined = function outputDefined () {
      return this.branches.every(function (ref) {
              ref[0];
              var out = ref[1];

              return out.outputDefined();
          }) && this.otherwise.outputDefined();
  };
  Case.prototype.serialize = function serialize () {
      var serialized = ['case'];
      this.eachChild(function (child) {
          serialized.push(child.serialize());
      });
      return serialized;
  };

  //      
  var Slice = function Slice(type, input, beginIndex, endIndex) {
      this.type = type;
      this.input = input;
      this.beginIndex = beginIndex;
      this.endIndex = endIndex;
  };
  Slice.parse = function parse (args, context) {
      if (args.length <= 2 || args.length >= 5) {
          return context.error(("Expected 3 or 4 arguments, but found " + (args.length - 1) + " instead."));
      }
      var input = context.parse(args[1], 1, ValueType);
      var beginIndex = context.parse(args[2], 2, NumberType);
      if (!input || !beginIndex)
          { return null; }
      if (!isValidType(input.type, [
              array$1(ValueType),
              StringType,
              ValueType
          ])) {
          return context.error(("Expected first argument to be of type array or string, but found " + (toString$1(input.type)) + " instead"));
      }
      if (args.length === 4) {
          var endIndex = context.parse(args[3], 3, NumberType);
          if (!endIndex)
              { return null; }
          return new Slice(input.type, input, beginIndex, endIndex);
      } else {
          return new Slice(input.type, input, beginIndex);
      }
  };
  Slice.prototype.evaluate = function evaluate (ctx) {
      var input = this.input.evaluate(ctx);
      var beginIndex = this.beginIndex.evaluate(ctx);
      if (!isValidNativeType(input, [
              'string',
              'array'
          ])) {
          throw new RuntimeError(("Expected first argument to be of type array or string, but found " + (toString$1(typeOf(input))) + " instead."));
      }
      if (this.endIndex) {
          var endIndex = this.endIndex.evaluate(ctx);
          return input.slice(beginIndex, endIndex);
      }
      return input.slice(beginIndex);
  };
  Slice.prototype.eachChild = function eachChild (fn) {
      fn(this.input);
      fn(this.beginIndex);
      if (this.endIndex) {
          fn(this.endIndex);
      }
  };
  Slice.prototype.outputDefined = function outputDefined () {
      return false;
  };
  Slice.prototype.serialize = function serialize () {
      if (this.endIndex != null && this.endIndex !== undefined) {
          var endIndex = this.endIndex.serialize();
          return [
              'slice',
              this.input.serialize(),
              this.beginIndex.serialize(),
              endIndex
          ];
      }
      return [
          'slice',
          this.input.serialize(),
          this.beginIndex.serialize()
      ];
  };

  //      
  function isComparableType(op, type) {
      if (op === '==' || op === '!=') {
          // equality operator
          return type.kind === 'boolean' || type.kind === 'string' || type.kind === 'number' || type.kind === 'null' || type.kind === 'value';
      } else {
          // ordering operator
          return type.kind === 'string' || type.kind === 'number' || type.kind === 'value';
      }
  }
  function eq(ctx, a, b) {
      return a === b;
  }
  function neq(ctx, a, b) {
      return a !== b;
  }
  function lt(ctx, a, b) {
      return a < b;
  }
  function gt(ctx, a, b) {
      return a > b;
  }
  function lteq(ctx, a, b) {
      return a <= b;
  }
  function gteq(ctx, a, b) {
      return a >= b;
  }
  function eqCollate(ctx, a, b, c) {
      return c.compare(a, b) === 0;
  }
  function neqCollate(ctx, a, b, c) {
      return !eqCollate(ctx, a, b, c);
  }
  function ltCollate(ctx, a, b, c) {
      return c.compare(a, b) < 0;
  }
  function gtCollate(ctx, a, b, c) {
      return c.compare(a, b) > 0;
  }
  function lteqCollate(ctx, a, b, c) {
      return c.compare(a, b) <= 0;
  }
  function gteqCollate(ctx, a, b, c) {
      return c.compare(a, b) >= 0;
  }
  /**
   * Special form for comparison operators, implementing the signatures:
   * - (T, T, ?Collator) => boolean
   * - (T, value, ?Collator) => boolean
   * - (value, T, ?Collator) => boolean
   *
   * For inequalities, T must be either value, string, or number. For ==/!=, it
   * can also be boolean or null.
   *
   * Equality semantics are equivalent to Javascript's strict equality (===/!==)
   * -- i.e., when the arguments' types don't match, == evaluates to false, != to
   * true.
   *
   * When types don't match in an ordering comparison, a runtime error is thrown.
   *
   * @private
   */
  function makeComparison(op, compareBasic, compareWithCollator) {
      var isOrderComparison = op !== '==' && op !== '!=';
      return /*@__PURE__*/(function () {
          function Comparison(lhs, rhs, collator) {
              this.type = BooleanType;
              this.lhs = lhs;
              this.rhs = rhs;
              this.collator = collator;
              this.hasUntypedArgument = lhs.type.kind === 'value' || rhs.type.kind === 'value';
          }
          Comparison.parse = function parse (args, context) {
              if (args.length !== 3 && args.length !== 4)
                  { return context.error("Expected two or three arguments."); }
              var op = args[0];
              var lhs = context.parse(args[1], 1, ValueType);
              if (!lhs)
                  { return null; }
              if (!isComparableType(op, lhs.type)) {
                  return context.concat(1).error(("\"" + op + "\" comparisons are not supported for type '" + (toString$1(lhs.type)) + "'."));
              }
              var rhs = context.parse(args[2], 2, ValueType);
              if (!rhs)
                  { return null; }
              if (!isComparableType(op, rhs.type)) {
                  return context.concat(2).error(("\"" + op + "\" comparisons are not supported for type '" + (toString$1(rhs.type)) + "'."));
              }
              if (lhs.type.kind !== rhs.type.kind && lhs.type.kind !== 'value' && rhs.type.kind !== 'value') {
                  return context.error(("Cannot compare types '" + (toString$1(lhs.type)) + "' and '" + (toString$1(rhs.type)) + "'."));
              }
              if (isOrderComparison) {
                  // typing rules specific to less/greater than operators
                  if (lhs.type.kind === 'value' && rhs.type.kind !== 'value') {
                      // (value, T)
                      lhs = new Assertion(rhs.type, [lhs]);
                  } else if (lhs.type.kind !== 'value' && rhs.type.kind === 'value') {
                      // (T, value)
                      rhs = new Assertion(lhs.type, [rhs]);
                  }
              }
              var collator = null;
              if (args.length === 4) {
                  if (lhs.type.kind !== 'string' && rhs.type.kind !== 'string' && lhs.type.kind !== 'value' && rhs.type.kind !== 'value') {
                      return context.error("Cannot use collator to compare non-string types.");
                  }
                  collator = context.parse(args[3], 3, CollatorType);
                  if (!collator)
                      { return null; }
              }
              return new Comparison(lhs, rhs, collator);
          };
          Comparison.prototype.evaluate = function evaluate (ctx) {
              var lhs = this.lhs.evaluate(ctx);
              var rhs = this.rhs.evaluate(ctx);
              if (isOrderComparison && this.hasUntypedArgument) {
                  var lt = typeOf(lhs);
                  var rt = typeOf(rhs);
                  // check that type is string or number, and equal
                  if (lt.kind !== rt.kind || !(lt.kind === 'string' || lt.kind === 'number')) {
                      throw new RuntimeError(("Expected arguments for \"" + op + "\" to be (string, string) or (number, number), but found (" + (lt.kind) + ", " + (rt.kind) + ") instead."));
                  }
              }
              if (this.collator && !isOrderComparison && this.hasUntypedArgument) {
                  var lt$1 = typeOf(lhs);
                  var rt$1 = typeOf(rhs);
                  if (lt$1.kind !== 'string' || rt$1.kind !== 'string') {
                      return compareBasic(ctx, lhs, rhs);
                  }
              }
              return this.collator ? compareWithCollator(ctx, lhs, rhs, this.collator.evaluate(ctx)) : compareBasic(ctx, lhs, rhs);
          };
          Comparison.prototype.eachChild = function eachChild (fn) {
              fn(this.lhs);
              fn(this.rhs);
              if (this.collator) {
                  fn(this.collator);
              }
          };
          Comparison.prototype.outputDefined = function outputDefined () {
              return true;
          };
          Comparison.prototype.serialize = function serialize () {
              var serialized = [op];
              this.eachChild(function (child) {
                  serialized.push(child.serialize());
              });
              return serialized;
          };

          return Comparison;
      }());
  }
  var Equals = makeComparison('==', eq, eqCollate);
  var NotEquals = makeComparison('!=', neq, neqCollate);
  var LessThan$1 = makeComparison('<', lt, ltCollate);
  var GreaterThan$1 = makeComparison('>', gt, gtCollate);
  var LessThanOrEqual = makeComparison('<=', lteq, lteqCollate);
  var GreaterThanOrEqual = makeComparison('>=', gteq, gteqCollate);

  //      
  var NumberFormat = function NumberFormat(number, locale, currency, minFractionDigits, maxFractionDigits) {
      this.type = StringType;
      this.number = number;
      this.locale = locale;
      this.currency = currency;
      this.minFractionDigits = minFractionDigits;
      this.maxFractionDigits = maxFractionDigits;
  };
  NumberFormat.parse = function parse (args, context) {
      if (args.length !== 3)
          { return context.error("Expected two arguments."); }
      var number = context.parse(args[1], 1, NumberType);
      if (!number)
          { return null; }
      var options = args[2];
      if (typeof options !== 'object' || Array.isArray(options))
          { return context.error("NumberFormat options argument must be an object."); }
      var locale = null;
      if (options['locale']) {
          locale = context.parse(options['locale'], 1, StringType);
          if (!locale)
              { return null; }
      }
      var currency = null;
      if (options['currency']) {
          currency = context.parse(options['currency'], 1, StringType);
          if (!currency)
              { return null; }
      }
      var minFractionDigits = null;
      if (options['min-fraction-digits']) {
          minFractionDigits = context.parse(options['min-fraction-digits'], 1, NumberType);
          if (!minFractionDigits)
              { return null; }
      }
      var maxFractionDigits = null;
      if (options['max-fraction-digits']) {
          maxFractionDigits = context.parse(options['max-fraction-digits'], 1, NumberType);
          if (!maxFractionDigits)
              { return null; }
      }
      return new NumberFormat(number, locale, currency, minFractionDigits, maxFractionDigits);
  };
  NumberFormat.prototype.evaluate = function evaluate (ctx) {
      return new Intl.NumberFormat(this.locale ? this.locale.evaluate(ctx) : [], {
          style: this.currency ? 'currency' : 'decimal',
          currency: this.currency ? this.currency.evaluate(ctx) : undefined,
          minimumFractionDigits: this.minFractionDigits ? this.minFractionDigits.evaluate(ctx) : undefined,
          maximumFractionDigits: this.maxFractionDigits ? this.maxFractionDigits.evaluate(ctx) : undefined
      }).format(this.number.evaluate(ctx));
  };
  NumberFormat.prototype.eachChild = function eachChild (fn) {
      fn(this.number);
      if (this.locale) {
          fn(this.locale);
      }
      if (this.currency) {
          fn(this.currency);
      }
      if (this.minFractionDigits) {
          fn(this.minFractionDigits);
      }
      if (this.maxFractionDigits) {
          fn(this.maxFractionDigits);
      }
  };
  NumberFormat.prototype.outputDefined = function outputDefined () {
      return false;
  };
  NumberFormat.prototype.serialize = function serialize () {
      var options = {};
      if (this.locale) {
          options['locale'] = this.locale.serialize();
      }
      if (this.currency) {
          options['currency'] = this.currency.serialize();
      }
      if (this.minFractionDigits) {
          options['min-fraction-digits'] = this.minFractionDigits.serialize();
      }
      if (this.maxFractionDigits) {
          options['max-fraction-digits'] = this.maxFractionDigits.serialize();
      }
      return [
          'number-format',
          this.number.serialize(),
          options
      ];
  };

  //      
  var Length = function Length(input) {
      this.type = NumberType;
      this.input = input;
  };
  Length.parse = function parse (args, context) {
      if (args.length !== 2)
          { return context.error(("Expected 1 argument, but found " + (args.length - 1) + " instead.")); }
      var input = context.parse(args[1], 1);
      if (!input)
          { return null; }
      if (input.type.kind !== 'array' && input.type.kind !== 'string' && input.type.kind !== 'value')
          { return context.error(("Expected argument of type string or array, but found " + (toString$1(input.type)) + " instead.")); }
      return new Length(input);
  };
  Length.prototype.evaluate = function evaluate (ctx) {
      var input = this.input.evaluate(ctx);
      if (typeof input === 'string') {
          return input.length;
      } else if (Array.isArray(input)) {
          return input.length;
      } else {
          throw new RuntimeError(("Expected value to be of type string or array, but found " + (toString$1(typeOf(input))) + " instead."));
      }
  };
  Length.prototype.eachChild = function eachChild (fn) {
      fn(this.input);
  };
  Length.prototype.outputDefined = function outputDefined () {
      return false;
  };
  Length.prototype.serialize = function serialize () {
      var serialized = ['length'];
      this.eachChild(function (child) {
          serialized.push(child.serialize());
      });
      return serialized;
  };

  //      
  var expressions = {
      // special forms
      '==': Equals,
      '!=': NotEquals,
      '>': GreaterThan$1,
      '<': LessThan$1,
      '>=': GreaterThanOrEqual,
      '<=': LessThanOrEqual,
      'array': Assertion,
      'at': At,
      'boolean': Assertion,
      'case': Case,
      'coalesce': Coalesce,
      'collator': CollatorExpression,
      'format': FormatExpression,
      'image': ImageExpression,
      'in': In,
      'index-of': IndexOf,
      'interpolate': Interpolate,
      'interpolate-hcl': Interpolate,
      'interpolate-lab': Interpolate,
      'length': Length,
      'let': Let,
      'literal': Literal,
      'match': Match,
      'number': Assertion,
      'number-format': NumberFormat,
      'object': Assertion,
      'slice': Slice,
      'step': Step,
      'string': Assertion,
      'to-boolean': Coercion,
      'to-color': Coercion,
      'to-number': Coercion,
      'to-string': Coercion,
      'var': Var,
      'within': Within$1
  };
  function rgba(ctx, ref) {
      var r = ref[0];
      var g = ref[1];
      var b = ref[2];
      var a = ref[3];

      r = r.evaluate(ctx);
      g = g.evaluate(ctx);
      b = b.evaluate(ctx);
      var alpha = a ? a.evaluate(ctx) : 1;
      var error = validateRGBA(r, g, b, alpha);
      if (error)
          { throw new RuntimeError(error); }
      return new Color(r / 255 * alpha, g / 255 * alpha, b / 255 * alpha, alpha);
  }
  function has(key, obj) {
      return key in obj;
  }
  function get(key, obj) {
      var v = obj[key];
      return typeof v === 'undefined' ? null : v;
  }
  function binarySearch(v, a, i, j) {
      while (i <= j) {
          var m = i + j >> 1;
          if (a[m] === v)
              { return true; }
          if (a[m] > v)
              { j = m - 1; }
          else
              { i = m + 1; }
      }
      return false;
  }
  function varargs(type) {
      return { type: type };
  }
  CompoundExpression.register(expressions, {
      'error': [
          ErrorType,
          [StringType],
          function (ctx, ref) {
              var v = ref[0];

              throw new RuntimeError(v.evaluate(ctx));
          }
      ],
      'typeof': [
          StringType,
          [ValueType],
          function (ctx, ref) {
              var v = ref[0];

              return toString$1(typeOf(v.evaluate(ctx)));
  }
      ],
      'to-rgba': [
          array$1(NumberType, 4),
          [ColorType],
          function (ctx, ref) {
              var v = ref[0];

              return v.evaluate(ctx).toArray();
          }
      ],
      'rgb': [
          ColorType,
          [
              NumberType,
              NumberType,
              NumberType
          ],
          rgba
      ],
      'rgba': [
          ColorType,
          [
              NumberType,
              NumberType,
              NumberType,
              NumberType
          ],
          rgba
      ],
      'has': {
          type: BooleanType,
          overloads: [
              [
                  [StringType],
                  function (ctx, ref) {
                      var key = ref[0];

                      return has(key.evaluate(ctx), ctx.properties());
  }
              ],
              [
                  [
                      StringType,
                      ObjectType
                  ],
                  function (ctx, ref) {
                      var key = ref[0];
                      var obj = ref[1];

                      return has(key.evaluate(ctx), obj.evaluate(ctx));
  }
              ]
          ]
      },
      'get': {
          type: ValueType,
          overloads: [
              [
                  [StringType],
                  function (ctx, ref) {
                      var key = ref[0];

                      return get(key.evaluate(ctx), ctx.properties());
  }
              ],
              [
                  [
                      StringType,
                      ObjectType
                  ],
                  function (ctx, ref) {
                      var key = ref[0];
                      var obj = ref[1];

                      return get(key.evaluate(ctx), obj.evaluate(ctx));
  }
              ]
          ]
      },
      'feature-state': [
          ValueType,
          [StringType],
          function (ctx, ref) {
              var key = ref[0];

              return get(key.evaluate(ctx), ctx.featureState || {});
  }
      ],
      'properties': [
          ObjectType,
          [],
          function (ctx) { return ctx.properties(); }
      ],
      'geometry-type': [
          StringType,
          [],
          function (ctx) { return ctx.geometryType(); }
      ],
      'id': [
          ValueType,
          [],
          function (ctx) { return ctx.id(); }
      ],
      'zoom': [
          NumberType,
          [],
          function (ctx) { return ctx.globals.zoom; }
      ],
      'pitch': [
          NumberType,
          [],
          function (ctx) { return ctx.globals.pitch || 0; }
      ],
      'distance-from-center': [
          NumberType,
          [],
          function (ctx) { return ctx.distanceFromCenter(); }
      ],
      'heatmap-density': [
          NumberType,
          [],
          function (ctx) { return ctx.globals.heatmapDensity || 0; }
      ],
      'line-progress': [
          NumberType,
          [],
          function (ctx) { return ctx.globals.lineProgress || 0; }
      ],
      'sky-radial-progress': [
          NumberType,
          [],
          function (ctx) { return ctx.globals.skyRadialProgress || 0; }
      ],
      'accumulated': [
          ValueType,
          [],
          function (ctx) { return ctx.globals.accumulated === undefined ? null : ctx.globals.accumulated; }
      ],
      '+': [
          NumberType,
          varargs(NumberType),
          function (ctx, args) {
              var result = 0;
              for (var i = 0, list = args; i < list.length; i += 1) {
                  var arg = list[i];

                  result += arg.evaluate(ctx);
              }
              return result;
          }
      ],
      '*': [
          NumberType,
          varargs(NumberType),
          function (ctx, args) {
              var result = 1;
              for (var i = 0, list = args; i < list.length; i += 1) {
                  var arg = list[i];

                  result *= arg.evaluate(ctx);
              }
              return result;
          }
      ],
      '-': {
          type: NumberType,
          overloads: [
              [
                  [
                      NumberType,
                      NumberType
                  ],
                  function (ctx, ref) {
                      var a = ref[0];
                      var b = ref[1];

                      return a.evaluate(ctx) - b.evaluate(ctx);
  }
              ],
              [
                  [NumberType],
                  function (ctx, ref) {
                      var a = ref[0];

                      return -a.evaluate(ctx);
  }
              ]
          ]
      },
      '/': [
          NumberType,
          [
              NumberType,
              NumberType
          ],
          function (ctx, ref) {
              var a = ref[0];
              var b = ref[1];

              return a.evaluate(ctx) / b.evaluate(ctx);
  }
      ],
      '%': [
          NumberType,
          [
              NumberType,
              NumberType
          ],
          function (ctx, ref) {
              var a = ref[0];
              var b = ref[1];

              return a.evaluate(ctx) % b.evaluate(ctx);
  }
      ],
      'ln2': [
          NumberType,
          [],
          function () { return Math.LN2; }
      ],
      'pi': [
          NumberType,
          [],
          function () { return Math.PI; }
      ],
      'e': [
          NumberType,
          [],
          function () { return Math.E; }
      ],
      '^': [
          NumberType,
          [
              NumberType,
              NumberType
          ],
          function (ctx, ref) {
              var b = ref[0];
              var e = ref[1];

              return Math.pow(b.evaluate(ctx), e.evaluate(ctx));
  }
      ],
      'sqrt': [
          NumberType,
          [NumberType],
          function (ctx, ref) {
              var x = ref[0];

              return Math.sqrt(x.evaluate(ctx));
  }
      ],
      'log10': [
          NumberType,
          [NumberType],
          function (ctx, ref) {
              var n = ref[0];

              return Math.log(n.evaluate(ctx)) / Math.LN10;
  }
      ],
      'ln': [
          NumberType,
          [NumberType],
          function (ctx, ref) {
              var n = ref[0];

              return Math.log(n.evaluate(ctx));
  }
      ],
      'log2': [
          NumberType,
          [NumberType],
          function (ctx, ref) {
              var n = ref[0];

              return Math.log(n.evaluate(ctx)) / Math.LN2;
  }
      ],
      'sin': [
          NumberType,
          [NumberType],
          function (ctx, ref) {
              var n = ref[0];

              return Math.sin(n.evaluate(ctx));
  }
      ],
      'cos': [
          NumberType,
          [NumberType],
          function (ctx, ref) {
              var n = ref[0];

              return Math.cos(n.evaluate(ctx));
  }
      ],
      'tan': [
          NumberType,
          [NumberType],
          function (ctx, ref) {
              var n = ref[0];

              return Math.tan(n.evaluate(ctx));
  }
      ],
      'asin': [
          NumberType,
          [NumberType],
          function (ctx, ref) {
              var n = ref[0];

              return Math.asin(n.evaluate(ctx));
  }
      ],
      'acos': [
          NumberType,
          [NumberType],
          function (ctx, ref) {
              var n = ref[0];

              return Math.acos(n.evaluate(ctx));
  }
      ],
      'atan': [
          NumberType,
          [NumberType],
          function (ctx, ref) {
              var n = ref[0];

              return Math.atan(n.evaluate(ctx));
  }
      ],
      'min': [
          NumberType,
          varargs(NumberType),
          function (ctx, args) { return Math.min.apply(Math, args.map(function (arg) { return arg.evaluate(ctx); })); }
      ],
      'max': [
          NumberType,
          varargs(NumberType),
          function (ctx, args) { return Math.max.apply(Math, args.map(function (arg) { return arg.evaluate(ctx); })); }
      ],
      'abs': [
          NumberType,
          [NumberType],
          function (ctx, ref) {
              var n = ref[0];

              return Math.abs(n.evaluate(ctx));
  }
      ],
      'round': [
          NumberType,
          [NumberType],
          function (ctx, ref) {
              var n = ref[0];

              var v = n.evaluate(ctx);
              // Javascript's Math.round() rounds towards +Infinity for halfway
              // values, even when they're negative. It's more common to round
              // away from 0 (e.g., this is what python and C++ do)
              return v < 0 ? -Math.round(-v) : Math.round(v);
          }
      ],
      'floor': [
          NumberType,
          [NumberType],
          function (ctx, ref) {
              var n = ref[0];

              return Math.floor(n.evaluate(ctx));
  }
      ],
      'ceil': [
          NumberType,
          [NumberType],
          function (ctx, ref) {
              var n = ref[0];

              return Math.ceil(n.evaluate(ctx));
  }
      ],
      'filter-==': [
          BooleanType,
          [
              StringType,
              ValueType
          ],
          function (ctx, ref) {
              var k = ref[0];
              var v = ref[1];

              return ctx.properties()[k.value] === v.value;
  }
      ],
      'filter-id-==': [
          BooleanType,
          [ValueType],
          function (ctx, ref) {
              var v = ref[0];

              return ctx.id() === v.value;
  }
      ],
      'filter-type-==': [
          BooleanType,
          [StringType],
          function (ctx, ref) {
              var v = ref[0];

              return ctx.geometryType() === v.value;
  }
      ],
      'filter-<': [
          BooleanType,
          [
              StringType,
              ValueType
          ],
          function (ctx, ref) {
              var k = ref[0];
              var v = ref[1];

              var a = ctx.properties()[k.value];
              var b = v.value;
              return typeof a === typeof b && a < b;
          }
      ],
      'filter-id-<': [
          BooleanType,
          [ValueType],
          function (ctx, ref) {
              var v = ref[0];

              var a = ctx.id();
              var b = v.value;
              return typeof a === typeof b && a < b;
          }
      ],
      'filter->': [
          BooleanType,
          [
              StringType,
              ValueType
          ],
          function (ctx, ref) {
              var k = ref[0];
              var v = ref[1];

              var a = ctx.properties()[k.value];
              var b = v.value;
              return typeof a === typeof b && a > b;
          }
      ],
      'filter-id->': [
          BooleanType,
          [ValueType],
          function (ctx, ref) {
              var v = ref[0];

              var a = ctx.id();
              var b = v.value;
              return typeof a === typeof b && a > b;
          }
      ],
      'filter-<=': [
          BooleanType,
          [
              StringType,
              ValueType
          ],
          function (ctx, ref) {
              var k = ref[0];
              var v = ref[1];

              var a = ctx.properties()[k.value];
              var b = v.value;
              return typeof a === typeof b && a <= b;
          }
      ],
      'filter-id-<=': [
          BooleanType,
          [ValueType],
          function (ctx, ref) {
              var v = ref[0];

              var a = ctx.id();
              var b = v.value;
              return typeof a === typeof b && a <= b;
          }
      ],
      'filter->=': [
          BooleanType,
          [
              StringType,
              ValueType
          ],
          function (ctx, ref) {
              var k = ref[0];
              var v = ref[1];

              var a = ctx.properties()[k.value];
              var b = v.value;
              return typeof a === typeof b && a >= b;
          }
      ],
      'filter-id->=': [
          BooleanType,
          [ValueType],
          function (ctx, ref) {
              var v = ref[0];

              var a = ctx.id();
              var b = v.value;
              return typeof a === typeof b && a >= b;
          }
      ],
      'filter-has': [
          BooleanType,
          [ValueType],
          function (ctx, ref) {
              var k = ref[0];

              return k.value in ctx.properties();
  }
      ],
      'filter-has-id': [
          BooleanType,
          [],
          function (ctx) { return ctx.id() !== null && ctx.id() !== undefined; }
      ],
      'filter-type-in': [
          BooleanType,
          [array$1(StringType)],
          function (ctx, ref) {
              var v = ref[0];

              return v.value.indexOf(ctx.geometryType()) >= 0;
  }
      ],
      'filter-id-in': [
          BooleanType,
          [array$1(ValueType)],
          function (ctx, ref) {
              var v = ref[0];

              return v.value.indexOf(ctx.id()) >= 0;
  }
      ],
      'filter-in-small': [
          BooleanType,
          [
              StringType,
              array$1(ValueType)
          ],
          // assumes v is an array literal
          function (ctx, ref) {
              var k = ref[0];
              var v = ref[1];

              return v.value.indexOf(ctx.properties()[k.value]) >= 0;
  }
      ],
      'filter-in-large': [
          BooleanType,
          [
              StringType,
              array$1(ValueType)
          ],
          // assumes v is a array literal with values sorted in ascending order and of a single type
          function (ctx, ref) {
              var k = ref[0];
              var v = ref[1];

              return binarySearch(ctx.properties()[k.value], v.value, 0, v.value.length - 1);
  }
      ],
      'all': {
          type: BooleanType,
          overloads: [
              [
                  [
                      BooleanType,
                      BooleanType
                  ],
                  function (ctx, ref) {
                      var a = ref[0];
                      var b = ref[1];

                      return a.evaluate(ctx) && b.evaluate(ctx);
  }
              ],
              [
                  varargs(BooleanType),
                  function (ctx, args) {
                      for (var i = 0, list = args; i < list.length; i += 1) {
                          var arg = list[i];

                          if (!arg.evaluate(ctx))
                              { return false; }
                      }
                      return true;
                  }
              ]
          ]
      },
      'any': {
          type: BooleanType,
          overloads: [
              [
                  [
                      BooleanType,
                      BooleanType
                  ],
                  function (ctx, ref) {
                      var a = ref[0];
                      var b = ref[1];

                      return a.evaluate(ctx) || b.evaluate(ctx);
  }
              ],
              [
                  varargs(BooleanType),
                  function (ctx, args) {
                      for (var i = 0, list = args; i < list.length; i += 1) {
                          var arg = list[i];

                          if (arg.evaluate(ctx))
                              { return true; }
                      }
                      return false;
                  }
              ]
          ]
      },
      '!': [
          BooleanType,
          [BooleanType],
          function (ctx, ref) {
              var b = ref[0];

              return !b.evaluate(ctx);
  }
      ],
      'is-supported-script': [
          BooleanType,
          [StringType],
          // At parse time this will always return true, so we need to exclude this expression with isGlobalPropertyConstant
          function (ctx, ref) {
              var s = ref[0];

              var isSupportedScript = ctx.globals && ctx.globals.isSupportedScript;
              if (isSupportedScript) {
                  return isSupportedScript(s.evaluate(ctx));
              }
              return true;
          }
      ],
      'upcase': [
          StringType,
          [StringType],
          function (ctx, ref) {
              var s = ref[0];

              return s.evaluate(ctx).toUpperCase();
  }
      ],
      'downcase': [
          StringType,
          [StringType],
          function (ctx, ref) {
              var s = ref[0];

              return s.evaluate(ctx).toLowerCase();
  }
      ],
      'concat': [
          StringType,
          varargs(ValueType),
          function (ctx, args) { return args.map(function (arg) { return toString(arg.evaluate(ctx)); }).join(''); }
      ],
      'resolved-locale': [
          StringType,
          [CollatorType],
          function (ctx, ref) {
              var collator = ref[0];

              return collator.evaluate(ctx).resolvedLocale();
  }
      ]
  });

  //      
  /**
   * A type used for returning and propagating errors. The first element of the union
   * represents success and contains a value, and the second represents an error and
   * contains an error value.
   * @private
   */
  function success(value) {
      return {
          result: 'success',
          value: value
      };
  }
  function error(value) {
      return {
          result: 'error',
          value: value
      };
  }

  //      
  function supportsPropertyExpression(spec) {
      return spec['property-type'] === 'data-driven' || spec['property-type'] === 'cross-faded-data-driven';
  }
  function supportsZoomExpression(spec) {
      return !!spec.expression && spec.expression.parameters.indexOf('zoom') > -1;
  }
  function supportsInterpolation(spec) {
      return !!spec.expression && spec.expression.interpolated;
  }

  function isFunction(value) {
      return typeof value === 'object' && value !== null && !Array.isArray(value);
  }

  var StyleExpression = function StyleExpression(expression, propertySpec) {
      this.expression = expression;
      this._warningHistory = {};
      this._evaluator = new EvaluationContext();
      this._defaultValue = propertySpec ? getDefaultValue(propertySpec) : null;
      this._enumValues = propertySpec && propertySpec.type === 'enum' ? propertySpec.values : null;
  };
  StyleExpression.prototype.evaluateWithoutErrorHandling = function evaluateWithoutErrorHandling (globals, feature, featureState, canonical, availableImages, formattedSection, featureTileCoord, featureDistanceData) {
      this._evaluator.globals = globals;
      this._evaluator.feature = feature;
      this._evaluator.featureState = featureState;
      this._evaluator.canonical = canonical || null;
      this._evaluator.availableImages = availableImages || null;
      this._evaluator.formattedSection = formattedSection;
      this._evaluator.featureTileCoord = featureTileCoord || null;
      this._evaluator.featureDistanceData = featureDistanceData || null;
      return this.expression.evaluate(this._evaluator);
  };
  StyleExpression.prototype.evaluate = function evaluate (globals, feature, featureState, canonical, availableImages, formattedSection, featureTileCoord, featureDistanceData) {
      this._evaluator.globals = globals;
      this._evaluator.feature = feature || null;
      this._evaluator.featureState = featureState || null;
      this._evaluator.canonical = canonical || null;
      this._evaluator.availableImages = availableImages || null;
      this._evaluator.formattedSection = formattedSection || null;
      this._evaluator.featureTileCoord = featureTileCoord || null;
      this._evaluator.featureDistanceData = featureDistanceData || null;
      try {
          var val = this.expression.evaluate(this._evaluator);
          // eslint-disable-next-line no-self-compare
          if (val === null || val === undefined || typeof val === 'number' && val !== val) {
              return this._defaultValue;
          }
          if (this._enumValues && !(val in this._enumValues)) {
              throw new RuntimeError(("Expected value to be one of " + (Object.keys(this._enumValues).map(function (v) { return JSON.stringify(v); }).join(', ')) + ", but found " + (JSON.stringify(val)) + " instead."));
          }
          return val;
      } catch (e) {
          if (!this._warningHistory[e.message]) {
              this._warningHistory[e.message] = true;
              if (typeof console !== 'undefined') {
                  console.warn(e.message);
              }
          }
          return this._defaultValue;
      }
  };
  function isExpression(expression) {
      return Array.isArray(expression) && expression.length > 0 && typeof expression[0] === 'string' && expression[0] in expressions;
  }
  /**
   * Parse and typecheck the given style spec JSON expression.  If
   * options.defaultValue is provided, then the resulting StyleExpression's
   * `evaluate()` method will handle errors by logging a warning (once per
   * message) and returning the default value.  Otherwise, it will throw
   * evaluation errors.
   *
   * @private
   */
  function createExpression(expression, propertySpec) {
      var parser = new ParsingContext(expressions, [], propertySpec ? getExpectedType(propertySpec) : undefined);
      // For string-valued properties, coerce to string at the top level rather than asserting.
      var parsed = parser.parse(expression, undefined, undefined, undefined, propertySpec && propertySpec.type === 'string' ? { typeAnnotation: 'coerce' } : undefined);
      if (!parsed) {
          return error(parser.errors);
      }
      return success(new StyleExpression(parsed, propertySpec));
  }
  var ZoomConstantExpression = function ZoomConstantExpression(kind, expression) {
      this.kind = kind;
      this._styleExpression = expression;
      this.isStateDependent = kind !== 'constant' && !isStateConstant(expression.expression);
  };
  ZoomConstantExpression.prototype.evaluateWithoutErrorHandling = function evaluateWithoutErrorHandling (globals, feature, featureState, canonical, availableImages, formattedSection) {
      return this._styleExpression.evaluateWithoutErrorHandling(globals, feature, featureState, canonical, availableImages, formattedSection);
  };
  ZoomConstantExpression.prototype.evaluate = function evaluate (globals, feature, featureState, canonical, availableImages, formattedSection) {
      return this._styleExpression.evaluate(globals, feature, featureState, canonical, availableImages, formattedSection);
  };
  var ZoomDependentExpression = function ZoomDependentExpression(kind, expression, zoomStops, interpolationType) {
      this.kind = kind;
      this.zoomStops = zoomStops;
      this._styleExpression = expression;
      this.isStateDependent = kind !== 'camera' && !isStateConstant(expression.expression);
      this.interpolationType = interpolationType;
  };
  ZoomDependentExpression.prototype.evaluateWithoutErrorHandling = function evaluateWithoutErrorHandling (globals, feature, featureState, canonical, availableImages, formattedSection) {
      return this._styleExpression.evaluateWithoutErrorHandling(globals, feature, featureState, canonical, availableImages, formattedSection);
  };
  ZoomDependentExpression.prototype.evaluate = function evaluate (globals, feature, featureState, canonical, availableImages, formattedSection) {
      return this._styleExpression.evaluate(globals, feature, featureState, canonical, availableImages, formattedSection);
  };
  ZoomDependentExpression.prototype.interpolationFactor = function interpolationFactor (input, lower, upper) {
      if (this.interpolationType) {
          return Interpolate.interpolationFactor(this.interpolationType, input, lower, upper);
      } else {
          return 0;
      }
  };
  function createPropertyExpression(expression, propertySpec) {
      expression = createExpression(expression, propertySpec);
      if (expression.result === 'error') {
          return expression;
      }
      var parsed = expression.value.expression;
      var isFeatureConstant$1 = isFeatureConstant(parsed);
      if (!isFeatureConstant$1 && !supportsPropertyExpression(propertySpec)) {
          return error([new ParsingError('', 'data expressions not supported')]);
      }
      var isZoomConstant = isGlobalPropertyConstant(parsed, [
          'zoom',
          'pitch',
          'distance-from-center'
      ]);
      if (!isZoomConstant && !supportsZoomExpression(propertySpec)) {
          return error([new ParsingError('', 'zoom expressions not supported')]);
      }
      var zoomCurve = findZoomCurve(parsed);
      if (!zoomCurve && !isZoomConstant) {
          return error([new ParsingError('', '"zoom" expression may only be used as input to a top-level "step" or "interpolate" expression.')]);
      } else if (zoomCurve instanceof ParsingError) {
          return error([zoomCurve]);
      } else if (zoomCurve instanceof Interpolate && !supportsInterpolation(propertySpec)) {
          return error([new ParsingError('', '"interpolate" expressions cannot be used with this property')]);
      }
      if (!zoomCurve) {
          return success(isFeatureConstant$1 ? new ZoomConstantExpression('constant', expression.value) : new ZoomConstantExpression('source', expression.value));
      }
      var interpolationType = zoomCurve instanceof Interpolate ? zoomCurve.interpolation : undefined;
      return success(isFeatureConstant$1 ? new ZoomDependentExpression('camera', expression.value, zoomCurve.labels, interpolationType) : new ZoomDependentExpression('composite', expression.value, zoomCurve.labels, interpolationType));
  }
  // Zoom-dependent expressions may only use ["zoom"] as the input to a top-level "step" or "interpolate"
  // expression (collectively referred to as a "curve"). The curve may be wrapped in one or more "let" or
  // "coalesce" expressions.
  function findZoomCurve(expression) {
      var result = null;
      if (expression instanceof Let) {
          result = findZoomCurve(expression.result);
      } else if (expression instanceof Coalesce) {
          for (var i = 0, list = expression.args; i < list.length; i += 1) {
              var arg = list[i];

              result = findZoomCurve(arg);
              if (result) {
                  break;
              }
          }
      } else if ((expression instanceof Step || expression instanceof Interpolate) && expression.input instanceof CompoundExpression && expression.input.name === 'zoom') {
          result = expression;
      }
      if (result instanceof ParsingError) {
          return result;
      }
      expression.eachChild(function (child) {
          var childResult = findZoomCurve(child);
          if (childResult instanceof ParsingError) {
              result = childResult;
          } else if (!result && childResult) {
              result = new ParsingError('', '"zoom" expression may only be used as input to a top-level "step" or "interpolate" expression.');
          } else if (result && childResult && result !== childResult) {
              result = new ParsingError('', 'Only one zoom-based "step" or "interpolate" subexpression may be used in an expression.');
          }
      });
      return result;
  }
  function getExpectedType(spec) {
      var types = {
          color: ColorType,
          string: StringType,
          number: NumberType,
          enum: StringType,
          boolean: BooleanType,
          formatted: FormattedType,
          resolvedImage: ResolvedImageType
      };
      if (spec.type === 'array') {
          return array$1(types[spec.value] || ValueType, spec.length);
      }
      return types[spec.type];
  }
  function getDefaultValue(spec) {
      if (spec.type === 'color' && (isFunction(spec.default) || Array.isArray(spec.default))) {
          // Special case for heatmap-color: it uses the 'default:' to define a
          // default color ramp, but createExpression expects a simple value to fall
          // back to in case of runtime errors
          return new Color(0, 0, 0, 0);
      } else if (spec.type === 'color') {
          return Color.parse(spec.default) || null;
      } else if (spec.default === undefined) {
          return null;
      } else {
          return spec.default;
      }
  }

  //      
  // Turn jsonlint-lines-primitives objects into primitive objects
  function unbundle(value) {
      if (value instanceof Number || value instanceof String || value instanceof Boolean) {
          return value.valueOf();
      } else {
          return value;
      }
  }
  function deepUnbundle(value) {
      if (Array.isArray(value)) {
          return value.map(deepUnbundle);
      } else if (value instanceof Object && !(value instanceof Number || value instanceof String || value instanceof Boolean)) {
          var unbundledValue = {};
          for (var key in value) {
              unbundledValue[key] = deepUnbundle(value[key]);
          }
          return unbundledValue;
      }
      return unbundle(value);
  }

  var spec = {"$version":8,"$root":{"version":{"required":true,"type":"enum","values":[8]},"name":{"type":"string"},"metadata":{"type":"*"},"center":{"type":"array","value":"number"},"zoom":{"type":"number"},"bearing":{"type":"number","default":0,"period":360,"units":"degrees"},"pitch":{"type":"number","default":0,"units":"degrees"},"light":{"type":"light"},"terrain":{"type":"terrain"},"fog":{"type":"fog"},"sources":{"required":true,"type":"sources"},"sprite":{"type":"string"},"glyphs":{"type":"string"},"transition":{"type":"transition"},"projection":{"type":"projection"},"layers":{"required":true,"type":"array","value":"layer"}},"sources":{"*":{"type":"source"}},"source":["source_vector","source_raster","source_raster_dem","source_geojson","source_video","source_image"],"source_vector":{"type":{"required":true,"type":"enum","values":{"vector":{}}},"url":{"type":"string"},"tiles":{"type":"array","value":"string"},"bounds":{"type":"array","value":"number","length":4,"default":[-180,-85.051129,180,85.051129]},"scheme":{"type":"enum","values":{"xyz":{},"tms":{}},"default":"xyz"},"minzoom":{"type":"number","default":0},"maxzoom":{"type":"number","default":22},"attribution":{"type":"string"},"promoteId":{"type":"promoteId"},"volatile":{"type":"boolean","default":false},"*":{"type":"*"}},"source_raster":{"type":{"required":true,"type":"enum","values":{"raster":{}}},"url":{"type":"string"},"tiles":{"type":"array","value":"string"},"bounds":{"type":"array","value":"number","length":4,"default":[-180,-85.051129,180,85.051129]},"minzoom":{"type":"number","default":0},"maxzoom":{"type":"number","default":22},"tileSize":{"type":"number","default":512,"units":"pixels"},"scheme":{"type":"enum","values":{"xyz":{},"tms":{}},"default":"xyz"},"attribution":{"type":"string"},"volatile":{"type":"boolean","default":false},"*":{"type":"*"}},"source_raster_dem":{"type":{"required":true,"type":"enum","values":{"raster-dem":{}}},"url":{"type":"string"},"tiles":{"type":"array","value":"string"},"bounds":{"type":"array","value":"number","length":4,"default":[-180,-85.051129,180,85.051129]},"minzoom":{"type":"number","default":0},"maxzoom":{"type":"number","default":22},"tileSize":{"type":"number","default":512,"units":"pixels"},"attribution":{"type":"string"},"encoding":{"type":"enum","values":{"terrarium":{},"mapbox":{}},"default":"mapbox"},"volatile":{"type":"boolean","default":false},"*":{"type":"*"}},"source_geojson":{"type":{"required":true,"type":"enum","values":{"geojson":{}}},"data":{"type":"*"},"maxzoom":{"type":"number","default":18},"attribution":{"type":"string"},"buffer":{"type":"number","default":128,"maximum":512,"minimum":0},"filter":{"type":"*"},"tolerance":{"type":"number","default":0.375},"cluster":{"type":"boolean","default":false},"clusterRadius":{"type":"number","default":50,"minimum":0},"clusterMaxZoom":{"type":"number"},"clusterMinPoints":{"type":"number"},"clusterProperties":{"type":"*"},"lineMetrics":{"type":"boolean","default":false},"generateId":{"type":"boolean","default":false},"promoteId":{"type":"promoteId"}},"source_video":{"type":{"required":true,"type":"enum","values":{"video":{}}},"urls":{"required":true,"type":"array","value":"string"},"coordinates":{"required":true,"type":"array","length":4,"value":{"type":"array","length":2,"value":"number"}}},"source_image":{"type":{"required":true,"type":"enum","values":{"image":{}}},"url":{"required":true,"type":"string"},"coordinates":{"required":true,"type":"array","length":4,"value":{"type":"array","length":2,"value":"number"}}},"layer":{"id":{"type":"string","required":true},"type":{"type":"enum","values":{"fill":{},"line":{},"symbol":{},"circle":{},"heatmap":{},"fill-extrusion":{},"raster":{},"hillshade":{},"background":{},"sky":{}},"required":true},"metadata":{"type":"*"},"source":{"type":"string"},"source-layer":{"type":"string"},"minzoom":{"type":"number","minimum":0,"maximum":24},"maxzoom":{"type":"number","minimum":0,"maximum":24},"filter":{"type":"filter"},"layout":{"type":"layout"},"paint":{"type":"paint"}},"layout":["layout_fill","layout_line","layout_circle","layout_heatmap","layout_fill-extrusion","layout_symbol","layout_raster","layout_hillshade","layout_background","layout_sky"],"layout_background":{"visibility":{"type":"enum","values":{"visible":{},"none":{}},"default":"visible","property-type":"constant"}},"layout_sky":{"visibility":{"type":"enum","values":{"visible":{},"none":{}},"default":"visible","property-type":"constant"}},"layout_fill":{"fill-sort-key":{"type":"number","expression":{"interpolated":false,"parameters":["zoom","feature"]},"property-type":"data-driven"},"visibility":{"type":"enum","values":{"visible":{},"none":{}},"default":"visible","property-type":"constant"}},"layout_circle":{"circle-sort-key":{"type":"number","expression":{"interpolated":false,"parameters":["zoom","feature"]},"property-type":"data-driven"},"visibility":{"type":"enum","values":{"visible":{},"none":{}},"default":"visible","property-type":"constant"}},"layout_heatmap":{"visibility":{"type":"enum","values":{"visible":{},"none":{}},"default":"visible","property-type":"constant"}},"layout_fill-extrusion":{"visibility":{"type":"enum","values":{"visible":{},"none":{}},"default":"visible","property-type":"constant"}},"layout_line":{"line-cap":{"type":"enum","values":{"butt":{},"round":{},"square":{}},"default":"butt","expression":{"interpolated":false,"parameters":["zoom","feature"]},"property-type":"data-driven"},"line-join":{"type":"enum","values":{"bevel":{},"round":{},"miter":{}},"default":"miter","expression":{"interpolated":false,"parameters":["zoom","feature"]},"property-type":"data-driven"},"line-miter-limit":{"type":"number","default":2,"requires":[{"line-join":"miter"}],"expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"},"line-round-limit":{"type":"number","default":1.05,"requires":[{"line-join":"round"}],"expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"},"line-sort-key":{"type":"number","expression":{"interpolated":false,"parameters":["zoom","feature"]},"property-type":"data-driven"},"visibility":{"type":"enum","values":{"visible":{},"none":{}},"default":"visible","property-type":"constant"}},"layout_symbol":{"symbol-placement":{"type":"enum","values":{"point":{},"line":{},"line-center":{}},"default":"point","expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"symbol-spacing":{"type":"number","default":250,"minimum":1,"units":"pixels","requires":[{"symbol-placement":"line"}],"expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"},"symbol-avoid-edges":{"type":"boolean","default":false,"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"symbol-sort-key":{"type":"number","expression":{"interpolated":false,"parameters":["zoom","feature"]},"property-type":"data-driven"},"symbol-z-order":{"type":"enum","values":{"auto":{},"viewport-y":{},"source":{}},"default":"auto","expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"icon-allow-overlap":{"type":"boolean","default":false,"requires":["icon-image"],"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"icon-ignore-placement":{"type":"boolean","default":false,"requires":["icon-image"],"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"icon-optional":{"type":"boolean","default":false,"requires":["icon-image","text-field"],"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"icon-rotation-alignment":{"type":"enum","values":{"map":{},"viewport":{},"auto":{}},"default":"auto","requires":["icon-image"],"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"icon-size":{"type":"number","default":1,"minimum":0,"units":"factor of the original icon size","requires":["icon-image"],"expression":{"interpolated":true,"parameters":["zoom","feature"]},"property-type":"data-driven"},"icon-text-fit":{"type":"enum","values":{"none":{},"width":{},"height":{},"both":{}},"default":"none","requires":["icon-image","text-field"],"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"icon-text-fit-padding":{"type":"array","value":"number","length":4,"default":[0,0,0,0],"units":"pixels","requires":["icon-image","text-field",{"icon-text-fit":["both","width","height"]}],"expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"},"icon-image":{"type":"resolvedImage","tokens":true,"expression":{"interpolated":false,"parameters":["zoom","feature"]},"property-type":"data-driven"},"icon-rotate":{"type":"number","default":0,"period":360,"units":"degrees","requires":["icon-image"],"expression":{"interpolated":true,"parameters":["zoom","feature"]},"property-type":"data-driven"},"icon-padding":{"type":"number","default":2,"minimum":0,"units":"pixels","requires":["icon-image"],"expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"},"icon-keep-upright":{"type":"boolean","default":false,"requires":["icon-image",{"icon-rotation-alignment":"map"},{"symbol-placement":["line","line-center"]}],"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"icon-offset":{"type":"array","value":"number","length":2,"default":[0,0],"requires":["icon-image"],"expression":{"interpolated":true,"parameters":["zoom","feature"]},"property-type":"data-driven"},"icon-anchor":{"type":"enum","values":{"center":{},"left":{},"right":{},"top":{},"bottom":{},"top-left":{},"top-right":{},"bottom-left":{},"bottom-right":{}},"default":"center","requires":["icon-image"],"expression":{"interpolated":false,"parameters":["zoom","feature"]},"property-type":"data-driven"},"icon-pitch-alignment":{"type":"enum","values":{"map":{},"viewport":{},"auto":{}},"default":"auto","requires":["icon-image"],"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"text-pitch-alignment":{"type":"enum","values":{"map":{},"viewport":{},"auto":{}},"default":"auto","requires":["text-field"],"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"text-rotation-alignment":{"type":"enum","values":{"map":{},"viewport":{},"auto":{}},"default":"auto","requires":["text-field"],"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"text-field":{"type":"formatted","default":"","tokens":true,"expression":{"interpolated":false,"parameters":["zoom","feature"]},"property-type":"data-driven"},"text-font":{"type":"array","value":"string","default":["Open Sans Regular","Arial Unicode MS Regular"],"requires":["text-field"],"expression":{"interpolated":false,"parameters":["zoom","feature"]},"property-type":"data-driven"},"text-size":{"type":"number","default":16,"minimum":0,"units":"pixels","requires":["text-field"],"expression":{"interpolated":true,"parameters":["zoom","feature"]},"property-type":"data-driven"},"text-max-width":{"type":"number","default":10,"minimum":0,"units":"ems","requires":["text-field",{"symbol-placement":["point"]}],"expression":{"interpolated":true,"parameters":["zoom","feature"]},"property-type":"data-driven"},"text-line-height":{"type":"number","default":1.2,"units":"ems","requires":["text-field"],"expression":{"interpolated":true,"parameters":["zoom","feature"]},"property-type":"data-driven"},"text-letter-spacing":{"type":"number","default":0,"units":"ems","requires":["text-field"],"expression":{"interpolated":true,"parameters":["zoom","feature"]},"property-type":"data-driven"},"text-justify":{"type":"enum","values":{"auto":{},"left":{},"center":{},"right":{}},"default":"center","requires":["text-field"],"expression":{"interpolated":false,"parameters":["zoom","feature"]},"property-type":"data-driven"},"text-radial-offset":{"type":"number","units":"ems","default":0,"requires":["text-field"],"property-type":"data-driven","expression":{"interpolated":true,"parameters":["zoom","feature"]}},"text-variable-anchor":{"type":"array","value":"enum","values":{"center":{},"left":{},"right":{},"top":{},"bottom":{},"top-left":{},"top-right":{},"bottom-left":{},"bottom-right":{}},"requires":["text-field",{"symbol-placement":["point"]}],"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"text-anchor":{"type":"enum","values":{"center":{},"left":{},"right":{},"top":{},"bottom":{},"top-left":{},"top-right":{},"bottom-left":{},"bottom-right":{}},"default":"center","requires":["text-field",{"!":"text-variable-anchor"}],"expression":{"interpolated":false,"parameters":["zoom","feature"]},"property-type":"data-driven"},"text-max-angle":{"type":"number","default":45,"units":"degrees","requires":["text-field",{"symbol-placement":["line","line-center"]}],"expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"},"text-writing-mode":{"type":"array","value":"enum","values":{"horizontal":{},"vertical":{}},"requires":["text-field"],"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"text-rotate":{"type":"number","default":0,"period":360,"units":"degrees","requires":["text-field"],"expression":{"interpolated":true,"parameters":["zoom","feature"]},"property-type":"data-driven"},"text-padding":{"type":"number","default":2,"minimum":0,"units":"pixels","requires":["text-field"],"expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"},"text-keep-upright":{"type":"boolean","default":true,"requires":["text-field",{"text-rotation-alignment":"map"},{"symbol-placement":["line","line-center"]}],"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"text-transform":{"type":"enum","values":{"none":{},"uppercase":{},"lowercase":{}},"default":"none","requires":["text-field"],"expression":{"interpolated":false,"parameters":["zoom","feature"]},"property-type":"data-driven"},"text-offset":{"type":"array","value":"number","units":"ems","length":2,"default":[0,0],"requires":["text-field",{"!":"text-radial-offset"}],"expression":{"interpolated":true,"parameters":["zoom","feature"]},"property-type":"data-driven"},"text-allow-overlap":{"type":"boolean","default":false,"requires":["text-field"],"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"text-ignore-placement":{"type":"boolean","default":false,"requires":["text-field"],"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"text-optional":{"type":"boolean","default":false,"requires":["text-field","icon-image"],"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"visibility":{"type":"enum","values":{"visible":{},"none":{}},"default":"visible","property-type":"constant"}},"layout_raster":{"visibility":{"type":"enum","values":{"visible":{},"none":{}},"default":"visible","property-type":"constant"}},"layout_hillshade":{"visibility":{"type":"enum","values":{"visible":{},"none":{}},"default":"visible","property-type":"constant"}},"filter":{"type":"array","value":"*"},"filter_symbol":{"type":"boolean","default":false,"transition":false,"property-type":"data-driven","expression":{"interpolated":false,"parameters":["zoom","feature","pitch","distance-from-center"]}},"filter_fill":{"type":"boolean","default":false,"transition":false,"property-type":"data-driven","expression":{"interpolated":false,"parameters":["zoom","feature"]}},"filter_line":{"type":"boolean","default":false,"transition":false,"property-type":"data-driven","expression":{"interpolated":false,"parameters":["zoom","feature"]}},"filter_circle":{"type":"boolean","default":false,"transition":false,"property-type":"data-driven","expression":{"interpolated":false,"parameters":["zoom","feature"]}},"filter_fill-extrusion":{"type":"boolean","default":false,"transition":false,"property-type":"data-driven","expression":{"interpolated":false,"parameters":["zoom","feature"]}},"filter_heatmap":{"type":"boolean","default":false,"transition":false,"property-type":"data-driven","expression":{"interpolated":false,"parameters":["zoom","feature"]}},"filter_operator":{"type":"enum","values":{"==":{},"!=":{},">":{},">=":{},"<":{},"<=":{},"in":{},"!in":{},"all":{},"any":{},"none":{},"has":{},"!has":{},"within":{}}},"geometry_type":{"type":"enum","values":{"Point":{},"LineString":{},"Polygon":{}}},"function":{"expression":{"type":"expression"},"stops":{"type":"array","value":"function_stop"},"base":{"type":"number","default":1,"minimum":0},"property":{"type":"string","default":"$zoom"},"type":{"type":"enum","values":{"identity":{},"exponential":{},"interval":{},"categorical":{}},"default":"exponential"},"colorSpace":{"type":"enum","values":{"rgb":{},"lab":{},"hcl":{}},"default":"rgb"},"default":{"type":"*","required":false}},"function_stop":{"type":"array","minimum":0,"maximum":24,"value":["number","color"],"length":2},"expression":{"type":"array","value":"*","minimum":1},"expression_name":{"type":"enum","values":{"let":{"group":"Variable binding"},"var":{"group":"Variable binding"},"literal":{"group":"Types"},"array":{"group":"Types"},"at":{"group":"Lookup"},"in":{"group":"Lookup"},"index-of":{"group":"Lookup"},"slice":{"group":"Lookup"},"case":{"group":"Decision"},"match":{"group":"Decision"},"coalesce":{"group":"Decision"},"step":{"group":"Ramps, scales, curves"},"interpolate":{"group":"Ramps, scales, curves"},"interpolate-hcl":{"group":"Ramps, scales, curves"},"interpolate-lab":{"group":"Ramps, scales, curves"},"ln2":{"group":"Math"},"pi":{"group":"Math"},"e":{"group":"Math"},"typeof":{"group":"Types"},"string":{"group":"Types"},"number":{"group":"Types"},"boolean":{"group":"Types"},"object":{"group":"Types"},"collator":{"group":"Types"},"format":{"group":"Types"},"image":{"group":"Types"},"number-format":{"group":"Types"},"to-string":{"group":"Types"},"to-number":{"group":"Types"},"to-boolean":{"group":"Types"},"to-rgba":{"group":"Color"},"to-color":{"group":"Types"},"rgb":{"group":"Color"},"rgba":{"group":"Color"},"get":{"group":"Lookup"},"has":{"group":"Lookup"},"length":{"group":"Lookup"},"properties":{"group":"Feature data"},"feature-state":{"group":"Feature data"},"geometry-type":{"group":"Feature data"},"id":{"group":"Feature data"},"zoom":{"group":"Camera"},"pitch":{"group":"Camera"},"distance-from-center":{"group":"Camera"},"heatmap-density":{"group":"Heatmap"},"line-progress":{"group":"Feature data"},"sky-radial-progress":{"group":"sky"},"accumulated":{"group":"Feature data"},"+":{"group":"Math"},"*":{"group":"Math"},"-":{"group":"Math"},"/":{"group":"Math"},"%":{"group":"Math"},"^":{"group":"Math"},"sqrt":{"group":"Math"},"log10":{"group":"Math"},"ln":{"group":"Math"},"log2":{"group":"Math"},"sin":{"group":"Math"},"cos":{"group":"Math"},"tan":{"group":"Math"},"asin":{"group":"Math"},"acos":{"group":"Math"},"atan":{"group":"Math"},"min":{"group":"Math"},"max":{"group":"Math"},"round":{"group":"Math"},"abs":{"group":"Math"},"ceil":{"group":"Math"},"floor":{"group":"Math"},"distance":{"group":"Math"},"==":{"group":"Decision"},"!=":{"group":"Decision"},">":{"group":"Decision"},"<":{"group":"Decision"},">=":{"group":"Decision"},"<=":{"group":"Decision"},"all":{"group":"Decision"},"any":{"group":"Decision"},"!":{"group":"Decision"},"within":{"group":"Decision"},"is-supported-script":{"group":"String"},"upcase":{"group":"String"},"downcase":{"group":"String"},"concat":{"group":"String"},"resolved-locale":{"group":"String"}}},"fog":{"range":{"type":"array","default":[0.5,10],"minimum":-20,"maximum":20,"length":2,"value":"number","property-type":"data-constant","transition":true,"expression":{"interpolated":true,"parameters":["zoom"]}},"color":{"type":"color","property-type":"data-constant","default":"#ffffff","expression":{"interpolated":true,"parameters":["zoom"]},"transition":true},"high-color":{"type":"color","property-type":"data-constant","default":"#245cdf","expression":{"interpolated":true,"parameters":["zoom"]},"transition":true},"space-color":{"type":"color","property-type":"data-constant","default":["interpolate",["linear"],["zoom"],4,"#010b19",7,"#367ab9"],"expression":{"interpolated":true,"parameters":["zoom"]},"transition":true},"horizon-blend":{"type":"number","property-type":"data-constant","default":["interpolate",["linear"],["zoom"],4,0.2,7,0.1],"minimum":0,"maximum":1,"expression":{"interpolated":true,"parameters":["zoom"]},"transition":true},"star-intensity":{"type":"number","property-type":"data-constant","default":["interpolate",["linear"],["zoom"],5,0.35,6,0],"minimum":0,"maximum":1,"expression":{"interpolated":true,"parameters":["zoom"]},"transition":true}},"light":{"anchor":{"type":"enum","default":"viewport","values":{"map":{},"viewport":{}},"property-type":"data-constant","transition":false,"expression":{"interpolated":false,"parameters":["zoom"]}},"position":{"type":"array","default":[1.15,210,30],"length":3,"value":"number","property-type":"data-constant","transition":true,"expression":{"interpolated":true,"parameters":["zoom"]}},"color":{"type":"color","property-type":"data-constant","default":"#ffffff","expression":{"interpolated":true,"parameters":["zoom"]},"transition":true},"intensity":{"type":"number","property-type":"data-constant","default":0.5,"minimum":0,"maximum":1,"expression":{"interpolated":true,"parameters":["zoom"]},"transition":true}},"projection":{"name":{"type":"enum","values":{"albers":{},"equalEarth":{},"equirectangular":{},"lambertConformalConic":{},"mercator":{},"naturalEarth":{},"winkelTripel":{},"globe":{}},"default":"mercator","required":true},"center":{"type":"array","length":2,"value":"number","property-type":"data-constant","minimum":[-180,-90],"maximum":[180,90],"transition":false,"requires":[{"name":["albers","lambertConformalConic"]}]},"parallels":{"type":"array","length":2,"value":"number","property-type":"data-constant","minimum":[-90,-90],"maximum":[90,90],"transition":false,"requires":[{"name":["albers","lambertConformalConic"]}]}},"terrain":{"source":{"type":"string","required":true},"exaggeration":{"type":"number","property-type":"data-constant","default":1,"minimum":0,"maximum":1000,"expression":{"interpolated":true,"parameters":["zoom"]},"transition":true,"requires":["source"]}},"paint":["paint_fill","paint_line","paint_circle","paint_heatmap","paint_fill-extrusion","paint_symbol","paint_raster","paint_hillshade","paint_background","paint_sky"],"paint_fill":{"fill-antialias":{"type":"boolean","default":true,"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"fill-opacity":{"type":"number","default":1,"minimum":0,"maximum":1,"transition":true,"expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"fill-color":{"type":"color","default":"#000000","transition":true,"requires":[{"!":"fill-pattern"}],"expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"fill-outline-color":{"type":"color","transition":true,"requires":[{"!":"fill-pattern"},{"fill-antialias":true}],"expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"fill-translate":{"type":"array","value":"number","length":2,"default":[0,0],"transition":true,"units":"pixels","expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"},"fill-translate-anchor":{"type":"enum","values":{"map":{},"viewport":{}},"default":"map","requires":["fill-translate"],"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"fill-pattern":{"type":"resolvedImage","transition":true,"expression":{"interpolated":false,"parameters":["zoom","feature"]},"property-type":"cross-faded-data-driven"}},"paint_fill-extrusion":{"fill-extrusion-opacity":{"type":"number","default":1,"minimum":0,"maximum":1,"transition":true,"expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"},"fill-extrusion-color":{"type":"color","default":"#000000","transition":true,"requires":[{"!":"fill-extrusion-pattern"}],"expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"fill-extrusion-translate":{"type":"array","value":"number","length":2,"default":[0,0],"transition":true,"units":"pixels","expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"},"fill-extrusion-translate-anchor":{"type":"enum","values":{"map":{},"viewport":{}},"default":"map","requires":["fill-extrusion-translate"],"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"fill-extrusion-pattern":{"type":"resolvedImage","transition":true,"expression":{"interpolated":false,"parameters":["zoom","feature"]},"property-type":"cross-faded-data-driven"},"fill-extrusion-height":{"type":"number","default":0,"minimum":0,"units":"meters","transition":true,"expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"fill-extrusion-base":{"type":"number","default":0,"minimum":0,"units":"meters","transition":true,"requires":["fill-extrusion-height"],"expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"fill-extrusion-vertical-gradient":{"type":"boolean","default":true,"transition":false,"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"}},"paint_line":{"line-opacity":{"type":"number","default":1,"minimum":0,"maximum":1,"transition":true,"expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"line-color":{"type":"color","default":"#000000","transition":true,"requires":[{"!":"line-pattern"}],"expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"line-translate":{"type":"array","value":"number","length":2,"default":[0,0],"transition":true,"units":"pixels","expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"},"line-translate-anchor":{"type":"enum","values":{"map":{},"viewport":{}},"default":"map","requires":["line-translate"],"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"line-width":{"type":"number","default":1,"minimum":0,"transition":true,"units":"pixels","expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"line-gap-width":{"type":"number","default":0,"minimum":0,"transition":true,"units":"pixels","expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"line-offset":{"type":"number","default":0,"transition":true,"units":"pixels","expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"line-blur":{"type":"number","default":0,"minimum":0,"transition":true,"units":"pixels","expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"line-dasharray":{"type":"array","value":"number","minimum":0,"transition":true,"units":"line widths","requires":[{"!":"line-pattern"}],"expression":{"interpolated":false,"parameters":["zoom","feature"]},"property-type":"cross-faded-data-driven"},"line-pattern":{"type":"resolvedImage","transition":true,"expression":{"interpolated":false,"parameters":["zoom","feature"]},"property-type":"cross-faded-data-driven"},"line-gradient":{"type":"color","transition":false,"requires":[{"!":"line-pattern"},{"source":"geojson","has":{"lineMetrics":true}}],"expression":{"interpolated":true,"parameters":["line-progress"]},"property-type":"color-ramp"},"line-trim-offset":{"type":"array","value":"number","length":2,"default":[0,0],"minimum":[0,0],"maximum":[1,1],"transition":false,"requires":[{"source":"geojson","has":{"lineMetrics":true}}],"property-type":"constant"}},"paint_circle":{"circle-radius":{"type":"number","default":5,"minimum":0,"transition":true,"units":"pixels","expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"circle-color":{"type":"color","default":"#000000","transition":true,"expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"circle-blur":{"type":"number","default":0,"transition":true,"expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"circle-opacity":{"type":"number","default":1,"minimum":0,"maximum":1,"transition":true,"expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"circle-translate":{"type":"array","value":"number","length":2,"default":[0,0],"transition":true,"units":"pixels","expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"},"circle-translate-anchor":{"type":"enum","values":{"map":{},"viewport":{}},"default":"map","requires":["circle-translate"],"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"circle-pitch-scale":{"type":"enum","values":{"map":{},"viewport":{}},"default":"map","expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"circle-pitch-alignment":{"type":"enum","values":{"map":{},"viewport":{}},"default":"viewport","expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"circle-stroke-width":{"type":"number","default":0,"minimum":0,"transition":true,"units":"pixels","expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"circle-stroke-color":{"type":"color","default":"#000000","transition":true,"expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"circle-stroke-opacity":{"type":"number","default":1,"minimum":0,"maximum":1,"transition":true,"expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"}},"paint_heatmap":{"heatmap-radius":{"type":"number","default":30,"minimum":1,"transition":true,"units":"pixels","expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"heatmap-weight":{"type":"number","default":1,"minimum":0,"transition":false,"expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"heatmap-intensity":{"type":"number","default":1,"minimum":0,"transition":true,"expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"},"heatmap-color":{"type":"color","default":["interpolate",["linear"],["heatmap-density"],0,"rgba(0, 0, 255, 0)",0.1,"royalblue",0.3,"cyan",0.5,"lime",0.7,"yellow",1,"red"],"transition":false,"expression":{"interpolated":true,"parameters":["heatmap-density"]},"property-type":"color-ramp"},"heatmap-opacity":{"type":"number","default":1,"minimum":0,"maximum":1,"transition":true,"expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"}},"paint_symbol":{"icon-opacity":{"type":"number","default":1,"minimum":0,"maximum":1,"transition":true,"requires":["icon-image"],"expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"icon-color":{"type":"color","default":"#000000","transition":true,"requires":["icon-image"],"expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"icon-halo-color":{"type":"color","default":"rgba(0, 0, 0, 0)","transition":true,"requires":["icon-image"],"expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"icon-halo-width":{"type":"number","default":0,"minimum":0,"transition":true,"units":"pixels","requires":["icon-image"],"expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"icon-halo-blur":{"type":"number","default":0,"minimum":0,"transition":true,"units":"pixels","requires":["icon-image"],"expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"icon-translate":{"type":"array","value":"number","length":2,"default":[0,0],"transition":true,"units":"pixels","requires":["icon-image"],"expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"},"icon-translate-anchor":{"type":"enum","values":{"map":{},"viewport":{}},"default":"map","requires":["icon-image","icon-translate"],"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"text-opacity":{"type":"number","default":1,"minimum":0,"maximum":1,"transition":true,"requires":["text-field"],"expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"text-color":{"type":"color","default":"#000000","transition":true,"overridable":true,"requires":["text-field"],"expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"text-halo-color":{"type":"color","default":"rgba(0, 0, 0, 0)","transition":true,"requires":["text-field"],"expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"text-halo-width":{"type":"number","default":0,"minimum":0,"transition":true,"units":"pixels","requires":["text-field"],"expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"text-halo-blur":{"type":"number","default":0,"minimum":0,"transition":true,"units":"pixels","requires":["text-field"],"expression":{"interpolated":true,"parameters":["zoom","feature","feature-state"]},"property-type":"data-driven"},"text-translate":{"type":"array","value":"number","length":2,"default":[0,0],"transition":true,"units":"pixels","requires":["text-field"],"expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"},"text-translate-anchor":{"type":"enum","values":{"map":{},"viewport":{}},"default":"map","requires":["text-field","text-translate"],"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"}},"paint_raster":{"raster-opacity":{"type":"number","default":1,"minimum":0,"maximum":1,"transition":true,"expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"},"raster-hue-rotate":{"type":"number","default":0,"period":360,"transition":true,"units":"degrees","expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"},"raster-brightness-min":{"type":"number","default":0,"minimum":0,"maximum":1,"transition":true,"expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"},"raster-brightness-max":{"type":"number","default":1,"minimum":0,"maximum":1,"transition":true,"expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"},"raster-saturation":{"type":"number","default":0,"minimum":-1,"maximum":1,"transition":true,"expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"},"raster-contrast":{"type":"number","default":0,"minimum":-1,"maximum":1,"transition":true,"expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"},"raster-resampling":{"type":"enum","values":{"linear":{},"nearest":{}},"default":"linear","expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"raster-fade-duration":{"type":"number","default":300,"minimum":0,"transition":false,"units":"milliseconds","expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"}},"paint_hillshade":{"hillshade-illumination-direction":{"type":"number","default":335,"minimum":0,"maximum":359,"transition":false,"expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"},"hillshade-illumination-anchor":{"type":"enum","values":{"map":{},"viewport":{}},"default":"viewport","expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"hillshade-exaggeration":{"type":"number","default":0.5,"minimum":0,"maximum":1,"transition":true,"expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"},"hillshade-shadow-color":{"type":"color","default":"#000000","transition":true,"expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"},"hillshade-highlight-color":{"type":"color","default":"#FFFFFF","transition":true,"expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"},"hillshade-accent-color":{"type":"color","default":"#000000","transition":true,"expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"}},"paint_background":{"background-color":{"type":"color","default":"#000000","transition":true,"requires":[{"!":"background-pattern"}],"expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"},"background-pattern":{"type":"resolvedImage","transition":true,"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"cross-faded"},"background-opacity":{"type":"number","default":1,"minimum":0,"maximum":1,"transition":true,"expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"}},"paint_sky":{"sky-type":{"type":"enum","values":{"gradient":{},"atmosphere":{}},"default":"atmosphere","expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"sky-atmosphere-sun":{"type":"array","value":"number","length":2,"units":"degrees","minimum":[0,0],"maximum":[360,180],"transition":false,"requires":[{"sky-type":"atmosphere"}],"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"sky-atmosphere-sun-intensity":{"type":"number","requires":[{"sky-type":"atmosphere"}],"default":10,"minimum":0,"maximum":100,"transition":false,"property-type":"data-constant"},"sky-gradient-center":{"type":"array","requires":[{"sky-type":"gradient"}],"value":"number","default":[0,0],"length":2,"units":"degrees","minimum":[0,0],"maximum":[360,180],"transition":false,"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"sky-gradient-radius":{"type":"number","requires":[{"sky-type":"gradient"}],"default":90,"minimum":0,"maximum":180,"transition":false,"expression":{"interpolated":false,"parameters":["zoom"]},"property-type":"data-constant"},"sky-gradient":{"type":"color","default":["interpolate",["linear"],["sky-radial-progress"],0.8,"#87ceeb",1,"white"],"transition":false,"requires":[{"sky-type":"gradient"}],"expression":{"interpolated":true,"parameters":["sky-radial-progress"]},"property-type":"color-ramp"},"sky-atmosphere-halo-color":{"type":"color","default":"white","transition":false,"requires":[{"sky-type":"atmosphere"}],"property-type":"data-constant"},"sky-atmosphere-color":{"type":"color","default":"white","transition":false,"requires":[{"sky-type":"atmosphere"}],"property-type":"data-constant"},"sky-opacity":{"type":"number","default":1,"minimum":0,"maximum":1,"transition":true,"expression":{"interpolated":true,"parameters":["zoom"]},"property-type":"data-constant"}},"transition":{"duration":{"type":"number","default":300,"minimum":0,"units":"milliseconds"},"delay":{"type":"number","default":0,"minimum":0,"units":"milliseconds"}},"property-type":{"data-driven":{"type":"property-type"},"cross-faded":{"type":"property-type"},"cross-faded-data-driven":{"type":"property-type"},"color-ramp":{"type":"property-type"},"data-constant":{"type":"property-type"},"constant":{"type":"property-type"}},"promoteId":{"*":{"type":"string"}}};

  //      
  function isExpressionFilter(filter) {
      if (filter === true || filter === false) {
          return true;
      }
      if (!Array.isArray(filter) || filter.length === 0) {
          return false;
      }
      switch (filter[0]) {
      case 'has':
          return filter.length >= 2 && filter[1] !== '$id' && filter[1] !== '$type';
      case 'in':
          return filter.length >= 3 && (typeof filter[1] !== 'string' || Array.isArray(filter[2]));
      case '!in':
      case '!has':
      case 'none':
          return false;
      case '==':
      case '!=':
      case '>':
      case '>=':
      case '<':
      case '<=':
          return filter.length !== 3 || (Array.isArray(filter[1]) || Array.isArray(filter[2]));
      case 'any':
      case 'all':
          for (var i = 0, list = filter.slice(1); i < list.length; i += 1) {
              var f = list[i];

          if (!isExpressionFilter(f) && typeof f !== 'boolean') {
                  return false;
              }
          }
          return true;
      default:
          return true;
      }
  }
  /**
   * Given a filter expressed as nested arrays, return a new function
   * that evaluates whether a given feature (with a .properties or .tags property)
   * passes its test.
   *
   * @private
   * @param {Array} filter mapbox gl filter
   * @param {string} layerType the type of the layer this filter will be applied to.
   * @returns {Function} filter-evaluating function
   */
  function createFilter(filter, layerType) {
      if ( layerType === void 0 ) layerType = 'fill';

      if (filter === null || filter === undefined) {
          return {
              filter: function () { return true; },
              needGeometry: false,
              needFeature: false
          };
      }
      if (!isExpressionFilter(filter)) {
          filter = convertFilter(filter);
      }
      var filterExp = filter;
      var staticFilter = true;
      try {
          staticFilter = extractStaticFilter(filterExp);
      } catch (e) {
          console.warn(("Failed to extract static filter. Filter will continue working, but at higher memory usage and slower framerate.\nThis is most likely a bug, please report this via https://github.com/mapbox/mapbox-gl-js/issues/new?assignees=&labels=&template=Bug_report.md\nand paste the contents of this message in the report.\nThank you!\nFilter Expression:\n" + (JSON.stringify(filterExp, null, 2)) + "\n        "));
      }
      // Compile the static component of the filter
      var filterSpec = spec[("filter_" + layerType)];
      var compiledStaticFilter = createExpression(staticFilter, filterSpec);
      var filterFunc = null;
      if (compiledStaticFilter.result === 'error') {
          throw new Error(compiledStaticFilter.value.map(function (err) { return ((err.key) + ": " + (err.message)); }).join(', '));
      } else {
          filterFunc = function (globalProperties, feature, canonical) { return compiledStaticFilter.value.evaluate(globalProperties, feature, {}, canonical); };
      }
      // If the static component is not equal to the entire filter then we have a dynamic component
      // Compile the dynamic component separately
      var dynamicFilterFunc = null;
      var needFeature = null;
      if (staticFilter !== filterExp) {
          var compiledDynamicFilter = createExpression(filterExp, filterSpec);
          if (compiledDynamicFilter.result === 'error') {
              throw new Error(compiledDynamicFilter.value.map(function (err) { return ((err.key) + ": " + (err.message)); }).join(', '));
          } else {
              dynamicFilterFunc = function (globalProperties, feature, canonical, featureTileCoord, featureDistanceData) { return compiledDynamicFilter.value.evaluate(globalProperties, feature, {}, canonical, undefined, undefined, featureTileCoord, featureDistanceData); };
              needFeature = !isFeatureConstant(compiledDynamicFilter.value.expression);
          }
      }
      filterFunc = filterFunc;
      var needGeometry = geometryNeeded(staticFilter);
      return {
          filter: filterFunc,
          dynamicFilter: dynamicFilterFunc ? dynamicFilterFunc : undefined,
          needGeometry: needGeometry,
          needFeature: !!needFeature
      };
  }
  function extractStaticFilter(filter) {
      if (!isDynamicFilter(filter)) {
          return filter;
      }
      // Shallow copy so we can replace expressions in-place
      var result = deepUnbundle(filter);
      // 1. Union branches
      unionDynamicBranches(result);
      // 2. Collapse dynamic conditions to  `true`
      result = collapseDynamicBooleanExpressions(result);
      return result;
  }
  function collapseDynamicBooleanExpressions(expression) {
      if (!Array.isArray(expression)) {
          return expression;
      }
      var collapsed = collapsedExpression(expression);
      if (collapsed === true) {
          return collapsed;
      } else {
          return collapsed.map(function (subExpression) { return collapseDynamicBooleanExpressions(subExpression); });
      }
  }
  /**
   * Traverses the expression and replaces all instances of branching on a
   * `dynamic` conditional (such as `['pitch']` or `['distance-from-center']`)
   * into an `any` expression.
   * This ensures that all possible outcomes of a `dynamic` branch are considered
   * when evaluating the expression upfront during filtering.
   *
   * @param {Array<any>} filter the filter expression mutated in-place.
   */
  function unionDynamicBranches(filter) {
      var isBranchingDynamically = false;
      var branches = [];
      if (filter[0] === 'case') {
          for (var i = 1; i < filter.length - 1; i += 2) {
              isBranchingDynamically = isBranchingDynamically || isDynamicFilter(filter[i]);
              branches.push(filter[i + 1]);
          }
          branches.push(filter[filter.length - 1]);
      } else if (filter[0] === 'match') {
          isBranchingDynamically = isBranchingDynamically || isDynamicFilter(filter[1]);
          for (var i$1 = 2; i$1 < filter.length - 1; i$1 += 2) {
              branches.push(filter[i$1 + 1]);
          }
          branches.push(filter[filter.length - 1]);
      } else if (filter[0] === 'step') {
          isBranchingDynamically = isBranchingDynamically || isDynamicFilter(filter[1]);
          for (var i$2 = 1; i$2 < filter.length - 1; i$2 += 2) {
              branches.push(filter[i$2 + 1]);
          }
      }
      if (isBranchingDynamically) {
          filter.length = 0;
          filter.push.apply(filter, [ 'any' ].concat( branches ));
      }
      // traverse and recurse into children
      for (var i$3 = 1; i$3 < filter.length; i$3++) {
          unionDynamicBranches(filter[i$3]);
      }
  }
  function isDynamicFilter(filter) {
      // Base Cases
      if (!Array.isArray(filter)) {
          return false;
      }
      if (isRootExpressionDynamic(filter[0])) {
          return true;
      }
      for (var i = 1; i < filter.length; i++) {
          var child = filter[i];
          if (isDynamicFilter(child)) {
              return true;
          }
      }
      return false;
  }
  function isRootExpressionDynamic(expression) {
      return expression === 'pitch' || expression === 'distance-from-center';
  }
  var dynamicConditionExpressions = new Set([
      'in',
      '==',
      '!=',
      '>',
      '>=',
      '<',
      '<=',
      'to-boolean'
  ]);
  function collapsedExpression(expression) {
      if (dynamicConditionExpressions.has(expression[0])) {
          for (var i = 1; i < expression.length; i++) {
              var param = expression[i];
              if (isDynamicFilter(param)) {
                  return true;
              }
          }
      }
      return expression;
  }
  // Comparison function to sort numbers and strings
  function compare(a, b) {
      return a < b ? -1 : a > b ? 1 : 0;
  }
  function geometryNeeded(filter) {
      if (!Array.isArray(filter))
          { return false; }
      if (filter[0] === 'within')
          { return true; }
      for (var index = 1; index < filter.length; index++) {
          if (geometryNeeded(filter[index]))
              { return true; }
      }
      return false;
  }
  function convertFilter(filter) {
      if (!filter)
          { return true; }
      var op = filter[0];
      if (filter.length <= 1)
          { return op !== 'any'; }
      var converted = op === '==' ? convertComparisonOp(filter[1], filter[2], '==') : op === '!=' ? convertNegation(convertComparisonOp(filter[1], filter[2], '==')) : op === '<' || op === '>' || op === '<=' || op === '>=' ? convertComparisonOp(filter[1], filter[2], op) : op === 'any' ? convertDisjunctionOp(filter.slice(1)) : op === 'all' ? ['all'].concat(filter.slice(1).map(convertFilter)) : op === 'none' ? ['all'].concat(filter.slice(1).map(convertFilter).map(convertNegation)) : op === 'in' ? convertInOp(filter[1], filter.slice(2)) : op === '!in' ? convertNegation(convertInOp(filter[1], filter.slice(2))) : op === 'has' ? convertHasOp(filter[1]) : op === '!has' ? convertNegation(convertHasOp(filter[1])) : op === 'within' ? filter : true;
      return converted;
  }
  function convertComparisonOp(property, value, op) {
      switch (property) {
      case '$type':
          return [
              ("filter-type-" + op),
              value
          ];
      case '$id':
          return [
              ("filter-id-" + op),
              value
          ];
      default:
          return [
              ("filter-" + op),
              property,
              value
          ];
      }
  }
  function convertDisjunctionOp(filters) {
      return ['any'].concat(filters.map(convertFilter));
  }
  function convertInOp(property, values) {
      if (values.length === 0) {
          return false;
      }
      switch (property) {
      case '$type':
          return [
              "filter-type-in",
              [
                  'literal',
                  values
              ]
          ];
      case '$id':
          return [
              "filter-id-in",
              [
                  'literal',
                  values
              ]
          ];
      default:
          if (values.length > 200 && !values.some(function (v) { return typeof v !== typeof values[0]; })) {
              return [
                  'filter-in-large',
                  property,
                  [
                      'literal',
                      values.sort(compare)
                  ]
              ];
          } else {
              return [
                  'filter-in-small',
                  property,
                  [
                      'literal',
                      values
                  ]
              ];
          }
      }
  }
  function convertHasOp(property) {
      switch (property) {
      case '$type':
          return true;
      case '$id':
          return ["filter-has-id"];
      default:
          return [
              "filter-has",
              property
          ];
      }
  }
  function convertNegation(filter) {
      return [
          '!',
          filter
      ];
  }

  //      
  var refProperties = [
      'type',
      'source',
      'source-layer',
      'minzoom',
      'maxzoom',
      'filter',
      'layout'
  ];

  //      
  function deref(layer, parent) {
      var result = {};
      for (var k in layer) {
          if (k !== 'ref') {
              result[k] = layer[k];
          }
      }
      refProperties.forEach(function (k) {
          if (k in parent) {
              result[k] = parent[k];
          }
      });
      return result;
  }
  /**
   * Given an array of layers, some of which may contain `ref` properties
   * whose value is the `id` of another property, return a new array where
   * such layers have been augmented with the 'type', 'source', etc. properties
   * from the parent layer, and the `ref` property has been removed.
   *
   * The input is not modified. The output may contain references to portions
   * of the input.
   *
   * @private
   * @param {Array<Layer>} layers
   * @returns {Array<Layer>}
   */
  function derefLayers(layers) {
      layers = layers.slice();
      var map = Object.create(null);
      for (var i = 0; i < layers.length; i++) {
          map[layers[i].id] = layers[i];
      }
      for (var i$1 = 0; i$1 < layers.length; i$1++) {
          if ('ref' in layers[i$1]) {
              layers[i$1] = deref(layers[i$1], map[layers[i$1].ref]);
          }
      }
      return layers;
  }

  var fontWeights = {
      thin: 100,
      hairline: 100,
      'ultra-light': 100,
      'extra-light': 100,
      light: 200,
      book: 300,
      regular: 400,
      normal: 400,
      plain: 400,
      roman: 400,
      standard: 400,
      medium: 500,
      'semi-bold': 600,
      'demi-bold': 600,
      bold: 700,
      heavy: 800,
      black: 800,
      'extra-bold': 800,
      'ultra-black': 900,
      'extra-black': 900,
      'ultra-bold': 900,
      'heavy-black': 900,
      fat: 900,
      poster: 900
  };
  var sp = ' ';
  var italicRE = /(italic|oblique)$/i;
  var fontCache = {};
  var mapboxToCssFont = function (fonts, size, lineHeight) {
      var cssData = fontCache[fonts];
      if (!cssData) {
          if (!Array.isArray(fonts)) {
              fonts = [fonts];
          }
          var weight = 400;
          var style = 'normal';
          var fontFamilies = [];
          var haveWeight, haveStyle;
          for (var i = 0, ii = fonts.length; i < ii; ++i) {
              var font = fonts[i];
              var parts = font.split(' ');
              var maybeWeight = parts[parts.length - 1].toLowerCase();
              if (maybeWeight == 'normal' || maybeWeight == 'italic' || maybeWeight == 'oblique') {
                  style = haveStyle ? style : maybeWeight;
                  parts.pop();
                  maybeWeight = parts[parts.length - 1].toLowerCase();
              } else if (italicRE.test(maybeWeight)) {
                  maybeWeight = maybeWeight.replace(italicRE, '');
                  style = haveStyle ? style : parts[parts.length - 1].replace(maybeWeight, '');
              }
              for (var w in fontWeights) {
                  var previousPart = parts.length > 1 ? parts[parts.length - 2].toLowerCase() : '';
                  if (maybeWeight == w || maybeWeight == w.replace('-', '') || previousPart + '-' + maybeWeight == w) {
                      weight = haveWeight ? weight : fontWeights[w];
                      parts.pop();
                      if (previousPart && w.startsWith(previousPart)) {
                          parts.pop();
                      }
                      break;
                  }
              }
              if (!haveWeight && typeof maybeWeight == 'number') {
                  weight = maybeWeight;
              }
              var fontFamily = parts.join(sp).replace('Klokantech Noto Sans', 'Noto Sans');
              if (fontFamily.indexOf(sp) !== -1) {
                  fontFamily = '"' + fontFamily + '"';
              }
              fontFamilies.push(fontFamily);
          }
          // CSS font property: font-style font-weight font-size/line-height font-family
          cssData = fontCache[fonts] = [
              style,
              weight,
              fontFamilies
          ];
      }
      return cssData[0] + sp + cssData[1] + sp + size + 'px' + (lineHeight ? '/' + lineHeight : '') + sp + cssData[2];
  };

  var mapboxBaseUrl = 'https://api.mapbox.com';
  /**
   * Gets the path from a mapbox:// URL.
   * @param {string} url The Mapbox URL.
   * @return {string} The path.
   * @private
   */
  function getMapboxPath(url) {
      var startsWith = 'mapbox://';
      if (url.indexOf(startsWith) !== 0) {
          return '';
      }
      return url.slice(startsWith.length);
  }
  /**
   * Turns mapbox:// sprite URLs into resolvable URLs.
   * @param {string} url The sprite URL.
   * @param {string} token The access token.
   * @param {string} styleUrl The style URL.
   * @return {string} A resolvable URL.
   * @private
   */
  function normalizeSpriteUrl(url, token, styleUrl) {
      var mapboxPath = getMapboxPath(url);
      if (!mapboxPath) {
          return decodeURI(new URL(url, styleUrl).href);
      }
      var startsWith = 'sprites/';
      if (mapboxPath.indexOf(startsWith) !== 0) {
          throw new Error(("unexpected sprites url: " + url));
      }
      var sprite = mapboxPath.slice(startsWith.length);
      return (mapboxBaseUrl + "/styles/v1/" + sprite + "/sprite?access_token=" + token);
  }
  /**
   * Turns mapbox:// style URLs into resolvable URLs.
   * @param {string} url The style URL.
   * @param {string} token The access token.
   * @return {string} A resolvable URL.
   * @private
   */
  function normalizeStyleUrl(url, token) {
      var mapboxPath = getMapboxPath(url);
      if (!mapboxPath) {
          return decodeURI(new URL(url, location.href).href);
      }
      var startsWith = 'styles/';
      if (mapboxPath.indexOf(startsWith) !== 0) {
          throw new Error(("unexpected style url: " + url));
      }
      var style = mapboxPath.slice(startsWith.length);
      return (mapboxBaseUrl + "/styles/v1/" + style + "?&access_token=" + token);
  }
  /**
   * Turns mapbox:// source URLs into vector tile URL templates.
   * @param {string} url The source URL.
   * @param {string} token The access token.
   * @param {string} tokenParam The access token key.
   * @param {string} styleUrl The style URL.
   * @return {string} A vector tile template.
   * @private
   */
  function normalizeSourceUrl(url, token, tokenParam, styleUrl) {
      var urlObject = new URL(url, styleUrl);
      var mapboxPath = getMapboxPath(url);
      if (!mapboxPath) {
          if (!token) {
              return decodeURI(urlObject.href);
          }
          urlObject.searchParams.set(tokenParam, token);
          return decodeURI(urlObject.href);
      }
      return ("https://{a-d}.tiles.mapbox.com/v4/" + mapboxPath + "/{z}/{x}/{y}.vector.pbf?access_token=" + token);
  }

  function deg2rad(degrees) {
      return degrees * Math.PI / 180;
  }
  var defaultResolutions = function () {
      var resolutions = [];
      for (var res = 78271.51696402048; resolutions.length <= 24; res /= 2) {
          resolutions.push(res);
      }
      return resolutions;
  }();
  /**
   * @param {number} width Width of the canvas.
   * @param {number} height Height of the canvas.
   * @return {HTMLCanvasElement} Canvas.
   */
  function createCanvas(width, height) {
      if (typeof WorkerGlobalScope !== 'undefined' && self instanceof WorkerGlobalScope && typeof OffscreenCanvas !== 'undefined') {
          // eslint-disable-line
          return new OffscreenCanvas(width, height);
      } else {
          var canvas = document.createElement('canvas');
          canvas.width = width;
          canvas.height = height;
          return canvas;
      }
  }
  function getZoomForResolution(resolution, resolutions) {
      var i = 0;
      var ii = resolutions.length;
      for (; i < ii; ++i) {
          var candidate = resolutions[i];
          if (candidate < resolution && i + 1 < ii) {
              var zoomFactor = resolutions[i] / resolutions[i + 1];
              return i + Math.log(resolutions[i] / resolution) / Math.log(zoomFactor);
          }
      }
      return ii - 1;
  }
  var pendingRequests = {};
  /**
   * @param {ResourceType} resourceType Type of resource to load.
   * @param {string} url Url of the resource.
   * @param {Options} [options={}] Options.
   * @return {Promise<Object|Response>} Promise that resolves with the loaded resource
   * or rejects with the Response object.
   * @private
   */
  function fetchResource(resourceType, url, options) {
      if ( options === void 0 ) options = {};

      if (url in pendingRequests) {
          return pendingRequests[url];
      } else {
          var request = options.transformRequest ? options.transformRequest(url, resourceType) || new Request(url) : new Request(url);
          if (!request.headers.get('Accept')) {
              request.headers.set('Accept', 'application/json');
          }
          var pendingRequest = fetch(request).then(function (response) {
              delete pendingRequests[url];
              return response.ok ? response.json() : Promise.reject(new Error('Error fetching source ' + url));
          }).catch(function (error) {
              delete pendingRequests[url];
              return Promise.reject(new Error('Error fetching source ' + url));
          });
          pendingRequests[url] = pendingRequest;
          return pendingRequest;
      }
  }
  function getGlStyle(glStyleOrUrl, options) {
      if (typeof glStyleOrUrl === 'string') {
          if (glStyleOrUrl.trim().startsWith('{')) {
              try {
                  var glStyle = JSON.parse(glStyleOrUrl);
                  return Promise.resolve(glStyle);
              } catch (error) {
                  return Promise.reject(error);
              }
          } else {
              glStyleOrUrl = normalizeStyleUrl(glStyleOrUrl, options.accessToken);
              return fetchResource('Style', glStyleOrUrl, options);
          }
      } else {
          return Promise.resolve(glStyleOrUrl);
      }
  }
  var tilejsonCache = {};
  /**
   * @param {Object} glSource glStyle source object.
   * @param {string} styleUrl Style URL.
   * @param {Options} options Options.
   * @return {Object} TileJson
   */
  function getTileJson(glSource, styleUrl, options) {
      if ( options === void 0 ) options = {};

      var cacheKey = [
          styleUrl,
          JSON.stringify(glSource)
      ].toString();
      var promise = tilejsonCache[cacheKey];
      if (!promise || options.transformRequest) {
          var url = glSource.url;
          if (url && !glSource.tiles) {
              var normalizedSourceUrl = normalizeSourceUrl(url, options.accessToken, options.accessTokenParam || 'access_token', styleUrl || location.href);
              if (url.startsWith('mapbox://')) {
                  promise = Promise.resolve(Object.assign({}, glSource, {
                      url: undefined,
                      tiles: normalizedSourceUrl
                  }));
              } else {
                  promise = fetchResource('Source', normalizedSourceUrl, options).then(function (tileJson) {
                      for (var i = 0, ii = tileJson.tiles.length; i < ii; ++i) {
                          var tileUrl = tileJson.tiles[i];
                          if (options.transformRequest) {
                              var request = options.transformRequest(normalizedSourceUrl, 'Source');
                              if (request) {
                                  normalizedSourceUrl = request.url;
                              }
                          }
                          var normalizedTileUrl = normalizeSourceUrl(tileUrl, options.accessToken, options.accessTokenParam || 'access_token', normalizedSourceUrl);
                          if (options.transformRequest) {
                              var transformedRequest = options.transformRequest(normalizedTileUrl, 'Tiles');
                              if (transformedRequest instanceof Request) {
                                  normalizedTileUrl = decodeURI(transformedRequest.url);
                              }
                          }
                          tileJson.tiles[i] = normalizedTileUrl;
                      }
                      return Promise.resolve(tileJson);
                  });
              }
          } else {
              glSource = Object.assign({}, glSource, {
                  tiles: glSource.tiles.map(function (tileUrl) {
                      return normalizeSourceUrl(tileUrl, options.accessToken, options.accessTokenParam || 'access_token', styleUrl || location.href);
                  })
              });
              promise = Promise.resolve(Object.assign({}, glSource));
          }
          tilejsonCache[cacheKey] = promise;
      }
      return promise;
  }    /**
   * @typedef {import("./apply.js").Options} Options
   * @typedef {import('./apply.js').ResourceType} ResourceType
   * @private
   */

  var hairSpacePool = Array(256).join('\u200A');
  function applyLetterSpacing(text, letterSpacing) {
      if (letterSpacing >= 0.05) {
          var textWithLetterSpacing = '';
          var lines = text.split('\n');
          var joinSpaceString = hairSpacePool.slice(0, Math.round(letterSpacing / 0.1));
          for (var l = 0, ll = lines.length; l < ll; ++l) {
              if (l > 0) {
                  textWithLetterSpacing += '\n';
              }
              textWithLetterSpacing += lines[l].split('').join(joinSpaceString);
          }
          return textWithLetterSpacing;
      }
      return text;
  }
  var measureContext;
  function getMeasureContext() {
      if (!measureContext) {
          measureContext = createCanvas(1, 1).getContext('2d');
      }
      return measureContext;
  }
  function measureText(text, letterSpacing) {
      return getMeasureContext().measureText(text).width + (text.length - 1) * letterSpacing;
  }
  var measureCache = {};
  function wrapText(text, font, em, letterSpacing) {
      if (text.indexOf('\n') !== -1) {
          var hardLines = text.split('\n');
          var lines = [];
          for (var i = 0, ii = hardLines.length; i < ii; ++i) {
              lines.push(wrapText(hardLines[i], font, em, letterSpacing));
          }
          return lines.join('\n');
      }
      var key = em + ',' + font + ',' + text + ',' + letterSpacing;
      var wrappedText = measureCache[key];
      if (!wrappedText) {
          var words = text.split(' ');
          if (words.length > 1) {
              var ctx = getMeasureContext();
              ctx.font = font;
              var oneEm = ctx.measureText('M').width;
              var maxWidth = oneEm * em;
              var line = '';
              var lines$1 = [];
              // Pass 1 - wrap lines to not exceed maxWidth
              for (var i$1 = 0, ii$1 = words.length; i$1 < ii$1; ++i$1) {
                  var word = words[i$1];
                  var testLine = line + (line ? ' ' : '') + word;
                  if (measureText(testLine, letterSpacing) <= maxWidth) {
                      line = testLine;
                  } else {
                      if (line) {
                          lines$1.push(line);
                      }
                      line = word;
                  }
              }
              if (line) {
                  lines$1.push(line);
              }
              // Pass 2 - add lines with a width of less than 30% of maxWidth to the previous or next line
              for (var i$2 = 0, ii$2 = lines$1.length; i$2 < ii$2 && ii$2 > 1; ++i$2) {
                  var line$1 = lines$1[i$2];
                  if (measureText(line$1, letterSpacing) < maxWidth * 0.35) {
                      var prevWidth = i$2 > 0 ? measureText(lines$1[i$2 - 1], letterSpacing) : Infinity;
                      var nextWidth = i$2 < ii$2 - 1 ? measureText(lines$1[i$2 + 1], letterSpacing) : Infinity;
                      lines$1.splice(i$2, 1);
                      ii$2 -= 1;
                      if (prevWidth < nextWidth) {
                          lines$1[i$2 - 1] += ' ' + line$1;
                          i$2 -= 1;
                      } else {
                          lines$1[i$2] = line$1 + ' ' + lines$1[i$2];
                      }
                  }
              }
              // Pass 3 - try to fill 80% of maxWidth for each line
              for (var i$3 = 0, ii$3 = lines$1.length - 1; i$3 < ii$3; ++i$3) {
                  var line$2 = lines$1[i$3];
                  var next = lines$1[i$3 + 1];
                  if (measureText(line$2, letterSpacing) > maxWidth * 0.7 && measureText(next, letterSpacing) < maxWidth * 0.6) {
                      var lineWords = line$2.split(' ');
                      var lastWord = lineWords.pop();
                      if (measureText(lastWord, letterSpacing) < maxWidth * 0.2) {
                          lines$1[i$3] = lineWords.join(' ');
                          lines$1[i$3 + 1] = lastWord + ' ' + next;
                      }
                      ii$3 -= 1;
                  }
              }
              wrappedText = lines$1.join('\n');
          } else {
              wrappedText = text;
          }
          wrappedText = applyLetterSpacing(wrappedText, letterSpacing);
          measureCache[key] = wrappedText;
      }
      return wrappedText;
  }
  var fontFamilyRegEx = /font-family: ?([^;]*);/;
  var stripQuotesRegEx = /("|')/g;
  var loadedFontFamilies;
  function hasFontFamily(family) {
      if (!loadedFontFamilies) {
          loadedFontFamilies = {};
          var styleSheets = document.styleSheets;
          for (var i = 0, ii = styleSheets.length; i < ii; ++i) {
              var styleSheet = styleSheets[i];
              try {
                  var cssRules = styleSheet.rules || styleSheet.cssRules;
                  if (cssRules) {
                      for (var j = 0, jj = cssRules.length; j < jj; ++j) {
                          var cssRule = cssRules[j];
                          if (cssRule.type == 5) {
                              var match = cssRule.cssText.match(fontFamilyRegEx);
                              loadedFontFamilies[match[1].replace(stripQuotesRegEx, '')] = true;
                          }
                      }
                  }
              } catch (e) {
              }
          }
      }
      return family in loadedFontFamilies;
  }
  var processedFontFamilies = {};
  /**
   * @param {Array} fonts Fonts.
   * @return {Array} Processed fonts.
   * @private
   */
  function getFonts(fonts) {
      var fontsKey = fonts.toString();
      if (fontsKey in processedFontFamilies) {
          return processedFontFamilies[fontsKey];
      }
      var googleFontDescriptions = [];
      for (var i = 0, ii = fonts.length; i < ii; ++i) {
          fonts[i] = fonts[i].replace('Arial Unicode MS', 'Arial');
          var font = fonts[i];
          var cssFont = mapboxToCssFont(font, 1);
          registerFont(cssFont);
          var parts = cssFont.split(' ');
          googleFontDescriptions.push([
              parts.slice(3).join(' ').replace(/"/g, ''),
              parts[1],
              parts[0]
          ]);
      }
      for (var i$1 = 0, ii$1 = googleFontDescriptions.length; i$1 < ii$1; ++i$1) {
          var googleFontDescription = googleFontDescriptions[i$1];
          var family = googleFontDescription[0];
          if (!hasFontFamily(family)) {
              if (checkedFonts.get(((googleFontDescription[2]) + "\n" + (googleFontDescription[1]) + " \n" + family)) !== 100) {
                  var fontUrl = 'https://fonts.googleapis.com/css?family=' + family.replace(/ /g, '+') + ':' + googleFontDescription[1] + googleFontDescription[2];
                  if (!document.querySelector('link[href="' + fontUrl + '"]')) {
                      var markup = document.createElement('link');
                      markup.href = fontUrl;
                      markup.rel = 'stylesheet';
                      document.head.appendChild(markup);
                  }
              }
          }
      }
      processedFontFamilies[fontsKey] = fonts;
      return fonts;
  }

  /*
  ol-mapbox-style - Use Mapbox Style objects with OpenLayers
  Copyright 2016-present ol-mapbox-style contributors
  License: https://raw.githubusercontent.com/openlayers/ol-mapbox-style/master/LICENSE
  */
  /**
   * @typedef {import("ol/layer/Vector").default} VectorLayer
   * @typedef {import("ol/layer/VectorTile").default} VectorTileLayer
   * @typedef {import("ol/style/Style").StyleFunction} StyleFunction
   */
  var types = {
      'Point': 1,
      'MultiPoint': 1,
      'LineString': 2,
      'MultiLineString': 2,
      'Polygon': 3,
      'MultiPolygon': 3
  };
  var anchor = {
      'center': [
          0.5,
          0.5
      ],
      'left': [
          0,
          0.5
      ],
      'right': [
          1,
          0.5
      ],
      'top': [
          0.5,
          0
      ],
      'bottom': [
          0.5,
          1
      ],
      'top-left': [
          0,
          0
      ],
      'top-right': [
          1,
          0
      ],
      'bottom-left': [
          0,
          1
      ],
      'bottom-right': [
          1,
          1
      ]
  };
  var expressionData = function (rawExpression, propertySpec) {
      var compiledExpression = createPropertyExpression(rawExpression, propertySpec);
      if (compiledExpression.result === 'error') {
          throw new Error(compiledExpression.value.map(function (err) { return ((err.key) + ": " + (err.message)); }).join(', '));
      }
      return compiledExpression.value;
  };
  var emptyObj$1 = {};
  var zoomObj = { zoom: 0 };
  var renderFeatureCoordinates, renderFeature;
  /**
   * @private
   * @param {Object} layer Gl object layer.
   * @param {string} layoutOrPaint 'layout' or 'paint'.
   * @param {string} property Feature property.
   * @param {number} zoom Zoom.
   * @param {Object} feature Gl feature.
   * @param {Object} [functionCache] Function cache.
   * @param {Object} [featureState] Feature state.
   * @return {?} Value.
   */
  function getValue(layer, layoutOrPaint, property, zoom, feature, functionCache, featureState) {
      var layerId = layer.id;
      if (!functionCache) {
          functionCache = {};
          console.warn('No functionCache provided to getValue()');    //eslint-disable-line no-console
      }
      if (!functionCache[layerId]) {
          functionCache[layerId] = {};
      }
      var functions = functionCache[layerId];
      if (!functions[property]) {
          var value = (layer[layoutOrPaint] || emptyObj$1)[property];
          var propertySpec = spec[(layoutOrPaint + "_" + (layer.type))][property];
          if (value === undefined) {
              value = propertySpec.default;
          }
          var isExpr = isExpression(value);
          if (!isExpr && isFunction(value)) {
              value = convertFunction(value, propertySpec);
              isExpr = true;
          }
          if (isExpr) {
              var compiledExpression = expressionData(value, propertySpec);
              functions[property] = compiledExpression.evaluate.bind(compiledExpression);
          } else {
              if (propertySpec.type == 'color') {
                  value = Color.parse(value);
              }
              functions[property] = function () {
                  return value;
              };
          }
      }
      zoomObj.zoom = zoom;
      return functions[property](zoomObj, feature, featureState);
  }
  /**
   * @private
   * @param {Object} layer Gl object layer.
   * @param {number} zoom Zoom.
   * @param {Object} feature Gl feature.
   * @param {Object} [functionCache] Function cache.
   * @return {"declutter"|"obstacle"|"none"} Value.
   */
  function getIconDeclutterMode(layer, zoom, feature, functionCache) {
      var allowOverlap = getValue(layer, 'layout', 'icon-allow-overlap', zoom, feature, functionCache);
      if (!allowOverlap) {
          return 'declutter';
      }
      var ignorePlacement = getValue(layer, 'layout', 'icon-ignore-placement', zoom, feature, functionCache);
      if (!ignorePlacement) {
          return 'obstacle';
      }
      return 'none';
  }
  /**
   * @private
   * @param {string} layerId Layer id.
   * @param {?} filter Filter.
   * @param {Object} feature Feature.
   * @param {number} zoom Zoom.
   * @param {Object} [filterCache] Filter cache.
   * @return {boolean} Filter result.
   */
  function evaluateFilter(layerId, filter, feature, zoom, filterCache) {
      if (!filterCache) {
          console.warn('No filterCache provided to evaluateFilter()');    //eslint-disable-line no-console
      }
      if (!(layerId in filterCache)) {
          filterCache[layerId] = createFilter(filter).filter;
      }
      zoomObj.zoom = zoom;
      return filterCache[layerId](zoomObj, feature);
  }
  /**
   * @private
   * @param {?} color Color.
   * @param {number} [opacity] Opacity.
   * @return {string} Color.
   */
  function colorWithOpacity(color, opacity) {
      if (color) {
          if ((color.a === 0 || opacity === 0)) {
              return undefined;
          }
          var a = color.a;
          opacity = opacity === undefined ? 1 : opacity;
          return a === 0 ? 'transparent' : 'rgba(' + Math.round(color.r * 255 / a) + ',' + Math.round(color.g * 255 / a) + ',' + Math.round(color.b * 255 / a) + ',' + a * opacity + ')';
      }
      return color;
  }
  var templateRegEx = /\{[^{}}]*\}/g;
  /**
   * @private
   * @param {string} text Text.
   * @param {Object} properties Properties.
   * @return {string} Text.
   */
  function fromTemplate(text, properties) {
      return text.replace(templateRegEx, function (match) {
          return properties[match.slice(1, -1)] || '';
      });
  }
  /**
   * Creates a style function from the `glStyle` object for all layers that use
   * the specified `source`, which needs to be a `"type": "vector"` or
   * `"type": "geojson"` source and applies it to the specified OpenLayers layer.
   *
   * Two additional properties will be set on the provided layer:
   *
   *  * `mapbox-source`: The `id` of the Mapbox Style document's source that the
   *    OpenLayers layer was created from. Usually `apply()` creates one
   *    OpenLayers layer per Mapbox Style source, unless the layer stack has
   *    layers from different sources in between.
   *  * `mapbox-layers`: The `id`s of the Mapbox Style document's layers that are
   *    included in the OpenLayers layer.
   *
   * This function also works in a web worker. In worker mode, the main thread needs
   * to listen to messages from the worker and respond with another message to make
   * sure that sprite image loading works:
   *
   * ```js
   *  worker.addEventListener('message', event => {
   *   if (event.data.action === 'loadImage') {
   *     const image = new Image();
   *     image.crossOrigin = 'anonymous';
   *     image.addEventListener('load', function() {
   *       createImageBitmap(image, 0, 0, image.width, image.height).then(imageBitmap => {
   *         worker.postMessage({
   *           action: 'imageLoaded',
   *           image: imageBitmap,
   *           src: event.data.src
   *         }, [imageBitmap]);
   *       });
   *     });
   *     image.src = event.data.src;
   *   }
   * });
   * ```
   *
   * @param {VectorLayer|VectorTileLayer} olLayer OpenLayers layer to
   * apply the style to. In addition to the style, the layer will get two
   * properties: `mapbox-source` will be the `id` of the `glStyle`'s source used
   * for the layer, and `mapbox-layers` will be an array of the `id`s of the
   * `glStyle`'s layers.
   * @param {string|Object} glStyle Mapbox Style object.
   * @param {string|Array<string>} sourceOrLayers `source` key or an array of layer `id`s
   * from the Mapbox Style object. When a `source` key is provided, all layers for
   * the specified source will be included in the style function. When layer `id`s
   * are provided, they must be from layers that use the same source.
   * @param {Array<number>} resolutions
   * Resolutions for mapping resolution to zoom level.
   * @param {Object} spriteData Sprite data from the url specified in
   * the Mapbox Style object's `sprite` property. Only required if a `sprite`
   * property is specified in the Mapbox Style object.
   * @param {string} spriteImageUrl Sprite image url for the sprite
   * specified in the Mapbox Style object's `sprite` property. Only required if a
   * `sprite` property is specified in the Mapbox Style object.
   * @param {function(Array<string>):Array<string>} getFonts Function that
   * receives a font stack as arguments, and returns a (modified) font stack that
   * is available. Font names are the names used in the Mapbox Style object. If
   * not provided, the font stack will be used as-is. This function can also be
   * used for loading web fonts.
   * @return {StyleFunction} Style function for use in
   * `ol.layer.Vector` or `ol.layer.VectorTile`.
   */
  function stylefunction(olLayer, glStyle, sourceOrLayers, resolutions, spriteData, spriteImageUrl, getFonts) {
      if ( resolutions === void 0 ) resolutions = defaultResolutions;
      if ( spriteData === void 0 ) spriteData = undefined;
      if ( spriteImageUrl === void 0 ) spriteImageUrl = undefined;
      if ( getFonts === void 0 ) getFonts = undefined;

      if (typeof glStyle == 'string') {
          glStyle = JSON.parse(glStyle);
      }
      if (glStyle.version != 8) {
          throw new Error('glStyle version 8 required.');
      }
      var spriteImage, spriteImgSize;
      if (spriteImageUrl) {
          if (typeof Image !== 'undefined') {
              var img = new Image();
              img.crossOrigin = 'anonymous';
              img.onload = function () {
                  spriteImage = img;
                  spriteImgSize = [
                      img.width,
                      img.height
                  ];
                  olLayer.changed();
                  img.onload = null;
              };
              img.src = spriteImageUrl;
          } else if (typeof WorkerGlobalScope !== 'undefined' && self instanceof WorkerGlobalScope) {
              //eslint-disable-line
              var worker = self;
              // Main thread needs to handle 'loadImage' and dispatch 'imageLoaded'
              worker.postMessage({
                  action: 'loadImage',
                  src: spriteImageUrl
              });
              worker.addEventListener('message', function handler(event) {
                  if (event.data.action === 'imageLoaded' && event.data.src === spriteImageUrl) {
                      spriteImage = event.data.image;
                      spriteImgSize = [
                          spriteImage.width,
                          spriteImage.height
                      ];
                  }
              });
          }
      }
      var allLayers = derefLayers(glStyle.layers);
      var layersBySourceLayer = {};
      var mapboxLayers = [];
      var iconImageCache = {};
      var patternCache = {};
      var functionCache = {};
      var filterCache = {};
      var mapboxSource;
      for (var i = 0, ii = allLayers.length; i < ii; ++i) {
          var layer = allLayers[i];
          var layerId = layer.id;
          if (typeof sourceOrLayers == 'string' && layer.source == sourceOrLayers || sourceOrLayers.indexOf(layerId) !== -1) {
              var sourceLayer = layer['source-layer'];
              if (!mapboxSource) {
                  mapboxSource = layer.source;
                  var source = glStyle.sources[mapboxSource];
                  if (!source) {
                      throw new Error(("Source \"" + mapboxSource + "\" is not defined"));
                  }
                  var type = source.type;
                  if (type !== 'vector' && type !== 'geojson') {
                      throw new Error(("Source \"" + mapboxSource + "\" is not of type \"vector\" or \"geojson\", but \"" + type + "\""));
                  }
              } else if (layer.source !== mapboxSource) {
                  throw new Error(("Layer \"" + layerId + "\" does not use source \"" + mapboxSource));
              }
              var layers = layersBySourceLayer[sourceLayer];
              if (!layers) {
                  layers = [];
                  layersBySourceLayer[sourceLayer] = layers;
              }
              layers.push({
                  layer: layer,
                  index: i
              });
              mapboxLayers.push(layerId);
          }
      }
      var textHalo = new Stroke();
      var textColor = new Fill();
      var styles = [];
      var styleFunction = function (feature, resolution) {
          var properties = feature.getProperties();
          var layers = layersBySourceLayer[properties.layer];
          if (!layers) {
              return;
          }
          var zoom = resolutions.indexOf(resolution);
          if (zoom == -1) {
              zoom = getZoomForResolution(resolution, resolutions);
          }
          var type = types[feature.getGeometry().getType()];
          var f = {
              properties: properties,
              type: type
          };
          var featureState = olLayer.get('mapbox-featurestate')[feature.getId()];
          var stylesLength = -1;
          for (var i = 0, ii = layers.length; i < ii; ++i) {
              var layerData = layers[i];
              var layer = layerData.layer;
              var layerId = layer.id;
              var layout = layer.layout || emptyObj$1;
              var paint = layer.paint || emptyObj$1;
              if (layout.visibility === 'none' || 'minzoom' in layer && zoom < layer.minzoom || 'maxzoom' in layer && zoom >= layer.maxzoom) {
                  continue;
              }
              var filter = layer.filter;
              if (!filter || evaluateFilter(layerId, filter, f, zoom, filterCache)) {
                  var color = (void 0), opacity = (void 0), fill = (void 0), stroke = (void 0), strokeColor = (void 0), style = (void 0);
                  var index = layerData.index;
                  if (type == 3 && (layer.type == 'fill' || layer.type == 'fill-extrusion')) {
                      opacity = getValue(layer, 'paint', layer.type + '-opacity', zoom, f, functionCache, featureState);
                      if (layer.type + '-pattern' in paint) {
                          var fillIcon = getValue(layer, 'paint', layer.type + '-pattern', zoom, f, functionCache, featureState);
                          if (fillIcon) {
                              var icon = typeof fillIcon === 'string' ? fromTemplate(fillIcon, properties) : fillIcon.toString();
                              if (spriteImage && spriteData && spriteData[icon]) {
                                  ++stylesLength;
                                  style = styles[stylesLength];
                                  if (!style || !style.getFill() || style.getStroke() || style.getText()) {
                                      style = new Style({ fill: new Fill() });
                                      styles[stylesLength] = style;
                                  }
                                  fill = style.getFill();
                                  style.setZIndex(index);
                                  var icon_cache_key = icon + '.' + opacity;
                                  var pattern = patternCache[icon_cache_key];
                                  if (!pattern) {
                                      var spriteImageData = spriteData[icon];
                                      var canvas = createCanvas(spriteImageData.width, spriteImageData.height);
                                      var ctx = canvas.getContext('2d');
                                      ctx.globalAlpha = opacity;
                                      ctx.drawImage(spriteImage, spriteImageData.x, spriteImageData.y, spriteImageData.width, spriteImageData.height, 0, 0, spriteImageData.width, spriteImageData.height);
                                      pattern = ctx.createPattern(canvas, 'repeat');
                                      patternCache[icon_cache_key] = pattern;
                                  }
                                  fill.setColor(pattern);
                              }
                          }
                      } else {
                          color = colorWithOpacity(getValue(layer, 'paint', layer.type + '-color', zoom, f, functionCache, featureState), opacity);
                          if (layer.type + '-outline-color' in paint) {
                              strokeColor = colorWithOpacity(getValue(layer, 'paint', layer.type + '-outline-color', zoom, f, functionCache, featureState), opacity);
                          }
                          if (!strokeColor) {
                              strokeColor = color;
                          }
                          if (color || strokeColor) {
                              ++stylesLength;
                              style = styles[stylesLength];
                              if (!style || color && !style.getFill() || !color && style.getFill() || strokeColor && !style.getStroke() || !strokeColor && style.getStroke() || style.getText()) {
                                  style = new Style({
                                      fill: color ? new Fill() : undefined,
                                      stroke: strokeColor ? new Stroke() : undefined
                                  });
                                  styles[stylesLength] = style;
                              }
                              if (color) {
                                  fill = style.getFill();
                                  fill.setColor(color);
                              }
                              if (strokeColor) {
                                  stroke = style.getStroke();
                                  stroke.setColor(strokeColor);
                                  stroke.setWidth(0.5);
                              }
                              style.setZIndex(index);
                          }
                      }
                  }
                  if (type != 1 && layer.type == 'line') {
                      color = !('line-pattern' in paint) && 'line-color' in paint ? colorWithOpacity(getValue(layer, 'paint', 'line-color', zoom, f, functionCache, featureState), getValue(layer, 'paint', 'line-opacity', zoom, f, functionCache, featureState)) : undefined;
                      var width = getValue(layer, 'paint', 'line-width', zoom, f, functionCache, featureState);
                      if (color && width > 0) {
                          ++stylesLength;
                          style = styles[stylesLength];
                          if (!style || !style.getStroke() || style.getFill() || style.getText()) {
                              style = new Style({ stroke: new Stroke() });
                              styles[stylesLength] = style;
                          }
                          stroke = style.getStroke();
                          stroke.setLineCap(getValue(layer, 'layout', 'line-cap', zoom, f, functionCache, featureState));
                          stroke.setLineJoin(getValue(layer, 'layout', 'line-join', zoom, f, functionCache, featureState));
                          stroke.setMiterLimit(getValue(layer, 'layout', 'line-miter-limit', zoom, f, functionCache, featureState));
                          stroke.setColor(color);
                          stroke.setWidth(width);
                          stroke.setLineDash(paint['line-dasharray'] ? getValue(layer, 'paint', 'line-dasharray', zoom, f, functionCache, featureState).map(function (x) {
                              return x * width;
                          }) : null);
                          style.setZIndex(index);
                      }
                  }
                  var hasImage = false;
                  var text = null;
                  var placementAngle = 0;
                  var icon$1 = (void 0), iconImg = (void 0), skipLabel = (void 0);
                  if ((type == 1 || type == 2) && 'icon-image' in layout) {
                      var iconImage = getValue(layer, 'layout', 'icon-image', zoom, f, functionCache, featureState);
                      if (iconImage) {
                          icon$1 = typeof iconImage === 'string' ? fromTemplate(iconImage, properties) : iconImage.toString();
                          var styleGeom = undefined;
                          if (spriteImage && spriteData && spriteData[icon$1]) {
                              var iconRotationAlignment = getValue(layer, 'layout', 'icon-rotation-alignment', zoom, f, functionCache, featureState);
                              if (type == 2) {
                                  var geom = feature.getGeometry();
                                  // ol package and ol-debug.js only
                                  if (geom.getFlatMidpoint || geom.getFlatMidpoints) {
                                      var extent = geom.getExtent();
                                      var size = Math.sqrt(Math.max(Math.pow((extent[2] - extent[0]) / resolution, 2), Math.pow((extent[3] - extent[1]) / resolution, 2)));
                                      if (size > 150) {
                                          //FIXME Do not hard-code a size of 150
                                          var midpoint = geom.getType() === 'MultiLineString' ? geom.getFlatMidpoints() : geom.getFlatMidpoint();
                                          if (!renderFeature) {
                                              renderFeatureCoordinates = [
                                                  NaN,
                                                  NaN
                                              ];
                                              renderFeature = new RenderFeature('Point', renderFeatureCoordinates, [], {}, null);
                                          }
                                          styleGeom = renderFeature;
                                          renderFeatureCoordinates[0] = midpoint[0];
                                          renderFeatureCoordinates[1] = midpoint[1];
                                          var placement = getValue(layer, 'layout', 'symbol-placement', zoom, f, functionCache, featureState);
                                          if (placement === 'line' && iconRotationAlignment === 'map') {
                                              var stride = geom.getStride();
                                              var coordinates = geom.getFlatCoordinates();
                                              for (var i$1 = 0, ii$1 = coordinates.length - stride; i$1 < ii$1; i$1 += stride) {
                                                  var x1 = coordinates[i$1];
                                                  var y1 = coordinates[i$1 + 1];
                                                  var x2 = coordinates[i$1 + stride];
                                                  var y2 = coordinates[i$1 + stride + 1];
                                                  var minX = Math.min(x1, x2);
                                                  var minY = Math.min(y1, y2);
                                                  var maxX = Math.max(x1, x2);
                                                  var maxY = Math.max(y1, y2);
                                                  if (midpoint[0] >= minX && midpoint[0] <= maxX && midpoint[1] >= minY && midpoint[1] <= maxY) {
                                                      placementAngle = Math.atan2(y1 - y2, x2 - x1);
                                                      break;
                                                  }
                                              }
                                          }
                                      }
                                  }
                              }
                              if (type !== 2 || styleGeom) {
                                  var iconSize = getValue(layer, 'layout', 'icon-size', zoom, f, functionCache, featureState);
                                  var iconColor = paint['icon-color'] !== undefined ? getValue(layer, 'paint', 'icon-color', zoom, f, functionCache, featureState) : null;
                                  if (!iconColor || iconColor.a !== 0) {
                                      var icon_cache_key$1 = icon$1 + '.' + iconSize;
                                      if (iconColor !== null) {
                                          icon_cache_key$1 += '.' + iconColor;
                                      }
                                      iconImg = iconImageCache[icon_cache_key$1];
                                      if (!iconImg) {
                                          var spriteImageData$1 = spriteData[icon$1];
                                          var declutterMode = getIconDeclutterMode(layer, zoom, f, functionCache);
                                          var displacement = (void 0);
                                          if ('icon-offset' in layout) {
                                              displacement = getValue(layer, 'layout', 'icon-offset', zoom, f, functionCache, featureState);
                                              displacement[1] *= -1;
                                          }
                                          iconImg = new Icon({
                                              color: iconColor ? [
                                                  iconColor.r * 255,
                                                  iconColor.g * 255,
                                                  iconColor.b * 255,
                                                  iconColor.a
                                              ] : undefined,
                                              img: spriteImage,
                                              imgSize: spriteImgSize,
                                              size: [
                                                  spriteImageData$1.width,
                                                  spriteImageData$1.height
                                              ],
                                              offset: [
                                                  spriteImageData$1.x,
                                                  spriteImageData$1.y
                                              ],
                                              rotateWithView: iconRotationAlignment === 'map',
                                              scale: iconSize / spriteImageData$1.pixelRatio,
                                              displacement: displacement,
                                              declutterMode: declutterMode
                                          });
                                          iconImageCache[icon_cache_key$1] = iconImg;
                                      }
                                  }
                                  if (iconImg) {
                                      ++stylesLength;
                                      style = styles[stylesLength];
                                      if (!style || !style.getImage() || style.getFill() || style.getStroke()) {
                                          style = new Style();
                                          styles[stylesLength] = style;
                                      }
                                      style.setGeometry(styleGeom);
                                      iconImg.setRotation(placementAngle + deg2rad(getValue(layer, 'layout', 'icon-rotate', zoom, f, functionCache, featureState)));
                                      iconImg.setOpacity(getValue(layer, 'paint', 'icon-opacity', zoom, f, functionCache, featureState));
                                      iconImg.setAnchor(anchor[getValue(layer, 'layout', 'icon-anchor', zoom, f, functionCache, featureState)]);
                                      style.setImage(iconImg);
                                      text = style.getText();
                                      style.setText(undefined);
                                      style.setZIndex(index);
                                      hasImage = true;
                                      skipLabel = false;
                                  }
                              } else {
                                  skipLabel = true;
                              }
                          }
                      }
                  }
                  if (type == 1 && layer.type === 'circle') {
                      ++stylesLength;
                      style = styles[stylesLength];
                      if (!style || !style.getImage() || style.getFill() || style.getStroke()) {
                          style = new Style();
                          styles[stylesLength] = style;
                      }
                      var circleRadius = 'circle-radius' in paint ? getValue(layer, 'paint', 'circle-radius', zoom, f, functionCache, featureState) : 5;
                      var circleStrokeColor = colorWithOpacity(getValue(layer, 'paint', 'circle-stroke-color', zoom, f, functionCache, featureState), getValue(layer, 'paint', 'circle-stroke-opacity', zoom, f, functionCache, featureState));
                      var circleColor = colorWithOpacity(getValue(layer, 'paint', 'circle-color', zoom, f, functionCache, featureState), getValue(layer, 'paint', 'circle-opacity', zoom, f, functionCache, featureState));
                      var circleStrokeWidth = getValue(layer, 'paint', 'circle-stroke-width', zoom, f, functionCache, featureState);
                      var cache_key = circleRadius + '.' + circleStrokeColor + '.' + circleColor + '.' + circleStrokeWidth;
                      iconImg = iconImageCache[cache_key];
                      if (!iconImg) {
                          iconImg = new CircleStyle({
                              radius: circleRadius,
                              stroke: circleStrokeColor && circleStrokeWidth > 0 ? new Stroke({
                                  width: circleStrokeWidth,
                                  color: circleStrokeColor
                              }) : undefined,
                              fill: circleColor ? new Fill({ color: circleColor }) : undefined,
                              declutterMode: 'none'
                          });
                          iconImageCache[cache_key] = iconImg;
                      }
                      style.setImage(iconImg);
                      text = style.getText();
                      style.setText(undefined);
                      style.setGeometry(undefined);
                      style.setZIndex(index);
                      hasImage = true;
                  }
                  var label = (void 0), font = (void 0), textLineHeight = (void 0), textSize = (void 0), letterSpacing = (void 0), maxTextWidth = (void 0);
                  if ('text-field' in layout) {
                      textSize = Math.round(getValue(layer, 'layout', 'text-size', zoom, f, functionCache, featureState));
                      var fontArray = getValue(layer, 'layout', 'text-font', zoom, f, functionCache, featureState);
                      textLineHeight = getValue(layer, 'layout', 'text-line-height', zoom, f, functionCache, featureState);
                      font = mapboxToCssFont(getFonts ? getFonts(fontArray) : fontArray, textSize, textLineHeight);
                      if (!font.includes('sans-serif')) {
                          font += ',sans-serif';
                      }
                      letterSpacing = getValue(layer, 'layout', 'text-letter-spacing', zoom, f, functionCache, featureState);
                      maxTextWidth = getValue(layer, 'layout', 'text-max-width', zoom, f, functionCache, featureState);
                      var textField = getValue(layer, 'layout', 'text-field', zoom, f, functionCache, featureState);
                      if (typeof textField === 'object' && textField.sections) {
                          if (textField.sections.length === 1) {
                              label = textField.toString();
                          } else {
                              label = textField.sections.reduce(function (acc, chunk, i) {
                                  var fonts = chunk.fontStack ? chunk.fontStack.split(',') : fontArray;
                                  var chunkFont = mapboxToCssFont(getFonts ? getFonts(fonts) : fonts, textSize * (chunk.scale || 1), textLineHeight);
                                  var text = chunk.text;
                                  if (text === '\n') {
                                      acc.push('\n', '');
                                      return acc;
                                  }
                                  if (type == 2) {
                                      acc.push(applyLetterSpacing(text, letterSpacing), chunkFont);
                                      return;
                                  }
                                  text = wrapText(text, chunkFont, maxTextWidth, letterSpacing).split('\n');
                                  for (var i$1 = 0, ii = text.length; i$1 < ii; ++i$1) {
                                      if (i$1 > 0) {
                                          acc.push('\n', '');
                                      }
                                      acc.push(text[i$1], chunkFont);
                                  }
                                  return acc;
                              }, []);
                          }
                      } else {
                          label = fromTemplate(textField, properties).trim();
                      }
                      opacity = getValue(layer, 'paint', 'text-opacity', zoom, f, functionCache, featureState);
                  }
                  if (label && opacity && !skipLabel) {
                      if (!hasImage) {
                          ++stylesLength;
                          style = styles[stylesLength];
                          if (!style || !style.getText() || style.getFill() || style.getStroke()) {
                              style = new Style();
                              styles[stylesLength] = style;
                          }
                          style.setImage(undefined);
                          style.setGeometry(undefined);
                      }
                      if (!style.getText()) {
                          style.setText(text || new Text({
                              padding: [
                                  2,
                                  2,
                                  2,
                                  2
                              ]
                          }));
                      }
                      text = style.getText();
                      var textTransform = layout['text-transform'];
                      if (textTransform == 'uppercase') {
                          label = Array.isArray(label) ? label.map(function (t, i) { return i % 2 ? t : t.toUpperCase(); }) : label.toUpperCase();
                      } else if (textTransform == 'lowercase') {
                          label = Array.isArray(label) ? label.map(function (t, i) { return i % 2 ? t : t.toLowerCase(); }) : label.toLowerCase();
                      }
                      var wrappedLabel = Array.isArray(label) ? label : type == 2 ? applyLetterSpacing(label, letterSpacing) : wrapText(label, font, maxTextWidth, letterSpacing);
                      text.setText(wrappedLabel);
                      text.setFont(font);
                      text.setRotation(deg2rad(getValue(layer, 'layout', 'text-rotate', zoom, f, functionCache, featureState)));
                      var textAnchor = getValue(layer, 'layout', 'text-anchor', zoom, f, functionCache, featureState);
                      var placement$1 = hasImage || type == 1 ? 'point' : getValue(layer, 'layout', 'symbol-placement', zoom, f, functionCache, featureState);
                      text.setPlacement(placement$1);
                      text.setOverflow(placement$1 === 'point');
                      var textHaloWidth = getValue(layer, 'paint', 'text-halo-width', zoom, f, functionCache, featureState);
                      var textOffset = getValue(layer, 'layout', 'text-offset', zoom, f, functionCache, featureState);
                      var textTranslate = getValue(layer, 'paint', 'text-translate', zoom, f, functionCache, featureState);
                      // Text offset has to take halo width and line height into account
                      var vOffset = 0;
                      var hOffset = 0;
                      if (placement$1 == 'point') {
                          var textAlign = 'center';
                          if (textAnchor.indexOf('left') !== -1) {
                              textAlign = 'left';
                              hOffset = textHaloWidth;
                          } else if (textAnchor.indexOf('right') !== -1) {
                              textAlign = 'right';
                              hOffset = -textHaloWidth;
                          }
                          text.setTextAlign(textAlign);
                          var textRotationAlignment = getValue(layer, 'layout', 'text-rotation-alignment', zoom, f, functionCache, featureState);
                          text.setRotateWithView(textRotationAlignment == 'map');
                      } else {
                          text.setMaxAngle(deg2rad(getValue(layer, 'layout', 'text-max-angle', zoom, f, functionCache, featureState)) * label.length / wrappedLabel.length);
                          text.setTextAlign();
                          text.setRotateWithView(false);
                      }
                      var textBaseline = 'middle';
                      if (textAnchor.indexOf('bottom') == 0) {
                          textBaseline = 'bottom';
                          vOffset = -textHaloWidth - 0.5 * (textLineHeight - 1) * textSize;
                      } else if (textAnchor.indexOf('top') == 0) {
                          textBaseline = 'top';
                          vOffset = textHaloWidth + 0.5 * (textLineHeight - 1) * textSize;
                      }
                      text.setTextBaseline(textBaseline);
                      var textJustify = getValue(layer, 'layout', 'text-justify', zoom, f, functionCache, featureState);
                      text.setJustify(textJustify === 'auto' ? undefined : textJustify);
                      text.setOffsetX(textOffset[0] * textSize + hOffset + textTranslate[0]);
                      text.setOffsetY(textOffset[1] * textSize + vOffset + textTranslate[1]);
                      textColor.setColor(colorWithOpacity(getValue(layer, 'paint', 'text-color', zoom, f, functionCache, featureState), opacity));
                      text.setFill(textColor);
                      var haloColor = colorWithOpacity(getValue(layer, 'paint', 'text-halo-color', zoom, f, functionCache, featureState), opacity);
                      if (haloColor) {
                          textHalo.setColor(haloColor);
                          // spec here : https://docs.mapbox.com/mapbox-gl-js/style-spec/#paint-symbol-text-halo-width
                          // Halo width must be doubled because it is applied around the center of the text outline
                          textHaloWidth *= 2;
                          // 1/4 of text size (spec) x 2
                          var halfTextSize = 0.5 * textSize;
                          textHalo.setWidth(textHaloWidth <= halfTextSize ? textHaloWidth : halfTextSize);
                          text.setStroke(textHalo);
                      } else {
                          text.setStroke(undefined);
                      }
                      var textPadding = getValue(layer, 'layout', 'text-padding', zoom, f, functionCache, featureState);
                      var padding = text.getPadding();
                      if (textPadding !== padding[0]) {
                          padding[0] = textPadding;
                          padding[1] = textPadding;
                          padding[2] = textPadding;
                          padding[3] = textPadding;
                      }
                      style.setZIndex(index);
                  }
              }
          }
          if (stylesLength > -1) {
              styles.length = stylesLength + 1;
              return styles;
          }
      };
      olLayer.setStyle(styleFunction);
      olLayer.set('mapbox-source', mapboxSource);
      olLayer.set('mapbox-layers', mapboxLayers);
      olLayer.set('mapbox-featurestate', {});
      return styleFunction;
  }

  /*
  ol-mapbox-style - Use Mapbox Style objects with OpenLayers
  Copyright 2016-present ol-mapbox-style contributors
  License: https://raw.githubusercontent.com/openlayers/ol-mapbox-style/master/LICENSE
  */
  /**
   * @typedef {Object} FeatureIdentifier
   * @property {string|number} id The feature id.
   * @property {string} source The source id.
   */
  /**
   * @typedef {Object} Options
   * @property {string} [accessToken] Access token for 'mapbox://' urls.
   * @property {function(string, ResourceType): (Request|void)} [transformRequest]
   * Function for controlling how `ol-mapbox-style` fetches resources. Can be used for modifying
   * the url, adding headers or setting credentials options. Called with the url and the resource
   * type as arguments, this function is supposed to return a `Request` object. Without a return value,
   * the original request will not be modified. For `Tiles` and `GeoJSON` resources, only the `url` of
   * the returned request will be respected.
   * @property {Array<number>} [resolutions] Resolutions for mapping resolution to zoom level.
   * Only needed when working with non-standard tile grids or projections.
   * @property {string} [styleUrl] URL of the Mapbox GL style. Required for styles that were provided
   * as object, when they contain a relative sprite url, or sources referencing data by relative url.
   * @property {string} [accessTokenParam='access_token'] Access token param. For internal use.
   */
  /** @typedef {'Style'|'Source'|'Sprite'|'SpriteImage'|'Tiles'|'GeoJSON'} ResourceType */
  /** @typedef {import("ol/layer/Layer").default} Layer */
  /** @typedef {import("ol/source/Source").default} Source */
  /**
   * @param {string} styleUrl Style URL.
   * @param {Options} options Options.
   * @return {Options} Completed options with accessToken and accessTokenParam.
   */
  function completeOptions(styleUrl, options) {
      if (!options.accessToken) {
          options = Object.assign({}, options);
          var searchParams = new URL(styleUrl).searchParams;
          // The last search parameter is the access token
          searchParams.forEach(function (value, key) {
              options.accessToken = value;
              options.accessTokenParam = key;
          });
      }
      return options;
  }
  /**
   * Applies a style function to an `ol/layer/VectorTile` or `ol/layer/Vector`
   * with an `ol/source/VectorTile` or an `ol/source/Vector`. If the layer does not have a source
   * yet, it will be created and populated from the information in the `glStyle`.
   *
   * **Example:**
   * ```js
   * import {applyStyle} from 'ol-mapbox-style';
   * import {VectorTile} from 'ol/layer.js';
   *
   * const layer = new VectorTile({declutter: true});
   * applyStyle(layer, 'https://api.maptiler.com/maps/basic/style.json?key=YOUR_OPENMAPTILES_TOKEN');
   * ```
   *
   * The style function will render all layers from the `glStyle` object that use the source
   * of the first layer, the specified `source`, or a subset of layers from the same source. The
   * source needs to be a `"type": "vector"` or `"type": "geojson"` source.
   *
   * Two additional properties will be set on the provided layer:
   *
   *  * `mapbox-source`: The `id` of the Mapbox Style document's source that the
   *    OpenLayers layer was created from. Usually `apply()` creates one
   *    OpenLayers layer per Mapbox Style source, unless the layer stack has
   *    layers from different sources in between.
   *  * `mapbox-layers`: The `id`s of the Mapbox Style document's layers that are
   *    included in the OpenLayers layer.
   *
   * @param {VectorTileLayer|VectorLayer} layer OpenLayers layer. When the layer has a source configured,
   * it will be modified to use the configuration from the glStyle's `source`. Options specified on the
   * layer's source will override those from the glStyle's `source`, except for `url`,
   * `tileUrlFunction` and `tileGrid` (exception: when the source projection is not `EPSG:3857`).
   * @param {string|Object} glStyle Mapbox Style object.
   * @param {string|Array<string>} sourceOrLayers `source` key or an array of layer `id`s from the
   * Mapbox Style object. When a `source` key is provided, all layers for the
   * specified source will be included in the style function. When layer `id`s
   * are provided, they must be from layers that use the same source. When not provided or a falsey
   * value, all layers using the first source specified in the glStyle will be rendered.
   * @param {Options|string} optionsOrPath Options. Alternatively the path of the style file
   * (only required when a relative path is used for the `"sprite"` property of the style).
   * @param {Array<number>} resolutions Resolutions for mapping resolution to zoom level.
   * Only needed when working with non-standard tile grids or projections.
   * @return {Promise} Promise which will be resolved when the style can be used
   * for rendering.
   */
  function applyStyle(layer, glStyle, sourceOrLayers, optionsOrPath, resolutions) {
      if ( sourceOrLayers === void 0 ) sourceOrLayers = '';
      if ( optionsOrPath === void 0 ) optionsOrPath = {};
      if ( resolutions === void 0 ) resolutions = undefined;

      var styleUrl, sourceId;
      /** @type {Options} */
      var options;
      if (typeof optionsOrPath === 'string') {
          styleUrl = optionsOrPath;
          options = {};
      } else {
          styleUrl = optionsOrPath.styleUrl;
          options = optionsOrPath;
      }
      if (!resolutions) {
          resolutions = options.resolutions;
      }
      if (!styleUrl && typeof glStyle === 'string' && !glStyle.trim().startsWith('{')) {
          styleUrl = glStyle;
      }
      if (styleUrl) {
          styleUrl = styleUrl.startsWith('data:') ? location.href : normalizeStyleUrl(styleUrl, options.accessToken);
          options = completeOptions(styleUrl, options);
      }
      return new Promise(function (resolve, reject) {
          // TODO: figure out where best place to check source type is
          // Note that the source arg is an array of gl layer ids and each must be
          // dereferenced to get source type to validate
          getGlStyle(glStyle, options).then(function (glStyle) {
              if (glStyle.version != 8) {
                  return reject(new Error('glStyle version 8 required.'));
              }
              if (!(layer instanceof VectorLayer || layer instanceof VectorTileLayer)) {
                  return reject(new Error('Can only apply to VectorLayer or VectorTileLayer'));
              }
              var type = layer instanceof VectorTileLayer ? 'vector' : 'geojson';
              if (!sourceOrLayers) {
                  sourceId = Object.keys(glStyle.sources).find(function (key) {
                      return glStyle.sources[key].type === type;
                  });
                  sourceOrLayers = sourceId;
              } else if (Array.isArray(sourceOrLayers)) {
                  sourceId = glStyle.layers.find(function (layer) {
                      return layer.id === sourceOrLayers[0];
                  }).source;
              } else {
                  sourceId = sourceOrLayers;
              }
              if (!sourceId) {
                  return reject(new Error(("No " + type + " source found in the glStyle.")));
              }
              function assignSource() {
                  if (layer instanceof VectorTileLayer) {
                      return setupVectorSource(glStyle.sources[sourceId], styleUrl, options).then(function (source) {
                          var targetSource = layer.getSource();
                          if (!targetSource) {
                              layer.setSource(source);
                          } else if (source !== targetSource) {
                              targetSource.setTileUrlFunction(source.getTileUrlFunction());
                              //@ts-ignore
                              if (!targetSource.format_) {
                                  //@ts-ignore
                                  targetSource.format_ = source.format_;
                              }
                              if (!targetSource.getAttributions()) {
                                  targetSource.setAttributions(source.getAttributions());
                              }
                              if (targetSource.getTileLoadFunction() === defaultLoadFunction) {
                                  targetSource.setTileLoadFunction(source.getTileLoadFunction());
                              }
                              if (equivalent(targetSource.getProjection(), source.getProjection())) {
                                  targetSource.tileGrid = source.getTileGrid();
                              }
                          }
                          if (!isFinite(layer.getMaxResolution()) && !isFinite(layer.getMinZoom())) {
                              var tileGrid = layer.getSource().getTileGrid();
                              layer.setMaxResolution(tileGrid.getResolution(tileGrid.getMinZoom()));
                          }
                      });
                  } else {
                      var glSource = glStyle.sources[sourceId];
                      var source = layer.getSource();
                      if (!source || source.get('mapbox-source') !== glSource) {
                          source = setupGeoJSONSource(glSource, styleUrl, options);
                      }
                      var targetSource = layer.getSource();
                      if (!targetSource) {
                          layer.setSource(source);
                      } else if (source !== targetSource) {
                          if (!targetSource.getAttributions()) {
                              targetSource.setAttributions(source.getAttributions());
                          }
                          //@ts-ignore
                          if (!targetSource.format_) {
                              //@ts-ignore
                              targetSource.format_ = source.getFormat();
                          }
                          //@ts-ignore
                          targetSource.url_ = source.getUrl();
                      }
                      return Promise.resolve();
                  }
              }
              var spriteScale, spriteData, spriteImageUrl, style;
              function onChange() {
                  if (!style && (!glStyle.sprite || spriteData)) {
                      style = stylefunction(layer, glStyle, sourceOrLayers, resolutions, spriteData, spriteImageUrl, getFonts);
                      if (!layer.getStyle()) {
                          reject(new Error(("Nothing to show for source [" + sourceId + "]")));
                      } else {
                          assignSource().then(resolve).catch(reject);
                      }
                  } else if (style) {
                      layer.setStyle(style);
                      assignSource().then(resolve).catch(reject);
                  } else {
                      reject(new Error('Something went wrong trying to apply style.'));
                  }
              }
              if (glStyle.sprite) {
                  var sprite = new URL(normalizeSpriteUrl(glStyle.sprite, options.accessToken, styleUrl || location.href));
                  spriteScale = window.devicePixelRatio >= 1.5 ? 0.5 : 1;
                  var sizeFactor = spriteScale == 0.5 ? '@2x' : '';
                  var spriteUrl = sprite.origin + sprite.pathname + sizeFactor + '.json' + sprite.search;
                  new Promise(function (resolve, reject) {
                      fetchResource('Sprite', spriteUrl, options).then(resolve).catch(function (error) {
                          spriteUrl = sprite.origin + sprite.pathname + '.json' + sprite.search;
                          fetchResource('Sprite', spriteUrl, options).then(resolve).catch(reject);
                      });
                  }).then(function (spritesJson) {
                      if (spritesJson === undefined) {
                          reject(new Error('No sprites found.'));
                      }
                      spriteData = spritesJson;
                      spriteImageUrl = sprite.origin + sprite.pathname + sizeFactor + '.png' + sprite.search;
                      if (options.transformRequest) {
                          var transformed = options.transformRequest(spriteImageUrl, 'SpriteImage');
                          if (transformed instanceof Request) {
                              spriteImageUrl = encodeURI(transformed.url);
                          }
                      }
                      onChange();
                  }).catch(function (err) {
                      reject(new Error(("Sprites cannot be loaded: " + spriteUrl + ": " + (err.message))));
                  });
              } else {
                  onChange();
              }
          }).catch(reject);
      });
  }
  var emptyObj = {};
  function setBackground(mapOrLayer, layer) {
      var background = {
          id: layer.id,
          type: layer.type
      };
      var functionCache = {};
      function updateStyle(resolution) {
          var layout = layer.layout || {};
          var paint = layer.paint || {};
          background['paint'] = paint;
          var zoom = typeof mapOrLayer.getSource === 'function' ? mapOrLayer.getSource().getTileGrid().getZForResolution(resolution) : mapOrLayer.getView().getZoom();
          var element = typeof mapOrLayer.getTargetElement === 'function' ? mapOrLayer.getTargetElement() : undefined;
          var bg, opacity;
          if (paint['background-color'] !== undefined) {
              bg = getValue(background, 'paint', 'background-color', zoom, emptyObj, functionCache);
              if (element) {
                  element.style.background = Color.parse(bg).toString();
              }
          }
          if (paint['background-opacity'] !== undefined) {
              opacity = getValue(background, 'paint', 'background-opacity', zoom, emptyObj, functionCache);
              if (element) {
                  element.style.opacity = opacity;
              }
          }
          if (layout.visibility == 'none') {
              if (element) {
                  element.style.backgroundColor = '';
                  element.style.opacity = '';
              }
              return undefined;
          }
          return colorWithOpacity(bg, opacity);
      }
      if (typeof mapOrLayer.getTargetElement === 'function') {
          if (mapOrLayer.getTargetElement()) {
              updateStyle();
          }
          mapOrLayer.on([
              'change:resolution',
              'change:target'
          ], updateStyle);
      } else if (typeof mapOrLayer.setBackground === 'function') {
          mapOrLayer.setBackground(updateStyle);
      } else {
          throw new Error('Unable to apply background.');
      }
  }
  function setFirstBackground(mapOrLayer, glStyle) {
      glStyle.layers.some(function (layer) {
          if (layer.type === 'background') {
              setBackground(mapOrLayer, layer);
              return true;
          }
      });
  }
  /**
   * Applies properties of the Mapbox Style's first `background` layer to the
   * provided map or VectorTile layer.
   *
   * **Example:**
   * ```js
   * import {applyBackground} from 'ol-mapbox-style';
   * import {Map} from 'ol';
   *
   * const map = new Map({target: 'map'});
   * applyBackground(map, 'https://api.maptiler.com/maps/basic/style.json?key=YOUR_OPENMAPTILES_TOKEN');
   * ```
   * @param {Map|VectorTileLayer} mapOrLayer OpenLayers Map or VectorTile layer.
   * @param {Object|string} glStyle Mapbox Style object or url.
   * @param {Options} options Options.
   * @return {Promise} Promise that resolves when the background is applied.
   */
  function applyBackground(mapOrLayer, glStyle, options) {
      if ( options === void 0 ) options = {};

      if (typeof glStyle === 'object') {
          setFirstBackground(mapOrLayer, glStyle);
          return Promise.resolve();
      }
      return getGlStyle(glStyle, options).then(function (glStyle) {
          setFirstBackground(mapOrLayer, glStyle);
      });
  }
  function extentFromTileJSON(tileJSON) {
      var bounds = tileJSON.bounds;
      if (bounds) {
          var ll = fromLonLat([
              bounds[0],
              bounds[1]
          ]);
          var tr = fromLonLat([
              bounds[2],
              bounds[3]
          ]);
          return [
              ll[0],
              ll[1],
              tr[0],
              tr[1]
          ];
      }
  }
  /**
   * Creates an OpenLayers VectorTile source for a gl source entry.
   * @param {Object} glSource "source" entry from a Mapbox Style object.
   * @param {string|undefined} styleUrl URL to use for the source. This is expected to be the complete http(s) url,
   * with access key applied.
   * @param {Options} options Options.
   * @return {Promise<import("ol/source/VectorTile").default>} Promise resolving to a VectorTile source.
   * @private
   */
  function setupVectorSource(glSource, styleUrl, options) {
      return new Promise(function (resolve, reject) {
          getTileJson(glSource, styleUrl, options).then(function (tileJSON) {
              var tileJSONSource = new TileJSON({ tileJSON: tileJSON });
              var tileJSONDoc = tileJSONSource.getTileJSON();
              var tileGrid = tileJSONSource.getTileGrid();
              var extent = extentFromTileJSON(tileJSONDoc);
              var minZoom = tileJSONDoc.minzoom || 0;
              var maxZoom = tileJSONDoc.maxzoom || 22;
              var sourceOptions = {
                  attributions: tileJSONSource.getAttributions(),
                  format: new MVT(),
                  tileGrid: new TileGrid({
                      origin: tileGrid.getOrigin(0),
                      extent: extent || tileGrid.getExtent(),
                      minZoom: minZoom,
                      resolutions: defaultResolutions.slice(0, maxZoom + 1),
                      tileSize: 512
                  })
              };
              if (Array.isArray(tileJSONDoc.tiles)) {
                  sourceOptions.urls = tileJSONDoc.tiles;
              } else {
                  sourceOptions.url = tileJSONDoc.tiles;
              }
              if (tileJSON.olSourceOptions) {
                  Object.assign(sourceOptions, tileJSON.olSourceOptions);
              }
              resolve(new VectorTile(sourceOptions));
          }).catch(reject);
      });
  }
  var geoJsonFormat = new GeoJSON();
  /**
   * @param {Object} glSource glStyle source.
   * @param {string} styleUrl Style URL.
   * @param {Options} options Options.
   * @return {VectorSource} Configured vector source.
   */
  function setupGeoJSONSource(glSource, styleUrl, options) {
      var data = glSource.data;
      var sourceOptions = {};
      if (typeof data == 'string') {
          var geoJsonUrl = normalizeSourceUrl(data, options.accessToken, options.accessTokenParam || 'access_token', styleUrl || location.href);
          if (options.transformRequest) {
              var transformed = options.transformRequest(geoJsonUrl, 'GeoJSON');
              if (transformed instanceof Request) {
                  geoJsonUrl = encodeURI(transformed.url);
              }
          }
          sourceOptions.url = geoJsonUrl;
      } else {
          sourceOptions.features = geoJsonFormat.readFeatures(data, { featureProjection: getUserProjection() || 'EPSG:3857' });
      }
      var source = new VectorSource(Object.assign({
          attributions: glSource.attribution,
          format: geoJsonFormat
      }, sourceOptions));
      source.set('mapbox-source', glSource);
      return source;
  }

  /**
   * @module ol/layer/MapboxVector
   */

  /**
   * @classdesc
   * Event emitted on configuration or loading error.
   */
  class ErrorEvent extends BaseEvent {
    /**
     * @param {Error} error error object.
     */
    constructor(error) {
      super(EventType$1.ERROR);

      /**
       * @type {Error}
       */
      this.error = error;
    }
  }

  /**
   * @typedef {Object} Options
   * @property {string} styleUrl The URL of the Mapbox style object to use for this layer.  For a
   * style created with Mapbox Studio and hosted on Mapbox, this will look like
   * 'mapbox://styles/you/your-style'.
   * @property {string} [accessToken] The access token for your Mapbox style. This has to be provided
   * for `mapbox://` style urls. For `https://` and other urls, any access key must be the last query
   * parameter of the style url.
   * @property {string} [source] If your style uses more than one source, you need to use either the
   * `source` property or the `layers` property to limit rendering to a single vector source.  The
   * `source` property corresponds to the id of a vector source in your Mapbox style.
   * @property {Array<string>} [layers] Limit rendering to the list of included layers.  All layers
   * must share the same vector source.  If your style uses more than one source, you need to use
   * either the `source` property or the `layers` property to limit rendering to a single vector
   * source.
   * @property {boolean} [declutter=true] Declutter images and text. Decluttering is applied to all
   * image and text styles of all Vector and VectorTile layers that have set this to `true`. The priority
   * is defined by the z-index of the layer, the `zIndex` of the style and the render order of features.
   * Higher z-index means higher priority. Within the same z-index, a feature rendered before another has
   * higher priority.
   *
   * As an optimization decluttered features from layers with the same `className` are rendered above
   * the fill and stroke styles of all of those layers regardless of z-index.  To opt out of this
   * behavior and place declutterd features with their own layer configure the layer with a `className`
   * other than `ol-layer`.
   * @property {import("./Base.js").BackgroundColor|false} [background] Background color for the layer.
   * If not specified, the background from the Mapbox style object will be used. Set to `false` to prevent
   * the Mapbox style's background from being used.
   * @property {string} [className='ol-layer'] A CSS class name to set to the layer element.
   * @property {number} [opacity=1] Opacity (0, 1).
   * @property {boolean} [visible=true] Visibility.
   * @property {import("../extent.js").Extent} [extent] The bounding extent for layer rendering.  The layer will not be
   * rendered outside of this extent.
   * @property {number} [zIndex] The z-index for layer rendering.  At rendering time, the layers
   * will be ordered, first by Z-index and then by position. When `undefined`, a `zIndex` of 0 is assumed
   * for layers that are added to the map's `layers` collection, or `Infinity` when the layer's `setMap()`
   * method was used.
   * @property {number} [minResolution] The minimum resolution (inclusive) at which this layer will be
   * visible.
   * @property {number} [maxResolution] The maximum resolution (exclusive) below which this layer will
   * be visible. If neither `maxResolution` nor `minZoom` are defined, the layer's `maxResolution` will
   * match the style source's `minzoom`.
   * @property {number} [minZoom] The minimum view zoom level (exclusive) above which this layer will
   * be visible. If neither `maxResolution` nor `minZoom` are defined, the layer's `minZoom` will match
   * the style source's `minzoom`.
   * @property {number} [maxZoom] The maximum view zoom level (inclusive) at which this layer will
   * be visible.
   * @property {import("../render.js").OrderFunction} [renderOrder] Render order. Function to be used when sorting
   * features before rendering. By default features are drawn in the order that they are created. Use
   * `null` to avoid the sort, but get an undefined draw order.
   * @property {number} [renderBuffer=100] The buffer in pixels around the tile extent used by the
   * renderer when getting features from the vector tile for the rendering or hit-detection.
   * Recommended value: Vector tiles are usually generated with a buffer, so this value should match
   * the largest possible buffer of the used tiles. It should be at least the size of the largest
   * point symbol or line width.
   * @property {import("./VectorTile.js").VectorTileRenderType} [renderMode='hybrid'] Render mode for vector tiles:
   *  * `'hybrid'`: Polygon and line elements are rendered as images, so pixels are scaled during zoom
   *    animations. Point symbols and texts are accurately rendered as vectors and can stay upright on
   *    rotated views.
   *  * `'vector'`: Everything is rendered as vectors. Use this mode for improved performance on vector
   *    tile layers with only a few rendered features (e.g. for highlighting a subset of features of
   *    another layer with the same source).
   * @property {import("../Map.js").default} [map] Sets the layer as overlay on a map. The map will not manage
   * this layer in its layers collection, and the layer will be rendered on top. This is useful for
   * temporary layers. The standard way to add a layer to a map and have it managed by the map is to
   * use [map.addLayer()]{@link import("../Map.js").default#addLayer}.
   * @property {boolean} [updateWhileAnimating=false] When set to `true`, feature batches will be
   * recreated during animations. This means that no vectors will be shown clipped, but the setting
   * will have a performance impact for large amounts of vector data. When set to `false`, batches
   * will be recreated when no animation is active.
   * @property {boolean} [updateWhileInteracting=false] When set to `true`, feature batches will be
   * recreated during interactions. See also `updateWhileAnimating`.
   * @property {number} [preload=0] Preload. Load low-resolution tiles up to `preload` levels. `0`
   * means no preloading.
   * @property {boolean} [useInterimTilesOnError=true] Use interim tiles on error.
   * @property {Object<string, *>} [properties] Arbitrary observable properties. Can be accessed with `#get()` and `#set()`.
   */

  /**
   * @classdesc
   * A vector tile layer based on a Mapbox style that uses a single vector source.  Configure
   * the layer with the `styleUrl` and `accessToken` shown in Mapbox Studio's share panel.
   * If the style uses more than one source, use the `source` property to choose a single
   * vector source.  If you want to render a subset of the layers in the style, use the `layers`
   * property (all layers must share the same vector source).  See the constructor options for
   * more detail.
   *
   *     const map = new Map({
   *       view: new View({
   *         center: [0, 0],
   *         zoom: 1,
   *       }),
   *       layers: [
   *         new MapboxVectorLayer({
   *           styleUrl: 'mapbox://styles/mapbox/bright-v9',
   *           accessToken: 'your-mapbox-access-token-here',
   *         }),
   *       ],
   *       target: 'map',
   *     });
   *
   * On configuration or loading error, the layer will trigger an `'error'` event.  Listeners
   * will receive an object with an `error` property that can be used to diagnose the problem.
   *
   * @param {Options} options Options.
   * @extends {VectorTileLayer}
   * @fires module:ol/events/Event~BaseEvent#event:error
   * @api
   */
  class MapboxVectorLayer extends VectorTileLayer {
    /**
     * @param {Options} options Layer options.  At a minimum, `styleUrl` and `accessToken`
     * must be provided.
     */
    constructor(options) {
      const declutter = 'declutter' in options ? options.declutter : true;
      const source = new VectorTile({
        state: 'loading',
        format: new MVT(),
      });

      super({
        source: source,
        background: options.background,
        declutter: declutter,
        className: options.className,
        opacity: options.opacity,
        visible: options.visible,
        zIndex: options.zIndex,
        minResolution: options.minResolution,
        maxResolution: options.maxResolution,
        minZoom: options.minZoom,
        maxZoom: options.maxZoom,
        renderOrder: options.renderOrder,
        renderBuffer: options.renderBuffer,
        renderMode: options.renderMode,
        map: options.map,
        updateWhileAnimating: options.updateWhileAnimating,
        updateWhileInteracting: options.updateWhileInteracting,
        preload: options.preload,
        useInterimTilesOnError: options.useInterimTilesOnError,
        properties: options.properties,
      });

      if (options.accessToken) {
        this.accessToken = options.accessToken;
      }
      const url = options.styleUrl;
      applyStyle(this, url, options.layers || options.source, {
        accessToken: this.accessToken,
      })
        .then(() => {
          source.setState('ready');
        })
        .catch((error) => {
          this.dispatchEvent(new ErrorEvent(error));
          const source = this.getSource();
          source.setState('error');
        });
      if (this.getBackground() === undefined) {
        applyBackground(this, options.styleUrl, {
          accessToken: this.accessToken,
        });
      }
    }
  }

  /**
   * @module ol/layer/VectorImage
   */

  /**
   * @template {import("../source/Vector.js").default} VectorSourceType
   * @typedef {Object} Options
   * @property {string} [className='ol-layer'] A CSS class name to set to the layer element.
   * @property {number} [opacity=1] Opacity (0, 1).
   * @property {boolean} [visible=true] Visibility.
   * @property {import("../extent.js").Extent} [extent] The bounding extent for layer rendering.  The layer will not be
   * rendered outside of this extent.
   * @property {number} [zIndex] The z-index for layer rendering.  At rendering time, the layers
   * will be ordered, first by Z-index and then by position. When `undefined`, a `zIndex` of 0 is assumed
   * for layers that are added to the map's `layers` collection, or `Infinity` when the layer's `setMap()`
   * method was used.
   * @property {number} [minResolution] The minimum resolution (inclusive) at which this layer will be
   * visible.
   * @property {number} [maxResolution] The maximum resolution (exclusive) below which this layer will
   * be visible.
   * @property {number} [minZoom] The minimum view zoom level (exclusive) above which this layer will be
   * visible.
   * @property {number} [maxZoom] The maximum view zoom level (inclusive) at which this layer will
   * be visible.
   * @property {import("../render.js").OrderFunction} [renderOrder] Render order. Function to be used when sorting
   * features before rendering. By default features are drawn in the order that they are created. Use
   * `null` to avoid the sort, but get an undefined draw order.
   * @property {number} [renderBuffer=100] The buffer in pixels around the viewport extent used by the
   * renderer when getting features from the vector source for the rendering or hit-detection.
   * Recommended value: the size of the largest symbol, line width or label.
   * @property {VectorSourceType} [source] Source.
   * @property {import("../Map.js").default} [map] Sets the layer as overlay on a map. The map will not manage
   * this layer in its layers collection, and the layer will be rendered on top. This is useful for
   * temporary layers. The standard way to add a layer to a map and have it managed by the map is to
   * use [map.addLayer()]{@link import("../Map.js").default#addLayer}.
   * @property {boolean} [declutter=false] Declutter images and text on this layer. The priority is defined
   * by the `zIndex` of the style and the render order of features. Higher z-index means higher priority.
   * Within the same z-index, a feature rendered before another has higher priority.
   * @property {import("../style/Style.js").StyleLike|null} [style] Layer style. When set to `null`, only
   * features that have their own style will be rendered. See {@link module:ol/style/Style~Style} for the default style
   * which will be used if this is not set.
   * @property {number} [imageRatio=1] Ratio by which the rendered extent should be larger than the
   * viewport extent. A larger ratio avoids cut images during panning, but will cause a decrease in performance.
   * @property {Object<string, *>} [properties] Arbitrary observable properties. Can be accessed with `#get()` and `#set()`.
   */

  /**
   * @classdesc
   * Vector data is rendered client-side, to an image. This layer type provides great performance
   * during panning and zooming, but point symbols and texts are always rotated with the view and
   * pixels are scaled during zoom animations. For more accurate rendering of vector data, use
   * {@link module:ol/layer/Vector~VectorLayer} instead.
   *
   * Note that any property set in the options is set as a {@link module:ol/Object~BaseObject}
   * property on the layer object; for example, setting `title: 'My Title'` in the
   * options means that `title` is observable, and has get/set accessors.
   *
   * @template {import("../source/Vector.js").default} VectorSourceType
   * @extends {BaseVectorLayer<VectorSourceType, CanvasVectorImageLayerRenderer>}
   * @api
   */
  class VectorImageLayer extends BaseVectorLayer {
    /**
     * @param {Options<VectorSourceType>} [options] Options.
     */
    constructor(options) {
      options = options ? options : {};

      const baseOptions = Object.assign({}, options);
      delete baseOptions.imageRatio;
      super(baseOptions);

      /**
       * @type {number}
       * @private
       */
      this.imageRatio_ =
        options.imageRatio !== undefined ? options.imageRatio : 1;
    }

    /**
     * @return {number} Ratio between rendered extent size and viewport extent size.
     */
    getImageRatio() {
      return this.imageRatio_;
    }

    createRenderer() {
      return new CanvasVectorImageLayerRenderer(this);
    }
  }

  /**
   * @module ol/layer/WebGLPoints
   */

  /**
   * @template {import("../source/Vector.js").default<import("../geom/Point.js").default>} VectorSourceType
   * @typedef {Object} Options
   * @property {import('../style/literal.js').LiteralStyle} style Literal style to apply to the layer features.
   * @property {string} [className='ol-layer'] A CSS class name to set to the layer element.
   * @property {number} [opacity=1] Opacity (0, 1).
   * @property {boolean} [visible=true] Visibility.
   * @property {import("../extent.js").Extent} [extent] The bounding extent for layer rendering.  The layer will not be
   * rendered outside of this extent.
   * @property {number} [zIndex] The z-index for layer rendering.  At rendering time, the layers
   * will be ordered, first by Z-index and then by position. When `undefined`, a `zIndex` of 0 is assumed
   * for layers that are added to the map's `layers` collection, or `Infinity` when the layer's `setMap()`
   * method was used.
   * @property {number} [minResolution] The minimum resolution (inclusive) at which this layer will be
   * visible.
   * @property {number} [maxResolution] The maximum resolution (exclusive) below which this layer will
   * be visible.
   * @property {number} [minZoom] The minimum view zoom level (exclusive) above which this layer will be
   * visible.
   * @property {number} [maxZoom] The maximum view zoom level (inclusive) at which this layer will
   * be visible.
   * @property {VectorSourceType} [source] Point source.
   * @property {boolean} [disableHitDetection=false] Setting this to true will provide a slight performance boost, but will
   * prevent all hit detection on the layer.
   * @property {Object<string, *>} [properties] Arbitrary observable properties. Can be accessed with `#get()` and `#set()`.
   */

  /**
   * @classdesc
   * Layer optimized for rendering large point datasets. Takes a `style` property which
   * is a serializable JSON object describing how the layer should be rendered.
   *
   * Here are a few samples of literal style objects:
   * ```js
   * const style = {
   *   symbol: {
   *     symbolType: 'circle',
   *     size: 8,
   *     color: '#33AAFF',
   *     opacity: 0.9
   *   }
   * }
   * ```
   *
   * ```js
   * const style = {
   *   symbol: {
   *     symbolType: 'image',
   *     offset: [0, 12],
   *     size: [4, 8],
   *     src: '../static/exclamation-mark.png'
   *   }
   * }
   * ```
   *
   * **Important: a `WebGLPoints` layer must be manually disposed when removed, otherwise the underlying WebGL context
   * will not be garbage collected.**
   *
   * Note that any property set in the options is set as a {@link module:ol/Object~BaseObject}
   * property on the layer object; for example, setting `title: 'My Title'` in the
   * options means that `title` is observable, and has get/set accessors.
   *
   * @template {import("../source/Vector.js").default<import("../geom/Point.js").default>} VectorSourceType
   * @extends {Layer<VectorSourceType, WebGLPointsLayerRenderer>}
   * @fires import("../render/Event.js").RenderEvent
   */
  class WebGLPointsLayer extends Layer {
    /**
     * @param {Options<VectorSourceType>} options Options.
     */
    constructor(options) {
      const baseOptions = Object.assign({}, options);

      super(baseOptions);

      /**
       * @private
       * @type {import('../webgl/ShaderBuilder.js').StyleParseResult}
       */
      this.parseResult_ = parseLiteralStyle(options.style);

      /**
       * @type {Object<string, (string|number)>}
       * @private
       */
      this.styleVariables_ = options.style.variables || {};

      /**
       * @private
       * @type {boolean}
       */
      this.hitDetectionDisabled_ = !!options.disableHitDetection;
    }

    createRenderer() {
      return new WebGLPointsLayerRenderer(this, {
        vertexShader: this.parseResult_.builder.getSymbolVertexShader(),
        fragmentShader: this.parseResult_.builder.getSymbolFragmentShader(),
        hitVertexShader:
          !this.hitDetectionDisabled_ &&
          this.parseResult_.builder.getSymbolVertexShader(true),
        hitFragmentShader:
          !this.hitDetectionDisabled_ &&
          this.parseResult_.builder.getSymbolFragmentShader(true),
        uniforms: this.parseResult_.uniforms,
        attributes: this.parseResult_.attributes,
      });
    }

    /**
     * Update any variables used by the layer style and trigger a re-render.
     * @param {Object<string, number>} variables Variables to update.
     */
    updateStyleVariables(variables) {
      Object.assign(this.styleVariables_, variables);
      this.changed();
    }
  }

  /**
   * @module ol/layer/WebGLTile
   */

  /**
   * @typedef {import("../source/DataTile.js").default|import("../source/TileImage.js").default} SourceType
   */

  /**
   * @typedef {Object} Style
   * Translates tile data to rendered pixels.
   *
   * @property {Object<string, (string|number)>} [variables] Style variables.  Each variable must hold a number or string.  These
   * variables can be used in the `color`, `brightness`, `contrast`, `exposure`, `saturation` and `gamma`
   * {@link import("../style/expressions.js").ExpressionValue expressions}, using the `['var', 'varName']` operator.
   * To update style variables, use the {@link import("./WebGLTile.js").default#updateStyleVariables} method.
   * @property {import("../style/expressions.js").ExpressionValue} [color] An expression applied to color values.
   * @property {import("../style/expressions.js").ExpressionValue} [brightness=0] Value used to decrease or increase
   * the layer brightness.  Values range from -1 to 1.
   * @property {import("../style/expressions.js").ExpressionValue} [contrast=0] Value used to decrease or increase
   * the layer contrast.  Values range from -1 to 1.
   * @property {import("../style/expressions.js").ExpressionValue} [exposure=0] Value used to decrease or increase
   * the layer exposure.  Values range from -1 to 1.
   * @property {import("../style/expressions.js").ExpressionValue} [saturation=0] Value used to decrease or increase
   * the layer saturation.  Values range from -1 to 1.
   * @property {import("../style/expressions.js").ExpressionValue} [gamma=1] Apply a gamma correction to the layer.
   * Values range from 0 to infinity.
   */

  /**
   * @typedef {Object} Options
   * @property {Style} [style] Style to apply to the layer.
   * @property {string} [className='ol-layer'] A CSS class name to set to the layer element.
   * @property {number} [opacity=1] Opacity (0, 1).
   * @property {boolean} [visible=true] Visibility.
   * @property {import("../extent.js").Extent} [extent] The bounding extent for layer rendering.  The layer will not be
   * rendered outside of this extent.
   * @property {number} [zIndex] The z-index for layer rendering.  At rendering time, the layers
   * will be ordered, first by Z-index and then by position. When `undefined`, a `zIndex` of 0 is assumed
   * for layers that are added to the map's `layers` collection, or `Infinity` when the layer's `setMap()`
   * method was used.
   * @property {number} [minResolution] The minimum resolution (inclusive) at which this layer will be
   * visible.
   * @property {number} [maxResolution] The maximum resolution (exclusive) below which this layer will
   * be visible.
   * @property {number} [minZoom] The minimum view zoom level (exclusive) above which this layer will be
   * visible.
   * @property {number} [maxZoom] The maximum view zoom level (inclusive) at which this layer will
   * be visible.
   * @property {number} [preload=0] Preload. Load low-resolution tiles up to `preload` levels. `0`
   * means no preloading.
   * @property {SourceType} [source] Source for this layer.
   * @property {Array<SourceType>|function(import("../extent.js").Extent, number):Array<SourceType>} [sources] Array
   * of sources for this layer. Takes precedence over `source`. Can either be an array of sources, or a function that
   * expects an extent and a resolution (in view projection units per pixel) and returns an array of sources. See
   * {@link module:ol/source.sourcesFromTileGrid} for a helper function to generate sources that are organized in a
   * pyramid following the same pattern as a tile grid. **Note:** All sources must have the same band count and content.
   * @property {import("../Map.js").default} [map] Sets the layer as overlay on a map. The map will not manage
   * this layer in its layers collection, and the layer will be rendered on top. This is useful for
   * temporary layers. The standard way to add a layer to a map and have it managed by the map is to
   * use {@link module:ol/Map~Map#addLayer}.
   * @property {boolean} [useInterimTilesOnError=true] Use interim tiles on error.
   * @property {number} [cacheSize=512] The internal texture cache size.  This needs to be large enough to render
   * two zoom levels worth of tiles.
   */

  /**
   * @typedef {Object} ParsedStyle
   * @property {string} vertexShader The vertex shader.
   * @property {string} fragmentShader The fragment shader.
   * @property {Object<string,import("../webgl/Helper.js").UniformValue>} uniforms Uniform definitions.
   * @property {Array<import("../webgl/PaletteTexture.js").default>} paletteTextures Palette textures.
   */

  /**
   * @param {Style} style The layer style.
   * @param {number} [bandCount] The number of bands.
   * @return {ParsedStyle} Shaders and uniforms generated from the style.
   */
  function parseStyle(style, bandCount) {
    const vertexShader = `
    attribute vec2 ${Attributes$3.TEXTURE_COORD};
    uniform mat4 ${Uniforms.TILE_TRANSFORM};
    uniform float ${Uniforms.TEXTURE_PIXEL_WIDTH};
    uniform float ${Uniforms.TEXTURE_PIXEL_HEIGHT};
    uniform float ${Uniforms.TEXTURE_RESOLUTION};
    uniform float ${Uniforms.TEXTURE_ORIGIN_X};
    uniform float ${Uniforms.TEXTURE_ORIGIN_Y};
    uniform float ${Uniforms.DEPTH};

    varying vec2 v_textureCoord;
    varying vec2 v_mapCoord;

    void main() {
      v_textureCoord = ${Attributes$3.TEXTURE_COORD};
      v_mapCoord = vec2(
        ${Uniforms.TEXTURE_ORIGIN_X} + ${Uniforms.TEXTURE_RESOLUTION} * ${Uniforms.TEXTURE_PIXEL_WIDTH} * v_textureCoord[0],
        ${Uniforms.TEXTURE_ORIGIN_Y} - ${Uniforms.TEXTURE_RESOLUTION} * ${Uniforms.TEXTURE_PIXEL_HEIGHT} * v_textureCoord[1]
      );
      gl_Position = ${Uniforms.TILE_TRANSFORM} * vec4(${Attributes$3.TEXTURE_COORD}, ${Uniforms.DEPTH}, 1.0);
    }
  `;

    /**
     * @type {import("../style/expressions.js").ParsingContext}
     */
    const context = {
      inFragmentShader: true,
      variables: [],
      attributes: [],
      stringLiteralsMap: {},
      functions: {},
      bandCount: bandCount,
    };

    const pipeline = [];

    if (style.color !== undefined) {
      const color = expressionToGlsl(context, style.color, ValueTypes.COLOR);
      pipeline.push(`color = ${color};`);
    }

    if (style.contrast !== undefined) {
      const contrast = expressionToGlsl(
        context,
        style.contrast,
        ValueTypes.NUMBER
      );
      pipeline.push(
        `color.rgb = clamp((${contrast} + 1.0) * color.rgb - (${contrast} / 2.0), vec3(0.0, 0.0, 0.0), vec3(1.0, 1.0, 1.0));`
      );
    }

    if (style.exposure !== undefined) {
      const exposure = expressionToGlsl(
        context,
        style.exposure,
        ValueTypes.NUMBER
      );
      pipeline.push(
        `color.rgb = clamp((${exposure} + 1.0) * color.rgb, vec3(0.0, 0.0, 0.0), vec3(1.0, 1.0, 1.0));`
      );
    }

    if (style.saturation !== undefined) {
      const saturation = expressionToGlsl(
        context,
        style.saturation,
        ValueTypes.NUMBER
      );
      pipeline.push(`
      float saturation = ${saturation} + 1.0;
      float sr = (1.0 - saturation) * 0.2126;
      float sg = (1.0 - saturation) * 0.7152;
      float sb = (1.0 - saturation) * 0.0722;
      mat3 saturationMatrix = mat3(
        sr + saturation, sr, sr,
        sg, sg + saturation, sg,
        sb, sb, sb + saturation
      );
      color.rgb = clamp(saturationMatrix * color.rgb, vec3(0.0, 0.0, 0.0), vec3(1.0, 1.0, 1.0));
    `);
    }

    if (style.gamma !== undefined) {
      const gamma = expressionToGlsl(context, style.gamma, ValueTypes.NUMBER);
      pipeline.push(`color.rgb = pow(color.rgb, vec3(1.0 / ${gamma}));`);
    }

    if (style.brightness !== undefined) {
      const brightness = expressionToGlsl(
        context,
        style.brightness,
        ValueTypes.NUMBER
      );
      pipeline.push(
        `color.rgb = clamp(color.rgb + ${brightness}, vec3(0.0, 0.0, 0.0), vec3(1.0, 1.0, 1.0));`
      );
    }

    /** @type {Object<string,import("../webgl/Helper").UniformValue>} */
    const uniforms = {};

    const numVariables = context.variables.length;
    if (numVariables > 1 && !style.variables) {
      throw new Error(
        `Missing variables in style (expected ${context.variables})`
      );
    }

    for (let i = 0; i < numVariables; ++i) {
      const variableName = context.variables[i];
      if (!(variableName in style.variables)) {
        throw new Error(`Missing '${variableName}' in style variables`);
      }
      const uniformName = uniformNameForVariable(variableName);
      uniforms[uniformName] = function () {
        let value = style.variables[variableName];
        if (typeof value === 'string') {
          value = getStringNumberEquivalent(context, value);
        }
        return value !== undefined ? value : -9999999; // to avoid matching with the first string literal
      };
    }

    const uniformDeclarations = Object.keys(uniforms).map(function (name) {
      return `uniform float ${name};`;
    });

    const textureCount = Math.ceil(bandCount / 4);
    uniformDeclarations.push(
      `uniform sampler2D ${Uniforms.TILE_TEXTURE_ARRAY}[${textureCount}];`
    );

    if (context.paletteTextures) {
      uniformDeclarations.push(
        `uniform sampler2D ${PALETTE_TEXTURE_ARRAY}[${context.paletteTextures.length}];`
      );
    }

    const functionDefintions = Object.keys(context.functions).map(function (
      name
    ) {
      return context.functions[name];
    });

    const fragmentShader = `
    #ifdef GL_FRAGMENT_PRECISION_HIGH
    precision highp float;
    #else
    precision mediump float;
    #endif

    varying vec2 v_textureCoord;
    varying vec2 v_mapCoord;
    uniform vec4 ${Uniforms.RENDER_EXTENT};
    uniform float ${Uniforms.TRANSITION_ALPHA};
    uniform float ${Uniforms.TEXTURE_PIXEL_WIDTH};
    uniform float ${Uniforms.TEXTURE_PIXEL_HEIGHT};
    uniform float ${Uniforms.RESOLUTION};
    uniform float ${Uniforms.ZOOM};

    ${uniformDeclarations.join('\n')}

    ${functionDefintions.join('\n')}

    void main() {
      if (
        v_mapCoord[0] < ${Uniforms.RENDER_EXTENT}[0] ||
        v_mapCoord[1] < ${Uniforms.RENDER_EXTENT}[1] ||
        v_mapCoord[0] > ${Uniforms.RENDER_EXTENT}[2] ||
        v_mapCoord[1] > ${Uniforms.RENDER_EXTENT}[3]
      ) {
        discard;
      }

      vec4 color = texture2D(${
        Uniforms.TILE_TEXTURE_ARRAY
      }[0],  v_textureCoord);

      ${pipeline.join('\n')}

      if (color.a == 0.0) {
        discard;
      }

      gl_FragColor = color;
      gl_FragColor.rgb *= gl_FragColor.a;
      gl_FragColor *= ${Uniforms.TRANSITION_ALPHA};
    }`;

    return {
      vertexShader: vertexShader,
      fragmentShader: fragmentShader,
      uniforms: uniforms,
      paletteTextures: context.paletteTextures,
    };
  }

  /**
   * @classdesc
   * For layer sources that provide pre-rendered, tiled images in grids that are
   * organized by zoom levels for specific resolutions.
   * Note that any property set in the options is set as a {@link module:ol/Object~BaseObject}
   * property on the layer object; for example, setting `title: 'My Title'` in the
   * options means that `title` is observable, and has get/set accessors.
   *
   * @extends BaseTileLayer<SourceType, WebGLTileLayerRenderer>
   * @fires import("../render/Event.js").RenderEvent
   * @api
   */
  class WebGLTileLayer extends BaseTileLayer {
    /**
     * @param {Options} options Tile layer options.
     */
    constructor(options) {
      options = options ? Object.assign({}, options) : {};

      const style = options.style || {};
      delete options.style;

      const cacheSize = options.cacheSize;
      delete options.cacheSize;

      super(options);

      /**
       * @type {Array<SourceType>|function(import("../extent.js").Extent, number):Array<SourceType>}
       * @private
       */
      this.sources_ = options.sources;

      /**
       * @type {SourceType|null}
       * @private
       */
      this.renderedSource_ = null;

      /**
       * @type {number}
       * @private
       */
      this.renderedResolution_ = NaN;

      /**
       * @type {Style}
       * @private
       */
      this.style_ = style;

      /**
       * @type {number}
       * @private
       */
      this.cacheSize_ = cacheSize;

      /**
       * @type {Object<string, (string|number)>}
       * @private
       */
      this.styleVariables_ = this.style_.variables || {};

      this.addChangeListener(LayerProperty.SOURCE, this.handleSourceUpdate_);
    }

    /**
     * Gets the sources for this layer, for a given extent and resolution.
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {number} resolution Resolution.
     * @return {Array<SourceType>} Sources.
     */
    getSources(extent, resolution) {
      const source = this.getSource();
      return this.sources_
        ? typeof this.sources_ === 'function'
          ? this.sources_(extent, resolution)
          : this.sources_
        : source
        ? [source]
        : [];
    }

    /**
     * @return {SourceType} The source being rendered.
     */
    getRenderSource() {
      return this.renderedSource_ || this.getSource();
    }

    /**
     * @return {import("../source/Source.js").State} Source state.
     */
    getSourceState() {
      const source = this.getRenderSource();
      return source ? source.getState() : 'undefined';
    }

    /**
     * @private
     */
    handleSourceUpdate_() {
      if (this.hasRenderer()) {
        this.getRenderer().clearCache();
      }
      if (this.getSource()) {
        this.setStyle(this.style_);
      }
    }

    /**
     * @private
     * @return {number} The number of source bands.
     */
    getSourceBandCount_() {
      const max = Number.MAX_SAFE_INTEGER;
      const sources = this.getSources([-max, -max, max, max], max);
      return sources && sources.length && 'bandCount' in sources[0]
        ? sources[0].bandCount
        : 4;
    }

    createRenderer() {
      const parsedStyle = parseStyle(this.style_, this.getSourceBandCount_());

      return new WebGLTileLayerRenderer(this, {
        vertexShader: parsedStyle.vertexShader,
        fragmentShader: parsedStyle.fragmentShader,
        uniforms: parsedStyle.uniforms,
        cacheSize: this.cacheSize_,
        paletteTextures: parsedStyle.paletteTextures,
      });
    }

    /**
     * @param {import("../Map").FrameState} frameState Frame state.
     * @param {Array<SourceType>} sources Sources.
     * @return {HTMLElement} Canvas.
     */
    renderSources(frameState, sources) {
      const layerRenderer = this.getRenderer();
      let canvas;
      for (let i = 0, ii = sources.length; i < ii; ++i) {
        this.renderedSource_ = sources[i];
        if (layerRenderer.prepareFrame(frameState)) {
          canvas = layerRenderer.renderFrame(frameState);
        }
      }
      return canvas;
    }

    /**
     * @param {?import("../Map.js").FrameState} frameState Frame state.
     * @param {HTMLElement} target Target which the renderer may (but need not) use
     * for rendering its content.
     * @return {HTMLElement} The rendered element.
     */
    render(frameState, target) {
      this.rendered = true;
      const viewState = frameState.viewState;
      const sources = this.getSources(frameState.extent, viewState.resolution);
      let ready = true;
      for (let i = 0, ii = sources.length; i < ii; ++i) {
        const source = sources[i];
        const sourceState = source.getState();
        if (sourceState == 'loading') {
          const onChange = () => {
            if (source.getState() == 'ready') {
              source.removeEventListener('change', onChange);
              this.changed();
            }
          };
          source.addEventListener('change', onChange);
        }
        ready = ready && sourceState == 'ready';
      }
      const canvas = this.renderSources(frameState, sources);
      if (this.getRenderer().renderComplete && ready) {
        // Fully rendered, done.
        this.renderedResolution_ = viewState.resolution;
        return canvas;
      }
      // Render sources from previously fully rendered frames
      if (this.renderedResolution_ > 0.5 * viewState.resolution) {
        const altSources = this.getSources(
          frameState.extent,
          this.renderedResolution_
        ).filter((source) => !sources.includes(source));
        if (altSources.length > 0) {
          return this.renderSources(frameState, altSources);
        }
      }
      return canvas;
    }

    /**
     * Update the layer style.  The `updateStyleVariables` function is a more efficient
     * way to update layer rendering.  In cases where the whole style needs to be updated,
     * this method may be called instead.  Note that calling this method will also replace
     * any previously set variables, so the new style also needs to include new variables,
     * if needed.
     * @param {Style} style The new style.
     */
    setStyle(style) {
      this.styleVariables_ = style.variables || {};
      this.style_ = style;
      const parsedStyle = parseStyle(this.style_, this.getSourceBandCount_());
      const renderer = this.getRenderer();
      renderer.reset({
        vertexShader: parsedStyle.vertexShader,
        fragmentShader: parsedStyle.fragmentShader,
        uniforms: parsedStyle.uniforms,
        paletteTextures: parsedStyle.paletteTextures,
      });
      this.changed();
    }

    /**
     * Update any variables used by the layer style and trigger a re-render.
     * @param {Object<string, number>} variables Variables to update.
     * @api
     */
    updateStyleVariables(variables) {
      Object.assign(this.styleVariables_, variables);
      this.changed();
    }
  }

  /**
   * Clean up underlying WebGL resources.
   * @function
   * @api
   */
  WebGLTileLayer.prototype.dispose;

  /**
   * @module ol/interaction/DragAndDrop
   */

  /**
   * @typedef {Object} Options
   * @property {Array<typeof import("../format/Feature.js").default|import("../format/Feature.js").default>} [formatConstructors] Format constructors
   * (and/or formats pre-constructed with options).
   * @property {import("../source/Vector.js").default} [source] Optional vector source where features will be added.  If a source is provided
   * all existing features will be removed and new features will be added when
   * they are dropped on the target.  If you want to add features to a vector
   * source without removing the existing features (append only), instead of
   * providing the source option listen for the "addfeatures" event.
   * @property {import("../proj.js").ProjectionLike} [projection] Target projection. By default, the map's view's projection is used.
   * @property {HTMLElement} [target] The element that is used as the drop target, default is the viewport element.
   */

  /**
   * @enum {string}
   */
  const DragAndDropEventType = {
    /**
     * Triggered when features are added
     * @event DragAndDropEvent#addfeatures
     * @api
     */
    ADD_FEATURES: 'addfeatures',
  };

  /**
   * @classdesc
   * Events emitted by {@link module:ol/interaction/DragAndDrop~DragAndDrop} instances are instances
   * of this type.
   */
  class DragAndDropEvent extends BaseEvent {
    /**
     * @param {DragAndDropEventType} type Type.
     * @param {File} file File.
     * @param {Array<import("../Feature.js").default>} [features] Features.
     * @param {import("../proj/Projection.js").default} [projection] Projection.
     */
    constructor(type, file, features, projection) {
      super(type);

      /**
       * The features parsed from dropped data.
       * @type {Array<import("../Feature.js").FeatureLike>|undefined}
       * @api
       */
      this.features = features;

      /**
       * The dropped file.
       * @type {File}
       * @api
       */
      this.file = file;

      /**
       * The feature projection.
       * @type {import("../proj/Projection.js").default|undefined}
       * @api
       */
      this.projection = projection;
    }
  }

  /***
   * @template Return
   * @typedef {import("../Observable").OnSignature<import("../Observable").EventTypes, import("../events/Event.js").default, Return> &
   *   import("../Observable").OnSignature<import("../ObjectEventType").Types|
   *     'change:active', import("../Object").ObjectEvent, Return> &
   *   import("../Observable").OnSignature<'addfeatures', DragAndDropEvent, Return> &
   *   import("../Observable").CombinedOnSignature<import("../Observable").EventTypes|import("../ObjectEventType").Types|
   *     'change:active'|'addfeatures', Return>} DragAndDropOnSignature
   */

  /**
   * @classdesc
   * Handles input of vector data by drag and drop.
   *
   * @api
   *
   * @fires DragAndDropEvent
   */
  class DragAndDrop extends Interaction {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      options = options ? options : {};

      super({
        handleEvent: TRUE,
      });

      /***
       * @type {DragAndDropOnSignature<import("../events").EventsKey>}
       */
      this.on;

      /***
       * @type {DragAndDropOnSignature<import("../events").EventsKey>}
       */
      this.once;

      /***
       * @type {DragAndDropOnSignature<void>}
       */
      this.un;

      /**
       * @private
       * @type {boolean}
       */
      this.readAsBuffer_ = false;

      /**
       * @private
       * @type {Array<import("../format/Feature.js").default>}
       */
      this.formats_ = [];
      const formatConstructors = options.formatConstructors
        ? options.formatConstructors
        : [];
      for (let i = 0, ii = formatConstructors.length; i < ii; ++i) {
        let format = formatConstructors[i];
        if (typeof format === 'function') {
          format = new format();
        }
        this.formats_.push(format);
        this.readAsBuffer_ =
          this.readAsBuffer_ || format.getType() === 'arraybuffer';
      }

      /**
       * @private
       * @type {import("../proj/Projection.js").default}
       */
      this.projection_ = options.projection
        ? get$2(options.projection)
        : null;

      /**
       * @private
       * @type {?Array<import("../events.js").EventsKey>}
       */
      this.dropListenKeys_ = null;

      /**
       * @private
       * @type {import("../source/Vector.js").default}
       */
      this.source_ = options.source || null;

      /**
       * @private
       * @type {HTMLElement|null}
       */
      this.target = options.target ? options.target : null;
    }

    /**
     * @param {File} file File.
     * @param {Event} event Load event.
     * @private
     */
    handleResult_(file, event) {
      const result = event.target.result;
      const map = this.getMap();
      let projection = this.projection_;
      if (!projection) {
        projection = getUserProjection();
        if (!projection) {
          const view = map.getView();
          projection = view.getProjection();
        }
      }

      let text;
      const formats = this.formats_;
      for (let i = 0, ii = formats.length; i < ii; ++i) {
        const format = formats[i];
        let input = result;
        if (this.readAsBuffer_ && format.getType() !== 'arraybuffer') {
          if (text === undefined) {
            text = new TextDecoder().decode(result);
          }
          input = text;
        }
        const features = this.tryReadFeatures_(format, input, {
          featureProjection: projection,
        });
        if (features && features.length > 0) {
          if (this.source_) {
            this.source_.clear();
            this.source_.addFeatures(features);
          }
          this.dispatchEvent(
            new DragAndDropEvent(
              DragAndDropEventType.ADD_FEATURES,
              file,
              features,
              projection
            )
          );
          break;
        }
      }
    }

    /**
     * @private
     */
    registerListeners_() {
      const map = this.getMap();
      if (map) {
        const dropArea = this.target ? this.target : map.getViewport();
        this.dropListenKeys_ = [
          listen(dropArea, EventType$1.DROP, this.handleDrop, this),
          listen(dropArea, EventType$1.DRAGENTER, this.handleStop, this),
          listen(dropArea, EventType$1.DRAGOVER, this.handleStop, this),
          listen(dropArea, EventType$1.DROP, this.handleStop, this),
        ];
      }
    }

    /**
     * Activate or deactivate the interaction.
     * @param {boolean} active Active.
     * @observable
     * @api
     */
    setActive(active) {
      if (!this.getActive() && active) {
        this.registerListeners_();
      }
      if (this.getActive() && !active) {
        this.unregisterListeners_();
      }
      super.setActive(active);
    }

    /**
     * Remove the interaction from its current map and attach it to the new map.
     * Subclasses may set up event handlers to get notified about changes to
     * the map here.
     * @param {import("../Map.js").default} map Map.
     */
    setMap(map) {
      this.unregisterListeners_();
      super.setMap(map);
      if (this.getActive()) {
        this.registerListeners_();
      }
    }

    /**
     * @param {import("../format/Feature.js").default} format Format.
     * @param {string} text Text.
     * @param {import("../format/Feature.js").ReadOptions} options Read options.
     * @private
     * @return {Array<import("../Feature.js").default>} Features.
     */
    tryReadFeatures_(format, text, options) {
      try {
        return (
          /** @type {Array<import("../Feature.js").default>} */
          (format.readFeatures(text, options))
        );
      } catch (e) {
        return null;
      }
    }

    /**
     * @private
     */
    unregisterListeners_() {
      if (this.dropListenKeys_) {
        this.dropListenKeys_.forEach(unlistenByKey);
        this.dropListenKeys_ = null;
      }
    }

    /**
     * @param {DragEvent} event Event.
     */
    handleDrop(event) {
      const files = event.dataTransfer.files;
      for (let i = 0, ii = files.length; i < ii; ++i) {
        const file = files.item(i);
        const reader = new FileReader();
        reader.addEventListener(
          EventType$1.LOAD,
          this.handleResult_.bind(this, file)
        );
        if (this.readAsBuffer_) {
          reader.readAsArrayBuffer(file);
        } else {
          reader.readAsText(file);
        }
      }
    }

    /**
     * @param {DragEvent} event Event.
     */
    handleStop(event) {
      event.stopPropagation();
      event.preventDefault();
      event.dataTransfer.dropEffect = 'copy';
    }
  }

  /**
   * @module ol/interaction/DragRotateAndZoom
   */

  /**
   * @typedef {Object} Options
   * @property {import("../events/condition.js").Condition} [condition] A function that
   * takes an {@link module:ol/MapBrowserEvent~MapBrowserEvent} and returns a
   * boolean to indicate whether that event should be handled.
   * Default is {@link module:ol/events/condition.shiftKeyOnly}.
   * @property {number} [duration=400] Animation duration in milliseconds.
   */

  /**
   * @classdesc
   * Allows the user to zoom and rotate the map by clicking and dragging
   * on the map.  By default, this interaction is limited to when the shift
   * key is held down.
   *
   * This interaction is only supported for mouse devices.
   *
   * And this interaction is not included in the default interactions.
   * @api
   */
  class DragRotateAndZoom extends PointerInteraction {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      options = options ? options : {};

      super(/** @type {import("./Pointer.js").Options} */ (options));

      /**
       * @private
       * @type {import("../events/condition.js").Condition}
       */
      this.condition_ = options.condition ? options.condition : shiftKeyOnly;

      /**
       * @private
       * @type {number|undefined}
       */
      this.lastAngle_ = undefined;

      /**
       * @private
       * @type {number|undefined}
       */
      this.lastMagnitude_ = undefined;

      /**
       * @private
       * @type {number}
       */
      this.lastScaleDelta_ = 0;

      /**
       * @private
       * @type {number}
       */
      this.duration_ = options.duration !== undefined ? options.duration : 400;
    }

    /**
     * Handle pointer drag events.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Event.
     */
    handleDragEvent(mapBrowserEvent) {
      if (!mouseOnly(mapBrowserEvent)) {
        return;
      }

      const map = mapBrowserEvent.map;
      const size = map.getSize();
      const offset = mapBrowserEvent.pixel;
      const deltaX = offset[0] - size[0] / 2;
      const deltaY = size[1] / 2 - offset[1];
      const theta = Math.atan2(deltaY, deltaX);
      const magnitude = Math.sqrt(deltaX * deltaX + deltaY * deltaY);
      const view = map.getView();
      if (this.lastAngle_ !== undefined) {
        const angleDelta = this.lastAngle_ - theta;
        view.adjustRotationInternal(angleDelta);
      }
      this.lastAngle_ = theta;
      if (this.lastMagnitude_ !== undefined) {
        view.adjustResolutionInternal(this.lastMagnitude_ / magnitude);
      }
      if (this.lastMagnitude_ !== undefined) {
        this.lastScaleDelta_ = this.lastMagnitude_ / magnitude;
      }
      this.lastMagnitude_ = magnitude;
    }

    /**
     * Handle pointer up events.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Event.
     * @return {boolean} If the event was consumed.
     */
    handleUpEvent(mapBrowserEvent) {
      if (!mouseOnly(mapBrowserEvent)) {
        return true;
      }

      const map = mapBrowserEvent.map;
      const view = map.getView();
      const direction = this.lastScaleDelta_ > 1 ? 1 : -1;
      view.endInteraction(this.duration_, direction);
      this.lastScaleDelta_ = 0;
      return false;
    }

    /**
     * Handle pointer down events.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Event.
     * @return {boolean} If the event was consumed.
     */
    handleDownEvent(mapBrowserEvent) {
      if (!mouseOnly(mapBrowserEvent)) {
        return false;
      }

      if (this.condition_(mapBrowserEvent)) {
        mapBrowserEvent.map.getView().beginInteraction();
        this.lastAngle_ = undefined;
        this.lastMagnitude_ = undefined;
        return true;
      } else {
        return false;
      }
    }
  }

  /**
   * @module ol/interaction/Draw
   */

  /**
   * @typedef {Object} Options
   * @property {import("../geom/Geometry.js").Type} type Geometry type of
   * the geometries being drawn with this instance.
   * @property {number} [clickTolerance=6] The maximum distance in pixels between
   * "down" and "up" for a "up" event to be considered a "click" event and
   * actually add a point/vertex to the geometry being drawn.  The default of `6`
   * was chosen for the draw interaction to behave correctly on mouse as well as
   * on touch devices.
   * @property {import("../Collection.js").default<Feature>} [features]
   * Destination collection for the drawn features.
   * @property {VectorSource} [source] Destination source for
   * the drawn features.
   * @property {number} [dragVertexDelay=500] Delay in milliseconds after pointerdown
   * before the current vertex can be dragged to its exact position.
   * @property {number} [snapTolerance=12] Pixel distance for snapping to the
   * drawing finish. Must be greater than `0`.
   * @property {boolean} [stopClick=false] Stop click, singleclick, and
   * doubleclick events from firing during drawing.
   * @property {number} [maxPoints] The number of points that can be drawn before
   * a polygon ring or line string is finished. By default there is no
   * restriction.
   * @property {number} [minPoints] The number of points that must be drawn
   * before a polygon ring or line string can be finished. Default is `3` for
   * polygon rings and `2` for line strings.
   * @property {import("../events/condition.js").Condition} [finishCondition] A function
   * that takes an {@link module:ol/MapBrowserEvent~MapBrowserEvent} and returns a
   * boolean to indicate whether the drawing can be finished. Not used when drawing
   * POINT or MULTI_POINT geometries.
   * @property {import("../style/Style.js").StyleLike|import("../style/flat.js").FlatStyleLike} [style]
   * Style for sketch features.
   * @property {GeometryFunction} [geometryFunction]
   * Function that is called when a geometry's coordinates are updated.
   * @property {string} [geometryName] Geometry name to use for features created
   * by the draw interaction.
   * @property {import("../events/condition.js").Condition} [condition] A function that
   * takes an {@link module:ol/MapBrowserEvent~MapBrowserEvent} and returns a
   * boolean to indicate whether that event should be handled.
   * By default {@link module:ol/events/condition.noModifierKeys}, i.e. a click,
   * adds a vertex or deactivates freehand drawing.
   * @property {boolean} [freehand=false] Operate in freehand mode for lines,
   * polygons, and circles.  This makes the interaction always operate in freehand
   * mode and takes precedence over any `freehandCondition` option.
   * @property {import("../events/condition.js").Condition} [freehandCondition]
   * Condition that activates freehand drawing for lines and polygons. This
   * function takes an {@link module:ol/MapBrowserEvent~MapBrowserEvent} and
   * returns a boolean to indicate whether that event should be handled. The
   * default is {@link module:ol/events/condition.shiftKeyOnly}, meaning that the
   * Shift key activates freehand drawing.
   * @property {boolean|import("../events/condition.js").Condition} [trace=false] Trace a portion of another geometry.
   * Ignored when in freehand mode.
   * @property {VectorSource} [traceSource] Source for features to trace.  If tracing is active and a `traceSource` is
   * not provided, the interaction's `source` will be used.  Tracing requires that the interaction is configured with
   * either a `traceSource` or a `source`.
   * @property {boolean} [wrapX=false] Wrap the world horizontally on the sketch
   * overlay.
   * @property {import("../geom/Geometry.js").GeometryLayout} [geometryLayout='XY'] Layout of the
   * feature geometries created by the draw interaction.
   */

  /**
   * Coordinate type when drawing points.
   * @typedef {import("../coordinate.js").Coordinate} PointCoordType
   */

  /**
   * Coordinate type when drawing lines.
   * @typedef {Array<import("../coordinate.js").Coordinate>} LineCoordType
   */

  /**
   * Coordinate type when drawing polygons.
   * @typedef {Array<Array<import("../coordinate.js").Coordinate>>} PolyCoordType
   */

  /**
   * Types used for drawing coordinates.
   * @typedef {PointCoordType|LineCoordType|PolyCoordType} SketchCoordType
   */

  /**
   * @typedef {Object} TraceState
   * @property {boolean} active Tracing active.
   * @property {import("../pixel.js").Pixel} [startPx] The initially clicked pixel location.
   * @property {Array<TraceTarget>} [targets] Targets available for tracing.
   * @property {number} [targetIndex] The index of the currently traced target.  A value of -1 indicates
   * that no trace target is active.
   */

  /**
   * @typedef {Object} TraceTarget
   * @property {Array<import("../coordinate.js").Coordinate>} coordinates Target coordinates.
   * @property {boolean} ring The target coordinates are a linear ring.
   * @property {number} startIndex The index of first traced coordinate.  A fractional index represents an
   * edge intersection.  Index values for rings will wrap (may be negative or larger than coordinates length).
   * @property {number} endIndex The index of last traced coordinate.  Details from startIndex also apply here.
   */

  /**
   * Function that takes an array of coordinates and an optional existing geometry
   * and a projection as arguments, and returns a geometry. The optional existing
   * geometry is the geometry that is returned when the function is called without
   * a second argument.
   * @typedef {function(!SketchCoordType, import("../geom/SimpleGeometry.js").default,
   *     import("../proj/Projection.js").default):
   *     import("../geom/SimpleGeometry.js").default} GeometryFunction
   */

  /**
   * @typedef {'Point' | 'LineString' | 'Polygon' | 'Circle'} Mode
   * Draw mode.  This collapses multi-part geometry types with their single-part
   * cousins.
   */

  /**
   * @enum {string}
   */
  const DrawEventType = {
    /**
     * Triggered upon feature draw start
     * @event DrawEvent#drawstart
     * @api
     */
    DRAWSTART: 'drawstart',
    /**
     * Triggered upon feature draw end
     * @event DrawEvent#drawend
     * @api
     */
    DRAWEND: 'drawend',
    /**
     * Triggered upon feature draw abortion
     * @event DrawEvent#drawabort
     * @api
     */
    DRAWABORT: 'drawabort',
  };

  /**
   * @classdesc
   * Events emitted by {@link module:ol/interaction/Draw~Draw} instances are
   * instances of this type.
   */
  class DrawEvent extends BaseEvent {
    /**
     * @param {DrawEventType} type Type.
     * @param {Feature} feature The feature drawn.
     */
    constructor(type, feature) {
      super(type);

      /**
       * The feature being drawn.
       * @type {Feature}
       * @api
       */
      this.feature = feature;
    }
  }

  /**
   * @param {import("../coordinate.js").Coordinate} coordinate The coordinate.
   * @param {Array<Feature>} features The candidate features.
   * @return {Array<TraceTarget>} The trace targets.
   */
  function getTraceTargets(coordinate, features) {
    /**
     * @type {Array<TraceTarget>}
     */
    const targets = [];

    for (let i = 0; i < features.length; ++i) {
      const feature = features[i];
      const geometry = feature.getGeometry();
      appendGeometryTraceTargets(coordinate, geometry, targets);
    }

    return targets;
  }

  /**
   * @param {import("../coordinate.js").Coordinate} a One coordinate.
   * @param {import("../coordinate.js").Coordinate} b Another coordinate.
   * @return {number} The squared distance between the two coordinates.
   */
  function getSquaredDistance(a, b) {
    return squaredDistance$1(a[0], a[1], b[0], b[1]);
  }

  /**
   * @param {LineCoordType} coordinates The ring coordinates.
   * @param {number} index The index.  May be wrapped.
   * @return {import("../coordinate.js").Coordinate} The coordinate.
   */
  function getCoordinate(coordinates, index) {
    const count = coordinates.length;
    if (index < 0) {
      return coordinates[index + count];
    }
    if (index >= count) {
      return coordinates[index - count];
    }
    return coordinates[index];
  }

  /**
   * Get the cumulative squared distance along a ring path.  The end index index may be "wrapped" and it may
   * be less than the start index to indicate the direction of travel.  The start and end index may have
   * a fractional part to indicate a point between two coordinates.
   * @param {LineCoordType} coordinates Ring coordinates.
   * @param {number} startIndex The start index.
   * @param {number} endIndex The end index.
   * @return {number} The cumulative squared distance along the ring path.
   */
  function getCumulativeSquaredDistance(coordinates, startIndex, endIndex) {
    let lowIndex, highIndex;
    if (startIndex < endIndex) {
      lowIndex = startIndex;
      highIndex = endIndex;
    } else {
      lowIndex = endIndex;
      highIndex = startIndex;
    }
    const lowWholeIndex = Math.ceil(lowIndex);
    const highWholeIndex = Math.floor(highIndex);

    if (lowWholeIndex > highWholeIndex) {
      // both start and end are on the same segment
      const start = interpolateCoordinate(coordinates, lowIndex);
      const end = interpolateCoordinate(coordinates, highIndex);
      return getSquaredDistance(start, end);
    }

    let sd = 0;

    if (lowIndex < lowWholeIndex) {
      const start = interpolateCoordinate(coordinates, lowIndex);
      const end = getCoordinate(coordinates, lowWholeIndex);
      sd += getSquaredDistance(start, end);
    }

    if (highWholeIndex < highIndex) {
      const start = getCoordinate(coordinates, highWholeIndex);
      const end = interpolateCoordinate(coordinates, highIndex);
      sd += getSquaredDistance(start, end);
    }

    for (let i = lowWholeIndex; i < highWholeIndex - 1; ++i) {
      const start = getCoordinate(coordinates, i);
      const end = getCoordinate(coordinates, i + 1);
      sd += getSquaredDistance(start, end);
    }

    return sd;
  }

  /**
   * @param {import("../coordinate.js").Coordinate} coordinate The coordinate.
   * @param {import("../geom/Geometry.js").default} geometry The candidate geometry.
   * @param {Array<TraceTarget>} targets The trace targets.
   */
  function appendGeometryTraceTargets(coordinate, geometry, targets) {
    if (geometry instanceof LineString) {
      appendTraceTarget(coordinate, geometry.getCoordinates(), false, targets);
      return;
    }
    if (geometry instanceof MultiLineString) {
      const coordinates = geometry.getCoordinates();
      for (let i = 0, ii = coordinates.length; i < ii; ++i) {
        appendTraceTarget(coordinate, coordinates[i], false, targets);
      }
      return;
    }
    if (geometry instanceof Polygon) {
      const coordinates = geometry.getCoordinates();
      for (let i = 0, ii = coordinates.length; i < ii; ++i) {
        appendTraceTarget(coordinate, coordinates[i], true, targets);
      }
      return;
    }
    if (geometry instanceof MultiPolygon) {
      const polys = geometry.getCoordinates();
      for (let i = 0, ii = polys.length; i < ii; ++i) {
        const coordinates = polys[i];
        for (let j = 0, jj = coordinates.length; j < jj; ++j) {
          appendTraceTarget(coordinate, coordinates[j], true, targets);
        }
      }
      return;
    }
    if (geometry instanceof GeometryCollection) {
      const geometries = geometry.getGeometries();
      for (let i = 0; i < geometries.length; ++i) {
        appendGeometryTraceTargets(coordinate, geometries[i], targets);
      }
      return;
    }
    // other types cannot be traced
  }

  /**
   * @typedef {Object} TraceTargetUpdateInfo
   * @property {number} index The new target index.
   * @property {number} endIndex The new segment end index.
   */

  /**
   * @type {TraceTargetUpdateInfo}
   */
  const sharedUpdateInfo = {index: -1, endIndex: NaN};

  /**
   * @param {import("../coordinate.js").Coordinate} coordinate The coordinate.
   * @param {TraceState} traceState The trace state.
   * @param {import("../Map.js").default} map The map.
   * @param {number} snapTolerance The snap tolerance.
   * @return {TraceTargetUpdateInfo} Information about the new trace target.  The returned
   * object is reused between calls and must not be modified by the caller.
   */
  function getTraceTargetUpdate(coordinate, traceState, map, snapTolerance) {
    const x = coordinate[0];
    const y = coordinate[1];

    let closestTargetDistance = Infinity;

    let newTargetIndex = -1;
    let newEndIndex = NaN;

    for (
      let targetIndex = 0;
      targetIndex < traceState.targets.length;
      ++targetIndex
    ) {
      const target = traceState.targets[targetIndex];
      const coordinates = target.coordinates;

      let minSegmentDistance = Infinity;
      let endIndex;
      for (
        let coordinateIndex = 0;
        coordinateIndex < coordinates.length - 1;
        ++coordinateIndex
      ) {
        const start = coordinates[coordinateIndex];
        const end = coordinates[coordinateIndex + 1];
        const rel = getPointSegmentRelationship(x, y, start, end);
        if (rel.squaredDistance < minSegmentDistance) {
          minSegmentDistance = rel.squaredDistance;
          endIndex = coordinateIndex + rel.along;
        }
      }

      if (minSegmentDistance < closestTargetDistance) {
        closestTargetDistance = minSegmentDistance;
        if (target.ring && traceState.targetIndex === targetIndex) {
          // same target, maintain the same trace direction
          if (target.endIndex > target.startIndex) {
            // forward trace
            if (endIndex < target.startIndex) {
              endIndex += coordinates.length;
            }
          } else if (target.endIndex < target.startIndex) {
            // reverse trace
            if (endIndex > target.startIndex) {
              endIndex -= coordinates.length;
            }
          }
        }
        newEndIndex = endIndex;
        newTargetIndex = targetIndex;
      }
    }

    const newTarget = traceState.targets[newTargetIndex];
    let considerBothDirections = newTarget.ring;
    if (traceState.targetIndex === newTargetIndex && considerBothDirections) {
      // only consider switching trace direction if close to the start
      const newCoordinate = interpolateCoordinate(
        newTarget.coordinates,
        newEndIndex
      );
      const pixel = map.getPixelFromCoordinate(newCoordinate);
      if (distance(pixel, traceState.startPx) > snapTolerance) {
        considerBothDirections = false;
      }
    }

    if (considerBothDirections) {
      const coordinates = newTarget.coordinates;
      const count = coordinates.length;
      const startIndex = newTarget.startIndex;
      const endIndex = newEndIndex;
      if (startIndex < endIndex) {
        const forwardDistance = getCumulativeSquaredDistance(
          coordinates,
          startIndex,
          endIndex
        );
        const reverseDistance = getCumulativeSquaredDistance(
          coordinates,
          startIndex,
          endIndex - count
        );
        if (reverseDistance < forwardDistance) {
          newEndIndex -= count;
        }
      } else {
        const reverseDistance = getCumulativeSquaredDistance(
          coordinates,
          startIndex,
          endIndex
        );
        const forwardDistance = getCumulativeSquaredDistance(
          coordinates,
          startIndex,
          endIndex + count
        );
        if (forwardDistance < reverseDistance) {
          newEndIndex += count;
        }
      }
    }

    sharedUpdateInfo.index = newTargetIndex;
    sharedUpdateInfo.endIndex = newEndIndex;
    return sharedUpdateInfo;
  }

  /**
   * @param {import("../coordinate.js").Coordinate} coordinate The clicked coordinate.
   * @param {Array<import("../coordinate.js").Coordinate>} coordinates The geometry component coordinates.
   * @param {boolean} ring The coordinates represent a linear ring.
   * @param {Array<TraceTarget>} targets The trace targets.
   */
  function appendTraceTarget(coordinate, coordinates, ring, targets) {
    const x = coordinate[0];
    const y = coordinate[1];
    for (let i = 0, ii = coordinates.length - 1; i < ii; ++i) {
      const start = coordinates[i];
      const end = coordinates[i + 1];
      const rel = getPointSegmentRelationship(x, y, start, end);
      if (rel.squaredDistance === 0) {
        const index = i + rel.along;
        targets.push({
          coordinates: coordinates,
          ring: ring,
          startIndex: index,
          endIndex: index,
        });
        return;
      }
    }
  }

  /**
   * @typedef {Object} PointSegmentRelationship
   * @property {number} along The closest point expressed as a fraction along the segment length.
   * @property {number} squaredDistance The squared distance of the point to the segment.
   */

  /**
   * @type {PointSegmentRelationship}
   */
  const sharedRel = {along: 0, squaredDistance: 0};

  /**
   * @param {number} x The point x.
   * @param {number} y The point y.
   * @param {import("../coordinate.js").Coordinate} start The segment start.
   * @param {import("../coordinate.js").Coordinate} end The segment end.
   * @return {PointSegmentRelationship} The point segment relationship.  The returned object is
   * shared between calls and must not be modified by the caller.
   */
  function getPointSegmentRelationship(x, y, start, end) {
    const x1 = start[0];
    const y1 = start[1];
    const x2 = end[0];
    const y2 = end[1];
    const dx = x2 - x1;
    const dy = y2 - y1;
    let along = 0;
    let px = x1;
    let py = y1;
    if (dx !== 0 || dy !== 0) {
      along = clamp(((x - x1) * dx + (y - y1) * dy) / (dx * dx + dy * dy), 0, 1);
      px += dx * along;
      py += dy * along;
    }

    sharedRel.along = along;
    sharedRel.squaredDistance = toFixed(squaredDistance$1(x, y, px, py), 10);
    return sharedRel;
  }

  /**
   * @param {LineCoordType} coordinates The coordinates.
   * @param {number} index The index.  May be fractional and may wrap.
   * @return {import("../coordinate.js").Coordinate} The interpolated coordinate.
   */
  function interpolateCoordinate(coordinates, index) {
    const count = coordinates.length;

    let startIndex = Math.floor(index);
    const along = index - startIndex;
    if (startIndex >= count) {
      startIndex -= count;
    } else if (startIndex < 0) {
      startIndex += count;
    }

    let endIndex = startIndex + 1;
    if (endIndex >= count) {
      endIndex -= count;
    }

    const start = coordinates[startIndex];
    const x0 = start[0];
    const y0 = start[1];
    const end = coordinates[endIndex];
    const dx = end[0] - x0;
    const dy = end[1] - y0;

    return [x0 + dx * along, y0 + dy * along];
  }

  /***
   * @template Return
   * @typedef {import("../Observable").OnSignature<import("../Observable").EventTypes, import("../events/Event.js").default, Return> &
   *   import("../Observable").OnSignature<import("../ObjectEventType").Types|
   *     'change:active', import("../Object").ObjectEvent, Return> &
   *   import("../Observable").OnSignature<'drawabort'|'drawend'|'drawstart', DrawEvent, Return> &
   *   import("../Observable").CombinedOnSignature<import("../Observable").EventTypes|import("../ObjectEventType").Types|
   *     'change:active'|'drawabort'|'drawend'|'drawstart', Return>} DrawOnSignature
   */

  /**
   * @classdesc
   * Interaction for drawing feature geometries.
   *
   * @fires DrawEvent
   * @api
   */
  class Draw extends PointerInteraction {
    /**
     * @param {Options} options Options.
     */
    constructor(options) {
      const pointerOptions = /** @type {import("./Pointer.js").Options} */ (
        options
      );
      if (!pointerOptions.stopDown) {
        pointerOptions.stopDown = FALSE;
      }

      super(pointerOptions);

      /***
       * @type {DrawOnSignature<import("../events").EventsKey>}
       */
      this.on;

      /***
       * @type {DrawOnSignature<import("../events").EventsKey>}
       */
      this.once;

      /***
       * @type {DrawOnSignature<void>}
       */
      this.un;

      /**
       * @type {boolean}
       * @private
       */
      this.shouldHandle_ = false;

      /**
       * @type {import("../pixel.js").Pixel}
       * @private
       */
      this.downPx_ = null;

      /**
       * @type {?}
       * @private
       */
      this.downTimeout_;

      /**
       * @type {number|undefined}
       * @private
       */
      this.lastDragTime_;

      /**
       * Pointer type of the last pointermove event
       * @type {string}
       * @private
       */
      this.pointerType_;

      /**
       * @type {boolean}
       * @private
       */
      this.freehand_ = false;

      /**
       * Target source for drawn features.
       * @type {VectorSource|null}
       * @private
       */
      this.source_ = options.source ? options.source : null;

      /**
       * Target collection for drawn features.
       * @type {import("../Collection.js").default<Feature>|null}
       * @private
       */
      this.features_ = options.features ? options.features : null;

      /**
       * Pixel distance for snapping.
       * @type {number}
       * @private
       */
      this.snapTolerance_ = options.snapTolerance ? options.snapTolerance : 12;

      /**
       * Geometry type.
       * @type {import("../geom/Geometry.js").Type}
       * @private
       */
      this.type_ = /** @type {import("../geom/Geometry.js").Type} */ (
        options.type
      );

      /**
       * Drawing mode (derived from geometry type.
       * @type {Mode}
       * @private
       */
      this.mode_ = getMode(this.type_);

      /**
       * Stop click, singleclick, and doubleclick events from firing during drawing.
       * Default is `false`.
       * @type {boolean}
       * @private
       */
      this.stopClick_ = !!options.stopClick;

      /**
       * The number of points that must be drawn before a polygon ring or line
       * string can be finished.  The default is 3 for polygon rings and 2 for
       * line strings.
       * @type {number}
       * @private
       */
      this.minPoints_ = options.minPoints
        ? options.minPoints
        : this.mode_ === 'Polygon'
        ? 3
        : 2;

      /**
       * The number of points that can be drawn before a polygon ring or line string
       * is finished. The default is no restriction.
       * @type {number}
       * @private
       */
      this.maxPoints_ =
        this.mode_ === 'Circle'
          ? 2
          : options.maxPoints
          ? options.maxPoints
          : Infinity;

      /**
       * A function to decide if a potential finish coordinate is permissible
       * @private
       * @type {import("../events/condition.js").Condition}
       */
      this.finishCondition_ = options.finishCondition
        ? options.finishCondition
        : TRUE;

      /**
       * @private
       * @type {import("../geom/Geometry.js").GeometryLayout}
       */
      this.geometryLayout_ = options.geometryLayout
        ? options.geometryLayout
        : 'XY';

      let geometryFunction = options.geometryFunction;
      if (!geometryFunction) {
        const mode = this.mode_;
        if (mode === 'Circle') {
          /**
           * @param {!LineCoordType} coordinates The coordinates.
           * @param {import("../geom/SimpleGeometry.js").default|undefined} geometry Optional geometry.
           * @param {import("../proj/Projection.js").default} projection The view projection.
           * @return {import("../geom/SimpleGeometry.js").default} A geometry.
           */
          geometryFunction = function (coordinates, geometry, projection) {
            const circle = geometry
              ? /** @type {Circle} */ (geometry)
              : new Circle([NaN, NaN]);
            const center = fromUserCoordinate(coordinates[0], projection);
            const squaredLength = squaredDistance(
              center,
              fromUserCoordinate(coordinates[coordinates.length - 1], projection)
            );
            circle.setCenterAndRadius(
              center,
              Math.sqrt(squaredLength),
              this.geometryLayout_
            );
            const userProjection = getUserProjection();
            if (userProjection) {
              circle.transform(projection, userProjection);
            }
            return circle;
          };
        } else {
          let Constructor;
          if (mode === 'Point') {
            Constructor = Point;
          } else if (mode === 'LineString') {
            Constructor = LineString;
          } else if (mode === 'Polygon') {
            Constructor = Polygon;
          }
          /**
           * @param {!LineCoordType} coordinates The coordinates.
           * @param {import("../geom/SimpleGeometry.js").default|undefined} geometry Optional geometry.
           * @param {import("../proj/Projection.js").default} projection The view projection.
           * @return {import("../geom/SimpleGeometry.js").default} A geometry.
           */
          geometryFunction = function (coordinates, geometry, projection) {
            if (geometry) {
              if (mode === 'Polygon') {
                if (coordinates[0].length) {
                  // Add a closing coordinate to match the first
                  geometry.setCoordinates(
                    [coordinates[0].concat([coordinates[0][0]])],
                    this.geometryLayout_
                  );
                } else {
                  geometry.setCoordinates([], this.geometryLayout_);
                }
              } else {
                geometry.setCoordinates(coordinates, this.geometryLayout_);
              }
            } else {
              geometry = new Constructor(coordinates, this.geometryLayout_);
            }
            return geometry;
          };
        }
      }

      /**
       * @type {GeometryFunction}
       * @private
       */
      this.geometryFunction_ = geometryFunction;

      /**
       * @type {number}
       * @private
       */
      this.dragVertexDelay_ =
        options.dragVertexDelay !== undefined ? options.dragVertexDelay : 500;

      /**
       * Finish coordinate for the feature (first point for polygons, last point for
       * linestrings).
       * @type {import("../coordinate.js").Coordinate}
       * @private
       */
      this.finishCoordinate_ = null;

      /**
       * Sketch feature.
       * @type {Feature<import('../geom/SimpleGeometry.js').default>}
       * @private
       */
      this.sketchFeature_ = null;

      /**
       * Sketch point.
       * @type {Feature<Point>}
       * @private
       */
      this.sketchPoint_ = null;

      /**
       * Sketch coordinates. Used when drawing a line or polygon.
       * @type {SketchCoordType}
       * @private
       */
      this.sketchCoords_ = null;

      /**
       * Sketch line. Used when drawing polygon.
       * @type {Feature<LineString>}
       * @private
       */
      this.sketchLine_ = null;

      /**
       * Sketch line coordinates. Used when drawing a polygon or circle.
       * @type {LineCoordType}
       * @private
       */
      this.sketchLineCoords_ = null;

      /**
       * Squared tolerance for handling up events.  If the squared distance
       * between a down and up event is greater than this tolerance, up events
       * will not be handled.
       * @type {number}
       * @private
       */
      this.squaredClickTolerance_ = options.clickTolerance
        ? options.clickTolerance * options.clickTolerance
        : 36;

      /**
       * Draw overlay where our sketch features are drawn.
       * @type {VectorLayer}
       * @private
       */
      this.overlay_ = new VectorLayer({
        source: new VectorSource({
          useSpatialIndex: false,
          wrapX: options.wrapX ? options.wrapX : false,
        }),
        style: options.style ? options.style : getDefaultStyleFunction$2(),
        updateWhileInteracting: true,
      });

      /**
       * Name of the geometry attribute for newly created features.
       * @type {string|undefined}
       * @private
       */
      this.geometryName_ = options.geometryName;

      /**
       * @private
       * @type {import("../events/condition.js").Condition}
       */
      this.condition_ = options.condition ? options.condition : noModifierKeys;

      /**
       * @private
       * @type {import("../events/condition.js").Condition}
       */
      this.freehandCondition_;
      if (options.freehand) {
        this.freehandCondition_ = always;
      } else {
        this.freehandCondition_ = options.freehandCondition
          ? options.freehandCondition
          : shiftKeyOnly;
      }

      /**
       * @type {import("../events/condition.js").Condition}
       * @private
       */
      this.traceCondition_;
      this.setTrace(options.trace || false);

      /**
       * @type {TraceState}
       * @private
       */
      this.traceState_ = {active: false};

      /**
       * @type {VectorSource|null}
       * @private
       */
      this.traceSource_ = options.traceSource || options.source || null;

      this.addChangeListener(InteractionProperty.ACTIVE, this.updateState_);
    }

    /**
     * Toggle tracing mode or set a tracing condition.
     *
     * @param {boolean|import("../events/condition.js").Condition} trace A boolean to toggle tracing mode or an event
     *     condition that will be checked when a feature is clicked to determine if tracing should be active.
     */
    setTrace(trace) {
      let condition;
      if (!trace) {
        condition = never;
      } else if (trace === true) {
        condition = always;
      } else {
        condition = trace;
      }
      this.traceCondition_ = condition;
    }

    /**
     * Remove the interaction from its current map and attach it to the new map.
     * Subclasses may set up event handlers to get notified about changes to
     * the map here.
     * @param {import("../Map.js").default} map Map.
     */
    setMap(map) {
      super.setMap(map);
      this.updateState_();
    }

    /**
     * Get the overlay layer that this interaction renders sketch features to.
     * @return {VectorLayer} Overlay layer.
     * @api
     */
    getOverlay() {
      return this.overlay_;
    }

    /**
     * Handles the {@link module:ol/MapBrowserEvent~MapBrowserEvent map browser event} and may actually draw or finish the drawing.
     * @param {import("../MapBrowserEvent.js").default} event Map browser event.
     * @return {boolean} `false` to stop event propagation.
     * @api
     */
    handleEvent(event) {
      if (event.originalEvent.type === EventType$1.CONTEXTMENU) {
        // Avoid context menu for long taps when drawing on mobile
        event.originalEvent.preventDefault();
      }
      this.freehand_ = this.mode_ !== 'Point' && this.freehandCondition_(event);
      let move = event.type === MapBrowserEventType.POINTERMOVE;
      let pass = true;
      if (
        !this.freehand_ &&
        this.lastDragTime_ &&
        event.type === MapBrowserEventType.POINTERDRAG
      ) {
        const now = Date.now();
        if (now - this.lastDragTime_ >= this.dragVertexDelay_) {
          this.downPx_ = event.pixel;
          this.shouldHandle_ = !this.freehand_;
          move = true;
        } else {
          this.lastDragTime_ = undefined;
        }
        if (this.shouldHandle_ && this.downTimeout_ !== undefined) {
          clearTimeout(this.downTimeout_);
          this.downTimeout_ = undefined;
        }
      }
      if (
        this.freehand_ &&
        event.type === MapBrowserEventType.POINTERDRAG &&
        this.sketchFeature_ !== null
      ) {
        this.addToDrawing_(event.coordinate);
        pass = false;
      } else if (
        this.freehand_ &&
        event.type === MapBrowserEventType.POINTERDOWN
      ) {
        pass = false;
      } else if (move && this.getPointerCount() < 2) {
        pass = event.type === MapBrowserEventType.POINTERMOVE;
        if (pass && this.freehand_) {
          this.handlePointerMove_(event);
          if (this.shouldHandle_) {
            // Avoid page scrolling when freehand drawing on mobile
            event.originalEvent.preventDefault();
          }
        } else if (
          event.originalEvent.pointerType === 'mouse' ||
          (event.type === MapBrowserEventType.POINTERDRAG &&
            this.downTimeout_ === undefined)
        ) {
          this.handlePointerMove_(event);
        }
      } else if (event.type === MapBrowserEventType.DBLCLICK) {
        pass = false;
      }

      return super.handleEvent(event) && pass;
    }

    /**
     * Handle pointer down events.
     * @param {import("../MapBrowserEvent.js").default} event Event.
     * @return {boolean} If the event was consumed.
     */
    handleDownEvent(event) {
      this.shouldHandle_ = !this.freehand_;

      if (this.freehand_) {
        this.downPx_ = event.pixel;
        if (!this.finishCoordinate_) {
          this.startDrawing_(event.coordinate);
        }
        return true;
      }

      if (!this.condition_(event)) {
        this.lastDragTime_ = undefined;
        return false;
      }

      this.lastDragTime_ = Date.now();
      this.downTimeout_ = setTimeout(
        function () {
          this.handlePointerMove_(
            new MapBrowserEvent(
              MapBrowserEventType.POINTERMOVE,
              event.map,
              event.originalEvent,
              false,
              event.frameState
            )
          );
        }.bind(this),
        this.dragVertexDelay_
      );
      this.downPx_ = event.pixel;
      return true;
    }

    /**
     * @private
     */
    deactivateTrace_() {
      this.traceState_ = {active: false};
    }

    /**
     * Activate or deactivate trace state based on a browser event.
     * @param {import("../MapBrowserEvent.js").default} event Event.
     * @private
     */
    toggleTraceState_(event) {
      if (!this.traceSource_ || !this.traceCondition_(event)) {
        return;
      }

      if (this.traceState_.active) {
        this.deactivateTrace_();
        return;
      }

      const map = this.getMap();
      const lowerLeft = map.getCoordinateFromPixel([
        event.pixel[0] - this.snapTolerance_,
        event.pixel[1] + this.snapTolerance_,
      ]);
      const upperRight = map.getCoordinateFromPixel([
        event.pixel[0] + this.snapTolerance_,
        event.pixel[1] - this.snapTolerance_,
      ]);
      const extent = boundingExtent([lowerLeft, upperRight]);
      const features = this.traceSource_.getFeaturesInExtent(extent);
      if (features.length === 0) {
        return;
      }

      const targets = getTraceTargets(event.coordinate, features);
      if (targets.length) {
        this.traceState_ = {
          active: true,
          startPx: event.pixel.slice(),
          targets: targets,
          targetIndex: -1,
        };
      }
    }

    /**
     * @param {TraceTarget} target The trace target.
     * @param {number} endIndex The new end index of the trace.
     * @private
     */
    addOrRemoveTracedCoordinates_(target, endIndex) {
      // three cases to handle:
      //  1. traced in the same direction and points need adding
      //  2. traced in the same direction and points need removing
      //  3. traced in a new direction
      const previouslyForward = target.startIndex <= target.endIndex;
      const currentlyForward = target.startIndex <= endIndex;
      if (previouslyForward === currentlyForward) {
        // same direction
        if (
          (previouslyForward && endIndex > target.endIndex) ||
          (!previouslyForward && endIndex < target.endIndex)
        ) {
          // case 1 - add new points
          this.addTracedCoordinates_(target, target.endIndex, endIndex);
        } else if (
          (previouslyForward && endIndex < target.endIndex) ||
          (!previouslyForward && endIndex > target.endIndex)
        ) {
          // case 2 - remove old points
          this.removeTracedCoordinates_(endIndex, target.endIndex);
        }
      } else {
        // case 3 - remove old points, add new points
        this.removeTracedCoordinates_(target.startIndex, target.endIndex);
        this.addTracedCoordinates_(target, target.startIndex, endIndex);
      }
    }

    /**
     * @param {number} fromIndex The start index.
     * @param {number} toIndex The end index.
     * @private
     */
    removeTracedCoordinates_(fromIndex, toIndex) {
      if (fromIndex === toIndex) {
        return;
      }

      let remove = 0;
      if (fromIndex < toIndex) {
        const start = Math.ceil(fromIndex);
        let end = Math.floor(toIndex);
        if (end === toIndex) {
          end -= 1;
        }
        remove = end - start + 1;
      } else {
        const start = Math.floor(fromIndex);
        let end = Math.ceil(toIndex);
        if (end === toIndex) {
          end += 1;
        }
        remove = start - end + 1;
      }

      if (remove > 0) {
        this.removeLastPoints_(remove);
      }
    }

    /**
     * @param {TraceTarget} target The trace target.
     * @param {number} fromIndex The start index.
     * @param {number} toIndex The end index.
     * @private
     */
    addTracedCoordinates_(target, fromIndex, toIndex) {
      if (fromIndex === toIndex) {
        return;
      }

      const coordinates = [];
      if (fromIndex < toIndex) {
        // forward trace
        const start = Math.ceil(fromIndex);
        let end = Math.floor(toIndex);
        if (end === toIndex) {
          // if end is snapped to a vertex, it will be added later
          end -= 1;
        }
        for (let i = start; i <= end; ++i) {
          coordinates.push(getCoordinate(target.coordinates, i));
        }
      } else {
        // reverse trace
        const start = Math.floor(fromIndex);
        let end = Math.ceil(toIndex);
        if (end === toIndex) {
          end += 1;
        }
        for (let i = start; i >= end; --i) {
          coordinates.push(getCoordinate(target.coordinates, i));
        }
      }
      if (coordinates.length) {
        this.appendCoordinates(coordinates);
      }
    }

    /**
     * Update the trace.
     * @param {import("../MapBrowserEvent.js").default} event Event.
     * @private
     */
    updateTrace_(event) {
      const traceState = this.traceState_;
      if (!traceState.active) {
        return;
      }

      if (traceState.targetIndex === -1) {
        // check if we are ready to pick a target
        if (distance(traceState.startPx, event.pixel) < this.snapTolerance_) {
          return;
        }
      }

      const updatedTraceTarget = getTraceTargetUpdate(
        event.coordinate,
        traceState,
        this.getMap(),
        this.snapTolerance_
      );

      if (traceState.targetIndex !== updatedTraceTarget.index) {
        // target changed
        if (traceState.targetIndex !== -1) {
          // remove points added during previous trace
          const oldTarget = traceState.targets[traceState.targetIndex];
          this.removeTracedCoordinates_(oldTarget.startIndex, oldTarget.endIndex);
        }
        // add points for the new target
        const newTarget = traceState.targets[updatedTraceTarget.index];
        this.addTracedCoordinates_(
          newTarget,
          newTarget.startIndex,
          updatedTraceTarget.endIndex
        );
      } else {
        // target stayed the same
        const target = traceState.targets[traceState.targetIndex];
        this.addOrRemoveTracedCoordinates_(target, updatedTraceTarget.endIndex);
      }

      // modify the state with updated info
      traceState.targetIndex = updatedTraceTarget.index;
      const target = traceState.targets[traceState.targetIndex];
      target.endIndex = updatedTraceTarget.endIndex;

      // update event coordinate and pixel to match end point of final segment
      const coordinate = interpolateCoordinate(
        target.coordinates,
        target.endIndex
      );
      const pixel = this.getMap().getPixelFromCoordinate(coordinate);
      event.coordinate = coordinate;
      event.pixel = [Math.round(pixel[0]), Math.round(pixel[1])];
    }

    /**
     * Handle pointer up events.
     * @param {import("../MapBrowserEvent.js").default} event Event.
     * @return {boolean} If the event was consumed.
     */
    handleUpEvent(event) {
      let pass = true;

      if (this.getPointerCount() === 0) {
        if (this.downTimeout_) {
          clearTimeout(this.downTimeout_);
          this.downTimeout_ = undefined;
        }

        this.handlePointerMove_(event);
        const tracing = this.traceState_.active;
        this.toggleTraceState_(event);

        if (this.shouldHandle_) {
          const startingToDraw = !this.finishCoordinate_;
          if (startingToDraw) {
            this.startDrawing_(event.coordinate);
          }
          if (!startingToDraw && this.freehand_) {
            this.finishDrawing();
          } else if (
            !this.freehand_ &&
            (!startingToDraw || this.mode_ === 'Point')
          ) {
            if (this.atFinish_(event.pixel, tracing)) {
              if (this.finishCondition_(event)) {
                this.finishDrawing();
              }
            } else {
              this.addToDrawing_(event.coordinate);
            }
          }
          pass = false;
        } else if (this.freehand_) {
          this.abortDrawing();
        }
      }

      if (!pass && this.stopClick_) {
        event.preventDefault();
      }
      return pass;
    }

    /**
     * Handle move events.
     * @param {import("../MapBrowserEvent.js").default} event A move event.
     * @private
     */
    handlePointerMove_(event) {
      this.pointerType_ = event.originalEvent.pointerType;
      if (
        this.downPx_ &&
        ((!this.freehand_ && this.shouldHandle_) ||
          (this.freehand_ && !this.shouldHandle_))
      ) {
        const downPx = this.downPx_;
        const clickPx = event.pixel;
        const dx = downPx[0] - clickPx[0];
        const dy = downPx[1] - clickPx[1];
        const squaredDistance = dx * dx + dy * dy;
        this.shouldHandle_ = this.freehand_
          ? squaredDistance > this.squaredClickTolerance_
          : squaredDistance <= this.squaredClickTolerance_;
        if (!this.shouldHandle_) {
          return;
        }
      }

      if (!this.finishCoordinate_) {
        this.createOrUpdateSketchPoint_(event.coordinate.slice());
        return;
      }

      this.updateTrace_(event);
      this.modifyDrawing_(event.coordinate);
    }

    /**
     * Determine if an event is within the snapping tolerance of the start coord.
     * @param {import("../pixel.js").Pixel} pixel Pixel.
     * @param {boolean} [tracing] Drawing in trace mode (only stop if at the starting point).
     * @return {boolean} The event is within the snapping tolerance of the start.
     * @private
     */
    atFinish_(pixel, tracing) {
      let at = false;
      if (this.sketchFeature_) {
        let potentiallyDone = false;
        let potentiallyFinishCoordinates = [this.finishCoordinate_];
        const mode = this.mode_;
        if (mode === 'Point') {
          at = true;
        } else if (mode === 'Circle') {
          at = this.sketchCoords_.length === 2;
        } else if (mode === 'LineString') {
          potentiallyDone =
            !tracing && this.sketchCoords_.length > this.minPoints_;
        } else if (mode === 'Polygon') {
          const sketchCoords = /** @type {PolyCoordType} */ (this.sketchCoords_);
          potentiallyDone = sketchCoords[0].length > this.minPoints_;
          potentiallyFinishCoordinates = [
            sketchCoords[0][0],
            sketchCoords[0][sketchCoords[0].length - 2],
          ];
          if (tracing) {
            potentiallyFinishCoordinates = [sketchCoords[0][0]];
          } else {
            potentiallyFinishCoordinates = [
              sketchCoords[0][0],
              sketchCoords[0][sketchCoords[0].length - 2],
            ];
          }
        }
        if (potentiallyDone) {
          const map = this.getMap();
          for (let i = 0, ii = potentiallyFinishCoordinates.length; i < ii; i++) {
            const finishCoordinate = potentiallyFinishCoordinates[i];
            const finishPixel = map.getPixelFromCoordinate(finishCoordinate);
            const dx = pixel[0] - finishPixel[0];
            const dy = pixel[1] - finishPixel[1];
            const snapTolerance = this.freehand_ ? 1 : this.snapTolerance_;
            at = Math.sqrt(dx * dx + dy * dy) <= snapTolerance;
            if (at) {
              this.finishCoordinate_ = finishCoordinate;
              break;
            }
          }
        }
      }
      return at;
    }

    /**
     * @param {import("../coordinate").Coordinate} coordinates Coordinate.
     * @private
     */
    createOrUpdateSketchPoint_(coordinates) {
      if (!this.sketchPoint_) {
        this.sketchPoint_ = new Feature(new Point(coordinates));
        this.updateSketchFeatures_();
      } else {
        const sketchPointGeom = this.sketchPoint_.getGeometry();
        sketchPointGeom.setCoordinates(coordinates);
      }
    }

    /**
     * @param {import("../geom/Polygon.js").default} geometry Polygon geometry.
     * @private
     */
    createOrUpdateCustomSketchLine_(geometry) {
      if (!this.sketchLine_) {
        this.sketchLine_ = new Feature();
      }
      const ring = geometry.getLinearRing(0);
      let sketchLineGeom = this.sketchLine_.getGeometry();
      if (!sketchLineGeom) {
        sketchLineGeom = new LineString(
          ring.getFlatCoordinates(),
          ring.getLayout()
        );
        this.sketchLine_.setGeometry(sketchLineGeom);
      } else {
        sketchLineGeom.setFlatCoordinates(
          ring.getLayout(),
          ring.getFlatCoordinates()
        );
        sketchLineGeom.changed();
      }
    }

    /**
     * Start the drawing.
     * @param {import("../coordinate.js").Coordinate} start Start coordinate.
     * @private
     */
    startDrawing_(start) {
      const projection = this.getMap().getView().getProjection();
      const stride = getStrideForLayout(this.geometryLayout_);
      while (start.length < stride) {
        start.push(0);
      }
      this.finishCoordinate_ = start;
      if (this.mode_ === 'Point') {
        this.sketchCoords_ = start.slice();
      } else if (this.mode_ === 'Polygon') {
        this.sketchCoords_ = [[start.slice(), start.slice()]];
        this.sketchLineCoords_ = this.sketchCoords_[0];
      } else {
        this.sketchCoords_ = [start.slice(), start.slice()];
      }
      if (this.sketchLineCoords_) {
        this.sketchLine_ = new Feature(new LineString(this.sketchLineCoords_));
      }
      const geometry = this.geometryFunction_(
        this.sketchCoords_,
        undefined,
        projection
      );
      this.sketchFeature_ = new Feature();
      if (this.geometryName_) {
        this.sketchFeature_.setGeometryName(this.geometryName_);
      }
      this.sketchFeature_.setGeometry(geometry);
      this.updateSketchFeatures_();
      this.dispatchEvent(
        new DrawEvent(DrawEventType.DRAWSTART, this.sketchFeature_)
      );
    }

    /**
     * Modify the drawing.
     * @param {import("../coordinate.js").Coordinate} coordinate Coordinate.
     * @private
     */
    modifyDrawing_(coordinate) {
      const map = this.getMap();
      const geometry = this.sketchFeature_.getGeometry();
      const projection = map.getView().getProjection();
      const stride = getStrideForLayout(this.geometryLayout_);
      let coordinates, last;
      while (coordinate.length < stride) {
        coordinate.push(0);
      }
      if (this.mode_ === 'Point') {
        last = this.sketchCoords_;
      } else if (this.mode_ === 'Polygon') {
        coordinates = /** @type {PolyCoordType} */ (this.sketchCoords_)[0];
        last = coordinates[coordinates.length - 1];
        if (this.atFinish_(map.getPixelFromCoordinate(coordinate))) {
          // snap to finish
          coordinate = this.finishCoordinate_.slice();
        }
      } else {
        coordinates = this.sketchCoords_;
        last = coordinates[coordinates.length - 1];
      }
      last[0] = coordinate[0];
      last[1] = coordinate[1];
      this.geometryFunction_(
        /** @type {!LineCoordType} */ (this.sketchCoords_),
        geometry,
        projection
      );
      if (this.sketchPoint_) {
        const sketchPointGeom = this.sketchPoint_.getGeometry();
        sketchPointGeom.setCoordinates(coordinate);
      }
      if (geometry.getType() === 'Polygon' && this.mode_ !== 'Polygon') {
        this.createOrUpdateCustomSketchLine_(/** @type {Polygon} */ (geometry));
      } else if (this.sketchLineCoords_) {
        const sketchLineGeom = this.sketchLine_.getGeometry();
        sketchLineGeom.setCoordinates(this.sketchLineCoords_);
      }
      this.updateSketchFeatures_();
    }

    /**
     * Add a new coordinate to the drawing.
     * @param {!PointCoordType} coordinate Coordinate
     * @private
     */
    addToDrawing_(coordinate) {
      const geometry = this.sketchFeature_.getGeometry();
      const projection = this.getMap().getView().getProjection();
      let done;
      let coordinates;
      const mode = this.mode_;
      if (mode === 'LineString' || mode === 'Circle') {
        this.finishCoordinate_ = coordinate.slice();
        coordinates = /** @type {LineCoordType} */ (this.sketchCoords_);
        if (coordinates.length >= this.maxPoints_) {
          if (this.freehand_) {
            coordinates.pop();
          } else {
            done = true;
          }
        }
        coordinates.push(coordinate.slice());
        this.geometryFunction_(coordinates, geometry, projection);
      } else if (mode === 'Polygon') {
        coordinates = /** @type {PolyCoordType} */ (this.sketchCoords_)[0];
        if (coordinates.length >= this.maxPoints_) {
          if (this.freehand_) {
            coordinates.pop();
          } else {
            done = true;
          }
        }
        coordinates.push(coordinate.slice());
        if (done) {
          this.finishCoordinate_ = coordinates[0];
        }
        this.geometryFunction_(this.sketchCoords_, geometry, projection);
      }
      this.createOrUpdateSketchPoint_(coordinate.slice());
      this.updateSketchFeatures_();
      if (done) {
        this.finishDrawing();
      }
    }

    /**
     * @param {number} n The number of points to remove.
     */
    removeLastPoints_(n) {
      if (!this.sketchFeature_) {
        return;
      }
      const geometry = this.sketchFeature_.getGeometry();
      const projection = this.getMap().getView().getProjection();
      const mode = this.mode_;
      for (let i = 0; i < n; ++i) {
        let coordinates;
        if (mode === 'LineString' || mode === 'Circle') {
          coordinates = /** @type {LineCoordType} */ (this.sketchCoords_);
          coordinates.splice(-2, 1);
          if (coordinates.length >= 2) {
            this.finishCoordinate_ = coordinates[coordinates.length - 2].slice();
            const finishCoordinate = this.finishCoordinate_.slice();
            coordinates[coordinates.length - 1] = finishCoordinate;
            this.createOrUpdateSketchPoint_(finishCoordinate);
          }
          this.geometryFunction_(coordinates, geometry, projection);
          if (geometry.getType() === 'Polygon' && this.sketchLine_) {
            this.createOrUpdateCustomSketchLine_(
              /** @type {Polygon} */ (geometry)
            );
          }
        } else if (mode === 'Polygon') {
          coordinates = /** @type {PolyCoordType} */ (this.sketchCoords_)[0];
          coordinates.splice(-2, 1);
          const sketchLineGeom = this.sketchLine_.getGeometry();
          if (coordinates.length >= 2) {
            const finishCoordinate = coordinates[coordinates.length - 2].slice();
            coordinates[coordinates.length - 1] = finishCoordinate;
            this.createOrUpdateSketchPoint_(finishCoordinate);
          }
          sketchLineGeom.setCoordinates(coordinates);
          this.geometryFunction_(this.sketchCoords_, geometry, projection);
        }

        if (coordinates.length === 1) {
          this.abortDrawing();
          break;
        }
      }

      this.updateSketchFeatures_();
    }

    /**
     * Remove last point of the feature currently being drawn. Does not do anything when
     * drawing POINT or MULTI_POINT geometries.
     * @api
     */
    removeLastPoint() {
      this.removeLastPoints_(1);
    }

    /**
     * Stop drawing and add the sketch feature to the target layer.
     * The {@link module:ol/interaction/Draw~DrawEventType.DRAWEND} event is
     * dispatched before inserting the feature.
     * @api
     */
    finishDrawing() {
      const sketchFeature = this.abortDrawing_();
      if (!sketchFeature) {
        return;
      }
      let coordinates = this.sketchCoords_;
      const geometry = sketchFeature.getGeometry();
      const projection = this.getMap().getView().getProjection();
      if (this.mode_ === 'LineString') {
        // remove the redundant last point
        coordinates.pop();
        this.geometryFunction_(coordinates, geometry, projection);
      } else if (this.mode_ === 'Polygon') {
        // remove the redundant last point in ring
        /** @type {PolyCoordType} */ (coordinates)[0].pop();
        this.geometryFunction_(coordinates, geometry, projection);
        coordinates = geometry.getCoordinates();
      }

      // cast multi-part geometries
      if (this.type_ === 'MultiPoint') {
        sketchFeature.setGeometry(
          new MultiPoint([/** @type {PointCoordType} */ (coordinates)])
        );
      } else if (this.type_ === 'MultiLineString') {
        sketchFeature.setGeometry(
          new MultiLineString([/** @type {LineCoordType} */ (coordinates)])
        );
      } else if (this.type_ === 'MultiPolygon') {
        sketchFeature.setGeometry(
          new MultiPolygon([/** @type {PolyCoordType} */ (coordinates)])
        );
      }

      // First dispatch event to allow full set up of feature
      this.dispatchEvent(new DrawEvent(DrawEventType.DRAWEND, sketchFeature));

      // Then insert feature
      if (this.features_) {
        this.features_.push(sketchFeature);
      }
      if (this.source_) {
        this.source_.addFeature(sketchFeature);
      }
    }

    /**
     * Stop drawing without adding the sketch feature to the target layer.
     * @return {Feature<import("../geom/SimpleGeometry.js").default>|null} The sketch feature (or null if none).
     * @private
     */
    abortDrawing_() {
      this.finishCoordinate_ = null;
      const sketchFeature = this.sketchFeature_;
      this.sketchFeature_ = null;
      this.sketchPoint_ = null;
      this.sketchLine_ = null;
      this.overlay_.getSource().clear(true);
      this.deactivateTrace_();
      return sketchFeature;
    }

    /**
     * Stop drawing without adding the sketch feature to the target layer.
     * @api
     */
    abortDrawing() {
      const sketchFeature = this.abortDrawing_();
      if (sketchFeature) {
        this.dispatchEvent(new DrawEvent(DrawEventType.DRAWABORT, sketchFeature));
      }
    }

    /**
     * Append coordinates to the end of the geometry that is currently being drawn.
     * This can be used when drawing LineStrings or Polygons. Coordinates will
     * either be appended to the current LineString or the outer ring of the current
     * Polygon. If no geometry is being drawn, a new one will be created.
     * @param {!LineCoordType} coordinates Linear coordinates to be appended to
     * the coordinate array.
     * @api
     */
    appendCoordinates(coordinates) {
      const mode = this.mode_;
      const newDrawing = !this.sketchFeature_;
      if (newDrawing) {
        this.startDrawing_(coordinates[0]);
      }
      /** @type {LineCoordType} */
      let sketchCoords;
      if (mode === 'LineString' || mode === 'Circle') {
        sketchCoords = /** @type {LineCoordType} */ (this.sketchCoords_);
      } else if (mode === 'Polygon') {
        sketchCoords =
          this.sketchCoords_ && this.sketchCoords_.length
            ? /** @type {PolyCoordType} */ (this.sketchCoords_)[0]
            : [];
      } else {
        return;
      }

      if (newDrawing) {
        sketchCoords.shift();
      }

      // Remove last coordinate from sketch drawing (this coordinate follows cursor position)
      sketchCoords.pop();

      // Append coordinate list
      for (let i = 0; i < coordinates.length; i++) {
        this.addToDrawing_(coordinates[i]);
      }

      const ending = coordinates[coordinates.length - 1];
      // Duplicate last coordinate for sketch drawing (cursor position)
      this.addToDrawing_(ending);
      this.modifyDrawing_(ending);
    }

    /**
     * Initiate draw mode by starting from an existing geometry which will
     * receive new additional points. This only works on features with
     * `LineString` geometries, where the interaction will extend lines by adding
     * points to the end of the coordinates array.
     * This will change the original feature, instead of drawing a copy.
     *
     * The function will dispatch a `drawstart` event.
     *
     * @param {!Feature<LineString>} feature Feature to be extended.
     * @api
     */
    extend(feature) {
      const geometry = feature.getGeometry();
      const lineString = geometry;
      this.sketchFeature_ = feature;
      this.sketchCoords_ = lineString.getCoordinates();
      const last = this.sketchCoords_[this.sketchCoords_.length - 1];
      this.finishCoordinate_ = last.slice();
      this.sketchCoords_.push(last.slice());
      this.sketchPoint_ = new Feature(new Point(last));
      this.updateSketchFeatures_();
      this.dispatchEvent(
        new DrawEvent(DrawEventType.DRAWSTART, this.sketchFeature_)
      );
    }

    /**
     * Redraw the sketch features.
     * @private
     */
    updateSketchFeatures_() {
      const sketchFeatures = [];
      if (this.sketchFeature_) {
        sketchFeatures.push(this.sketchFeature_);
      }
      if (this.sketchLine_) {
        sketchFeatures.push(this.sketchLine_);
      }
      if (this.sketchPoint_) {
        sketchFeatures.push(this.sketchPoint_);
      }
      const overlaySource = this.overlay_.getSource();
      overlaySource.clear(true);
      overlaySource.addFeatures(sketchFeatures);
    }

    /**
     * @private
     */
    updateState_() {
      const map = this.getMap();
      const active = this.getActive();
      if (!map || !active) {
        this.abortDrawing();
      }
      this.overlay_.setMap(active ? map : null);
    }
  }

  /**
   * @return {import("../style/Style.js").StyleFunction} Styles.
   */
  function getDefaultStyleFunction$2() {
    const styles = createEditingStyle();
    return function (feature, resolution) {
      return styles[feature.getGeometry().getType()];
    };
  }

  /**
   * Create a `geometryFunction` for `type: 'Circle'` that will create a regular
   * polygon with a user specified number of sides and start angle instead of a
   * {@link import("../geom/Circle.js").Circle} geometry.
   * @param {number} [sides] Number of sides of the regular polygon.
   *     Default is 32.
   * @param {number} [angle] Angle of the first point in counter-clockwise
   *     radians. 0 means East.
   *     Default is the angle defined by the heading from the center of the
   *     regular polygon to the current pointer position.
   * @return {GeometryFunction} Function that draws a polygon.
   * @api
   */
  function createRegularPolygon(sides, angle) {
    return function (coordinates, geometry, projection) {
      const center = fromUserCoordinate(
        /** @type {LineCoordType} */ (coordinates)[0],
        projection
      );
      const end = fromUserCoordinate(
        /** @type {LineCoordType} */ (coordinates)[coordinates.length - 1],
        projection
      );
      const radius = Math.sqrt(squaredDistance(center, end));
      geometry = geometry || fromCircle(new Circle(center), sides);

      let internalAngle = angle;
      if (!angle && angle !== 0) {
        const x = end[0] - center[0];
        const y = end[1] - center[1];
        internalAngle = Math.atan2(y, x);
      }
      makeRegular(
        /** @type {Polygon} */ (geometry),
        center,
        radius,
        internalAngle
      );

      const userProjection = getUserProjection();
      if (userProjection) {
        geometry.transform(projection, userProjection);
      }
      return geometry;
    };
  }

  /**
   * Create a `geometryFunction` that will create a box-shaped polygon (aligned
   * with the coordinate system axes).  Use this with the draw interaction and
   * `type: 'Circle'` to return a box instead of a circle geometry.
   * @return {GeometryFunction} Function that draws a box-shaped polygon.
   * @api
   */
  function createBox() {
    return function (coordinates, geometry, projection) {
      const extent = boundingExtent(
        /** @type {LineCoordType} */ ([
          coordinates[0],
          coordinates[coordinates.length - 1],
        ]).map(function (coordinate) {
          return fromUserCoordinate(coordinate, projection);
        })
      );
      const boxCoordinates = [
        [
          getBottomLeft(extent),
          getBottomRight(extent),
          getTopRight(extent),
          getTopLeft(extent),
          getBottomLeft(extent),
        ],
      ];
      if (geometry) {
        geometry.setCoordinates(boxCoordinates);
      } else {
        geometry = new Polygon(boxCoordinates);
      }
      const userProjection = getUserProjection();
      if (userProjection) {
        geometry.transform(projection, userProjection);
      }
      return geometry;
    };
  }

  /**
   * Get the drawing mode.  The mode for multi-part geometries is the same as for
   * their single-part cousins.
   * @param {import("../geom/Geometry.js").Type} type Geometry type.
   * @return {Mode} Drawing mode.
   */
  function getMode(type) {
    switch (type) {
      case 'Point':
      case 'MultiPoint':
        return 'Point';
      case 'LineString':
      case 'MultiLineString':
        return 'LineString';
      case 'Polygon':
      case 'MultiPolygon':
        return 'Polygon';
      case 'Circle':
        return 'Circle';
      default:
        throw new Error('Invalid type: ' + type);
    }
  }

  /**
   * @module ol/interaction/Extent
   */

  /**
   * @typedef {Object} Options
   * @property {import("../events/condition.js").Condition} [condition] A function that
   * takes an {@link module:ol/MapBrowserEvent~MapBrowserEvent} and returns a
   * boolean to indicate whether that event should be handled.
   * Default is {@link module:ol/events/condition.always}.
   * @property {import("../extent.js").Extent} [extent] Initial extent. Defaults to no
   * initial extent.
   * @property {import("../style/Style.js").StyleLike} [boxStyle]
   * Style for the drawn extent box. Defaults to the `Polygon` editing style
   * documented in {@link module:ol/style/Style~Style}
   * @property {number} [pixelTolerance=10] Pixel tolerance for considering the
   * pointer close enough to a segment or vertex for editing.
   * @property {import("../style/Style.js").StyleLike} [pointerStyle]
   * Style for the cursor used to draw the extent. Defaults to the `Point` editing style
   * documented in {@link module:ol/style/Style~Style}
   * @property {boolean} [wrapX=false] Wrap the drawn extent across multiple maps
   * in the X direction? Only affects visuals, not functionality.
   */

  /**
   * @enum {string}
   */
  const ExtentEventType = {
    /**
     * Triggered after the extent is changed
     * @event ExtentEvent#extentchanged
     * @api
     */
    EXTENTCHANGED: 'extentchanged',
  };

  /**
   * @classdesc
   * Events emitted by {@link module:ol/interaction/Extent~Extent} instances are
   * instances of this type.
   */
  class ExtentEvent extends BaseEvent {
    /**
     * @param {import("../extent.js").Extent} extent the new extent
     */
    constructor(extent) {
      super(ExtentEventType.EXTENTCHANGED);

      /**
       * The current extent.
       * @type {import("../extent.js").Extent}
       * @api
       */
      this.extent = extent;
    }
  }

  /***
   * @template Return
   * @typedef {import("../Observable").OnSignature<import("../Observable").EventTypes, import("../events/Event.js").default, Return> &
   *   import("../Observable").OnSignature<import("../ObjectEventType").Types|
   *     'change:active', import("../Object").ObjectEvent, Return> &
   *   import("../Observable").OnSignature<'extentchanged', ExtentEvent, Return> &
   *   import("../Observable").CombinedOnSignature<import("../Observable").EventTypes|import("../ObjectEventType").Types|
   *     'change:active'|'extentchanged', Return>} ExtentOnSignature
   */

  /**
   * @classdesc
   * Allows the user to draw a vector box by clicking and dragging on the map.
   * Once drawn, the vector box can be modified by dragging its vertices or edges.
   * This interaction is only supported for mouse devices.
   *
   * @fires ExtentEvent
   * @api
   */
  class Extent extends PointerInteraction {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      options = options || {};

      super(/** @type {import("./Pointer.js").Options} */ (options));

      /***
       * @type {ExtentOnSignature<import("../events").EventsKey>}
       */
      this.on;

      /***
       * @type {ExtentOnSignature<import("../events").EventsKey>}
       */
      this.once;

      /***
       * @type {ExtentOnSignature<void>}
       */
      this.un;

      /**
       * Condition
       * @type {import("../events/condition.js").Condition}
       * @private
       */
      this.condition_ = options.condition ? options.condition : always;

      /**
       * Extent of the drawn box
       * @type {import("../extent.js").Extent}
       * @private
       */
      this.extent_ = null;

      /**
       * Handler for pointer move events
       * @type {function (import("../coordinate.js").Coordinate): import("../extent.js").Extent|null}
       * @private
       */
      this.pointerHandler_ = null;

      /**
       * Pixel threshold to snap to extent
       * @type {number}
       * @private
       */
      this.pixelTolerance_ =
        options.pixelTolerance !== undefined ? options.pixelTolerance : 10;

      /**
       * Is the pointer snapped to an extent vertex
       * @type {boolean}
       * @private
       */
      this.snappedToVertex_ = false;

      /**
       * Feature for displaying the visible extent
       * @type {Feature}
       * @private
       */
      this.extentFeature_ = null;

      /**
       * Feature for displaying the visible pointer
       * @type {Feature<Point>}
       * @private
       */
      this.vertexFeature_ = null;

      if (!options) {
        options = {};
      }

      /**
       * Layer for the extentFeature
       * @type {VectorLayer}
       * @private
       */
      this.extentOverlay_ = new VectorLayer({
        source: new VectorSource({
          useSpatialIndex: false,
          wrapX: !!options.wrapX,
        }),
        style: options.boxStyle
          ? options.boxStyle
          : getDefaultExtentStyleFunction(),
        updateWhileAnimating: true,
        updateWhileInteracting: true,
      });

      /**
       * Layer for the vertexFeature
       * @type {VectorLayer}
       * @private
       */
      this.vertexOverlay_ = new VectorLayer({
        source: new VectorSource({
          useSpatialIndex: false,
          wrapX: !!options.wrapX,
        }),
        style: options.pointerStyle
          ? options.pointerStyle
          : getDefaultPointerStyleFunction(),
        updateWhileAnimating: true,
        updateWhileInteracting: true,
      });

      if (options.extent) {
        this.setExtent(options.extent);
      }
    }

    /**
     * @param {import("../pixel.js").Pixel} pixel cursor location
     * @param {import("../Map.js").default} map map
     * @return {import("../coordinate.js").Coordinate|null} snapped vertex on extent
     * @private
     */
    snapToVertex_(pixel, map) {
      const pixelCoordinate = map.getCoordinateFromPixelInternal(pixel);
      const sortByDistance = function (a, b) {
        return (
          squaredDistanceToSegment(pixelCoordinate, a) -
          squaredDistanceToSegment(pixelCoordinate, b)
        );
      };
      const extent = this.getExtentInternal();
      if (extent) {
        //convert extents to line segments and find the segment closest to pixelCoordinate
        const segments = getSegments(extent);
        segments.sort(sortByDistance);
        const closestSegment = segments[0];

        let vertex = closestOnSegment(pixelCoordinate, closestSegment);
        const vertexPixel = map.getPixelFromCoordinateInternal(vertex);

        //if the distance is within tolerance, snap to the segment
        if (distance(pixel, vertexPixel) <= this.pixelTolerance_) {
          //test if we should further snap to a vertex
          const pixel1 = map.getPixelFromCoordinateInternal(closestSegment[0]);
          const pixel2 = map.getPixelFromCoordinateInternal(closestSegment[1]);
          const squaredDist1 = squaredDistance(vertexPixel, pixel1);
          const squaredDist2 = squaredDistance(vertexPixel, pixel2);
          const dist = Math.sqrt(Math.min(squaredDist1, squaredDist2));
          this.snappedToVertex_ = dist <= this.pixelTolerance_;
          if (this.snappedToVertex_) {
            vertex =
              squaredDist1 > squaredDist2 ? closestSegment[1] : closestSegment[0];
          }
          return vertex;
        }
      }
      return null;
    }

    /**
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent pointer move event
     * @private
     */
    handlePointerMove_(mapBrowserEvent) {
      const pixel = mapBrowserEvent.pixel;
      const map = mapBrowserEvent.map;

      let vertex = this.snapToVertex_(pixel, map);
      if (!vertex) {
        vertex = map.getCoordinateFromPixelInternal(pixel);
      }
      this.createOrUpdatePointerFeature_(vertex);
    }

    /**
     * @param {import("../extent.js").Extent} extent extent
     * @return {Feature} extent as featrue
     * @private
     */
    createOrUpdateExtentFeature_(extent) {
      let extentFeature = this.extentFeature_;

      if (!extentFeature) {
        if (!extent) {
          extentFeature = new Feature({});
        } else {
          extentFeature = new Feature(fromExtent(extent));
        }
        this.extentFeature_ = extentFeature;
        this.extentOverlay_.getSource().addFeature(extentFeature);
      } else {
        if (!extent) {
          extentFeature.setGeometry(undefined);
        } else {
          extentFeature.setGeometry(fromExtent(extent));
        }
      }
      return extentFeature;
    }

    /**
     * @param {import("../coordinate.js").Coordinate} vertex location of feature
     * @return {Feature} vertex as feature
     * @private
     */
    createOrUpdatePointerFeature_(vertex) {
      let vertexFeature = this.vertexFeature_;
      if (!vertexFeature) {
        vertexFeature = new Feature(new Point(vertex));
        this.vertexFeature_ = vertexFeature;
        this.vertexOverlay_.getSource().addFeature(vertexFeature);
      } else {
        const geometry = vertexFeature.getGeometry();
        geometry.setCoordinates(vertex);
      }
      return vertexFeature;
    }

    /**
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Map browser event.
     * @return {boolean} `false` to stop event propagation.
     */
    handleEvent(mapBrowserEvent) {
      if (!mapBrowserEvent.originalEvent || !this.condition_(mapBrowserEvent)) {
        return true;
      }
      //display pointer (if not dragging)
      if (
        mapBrowserEvent.type == MapBrowserEventType.POINTERMOVE &&
        !this.handlingDownUpSequence
      ) {
        this.handlePointerMove_(mapBrowserEvent);
      }
      //call pointer to determine up/down/drag
      super.handleEvent(mapBrowserEvent);
      //return false to stop propagation
      return false;
    }

    /**
     * Handle pointer down events.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Event.
     * @return {boolean} If the event was consumed.
     */
    handleDownEvent(mapBrowserEvent) {
      const pixel = mapBrowserEvent.pixel;
      const map = mapBrowserEvent.map;

      const extent = this.getExtentInternal();
      let vertex = this.snapToVertex_(pixel, map);

      //find the extent corner opposite the passed corner
      const getOpposingPoint = function (point) {
        let x_ = null;
        let y_ = null;
        if (point[0] == extent[0]) {
          x_ = extent[2];
        } else if (point[0] == extent[2]) {
          x_ = extent[0];
        }
        if (point[1] == extent[1]) {
          y_ = extent[3];
        } else if (point[1] == extent[3]) {
          y_ = extent[1];
        }
        if (x_ !== null && y_ !== null) {
          return [x_, y_];
        }
        return null;
      };
      if (vertex && extent) {
        const x =
          vertex[0] == extent[0] || vertex[0] == extent[2] ? vertex[0] : null;
        const y =
          vertex[1] == extent[1] || vertex[1] == extent[3] ? vertex[1] : null;

        //snap to point
        if (x !== null && y !== null) {
          this.pointerHandler_ = getPointHandler(getOpposingPoint(vertex));
          //snap to edge
        } else if (x !== null) {
          this.pointerHandler_ = getEdgeHandler(
            getOpposingPoint([x, extent[1]]),
            getOpposingPoint([x, extent[3]])
          );
        } else if (y !== null) {
          this.pointerHandler_ = getEdgeHandler(
            getOpposingPoint([extent[0], y]),
            getOpposingPoint([extent[2], y])
          );
        }
        //no snap - new bbox
      } else {
        vertex = map.getCoordinateFromPixelInternal(pixel);
        this.setExtent([vertex[0], vertex[1], vertex[0], vertex[1]]);
        this.pointerHandler_ = getPointHandler(vertex);
      }
      return true; //event handled; start downup sequence
    }

    /**
     * Handle pointer drag events.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Event.
     */
    handleDragEvent(mapBrowserEvent) {
      if (this.pointerHandler_) {
        const pixelCoordinate = mapBrowserEvent.coordinate;
        this.setExtent(this.pointerHandler_(pixelCoordinate));
        this.createOrUpdatePointerFeature_(pixelCoordinate);
      }
    }

    /**
     * Handle pointer up events.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Event.
     * @return {boolean} If the event was consumed.
     */
    handleUpEvent(mapBrowserEvent) {
      this.pointerHandler_ = null;
      //If bbox is zero area, set to null;
      const extent = this.getExtentInternal();
      if (!extent || getArea$1(extent) === 0) {
        this.setExtent(null);
      }
      return false; //Stop handling downup sequence
    }

    /**
     * Remove the interaction from its current map and attach it to the new map.
     * Subclasses may set up event handlers to get notified about changes to
     * the map here.
     * @param {import("../Map.js").default} map Map.
     */
    setMap(map) {
      this.extentOverlay_.setMap(map);
      this.vertexOverlay_.setMap(map);
      super.setMap(map);
    }

    /**
     * Returns the current drawn extent in the view projection (or user projection if set)
     *
     * @return {import("../extent.js").Extent} Drawn extent in the view projection.
     * @api
     */
    getExtent() {
      return toUserExtent(
        this.getExtentInternal(),
        this.getMap().getView().getProjection()
      );
    }

    /**
     * Returns the current drawn extent in the view projection
     *
     * @return {import("../extent.js").Extent} Drawn extent in the view projection.
     * @api
     */
    getExtentInternal() {
      return this.extent_;
    }

    /**
     * Manually sets the drawn extent, using the view projection.
     *
     * @param {import("../extent.js").Extent} extent Extent
     * @api
     */
    setExtent(extent) {
      //Null extent means no bbox
      this.extent_ = extent ? extent : null;
      this.createOrUpdateExtentFeature_(extent);
      this.dispatchEvent(new ExtentEvent(this.extent_));
    }
  }

  /**
   * Returns the default style for the drawn bbox
   *
   * @return {import("../style/Style.js").StyleFunction} Default Extent style
   */
  function getDefaultExtentStyleFunction() {
    const style = createEditingStyle();
    return function (feature, resolution) {
      return style['Polygon'];
    };
  }

  /**
   * Returns the default style for the pointer
   *
   * @return {import("../style/Style.js").StyleFunction} Default pointer style
   */
  function getDefaultPointerStyleFunction() {
    const style = createEditingStyle();
    return function (feature, resolution) {
      return style['Point'];
    };
  }

  /**
   * @param {import("../coordinate.js").Coordinate} fixedPoint corner that will be unchanged in the new extent
   * @return {function (import("../coordinate.js").Coordinate): import("../extent.js").Extent} event handler
   */
  function getPointHandler(fixedPoint) {
    return function (point) {
      return boundingExtent([fixedPoint, point]);
    };
  }

  /**
   * @param {import("../coordinate.js").Coordinate} fixedP1 first corner that will be unchanged in the new extent
   * @param {import("../coordinate.js").Coordinate} fixedP2 second corner that will be unchanged in the new extent
   * @return {function (import("../coordinate.js").Coordinate): import("../extent.js").Extent|null} event handler
   */
  function getEdgeHandler(fixedP1, fixedP2) {
    if (fixedP1[0] == fixedP2[0]) {
      return function (point) {
        return boundingExtent([fixedP1, [point[0], fixedP2[1]]]);
      };
    } else if (fixedP1[1] == fixedP2[1]) {
      return function (point) {
        return boundingExtent([fixedP1, [fixedP2[0], point[1]]]);
      };
    } else {
      return null;
    }
  }

  /**
   * @param {import("../extent.js").Extent} extent extent
   * @return {Array<Array<import("../coordinate.js").Coordinate>>} extent line segments
   */
  function getSegments(extent) {
    return [
      [
        [extent[0], extent[1]],
        [extent[0], extent[3]],
      ],
      [
        [extent[0], extent[3]],
        [extent[2], extent[3]],
      ],
      [
        [extent[2], extent[3]],
        [extent[2], extent[1]],
      ],
      [
        [extent[2], extent[1]],
        [extent[0], extent[1]],
      ],
    ];
  }

  /**
   * @module ol/interaction/Link
   */

  /**
   * @param {number} number A number.
   * @return {number} A number with at most 5 decimal places.
   */
  function to5(number) {
    return toFixed(number, 5);
  }

  /**
   * @param {string} string A string.
   * @return {number} A number representing the string.
   */
  function readNumber(string) {
    return parseFloat(string);
  }

  /**
   * @param {number} number A number.
   * @return {string} A string representing the number.
   */
  function writeNumber(number) {
    return to5(number).toString();
  }

  /**
   * @param {number} a A number.
   * @param {number} b A number.
   * @return {boolean} The numbers are different.
   */
  function differentNumber(a, b) {
    if (isNaN(a)) {
      return false;
    }
    return a !== readNumber(writeNumber(b));
  }

  /**
   * @param {Array<number>} a An array of two numbers.
   * @param {Array<number>} b An array of two numbers.
   * @return {boolean} The arrays are different.
   */
  function differentArray(a, b) {
    return differentNumber(a[0], b[0]) || differentNumber(a[1], b[1]);
  }

  /**
   * @typedef {Object} Options
   * @property {boolean|import('../View.js').AnimationOptions} [animate=true] Animate view transitions.
   * @property {boolean} [replace=false] Replace the current URL without creating the new entry in browser history.
   * By default, changes in the map state result in a new entry being added to the browser history.
   * @property {string} [prefix=''] By default, the URL will be updated with search parameters x, y, z, and r.  To
   * avoid collisions with existing search parameters that your application uses, you can supply a custom prefix for
   * the ones used by this interaction (e.g. 'ol:').
   */

  /**
   * @classdesc
   * An interaction that synchronizes the map state with the URL.
   *
   * @api
   */
  class Link extends Interaction {
    /**
     * @param {Options} [options] Link options.
     */
    constructor(options) {
      super();

      options = Object.assign(
        {animate: true, replace: false, prefix: ''},
        options || {}
      );

      let animationOptions;
      if (options.animate === true) {
        animationOptions = {duration: 250};
      } else if (!options.animate) {
        animationOptions = null;
      } else {
        animationOptions = options.animate;
      }

      /**
       * @type {import('../View.js').AnimationOptions|null}
       * @private
       */
      this.animationOptions_ = animationOptions;

      /**
       * @private
       * @type {boolean}
       */
      this.replace_ = options.replace;

      /**
       * @private
       * @type {string}
       */
      this.prefix_ = options.prefix;

      /**
       * @private
       * @type {!Array<import("../events.js").EventsKey>}
       */
      this.listenerKeys_ = [];

      /**
       * @private
       * @type {boolean}
       */
      this.initial_ = true;

      this.updateState_ = this.updateState_.bind(this);
    }

    /**
     * @private
     * @param {string} name A parameter name.
     * @return {string} A name with the prefix applied.
     */
    getParamName_(name) {
      if (!this.prefix_) {
        return name;
      }
      return this.prefix_ + name;
    }

    /**
     * @private
     * @param {URLSearchParams} params The search params.
     * @param {string} name The unprefixed parameter name.
     * @return {string|null} The parameter value.
     */
    get_(params, name) {
      return params.get(this.getParamName_(name));
    }

    /**
     * @private
     * @param {URLSearchParams} params The search params.
     * @param {string} name The unprefixed parameter name.
     * @param {string} value The param value.
     */
    set_(params, name, value) {
      params.set(this.getParamName_(name), value);
    }

    /**
     * @private
     * @param {URLSearchParams} params The search params.
     * @param {string} name The unprefixed parameter name.
     */
    delete_(params, name) {
      params.delete(this.getParamName_(name));
    }

    /**
     * @param {import("../Map.js").default|null} map Map.
     */
    setMap(map) {
      const oldMap = this.getMap();
      super.setMap(map);
      if (map === oldMap) {
        return;
      }
      if (oldMap) {
        this.unregisterListeners_(oldMap);
      }
      if (map) {
        this.initial_ = true;
        this.updateState_();
        this.registerListeners_(map);
      }
    }

    /**
     * @param {import("../Map.js").default} map Map.
     * @private
     */
    registerListeners_(map) {
      this.listenerKeys_.push(
        listen(map, MapEventType.MOVEEND, this.updateUrl_, this),
        listen(map.getLayerGroup(), EventType$1.CHANGE, this.updateUrl_, this),
        listen(map, 'change:layergroup', this.handleChangeLayerGroup_, this)
      );

      if (!this.replace_) {
        addEventListener('popstate', this.updateState_);
      }
    }

    /**
     * @param {import("../Map.js").default} map Map.
     * @private
     */
    unregisterListeners_(map) {
      for (let i = 0, ii = this.listenerKeys_.length; i < ii; ++i) {
        unlistenByKey(this.listenerKeys_[i]);
      }
      this.listenerKeys_.length = 0;

      if (!this.replace_) {
        removeEventListener('popstate', this.updateState_);
      }

      const url = new URL(window.location.href);
      const params = url.searchParams;
      this.delete_(params, 'x');
      this.delete_(params, 'y');
      this.delete_(params, 'z');
      this.delete_(params, 'r');
      this.delete_(params, 'l');
      window.history.replaceState(null, '', url);
    }

    /**
     * @private
     */
    handleChangeLayerGroup_() {
      const map = this.getMap();
      if (!map) {
        return;
      }
      this.unregisterListeners_(map);
      this.registerListeners_(map);
      this.initial_ = true;
      this.updateUrl_();
    }

    /**
     * @private
     */
    updateState_() {
      const map = this.getMap();
      if (!map) {
        return;
      }
      const view = map.getView();
      if (!view) {
        return;
      }
      const url = new URL(window.location.href);
      const params = url.searchParams;

      let updateView = false;

      /**
       * @type {import('../View.js').AnimationOptions}
       */
      const viewProperties = {};

      const zoom = readNumber(this.get_(params, 'z'));
      if (differentNumber(zoom, view.getZoom())) {
        updateView = true;
        viewProperties.zoom = zoom;
      }

      const rotation = readNumber(this.get_(params, 'r'));
      if (differentNumber(rotation, view.getRotation())) {
        updateView = true;
        viewProperties.rotation = rotation;
      }

      const center = [
        readNumber(this.get_(params, 'x')),
        readNumber(this.get_(params, 'y')),
      ];
      if (differentArray(center, view.getCenter())) {
        updateView = true;
        viewProperties.center = center;
      }

      if (updateView) {
        if (!this.initial_ && this.animationOptions_) {
          view.animate(Object.assign(viewProperties, this.animationOptions_));
        } else {
          if (viewProperties.center) {
            view.setCenter(viewProperties.center);
          }
          if ('zoom' in viewProperties) {
            view.setZoom(viewProperties.zoom);
          }
          if ('rotation' in viewProperties) {
            view.setRotation(viewProperties.rotation);
          }
        }
      }

      const layers = map.getAllLayers();
      const layersParam = this.get_(params, 'l');
      if (layersParam && layersParam.length === layers.length) {
        for (let i = 0, ii = layers.length; i < ii; ++i) {
          const value = parseInt(layersParam[i]);
          if (!isNaN(value)) {
            const visible = Boolean(value);
            const layer = layers[i];
            if (layer.getVisible() !== visible) {
              layer.setVisible(visible);
            }
          }
        }
      }
    }

    /**
     * @private
     */
    updateUrl_() {
      const map = this.getMap();
      if (!map) {
        return;
      }
      const view = map.getView();
      if (!view) {
        return;
      }
      const initial = this.initial_;
      this.initial_ = false;

      const center = view.getCenter();
      const zoom = view.getZoom();
      const rotation = view.getRotation();

      const layers = map.getAllLayers();
      const visibilities = new Array(layers.length);
      for (let i = 0, ii = layers.length; i < ii; ++i) {
        visibilities[i] = layers[i].getVisible() ? '1' : '0';
      }

      const url = new URL(window.location.href);
      const params = url.searchParams;

      this.set_(params, 'x', writeNumber(center[0]));
      this.set_(params, 'y', writeNumber(center[1]));
      this.set_(params, 'z', writeNumber(zoom));
      this.set_(params, 'r', writeNumber(rotation));
      this.set_(params, 'l', visibilities.join(''));

      if (url.href !== window.location.href) {
        if (initial || this.replace_) {
          window.history.replaceState(null, '', url);
        } else {
          window.history.pushState(null, '', url);
        }
      }
    }
  }

  /**
   * @module ol/interaction/Modify
   */

  /**
   * The segment index assigned to a circle's center when
   * breaking up a circle into ModifySegmentDataType segments.
   * @type {number}
   */
  const CIRCLE_CENTER_INDEX = 0;

  /**
   * The segment index assigned to a circle's circumference when
   * breaking up a circle into ModifySegmentDataType segments.
   * @type {number}
   */
  const CIRCLE_CIRCUMFERENCE_INDEX = 1;

  const tempExtent = [0, 0, 0, 0];
  const tempSegment$1 = [];

  /**
   * @enum {string}
   */
  const ModifyEventType = {
    /**
     * Triggered upon feature modification start
     * @event ModifyEvent#modifystart
     * @api
     */
    MODIFYSTART: 'modifystart',
    /**
     * Triggered upon feature modification end
     * @event ModifyEvent#modifyend
     * @api
     */
    MODIFYEND: 'modifyend',
  };

  /**
   * @typedef {Object} SegmentData
   * @property {Array<number>} [depth] Depth.
   * @property {import("../Feature").FeatureLike} feature Feature.
   * @property {import("../geom/SimpleGeometry.js").default} geometry Geometry.
   * @property {number} [index] Index.
   * @property {Array<Array<number>>} segment Segment.
   * @property {Array<SegmentData>} [featureSegments] FeatureSegments.
   */

  /**
   * @typedef {Object} Options
   * @property {import("../events/condition.js").Condition} [condition] A function that
   * takes an {@link module:ol/MapBrowserEvent~MapBrowserEvent} and returns a
   * boolean to indicate whether that event will be considered to add or move a
   * vertex to the sketch. Default is
   * {@link module:ol/events/condition.primaryAction}.
   * @property {import("../events/condition.js").Condition} [deleteCondition] A function
   * that takes an {@link module:ol/MapBrowserEvent~MapBrowserEvent} and returns a
   * boolean to indicate whether that event should be handled. By default,
   * {@link module:ol/events/condition.singleClick} with
   * {@link module:ol/events/condition.altKeyOnly} results in a vertex deletion.
   * @property {import("../events/condition.js").Condition} [insertVertexCondition] A
   * function that takes an {@link module:ol/MapBrowserEvent~MapBrowserEvent} and
   * returns a boolean to indicate whether a new vertex should be added to the sketch
   * features. Default is {@link module:ol/events/condition.always}.
   * @property {number} [pixelTolerance=10] Pixel tolerance for considering the
   * pointer close enough to a segment or vertex for editing.
   * @property {import("../style/Style.js").StyleLike|import("../style/flat.js").FlatStyleLike} [style]
   * Style used for the modification point or vertex. For linestrings and polygons, this will
   * be the affected vertex, for circles a point along the circle, and for points the actual
   * point. If not configured, the default edit style is used (see {@link module:ol/style/Style~Style}).
   * When using a style function, the point feature passed to the function will have a `features`
   * property - an array whose entries are the features that are being modified, and a `geometries`
   * property - an array whose entries are the geometries that are being modified. Both arrays are
   * in the same order. The `geometries` are only useful when modifying geometry collections, where
   * the geometry will be the particular geometry from the collection that is being modified.
   * @property {VectorSource} [source] The vector source with
   * features to modify.  If a vector source is not provided, a feature collection
   * must be provided with the `features` option.
   * @property {boolean|import("../layer/BaseVector").default} [hitDetection] When configured, point
   * features will be considered for modification based on their visual appearance, instead of being within
   * the `pixelTolerance` from the pointer location. When a {@link module:ol/layer/BaseVector~BaseVectorLayer} is
   * provided, only the rendered representation of the features on that layer will be considered.
   * @property {Collection<Feature>} [features]
   * The features the interaction works on.  If a feature collection is not
   * provided, a vector source must be provided with the `source` option.
   * @property {boolean} [wrapX=false] Wrap the world horizontally on the sketch
   * overlay.
   * @property {boolean} [snapToPointer=!hitDetection] The vertex, point or segment being modified snaps to the
   * pointer coordinate when clicked within the `pixelTolerance`.
   */

  /**
   * @classdesc
   * Events emitted by {@link module:ol/interaction/Modify~Modify} instances are
   * instances of this type.
   */
  class ModifyEvent extends BaseEvent {
    /**
     * @param {ModifyEventType} type Type.
     * @param {Collection<import("../Feature").FeatureLike>} features
     * The features modified.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent
     * Associated {@link module:ol/MapBrowserEvent~MapBrowserEvent}.
     */
    constructor(type, features, mapBrowserEvent) {
      super(type);

      /**
       * The features being modified.
       * @type {Collection<import("../Feature").FeatureLike>}
       * @api
       */
      this.features = features;

      /**
       * Associated {@link module:ol/MapBrowserEvent~MapBrowserEvent}.
       * @type {import("../MapBrowserEvent.js").default}
       * @api
       */
      this.mapBrowserEvent = mapBrowserEvent;
    }
  }

  /***
   * @template Return
   * @typedef {import("../Observable").OnSignature<import("../Observable").EventTypes, import("../events/Event.js").default, Return> &
   *   import("../Observable").OnSignature<import("../ObjectEventType").Types|
   *     'change:active', import("../Object").ObjectEvent, Return> &
   *   import("../Observable").OnSignature<'modifyend'|'modifystart', ModifyEvent, Return> &
   *   import("../Observable").CombinedOnSignature<import("../Observable").EventTypes|import("../ObjectEventType").Types|
   *     'change:active'|'modifyend'|'modifystart', Return>} ModifyOnSignature
   */

  /**
   * @classdesc
   * Interaction for modifying feature geometries.  To modify features that have
   * been added to an existing source, construct the modify interaction with the
   * `source` option.  If you want to modify features in a collection (for example,
   * the collection used by a select interaction), construct the interaction with
   * the `features` option.  The interaction must be constructed with either a
   * `source` or `features` option.
   *
   * Cartesian distance from the pointer is used to determine the features that
   * will be modified. This means that geometries will only be considered for
   * modification when they are within the configured `pixelTolerance`. For point
   * geometries, the `hitDetection` option can be used to match their visual
   * appearance.
   *
   * By default, the interaction will allow deletion of vertices when the `alt`
   * key is pressed.  To configure the interaction with a different condition
   * for deletion, use the `deleteCondition` option.
   * @fires ModifyEvent
   * @api
   */
  class Modify extends PointerInteraction {
    /**
     * @param {Options} options Options.
     */
    constructor(options) {
      super(/** @type {import("./Pointer.js").Options} */ (options));

      /***
       * @type {ModifyOnSignature<import("../events").EventsKey>}
       */
      this.on;

      /***
       * @type {ModifyOnSignature<import("../events").EventsKey>}
       */
      this.once;

      /***
       * @type {ModifyOnSignature<void>}
       */
      this.un;

      /** @private */
      this.boundHandleFeatureChange_ = this.handleFeatureChange_.bind(this);

      /**
       * @private
       * @type {import("../events/condition.js").Condition}
       */
      this.condition_ = options.condition ? options.condition : primaryAction;

      /**
       * @private
       * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Browser event.
       * @return {boolean} Combined condition result.
       */
      this.defaultDeleteCondition_ = function (mapBrowserEvent) {
        return altKeyOnly(mapBrowserEvent) && singleClick(mapBrowserEvent);
      };

      /**
       * @type {import("../events/condition.js").Condition}
       * @private
       */
      this.deleteCondition_ = options.deleteCondition
        ? options.deleteCondition
        : this.defaultDeleteCondition_;

      /**
       * @type {import("../events/condition.js").Condition}
       * @private
       */
      this.insertVertexCondition_ = options.insertVertexCondition
        ? options.insertVertexCondition
        : always;

      /**
       * Editing vertex.
       * @type {Feature<Point>}
       * @private
       */
      this.vertexFeature_ = null;

      /**
       * Segments intersecting {@link this.vertexFeature_} by segment uid.
       * @type {Object<string, boolean>}
       * @private
       */
      this.vertexSegments_ = null;

      /**
       * @type {import("../pixel.js").Pixel}
       * @private
       */
      this.lastPixel_ = [0, 0];

      /**
       * Tracks if the next `singleclick` event should be ignored to prevent
       * accidental deletion right after vertex creation.
       * @type {boolean}
       * @private
       */
      this.ignoreNextSingleClick_ = false;

      /**
       * @type {Collection<import("../Feature").FeatureLike>}
       * @private
       */
      this.featuresBeingModified_ = null;

      /**
       * Segment RTree for each layer
       * @type {RBush<SegmentData>}
       * @private
       */
      this.rBush_ = new RBush();

      /**
       * @type {number}
       * @private
       */
      this.pixelTolerance_ =
        options.pixelTolerance !== undefined ? options.pixelTolerance : 10;

      /**
       * @type {boolean}
       * @private
       */
      this.snappedToVertex_ = false;

      /**
       * Indicate whether the interaction is currently changing a feature's
       * coordinates.
       * @type {boolean}
       * @private
       */
      this.changingFeature_ = false;

      /**
       * @type {Array}
       * @private
       */
      this.dragSegments_ = [];

      /**
       * Draw overlay where sketch features are drawn.
       * @type {VectorLayer}
       * @private
       */
      this.overlay_ = new VectorLayer({
        source: new VectorSource({
          useSpatialIndex: false,
          wrapX: !!options.wrapX,
        }),
        style: options.style ? options.style : getDefaultStyleFunction$1(),
        updateWhileAnimating: true,
        updateWhileInteracting: true,
      });

      /**
       * @const
       * @private
       * @type {!Object<string, function(Feature, import("../geom/Geometry.js").default): void>}
       */
      this.SEGMENT_WRITERS_ = {
        'Point': this.writePointGeometry_.bind(this),
        'LineString': this.writeLineStringGeometry_.bind(this),
        'LinearRing': this.writeLineStringGeometry_.bind(this),
        'Polygon': this.writePolygonGeometry_.bind(this),
        'MultiPoint': this.writeMultiPointGeometry_.bind(this),
        'MultiLineString': this.writeMultiLineStringGeometry_.bind(this),
        'MultiPolygon': this.writeMultiPolygonGeometry_.bind(this),
        'Circle': this.writeCircleGeometry_.bind(this),
        'GeometryCollection': this.writeGeometryCollectionGeometry_.bind(this),
      };

      /**
       * @type {VectorSource}
       * @private
       */
      this.source_ = null;

      /**
       * @type {boolean|import("../layer/BaseVector").default}
       */
      this.hitDetection_ = null;

      /** @type {Collection<Feature>} */
      let features;
      if (options.features) {
        features = options.features;
      } else if (options.source) {
        this.source_ = options.source;
        features = new Collection(this.source_.getFeatures());
        this.source_.addEventListener(
          VectorEventType.ADDFEATURE,
          this.handleSourceAdd_.bind(this)
        );
        this.source_.addEventListener(
          VectorEventType.REMOVEFEATURE,
          this.handleSourceRemove_.bind(this)
        );
      }
      if (!features) {
        throw new Error(
          'The modify interaction requires features, a source or a layer'
        );
      }
      if (options.hitDetection) {
        this.hitDetection_ = options.hitDetection;
      }

      /**
       * @type {Collection<import("../Feature.js").FeatureLike>}
       * @private
       */
      this.features_ = features;

      this.features_.forEach(this.addFeature_.bind(this));
      this.features_.addEventListener(
        CollectionEventType.ADD,
        this.handleFeatureAdd_.bind(this)
      );
      this.features_.addEventListener(
        CollectionEventType.REMOVE,
        this.handleFeatureRemove_.bind(this)
      );

      /**
       * @type {import("../MapBrowserEvent.js").default}
       * @private
       */
      this.lastPointerEvent_ = null;

      /**
       * Delta (x, y in map units) between matched rtree vertex and pointer vertex.
       * @type {Array<number>}
       */
      this.delta_ = [0, 0];

      /**
       * @private
       */
      this.snapToPointer_ =
        options.snapToPointer === undefined
          ? !this.hitDetection_
          : options.snapToPointer;
    }

    /**
     * @param {Feature} feature Feature.
     * @private
     */
    addFeature_(feature) {
      const geometry = feature.getGeometry();
      if (geometry) {
        const writer = this.SEGMENT_WRITERS_[geometry.getType()];
        if (writer) {
          writer(feature, geometry);
        }
      }
      const map = this.getMap();
      if (map && map.isRendered() && this.getActive()) {
        this.handlePointerAtPixel_(this.lastPixel_, map);
      }
      feature.addEventListener(EventType$1.CHANGE, this.boundHandleFeatureChange_);
    }

    /**
     * @param {import("../MapBrowserEvent.js").default} evt Map browser event.
     * @param {Array<Array<SegmentData>>} segments The segments subject to modification.
     * @private
     */
    willModifyFeatures_(evt, segments) {
      if (!this.featuresBeingModified_) {
        this.featuresBeingModified_ = new Collection();
        const features = this.featuresBeingModified_.getArray();
        for (let i = 0, ii = segments.length; i < ii; ++i) {
          const segment = segments[i];
          for (let s = 0, ss = segment.length; s < ss; ++s) {
            const feature = segment[s].feature;
            if (feature && !features.includes(feature)) {
              this.featuresBeingModified_.push(feature);
            }
          }
        }
        if (this.featuresBeingModified_.getLength() === 0) {
          this.featuresBeingModified_ = null;
        } else {
          this.dispatchEvent(
            new ModifyEvent(
              ModifyEventType.MODIFYSTART,
              this.featuresBeingModified_,
              evt
            )
          );
        }
      }
    }

    /**
     * @param {Feature} feature Feature.
     * @private
     */
    removeFeature_(feature) {
      this.removeFeatureSegmentData_(feature);
      // Remove the vertex feature if the collection of candidate features is empty.
      if (this.vertexFeature_ && this.features_.getLength() === 0) {
        this.overlay_.getSource().removeFeature(this.vertexFeature_);
        this.vertexFeature_ = null;
      }
      feature.removeEventListener(
        EventType$1.CHANGE,
        this.boundHandleFeatureChange_
      );
    }

    /**
     * @param {Feature} feature Feature.
     * @private
     */
    removeFeatureSegmentData_(feature) {
      const rBush = this.rBush_;
      /** @type {Array<SegmentData>} */
      const nodesToRemove = [];
      rBush.forEach(
        /**
         * @param {SegmentData} node RTree node.
         */
        function (node) {
          if (feature === node.feature) {
            nodesToRemove.push(node);
          }
        }
      );
      for (let i = nodesToRemove.length - 1; i >= 0; --i) {
        const nodeToRemove = nodesToRemove[i];
        for (let j = this.dragSegments_.length - 1; j >= 0; --j) {
          if (this.dragSegments_[j][0] === nodeToRemove) {
            this.dragSegments_.splice(j, 1);
          }
        }
        rBush.remove(nodeToRemove);
      }
    }

    /**
     * Activate or deactivate the interaction.
     * @param {boolean} active Active.
     * @observable
     * @api
     */
    setActive(active) {
      if (this.vertexFeature_ && !active) {
        this.overlay_.getSource().removeFeature(this.vertexFeature_);
        this.vertexFeature_ = null;
      }
      super.setActive(active);
    }

    /**
     * Remove the interaction from its current map and attach it to the new map.
     * Subclasses may set up event handlers to get notified about changes to
     * the map here.
     * @param {import("../Map.js").default} map Map.
     */
    setMap(map) {
      this.overlay_.setMap(map);
      super.setMap(map);
    }

    /**
     * Get the overlay layer that this interaction renders the modification point or vertex to.
     * @return {VectorLayer} Overlay layer.
     * @api
     */
    getOverlay() {
      return this.overlay_;
    }

    /**
     * @param {import("../source/Vector.js").VectorSourceEvent} event Event.
     * @private
     */
    handleSourceAdd_(event) {
      if (event.feature) {
        this.features_.push(event.feature);
      }
    }

    /**
     * @param {import("../source/Vector.js").VectorSourceEvent} event Event.
     * @private
     */
    handleSourceRemove_(event) {
      if (event.feature) {
        this.features_.remove(event.feature);
      }
    }

    /**
     * @param {import("../Collection.js").CollectionEvent<Feature>} evt Event.
     * @private
     */
    handleFeatureAdd_(evt) {
      this.addFeature_(evt.element);
    }

    /**
     * @param {import("../events/Event.js").default} evt Event.
     * @private
     */
    handleFeatureChange_(evt) {
      if (!this.changingFeature_) {
        const feature = /** @type {Feature} */ (evt.target);
        this.removeFeature_(feature);
        this.addFeature_(feature);
      }
    }

    /**
     * @param {import("../Collection.js").CollectionEvent<Feature>} evt Event.
     * @private
     */
    handleFeatureRemove_(evt) {
      this.removeFeature_(evt.element);
    }

    /**
     * @param {Feature} feature Feature
     * @param {Point} geometry Geometry.
     * @private
     */
    writePointGeometry_(feature, geometry) {
      const coordinates = geometry.getCoordinates();

      /** @type {SegmentData} */
      const segmentData = {
        feature: feature,
        geometry: geometry,
        segment: [coordinates, coordinates],
      };

      this.rBush_.insert(geometry.getExtent(), segmentData);
    }

    /**
     * @param {Feature} feature Feature
     * @param {import("../geom/MultiPoint.js").default} geometry Geometry.
     * @private
     */
    writeMultiPointGeometry_(feature, geometry) {
      const points = geometry.getCoordinates();
      for (let i = 0, ii = points.length; i < ii; ++i) {
        const coordinates = points[i];

        /** @type {SegmentData} */
        const segmentData = {
          feature: feature,
          geometry: geometry,
          depth: [i],
          index: i,
          segment: [coordinates, coordinates],
        };

        this.rBush_.insert(geometry.getExtent(), segmentData);
      }
    }

    /**
     * @param {Feature} feature Feature
     * @param {import("../geom/LineString.js").default} geometry Geometry.
     * @private
     */
    writeLineStringGeometry_(feature, geometry) {
      const coordinates = geometry.getCoordinates();
      for (let i = 0, ii = coordinates.length - 1; i < ii; ++i) {
        const segment = coordinates.slice(i, i + 2);

        /** @type {SegmentData} */
        const segmentData = {
          feature: feature,
          geometry: geometry,
          index: i,
          segment: segment,
        };

        this.rBush_.insert(boundingExtent(segment), segmentData);
      }
    }

    /**
     * @param {Feature} feature Feature
     * @param {import("../geom/MultiLineString.js").default} geometry Geometry.
     * @private
     */
    writeMultiLineStringGeometry_(feature, geometry) {
      const lines = geometry.getCoordinates();
      for (let j = 0, jj = lines.length; j < jj; ++j) {
        const coordinates = lines[j];
        for (let i = 0, ii = coordinates.length - 1; i < ii; ++i) {
          const segment = coordinates.slice(i, i + 2);

          /** @type {SegmentData} */
          const segmentData = {
            feature: feature,
            geometry: geometry,
            depth: [j],
            index: i,
            segment: segment,
          };

          this.rBush_.insert(boundingExtent(segment), segmentData);
        }
      }
    }

    /**
     * @param {Feature} feature Feature
     * @param {import("../geom/Polygon.js").default} geometry Geometry.
     * @private
     */
    writePolygonGeometry_(feature, geometry) {
      const rings = geometry.getCoordinates();
      for (let j = 0, jj = rings.length; j < jj; ++j) {
        const coordinates = rings[j];
        for (let i = 0, ii = coordinates.length - 1; i < ii; ++i) {
          const segment = coordinates.slice(i, i + 2);

          /** @type {SegmentData} */
          const segmentData = {
            feature: feature,
            geometry: geometry,
            depth: [j],
            index: i,
            segment: segment,
          };

          this.rBush_.insert(boundingExtent(segment), segmentData);
        }
      }
    }

    /**
     * @param {Feature} feature Feature
     * @param {import("../geom/MultiPolygon.js").default} geometry Geometry.
     * @private
     */
    writeMultiPolygonGeometry_(feature, geometry) {
      const polygons = geometry.getCoordinates();
      for (let k = 0, kk = polygons.length; k < kk; ++k) {
        const rings = polygons[k];
        for (let j = 0, jj = rings.length; j < jj; ++j) {
          const coordinates = rings[j];
          for (let i = 0, ii = coordinates.length - 1; i < ii; ++i) {
            const segment = coordinates.slice(i, i + 2);

            /** @type {SegmentData} */
            const segmentData = {
              feature: feature,
              geometry: geometry,
              depth: [j, k],
              index: i,
              segment: segment,
            };

            this.rBush_.insert(boundingExtent(segment), segmentData);
          }
        }
      }
    }

    /**
     * We convert a circle into two segments.  The segment at index
     * {@link CIRCLE_CENTER_INDEX} is the
     * circle's center (a point).  The segment at index
     * {@link CIRCLE_CIRCUMFERENCE_INDEX} is
     * the circumference, and is not a line segment.
     *
     * @param {Feature} feature Feature.
     * @param {import("../geom/Circle.js").default} geometry Geometry.
     * @private
     */
    writeCircleGeometry_(feature, geometry) {
      const coordinates = geometry.getCenter();

      /** @type {SegmentData} */
      const centerSegmentData = {
        feature: feature,
        geometry: geometry,
        index: CIRCLE_CENTER_INDEX,
        segment: [coordinates, coordinates],
      };

      /** @type {SegmentData} */
      const circumferenceSegmentData = {
        feature: feature,
        geometry: geometry,
        index: CIRCLE_CIRCUMFERENCE_INDEX,
        segment: [coordinates, coordinates],
      };

      const featureSegments = [centerSegmentData, circumferenceSegmentData];
      centerSegmentData.featureSegments = featureSegments;
      circumferenceSegmentData.featureSegments = featureSegments;
      this.rBush_.insert(createOrUpdateFromCoordinate(coordinates), centerSegmentData);
      let circleGeometry = /** @type {import("../geom/Geometry.js").default} */ (
        geometry
      );
      const userProjection = getUserProjection();
      if (userProjection && this.getMap()) {
        const projection = this.getMap().getView().getProjection();
        circleGeometry = circleGeometry
          .clone()
          .transform(userProjection, projection);
        circleGeometry = fromCircle(
          /** @type {import("../geom/Circle.js").default} */ (circleGeometry)
        ).transform(projection, userProjection);
      }
      this.rBush_.insert(circleGeometry.getExtent(), circumferenceSegmentData);
    }

    /**
     * @param {Feature} feature Feature
     * @param {import("../geom/GeometryCollection.js").default} geometry Geometry.
     * @private
     */
    writeGeometryCollectionGeometry_(feature, geometry) {
      const geometries = geometry.getGeometriesArray();
      for (let i = 0; i < geometries.length; ++i) {
        const geometry = geometries[i];
        const writer = this.SEGMENT_WRITERS_[geometry.getType()];
        writer(feature, geometry);
      }
    }

    /**
     * @param {import("../coordinate.js").Coordinate} coordinates Coordinates.
     * @param {Array<import("../Feature").FeatureLike>} features The features being modified.
     * @param {Array<import("../geom/SimpleGeometry.js").default>} geometries The geometries being modified.
     * @return {Feature} Vertex feature.
     * @private
     */
    createOrUpdateVertexFeature_(coordinates, features, geometries) {
      let vertexFeature = this.vertexFeature_;
      if (!vertexFeature) {
        vertexFeature = new Feature(new Point(coordinates));
        this.vertexFeature_ = vertexFeature;
        this.overlay_.getSource().addFeature(vertexFeature);
      } else {
        const geometry = vertexFeature.getGeometry();
        geometry.setCoordinates(coordinates);
      }
      vertexFeature.set('features', features);
      vertexFeature.set('geometries', geometries);
      return vertexFeature;
    }

    /**
     * Handles the {@link module:ol/MapBrowserEvent~MapBrowserEvent map browser event} and may modify the geometry.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Map browser event.
     * @return {boolean} `false` to stop event propagation.
     */
    handleEvent(mapBrowserEvent) {
      if (!mapBrowserEvent.originalEvent) {
        return true;
      }
      this.lastPointerEvent_ = mapBrowserEvent;

      let handled;
      if (
        !mapBrowserEvent.map.getView().getInteracting() &&
        mapBrowserEvent.type == MapBrowserEventType.POINTERMOVE &&
        !this.handlingDownUpSequence
      ) {
        this.handlePointerMove_(mapBrowserEvent);
      }
      if (this.vertexFeature_ && this.deleteCondition_(mapBrowserEvent)) {
        if (
          mapBrowserEvent.type != MapBrowserEventType.SINGLECLICK ||
          !this.ignoreNextSingleClick_
        ) {
          handled = this.removePoint();
        } else {
          handled = true;
        }
      }

      if (mapBrowserEvent.type == MapBrowserEventType.SINGLECLICK) {
        this.ignoreNextSingleClick_ = false;
      }

      return super.handleEvent(mapBrowserEvent) && !handled;
    }

    /**
     * Handle pointer drag events.
     * @param {import("../MapBrowserEvent.js").default} evt Event.
     */
    handleDragEvent(evt) {
      this.ignoreNextSingleClick_ = false;
      this.willModifyFeatures_(evt, this.dragSegments_);

      const vertex = [
        evt.coordinate[0] + this.delta_[0],
        evt.coordinate[1] + this.delta_[1],
      ];
      const features = [];
      const geometries = [];
      for (let i = 0, ii = this.dragSegments_.length; i < ii; ++i) {
        const dragSegment = this.dragSegments_[i];
        const segmentData = dragSegment[0];
        const feature = segmentData.feature;
        if (!features.includes(feature)) {
          features.push(feature);
        }
        const geometry = segmentData.geometry;
        if (!geometries.includes(geometry)) {
          geometries.push(geometry);
        }
        const depth = segmentData.depth;
        let coordinates;
        const segment = segmentData.segment;
        const index = dragSegment[1];

        while (vertex.length < geometry.getStride()) {
          vertex.push(segment[index][vertex.length]);
        }

        switch (geometry.getType()) {
          case 'Point':
            coordinates = vertex;
            segment[0] = vertex;
            segment[1] = vertex;
            break;
          case 'MultiPoint':
            coordinates = geometry.getCoordinates();
            coordinates[segmentData.index] = vertex;
            segment[0] = vertex;
            segment[1] = vertex;
            break;
          case 'LineString':
            coordinates = geometry.getCoordinates();
            coordinates[segmentData.index + index] = vertex;
            segment[index] = vertex;
            break;
          case 'MultiLineString':
            coordinates = geometry.getCoordinates();
            coordinates[depth[0]][segmentData.index + index] = vertex;
            segment[index] = vertex;
            break;
          case 'Polygon':
            coordinates = geometry.getCoordinates();
            coordinates[depth[0]][segmentData.index + index] = vertex;
            segment[index] = vertex;
            break;
          case 'MultiPolygon':
            coordinates = geometry.getCoordinates();
            coordinates[depth[1]][depth[0]][segmentData.index + index] = vertex;
            segment[index] = vertex;
            break;
          case 'Circle':
            segment[0] = vertex;
            segment[1] = vertex;
            if (segmentData.index === CIRCLE_CENTER_INDEX) {
              this.changingFeature_ = true;
              geometry.setCenter(vertex);
              this.changingFeature_ = false;
            } else {
              // We're dragging the circle's circumference:
              this.changingFeature_ = true;
              const projection = evt.map.getView().getProjection();
              let radius = distance(
                fromUserCoordinate(geometry.getCenter(), projection),
                fromUserCoordinate(vertex, projection)
              );
              const userProjection = getUserProjection();
              if (userProjection) {
                const circleGeometry = geometry
                  .clone()
                  .transform(userProjection, projection);
                circleGeometry.setRadius(radius);
                radius = circleGeometry
                  .transform(projection, userProjection)
                  .getRadius();
              }
              geometry.setRadius(radius);
              this.changingFeature_ = false;
            }
            break;
          // pass
        }

        if (coordinates) {
          this.setGeometryCoordinates_(geometry, coordinates);
        }
      }
      this.createOrUpdateVertexFeature_(vertex, features, geometries);
    }

    /**
     * Handle pointer down events.
     * @param {import("../MapBrowserEvent.js").default} evt Event.
     * @return {boolean} If the event was consumed.
     */
    handleDownEvent(evt) {
      if (!this.condition_(evt)) {
        return false;
      }
      const pixelCoordinate = evt.coordinate;
      this.handlePointerAtPixel_(evt.pixel, evt.map, pixelCoordinate);
      this.dragSegments_.length = 0;
      this.featuresBeingModified_ = null;
      const vertexFeature = this.vertexFeature_;
      if (vertexFeature) {
        const projection = evt.map.getView().getProjection();
        const insertVertices = [];
        const vertex = vertexFeature.getGeometry().getCoordinates();
        const vertexExtent = boundingExtent([vertex]);
        const segmentDataMatches = this.rBush_.getInExtent(vertexExtent);
        const componentSegments = {};
        segmentDataMatches.sort(compareIndexes);
        for (let i = 0, ii = segmentDataMatches.length; i < ii; ++i) {
          const segmentDataMatch = segmentDataMatches[i];
          const segment = segmentDataMatch.segment;
          let uid = getUid(segmentDataMatch.geometry);
          const depth = segmentDataMatch.depth;
          if (depth) {
            uid += '-' + depth.join('-'); // separate feature components
          }
          if (!componentSegments[uid]) {
            componentSegments[uid] = new Array(2);
          }

          if (
            segmentDataMatch.geometry.getType() === 'Circle' &&
            segmentDataMatch.index === CIRCLE_CIRCUMFERENCE_INDEX
          ) {
            const closestVertex = closestOnSegmentData(
              pixelCoordinate,
              segmentDataMatch,
              projection
            );
            if (
              equals$1(closestVertex, vertex) &&
              !componentSegments[uid][0]
            ) {
              this.dragSegments_.push([segmentDataMatch, 0]);
              componentSegments[uid][0] = segmentDataMatch;
            }
            continue;
          }

          if (
            equals$1(segment[0], vertex) &&
            !componentSegments[uid][0]
          ) {
            this.dragSegments_.push([segmentDataMatch, 0]);
            componentSegments[uid][0] = segmentDataMatch;
            continue;
          }

          if (
            equals$1(segment[1], vertex) &&
            !componentSegments[uid][1]
          ) {
            if (
              componentSegments[uid][0] &&
              componentSegments[uid][0].index === 0
            ) {
              let coordinates = segmentDataMatch.geometry.getCoordinates();
              switch (segmentDataMatch.geometry.getType()) {
                // prevent dragging closed linestrings by the connecting node
                case 'LineString':
                case 'MultiLineString':
                  continue;
                // if dragging the first vertex of a polygon, ensure the other segment
                // belongs to the closing vertex of the linear ring
                case 'MultiPolygon':
                  coordinates = coordinates[depth[1]];
                /* falls through */
                case 'Polygon':
                  if (
                    segmentDataMatch.index !==
                    coordinates[depth[0]].length - 2
                  ) {
                    continue;
                  }
                  break;
                // pass
              }
            }

            this.dragSegments_.push([segmentDataMatch, 1]);
            componentSegments[uid][1] = segmentDataMatch;
            continue;
          }

          if (
            getUid(segment) in this.vertexSegments_ &&
            !componentSegments[uid][0] &&
            !componentSegments[uid][1] &&
            this.insertVertexCondition_(evt)
          ) {
            insertVertices.push(segmentDataMatch);
          }
        }

        if (insertVertices.length) {
          this.willModifyFeatures_(evt, [insertVertices]);
        }

        for (let j = insertVertices.length - 1; j >= 0; --j) {
          this.insertVertex_(insertVertices[j], vertex);
        }
      }
      return !!this.vertexFeature_;
    }

    /**
     * Handle pointer up events.
     * @param {import("../MapBrowserEvent.js").default} evt Event.
     * @return {boolean} If the event was consumed.
     */
    handleUpEvent(evt) {
      for (let i = this.dragSegments_.length - 1; i >= 0; --i) {
        const segmentData = this.dragSegments_[i][0];
        const geometry = segmentData.geometry;
        if (geometry.getType() === 'Circle') {
          // Update a circle object in the R* bush:
          const coordinates = geometry.getCenter();
          const centerSegmentData = segmentData.featureSegments[0];
          const circumferenceSegmentData = segmentData.featureSegments[1];
          centerSegmentData.segment[0] = coordinates;
          centerSegmentData.segment[1] = coordinates;
          circumferenceSegmentData.segment[0] = coordinates;
          circumferenceSegmentData.segment[1] = coordinates;
          this.rBush_.update(createOrUpdateFromCoordinate(coordinates), centerSegmentData);
          let circleGeometry = geometry;
          const userProjection = getUserProjection();
          if (userProjection) {
            const projection = evt.map.getView().getProjection();
            circleGeometry = circleGeometry
              .clone()
              .transform(userProjection, projection);
            circleGeometry = fromCircle(circleGeometry).transform(
              projection,
              userProjection
            );
          }
          this.rBush_.update(
            circleGeometry.getExtent(),
            circumferenceSegmentData
          );
        } else {
          this.rBush_.update(boundingExtent(segmentData.segment), segmentData);
        }
      }
      if (this.featuresBeingModified_) {
        this.dispatchEvent(
          new ModifyEvent(
            ModifyEventType.MODIFYEND,
            this.featuresBeingModified_,
            evt
          )
        );
        this.featuresBeingModified_ = null;
      }
      return false;
    }

    /**
     * @param {import("../MapBrowserEvent.js").default} evt Event.
     * @private
     */
    handlePointerMove_(evt) {
      this.lastPixel_ = evt.pixel;
      this.handlePointerAtPixel_(evt.pixel, evt.map, evt.coordinate);
    }

    /**
     * @param {import("../pixel.js").Pixel} pixel Pixel
     * @param {import("../Map.js").default} map Map.
     * @param {import("../coordinate.js").Coordinate} [coordinate] The pixel Coordinate.
     * @private
     */
    handlePointerAtPixel_(pixel, map, coordinate) {
      const pixelCoordinate = coordinate || map.getCoordinateFromPixel(pixel);
      const projection = map.getView().getProjection();
      const sortByDistance = function (a, b) {
        return (
          projectedDistanceToSegmentDataSquared(pixelCoordinate, a, projection) -
          projectedDistanceToSegmentDataSquared(pixelCoordinate, b, projection)
        );
      };

      /** @type {Array<SegmentData>|undefined} */
      let nodes;
      let hitPointGeometry;
      if (this.hitDetection_) {
        const layerFilter =
          typeof this.hitDetection_ === 'object'
            ? (layer) => layer === this.hitDetection_
            : undefined;
        map.forEachFeatureAtPixel(
          pixel,
          (feature, layer, geometry) => {
            geometry =
              geometry ||
              /** @type {import("../geom/SimpleGeometry").default} */ (
                feature.getGeometry()
              );
            if (
              geometry.getType() === 'Point' &&
              this.features_.getArray().includes(feature)
            ) {
              hitPointGeometry = geometry;
              const coordinate = geometry.getFlatCoordinates().slice(0, 2);
              nodes = [
                {
                  feature,
                  geometry,
                  segment: [coordinate, coordinate],
                },
              ];
            }
            return true;
          },
          {layerFilter}
        );
      }
      if (!nodes) {
        const viewExtent = fromUserExtent(
          createOrUpdateFromCoordinate(pixelCoordinate, tempExtent),
          projection
        );
        const buffer = map.getView().getResolution() * this.pixelTolerance_;
        const box = toUserExtent(
          buffer$1(viewExtent, buffer, tempExtent),
          projection
        );
        nodes = this.rBush_.getInExtent(box);
      }

      if (nodes && nodes.length > 0) {
        const node = nodes.sort(sortByDistance)[0];
        const closestSegment = node.segment;
        let vertex = closestOnSegmentData(pixelCoordinate, node, projection);
        const vertexPixel = map.getPixelFromCoordinate(vertex);
        let dist = distance(pixel, vertexPixel);
        if (hitPointGeometry || dist <= this.pixelTolerance_) {
          /** @type {Object<string, boolean>} */
          const vertexSegments = {};
          vertexSegments[getUid(closestSegment)] = true;

          if (!this.snapToPointer_) {
            this.delta_[0] = vertex[0] - pixelCoordinate[0];
            this.delta_[1] = vertex[1] - pixelCoordinate[1];
          }
          if (
            node.geometry.getType() === 'Circle' &&
            node.index === CIRCLE_CIRCUMFERENCE_INDEX
          ) {
            this.snappedToVertex_ = true;
            this.createOrUpdateVertexFeature_(
              vertex,
              [node.feature],
              [node.geometry]
            );
          } else {
            const pixel1 = map.getPixelFromCoordinate(closestSegment[0]);
            const pixel2 = map.getPixelFromCoordinate(closestSegment[1]);
            const squaredDist1 = squaredDistance(vertexPixel, pixel1);
            const squaredDist2 = squaredDistance(vertexPixel, pixel2);
            dist = Math.sqrt(Math.min(squaredDist1, squaredDist2));
            this.snappedToVertex_ = dist <= this.pixelTolerance_;
            if (this.snappedToVertex_) {
              vertex =
                squaredDist1 > squaredDist2
                  ? closestSegment[1]
                  : closestSegment[0];
            }
            this.createOrUpdateVertexFeature_(
              vertex,
              [node.feature],
              [node.geometry]
            );
            const geometries = {};
            geometries[getUid(node.geometry)] = true;
            for (let i = 1, ii = nodes.length; i < ii; ++i) {
              const segment = nodes[i].segment;
              if (
                (equals$1(closestSegment[0], segment[0]) &&
                  equals$1(closestSegment[1], segment[1])) ||
                (equals$1(closestSegment[0], segment[1]) &&
                  equals$1(closestSegment[1], segment[0]))
              ) {
                const geometryUid = getUid(nodes[i].geometry);
                if (!(geometryUid in geometries)) {
                  geometries[geometryUid] = true;
                  vertexSegments[getUid(segment)] = true;
                }
              } else {
                break;
              }
            }
          }

          this.vertexSegments_ = vertexSegments;
          return;
        }
      }
      if (this.vertexFeature_) {
        this.overlay_.getSource().removeFeature(this.vertexFeature_);
        this.vertexFeature_ = null;
      }
    }

    /**
     * @param {SegmentData} segmentData Segment data.
     * @param {import("../coordinate.js").Coordinate} vertex Vertex.
     * @private
     */
    insertVertex_(segmentData, vertex) {
      const segment = segmentData.segment;
      const feature = segmentData.feature;
      const geometry = segmentData.geometry;
      const depth = segmentData.depth;
      const index = segmentData.index;
      let coordinates;

      while (vertex.length < geometry.getStride()) {
        vertex.push(0);
      }

      switch (geometry.getType()) {
        case 'MultiLineString':
          coordinates = geometry.getCoordinates();
          coordinates[depth[0]].splice(index + 1, 0, vertex);
          break;
        case 'Polygon':
          coordinates = geometry.getCoordinates();
          coordinates[depth[0]].splice(index + 1, 0, vertex);
          break;
        case 'MultiPolygon':
          coordinates = geometry.getCoordinates();
          coordinates[depth[1]][depth[0]].splice(index + 1, 0, vertex);
          break;
        case 'LineString':
          coordinates = geometry.getCoordinates();
          coordinates.splice(index + 1, 0, vertex);
          break;
        default:
          return;
      }

      this.setGeometryCoordinates_(geometry, coordinates);
      const rTree = this.rBush_;
      rTree.remove(segmentData);
      this.updateSegmentIndices_(geometry, index, depth, 1);

      /** @type {SegmentData} */
      const newSegmentData = {
        segment: [segment[0], vertex],
        feature: feature,
        geometry: geometry,
        depth: depth,
        index: index,
      };

      rTree.insert(boundingExtent(newSegmentData.segment), newSegmentData);
      this.dragSegments_.push([newSegmentData, 1]);

      /** @type {SegmentData} */
      const newSegmentData2 = {
        segment: [vertex, segment[1]],
        feature: feature,
        geometry: geometry,
        depth: depth,
        index: index + 1,
      };

      rTree.insert(boundingExtent(newSegmentData2.segment), newSegmentData2);
      this.dragSegments_.push([newSegmentData2, 0]);
      this.ignoreNextSingleClick_ = true;
    }

    /**
     * Removes the vertex currently being pointed.
     * @return {boolean} True when a vertex was removed.
     * @api
     */
    removePoint() {
      if (
        this.lastPointerEvent_ &&
        this.lastPointerEvent_.type != MapBrowserEventType.POINTERDRAG
      ) {
        const evt = this.lastPointerEvent_;
        this.willModifyFeatures_(evt, this.dragSegments_);
        const removed = this.removeVertex_();
        if (this.featuresBeingModified_) {
          this.dispatchEvent(
            new ModifyEvent(
              ModifyEventType.MODIFYEND,
              this.featuresBeingModified_,
              evt
            )
          );
        }

        this.featuresBeingModified_ = null;
        return removed;
      }
      return false;
    }

    /**
     * Removes a vertex from all matching features.
     * @return {boolean} True when a vertex was removed.
     * @private
     */
    removeVertex_() {
      const dragSegments = this.dragSegments_;
      const segmentsByFeature = {};
      let deleted = false;
      let component, coordinates, dragSegment, geometry, i, index, left;
      let newIndex, right, segmentData, uid;
      for (i = dragSegments.length - 1; i >= 0; --i) {
        dragSegment = dragSegments[i];
        segmentData = dragSegment[0];
        uid = getUid(segmentData.feature);
        if (segmentData.depth) {
          // separate feature components
          uid += '-' + segmentData.depth.join('-');
        }
        if (!(uid in segmentsByFeature)) {
          segmentsByFeature[uid] = {};
        }
        if (dragSegment[1] === 0) {
          segmentsByFeature[uid].right = segmentData;
          segmentsByFeature[uid].index = segmentData.index;
        } else if (dragSegment[1] == 1) {
          segmentsByFeature[uid].left = segmentData;
          segmentsByFeature[uid].index = segmentData.index + 1;
        }
      }
      for (uid in segmentsByFeature) {
        right = segmentsByFeature[uid].right;
        left = segmentsByFeature[uid].left;
        index = segmentsByFeature[uid].index;
        newIndex = index - 1;
        if (left !== undefined) {
          segmentData = left;
        } else {
          segmentData = right;
        }
        if (newIndex < 0) {
          newIndex = 0;
        }
        geometry = segmentData.geometry;
        coordinates = geometry.getCoordinates();
        component = coordinates;
        deleted = false;
        switch (geometry.getType()) {
          case 'MultiLineString':
            if (coordinates[segmentData.depth[0]].length > 2) {
              coordinates[segmentData.depth[0]].splice(index, 1);
              deleted = true;
            }
            break;
          case 'LineString':
            if (coordinates.length > 2) {
              coordinates.splice(index, 1);
              deleted = true;
            }
            break;
          case 'MultiPolygon':
            component = component[segmentData.depth[1]];
          /* falls through */
          case 'Polygon':
            component = component[segmentData.depth[0]];
            if (component.length > 4) {
              if (index == component.length - 1) {
                index = 0;
              }
              component.splice(index, 1);
              deleted = true;
              if (index === 0) {
                // close the ring again
                component.pop();
                component.push(component[0]);
                newIndex = component.length - 1;
              }
            }
            break;
          // pass
        }

        if (deleted) {
          this.setGeometryCoordinates_(geometry, coordinates);
          const segments = [];
          if (left !== undefined) {
            this.rBush_.remove(left);
            segments.push(left.segment[0]);
          }
          if (right !== undefined) {
            this.rBush_.remove(right);
            segments.push(right.segment[1]);
          }
          if (left !== undefined && right !== undefined) {
            /** @type {SegmentData} */
            const newSegmentData = {
              depth: segmentData.depth,
              feature: segmentData.feature,
              geometry: segmentData.geometry,
              index: newIndex,
              segment: segments,
            };

            this.rBush_.insert(
              boundingExtent(newSegmentData.segment),
              newSegmentData
            );
          }
          this.updateSegmentIndices_(geometry, index, segmentData.depth, -1);
          if (this.vertexFeature_) {
            this.overlay_.getSource().removeFeature(this.vertexFeature_);
            this.vertexFeature_ = null;
          }
          dragSegments.length = 0;
        }
      }
      return deleted;
    }

    /**
     * @param {import("../geom/SimpleGeometry.js").default} geometry Geometry.
     * @param {Array} coordinates Coordinates.
     * @private
     */
    setGeometryCoordinates_(geometry, coordinates) {
      this.changingFeature_ = true;
      geometry.setCoordinates(coordinates);
      this.changingFeature_ = false;
    }

    /**
     * @param {import("../geom/SimpleGeometry.js").default} geometry Geometry.
     * @param {number} index Index.
     * @param {Array<number>|undefined} depth Depth.
     * @param {number} delta Delta (1 or -1).
     * @private
     */
    updateSegmentIndices_(geometry, index, depth, delta) {
      this.rBush_.forEachInExtent(
        geometry.getExtent(),
        function (segmentDataMatch) {
          if (
            segmentDataMatch.geometry === geometry &&
            (depth === undefined ||
              segmentDataMatch.depth === undefined ||
              equals$3(segmentDataMatch.depth, depth)) &&
            segmentDataMatch.index > index
          ) {
            segmentDataMatch.index += delta;
          }
        }
      );
    }
  }

  /**
   * @param {SegmentData} a The first segment data.
   * @param {SegmentData} b The second segment data.
   * @return {number} The difference in indexes.
   */
  function compareIndexes(a, b) {
    return a.index - b.index;
  }

  /**
   * Returns the distance from a point to a line segment.
   *
   * @param {import("../coordinate.js").Coordinate} pointCoordinates The coordinates of the point from
   *        which to calculate the distance.
   * @param {SegmentData} segmentData The object describing the line
   *        segment we are calculating the distance to.
   * @param {import("../proj/Projection.js").default} projection The view projection.
   * @return {number} The square of the distance between a point and a line segment.
   */
  function projectedDistanceToSegmentDataSquared(
    pointCoordinates,
    segmentData,
    projection
  ) {
    const geometry = segmentData.geometry;

    if (geometry.getType() === 'Circle') {
      let circleGeometry = /** @type {import("../geom/Circle.js").default} */ (
        geometry
      );

      if (segmentData.index === CIRCLE_CIRCUMFERENCE_INDEX) {
        const userProjection = getUserProjection();
        if (userProjection) {
          circleGeometry = /** @type {import("../geom/Circle.js").default} */ (
            circleGeometry.clone().transform(userProjection, projection)
          );
        }
        const distanceToCenterSquared = squaredDistance(
          circleGeometry.getCenter(),
          fromUserCoordinate(pointCoordinates, projection)
        );
        const distanceToCircumference =
          Math.sqrt(distanceToCenterSquared) - circleGeometry.getRadius();
        return distanceToCircumference * distanceToCircumference;
      }
    }

    const coordinate = fromUserCoordinate(pointCoordinates, projection);
    tempSegment$1[0] = fromUserCoordinate(segmentData.segment[0], projection);
    tempSegment$1[1] = fromUserCoordinate(segmentData.segment[1], projection);
    return squaredDistanceToSegment(coordinate, tempSegment$1);
  }

  /**
   * Returns the point closest to a given line segment.
   *
   * @param {import("../coordinate.js").Coordinate} pointCoordinates The point to which a closest point
   *        should be found.
   * @param {SegmentData} segmentData The object describing the line
   *        segment which should contain the closest point.
   * @param {import("../proj/Projection.js").default} projection The view projection.
   * @return {import("../coordinate.js").Coordinate} The point closest to the specified line segment.
   */
  function closestOnSegmentData(pointCoordinates, segmentData, projection) {
    const geometry = segmentData.geometry;

    if (
      geometry.getType() === 'Circle' &&
      segmentData.index === CIRCLE_CIRCUMFERENCE_INDEX
    ) {
      let circleGeometry = /** @type {import("../geom/Circle.js").default} */ (
        geometry
      );
      const userProjection = getUserProjection();
      if (userProjection) {
        circleGeometry = /** @type {import("../geom/Circle.js").default} */ (
          circleGeometry.clone().transform(userProjection, projection)
        );
      }
      return toUserCoordinate(
        circleGeometry.getClosestPoint(
          fromUserCoordinate(pointCoordinates, projection)
        ),
        projection
      );
    }
    const coordinate = fromUserCoordinate(pointCoordinates, projection);
    tempSegment$1[0] = fromUserCoordinate(segmentData.segment[0], projection);
    tempSegment$1[1] = fromUserCoordinate(segmentData.segment[1], projection);
    return toUserCoordinate(
      closestOnSegment(coordinate, tempSegment$1),
      projection
    );
  }

  /**
   * @return {import("../style/Style.js").StyleFunction} Styles.
   */
  function getDefaultStyleFunction$1() {
    const style = createEditingStyle();
    return function (feature, resolution) {
      return style['Point'];
    };
  }

  /**
   * @module ol/interaction/Select
   */

  /**
   * @enum {string}
   */
  const SelectEventType = {
    /**
     * Triggered when feature(s) has been (de)selected.
     * @event SelectEvent#select
     * @api
     */
    SELECT: 'select',
  };

  /**
   * A function that takes an {@link module:ol/Feature~Feature} and returns `true` if the feature may be
   * selected or `false` otherwise.
   * @typedef {function(import("../Feature.js").default, import("../layer/Layer.js").default<import("../source/Source").default>):boolean} FilterFunction
   */

  /**
   * @typedef {Object} Options
   * @property {import("../events/condition.js").Condition} [addCondition] A function
   * that takes an {@link module:ol/MapBrowserEvent~MapBrowserEvent} and returns a
   * boolean to indicate whether that event should be handled.
   * By default, this is {@link module:ol/events/condition.never}. Use this if you
   * want to use different events for add and remove instead of `toggle`.
   * @property {import("../events/condition.js").Condition} [condition] A function that
   * takes an {@link module:ol/MapBrowserEvent~MapBrowserEvent} and returns a
   * boolean to indicate whether that event should be handled. This is the event
   * for the selected features as a whole. By default, this is
   * {@link module:ol/events/condition.singleClick}. Clicking on a feature selects that
   * feature and removes any that were in the selection. Clicking outside any
   * feature removes all from the selection.
   * See `toggle`, `add`, `remove` options for adding/removing extra features to/
   * from the selection.
   * @property {Array<import("../layer/Layer.js").default>|function(import("../layer/Layer.js").default<import("../source/Source").default>): boolean} [layers]
   * A list of layers from which features should be selected. Alternatively, a
   * filter function can be provided. The function will be called for each layer
   * in the map and should return `true` for layers that you want to be
   * selectable. If the option is absent, all visible layers will be considered
   * selectable.
   * @property {import("../style/Style.js").StyleLike|null} [style]
   * Style for the selected features. By default the default edit style is used
   * (see {@link module:ol/style/Style~Style}). Set to `null` if this interaction should not apply
   * any style changes for selected features.
   * If set to a falsey value, the selected feature's style will not change.
   * @property {import("../events/condition.js").Condition} [removeCondition] A function
   * that takes an {@link module:ol/MapBrowserEvent~MapBrowserEvent} and returns a
   * boolean to indicate whether that event should be handled.
   * By default, this is {@link module:ol/events/condition.never}. Use this if you
   * want to use different events for add and remove instead of `toggle`.
   * @property {import("../events/condition.js").Condition} [toggleCondition] A function
   * that takes an {@link module:ol/MapBrowserEvent~MapBrowserEvent} and returns a
   * boolean to indicate whether that event should be handled. This is in addition
   * to the `condition` event. By default,
   * {@link module:ol/events/condition.shiftKeyOnly}, i.e. pressing `shift` as
   * well as the `condition` event, adds that feature to the current selection if
   * it is not currently selected, and removes it if it is. See `add` and `remove`
   * if you want to use different events instead of a toggle.
   * @property {boolean} [multi=false] A boolean that determines if the default
   * behaviour should select only single features or all (overlapping) features at
   * the clicked map position. The default of `false` means single select.
   * @property {Collection<Feature>} [features]
   * Collection where the interaction will place selected features. Optional. If
   * not set the interaction will create a collection. In any case the collection
   * used by the interaction is returned by
   * {@link module:ol/interaction/Select~Select#getFeatures}.
   * @property {FilterFunction} [filter] A function
   * that takes an {@link module:ol/Feature~Feature} and an
   * {@link module:ol/layer/Layer~Layer} and returns `true` if the feature may be
   * selected or `false` otherwise.
   * @property {number} [hitTolerance=0] Hit-detection tolerance. Pixels inside
   * the radius around the given position will be checked for features.
   */

  /**
   * @classdesc
   * Events emitted by {@link module:ol/interaction/Select~Select} instances are instances of
   * this type.
   */
  class SelectEvent extends BaseEvent {
    /**
     * @param {SelectEventType} type The event type.
     * @param {Array<import("../Feature.js").default>} selected Selected features.
     * @param {Array<import("../Feature.js").default>} deselected Deselected features.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Associated
     *     {@link module:ol/MapBrowserEvent~MapBrowserEvent}.
     */
    constructor(type, selected, deselected, mapBrowserEvent) {
      super(type);

      /**
       * Selected features array.
       * @type {Array<import("../Feature.js").default>}
       * @api
       */
      this.selected = selected;

      /**
       * Deselected features array.
       * @type {Array<import("../Feature.js").default>}
       * @api
       */
      this.deselected = deselected;

      /**
       * Associated {@link module:ol/MapBrowserEvent~MapBrowserEvent}.
       * @type {import("../MapBrowserEvent.js").default}
       * @api
       */
      this.mapBrowserEvent = mapBrowserEvent;
    }
  }

  /**
   * Original feature styles to reset to when features are no longer selected.
   * @type {Object<number, import("../style/Style.js").default|Array<import("../style/Style.js").default>|import("../style/Style.js").StyleFunction>}
   */
  const originalFeatureStyles = {};

  /***
   * @template Return
   * @typedef {import("../Observable").OnSignature<import("../Observable").EventTypes, import("../events/Event.js").default, Return> &
   *   import("../Observable").OnSignature<import("../ObjectEventType").Types|
   *     'change:active', import("../Object").ObjectEvent, Return> &
   *   import("../Observable").OnSignature<'select', SelectEvent, Return> &
   *   import("../Observable").CombinedOnSignature<import("../Observable").EventTypes|import("../ObjectEventType").Types|
   *     'change:active'|'select', Return>} SelectOnSignature
   */

  /**
   * @classdesc
   * Interaction for selecting vector features. By default, selected features are
   * styled differently, so this interaction can be used for visual highlighting,
   * as well as selecting features for other actions, such as modification or
   * output. There are three ways of controlling which features are selected:
   * using the browser event as defined by the `condition` and optionally the
   * `toggle`, `add`/`remove`, and `multi` options; a `layers` filter; and a
   * further feature filter using the `filter` option.
   *
   * @fires SelectEvent
   * @api
   */
  class Select extends Interaction {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      super();

      /***
       * @type {SelectOnSignature<import("../events").EventsKey>}
       */
      this.on;

      /***
       * @type {SelectOnSignature<import("../events").EventsKey>}
       */
      this.once;

      /***
       * @type {SelectOnSignature<void>}
       */
      this.un;

      options = options ? options : {};

      /**
       * @private
       */
      this.boundAddFeature_ = this.addFeature_.bind(this);

      /**
       * @private
       */
      this.boundRemoveFeature_ = this.removeFeature_.bind(this);

      /**
       * @private
       * @type {import("../events/condition.js").Condition}
       */
      this.condition_ = options.condition ? options.condition : singleClick;

      /**
       * @private
       * @type {import("../events/condition.js").Condition}
       */
      this.addCondition_ = options.addCondition ? options.addCondition : never;

      /**
       * @private
       * @type {import("../events/condition.js").Condition}
       */
      this.removeCondition_ = options.removeCondition
        ? options.removeCondition
        : never;

      /**
       * @private
       * @type {import("../events/condition.js").Condition}
       */
      this.toggleCondition_ = options.toggleCondition
        ? options.toggleCondition
        : shiftKeyOnly;

      /**
       * @private
       * @type {boolean}
       */
      this.multi_ = options.multi ? options.multi : false;

      /**
       * @private
       * @type {FilterFunction}
       */
      this.filter_ = options.filter ? options.filter : TRUE;

      /**
       * @private
       * @type {number}
       */
      this.hitTolerance_ = options.hitTolerance ? options.hitTolerance : 0;

      /**
       * @private
       * @type {import("../style/Style.js").default|Array<import("../style/Style.js").default>|import("../style/Style.js").StyleFunction|null}
       */
      this.style_ =
        options.style !== undefined ? options.style : getDefaultStyleFunction();

      /**
       * @private
       * @type {Collection<Feature>}
       */
      this.features_ = options.features || new Collection();

      /** @type {function(import("../layer/Layer.js").default<import("../source/Source").default>): boolean} */
      let layerFilter;
      if (options.layers) {
        if (typeof options.layers === 'function') {
          layerFilter = options.layers;
        } else {
          const layers = options.layers;
          layerFilter = function (layer) {
            return layers.includes(layer);
          };
        }
      } else {
        layerFilter = TRUE;
      }

      /**
       * @private
       * @type {function(import("../layer/Layer.js").default<import("../source/Source").default>): boolean}
       */
      this.layerFilter_ = layerFilter;

      /**
       * An association between selected feature (key)
       * and layer (value)
       * @private
       * @type {Object<string, import("../layer/Layer.js").default>}
       */
      this.featureLayerAssociation_ = {};
    }

    /**
     * @param {import("../Feature.js").default} feature Feature.
     * @param {import("../layer/Layer.js").default} layer Layer.
     * @private
     */
    addFeatureLayerAssociation_(feature, layer) {
      this.featureLayerAssociation_[getUid(feature)] = layer;
    }

    /**
     * Get the selected features.
     * @return {Collection<Feature>} Features collection.
     * @api
     */
    getFeatures() {
      return this.features_;
    }

    /**
     * Returns the Hit-detection tolerance.
     * @return {number} Hit tolerance in pixels.
     * @api
     */
    getHitTolerance() {
      return this.hitTolerance_;
    }

    /**
     * Returns the associated {@link module:ol/layer/Vector~VectorLayer vector layer} of
     * a selected feature.
     * @param {import("../Feature.js").default} feature Feature
     * @return {import('../layer/Vector.js').default} Layer.
     * @api
     */
    getLayer(feature) {
      return /** @type {import('../layer/Vector.js').default} */ (
        this.featureLayerAssociation_[getUid(feature)]
      );
    }

    /**
     * Hit-detection tolerance. Pixels inside the radius around the given position
     * will be checked for features.
     * @param {number} hitTolerance Hit tolerance in pixels.
     * @api
     */
    setHitTolerance(hitTolerance) {
      this.hitTolerance_ = hitTolerance;
    }

    /**
     * Remove the interaction from its current map, if any,  and attach it to a new
     * map, if any. Pass `null` to just remove the interaction from the current map.
     * @param {import("../Map.js").default|null} map Map.
     * @api
     */
    setMap(map) {
      const currentMap = this.getMap();
      if (currentMap && this.style_) {
        this.features_.forEach(this.restorePreviousStyle_.bind(this));
      }
      super.setMap(map);
      if (map) {
        this.features_.addEventListener(
          CollectionEventType.ADD,
          this.boundAddFeature_
        );
        this.features_.addEventListener(
          CollectionEventType.REMOVE,
          this.boundRemoveFeature_
        );

        if (this.style_) {
          this.features_.forEach(this.applySelectedStyle_.bind(this));
        }
      } else {
        this.features_.removeEventListener(
          CollectionEventType.ADD,
          this.boundAddFeature_
        );
        this.features_.removeEventListener(
          CollectionEventType.REMOVE,
          this.boundRemoveFeature_
        );
      }
    }

    /**
     * @param {import("../Collection.js").CollectionEvent<Feature>} evt Event.
     * @private
     */
    addFeature_(evt) {
      const feature = evt.element;
      if (this.style_) {
        this.applySelectedStyle_(feature);
      }
      if (!this.getLayer(feature)) {
        const layer = /** @type {VectorLayer} */ (
          this.getMap()
            .getAllLayers()
            .find(function (layer) {
              if (
                layer instanceof VectorLayer &&
                layer.getSource() &&
                layer.getSource().hasFeature(feature)
              ) {
                return layer;
              }
            })
        );
        if (layer) {
          this.addFeatureLayerAssociation_(feature, layer);
        }
      }
    }

    /**
     * @param {import("../Collection.js").CollectionEvent<Feature>} evt Event.
     * @private
     */
    removeFeature_(evt) {
      if (this.style_) {
        this.restorePreviousStyle_(evt.element);
      }
    }

    /**
     * @return {import("../style/Style.js").StyleLike|null} Select style.
     */
    getStyle() {
      return this.style_;
    }

    /**
     * @param {Feature} feature Feature
     * @private
     */
    applySelectedStyle_(feature) {
      const key = getUid(feature);
      if (!(key in originalFeatureStyles)) {
        originalFeatureStyles[key] = feature.getStyle();
      }
      feature.setStyle(this.style_);
    }

    /**
     * @param {Feature} feature Feature
     * @private
     */
    restorePreviousStyle_(feature) {
      const interactions = this.getMap().getInteractions().getArray();
      for (let i = interactions.length - 1; i >= 0; --i) {
        const interaction = interactions[i];
        if (
          interaction !== this &&
          interaction instanceof Select &&
          interaction.getStyle() &&
          interaction.getFeatures().getArray().lastIndexOf(feature) !== -1
        ) {
          feature.setStyle(interaction.getStyle());
          return;
        }
      }

      const key = getUid(feature);
      feature.setStyle(originalFeatureStyles[key]);
      delete originalFeatureStyles[key];
    }

    /**
     * @param {Feature} feature Feature.
     * @private
     */
    removeFeatureLayerAssociation_(feature) {
      delete this.featureLayerAssociation_[getUid(feature)];
    }

    /**
     * Handles the {@link module:ol/MapBrowserEvent~MapBrowserEvent map browser event} and may change the
     * selected state of features.
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Map browser event.
     * @return {boolean} `false` to stop event propagation.
     * @this {Select}
     */
    handleEvent(mapBrowserEvent) {
      if (!this.condition_(mapBrowserEvent)) {
        return true;
      }
      const add = this.addCondition_(mapBrowserEvent);
      const remove = this.removeCondition_(mapBrowserEvent);
      const toggle = this.toggleCondition_(mapBrowserEvent);
      const set = !add && !remove && !toggle;
      const map = mapBrowserEvent.map;
      const features = this.getFeatures();

      /**
       * @type {Array<Feature>}
       */
      const deselected = [];

      /**
       * @type {Array<Feature>}
       */
      const selected = [];

      if (set) {
        // Replace the currently selected feature(s) with the feature(s) at the
        // pixel, or clear the selected feature(s) if there is no feature at
        // the pixel.
        clear$2(this.featureLayerAssociation_);
        map.forEachFeatureAtPixel(
          mapBrowserEvent.pixel,
          /**
           * @param {import("../Feature.js").FeatureLike} feature Feature.
           * @param {import("../layer/Layer.js").default} layer Layer.
           * @return {boolean|undefined} Continue to iterate over the features.
           */
          function (feature, layer) {
            if (!(feature instanceof Feature) || !this.filter_(feature, layer)) {
              return;
            }
            this.addFeatureLayerAssociation_(feature, layer);
            selected.push(feature);
            return !this.multi_;
          }.bind(this),
          {
            layerFilter: this.layerFilter_,
            hitTolerance: this.hitTolerance_,
          }
        );
        for (let i = features.getLength() - 1; i >= 0; --i) {
          const feature = features.item(i);
          const index = selected.indexOf(feature);
          if (index > -1) {
            // feature is already selected
            selected.splice(index, 1);
          } else {
            features.remove(feature);
            deselected.push(feature);
          }
        }
        if (selected.length !== 0) {
          features.extend(selected);
        }
      } else {
        // Modify the currently selected feature(s).
        map.forEachFeatureAtPixel(
          mapBrowserEvent.pixel,
          /**
           * @param {import("../Feature.js").FeatureLike} feature Feature.
           * @param {import("../layer/Layer.js").default} layer Layer.
           * @return {boolean|undefined} Continue to iterate over the features.
           */
          function (feature, layer) {
            if (!(feature instanceof Feature) || !this.filter_(feature, layer)) {
              return;
            }
            if ((add || toggle) && !features.getArray().includes(feature)) {
              this.addFeatureLayerAssociation_(feature, layer);
              selected.push(feature);
            } else if (
              (remove || toggle) &&
              features.getArray().includes(feature)
            ) {
              deselected.push(feature);
              this.removeFeatureLayerAssociation_(feature);
            }
            return !this.multi_;
          }.bind(this),
          {
            layerFilter: this.layerFilter_,
            hitTolerance: this.hitTolerance_,
          }
        );
        for (let j = deselected.length - 1; j >= 0; --j) {
          features.remove(deselected[j]);
        }
        features.extend(selected);
      }
      if (selected.length > 0 || deselected.length > 0) {
        this.dispatchEvent(
          new SelectEvent(
            SelectEventType.SELECT,
            selected,
            deselected,
            mapBrowserEvent
          )
        );
      }
      return true;
    }
  }

  /**
   * @return {import("../style/Style.js").StyleFunction} Styles.
   */
  function getDefaultStyleFunction() {
    const styles = createEditingStyle();
    extend$2(styles['Polygon'], styles['LineString']);
    extend$2(styles['GeometryCollection'], styles['LineString']);

    return function (feature) {
      if (!feature.getGeometry()) {
        return null;
      }
      return styles[feature.getGeometry().getType()];
    };
  }

  /**
   * @module ol/interaction/Snap
   */

  /**
   * @typedef {Object} Result
   * @property {import("../coordinate.js").Coordinate|null} vertex Vertex.
   * @property {import("../pixel.js").Pixel|null} vertexPixel VertexPixel.
   */

  /**
   * @typedef {Object} SegmentData
   * @property {import("../Feature.js").default} feature Feature.
   * @property {Array<import("../coordinate.js").Coordinate>} segment Segment.
   */

  /**
   * @typedef {Object} Options
   * @property {import("../Collection.js").default<import("../Feature.js").default>} [features] Snap to these features. Either this option or source should be provided.
   * @property {boolean} [edge=true] Snap to edges.
   * @property {boolean} [vertex=true] Snap to vertices.
   * @property {number} [pixelTolerance=10] Pixel tolerance for considering the pointer close enough to a segment or
   * vertex for snapping.
   * @property {import("../source/Vector.js").default} [source] Snap to features from this source. Either this option or features should be provided
   */

  /**
   * @param  {import("../source/Vector.js").VectorSourceEvent|import("../Collection.js").CollectionEvent<import("../Feature.js").default>} evt Event.
   * @return {import("../Feature.js").default} Feature.
   */
  function getFeatureFromEvent(evt) {
    if (
      /** @type {import("../source/Vector.js").VectorSourceEvent} */ (evt).feature
    ) {
      return /** @type {import("../source/Vector.js").VectorSourceEvent} */ (evt)
        .feature;
    } else if (
      /** @type {import("../Collection.js").CollectionEvent<import("../Feature.js").default>} */ (
        evt
      ).element
    ) {
      return /** @type {import("../Collection.js").CollectionEvent<import("../Feature.js").default>} */ (
        evt
      ).element;
    }
  }

  const tempSegment = [];

  /**
   * @classdesc
   * Handles snapping of vector features while modifying or drawing them.  The
   * features can come from a {@link module:ol/source/Vector~VectorSource} or {@link module:ol/Collection~Collection}
   * Any interaction object that allows the user to interact
   * with the features using the mouse can benefit from the snapping, as long
   * as it is added before.
   *
   * The snap interaction modifies map browser event `coordinate` and `pixel`
   * properties to force the snap to occur to any interaction that them.
   *
   * Example:
   *
   *     import Snap from 'ol/interaction/Snap';
   *
   *     const snap = new Snap({
   *       source: source
   *     });
   *
   *     map.addInteraction(snap);
   *
   * @api
   */
  class Snap extends PointerInteraction {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      options = options ? options : {};

      const pointerOptions = /** @type {import("./Pointer.js").Options} */ (
        options
      );

      if (!pointerOptions.handleDownEvent) {
        pointerOptions.handleDownEvent = TRUE;
      }

      if (!pointerOptions.stopDown) {
        pointerOptions.stopDown = FALSE;
      }

      super(pointerOptions);

      /**
       * @type {import("../source/Vector.js").default|null}
       * @private
       */
      this.source_ = options.source ? options.source : null;

      /**
       * @private
       * @type {boolean}
       */
      this.vertex_ = options.vertex !== undefined ? options.vertex : true;

      /**
       * @private
       * @type {boolean}
       */
      this.edge_ = options.edge !== undefined ? options.edge : true;

      /**
       * @type {import("../Collection.js").default<import("../Feature.js").default>|null}
       * @private
       */
      this.features_ = options.features ? options.features : null;

      /**
       * @type {Array<import("../events.js").EventsKey>}
       * @private
       */
      this.featuresListenerKeys_ = [];

      /**
       * @type {Object<string, import("../events.js").EventsKey>}
       * @private
       */
      this.featureChangeListenerKeys_ = {};

      /**
       * Extents are preserved so indexed segment can be quickly removed
       * when its feature geometry changes
       * @type {Object<string, import("../extent.js").Extent>}
       * @private
       */
      this.indexedFeaturesExtents_ = {};

      /**
       * If a feature geometry changes while a pointer drag|move event occurs, the
       * feature doesn't get updated right away.  It will be at the next 'pointerup'
       * event fired.
       * @type {!Object<string, import("../Feature.js").default>}
       * @private
       */
      this.pendingFeatures_ = {};

      /**
       * @type {number}
       * @private
       */
      this.pixelTolerance_ =
        options.pixelTolerance !== undefined ? options.pixelTolerance : 10;

      /**
       * Segment RTree for each layer
       * @type {import("../structs/RBush.js").default<SegmentData>}
       * @private
       */
      this.rBush_ = new RBush();

      /**
       * @const
       * @private
       * @type {Object<string, function(Array<Array<import('../coordinate.js').Coordinate>>, import("../geom/Geometry.js").default): void>}
       */
      this.GEOMETRY_SEGMENTERS_ = {
        'Point': this.segmentPointGeometry_.bind(this),
        'LineString': this.segmentLineStringGeometry_.bind(this),
        'LinearRing': this.segmentLineStringGeometry_.bind(this),
        'Polygon': this.segmentPolygonGeometry_.bind(this),
        'MultiPoint': this.segmentMultiPointGeometry_.bind(this),
        'MultiLineString': this.segmentMultiLineStringGeometry_.bind(this),
        'MultiPolygon': this.segmentMultiPolygonGeometry_.bind(this),
        'GeometryCollection': this.segmentGeometryCollectionGeometry_.bind(this),
        'Circle': this.segmentCircleGeometry_.bind(this),
      };
    }

    /**
     * Add a feature to the collection of features that we may snap to.
     * @param {import("../Feature.js").default} feature Feature.
     * @param {boolean} [register] Whether to listen to the feature change or not
     *     Defaults to `true`.
     * @api
     */
    addFeature(feature, register) {
      register = register !== undefined ? register : true;
      const feature_uid = getUid(feature);
      const geometry = feature.getGeometry();
      if (geometry) {
        const segmenter = this.GEOMETRY_SEGMENTERS_[geometry.getType()];
        if (segmenter) {
          this.indexedFeaturesExtents_[feature_uid] = geometry.getExtent(
            createEmpty()
          );
          const segments =
            /** @type {Array<Array<import('../coordinate.js').Coordinate>>} */ ([]);
          segmenter(segments, geometry);
          if (segments.length === 1) {
            this.rBush_.insert(boundingExtent(segments[0]), {
              feature: feature,
              segment: segments[0],
            });
          } else if (segments.length > 1) {
            const extents = segments.map((s) => boundingExtent(s));
            const segmentsData = segments.map((segment) => ({
              feature: feature,
              segment: segment,
            }));
            this.rBush_.load(extents, segmentsData);
          }
        }
      }

      if (register) {
        this.featureChangeListenerKeys_[feature_uid] = listen(
          feature,
          EventType$1.CHANGE,
          this.handleFeatureChange_,
          this
        );
      }
    }

    /**
     * @param {import("../Feature.js").default} feature Feature.
     * @private
     */
    forEachFeatureAdd_(feature) {
      this.addFeature(feature);
    }

    /**
     * @param {import("../Feature.js").default} feature Feature.
     * @private
     */
    forEachFeatureRemove_(feature) {
      this.removeFeature(feature);
    }

    /**
     * @return {import("../Collection.js").default<import("../Feature.js").default>|Array<import("../Feature.js").default>} Features.
     * @private
     */
    getFeatures_() {
      /** @type {import("../Collection.js").default<import("../Feature.js").default>|Array<import("../Feature.js").default>} */
      let features;
      if (this.features_) {
        features = this.features_;
      } else if (this.source_) {
        features = this.source_.getFeatures();
      }
      return features;
    }

    /**
     * @param {import("../MapBrowserEvent.js").default} evt Map browser event.
     * @return {boolean} `false` to stop event propagation.
     */
    handleEvent(evt) {
      const result = this.snapTo(evt.pixel, evt.coordinate, evt.map);
      if (result) {
        evt.coordinate = result.vertex.slice(0, 2);
        evt.pixel = result.vertexPixel;
      }
      return super.handleEvent(evt);
    }

    /**
     * @param {import("../source/Vector.js").VectorSourceEvent|import("../Collection.js").CollectionEvent<import("../Feature.js").default>} evt Event.
     * @private
     */
    handleFeatureAdd_(evt) {
      const feature = getFeatureFromEvent(evt);
      this.addFeature(feature);
    }

    /**
     * @param {import("../source/Vector.js").VectorSourceEvent|import("../Collection.js").CollectionEvent<import("../Feature.js").default>} evt Event.
     * @private
     */
    handleFeatureRemove_(evt) {
      const feature = getFeatureFromEvent(evt);
      this.removeFeature(feature);
    }

    /**
     * @param {import("../events/Event.js").default} evt Event.
     * @private
     */
    handleFeatureChange_(evt) {
      const feature = /** @type {import("../Feature.js").default} */ (evt.target);
      if (this.handlingDownUpSequence) {
        const uid = getUid(feature);
        if (!(uid in this.pendingFeatures_)) {
          this.pendingFeatures_[uid] = feature;
        }
      } else {
        this.updateFeature_(feature);
      }
    }

    /**
     * Handle pointer up events.
     * @param {import("../MapBrowserEvent.js").default} evt Event.
     * @return {boolean} If the event was consumed.
     */
    handleUpEvent(evt) {
      const featuresToUpdate = Object.values(this.pendingFeatures_);
      if (featuresToUpdate.length) {
        featuresToUpdate.forEach(this.updateFeature_.bind(this));
        this.pendingFeatures_ = {};
      }
      return false;
    }

    /**
     * Remove a feature from the collection of features that we may snap to.
     * @param {import("../Feature.js").default} feature Feature
     * @param {boolean} [unlisten] Whether to unlisten to the feature change
     *     or not. Defaults to `true`.
     * @api
     */
    removeFeature(feature, unlisten) {
      const unregister = unlisten !== undefined ? unlisten : true;
      const feature_uid = getUid(feature);
      const extent = this.indexedFeaturesExtents_[feature_uid];
      if (extent) {
        const rBush = this.rBush_;
        const nodesToRemove = [];
        rBush.forEachInExtent(extent, function (node) {
          if (feature === node.feature) {
            nodesToRemove.push(node);
          }
        });
        for (let i = nodesToRemove.length - 1; i >= 0; --i) {
          rBush.remove(nodesToRemove[i]);
        }
      }

      if (unregister) {
        unlistenByKey(this.featureChangeListenerKeys_[feature_uid]);
        delete this.featureChangeListenerKeys_[feature_uid];
      }
    }

    /**
     * Remove the interaction from its current map and attach it to the new map.
     * Subclasses may set up event handlers to get notified about changes to
     * the map here.
     * @param {import("../Map.js").default} map Map.
     */
    setMap(map) {
      const currentMap = this.getMap();
      const keys = this.featuresListenerKeys_;
      const features = /** @type {Array<import("../Feature.js").default>} */ (
        this.getFeatures_()
      );

      if (currentMap) {
        keys.forEach(unlistenByKey);
        keys.length = 0;
        features.forEach(this.forEachFeatureRemove_.bind(this));
      }
      super.setMap(map);

      if (map) {
        if (this.features_) {
          keys.push(
            listen(
              this.features_,
              CollectionEventType.ADD,
              this.handleFeatureAdd_,
              this
            ),
            listen(
              this.features_,
              CollectionEventType.REMOVE,
              this.handleFeatureRemove_,
              this
            )
          );
        } else if (this.source_) {
          keys.push(
            listen(
              this.source_,
              VectorEventType.ADDFEATURE,
              this.handleFeatureAdd_,
              this
            ),
            listen(
              this.source_,
              VectorEventType.REMOVEFEATURE,
              this.handleFeatureRemove_,
              this
            )
          );
        }
        features.forEach(this.forEachFeatureAdd_.bind(this));
      }
    }

    /**
     * @param {import("../pixel.js").Pixel} pixel Pixel
     * @param {import("../coordinate.js").Coordinate} pixelCoordinate Coordinate
     * @param {import("../Map.js").default} map Map.
     * @return {Result|null} Snap result
     */
    snapTo(pixel, pixelCoordinate, map) {
      const lowerLeft = map.getCoordinateFromPixel([
        pixel[0] - this.pixelTolerance_,
        pixel[1] + this.pixelTolerance_,
      ]);
      const upperRight = map.getCoordinateFromPixel([
        pixel[0] + this.pixelTolerance_,
        pixel[1] - this.pixelTolerance_,
      ]);
      const box = boundingExtent([lowerLeft, upperRight]);

      const segments = this.rBush_.getInExtent(box);

      const segmentsLength = segments.length;
      if (segmentsLength === 0) {
        return null;
      }

      const projection = map.getView().getProjection();
      const projectedCoordinate = fromUserCoordinate(pixelCoordinate, projection);

      let closestVertex;
      let minSquaredDistance = Infinity;

      const squaredPixelTolerance = this.pixelTolerance_ * this.pixelTolerance_;
      const getResult = () => {
        if (closestVertex) {
          const vertexPixel = map.getPixelFromCoordinate(closestVertex);
          const squaredPixelDistance = squaredDistance(pixel, vertexPixel);
          if (squaredPixelDistance <= squaredPixelTolerance) {
            return {
              vertex: closestVertex,
              vertexPixel: [
                Math.round(vertexPixel[0]),
                Math.round(vertexPixel[1]),
              ],
            };
          }
        }
        return null;
      };

      if (this.vertex_) {
        for (let i = 0; i < segmentsLength; ++i) {
          const segmentData = segments[i];
          if (segmentData.feature.getGeometry().getType() !== 'Circle') {
            segmentData.segment.forEach((vertex) => {
              const tempVertexCoord = fromUserCoordinate(vertex, projection);
              const delta = squaredDistance(projectedCoordinate, tempVertexCoord);
              if (delta < minSquaredDistance) {
                closestVertex = vertex;
                minSquaredDistance = delta;
              }
            });
          }
        }
        const result = getResult();
        if (result) {
          return result;
        }
      }

      if (this.edge_) {
        for (let i = 0; i < segmentsLength; ++i) {
          let vertex = null;
          const segmentData = segments[i];
          if (segmentData.feature.getGeometry().getType() === 'Circle') {
            let circleGeometry = segmentData.feature.getGeometry();
            const userProjection = getUserProjection();
            if (userProjection) {
              circleGeometry = circleGeometry
                .clone()
                .transform(userProjection, projection);
            }
            vertex = toUserCoordinate(
              closestOnCircle(
                projectedCoordinate,
                /** @type {import("../geom/Circle.js").default} */ (
                  circleGeometry
                )
              ),
              projection
            );
          } else {
            const [segmentStart, segmentEnd] = segmentData.segment;
            // points have only one coordinate
            if (segmentEnd) {
              tempSegment[0] = fromUserCoordinate(segmentStart, projection);
              tempSegment[1] = fromUserCoordinate(segmentEnd, projection);
              vertex = closestOnSegment(projectedCoordinate, tempSegment);
            }
          }
          if (vertex) {
            const delta = squaredDistance(projectedCoordinate, vertex);
            if (delta < minSquaredDistance) {
              closestVertex = vertex;
              minSquaredDistance = delta;
            }
          }
        }

        const result = getResult();
        if (result) {
          return result;
        }
      }

      return null;
    }

    /**
     * @param {import("../Feature.js").default} feature Feature
     * @private
     */
    updateFeature_(feature) {
      this.removeFeature(feature, false);
      this.addFeature(feature, false);
    }

    /**
     * @param {Array<Array<import('../coordinate.js').Coordinate>>} segments Segments
     * @param {import("../geom/Circle.js").default} geometry Geometry.
     * @private
     */
    segmentCircleGeometry_(segments, geometry) {
      const projection = this.getMap().getView().getProjection();
      let circleGeometry = geometry;
      const userProjection = getUserProjection();
      if (userProjection) {
        circleGeometry = /** @type {import("../geom/Circle.js").default} */ (
          circleGeometry.clone().transform(userProjection, projection)
        );
      }
      const polygon = fromCircle(circleGeometry);
      if (userProjection) {
        polygon.transform(projection, userProjection);
      }
      const coordinates = polygon.getCoordinates()[0];
      for (let i = 0, ii = coordinates.length - 1; i < ii; ++i) {
        segments.push(coordinates.slice(i, i + 2));
      }
    }

    /**
     * @param {Array<Array<import('../coordinate.js').Coordinate>>} segments Segments
     * @param {import("../geom/GeometryCollection.js").default} geometry Geometry.
     * @private
     */
    segmentGeometryCollectionGeometry_(segments, geometry) {
      const geometries = geometry.getGeometriesArray();
      for (let i = 0; i < geometries.length; ++i) {
        const segmenter = this.GEOMETRY_SEGMENTERS_[geometries[i].getType()];
        if (segmenter) {
          segmenter(segments, geometries[i]);
        }
      }
    }

    /**
     * @param {Array<Array<import('../coordinate.js').Coordinate>>} segments Segments
     * @param {import("../geom/LineString.js").default} geometry Geometry.
     * @private
     */
    segmentLineStringGeometry_(segments, geometry) {
      const coordinates = geometry.getCoordinates();
      for (let i = 0, ii = coordinates.length - 1; i < ii; ++i) {
        segments.push(coordinates.slice(i, i + 2));
      }
    }

    /**
     * @param {Array<Array<import('../coordinate.js').Coordinate>>} segments Segments
     * @param {import("../geom/MultiLineString.js").default} geometry Geometry.
     * @private
     */
    segmentMultiLineStringGeometry_(segments, geometry) {
      const lines = geometry.getCoordinates();
      for (let j = 0, jj = lines.length; j < jj; ++j) {
        const coordinates = lines[j];
        for (let i = 0, ii = coordinates.length - 1; i < ii; ++i) {
          segments.push(coordinates.slice(i, i + 2));
        }
      }
    }

    /**
     * @param {Array<Array<import('../coordinate.js').Coordinate>>} segments Segments
     * @param {import("../geom/MultiPoint.js").default} geometry Geometry.
     * @private
     */
    segmentMultiPointGeometry_(segments, geometry) {
      geometry.getCoordinates().forEach((point) => {
        segments.push([point]);
      });
    }

    /**
     * @param {Array<Array<import('../coordinate.js').Coordinate>>} segments Segments
     * @param {import("../geom/MultiPolygon.js").default} geometry Geometry.
     * @private
     */
    segmentMultiPolygonGeometry_(segments, geometry) {
      const polygons = geometry.getCoordinates();
      for (let k = 0, kk = polygons.length; k < kk; ++k) {
        const rings = polygons[k];
        for (let j = 0, jj = rings.length; j < jj; ++j) {
          const coordinates = rings[j];
          for (let i = 0, ii = coordinates.length - 1; i < ii; ++i) {
            segments.push(coordinates.slice(i, i + 2));
          }
        }
      }
    }

    /**
     * @param {Array<Array<import('../coordinate.js').Coordinate>>} segments Segments
     * @param {import("../geom/Point.js").default} geometry Geometry.
     * @private
     */
    segmentPointGeometry_(segments, geometry) {
      segments.push([geometry.getCoordinates()]);
    }

    /**
     * @param {Array<Array<import('../coordinate.js').Coordinate>>} segments Segments
     * @param {import("../geom/Polygon.js").default} geometry Geometry.
     * @private
     */
    segmentPolygonGeometry_(segments, geometry) {
      const rings = geometry.getCoordinates();
      for (let j = 0, jj = rings.length; j < jj; ++j) {
        const coordinates = rings[j];
        for (let i = 0, ii = coordinates.length - 1; i < ii; ++i) {
          segments.push(coordinates.slice(i, i + 2));
        }
      }
    }
  }

  /**
   * @module ol/interaction/Translate
   */

  /**
   * @enum {string}
   */
  const TranslateEventType = {
    /**
     * Triggered upon feature translation start.
     * @event TranslateEvent#translatestart
     * @api
     */
    TRANSLATESTART: 'translatestart',
    /**
     * Triggered upon feature translation.
     * @event TranslateEvent#translating
     * @api
     */
    TRANSLATING: 'translating',
    /**
     * Triggered upon feature translation end.
     * @event TranslateEvent#translateend
     * @api
     */
    TRANSLATEEND: 'translateend',
  };

  /**
   * A function that takes an {@link module:ol/Feature~Feature} or
   * {@link module:ol/render/Feature~RenderFeature} and an
   * {@link module:ol/layer/Layer~Layer} and returns `true` if the feature may be
   * translated or `false` otherwise.
   * @typedef {function(import("../Feature.js").FeatureLike, import("../layer/Layer.js").default<import("../source/Source").default>):boolean} FilterFunction
   */

  /**
   * @typedef {Object} Options
   * @property {import("../events/condition.js").Condition} [condition] A function that
   * takes an {@link module:ol/MapBrowserEvent~MapBrowserEvent} and returns a
   * boolean to indicate whether that event should be handled.
   * Default is {@link module:ol/events/condition.always}.
   * @property {Collection<import("../Feature.js").default>} [features] Features contained in this collection will be able to be translated together.
   * @property {Array<import("../layer/Layer.js").default>|function(import("../layer/Layer.js").default<import("../source/Source").default>): boolean} [layers] A list of layers from which features should be
   * translated. Alternatively, a filter function can be provided. The
   * function will be called for each layer in the map and should return
   * `true` for layers that you want to be translatable. If the option is
   * absent, all visible layers will be considered translatable.
   * Not used if `features` is provided.
   * @property {FilterFunction} [filter] A function
   * that takes an {@link module:ol/Feature~Feature} and an
   * {@link module:ol/layer/Layer~Layer} and returns `true` if the feature may be
   * translated or `false` otherwise. Not used if `features` is provided.
   * @property {number} [hitTolerance=0] Hit-detection tolerance. Pixels inside the radius around the given position
   * will be checked for features.
   */

  /**
   * @classdesc
   * Events emitted by {@link module:ol/interaction/Translate~Translate} instances
   * are instances of this type.
   */
  class TranslateEvent extends BaseEvent {
    /**
     * @param {TranslateEventType} type Type.
     * @param {Collection<import("../Feature.js").default>} features The features translated.
     * @param {import("../coordinate.js").Coordinate} coordinate The event coordinate.
     * @param {import("../coordinate.js").Coordinate} startCoordinate The original coordinates before.translation started
     * @param {import("../MapBrowserEvent.js").default} mapBrowserEvent Map browser event.
     */
    constructor(type, features, coordinate, startCoordinate, mapBrowserEvent) {
      super(type);

      /**
       * The features being translated.
       * @type {Collection<import("../Feature.js").default>}
       * @api
       */
      this.features = features;

      /**
       * The coordinate of the drag event.
       * @const
       * @type {import("../coordinate.js").Coordinate}
       * @api
       */
      this.coordinate = coordinate;

      /**
       * The coordinate of the start position before translation started.
       * @const
       * @type {import("../coordinate.js").Coordinate}
       * @api
       */
      this.startCoordinate = startCoordinate;

      /**
       * Associated {@link module:ol/MapBrowserEvent~MapBrowserEvent}.
       * @type {import("../MapBrowserEvent.js").default}
       * @api
       */
      this.mapBrowserEvent = mapBrowserEvent;
    }
  }

  /***
   * @template Return
   * @typedef {import("../Observable").OnSignature<import("../Observable").EventTypes, import("../events/Event.js").default, Return> &
   *   import("../Observable").OnSignature<import("../ObjectEventType").Types|
   *     'change:active', import("../Object").ObjectEvent, Return> &
   *   import("../Observable").OnSignature<'translateend'|'translatestart'|'translating', TranslateEvent, Return> &
   *   import("../Observable").CombinedOnSignature<import("../Observable").EventTypes|import("../ObjectEventType").Types|
   *     'change:active'|'translateend'|'translatestart'|'translating', Return>} TranslateOnSignature
   */

  /**
   * @classdesc
   * Interaction for translating (moving) features.
   * If you want to translate multiple features in a single action (for example,
   * the collection used by a select interaction), construct the interaction with
   * the `features` option.
   *
   * @fires TranslateEvent
   * @api
   */
  class Translate extends PointerInteraction {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      options = options ? options : {};

      super(/** @type {import("./Pointer.js").Options} */ (options));

      /***
       * @type {TranslateOnSignature<import("../events").EventsKey>}
       */
      this.on;

      /***
       * @type {TranslateOnSignature<import("../events").EventsKey>}
       */
      this.once;

      /***
       * @type {TranslateOnSignature<void>}
       */
      this.un;

      /**
       * The last position we translated to.
       * @type {import("../coordinate.js").Coordinate}
       * @private
       */
      this.lastCoordinate_ = null;

      /**
       * The start position before translation started.
       * @type {import("../coordinate.js").Coordinate}
       * @private
       */
      this.startCoordinate_ = null;

      /**
       * @type {Collection<import("../Feature.js").default>|null}
       * @private
       */
      this.features_ = options.features !== undefined ? options.features : null;

      /** @type {function(import("../layer/Layer.js").default<import("../source/Source").default>): boolean} */
      let layerFilter;
      if (options.layers && !this.features_) {
        if (typeof options.layers === 'function') {
          layerFilter = options.layers;
        } else {
          const layers = options.layers;
          layerFilter = function (layer) {
            return layers.includes(layer);
          };
        }
      } else {
        layerFilter = TRUE;
      }

      /**
       * @private
       * @type {function(import("../layer/Layer.js").default<import("../source/Source").default>): boolean}
       */
      this.layerFilter_ = layerFilter;

      /**
       * @private
       * @type {FilterFunction}
       */
      this.filter_ = options.filter && !this.features_ ? options.filter : TRUE;

      /**
       * @private
       * @type {number}
       */
      this.hitTolerance_ = options.hitTolerance ? options.hitTolerance : 0;

      /**
       * @private
       * @type {import("../events/condition.js").Condition}
       */
      this.condition_ = options.condition ? options.condition : always;

      /**
       * @type {import("../Feature.js").default}
       * @private
       */
      this.lastFeature_ = null;

      this.addChangeListener(
        InteractionProperty.ACTIVE,
        this.handleActiveChanged_
      );
    }

    /**
     * Handle pointer down events.
     * @param {import("../MapBrowserEvent.js").default} event Event.
     * @return {boolean} If the event was consumed.
     */
    handleDownEvent(event) {
      if (!event.originalEvent || !this.condition_(event)) {
        return false;
      }
      this.lastFeature_ = this.featuresAtPixel_(event.pixel, event.map);
      if (!this.lastCoordinate_ && this.lastFeature_) {
        this.startCoordinate_ = event.coordinate;
        this.lastCoordinate_ = event.coordinate;
        this.handleMoveEvent(event);

        const features = this.features_ || new Collection([this.lastFeature_]);

        this.dispatchEvent(
          new TranslateEvent(
            TranslateEventType.TRANSLATESTART,
            features,
            event.coordinate,
            this.startCoordinate_,
            event
          )
        );
        return true;
      }
      return false;
    }

    /**
     * Handle pointer up events.
     * @param {import("../MapBrowserEvent.js").default} event Event.
     * @return {boolean} If the event was consumed.
     */
    handleUpEvent(event) {
      if (this.lastCoordinate_) {
        this.lastCoordinate_ = null;
        this.handleMoveEvent(event);

        const features = this.features_ || new Collection([this.lastFeature_]);

        this.dispatchEvent(
          new TranslateEvent(
            TranslateEventType.TRANSLATEEND,
            features,
            event.coordinate,
            this.startCoordinate_,
            event
          )
        );
        // cleanup
        this.startCoordinate_ = null;
        return true;
      }
      return false;
    }

    /**
     * Handle pointer drag events.
     * @param {import("../MapBrowserEvent.js").default} event Event.
     */
    handleDragEvent(event) {
      if (this.lastCoordinate_) {
        const newCoordinate = event.coordinate;
        const deltaX = newCoordinate[0] - this.lastCoordinate_[0];
        const deltaY = newCoordinate[1] - this.lastCoordinate_[1];

        const features = this.features_ || new Collection([this.lastFeature_]);

        features.forEach(function (feature) {
          const geom = feature.getGeometry();
          geom.translate(deltaX, deltaY);
          feature.setGeometry(geom);
        });

        this.lastCoordinate_ = newCoordinate;

        this.dispatchEvent(
          new TranslateEvent(
            TranslateEventType.TRANSLATING,
            features,
            newCoordinate,
            this.startCoordinate_,
            event
          )
        );
      }
    }

    /**
     * Handle pointer move events.
     * @param {import("../MapBrowserEvent.js").default} event Event.
     */
    handleMoveEvent(event) {
      const elem = event.map.getViewport();

      // Change the cursor to grab/grabbing if hovering any of the features managed
      // by the interaction
      if (this.featuresAtPixel_(event.pixel, event.map)) {
        elem.classList.remove(this.lastCoordinate_ ? 'ol-grab' : 'ol-grabbing');
        elem.classList.add(this.lastCoordinate_ ? 'ol-grabbing' : 'ol-grab');
      } else {
        elem.classList.remove('ol-grab', 'ol-grabbing');
      }
    }

    /**
     * Tests to see if the given coordinates intersects any of our selected
     * features.
     * @param {import("../pixel.js").Pixel} pixel Pixel coordinate to test for intersection.
     * @param {import("../Map.js").default} map Map to test the intersection on.
     * @return {import("../Feature.js").default} Returns the feature found at the specified pixel
     * coordinates.
     * @private
     */
    featuresAtPixel_(pixel, map) {
      return map.forEachFeatureAtPixel(
        pixel,
        function (feature, layer) {
          if (this.filter_(feature, layer)) {
            if (!this.features_ || this.features_.getArray().includes(feature)) {
              return feature;
            }
          }
        }.bind(this),
        {
          layerFilter: this.layerFilter_,
          hitTolerance: this.hitTolerance_,
        }
      );
    }

    /**
     * Returns the Hit-detection tolerance.
     * @return {number} Hit tolerance in pixels.
     * @api
     */
    getHitTolerance() {
      return this.hitTolerance_;
    }

    /**
     * Hit-detection tolerance. Pixels inside the radius around the given position
     * will be checked for features.
     * @param {number} hitTolerance Hit tolerance in pixels.
     * @api
     */
    setHitTolerance(hitTolerance) {
      this.hitTolerance_ = hitTolerance;
    }

    /**
     * Remove the interaction from its current map and attach it to the new map.
     * Subclasses may set up event handlers to get notified about changes to
     * the map here.
     * @param {import("../Map.js").default} map Map.
     */
    setMap(map) {
      const oldMap = this.getMap();
      super.setMap(map);
      this.updateState_(oldMap);
    }

    /**
     * @private
     */
    handleActiveChanged_() {
      this.updateState_(null);
    }

    /**
     * @param {import("../Map.js").default} oldMap Old map.
     * @private
     */
    updateState_(oldMap) {
      let map = this.getMap();
      const active = this.getActive();
      if (!map || !active) {
        map = map || oldMap;
        if (map) {
          const elem = map.getViewport();
          elem.classList.remove('ol-grab', 'ol-grabbing');
        }
      }
    }
  }

  /**
   * @module ol/geom/flat/flip
   */

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   * @param {Array<number>} [dest] Destination.
   * @param {number} [destOffset] Destination offset.
   * @return {Array<number>} Flat coordinates.
   */
  function flipXY(flatCoordinates, offset, end, stride, dest, destOffset) {
    if (dest !== undefined) {
      dest = dest;
      destOffset = destOffset !== undefined ? destOffset : 0;
    } else {
      dest = [];
      destOffset = 0;
    }
    let j = offset;
    while (j < end) {
      const x = flatCoordinates[j++];
      dest[destOffset++] = flatCoordinates[j++];
      dest[destOffset++] = x;
      for (let k = 2; k < stride; ++k) {
        dest[destOffset++] = flatCoordinates[j++];
      }
    }
    dest.length = destOffset;
    return dest;
  }

  /**
   * @module ol/geom/flat/topology
   */

  /**
   * Check if the linestring is a boundary.
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {number} offset Offset.
   * @param {number} end End.
   * @param {number} stride Stride.
   * @return {boolean} The linestring is a boundary.
   */
  function lineStringIsClosed(flatCoordinates, offset, end, stride) {
    const lastCoord = end - stride;
    if (
      flatCoordinates[offset] === flatCoordinates[lastCoord] &&
      flatCoordinates[offset + 1] === flatCoordinates[lastCoord + 1] &&
      (end - offset) / stride > 3
    ) {
      return !!linearRing(flatCoordinates, offset, end, stride);
    }
    return false;
  }

  /**
   * @module ol/format/EsriJSON
   */

  /**
   * @typedef {import("arcgis-rest-api").Feature} EsriJSONFeature
   * @typedef {import("arcgis-rest-api").FeatureSet} EsriJSONFeatureSet
   * @typedef {import("arcgis-rest-api").Geometry} EsriJSONGeometry
   * @typedef {import("arcgis-rest-api").Point} EsriJSONPoint
   * @typedef {import("arcgis-rest-api").Polyline} EsriJSONPolyline
   * @typedef {import("arcgis-rest-api").Polygon} EsriJSONPolygon
   * @typedef {import("arcgis-rest-api").Multipoint} EsriJSONMultipoint
   * @typedef {import("arcgis-rest-api").HasZM} EsriJSONHasZM
   * @typedef {import("arcgis-rest-api").Position} EsriJSONPosition
   * @typedef {import("arcgis-rest-api").SpatialReferenceWkid} EsriJSONSpatialReferenceWkid
   */

  /**
   * @typedef {Object} EsriJSONMultiPolygon
   * @property {Array<Array<Array<Array<number>>>>} rings Rings for the MultiPolygon.
   * @property {boolean} [hasM] If the polygon coordinates have an M value.
   * @property {boolean} [hasZ] If the polygon coordinates have a Z value.
   * @property {EsriJSONSpatialReferenceWkid} [spatialReference] The coordinate reference system.
   */

  /**
   * @const
   * @type {Object<import("../geom/Geometry.js").Type, function(EsriJSONGeometry): import("../geom/Geometry.js").default>}
   */
  const GEOMETRY_READERS$1 = {
    Point: readPointGeometry$1,
    LineString: readLineStringGeometry$1,
    Polygon: readPolygonGeometry$1,
    MultiPoint: readMultiPointGeometry$1,
    MultiLineString: readMultiLineStringGeometry$1,
    MultiPolygon: readMultiPolygonGeometry$1,
  };

  /**
   * @const
   * @type {Object<import("../geom/Geometry.js").Type, function(import("../geom/Geometry.js").default, import("./Feature.js").WriteOptions=): (EsriJSONGeometry)>}
   */
  const GEOMETRY_WRITERS = {
    Point: writePointGeometry,
    LineString: writeLineStringGeometry,
    Polygon: writePolygonGeometry,
    MultiPoint: writeMultiPointGeometry,
    MultiLineString: writeMultiLineStringGeometry,
    MultiPolygon: writeMultiPolygonGeometry,
  };

  /**
   * @typedef {Object} Options
   * @property {string} [geometryName] Geometry name to use when creating features.
   */

  /**
   * @classdesc
   * Feature format for reading and writing data in the EsriJSON format.
   *
   * @api
   */
  class EsriJSON extends JSONFeature {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      options = options ? options : {};

      super();

      /**
       * Name of the geometry attribute for features.
       * @type {string|undefined}
       * @private
       */
      this.geometryName_ = options.geometryName;
    }

    /**
     * @param {Object} object Object.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @param {string} [idField] Name of the field where to get the id from.
     * @protected
     * @return {import("../Feature.js").default} Feature.
     */
    readFeatureFromObject(object, options, idField) {
      const esriJSONFeature = /** @type {EsriJSONFeature} */ (object);
      const geometry = readGeometry(esriJSONFeature.geometry, options);
      const feature = new Feature();
      if (this.geometryName_) {
        feature.setGeometryName(this.geometryName_);
      }
      feature.setGeometry(geometry);
      if (esriJSONFeature.attributes) {
        feature.setProperties(esriJSONFeature.attributes, true);
        const id = esriJSONFeature.attributes[idField];
        if (id !== undefined) {
          feature.setId(/** @type {number} */ (id));
        }
      }
      return feature;
    }

    /**
     * @param {Object} object Object.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @protected
     * @return {Array<Feature>} Features.
     */
    readFeaturesFromObject(object, options) {
      options = options ? options : {};
      if (object['features']) {
        const esriJSONFeatureSet = /** @type {EsriJSONFeatureSet} */ (object);
        /** @type {Array<import("../Feature.js").default>} */
        const features = [];
        const esriJSONFeatures = esriJSONFeatureSet.features;
        for (let i = 0, ii = esriJSONFeatures.length; i < ii; ++i) {
          features.push(
            this.readFeatureFromObject(
              esriJSONFeatures[i],
              options,
              object.objectIdFieldName
            )
          );
        }
        return features;
      } else {
        return [this.readFeatureFromObject(object, options)];
      }
    }

    /**
     * @param {EsriJSONGeometry} object Object.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @protected
     * @return {import("../geom/Geometry.js").default} Geometry.
     */
    readGeometryFromObject(object, options) {
      return readGeometry(object, options);
    }

    /**
     * @param {Object} object Object.
     * @protected
     * @return {import("../proj/Projection.js").default} Projection.
     */
    readProjectionFromObject(object) {
      if (
        object['spatialReference'] &&
        object['spatialReference']['wkid'] !== undefined
      ) {
        const spatialReference = /** @type {EsriJSONSpatialReferenceWkid} */ (
          object['spatialReference']
        );
        const crs = spatialReference.wkid;
        return get$2('EPSG:' + crs);
      } else {
        return null;
      }
    }

    /**
     * Encode a geometry as a EsriJSON object.
     *
     * @param {import("../geom/Geometry.js").default} geometry Geometry.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @return {EsriJSONGeometry} Object.
     * @api
     */
    writeGeometryObject(geometry, options) {
      return writeGeometry(geometry, this.adaptOptions(options));
    }

    /**
     * Encode a feature as a esriJSON Feature object.
     *
     * @param {import("../Feature.js").default} feature Feature.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @return {Object} Object.
     * @api
     */
    writeFeatureObject(feature, options) {
      options = this.adaptOptions(options);
      const object = {};
      if (!feature.hasProperties()) {
        object['attributes'] = {};
        return object;
      }
      const properties = feature.getProperties();
      const geometry = feature.getGeometry();
      if (geometry) {
        object['geometry'] = writeGeometry(geometry, options);
        const projection =
          options && (options.dataProjection || options.featureProjection);
        if (projection) {
          object['geometry']['spatialReference'] =
            /** @type {EsriJSONSpatialReferenceWkid} */ ({
              wkid: Number(get$2(projection).getCode().split(':').pop()),
            });
        }
        delete properties[feature.getGeometryName()];
      }
      if (!isEmpty$1(properties)) {
        object['attributes'] = properties;
      } else {
        object['attributes'] = {};
      }
      return object;
    }

    /**
     * Encode an array of features as a EsriJSON object.
     *
     * @param {Array<import("../Feature.js").default>} features Features.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @return {EsriJSONFeatureSet} EsriJSON Object.
     * @api
     */
    writeFeaturesObject(features, options) {
      options = this.adaptOptions(options);
      const objects = [];
      for (let i = 0, ii = features.length; i < ii; ++i) {
        objects.push(this.writeFeatureObject(features[i], options));
      }
      return {
        'features': objects,
      };
    }
  }

  /**
   * @param {EsriJSONGeometry} object Object.
   * @param {import("./Feature.js").ReadOptions} [options] Read options.
   * @return {import("../geom/Geometry.js").default} Geometry.
   */
  function readGeometry(object, options) {
    if (!object) {
      return null;
    }
    /** @type {import("../geom/Geometry.js").Type} */
    let type;
    if (typeof object['x'] === 'number' && typeof object['y'] === 'number') {
      type = 'Point';
    } else if (object['points']) {
      type = 'MultiPoint';
    } else if (object['paths']) {
      const esriJSONPolyline = /** @type {EsriJSONPolyline} */ (object);
      if (esriJSONPolyline.paths.length === 1) {
        type = 'LineString';
      } else {
        type = 'MultiLineString';
      }
    } else if (object['rings']) {
      const esriJSONPolygon = /** @type {EsriJSONPolygon} */ (object);
      const layout = getGeometryLayout(esriJSONPolygon);
      const rings = convertRings(esriJSONPolygon.rings, layout);
      if (rings.length === 1) {
        type = 'Polygon';
        object = Object.assign({}, object, {['rings']: rings[0]});
      } else {
        type = 'MultiPolygon';
        object = Object.assign({}, object, {['rings']: rings});
      }
    }
    const geometryReader = GEOMETRY_READERS$1[type];
    return transformGeometryWithOptions(geometryReader(object), false, options);
  }

  /**
   * Determines inner and outer rings.
   * Checks if any polygons in this array contain any other polygons in this
   * array. It is used for checking for holes.
   * Logic inspired by: https://github.com/Esri/terraformer-arcgis-parser
   * @param {Array<!Array<!Array<number>>>} rings Rings.
   * @param {import("../geom/Geometry.js").GeometryLayout} layout Geometry layout.
   * @return {Array<!Array<!Array<!Array<number>>>>} Transformed rings.
   */
  function convertRings(rings, layout) {
    const flatRing = [];
    const outerRings = [];
    const holes = [];
    let i, ii;
    for (i = 0, ii = rings.length; i < ii; ++i) {
      flatRing.length = 0;
      deflateCoordinates(flatRing, 0, rings[i], layout.length);
      // is this ring an outer ring? is it clockwise?
      const clockwise = linearRingIsClockwise(
        flatRing,
        0,
        flatRing.length,
        layout.length
      );
      if (clockwise) {
        outerRings.push([rings[i]]);
      } else {
        holes.push(rings[i]);
      }
    }
    while (holes.length) {
      const hole = holes.shift();
      let matched = false;
      // loop over all outer rings and see if they contain our hole.
      for (i = outerRings.length - 1; i >= 0; i--) {
        const outerRing = outerRings[i][0];
        const containsHole = containsExtent(
          new LinearRing(outerRing).getExtent(),
          new LinearRing(hole).getExtent()
        );
        if (containsHole) {
          // the hole is contained push it into our polygon
          outerRings[i].push(hole);
          matched = true;
          break;
        }
      }
      if (!matched) {
        // no outer rings contain this hole turn it into and outer
        // ring (reverse it)
        outerRings.push([hole.reverse()]);
      }
    }
    return outerRings;
  }

  /**
   * @param {EsriJSONPoint} object Object.
   * @return {import("../geom/Geometry.js").default} Point.
   */
  function readPointGeometry$1(object) {
    let point;
    if (object.m !== undefined && object.z !== undefined) {
      point = new Point([object.x, object.y, object.z, object.m], 'XYZM');
    } else if (object.z !== undefined) {
      point = new Point([object.x, object.y, object.z], 'XYZ');
    } else if (object.m !== undefined) {
      point = new Point([object.x, object.y, object.m], 'XYM');
    } else {
      point = new Point([object.x, object.y]);
    }
    return point;
  }

  /**
   * @param {EsriJSONPolyline} object Object.
   * @return {import("../geom/Geometry.js").default} LineString.
   */
  function readLineStringGeometry$1(object) {
    const layout = getGeometryLayout(object);
    return new LineString(object.paths[0], layout);
  }

  /**
   * @param {EsriJSONPolyline} object Object.
   * @return {import("../geom/Geometry.js").default} MultiLineString.
   */
  function readMultiLineStringGeometry$1(object) {
    const layout = getGeometryLayout(object);
    return new MultiLineString(object.paths, layout);
  }

  /**
   * @param {EsriJSONHasZM} object Object.
   * @return {import("../geom/Geometry.js").GeometryLayout} The geometry layout to use.
   */
  function getGeometryLayout(object) {
    /** @type {import("../geom/Geometry.js").GeometryLayout} */
    let layout = 'XY';
    if (object.hasZ === true && object.hasM === true) {
      layout = 'XYZM';
    } else if (object.hasZ === true) {
      layout = 'XYZ';
    } else if (object.hasM === true) {
      layout = 'XYM';
    }
    return layout;
  }

  /**
   * @param {EsriJSONMultipoint} object Object.
   * @return {import("../geom/Geometry.js").default} MultiPoint.
   */
  function readMultiPointGeometry$1(object) {
    const layout = getGeometryLayout(object);
    return new MultiPoint(object.points, layout);
  }

  /**
   * @param {EsriJSONMultiPolygon} object Object.
   * @return {import("../geom/Geometry.js").default} MultiPolygon.
   */
  function readMultiPolygonGeometry$1(object) {
    const layout = getGeometryLayout(object);
    return new MultiPolygon(object.rings, layout);
  }

  /**
   * @param {EsriJSONPolygon} object Object.
   * @return {import("../geom/Geometry.js").default} Polygon.
   */
  function readPolygonGeometry$1(object) {
    const layout = getGeometryLayout(object);
    return new Polygon(object.rings, layout);
  }

  /**
   * @param {import("../geom/Point.js").default} geometry Geometry.
   * @param {import("./Feature.js").WriteOptions} [options] Write options.
   * @return {EsriJSONPoint} EsriJSON geometry.
   */
  function writePointGeometry(geometry, options) {
    const coordinates = geometry.getCoordinates();
    /** @type {EsriJSONPoint} */
    let esriJSON;
    const layout = geometry.getLayout();
    if (layout === 'XYZ') {
      esriJSON = {
        x: coordinates[0],
        y: coordinates[1],
        z: coordinates[2],
      };
    } else if (layout === 'XYM') {
      esriJSON = {
        x: coordinates[0],
        y: coordinates[1],
        m: coordinates[2],
      };
    } else if (layout === 'XYZM') {
      esriJSON = {
        x: coordinates[0],
        y: coordinates[1],
        z: coordinates[2],
        m: coordinates[3],
      };
    } else if (layout === 'XY') {
      esriJSON = {
        x: coordinates[0],
        y: coordinates[1],
      };
    } else {
      assert(false, 34); // Invalid geometry layout
    }
    return esriJSON;
  }

  /**
   * @param {import("../geom/SimpleGeometry.js").default} geometry Geometry.
   * @return {Object} Object with boolean hasZ and hasM keys.
   */
  function getHasZM(geometry) {
    const layout = geometry.getLayout();
    return {
      hasZ: layout === 'XYZ' || layout === 'XYZM',
      hasM: layout === 'XYM' || layout === 'XYZM',
    };
  }

  /**
   * @param {import("../geom/LineString.js").default} lineString Geometry.
   * @param {import("./Feature.js").WriteOptions} [options] Write options.
   * @return {EsriJSONPolyline} EsriJSON geometry.
   */
  function writeLineStringGeometry(lineString, options) {
    const hasZM = getHasZM(lineString);
    return {
      hasZ: hasZM.hasZ,
      hasM: hasZM.hasM,
      paths: [
        /** @type {Array<EsriJSONPosition>} */ (lineString.getCoordinates()),
      ],
    };
  }

  /**
   * @param {import("../geom/Polygon.js").default} polygon Geometry.
   * @param {import("./Feature.js").WriteOptions} [options] Write options.
   * @return {EsriJSONPolygon} EsriJSON geometry.
   */
  function writePolygonGeometry(polygon, options) {
    // Esri geometries use the left-hand rule
    const hasZM = getHasZM(polygon);
    return {
      hasZ: hasZM.hasZ,
      hasM: hasZM.hasM,
      rings: /** @type {Array<Array<EsriJSONPosition>>} */ (
        polygon.getCoordinates(false)
      ),
    };
  }

  /**
   * @param {import("../geom/MultiLineString.js").default} multiLineString Geometry.
   * @param {import("./Feature.js").WriteOptions} [options] Write options.
   * @return {EsriJSONPolyline} EsriJSON geometry.
   */
  function writeMultiLineStringGeometry(multiLineString, options) {
    const hasZM = getHasZM(multiLineString);
    return {
      hasZ: hasZM.hasZ,
      hasM: hasZM.hasM,
      paths: /** @type {Array<Array<EsriJSONPosition>>} */ (
        multiLineString.getCoordinates()
      ),
    };
  }

  /**
   * @param {import("../geom/MultiPoint.js").default} multiPoint Geometry.
   * @param {import("./Feature.js").WriteOptions} [options] Write options.
   * @return {EsriJSONMultipoint} EsriJSON geometry.
   */
  function writeMultiPointGeometry(multiPoint, options) {
    const hasZM = getHasZM(multiPoint);
    return {
      hasZ: hasZM.hasZ,
      hasM: hasZM.hasM,
      points: /** @type {Array<EsriJSONPosition>} */ (
        multiPoint.getCoordinates()
      ),
    };
  }

  /**
   * @param {import("../geom/MultiPolygon.js").default} geometry Geometry.
   * @param {import("./Feature.js").WriteOptions} [options] Write options.
   * @return {EsriJSONPolygon} EsriJSON geometry.
   */
  function writeMultiPolygonGeometry(geometry, options) {
    const hasZM = getHasZM(geometry);
    const coordinates = geometry.getCoordinates(false);
    const output = [];
    for (let i = 0; i < coordinates.length; i++) {
      for (let x = coordinates[i].length - 1; x >= 0; x--) {
        output.push(coordinates[i][x]);
      }
    }
    return {
      hasZ: hasZM.hasZ,
      hasM: hasZM.hasM,
      rings: /** @type {Array<Array<EsriJSONPosition>>} */ (output),
    };
  }

  /**
   * @param {import("../geom/Geometry.js").default} geometry Geometry.
   * @param {import("./Feature.js").WriteOptions} [options] Write options.
   * @return {EsriJSONGeometry} EsriJSON geometry.
   */
  function writeGeometry(geometry, options) {
    const geometryWriter = GEOMETRY_WRITERS[geometry.getType()];
    return geometryWriter(
      transformGeometryWithOptions(geometry, true, options),
      options
    );
  }

  /**
   * @module ol/format/XMLFeature
   */

  /**
   * @classdesc
   * Abstract base class; normally only used for creating subclasses and not
   * instantiated in apps.
   * Base class for XML feature formats.
   *
   * @abstract
   */
  class XMLFeature extends FeatureFormat {
    constructor() {
      super();

      /**
       * @type {XMLSerializer}
       * @private
       */
      this.xmlSerializer_ = getXMLSerializer();
    }

    /**
     * @return {import("./Feature.js").Type} Format.
     */
    getType() {
      return 'xml';
    }

    /**
     * Read a single feature.
     *
     * @param {Document|Element|Object|string} source Source.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @return {import("../Feature.js").default} Feature.
     * @api
     */
    readFeature(source, options) {
      if (!source) {
        return null;
      } else if (typeof source === 'string') {
        const doc = parse(source);
        return this.readFeatureFromDocument(doc, options);
      } else if (isDocument(source)) {
        return this.readFeatureFromDocument(
          /** @type {Document} */ (source),
          options
        );
      } else {
        return this.readFeatureFromNode(/** @type {Element} */ (source), options);
      }
    }

    /**
     * @param {Document} doc Document.
     * @param {import("./Feature.js").ReadOptions} [options] Options.
     * @return {import("../Feature.js").default} Feature.
     */
    readFeatureFromDocument(doc, options) {
      const features = this.readFeaturesFromDocument(doc, options);
      if (features.length > 0) {
        return features[0];
      } else {
        return null;
      }
    }

    /**
     * @param {Element} node Node.
     * @param {import("./Feature.js").ReadOptions} [options] Options.
     * @return {import("../Feature.js").default} Feature.
     */
    readFeatureFromNode(node, options) {
      return null; // not implemented
    }

    /**
     * Read all features from a feature collection.
     *
     * @param {Document|Element|Object|string} source Source.
     * @param {import("./Feature.js").ReadOptions} [options] Options.
     * @return {Array<import("../Feature.js").default>} Features.
     * @api
     */
    readFeatures(source, options) {
      if (!source) {
        return [];
      } else if (typeof source === 'string') {
        const doc = parse(source);
        return this.readFeaturesFromDocument(doc, options);
      } else if (isDocument(source)) {
        return this.readFeaturesFromDocument(
          /** @type {Document} */ (source),
          options
        );
      } else {
        return this.readFeaturesFromNode(
          /** @type {Element} */ (source),
          options
        );
      }
    }

    /**
     * @param {Document} doc Document.
     * @param {import("./Feature.js").ReadOptions} [options] Options.
     * @protected
     * @return {Array<import("../Feature.js").default>} Features.
     */
    readFeaturesFromDocument(doc, options) {
      /** @type {Array<import("../Feature.js").default>} */
      const features = [];
      for (let n = doc.firstChild; n; n = n.nextSibling) {
        if (n.nodeType == Node.ELEMENT_NODE) {
          extend$2(
            features,
            this.readFeaturesFromNode(/** @type {Element} */ (n), options)
          );
        }
      }
      return features;
    }

    /**
     * @abstract
     * @param {Element} node Node.
     * @param {import("./Feature.js").ReadOptions} [options] Options.
     * @protected
     * @return {Array<import("../Feature.js").default>} Features.
     */
    readFeaturesFromNode(node, options) {
      return abstract();
    }

    /**
     * Read a single geometry from a source.
     *
     * @param {Document|Element|Object|string} source Source.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @return {import("../geom/Geometry.js").default} Geometry.
     */
    readGeometry(source, options) {
      if (!source) {
        return null;
      } else if (typeof source === 'string') {
        const doc = parse(source);
        return this.readGeometryFromDocument(doc, options);
      } else if (isDocument(source)) {
        return this.readGeometryFromDocument(
          /** @type {Document} */ (source),
          options
        );
      } else {
        return this.readGeometryFromNode(
          /** @type {Element} */ (source),
          options
        );
      }
    }

    /**
     * @param {Document} doc Document.
     * @param {import("./Feature.js").ReadOptions} [options] Options.
     * @protected
     * @return {import("../geom/Geometry.js").default} Geometry.
     */
    readGeometryFromDocument(doc, options) {
      return null; // not implemented
    }

    /**
     * @param {Element} node Node.
     * @param {import("./Feature.js").ReadOptions} [options] Options.
     * @protected
     * @return {import("../geom/Geometry.js").default} Geometry.
     */
    readGeometryFromNode(node, options) {
      return null; // not implemented
    }

    /**
     * Read the projection from the source.
     *
     * @param {Document|Element|Object|string} source Source.
     * @return {import("../proj/Projection.js").default} Projection.
     * @api
     */
    readProjection(source) {
      if (!source) {
        return null;
      } else if (typeof source === 'string') {
        const doc = parse(source);
        return this.readProjectionFromDocument(doc);
      } else if (isDocument(source)) {
        return this.readProjectionFromDocument(/** @type {Document} */ (source));
      } else {
        return this.readProjectionFromNode(/** @type {Element} */ (source));
      }
    }

    /**
     * @param {Document} doc Document.
     * @protected
     * @return {import("../proj/Projection.js").default} Projection.
     */
    readProjectionFromDocument(doc) {
      return this.dataProjection;
    }

    /**
     * @param {Element} node Node.
     * @protected
     * @return {import("../proj/Projection.js").default} Projection.
     */
    readProjectionFromNode(node) {
      return this.dataProjection;
    }

    /**
     * Encode a feature as string.
     *
     * @param {import("../Feature.js").default} feature Feature.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @return {string} Encoded feature.
     */
    writeFeature(feature, options) {
      const node = this.writeFeatureNode(feature, options);
      return this.xmlSerializer_.serializeToString(node);
    }

    /**
     * @param {import("../Feature.js").default} feature Feature.
     * @param {import("./Feature.js").WriteOptions} [options] Options.
     * @protected
     * @return {Node} Node.
     */
    writeFeatureNode(feature, options) {
      return null; // not implemented
    }

    /**
     * Encode an array of features as string.
     *
     * @param {Array<import("../Feature.js").default>} features Features.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @return {string} Result.
     * @api
     */
    writeFeatures(features, options) {
      const node = this.writeFeaturesNode(features, options);
      return this.xmlSerializer_.serializeToString(node);
    }

    /**
     * @param {Array<import("../Feature.js").default>} features Features.
     * @param {import("./Feature.js").WriteOptions} [options] Options.
     * @return {Node} Node.
     */
    writeFeaturesNode(features, options) {
      return null; // not implemented
    }

    /**
     * Encode a geometry as string.
     *
     * @param {import("../geom/Geometry.js").default} geometry Geometry.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @return {string} Encoded geometry.
     */
    writeGeometry(geometry, options) {
      const node = this.writeGeometryNode(geometry, options);
      return this.xmlSerializer_.serializeToString(node);
    }

    /**
     * @param {import("../geom/Geometry.js").default} geometry Geometry.
     * @param {import("./Feature.js").WriteOptions} [options] Options.
     * @return {Node} Node.
     */
    writeGeometryNode(geometry, options) {
      return null; // not implemented
    }
  }

  /**
   * @module ol/format/GMLBase
   */

  /**
   * @const
   * @type {string}
   */
  const GMLNS = 'http://www.opengis.net/gml';

  /**
   * A regular expression that matches if a string only contains whitespace
   * characters. It will e.g. match `''`, `' '`, `'\n'` etc.
   *
   * @const
   * @type {RegExp}
   */
  const ONLY_WHITESPACE_RE = /^\s*$/;

  /**
   * @typedef {Object} Options
   * @property {Object<string, string>|string} [featureNS] Feature
   * namespace. If not defined will be derived from GML. If multiple
   * feature types have been configured which come from different feature
   * namespaces, this will be an object with the keys being the prefixes used
   * in the entries of featureType array. The values of the object will be the
   * feature namespaces themselves. So for instance there might be a featureType
   * item `topp:states` in the `featureType` array and then there will be a key
   * `topp` in the featureNS object with value `http://www.openplans.org/topp`.
   * @property {Array<string>|string} [featureType] Feature type(s) to parse.
   * If multiple feature types need to be configured
   * which come from different feature namespaces, `featureNS` will be an object
   * with the keys being the prefixes used in the entries of featureType array.
   * The values of the object will be the feature namespaces themselves.
   * So for instance there might be a featureType item `topp:states` and then
   * there will be a key named `topp` in the featureNS object with value
   * `http://www.openplans.org/topp`.
   * @property {string} [srsName] srsName to use when writing geometries.
   * @property {boolean} [surface=false] Write gml:Surface instead of gml:Polygon
   * elements. This also affects the elements in multi-part geometries.
   * @property {boolean} [curve=false] Write gml:Curve instead of gml:LineString
   * elements. This also affects the elements in multi-part geometries.
   * @property {boolean} [multiCurve=true] Write gml:MultiCurve instead of gml:MultiLineString.
   * Since the latter is deprecated in GML 3.
   * @property {boolean} [multiSurface=true] Write gml:multiSurface instead of
   * gml:MultiPolygon. Since the latter is deprecated in GML 3.
   * @property {string} [schemaLocation] Optional schemaLocation to use when
   * writing out the GML, this will override the default provided.
   * @property {boolean} [hasZ=false] If coordinates have a Z value.
   */

  /**
   * @classdesc
   * Abstract base class; normally only used for creating subclasses and not
   * instantiated in apps.
   * Feature base format for reading and writing data in the GML format.
   * This class cannot be instantiated, it contains only base content that
   * is shared with versioned format classes GML2 and GML3.
   *
   * @abstract
   * @api
   */
  class GMLBase extends XMLFeature {
    /**
     * @param {Options} [options] Optional configuration object.
     */
    constructor(options) {
      super();

      options = options ? options : {};

      /**
       * @protected
       * @type {Array<string>|string|undefined}
       */
      this.featureType = options.featureType;

      /**
       * @protected
       * @type {Object<string, string>|string|undefined}
       */
      this.featureNS = options.featureNS;

      /**
       * @protected
       * @type {string|undefined}
       */
      this.srsName = options.srsName;

      /**
       * @protected
       * @type {string}
       */
      this.schemaLocation = '';

      /**
       * @type {Object<string, Object<string, Object>>}
       */
      this.FEATURE_COLLECTION_PARSERS = {};
      this.FEATURE_COLLECTION_PARSERS[this.namespace] = {
        'featureMember': makeArrayPusher(this.readFeaturesInternal),
        'featureMembers': makeReplacer(this.readFeaturesInternal),
      };

      this.supportedMediaTypes = ['application/gml+xml'];
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {Array<Feature> | undefined} Features.
     */
    readFeaturesInternal(node, objectStack) {
      const localName = node.localName;
      let features = null;
      if (localName == 'FeatureCollection') {
        features = pushParseAndPop(
          [],
          this.FEATURE_COLLECTION_PARSERS,
          node,
          objectStack,
          this
        );
      } else if (
        localName == 'featureMembers' ||
        localName == 'featureMember' ||
        localName == 'member'
      ) {
        const context = objectStack[0];
        let featureType = context['featureType'];
        let featureNS = context['featureNS'];
        const prefix = 'p';
        const defaultPrefix = 'p0';
        if (!featureType && node.childNodes) {
          (featureType = []), (featureNS = {});
          for (let i = 0, ii = node.childNodes.length; i < ii; ++i) {
            const child = /** @type {Element} */ (node.childNodes[i]);
            if (child.nodeType === 1) {
              const ft = child.nodeName.split(':').pop();
              if (!featureType.includes(ft)) {
                let key = '';
                let count = 0;
                const uri = child.namespaceURI;
                for (const candidate in featureNS) {
                  if (featureNS[candidate] === uri) {
                    key = candidate;
                    break;
                  }
                  ++count;
                }
                if (!key) {
                  key = prefix + count;
                  featureNS[key] = uri;
                }
                featureType.push(key + ':' + ft);
              }
            }
          }
          if (localName != 'featureMember') {
            // recheck featureType for each featureMember
            context['featureType'] = featureType;
            context['featureNS'] = featureNS;
          }
        }
        if (typeof featureNS === 'string') {
          const ns = featureNS;
          featureNS = {};
          featureNS[defaultPrefix] = ns;
        }
        /** @type {Object<string, Object<string, import("../xml.js").Parser>>} */
        const parsersNS = {};
        const featureTypes = Array.isArray(featureType)
          ? featureType
          : [featureType];
        for (const p in featureNS) {
          /** @type {Object<string, import("../xml.js").Parser>} */
          const parsers = {};
          for (let i = 0, ii = featureTypes.length; i < ii; ++i) {
            const featurePrefix = featureTypes[i].includes(':')
              ? featureTypes[i].split(':')[0]
              : defaultPrefix;
            if (featurePrefix === p) {
              parsers[featureTypes[i].split(':').pop()] =
                localName == 'featureMembers'
                  ? makeArrayPusher(this.readFeatureElement, this)
                  : makeReplacer(this.readFeatureElement, this);
            }
          }
          parsersNS[featureNS[p]] = parsers;
        }
        if (localName == 'featureMember' || localName == 'member') {
          features = pushParseAndPop(undefined, parsersNS, node, objectStack);
        } else {
          features = pushParseAndPop([], parsersNS, node, objectStack);
        }
      }
      if (features === null) {
        features = [];
      }
      return features;
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {import("../geom/Geometry.js").default|import("../extent.js").Extent|undefined} Geometry.
     */
    readGeometryOrExtent(node, objectStack) {
      const context = /** @type {Object} */ (objectStack[0]);
      context['srsName'] = node.firstElementChild.getAttribute('srsName');
      context['srsDimension'] =
        node.firstElementChild.getAttribute('srsDimension');
      return pushParseAndPop(
        null,
        this.GEOMETRY_PARSERS,
        node,
        objectStack,
        this
      );
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {import("../extent.js").Extent|undefined} Geometry.
     */
    readExtentElement(node, objectStack) {
      const context = /** @type {Object} */ (objectStack[0]);
      const extent = /** @type {import("../extent.js").Extent} */ (
        this.readGeometryOrExtent(node, objectStack)
      );
      return extent ? transformExtentWithOptions(extent, context) : undefined;
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {import("../geom/Geometry.js").default|undefined} Geometry.
     */
    readGeometryElement(node, objectStack) {
      const context = /** @type {Object} */ (objectStack[0]);
      const geometry = /** @type {import("../geom/Geometry.js").default} */ (
        this.readGeometryOrExtent(node, objectStack)
      );
      return geometry
        ? transformGeometryWithOptions(geometry, false, context)
        : undefined;
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @param {boolean} asFeature whether result should be wrapped as a feature.
     * @return {Feature|Object} Feature
     */
    readFeatureElementInternal(node, objectStack, asFeature) {
      let geometryName;
      const values = {};
      for (let n = node.firstElementChild; n; n = n.nextElementSibling) {
        let value;
        const localName = n.localName;
        // first, check if it is simple attribute
        if (
          n.childNodes.length === 0 ||
          (n.childNodes.length === 1 &&
            (n.firstChild.nodeType === 3 || n.firstChild.nodeType === 4))
        ) {
          value = getAllTextContent(n, false);
          if (ONLY_WHITESPACE_RE.test(value)) {
            value = undefined;
          }
        } else {
          if (asFeature) {
            //if feature, try it as a geometry or extent
            value =
              localName === 'boundedBy'
                ? this.readExtentElement(n, objectStack)
                : this.readGeometryElement(n, objectStack);
          }
          if (!value) {
            //if not a geometry or not a feature, treat it as a complex attribute
            value = this.readFeatureElementInternal(n, objectStack, false);
          } else if (localName !== 'boundedBy') {
            // boundedBy is an extent and must not be considered as a geometry
            geometryName = localName;
          }
        }

        const len = n.attributes.length;
        if (len > 0) {
          value = {_content_: value};
          for (let i = 0; i < len; i++) {
            const attName = n.attributes[i].name;
            value[attName] = n.attributes[i].value;
          }
        }

        if (values[localName]) {
          if (!(values[localName] instanceof Array)) {
            values[localName] = [values[localName]];
          }
          values[localName].push(value);
        } else {
          values[localName] = value;
        }
      }
      if (!asFeature) {
        return values;
      } else {
        const feature = new Feature(values);
        if (geometryName) {
          feature.setGeometryName(geometryName);
        }
        const fid =
          node.getAttribute('fid') || getAttributeNS(node, this.namespace, 'id');
        if (fid) {
          feature.setId(fid);
        }
        return feature;
      }
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {Feature} Feature.
     */
    readFeatureElement(node, objectStack) {
      return this.readFeatureElementInternal(node, objectStack, true);
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {Point|undefined} Point.
     */
    readPoint(node, objectStack) {
      const flatCoordinates = this.readFlatCoordinatesFromNode(node, objectStack);
      if (flatCoordinates) {
        return new Point(flatCoordinates, 'XYZ');
      }
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {MultiPoint|undefined} MultiPoint.
     */
    readMultiPoint(node, objectStack) {
      /** @type {Array<Array<number>>} */
      const coordinates = pushParseAndPop(
        [],
        this.MULTIPOINT_PARSERS,
        node,
        objectStack,
        this
      );
      if (coordinates) {
        return new MultiPoint(coordinates);
      } else {
        return undefined;
      }
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {MultiLineString|undefined} MultiLineString.
     */
    readMultiLineString(node, objectStack) {
      /** @type {Array<LineString>} */
      const lineStrings = pushParseAndPop(
        [],
        this.MULTILINESTRING_PARSERS,
        node,
        objectStack,
        this
      );
      if (lineStrings) {
        return new MultiLineString(lineStrings);
      }
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {MultiPolygon|undefined} MultiPolygon.
     */
    readMultiPolygon(node, objectStack) {
      /** @type {Array<Polygon>} */
      const polygons = pushParseAndPop(
        [],
        this.MULTIPOLYGON_PARSERS,
        node,
        objectStack,
        this
      );
      if (polygons) {
        return new MultiPolygon(polygons);
      }
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     */
    pointMemberParser(node, objectStack) {
      parseNode(this.POINTMEMBER_PARSERS, node, objectStack, this);
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     */
    lineStringMemberParser(node, objectStack) {
      parseNode(this.LINESTRINGMEMBER_PARSERS, node, objectStack, this);
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     */
    polygonMemberParser(node, objectStack) {
      parseNode(this.POLYGONMEMBER_PARSERS, node, objectStack, this);
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {LineString|undefined} LineString.
     */
    readLineString(node, objectStack) {
      const flatCoordinates = this.readFlatCoordinatesFromNode(node, objectStack);
      if (flatCoordinates) {
        const lineString = new LineString(flatCoordinates, 'XYZ');
        return lineString;
      } else {
        return undefined;
      }
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {Array<number>|undefined} LinearRing flat coordinates.
     */
    readFlatLinearRing(node, objectStack) {
      const ring = pushParseAndPop(
        null,
        this.GEOMETRY_FLAT_COORDINATES_PARSERS,
        node,
        objectStack,
        this
      );
      if (ring) {
        return ring;
      } else {
        return undefined;
      }
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {LinearRing|undefined} LinearRing.
     */
    readLinearRing(node, objectStack) {
      const flatCoordinates = this.readFlatCoordinatesFromNode(node, objectStack);
      if (flatCoordinates) {
        return new LinearRing(flatCoordinates, 'XYZ');
      }
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {Polygon|undefined} Polygon.
     */
    readPolygon(node, objectStack) {
      /** @type {Array<Array<number>>} */
      const flatLinearRings = pushParseAndPop(
        [null],
        this.FLAT_LINEAR_RINGS_PARSERS,
        node,
        objectStack,
        this
      );
      if (flatLinearRings && flatLinearRings[0]) {
        const flatCoordinates = flatLinearRings[0];
        const ends = [flatCoordinates.length];
        let i, ii;
        for (i = 1, ii = flatLinearRings.length; i < ii; ++i) {
          extend$2(flatCoordinates, flatLinearRings[i]);
          ends.push(flatCoordinates.length);
        }
        return new Polygon(flatCoordinates, 'XYZ', ends);
      } else {
        return undefined;
      }
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {Array<number>} Flat coordinates.
     */
    readFlatCoordinatesFromNode(node, objectStack) {
      return pushParseAndPop(
        null,
        this.GEOMETRY_FLAT_COORDINATES_PARSERS,
        node,
        objectStack,
        this
      );
    }

    /**
     * @param {Element} node Node.
     * @param {import("./Feature.js").ReadOptions} [options] Options.
     * @protected
     * @return {import("../geom/Geometry.js").default} Geometry.
     */
    readGeometryFromNode(node, options) {
      const geometry = this.readGeometryElement(node, [
        this.getReadOptions(node, options ? options : {}),
      ]);
      return geometry ? geometry : null;
    }

    /**
     * @param {Element} node Node.
     * @param {import("./Feature.js").ReadOptions} [options] Options.
     * @return {Array<import("../Feature.js").default>} Features.
     */
    readFeaturesFromNode(node, options) {
      const internalOptions = {
        featureType: this.featureType,
        featureNS: this.featureNS,
      };
      if (internalOptions) {
        Object.assign(internalOptions, this.getReadOptions(node, options));
      }
      const features = this.readFeaturesInternal(node, [internalOptions]);
      return features || [];
    }

    /**
     * @param {Element} node Node.
     * @return {import("../proj/Projection.js").default} Projection.
     */
    readProjectionFromNode(node) {
      return get$2(
        this.srsName
          ? this.srsName
          : node.firstElementChild.getAttribute('srsName')
      );
    }
  }

  GMLBase.prototype.namespace = GMLNS;

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GMLBase.prototype.FLAT_LINEAR_RINGS_PARSERS = {
    'http://www.opengis.net/gml': {},
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GMLBase.prototype.GEOMETRY_FLAT_COORDINATES_PARSERS = {
    'http://www.opengis.net/gml': {},
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GMLBase.prototype.GEOMETRY_PARSERS = {
    'http://www.opengis.net/gml': {},
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GMLBase.prototype.MULTIPOINT_PARSERS = {
    'http://www.opengis.net/gml': {
      'pointMember': makeArrayPusher(GMLBase.prototype.pointMemberParser),
      'pointMembers': makeArrayPusher(GMLBase.prototype.pointMemberParser),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GMLBase.prototype.MULTILINESTRING_PARSERS = {
    'http://www.opengis.net/gml': {
      'lineStringMember': makeArrayPusher(
        GMLBase.prototype.lineStringMemberParser
      ),
      'lineStringMembers': makeArrayPusher(
        GMLBase.prototype.lineStringMemberParser
      ),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GMLBase.prototype.MULTIPOLYGON_PARSERS = {
    'http://www.opengis.net/gml': {
      'polygonMember': makeArrayPusher(GMLBase.prototype.polygonMemberParser),
      'polygonMembers': makeArrayPusher(GMLBase.prototype.polygonMemberParser),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GMLBase.prototype.POINTMEMBER_PARSERS = {
    'http://www.opengis.net/gml': {
      'Point': makeArrayPusher(GMLBase.prototype.readFlatCoordinatesFromNode),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GMLBase.prototype.LINESTRINGMEMBER_PARSERS = {
    'http://www.opengis.net/gml': {
      'LineString': makeArrayPusher(GMLBase.prototype.readLineString),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GMLBase.prototype.POLYGONMEMBER_PARSERS = {
    'http://www.opengis.net/gml': {
      'Polygon': makeArrayPusher(GMLBase.prototype.readPolygon),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GMLBase.prototype.RING_PARSERS = {
    'http://www.opengis.net/gml': {
      'LinearRing': makeReplacer(GMLBase.prototype.readFlatLinearRing),
    },
  };

  /**
   * @module ol/format/xsd
   */

  /**
   * @param {Node} node Node.
   * @return {boolean|undefined} Boolean.
   */
  function readBoolean(node) {
    const s = getAllTextContent(node, false);
    return readBooleanString(s);
  }

  /**
   * @param {string} string String.
   * @return {boolean|undefined} Boolean.
   */
  function readBooleanString(string) {
    const m = /^\s*(true|1)|(false|0)\s*$/.exec(string);
    if (m) {
      return m[1] !== undefined || false;
    } else {
      return undefined;
    }
  }

  /**
   * @param {Node} node Node.
   * @return {number|undefined} DateTime in seconds.
   */
  function readDateTime(node) {
    const s = getAllTextContent(node, false);
    const dateTime = Date.parse(s);
    return isNaN(dateTime) ? undefined : dateTime / 1000;
  }

  /**
   * @param {Node} node Node.
   * @return {number|undefined} Decimal.
   */
  function readDecimal(node) {
    const s = getAllTextContent(node, false);
    return readDecimalString(s);
  }

  /**
   * @param {string} string String.
   * @return {number|undefined} Decimal.
   */
  function readDecimalString(string) {
    // FIXME check spec
    const m = /^\s*([+\-]?\d*\.?\d+(?:e[+\-]?\d+)?)\s*$/i.exec(string);
    if (m) {
      return parseFloat(m[1]);
    } else {
      return undefined;
    }
  }

  /**
   * @param {Node} node Node.
   * @return {number|undefined} Non negative integer.
   */
  function readPositiveInteger(node) {
    const s = getAllTextContent(node, false);
    return readNonNegativeIntegerString(s);
  }

  /**
   * @param {string} string String.
   * @return {number|undefined} Non negative integer.
   */
  function readNonNegativeIntegerString(string) {
    const m = /^\s*(\d+)\s*$/.exec(string);
    if (m) {
      return parseInt(m[1], 10);
    } else {
      return undefined;
    }
  }

  /**
   * @param {Node} node Node.
   * @return {string|undefined} String.
   */
  function readString(node) {
    return getAllTextContent(node, false).trim();
  }

  /**
   * @param {Node} node Node to append a TextNode with the boolean to.
   * @param {boolean} bool Boolean.
   */
  function writeBooleanTextNode(node, bool) {
    writeStringTextNode(node, bool ? '1' : '0');
  }

  /**
   * @param {Node} node Node to append a CDATA Section with the string to.
   * @param {string} string String.
   */
  function writeCDATASection(node, string) {
    node.appendChild(getDocument().createCDATASection(string));
  }

  /**
   * @param {Node} node Node to append a TextNode with the dateTime to.
   * @param {number} dateTime DateTime in seconds.
   */
  function writeDateTimeTextNode(node, dateTime) {
    const date = new Date(dateTime * 1000);
    const string =
      date.getUTCFullYear() +
      '-' +
      padNumber(date.getUTCMonth() + 1, 2) +
      '-' +
      padNumber(date.getUTCDate(), 2) +
      'T' +
      padNumber(date.getUTCHours(), 2) +
      ':' +
      padNumber(date.getUTCMinutes(), 2) +
      ':' +
      padNumber(date.getUTCSeconds(), 2) +
      'Z';
    node.appendChild(getDocument().createTextNode(string));
  }

  /**
   * @param {Node} node Node to append a TextNode with the decimal to.
   * @param {number} decimal Decimal.
   */
  function writeDecimalTextNode(node, decimal) {
    const string = decimal.toPrecision();
    node.appendChild(getDocument().createTextNode(string));
  }

  /**
   * @param {Node} node Node to append a TextNode with the decimal to.
   * @param {number} nonNegativeInteger Non negative integer.
   */
  function writeNonNegativeIntegerTextNode(node, nonNegativeInteger) {
    const string = nonNegativeInteger.toString();
    node.appendChild(getDocument().createTextNode(string));
  }

  /**
   * @param {Node} node Node to append a TextNode with the string to.
   * @param {string} string String.
   */
  function writeStringTextNode(node, string) {
    node.appendChild(getDocument().createTextNode(string));
  }

  /**
   * @module ol/format/GML2
   */

  /**
   * @const
   * @type {string}
   */
  const schemaLocation$1 =
    GMLNS + ' http://schemas.opengis.net/gml/2.1.2/feature.xsd';

  /**
   * @const
   * @type {Object<string, string>}
   */
  const MULTIGEOMETRY_TO_MEMBER_NODENAME$1 = {
    'MultiLineString': 'lineStringMember',
    'MultiCurve': 'curveMember',
    'MultiPolygon': 'polygonMember',
    'MultiSurface': 'surfaceMember',
  };

  /**
   * @classdesc
   * Feature format for reading and writing data in the GML format,
   * version 2.1.2.
   *
   * @api
   */
  class GML2 extends GMLBase {
    /**
     * @param {import("./GMLBase.js").Options} [options] Optional configuration object.
     */
    constructor(options) {
      options = options ? options : {};

      super(options);

      this.FEATURE_COLLECTION_PARSERS[GMLNS]['featureMember'] = makeArrayPusher(
        this.readFeaturesInternal
      );

      /**
       * @type {string}
       */
      this.schemaLocation = options.schemaLocation
        ? options.schemaLocation
        : schemaLocation$1;
    }

    /**
     * @param {Node} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {Array<number>|undefined} Flat coordinates.
     */
    readFlatCoordinates(node, objectStack) {
      const s = getAllTextContent(node, false).replace(/^\s*|\s*$/g, '');
      const context = /** @type {import("../xml.js").NodeStackItem} */ (
        objectStack[0]
      );
      const containerSrs = context['srsName'];
      let axisOrientation = 'enu';
      if (containerSrs) {
        const proj = get$2(containerSrs);
        if (proj) {
          axisOrientation = proj.getAxisOrientation();
        }
      }
      const coordsGroups = s.trim().split(/\s+/);
      const flatCoordinates = [];
      for (let i = 0, ii = coordsGroups.length; i < ii; i++) {
        const coords = coordsGroups[i].split(/,+/);
        const x = parseFloat(coords[0]);
        const y = parseFloat(coords[1]);
        const z = coords.length === 3 ? parseFloat(coords[2]) : 0;
        if (axisOrientation.substr(0, 2) === 'en') {
          flatCoordinates.push(x, y, z);
        } else {
          flatCoordinates.push(y, x, z);
        }
      }
      return flatCoordinates;
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {import("../extent.js").Extent|undefined} Envelope.
     */
    readBox(node, objectStack) {
      /** @type {Array<number>} */
      const flatCoordinates = pushParseAndPop(
        [null],
        this.BOX_PARSERS_,
        node,
        objectStack,
        this
      );
      return createOrUpdate$2(
        flatCoordinates[1][0],
        flatCoordinates[1][1],
        flatCoordinates[1][3],
        flatCoordinates[1][4]
      );
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     */
    innerBoundaryIsParser(node, objectStack) {
      /** @type {Array<number>|undefined} */
      const flatLinearRing = pushParseAndPop(
        undefined,
        this.RING_PARSERS,
        node,
        objectStack,
        this
      );
      if (flatLinearRing) {
        const flatLinearRings =
          /** @type {Array<Array<number>>} */
          (objectStack[objectStack.length - 1]);
        flatLinearRings.push(flatLinearRing);
      }
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     */
    outerBoundaryIsParser(node, objectStack) {
      /** @type {Array<number>|undefined} */
      const flatLinearRing = pushParseAndPop(
        undefined,
        this.RING_PARSERS,
        node,
        objectStack,
        this
      );
      if (flatLinearRing) {
        const flatLinearRings =
          /** @type {Array<Array<number>>} */
          (objectStack[objectStack.length - 1]);
        flatLinearRings[0] = flatLinearRing;
      }
    }

    /**
     * @const
     * @param {*} value Value.
     * @param {Array<*>} objectStack Object stack.
     * @param {string} [nodeName] Node name.
     * @return {Element|undefined} Node.
     * @private
     */
    GEOMETRY_NODE_FACTORY_(value, objectStack, nodeName) {
      const context = objectStack[objectStack.length - 1];
      const multiSurface = context['multiSurface'];
      const surface = context['surface'];
      const multiCurve = context['multiCurve'];
      if (!Array.isArray(value)) {
        nodeName = /** @type {import("../geom/Geometry.js").default} */ (
          value
        ).getType();
        if (nodeName === 'MultiPolygon' && multiSurface === true) {
          nodeName = 'MultiSurface';
        } else if (nodeName === 'Polygon' && surface === true) {
          nodeName = 'Surface';
        } else if (nodeName === 'MultiLineString' && multiCurve === true) {
          nodeName = 'MultiCurve';
        }
      } else {
        nodeName = 'Envelope';
      }
      return createElementNS('http://www.opengis.net/gml', nodeName);
    }

    /**
     * @param {Element} node Node.
     * @param {import("../Feature.js").default} feature Feature.
     * @param {Array<*>} objectStack Node stack.
     */
    writeFeatureElement(node, feature, objectStack) {
      const fid = feature.getId();
      if (fid) {
        node.setAttribute('fid', /** @type {string} */ (fid));
      }
      const context = /** @type {Object} */ (objectStack[objectStack.length - 1]);
      const featureNS = context['featureNS'];
      const geometryName = feature.getGeometryName();
      if (!context.serializers) {
        context.serializers = {};
        context.serializers[featureNS] = {};
      }
      const keys = [];
      const values = [];
      if (feature.hasProperties()) {
        const properties = feature.getProperties();
        for (const key in properties) {
          const value = properties[key];
          if (value !== null) {
            keys.push(key);
            values.push(value);
            if (
              key == geometryName ||
              typeof (/** @type {?} */ (value).getSimplifiedGeometry) ===
                'function'
            ) {
              if (!(key in context.serializers[featureNS])) {
                context.serializers[featureNS][key] = makeChildAppender(
                  this.writeGeometryElement,
                  this
                );
              }
            } else {
              if (!(key in context.serializers[featureNS])) {
                context.serializers[featureNS][key] =
                  makeChildAppender(writeStringTextNode);
              }
            }
          }
        }
      }
      const item = Object.assign({}, context);
      item.node = node;
      pushSerializeAndPop(
        /** @type {import("../xml.js").NodeStackItem} */
        (item),
        context.serializers,
        makeSimpleNodeFactory(undefined, featureNS),
        values,
        objectStack,
        keys
      );
    }

    /**
     * @param {Element} node Node.
     * @param {import("../geom/LineString.js").default} geometry LineString geometry.
     * @param {Array<*>} objectStack Node stack.
     */
    writeCurveOrLineString(node, geometry, objectStack) {
      const context = objectStack[objectStack.length - 1];
      const srsName = context['srsName'];
      if (node.nodeName !== 'LineStringSegment' && srsName) {
        node.setAttribute('srsName', srsName);
      }
      if (
        node.nodeName === 'LineString' ||
        node.nodeName === 'LineStringSegment'
      ) {
        const coordinates = this.createCoordinatesNode_(node.namespaceURI);
        node.appendChild(coordinates);
        this.writeCoordinates_(coordinates, geometry, objectStack);
      } else if (node.nodeName === 'Curve') {
        const segments = createElementNS(node.namespaceURI, 'segments');
        node.appendChild(segments);
        this.writeCurveSegments_(segments, geometry, objectStack);
      }
    }

    /**
     * @param {Element} node Node.
     * @param {import("../geom/LineString.js").default} line LineString geometry.
     * @param {Array<*>} objectStack Node stack.
     */
    writeLineStringOrCurveMember(node, line, objectStack) {
      const child = this.GEOMETRY_NODE_FACTORY_(line, objectStack);
      if (child) {
        node.appendChild(child);
        this.writeCurveOrLineString(child, line, objectStack);
      }
    }

    /**
     * @param {Element} node Node.
     * @param {import("../geom/MultiLineString.js").default} geometry MultiLineString geometry.
     * @param {Array<*>} objectStack Node stack.
     */
    writeMultiCurveOrLineString(node, geometry, objectStack) {
      const context = objectStack[objectStack.length - 1];
      const hasZ = context['hasZ'];
      const srsName = context['srsName'];
      const curve = context['curve'];
      if (srsName) {
        node.setAttribute('srsName', srsName);
      }
      const lines = geometry.getLineStrings();
      pushSerializeAndPop(
        {node: node, hasZ: hasZ, srsName: srsName, curve: curve},
        this.LINESTRINGORCURVEMEMBER_SERIALIZERS,
        this.MULTIGEOMETRY_MEMBER_NODE_FACTORY_,
        lines,
        objectStack,
        undefined,
        this
      );
    }

    /**
     * @param {Node} node Node.
     * @param {import("../geom/Geometry.js").default|import("../extent.js").Extent} geometry Geometry.
     * @param {Array<*>} objectStack Node stack.
     */
    writeGeometryElement(node, geometry, objectStack) {
      const context = /** @type {import("./Feature.js").WriteOptions} */ (
        objectStack[objectStack.length - 1]
      );
      const item = Object.assign({}, context);
      item['node'] = node;
      let value;
      if (Array.isArray(geometry)) {
        value = transformExtentWithOptions(
          /** @type {import("../extent.js").Extent} */ (geometry),
          context
        );
      } else {
        value = transformGeometryWithOptions(
          /** @type {import("../geom/Geometry.js").default} */ (geometry),
          true,
          context
        );
      }
      pushSerializeAndPop(
        /** @type {import("../xml.js").NodeStackItem} */
        (item),
        this.GEOMETRY_SERIALIZERS,
        this.GEOMETRY_NODE_FACTORY_,
        [value],
        objectStack,
        undefined,
        this
      );
    }

    /**
     * @param {string} namespaceURI XML namespace.
     * @return {Element} coordinates node.
     * @private
     */
    createCoordinatesNode_(namespaceURI) {
      const coordinates = createElementNS(namespaceURI, 'coordinates');
      coordinates.setAttribute('decimal', '.');
      coordinates.setAttribute('cs', ',');
      coordinates.setAttribute('ts', ' ');

      return coordinates;
    }

    /**
     * @param {Node} node Node.
     * @param {import("../geom/LineString.js").default|import("../geom/LinearRing.js").default} value Geometry.
     * @param {Array<*>} objectStack Node stack.
     * @private
     */
    writeCoordinates_(node, value, objectStack) {
      const context = objectStack[objectStack.length - 1];
      const hasZ = context['hasZ'];
      const srsName = context['srsName'];
      // only 2d for simple features profile
      const points = value.getCoordinates();
      const len = points.length;
      const parts = new Array(len);
      for (let i = 0; i < len; ++i) {
        const point = points[i];
        parts[i] = this.getCoords_(point, srsName, hasZ);
      }
      writeStringTextNode(node, parts.join(' '));
    }

    /**
     * @param {Element} node Node.
     * @param {import("../geom/LineString.js").default} line LineString geometry.
     * @param {Array<*>} objectStack Node stack.
     * @private
     */
    writeCurveSegments_(node, line, objectStack) {
      const child = createElementNS(node.namespaceURI, 'LineStringSegment');
      node.appendChild(child);
      this.writeCurveOrLineString(child, line, objectStack);
    }

    /**
     * @param {Element} node Node.
     * @param {import("../geom/Polygon.js").default} geometry Polygon geometry.
     * @param {Array<*>} objectStack Node stack.
     */
    writeSurfaceOrPolygon(node, geometry, objectStack) {
      const context = objectStack[objectStack.length - 1];
      const hasZ = context['hasZ'];
      const srsName = context['srsName'];
      if (node.nodeName !== 'PolygonPatch' && srsName) {
        node.setAttribute('srsName', srsName);
      }
      if (node.nodeName === 'Polygon' || node.nodeName === 'PolygonPatch') {
        const rings = geometry.getLinearRings();
        pushSerializeAndPop(
          {node: node, hasZ: hasZ, srsName: srsName},
          this.RING_SERIALIZERS,
          this.RING_NODE_FACTORY_,
          rings,
          objectStack,
          undefined,
          this
        );
      } else if (node.nodeName === 'Surface') {
        const patches = createElementNS(node.namespaceURI, 'patches');
        node.appendChild(patches);
        this.writeSurfacePatches_(patches, geometry, objectStack);
      }
    }

    /**
     * @param {*} value Value.
     * @param {Array<*>} objectStack Object stack.
     * @param {string} [nodeName] Node name.
     * @return {Node} Node.
     * @private
     */
    RING_NODE_FACTORY_(value, objectStack, nodeName) {
      const context = objectStack[objectStack.length - 1];
      const parentNode = context.node;
      const exteriorWritten = context['exteriorWritten'];
      if (exteriorWritten === undefined) {
        context['exteriorWritten'] = true;
      }
      return createElementNS(
        parentNode.namespaceURI,
        exteriorWritten !== undefined ? 'innerBoundaryIs' : 'outerBoundaryIs'
      );
    }

    /**
     * @param {Element} node Node.
     * @param {import("../geom/Polygon.js").default} polygon Polygon geometry.
     * @param {Array<*>} objectStack Node stack.
     * @private
     */
    writeSurfacePatches_(node, polygon, objectStack) {
      const child = createElementNS(node.namespaceURI, 'PolygonPatch');
      node.appendChild(child);
      this.writeSurfaceOrPolygon(child, polygon, objectStack);
    }

    /**
     * @param {Element} node Node.
     * @param {import("../geom/LinearRing.js").default} ring LinearRing geometry.
     * @param {Array<*>} objectStack Node stack.
     */
    writeRing(node, ring, objectStack) {
      const linearRing = createElementNS(node.namespaceURI, 'LinearRing');
      node.appendChild(linearRing);
      this.writeLinearRing(linearRing, ring, objectStack);
    }

    /**
     * @param {Array<number>} point Point geometry.
     * @param {string} [srsName] Optional srsName
     * @param {boolean} [hasZ] whether the geometry has a Z coordinate (is 3D) or not.
     * @return {string} The coords string.
     * @private
     */
    getCoords_(point, srsName, hasZ) {
      let axisOrientation = 'enu';
      if (srsName) {
        axisOrientation = get$2(srsName).getAxisOrientation();
      }
      let coords =
        axisOrientation.substr(0, 2) === 'en'
          ? point[0] + ',' + point[1]
          : point[1] + ',' + point[0];
      if (hasZ) {
        // For newly created points, Z can be undefined.
        const z = point[2] || 0;
        coords += ',' + z;
      }

      return coords;
    }

    /**
     * @param {Element} node Node.
     * @param {import("../geom/Point.js").default} geometry Point geometry.
     * @param {Array<*>} objectStack Node stack.
     */
    writePoint(node, geometry, objectStack) {
      const context = objectStack[objectStack.length - 1];
      const hasZ = context['hasZ'];
      const srsName = context['srsName'];
      if (srsName) {
        node.setAttribute('srsName', srsName);
      }
      const coordinates = this.createCoordinatesNode_(node.namespaceURI);
      node.appendChild(coordinates);
      const point = geometry.getCoordinates();
      const coord = this.getCoords_(point, srsName, hasZ);
      writeStringTextNode(coordinates, coord);
    }

    /**
     * @param {Element} node Node.
     * @param {import("../geom/MultiPoint.js").default} geometry MultiPoint geometry.
     * @param {Array<*>} objectStack Node stack.
     */
    writeMultiPoint(node, geometry, objectStack) {
      const context = objectStack[objectStack.length - 1];
      const hasZ = context['hasZ'];
      const srsName = context['srsName'];
      if (srsName) {
        node.setAttribute('srsName', srsName);
      }
      const points = geometry.getPoints();
      pushSerializeAndPop(
        {node: node, hasZ: hasZ, srsName: srsName},
        this.POINTMEMBER_SERIALIZERS,
        makeSimpleNodeFactory('pointMember'),
        points,
        objectStack,
        undefined,
        this
      );
    }

    /**
     * @param {Element} node Node.
     * @param {import("../geom/Point.js").default} point Point geometry.
     * @param {Array<*>} objectStack Node stack.
     */
    writePointMember(node, point, objectStack) {
      const child = createElementNS(node.namespaceURI, 'Point');
      node.appendChild(child);
      this.writePoint(child, point, objectStack);
    }

    /**
     * @param {Element} node Node.
     * @param {import("../geom/LinearRing.js").default} geometry LinearRing geometry.
     * @param {Array<*>} objectStack Node stack.
     */
    writeLinearRing(node, geometry, objectStack) {
      const context = objectStack[objectStack.length - 1];
      const srsName = context['srsName'];
      if (srsName) {
        node.setAttribute('srsName', srsName);
      }
      const coordinates = this.createCoordinatesNode_(node.namespaceURI);
      node.appendChild(coordinates);
      this.writeCoordinates_(coordinates, geometry, objectStack);
    }

    /**
     * @param {Element} node Node.
     * @param {import("../geom/MultiPolygon.js").default} geometry MultiPolygon geometry.
     * @param {Array<*>} objectStack Node stack.
     */
    writeMultiSurfaceOrPolygon(node, geometry, objectStack) {
      const context = objectStack[objectStack.length - 1];
      const hasZ = context['hasZ'];
      const srsName = context['srsName'];
      const surface = context['surface'];
      if (srsName) {
        node.setAttribute('srsName', srsName);
      }
      const polygons = geometry.getPolygons();
      pushSerializeAndPop(
        {node: node, hasZ: hasZ, srsName: srsName, surface: surface},
        this.SURFACEORPOLYGONMEMBER_SERIALIZERS,
        this.MULTIGEOMETRY_MEMBER_NODE_FACTORY_,
        polygons,
        objectStack,
        undefined,
        this
      );
    }

    /**
     * @param {Node} node Node.
     * @param {import("../geom/Polygon.js").default} polygon Polygon geometry.
     * @param {Array<*>} objectStack Node stack.
     */
    writeSurfaceOrPolygonMember(node, polygon, objectStack) {
      const child = this.GEOMETRY_NODE_FACTORY_(polygon, objectStack);
      if (child) {
        node.appendChild(child);
        this.writeSurfaceOrPolygon(child, polygon, objectStack);
      }
    }

    /**
     * @param {Element} node Node.
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {Array<*>} objectStack Node stack.
     */
    writeEnvelope(node, extent, objectStack) {
      const context = objectStack[objectStack.length - 1];
      const srsName = context['srsName'];
      if (srsName) {
        node.setAttribute('srsName', srsName);
      }
      const keys = ['lowerCorner', 'upperCorner'];
      const values = [extent[0] + ' ' + extent[1], extent[2] + ' ' + extent[3]];
      pushSerializeAndPop(
        /** @type {import("../xml.js").NodeStackItem} */
        ({node: node}),
        this.ENVELOPE_SERIALIZERS,
        OBJECT_PROPERTY_NODE_FACTORY,
        values,
        objectStack,
        keys,
        this
      );
    }

    /**
     * @const
     * @param {*} value Value.
     * @param {Array<*>} objectStack Object stack.
     * @param {string} [nodeName] Node name.
     * @return {Node|undefined} Node.
     * @private
     */
    MULTIGEOMETRY_MEMBER_NODE_FACTORY_(value, objectStack, nodeName) {
      const parentNode = objectStack[objectStack.length - 1].node;
      return createElementNS(
        'http://www.opengis.net/gml',
        MULTIGEOMETRY_TO_MEMBER_NODENAME$1[parentNode.nodeName]
      );
    }
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML2.prototype.GEOMETRY_FLAT_COORDINATES_PARSERS = {
    'http://www.opengis.net/gml': {
      'coordinates': makeReplacer(GML2.prototype.readFlatCoordinates),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML2.prototype.FLAT_LINEAR_RINGS_PARSERS = {
    'http://www.opengis.net/gml': {
      'innerBoundaryIs': GML2.prototype.innerBoundaryIsParser,
      'outerBoundaryIs': GML2.prototype.outerBoundaryIsParser,
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML2.prototype.BOX_PARSERS_ = {
    'http://www.opengis.net/gml': {
      'coordinates': makeArrayPusher(GML2.prototype.readFlatCoordinates),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML2.prototype.GEOMETRY_PARSERS = {
    'http://www.opengis.net/gml': {
      'Point': makeReplacer(GMLBase.prototype.readPoint),
      'MultiPoint': makeReplacer(GMLBase.prototype.readMultiPoint),
      'LineString': makeReplacer(GMLBase.prototype.readLineString),
      'MultiLineString': makeReplacer(GMLBase.prototype.readMultiLineString),
      'LinearRing': makeReplacer(GMLBase.prototype.readLinearRing),
      'Polygon': makeReplacer(GMLBase.prototype.readPolygon),
      'MultiPolygon': makeReplacer(GMLBase.prototype.readMultiPolygon),
      'Box': makeReplacer(GML2.prototype.readBox),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  GML2.prototype.GEOMETRY_SERIALIZERS = {
    'http://www.opengis.net/gml': {
      'Curve': makeChildAppender(GML2.prototype.writeCurveOrLineString),
      'MultiCurve': makeChildAppender(GML2.prototype.writeMultiCurveOrLineString),
      'Point': makeChildAppender(GML2.prototype.writePoint),
      'MultiPoint': makeChildAppender(GML2.prototype.writeMultiPoint),
      'LineString': makeChildAppender(GML2.prototype.writeCurveOrLineString),
      'MultiLineString': makeChildAppender(
        GML2.prototype.writeMultiCurveOrLineString
      ),
      'LinearRing': makeChildAppender(GML2.prototype.writeLinearRing),
      'Polygon': makeChildAppender(GML2.prototype.writeSurfaceOrPolygon),
      'MultiPolygon': makeChildAppender(
        GML2.prototype.writeMultiSurfaceOrPolygon
      ),
      'Surface': makeChildAppender(GML2.prototype.writeSurfaceOrPolygon),
      'MultiSurface': makeChildAppender(
        GML2.prototype.writeMultiSurfaceOrPolygon
      ),
      'Envelope': makeChildAppender(GML2.prototype.writeEnvelope),
    },
  };

  /**
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  GML2.prototype.LINESTRINGORCURVEMEMBER_SERIALIZERS = {
    'http://www.opengis.net/gml': {
      'lineStringMember': makeChildAppender(
        GML2.prototype.writeLineStringOrCurveMember
      ),
      'curveMember': makeChildAppender(
        GML2.prototype.writeLineStringOrCurveMember
      ),
    },
  };

  /**
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  GML2.prototype.RING_SERIALIZERS = {
    'http://www.opengis.net/gml': {
      'outerBoundaryIs': makeChildAppender(GML2.prototype.writeRing),
      'innerBoundaryIs': makeChildAppender(GML2.prototype.writeRing),
    },
  };

  /**
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  GML2.prototype.POINTMEMBER_SERIALIZERS = {
    'http://www.opengis.net/gml': {
      'pointMember': makeChildAppender(GML2.prototype.writePointMember),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  GML2.prototype.SURFACEORPOLYGONMEMBER_SERIALIZERS = {
    'http://www.opengis.net/gml': {
      'surfaceMember': makeChildAppender(
        GML2.prototype.writeSurfaceOrPolygonMember
      ),
      'polygonMember': makeChildAppender(
        GML2.prototype.writeSurfaceOrPolygonMember
      ),
    },
  };

  /**
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  GML2.prototype.ENVELOPE_SERIALIZERS = {
    'http://www.opengis.net/gml': {
      'lowerCorner': makeChildAppender(writeStringTextNode),
      'upperCorner': makeChildAppender(writeStringTextNode),
    },
  };

  /**
   * @module ol/format/GML3
   */

  /**
   * @const
   * @type {string}
   * @private
   */
  const schemaLocation =
    GMLNS +
    ' http://schemas.opengis.net/gml/3.1.1/profiles/gmlsfProfile/' +
    '1.0.0/gmlsf.xsd';

  /**
   * @const
   * @type {Object<string, string>}
   */
  const MULTIGEOMETRY_TO_MEMBER_NODENAME = {
    'MultiLineString': 'lineStringMember',
    'MultiCurve': 'curveMember',
    'MultiPolygon': 'polygonMember',
    'MultiSurface': 'surfaceMember',
  };

  /**
   * @classdesc
   * Feature format for reading and writing data in the GML format
   * version 3.1.1.
   * Currently only supports GML 3.1.1 Simple Features profile.
   *
   * @api
   */
  class GML3 extends GMLBase {
    /**
     * @param {import("./GMLBase.js").Options} [options] Optional configuration object.
     */
    constructor(options) {
      options = options ? options : {};

      super(options);

      /**
       * @private
       * @type {boolean}
       */
      this.surface_ = options.surface !== undefined ? options.surface : false;

      /**
       * @private
       * @type {boolean}
       */
      this.curve_ = options.curve !== undefined ? options.curve : false;

      /**
       * @private
       * @type {boolean}
       */
      this.multiCurve_ =
        options.multiCurve !== undefined ? options.multiCurve : true;

      /**
       * @private
       * @type {boolean}
       */
      this.multiSurface_ =
        options.multiSurface !== undefined ? options.multiSurface : true;

      /**
       * @type {string}
       */
      this.schemaLocation = options.schemaLocation
        ? options.schemaLocation
        : schemaLocation;

      /**
       * @private
       * @type {boolean}
       */
      this.hasZ = options.hasZ !== undefined ? options.hasZ : false;
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {MultiLineString|undefined} MultiLineString.
     */
    readMultiCurve(node, objectStack) {
      /** @type {Array<LineString>} */
      const lineStrings = pushParseAndPop(
        [],
        this.MULTICURVE_PARSERS,
        node,
        objectStack,
        this
      );
      if (lineStrings) {
        const multiLineString = new MultiLineString(lineStrings);
        return multiLineString;
      } else {
        return undefined;
      }
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {Array<number>|undefined} Polygon.
     */
    readFlatCurveRing(node, objectStack) {
      /** @type {Array<LineString>} */
      const lineStrings = pushParseAndPop(
        [],
        this.MULTICURVE_PARSERS,
        node,
        objectStack,
        this
      );
      const flatCoordinates = [];
      for (let i = 0, ii = lineStrings.length; i < ii; ++i) {
        extend$2(flatCoordinates, lineStrings[i].getFlatCoordinates());
      }
      return flatCoordinates;
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {MultiPolygon|undefined} MultiPolygon.
     */
    readMultiSurface(node, objectStack) {
      /** @type {Array<Polygon>} */
      const polygons = pushParseAndPop(
        [],
        this.MULTISURFACE_PARSERS,
        node,
        objectStack,
        this
      );
      if (polygons) {
        return new MultiPolygon(polygons);
      }
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     */
    curveMemberParser(node, objectStack) {
      parseNode(this.CURVEMEMBER_PARSERS, node, objectStack, this);
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     */
    surfaceMemberParser(node, objectStack) {
      parseNode(this.SURFACEMEMBER_PARSERS, node, objectStack, this);
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {Array<(Array<number>)>|undefined} flat coordinates.
     */
    readPatch(node, objectStack) {
      return pushParseAndPop(
        [null],
        this.PATCHES_PARSERS,
        node,
        objectStack,
        this
      );
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {Array<number>|undefined} flat coordinates.
     */
    readSegment(node, objectStack) {
      return pushParseAndPop([], this.SEGMENTS_PARSERS, node, objectStack, this);
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {Array<(Array<number>)>|undefined} flat coordinates.
     */
    readPolygonPatch(node, objectStack) {
      return pushParseAndPop(
        [null],
        this.FLAT_LINEAR_RINGS_PARSERS,
        node,
        objectStack,
        this
      );
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {Array<number>|undefined} flat coordinates.
     */
    readLineStringSegment(node, objectStack) {
      return pushParseAndPop(
        [null],
        this.GEOMETRY_FLAT_COORDINATES_PARSERS,
        node,
        objectStack,
        this
      );
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     */
    interiorParser(node, objectStack) {
      /** @type {Array<number>|undefined} */
      const flatLinearRing = pushParseAndPop(
        undefined,
        this.RING_PARSERS,
        node,
        objectStack,
        this
      );
      if (flatLinearRing) {
        const flatLinearRings =
          /** @type {Array<Array<number>>} */
          (objectStack[objectStack.length - 1]);
        flatLinearRings.push(flatLinearRing);
      }
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     */
    exteriorParser(node, objectStack) {
      /** @type {Array<number>|undefined} */
      const flatLinearRing = pushParseAndPop(
        undefined,
        this.RING_PARSERS,
        node,
        objectStack,
        this
      );
      if (flatLinearRing) {
        const flatLinearRings =
          /** @type {Array<Array<number>>} */
          (objectStack[objectStack.length - 1]);
        flatLinearRings[0] = flatLinearRing;
      }
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {Polygon|undefined} Polygon.
     */
    readSurface(node, objectStack) {
      /** @type {Array<Array<number>>} */
      const flatLinearRings = pushParseAndPop(
        [null],
        this.SURFACE_PARSERS,
        node,
        objectStack,
        this
      );
      if (flatLinearRings && flatLinearRings[0]) {
        const flatCoordinates = flatLinearRings[0];
        const ends = [flatCoordinates.length];
        let i, ii;
        for (i = 1, ii = flatLinearRings.length; i < ii; ++i) {
          extend$2(flatCoordinates, flatLinearRings[i]);
          ends.push(flatCoordinates.length);
        }
        return new Polygon(flatCoordinates, 'XYZ', ends);
      } else {
        return undefined;
      }
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {LineString|undefined} LineString.
     */
    readCurve(node, objectStack) {
      /** @type {Array<number>} */
      const flatCoordinates = pushParseAndPop(
        [null],
        this.CURVE_PARSERS,
        node,
        objectStack,
        this
      );
      if (flatCoordinates) {
        const lineString = new LineString(flatCoordinates, 'XYZ');
        return lineString;
      } else {
        return undefined;
      }
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {import("../extent.js").Extent|undefined} Envelope.
     */
    readEnvelope(node, objectStack) {
      /** @type {Array<number>} */
      const flatCoordinates = pushParseAndPop(
        [null],
        this.ENVELOPE_PARSERS,
        node,
        objectStack,
        this
      );
      return createOrUpdate$2(
        flatCoordinates[1][0],
        flatCoordinates[1][1],
        flatCoordinates[2][0],
        flatCoordinates[2][1]
      );
    }

    /**
     * @param {Node} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {Array<number>|undefined} Flat coordinates.
     */
    readFlatPos(node, objectStack) {
      let s = getAllTextContent(node, false);
      const re = /^\s*([+\-]?\d*\.?\d+(?:[eE][+\-]?\d+)?)\s*/;
      /** @type {Array<number>} */
      const flatCoordinates = [];
      let m;
      while ((m = re.exec(s))) {
        flatCoordinates.push(parseFloat(m[1]));
        s = s.substr(m[0].length);
      }
      if (s !== '') {
        return undefined;
      }
      const context = objectStack[0];
      const containerSrs = context['srsName'];
      let axisOrientation = 'enu';
      if (containerSrs) {
        const proj = get$2(containerSrs);
        axisOrientation = proj.getAxisOrientation();
      }
      if (axisOrientation === 'neu') {
        let i, ii;
        for (i = 0, ii = flatCoordinates.length; i < ii; i += 3) {
          const y = flatCoordinates[i];
          const x = flatCoordinates[i + 1];
          flatCoordinates[i] = x;
          flatCoordinates[i + 1] = y;
        }
      }
      const len = flatCoordinates.length;
      if (len == 2) {
        flatCoordinates.push(0);
      }
      if (len === 0) {
        return undefined;
      }
      return flatCoordinates;
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {Array<number>|undefined} Flat coordinates.
     */
    readFlatPosList(node, objectStack) {
      const s = getAllTextContent(node, false).replace(/^\s*|\s*$/g, '');
      const context = objectStack[0];
      const containerSrs = context['srsName'];
      const contextDimension = context['srsDimension'];
      let axisOrientation = 'enu';
      if (containerSrs) {
        const proj = get$2(containerSrs);
        axisOrientation = proj.getAxisOrientation();
      }
      const coords = s.split(/\s+/);
      // The "dimension" attribute is from the GML 3.0.1 spec.
      let dim = 2;
      if (node.getAttribute('srsDimension')) {
        dim = readNonNegativeIntegerString(node.getAttribute('srsDimension'));
      } else if (node.getAttribute('dimension')) {
        dim = readNonNegativeIntegerString(node.getAttribute('dimension'));
      } else if (
        /** @type {Element} */ (node.parentNode).getAttribute('srsDimension')
      ) {
        dim = readNonNegativeIntegerString(
          /** @type {Element} */ (node.parentNode).getAttribute('srsDimension')
        );
      } else if (contextDimension) {
        dim = readNonNegativeIntegerString(contextDimension);
      }
      let x, y, z;
      const flatCoordinates = [];
      for (let i = 0, ii = coords.length; i < ii; i += dim) {
        x = parseFloat(coords[i]);
        y = parseFloat(coords[i + 1]);
        z = dim === 3 ? parseFloat(coords[i + 2]) : 0;
        if (axisOrientation.substr(0, 2) === 'en') {
          flatCoordinates.push(x, y, z);
        } else {
          flatCoordinates.push(y, x, z);
        }
      }
      return flatCoordinates;
    }

    /**
     * @param {Element} node Node.
     * @param {import("../geom/Point.js").default} value Point geometry.
     * @param {Array<*>} objectStack Node stack.
     * @private
     */
    writePos_(node, value, objectStack) {
      const context = objectStack[objectStack.length - 1];
      const hasZ = context['hasZ'];
      const srsDimension = hasZ ? '3' : '2';
      node.setAttribute('srsDimension', srsDimension);
      const srsName = context['srsName'];
      let axisOrientation = 'enu';
      if (srsName) {
        axisOrientation = get$2(srsName).getAxisOrientation();
      }
      const point = value.getCoordinates();
      let coords;
      // only 2d for simple features profile
      if (axisOrientation.substr(0, 2) === 'en') {
        coords = point[0] + ' ' + point[1];
      } else {
        coords = point[1] + ' ' + point[0];
      }
      if (hasZ) {
        // For newly created points, Z can be undefined.
        const z = point[2] || 0;
        coords += ' ' + z;
      }
      writeStringTextNode(node, coords);
    }

    /**
     * @param {Array<number>} point Point geometry.
     * @param {string} [srsName] Optional srsName
     * @param {boolean} [hasZ] whether the geometry has a Z coordinate (is 3D) or not.
     * @return {string} The coords string.
     * @private
     */
    getCoords_(point, srsName, hasZ) {
      let axisOrientation = 'enu';
      if (srsName) {
        axisOrientation = get$2(srsName).getAxisOrientation();
      }
      let coords =
        axisOrientation.substr(0, 2) === 'en'
          ? point[0] + ' ' + point[1]
          : point[1] + ' ' + point[0];
      if (hasZ) {
        // For newly created points, Z can be undefined.
        const z = point[2] || 0;
        coords += ' ' + z;
      }

      return coords;
    }

    /**
     * @param {Element} node Node.
     * @param {LineString|import("../geom/LinearRing.js").default} value Geometry.
     * @param {Array<*>} objectStack Node stack.
     * @private
     */
    writePosList_(node, value, objectStack) {
      const context = objectStack[objectStack.length - 1];
      const hasZ = context['hasZ'];
      const srsDimension = hasZ ? '3' : '2';
      node.setAttribute('srsDimension', srsDimension);
      const srsName = context['srsName'];
      // only 2d for simple features profile
      const points = value.getCoordinates();
      const len = points.length;
      const parts = new Array(len);
      let point;
      for (let i = 0; i < len; ++i) {
        point = points[i];
        parts[i] = this.getCoords_(point, srsName, hasZ);
      }
      writeStringTextNode(node, parts.join(' '));
    }

    /**
     * @param {Element} node Node.
     * @param {import("../geom/Point.js").default} geometry Point geometry.
     * @param {Array<*>} objectStack Node stack.
     */
    writePoint(node, geometry, objectStack) {
      const context = objectStack[objectStack.length - 1];
      const srsName = context['srsName'];
      if (srsName) {
        node.setAttribute('srsName', srsName);
      }
      const pos = createElementNS(node.namespaceURI, 'pos');
      node.appendChild(pos);
      this.writePos_(pos, geometry, objectStack);
    }

    /**
     * @param {Element} node Node.
     * @param {import("../extent.js").Extent} extent Extent.
     * @param {Array<*>} objectStack Node stack.
     */
    writeEnvelope(node, extent, objectStack) {
      const context = objectStack[objectStack.length - 1];
      const srsName = context['srsName'];
      if (srsName) {
        node.setAttribute('srsName', srsName);
      }
      const keys = ['lowerCorner', 'upperCorner'];
      const values = [extent[0] + ' ' + extent[1], extent[2] + ' ' + extent[3]];
      pushSerializeAndPop(
        /** @type {import("../xml.js").NodeStackItem} */
        ({node: node}),
        this.ENVELOPE_SERIALIZERS,
        OBJECT_PROPERTY_NODE_FACTORY,
        values,
        objectStack,
        keys,
        this
      );
    }

    /**
     * @param {Element} node Node.
     * @param {import("../geom/LinearRing.js").default} geometry LinearRing geometry.
     * @param {Array<*>} objectStack Node stack.
     */
    writeLinearRing(node, geometry, objectStack) {
      const context = objectStack[objectStack.length - 1];
      const srsName = context['srsName'];
      if (srsName) {
        node.setAttribute('srsName', srsName);
      }
      const posList = createElementNS(node.namespaceURI, 'posList');
      node.appendChild(posList);
      this.writePosList_(posList, geometry, objectStack);
    }

    /**
     * @param {*} value Value.
     * @param {Array<*>} objectStack Object stack.
     * @param {string} [nodeName] Node name.
     * @return {Node} Node.
     * @private
     */
    RING_NODE_FACTORY_(value, objectStack, nodeName) {
      const context = objectStack[objectStack.length - 1];
      const parentNode = context.node;
      const exteriorWritten = context['exteriorWritten'];
      if (exteriorWritten === undefined) {
        context['exteriorWritten'] = true;
      }
      return createElementNS(
        parentNode.namespaceURI,
        exteriorWritten !== undefined ? 'interior' : 'exterior'
      );
    }

    /**
     * @param {Element} node Node.
     * @param {Polygon} geometry Polygon geometry.
     * @param {Array<*>} objectStack Node stack.
     */
    writeSurfaceOrPolygon(node, geometry, objectStack) {
      const context = objectStack[objectStack.length - 1];
      const hasZ = context['hasZ'];
      const srsName = context['srsName'];
      if (node.nodeName !== 'PolygonPatch' && srsName) {
        node.setAttribute('srsName', srsName);
      }
      if (node.nodeName === 'Polygon' || node.nodeName === 'PolygonPatch') {
        const rings = geometry.getLinearRings();
        pushSerializeAndPop(
          {node: node, hasZ: hasZ, srsName: srsName},
          this.RING_SERIALIZERS,
          this.RING_NODE_FACTORY_,
          rings,
          objectStack,
          undefined,
          this
        );
      } else if (node.nodeName === 'Surface') {
        const patches = createElementNS(node.namespaceURI, 'patches');
        node.appendChild(patches);
        this.writeSurfacePatches_(patches, geometry, objectStack);
      }
    }

    /**
     * @param {Element} node Node.
     * @param {LineString} geometry LineString geometry.
     * @param {Array<*>} objectStack Node stack.
     */
    writeCurveOrLineString(node, geometry, objectStack) {
      const context = objectStack[objectStack.length - 1];
      const srsName = context['srsName'];
      if (node.nodeName !== 'LineStringSegment' && srsName) {
        node.setAttribute('srsName', srsName);
      }
      if (
        node.nodeName === 'LineString' ||
        node.nodeName === 'LineStringSegment'
      ) {
        const posList = createElementNS(node.namespaceURI, 'posList');
        node.appendChild(posList);
        this.writePosList_(posList, geometry, objectStack);
      } else if (node.nodeName === 'Curve') {
        const segments = createElementNS(node.namespaceURI, 'segments');
        node.appendChild(segments);
        this.writeCurveSegments_(segments, geometry, objectStack);
      }
    }

    /**
     * @param {Element} node Node.
     * @param {MultiPolygon} geometry MultiPolygon geometry.
     * @param {Array<*>} objectStack Node stack.
     */
    writeMultiSurfaceOrPolygon(node, geometry, objectStack) {
      const context = objectStack[objectStack.length - 1];
      const hasZ = context['hasZ'];
      const srsName = context['srsName'];
      const surface = context['surface'];
      if (srsName) {
        node.setAttribute('srsName', srsName);
      }
      const polygons = geometry.getPolygons();
      pushSerializeAndPop(
        {node: node, hasZ: hasZ, srsName: srsName, surface: surface},
        this.SURFACEORPOLYGONMEMBER_SERIALIZERS,
        this.MULTIGEOMETRY_MEMBER_NODE_FACTORY_,
        polygons,
        objectStack,
        undefined,
        this
      );
    }

    /**
     * @param {Element} node Node.
     * @param {import("../geom/MultiPoint.js").default} geometry MultiPoint geometry.
     * @param {Array<*>} objectStack Node stack.
     */
    writeMultiPoint(node, geometry, objectStack) {
      const context = objectStack[objectStack.length - 1];
      const srsName = context['srsName'];
      const hasZ = context['hasZ'];
      if (srsName) {
        node.setAttribute('srsName', srsName);
      }
      const points = geometry.getPoints();
      pushSerializeAndPop(
        {node: node, hasZ: hasZ, srsName: srsName},
        this.POINTMEMBER_SERIALIZERS,
        makeSimpleNodeFactory('pointMember'),
        points,
        objectStack,
        undefined,
        this
      );
    }

    /**
     * @param {Element} node Node.
     * @param {MultiLineString} geometry MultiLineString geometry.
     * @param {Array<*>} objectStack Node stack.
     */
    writeMultiCurveOrLineString(node, geometry, objectStack) {
      const context = objectStack[objectStack.length - 1];
      const hasZ = context['hasZ'];
      const srsName = context['srsName'];
      const curve = context['curve'];
      if (srsName) {
        node.setAttribute('srsName', srsName);
      }
      const lines = geometry.getLineStrings();
      pushSerializeAndPop(
        {node: node, hasZ: hasZ, srsName: srsName, curve: curve},
        this.LINESTRINGORCURVEMEMBER_SERIALIZERS,
        this.MULTIGEOMETRY_MEMBER_NODE_FACTORY_,
        lines,
        objectStack,
        undefined,
        this
      );
    }

    /**
     * @param {Element} node Node.
     * @param {import("../geom/LinearRing.js").default} ring LinearRing geometry.
     * @param {Array<*>} objectStack Node stack.
     */
    writeRing(node, ring, objectStack) {
      const linearRing = createElementNS(node.namespaceURI, 'LinearRing');
      node.appendChild(linearRing);
      this.writeLinearRing(linearRing, ring, objectStack);
    }

    /**
     * @param {Node} node Node.
     * @param {Polygon} polygon Polygon geometry.
     * @param {Array<*>} objectStack Node stack.
     */
    writeSurfaceOrPolygonMember(node, polygon, objectStack) {
      const child = this.GEOMETRY_NODE_FACTORY_(polygon, objectStack);
      if (child) {
        node.appendChild(child);
        this.writeSurfaceOrPolygon(child, polygon, objectStack);
      }
    }

    /**
     * @param {Element} node Node.
     * @param {import("../geom/Point.js").default} point Point geometry.
     * @param {Array<*>} objectStack Node stack.
     */
    writePointMember(node, point, objectStack) {
      const child = createElementNS(node.namespaceURI, 'Point');
      node.appendChild(child);
      this.writePoint(child, point, objectStack);
    }

    /**
     * @param {Node} node Node.
     * @param {LineString} line LineString geometry.
     * @param {Array<*>} objectStack Node stack.
     */
    writeLineStringOrCurveMember(node, line, objectStack) {
      const child = this.GEOMETRY_NODE_FACTORY_(line, objectStack);
      if (child) {
        node.appendChild(child);
        this.writeCurveOrLineString(child, line, objectStack);
      }
    }

    /**
     * @param {Element} node Node.
     * @param {Polygon} polygon Polygon geometry.
     * @param {Array<*>} objectStack Node stack.
     * @private
     */
    writeSurfacePatches_(node, polygon, objectStack) {
      const child = createElementNS(node.namespaceURI, 'PolygonPatch');
      node.appendChild(child);
      this.writeSurfaceOrPolygon(child, polygon, objectStack);
    }

    /**
     * @param {Element} node Node.
     * @param {LineString} line LineString geometry.
     * @param {Array<*>} objectStack Node stack.
     * @private
     */
    writeCurveSegments_(node, line, objectStack) {
      const child = createElementNS(node.namespaceURI, 'LineStringSegment');
      node.appendChild(child);
      this.writeCurveOrLineString(child, line, objectStack);
    }

    /**
     * @param {Node} node Node.
     * @param {import("../geom/Geometry.js").default|import("../extent.js").Extent} geometry Geometry.
     * @param {Array<*>} objectStack Node stack.
     */
    writeGeometryElement(node, geometry, objectStack) {
      const context = /** @type {import("./Feature.js").WriteOptions} */ (
        objectStack[objectStack.length - 1]
      );
      const item = Object.assign({}, context);
      item['node'] = node;
      let value;
      if (Array.isArray(geometry)) {
        value = transformExtentWithOptions(
          /** @type {import("../extent.js").Extent} */ (geometry),
          context
        );
      } else {
        value = transformGeometryWithOptions(
          /** @type {import("../geom/Geometry.js").default} */ (geometry),
          true,
          context
        );
      }
      pushSerializeAndPop(
        /** @type {import("../xml.js").NodeStackItem} */
        (item),
        this.GEOMETRY_SERIALIZERS,
        this.GEOMETRY_NODE_FACTORY_,
        [value],
        objectStack,
        undefined,
        this
      );
    }

    /**
     * @param {Element} node Node.
     * @param {import("../Feature.js").default} feature Feature.
     * @param {Array<*>} objectStack Node stack.
     */
    writeFeatureElement(node, feature, objectStack) {
      const fid = feature.getId();
      if (fid) {
        node.setAttribute('fid', /** @type {string} */ (fid));
      }
      const context = /** @type {Object} */ (objectStack[objectStack.length - 1]);
      const featureNS = context['featureNS'];
      const geometryName = feature.getGeometryName();
      if (!context.serializers) {
        context.serializers = {};
        context.serializers[featureNS] = {};
      }
      const keys = [];
      const values = [];
      if (feature.hasProperties()) {
        const properties = feature.getProperties();
        for (const key in properties) {
          const value = properties[key];
          if (value !== null) {
            keys.push(key);
            values.push(value);
            if (
              key == geometryName ||
              typeof (/** @type {?} */ (value).getSimplifiedGeometry) ===
                'function'
            ) {
              if (!(key in context.serializers[featureNS])) {
                context.serializers[featureNS][key] = makeChildAppender(
                  this.writeGeometryElement,
                  this
                );
              }
            } else {
              if (!(key in context.serializers[featureNS])) {
                context.serializers[featureNS][key] =
                  makeChildAppender(writeStringTextNode);
              }
            }
          }
        }
      }
      const item = Object.assign({}, context);
      item.node = node;
      pushSerializeAndPop(
        /** @type {import("../xml.js").NodeStackItem} */
        (item),
        context.serializers,
        makeSimpleNodeFactory(undefined, featureNS),
        values,
        objectStack,
        keys
      );
    }

    /**
     * @param {Node} node Node.
     * @param {Array<import("../Feature.js").default>} features Features.
     * @param {Array<*>} objectStack Node stack.
     * @private
     */
    writeFeatureMembers_(node, features, objectStack) {
      const context = /** @type {Object} */ (objectStack[objectStack.length - 1]);
      const featureType = context['featureType'];
      const featureNS = context['featureNS'];
      /** @type {Object<string, Object<string, import("../xml.js").Serializer>>} */
      const serializers = {};
      serializers[featureNS] = {};
      serializers[featureNS][featureType] = makeChildAppender(
        this.writeFeatureElement,
        this
      );
      const item = Object.assign({}, context);
      item.node = node;
      pushSerializeAndPop(
        /** @type {import("../xml.js").NodeStackItem} */
        (item),
        serializers,
        makeSimpleNodeFactory(featureType, featureNS),
        features,
        objectStack
      );
    }

    /**
     * @const
     * @param {*} value Value.
     * @param {Array<*>} objectStack Object stack.
     * @param {string} [nodeName] Node name.
     * @return {Node|undefined} Node.
     * @private
     */
    MULTIGEOMETRY_MEMBER_NODE_FACTORY_(value, objectStack, nodeName) {
      const parentNode = objectStack[objectStack.length - 1].node;
      return createElementNS(
        this.namespace,
        MULTIGEOMETRY_TO_MEMBER_NODENAME[parentNode.nodeName]
      );
    }

    /**
     * @const
     * @param {*} value Value.
     * @param {Array<*>} objectStack Object stack.
     * @param {string} [nodeName] Node name.
     * @return {Element|undefined} Node.
     * @private
     */
    GEOMETRY_NODE_FACTORY_(value, objectStack, nodeName) {
      const context = objectStack[objectStack.length - 1];
      const multiSurface = context['multiSurface'];
      const surface = context['surface'];
      const curve = context['curve'];
      const multiCurve = context['multiCurve'];
      if (!Array.isArray(value)) {
        nodeName = /** @type {import("../geom/Geometry.js").default} */ (
          value
        ).getType();
        if (nodeName === 'MultiPolygon' && multiSurface === true) {
          nodeName = 'MultiSurface';
        } else if (nodeName === 'Polygon' && surface === true) {
          nodeName = 'Surface';
        } else if (nodeName === 'LineString' && curve === true) {
          nodeName = 'Curve';
        } else if (nodeName === 'MultiLineString' && multiCurve === true) {
          nodeName = 'MultiCurve';
        }
      } else {
        nodeName = 'Envelope';
      }
      return createElementNS(this.namespace, nodeName);
    }

    /**
     * Encode a geometry in GML 3.1.1 Simple Features.
     *
     * @param {import("../geom/Geometry.js").default} geometry Geometry.
     * @param {import("./Feature.js").WriteOptions} [options] Options.
     * @return {Node} Node.
     * @api
     */
    writeGeometryNode(geometry, options) {
      options = this.adaptOptions(options);
      const geom = createElementNS(this.namespace, 'geom');
      const context = {
        node: geom,
        hasZ: this.hasZ,
        srsName: this.srsName,
        curve: this.curve_,
        surface: this.surface_,
        multiSurface: this.multiSurface_,
        multiCurve: this.multiCurve_,
      };
      if (options) {
        Object.assign(context, options);
      }
      this.writeGeometryElement(geom, geometry, [context]);
      return geom;
    }

    /**
     * Encode an array of features in the GML 3.1.1 format as an XML node.
     *
     * @param {Array<import("../Feature.js").default>} features Features.
     * @param {import("./Feature.js").WriteOptions} [options] Options.
     * @return {Element} Node.
     * @api
     */
    writeFeaturesNode(features, options) {
      options = this.adaptOptions(options);
      const node = createElementNS(this.namespace, 'featureMembers');
      node.setAttributeNS(
        XML_SCHEMA_INSTANCE_URI,
        'xsi:schemaLocation',
        this.schemaLocation
      );
      const context = {
        srsName: this.srsName,
        hasZ: this.hasZ,
        curve: this.curve_,
        surface: this.surface_,
        multiSurface: this.multiSurface_,
        multiCurve: this.multiCurve_,
        featureNS: this.featureNS,
        featureType: this.featureType,
      };
      if (options) {
        Object.assign(context, options);
      }
      this.writeFeatureMembers_(node, features, [context]);
      return node;
    }
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML3.prototype.GEOMETRY_FLAT_COORDINATES_PARSERS = {
    'http://www.opengis.net/gml': {
      'pos': makeReplacer(GML3.prototype.readFlatPos),
      'posList': makeReplacer(GML3.prototype.readFlatPosList),
      'coordinates': makeReplacer(GML2.prototype.readFlatCoordinates),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML3.prototype.FLAT_LINEAR_RINGS_PARSERS = {
    'http://www.opengis.net/gml': {
      'interior': GML3.prototype.interiorParser,
      'exterior': GML3.prototype.exteriorParser,
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML3.prototype.GEOMETRY_PARSERS = {
    'http://www.opengis.net/gml': {
      'Point': makeReplacer(GMLBase.prototype.readPoint),
      'MultiPoint': makeReplacer(GMLBase.prototype.readMultiPoint),
      'LineString': makeReplacer(GMLBase.prototype.readLineString),
      'MultiLineString': makeReplacer(GMLBase.prototype.readMultiLineString),
      'LinearRing': makeReplacer(GMLBase.prototype.readLinearRing),
      'Polygon': makeReplacer(GMLBase.prototype.readPolygon),
      'MultiPolygon': makeReplacer(GMLBase.prototype.readMultiPolygon),
      'Surface': makeReplacer(GML3.prototype.readSurface),
      'MultiSurface': makeReplacer(GML3.prototype.readMultiSurface),
      'Curve': makeReplacer(GML3.prototype.readCurve),
      'MultiCurve': makeReplacer(GML3.prototype.readMultiCurve),
      'Envelope': makeReplacer(GML3.prototype.readEnvelope),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML3.prototype.MULTICURVE_PARSERS = {
    'http://www.opengis.net/gml': {
      'curveMember': makeArrayPusher(GML3.prototype.curveMemberParser),
      'curveMembers': makeArrayPusher(GML3.prototype.curveMemberParser),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML3.prototype.MULTISURFACE_PARSERS = {
    'http://www.opengis.net/gml': {
      'surfaceMember': makeArrayPusher(GML3.prototype.surfaceMemberParser),
      'surfaceMembers': makeArrayPusher(GML3.prototype.surfaceMemberParser),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML3.prototype.CURVEMEMBER_PARSERS = {
    'http://www.opengis.net/gml': {
      'LineString': makeArrayPusher(GMLBase.prototype.readLineString),
      'Curve': makeArrayPusher(GML3.prototype.readCurve),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML3.prototype.SURFACEMEMBER_PARSERS = {
    'http://www.opengis.net/gml': {
      'Polygon': makeArrayPusher(GMLBase.prototype.readPolygon),
      'Surface': makeArrayPusher(GML3.prototype.readSurface),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML3.prototype.SURFACE_PARSERS = {
    'http://www.opengis.net/gml': {
      'patches': makeReplacer(GML3.prototype.readPatch),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML3.prototype.CURVE_PARSERS = {
    'http://www.opengis.net/gml': {
      'segments': makeReplacer(GML3.prototype.readSegment),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML3.prototype.ENVELOPE_PARSERS = {
    'http://www.opengis.net/gml': {
      'lowerCorner': makeArrayPusher(GML3.prototype.readFlatPosList),
      'upperCorner': makeArrayPusher(GML3.prototype.readFlatPosList),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML3.prototype.PATCHES_PARSERS = {
    'http://www.opengis.net/gml': {
      'PolygonPatch': makeReplacer(GML3.prototype.readPolygonPatch),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML3.prototype.SEGMENTS_PARSERS = {
    'http://www.opengis.net/gml': {
      'LineStringSegment': makeArrayExtender(
        GML3.prototype.readLineStringSegment
      ),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GMLBase.prototype.RING_PARSERS = {
    'http://www.opengis.net/gml': {
      'LinearRing': makeReplacer(GMLBase.prototype.readFlatLinearRing),
      'Ring': makeReplacer(GML3.prototype.readFlatCurveRing),
    },
  };

  /**
   * Encode an array of features in GML 3.1.1 Simple Features.
   *
   * @function
   * @param {Array<import("../Feature.js").default>} features Features.
   * @param {import("./Feature.js").WriteOptions} [options] Options.
   * @return {string} Result.
   * @api
   */
  GML3.prototype.writeFeatures;

  /**
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  GML3.prototype.RING_SERIALIZERS = {
    'http://www.opengis.net/gml': {
      'exterior': makeChildAppender(GML3.prototype.writeRing),
      'interior': makeChildAppender(GML3.prototype.writeRing),
    },
  };

  /**
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  GML3.prototype.ENVELOPE_SERIALIZERS = {
    'http://www.opengis.net/gml': {
      'lowerCorner': makeChildAppender(writeStringTextNode),
      'upperCorner': makeChildAppender(writeStringTextNode),
    },
  };

  /**
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  GML3.prototype.SURFACEORPOLYGONMEMBER_SERIALIZERS = {
    'http://www.opengis.net/gml': {
      'surfaceMember': makeChildAppender(
        GML3.prototype.writeSurfaceOrPolygonMember
      ),
      'polygonMember': makeChildAppender(
        GML3.prototype.writeSurfaceOrPolygonMember
      ),
    },
  };

  /**
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  GML3.prototype.POINTMEMBER_SERIALIZERS = {
    'http://www.opengis.net/gml': {
      'pointMember': makeChildAppender(GML3.prototype.writePointMember),
    },
  };

  /**
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  GML3.prototype.LINESTRINGORCURVEMEMBER_SERIALIZERS = {
    'http://www.opengis.net/gml': {
      'lineStringMember': makeChildAppender(
        GML3.prototype.writeLineStringOrCurveMember
      ),
      'curveMember': makeChildAppender(
        GML3.prototype.writeLineStringOrCurveMember
      ),
    },
  };

  /**
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  GML3.prototype.GEOMETRY_SERIALIZERS = {
    'http://www.opengis.net/gml': {
      'Curve': makeChildAppender(GML3.prototype.writeCurveOrLineString),
      'MultiCurve': makeChildAppender(GML3.prototype.writeMultiCurveOrLineString),
      'Point': makeChildAppender(GML3.prototype.writePoint),
      'MultiPoint': makeChildAppender(GML3.prototype.writeMultiPoint),
      'LineString': makeChildAppender(GML3.prototype.writeCurveOrLineString),
      'MultiLineString': makeChildAppender(
        GML3.prototype.writeMultiCurveOrLineString
      ),
      'LinearRing': makeChildAppender(GML3.prototype.writeLinearRing),
      'Polygon': makeChildAppender(GML3.prototype.writeSurfaceOrPolygon),
      'MultiPolygon': makeChildAppender(
        GML3.prototype.writeMultiSurfaceOrPolygon
      ),
      'Surface': makeChildAppender(GML3.prototype.writeSurfaceOrPolygon),
      'MultiSurface': makeChildAppender(
        GML3.prototype.writeMultiSurfaceOrPolygon
      ),
      'Envelope': makeChildAppender(GML3.prototype.writeEnvelope),
    },
  };

  /**
   * @module ol/format/GML
   */

  /**
   * Feature format for reading and writing data in the GML format
   * version 3.1.1.
   * Currently only supports GML 3.1.1 Simple Features profile.
   *
   * @class
   * @param {import("./GMLBase.js").Options} [options]
   *     Optional configuration object.
   * @api
   */
  const GML = GML3;

  /**
   * Encode an array of features in GML 3.1.1 Simple Features.
   *
   * @function
   * @param {Array<import("../Feature.js").default>} features Features.
   * @param {import("./Feature.js").WriteOptions} [options] Options.
   * @return {string} Result.
   * @api
   */
  GML.prototype.writeFeatures;

  /**
   * Encode an array of features in the GML 3.1.1 format as an XML node.
   *
   * @function
   * @param {Array<import("../Feature.js").default>} features Features.
   * @param {import("./Feature.js").WriteOptions} [options] Options.
   * @return {Node} Node.
   * @api
   */
  GML.prototype.writeFeaturesNode;

  /**
   * @module ol/format/GML32
   */

  /**
   * @classdesc Feature format for reading and writing data in the GML format
   *            version 3.2.1.
   * @api
   */
  class GML32 extends GML3 {
    /**
     * @param {import("./GMLBase.js").Options} [options] Optional configuration object.
     */
    constructor(options) {
      options = options ? options : {};

      super(options);

      /**
       * @type {string}
       */
      this.schemaLocation = options.schemaLocation
        ? options.schemaLocation
        : this.namespace + ' http://schemas.opengis.net/gml/3.2.1/gml.xsd';
    }
  }

  GML32.prototype.namespace = 'http://www.opengis.net/gml/3.2';

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML32.prototype.GEOMETRY_FLAT_COORDINATES_PARSERS = {
    'http://www.opengis.net/gml/3.2': {
      'pos': makeReplacer(GML3.prototype.readFlatPos),
      'posList': makeReplacer(GML3.prototype.readFlatPosList),
      'coordinates': makeReplacer(GML2.prototype.readFlatCoordinates),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML32.prototype.FLAT_LINEAR_RINGS_PARSERS = {
    'http://www.opengis.net/gml/3.2': {
      'interior': GML3.prototype.interiorParser,
      'exterior': GML3.prototype.exteriorParser,
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML32.prototype.GEOMETRY_PARSERS = {
    'http://www.opengis.net/gml/3.2': {
      'Point': makeReplacer(GMLBase.prototype.readPoint),
      'MultiPoint': makeReplacer(GMLBase.prototype.readMultiPoint),
      'LineString': makeReplacer(GMLBase.prototype.readLineString),
      'MultiLineString': makeReplacer(GMLBase.prototype.readMultiLineString),
      'LinearRing': makeReplacer(GMLBase.prototype.readLinearRing),
      'Polygon': makeReplacer(GMLBase.prototype.readPolygon),
      'MultiPolygon': makeReplacer(GMLBase.prototype.readMultiPolygon),
      'Surface': makeReplacer(GML32.prototype.readSurface),
      'MultiSurface': makeReplacer(GML3.prototype.readMultiSurface),
      'Curve': makeReplacer(GML32.prototype.readCurve),
      'MultiCurve': makeReplacer(GML3.prototype.readMultiCurve),
      'Envelope': makeReplacer(GML32.prototype.readEnvelope),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML32.prototype.MULTICURVE_PARSERS = {
    'http://www.opengis.net/gml/3.2': {
      'curveMember': makeArrayPusher(GML3.prototype.curveMemberParser),
      'curveMembers': makeArrayPusher(GML3.prototype.curveMemberParser),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML32.prototype.MULTISURFACE_PARSERS = {
    'http://www.opengis.net/gml/3.2': {
      'surfaceMember': makeArrayPusher(GML3.prototype.surfaceMemberParser),
      'surfaceMembers': makeArrayPusher(GML3.prototype.surfaceMemberParser),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML32.prototype.CURVEMEMBER_PARSERS = {
    'http://www.opengis.net/gml/3.2': {
      'LineString': makeArrayPusher(GMLBase.prototype.readLineString),
      'Curve': makeArrayPusher(GML3.prototype.readCurve),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML32.prototype.SURFACEMEMBER_PARSERS = {
    'http://www.opengis.net/gml/3.2': {
      'Polygon': makeArrayPusher(GMLBase.prototype.readPolygon),
      'Surface': makeArrayPusher(GML3.prototype.readSurface),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML32.prototype.SURFACE_PARSERS = {
    'http://www.opengis.net/gml/3.2': {
      'patches': makeReplacer(GML3.prototype.readPatch),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML32.prototype.CURVE_PARSERS = {
    'http://www.opengis.net/gml/3.2': {
      'segments': makeReplacer(GML3.prototype.readSegment),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML32.prototype.ENVELOPE_PARSERS = {
    'http://www.opengis.net/gml/3.2': {
      'lowerCorner': makeArrayPusher(GML3.prototype.readFlatPosList),
      'upperCorner': makeArrayPusher(GML3.prototype.readFlatPosList),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML32.prototype.PATCHES_PARSERS = {
    'http://www.opengis.net/gml/3.2': {
      'PolygonPatch': makeReplacer(GML3.prototype.readPolygonPatch),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML32.prototype.SEGMENTS_PARSERS = {
    'http://www.opengis.net/gml/3.2': {
      'LineStringSegment': makeArrayExtender(
        GML3.prototype.readLineStringSegment
      ),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML32.prototype.MULTIPOINT_PARSERS = {
    'http://www.opengis.net/gml/3.2': {
      'pointMember': makeArrayPusher(GMLBase.prototype.pointMemberParser),
      'pointMembers': makeArrayPusher(GMLBase.prototype.pointMemberParser),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML32.prototype.MULTILINESTRING_PARSERS = {
    'http://www.opengis.net/gml/3.2': {
      'lineStringMember': makeArrayPusher(
        GMLBase.prototype.lineStringMemberParser
      ),
      'lineStringMembers': makeArrayPusher(
        GMLBase.prototype.lineStringMemberParser
      ),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML32.prototype.MULTIPOLYGON_PARSERS = {
    'http://www.opengis.net/gml/3.2': {
      'polygonMember': makeArrayPusher(GMLBase.prototype.polygonMemberParser),
      'polygonMembers': makeArrayPusher(GMLBase.prototype.polygonMemberParser),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML32.prototype.POINTMEMBER_PARSERS = {
    'http://www.opengis.net/gml/3.2': {
      'Point': makeArrayPusher(GMLBase.prototype.readFlatCoordinatesFromNode),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML32.prototype.LINESTRINGMEMBER_PARSERS = {
    'http://www.opengis.net/gml/3.2': {
      'LineString': makeArrayPusher(GMLBase.prototype.readLineString),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML32.prototype.POLYGONMEMBER_PARSERS = {
    'http://www.opengis.net/gml/3.2': {
      'Polygon': makeArrayPusher(GMLBase.prototype.readPolygon),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  GML32.prototype.RING_PARSERS = {
    'http://www.opengis.net/gml/3.2': {
      'LinearRing': makeReplacer(GMLBase.prototype.readFlatLinearRing),
      'Ring': makeReplacer(GML32.prototype.readFlatCurveRing),
    },
  };

  /**
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  GML32.prototype.RING_SERIALIZERS = {
    'http://www.opengis.net/gml/3.2': {
      'exterior': makeChildAppender(GML3.prototype.writeRing),
      'interior': makeChildAppender(GML3.prototype.writeRing),
    },
  };

  /**
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  GML32.prototype.ENVELOPE_SERIALIZERS = {
    'http://www.opengis.net/gml/3.2': {
      'lowerCorner': makeChildAppender(writeStringTextNode),
      'upperCorner': makeChildAppender(writeStringTextNode),
    },
  };

  /**
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  GML32.prototype.SURFACEORPOLYGONMEMBER_SERIALIZERS = {
    'http://www.opengis.net/gml/3.2': {
      'surfaceMember': makeChildAppender(
        GML3.prototype.writeSurfaceOrPolygonMember
      ),
      'polygonMember': makeChildAppender(
        GML3.prototype.writeSurfaceOrPolygonMember
      ),
    },
  };

  /**
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  GML32.prototype.POINTMEMBER_SERIALIZERS = {
    'http://www.opengis.net/gml/3.2': {
      'pointMember': makeChildAppender(GML3.prototype.writePointMember),
    },
  };

  /**
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  GML32.prototype.LINESTRINGORCURVEMEMBER_SERIALIZERS = {
    'http://www.opengis.net/gml/3.2': {
      'lineStringMember': makeChildAppender(
        GML3.prototype.writeLineStringOrCurveMember
      ),
      'curveMember': makeChildAppender(
        GML3.prototype.writeLineStringOrCurveMember
      ),
    },
  };

  /**
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  GML32.prototype.GEOMETRY_SERIALIZERS = {
    'http://www.opengis.net/gml/3.2': {
      'Curve': makeChildAppender(GML3.prototype.writeCurveOrLineString),
      'MultiCurve': makeChildAppender(GML3.prototype.writeMultiCurveOrLineString),
      'Point': makeChildAppender(GML32.prototype.writePoint),
      'MultiPoint': makeChildAppender(GML3.prototype.writeMultiPoint),
      'LineString': makeChildAppender(GML3.prototype.writeCurveOrLineString),
      'MultiLineString': makeChildAppender(
        GML3.prototype.writeMultiCurveOrLineString
      ),
      'LinearRing': makeChildAppender(GML3.prototype.writeLinearRing),
      'Polygon': makeChildAppender(GML3.prototype.writeSurfaceOrPolygon),
      'MultiPolygon': makeChildAppender(
        GML3.prototype.writeMultiSurfaceOrPolygon
      ),
      'Surface': makeChildAppender(GML3.prototype.writeSurfaceOrPolygon),
      'MultiSurface': makeChildAppender(
        GML3.prototype.writeMultiSurfaceOrPolygon
      ),
      'Envelope': makeChildAppender(GML3.prototype.writeEnvelope),
    },
  };

  /**
   * @module ol/format/GPX
   */

  /**
   * @const
   * @type {Array<null|string>}
   */
  const NAMESPACE_URIS$5 = [
    null,
    'http://www.topografix.com/GPX/1/0',
    'http://www.topografix.com/GPX/1/1',
  ];

  /**
   * @const
   * @type {string}
   */
  const SCHEMA_LOCATION$1 =
    'http://www.topografix.com/GPX/1/1 ' +
    'http://www.topografix.com/GPX/1/1/gpx.xsd';

  /**
   * @const
   * @type {Object<string, function(Node, Array<*>): (Feature|undefined)>}
   */
  const FEATURE_READER = {
    'rte': readRte,
    'trk': readTrk,
    'wpt': readWpt,
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const GPX_PARSERS = makeStructureNS(NAMESPACE_URIS$5, {
    'rte': makeArrayPusher(readRte),
    'trk': makeArrayPusher(readTrk),
    'wpt': makeArrayPusher(readWpt),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const LINK_PARSERS$1 = makeStructureNS(NAMESPACE_URIS$5, {
    'text': makeObjectPropertySetter(readString, 'linkText'),
    'type': makeObjectPropertySetter(readString, 'linkType'),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  // @ts-ignore
  const GPX_SERIALIZERS = makeStructureNS(NAMESPACE_URIS$5, {
    'rte': makeChildAppender(writeRte),
    'trk': makeChildAppender(writeTrk),
    'wpt': makeChildAppender(writeWpt),
  });

  /**
   * @typedef {Object} Options
   * @property {function(Feature, Node):void} [readExtensions] Callback function
   * to process `extensions` nodes. To prevent memory leaks, this callback function must
   * not store any references to the node. Note that the `extensions`
   * node is not allowed in GPX 1.0. Moreover, only `extensions`
   * nodes from `wpt`, `rte` and `trk` can be processed, as those are
   * directly mapped to a feature.
   */

  /**
   * @typedef {Object} LayoutOptions
   * @property {boolean} [hasZ] HasZ.
   * @property {boolean} [hasM] HasM.
   */

  /**
   * @classdesc
   * Feature format for reading and writing data in the GPX format.
   *
   * Note that {@link module:ol/format/GPX~GPX#readFeature} only reads the first
   * feature of the source.
   *
   * When reading, routes (`<rte>`) are converted into LineString geometries, and
   * tracks (`<trk>`) into MultiLineString. Any properties on route and track
   * waypoints are ignored.
   *
   * When writing, LineString geometries are output as routes (`<rte>`), and
   * MultiLineString as tracks (`<trk>`).
   *
   * @api
   */
  class GPX extends XMLFeature {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      super();

      options = options ? options : {};

      /**
       * @type {import("../proj/Projection.js").default}
       */
      this.dataProjection = get$2('EPSG:4326');

      /**
       * @type {function(Feature, Node): void|undefined}
       * @private
       */
      this.readExtensions_ = options.readExtensions;
    }

    /**
     * @param {Array<Feature>} features List of features.
     * @private
     */
    handleReadExtensions_(features) {
      if (!features) {
        features = [];
      }
      for (let i = 0, ii = features.length; i < ii; ++i) {
        const feature = features[i];
        if (this.readExtensions_) {
          const extensionsNode = feature.get('extensionsNode_') || null;
          this.readExtensions_(feature, extensionsNode);
        }
        feature.set('extensionsNode_', undefined);
      }
    }

    /**
     * @param {Element} node Node.
     * @param {import("./Feature.js").ReadOptions} [options] Options.
     * @return {import("../Feature.js").default} Feature.
     */
    readFeatureFromNode(node, options) {
      if (!NAMESPACE_URIS$5.includes(node.namespaceURI)) {
        return null;
      }
      const featureReader = FEATURE_READER[node.localName];
      if (!featureReader) {
        return null;
      }
      const feature = featureReader(node, [this.getReadOptions(node, options)]);
      if (!feature) {
        return null;
      }
      this.handleReadExtensions_([feature]);
      return feature;
    }

    /**
     * @param {Element} node Node.
     * @param {import("./Feature.js").ReadOptions} [options] Options.
     * @return {Array<import("../Feature.js").default>} Features.
     */
    readFeaturesFromNode(node, options) {
      if (!NAMESPACE_URIS$5.includes(node.namespaceURI)) {
        return [];
      }
      if (node.localName == 'gpx') {
        /** @type {Array<Feature>} */
        const features = pushParseAndPop([], GPX_PARSERS, node, [
          this.getReadOptions(node, options),
        ]);
        if (features) {
          this.handleReadExtensions_(features);
          return features;
        } else {
          return [];
        }
      }
      return [];
    }

    /**
     * Encode an array of features in the GPX format as an XML node.
     * LineString geometries are output as routes (`<rte>`), and MultiLineString
     * as tracks (`<trk>`).
     *
     * @param {Array<Feature>} features Features.
     * @param {import("./Feature.js").WriteOptions} [options] Options.
     * @return {Node} Node.
     * @api
     */
    writeFeaturesNode(features, options) {
      options = this.adaptOptions(options);
      //FIXME Serialize metadata
      const gpx = createElementNS('http://www.topografix.com/GPX/1/1', 'gpx');
      const xmlnsUri = 'http://www.w3.org/2000/xmlns/';
      gpx.setAttributeNS(xmlnsUri, 'xmlns:xsi', XML_SCHEMA_INSTANCE_URI);
      gpx.setAttributeNS(
        XML_SCHEMA_INSTANCE_URI,
        'xsi:schemaLocation',
        SCHEMA_LOCATION$1
      );
      gpx.setAttribute('version', '1.1');
      gpx.setAttribute('creator', 'OpenLayers');

      pushSerializeAndPop(
        /** @type {import("../xml.js").NodeStackItem} */
        ({node: gpx}),
        GPX_SERIALIZERS,
        GPX_NODE_FACTORY,
        features,
        [options]
      );
      return gpx;
    }
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const RTE_PARSERS = makeStructureNS(NAMESPACE_URIS$5, {
    'name': makeObjectPropertySetter(readString),
    'cmt': makeObjectPropertySetter(readString),
    'desc': makeObjectPropertySetter(readString),
    'src': makeObjectPropertySetter(readString),
    'link': parseLink,
    'number': makeObjectPropertySetter(readPositiveInteger),
    'extensions': parseExtensions,
    'type': makeObjectPropertySetter(readString),
    'rtept': parseRtePt,
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const RTEPT_PARSERS = makeStructureNS(NAMESPACE_URIS$5, {
    'ele': makeObjectPropertySetter(readDecimal),
    'time': makeObjectPropertySetter(readDateTime),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const TRK_PARSERS = makeStructureNS(NAMESPACE_URIS$5, {
    'name': makeObjectPropertySetter(readString),
    'cmt': makeObjectPropertySetter(readString),
    'desc': makeObjectPropertySetter(readString),
    'src': makeObjectPropertySetter(readString),
    'link': parseLink,
    'number': makeObjectPropertySetter(readPositiveInteger),
    'type': makeObjectPropertySetter(readString),
    'extensions': parseExtensions,
    'trkseg': parseTrkSeg,
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const TRKSEG_PARSERS = makeStructureNS(NAMESPACE_URIS$5, {
    'trkpt': parseTrkPt,
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const TRKPT_PARSERS = makeStructureNS(NAMESPACE_URIS$5, {
    'ele': makeObjectPropertySetter(readDecimal),
    'time': makeObjectPropertySetter(readDateTime),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const WPT_PARSERS = makeStructureNS(NAMESPACE_URIS$5, {
    'ele': makeObjectPropertySetter(readDecimal),
    'time': makeObjectPropertySetter(readDateTime),
    'magvar': makeObjectPropertySetter(readDecimal),
    'geoidheight': makeObjectPropertySetter(readDecimal),
    'name': makeObjectPropertySetter(readString),
    'cmt': makeObjectPropertySetter(readString),
    'desc': makeObjectPropertySetter(readString),
    'src': makeObjectPropertySetter(readString),
    'link': parseLink,
    'sym': makeObjectPropertySetter(readString),
    'type': makeObjectPropertySetter(readString),
    'fix': makeObjectPropertySetter(readString),
    'sat': makeObjectPropertySetter(readPositiveInteger),
    'hdop': makeObjectPropertySetter(readDecimal),
    'vdop': makeObjectPropertySetter(readDecimal),
    'pdop': makeObjectPropertySetter(readDecimal),
    'ageofdgpsdata': makeObjectPropertySetter(readDecimal),
    'dgpsid': makeObjectPropertySetter(readPositiveInteger),
    'extensions': parseExtensions,
  });

  /**
   * @const
   * @type {Array<string>}
   */
  const LINK_SEQUENCE = ['text', 'type'];

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  // @ts-ignore
  const LINK_SERIALIZERS = makeStructureNS(NAMESPACE_URIS$5, {
    'text': makeChildAppender(writeStringTextNode),
    'type': makeChildAppender(writeStringTextNode),
  });

  /**
   * @const
   * @type {Object<string, Array<string>>}
   */
  // @ts-ignore
  const RTE_SEQUENCE = makeStructureNS(NAMESPACE_URIS$5, [
    'name',
    'cmt',
    'desc',
    'src',
    'link',
    'number',
    'type',
    'rtept',
  ]);

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  // @ts-ignore
  const RTE_SERIALIZERS = makeStructureNS(NAMESPACE_URIS$5, {
    'name': makeChildAppender(writeStringTextNode),
    'cmt': makeChildAppender(writeStringTextNode),
    'desc': makeChildAppender(writeStringTextNode),
    'src': makeChildAppender(writeStringTextNode),
    'link': makeChildAppender(writeLink),
    'number': makeChildAppender(writeNonNegativeIntegerTextNode),
    'type': makeChildAppender(writeStringTextNode),
    'rtept': makeArraySerializer(makeChildAppender(writeWptType)),
  });

  /**
   * @const
   * @type {Object<string, Array<string>>}
   */
  // @ts-ignore
  const RTEPT_TYPE_SEQUENCE = makeStructureNS(NAMESPACE_URIS$5, ['ele', 'time']);

  /**
   * @const
   * @type {Object<string, Array<string>>}
   */
  // @ts-ignore
  const TRK_SEQUENCE = makeStructureNS(NAMESPACE_URIS$5, [
    'name',
    'cmt',
    'desc',
    'src',
    'link',
    'number',
    'type',
    'trkseg',
  ]);

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  // @ts-ignore
  const TRK_SERIALIZERS = makeStructureNS(NAMESPACE_URIS$5, {
    'name': makeChildAppender(writeStringTextNode),
    'cmt': makeChildAppender(writeStringTextNode),
    'desc': makeChildAppender(writeStringTextNode),
    'src': makeChildAppender(writeStringTextNode),
    'link': makeChildAppender(writeLink),
    'number': makeChildAppender(writeNonNegativeIntegerTextNode),
    'type': makeChildAppender(writeStringTextNode),
    'trkseg': makeArraySerializer(makeChildAppender(writeTrkSeg)),
  });

  /**
   * @const
   * @type {function(*, Array<*>, string=): (Node|undefined)}
   */
  const TRKSEG_NODE_FACTORY = makeSimpleNodeFactory('trkpt');

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  // @ts-ignore
  const TRKSEG_SERIALIZERS = makeStructureNS(NAMESPACE_URIS$5, {
    'trkpt': makeChildAppender(writeWptType),
  });

  /**
   * @const
   * @type {Object<string, Array<string>>}
   */
  // @ts-ignore
  const WPT_TYPE_SEQUENCE = makeStructureNS(NAMESPACE_URIS$5, [
    'ele',
    'time',
    'magvar',
    'geoidheight',
    'name',
    'cmt',
    'desc',
    'src',
    'link',
    'sym',
    'type',
    'fix',
    'sat',
    'hdop',
    'vdop',
    'pdop',
    'ageofdgpsdata',
    'dgpsid',
  ]);

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  // @ts-ignore
  const WPT_TYPE_SERIALIZERS = makeStructureNS(NAMESPACE_URIS$5, {
    'ele': makeChildAppender(writeDecimalTextNode),
    'time': makeChildAppender(writeDateTimeTextNode),
    'magvar': makeChildAppender(writeDecimalTextNode),
    'geoidheight': makeChildAppender(writeDecimalTextNode),
    'name': makeChildAppender(writeStringTextNode),
    'cmt': makeChildAppender(writeStringTextNode),
    'desc': makeChildAppender(writeStringTextNode),
    'src': makeChildAppender(writeStringTextNode),
    'link': makeChildAppender(writeLink),
    'sym': makeChildAppender(writeStringTextNode),
    'type': makeChildAppender(writeStringTextNode),
    'fix': makeChildAppender(writeStringTextNode),
    'sat': makeChildAppender(writeNonNegativeIntegerTextNode),
    'hdop': makeChildAppender(writeDecimalTextNode),
    'vdop': makeChildAppender(writeDecimalTextNode),
    'pdop': makeChildAppender(writeDecimalTextNode),
    'ageofdgpsdata': makeChildAppender(writeDecimalTextNode),
    'dgpsid': makeChildAppender(writeNonNegativeIntegerTextNode),
  });

  /**
   * @const
   * @type {Object<string, string>}
   */
  const GEOMETRY_TYPE_TO_NODENAME$1 = {
    'Point': 'wpt',
    'LineString': 'rte',
    'MultiLineString': 'trk',
  };

  /**
   * @param {*} value Value.
   * @param {Array<*>} objectStack Object stack.
   * @param {string} [nodeName] Node name.
   * @return {Node|undefined} Node.
   */
  function GPX_NODE_FACTORY(value, objectStack, nodeName) {
    const geometry = /** @type {Feature} */ (value).getGeometry();
    if (geometry) {
      const nodeName = GEOMETRY_TYPE_TO_NODENAME$1[geometry.getType()];
      if (nodeName) {
        const parentNode = objectStack[objectStack.length - 1].node;
        return createElementNS(parentNode.namespaceURI, nodeName);
      }
    }
  }

  /**
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {LayoutOptions} layoutOptions Layout options.
   * @param {Element} node Node.
   * @param {!Object} values Values.
   * @return {Array<number>} Flat coordinates.
   */
  function appendCoordinate(flatCoordinates, layoutOptions, node, values) {
    flatCoordinates.push(
      parseFloat(node.getAttribute('lon')),
      parseFloat(node.getAttribute('lat'))
    );
    if ('ele' in values) {
      flatCoordinates.push(/** @type {number} */ (values['ele']));
      delete values['ele'];
      layoutOptions.hasZ = true;
    } else {
      flatCoordinates.push(0);
    }
    if ('time' in values) {
      flatCoordinates.push(/** @type {number} */ (values['time']));
      delete values['time'];
      layoutOptions.hasM = true;
    } else {
      flatCoordinates.push(0);
    }
    return flatCoordinates;
  }

  /**
   * Choose GeometryLayout based on flags in layoutOptions and adjust flatCoordinates
   * and ends arrays by shrinking them accordingly (removing unused zero entries).
   *
   * @param {LayoutOptions} layoutOptions Layout options.
   * @param {Array<number>} flatCoordinates Flat coordinates.
   * @param {Array<number>} [ends] Ends.
   * @return {import("../geom/Geometry.js").GeometryLayout} Layout.
   */
  function applyLayoutOptions(layoutOptions, flatCoordinates, ends) {
    /** @type {import("../geom/Geometry.js").GeometryLayout} */
    let layout = 'XY';
    let stride = 2;
    if (layoutOptions.hasZ && layoutOptions.hasM) {
      layout = 'XYZM';
      stride = 4;
    } else if (layoutOptions.hasZ) {
      layout = 'XYZ';
      stride = 3;
    } else if (layoutOptions.hasM) {
      layout = 'XYM';
      stride = 3;
    }
    if (stride !== 4) {
      for (let i = 0, ii = flatCoordinates.length / 4; i < ii; i++) {
        flatCoordinates[i * stride] = flatCoordinates[i * 4];
        flatCoordinates[i * stride + 1] = flatCoordinates[i * 4 + 1];
        if (layoutOptions.hasZ) {
          flatCoordinates[i * stride + 2] = flatCoordinates[i * 4 + 2];
        }
        if (layoutOptions.hasM) {
          flatCoordinates[i * stride + 2] = flatCoordinates[i * 4 + 3];
        }
      }
      flatCoordinates.length = (flatCoordinates.length / 4) * stride;
      if (ends) {
        for (let i = 0, ii = ends.length; i < ii; i++) {
          ends[i] = (ends[i] / 4) * stride;
        }
      }
    }
    return layout;
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function parseLink(node, objectStack) {
    const values = /** @type {Object} */ (objectStack[objectStack.length - 1]);
    const href = node.getAttribute('href');
    if (href !== null) {
      values['link'] = href;
    }
    parseNode(LINK_PARSERS$1, node, objectStack);
  }

  /**
   * @param {Node} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function parseExtensions(node, objectStack) {
    const values = /** @type {Object} */ (objectStack[objectStack.length - 1]);
    values['extensionsNode_'] = node;
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function parseRtePt(node, objectStack) {
    const values = pushParseAndPop({}, RTEPT_PARSERS, node, objectStack);
    if (values) {
      const rteValues = /** @type {!Object} */ (
        objectStack[objectStack.length - 1]
      );
      const flatCoordinates = /** @type {Array<number>} */ (
        rteValues['flatCoordinates']
      );
      const layoutOptions = /** @type {LayoutOptions} */ (
        rteValues['layoutOptions']
      );
      appendCoordinate(flatCoordinates, layoutOptions, node, values);
    }
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function parseTrkPt(node, objectStack) {
    const values = pushParseAndPop({}, TRKPT_PARSERS, node, objectStack);
    if (values) {
      const trkValues = /** @type {!Object} */ (
        objectStack[objectStack.length - 1]
      );
      const flatCoordinates = /** @type {Array<number>} */ (
        trkValues['flatCoordinates']
      );
      const layoutOptions = /** @type {LayoutOptions} */ (
        trkValues['layoutOptions']
      );
      appendCoordinate(flatCoordinates, layoutOptions, node, values);
    }
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function parseTrkSeg(node, objectStack) {
    const values = /** @type {Object} */ (objectStack[objectStack.length - 1]);
    parseNode(TRKSEG_PARSERS, node, objectStack);
    const flatCoordinates =
      /** @type {Array<number>} */
      (values['flatCoordinates']);
    const ends = /** @type {Array<number>} */ (values['ends']);
    ends.push(flatCoordinates.length);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Feature|undefined} Track.
   */
  function readRte(node, objectStack) {
    const options = /** @type {import("./Feature.js").ReadOptions} */ (
      objectStack[0]
    );
    const values = pushParseAndPop(
      {
        'flatCoordinates': [],
        'layoutOptions': {},
      },
      RTE_PARSERS,
      node,
      objectStack
    );
    if (!values) {
      return undefined;
    }
    const flatCoordinates =
      /** @type {Array<number>} */
      (values['flatCoordinates']);
    delete values['flatCoordinates'];
    const layoutOptions = /** @type {LayoutOptions} */ (values['layoutOptions']);
    delete values['layoutOptions'];
    const layout = applyLayoutOptions(layoutOptions, flatCoordinates);
    const geometry = new LineString(flatCoordinates, layout);
    transformGeometryWithOptions(geometry, false, options);
    const feature = new Feature(geometry);
    feature.setProperties(values, true);
    return feature;
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Feature|undefined} Track.
   */
  function readTrk(node, objectStack) {
    const options = /** @type {import("./Feature.js").ReadOptions} */ (
      objectStack[0]
    );
    const values = pushParseAndPop(
      {
        'flatCoordinates': [],
        'ends': [],
        'layoutOptions': {},
      },
      TRK_PARSERS,
      node,
      objectStack
    );
    if (!values) {
      return undefined;
    }
    const flatCoordinates =
      /** @type {Array<number>} */
      (values['flatCoordinates']);
    delete values['flatCoordinates'];
    const ends = /** @type {Array<number>} */ (values['ends']);
    delete values['ends'];
    const layoutOptions = /** @type {LayoutOptions} */ (values['layoutOptions']);
    delete values['layoutOptions'];
    const layout = applyLayoutOptions(layoutOptions, flatCoordinates, ends);
    const geometry = new MultiLineString(flatCoordinates, layout, ends);
    transformGeometryWithOptions(geometry, false, options);
    const feature = new Feature(geometry);
    feature.setProperties(values, true);
    return feature;
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Feature|undefined} Waypoint.
   */
  function readWpt(node, objectStack) {
    const options = /** @type {import("./Feature.js").ReadOptions} */ (
      objectStack[0]
    );
    const values = pushParseAndPop({}, WPT_PARSERS, node, objectStack);
    if (!values) {
      return undefined;
    }
    const layoutOptions = /** @type {LayoutOptions} */ ({});
    const coordinates = appendCoordinate([], layoutOptions, node, values);
    const layout = applyLayoutOptions(layoutOptions, coordinates);
    const geometry = new Point(coordinates, layout);
    transformGeometryWithOptions(geometry, false, options);
    const feature = new Feature(geometry);
    feature.setProperties(values, true);
    return feature;
  }

  /**
   * @param {Element} node Node.
   * @param {string} value Value for the link's `href` attribute.
   * @param {Array<*>} objectStack Node stack.
   */
  function writeLink(node, value, objectStack) {
    node.setAttribute('href', value);
    const context = objectStack[objectStack.length - 1];
    const properties = context['properties'];
    const link = [properties['linkText'], properties['linkType']];
    pushSerializeAndPop(
      /** @type {import("../xml.js").NodeStackItem} */ ({node: node}),
      LINK_SERIALIZERS,
      OBJECT_PROPERTY_NODE_FACTORY,
      link,
      objectStack,
      LINK_SEQUENCE
    );
  }

  /**
   * @param {Element} node Node.
   * @param {import("../coordinate.js").Coordinate} coordinate Coordinate.
   * @param {Array<*>} objectStack Object stack.
   */
  function writeWptType(node, coordinate, objectStack) {
    const context = objectStack[objectStack.length - 1];
    const parentNode = context.node;
    const namespaceURI = parentNode.namespaceURI;
    const properties = context['properties'];
    //FIXME Projection handling
    node.setAttributeNS(null, 'lat', String(coordinate[1]));
    node.setAttributeNS(null, 'lon', String(coordinate[0]));
    const geometryLayout = context['geometryLayout'];
    switch (geometryLayout) {
      case 'XYZM':
        if (coordinate[3] !== 0) {
          properties['time'] = coordinate[3];
        }
      // fall through
      case 'XYZ':
        if (coordinate[2] !== 0) {
          properties['ele'] = coordinate[2];
        }
        break;
      case 'XYM':
        if (coordinate[2] !== 0) {
          properties['time'] = coordinate[2];
        }
        break;
      // pass
    }
    const orderedKeys =
      node.nodeName == 'rtept'
        ? RTEPT_TYPE_SEQUENCE[namespaceURI]
        : WPT_TYPE_SEQUENCE[namespaceURI];
    const values = makeSequence(properties, orderedKeys);
    pushSerializeAndPop(
      /** @type {import("../xml.js").NodeStackItem} */
      ({node: node, 'properties': properties}),
      WPT_TYPE_SERIALIZERS,
      OBJECT_PROPERTY_NODE_FACTORY,
      values,
      objectStack,
      orderedKeys
    );
  }

  /**
   * @param {Node} node Node.
   * @param {Feature} feature Feature.
   * @param {Array<*>} objectStack Object stack.
   */
  function writeRte(node, feature, objectStack) {
    const options = /** @type {import("./Feature.js").WriteOptions} */ (
      objectStack[0]
    );
    const properties = feature.getProperties();
    const context = {node: node};
    context['properties'] = properties;
    const geometry = feature.getGeometry();
    if (geometry.getType() == 'LineString') {
      const lineString = /** @type {LineString} */ (
        transformGeometryWithOptions(geometry, true, options)
      );
      context['geometryLayout'] = lineString.getLayout();
      properties['rtept'] = lineString.getCoordinates();
    }
    const parentNode = objectStack[objectStack.length - 1].node;
    const orderedKeys = RTE_SEQUENCE[parentNode.namespaceURI];
    const values = makeSequence(properties, orderedKeys);
    pushSerializeAndPop(
      context,
      RTE_SERIALIZERS,
      OBJECT_PROPERTY_NODE_FACTORY,
      values,
      objectStack,
      orderedKeys
    );
  }

  /**
   * @param {Element} node Node.
   * @param {Feature} feature Feature.
   * @param {Array<*>} objectStack Object stack.
   */
  function writeTrk(node, feature, objectStack) {
    const options = /** @type {import("./Feature.js").WriteOptions} */ (
      objectStack[0]
    );
    const properties = feature.getProperties();
    /** @type {import("../xml.js").NodeStackItem} */
    const context = {node: node};
    context['properties'] = properties;
    const geometry = feature.getGeometry();
    if (geometry.getType() == 'MultiLineString') {
      const multiLineString = /** @type {MultiLineString} */ (
        transformGeometryWithOptions(geometry, true, options)
      );
      properties['trkseg'] = multiLineString.getLineStrings();
    }
    const parentNode = objectStack[objectStack.length - 1].node;
    const orderedKeys = TRK_SEQUENCE[parentNode.namespaceURI];
    const values = makeSequence(properties, orderedKeys);
    pushSerializeAndPop(
      context,
      TRK_SERIALIZERS,
      OBJECT_PROPERTY_NODE_FACTORY,
      values,
      objectStack,
      orderedKeys
    );
  }

  /**
   * @param {Element} node Node.
   * @param {LineString} lineString LineString.
   * @param {Array<*>} objectStack Object stack.
   */
  function writeTrkSeg(node, lineString, objectStack) {
    /** @type {import("../xml.js").NodeStackItem} */
    const context = {node: node};
    context['geometryLayout'] = lineString.getLayout();
    context['properties'] = {};
    pushSerializeAndPop(
      context,
      TRKSEG_SERIALIZERS,
      TRKSEG_NODE_FACTORY,
      lineString.getCoordinates(),
      objectStack
    );
  }

  /**
   * @param {Element} node Node.
   * @param {Feature} feature Feature.
   * @param {Array<*>} objectStack Object stack.
   */
  function writeWpt(node, feature, objectStack) {
    const options = /** @type {import("./Feature.js").WriteOptions} */ (
      objectStack[0]
    );
    const context = objectStack[objectStack.length - 1];
    context['properties'] = feature.getProperties();
    const geometry = feature.getGeometry();
    if (geometry.getType() == 'Point') {
      const point = /** @type {Point} */ (
        transformGeometryWithOptions(geometry, true, options)
      );
      context['geometryLayout'] = point.getLayout();
      writeWptType(node, point.getCoordinates(), objectStack);
    }
  }

  /**
   * @module ol/format/TextFeature
   */

  /**
   * @classdesc
   * Abstract base class; normally only used for creating subclasses and not
   * instantiated in apps.
   * Base class for text feature formats.
   *
   * @abstract
   */
  class TextFeature extends FeatureFormat {
    constructor() {
      super();
    }

    /**
     * @return {import("./Feature.js").Type} Format.
     */
    getType() {
      return 'text';
    }

    /**
     * Read the feature from the source.
     *
     * @param {Document|Element|Object|string} source Source.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @return {import("../Feature.js").default} Feature.
     * @api
     */
    readFeature(source, options) {
      return this.readFeatureFromText(
        getText(source),
        this.adaptOptions(options)
      );
    }

    /**
     * @abstract
     * @param {string} text Text.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @protected
     * @return {import("../Feature.js").default} Feature.
     */
    readFeatureFromText(text, options) {
      return abstract();
    }

    /**
     * Read the features from the source.
     *
     * @param {Document|Element|Object|string} source Source.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @return {Array<import("../Feature.js").default>} Features.
     * @api
     */
    readFeatures(source, options) {
      return this.readFeaturesFromText(
        getText(source),
        this.adaptOptions(options)
      );
    }

    /**
     * @abstract
     * @param {string} text Text.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @protected
     * @return {Array<import("../Feature.js").default>} Features.
     */
    readFeaturesFromText(text, options) {
      return abstract();
    }

    /**
     * Read the geometry from the source.
     *
     * @param {Document|Element|Object|string} source Source.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @return {import("../geom/Geometry.js").default} Geometry.
     * @api
     */
    readGeometry(source, options) {
      return this.readGeometryFromText(
        getText(source),
        this.adaptOptions(options)
      );
    }

    /**
     * @abstract
     * @param {string} text Text.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @protected
     * @return {import("../geom/Geometry.js").default} Geometry.
     */
    readGeometryFromText(text, options) {
      return abstract();
    }

    /**
     * Read the projection from the source.
     *
     * @param {Document|Element|Object|string} source Source.
     * @return {import("../proj/Projection.js").default|undefined} Projection.
     * @api
     */
    readProjection(source) {
      return this.readProjectionFromText(getText(source));
    }

    /**
     * @param {string} text Text.
     * @protected
     * @return {import("../proj/Projection.js").default|undefined} Projection.
     */
    readProjectionFromText(text) {
      return this.dataProjection;
    }

    /**
     * Encode a feature as a string.
     *
     * @param {import("../Feature.js").default} feature Feature.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @return {string} Encoded feature.
     * @api
     */
    writeFeature(feature, options) {
      return this.writeFeatureText(feature, this.adaptOptions(options));
    }

    /**
     * @abstract
     * @param {import("../Feature.js").default} feature Features.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @protected
     * @return {string} Text.
     */
    writeFeatureText(feature, options) {
      return abstract();
    }

    /**
     * Encode an array of features as string.
     *
     * @param {Array<import("../Feature.js").default>} features Features.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @return {string} Encoded features.
     * @api
     */
    writeFeatures(features, options) {
      return this.writeFeaturesText(features, this.adaptOptions(options));
    }

    /**
     * @abstract
     * @param {Array<import("../Feature.js").default>} features Features.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @protected
     * @return {string} Text.
     */
    writeFeaturesText(features, options) {
      return abstract();
    }

    /**
     * Write a single geometry.
     *
     * @param {import("../geom/Geometry.js").default} geometry Geometry.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @return {string} Geometry.
     * @api
     */
    writeGeometry(geometry, options) {
      return this.writeGeometryText(geometry, this.adaptOptions(options));
    }

    /**
     * @abstract
     * @param {import("../geom/Geometry.js").default} geometry Geometry.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @protected
     * @return {string} Text.
     */
    writeGeometryText(geometry, options) {
      return abstract();
    }
  }

  /**
   * @param {Document|Element|Object|string} source Source.
   * @return {string} Text.
   */
  function getText(source) {
    if (typeof source === 'string') {
      return source;
    } else {
      return '';
    }
  }

  /**
   * @module ol/format/IGC
   */

  /**
   * @typedef {'barometric' | 'gps' | 'none'} IGCZ
   * IGC altitude/z. One of 'barometric', 'gps', 'none'.
   */

  /**
   * @const
   * @type {RegExp}
   */
  const B_RECORD_RE =
    /^B(\d{2})(\d{2})(\d{2})(\d{2})(\d{5})([NS])(\d{3})(\d{5})([EW])([AV])(\d{5})(\d{5})/;

  /**
   * @const
   * @type {RegExp}
   */
  const H_RECORD_RE = /^H.([A-Z]{3}).*?:(.*)/;

  /**
   * @const
   * @type {RegExp}
   */
  const HFDTE_RECORD_RE = /^HFDTE(\d{2})(\d{2})(\d{2})/;

  /**
   * A regular expression matching the newline characters `\r\n`, `\r` and `\n`.
   *
   * @const
   * @type {RegExp}
   */
  const NEWLINE_RE = /\r\n|\r|\n/;

  /**
   * @typedef {Object} Options
   * @property {IGCZ} [altitudeMode='none'] Altitude mode. Possible
   * values are `'barometric'`, `'gps'`, and `'none'`.
   */

  /**
   * @classdesc
   * Feature format for `*.igc` flight recording files.
   *
   * As IGC sources contain a single feature,
   * {@link module:ol/format/IGC~IGC#readFeatures} will return the feature in an
   * array
   *
   * @api
   */
  class IGC extends TextFeature {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      super();

      options = options ? options : {};

      /**
       * @type {import("../proj/Projection.js").default}
       */
      this.dataProjection = get$2('EPSG:4326');

      /**
       * @private
       * @type {IGCZ}
       */
      this.altitudeMode_ = options.altitudeMode ? options.altitudeMode : 'none';
    }

    /**
     * @protected
     * @param {string} text Text.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @return {import("../Feature.js").default} Feature.
     */
    readFeatureFromText(text, options) {
      const altitudeMode = this.altitudeMode_;
      const lines = text.split(NEWLINE_RE);
      /** @type {Object<string, string>} */
      const properties = {};
      const flatCoordinates = [];
      let year = 2000;
      let month = 0;
      let day = 1;
      let lastDateTime = -1;
      let i, ii;
      for (i = 0, ii = lines.length; i < ii; ++i) {
        const line = lines[i];
        let m;
        if (line.charAt(0) == 'B') {
          m = B_RECORD_RE.exec(line);
          if (m) {
            const hour = parseInt(m[1], 10);
            const minute = parseInt(m[2], 10);
            const second = parseInt(m[3], 10);
            let y = parseInt(m[4], 10) + parseInt(m[5], 10) / 60000;
            if (m[6] == 'S') {
              y = -y;
            }
            let x = parseInt(m[7], 10) + parseInt(m[8], 10) / 60000;
            if (m[9] == 'W') {
              x = -x;
            }
            flatCoordinates.push(x, y);
            if (altitudeMode != 'none') {
              let z;
              if (altitudeMode == 'gps') {
                z = parseInt(m[11], 10);
              } else if (altitudeMode == 'barometric') {
                z = parseInt(m[12], 10);
              } else {
                z = 0;
              }
              flatCoordinates.push(z);
            }
            let dateTime = Date.UTC(year, month, day, hour, minute, second);
            // Detect UTC midnight wrap around.
            if (dateTime < lastDateTime) {
              dateTime = Date.UTC(year, month, day + 1, hour, minute, second);
            }
            flatCoordinates.push(dateTime / 1000);
            lastDateTime = dateTime;
          }
        } else if (line.charAt(0) == 'H') {
          m = HFDTE_RECORD_RE.exec(line);
          if (m) {
            day = parseInt(m[1], 10);
            month = parseInt(m[2], 10) - 1;
            year = 2000 + parseInt(m[3], 10);
          } else {
            m = H_RECORD_RE.exec(line);
            if (m) {
              properties[m[1]] = m[2].trim();
            }
          }
        }
      }
      if (flatCoordinates.length === 0) {
        return null;
      }
      const layout = altitudeMode == 'none' ? 'XYM' : 'XYZM';
      const lineString = new LineString(flatCoordinates, layout);
      const feature = new Feature(
        transformGeometryWithOptions(lineString, false, options)
      );
      feature.setProperties(properties, true);
      return feature;
    }

    /**
     * @param {string} text Text.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @protected
     * @return {Array<Feature>} Features.
     */
    readFeaturesFromText(text, options) {
      const feature = this.readFeatureFromText(text, options);
      if (feature) {
        return [feature];
      } else {
        return [];
      }
    }
  }

  /**
   * @module ol/format/KML
   */

  /**
   * @typedef {Object} Vec2
   * @property {number} x X coordinate.
   * @property {import("../style/Icon.js").IconAnchorUnits} xunits Units of x.
   * @property {number} y Y coordinate.
   * @property {import("../style/Icon.js").IconAnchorUnits} yunits Units of Y.
   * @property {import("../style/Icon.js").IconOrigin} [origin] Origin.
   */

  /**
   * @typedef {Object} GxTrackObject
   * @property {Array<Array<number>>} coordinates Coordinates.
   * @property {Array<number>} whens Whens.
   */

  /**
   * @const
   * @type {Array<string>}
   */
  const GX_NAMESPACE_URIS = ['http://www.google.com/kml/ext/2.2'];

  /**
   * @const
   * @type {Array<null|string>}
   */
  const NAMESPACE_URIS$4 = [
    null,
    'http://earth.google.com/kml/2.0',
    'http://earth.google.com/kml/2.1',
    'http://earth.google.com/kml/2.2',
    'http://www.opengis.net/kml/2.2',
  ];

  /**
   * @const
   * @type {string}
   */
  const SCHEMA_LOCATION =
    'http://www.opengis.net/kml/2.2 ' +
    'https://developers.google.com/kml/schema/kml22gx.xsd';

  /**
   * @type {Object<string, import("../style/Icon.js").IconAnchorUnits>}
   */
  const ICON_ANCHOR_UNITS_MAP = {
    'fraction': 'fraction',
    'pixels': 'pixels',
    'insetPixels': 'pixels',
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const PLACEMARK_PARSERS = makeStructureNS(
    NAMESPACE_URIS$4,
    {
      'ExtendedData': extendedDataParser,
      'Region': regionParser,
      'MultiGeometry': makeObjectPropertySetter(readMultiGeometry, 'geometry'),
      'LineString': makeObjectPropertySetter(readLineString, 'geometry'),
      'LinearRing': makeObjectPropertySetter(readLinearRing, 'geometry'),
      'Point': makeObjectPropertySetter(readPoint, 'geometry'),
      'Polygon': makeObjectPropertySetter(readPolygon, 'geometry'),
      'Style': makeObjectPropertySetter(readStyle$2),
      'StyleMap': placemarkStyleMapParser,
      'address': makeObjectPropertySetter(readString),
      'description': makeObjectPropertySetter(readString),
      'name': makeObjectPropertySetter(readString),
      'open': makeObjectPropertySetter(readBoolean),
      'phoneNumber': makeObjectPropertySetter(readString),
      'styleUrl': makeObjectPropertySetter(readStyleURL),
      'visibility': makeObjectPropertySetter(readBoolean),
    },
    makeStructureNS(GX_NAMESPACE_URIS, {
      'MultiTrack': makeObjectPropertySetter(readGxMultiTrack, 'geometry'),
      'Track': makeObjectPropertySetter(readGxTrack, 'geometry'),
    })
  );

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const NETWORK_LINK_PARSERS = makeStructureNS(NAMESPACE_URIS$4, {
    'ExtendedData': extendedDataParser,
    'Region': regionParser,
    'Link': linkParser,
    'address': makeObjectPropertySetter(readString),
    'description': makeObjectPropertySetter(readString),
    'name': makeObjectPropertySetter(readString),
    'open': makeObjectPropertySetter(readBoolean),
    'phoneNumber': makeObjectPropertySetter(readString),
    'visibility': makeObjectPropertySetter(readBoolean),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const LINK_PARSERS = makeStructureNS(NAMESPACE_URIS$4, {
    'href': makeObjectPropertySetter(readURI),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const REGION_PARSERS = makeStructureNS(NAMESPACE_URIS$4, {
    'LatLonAltBox': latLonAltBoxParser,
    'Lod': lodParser,
  });

  /**
   * @const
   * @type {Object<string, Array<string>>}
   */
  // @ts-ignore
  const KML_SEQUENCE = makeStructureNS(NAMESPACE_URIS$4, ['Document', 'Placemark']);

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  // @ts-ignore
  const KML_SERIALIZERS = makeStructureNS(NAMESPACE_URIS$4, {
    'Document': makeChildAppender(writeDocument),
    'Placemark': makeChildAppender(writePlacemark),
  });

  /**
   * @type {import("../color.js").Color}
   */
  let DEFAULT_COLOR;

  /**
   * @type {Fill|null}
   */
  let DEFAULT_FILL_STYLE = null;

  /**
   * Get the default fill style (or null if not yet set).
   * @return {Fill|null} The default fill style.
   */
  function getDefaultFillStyle() {
    return DEFAULT_FILL_STYLE;
  }

  /**
   * @type {import("../size.js").Size}
   */
  let DEFAULT_IMAGE_STYLE_ANCHOR;

  /**
   * @type {import("../style/Icon.js").IconAnchorUnits}
   */
  let DEFAULT_IMAGE_STYLE_ANCHOR_X_UNITS;

  /**
   * @type {import("../style/Icon.js").IconAnchorUnits}
   */
  let DEFAULT_IMAGE_STYLE_ANCHOR_Y_UNITS;

  /**
   * @type {import("../size.js").Size}
   */
  let DEFAULT_IMAGE_STYLE_SIZE;

  /**
   * @type {string}
   */
  let DEFAULT_IMAGE_STYLE_SRC;

  /**
   * @type {import("../style/Image.js").default|null}
   */
  let DEFAULT_IMAGE_STYLE = null;

  /**
   * Get the default image style (or null if not yet set).
   * @return {import("../style/Image.js").default|null} The default image style.
   */
  function getDefaultImageStyle() {
    return DEFAULT_IMAGE_STYLE;
  }

  /**
   * @type {string}
   */
  let DEFAULT_NO_IMAGE_STYLE;

  /**
   * @type {Stroke|null}
   */
  let DEFAULT_STROKE_STYLE = null;

  /**
   * Get the default stroke style (or null if not yet set).
   * @return {Stroke|null} The default stroke style.
   */
  function getDefaultStrokeStyle() {
    return DEFAULT_STROKE_STYLE;
  }

  /**
   * @type {Stroke}
   */
  let DEFAULT_TEXT_STROKE_STYLE;

  /**
   * @type {Text|null}
   */
  let DEFAULT_TEXT_STYLE = null;

  /**
   * Get the default text style (or null if not yet set).
   * @return {Text|null} The default text style.
   */
  function getDefaultTextStyle() {
    return DEFAULT_TEXT_STYLE;
  }

  /**
   * @type {Style|null}
   */
  let DEFAULT_STYLE = null;

  /**
   * Get the default style (or null if not yet set).
   * @return {Style|null} The default style.
   */
  function getDefaultStyle() {
    return DEFAULT_STYLE;
  }

  /**
   * @type {Array<Style>|null}
   */
  let DEFAULT_STYLE_ARRAY = null;

  /**
   * Get the default style array (or null if not yet set).
   * @return {Array<Style>|null} The default style.
   */
  function getDefaultStyleArray() {
    return DEFAULT_STYLE_ARRAY;
  }

  /**
   * Function that returns the scale needed to normalize an icon image to 32 pixels.
   * @param {import("../size.js").Size} size Image size.
   * @return {number} Scale.
   */
  function scaleForSize(size) {
    return 32 / Math.min(size[0], size[1]);
  }

  function createStyleDefaults() {
    DEFAULT_COLOR = [255, 255, 255, 1];

    DEFAULT_FILL_STYLE = new Fill({
      color: DEFAULT_COLOR,
    });

    DEFAULT_IMAGE_STYLE_ANCHOR = [20, 2];

    DEFAULT_IMAGE_STYLE_ANCHOR_X_UNITS = 'pixels';

    DEFAULT_IMAGE_STYLE_ANCHOR_Y_UNITS = 'pixels';

    DEFAULT_IMAGE_STYLE_SIZE = [64, 64];

    DEFAULT_IMAGE_STYLE_SRC =
      'https://maps.google.com/mapfiles/kml/pushpin/ylw-pushpin.png';

    DEFAULT_IMAGE_STYLE = new Icon({
      anchor: DEFAULT_IMAGE_STYLE_ANCHOR,
      anchorOrigin: 'bottom-left',
      anchorXUnits: DEFAULT_IMAGE_STYLE_ANCHOR_X_UNITS,
      anchorYUnits: DEFAULT_IMAGE_STYLE_ANCHOR_Y_UNITS,
      crossOrigin: 'anonymous',
      rotation: 0,
      scale: scaleForSize(DEFAULT_IMAGE_STYLE_SIZE),
      size: DEFAULT_IMAGE_STYLE_SIZE,
      src: DEFAULT_IMAGE_STYLE_SRC,
    });

    DEFAULT_NO_IMAGE_STYLE = 'NO_IMAGE';

    DEFAULT_STROKE_STYLE = new Stroke({
      color: DEFAULT_COLOR,
      width: 1,
    });

    DEFAULT_TEXT_STROKE_STYLE = new Stroke({
      color: [51, 51, 51, 1],
      width: 2,
    });

    DEFAULT_TEXT_STYLE = new Text({
      font: 'bold 16px Helvetica',
      fill: DEFAULT_FILL_STYLE,
      stroke: DEFAULT_TEXT_STROKE_STYLE,
      scale: 0.8,
    });

    DEFAULT_STYLE = new Style({
      fill: DEFAULT_FILL_STYLE,
      image: DEFAULT_IMAGE_STYLE,
      text: DEFAULT_TEXT_STYLE,
      stroke: DEFAULT_STROKE_STYLE,
      zIndex: 0,
    });

    DEFAULT_STYLE_ARRAY = [DEFAULT_STYLE];
  }

  /**
   * @type {HTMLTextAreaElement}
   */
  let TEXTAREA;

  /**
   * A function that takes a url `{string}` and returns a url `{string}`.
   * Might be used to change an icon path or to substitute a
   * data url obtained from a KMZ array buffer.
   *
   * @typedef {function(string):string} IconUrlFunction
   * @api
   */

  /**
   * Function that returns a url unchanged.
   * @param {string} href Input url.
   * @return {string} Output url.
   */
  function defaultIconUrlFunction(href) {
    return href;
  }

  /**
   * @typedef {Object} Options
   * @property {boolean} [extractStyles=true] Extract styles from the KML.
   * @property {boolean} [showPointNames=true] Show names as labels for placemarks which contain points.
   * @property {Array<Style>} [defaultStyle] Default style. The
   * default default style is the same as Google Earth.
   * @property {boolean} [writeStyles=true] Write styles into KML.
   * @property {null|string} [crossOrigin='anonymous'] The `crossOrigin` attribute for loaded images. Note that you must provide a
   * `crossOrigin` value if you want to access pixel data with the Canvas renderer.
   * @property {IconUrlFunction} [iconUrlFunction] Function that takes a url string and returns a url string.
   * Might be used to change an icon path or to substitute a data url obtained from a KMZ array buffer.
   */

  /**
   * @classdesc
   * Feature format for reading and writing data in the KML format.
   *
   * {@link module:ol/format/KML~KML#readFeature} will read the first feature from
   * a KML source.
   *
   * MultiGeometries are converted into GeometryCollections if they are a mix of
   * geometry types, and into MultiPoint/MultiLineString/MultiPolygon if they are
   * all of the same type.
   *
   * @api
   */
  class KML extends XMLFeature {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      super();

      options = options ? options : {};

      if (!DEFAULT_STYLE_ARRAY) {
        createStyleDefaults();
      }

      /**
       * @type {import("../proj/Projection.js").default}
       */
      this.dataProjection = get$2('EPSG:4326');

      /**
       * @private
       * @type {Array<Style>}
       */
      this.defaultStyle_ = options.defaultStyle
        ? options.defaultStyle
        : DEFAULT_STYLE_ARRAY;

      /**
       * @private
       * @type {boolean}
       */
      this.extractStyles_ =
        options.extractStyles !== undefined ? options.extractStyles : true;

      /**
       * @type {boolean}
       */
      this.writeStyles_ =
        options.writeStyles !== undefined ? options.writeStyles : true;

      /**
       * @private
       * @type {!Object<string, (Array<Style>|string)>}
       */
      this.sharedStyles_ = {};

      /**
       * @private
       * @type {boolean}
       */
      this.showPointNames_ =
        options.showPointNames !== undefined ? options.showPointNames : true;

      /**
       * @type {null|string}
       */
      this.crossOrigin_ =
        options.crossOrigin !== undefined ? options.crossOrigin : 'anonymous';

      /**
       * @type {IconUrlFunction}
       */
      this.iconUrlFunction_ = options.iconUrlFunction
        ? options.iconUrlFunction
        : defaultIconUrlFunction;

      this.supportedMediaTypes = ['application/vnd.google-earth.kml+xml'];
    }

    /**
     * @param {Node} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @private
     * @return {Array<Feature>|undefined} Features.
     */
    readDocumentOrFolder_(node, objectStack) {
      // FIXME use scope somehow
      const parsersNS = makeStructureNS(NAMESPACE_URIS$4, {
        'Document': makeArrayExtender(this.readDocumentOrFolder_, this),
        'Folder': makeArrayExtender(this.readDocumentOrFolder_, this),
        'Placemark': makeArrayPusher(this.readPlacemark_, this),
        'Style': this.readSharedStyle_.bind(this),
        'StyleMap': this.readSharedStyleMap_.bind(this),
      });
      /** @type {Array<Feature>} */
      // @ts-ignore
      const features = pushParseAndPop([], parsersNS, node, objectStack, this);
      if (features) {
        return features;
      } else {
        return undefined;
      }
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @private
     * @return {Feature|undefined} Feature.
     */
    readPlacemark_(node, objectStack) {
      const object = pushParseAndPop(
        {'geometry': null},
        PLACEMARK_PARSERS,
        node,
        objectStack,
        this
      );
      if (!object) {
        return undefined;
      }
      const feature = new Feature();
      const id = node.getAttribute('id');
      if (id !== null) {
        feature.setId(id);
      }
      const options = /** @type {import("./Feature.js").ReadOptions} */ (
        objectStack[0]
      );

      const geometry = object['geometry'];
      if (geometry) {
        transformGeometryWithOptions(geometry, false, options);
      }
      feature.setGeometry(geometry);
      delete object['geometry'];

      if (this.extractStyles_) {
        const style = object['Style'];
        const styleUrl = object['styleUrl'];
        const styleFunction = createFeatureStyleFunction(
          style,
          styleUrl,
          this.defaultStyle_,
          this.sharedStyles_,
          this.showPointNames_
        );
        feature.setStyle(styleFunction);
      }
      delete object['Style'];
      // we do not remove the styleUrl property from the object, so it
      // gets stored on feature when setProperties is called

      feature.setProperties(object, true);

      return feature;
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @private
     */
    readSharedStyle_(node, objectStack) {
      const id = node.getAttribute('id');
      if (id !== null) {
        const style = readStyle$2.call(this, node, objectStack);
        if (style) {
          let styleUri;
          let baseURI = node.baseURI;
          if (!baseURI || baseURI == 'about:blank') {
            baseURI = window.location.href;
          }
          if (baseURI) {
            const url = new URL('#' + id, baseURI);
            styleUri = url.href;
          } else {
            styleUri = '#' + id;
          }
          this.sharedStyles_[styleUri] = style;
        }
      }
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @private
     */
    readSharedStyleMap_(node, objectStack) {
      const id = node.getAttribute('id');
      if (id === null) {
        return;
      }
      const styleMapValue = readStyleMapValue.call(this, node, objectStack);
      if (!styleMapValue) {
        return;
      }
      let styleUri;
      let baseURI = node.baseURI;
      if (!baseURI || baseURI == 'about:blank') {
        baseURI = window.location.href;
      }
      if (baseURI) {
        const url = new URL('#' + id, baseURI);
        styleUri = url.href;
      } else {
        styleUri = '#' + id;
      }
      this.sharedStyles_[styleUri] = styleMapValue;
    }

    /**
     * @param {Element} node Node.
     * @param {import("./Feature.js").ReadOptions} [options] Options.
     * @return {import("../Feature.js").default} Feature.
     */
    readFeatureFromNode(node, options) {
      if (!NAMESPACE_URIS$4.includes(node.namespaceURI)) {
        return null;
      }
      const feature = this.readPlacemark_(node, [
        this.getReadOptions(node, options),
      ]);
      if (feature) {
        return feature;
      } else {
        return null;
      }
    }

    /**
     * @protected
     * @param {Element} node Node.
     * @param {import("./Feature.js").ReadOptions} [options] Options.
     * @return {Array<import("../Feature.js").default>} Features.
     */
    readFeaturesFromNode(node, options) {
      if (!NAMESPACE_URIS$4.includes(node.namespaceURI)) {
        return [];
      }
      let features;
      const localName = node.localName;
      if (localName == 'Document' || localName == 'Folder') {
        features = this.readDocumentOrFolder_(node, [
          this.getReadOptions(node, options),
        ]);
        if (features) {
          return features;
        } else {
          return [];
        }
      } else if (localName == 'Placemark') {
        const feature = this.readPlacemark_(node, [
          this.getReadOptions(node, options),
        ]);
        if (feature) {
          return [feature];
        } else {
          return [];
        }
      } else if (localName == 'kml') {
        features = [];
        for (let n = node.firstElementChild; n; n = n.nextElementSibling) {
          const fs = this.readFeaturesFromNode(n, options);
          if (fs) {
            extend$2(features, fs);
          }
        }
        return features;
      } else {
        return [];
      }
    }

    /**
     * Read the name of the KML.
     *
     * @param {Document|Element|string} source Source.
     * @return {string|undefined} Name.
     * @api
     */
    readName(source) {
      if (!source) {
        return undefined;
      } else if (typeof source === 'string') {
        const doc = parse(source);
        return this.readNameFromDocument(doc);
      } else if (isDocument(source)) {
        return this.readNameFromDocument(/** @type {Document} */ (source));
      } else {
        return this.readNameFromNode(/** @type {Element} */ (source));
      }
    }

    /**
     * @param {Document} doc Document.
     * @return {string|undefined} Name.
     */
    readNameFromDocument(doc) {
      for (let n = /** @type {Node} */ (doc.firstChild); n; n = n.nextSibling) {
        if (n.nodeType == Node.ELEMENT_NODE) {
          const name = this.readNameFromNode(/** @type {Element} */ (n));
          if (name) {
            return name;
          }
        }
      }
      return undefined;
    }

    /**
     * @param {Element} node Node.
     * @return {string|undefined} Name.
     */
    readNameFromNode(node) {
      for (let n = node.firstElementChild; n; n = n.nextElementSibling) {
        if (NAMESPACE_URIS$4.includes(n.namespaceURI) && n.localName == 'name') {
          return readString(n);
        }
      }
      for (let n = node.firstElementChild; n; n = n.nextElementSibling) {
        const localName = n.localName;
        if (
          NAMESPACE_URIS$4.includes(n.namespaceURI) &&
          (localName == 'Document' ||
            localName == 'Folder' ||
            localName == 'Placemark' ||
            localName == 'kml')
        ) {
          const name = this.readNameFromNode(n);
          if (name) {
            return name;
          }
        }
      }
      return undefined;
    }

    /**
     * Read the network links of the KML.
     *
     * @param {Document|Element|string} source Source.
     * @return {Array<Object>} Network links.
     * @api
     */
    readNetworkLinks(source) {
      const networkLinks = [];
      if (typeof source === 'string') {
        const doc = parse(source);
        extend$2(networkLinks, this.readNetworkLinksFromDocument(doc));
      } else if (isDocument(source)) {
        extend$2(
          networkLinks,
          this.readNetworkLinksFromDocument(/** @type {Document} */ (source))
        );
      } else {
        extend$2(
          networkLinks,
          this.readNetworkLinksFromNode(/** @type {Element} */ (source))
        );
      }
      return networkLinks;
    }

    /**
     * @param {Document} doc Document.
     * @return {Array<Object>} Network links.
     */
    readNetworkLinksFromDocument(doc) {
      const networkLinks = [];
      for (let n = /** @type {Node} */ (doc.firstChild); n; n = n.nextSibling) {
        if (n.nodeType == Node.ELEMENT_NODE) {
          extend$2(
            networkLinks,
            this.readNetworkLinksFromNode(/** @type {Element} */ (n))
          );
        }
      }
      return networkLinks;
    }

    /**
     * @param {Element} node Node.
     * @return {Array<Object>} Network links.
     */
    readNetworkLinksFromNode(node) {
      const networkLinks = [];
      for (let n = node.firstElementChild; n; n = n.nextElementSibling) {
        if (
          NAMESPACE_URIS$4.includes(n.namespaceURI) &&
          n.localName == 'NetworkLink'
        ) {
          const obj = pushParseAndPop({}, NETWORK_LINK_PARSERS, n, []);
          networkLinks.push(obj);
        }
      }
      for (let n = node.firstElementChild; n; n = n.nextElementSibling) {
        const localName = n.localName;
        if (
          NAMESPACE_URIS$4.includes(n.namespaceURI) &&
          (localName == 'Document' || localName == 'Folder' || localName == 'kml')
        ) {
          extend$2(networkLinks, this.readNetworkLinksFromNode(n));
        }
      }
      return networkLinks;
    }

    /**
     * Read the regions of the KML.
     *
     * @param {Document|Element|string} source Source.
     * @return {Array<Object>} Regions.
     * @api
     */
    readRegion(source) {
      const regions = [];
      if (typeof source === 'string') {
        const doc = parse(source);
        extend$2(regions, this.readRegionFromDocument(doc));
      } else if (isDocument(source)) {
        extend$2(
          regions,
          this.readRegionFromDocument(/** @type {Document} */ (source))
        );
      } else {
        extend$2(regions, this.readRegionFromNode(/** @type {Element} */ (source)));
      }
      return regions;
    }

    /**
     * @param {Document} doc Document.
     * @return {Array<Object>} Region.
     */
    readRegionFromDocument(doc) {
      const regions = [];
      for (let n = /** @type {Node} */ (doc.firstChild); n; n = n.nextSibling) {
        if (n.nodeType == Node.ELEMENT_NODE) {
          extend$2(regions, this.readRegionFromNode(/** @type {Element} */ (n)));
        }
      }
      return regions;
    }

    /**
     * @param {Element} node Node.
     * @return {Array<Object>} Region.
     * @api
     */
    readRegionFromNode(node) {
      const regions = [];
      for (let n = node.firstElementChild; n; n = n.nextElementSibling) {
        if (NAMESPACE_URIS$4.includes(n.namespaceURI) && n.localName == 'Region') {
          const obj = pushParseAndPop({}, REGION_PARSERS, n, []);
          regions.push(obj);
        }
      }
      for (let n = node.firstElementChild; n; n = n.nextElementSibling) {
        const localName = n.localName;
        if (
          NAMESPACE_URIS$4.includes(n.namespaceURI) &&
          (localName == 'Document' || localName == 'Folder' || localName == 'kml')
        ) {
          extend$2(regions, this.readRegionFromNode(n));
        }
      }
      return regions;
    }

    /**
     * Encode an array of features in the KML format as an XML node. GeometryCollections,
     * MultiPoints, MultiLineStrings, and MultiPolygons are output as MultiGeometries.
     *
     * @param {Array<Feature>} features Features.
     * @param {import("./Feature.js").WriteOptions} [options] Options.
     * @return {Node} Node.
     * @api
     */
    writeFeaturesNode(features, options) {
      options = this.adaptOptions(options);
      const kml = createElementNS(NAMESPACE_URIS$4[4], 'kml');
      const xmlnsUri = 'http://www.w3.org/2000/xmlns/';
      kml.setAttributeNS(xmlnsUri, 'xmlns:gx', GX_NAMESPACE_URIS[0]);
      kml.setAttributeNS(xmlnsUri, 'xmlns:xsi', XML_SCHEMA_INSTANCE_URI);
      kml.setAttributeNS(
        XML_SCHEMA_INSTANCE_URI,
        'xsi:schemaLocation',
        SCHEMA_LOCATION
      );

      const /** @type {import("../xml.js").NodeStackItem} */ context = {
          node: kml,
        };
      /** @type {!Object<string, (Array<Feature>|Feature|undefined)>} */
      const properties = {};
      if (features.length > 1) {
        properties['Document'] = features;
      } else if (features.length == 1) {
        properties['Placemark'] = features[0];
      }
      const orderedKeys = KML_SEQUENCE[kml.namespaceURI];
      const values = makeSequence(properties, orderedKeys);
      pushSerializeAndPop(
        context,
        KML_SERIALIZERS,
        OBJECT_PROPERTY_NODE_FACTORY,
        values,
        [options],
        orderedKeys,
        this
      );
      return kml;
    }
  }

  /**
   * @param {Style|undefined} foundStyle Style.
   * @param {string} name Name.
   * @return {Style} style Style.
   */
  function createNameStyleFunction(foundStyle, name) {
    const textOffset = [0, 0];
    /** @type {CanvasTextAlign} */
    let textAlign = 'start';
    const imageStyle = foundStyle.getImage();
    if (imageStyle) {
      const imageSize = imageStyle.getSize();
      if (imageSize && imageSize.length == 2) {
        const imageScale = imageStyle.getScaleArray();
        const anchor = imageStyle.getAnchor();
        // Offset the label to be centered to the right of the icon,
        // if there is one.
        textOffset[0] = imageScale[0] * (imageSize[0] - anchor[0]);
        textOffset[1] = imageScale[1] * (imageSize[1] / 2 - anchor[1]);
        textAlign = 'left';
      }
    }
    let textStyle = foundStyle.getText();
    if (textStyle) {
      // clone the text style, customizing it with name, alignments and offset.
      // Note that kml does not support many text options that OpenLayers does (rotation, textBaseline).
      textStyle = textStyle.clone();
      textStyle.setFont(textStyle.getFont() || DEFAULT_TEXT_STYLE.getFont());
      textStyle.setScale(textStyle.getScale() || DEFAULT_TEXT_STYLE.getScale());
      textStyle.setFill(textStyle.getFill() || DEFAULT_TEXT_STYLE.getFill());
      textStyle.setStroke(textStyle.getStroke() || DEFAULT_TEXT_STROKE_STYLE);
    } else {
      textStyle = DEFAULT_TEXT_STYLE.clone();
    }
    textStyle.setText(name);
    textStyle.setOffsetX(textOffset[0]);
    textStyle.setOffsetY(textOffset[1]);
    textStyle.setTextAlign(textAlign);

    const nameStyle = new Style({
      image: imageStyle,
      text: textStyle,
    });
    return nameStyle;
  }

  /**
   * @param {Array<Style>|undefined} style Style.
   * @param {string} styleUrl Style URL.
   * @param {Array<Style>} defaultStyle Default style.
   * @param {!Object<string, (Array<Style>|string)>} sharedStyles Shared styles.
   * @param {boolean|undefined} showPointNames true to show names for point placemarks.
   * @return {import("../style/Style.js").StyleFunction} Feature style function.
   */
  function createFeatureStyleFunction(
    style,
    styleUrl,
    defaultStyle,
    sharedStyles,
    showPointNames
  ) {
    return (
      /**
       * @param {Feature} feature feature.
       * @param {number} resolution Resolution.
       * @return {Array<Style>|Style} Style.
       */
      function (feature, resolution) {
        let drawName = showPointNames;
        let name = '';
        let multiGeometryPoints = [];
        if (drawName) {
          const geometry = feature.getGeometry();
          if (geometry) {
            if (geometry instanceof GeometryCollection) {
              multiGeometryPoints = geometry
                .getGeometriesArrayRecursive()
                .filter(function (geometry) {
                  const type = geometry.getType();
                  return type === 'Point' || type === 'MultiPoint';
                });
              drawName = multiGeometryPoints.length > 0;
            } else {
              const type = geometry.getType();
              drawName = type === 'Point' || type === 'MultiPoint';
            }
          }
        }

        if (drawName) {
          name = /** @type {string} */ (feature.get('name'));
          drawName = drawName && !!name;
          // convert any html character codes
          if (drawName && /&[^&]+;/.test(name)) {
            if (!TEXTAREA) {
              TEXTAREA = document.createElement('textarea');
            }
            TEXTAREA.innerHTML = name;
            name = TEXTAREA.value;
          }
        }

        let featureStyle = defaultStyle;
        if (style) {
          featureStyle = style;
        } else if (styleUrl) {
          featureStyle = findStyle(styleUrl, defaultStyle, sharedStyles);
        }
        if (drawName) {
          const nameStyle = createNameStyleFunction(featureStyle[0], name);
          if (multiGeometryPoints.length > 0) {
            // in multigeometries restrict the name style to points and create a
            // style without image or text for geometries requiring fill or stroke
            // including any polygon specific style if there is one
            nameStyle.setGeometry(new GeometryCollection(multiGeometryPoints));
            const baseStyle = new Style({
              geometry: featureStyle[0].getGeometry(),
              image: null,
              fill: featureStyle[0].getFill(),
              stroke: featureStyle[0].getStroke(),
              text: null,
            });
            return [nameStyle, baseStyle].concat(featureStyle.slice(1));
          }
          return nameStyle;
        }
        return featureStyle;
      }
    );
  }

  /**
   * @param {Array<Style>|string|undefined} styleValue Style value.
   * @param {Array<Style>} defaultStyle Default style.
   * @param {!Object<string, (Array<Style>|string)>} sharedStyles
   * Shared styles.
   * @return {Array<Style>} Style.
   */
  function findStyle(styleValue, defaultStyle, sharedStyles) {
    if (Array.isArray(styleValue)) {
      return styleValue;
    } else if (typeof styleValue === 'string') {
      return findStyle(sharedStyles[styleValue], defaultStyle, sharedStyles);
    } else {
      return defaultStyle;
    }
  }

  /**
   * @param {Node} node Node.
   * @return {import("../color.js").Color|undefined} Color.
   */
  function readColor(node) {
    const s = getAllTextContent(node, false);
    // The KML specification states that colors should not include a leading `#`
    // but we tolerate them.
    const m = /^\s*#?\s*([0-9A-Fa-f]{8})\s*$/.exec(s);
    if (m) {
      const hexColor = m[1];
      return [
        parseInt(hexColor.substr(6, 2), 16),
        parseInt(hexColor.substr(4, 2), 16),
        parseInt(hexColor.substr(2, 2), 16),
        parseInt(hexColor.substr(0, 2), 16) / 255,
      ];
    } else {
      return undefined;
    }
  }

  /**
   * @param {Node} node Node.
   * @return {Array<number>|undefined} Flat coordinates.
   */
  function readFlatCoordinates(node) {
    let s = getAllTextContent(node, false);
    const flatCoordinates = [];
    // The KML specification states that coordinate tuples should not include
    // spaces, but we tolerate them.
    s = s.replace(/\s*,\s*/g, ',');
    const re =
      /^\s*([+\-]?\d*\.?\d+(?:e[+\-]?\d+)?),([+\-]?\d*\.?\d+(?:e[+\-]?\d+)?)(?:\s+|,|$)(?:([+\-]?\d*\.?\d+(?:e[+\-]?\d+)?)(?:\s+|$))?\s*/i;
    let m;
    while ((m = re.exec(s))) {
      const x = parseFloat(m[1]);
      const y = parseFloat(m[2]);
      const z = m[3] ? parseFloat(m[3]) : 0;
      flatCoordinates.push(x, y, z);
      s = s.substr(m[0].length);
    }
    if (s !== '') {
      return undefined;
    }
    return flatCoordinates;
  }

  /**
   * @param {Node} node Node.
   * @return {string} URI.
   */
  function readURI(node) {
    const s = getAllTextContent(node, false).trim();
    let baseURI = node.baseURI;
    if (!baseURI || baseURI == 'about:blank') {
      baseURI = window.location.href;
    }
    if (baseURI) {
      const url = new URL(s, baseURI);
      return url.href;
    } else {
      return s;
    }
  }

  /**
   * @param {Node} node Node.
   * @return {string} URI.
   */
  function readStyleURL(node) {
    // KML files in the wild occasionally forget the leading
    // `#` on styleUrlsdefined in the same document.
    const s = getAllTextContent(node, false)
      .trim()
      .replace(/^(?!.*#)/, '#');
    let baseURI = node.baseURI;
    if (!baseURI || baseURI == 'about:blank') {
      baseURI = window.location.href;
    }
    if (baseURI) {
      const url = new URL(s, baseURI);
      return url.href;
    } else {
      return s;
    }
  }

  /**
   * @param {Element} node Node.
   * @return {Vec2} Vec2.
   */
  function readVec2(node) {
    const xunits = node.getAttribute('xunits');
    const yunits = node.getAttribute('yunits');
    /** @type {import('../style/Icon.js').IconOrigin} */
    let origin;
    if (xunits !== 'insetPixels') {
      if (yunits !== 'insetPixels') {
        origin = 'bottom-left';
      } else {
        origin = 'top-left';
      }
    } else {
      if (yunits !== 'insetPixels') {
        origin = 'bottom-right';
      } else {
        origin = 'top-right';
      }
    }
    return {
      x: parseFloat(node.getAttribute('x')),
      xunits: ICON_ANCHOR_UNITS_MAP[xunits],
      y: parseFloat(node.getAttribute('y')),
      yunits: ICON_ANCHOR_UNITS_MAP[yunits],
      origin: origin,
    };
  }

  /**
   * @param {Node} node Node.
   * @return {number|undefined} Scale.
   */
  function readScale(node) {
    return readDecimal(node);
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const STYLE_MAP_PARSERS = makeStructureNS(NAMESPACE_URIS$4, {
    'Pair': pairDataParser,
  });

  /**
   * @this {KML}
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Array<Style>|string|undefined} StyleMap.
   */
  function readStyleMapValue(node, objectStack) {
    return pushParseAndPop(undefined, STYLE_MAP_PARSERS, node, objectStack, this);
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const ICON_STYLE_PARSERS = makeStructureNS(NAMESPACE_URIS$4, {
    'Icon': makeObjectPropertySetter(readIcon),
    'color': makeObjectPropertySetter(readColor),
    'heading': makeObjectPropertySetter(readDecimal),
    'hotSpot': makeObjectPropertySetter(readVec2),
    'scale': makeObjectPropertySetter(readScale),
  });

  /**
   * @this {KML}
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function iconStyleParser(node, objectStack) {
    // FIXME refreshMode
    // FIXME refreshInterval
    // FIXME viewRefreshTime
    // FIXME viewBoundScale
    // FIXME viewFormat
    // FIXME httpQuery
    const object = pushParseAndPop({}, ICON_STYLE_PARSERS, node, objectStack);
    if (!object) {
      return;
    }
    const styleObject = /** @type {Object} */ (
      objectStack[objectStack.length - 1]
    );
    const IconObject = 'Icon' in object ? object['Icon'] : {};
    const drawIcon = !('Icon' in object) || Object.keys(IconObject).length > 0;
    let src;
    const href = /** @type {string|undefined} */ (IconObject['href']);
    if (href) {
      src = href;
    } else if (drawIcon) {
      src = DEFAULT_IMAGE_STYLE_SRC;
    }
    let anchor, anchorXUnits, anchorYUnits;
    /** @type {import('../style/Icon.js').IconOrigin|undefined} */
    let anchorOrigin = 'bottom-left';
    const hotSpot = /** @type {Vec2|undefined} */ (object['hotSpot']);
    if (hotSpot) {
      anchor = [hotSpot.x, hotSpot.y];
      anchorXUnits = hotSpot.xunits;
      anchorYUnits = hotSpot.yunits;
      anchorOrigin = hotSpot.origin;
    } else if (/^https?:\/\/maps\.(?:google|gstatic)\.com\//.test(src)) {
      // Google hotspots from https://kml4earth.appspot.com/icons.html#notes
      if (src.includes('pushpin')) {
        anchor = DEFAULT_IMAGE_STYLE_ANCHOR;
        anchorXUnits = DEFAULT_IMAGE_STYLE_ANCHOR_X_UNITS;
        anchorYUnits = DEFAULT_IMAGE_STYLE_ANCHOR_Y_UNITS;
      } else if (src.includes('arrow-reverse')) {
        anchor = [54, 42];
        anchorXUnits = DEFAULT_IMAGE_STYLE_ANCHOR_X_UNITS;
        anchorYUnits = DEFAULT_IMAGE_STYLE_ANCHOR_Y_UNITS;
      } else if (src.includes('paddle')) {
        anchor = [32, 1];
        anchorXUnits = DEFAULT_IMAGE_STYLE_ANCHOR_X_UNITS;
        anchorYUnits = DEFAULT_IMAGE_STYLE_ANCHOR_Y_UNITS;
      }
    }

    let offset;
    const x = /** @type {number|undefined} */ (IconObject['x']);
    const y = /** @type {number|undefined} */ (IconObject['y']);
    if (x !== undefined && y !== undefined) {
      offset = [x, y];
    }

    let size;
    const w = /** @type {number|undefined} */ (IconObject['w']);
    const h = /** @type {number|undefined} */ (IconObject['h']);
    if (w !== undefined && h !== undefined) {
      size = [w, h];
    }

    let rotation;
    const heading = /** @type {number} */ (object['heading']);
    if (heading !== undefined) {
      rotation = toRadians(heading);
    }

    const scale = /** @type {number|undefined} */ (object['scale']);

    const color = /** @type {Array<number>|undefined} */ (object['color']);

    if (drawIcon) {
      if (src == DEFAULT_IMAGE_STYLE_SRC) {
        size = DEFAULT_IMAGE_STYLE_SIZE;
      }

      const imageStyle = new Icon({
        anchor: anchor,
        anchorOrigin: anchorOrigin,
        anchorXUnits: anchorXUnits,
        anchorYUnits: anchorYUnits,
        crossOrigin: this.crossOrigin_,
        offset: offset,
        offsetOrigin: 'bottom-left',
        rotation: rotation,
        scale: scale,
        size: size,
        src: this.iconUrlFunction_(src),
        color: color,
      });

      const imageScale = imageStyle.getScaleArray()[0];
      const imageSize = imageStyle.getSize();
      if (imageSize === null) {
        const imageState = imageStyle.getImageState();
        if (imageState === ImageState.IDLE || imageState === ImageState.LOADING) {
          const listener = function () {
            const imageState = imageStyle.getImageState();
            if (
              !(
                imageState === ImageState.IDLE ||
                imageState === ImageState.LOADING
              )
            ) {
              const imageSize = imageStyle.getSize();
              if (imageSize && imageSize.length == 2) {
                const resizeScale = scaleForSize(imageSize);
                imageStyle.setScale(imageScale * resizeScale);
              }
              imageStyle.unlistenImageChange(listener);
            }
          };
          imageStyle.listenImageChange(listener);
          if (imageState === ImageState.IDLE) {
            imageStyle.load();
          }
        }
      } else if (imageSize.length == 2) {
        const resizeScale = scaleForSize(imageSize);
        imageStyle.setScale(imageScale * resizeScale);
      }
      styleObject['imageStyle'] = imageStyle;
    } else {
      // handle the case when we explicitly want to draw no icon.
      styleObject['imageStyle'] = DEFAULT_NO_IMAGE_STYLE;
    }
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const LABEL_STYLE_PARSERS = makeStructureNS(NAMESPACE_URIS$4, {
    'color': makeObjectPropertySetter(readColor),
    'scale': makeObjectPropertySetter(readScale),
  });

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function labelStyleParser(node, objectStack) {
    // FIXME colorMode
    const object = pushParseAndPop({}, LABEL_STYLE_PARSERS, node, objectStack);
    if (!object) {
      return;
    }
    const styleObject = objectStack[objectStack.length - 1];
    const textStyle = new Text({
      fill: new Fill({
        color:
          /** @type {import("../color.js").Color} */
          ('color' in object ? object['color'] : DEFAULT_COLOR),
      }),
      scale: /** @type {number|undefined} */ (object['scale']),
    });
    styleObject['textStyle'] = textStyle;
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const LINE_STYLE_PARSERS = makeStructureNS(NAMESPACE_URIS$4, {
    'color': makeObjectPropertySetter(readColor),
    'width': makeObjectPropertySetter(readDecimal),
  });

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function lineStyleParser(node, objectStack) {
    // FIXME colorMode
    // FIXME gx:outerColor
    // FIXME gx:outerWidth
    // FIXME gx:physicalWidth
    // FIXME gx:labelVisibility
    const object = pushParseAndPop({}, LINE_STYLE_PARSERS, node, objectStack);
    if (!object) {
      return;
    }
    const styleObject = objectStack[objectStack.length - 1];
    const strokeStyle = new Stroke({
      color:
        /** @type {import("../color.js").Color} */
        ('color' in object ? object['color'] : DEFAULT_COLOR),
      width: /** @type {number} */ ('width' in object ? object['width'] : 1),
    });
    styleObject['strokeStyle'] = strokeStyle;
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const POLY_STYLE_PARSERS = makeStructureNS(NAMESPACE_URIS$4, {
    'color': makeObjectPropertySetter(readColor),
    'fill': makeObjectPropertySetter(readBoolean),
    'outline': makeObjectPropertySetter(readBoolean),
  });

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function polyStyleParser(node, objectStack) {
    // FIXME colorMode
    const object = pushParseAndPop({}, POLY_STYLE_PARSERS, node, objectStack);
    if (!object) {
      return;
    }
    const styleObject = objectStack[objectStack.length - 1];
    const fillStyle = new Fill({
      color:
        /** @type {import("../color.js").Color} */
        ('color' in object ? object['color'] : DEFAULT_COLOR),
    });
    styleObject['fillStyle'] = fillStyle;
    const fill = /** @type {boolean|undefined} */ (object['fill']);
    if (fill !== undefined) {
      styleObject['fill'] = fill;
    }
    const outline = /** @type {boolean|undefined} */ (object['outline']);
    if (outline !== undefined) {
      styleObject['outline'] = outline;
    }
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const FLAT_LINEAR_RING_PARSERS = makeStructureNS(NAMESPACE_URIS$4, {
    'coordinates': makeReplacer(readFlatCoordinates),
  });

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Array<number>} LinearRing flat coordinates.
   */
  function readFlatLinearRing(node, objectStack) {
    return pushParseAndPop(null, FLAT_LINEAR_RING_PARSERS, node, objectStack);
  }

  /**
   * @param {Node} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function gxCoordParser(node, objectStack) {
    const gxTrackObject =
      /** @type {GxTrackObject} */
      (objectStack[objectStack.length - 1]);
    const coordinates = gxTrackObject.coordinates;
    const s = getAllTextContent(node, false);
    const re =
      /^\s*([+\-]?\d+(?:\.\d*)?(?:e[+\-]?\d*)?)\s+([+\-]?\d+(?:\.\d*)?(?:e[+\-]?\d*)?)\s+([+\-]?\d+(?:\.\d*)?(?:e[+\-]?\d*)?)\s*$/i;
    const m = re.exec(s);
    if (m) {
      const x = parseFloat(m[1]);
      const y = parseFloat(m[2]);
      const z = parseFloat(m[3]);
      coordinates.push([x, y, z]);
    } else {
      coordinates.push([]);
    }
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const GX_MULTITRACK_GEOMETRY_PARSERS = makeStructureNS(GX_NAMESPACE_URIS, {
    'Track': makeArrayPusher(readGxTrack),
  });

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {MultiLineString|undefined} MultiLineString.
   */
  function readGxMultiTrack(node, objectStack) {
    const lineStrings = pushParseAndPop(
      [],
      GX_MULTITRACK_GEOMETRY_PARSERS,
      node,
      objectStack
    );
    if (!lineStrings) {
      return undefined;
    }
    return new MultiLineString(lineStrings);
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const GX_TRACK_PARSERS = makeStructureNS(
    NAMESPACE_URIS$4,
    {
      'when': whenParser,
    },
    makeStructureNS(GX_NAMESPACE_URIS, {
      'coord': gxCoordParser,
    })
  );

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {LineString|undefined} LineString.
   */
  function readGxTrack(node, objectStack) {
    const gxTrackObject = pushParseAndPop(
      /** @type {GxTrackObject} */ ({
        coordinates: [],
        whens: [],
      }),
      GX_TRACK_PARSERS,
      node,
      objectStack
    );
    if (!gxTrackObject) {
      return undefined;
    }
    const flatCoordinates = [];
    const coordinates = gxTrackObject.coordinates;
    const whens = gxTrackObject.whens;
    for (
      let i = 0, ii = Math.min(coordinates.length, whens.length);
      i < ii;
      ++i
    ) {
      if (coordinates[i].length == 3) {
        flatCoordinates.push(
          coordinates[i][0],
          coordinates[i][1],
          coordinates[i][2],
          whens[i]
        );
      }
    }
    return new LineString(flatCoordinates, 'XYZM');
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const ICON_PARSERS = makeStructureNS(
    NAMESPACE_URIS$4,
    {
      'href': makeObjectPropertySetter(readURI),
    },
    makeStructureNS(GX_NAMESPACE_URIS, {
      'x': makeObjectPropertySetter(readDecimal),
      'y': makeObjectPropertySetter(readDecimal),
      'w': makeObjectPropertySetter(readDecimal),
      'h': makeObjectPropertySetter(readDecimal),
    })
  );

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object} Icon object.
   */
  function readIcon(node, objectStack) {
    const iconObject = pushParseAndPop({}, ICON_PARSERS, node, objectStack);
    if (iconObject) {
      return iconObject;
    } else {
      return null;
    }
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const GEOMETRY_FLAT_COORDINATES_PARSERS = makeStructureNS(NAMESPACE_URIS$4, {
    'coordinates': makeReplacer(readFlatCoordinates),
  });

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Array<number>} Flat coordinates.
   */
  function readFlatCoordinatesFromNode(node, objectStack) {
    return pushParseAndPop(
      null,
      GEOMETRY_FLAT_COORDINATES_PARSERS,
      node,
      objectStack
    );
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const EXTRUDE_AND_ALTITUDE_MODE_PARSERS = makeStructureNS(NAMESPACE_URIS$4, {
    'extrude': makeObjectPropertySetter(readBoolean),
    'tessellate': makeObjectPropertySetter(readBoolean),
    'altitudeMode': makeObjectPropertySetter(readString),
  });

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {LineString|undefined} LineString.
   */
  function readLineString(node, objectStack) {
    const properties = pushParseAndPop(
      {},
      EXTRUDE_AND_ALTITUDE_MODE_PARSERS,
      node,
      objectStack
    );
    const flatCoordinates = readFlatCoordinatesFromNode(node, objectStack);
    if (flatCoordinates) {
      const lineString = new LineString(flatCoordinates, 'XYZ');
      lineString.setProperties(properties, true);
      return lineString;
    } else {
      return undefined;
    }
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Polygon|undefined} Polygon.
   */
  function readLinearRing(node, objectStack) {
    const properties = pushParseAndPop(
      {},
      EXTRUDE_AND_ALTITUDE_MODE_PARSERS,
      node,
      objectStack
    );
    const flatCoordinates = readFlatCoordinatesFromNode(node, objectStack);
    if (flatCoordinates) {
      const polygon = new Polygon(flatCoordinates, 'XYZ', [
        flatCoordinates.length,
      ]);
      polygon.setProperties(properties, true);
      return polygon;
    } else {
      return undefined;
    }
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const MULTI_GEOMETRY_PARSERS = makeStructureNS(NAMESPACE_URIS$4, {
    'LineString': makeArrayPusher(readLineString),
    'LinearRing': makeArrayPusher(readLinearRing),
    'MultiGeometry': makeArrayPusher(readMultiGeometry),
    'Point': makeArrayPusher(readPoint),
    'Polygon': makeArrayPusher(readPolygon),
  });

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {import("../geom/Geometry.js").default} Geometry.
   */
  function readMultiGeometry(node, objectStack) {
    const geometries = pushParseAndPop(
      [],
      MULTI_GEOMETRY_PARSERS,
      node,
      objectStack
    );
    if (!geometries) {
      return null;
    }
    if (geometries.length === 0) {
      return new GeometryCollection(geometries);
    }
    let multiGeometry;
    let homogeneous = true;
    const type = geometries[0].getType();
    let geometry;
    for (let i = 1, ii = geometries.length; i < ii; ++i) {
      geometry = geometries[i];
      if (geometry.getType() != type) {
        homogeneous = false;
        break;
      }
    }
    if (homogeneous) {
      let layout;
      let flatCoordinates;
      if (type == 'Point') {
        const point = geometries[0];
        layout = point.getLayout();
        flatCoordinates = point.getFlatCoordinates();
        for (let i = 1, ii = geometries.length; i < ii; ++i) {
          geometry = geometries[i];
          extend$2(flatCoordinates, geometry.getFlatCoordinates());
        }
        multiGeometry = new MultiPoint(flatCoordinates, layout);
        setCommonGeometryProperties(multiGeometry, geometries);
      } else if (type == 'LineString') {
        multiGeometry = new MultiLineString(geometries);
        setCommonGeometryProperties(multiGeometry, geometries);
      } else if (type == 'Polygon') {
        multiGeometry = new MultiPolygon(geometries);
        setCommonGeometryProperties(multiGeometry, geometries);
      } else if (type == 'GeometryCollection') {
        multiGeometry = new GeometryCollection(geometries);
      } else {
        assert(false, 37); // Unknown geometry type found
      }
    } else {
      multiGeometry = new GeometryCollection(geometries);
    }
    return /** @type {import("../geom/Geometry.js").default} */ (multiGeometry);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Point|undefined} Point.
   */
  function readPoint(node, objectStack) {
    const properties = pushParseAndPop(
      {},
      EXTRUDE_AND_ALTITUDE_MODE_PARSERS,
      node,
      objectStack
    );
    const flatCoordinates = readFlatCoordinatesFromNode(node, objectStack);
    if (flatCoordinates) {
      const point = new Point(flatCoordinates, 'XYZ');
      point.setProperties(properties, true);
      return point;
    } else {
      return undefined;
    }
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const FLAT_LINEAR_RINGS_PARSERS = makeStructureNS(NAMESPACE_URIS$4, {
    'innerBoundaryIs': innerBoundaryIsParser,
    'outerBoundaryIs': outerBoundaryIsParser,
  });

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Polygon|undefined} Polygon.
   */
  function readPolygon(node, objectStack) {
    const properties = pushParseAndPop(
      /** @type {Object<string,*>} */ ({}),
      EXTRUDE_AND_ALTITUDE_MODE_PARSERS,
      node,
      objectStack
    );
    const flatLinearRings = pushParseAndPop(
      [null],
      FLAT_LINEAR_RINGS_PARSERS,
      node,
      objectStack
    );
    if (flatLinearRings && flatLinearRings[0]) {
      const flatCoordinates = flatLinearRings[0];
      const ends = [flatCoordinates.length];
      for (let i = 1, ii = flatLinearRings.length; i < ii; ++i) {
        extend$2(flatCoordinates, flatLinearRings[i]);
        ends.push(flatCoordinates.length);
      }
      const polygon = new Polygon(flatCoordinates, 'XYZ', ends);
      polygon.setProperties(properties, true);
      return polygon;
    } else {
      return undefined;
    }
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const STYLE_PARSERS$2 = makeStructureNS(NAMESPACE_URIS$4, {
    'IconStyle': iconStyleParser,
    'LabelStyle': labelStyleParser,
    'LineStyle': lineStyleParser,
    'PolyStyle': polyStyleParser,
  });

  /**
   * @this {KML}
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Array<Style>} Style.
   */
  function readStyle$2(node, objectStack) {
    const styleObject = pushParseAndPop(
      {},
      STYLE_PARSERS$2,
      node,
      objectStack,
      this
    );
    if (!styleObject) {
      return null;
    }
    let fillStyle =
      /** @type {Fill} */
      (
        'fillStyle' in styleObject ? styleObject['fillStyle'] : DEFAULT_FILL_STYLE
      );
    const fill = /** @type {boolean|undefined} */ (styleObject['fill']);
    if (fill !== undefined && !fill) {
      fillStyle = null;
    }
    let imageStyle;
    if ('imageStyle' in styleObject) {
      if (styleObject['imageStyle'] != DEFAULT_NO_IMAGE_STYLE) {
        imageStyle = styleObject['imageStyle'];
      }
    } else {
      imageStyle = DEFAULT_IMAGE_STYLE;
    }
    const textStyle =
      /** @type {Text} */
      (
        'textStyle' in styleObject ? styleObject['textStyle'] : DEFAULT_TEXT_STYLE
      );
    const strokeStyle =
      /** @type {Stroke} */
      (
        'strokeStyle' in styleObject
          ? styleObject['strokeStyle']
          : DEFAULT_STROKE_STYLE
      );
    const outline = /** @type {boolean|undefined} */ (styleObject['outline']);
    if (outline !== undefined && !outline) {
      // if the polystyle specifies no outline two styles are needed,
      // one for non-polygon geometries where linestrings require a stroke
      // and one for polygons where there should be no stroke
      return [
        new Style({
          geometry: function (feature) {
            const geometry = feature.getGeometry();
            const type = geometry.getType();
            if (type === 'GeometryCollection') {
              const collection =
                /** @type {import("../geom/GeometryCollection").default} */ (
                  geometry
                );
              return new GeometryCollection(
                collection
                  .getGeometriesArrayRecursive()
                  .filter(function (geometry) {
                    const type = geometry.getType();
                    return type !== 'Polygon' && type !== 'MultiPolygon';
                  })
              );
            } else if (type !== 'Polygon' && type !== 'MultiPolygon') {
              return geometry;
            }
          },
          fill: fillStyle,
          image: imageStyle,
          stroke: strokeStyle,
          text: textStyle,
          zIndex: undefined, // FIXME
        }),
        new Style({
          geometry: function (feature) {
            const geometry = feature.getGeometry();
            const type = geometry.getType();
            if (type === 'GeometryCollection') {
              const collection =
                /** @type {import("../geom/GeometryCollection").default} */ (
                  geometry
                );
              return new GeometryCollection(
                collection
                  .getGeometriesArrayRecursive()
                  .filter(function (geometry) {
                    const type = geometry.getType();
                    return type === 'Polygon' || type === 'MultiPolygon';
                  })
              );
            } else if (type === 'Polygon' || type === 'MultiPolygon') {
              return geometry;
            }
          },
          fill: fillStyle,
          stroke: null,
          zIndex: undefined, // FIXME
        }),
      ];
    }
    return [
      new Style({
        fill: fillStyle,
        image: imageStyle,
        stroke: strokeStyle,
        text: textStyle,
        zIndex: undefined, // FIXME
      }),
    ];
  }

  /**
   * Reads an array of geometries and creates arrays for common geometry
   * properties. Then sets them to the multi geometry.
   * @param {MultiPoint|MultiLineString|MultiPolygon} multiGeometry A multi-geometry.
   * @param {Array<import("../geom/Geometry.js").default>} geometries List of geometries.
   */
  function setCommonGeometryProperties(multiGeometry, geometries) {
    const ii = geometries.length;
    const extrudes = new Array(geometries.length);
    const tessellates = new Array(geometries.length);
    const altitudeModes = new Array(geometries.length);
    let hasExtrude, hasTessellate, hasAltitudeMode;
    hasExtrude = false;
    hasTessellate = false;
    hasAltitudeMode = false;
    for (let i = 0; i < ii; ++i) {
      const geometry = geometries[i];
      extrudes[i] = geometry.get('extrude');
      tessellates[i] = geometry.get('tessellate');
      altitudeModes[i] = geometry.get('altitudeMode');
      hasExtrude = hasExtrude || extrudes[i] !== undefined;
      hasTessellate = hasTessellate || tessellates[i] !== undefined;
      hasAltitudeMode = hasAltitudeMode || altitudeModes[i];
    }
    if (hasExtrude) {
      multiGeometry.set('extrude', extrudes);
    }
    if (hasTessellate) {
      multiGeometry.set('tessellate', tessellates);
    }
    if (hasAltitudeMode) {
      multiGeometry.set('altitudeMode', altitudeModes);
    }
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const DATA_PARSERS = makeStructureNS(NAMESPACE_URIS$4, {
    'displayName': makeObjectPropertySetter(readString),
    'value': makeObjectPropertySetter(readString),
  });

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function dataParser(node, objectStack) {
    const name = node.getAttribute('name');
    parseNode(DATA_PARSERS, node, objectStack);
    const featureObject = /** @type {Object} */ (
      objectStack[objectStack.length - 1]
    );
    if (name && featureObject.displayName) {
      featureObject[name] = {
        value: featureObject.value,
        displayName: featureObject.displayName,
        toString: function () {
          return featureObject.value;
        },
      };
    } else if (name !== null) {
      featureObject[name] = featureObject.value;
    } else if (featureObject.displayName !== null) {
      featureObject[featureObject.displayName] = featureObject.value;
    }
    delete featureObject['value'];
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const EXTENDED_DATA_PARSERS = makeStructureNS(NAMESPACE_URIS$4, {
    'Data': dataParser,
    'SchemaData': schemaDataParser,
  });

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function extendedDataParser(node, objectStack) {
    parseNode(EXTENDED_DATA_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function regionParser(node, objectStack) {
    parseNode(REGION_PARSERS, node, objectStack);
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const PAIR_PARSERS = makeStructureNS(NAMESPACE_URIS$4, {
    'Style': makeObjectPropertySetter(readStyle$2),
    'key': makeObjectPropertySetter(readString),
    'styleUrl': makeObjectPropertySetter(readStyleURL),
  });

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function pairDataParser(node, objectStack) {
    const pairObject = pushParseAndPop({}, PAIR_PARSERS, node, objectStack, this);
    if (!pairObject) {
      return;
    }
    const key = /** @type {string|undefined} */ (pairObject['key']);
    if (key && key == 'normal') {
      const styleUrl = /** @type {string|undefined} */ (pairObject['styleUrl']);
      if (styleUrl) {
        objectStack[objectStack.length - 1] = styleUrl;
      }
      const style = /** @type {Style} */ (pairObject['Style']);
      if (style) {
        objectStack[objectStack.length - 1] = style;
      }
    }
  }

  /**
   * @this {KML}
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function placemarkStyleMapParser(node, objectStack) {
    const styleMapValue = readStyleMapValue.call(this, node, objectStack);
    if (!styleMapValue) {
      return;
    }
    const placemarkObject = objectStack[objectStack.length - 1];
    if (Array.isArray(styleMapValue)) {
      placemarkObject['Style'] = styleMapValue;
    } else if (typeof styleMapValue === 'string') {
      placemarkObject['styleUrl'] = styleMapValue;
    } else {
      assert(false, 38); // `styleMapValue` has an unknown type
    }
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const SCHEMA_DATA_PARSERS = makeStructureNS(NAMESPACE_URIS$4, {
    'SimpleData': simpleDataParser,
  });

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function schemaDataParser(node, objectStack) {
    parseNode(SCHEMA_DATA_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function simpleDataParser(node, objectStack) {
    const name = node.getAttribute('name');
    if (name !== null) {
      const data = readString(node);
      const featureObject = /** @type {Object} */ (
        objectStack[objectStack.length - 1]
      );
      featureObject[name] = data;
    }
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const LAT_LON_ALT_BOX_PARSERS = makeStructureNS(NAMESPACE_URIS$4, {
    'altitudeMode': makeObjectPropertySetter(readString),
    'minAltitude': makeObjectPropertySetter(readDecimal),
    'maxAltitude': makeObjectPropertySetter(readDecimal),
    'north': makeObjectPropertySetter(readDecimal),
    'south': makeObjectPropertySetter(readDecimal),
    'east': makeObjectPropertySetter(readDecimal),
    'west': makeObjectPropertySetter(readDecimal),
  });

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function latLonAltBoxParser(node, objectStack) {
    const object = pushParseAndPop(
      {},
      LAT_LON_ALT_BOX_PARSERS,
      node,
      objectStack
    );
    if (!object) {
      return;
    }
    const regionObject = /** @type {Object} */ (
      objectStack[objectStack.length - 1]
    );
    const extent = [
      parseFloat(object['west']),
      parseFloat(object['south']),
      parseFloat(object['east']),
      parseFloat(object['north']),
    ];
    regionObject['extent'] = extent;
    regionObject['altitudeMode'] = object['altitudeMode'];
    regionObject['minAltitude'] = parseFloat(object['minAltitude']);
    regionObject['maxAltitude'] = parseFloat(object['maxAltitude']);
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const LOD_PARSERS = makeStructureNS(NAMESPACE_URIS$4, {
    'minLodPixels': makeObjectPropertySetter(readDecimal),
    'maxLodPixels': makeObjectPropertySetter(readDecimal),
    'minFadeExtent': makeObjectPropertySetter(readDecimal),
    'maxFadeExtent': makeObjectPropertySetter(readDecimal),
  });

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function lodParser(node, objectStack) {
    const object = pushParseAndPop({}, LOD_PARSERS, node, objectStack);
    if (!object) {
      return;
    }
    const lodObject = /** @type {Object} */ (objectStack[objectStack.length - 1]);
    lodObject['minLodPixels'] = parseFloat(object['minLodPixels']);
    lodObject['maxLodPixels'] = parseFloat(object['maxLodPixels']);
    lodObject['minFadeExtent'] = parseFloat(object['minFadeExtent']);
    lodObject['maxFadeExtent'] = parseFloat(object['maxFadeExtent']);
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const INNER_BOUNDARY_IS_PARSERS = makeStructureNS(NAMESPACE_URIS$4, {
    // KML spec only allows one LinearRing  per innerBoundaryIs, but Google Earth
    // allows multiple, so we parse multiple here too.
    'LinearRing': makeArrayPusher(readFlatLinearRing),
  });

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function innerBoundaryIsParser(node, objectStack) {
    const innerBoundaryFlatLinearRings = pushParseAndPop(
      /** @type {Array<Array<number>>} */ ([]),
      INNER_BOUNDARY_IS_PARSERS,
      node,
      objectStack
    );
    if (innerBoundaryFlatLinearRings.length > 0) {
      const flatLinearRings =
        /** @type {Array<Array<number>>} */
        (objectStack[objectStack.length - 1]);
      flatLinearRings.push(...innerBoundaryFlatLinearRings);
    }
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const OUTER_BOUNDARY_IS_PARSERS = makeStructureNS(NAMESPACE_URIS$4, {
    'LinearRing': makeReplacer(readFlatLinearRing),
  });

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function outerBoundaryIsParser(node, objectStack) {
    /** @type {Array<number>|undefined} */
    const flatLinearRing = pushParseAndPop(
      undefined,
      OUTER_BOUNDARY_IS_PARSERS,
      node,
      objectStack
    );
    if (flatLinearRing) {
      const flatLinearRings =
        /** @type {Array<Array<number>>} */
        (objectStack[objectStack.length - 1]);
      flatLinearRings[0] = flatLinearRing;
    }
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function linkParser(node, objectStack) {
    parseNode(LINK_PARSERS, node, objectStack);
  }

  /**
   * @param {Node} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function whenParser(node, objectStack) {
    const gxTrackObject =
      /** @type {GxTrackObject} */
      (objectStack[objectStack.length - 1]);
    const whens = gxTrackObject.whens;
    const s = getAllTextContent(node, false);
    const when = Date.parse(s);
    whens.push(isNaN(when) ? 0 : when);
  }

  /**
   * @param {Node} node Node to append a TextNode with the color to.
   * @param {import("../color.js").Color|string} color Color.
   */
  function writeColorTextNode(node, color) {
    const rgba = asArray(color);
    const opacity = rgba.length == 4 ? rgba[3] : 1;
    /** @type {Array<string|number>} */
    const abgr = [opacity * 255, rgba[2], rgba[1], rgba[0]];
    for (let i = 0; i < 4; ++i) {
      const hex = Math.floor(/** @type {number} */ (abgr[i])).toString(16);
      abgr[i] = hex.length == 1 ? '0' + hex : hex;
    }
    writeStringTextNode(node, abgr.join(''));
  }

  /**
   * @param {Node} node Node to append a TextNode with the coordinates to.
   * @param {Array<number>} coordinates Coordinates.
   * @param {Array<*>} objectStack Object stack.
   */
  function writeCoordinatesTextNode(node, coordinates, objectStack) {
    const context = objectStack[objectStack.length - 1];

    const layout = context['layout'];
    const stride = context['stride'];

    let dimension;
    if (layout == 'XY' || layout == 'XYM') {
      dimension = 2;
    } else if (layout == 'XYZ' || layout == 'XYZM') {
      dimension = 3;
    } else {
      assert(false, 34); // Invalid geometry layout
    }

    const ii = coordinates.length;
    let text = '';
    if (ii > 0) {
      text += coordinates[0];
      for (let d = 1; d < dimension; ++d) {
        text += ',' + coordinates[d];
      }
      for (let i = stride; i < ii; i += stride) {
        text += ' ' + coordinates[i];
        for (let d = 1; d < dimension; ++d) {
          text += ',' + coordinates[i + d];
        }
      }
    }
    writeStringTextNode(node, text);
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  // @ts-ignore
  const EXTENDEDDATA_NODE_SERIALIZERS = makeStructureNS(NAMESPACE_URIS$4, {
    'Data': makeChildAppender(writeDataNode),
    'value': makeChildAppender(writeDataNodeValue),
    'displayName': makeChildAppender(writeDataNodeName),
  });

  /**
   * @param {Element} node Node.
   * @param {{name: *, value: *}} pair Name value pair.
   * @param {Array<*>} objectStack Object stack.
   */
  function writeDataNode(node, pair, objectStack) {
    node.setAttribute('name', pair.name);
    const /** @type {import("../xml.js").NodeStackItem} */ context = {node: node};
    const value = pair.value;

    if (typeof value == 'object') {
      if (value !== null && value.displayName) {
        pushSerializeAndPop(
          context,
          EXTENDEDDATA_NODE_SERIALIZERS,
          OBJECT_PROPERTY_NODE_FACTORY,
          [value.displayName],
          objectStack,
          ['displayName']
        );
      }

      if (value !== null && value.value) {
        pushSerializeAndPop(
          context,
          EXTENDEDDATA_NODE_SERIALIZERS,
          OBJECT_PROPERTY_NODE_FACTORY,
          [value.value],
          objectStack,
          ['value']
        );
      }
    } else {
      pushSerializeAndPop(
        context,
        EXTENDEDDATA_NODE_SERIALIZERS,
        OBJECT_PROPERTY_NODE_FACTORY,
        [value],
        objectStack,
        ['value']
      );
    }
  }

  /**
   * @param {Node} node Node to append a TextNode with the name to.
   * @param {string} name DisplayName.
   */
  function writeDataNodeName(node, name) {
    writeCDATASection(node, name);
  }

  /**
   * @param {Node} node Node to append a CDATA Section with the value to.
   * @param {string} value Value.
   */
  function writeDataNodeValue(node, value) {
    writeStringTextNode(node, value);
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  // @ts-ignore
  const DOCUMENT_SERIALIZERS = makeStructureNS(NAMESPACE_URIS$4, {
    'Placemark': makeChildAppender(writePlacemark),
  });

  /**
   * @const
   * @param {*} value Value.
   * @param {Array<*>} objectStack Object stack.
   * @param {string} [nodeName] Node name.
   * @return {Node|undefined} Node.
   */
  const DOCUMENT_NODE_FACTORY = function (value, objectStack, nodeName) {
    const parentNode = objectStack[objectStack.length - 1].node;
    return createElementNS(parentNode.namespaceURI, 'Placemark');
  };

  /**
   * @param {Element} node Node.
   * @param {Array<Feature>} features Features.
   * @param {Array<*>} objectStack Object stack.
   * @this {KML}
   */
  function writeDocument(node, features, objectStack) {
    const /** @type {import("../xml.js").NodeStackItem} */ context = {node: node};
    pushSerializeAndPop(
      context,
      DOCUMENT_SERIALIZERS,
      DOCUMENT_NODE_FACTORY,
      features,
      objectStack,
      undefined,
      this
    );
  }

  /**
   * A factory for creating Data nodes.
   * @const
   * @type {function(*, Array<*>): (Node|undefined)}
   */
  const DATA_NODE_FACTORY = makeSimpleNodeFactory('Data');

  /**
   * @param {Element} node Node.
   * @param {{names: Array<string>, values: (Array<*>)}} namesAndValues Names and values.
   * @param {Array<*>} objectStack Object stack.
   */
  function writeExtendedData(node, namesAndValues, objectStack) {
    const /** @type {import("../xml.js").NodeStackItem} */ context = {node: node};
    const names = namesAndValues.names;
    const values = namesAndValues.values;
    const length = names.length;

    for (let i = 0; i < length; i++) {
      pushSerializeAndPop(
        context,
        EXTENDEDDATA_NODE_SERIALIZERS,
        DATA_NODE_FACTORY,
        [{name: names[i], value: values[i]}],
        objectStack
      );
    }
  }

  /**
   * @const
   * @type {Object<string, Array<string>>}
   */
  // @ts-ignore
  const ICON_SEQUENCE = makeStructureNS(
    NAMESPACE_URIS$4,
    ['href'],
    makeStructureNS(GX_NAMESPACE_URIS, ['x', 'y', 'w', 'h'])
  );

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  // @ts-ignore
  const ICON_SERIALIZERS = makeStructureNS(
    NAMESPACE_URIS$4,
    {
      'href': makeChildAppender(writeStringTextNode),
    },
    makeStructureNS(GX_NAMESPACE_URIS, {
      'x': makeChildAppender(writeDecimalTextNode),
      'y': makeChildAppender(writeDecimalTextNode),
      'w': makeChildAppender(writeDecimalTextNode),
      'h': makeChildAppender(writeDecimalTextNode),
    })
  );

  /**
   * @const
   * @param {*} value Value.
   * @param {Array<*>} objectStack Object stack.
   * @param {string} [nodeName] Node name.
   * @return {Node|undefined} Node.
   */
  const GX_NODE_FACTORY = function (value, objectStack, nodeName) {
    return createElementNS(GX_NAMESPACE_URIS[0], 'gx:' + nodeName);
  };

  /**
   * @param {Element} node Node.
   * @param {Object} icon Icon object.
   * @param {Array<*>} objectStack Object stack.
   */
  function writeIcon(node, icon, objectStack) {
    const /** @type {import("../xml.js").NodeStackItem} */ context = {node: node};
    const parentNode = objectStack[objectStack.length - 1].node;
    let orderedKeys = ICON_SEQUENCE[parentNode.namespaceURI];
    let values = makeSequence(icon, orderedKeys);
    pushSerializeAndPop(
      context,
      ICON_SERIALIZERS,
      OBJECT_PROPERTY_NODE_FACTORY,
      values,
      objectStack,
      orderedKeys
    );
    orderedKeys = ICON_SEQUENCE[GX_NAMESPACE_URIS[0]];
    values = makeSequence(icon, orderedKeys);
    pushSerializeAndPop(
      context,
      ICON_SERIALIZERS,
      GX_NODE_FACTORY,
      values,
      objectStack,
      orderedKeys
    );
  }

  /**
   * @const
   * @type {Object<string, Array<string>>}
   */
  // @ts-ignore
  const ICON_STYLE_SEQUENCE = makeStructureNS(NAMESPACE_URIS$4, [
    'scale',
    'heading',
    'Icon',
    'color',
    'hotSpot',
  ]);

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  // @ts-ignore
  const ICON_STYLE_SERIALIZERS = makeStructureNS(NAMESPACE_URIS$4, {
    'Icon': makeChildAppender(writeIcon),
    'color': makeChildAppender(writeColorTextNode),
    'heading': makeChildAppender(writeDecimalTextNode),
    'hotSpot': makeChildAppender(writeVec2),
    'scale': makeChildAppender(writeScaleTextNode),
  });

  /**
   * @param {Element} node Node.
   * @param {import("../style/Icon.js").default} style Icon style.
   * @param {Array<*>} objectStack Object stack.
   */
  function writeIconStyle(node, style, objectStack) {
    const /** @type {import("../xml.js").NodeStackItem} */ context = {node: node};
    const /** @type {Object<string, any>} */ properties = {};
    const src = style.getSrc();
    const size = style.getSize();
    const iconImageSize = style.getImageSize();
    const iconProperties = {
      'href': src,
    };

    if (size) {
      iconProperties['w'] = size[0];
      iconProperties['h'] = size[1];
      const anchor = style.getAnchor(); // top-left
      const origin = style.getOrigin(); // top-left

      if (origin && iconImageSize && origin[0] !== 0 && origin[1] !== size[1]) {
        iconProperties['x'] = origin[0];
        iconProperties['y'] = iconImageSize[1] - (origin[1] + size[1]);
      }

      if (anchor && (anchor[0] !== size[0] / 2 || anchor[1] !== size[1] / 2)) {
        const /** @type {Vec2} */ hotSpot = {
            x: anchor[0],
            xunits: 'pixels',
            y: size[1] - anchor[1],
            yunits: 'pixels',
          };
        properties['hotSpot'] = hotSpot;
      }
    }

    properties['Icon'] = iconProperties;

    let scale = style.getScaleArray()[0];
    let imageSize = size;
    if (imageSize === null) {
      imageSize = DEFAULT_IMAGE_STYLE_SIZE;
    }
    if (imageSize.length == 2) {
      const resizeScale = scaleForSize(imageSize);
      scale = scale / resizeScale;
    }
    if (scale !== 1) {
      properties['scale'] = scale;
    }

    const rotation = style.getRotation();
    if (rotation !== 0) {
      properties['heading'] = rotation; // 0-360
    }

    const color = style.getColor();
    if (color) {
      properties['color'] = color;
    }

    const parentNode = objectStack[objectStack.length - 1].node;
    const orderedKeys = ICON_STYLE_SEQUENCE[parentNode.namespaceURI];
    const values = makeSequence(properties, orderedKeys);
    pushSerializeAndPop(
      context,
      ICON_STYLE_SERIALIZERS,
      OBJECT_PROPERTY_NODE_FACTORY,
      values,
      objectStack,
      orderedKeys
    );
  }

  /**
   * @const
   * @type {Object<string, Array<string>>}
   */
  // @ts-ignore
  const LABEL_STYLE_SEQUENCE = makeStructureNS(NAMESPACE_URIS$4, [
    'color',
    'scale',
  ]);

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  // @ts-ignore
  const LABEL_STYLE_SERIALIZERS = makeStructureNS(NAMESPACE_URIS$4, {
    'color': makeChildAppender(writeColorTextNode),
    'scale': makeChildAppender(writeScaleTextNode),
  });

  /**
   * @param {Element} node Node.
   * @param {Text} style style.
   * @param {Array<*>} objectStack Object stack.
   */
  function writeLabelStyle(node, style, objectStack) {
    const /** @type {import("../xml.js").NodeStackItem} */ context = {node: node};
    const properties = {};
    const fill = style.getFill();
    if (fill) {
      properties['color'] = fill.getColor();
    }
    const scale = style.getScale();
    if (scale && scale !== 1) {
      properties['scale'] = scale;
    }
    const parentNode = objectStack[objectStack.length - 1].node;
    const orderedKeys = LABEL_STYLE_SEQUENCE[parentNode.namespaceURI];
    const values = makeSequence(properties, orderedKeys);
    pushSerializeAndPop(
      context,
      LABEL_STYLE_SERIALIZERS,
      OBJECT_PROPERTY_NODE_FACTORY,
      values,
      objectStack,
      orderedKeys
    );
  }

  /**
   * @const
   * @type {Object<string, Array<string>>}
   */
  // @ts-ignore
  const LINE_STYLE_SEQUENCE = makeStructureNS(NAMESPACE_URIS$4, ['color', 'width']);

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  // @ts-ignore
  const LINE_STYLE_SERIALIZERS = makeStructureNS(NAMESPACE_URIS$4, {
    'color': makeChildAppender(writeColorTextNode),
    'width': makeChildAppender(writeDecimalTextNode),
  });

  /**
   * @param {Element} node Node.
   * @param {Stroke} style style.
   * @param {Array<*>} objectStack Object stack.
   */
  function writeLineStyle(node, style, objectStack) {
    const /** @type {import("../xml.js").NodeStackItem} */ context = {node: node};
    const properties = {
      'color': style.getColor(),
      'width': Number(style.getWidth()) || 1,
    };
    const parentNode = objectStack[objectStack.length - 1].node;
    const orderedKeys = LINE_STYLE_SEQUENCE[parentNode.namespaceURI];
    const values = makeSequence(properties, orderedKeys);
    pushSerializeAndPop(
      context,
      LINE_STYLE_SERIALIZERS,
      OBJECT_PROPERTY_NODE_FACTORY,
      values,
      objectStack,
      orderedKeys
    );
  }

  /**
   * @const
   * @type {Object<string, string>}
   */
  const GEOMETRY_TYPE_TO_NODENAME = {
    'Point': 'Point',
    'LineString': 'LineString',
    'LinearRing': 'LinearRing',
    'Polygon': 'Polygon',
    'MultiPoint': 'MultiGeometry',
    'MultiLineString': 'MultiGeometry',
    'MultiPolygon': 'MultiGeometry',
    'GeometryCollection': 'MultiGeometry',
  };

  /**
   * @const
   * @param {*} value Value.
   * @param {Array<*>} objectStack Object stack.
   * @param {string} [nodeName] Node name.
   * @return {Node|undefined} Node.
   */
  const GEOMETRY_NODE_FACTORY = function (value, objectStack, nodeName) {
    if (value) {
      const parentNode = objectStack[objectStack.length - 1].node;
      return createElementNS(
        parentNode.namespaceURI,
        GEOMETRY_TYPE_TO_NODENAME[
          /** @type {import("../geom/Geometry.js").default} */ (value).getType()
        ]
      );
    }
  };

  /**
   * A factory for creating Point nodes.
   * @const
   * @type {function(*, Array<*>, string=): (Node|undefined)}
   */
  const POINT_NODE_FACTORY = makeSimpleNodeFactory('Point');

  /**
   * A factory for creating LineString nodes.
   * @const
   * @type {function(*, Array<*>, string=): (Node|undefined)}
   */
  const LINE_STRING_NODE_FACTORY = makeSimpleNodeFactory('LineString');

  /**
   * A factory for creating LinearRing nodes.
   * @const
   * @type {function(*, Array<*>, string=): (Node|undefined)}
   */
  const LINEAR_RING_NODE_FACTORY = makeSimpleNodeFactory('LinearRing');

  /**
   * A factory for creating Polygon nodes.
   * @const
   * @type {function(*, Array<*>, string=): (Node|undefined)}
   */
  const POLYGON_NODE_FACTORY = makeSimpleNodeFactory('Polygon');

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  // @ts-ignore
  const MULTI_GEOMETRY_SERIALIZERS = makeStructureNS(NAMESPACE_URIS$4, {
    'LineString': makeChildAppender(writePrimitiveGeometry),
    'Point': makeChildAppender(writePrimitiveGeometry),
    'Polygon': makeChildAppender(writePolygon),
    'GeometryCollection': makeChildAppender(writeMultiGeometry),
  });

  /**
   * @param {Element} node Node.
   * @param {import("../geom/Geometry.js").default} geometry Geometry.
   * @param {Array<*>} objectStack Object stack.
   */
  function writeMultiGeometry(node, geometry, objectStack) {
    /** @type {import("../xml.js").NodeStackItem} */
    const context = {node: node};
    const type = geometry.getType();
    /** @type {Array<import("../geom/Geometry.js").default>} */
    let geometries = [];
    /** @type {function(*, Array<*>, string=): (Node|undefined)} */
    let factory;
    if (type === 'GeometryCollection') {
      /** @type {GeometryCollection} */ (geometry)
        .getGeometriesArrayRecursive()
        .forEach(function (geometry) {
          const type = geometry.getType();
          if (type === 'MultiPoint') {
            geometries = geometries.concat(
              /** @type {MultiPoint} */ (geometry).getPoints()
            );
          } else if (type === 'MultiLineString') {
            geometries = geometries.concat(
              /** @type {MultiLineString} */ (geometry).getLineStrings()
            );
          } else if (type === 'MultiPolygon') {
            geometries = geometries.concat(
              /** @type {MultiPolygon} */ (geometry).getPolygons()
            );
          } else if (
            type === 'Point' ||
            type === 'LineString' ||
            type === 'Polygon'
          ) {
            geometries.push(geometry);
          } else {
            assert(false, 39); // Unknown geometry type
          }
        });
      factory = GEOMETRY_NODE_FACTORY;
    } else if (type === 'MultiPoint') {
      geometries = /** @type {MultiPoint} */ (geometry).getPoints();
      factory = POINT_NODE_FACTORY;
    } else if (type === 'MultiLineString') {
      geometries = /** @type {MultiLineString} */ (geometry).getLineStrings();
      factory = LINE_STRING_NODE_FACTORY;
    } else if (type === 'MultiPolygon') {
      geometries = /** @type {MultiPolygon} */ (geometry).getPolygons();
      factory = POLYGON_NODE_FACTORY;
    } else {
      assert(false, 39); // Unknown geometry type
    }
    pushSerializeAndPop(
      context,
      MULTI_GEOMETRY_SERIALIZERS,
      factory,
      geometries,
      objectStack
    );
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  // @ts-ignore
  const BOUNDARY_IS_SERIALIZERS = makeStructureNS(NAMESPACE_URIS$4, {
    'LinearRing': makeChildAppender(writePrimitiveGeometry),
  });

  /**
   * @param {Element} node Node.
   * @param {import("../geom/LinearRing.js").default} linearRing Linear ring.
   * @param {Array<*>} objectStack Object stack.
   */
  function writeBoundaryIs(node, linearRing, objectStack) {
    const /** @type {import("../xml.js").NodeStackItem} */ context = {node: node};
    pushSerializeAndPop(
      context,
      BOUNDARY_IS_SERIALIZERS,
      LINEAR_RING_NODE_FACTORY,
      [linearRing],
      objectStack
    );
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  // @ts-ignore
  const PLACEMARK_SERIALIZERS = makeStructureNS(NAMESPACE_URIS$4, {
    'ExtendedData': makeChildAppender(writeExtendedData),
    'MultiGeometry': makeChildAppender(writeMultiGeometry),
    'LineString': makeChildAppender(writePrimitiveGeometry),
    'LinearRing': makeChildAppender(writePrimitiveGeometry),
    'Point': makeChildAppender(writePrimitiveGeometry),
    'Polygon': makeChildAppender(writePolygon),
    'Style': makeChildAppender(writeStyle),
    'address': makeChildAppender(writeStringTextNode),
    'description': makeChildAppender(writeStringTextNode),
    'name': makeChildAppender(writeStringTextNode),
    'open': makeChildAppender(writeBooleanTextNode),
    'phoneNumber': makeChildAppender(writeStringTextNode),
    'styleUrl': makeChildAppender(writeStringTextNode),
    'visibility': makeChildAppender(writeBooleanTextNode),
  });

  /**
   * @const
   * @type {Object<string, Array<string>>}
   */
  // @ts-ignore
  const PLACEMARK_SEQUENCE = makeStructureNS(NAMESPACE_URIS$4, [
    'name',
    'open',
    'visibility',
    'address',
    'phoneNumber',
    'description',
    'styleUrl',
    'Style',
  ]);

  /**
   * A factory for creating ExtendedData nodes.
   * @const
   * @type {function(*, Array<*>): (Node|undefined)}
   */
  const EXTENDEDDATA_NODE_FACTORY = makeSimpleNodeFactory('ExtendedData');

  /**
   * FIXME currently we do serialize arbitrary/custom feature properties
   * (ExtendedData).
   * @param {Element} node Node.
   * @param {Feature} feature Feature.
   * @param {Array<*>} objectStack Object stack.
   * @this {KML}
   */
  function writePlacemark(node, feature, objectStack) {
    const /** @type {import("../xml.js").NodeStackItem} */ context = {node: node};

    // set id
    if (feature.getId()) {
      node.setAttribute('id', /** @type {string} */ (feature.getId()));
    }

    // serialize properties (properties unknown to KML are not serialized)
    const properties = feature.getProperties();

    // don't export these to ExtendedData
    const filter = {
      'address': 1,
      'description': 1,
      'name': 1,
      'open': 1,
      'phoneNumber': 1,
      'styleUrl': 1,
      'visibility': 1,
    };
    filter[feature.getGeometryName()] = 1;
    const keys = Object.keys(properties || {})
      .sort()
      .filter(function (v) {
        return !filter[v];
      });

    const styleFunction = feature.getStyleFunction();
    if (styleFunction) {
      // FIXME the styles returned by the style function are supposed to be
      // resolution-independent here
      const styles = styleFunction(feature, 0);
      if (styles) {
        const styleArray = Array.isArray(styles) ? styles : [styles];
        let pointStyles = styleArray;
        if (feature.getGeometry()) {
          pointStyles = styleArray.filter(function (style) {
            const geometry = style.getGeometryFunction()(feature);
            if (geometry) {
              const type = geometry.getType();
              if (type === 'GeometryCollection') {
                return /** @type {GeometryCollection} */ (geometry)
                  .getGeometriesArrayRecursive()
                  .filter(function (geometry) {
                    const type = geometry.getType();
                    return type === 'Point' || type === 'MultiPoint';
                  }).length;
              }
              return type === 'Point' || type === 'MultiPoint';
            }
          });
        }
        if (this.writeStyles_) {
          let lineStyles = styleArray;
          let polyStyles = styleArray;
          if (feature.getGeometry()) {
            lineStyles = styleArray.filter(function (style) {
              const geometry = style.getGeometryFunction()(feature);
              if (geometry) {
                const type = geometry.getType();
                if (type === 'GeometryCollection') {
                  return /** @type {GeometryCollection} */ (geometry)
                    .getGeometriesArrayRecursive()
                    .filter(function (geometry) {
                      const type = geometry.getType();
                      return type === 'LineString' || type === 'MultiLineString';
                    }).length;
                }
                return type === 'LineString' || type === 'MultiLineString';
              }
            });
            polyStyles = styleArray.filter(function (style) {
              const geometry = style.getGeometryFunction()(feature);
              if (geometry) {
                const type = geometry.getType();
                if (type === 'GeometryCollection') {
                  return /** @type {GeometryCollection} */ (geometry)
                    .getGeometriesArrayRecursive()
                    .filter(function (geometry) {
                      const type = geometry.getType();
                      return type === 'Polygon' || type === 'MultiPolygon';
                    }).length;
                }
                return type === 'Polygon' || type === 'MultiPolygon';
              }
            });
          }
          properties['Style'] = {
            pointStyles: pointStyles,
            lineStyles: lineStyles,
            polyStyles: polyStyles,
          };
        }
        if (pointStyles.length && properties['name'] === undefined) {
          const textStyle = pointStyles[0].getText();
          if (textStyle) {
            properties['name'] = textStyle.getText();
          }
        }
      }
    }
    const parentNode = objectStack[objectStack.length - 1].node;
    const orderedKeys = PLACEMARK_SEQUENCE[parentNode.namespaceURI];
    const values = makeSequence(properties, orderedKeys);
    pushSerializeAndPop(
      context,
      PLACEMARK_SERIALIZERS,
      OBJECT_PROPERTY_NODE_FACTORY,
      values,
      objectStack,
      orderedKeys
    );

    if (keys.length > 0) {
      const sequence = makeSequence(properties, keys);
      const namesAndValues = {names: keys, values: sequence};
      pushSerializeAndPop(
        context,
        PLACEMARK_SERIALIZERS,
        EXTENDEDDATA_NODE_FACTORY,
        [namesAndValues],
        objectStack
      );
    }

    // serialize geometry
    const options = /** @type {import("./Feature.js").WriteOptions} */ (
      objectStack[0]
    );
    let geometry = feature.getGeometry();
    if (geometry) {
      geometry = transformGeometryWithOptions(geometry, true, options);
    }
    pushSerializeAndPop(
      context,
      PLACEMARK_SERIALIZERS,
      GEOMETRY_NODE_FACTORY,
      [geometry],
      objectStack
    );
  }

  /**
   * @const
   * @type {Object<string, Array<string>>}
   */
  // @ts-ignore
  const PRIMITIVE_GEOMETRY_SEQUENCE = makeStructureNS(NAMESPACE_URIS$4, [
    'extrude',
    'tessellate',
    'altitudeMode',
    'coordinates',
  ]);

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  // @ts-ignore
  const PRIMITIVE_GEOMETRY_SERIALIZERS = makeStructureNS(NAMESPACE_URIS$4, {
    'extrude': makeChildAppender(writeBooleanTextNode),
    'tessellate': makeChildAppender(writeBooleanTextNode),
    'altitudeMode': makeChildAppender(writeStringTextNode),
    'coordinates': makeChildAppender(writeCoordinatesTextNode),
  });

  /**
   * @param {Element} node Node.
   * @param {import("../geom/SimpleGeometry.js").default} geometry Geometry.
   * @param {Array<*>} objectStack Object stack.
   */
  function writePrimitiveGeometry(node, geometry, objectStack) {
    const flatCoordinates = geometry.getFlatCoordinates();
    const /** @type {import("../xml.js").NodeStackItem} */ context = {node: node};
    context['layout'] = geometry.getLayout();
    context['stride'] = geometry.getStride();

    // serialize properties (properties unknown to KML are not serialized)
    const properties = geometry.getProperties();
    properties.coordinates = flatCoordinates;

    const parentNode = objectStack[objectStack.length - 1].node;
    const orderedKeys = PRIMITIVE_GEOMETRY_SEQUENCE[parentNode.namespaceURI];
    const values = makeSequence(properties, orderedKeys);
    pushSerializeAndPop(
      context,
      PRIMITIVE_GEOMETRY_SERIALIZERS,
      OBJECT_PROPERTY_NODE_FACTORY,
      values,
      objectStack,
      orderedKeys
    );
  }

  /**
   * @const
   * @type {Object<string, Array<string>>}
   */
  // @ts-ignore
  const POLY_STYLE_SEQUENCE = makeStructureNS(NAMESPACE_URIS$4, [
    'color',
    'fill',
    'outline',
  ]);

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  // @ts-ignore
  const POLYGON_SERIALIZERS = makeStructureNS(NAMESPACE_URIS$4, {
    'outerBoundaryIs': makeChildAppender(writeBoundaryIs),
    'innerBoundaryIs': makeChildAppender(writeBoundaryIs),
  });

  /**
   * A factory for creating innerBoundaryIs nodes.
   * @const
   * @type {function(*, Array<*>, string=): (Node|undefined)}
   */
  const INNER_BOUNDARY_NODE_FACTORY = makeSimpleNodeFactory('innerBoundaryIs');

  /**
   * A factory for creating outerBoundaryIs nodes.
   * @const
   * @type {function(*, Array<*>, string=): (Node|undefined)}
   */
  const OUTER_BOUNDARY_NODE_FACTORY = makeSimpleNodeFactory('outerBoundaryIs');

  /**
   * @param {Element} node Node.
   * @param {Polygon} polygon Polygon.
   * @param {Array<*>} objectStack Object stack.
   */
  function writePolygon(node, polygon, objectStack) {
    const linearRings = polygon.getLinearRings();
    const outerRing = linearRings.shift();
    const /** @type {import("../xml.js").NodeStackItem} */ context = {node: node};
    // inner rings
    pushSerializeAndPop(
      context,
      POLYGON_SERIALIZERS,
      INNER_BOUNDARY_NODE_FACTORY,
      linearRings,
      objectStack
    );
    // outer ring
    pushSerializeAndPop(
      context,
      POLYGON_SERIALIZERS,
      OUTER_BOUNDARY_NODE_FACTORY,
      [outerRing],
      objectStack
    );
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  // @ts-ignore
  const POLY_STYLE_SERIALIZERS = makeStructureNS(NAMESPACE_URIS$4, {
    'color': makeChildAppender(writeColorTextNode),
    'fill': makeChildAppender(writeBooleanTextNode),
    'outline': makeChildAppender(writeBooleanTextNode),
  });

  /**
   * @param {Element} node Node.
   * @param {Style} style Style.
   * @param {Array<*>} objectStack Object stack.
   */
  function writePolyStyle(node, style, objectStack) {
    const /** @type {import("../xml.js").NodeStackItem} */ context = {node: node};
    const fill = style.getFill();
    const stroke = style.getStroke();
    const properties = {
      'color': fill ? fill.getColor() : undefined,
      'fill': fill ? undefined : false,
      'outline': stroke ? undefined : false,
    };
    const parentNode = objectStack[objectStack.length - 1].node;
    const orderedKeys = POLY_STYLE_SEQUENCE[parentNode.namespaceURI];
    const values = makeSequence(properties, orderedKeys);
    pushSerializeAndPop(
      context,
      POLY_STYLE_SERIALIZERS,
      OBJECT_PROPERTY_NODE_FACTORY,
      values,
      objectStack,
      orderedKeys
    );
  }

  /**
   * @param {Node} node Node to append a TextNode with the scale to.
   * @param {number|undefined} scale Scale.
   */
  function writeScaleTextNode(node, scale) {
    // the Math is to remove any excess decimals created by float arithmetic
    writeDecimalTextNode(node, Math.round(scale * 1e6) / 1e6);
  }

  /**
   * @const
   * @type {Object<string, Array<string>>}
   */
  // @ts-ignore
  const STYLE_SEQUENCE = makeStructureNS(NAMESPACE_URIS$4, [
    'IconStyle',
    'LabelStyle',
    'LineStyle',
    'PolyStyle',
  ]);

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  // @ts-ignore
  const STYLE_SERIALIZERS = makeStructureNS(NAMESPACE_URIS$4, {
    'IconStyle': makeChildAppender(writeIconStyle),
    'LabelStyle': makeChildAppender(writeLabelStyle),
    'LineStyle': makeChildAppender(writeLineStyle),
    'PolyStyle': makeChildAppender(writePolyStyle),
  });

  /**
   * @param {Element} node Node.
   * @param {Object<string, Array<Style>>} styles Styles.
   * @param {Array<*>} objectStack Object stack.
   */
  function writeStyle(node, styles, objectStack) {
    const /** @type {import("../xml.js").NodeStackItem} */ context = {node: node};
    const properties = {};
    if (styles.pointStyles.length) {
      const textStyle = styles.pointStyles[0].getText();
      if (textStyle) {
        properties['LabelStyle'] = textStyle;
      }
      const imageStyle = styles.pointStyles[0].getImage();
      if (
        imageStyle &&
        typeof (/** @type {?} */ (imageStyle).getSrc) === 'function'
      ) {
        properties['IconStyle'] = imageStyle;
      }
    }
    if (styles.lineStyles.length) {
      const strokeStyle = styles.lineStyles[0].getStroke();
      if (strokeStyle) {
        properties['LineStyle'] = strokeStyle;
      }
    }
    if (styles.polyStyles.length) {
      const strokeStyle = styles.polyStyles[0].getStroke();
      if (strokeStyle && !properties['LineStyle']) {
        properties['LineStyle'] = strokeStyle;
      }
      properties['PolyStyle'] = styles.polyStyles[0];
    }
    const parentNode = objectStack[objectStack.length - 1].node;
    const orderedKeys = STYLE_SEQUENCE[parentNode.namespaceURI];
    const values = makeSequence(properties, orderedKeys);
    pushSerializeAndPop(
      context,
      STYLE_SERIALIZERS,
      OBJECT_PROPERTY_NODE_FACTORY,
      values,
      objectStack,
      orderedKeys
    );
  }

  /**
   * @param {Element} node Node to append a TextNode with the Vec2 to.
   * @param {Vec2} vec2 Vec2.
   */
  function writeVec2(node, vec2) {
    node.setAttribute('x', String(vec2.x));
    node.setAttribute('y', String(vec2.y));
    node.setAttribute('xunits', vec2.xunits);
    node.setAttribute('yunits', vec2.yunits);
  }

  /**
   * @module ol/format/OSMXML
   */

  /**
   * @const
   * @type {Array<null>}
   */
  const NAMESPACE_URIS$3 = [null];

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const WAY_PARSERS = makeStructureNS(NAMESPACE_URIS$3, {
    'nd': readNd,
    'tag': readTag,
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const PARSERS$3 = makeStructureNS(NAMESPACE_URIS$3, {
    'node': readNode,
    'way': readWay,
  });

  /**
   * @classdesc
   * Feature format for reading data in the
   * [OSMXML format](https://wiki.openstreetmap.org/wiki/OSM_XML).
   *
   * @api
   */
  class OSMXML extends XMLFeature {
    constructor() {
      super();

      /**
       * @type {import("../proj/Projection.js").default}
       */
      this.dataProjection = get$2('EPSG:4326');
    }

    /**
     * @protected
     * @param {Element} node Node.
     * @param {import("./Feature.js").ReadOptions} [options] Options.
     * @return {Array<import("../Feature.js").default>} Features.
     */
    readFeaturesFromNode(node, options) {
      options = this.getReadOptions(node, options);
      if (node.localName == 'osm') {
        const state = pushParseAndPop(
          {
            nodes: {},
            ways: [],
            features: [],
          },
          PARSERS$3,
          node,
          [options]
        );
        // parse nodes in ways
        for (let j = 0; j < state.ways.length; j++) {
          const values = /** @type {Object} */ (state.ways[j]);
          /** @type {Array<number>} */
          const flatCoordinates = [];
          for (let i = 0, ii = values.ndrefs.length; i < ii; i++) {
            const point = state.nodes[values.ndrefs[i]];
            extend$2(flatCoordinates, point);
          }
          let geometry;
          if (values.ndrefs[0] == values.ndrefs[values.ndrefs.length - 1]) {
            // closed way
            geometry = new Polygon(flatCoordinates, 'XY', [
              flatCoordinates.length,
            ]);
          } else {
            geometry = new LineString(flatCoordinates, 'XY');
          }
          transformGeometryWithOptions(geometry, false, options);
          const feature = new Feature(geometry);
          if (values.id !== undefined) {
            feature.setId(values.id);
          }
          feature.setProperties(values.tags, true);
          state.features.push(feature);
        }
        if (state.features) {
          return state.features;
        }
      }
      return [];
    }
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const NODE_PARSERS = makeStructureNS(NAMESPACE_URIS$3, {
    'tag': readTag,
  });

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function readNode(node, objectStack) {
    const options = /** @type {import("./Feature.js").ReadOptions} */ (
      objectStack[0]
    );
    const state = /** @type {Object} */ (objectStack[objectStack.length - 1]);
    const id = node.getAttribute('id');
    /** @type {import("../coordinate.js").Coordinate} */
    const coordinates = [
      parseFloat(node.getAttribute('lon')),
      parseFloat(node.getAttribute('lat')),
    ];
    state.nodes[id] = coordinates;

    const values = pushParseAndPop(
      {
        tags: {},
      },
      NODE_PARSERS,
      node,
      objectStack
    );
    if (!isEmpty$1(values.tags)) {
      const geometry = new Point(coordinates);
      transformGeometryWithOptions(geometry, false, options);
      const feature = new Feature(geometry);
      if (id !== undefined) {
        feature.setId(id);
      }
      feature.setProperties(values.tags, true);
      state.features.push(feature);
    }
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function readWay(node, objectStack) {
    const id = node.getAttribute('id');
    const values = pushParseAndPop(
      {
        id: id,
        ndrefs: [],
        tags: {},
      },
      WAY_PARSERS,
      node,
      objectStack
    );
    const state = /** @type {Object} */ (objectStack[objectStack.length - 1]);
    state.ways.push(values);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function readNd(node, objectStack) {
    const values = /** @type {Object} */ (objectStack[objectStack.length - 1]);
    values.ndrefs.push(node.getAttribute('ref'));
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function readTag(node, objectStack) {
    const values = /** @type {Object} */ (objectStack[objectStack.length - 1]);
    values.tags[node.getAttribute('k')] = node.getAttribute('v');
  }

  /**
   * @module ol/format/XML
   */

  /**
   * @classdesc
   * Generic format for reading non-feature XML data
   *
   * @abstract
   */
  class XML {
    /**
     * Read the source document.
     *
     * @param {Document|Element|string} source The XML source.
     * @return {Object} An object representing the source.
     * @api
     */
    read(source) {
      if (!source) {
        return null;
      } else if (typeof source === 'string') {
        const doc = parse(source);
        return this.readFromDocument(doc);
      } else if (isDocument(source)) {
        return this.readFromDocument(/** @type {Document} */ (source));
      } else {
        return this.readFromNode(/** @type {Element} */ (source));
      }
    }

    /**
     * @param {Document} doc Document.
     * @return {Object} Object
     */
    readFromDocument(doc) {
      for (let n = doc.firstChild; n; n = n.nextSibling) {
        if (n.nodeType == Node.ELEMENT_NODE) {
          return this.readFromNode(/** @type {Element} */ (n));
        }
      }
      return null;
    }

    /**
     * @abstract
     * @param {Element} node Node.
     * @return {Object} Object
     */
    readFromNode(node) {}
  }

  /**
   * @module ol/format/xlink
   */

  /**
   * @const
   * @type {string}
   */
  const NAMESPACE_URI = 'http://www.w3.org/1999/xlink';

  /**
   * @param {Element} node Node.
   * @return {string|undefined} href.
   */
  function readHref(node) {
    return node.getAttributeNS(NAMESPACE_URI, 'href');
  }

  /**
   * @module ol/format/OWS
   */

  /**
   * @const
   * @type {Array<null|string>}
   */
  const NAMESPACE_URIS$2 = [null, 'http://www.opengis.net/ows/1.1'];

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const PARSERS$2 = makeStructureNS(NAMESPACE_URIS$2, {
    'ServiceIdentification': makeObjectPropertySetter(readServiceIdentification),
    'ServiceProvider': makeObjectPropertySetter(readServiceProvider),
    'OperationsMetadata': makeObjectPropertySetter(readOperationsMetadata),
  });

  class OWS extends XML {
    constructor() {
      super();
    }

    /**
     * @param {Element} node Node.
     * @return {Object} Object
     */
    readFromNode(node) {
      const owsObject = pushParseAndPop({}, PARSERS$2, node, []);
      return owsObject ? owsObject : null;
    }
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const ADDRESS_PARSERS = makeStructureNS(NAMESPACE_URIS$2, {
    'DeliveryPoint': makeObjectPropertySetter(readString),
    'City': makeObjectPropertySetter(readString),
    'AdministrativeArea': makeObjectPropertySetter(readString),
    'PostalCode': makeObjectPropertySetter(readString),
    'Country': makeObjectPropertySetter(readString),
    'ElectronicMailAddress': makeObjectPropertySetter(readString),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const ALLOWED_VALUES_PARSERS = makeStructureNS(NAMESPACE_URIS$2, {
    'Value': makeObjectPropertyPusher(readValue),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const CONSTRAINT_PARSERS = makeStructureNS(NAMESPACE_URIS$2, {
    'AllowedValues': makeObjectPropertySetter(readAllowedValues),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const CONTACT_INFO_PARSERS = makeStructureNS(NAMESPACE_URIS$2, {
    'Phone': makeObjectPropertySetter(readPhone),
    'Address': makeObjectPropertySetter(readAddress),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const DCP_PARSERS = makeStructureNS(NAMESPACE_URIS$2, {
    'HTTP': makeObjectPropertySetter(readHttp),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const HTTP_PARSERS$1 = makeStructureNS(NAMESPACE_URIS$2, {
    'Get': makeObjectPropertyPusher(readGet),
    'Post': undefined, // TODO
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const OPERATION_PARSERS = makeStructureNS(NAMESPACE_URIS$2, {
    'DCP': makeObjectPropertySetter(readDcp),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const OPERATIONS_METADATA_PARSERS = makeStructureNS(NAMESPACE_URIS$2, {
    'Operation': readOperation,
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const PHONE_PARSERS = makeStructureNS(NAMESPACE_URIS$2, {
    'Voice': makeObjectPropertySetter(readString),
    'Facsimile': makeObjectPropertySetter(readString),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const REQUEST_METHOD_PARSERS = makeStructureNS(NAMESPACE_URIS$2, {
    'Constraint': makeObjectPropertyPusher(readConstraint),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const SERVICE_CONTACT_PARSERS = makeStructureNS(NAMESPACE_URIS$2, {
    'IndividualName': makeObjectPropertySetter(readString),
    'PositionName': makeObjectPropertySetter(readString),
    'ContactInfo': makeObjectPropertySetter(readContactInfo),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const SERVICE_IDENTIFICATION_PARSERS = makeStructureNS(NAMESPACE_URIS$2, {
    'Abstract': makeObjectPropertySetter(readString),
    'AccessConstraints': makeObjectPropertySetter(readString),
    'Fees': makeObjectPropertySetter(readString),
    'Title': makeObjectPropertySetter(readString),
    'ServiceTypeVersion': makeObjectPropertySetter(readString),
    'ServiceType': makeObjectPropertySetter(readString),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const SERVICE_PROVIDER_PARSERS = makeStructureNS(NAMESPACE_URIS$2, {
    'ProviderName': makeObjectPropertySetter(readString),
    'ProviderSite': makeObjectPropertySetter(readHref),
    'ServiceContact': makeObjectPropertySetter(readServiceContact),
  });

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} The address.
   */
  function readAddress(node, objectStack) {
    return pushParseAndPop({}, ADDRESS_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} The values.
   */
  function readAllowedValues(node, objectStack) {
    return pushParseAndPop({}, ALLOWED_VALUES_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} The constraint.
   */
  function readConstraint(node, objectStack) {
    const name = node.getAttribute('name');
    if (!name) {
      return undefined;
    }
    return pushParseAndPop({'name': name}, CONSTRAINT_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} The contact info.
   */
  function readContactInfo(node, objectStack) {
    return pushParseAndPop({}, CONTACT_INFO_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} The DCP.
   */
  function readDcp(node, objectStack) {
    return pushParseAndPop({}, DCP_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} The GET object.
   */
  function readGet(node, objectStack) {
    const href = readHref(node);
    if (!href) {
      return undefined;
    }
    return pushParseAndPop(
      {'href': href},
      REQUEST_METHOD_PARSERS,
      node,
      objectStack
    );
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} The HTTP object.
   */
  function readHttp(node, objectStack) {
    return pushParseAndPop({}, HTTP_PARSERS$1, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} The operation.
   */
  function readOperation(node, objectStack) {
    const name = node.getAttribute('name');
    const value = pushParseAndPop({}, OPERATION_PARSERS, node, objectStack);
    if (!value) {
      return undefined;
    }
    const object = /** @type {Object} */ (objectStack[objectStack.length - 1]);
    object[name] = value;
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} The operations metadata.
   */
  function readOperationsMetadata(node, objectStack) {
    return pushParseAndPop({}, OPERATIONS_METADATA_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} The phone.
   */
  function readPhone(node, objectStack) {
    return pushParseAndPop({}, PHONE_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} The service identification.
   */
  function readServiceIdentification(node, objectStack) {
    return pushParseAndPop({}, SERVICE_IDENTIFICATION_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} The service contact.
   */
  function readServiceContact(node, objectStack) {
    return pushParseAndPop({}, SERVICE_CONTACT_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} The service provider.
   */
  function readServiceProvider(node, objectStack) {
    return pushParseAndPop({}, SERVICE_PROVIDER_PARSERS, node, objectStack);
  }

  /**
   * @param {Node} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {string|undefined} The value.
   */
  function readValue(node, objectStack) {
    return readString(node);
  }

  /**
   * @module ol/format/Polyline
   */

  /**
   * @typedef {Object} Options
   * @property {number} [factor=1e5] The factor by which the coordinates values will be scaled.
   * @property {import("../geom/Geometry.js").GeometryLayout} [geometryLayout='XY'] Layout of the
   * feature geometries created by the format reader.
   */

  /**
   * @classdesc
   * Feature format for reading and writing data in the Encoded
   * Polyline Algorithm Format.
   *
   * When reading features, the coordinates are assumed to be in two dimensions
   * and in [latitude, longitude] order.
   *
   * As Polyline sources contain a single feature,
   * {@link module:ol/format/Polyline~Polyline#readFeatures} will return the
   * feature in an array.
   *
   * @api
   */
  class Polyline extends TextFeature {
    /**
     * @param {Options} [options] Optional configuration object.
     */
    constructor(options) {
      super();

      options = options ? options : {};

      /**
       * @type {import("../proj/Projection.js").default}
       */
      this.dataProjection = get$2('EPSG:4326');

      /**
       * @private
       * @type {number}
       */
      this.factor_ = options.factor ? options.factor : 1e5;

      /**
       * @private
       * @type {import("../geom/Geometry.js").GeometryLayout}
       */
      this.geometryLayout_ = options.geometryLayout
        ? options.geometryLayout
        : 'XY';
    }

    /**
     * @protected
     * @param {string} text Text.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @return {import("../Feature.js").default} Feature.
     */
    readFeatureFromText(text, options) {
      const geometry = this.readGeometryFromText(text, options);
      return new Feature(geometry);
    }

    /**
     * @param {string} text Text.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @protected
     * @return {Array<Feature>} Features.
     */
    readFeaturesFromText(text, options) {
      const feature = this.readFeatureFromText(text, options);
      return [feature];
    }

    /**
     * @param {string} text Text.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @protected
     * @return {import("../geom/Geometry.js").default} Geometry.
     */
    readGeometryFromText(text, options) {
      const stride = getStrideForLayout(this.geometryLayout_);
      const flatCoordinates = decodeDeltas(text, stride, this.factor_);
      flipXY(flatCoordinates, 0, flatCoordinates.length, stride, flatCoordinates);
      const coordinates = inflateCoordinates(
        flatCoordinates,
        0,
        flatCoordinates.length,
        stride
      );
      const lineString = new LineString(coordinates, this.geometryLayout_);

      return transformGeometryWithOptions(
        lineString,
        false,
        this.adaptOptions(options)
      );
    }

    /**
     * @param {import("../Feature.js").default<LineString>} feature Features.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @protected
     * @return {string} Text.
     */
    writeFeatureText(feature, options) {
      const geometry = feature.getGeometry();
      if (geometry) {
        return this.writeGeometryText(geometry, options);
      } else {
        assert(false, 40); // Expected `feature` to have a geometry
        return '';
      }
    }

    /**
     * @param {Array<import("../Feature.js").default<LineString>>} features Features.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @protected
     * @return {string} Text.
     */
    writeFeaturesText(features, options) {
      return this.writeFeatureText(features[0], options);
    }

    /**
     * @param {LineString} geometry Geometry.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @protected
     * @return {string} Text.
     */
    writeGeometryText(geometry, options) {
      geometry =
        /** @type {LineString} */
        (
          transformGeometryWithOptions(geometry, true, this.adaptOptions(options))
        );
      const flatCoordinates = geometry.getFlatCoordinates();
      const stride = geometry.getStride();
      flipXY(flatCoordinates, 0, flatCoordinates.length, stride, flatCoordinates);
      return encodeDeltas(flatCoordinates, stride, this.factor_);
    }
  }

  /**
   * Encode a list of n-dimensional points and return an encoded string
   *
   * Attention: This function will modify the passed array!
   *
   * @param {Array<number>} numbers A list of n-dimensional points.
   * @param {number} stride The number of dimension of the points in the list.
   * @param {number} [factor] The factor by which the numbers will be
   *     multiplied. The remaining decimal places will get rounded away.
   *     Default is `1e5`.
   * @return {string} The encoded string.
   * @api
   */
  function encodeDeltas(numbers, stride, factor) {
    factor = factor ? factor : 1e5;
    let d;

    const lastNumbers = new Array(stride);
    for (d = 0; d < stride; ++d) {
      lastNumbers[d] = 0;
    }

    for (let i = 0, ii = numbers.length; i < ii; ) {
      for (d = 0; d < stride; ++d, ++i) {
        const num = numbers[i];
        const delta = num - lastNumbers[d];
        lastNumbers[d] = num;

        numbers[i] = delta;
      }
    }

    return encodeFloats(numbers, factor);
  }

  /**
   * Decode a list of n-dimensional points from an encoded string
   *
   * @param {string} encoded An encoded string.
   * @param {number} stride The number of dimension of the points in the
   *     encoded string.
   * @param {number} [factor] The factor by which the resulting numbers will
   *     be divided. Default is `1e5`.
   * @return {Array<number>} A list of n-dimensional points.
   * @api
   */
  function decodeDeltas(encoded, stride, factor) {
    factor = factor ? factor : 1e5;
    let d;

    /** @type {Array<number>} */
    const lastNumbers = new Array(stride);
    for (d = 0; d < stride; ++d) {
      lastNumbers[d] = 0;
    }

    const numbers = decodeFloats(encoded, factor);

    for (let i = 0, ii = numbers.length; i < ii; ) {
      for (d = 0; d < stride; ++d, ++i) {
        lastNumbers[d] += numbers[i];

        numbers[i] = lastNumbers[d];
      }
    }

    return numbers;
  }

  /**
   * Encode a list of floating point numbers and return an encoded string
   *
   * Attention: This function will modify the passed array!
   *
   * @param {Array<number>} numbers A list of floating point numbers.
   * @param {number} [factor] The factor by which the numbers will be
   *     multiplied. The remaining decimal places will get rounded away.
   *     Default is `1e5`.
   * @return {string} The encoded string.
   * @api
   */
  function encodeFloats(numbers, factor) {
    factor = factor ? factor : 1e5;
    for (let i = 0, ii = numbers.length; i < ii; ++i) {
      numbers[i] = Math.round(numbers[i] * factor);
    }

    return encodeSignedIntegers(numbers);
  }

  /**
   * Decode a list of floating point numbers from an encoded string
   *
   * @param {string} encoded An encoded string.
   * @param {number} [factor] The factor by which the result will be divided.
   *     Default is `1e5`.
   * @return {Array<number>} A list of floating point numbers.
   * @api
   */
  function decodeFloats(encoded, factor) {
    factor = factor ? factor : 1e5;
    const numbers = decodeSignedIntegers(encoded);
    for (let i = 0, ii = numbers.length; i < ii; ++i) {
      numbers[i] /= factor;
    }
    return numbers;
  }

  /**
   * Encode a list of signed integers and return an encoded string
   *
   * Attention: This function will modify the passed array!
   *
   * @param {Array<number>} numbers A list of signed integers.
   * @return {string} The encoded string.
   */
  function encodeSignedIntegers(numbers) {
    for (let i = 0, ii = numbers.length; i < ii; ++i) {
      const num = numbers[i];
      numbers[i] = num < 0 ? ~(num << 1) : num << 1;
    }
    return encodeUnsignedIntegers(numbers);
  }

  /**
   * Decode a list of signed integers from an encoded string
   *
   * @param {string} encoded An encoded string.
   * @return {Array<number>} A list of signed integers.
   */
  function decodeSignedIntegers(encoded) {
    const numbers = decodeUnsignedIntegers(encoded);
    for (let i = 0, ii = numbers.length; i < ii; ++i) {
      const num = numbers[i];
      numbers[i] = num & 1 ? ~(num >> 1) : num >> 1;
    }
    return numbers;
  }

  /**
   * Encode a list of unsigned integers and return an encoded string
   *
   * @param {Array<number>} numbers A list of unsigned integers.
   * @return {string} The encoded string.
   */
  function encodeUnsignedIntegers(numbers) {
    let encoded = '';
    for (let i = 0, ii = numbers.length; i < ii; ++i) {
      encoded += encodeUnsignedInteger(numbers[i]);
    }
    return encoded;
  }

  /**
   * Decode a list of unsigned integers from an encoded string
   *
   * @param {string} encoded An encoded string.
   * @return {Array<number>} A list of unsigned integers.
   */
  function decodeUnsignedIntegers(encoded) {
    const numbers = [];
    let current = 0;
    let shift = 0;
    for (let i = 0, ii = encoded.length; i < ii; ++i) {
      const b = encoded.charCodeAt(i) - 63;
      current |= (b & 0x1f) << shift;
      if (b < 0x20) {
        numbers.push(current);
        current = 0;
        shift = 0;
      } else {
        shift += 5;
      }
    }
    return numbers;
  }

  /**
   * Encode one single unsigned integer and return an encoded string
   *
   * @param {number} num Unsigned integer that should be encoded.
   * @return {string} The encoded string.
   */
  function encodeUnsignedInteger(num) {
    let value,
      encoded = '';
    while (num >= 0x20) {
      value = (0x20 | (num & 0x1f)) + 63;
      encoded += String.fromCharCode(value);
      num >>= 5;
    }
    value = num + 63;
    encoded += String.fromCharCode(value);
    return encoded;
  }

  /**
   * @module ol/format/TopoJSON
   */

  /**
   * @typedef {import("topojson-specification").Topology} TopoJSONTopology
   * @typedef {import("topojson-specification").GeometryCollection} TopoJSONGeometryCollection
   * @typedef {import("topojson-specification").GeometryObject} TopoJSONGeometry
   * @typedef {import("topojson-specification").Point} TopoJSONPoint
   * @typedef {import("topojson-specification").MultiPoint} TopoJSONMultiPoint
   * @typedef {import("topojson-specification").LineString} TopoJSONLineString
   * @typedef {import("topojson-specification").MultiLineString} TopoJSONMultiLineString
   * @typedef {import("topojson-specification").Polygon} TopoJSONPolygon
   * @typedef {import("topojson-specification").MultiPolygon} TopoJSONMultiPolygon
   */

  /**
   * @typedef {Object} Options
   * @property {import("../proj.js").ProjectionLike} [dataProjection='EPSG:4326'] Default data projection.
   * @property {string} [layerName] Set the name of the TopoJSON topology
   * `objects`'s children as feature property with the specified name. This means
   * that when set to `'layer'`, a topology like
   * ```
   * {
   *   "type": "Topology",
   *   "objects": {
   *     "example": {
   *       "type": "GeometryCollection",
   *       "geometries": []
   *     }
   *   }
   * }
   * ```
   * will result in features that have a property `'layer'` set to `'example'`.
   * When not set, no property will be added to features.
   * @property {Array<string>} [layers] Names of the TopoJSON topology's
   * `objects`'s children to read features from.  If not provided, features will
   * be read from all children.
   */

  /**
   * @classdesc
   * Feature format for reading data in the TopoJSON format.
   *
   * @api
   */
  class TopoJSON extends JSONFeature {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      super();

      options = options ? options : {};

      /**
       * @private
       * @type {string|undefined}
       */
      this.layerName_ = options.layerName;

      /**
       * @private
       * @type {?Array<string>}
       */
      this.layers_ = options.layers ? options.layers : null;

      /**
       * @type {import("../proj/Projection.js").default}
       */
      this.dataProjection = get$2(
        options.dataProjection ? options.dataProjection : 'EPSG:4326'
      );
    }

    /**
     * @param {Object} object Object.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @protected
     * @return {Array<Feature>} Features.
     */
    readFeaturesFromObject(object, options) {
      if (object.type == 'Topology') {
        const topoJSONTopology = /** @type {TopoJSONTopology} */ (object);
        let transform,
          scale = null,
          translate = null;
        if (topoJSONTopology['transform']) {
          transform = topoJSONTopology['transform'];
          scale = transform['scale'];
          translate = transform['translate'];
        }
        const arcs = topoJSONTopology['arcs'];
        if (transform) {
          transformArcs(arcs, scale, translate);
        }
        /** @type {Array<Feature>} */
        const features = [];
        const topoJSONFeatures = topoJSONTopology['objects'];
        const property = this.layerName_;
        let feature;
        for (const objectName in topoJSONFeatures) {
          if (this.layers_ && !this.layers_.includes(objectName)) {
            continue;
          }
          if (topoJSONFeatures[objectName].type === 'GeometryCollection') {
            feature = /** @type {TopoJSONGeometryCollection} */ (
              topoJSONFeatures[objectName]
            );
            features.push.apply(
              features,
              readFeaturesFromGeometryCollection(
                feature,
                arcs,
                scale,
                translate,
                property,
                objectName,
                options
              )
            );
          } else {
            feature = /** @type {TopoJSONGeometry} */ (
              topoJSONFeatures[objectName]
            );
            features.push(
              readFeatureFromGeometry(
                feature,
                arcs,
                scale,
                translate,
                property,
                objectName,
                options
              )
            );
          }
        }
        return features;
      } else {
        return [];
      }
    }

    /**
     * @param {Object} object Object.
     * @protected
     * @return {import("../proj/Projection.js").default} Projection.
     */
    readProjectionFromObject(object) {
      return this.dataProjection;
    }
  }

  /**
   * @const
   * @type {Object<string, function(TopoJSONGeometry, Array, ...Array=): import("../geom/Geometry.js").default>}
   */
  const GEOMETRY_READERS = {
    'Point': readPointGeometry,
    'LineString': readLineStringGeometry,
    'Polygon': readPolygonGeometry,
    'MultiPoint': readMultiPointGeometry,
    'MultiLineString': readMultiLineStringGeometry,
    'MultiPolygon': readMultiPolygonGeometry,
  };

  /**
   * Concatenate arcs into a coordinate array.
   * @param {Array<number>} indices Indices of arcs to concatenate.  Negative
   *     values indicate arcs need to be reversed.
   * @param {Array<Array<import("../coordinate.js").Coordinate>>} arcs Array of arcs (already
   *     transformed).
   * @return {Array<import("../coordinate.js").Coordinate>} Coordinates array.
   */
  function concatenateArcs(indices, arcs) {
    /** @type {Array<import("../coordinate.js").Coordinate>} */
    const coordinates = [];
    let index;
    for (let i = 0, ii = indices.length; i < ii; ++i) {
      index = indices[i];
      if (i > 0) {
        // splicing together arcs, discard last point
        coordinates.pop();
      }
      if (index >= 0) {
        // forward arc
        const arc = arcs[index];
        for (let j = 0, jj = arc.length; j < jj; ++j) {
          coordinates.push(arc[j].slice(0));
        }
      } else {
        // reverse arc
        const arc = arcs[~index];
        for (let j = arc.length - 1; j >= 0; --j) {
          coordinates.push(arc[j].slice(0));
        }
      }
    }
    return coordinates;
  }

  /**
   * Create a point from a TopoJSON geometry object.
   *
   * @param {TopoJSONPoint} object TopoJSON object.
   * @param {Array<number>} scale Scale for each dimension.
   * @param {Array<number>} translate Translation for each dimension.
   * @return {Point} Geometry.
   */
  function readPointGeometry(object, scale, translate) {
    const coordinates = object['coordinates'];
    if (scale && translate) {
      transformVertex(coordinates, scale, translate);
    }
    return new Point(coordinates);
  }

  /**
   * Create a multi-point from a TopoJSON geometry object.
   *
   * @param {TopoJSONMultiPoint} object TopoJSON object.
   * @param {Array<number>} scale Scale for each dimension.
   * @param {Array<number>} translate Translation for each dimension.
   * @return {MultiPoint} Geometry.
   */
  function readMultiPointGeometry(object, scale, translate) {
    const coordinates = object['coordinates'];
    if (scale && translate) {
      for (let i = 0, ii = coordinates.length; i < ii; ++i) {
        transformVertex(coordinates[i], scale, translate);
      }
    }
    return new MultiPoint(coordinates);
  }

  /**
   * Create a linestring from a TopoJSON geometry object.
   *
   * @param {TopoJSONLineString} object TopoJSON object.
   * @param {Array<Array<import("../coordinate.js").Coordinate>>} arcs Array of arcs.
   * @return {LineString} Geometry.
   */
  function readLineStringGeometry(object, arcs) {
    const coordinates = concatenateArcs(object['arcs'], arcs);
    return new LineString(coordinates);
  }

  /**
   * Create a multi-linestring from a TopoJSON geometry object.
   *
   * @param {TopoJSONMultiLineString} object TopoJSON object.
   * @param {Array<Array<import("../coordinate.js").Coordinate>>} arcs Array of arcs.
   * @return {MultiLineString} Geometry.
   */
  function readMultiLineStringGeometry(object, arcs) {
    const coordinates = [];
    for (let i = 0, ii = object['arcs'].length; i < ii; ++i) {
      coordinates[i] = concatenateArcs(object['arcs'][i], arcs);
    }
    return new MultiLineString(coordinates);
  }

  /**
   * Create a polygon from a TopoJSON geometry object.
   *
   * @param {TopoJSONPolygon} object TopoJSON object.
   * @param {Array<Array<import("../coordinate.js").Coordinate>>} arcs Array of arcs.
   * @return {Polygon} Geometry.
   */
  function readPolygonGeometry(object, arcs) {
    const coordinates = [];
    for (let i = 0, ii = object['arcs'].length; i < ii; ++i) {
      coordinates[i] = concatenateArcs(object['arcs'][i], arcs);
    }
    return new Polygon(coordinates);
  }

  /**
   * Create a multi-polygon from a TopoJSON geometry object.
   *
   * @param {TopoJSONMultiPolygon} object TopoJSON object.
   * @param {Array<Array<import("../coordinate.js").Coordinate>>} arcs Array of arcs.
   * @return {MultiPolygon} Geometry.
   */
  function readMultiPolygonGeometry(object, arcs) {
    const coordinates = [];
    for (let i = 0, ii = object['arcs'].length; i < ii; ++i) {
      // for each polygon
      const polyArray = object['arcs'][i];
      const ringCoords = [];
      for (let j = 0, jj = polyArray.length; j < jj; ++j) {
        // for each ring
        ringCoords[j] = concatenateArcs(polyArray[j], arcs);
      }
      coordinates[i] = ringCoords;
    }
    return new MultiPolygon(coordinates);
  }

  /**
   * Create features from a TopoJSON GeometryCollection object.
   *
   * @param {TopoJSONGeometryCollection} collection TopoJSON Geometry
   *     object.
   * @param {Array<Array<import("../coordinate.js").Coordinate>>} arcs Array of arcs.
   * @param {Array<number>} scale Scale for each dimension.
   * @param {Array<number>} translate Translation for each dimension.
   * @param {string|undefined} property Property to set the `GeometryCollection`'s parent
   *     object to.
   * @param {string} name Name of the `Topology`'s child object.
   * @param {import("./Feature.js").ReadOptions} [options] Read options.
   * @return {Array<Feature>} Array of features.
   */
  function readFeaturesFromGeometryCollection(
    collection,
    arcs,
    scale,
    translate,
    property,
    name,
    options
  ) {
    const geometries = collection['geometries'];
    const features = [];
    for (let i = 0, ii = geometries.length; i < ii; ++i) {
      features[i] = readFeatureFromGeometry(
        geometries[i],
        arcs,
        scale,
        translate,
        property,
        name,
        options
      );
    }
    return features;
  }

  /**
   * Create a feature from a TopoJSON geometry object.
   *
   * @param {TopoJSONGeometry} object TopoJSON geometry object.
   * @param {Array<Array<import("../coordinate.js").Coordinate>>} arcs Array of arcs.
   * @param {Array<number>} scale Scale for each dimension.
   * @param {Array<number>} translate Translation for each dimension.
   * @param {string|undefined} property Property to set the `GeometryCollection`'s parent
   *     object to.
   * @param {string} name Name of the `Topology`'s child object.
   * @param {import("./Feature.js").ReadOptions} [options] Read options.
   * @return {Feature} Feature.
   */
  function readFeatureFromGeometry(
    object,
    arcs,
    scale,
    translate,
    property,
    name,
    options
  ) {
    let geometry = null;
    const type = object.type;
    if (type) {
      const geometryReader = GEOMETRY_READERS[type];
      if (type === 'Point' || type === 'MultiPoint') {
        geometry = geometryReader(object, scale, translate);
      } else {
        geometry = geometryReader(object, arcs);
      }
      geometry = transformGeometryWithOptions(geometry, false, options);
    }
    const feature = new Feature({geometry: geometry});
    if (object.id !== undefined) {
      feature.setId(object.id);
    }
    let properties = object.properties;
    if (property) {
      if (!properties) {
        properties = {};
      }
      properties[property] = name;
    }
    if (properties) {
      feature.setProperties(properties, true);
    }
    return feature;
  }

  /**
   * Apply a linear transform to array of arcs.  The provided array of arcs is
   * modified in place.
   *
   * @param {Array<Array<import("../coordinate.js").Coordinate>>} arcs Array of arcs.
   * @param {Array<number>} scale Scale for each dimension.
   * @param {Array<number>} translate Translation for each dimension.
   */
  function transformArcs(arcs, scale, translate) {
    for (let i = 0, ii = arcs.length; i < ii; ++i) {
      transformArc(arcs[i], scale, translate);
    }
  }

  /**
   * Apply a linear transform to an arc.  The provided arc is modified in place.
   *
   * @param {Array<import("../coordinate.js").Coordinate>} arc Arc.
   * @param {Array<number>} scale Scale for each dimension.
   * @param {Array<number>} translate Translation for each dimension.
   */
  function transformArc(arc, scale, translate) {
    let x = 0;
    let y = 0;
    for (let i = 0, ii = arc.length; i < ii; ++i) {
      const vertex = arc[i];
      x += vertex[0];
      y += vertex[1];
      vertex[0] = x;
      vertex[1] = y;
      transformVertex(vertex, scale, translate);
    }
  }

  /**
   * Apply a linear transform to a vertex.  The provided vertex is modified in
   * place.
   *
   * @param {import("../coordinate.js").Coordinate} vertex Vertex.
   * @param {Array<number>} scale Scale for each dimension.
   * @param {Array<number>} translate Translation for each dimension.
   */
  function transformVertex(vertex, scale, translate) {
    vertex[0] = vertex[0] * scale[0] + translate[0];
    vertex[1] = vertex[1] * scale[1] + translate[1];
  }

  /**
   * @module ol/format/filter/Filter
   */

  /**
   * @classdesc
   * Abstract class; normally only used for creating subclasses and not instantiated in apps.
   * Base class for WFS GetFeature filters.
   *
   * @abstract
   */
  class Filter {
    /**
     * @param {!string} tagName The XML tag name for this filter.
     */
    constructor(tagName) {
      /**
       * @private
       * @type {!string}
       */
      this.tagName_ = tagName;
    }

    /**
     * The XML tag name for a filter.
     * @return {!string} Name.
     */
    getTagName() {
      return this.tagName_;
    }
  }

  /**
   * @module ol/format/filter/LogicalNary
   */

  /**
   * @classdesc
   * Abstract class; normally only used for creating subclasses and not instantiated in apps.
   * Base class for WFS GetFeature n-ary logical filters.
   *
   * @abstract
   */
  class LogicalNary extends Filter {
    /**
     * @param {!string} tagName The XML tag name for this filter.
     * @param {Array<import("./Filter.js").default>} conditions Conditions.
     */
    constructor(tagName, conditions) {
      super(tagName);

      /**
       * @type {Array<import("./Filter.js").default>}
       */
      this.conditions = conditions;
      assert(this.conditions.length >= 2, 57); // At least 2 conditions are required.
    }
  }

  /**
   * @module ol/format/filter/And
   */

  /**
   * @classdesc
   * Represents a logical `<And>` operator between two or more filter conditions.
   *
   * @abstract
   */
  class And extends LogicalNary {
    /**
     * @param {...import("./Filter.js").default} conditions Conditions.
     */
    constructor(conditions) {
      super('And', Array.prototype.slice.call(arguments));
    }
  }

  /**
   * @module ol/format/filter/Bbox
   */

  /**
   * @classdesc
   * Represents a `<BBOX>` operator to test whether a geometry-valued property
   * intersects a fixed bounding box
   *
   * @api
   */
  class Bbox extends Filter {
    /**
     * @param {!string} geometryName Geometry name to use.
     * @param {!import("../../extent.js").Extent} extent Extent.
     * @param {string} [srsName] SRS name. No srsName attribute will be set
     * on geometries when this is not provided.
     */
    constructor(geometryName, extent, srsName) {
      super('BBOX');

      /**
       * @type {!string}
       */
      this.geometryName = geometryName;

      /**
       * @type {import("../../extent.js").Extent}
       */
      this.extent = extent;
      if (extent.length !== 4) {
        throw new Error(
          'Expected an extent with four values ([minX, minY, maxX, maxY])'
        );
      }

      /**
       * @type {string|undefined}
       */
      this.srsName = srsName;
    }
  }

  /**
   * @module ol/format/filter/Spatial
   */

  /**
   * @classdesc
   * Abstract class; normally only used for creating subclasses and not instantiated in apps.
   * Represents a spatial operator to test whether a geometry-valued property
   * relates to a given geometry.
   *
   * @abstract
   */
  class Spatial extends Filter {
    /**
     * @param {!string} tagName The XML tag name for this filter.
     * @param {!string} geometryName Geometry name to use.
     * @param {!import("../../geom/Geometry.js").default} geometry Geometry.
     * @param {string} [srsName] SRS name. No srsName attribute will be
     *    set on geometries when this is not provided.
     */
    constructor(tagName, geometryName, geometry, srsName) {
      super(tagName);

      /**
       * @type {!string}
       */
      this.geometryName = geometryName || 'the_geom';

      /**
       * @type {import("../../geom/Geometry.js").default}
       */
      this.geometry = geometry;

      /**
       * @type {string|undefined}
       */
      this.srsName = srsName;
    }
  }

  /**
   * @module ol/format/filter/Contains
   */

  /**
   * @classdesc
   * Represents a `<Contains>` operator to test whether a geometry-valued property
   * contains a given geometry.
   * @api
   */
  class Contains extends Spatial {
    /**
     * @param {!string} geometryName Geometry name to use.
     * @param {!import("../../geom/Geometry.js").default} geometry Geometry.
     * @param {string} [srsName] SRS name. No srsName attribute will be
     *    set on geometries when this is not provided.
     */
    constructor(geometryName, geometry, srsName) {
      super('Contains', geometryName, geometry, srsName);
    }
  }

  /**
   * @module ol/format/filter/DWithin
   */

  /**
   * @classdesc
   * Represents a `<DWithin>` operator to test whether a geometry-valued property
   * is within a distance to a given geometry.
   * @api
   */
  class DWithin extends Spatial {
    /**
     * @param {!string} geometryName Geometry name to use.
     * @param {!import("../../geom/Geometry.js").default} geometry Geometry.
     * @param {!number} distance Distance.
     * @param {!string} unit Unit.
     * @param {string} [srsName] SRS name. No srsName attribute will be
     *    set on geometries when this is not provided.
     */
    constructor(geometryName, geometry, distance, unit, srsName) {
      super('DWithin', geometryName, geometry, srsName);

      /**
       * @public
       * @type {!number}
       */
      this.distance = distance;

      /**
       * @public
       * @type {!string}
       */
      this.unit = unit;
    }
  }

  /**
   * @module ol/format/filter/Disjoint
   */

  /**
   * @classdesc
   * Represents a `<Disjoint>` operator to test whether a geometry-valued property
   * is disjoint to a given geometry.
   * @api
   */
  class Disjoint extends Spatial {
    /**
     * @param {!string} geometryName Geometry name to use.
     * @param {!import("../../geom/Geometry.js").default} geometry Geometry.
     * @param {string} [srsName] SRS name. No srsName attribute will be
     *    set on geometries when this is not provided.
     */
    constructor(geometryName, geometry, srsName) {
      super('Disjoint', geometryName, geometry, srsName);
    }
  }

  /**
   * @module ol/format/filter/Comparison
   */

  /**
   * @classdesc
   * Abstract class; normally only used for creating subclasses and not instantiated in apps.
   * Base class for WFS GetFeature property comparison filters.
   *
   * @abstract
   */
  class Comparison extends Filter {
    /**
     * @param {!string} tagName The XML tag name for this filter.
     * @param {!string} propertyName Name of the context property to compare.
     */
    constructor(tagName, propertyName) {
      super(tagName);

      /**
       * @type {!string}
       */
      this.propertyName = propertyName;
    }
  }

  /**
   * @module ol/format/filter/During
   */

  /**
   * @classdesc
   * Represents a `<During>` comparison operator.
   * @api
   */
  class During extends Comparison {
    /**
     * @param {!string} propertyName Name of the context property to compare.
     * @param {!string} begin The begin date in ISO-8601 format.
     * @param {!string} end The end date in ISO-8601 format.
     */
    constructor(propertyName, begin, end) {
      super('During', propertyName);

      /**
       * @type {!string}
       */
      this.begin = begin;

      /**
       * @type {!string}
       */
      this.end = end;
    }
  }

  /**
   * @module ol/format/filter/ComparisonBinary
   */

  /**
   * @classdesc
   * Abstract class; normally only used for creating subclasses and not instantiated in apps.
   * Base class for WFS GetFeature property binary comparison filters.
   *
   * @abstract
   */
  class ComparisonBinary extends Comparison {
    /**
     * @param {!string} tagName The XML tag name for this filter.
     * @param {!string} propertyName Name of the context property to compare.
     * @param {!(string|number)} expression The value to compare.
     * @param {boolean} [matchCase] Case-sensitive?
     */
    constructor(tagName, propertyName, expression, matchCase) {
      super(tagName, propertyName);

      /**
       * @type {!(string|number)}
       */
      this.expression = expression;

      /**
       * @type {boolean|undefined}
       */
      this.matchCase = matchCase;
    }
  }

  /**
   * @module ol/format/filter/EqualTo
   */

  /**
   * @classdesc
   * Represents a `<PropertyIsEqualTo>` comparison operator.
   * @api
   */
  class EqualTo extends ComparisonBinary {
    /**
     * @param {!string} propertyName Name of the context property to compare.
     * @param {!(string|number)} expression The value to compare.
     * @param {boolean} [matchCase] Case-sensitive?
     */
    constructor(propertyName, expression, matchCase) {
      super('PropertyIsEqualTo', propertyName, expression, matchCase);
    }
  }

  /**
   * @module ol/format/filter/GreaterThan
   */

  /**
   * @classdesc
   * Represents a `<PropertyIsGreaterThan>` comparison operator.
   * @api
   */
  class GreaterThan extends ComparisonBinary {
    /**
     * @param {!string} propertyName Name of the context property to compare.
     * @param {!number} expression The value to compare.
     */
    constructor(propertyName, expression) {
      super('PropertyIsGreaterThan', propertyName, expression);
    }
  }

  /**
   * @module ol/format/filter/GreaterThanOrEqualTo
   */

  /**
   * @classdesc
   * Represents a `<PropertyIsGreaterThanOrEqualTo>` comparison operator.
   * @api
   */
  class GreaterThanOrEqualTo extends ComparisonBinary {
    /**
     * @param {!string} propertyName Name of the context property to compare.
     * @param {!number} expression The value to compare.
     */
    constructor(propertyName, expression) {
      super('PropertyIsGreaterThanOrEqualTo', propertyName, expression);
    }
  }

  /**
   * @module ol/format/filter/Intersects
   */

  /**
   * @classdesc
   * Represents a `<Intersects>` operator to test whether a geometry-valued property
   * intersects a given geometry.
   * @api
   */
  class Intersects extends Spatial {
    /**
     * @param {!string} geometryName Geometry name to use.
     * @param {!import("../../geom/Geometry.js").default} geometry Geometry.
     * @param {string} [srsName] SRS name. No srsName attribute will be
     *    set on geometries when this is not provided.
     */
    constructor(geometryName, geometry, srsName) {
      super('Intersects', geometryName, geometry, srsName);
    }
  }

  /**
   * @module ol/format/filter/IsBetween
   */

  /**
   * @classdesc
   * Represents a `<PropertyIsBetween>` comparison operator.
   * @api
   */
  class IsBetween extends Comparison {
    /**
     * @param {!string} propertyName Name of the context property to compare.
     * @param {!number} lowerBoundary The lower bound of the range.
     * @param {!number} upperBoundary The upper bound of the range.
     */
    constructor(propertyName, lowerBoundary, upperBoundary) {
      super('PropertyIsBetween', propertyName);

      /**
       * @type {!number}
       */
      this.lowerBoundary = lowerBoundary;

      /**
       * @type {!number}
       */
      this.upperBoundary = upperBoundary;
    }
  }

  /**
   * @module ol/format/filter/IsLike
   */

  /**
   * @classdesc
   * Represents a `<PropertyIsLike>` comparison operator.
   * @api
   */
  class IsLike extends Comparison {
    /**
     * [constructor description]
     * @param {!string} propertyName Name of the context property to compare.
     * @param {!string} pattern Text pattern.
     * @param {string} [wildCard] Pattern character which matches any sequence of
     *    zero or more string characters. Default is '*'.
     * @param {string} [singleChar] pattern character which matches any single
     *    string character. Default is '.'.
     * @param {string} [escapeChar] Escape character which can be used to escape
     *    the pattern characters. Default is '!'.
     * @param {boolean} [matchCase] Case-sensitive?
     */
    constructor(
      propertyName,
      pattern,
      wildCard,
      singleChar,
      escapeChar,
      matchCase
    ) {
      super('PropertyIsLike', propertyName);

      /**
       * @type {!string}
       */
      this.pattern = pattern;

      /**
       * @type {!string}
       */
      this.wildCard = wildCard !== undefined ? wildCard : '*';

      /**
       * @type {!string}
       */
      this.singleChar = singleChar !== undefined ? singleChar : '.';

      /**
       * @type {!string}
       */
      this.escapeChar = escapeChar !== undefined ? escapeChar : '!';

      /**
       * @type {boolean|undefined}
       */
      this.matchCase = matchCase;
    }
  }

  /**
   * @module ol/format/filter/IsNull
   */

  /**
   * @classdesc
   * Represents a `<PropertyIsNull>` comparison operator.
   * @api
   */
  class IsNull extends Comparison {
    /**
     * @param {!string} propertyName Name of the context property to compare.
     */
    constructor(propertyName) {
      super('PropertyIsNull', propertyName);
    }
  }

  /**
   * @module ol/format/filter/LessThan
   */

  /**
   * @classdesc
   * Represents a `<PropertyIsLessThan>` comparison operator.
   * @api
   */
  class LessThan extends ComparisonBinary {
    /**
     * @param {!string} propertyName Name of the context property to compare.
     * @param {!number} expression The value to compare.
     */
    constructor(propertyName, expression) {
      super('PropertyIsLessThan', propertyName, expression);
    }
  }

  /**
   * @module ol/format/filter/LessThanOrEqualTo
   */

  /**
   * @classdesc
   * Represents a `<PropertyIsLessThanOrEqualTo>` comparison operator.
   * @api
   */
  class LessThanOrEqualTo extends ComparisonBinary {
    /**
     * @param {!string} propertyName Name of the context property to compare.
     * @param {!number} expression The value to compare.
     */
    constructor(propertyName, expression) {
      super('PropertyIsLessThanOrEqualTo', propertyName, expression);
    }
  }

  /**
   * @module ol/format/filter/Not
   */

  /**
   * @classdesc
   * Represents a logical `<Not>` operator for a filter condition.
   * @api
   */
  class Not extends Filter {
    /**
     * @param {!import("./Filter.js").default} condition Filter condition.
     */
    constructor(condition) {
      super('Not');

      /**
       * @type {!import("./Filter.js").default}
       */
      this.condition = condition;
    }
  }

  /**
   * @module ol/format/filter/NotEqualTo
   */

  /**
   * @classdesc
   * Represents a `<PropertyIsNotEqualTo>` comparison operator.
   * @api
   */
  class NotEqualTo extends ComparisonBinary {
    /**
     * @param {!string} propertyName Name of the context property to compare.
     * @param {!(string|number)} expression The value to compare.
     * @param {boolean} [matchCase] Case-sensitive?
     */
    constructor(propertyName, expression, matchCase) {
      super('PropertyIsNotEqualTo', propertyName, expression, matchCase);
    }
  }

  /**
   * @module ol/format/filter/Or
   */

  /**
   * @classdesc
   * Represents a logical `<Or>` operator between two ore more filter conditions.
   * @api
   */
  class Or extends LogicalNary {
    /**
     * @param {...import("./Filter.js").default} conditions Conditions.
     */
    constructor(conditions) {
      super('Or', Array.prototype.slice.call(arguments));
    }
  }

  /**
   * @module ol/format/filter/ResourceId
   */

  /**
   * @classdesc
   *
   * @abstract
   */
  class ResourceId extends Filter {
    /**
     * @param {!string} rid Resource ID.
     */
    constructor(rid) {
      super('ResourceId');

      /**
       * @type {!string}
       */
      this.rid = rid;
    }
  }

  /**
   * @module ol/format/filter/Within
   */

  /**
   * @classdesc
   * Represents a `<Within>` operator to test whether a geometry-valued property
   * is within a given geometry.
   * @api
   */
  class Within extends Spatial {
    /**
     * @param {!string} geometryName Geometry name to use.
     * @param {!import("../../geom/Geometry.js").default} geometry Geometry.
     * @param {string} [srsName] SRS name. No srsName attribute will be
     *    set on geometries when this is not provided.
     */
    constructor(geometryName, geometry, srsName) {
      super('Within', geometryName, geometry, srsName);
    }
  }

  /**
   * @module ol/format/filter
   */

  /**
   * Create a logical `<And>` operator between two or more filter conditions.
   *
   * @param {...import("./filter/Filter.js").default} conditions Filter conditions.
   * @return {!And} `<And>` operator.
   * @api
   */
  function and(conditions) {
    const params = [null].concat(Array.prototype.slice.call(arguments));
    return new (Function.prototype.bind.apply(And, params))();
  }

  /**
   * Create a logical `<Or>` operator between two or more filter conditions.
   *
   * @param {...import("./filter/Filter.js").default} conditions Filter conditions.
   * @return {!Or} `<Or>` operator.
   * @api
   */
  function or(conditions) {
    const params = [null].concat(Array.prototype.slice.call(arguments));
    return new (Function.prototype.bind.apply(Or, params))();
  }

  /**
   * Represents a logical `<Not>` operator for a filter condition.
   *
   * @param {!import("./filter/Filter.js").default} condition Filter condition.
   * @return {!Not} `<Not>` operator.
   * @api
   */
  function not(condition) {
    return new Not(condition);
  }

  /**
   * Create a `<BBOX>` operator to test whether a geometry-valued property
   * intersects a fixed bounding box
   *
   * @param {!string} geometryName Geometry name to use.
   * @param {!import("../extent.js").Extent} extent Extent.
   * @param {string} [srsName] SRS name. No srsName attribute will be
   *    set on geometries when this is not provided.
   * @return {!Bbox} `<BBOX>` operator.
   * @api
   */
  function bbox(geometryName, extent, srsName) {
    return new Bbox(geometryName, extent, srsName);
  }

  /**
   * Create a `<Contains>` operator to test whether a geometry-valued property
   * contains a given geometry.
   *
   * @param {!string} geometryName Geometry name to use.
   * @param {!import("../geom/Geometry.js").default} geometry Geometry.
   * @param {string} [srsName] SRS name. No srsName attribute will be
   *    set on geometries when this is not provided.
   * @return {!Contains} `<Contains>` operator.
   * @api
   */
  function contains(geometryName, geometry, srsName) {
    return new Contains(geometryName, geometry, srsName);
  }

  /**
   * Create a `<Intersects>` operator to test whether a geometry-valued property
   * intersects a given geometry.
   *
   * @param {!string} geometryName Geometry name to use.
   * @param {!import("../geom/Geometry.js").default} geometry Geometry.
   * @param {string} [srsName] SRS name. No srsName attribute will be
   *    set on geometries when this is not provided.
   * @return {!Intersects} `<Intersects>` operator.
   * @api
   */
  function intersects(geometryName, geometry, srsName) {
    return new Intersects(geometryName, geometry, srsName);
  }

  /**
   * Create a `<Disjoint>` operator to test whether a geometry-valued property
   * is disjoint to a given geometry.
   *
   * @param {!string} geometryName Geometry name to use.
   * @param {!import("../geom/Geometry.js").default} geometry Geometry.
   * @param {string} [srsName] SRS name. No srsName attribute will be
   *    set on geometries when this is not provided.
   * @return {!Disjoint} `<Disjoint>` operator.
   * @api
   */
  function disjoint(geometryName, geometry, srsName) {
    return new Disjoint(geometryName, geometry, srsName);
  }

  /**
   * Create a `<Within>` operator to test whether a geometry-valued property
   * is within a given geometry.
   *
   * @param {!string} geometryName Geometry name to use.
   * @param {!import("../geom/Geometry.js").default} geometry Geometry.
   * @param {string} [srsName] SRS name. No srsName attribute will be
   *    set on geometries when this is not provided.
   * @return {!Within} `<Within>` operator.
   * @api
   */
  function within(geometryName, geometry, srsName) {
    return new Within(geometryName, geometry, srsName);
  }

  /**
   * Create a `<DWithin>` operator to test whether a geometry-valued property
   * is within a distance to a given geometry.
   *
   * @param {!string} geometryName Geometry name to use.
   * @param {!import("../geom/Geometry.js").default} geometry Geometry.
   * @param {!number} distance Distance.
   * @param {!string} unit Unit.
   * @param {string} [srsName] SRS name. No srsName attribute will be
   *    set on geometries when this is not provided.
   * @return {!DWithin} `<DWithin>` operator.
   * @api
   */
  function dwithin(geometryName, geometry, distance, unit, srsName) {
    return new DWithin(geometryName, geometry, distance, unit, srsName);
  }

  /**
   * Creates a `<PropertyIsEqualTo>` comparison operator.
   *
   * @param {!string} propertyName Name of the context property to compare.
   * @param {!(string|number)} expression The value to compare.
   * @param {boolean} [matchCase] Case-sensitive?
   * @return {!EqualTo} `<PropertyIsEqualTo>` operator.
   * @api
   */
  function equalTo(propertyName, expression, matchCase) {
    return new EqualTo(propertyName, expression, matchCase);
  }

  /**
   * Creates a `<PropertyIsNotEqualTo>` comparison operator.
   *
   * @param {!string} propertyName Name of the context property to compare.
   * @param {!(string|number)} expression The value to compare.
   * @param {boolean} [matchCase] Case-sensitive?
   * @return {!NotEqualTo} `<PropertyIsNotEqualTo>` operator.
   * @api
   */
  function notEqualTo(propertyName, expression, matchCase) {
    return new NotEqualTo(propertyName, expression, matchCase);
  }

  /**
   * Creates a `<PropertyIsLessThan>` comparison operator.
   *
   * @param {!string} propertyName Name of the context property to compare.
   * @param {!number} expression The value to compare.
   * @return {!LessThan} `<PropertyIsLessThan>` operator.
   * @api
   */
  function lessThan(propertyName, expression) {
    return new LessThan(propertyName, expression);
  }

  /**
   * Creates a `<PropertyIsLessThanOrEqualTo>` comparison operator.
   *
   * @param {!string} propertyName Name of the context property to compare.
   * @param {!number} expression The value to compare.
   * @return {!LessThanOrEqualTo} `<PropertyIsLessThanOrEqualTo>` operator.
   * @api
   */
  function lessThanOrEqualTo(propertyName, expression) {
    return new LessThanOrEqualTo(propertyName, expression);
  }

  /**
   * Creates a `<PropertyIsGreaterThan>` comparison operator.
   *
   * @param {!string} propertyName Name of the context property to compare.
   * @param {!number} expression The value to compare.
   * @return {!GreaterThan} `<PropertyIsGreaterThan>` operator.
   * @api
   */
  function greaterThan(propertyName, expression) {
    return new GreaterThan(propertyName, expression);
  }

  /**
   * Creates a `<PropertyIsGreaterThanOrEqualTo>` comparison operator.
   *
   * @param {!string} propertyName Name of the context property to compare.
   * @param {!number} expression The value to compare.
   * @return {!GreaterThanOrEqualTo} `<PropertyIsGreaterThanOrEqualTo>` operator.
   * @api
   */
  function greaterThanOrEqualTo(propertyName, expression) {
    return new GreaterThanOrEqualTo(propertyName, expression);
  }

  /**
   * Creates a `<PropertyIsNull>` comparison operator to test whether a property value
   * is null.
   *
   * @param {!string} propertyName Name of the context property to compare.
   * @return {!IsNull} `<PropertyIsNull>` operator.
   * @api
   */
  function isNull(propertyName) {
    return new IsNull(propertyName);
  }

  /**
   * Creates a `<PropertyIsBetween>` comparison operator to test whether an expression
   * value lies within a range given by a lower and upper bound (inclusive).
   *
   * @param {!string} propertyName Name of the context property to compare.
   * @param {!number} lowerBoundary The lower bound of the range.
   * @param {!number} upperBoundary The upper bound of the range.
   * @return {!IsBetween} `<PropertyIsBetween>` operator.
   * @api
   */
  function between(propertyName, lowerBoundary, upperBoundary) {
    return new IsBetween(propertyName, lowerBoundary, upperBoundary);
  }

  /**
   * Represents a `<PropertyIsLike>` comparison operator that matches a string property
   * value against a text pattern.
   *
   * @param {!string} propertyName Name of the context property to compare.
   * @param {!string} pattern Text pattern.
   * @param {string} [wildCard] Pattern character which matches any sequence of
   *    zero or more string characters. Default is '*'.
   * @param {string} [singleChar] pattern character which matches any single
   *    string character. Default is '.'.
   * @param {string} [escapeChar] Escape character which can be used to escape
   *    the pattern characters. Default is '!'.
   * @param {boolean} [matchCase] Case-sensitive?
   * @return {!IsLike} `<PropertyIsLike>` operator.
   * @api
   */
  function like(
    propertyName,
    pattern,
    wildCard,
    singleChar,
    escapeChar,
    matchCase
  ) {
    return new IsLike(
      propertyName,
      pattern,
      wildCard,
      singleChar,
      escapeChar,
      matchCase
    );
  }

  /**
   * Create a `<During>` temporal operator.
   *
   * @param {!string} propertyName Name of the context property to compare.
   * @param {!string} begin The begin date in ISO-8601 format.
   * @param {!string} end The end date in ISO-8601 format.
   * @return {!During} `<During>` operator.
   * @api
   */
  function during(propertyName, begin, end) {
    return new During(propertyName, begin, end);
  }

  function resourceId(rid) {
    return new ResourceId(rid);
  }

  /**
   * @module ol/format/WFS
   */

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  const FEATURE_COLLECTION_PARSERS = {
    'http://www.opengis.net/gml': {
      'boundedBy': makeObjectPropertySetter(
        GMLBase.prototype.readExtentElement,
        'bounds'
      ),
    },
    'http://www.opengis.net/wfs/2.0': {
      'member': makeArrayPusher(GMLBase.prototype.readFeaturesInternal),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  const TRANSACTION_SUMMARY_PARSERS = {
    'http://www.opengis.net/wfs': {
      'totalInserted': makeObjectPropertySetter(readPositiveInteger),
      'totalUpdated': makeObjectPropertySetter(readPositiveInteger),
      'totalDeleted': makeObjectPropertySetter(readPositiveInteger),
    },
    'http://www.opengis.net/wfs/2.0': {
      'totalInserted': makeObjectPropertySetter(readPositiveInteger),
      'totalUpdated': makeObjectPropertySetter(readPositiveInteger),
      'totalDeleted': makeObjectPropertySetter(readPositiveInteger),
    },
  };

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  const TRANSACTION_RESPONSE_PARSERS = {
    'http://www.opengis.net/wfs': {
      'TransactionSummary': makeObjectPropertySetter(
        readTransactionSummary,
        'transactionSummary'
      ),
      'InsertResults': makeObjectPropertySetter(readInsertResults, 'insertIds'),
    },
    'http://www.opengis.net/wfs/2.0': {
      'TransactionSummary': makeObjectPropertySetter(
        readTransactionSummary,
        'transactionSummary'
      ),
      'InsertResults': makeObjectPropertySetter(readInsertResults, 'insertIds'),
    },
  };

  /**
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  const QUERY_SERIALIZERS = {
    'http://www.opengis.net/wfs': {
      'PropertyName': makeChildAppender(writeStringTextNode),
    },
    'http://www.opengis.net/wfs/2.0': {
      'PropertyName': makeChildAppender(writeStringTextNode),
    },
  };

  /**
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  const TRANSACTION_SERIALIZERS = {
    'http://www.opengis.net/wfs': {
      'Insert': makeChildAppender(writeFeature),
      'Update': makeChildAppender(writeUpdate),
      'Delete': makeChildAppender(writeDelete),
      'Property': makeChildAppender(writeProperty),
      'Native': makeChildAppender(writeNative),
    },
    'http://www.opengis.net/wfs/2.0': {
      'Insert': makeChildAppender(writeFeature),
      'Update': makeChildAppender(writeUpdate),
      'Delete': makeChildAppender(writeDelete),
      'Property': makeChildAppender(writeProperty),
      'Native': makeChildAppender(writeNative),
    },
  };

  /**
   * @typedef {Object} Options
   * @property {Object<string, string>|string} [featureNS] The namespace URI used for features.
   * @property {Array<string>|string} [featureType] The feature type to parse. Only used for read operations.
   * @property {GMLBase} [gmlFormat] The GML format to use to parse the response.
   * Default is `ol/format/GML2` for WFS 1.0.0, `ol/format/GML3` for WFS 1.1.0 and `ol/format/GML32` for WFS 2.0.0.
   * @property {string} [schemaLocation] Optional schemaLocation to use for serialization, this will override the default.
   * @property {string} [version='1.1.0'] WFS version to use. Can be either `1.0.0`, `1.1.0` or `2.0.0`.
   */

  /**
   * @typedef {Object} WriteGetFeatureOptions
   * @property {string} featureNS The namespace URI used for features.
   * @property {string} featurePrefix The prefix for the feature namespace.
   * @property {Array<string|FeatureType>} featureTypes The feature type names or FeatureType objects to
   * define a unique bbox filter per feature type name (in this case, options `bbox` and `geometryName` are
   * ignored.).
   * @property {string} [srsName] SRS name. No srsName attribute will be set on
   * geometries when this is not provided.
   * @property {string} [handle] Handle.
   * @property {string} [outputFormat] Output format.
   * @property {number} [maxFeatures] Maximum number of features to fetch.
   * @property {string} [geometryName] Geometry name to use in a BBOX filter.
   * @property {Array<string>} [propertyNames] Optional list of property names to serialize.
   * @property {string} [viewParams] viewParams GeoServer vendor parameter.
   * @property {number} [startIndex] Start index to use for WFS paging. This is a
   * WFS 2.0 feature backported to WFS 1.1.0 by some Web Feature Services.
   * @property {number} [count] Number of features to retrieve when paging. This is a
   * WFS 2.0 feature backported to WFS 1.1.0 by some Web Feature Services. Please note that some
   * Web Feature Services have repurposed `maxfeatures` instead.
   * @property {import("../extent.js").Extent} [bbox] Extent to use for the BBOX filter. The `geometryName`
   * option must be set.
   * @property {import("./filter/Filter.js").default} [filter] Filter condition. See
   * {@link module:ol/format/filter} for more information.
   * @property {string} [resultType] Indicates what response should be returned,
   * e.g. `hits` only includes the `numberOfFeatures` attribute in the response and no features.
   */

  /**
   * @typedef {Object} FeatureType
   * @property {!string} name The feature type name.
   * @property {!import("../extent.js").Extent} bbox Extent to use for the BBOX filter.
   * @property {!string} geometryName Geometry name to use in the BBOX filter.
   */

  /**
   * @typedef {Object} WriteTransactionOptions
   * @property {string} featureNS The namespace URI used for features.
   * @property {string} featurePrefix The prefix for the feature namespace.
   * @property {string} featureType The feature type name.
   * @property {string} [srsName] SRS name. No srsName attribute will be set on
   * geometries when this is not provided.
   * @property {string} [handle] Handle.
   * @property {boolean} [hasZ] Must be set to true if the transaction is for
   * a 3D layer. This will allow the Z coordinate to be included in the transaction.
   * @property {Array<Object>} nativeElements Native elements. Currently not supported.
   * @property {import("./GMLBase.js").Options} [gmlOptions] GML options for the WFS transaction writer.
   * @property {string} [version='1.1.0'] WFS version to use for the transaction. Can be either `1.0.0`, `1.1.0` or `2.0.0`.
   */

  /**
   * Number of features; bounds/extent.
   * @typedef {Object} FeatureCollectionMetadata
   * @property {number} numberOfFeatures NumberOfFeatures.
   * @property {import("../extent.js").Extent} bounds Bounds.
   */

  /**
   * Total deleted; total inserted; total updated; array of insert ids.
   * @typedef {Object} TransactionResponse
   * @property {number} totalDeleted TotalDeleted.
   * @property {number} totalInserted TotalInserted.
   * @property {number} totalUpdated TotalUpdated.
   * @property {Array<string>} insertIds InsertIds.
   */

  /**
   * @type {string}
   */
  const FEATURE_PREFIX = 'feature';

  /**
   * @type {string}
   */
  const XMLNS = 'http://www.w3.org/2000/xmlns/';

  /**
   * @type {Object<string, string>}
   */
  const OGCNS = {
    '2.0.0': 'http://www.opengis.net/ogc/1.1',
    '1.1.0': 'http://www.opengis.net/ogc',
    '1.0.0': 'http://www.opengis.net/ogc',
  };

  /**
   * @type {Object<string, string>}
   */
  const WFSNS = {
    '2.0.0': 'http://www.opengis.net/wfs/2.0',
    '1.1.0': 'http://www.opengis.net/wfs',
    '1.0.0': 'http://www.opengis.net/wfs',
  };

  /**
   * @type {Object<string, string>}
   */
  const FESNS = {
    '2.0.0': 'http://www.opengis.net/fes/2.0',
    '1.1.0': 'http://www.opengis.net/fes',
    '1.0.0': 'http://www.opengis.net/fes',
  };

  /**
   * @type {Object<string, string>}
   */
  const SCHEMA_LOCATIONS = {
    '2.0.0':
      'http://www.opengis.net/wfs/2.0 http://schemas.opengis.net/wfs/2.0/wfs.xsd',
    '1.1.0':
      'http://www.opengis.net/wfs http://schemas.opengis.net/wfs/1.1.0/wfs.xsd',
    '1.0.0':
      'http://www.opengis.net/wfs http://schemas.opengis.net/wfs/1.0.0/wfs.xsd',
  };

  /**
   * @type {Object<string, object>}
   */
  const GML_FORMATS = {
    '2.0.0': GML32,
    '1.1.0': GML3,
    '1.0.0': GML2,
  };

  /**
   * @const
   * @type {string}
   */
  const DEFAULT_VERSION = '1.1.0';

  /**
   * @classdesc
   * Feature format for reading and writing data in the WFS format.
   * By default, supports WFS version 1.1.0. You can pass a GML format
   * as option to override the default.
   * Also see {@link module:ol/format/GMLBase~GMLBase} which is used by this format.
   *
   * @api
   */
  class WFS extends XMLFeature {
    /**
     * @param {Options} [options] Optional configuration object.
     */
    constructor(options) {
      super();

      options = options ? options : {};

      /**
       * @private
       * @type {string}
       */
      this.version_ = options.version ? options.version : DEFAULT_VERSION;

      /**
       * @private
       * @type {Array<string>|string|undefined}
       */
      this.featureType_ = options.featureType;

      /**
       * @private
       * @type {Object<string, string>|string|undefined}
       */
      this.featureNS_ = options.featureNS;

      /**
       * @private
       * @type {GMLBase}
       */
      this.gmlFormat_ = options.gmlFormat
        ? options.gmlFormat
        : new GML_FORMATS[this.version_]();

      /**
       * @private
       * @type {string}
       */
      this.schemaLocation_ = options.schemaLocation
        ? options.schemaLocation
        : SCHEMA_LOCATIONS[this.version_];
    }

    /**
     * @return {Array<string>|string|undefined} featureType
     */
    getFeatureType() {
      return this.featureType_;
    }

    /**
     * @param {Array<string>|string|undefined} featureType Feature type(s) to parse.
     */
    setFeatureType(featureType) {
      this.featureType_ = featureType;
    }

    /**
     * @protected
     * @param {Element} node Node.
     * @param {import("./Feature.js").ReadOptions} [options] Options.
     * @return {Array<import("../Feature.js").default>} Features.
     */
    readFeaturesFromNode(node, options) {
      /** @type {import("../xml.js").NodeStackItem} */
      const context = {
        node,
      };
      Object.assign(context, {
        'featureType': this.featureType_,
        'featureNS': this.featureNS_,
      });

      Object.assign(context, this.getReadOptions(node, options ? options : {}));
      const objectStack = [context];
      let featuresNS;
      if (this.version_ === '2.0.0') {
        featuresNS = FEATURE_COLLECTION_PARSERS;
      } else {
        featuresNS = this.gmlFormat_.FEATURE_COLLECTION_PARSERS;
      }
      let features = pushParseAndPop(
        [],
        featuresNS,
        node,
        objectStack,
        this.gmlFormat_
      );
      if (!features) {
        features = [];
      }
      return features;
    }

    /**
     * Read transaction response of the source.
     *
     * @param {Document|Element|Object|string} source Source.
     * @return {TransactionResponse|undefined} Transaction response.
     * @api
     */
    readTransactionResponse(source) {
      if (!source) {
        return undefined;
      } else if (typeof source === 'string') {
        const doc = parse(source);
        return this.readTransactionResponseFromDocument(doc);
      } else if (isDocument(source)) {
        return this.readTransactionResponseFromDocument(
          /** @type {Document} */ (source)
        );
      } else {
        return this.readTransactionResponseFromNode(
          /** @type {Element} */ (source)
        );
      }
    }

    /**
     * Read feature collection metadata of the source.
     *
     * @param {Document|Element|Object|string} source Source.
     * @return {FeatureCollectionMetadata|undefined}
     *     FeatureCollection metadata.
     * @api
     */
    readFeatureCollectionMetadata(source) {
      if (!source) {
        return undefined;
      } else if (typeof source === 'string') {
        const doc = parse(source);
        return this.readFeatureCollectionMetadataFromDocument(doc);
      } else if (isDocument(source)) {
        return this.readFeatureCollectionMetadataFromDocument(
          /** @type {Document} */ (source)
        );
      } else {
        return this.readFeatureCollectionMetadataFromNode(
          /** @type {Element} */ (source)
        );
      }
    }

    /**
     * @param {Document} doc Document.
     * @return {FeatureCollectionMetadata|undefined}
     *     FeatureCollection metadata.
     */
    readFeatureCollectionMetadataFromDocument(doc) {
      for (let n = /** @type {Node} */ (doc.firstChild); n; n = n.nextSibling) {
        if (n.nodeType == Node.ELEMENT_NODE) {
          return this.readFeatureCollectionMetadataFromNode(
            /** @type {Element} */ (n)
          );
        }
      }
      return undefined;
    }

    /**
     * @param {Element} node Node.
     * @return {FeatureCollectionMetadata|undefined}
     *     FeatureCollection metadata.
     */
    readFeatureCollectionMetadataFromNode(node) {
      const result = {};
      const value = readNonNegativeIntegerString(
        node.getAttribute('numberOfFeatures')
      );
      result['numberOfFeatures'] = value;
      return pushParseAndPop(
        /** @type {FeatureCollectionMetadata} */ (result),
        FEATURE_COLLECTION_PARSERS,
        node,
        [],
        this.gmlFormat_
      );
    }

    /**
     * @param {Document} doc Document.
     * @return {TransactionResponse|undefined} Transaction response.
     */
    readTransactionResponseFromDocument(doc) {
      for (let n = /** @type {Node} */ (doc.firstChild); n; n = n.nextSibling) {
        if (n.nodeType == Node.ELEMENT_NODE) {
          return this.readTransactionResponseFromNode(/** @type {Element} */ (n));
        }
      }
      return undefined;
    }

    /**
     * @param {Element} node Node.
     * @return {TransactionResponse|undefined} Transaction response.
     */
    readTransactionResponseFromNode(node) {
      return pushParseAndPop(
        /** @type {TransactionResponse} */ ({}),
        TRANSACTION_RESPONSE_PARSERS,
        node,
        []
      );
    }

    /**
     * Encode format as WFS `GetFeature` and return the Node.
     *
     * @param {WriteGetFeatureOptions} options Options.
     * @return {Node} Result.
     * @api
     */
    writeGetFeature(options) {
      const node = createElementNS(WFSNS[this.version_], 'GetFeature');
      node.setAttribute('service', 'WFS');
      node.setAttribute('version', this.version_);
      if (options.handle) {
        node.setAttribute('handle', options.handle);
      }
      if (options.outputFormat) {
        node.setAttribute('outputFormat', options.outputFormat);
      }
      if (options.maxFeatures !== undefined) {
        node.setAttribute('maxFeatures', String(options.maxFeatures));
      }
      if (options.resultType) {
        node.setAttribute('resultType', options.resultType);
      }
      if (options.startIndex !== undefined) {
        node.setAttribute('startIndex', String(options.startIndex));
      }
      if (options.count !== undefined) {
        node.setAttribute('count', String(options.count));
      }
      if (options.viewParams !== undefined) {
        node.setAttribute('viewParams', options.viewParams);
      }
      node.setAttributeNS(
        XML_SCHEMA_INSTANCE_URI,
        'xsi:schemaLocation',
        this.schemaLocation_
      );
      /** @type {import("../xml.js").NodeStackItem} */
      const context = {
        node,
      };
      Object.assign(context, {
        'version': this.version_,
        'srsName': options.srsName,
        'featureNS': options.featureNS ? options.featureNS : this.featureNS_,
        'featurePrefix': options.featurePrefix,
        'propertyNames': options.propertyNames ? options.propertyNames : [],
      });
      assert(Array.isArray(options.featureTypes), 11); // `options.featureTypes` must be an Array
      if (typeof options.featureTypes[0] === 'string') {
        let filter = options.filter;
        if (options.bbox) {
          assert(options.geometryName, 12); // `options.geometryName` must also be provided when `options.bbox` is set
          filter = this.combineBboxAndFilter(
            options.geometryName,
            options.bbox,
            options.srsName,
            filter
          );
        }
        Object.assign(context, {
          'geometryName': options.geometryName,
          'filter': filter,
        });
        writeGetFeature(
          node,
          /** @type {!Array<string>} */ (options.featureTypes),
          [context]
        );
      } else {
        // Write one query node per element in featuresType.
        options.featureTypes.forEach((/** @type {FeatureType} */ featureType) => {
          const completeFilter = this.combineBboxAndFilter(
            featureType.geometryName,
            featureType.bbox,
            options.srsName,
            options.filter
          );
          Object.assign(context, {
            'geometryName': featureType.geometryName,
            'filter': completeFilter,
          });
          writeGetFeature(node, [featureType.name], [context]);
        });
      }
      return node;
    }

    /**
     * Create a bbox filter and combine it with another optional filter.
     *
     * @param {!string} geometryName Geometry name to use.
     * @param {!import("../extent.js").Extent} extent Extent.
     * @param {string} [srsName] SRS name. No srsName attribute will be
     *    set on geometries when this is not provided.
     * @param {import("./filter/Filter.js").default} [filter] Filter condition.
     * @return {import("./filter/Filter.js").default} The filter.
     */
    combineBboxAndFilter(geometryName, extent, srsName, filter) {
      const bboxFilter = bbox(geometryName, extent, srsName);
      if (filter) {
        // if bbox and filter are both set, combine the two into a single filter
        return and(filter, bboxFilter);
      }
      return bboxFilter;
    }

    /**
     * Encode format as WFS `Transaction` and return the Node.
     *
     * @param {Array<import("../Feature.js").default>} inserts The features to insert.
     * @param {Array<import("../Feature.js").default>} updates The features to update.
     * @param {Array<import("../Feature.js").default>} deletes The features to delete.
     * @param {WriteTransactionOptions} options Write options.
     * @return {Node} Result.
     * @api
     */
    writeTransaction(inserts, updates, deletes, options) {
      const objectStack = [];
      const version = options.version ? options.version : this.version_;
      const node = createElementNS(WFSNS[version], 'Transaction');

      node.setAttribute('service', 'WFS');
      node.setAttribute('version', version);
      let baseObj;
      /** @type {import("../xml.js").NodeStackItem} */
      if (options) {
        baseObj = options.gmlOptions ? options.gmlOptions : {};
        if (options.handle) {
          node.setAttribute('handle', options.handle);
        }
      }
      node.setAttributeNS(
        XML_SCHEMA_INSTANCE_URI,
        'xsi:schemaLocation',
        SCHEMA_LOCATIONS[version]
      );

      const request = createTransactionRequest(node, baseObj, version, options);
      if (inserts) {
        serializeTransactionRequest('Insert', inserts, objectStack, request);
      }
      if (updates) {
        serializeTransactionRequest('Update', updates, objectStack, request);
      }
      if (deletes) {
        serializeTransactionRequest('Delete', deletes, objectStack, request);
      }
      if (options.nativeElements) {
        serializeTransactionRequest(
          'Native',
          options.nativeElements,
          objectStack,
          request
        );
      }
      return node;
    }

    /**
     * @param {Document} doc Document.
     * @return {import("../proj/Projection.js").default} Projection.
     */
    readProjectionFromDocument(doc) {
      for (let n = doc.firstChild; n; n = n.nextSibling) {
        if (n.nodeType == Node.ELEMENT_NODE) {
          return this.readProjectionFromNode(/** @type {Element} */ (n));
        }
      }
      return null;
    }

    /**
     * @param {Element} node Node.
     * @return {import("../proj/Projection.js").default} Projection.
     */
    readProjectionFromNode(node) {
      if (node.firstElementChild && node.firstElementChild.firstElementChild) {
        node = node.firstElementChild.firstElementChild;
        for (let n = node.firstElementChild; n; n = n.nextElementSibling) {
          if (
            !(
              n.childNodes.length === 0 ||
              (n.childNodes.length === 1 && n.firstChild.nodeType === 3)
            )
          ) {
            const objectStack = [{}];
            this.gmlFormat_.readGeometryElement(n, objectStack);
            return get$2(objectStack.pop().srsName);
          }
        }
      }

      return null;
    }
  }

  /**
   * @param {Element} node Node.
   * @param {*} baseObj Base object.
   * @param {string} version Version.
   * @param {WriteTransactionOptions} options Options.
   * @return {Object} Request object.
   */
  function createTransactionRequest(node, baseObj, version, options) {
    const featurePrefix = options.featurePrefix
      ? options.featurePrefix
      : FEATURE_PREFIX;
    let gmlVersion;
    if (version === '1.0.0') {
      gmlVersion = 2;
    } else if (version === '1.1.0') {
      gmlVersion = 3;
    } else if (version === '2.0.0') {
      gmlVersion = 3.2;
    }
    const obj = Object.assign(
      {node},
      {
        version,
        'featureNS': options.featureNS,
        'featureType': options.featureType,
        'featurePrefix': featurePrefix,
        'gmlVersion': gmlVersion,
        'hasZ': options.hasZ,
        'srsName': options.srsName,
      },
      baseObj
    );
    return obj;
  }

  /**
   * @param {string} type Request type.
   * @param {Array<import("../Feature.js").default>} features Features.
   * @param {Array<*>} objectStack Object stack.
   * @param {Element} request Transaction Request.
   */
  function serializeTransactionRequest(type, features, objectStack, request) {
    pushSerializeAndPop(
      request,
      TRANSACTION_SERIALIZERS,
      makeSimpleNodeFactory(type),
      features,
      objectStack
    );
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} Transaction Summary.
   */
  function readTransactionSummary(node, objectStack) {
    return pushParseAndPop({}, TRANSACTION_SUMMARY_PARSERS, node, objectStack);
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  const OGC_FID_PARSERS = {
    'http://www.opengis.net/ogc': {
      'FeatureId': makeArrayPusher(function (node, objectStack) {
        return node.getAttribute('fid');
      }),
    },
    'http://www.opengis.net/ogc/1.1': {
      'FeatureId': makeArrayPusher(function (node, objectStack) {
        return node.getAttribute('fid');
      }),
    },
  };

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   */
  function fidParser(node, objectStack) {
    parseNode(OGC_FID_PARSERS, node, objectStack);
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  const INSERT_RESULTS_PARSERS = {
    'http://www.opengis.net/wfs': {
      'Feature': fidParser,
    },
    'http://www.opengis.net/wfs/2.0': {
      'Feature': fidParser,
    },
  };

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Array<string>|undefined} Insert results.
   */
  function readInsertResults(node, objectStack) {
    return pushParseAndPop([], INSERT_RESULTS_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {import("../Feature.js").default} feature Feature.
   * @param {Array<*>} objectStack Node stack.
   */
  function writeFeature(node, feature, objectStack) {
    const context = objectStack[objectStack.length - 1];
    const featureType = context['featureType'];
    const featureNS = context['featureNS'];
    const gmlVersion = context['gmlVersion'];
    const child = createElementNS(featureNS, featureType);
    node.appendChild(child);
    if (gmlVersion === 2) {
      GML2.prototype.writeFeatureElement(child, feature, objectStack);
    } else if (gmlVersion === 3) {
      GML3.prototype.writeFeatureElement(child, feature, objectStack);
    } else {
      GML32.prototype.writeFeatureElement(child, feature, objectStack);
    }
  }

  /**
   * @param {Node} node Node.
   * @param {number|string} fid Feature identifier.
   * @param {Array<*>} objectStack Node stack.
   */
  function writeOgcFidFilter(node, fid, objectStack) {
    const context = objectStack[objectStack.length - 1];
    const version = context['version'];
    const ns = OGCNS[version];
    const filter = createElementNS(ns, 'Filter');
    const child = createElementNS(ns, 'FeatureId');
    filter.appendChild(child);
    child.setAttribute('fid', /** @type {string} */ (fid));
    node.appendChild(filter);
  }

  /**
   * @param {string|undefined} featurePrefix The prefix of the feature.
   * @param {string} featureType The type of the feature.
   * @return {string} The value of the typeName property.
   */
  function getTypeName(featurePrefix, featureType) {
    featurePrefix = featurePrefix ? featurePrefix : FEATURE_PREFIX;
    const prefix = featurePrefix + ':';
    // The featureType already contains the prefix.
    if (featureType.startsWith(prefix)) {
      return featureType;
    } else {
      return prefix + featureType;
    }
  }

  /**
   * @param {Element} node Node.
   * @param {import("../Feature.js").default} feature Feature.
   * @param {Array<*>} objectStack Node stack.
   */
  function writeDelete(node, feature, objectStack) {
    const context = objectStack[objectStack.length - 1];
    assert(feature.getId() !== undefined, 26); // Features must have an id set
    const featureType = context['featureType'];
    const featurePrefix = context['featurePrefix'];
    const featureNS = context['featureNS'];
    const typeName = getTypeName(featurePrefix, featureType);
    node.setAttribute('typeName', typeName);
    node.setAttributeNS(XMLNS, 'xmlns:' + featurePrefix, featureNS);
    const fid = feature.getId();
    if (fid !== undefined) {
      writeOgcFidFilter(node, fid, objectStack);
    }
  }

  /**
   * @param {Element} node Node.
   * @param {import("../Feature.js").default} feature Feature.
   * @param {Array<*>} objectStack Node stack.
   */
  function writeUpdate(node, feature, objectStack) {
    const context = objectStack[objectStack.length - 1];
    assert(feature.getId() !== undefined, 27); // Features must have an id set
    const version = context['version'];
    const featureType = context['featureType'];
    const featurePrefix = context['featurePrefix'];
    const featureNS = context['featureNS'];
    const typeName = getTypeName(featurePrefix, featureType);
    const geometryName = feature.getGeometryName();
    node.setAttribute('typeName', typeName);
    node.setAttributeNS(XMLNS, 'xmlns:' + featurePrefix, featureNS);
    const fid = feature.getId();
    if (fid !== undefined) {
      const keys = feature.getKeys();
      const values = [];
      for (let i = 0, ii = keys.length; i < ii; i++) {
        const value = feature.get(keys[i]);
        if (value !== undefined) {
          let name = keys[i];
          if (
            value &&
            typeof (/** @type {?} */ (value).getSimplifiedGeometry) === 'function'
          ) {
            name = geometryName;
          }
          values.push({name: name, value: value});
        }
      }
      pushSerializeAndPop(
        /** @type {import("../xml.js").NodeStackItem} */ ({
          version,
          'gmlVersion': context['gmlVersion'],
          node,
          'hasZ': context['hasZ'],
          'srsName': context['srsName'],
        }),
        TRANSACTION_SERIALIZERS,
        makeSimpleNodeFactory('Property'),
        values,
        objectStack
      );
      writeOgcFidFilter(node, fid, objectStack);
    }
  }

  /**
   * @param {Node} node Node.
   * @param {Object} pair Property name and value.
   * @param {Array<*>} objectStack Node stack.
   */
  function writeProperty(node, pair, objectStack) {
    const context = objectStack[objectStack.length - 1];
    const version = context['version'];
    const ns = WFSNS[version];
    const name = createElementNS(ns, 'Name');
    const gmlVersion = context['gmlVersion'];
    node.appendChild(name);
    writeStringTextNode(name, pair.name);
    if (pair.value !== undefined && pair.value !== null) {
      const value = createElementNS(ns, 'Value');
      node.appendChild(value);
      if (
        pair.value &&
        typeof (/** @type {?} */ (pair.value).getSimplifiedGeometry) ===
          'function'
      ) {
        if (gmlVersion === 2) {
          GML2.prototype.writeGeometryElement(value, pair.value, objectStack);
        } else if (gmlVersion === 3) {
          GML3.prototype.writeGeometryElement(value, pair.value, objectStack);
        } else {
          GML32.prototype.writeGeometryElement(value, pair.value, objectStack);
        }
      } else {
        writeStringTextNode(value, pair.value);
      }
    }
  }

  /**
   * @param {Element} node Node.
   * @param {{vendorId: string, safeToIgnore: boolean, value: string}} nativeElement The native element.
   * @param {Array<*>} objectStack Node stack.
   */
  function writeNative(node, nativeElement, objectStack) {
    if (nativeElement.vendorId) {
      node.setAttribute('vendorId', nativeElement.vendorId);
    }
    if (nativeElement.safeToIgnore !== undefined) {
      node.setAttribute('safeToIgnore', String(nativeElement.safeToIgnore));
    }
    if (nativeElement.value !== undefined) {
      writeStringTextNode(node, nativeElement.value);
    }
  }

  /**
   * @type {Object<string, Object<string, import("../xml.js").Serializer>>}
   */
  const GETFEATURE_SERIALIZERS = {
    'http://www.opengis.net/wfs': {
      'Query': makeChildAppender(writeQuery),
    },
    'http://www.opengis.net/wfs/2.0': {
      'Query': makeChildAppender(writeQuery),
    },
    'http://www.opengis.net/ogc': {
      'During': makeChildAppender(writeDuringFilter),
      'And': makeChildAppender(writeLogicalFilter),
      'Or': makeChildAppender(writeLogicalFilter),
      'Not': makeChildAppender(writeNotFilter),
      'BBOX': makeChildAppender(writeBboxFilter),
      'Contains': makeChildAppender(writeSpatialFilter),
      'Intersects': makeChildAppender(writeSpatialFilter),
      'Within': makeChildAppender(writeSpatialFilter),
      'DWithin': makeChildAppender(writeDWithinFilter),
      'PropertyIsEqualTo': makeChildAppender(writeComparisonFilter),
      'PropertyIsNotEqualTo': makeChildAppender(writeComparisonFilter),
      'PropertyIsLessThan': makeChildAppender(writeComparisonFilter),
      'PropertyIsLessThanOrEqualTo': makeChildAppender(writeComparisonFilter),
      'PropertyIsGreaterThan': makeChildAppender(writeComparisonFilter),
      'PropertyIsGreaterThanOrEqualTo': makeChildAppender(writeComparisonFilter),
      'PropertyIsNull': makeChildAppender(writeIsNullFilter),
      'PropertyIsBetween': makeChildAppender(writeIsBetweenFilter),
      'PropertyIsLike': makeChildAppender(writeIsLikeFilter),
    },
    'http://www.opengis.net/fes/2.0': {
      'During': makeChildAppender(writeDuringFilter),
      'And': makeChildAppender(writeLogicalFilter),
      'Or': makeChildAppender(writeLogicalFilter),
      'Not': makeChildAppender(writeNotFilter),
      'BBOX': makeChildAppender(writeBboxFilter),
      'Contains': makeChildAppender(writeSpatialFilter),
      'Disjoint': makeChildAppender(writeSpatialFilter),
      'Intersects': makeChildAppender(writeSpatialFilter),
      'ResourceId': makeChildAppender(writeResourceIdFilter),
      'Within': makeChildAppender(writeSpatialFilter),
      'DWithin': makeChildAppender(writeDWithinFilter),
      'PropertyIsEqualTo': makeChildAppender(writeComparisonFilter),
      'PropertyIsNotEqualTo': makeChildAppender(writeComparisonFilter),
      'PropertyIsLessThan': makeChildAppender(writeComparisonFilter),
      'PropertyIsLessThanOrEqualTo': makeChildAppender(writeComparisonFilter),
      'PropertyIsGreaterThan': makeChildAppender(writeComparisonFilter),
      'PropertyIsGreaterThanOrEqualTo': makeChildAppender(writeComparisonFilter),
      'PropertyIsNull': makeChildAppender(writeIsNullFilter),
      'PropertyIsBetween': makeChildAppender(writeIsBetweenFilter),
      'PropertyIsLike': makeChildAppender(writeIsLikeFilter),
    },
  };

  /**
   * @param {Element} node Node.
   * @param {string} featureType Feature type.
   * @param {Array<*>} objectStack Node stack.
   */
  function writeQuery(node, featureType, objectStack) {
    const context = /** @type {Object} */ (objectStack[objectStack.length - 1]);
    const version = context['version'];
    const featurePrefix = context['featurePrefix'];
    const featureNS = context['featureNS'];
    const propertyNames = context['propertyNames'];
    const srsName = context['srsName'];
    let typeName;
    // If feature prefix is not defined, we must not use the default prefix.
    if (featurePrefix) {
      typeName = getTypeName(featurePrefix, featureType);
    } else {
      typeName = featureType;
    }
    let typeNameAttr;
    if (version === '2.0.0') {
      typeNameAttr = 'typeNames';
    } else {
      typeNameAttr = 'typeName';
    }
    node.setAttribute(typeNameAttr, typeName);
    if (srsName) {
      node.setAttribute('srsName', srsName);
    }
    if (featureNS) {
      node.setAttributeNS(XMLNS, 'xmlns:' + featurePrefix, featureNS);
    }
    const item = /** @type {import("../xml.js").NodeStackItem} */ (
      Object.assign({}, context)
    );
    item.node = node;
    pushSerializeAndPop(
      item,
      QUERY_SERIALIZERS,
      makeSimpleNodeFactory('PropertyName'),
      propertyNames,
      objectStack
    );
    const filter = context['filter'];
    if (filter) {
      const child = createElementNS(getFilterNS(version), 'Filter');
      node.appendChild(child);
      writeFilterCondition(child, filter, objectStack);
    }
  }

  /**
   * @param {Element} node Node.
   * @param {import("./filter/Filter.js").default} filter Filter.
   * @param {Array<*>} objectStack Node stack.
   */
  function writeFilterCondition(node, filter, objectStack) {
    const context = /** @type {Object} */ (objectStack[objectStack.length - 1]);
    /** @type {import("../xml.js").NodeStackItem} */
    const item = {node};
    Object.assign(item, {context});
    pushSerializeAndPop(
      item,
      GETFEATURE_SERIALIZERS,
      makeSimpleNodeFactory(filter.getTagName()),
      [filter],
      objectStack
    );
  }

  /**
   * @param {Node} node Node.
   * @param {import("./filter/Bbox.js").default} filter Filter.
   * @param {Array<*>} objectStack Node stack.
   */
  function writeBboxFilter(node, filter, objectStack) {
    const parent = /** @type {Object} */ (objectStack[objectStack.length - 1]);
    const context = parent['context'];
    const version = context['version'];
    parent['srsName'] = filter.srsName;
    const format = GML_FORMATS[version];

    writePropertyName(version, node, filter.geometryName);
    format.prototype.writeGeometryElement(node, filter.extent, objectStack);
  }

  /**
   * @param {Element} node Element.
   * @param {import("./filter/ResourceId.js").default} filter Filter.
   * @param {Array<*>} objectStack Node stack.
   */
  function writeResourceIdFilter(node, filter, objectStack) {
    node.setAttribute('rid', /** @type {string} */ (filter.rid));
  }

  /**
   * @param {Node} node Node.
   * @param {import("./filter/Spatial.js").default} filter Filter.
   * @param {Array<*>} objectStack Node stack.
   */
  function writeSpatialFilter(node, filter, objectStack) {
    const parent = /** @type {Object} */ (objectStack[objectStack.length - 1]);
    const context = parent['context'];
    const version = context['version'];
    parent['srsName'] = filter.srsName;
    const format = GML_FORMATS[version];

    writePropertyName(version, node, filter.geometryName);
    format.prototype.writeGeometryElement(node, filter.geometry, objectStack);
  }

  /**
   * @param {Node} node Node.
   * @param {import("./filter/DWithin.js").default} filter Filter.
   * @param {Array<*>} objectStack Node stack.
   */
  function writeDWithinFilter(node, filter, objectStack) {
    const parent = /** @type {Object} */ (objectStack[objectStack.length - 1]);
    const context = parent['context'];
    const version = context['version'];
    writeSpatialFilter(node, filter, objectStack);
    const distance = createElementNS(getFilterNS(version), 'Distance');
    writeStringTextNode(distance, filter.distance.toString());
    if (version === '2.0.0') {
      distance.setAttribute('uom', filter.unit);
    } else {
      distance.setAttribute('units', filter.unit);
    }
    node.appendChild(distance);
  }

  /**
   * @param {Node} node Node.
   * @param {import("./filter/During.js").default} filter Filter.
   * @param {Array<*>} objectStack Node stack.
   */
  function writeDuringFilter(node, filter, objectStack) {
    const parent = /** @type {Object} */ (objectStack[objectStack.length - 1]);
    const context = parent['context'];
    const version = context['version'];

    writeExpression(FESNS[version], 'ValueReference', node, filter.propertyName);
    const timePeriod = createElementNS(GMLNS, 'TimePeriod');

    node.appendChild(timePeriod);

    const begin = createElementNS(GMLNS, 'begin');
    timePeriod.appendChild(begin);
    writeTimeInstant(begin, filter.begin);

    const end = createElementNS(GMLNS, 'end');
    timePeriod.appendChild(end);
    writeTimeInstant(end, filter.end);
  }

  /**
   * @param {Element} node Node.
   * @param {import("./filter/LogicalNary.js").default} filter Filter.
   * @param {Array<*>} objectStack Node stack.
   */
  function writeLogicalFilter(node, filter, objectStack) {
    const parent = /** @type {Object} */ (objectStack[objectStack.length - 1]);
    const context = parent['context'];
    /** @type {import("../xml.js").NodeStackItem} */
    const item = {node};
    Object.assign(item, {context});
    const conditions = filter.conditions;
    for (let i = 0, ii = conditions.length; i < ii; ++i) {
      const condition = conditions[i];
      pushSerializeAndPop(
        item,
        GETFEATURE_SERIALIZERS,
        makeSimpleNodeFactory(condition.getTagName()),
        [condition],
        objectStack
      );
    }
  }

  /**
   * @param {Element} node Node.
   * @param {import("./filter/Not.js").default} filter Filter.
   * @param {Array<*>} objectStack Node stack.
   */
  function writeNotFilter(node, filter, objectStack) {
    const parent = /** @type {Object} */ (objectStack[objectStack.length - 1]);
    const context = parent['context'];
    /** @type {import("../xml.js").NodeStackItem} */
    const item = {node};
    Object.assign(item, {context});
    const condition = filter.condition;
    pushSerializeAndPop(
      item,
      GETFEATURE_SERIALIZERS,
      makeSimpleNodeFactory(condition.getTagName()),
      [condition],
      objectStack
    );
  }

  /**
   * @param {Element} node Node.
   * @param {import("./filter/ComparisonBinary.js").default} filter Filter.
   * @param {Array<*>} objectStack Node stack.
   */
  function writeComparisonFilter(node, filter, objectStack) {
    const parent = /** @type {Object} */ (objectStack[objectStack.length - 1]);
    const context = parent['context'];
    const version = context['version'];
    if (filter.matchCase !== undefined) {
      node.setAttribute('matchCase', filter.matchCase.toString());
    }
    writePropertyName(version, node, filter.propertyName);
    writeLiteral(version, node, '' + filter.expression);
  }

  /**
   * @param {Node} node Node.
   * @param {import("./filter/IsNull.js").default} filter Filter.
   * @param {Array<*>} objectStack Node stack.
   */
  function writeIsNullFilter(node, filter, objectStack) {
    const parent = /** @type {Object} */ (objectStack[objectStack.length - 1]);
    const context = parent['context'];
    const version = context['version'];
    writePropertyName(version, node, filter.propertyName);
  }

  /**
   * @param {Node} node Node.
   * @param {import("./filter/IsBetween.js").default} filter Filter.
   * @param {Array<*>} objectStack Node stack.
   */
  function writeIsBetweenFilter(node, filter, objectStack) {
    const parent = /** @type {Object} */ (objectStack[objectStack.length - 1]);
    const context = parent['context'];
    const version = context['version'];
    const ns = getFilterNS(version);

    writePropertyName(version, node, filter.propertyName);

    const lowerBoundary = createElementNS(ns, 'LowerBoundary');
    node.appendChild(lowerBoundary);
    writeLiteral(version, lowerBoundary, '' + filter.lowerBoundary);

    const upperBoundary = createElementNS(ns, 'UpperBoundary');
    node.appendChild(upperBoundary);
    writeLiteral(version, upperBoundary, '' + filter.upperBoundary);
  }

  /**
   * @param {Element} node Node.
   * @param {import("./filter/IsLike.js").default} filter Filter.
   * @param {Array<*>} objectStack Node stack.
   */
  function writeIsLikeFilter(node, filter, objectStack) {
    const parent = /** @type {Object} */ (objectStack[objectStack.length - 1]);
    const context = parent['context'];
    const version = context['version'];
    node.setAttribute('wildCard', filter.wildCard);
    node.setAttribute('singleChar', filter.singleChar);
    node.setAttribute('escapeChar', filter.escapeChar);
    if (filter.matchCase !== undefined) {
      node.setAttribute('matchCase', filter.matchCase.toString());
    }
    writePropertyName(version, node, filter.propertyName);
    writeLiteral(version, node, '' + filter.pattern);
  }

  /**
   * @param {string} ns Namespace.
   * @param {string} tagName Tag name.
   * @param {Node} node Node.
   * @param {string} value Value.
   */
  function writeExpression(ns, tagName, node, value) {
    const property = createElementNS(ns, tagName);
    writeStringTextNode(property, value);
    node.appendChild(property);
  }

  /**
   * @param {string} version Version.
   * @param {Node} node Node.
   * @param {string} value PropertyName value.
   */
  function writeLiteral(version, node, value) {
    writeExpression(getFilterNS(version), 'Literal', node, value);
  }

  /**
   * @param {string} version Version.
   * @param {Node} node Node.
   * @param {string} value PropertyName value.
   */
  function writePropertyName(version, node, value) {
    if (version === '2.0.0') {
      writeExpression(FESNS[version], 'ValueReference', node, value);
    } else {
      writeExpression(OGCNS[version], 'PropertyName', node, value);
    }
  }

  /**
   * @param {Node} node Node.
   * @param {string} time PropertyName value.
   */
  function writeTimeInstant(node, time) {
    const timeInstant = createElementNS(GMLNS, 'TimeInstant');
    node.appendChild(timeInstant);

    const timePosition = createElementNS(GMLNS, 'timePosition');
    timeInstant.appendChild(timePosition);
    writeStringTextNode(timePosition, time);
  }

  /**
   * Encode filter as WFS `Filter` and return the Node.
   *
   * @param {import("./filter/Filter.js").default} filter Filter.
   * @param {string} version WFS version. If not provided defaults to '1.1.0'
   * @return {Node} Result.
   * @api
   */
  function writeFilter(filter, version) {
    version = version || '1.1.0';
    const child = createElementNS(getFilterNS(version), 'Filter');
    const context = {
      node: child,
    };
    Object.assign(context, {
      'version': version,
      'filter': filter,
    });
    writeFilterCondition(child, filter, [context]);
    return child;
  }

  /**
   * @param {Element} node Node.
   * @param {Array<string>} featureTypes Feature types.
   * @param {Array<*>} objectStack Node stack.
   */
  function writeGetFeature(node, featureTypes, objectStack) {
    const context = /** @type {Object} */ (objectStack[objectStack.length - 1]);
    const item = /** @type {import("../xml.js").NodeStackItem} */ (
      Object.assign({}, context)
    );
    item.node = node;
    pushSerializeAndPop(
      item,
      GETFEATURE_SERIALIZERS,
      makeSimpleNodeFactory('Query'),
      featureTypes,
      objectStack
    );
  }

  function getFilterNS(version) {
    let ns;
    if (version === '2.0.0') {
      ns = FESNS[version];
    } else {
      ns = OGCNS[version];
    }
    return ns;
  }

  /**
   * @module ol/format/WKB
   */

  // WKB spec: https://www.ogc.org/standards/sfa
  // EWKB spec: https://raw.githubusercontent.com/postgis/postgis/2.1.0/doc/ZMSgeoms.txt

  /**
   * @const
   * @enum {number}
   */
  const WKBGeometryType = {
    POINT: 1,
    LINE_STRING: 2,
    POLYGON: 3,
    MULTI_POINT: 4,
    MULTI_LINE_STRING: 5,
    MULTI_POLYGON: 6,
    GEOMETRY_COLLECTION: 7,

    /*
    CIRCULAR_STRING: 8,
    COMPOUND_CURVE: 9,
    CURVE_POLYGON: 10,

    MULTI_CURVE: 11,
    MULTI_SURFACE: 12,
    CURVE: 13,
    SURFACE: 14,
    */

    POLYHEDRAL_SURFACE: 15,
    TIN: 16,
    TRIANGLE: 17,
  };

  class WkbReader {
    /**
     * @param {DataView} view source to read
     */
    constructor(view) {
      /** @private */
      this.view_ = view;

      /**
       * @type {number}
       * @private
       */
      this.pos_ = 0;

      /**
       * @type {boolean}
       * @private
       */
      this.initialized_ = false;

      /**
       * @type {boolean}
       * @private
       */
      this.isLittleEndian_ = false;

      /**
       * @type {boolean}
       * @private
       */
      this.hasZ_ = false;

      /**
       * @type {boolean}
       * @private
       */
      this.hasM_ = false;

      /**
       * @type {number|null}
       * @private
       */
      this.srid_ = null;

      /**
       * @type {import("../geom/Geometry.js").GeometryLayout}
       * @private
       */
      this.layout_ = 'XY';
    }

    /**
     * @return {number} value
     */
    readUint8() {
      return this.view_.getUint8(this.pos_++);
    }

    /**
     * @param {boolean} [isLittleEndian] Whether read value as little endian
     * @return {number} value
     */
    readUint32(isLittleEndian) {
      return this.view_.getUint32(
        (this.pos_ += 4) - 4,
        isLittleEndian !== undefined ? isLittleEndian : this.isLittleEndian_
      );
    }

    /**
     * @param {boolean} [isLittleEndian] Whether read value as little endian
     * @return {number} value
     */
    readDouble(isLittleEndian) {
      return this.view_.getFloat64(
        (this.pos_ += 8) - 8,
        isLittleEndian !== undefined ? isLittleEndian : this.isLittleEndian_
      );
    }

    /**
     * @return {import('../coordinate.js').Coordinate} coords for Point
     */
    readPoint() {
      /** @type import('../coordinate.js').Coordinate */
      const coords = [];

      coords.push(this.readDouble());
      coords.push(this.readDouble());
      if (this.hasZ_) {
        coords.push(this.readDouble());
      }
      if (this.hasM_) {
        coords.push(this.readDouble());
      }

      return coords;
    }

    /**
     * @return {Array<import('../coordinate.js').Coordinate>} coords for LineString / LinearRing
     */
    readLineString() {
      const numPoints = this.readUint32();

      /** @type Array<import('../coordinate.js').Coordinate> */
      const coords = [];
      for (let i = 0; i < numPoints; i++) {
        coords.push(this.readPoint());
      }

      return coords;
    }

    /**
     * @return {Array<Array<import('../coordinate.js').Coordinate>>} coords for Polygon like
     */
    readPolygon() {
      const numRings = this.readUint32();

      /** @type Array<Array<import('../coordinate.js').Coordinate>> */
      const rings = [];
      for (let i = 0; i < numRings; i++) {
        rings.push(this.readLineString()); // as a LinearRing
      }

      return rings;
    }

    /**
     * @param {number} [expectedTypeId] Expected WKB Type ID
     * @return {number} WKB Type ID
     */
    readWkbHeader(expectedTypeId) {
      const byteOrder = this.readUint8();
      const isLittleEndian = byteOrder > 0;

      const wkbType = this.readUint32(isLittleEndian);
      const wkbTypeThousandth = Math.floor((wkbType & 0x0fffffff) / 1000);
      const hasZ =
        Boolean(wkbType & 0x80000000) ||
        wkbTypeThousandth === 1 ||
        wkbTypeThousandth === 3;
      const hasM =
        Boolean(wkbType & 0x40000000) ||
        wkbTypeThousandth === 2 ||
        wkbTypeThousandth === 3;
      const hasSRID = Boolean(wkbType & 0x20000000);
      const typeId = (wkbType & 0x0fffffff) % 1000; // Assume 1000 is an upper limit for type ID
      const layout = /** @type {import("../geom/Geometry.js").GeometryLayout} */ (
        ['XY', hasZ ? 'Z' : '', hasM ? 'M' : ''].join('')
      );

      const srid = hasSRID ? this.readUint32(isLittleEndian) : null;

      if (expectedTypeId !== undefined && expectedTypeId !== typeId) {
        throw new Error('Unexpected WKB geometry type ' + typeId);
      }

      if (this.initialized_) {
        // sanity checks
        if (this.isLittleEndian_ !== isLittleEndian) {
          throw new Error('Inconsistent endian');
        }
        if (this.layout_ !== layout) {
          throw new Error('Inconsistent geometry layout');
        }
        if (srid && this.srid_ !== srid) {
          throw new Error('Inconsistent coordinate system (SRID)');
        }
      } else {
        this.isLittleEndian_ = isLittleEndian;
        this.hasZ_ = hasZ;
        this.hasM_ = hasM;
        this.layout_ = layout;
        this.srid_ = srid;
        this.initialized_ = true;
      }

      return typeId;
    }

    /**
     * @param {number} typeId WKB Type ID
     * @return {any} values read
     */
    readWkbPayload(typeId) {
      switch (typeId) {
        case WKBGeometryType.POINT:
          return this.readPoint();

        case WKBGeometryType.LINE_STRING:
          return this.readLineString();

        case WKBGeometryType.POLYGON:
        case WKBGeometryType.TRIANGLE:
          return this.readPolygon();

        case WKBGeometryType.MULTI_POINT:
          return this.readMultiPoint();

        case WKBGeometryType.MULTI_LINE_STRING:
          return this.readMultiLineString();

        case WKBGeometryType.MULTI_POLYGON:
        case WKBGeometryType.POLYHEDRAL_SURFACE:
        case WKBGeometryType.TIN:
          return this.readMultiPolygon();

        case WKBGeometryType.GEOMETRY_COLLECTION:
          return this.readGeometryCollection();

        default:
          throw new Error(
            'Unsupported WKB geometry type ' + typeId + ' is found'
          );
      }
    }

    /**
     * @param {number} expectedTypeId Expected WKB Type ID
     * @return {any} values read
     */
    readWkbBlock(expectedTypeId) {
      return this.readWkbPayload(this.readWkbHeader(expectedTypeId));
    }

    /**
     * @param {Function} reader reader function for each item
     * @param {number} [expectedTypeId] Expected WKB Type ID
     * @return {any} values read
     */
    readWkbCollection(reader, expectedTypeId) {
      const num = this.readUint32();

      const items = [];
      for (let i = 0; i < num; i++) {
        const result = reader.call(this, expectedTypeId);
        if (result) {
          items.push(result);
        }
      }

      return items;
    }

    /**
     * @return {Array<import('../coordinate.js').Coordinate>} coords for MultiPoint
     */
    readMultiPoint() {
      return this.readWkbCollection(this.readWkbBlock, WKBGeometryType.POINT);
    }

    /**
     * @return {Array<Array<import('../coordinate.js').Coordinate>>} coords for MultiLineString like
     */
    readMultiLineString() {
      return this.readWkbCollection(
        this.readWkbBlock,
        WKBGeometryType.LINE_STRING
      );
    }

    /**
     * @return {Array<Array<Array<import('../coordinate.js').Coordinate>>>} coords for MultiPolygon like
     */
    readMultiPolygon() {
      return this.readWkbCollection(this.readWkbBlock, WKBGeometryType.POLYGON);
    }

    /**
     * @return {Array<import('../geom/Geometry.js').default>} array of geometries
     */
    readGeometryCollection() {
      return this.readWkbCollection(this.readGeometry);
    }

    /**
     * @return {import('../geom/Geometry.js').default} geometry
     */
    readGeometry() {
      const typeId = this.readWkbHeader();
      const result = this.readWkbPayload(typeId);

      switch (typeId) {
        case WKBGeometryType.POINT:
          return new Point(
            /** @type {import('../coordinate.js').Coordinate} */ (result),
            this.layout_
          );

        case WKBGeometryType.LINE_STRING:
          return new LineString(
            /** @type {Array<import('../coordinate.js').Coordinate>} */ (result),
            this.layout_
          );

        case WKBGeometryType.POLYGON:
        case WKBGeometryType.TRIANGLE:
          return new Polygon(
            /** @type {Array<Array<import('../coordinate.js').Coordinate>>} */ (
              result
            ),
            this.layout_
          );

        case WKBGeometryType.MULTI_POINT:
          return new MultiPoint(
            /** @type {Array<import('../coordinate.js').Coordinate>} */ (result),
            this.layout_
          );

        case WKBGeometryType.MULTI_LINE_STRING:
          return new MultiLineString(
            /** @type {Array<Array<import('../coordinate.js').Coordinate>>} */ (
              result
            ),
            this.layout_
          );

        case WKBGeometryType.MULTI_POLYGON:
        case WKBGeometryType.POLYHEDRAL_SURFACE:
        case WKBGeometryType.TIN:
          return new MultiPolygon(
            /** @type {Array<Array<Array<import('../coordinate.js').Coordinate>>>} */ (
              result
            ),
            this.layout_
          );

        case WKBGeometryType.GEOMETRY_COLLECTION:
          return new GeometryCollection(
            /** @type {Array<import('../geom/Geometry.js').default>} */ (result)
          );

        default:
          return null;
      }
    }

    /**
     * @return {number|null} SRID in the EWKB. `null` if not defined.
     */
    getSrid() {
      return this.srid_;
    }
  }

  class WkbWriter {
    /**
     * @type {Object}
     * @property {string} [layout] geometryLayout
     * @property {boolean} [littleEndian=true] littleEndian
     * @property {boolean} [ewkb=true] Whether writes in EWKB format
     * @property {Object} [nodata] NoData value for each axes
     * @param {Object} opts options
     */
    constructor(opts) {
      opts = opts || {};

      /** @type {string} */
      this.layout_ = opts.layout;
      this.isLittleEndian_ = opts.littleEndian !== false;

      this.isEWKB_ = opts.ewkb !== false;

      /** @type {Array<Array<number>>} */
      this.writeQueue_ = [];

      /**
       * @type {Object}
       * @property {number} X NoData value for X
       * @property {number} Y NoData value for Y
       * @property {number} Z NoData value for Z
       * @property {number} M NoData value for M
       */
      this.nodata_ = Object.assign({X: 0, Y: 0, Z: 0, M: 0}, opts.nodata);
    }

    /**
     * @param {number} value value
     */
    writeUint8(value) {
      this.writeQueue_.push([1, value]);
    }

    /**
     * @param {number} value value
     */
    writeUint32(value) {
      this.writeQueue_.push([4, value]);
    }

    /**
     * @param {number} value value
     */
    writeDouble(value) {
      this.writeQueue_.push([8, value]);
    }

    /**
     * @param {import('../coordinate.js').Coordinate} coords coords
     * @param {import("../geom/Geometry.js").GeometryLayout} layout layout
     */
    writePoint(coords, layout) {
      /**
       * @type {Object}
       * @property {number} X NoData value for X
       * @property {number} Y NoData value for Y
       * @property {number} [Z] NoData value for Z
       * @property {number} [M] NoData value for M
       */
      const coordsObj = Object.assign.apply(
        null,
        layout.split('').map((axis, idx) => ({[axis]: coords[idx]}))
      );

      for (const axis of this.layout_) {
        this.writeDouble(
          axis in coordsObj ? coordsObj[axis] : this.nodata_[axis]
        );
      }
    }

    /**
     * @param {Array<import('../coordinate.js').Coordinate>} coords coords
     * @param {import("../geom/Geometry.js").GeometryLayout} layout layout
     */
    writeLineString(coords, layout) {
      this.writeUint32(coords.length); // numPoints
      for (let i = 0; i < coords.length; i++) {
        this.writePoint(coords[i], layout);
      }
    }

    /**
     * @param {Array<Array<import('../coordinate.js').Coordinate>>} rings rings
     * @param {import("../geom/Geometry.js").GeometryLayout} layout layout
     */
    writePolygon(rings, layout) {
      this.writeUint32(rings.length); // numRings
      for (let i = 0; i < rings.length; i++) {
        this.writeLineString(rings[i], layout); // as a LinearRing
      }
    }

    /**
     * @param {number} wkbType WKB Type ID
     * @param {number} [srid] SRID
     */
    writeWkbHeader(wkbType, srid) {
      wkbType %= 1000; // Assume 1000 is an upper limit for type ID
      if (this.layout_.includes('Z')) {
        wkbType += this.isEWKB_ ? 0x80000000 : 1000;
      }
      if (this.layout_.includes('M')) {
        wkbType += this.isEWKB_ ? 0x40000000 : 2000;
      }
      if (this.isEWKB_ && Number.isInteger(srid)) {
        wkbType |= 0x20000000;
      }

      this.writeUint8(this.isLittleEndian_ ? 1 : 0);
      this.writeUint32(wkbType);
      if (this.isEWKB_ && Number.isInteger(srid)) {
        this.writeUint32(srid);
      }
    }

    /**
     * @param {Array<import('../coordinate.js').Coordinate>} coords coords
     * @param {import("../geom/Geometry.js").GeometryLayout} layout layout
     */
    writeMultiPoint(coords, layout) {
      this.writeUint32(coords.length); // numItems
      for (let i = 0; i < coords.length; i++) {
        this.writeWkbHeader(1);
        this.writePoint(coords[i], layout);
      }
    }

    /**
     * @param {Array<Array<import('../coordinate.js').Coordinate>>} coords coords
     * @param {import("../geom/Geometry.js").GeometryLayout} layout layout
     */
    writeMultiLineString(coords, layout) {
      this.writeUint32(coords.length); // numItems
      for (let i = 0; i < coords.length; i++) {
        this.writeWkbHeader(2);
        this.writeLineString(coords[i], layout);
      }
    }

    /**
     * @param {Array<Array<Array<import('../coordinate.js').Coordinate>>>} coords coords
     * @param {import("../geom/Geometry.js").GeometryLayout} layout layout
     */
    writeMultiPolygon(coords, layout) {
      this.writeUint32(coords.length); // numItems
      for (let i = 0; i < coords.length; i++) {
        this.writeWkbHeader(3);
        this.writePolygon(coords[i], layout);
      }
    }

    /**
     * @param {Array<import('../geom/Geometry.js').default>} geometries geometries
     */
    writeGeometryCollection(geometries) {
      this.writeUint32(geometries.length); // numItems

      for (let i = 0; i < geometries.length; i++) {
        this.writeGeometry(geometries[i]);
      }
    }

    /**
     * @param {import("../geom/Geometry.js").default} geom geometry
     * @param {import("../geom/Geometry.js").GeometryLayout} [layout] layout
     * @return {import("../geom/Geometry.js").GeometryLayout} minumum layout made by common axes
     */
    findMinimumLayout(geom, layout = 'XYZM') {
      /**
       * @param {import("../geom/Geometry.js").GeometryLayout} a A
       * @param {import("../geom/Geometry.js").GeometryLayout} b B
       * @return {import("../geom/Geometry.js").GeometryLayout} minumum layout made by common axes
       */
      const GeometryLayout_min = (a, b) => {
        if (a === b) {
          return a;
        }

        if (a === 'XYZM') {
          // anything `b` is minimum
          return b;
        }
        if (b === 'XYZM') {
          // anything `a` is minimum
          return a;
        }

        // otherwise, incompatible
        return 'XY';
      };

      if (geom instanceof SimpleGeometry) {
        return GeometryLayout_min(geom.getLayout(), layout);
      }

      if (geom instanceof GeometryCollection) {
        const geoms = geom.getGeometriesArray();
        for (let i = 0; i < geoms.length && layout !== 'XY'; i++) {
          layout = this.findMinimumLayout(geoms[i], layout);
        }
      }

      return layout;
    }

    /**
     * @param {import("../geom/Geometry.js").default} geom geometry
     * @param {number} [srid] SRID
     */
    writeGeometry(geom, srid) {
      /**
       * @type {Object<import("../geom/Geometry.js").Type, WKBGeometryType>}
       */
      const wkblut = {
        Point: WKBGeometryType.POINT,
        LineString: WKBGeometryType.LINE_STRING,
        Polygon: WKBGeometryType.POLYGON,
        MultiPoint: WKBGeometryType.MULTI_POINT,
        MultiLineString: WKBGeometryType.MULTI_LINE_STRING,
        MultiPolygon: WKBGeometryType.MULTI_POLYGON,
        GeometryCollection: WKBGeometryType.GEOMETRY_COLLECTION,
      };
      const geomType = geom.getType();
      const typeId = wkblut[geomType];

      if (!typeId) {
        throw new Error('GeometryType ' + geomType + ' is not supported');
      }

      // first call of writeGeometry() traverse whole geometries to determine its output layout if not specified on constructor.
      if (!this.layout_) {
        this.layout_ = this.findMinimumLayout(geom);
      }

      this.writeWkbHeader(typeId, srid);

      if (geom instanceof SimpleGeometry) {
        const writerLUT = {
          Point: this.writePoint,
          LineString: this.writeLineString,
          Polygon: this.writePolygon,
          MultiPoint: this.writeMultiPoint,
          MultiLineString: this.writeMultiLineString,
          MultiPolygon: this.writeMultiPolygon,
        };
        writerLUT[geomType].call(this, geom.getCoordinates(), geom.getLayout());
      } else if (geom instanceof GeometryCollection) {
        this.writeGeometryCollection(geom.getGeometriesArray());
      }
    }

    getBuffer() {
      const byteLength = this.writeQueue_.reduce((acc, item) => acc + item[0], 0);
      const buffer = new ArrayBuffer(byteLength);
      const view = new DataView(buffer);

      let pos = 0;
      this.writeQueue_.forEach((item) => {
        switch (item[0]) {
          case 1:
            view.setUint8(pos, item[1]);
            break;
          case 4:
            view.setUint32(pos, item[1], this.isLittleEndian_);
            break;
          case 8:
            view.setFloat64(pos, item[1], this.isLittleEndian_);
            break;
        }

        pos += item[0];
      });

      return buffer;
    }
  }

  /**
   * @typedef {Object} Options
   * @property {boolean} [splitCollection=false] Whether to split GeometryCollections into multiple features on reading.
   * @property {boolean} [hex=true] Returns hex string instead of ArrayBuffer for output. This also is used as a hint internally whether it should load contents as text or ArrayBuffer on reading.
   * @property {boolean} [littleEndian=true] Use littleEndian for output.
   * @property {boolean} [ewkb=true] Use EWKB format for output.
   * @property {import("../geom/Geometry.js").GeometryLayout} [geometryLayout=null] Use specific coordinate layout for output features (null: auto detect)
   * @property {number} [nodataZ=0] If the `geometryLayout` doesn't match with geometry to be output, this value is used to fill missing coordinate value of Z.
   * @property {number} [nodataM=0] If the `geometryLayout` doesn't match with geometry to be output, this value is used to fill missing coordinate value of M.
   * @property {number|boolean} [srid=true] SRID for output. Specify integer value to enforce the value as a SRID. Specify `true` to extract from `dataProjection`. `false` to suppress the output. This option only takes effect when `ewkb` is `true`.
   */

  /**
   * @classdesc
   * Geometry format for reading and writing data in the `Well-Known Binary` (WKB) format.
   * Also supports `Extended Well-Known Binary` (EWKB) format, used in PostGIS for example.
   *
   * @api
   */
  class WKB extends FeatureFormat {
    /**
     * @param {Options} [options] Optional configuration object.
     */
    constructor(options) {
      super();

      options = options ? options : {};

      this.splitCollection = Boolean(options.splitCollection);

      this.viewCache_ = null;

      this.hex_ = options.hex !== false;
      this.littleEndian_ = options.littleEndian !== false;
      this.ewkb_ = options.ewkb !== false;

      this.layout_ = options.geometryLayout; // null for auto detect
      this.nodataZ_ = options.nodataZ || 0;
      this.nodataM_ = options.nodataM || 0;

      this.srid_ = options.srid;
    }

    /**
     * @return {import("./Feature.js").Type} Format.
     */
    getType() {
      return this.hex_ ? 'text' : 'arraybuffer';
    }

    /**
     * Read a single feature from a source.
     *
     * @param {string|ArrayBuffer|ArrayBufferView} source Source.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @return {import("../Feature.js").default} Feature.
     * @api
     */
    readFeature(source, options) {
      return new Feature({
        geometry: this.readGeometry(source, options),
      });
    }

    /**
     * Read all features from a source.
     *
     * @param {string|ArrayBuffer|ArrayBufferView} source Source.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @return {Array<import("../Feature.js").default>} Features.
     * @api
     */
    readFeatures(source, options) {
      let geometries = [];
      const geometry = this.readGeometry(source, options);
      if (this.splitCollection && geometry instanceof GeometryCollection) {
        geometries = geometry.getGeometriesArray();
      } else {
        geometries = [geometry];
      }
      return geometries.map((geometry) => new Feature({geometry}));
    }

    /**
     * Read a single geometry from a source.
     *
     * @param {string|ArrayBuffer|ArrayBufferView} source Source.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @return {import("../geom/Geometry.js").default} Geometry.
     * @api
     */
    readGeometry(source, options) {
      const view = getDataView(source);
      if (!view) {
        return null;
      }

      const reader = new WkbReader(view);
      const geometry = reader.readGeometry();

      this.viewCache_ = view; // cache for internal subsequent call of readProjection()
      options = this.getReadOptions(source, options);
      this.viewCache_ = null; // release

      return transformGeometryWithOptions(geometry, false, options);
    }

    /**
     * Read the projection from a source.
     *
     * @param {string|ArrayBuffer|ArrayBufferView} source Source.
     * @return {import("../proj/Projection.js").default|undefined} Projection.
     * @api
     */
    readProjection(source) {
      const view = this.viewCache_ || getDataView(source);
      if (!view) {
        return undefined;
      }

      const reader = new WkbReader(view);
      reader.readWkbHeader();

      return (
        (reader.getSrid() && get$2('EPSG:' + reader.getSrid())) ||
        undefined
      );
    }

    /**
     * Encode a feature in this format.
     *
     * @param {import("../Feature.js").default} feature Feature.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @return {string|ArrayBuffer} Result.
     * @api
     */
    writeFeature(feature, options) {
      return this.writeGeometry(feature.getGeometry(), options);
    }

    /**
     * Encode an array of features in this format.
     *
     * @param {Array<import("../Feature.js").default>} features Features.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @return {string|ArrayBuffer} Result.
     * @api
     */
    writeFeatures(features, options) {
      return this.writeGeometry(
        new GeometryCollection(features.map((f) => f.getGeometry())),
        options
      );
    }

    /**
     * Write a single geometry in this format.
     *
     * @param {import("../geom/Geometry.js").default} geometry Geometry.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @return {string|ArrayBuffer} Result.
     * @api
     */
    writeGeometry(geometry, options) {
      options = this.adaptOptions(options);

      const writer = new WkbWriter({
        layout: this.layout_,
        littleEndian: this.littleEndian_,
        ewkb: this.ewkb_,

        nodata: {
          Z: this.nodataZ_,
          M: this.nodataM_,
        },
      });

      // extract SRID from `dataProjection`
      let srid = Number.isInteger(this.srid_) ? Number(this.srid_) : null;
      if (this.srid_ !== false && !Number.isInteger(this.srid_)) {
        const dataProjection =
          options.dataProjection && get$2(options.dataProjection);
        if (dataProjection) {
          const code = dataProjection.getCode();
          if (code.startsWith('EPSG:')) {
            srid = Number(code.substring(5));
          }
        }
      }

      writer.writeGeometry(
        transformGeometryWithOptions(geometry, true, options),
        srid
      );
      const buffer = writer.getBuffer();

      return this.hex_ ? encodeHexString(buffer) : buffer;
    }
  }

  /**
   * @param {ArrayBuffer} buffer source buffer
   * @return {string} encoded hex string
   */
  function encodeHexString(buffer) {
    const view = new Uint8Array(buffer);
    return Array.from(view.values())
      .map((x) => (x < 16 ? '0' : '') + Number(x).toString(16).toUpperCase())
      .join('');
  }

  /**
   * @param {string} text source text
   * @return {DataView} decoded binary buffer
   */
  function decodeHexString(text) {
    const buffer = new Uint8Array(text.length / 2);
    for (let i = 0; i < text.length / 2; i++) {
      buffer[i] = parseInt(text.substr(i * 2, 2), 16);
    }
    return new DataView(buffer.buffer);
  }

  /**
   * @param {string | ArrayBuffer | ArrayBufferView} source source
   * @return {DataView} data view
   */
  function getDataView(source) {
    if (typeof source === 'string') {
      return decodeHexString(source);
    } else if (ArrayBuffer.isView(source)) {
      if (source instanceof DataView) {
        return source;
      }
      return new DataView(source.buffer, source.byteOffset, source.byteLength);
    } else if (source instanceof ArrayBuffer) {
      return new DataView(source);
    } else {
      return null;
    }
  }

  /**
   * @module ol/format/WKT
   */

  /**
   * Geometry constructors
   * @enum {function (new:import("../geom/Geometry.js").default, Array, import("../geom/Geometry.js").GeometryLayout)}
   */
  const GeometryConstructor = {
    'POINT': Point,
    'LINESTRING': LineString,
    'POLYGON': Polygon,
    'MULTIPOINT': MultiPoint,
    'MULTILINESTRING': MultiLineString,
    'MULTIPOLYGON': MultiPolygon,
  };

  /**
   * @typedef {Object} Options
   * @property {boolean} [splitCollection=false] Whether to split GeometryCollections into
   * multiple features on reading.
   */

  /**
   * @typedef {Object} Token
   * @property {number} type Type.
   * @property {number|string} [value] Value.
   * @property {number} position Position.
   */

  /**
   * @const
   * @type {string}
   */
  const EMPTY = 'EMPTY';

  /**
   * @const
   * @type {string}
   */
  const Z = 'Z';

  /**
   * @const
   * @type {string}
   */
  const M = 'M';

  /**
   * @const
   * @type {string}
   */
  const ZM = 'ZM';

  /**
   * @const
   * @enum {number}
   */
  const TokenType = {
    START: 0,
    TEXT: 1,
    LEFT_PAREN: 2,
    RIGHT_PAREN: 3,
    NUMBER: 4,
    COMMA: 5,
    EOF: 6,
  };

  /**
   * @type {Object<import("../geom/Geometry.js").Type, string>}
   */
  const wktTypeLookup = {
    Point: 'POINT',
    LineString: 'LINESTRING',
    Polygon: 'POLYGON',
    MultiPoint: 'MULTIPOINT',
    MultiLineString: 'MULTILINESTRING',
    MultiPolygon: 'MULTIPOLYGON',
    GeometryCollection: 'GEOMETRYCOLLECTION',
    Circle: 'CIRCLE',
  };

  /**
   * Class to tokenize a WKT string.
   */
  class Lexer {
    /**
     * @param {string} wkt WKT string.
     */
    constructor(wkt) {
      /**
       * @type {string}
       */
      this.wkt = wkt;

      /**
       * @type {number}
       * @private
       */
      this.index_ = -1;
    }

    /**
     * @param {string} c Character.
     * @return {boolean} Whether the character is alphabetic.
     * @private
     */
    isAlpha_(c) {
      return (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z');
    }

    /**
     * @param {string} c Character.
     * @param {boolean} [decimal] Whether the string number
     *     contains a dot, i.e. is a decimal number.
     * @return {boolean} Whether the character is numeric.
     * @private
     */
    isNumeric_(c, decimal) {
      decimal = decimal !== undefined ? decimal : false;
      return (c >= '0' && c <= '9') || (c == '.' && !decimal);
    }

    /**
     * @param {string} c Character.
     * @return {boolean} Whether the character is whitespace.
     * @private
     */
    isWhiteSpace_(c) {
      return c == ' ' || c == '\t' || c == '\r' || c == '\n';
    }

    /**
     * @return {string} Next string character.
     * @private
     */
    nextChar_() {
      return this.wkt.charAt(++this.index_);
    }

    /**
     * Fetch and return the next token.
     * @return {Token} Next string token.
     */
    nextToken() {
      const c = this.nextChar_();
      const position = this.index_;
      /** @type {number|string} */
      let value = c;
      let type;

      if (c == '(') {
        type = TokenType.LEFT_PAREN;
      } else if (c == ',') {
        type = TokenType.COMMA;
      } else if (c == ')') {
        type = TokenType.RIGHT_PAREN;
      } else if (this.isNumeric_(c) || c == '-') {
        type = TokenType.NUMBER;
        value = this.readNumber_();
      } else if (this.isAlpha_(c)) {
        type = TokenType.TEXT;
        value = this.readText_();
      } else if (this.isWhiteSpace_(c)) {
        return this.nextToken();
      } else if (c === '') {
        type = TokenType.EOF;
      } else {
        throw new Error('Unexpected character: ' + c);
      }

      return {position: position, value: value, type: type};
    }

    /**
     * @return {number} Numeric token value.
     * @private
     */
    readNumber_() {
      let c;
      const index = this.index_;
      let decimal = false;
      let scientificNotation = false;
      do {
        if (c == '.') {
          decimal = true;
        } else if (c == 'e' || c == 'E') {
          scientificNotation = true;
        }
        c = this.nextChar_();
      } while (
        this.isNumeric_(c, decimal) ||
        // if we haven't detected a scientific number before, 'e' or 'E'
        // hint that we should continue to read
        (!scientificNotation && (c == 'e' || c == 'E')) ||
        // once we know that we have a scientific number, both '-' and '+'
        // are allowed
        (scientificNotation && (c == '-' || c == '+'))
      );
      return parseFloat(this.wkt.substring(index, this.index_--));
    }

    /**
     * @return {string} String token value.
     * @private
     */
    readText_() {
      let c;
      const index = this.index_;
      do {
        c = this.nextChar_();
      } while (this.isAlpha_(c));
      return this.wkt.substring(index, this.index_--).toUpperCase();
    }
  }

  /**
   * Class to parse the tokens from the WKT string.
   */
  class Parser {
    /**
     * @param {Lexer} lexer The lexer.
     */
    constructor(lexer) {
      /**
       * @type {Lexer}
       * @private
       */
      this.lexer_ = lexer;

      /**
       * @type {Token}
       * @private
       */
      this.token_ = {
        position: 0,
        type: TokenType.START,
      };

      /**
       * @type {import("../geom/Geometry.js").GeometryLayout}
       * @private
       */
      this.layout_ = 'XY';
    }

    /**
     * Fetch the next token form the lexer and replace the active token.
     * @private
     */
    consume_() {
      this.token_ = this.lexer_.nextToken();
    }

    /**
     * Tests if the given type matches the type of the current token.
     * @param {TokenType} type Token type.
     * @return {boolean} Whether the token matches the given type.
     */
    isTokenType(type) {
      return this.token_.type == type;
    }

    /**
     * If the given type matches the current token, consume it.
     * @param {TokenType} type Token type.
     * @return {boolean} Whether the token matches the given type.
     */
    match(type) {
      const isMatch = this.isTokenType(type);
      if (isMatch) {
        this.consume_();
      }
      return isMatch;
    }

    /**
     * Try to parse the tokens provided by the lexer.
     * @return {import("../geom/Geometry.js").default} The geometry.
     */
    parse() {
      this.consume_();
      return this.parseGeometry_();
    }

    /**
     * Try to parse the dimensional info.
     * @return {import("../geom/Geometry.js").GeometryLayout} The layout.
     * @private
     */
    parseGeometryLayout_() {
      /** @type {import("../geom/Geometry.js").GeometryLayout} */
      let layout = 'XY';
      const dimToken = this.token_;
      if (this.isTokenType(TokenType.TEXT)) {
        const dimInfo = dimToken.value;
        if (dimInfo === Z) {
          layout = 'XYZ';
        } else if (dimInfo === M) {
          layout = 'XYM';
        } else if (dimInfo === ZM) {
          layout = 'XYZM';
        }
        if (layout !== 'XY') {
          this.consume_();
        }
      }
      return layout;
    }

    /**
     * @return {Array<import("../geom/Geometry.js").default>} A collection of geometries.
     * @private
     */
    parseGeometryCollectionText_() {
      if (this.match(TokenType.LEFT_PAREN)) {
        const geometries = [];
        do {
          geometries.push(this.parseGeometry_());
        } while (this.match(TokenType.COMMA));
        if (this.match(TokenType.RIGHT_PAREN)) {
          return geometries;
        }
      }
      throw new Error(this.formatErrorMessage_());
    }

    /**
     * @return {Array<number>} All values in a point.
     * @private
     */
    parsePointText_() {
      if (this.match(TokenType.LEFT_PAREN)) {
        const coordinates = this.parsePoint_();
        if (this.match(TokenType.RIGHT_PAREN)) {
          return coordinates;
        }
      }
      throw new Error(this.formatErrorMessage_());
    }

    /**
     * @return {Array<Array<number>>} All points in a linestring.
     * @private
     */
    parseLineStringText_() {
      if (this.match(TokenType.LEFT_PAREN)) {
        const coordinates = this.parsePointList_();
        if (this.match(TokenType.RIGHT_PAREN)) {
          return coordinates;
        }
      }
      throw new Error(this.formatErrorMessage_());
    }

    /**
     * @return {Array<Array<Array<number>>>} All points in a polygon.
     * @private
     */
    parsePolygonText_() {
      if (this.match(TokenType.LEFT_PAREN)) {
        const coordinates = this.parseLineStringTextList_();
        if (this.match(TokenType.RIGHT_PAREN)) {
          return coordinates;
        }
      }
      throw new Error(this.formatErrorMessage_());
    }

    /**
     * @return {Array<Array<number>>} All points in a multipoint.
     * @private
     */
    parseMultiPointText_() {
      if (this.match(TokenType.LEFT_PAREN)) {
        let coordinates;
        if (this.token_.type == TokenType.LEFT_PAREN) {
          coordinates = this.parsePointTextList_();
        } else {
          coordinates = this.parsePointList_();
        }
        if (this.match(TokenType.RIGHT_PAREN)) {
          return coordinates;
        }
      }
      throw new Error(this.formatErrorMessage_());
    }

    /**
     * @return {Array<Array<Array<number>>>} All linestring points
     *                                          in a multilinestring.
     * @private
     */
    parseMultiLineStringText_() {
      if (this.match(TokenType.LEFT_PAREN)) {
        const coordinates = this.parseLineStringTextList_();
        if (this.match(TokenType.RIGHT_PAREN)) {
          return coordinates;
        }
      }
      throw new Error(this.formatErrorMessage_());
    }

    /**
     * @return {Array<Array<Array<Array<number>>>>} All polygon points in a multipolygon.
     * @private
     */
    parseMultiPolygonText_() {
      if (this.match(TokenType.LEFT_PAREN)) {
        const coordinates = this.parsePolygonTextList_();
        if (this.match(TokenType.RIGHT_PAREN)) {
          return coordinates;
        }
      }
      throw new Error(this.formatErrorMessage_());
    }

    /**
     * @return {Array<number>} A point.
     * @private
     */
    parsePoint_() {
      const coordinates = [];
      const dimensions = this.layout_.length;
      for (let i = 0; i < dimensions; ++i) {
        const token = this.token_;
        if (this.match(TokenType.NUMBER)) {
          coordinates.push(/** @type {number} */ (token.value));
        } else {
          break;
        }
      }
      if (coordinates.length == dimensions) {
        return coordinates;
      }
      throw new Error(this.formatErrorMessage_());
    }

    /**
     * @return {Array<Array<number>>} An array of points.
     * @private
     */
    parsePointList_() {
      const coordinates = [this.parsePoint_()];
      while (this.match(TokenType.COMMA)) {
        coordinates.push(this.parsePoint_());
      }
      return coordinates;
    }

    /**
     * @return {Array<Array<number>>} An array of points.
     * @private
     */
    parsePointTextList_() {
      const coordinates = [this.parsePointText_()];
      while (this.match(TokenType.COMMA)) {
        coordinates.push(this.parsePointText_());
      }
      return coordinates;
    }

    /**
     * @return {Array<Array<Array<number>>>} An array of points.
     * @private
     */
    parseLineStringTextList_() {
      const coordinates = [this.parseLineStringText_()];
      while (this.match(TokenType.COMMA)) {
        coordinates.push(this.parseLineStringText_());
      }
      return coordinates;
    }

    /**
     * @return {Array<Array<Array<Array<number>>>>} An array of points.
     * @private
     */
    parsePolygonTextList_() {
      const coordinates = [this.parsePolygonText_()];
      while (this.match(TokenType.COMMA)) {
        coordinates.push(this.parsePolygonText_());
      }
      return coordinates;
    }

    /**
     * @return {boolean} Whether the token implies an empty geometry.
     * @private
     */
    isEmptyGeometry_() {
      const isEmpty =
        this.isTokenType(TokenType.TEXT) && this.token_.value == EMPTY;
      if (isEmpty) {
        this.consume_();
      }
      return isEmpty;
    }

    /**
     * Create an error message for an unexpected token error.
     * @return {string} Error message.
     * @private
     */
    formatErrorMessage_() {
      return (
        'Unexpected `' +
        this.token_.value +
        '` at position ' +
        this.token_.position +
        ' in `' +
        this.lexer_.wkt +
        '`'
      );
    }

    /**
     * @return {import("../geom/Geometry.js").default} The geometry.
     * @private
     */
    parseGeometry_() {
      const token = this.token_;
      if (this.match(TokenType.TEXT)) {
        const geomType = /** @type {string} */ (token.value);
        this.layout_ = this.parseGeometryLayout_();
        const isEmpty = this.isEmptyGeometry_();
        if (geomType == 'GEOMETRYCOLLECTION') {
          if (isEmpty) {
            return new GeometryCollection([]);
          }
          const geometries = this.parseGeometryCollectionText_();
          return new GeometryCollection(geometries);
        } else {
          const ctor = GeometryConstructor[geomType];
          if (!ctor) {
            throw new Error('Invalid geometry type: ' + geomType);
          }

          let coordinates;

          if (isEmpty) {
            if (geomType == 'POINT') {
              coordinates = [NaN, NaN];
            } else {
              coordinates = [];
            }
          } else {
            switch (geomType) {
              case 'POINT': {
                coordinates = this.parsePointText_();
                break;
              }
              case 'LINESTRING': {
                coordinates = this.parseLineStringText_();
                break;
              }
              case 'POLYGON': {
                coordinates = this.parsePolygonText_();
                break;
              }
              case 'MULTIPOINT': {
                coordinates = this.parseMultiPointText_();
                break;
              }
              case 'MULTILINESTRING': {
                coordinates = this.parseMultiLineStringText_();
                break;
              }
              case 'MULTIPOLYGON': {
                coordinates = this.parseMultiPolygonText_();
                break;
              }
            }
          }

          return new ctor(coordinates, this.layout_);
        }
      }
      throw new Error(this.formatErrorMessage_());
    }
  }

  /**
   * @classdesc
   * Geometry format for reading and writing data in the `WellKnownText` (WKT)
   * format.
   *
   * @api
   */
  class WKT extends TextFeature {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      super();

      options = options ? options : {};

      /**
       * Split GeometryCollection into multiple features.
       * @type {boolean}
       * @private
       */
      this.splitCollection_ =
        options.splitCollection !== undefined ? options.splitCollection : false;
    }

    /**
     * Parse a WKT string.
     * @param {string} wkt WKT string.
     * @return {import("../geom/Geometry.js").default}
     *     The geometry created.
     * @private
     */
    parse_(wkt) {
      const lexer = new Lexer(wkt);
      const parser = new Parser(lexer);
      return parser.parse();
    }

    /**
     * @protected
     * @param {string} text Text.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @return {import("../Feature.js").default} Feature.
     */
    readFeatureFromText(text, options) {
      const geom = this.readGeometryFromText(text, options);
      const feature = new Feature();
      feature.setGeometry(geom);
      return feature;
    }

    /**
     * @param {string} text Text.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @protected
     * @return {Array<Feature>} Features.
     */
    readFeaturesFromText(text, options) {
      let geometries = [];
      const geometry = this.readGeometryFromText(text, options);
      if (this.splitCollection_ && geometry.getType() == 'GeometryCollection') {
        geometries = /** @type {GeometryCollection} */ (
          geometry
        ).getGeometriesArray();
      } else {
        geometries = [geometry];
      }
      const features = [];
      for (let i = 0, ii = geometries.length; i < ii; ++i) {
        const feature = new Feature();
        feature.setGeometry(geometries[i]);
        features.push(feature);
      }
      return features;
    }

    /**
     * @param {string} text Text.
     * @param {import("./Feature.js").ReadOptions} [options] Read options.
     * @protected
     * @return {import("../geom/Geometry.js").default} Geometry.
     */
    readGeometryFromText(text, options) {
      const geometry = this.parse_(text);
      return transformGeometryWithOptions(geometry, false, options);
    }

    /**
     * @param {import("../Feature.js").default} feature Features.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @protected
     * @return {string} Text.
     */
    writeFeatureText(feature, options) {
      const geometry = feature.getGeometry();
      if (geometry) {
        return this.writeGeometryText(geometry, options);
      }
      return '';
    }

    /**
     * @param {Array<import("../Feature.js").default>} features Features.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @protected
     * @return {string} Text.
     */
    writeFeaturesText(features, options) {
      if (features.length == 1) {
        return this.writeFeatureText(features[0], options);
      }
      const geometries = [];
      for (let i = 0, ii = features.length; i < ii; ++i) {
        geometries.push(features[i].getGeometry());
      }
      const collection = new GeometryCollection(geometries);
      return this.writeGeometryText(collection, options);
    }

    /**
     * @param {import("../geom/Geometry.js").default} geometry Geometry.
     * @param {import("./Feature.js").WriteOptions} [options] Write options.
     * @protected
     * @return {string} Text.
     */
    writeGeometryText(geometry, options) {
      return encode(transformGeometryWithOptions(geometry, true, options));
    }
  }

  /**
   * @param {Point} geom Point geometry.
   * @return {string} Coordinates part of Point as WKT.
   */
  function encodePointGeometry(geom) {
    const coordinates = geom.getCoordinates();
    if (coordinates.length === 0) {
      return '';
    }
    return coordinates.join(' ');
  }

  /**
   * @param {MultiPoint} geom MultiPoint geometry.
   * @return {string} Coordinates part of MultiPoint as WKT.
   */
  function encodeMultiPointGeometry(geom) {
    const array = [];
    const components = geom.getPoints();
    for (let i = 0, ii = components.length; i < ii; ++i) {
      array.push('(' + encodePointGeometry(components[i]) + ')');
    }
    return array.join(',');
  }

  /**
   * @param {GeometryCollection} geom GeometryCollection geometry.
   * @return {string} Coordinates part of GeometryCollection as WKT.
   */
  function encodeGeometryCollectionGeometry(geom) {
    const array = [];
    const geoms = geom.getGeometries();
    for (let i = 0, ii = geoms.length; i < ii; ++i) {
      array.push(encode(geoms[i]));
    }
    return array.join(',');
  }

  /**
   * @param {LineString|import("../geom/LinearRing.js").default} geom LineString geometry.
   * @return {string} Coordinates part of LineString as WKT.
   */
  function encodeLineStringGeometry(geom) {
    const coordinates = geom.getCoordinates();
    const array = [];
    for (let i = 0, ii = coordinates.length; i < ii; ++i) {
      array.push(coordinates[i].join(' '));
    }
    return array.join(',');
  }

  /**
   * @param {MultiLineString} geom MultiLineString geometry.
   * @return {string} Coordinates part of MultiLineString as WKT.
   */
  function encodeMultiLineStringGeometry(geom) {
    const array = [];
    const components = geom.getLineStrings();
    for (let i = 0, ii = components.length; i < ii; ++i) {
      array.push('(' + encodeLineStringGeometry(components[i]) + ')');
    }
    return array.join(',');
  }

  /**
   * @param {Polygon} geom Polygon geometry.
   * @return {string} Coordinates part of Polygon as WKT.
   */
  function encodePolygonGeometry(geom) {
    const array = [];
    const rings = geom.getLinearRings();
    for (let i = 0, ii = rings.length; i < ii; ++i) {
      array.push('(' + encodeLineStringGeometry(rings[i]) + ')');
    }
    return array.join(',');
  }

  /**
   * @param {MultiPolygon} geom MultiPolygon geometry.
   * @return {string} Coordinates part of MultiPolygon as WKT.
   */
  function encodeMultiPolygonGeometry(geom) {
    const array = [];
    const components = geom.getPolygons();
    for (let i = 0, ii = components.length; i < ii; ++i) {
      array.push('(' + encodePolygonGeometry(components[i]) + ')');
    }
    return array.join(',');
  }

  /**
   * @param {import("../geom/SimpleGeometry.js").default} geom SimpleGeometry geometry.
   * @return {string} Potential dimensional information for WKT type.
   */
  function encodeGeometryLayout(geom) {
    const layout = geom.getLayout();
    let dimInfo = '';
    if (layout === 'XYZ' || layout === 'XYZM') {
      dimInfo += Z;
    }
    if (layout === 'XYM' || layout === 'XYZM') {
      dimInfo += M;
    }
    return dimInfo;
  }

  /**
   * @const
   * @type {Object<string, function(import("../geom/Geometry.js").default): string>}
   */
  const GeometryEncoder = {
    'Point': encodePointGeometry,
    'LineString': encodeLineStringGeometry,
    'Polygon': encodePolygonGeometry,
    'MultiPoint': encodeMultiPointGeometry,
    'MultiLineString': encodeMultiLineStringGeometry,
    'MultiPolygon': encodeMultiPolygonGeometry,
    'GeometryCollection': encodeGeometryCollectionGeometry,
  };

  /**
   * Encode a geometry as WKT.
   * @param {import("../geom/Geometry.js").default} geom The geometry to encode.
   * @return {string} WKT string for the geometry.
   */
  function encode(geom) {
    const type = geom.getType();
    const geometryEncoder = GeometryEncoder[type];
    const enc = geometryEncoder(geom);
    let wktType = wktTypeLookup[type];
    if (typeof (/** @type {?} */ (geom).getFlatCoordinates) === 'function') {
      const dimInfo = encodeGeometryLayout(
        /** @type {import("../geom/SimpleGeometry.js").default} */ (geom)
      );
      if (dimInfo.length > 0) {
        wktType += ' ' + dimInfo;
      }
    }
    if (enc.length === 0) {
      return wktType + ' ' + EMPTY;
    }
    return wktType + '(' + enc + ')';
  }

  /**
   * @module ol/format/WMSCapabilities
   */

  /**
   * @const
   * @type {Array<null|string>}
   */
  const NAMESPACE_URIS$1 = [null, 'http://www.opengis.net/wms'];

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const PARSERS$1 = makeStructureNS(NAMESPACE_URIS$1, {
    'Service': makeObjectPropertySetter(readService),
    'Capability': makeObjectPropertySetter(readCapability),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const CAPABILITY_PARSERS = makeStructureNS(NAMESPACE_URIS$1, {
    'Request': makeObjectPropertySetter(readRequest),
    'Exception': makeObjectPropertySetter(readException),
    'Layer': makeObjectPropertySetter(readCapabilityLayer),
  });

  /**
   * @classdesc
   * Format for reading WMS capabilities data
   *
   * @api
   */
  class WMSCapabilities extends XML {
    constructor() {
      super();

      /**
       * @type {string|undefined}
       */
      this.version = undefined;
    }

    /**
     * @param {Element} node Node.
     * @return {Object} Object
     */
    readFromNode(node) {
      this.version = node.getAttribute('version').trim();
      const wmsCapabilityObject = pushParseAndPop(
        {
          'version': this.version,
        },
        PARSERS$1,
        node,
        []
      );
      return wmsCapabilityObject ? wmsCapabilityObject : null;
    }
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const SERVICE_PARSERS = makeStructureNS(NAMESPACE_URIS$1, {
    'Name': makeObjectPropertySetter(readString),
    'Title': makeObjectPropertySetter(readString),
    'Abstract': makeObjectPropertySetter(readString),
    'KeywordList': makeObjectPropertySetter(readKeywordList),
    'OnlineResource': makeObjectPropertySetter(readHref),
    'ContactInformation': makeObjectPropertySetter(readContactInformation),
    'Fees': makeObjectPropertySetter(readString),
    'AccessConstraints': makeObjectPropertySetter(readString),
    'LayerLimit': makeObjectPropertySetter(readPositiveInteger),
    'MaxWidth': makeObjectPropertySetter(readPositiveInteger),
    'MaxHeight': makeObjectPropertySetter(readPositiveInteger),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const CONTACT_INFORMATION_PARSERS = makeStructureNS(NAMESPACE_URIS$1, {
    'ContactPersonPrimary': makeObjectPropertySetter(readContactPersonPrimary),
    'ContactPosition': makeObjectPropertySetter(readString),
    'ContactAddress': makeObjectPropertySetter(readContactAddress),
    'ContactVoiceTelephone': makeObjectPropertySetter(readString),
    'ContactFacsimileTelephone': makeObjectPropertySetter(readString),
    'ContactElectronicMailAddress': makeObjectPropertySetter(readString),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const CONTACT_PERSON_PARSERS = makeStructureNS(NAMESPACE_URIS$1, {
    'ContactPerson': makeObjectPropertySetter(readString),
    'ContactOrganization': makeObjectPropertySetter(readString),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const CONTACT_ADDRESS_PARSERS = makeStructureNS(NAMESPACE_URIS$1, {
    'AddressType': makeObjectPropertySetter(readString),
    'Address': makeObjectPropertySetter(readString),
    'City': makeObjectPropertySetter(readString),
    'StateOrProvince': makeObjectPropertySetter(readString),
    'PostCode': makeObjectPropertySetter(readString),
    'Country': makeObjectPropertySetter(readString),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const EXCEPTION_PARSERS = makeStructureNS(NAMESPACE_URIS$1, {
    'Format': makeArrayPusher(readString),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const LAYER_PARSERS$1 = makeStructureNS(NAMESPACE_URIS$1, {
    'Name': makeObjectPropertySetter(readString),
    'Title': makeObjectPropertySetter(readString),
    'Abstract': makeObjectPropertySetter(readString),
    'KeywordList': makeObjectPropertySetter(readKeywordList),
    'CRS': makeObjectPropertyPusher(readString),
    'EX_GeographicBoundingBox': makeObjectPropertySetter(
      readEXGeographicBoundingBox
    ),
    'BoundingBox': makeObjectPropertyPusher(readBoundingBox$1),
    'Dimension': makeObjectPropertyPusher(readDimension),
    'Attribution': makeObjectPropertySetter(readAttribution),
    'AuthorityURL': makeObjectPropertyPusher(readAuthorityURL),
    'Identifier': makeObjectPropertyPusher(readString),
    'MetadataURL': makeObjectPropertyPusher(readMetadataURL),
    'DataURL': makeObjectPropertyPusher(readFormatOnlineresource),
    'FeatureListURL': makeObjectPropertyPusher(readFormatOnlineresource),
    'Style': makeObjectPropertyPusher(readStyle$1),
    'MinScaleDenominator': makeObjectPropertySetter(readDecimal),
    'MaxScaleDenominator': makeObjectPropertySetter(readDecimal),
    'Layer': makeObjectPropertyPusher(readLayer$1),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const ATTRIBUTION_PARSERS = makeStructureNS(NAMESPACE_URIS$1, {
    'Title': makeObjectPropertySetter(readString),
    'OnlineResource': makeObjectPropertySetter(readHref),
    'LogoURL': makeObjectPropertySetter(readSizedFormatOnlineresource),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const EX_GEOGRAPHIC_BOUNDING_BOX_PARSERS = makeStructureNS(NAMESPACE_URIS$1, {
    'westBoundLongitude': makeObjectPropertySetter(readDecimal),
    'eastBoundLongitude': makeObjectPropertySetter(readDecimal),
    'southBoundLatitude': makeObjectPropertySetter(readDecimal),
    'northBoundLatitude': makeObjectPropertySetter(readDecimal),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const REQUEST_PARSERS = makeStructureNS(NAMESPACE_URIS$1, {
    'GetCapabilities': makeObjectPropertySetter(readOperationType),
    'GetMap': makeObjectPropertySetter(readOperationType),
    'GetFeatureInfo': makeObjectPropertySetter(readOperationType),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const OPERATIONTYPE_PARSERS = makeStructureNS(NAMESPACE_URIS$1, {
    'Format': makeObjectPropertyPusher(readString),
    'DCPType': makeObjectPropertyPusher(readDCPType),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const DCPTYPE_PARSERS = makeStructureNS(NAMESPACE_URIS$1, {
    'HTTP': makeObjectPropertySetter(readHTTP),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const HTTP_PARSERS = makeStructureNS(NAMESPACE_URIS$1, {
    'Get': makeObjectPropertySetter(readFormatOnlineresource),
    'Post': makeObjectPropertySetter(readFormatOnlineresource),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const STYLE_PARSERS$1 = makeStructureNS(NAMESPACE_URIS$1, {
    'Name': makeObjectPropertySetter(readString),
    'Title': makeObjectPropertySetter(readString),
    'Abstract': makeObjectPropertySetter(readString),
    'LegendURL': makeObjectPropertyPusher(readSizedFormatOnlineresource),
    'StyleSheetURL': makeObjectPropertySetter(readFormatOnlineresource),
    'StyleURL': makeObjectPropertySetter(readFormatOnlineresource),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const FORMAT_ONLINERESOURCE_PARSERS = makeStructureNS(NAMESPACE_URIS$1, {
    'Format': makeObjectPropertySetter(readString),
    'OnlineResource': makeObjectPropertySetter(readHref),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const KEYWORDLIST_PARSERS = makeStructureNS(NAMESPACE_URIS$1, {
    'Keyword': makeArrayPusher(readString),
  });

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} Attribution object.
   */
  function readAttribution(node, objectStack) {
    return pushParseAndPop({}, ATTRIBUTION_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object} Bounding box object.
   */
  function readBoundingBox$1(node, objectStack) {
    const extent = [
      readDecimalString(node.getAttribute('minx')),
      readDecimalString(node.getAttribute('miny')),
      readDecimalString(node.getAttribute('maxx')),
      readDecimalString(node.getAttribute('maxy')),
    ];

    const resolutions = [
      readDecimalString(node.getAttribute('resx')),
      readDecimalString(node.getAttribute('resy')),
    ];

    return {
      'crs': node.getAttribute('CRS'),
      'extent': extent,
      'res': resolutions,
    };
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {import("../extent.js").Extent|undefined} Bounding box object.
   */
  function readEXGeographicBoundingBox(node, objectStack) {
    const geographicBoundingBox = pushParseAndPop(
      {},
      EX_GEOGRAPHIC_BOUNDING_BOX_PARSERS,
      node,
      objectStack
    );
    if (!geographicBoundingBox) {
      return undefined;
    }
    const westBoundLongitude =
      /** @type {number|undefined} */
      (geographicBoundingBox['westBoundLongitude']);
    const southBoundLatitude =
      /** @type {number|undefined} */
      (geographicBoundingBox['southBoundLatitude']);
    const eastBoundLongitude =
      /** @type {number|undefined} */
      (geographicBoundingBox['eastBoundLongitude']);
    const northBoundLatitude =
      /** @type {number|undefined} */
      (geographicBoundingBox['northBoundLatitude']);
    if (
      westBoundLongitude === undefined ||
      southBoundLatitude === undefined ||
      eastBoundLongitude === undefined ||
      northBoundLatitude === undefined
    ) {
      return undefined;
    }
    return [
      westBoundLongitude,
      southBoundLatitude,
      eastBoundLongitude,
      northBoundLatitude,
    ];
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} Capability object.
   */
  function readCapability(node, objectStack) {
    return pushParseAndPop({}, CAPABILITY_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} Service object.
   */
  function readService(node, objectStack) {
    return pushParseAndPop({}, SERVICE_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} Contact information object.
   */
  function readContactInformation(node, objectStack) {
    return pushParseAndPop({}, CONTACT_INFORMATION_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} Contact person object.
   */
  function readContactPersonPrimary(node, objectStack) {
    return pushParseAndPop({}, CONTACT_PERSON_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} Contact address object.
   */
  function readContactAddress(node, objectStack) {
    return pushParseAndPop({}, CONTACT_ADDRESS_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Array<string>|undefined} Format array.
   */
  function readException(node, objectStack) {
    return pushParseAndPop([], EXCEPTION_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} Layer object.
   */
  function readCapabilityLayer(node, objectStack) {
    const layerObject = pushParseAndPop({}, LAYER_PARSERS$1, node, objectStack);

    if (layerObject['Layer'] === undefined) {
      return Object.assign(layerObject, readLayer$1(node, objectStack));
    }

    return layerObject;
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} Layer object.
   */
  function readLayer$1(node, objectStack) {
    const parentLayerObject = /**  @type {!Object<string,*>} */ (
      objectStack[objectStack.length - 1]
    );

    const layerObject = pushParseAndPop({}, LAYER_PARSERS$1, node, objectStack);

    if (!layerObject) {
      return undefined;
    }
    let queryable = readBooleanString(node.getAttribute('queryable'));
    if (queryable === undefined) {
      queryable = parentLayerObject['queryable'];
    }
    layerObject['queryable'] = queryable !== undefined ? queryable : false;

    let cascaded = readNonNegativeIntegerString(node.getAttribute('cascaded'));
    if (cascaded === undefined) {
      cascaded = parentLayerObject['cascaded'];
    }
    layerObject['cascaded'] = cascaded;

    let opaque = readBooleanString(node.getAttribute('opaque'));
    if (opaque === undefined) {
      opaque = parentLayerObject['opaque'];
    }
    layerObject['opaque'] = opaque !== undefined ? opaque : false;

    let noSubsets = readBooleanString(node.getAttribute('noSubsets'));
    if (noSubsets === undefined) {
      noSubsets = parentLayerObject['noSubsets'];
    }
    layerObject['noSubsets'] = noSubsets !== undefined ? noSubsets : false;

    let fixedWidth = readDecimalString(node.getAttribute('fixedWidth'));
    if (!fixedWidth) {
      fixedWidth = parentLayerObject['fixedWidth'];
    }
    layerObject['fixedWidth'] = fixedWidth;

    let fixedHeight = readDecimalString(node.getAttribute('fixedHeight'));
    if (!fixedHeight) {
      fixedHeight = parentLayerObject['fixedHeight'];
    }
    layerObject['fixedHeight'] = fixedHeight;

    // See 7.2.4.8
    const addKeys = ['Style', 'CRS', 'AuthorityURL'];
    addKeys.forEach(function (key) {
      if (key in parentLayerObject) {
        const childValue = layerObject[key] || [];
        layerObject[key] = childValue.concat(parentLayerObject[key]);
      }
    });

    const replaceKeys = [
      'EX_GeographicBoundingBox',
      'BoundingBox',
      'Dimension',
      'Attribution',
      'MinScaleDenominator',
      'MaxScaleDenominator',
    ];
    replaceKeys.forEach(function (key) {
      if (!(key in layerObject)) {
        const parentValue = parentLayerObject[key];
        layerObject[key] = parentValue;
      }
    });

    return layerObject;
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object} Dimension object.
   */
  function readDimension(node, objectStack) {
    const dimensionObject = {
      'name': node.getAttribute('name'),
      'units': node.getAttribute('units'),
      'unitSymbol': node.getAttribute('unitSymbol'),
      'default': node.getAttribute('default'),
      'multipleValues': readBooleanString(node.getAttribute('multipleValues')),
      'nearestValue': readBooleanString(node.getAttribute('nearestValue')),
      'current': readBooleanString(node.getAttribute('current')),
      'values': readString(node),
    };
    return dimensionObject;
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} Online resource object.
   */
  function readFormatOnlineresource(node, objectStack) {
    return pushParseAndPop({}, FORMAT_ONLINERESOURCE_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} Request object.
   */
  function readRequest(node, objectStack) {
    return pushParseAndPop({}, REQUEST_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} DCP type object.
   */
  function readDCPType(node, objectStack) {
    return pushParseAndPop({}, DCPTYPE_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} HTTP object.
   */
  function readHTTP(node, objectStack) {
    return pushParseAndPop({}, HTTP_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} Operation type object.
   */
  function readOperationType(node, objectStack) {
    return pushParseAndPop({}, OPERATIONTYPE_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} Online resource object.
   */
  function readSizedFormatOnlineresource(node, objectStack) {
    const formatOnlineresource = readFormatOnlineresource(node, objectStack);
    if (formatOnlineresource) {
      const size = [
        readNonNegativeIntegerString(node.getAttribute('width')),
        readNonNegativeIntegerString(node.getAttribute('height')),
      ];
      formatOnlineresource['size'] = size;
      return formatOnlineresource;
    }
    return undefined;
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} Authority URL object.
   */
  function readAuthorityURL(node, objectStack) {
    const authorityObject = readFormatOnlineresource(node, objectStack);
    if (authorityObject) {
      authorityObject['name'] = node.getAttribute('name');
      return authorityObject;
    }
    return undefined;
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} Metadata URL object.
   */
  function readMetadataURL(node, objectStack) {
    const metadataObject = readFormatOnlineresource(node, objectStack);
    if (metadataObject) {
      metadataObject['type'] = node.getAttribute('type');
      return metadataObject;
    }
    return undefined;
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} Style object.
   */
  function readStyle$1(node, objectStack) {
    return pushParseAndPop({}, STYLE_PARSERS$1, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Array<string>|undefined} Keyword list.
   */
  function readKeywordList(node, objectStack) {
    return pushParseAndPop([], KEYWORDLIST_PARSERS, node, objectStack);
  }

  /**
   * @module ol/format/WMSGetFeatureInfo
   */

  /**
   * @typedef {Object} Options
   * @property {Array<string>} [layers] If set, only features of the given layers will be returned by the format when read.
   */

  /**
   * @const
   * @type {string}
   */
  const featureIdentifier = '_feature';

  /**
   * @const
   * @type {string}
   */
  const layerIdentifier = '_layer';

  /**
   * @classdesc
   * Format for reading WMSGetFeatureInfo format. It uses
   * {@link module:ol/format/GML2~GML2} to read features.
   *
   * @api
   */
  class WMSGetFeatureInfo extends XMLFeature {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      super();

      options = options ? options : {};

      /**
       * @private
       * @type {string}
       */
      this.featureNS_ = 'http://mapserver.gis.umn.edu/mapserver';

      /**
       * @private
       * @type {GML2}
       */
      this.gmlFormat_ = new GML2();

      /**
       * @private
       * @type {Array<string>|null}
       */
      this.layers_ = options.layers ? options.layers : null;
    }

    /**
     * @return {Array<string>|null} layers
     */
    getLayers() {
      return this.layers_;
    }

    /**
     * @param {Array<string>|null} layers Layers to parse.
     */
    setLayers(layers) {
      this.layers_ = layers;
    }

    /**
     * @param {Element} node Node.
     * @param {Array<*>} objectStack Object stack.
     * @return {Array<import("../Feature.js").default>} Features.
     * @private
     */
    readFeatures_(node, objectStack) {
      node.setAttribute('namespaceURI', this.featureNS_);
      const localName = node.localName;
      /** @type {Array<import("../Feature.js").default>} */
      let features = [];
      if (node.childNodes.length === 0) {
        return features;
      }
      if (localName == 'msGMLOutput') {
        for (let i = 0, ii = node.childNodes.length; i < ii; i++) {
          const layer = node.childNodes[i];
          if (layer.nodeType !== Node.ELEMENT_NODE) {
            continue;
          }

          const layerElement = /** @type {Element} */ (layer);
          const context = objectStack[0];

          const toRemove = layerIdentifier;
          const layerName = layerElement.localName.replace(toRemove, '');

          if (this.layers_ && !this.layers_.includes(layerName)) {
            continue;
          }

          const featureType = layerName + featureIdentifier;

          context['featureType'] = featureType;
          context['featureNS'] = this.featureNS_;

          /** @type {Object<string, import("../xml.js").Parser>} */
          const parsers = {};
          parsers[featureType] = makeArrayPusher(
            this.gmlFormat_.readFeatureElement,
            this.gmlFormat_
          );
          const parsersNS = makeStructureNS(
            [context['featureNS'], null],
            parsers
          );
          layerElement.setAttribute('namespaceURI', this.featureNS_);
          const layerFeatures = pushParseAndPop(
            [],
            // @ts-ignore
            parsersNS,
            layerElement,
            objectStack,
            this.gmlFormat_
          );
          if (layerFeatures) {
            extend$2(features, layerFeatures);
          }
        }
      }
      if (localName == 'FeatureCollection') {
        const gmlFeatures = pushParseAndPop(
          [],
          this.gmlFormat_.FEATURE_COLLECTION_PARSERS,
          node,
          [{}],
          this.gmlFormat_
        );
        if (gmlFeatures) {
          features = gmlFeatures;
        }
      }
      return features;
    }

    /**
     * @protected
     * @param {Element} node Node.
     * @param {import("./Feature.js").ReadOptions} [options] Options.
     * @return {Array<import("../Feature.js").default>} Features.
     */
    readFeaturesFromNode(node, options) {
      const internalOptions = {};
      if (options) {
        Object.assign(internalOptions, this.getReadOptions(node, options));
      }
      return this.readFeatures_(node, [internalOptions]);
    }
  }

  /**
   * @module ol/format/WMTSCapabilities
   */

  /**
   * @const
   * @type {Array<null|string>}
   */
  const NAMESPACE_URIS = [null, 'http://www.opengis.net/wmts/1.0'];

  /**
   * @const
   * @type {Array<null|string>}
   */
  const OWS_NAMESPACE_URIS = [null, 'http://www.opengis.net/ows/1.1'];

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const PARSERS = makeStructureNS(NAMESPACE_URIS, {
    'Contents': makeObjectPropertySetter(readContents),
  });

  /**
   * @classdesc
   * Format for reading WMTS capabilities data.
   *
   * @api
   */
  class WMTSCapabilities extends XML {
    constructor() {
      super();

      /**
       * @type {OWS}
       * @private
       */
      this.owsParser_ = new OWS();
    }

    /**
     * @param {Element} node Node.
     * @return {Object} Object
     */
    readFromNode(node) {
      let version = node.getAttribute('version');
      if (version) {
        version = version.trim();
      }
      let WMTSCapabilityObject = this.owsParser_.readFromNode(node);
      if (!WMTSCapabilityObject) {
        return null;
      }
      WMTSCapabilityObject['version'] = version;
      WMTSCapabilityObject = pushParseAndPop(
        WMTSCapabilityObject,
        PARSERS,
        node,
        []
      );
      return WMTSCapabilityObject ? WMTSCapabilityObject : null;
    }
  }

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const CONTENTS_PARSERS = makeStructureNS(NAMESPACE_URIS, {
    'Layer': makeObjectPropertyPusher(readLayer),
    'TileMatrixSet': makeObjectPropertyPusher(readTileMatrixSet),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const LAYER_PARSERS = makeStructureNS(
    NAMESPACE_URIS,
    {
      'Style': makeObjectPropertyPusher(readStyle),
      'Format': makeObjectPropertyPusher(readString),
      'TileMatrixSetLink': makeObjectPropertyPusher(readTileMatrixSetLink),
      'Dimension': makeObjectPropertyPusher(readDimensions),
      'ResourceURL': makeObjectPropertyPusher(readResourceUrl),
    },
    makeStructureNS(OWS_NAMESPACE_URIS, {
      'Title': makeObjectPropertySetter(readString),
      'Abstract': makeObjectPropertySetter(readString),
      'WGS84BoundingBox': makeObjectPropertySetter(readBoundingBox),
      'Identifier': makeObjectPropertySetter(readString),
    })
  );

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const STYLE_PARSERS = makeStructureNS(
    NAMESPACE_URIS,
    {
      'LegendURL': makeObjectPropertyPusher(readLegendUrl),
    },
    makeStructureNS(OWS_NAMESPACE_URIS, {
      'Title': makeObjectPropertySetter(readString),
      'Identifier': makeObjectPropertySetter(readString),
    })
  );

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const TMS_LINKS_PARSERS = makeStructureNS(NAMESPACE_URIS, {
    'TileMatrixSet': makeObjectPropertySetter(readString),
    'TileMatrixSetLimits': makeObjectPropertySetter(readTileMatrixLimitsList),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const TMS_LIMITS_LIST_PARSERS = makeStructureNS(NAMESPACE_URIS, {
    'TileMatrixLimits': makeArrayPusher(readTileMatrixLimits),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const TMS_LIMITS_PARSERS = makeStructureNS(NAMESPACE_URIS, {
    'TileMatrix': makeObjectPropertySetter(readString),
    'MinTileRow': makeObjectPropertySetter(readPositiveInteger),
    'MaxTileRow': makeObjectPropertySetter(readPositiveInteger),
    'MinTileCol': makeObjectPropertySetter(readPositiveInteger),
    'MaxTileCol': makeObjectPropertySetter(readPositiveInteger),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const DIMENSION_PARSERS = makeStructureNS(
    NAMESPACE_URIS,
    {
      'Default': makeObjectPropertySetter(readString),
      'Value': makeObjectPropertyPusher(readString),
    },
    makeStructureNS(OWS_NAMESPACE_URIS, {
      'Identifier': makeObjectPropertySetter(readString),
    })
  );

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const WGS84_BBOX_READERS = makeStructureNS(OWS_NAMESPACE_URIS, {
    'LowerCorner': makeArrayPusher(readCoordinates),
    'UpperCorner': makeArrayPusher(readCoordinates),
  });

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const TMS_PARSERS = makeStructureNS(
    NAMESPACE_URIS,
    {
      'WellKnownScaleSet': makeObjectPropertySetter(readString),
      'TileMatrix': makeObjectPropertyPusher(readTileMatrix),
    },
    makeStructureNS(OWS_NAMESPACE_URIS, {
      'SupportedCRS': makeObjectPropertySetter(readString),
      'Identifier': makeObjectPropertySetter(readString),
      'BoundingBox': makeObjectPropertySetter(readBoundingBox),
    })
  );

  /**
   * @const
   * @type {Object<string, Object<string, import("../xml.js").Parser>>}
   */
  // @ts-ignore
  const TM_PARSERS = makeStructureNS(
    NAMESPACE_URIS,
    {
      'TopLeftCorner': makeObjectPropertySetter(readCoordinates),
      'ScaleDenominator': makeObjectPropertySetter(readDecimal),
      'TileWidth': makeObjectPropertySetter(readPositiveInteger),
      'TileHeight': makeObjectPropertySetter(readPositiveInteger),
      'MatrixWidth': makeObjectPropertySetter(readPositiveInteger),
      'MatrixHeight': makeObjectPropertySetter(readPositiveInteger),
    },
    makeStructureNS(OWS_NAMESPACE_URIS, {
      'Identifier': makeObjectPropertySetter(readString),
    })
  );

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} Attribution object.
   */
  function readContents(node, objectStack) {
    return pushParseAndPop({}, CONTENTS_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} Layers object.
   */
  function readLayer(node, objectStack) {
    return pushParseAndPop({}, LAYER_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} Tile Matrix Set object.
   */
  function readTileMatrixSet(node, objectStack) {
    return pushParseAndPop({}, TMS_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} Style object.
   */
  function readStyle(node, objectStack) {
    const style = pushParseAndPop({}, STYLE_PARSERS, node, objectStack);
    if (!style) {
      return undefined;
    }
    const isDefault = node.getAttribute('isDefault') === 'true';
    style['isDefault'] = isDefault;
    return style;
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} Tile Matrix Set Link object.
   */
  function readTileMatrixSetLink(node, objectStack) {
    return pushParseAndPop({}, TMS_LINKS_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} Dimension object.
   */
  function readDimensions(node, objectStack) {
    return pushParseAndPop({}, DIMENSION_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} Resource URL object.
   */
  function readResourceUrl(node, objectStack) {
    const format = node.getAttribute('format');
    const template = node.getAttribute('template');
    const resourceType = node.getAttribute('resourceType');
    const resource = {};
    if (format) {
      resource['format'] = format;
    }
    if (template) {
      resource['template'] = template;
    }
    if (resourceType) {
      resource['resourceType'] = resourceType;
    }
    return resource;
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} BBox object.
   */
  function readBoundingBox(node, objectStack) {
    const coordinates = pushParseAndPop(
      [],
      WGS84_BBOX_READERS,
      node,
      objectStack
    );
    if (coordinates.length != 2) {
      return undefined;
    }
    return boundingExtent(coordinates);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} Legend object.
   */
  function readLegendUrl(node, objectStack) {
    const legend = {};
    legend['format'] = node.getAttribute('format');
    legend['href'] = readHref(node);
    return legend;
  }

  /**
   * @param {Node} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} Coordinates object.
   */
  function readCoordinates(node, objectStack) {
    const coordinates = readString(node).split(/\s+/);
    if (!coordinates || coordinates.length != 2) {
      return undefined;
    }
    const x = +coordinates[0];
    const y = +coordinates[1];
    if (isNaN(x) || isNaN(y)) {
      return undefined;
    }
    return [x, y];
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} TileMatrix object.
   */
  function readTileMatrix(node, objectStack) {
    return pushParseAndPop({}, TM_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} TileMatrixSetLimits Object.
   */
  function readTileMatrixLimitsList(node, objectStack) {
    return pushParseAndPop([], TMS_LIMITS_LIST_PARSERS, node, objectStack);
  }

  /**
   * @param {Element} node Node.
   * @param {Array<*>} objectStack Object stack.
   * @return {Object|undefined} TileMatrixLimits Array.
   */
  function readTileMatrixLimits(node, objectStack) {
    return pushParseAndPop({}, TMS_LIMITS_PARSERS, node, objectStack);
  }

  /**
   * @module ol/control/FullScreen
   */

  const events = [
    'fullscreenchange',
    'webkitfullscreenchange',
    'MSFullscreenChange',
  ];

  /**
   * @enum {string}
   */
  const FullScreenEventType = {
    /**
     * Triggered after the map entered fullscreen.
     * @event FullScreenEventType#enterfullscreen
     * @api
     */
    ENTERFULLSCREEN: 'enterfullscreen',

    /**
     * Triggered after the map leave fullscreen.
     * @event FullScreenEventType#leavefullscreen
     * @api
     */
    LEAVEFULLSCREEN: 'leavefullscreen',
  };

  /***
   * @template Return
   * @typedef {import("../Observable").OnSignature<import("../Observable").EventTypes|
   *     'enterfullscreen'|'leavefullscreen', import("../events/Event.js").default, Return> &
   *   import("../Observable").OnSignature<import("../ObjectEventType").Types, import("../Object").ObjectEvent, Return> &
   *   import("../Observable").CombinedOnSignature<import("../Observable").EventTypes|
   *     'enterfullscreen'|'leavefullscreen'|import("../ObjectEventType").Types, Return>} FullScreenOnSignature
   */

  /**
   * @typedef {Object} Options
   * @property {string} [className='ol-full-screen'] CSS class name.
   * @property {string|Text|HTMLElement} [label='\u2922'] Text label to use for the button.
   * Instead of text, also an element (e.g. a `span` element) can be used.
   * @property {string|Text|HTMLElement} [labelActive='\u00d7'] Text label to use for the
   * button when full-screen is active.
   * Instead of text, also an element (e.g. a `span` element) can be used.
   * @property {string} [activeClassName=className + '-true'] CSS class name for the button
   * when full-screen is active.
   * @property {string} [inactiveClassName=className + '-false'] CSS class name for the button
   * when full-screen is inactive.
   * @property {string} [tipLabel='Toggle full-screen'] Text label to use for the button tip.
   * @property {boolean} [keys=false] Full keyboard access.
   * @property {HTMLElement|string} [target] Specify a target if you want the
   * control to be rendered outside of the map's viewport.
   * @property {HTMLElement|string} [source] The element to be displayed
   * fullscreen. When not provided, the element containing the map viewport will
   * be displayed fullscreen.
   */

  /**
   * @classdesc
   * Provides a button that when clicked fills up the full screen with the map.
   * The full screen source element is by default the element containing the map viewport unless
   * overridden by providing the `source` option. In which case, the dom
   * element introduced using this parameter will be displayed in full screen.
   *
   * When in full screen mode, a close button is shown to exit full screen mode.
   * The [Fullscreen API](https://www.w3.org/TR/fullscreen/) is used to
   * toggle the map in full screen mode.
   *
   * @fires FullScreenEventType#enterfullscreen
   * @fires FullScreenEventType#leavefullscreen
   * @api
   */
  class FullScreen extends Control {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      options = options ? options : {};

      super({
        element: document.createElement('div'),
        target: options.target,
      });

      /***
       * @type {FullScreenOnSignature<import("../events").EventsKey>}
       */
      this.on;

      /***
       * @type {FullScreenOnSignature<import("../events").EventsKey>}
       */
      this.once;

      /***
       * @type {FullScreenOnSignature<void>}
       */
      this.un;

      /**
       * @private
       * @type {boolean}
       */
      this.keys_ = options.keys !== undefined ? options.keys : false;

      /**
       * @private
       * @type {HTMLElement|string|undefined}
       */
      this.source_ = options.source;

      /**
       * @type {boolean}
       * @private
       */
      this.isInFullscreen_ = false;

      /**
       * @private
       */
      this.boundHandleMapTargetChange_ = this.handleMapTargetChange_.bind(this);

      /**
       * @private
       * @type {string}
       */
      this.cssClassName_ =
        options.className !== undefined ? options.className : 'ol-full-screen';

      /**
       * @private
       * @type {Array<import("../events.js").EventsKey>}
       */
      this.documentListeners_ = [];

      /**
       * @private
       * @type {Array<string>}
       */
      this.activeClassName_ =
        options.activeClassName !== undefined
          ? options.activeClassName.split(' ')
          : [this.cssClassName_ + '-true'];

      /**
       * @private
       * @type {Array<string>}
       */
      this.inactiveClassName_ =
        options.inactiveClassName !== undefined
          ? options.inactiveClassName.split(' ')
          : [this.cssClassName_ + '-false'];

      const label = options.label !== undefined ? options.label : '\u2922';

      /**
       * @private
       * @type {Text|HTMLElement}
       */
      this.labelNode_ =
        typeof label === 'string' ? document.createTextNode(label) : label;

      const labelActive =
        options.labelActive !== undefined ? options.labelActive : '\u00d7';

      /**
       * @private
       * @type {Text|HTMLElement}
       */
      this.labelActiveNode_ =
        typeof labelActive === 'string'
          ? document.createTextNode(labelActive)
          : labelActive;

      const tipLabel = options.tipLabel ? options.tipLabel : 'Toggle full-screen';

      /**
       * @private
       * @type {HTMLElement}
       */
      this.button_ = document.createElement('button');
      this.button_.title = tipLabel;
      this.button_.setAttribute('type', 'button');
      this.button_.appendChild(this.labelNode_);
      this.button_.addEventListener(
        EventType$1.CLICK,
        this.handleClick_.bind(this),
        false
      );
      this.setClassName_(this.button_, this.isInFullscreen_);

      this.element.className = `${this.cssClassName_} ${CLASS_UNSELECTABLE} ${CLASS_CONTROL}`;
      this.element.appendChild(this.button_);
    }

    /**
     * @param {MouseEvent} event The event to handle
     * @private
     */
    handleClick_(event) {
      event.preventDefault();
      this.handleFullScreen_();
    }

    /**
     * @private
     */
    handleFullScreen_() {
      const map = this.getMap();
      if (!map) {
        return;
      }
      const doc = map.getOwnerDocument();
      if (!isFullScreenSupported(doc)) {
        return;
      }
      if (isFullScreen(doc)) {
        exitFullScreen(doc);
      } else {
        let element;
        if (this.source_) {
          element =
            typeof this.source_ === 'string'
              ? doc.getElementById(this.source_)
              : this.source_;
        } else {
          element = map.getTargetElement();
        }
        if (this.keys_) {
          requestFullScreenWithKeys(element);
        } else {
          requestFullScreen(element);
        }
      }
    }

    /**
     * @private
     */
    handleFullScreenChange_() {
      const map = this.getMap();
      if (!map) {
        return;
      }
      const wasInFullscreen = this.isInFullscreen_;
      this.isInFullscreen_ = isFullScreen(map.getOwnerDocument());
      if (wasInFullscreen !== this.isInFullscreen_) {
        this.setClassName_(this.button_, this.isInFullscreen_);
        if (this.isInFullscreen_) {
          replaceNode(this.labelActiveNode_, this.labelNode_);
          this.dispatchEvent(FullScreenEventType.ENTERFULLSCREEN);
        } else {
          replaceNode(this.labelNode_, this.labelActiveNode_);
          this.dispatchEvent(FullScreenEventType.LEAVEFULLSCREEN);
        }
        map.updateSize();
      }
    }

    /**
     * @param {HTMLElement} element Target element
     * @param {boolean} fullscreen True if fullscreen class name should be active
     * @private
     */
    setClassName_(element, fullscreen) {
      if (fullscreen) {
        element.classList.remove(...this.inactiveClassName_);
        element.classList.add(...this.activeClassName_);
      } else {
        element.classList.remove(...this.activeClassName_);
        element.classList.add(...this.inactiveClassName_);
      }
    }

    /**
     * Remove the control from its current map and attach it to the new map.
     * Pass `null` to just remove the control from the current map.
     * Subclasses may set up event handlers to get notified about changes to
     * the map here.
     * @param {import("../Map.js").default|null} map Map.
     * @api
     */
    setMap(map) {
      const oldMap = this.getMap();
      if (oldMap) {
        oldMap.removeChangeListener(
          MapProperty.TARGET,
          this.boundHandleMapTargetChange_
        );
      }

      super.setMap(map);

      this.handleMapTargetChange_();
      if (map) {
        map.addChangeListener(
          MapProperty.TARGET,
          this.boundHandleMapTargetChange_
        );
      }
    }

    /**
     * @private
     */
    handleMapTargetChange_() {
      const listeners = this.documentListeners_;
      for (let i = 0, ii = listeners.length; i < ii; ++i) {
        unlistenByKey(listeners[i]);
      }
      listeners.length = 0;

      const map = this.getMap();
      if (map) {
        const doc = map.getOwnerDocument();
        if (isFullScreenSupported(doc)) {
          this.element.classList.remove(CLASS_UNSUPPORTED);
        } else {
          this.element.classList.add(CLASS_UNSUPPORTED);
        }

        for (let i = 0, ii = events.length; i < ii; ++i) {
          listeners.push(
            listen(doc, events[i], this.handleFullScreenChange_, this)
          );
        }
        this.handleFullScreenChange_();
      }
    }
  }

  /**
   * @param {Document} doc The root document to check.
   * @return {boolean} Fullscreen is supported by the current platform.
   */
  function isFullScreenSupported(doc) {
    const body = doc.body;
    return !!(
      body['webkitRequestFullscreen'] ||
      (body.requestFullscreen && doc.fullscreenEnabled)
    );
  }

  /**
   * @param {Document} doc The root document to check.
   * @return {boolean} Element is currently in fullscreen.
   */
  function isFullScreen(doc) {
    return !!(doc['webkitIsFullScreen'] || doc.fullscreenElement);
  }

  /**
   * Request to fullscreen an element.
   * @param {HTMLElement} element Element to request fullscreen
   */
  function requestFullScreen(element) {
    if (element.requestFullscreen) {
      element.requestFullscreen();
    } else if (element['webkitRequestFullscreen']) {
      element['webkitRequestFullscreen']();
    }
  }

  /**
   * Request to fullscreen an element with keyboard input.
   * @param {HTMLElement} element Element to request fullscreen
   */
  function requestFullScreenWithKeys(element) {
    if (element['webkitRequestFullscreen']) {
      element['webkitRequestFullscreen']();
    } else {
      requestFullScreen(element);
    }
  }

  /**
   * Exit fullscreen.
   * @param {Document} doc The document to exit fullscren from
   */
  function exitFullScreen(doc) {
    if (doc.exitFullscreen) {
      doc.exitFullscreen();
    } else if (doc['webkitExitFullscreen']) {
      doc['webkitExitFullscreen']();
    }
  }

  /**
   * @module ol/control/MousePosition
   */

  /**
   * @type {string}
   */
  const PROJECTION = 'projection';

  /**
   * @type {string}
   */
  const COORDINATE_FORMAT = 'coordinateFormat';

  /***
   * @template Return
   * @typedef {import("../Observable").OnSignature<import("../Observable").EventTypes, import("../events/Event.js").default, Return> &
   *   import("../Observable").OnSignature<import("../ObjectEventType").Types|
   *     'change:coordinateFormat'|'change:projection', import("../Object").ObjectEvent, Return> &
   *   import("../Observable").CombinedOnSignature<import("../Observable").EventTypes|import("../ObjectEventType").Types|
   *     'change:coordinateFormat'|'change:projection', Return>} MousePositionOnSignature
   */

  /**
   * @typedef {Object} Options
   * @property {string} [className='ol-mouse-position'] CSS class name.
   * @property {import("../coordinate.js").CoordinateFormat} [coordinateFormat] Coordinate format.
   * @property {import("../proj.js").ProjectionLike} [projection] Projection. Default is the view projection.
   * @property {function(import("../MapEvent.js").default):void} [render] Function called when the
   * control should be re-rendered. This is called in a `requestAnimationFrame`
   * callback.
   * @property {HTMLElement|string} [target] Specify a target if you want the
   * control to be rendered outside of the map's viewport.
   * @property {string} [placeholder] Markup to show when the mouse position is not
   * available (e.g. when the pointer leaves the map viewport).  By default, a non-breaking space is rendered
   * initially and the last position is retained when the mouse leaves the viewport.
   * When a string is provided (e.g. `'no position'` or `''` for an empty string) it is used as a
   * placeholder.
   */

  /**
   * @classdesc
   * A control to show the 2D coordinates of the mouse cursor. By default, these
   * are in the view projection, but can be in any supported projection.
   * By default the control is shown in the top right corner of the map, but this
   * can be changed by using the css selector `.ol-mouse-position`.
   *
   * On touch devices, which usually do not have a mouse cursor, the coordinates
   * of the currently touched position are shown.
   *
   * @api
   */
  class MousePosition extends Control {
    /**
     * @param {Options} [options] Mouse position options.
     */
    constructor(options) {
      options = options ? options : {};

      const element = document.createElement('div');
      element.className =
        options.className !== undefined ? options.className : 'ol-mouse-position';

      super({
        element: element,
        render: options.render,
        target: options.target,
      });

      /***
       * @type {MousePositionOnSignature<import("../events").EventsKey>}
       */
      this.on;

      /***
       * @type {MousePositionOnSignature<import("../events").EventsKey>}
       */
      this.once;

      /***
       * @type {MousePositionOnSignature<void>}
       */
      this.un;

      this.addChangeListener(PROJECTION, this.handleProjectionChanged_);

      if (options.coordinateFormat) {
        this.setCoordinateFormat(options.coordinateFormat);
      }
      if (options.projection) {
        this.setProjection(options.projection);
      }

      /**
       * @private
       * @type {boolean}
       */
      this.renderOnMouseOut_ = options.placeholder !== undefined;

      /**
       * @private
       * @type {string}
       */
      this.placeholder_ = this.renderOnMouseOut_ ? options.placeholder : '&#160;';

      /**
       * @private
       * @type {string}
       */
      this.renderedHTML_ = element.innerHTML;

      /**
       * @private
       * @type {?import("../proj/Projection.js").default}
       */
      this.mapProjection_ = null;

      /**
       * @private
       * @type {?import("../proj.js").TransformFunction}
       */
      this.transform_ = null;
    }

    /**
     * @private
     */
    handleProjectionChanged_() {
      this.transform_ = null;
    }

    /**
     * Return the coordinate format type used to render the current position or
     * undefined.
     * @return {import("../coordinate.js").CoordinateFormat|undefined} The format to render the current
     *     position in.
     * @observable
     * @api
     */
    getCoordinateFormat() {
      return /** @type {import("../coordinate.js").CoordinateFormat|undefined} */ (
        this.get(COORDINATE_FORMAT)
      );
    }

    /**
     * Return the projection that is used to report the mouse position.
     * @return {import("../proj/Projection.js").default|undefined} The projection to report mouse
     *     position in.
     * @observable
     * @api
     */
    getProjection() {
      return /** @type {import("../proj/Projection.js").default|undefined} */ (
        this.get(PROJECTION)
      );
    }

    /**
     * @param {MouseEvent} event Browser event.
     * @protected
     */
    handleMouseMove(event) {
      const map = this.getMap();
      this.updateHTML_(map.getEventPixel(event));
    }

    /**
     * @param {Event} event Browser event.
     * @protected
     */
    handleMouseOut(event) {
      this.updateHTML_(null);
    }

    /**
     * Remove the control from its current map and attach it to the new map.
     * Pass `null` to just remove the control from the current map.
     * Subclasses may set up event handlers to get notified about changes to
     * the map here.
     * @param {import("../Map.js").default|null} map Map.
     * @api
     */
    setMap(map) {
      super.setMap(map);
      if (map) {
        const viewport = map.getViewport();
        this.listenerKeys.push(
          listen(viewport, PointerEventType.POINTERMOVE, this.handleMouseMove, this)
        );
        if (this.renderOnMouseOut_) {
          this.listenerKeys.push(
            listen(viewport, PointerEventType.POINTEROUT, this.handleMouseOut, this)
          );
        }
        this.updateHTML_(null);
      }
    }

    /**
     * Set the coordinate format type used to render the current position.
     * @param {import("../coordinate.js").CoordinateFormat} format The format to render the current
     *     position in.
     * @observable
     * @api
     */
    setCoordinateFormat(format) {
      this.set(COORDINATE_FORMAT, format);
    }

    /**
     * Set the projection that is used to report the mouse position.
     * @param {import("../proj.js").ProjectionLike} projection The projection to report mouse
     *     position in.
     * @observable
     * @api
     */
    setProjection(projection) {
      this.set(PROJECTION, get$2(projection));
    }

    /**
     * @param {?import("../pixel.js").Pixel} pixel Pixel.
     * @private
     */
    updateHTML_(pixel) {
      let html = this.placeholder_;
      if (pixel && this.mapProjection_) {
        if (!this.transform_) {
          const projection = this.getProjection();
          if (projection) {
            this.transform_ = getTransformFromProjections(
              this.mapProjection_,
              projection
            );
          } else {
            this.transform_ = identityTransform;
          }
        }
        const map = this.getMap();
        const coordinate = map.getCoordinateFromPixelInternal(pixel);
        if (coordinate) {
          const userProjection = getUserProjection();
          if (userProjection) {
            this.transform_ = getTransformFromProjections(
              this.mapProjection_,
              userProjection
            );
          }
          this.transform_(coordinate, coordinate);
          const coordinateFormat = this.getCoordinateFormat();
          if (coordinateFormat) {
            html = coordinateFormat(coordinate);
          } else {
            html = coordinate.toString();
          }
        }
      }
      if (!this.renderedHTML_ || html !== this.renderedHTML_) {
        this.element.innerHTML = html;
        this.renderedHTML_ = html;
      }
    }

    /**
     * Update the projection. Rendering of the coordinates is done in
     * `handleMouseMove` and `handleMouseUp`.
     * @param {import("../MapEvent.js").default} mapEvent Map event.
     * @override
     */
    render(mapEvent) {
      const frameState = mapEvent.frameState;
      if (!frameState) {
        this.mapProjection_ = null;
      } else {
        if (this.mapProjection_ != frameState.viewState.projection) {
          this.mapProjection_ = frameState.viewState.projection;
          this.transform_ = null;
        }
      }
    }
  }

  /**
   * @module ol/control/OverviewMap
   */

  /**
   * Maximum width and/or height extent ratio that determines when the overview
   * map should be zoomed out.
   * @type {number}
   */
  const MAX_RATIO = 0.75;

  /**
   * Minimum width and/or height extent ratio that determines when the overview
   * map should be zoomed in.
   * @type {number}
   */
  const MIN_RATIO = 0.1;

  /**
   * @typedef {Object} Options
   * @property {string} [className='ol-overviewmap'] CSS class name.
   * @property {boolean} [collapsed=true] Whether the control should start collapsed or not (expanded).
   * @property {string|HTMLElement} [collapseLabel='‹'] Text label to use for the
   * expanded overviewmap button. Instead of text, also an element (e.g. a `span` element) can be used.
   * @property {boolean} [collapsible=true] Whether the control can be collapsed or not.
   * @property {string|HTMLElement} [label='›'] Text label to use for the collapsed
   * overviewmap button. Instead of text, also an element (e.g. a `span` element) can be used.
   * @property {Array<import("../layer/Base.js").default>|import("../Collection.js").default<import("../layer/Base.js").default>} [layers]
   * Layers for the overview map.
   * @property {function(import("../MapEvent.js").default):void} [render] Function called when the control
   * should be re-rendered. This is called in a `requestAnimationFrame` callback.
   * @property {boolean} [rotateWithView=false] Whether the control view should rotate with the main map view.
   * @property {HTMLElement|string} [target] Specify a target if you want the control
   * to be rendered outside of the map's viewport.
   * @property {string} [tipLabel='Overview map'] Text label to use for the button tip.
   * @property {View} [view] Custom view for the overview map (should use same projection as main map). If not provided,
   * a default view with the same projection as the main map will be used.
   */

  /**
   * Create a new control with a map acting as an overview map for another
   * defined map.
   *
   * @api
   */
  class OverviewMap extends Control {
    /**
     * @param {Options} [options] OverviewMap options.
     */
    constructor(options) {
      options = options ? options : {};

      super({
        element: document.createElement('div'),
        render: options.render,
        target: options.target,
      });

      /**
       * @private
       */
      this.boundHandleRotationChanged_ = this.handleRotationChanged_.bind(this);

      /**
       * @type {boolean}
       * @private
       */
      this.collapsed_ =
        options.collapsed !== undefined ? options.collapsed : true;

      /**
       * @private
       * @type {boolean}
       */
      this.collapsible_ =
        options.collapsible !== undefined ? options.collapsible : true;

      if (!this.collapsible_) {
        this.collapsed_ = false;
      }

      /**
       * @private
       * @type {boolean}
       */
      this.rotateWithView_ =
        options.rotateWithView !== undefined ? options.rotateWithView : false;

      /**
       * @private
       * @type {import("../extent.js").Extent|undefined}
       */
      this.viewExtent_ = undefined;

      const className =
        options.className !== undefined ? options.className : 'ol-overviewmap';

      const tipLabel =
        options.tipLabel !== undefined ? options.tipLabel : 'Overview map';

      const collapseLabel =
        options.collapseLabel !== undefined ? options.collapseLabel : '\u2039';

      if (typeof collapseLabel === 'string') {
        /**
         * @private
         * @type {HTMLElement}
         */
        this.collapseLabel_ = document.createElement('span');
        this.collapseLabel_.textContent = collapseLabel;
      } else {
        this.collapseLabel_ = collapseLabel;
      }

      const label = options.label !== undefined ? options.label : '\u203A';

      if (typeof label === 'string') {
        /**
         * @private
         * @type {HTMLElement}
         */
        this.label_ = document.createElement('span');
        this.label_.textContent = label;
      } else {
        this.label_ = label;
      }

      const activeLabel =
        this.collapsible_ && !this.collapsed_ ? this.collapseLabel_ : this.label_;
      const button = document.createElement('button');
      button.setAttribute('type', 'button');
      button.title = tipLabel;
      button.appendChild(activeLabel);

      button.addEventListener(
        EventType$1.CLICK,
        this.handleClick_.bind(this),
        false
      );

      /**
       * @type {HTMLElement}
       * @private
       */
      this.ovmapDiv_ = document.createElement('div');
      this.ovmapDiv_.className = 'ol-overviewmap-map';

      /**
       * Explicitly given view to be used instead of a view derived from the main map.
       * @type {View}
       * @private
       */
      this.view_ = options.view;

      const ovmap = new Map({
        view: options.view,
        controls: new Collection(),
        interactions: new Collection(),
      });

      /**
       * @type {Map}
       * @private
       */
      this.ovmap_ = ovmap;

      if (options.layers) {
        options.layers.forEach(function (layer) {
          ovmap.addLayer(layer);
        });
      }

      const box = document.createElement('div');
      box.className = 'ol-overviewmap-box';
      box.style.boxSizing = 'border-box';

      /**
       * @type {import("../Overlay.js").default}
       * @private
       */
      this.boxOverlay_ = new Overlay({
        position: [0, 0],
        positioning: 'center-center',
        element: box,
      });
      this.ovmap_.addOverlay(this.boxOverlay_);

      const cssClasses =
        className +
        ' ' +
        CLASS_UNSELECTABLE +
        ' ' +
        CLASS_CONTROL +
        (this.collapsed_ && this.collapsible_ ? ' ' + CLASS_COLLAPSED : '') +
        (this.collapsible_ ? '' : ' ol-uncollapsible');
      const element = this.element;
      element.className = cssClasses;
      element.appendChild(this.ovmapDiv_);
      element.appendChild(button);

      /* Interactive map */

      const scope = this;

      const overlay = this.boxOverlay_;
      const overlayBox = this.boxOverlay_.getElement();

      /* Functions definition */

      const computeDesiredMousePosition = function (mousePosition) {
        return {
          clientX: mousePosition.clientX,
          clientY: mousePosition.clientY,
        };
      };

      const move = function (event) {
        const position = /** @type {?} */ (computeDesiredMousePosition(event));
        const coordinates = ovmap.getEventCoordinateInternal(
          /** @type {MouseEvent} */ (position)
        );

        overlay.setPosition(coordinates);
      };

      const endMoving = function (event) {
        const coordinates = ovmap.getEventCoordinateInternal(event);

        scope.getMap().getView().setCenterInternal(coordinates);

        window.removeEventListener('mousemove', move);
        window.removeEventListener('mouseup', endMoving);
      };

      /* Binding */

      overlayBox.addEventListener('mousedown', function () {
        window.addEventListener('mousemove', move);
        window.addEventListener('mouseup', endMoving);
      });
    }

    /**
     * Remove the control from its current map and attach it to the new map.
     * Pass `null` to just remove the control from the current map.
     * Subclasses may set up event handlers to get notified about changes to
     * the map here.
     * @param {import("../Map.js").default|null} map Map.
     * @api
     */
    setMap(map) {
      const oldMap = this.getMap();
      if (map === oldMap) {
        return;
      }
      if (oldMap) {
        const oldView = oldMap.getView();
        if (oldView) {
          this.unbindView_(oldView);
        }
        this.ovmap_.setTarget(null);
      }
      super.setMap(map);

      if (map) {
        this.ovmap_.setTarget(this.ovmapDiv_);
        this.listenerKeys.push(
          listen(
            map,
            ObjectEventType.PROPERTYCHANGE,
            this.handleMapPropertyChange_,
            this
          )
        );

        const view = map.getView();
        if (view) {
          this.bindView_(view);
          if (view.isDef()) {
            this.ovmap_.updateSize();
            this.resetExtent_();
          }
        }

        if (!this.ovmap_.isRendered()) {
          this.updateBoxAfterOvmapIsRendered_();
        }
      }
    }

    /**
     * Handle map property changes.  This only deals with changes to the map's view.
     * @param {import("../Object.js").ObjectEvent} event The propertychange event.
     * @private
     */
    handleMapPropertyChange_(event) {
      if (event.key === MapProperty.VIEW) {
        const oldView = /** @type {import("../View.js").default} */ (
          event.oldValue
        );
        if (oldView) {
          this.unbindView_(oldView);
        }
        const newView = this.getMap().getView();
        this.bindView_(newView);
      } else if (
        !this.ovmap_.isRendered() &&
        (event.key === MapProperty.TARGET || event.key === MapProperty.SIZE)
      ) {
        this.ovmap_.updateSize();
      }
    }

    /**
     * Register listeners for view property changes.
     * @param {import("../View.js").default} view The view.
     * @private
     */
    bindView_(view) {
      if (!this.view_) {
        // Unless an explicit view definition was given, derive default from whatever main map uses.
        const newView = new View({
          projection: view.getProjection(),
        });
        this.ovmap_.setView(newView);
      }

      view.addChangeListener(
        ViewProperty.ROTATION,
        this.boundHandleRotationChanged_
      );
      // Sync once with the new view
      this.handleRotationChanged_();
    }

    /**
     * Unregister listeners for view property changes.
     * @param {import("../View.js").default} view The view.
     * @private
     */
    unbindView_(view) {
      view.removeChangeListener(
        ViewProperty.ROTATION,
        this.boundHandleRotationChanged_
      );
    }

    /**
     * Handle rotation changes to the main map.
     * @private
     */
    handleRotationChanged_() {
      if (this.rotateWithView_) {
        this.ovmap_.getView().setRotation(this.getMap().getView().getRotation());
      }
    }

    /**
     * Reset the overview map extent if the box size (width or
     * height) is less than the size of the overview map size times minRatio
     * or is greater than the size of the overview size times maxRatio.
     *
     * If the map extent was not reset, the box size can fits in the defined
     * ratio sizes. This method then checks if is contained inside the overview
     * map current extent. If not, recenter the overview map to the current
     * main map center location.
     * @private
     */
    validateExtent_() {
      const map = this.getMap();
      const ovmap = this.ovmap_;

      if (!map.isRendered() || !ovmap.isRendered()) {
        return;
      }

      const mapSize = /** @type {import("../size.js").Size} */ (map.getSize());

      const view = map.getView();
      const extent = view.calculateExtentInternal(mapSize);

      if (this.viewExtent_ && equals$2(extent, this.viewExtent_)) {
        // repeats of the same extent may indicate constraint conflicts leading to an endless cycle
        return;
      }
      this.viewExtent_ = extent;

      const ovmapSize = /** @type {import("../size.js").Size} */ (
        ovmap.getSize()
      );

      const ovview = ovmap.getView();
      const ovextent = ovview.calculateExtentInternal(ovmapSize);

      const topLeftPixel = ovmap.getPixelFromCoordinateInternal(
        getTopLeft(extent)
      );
      const bottomRightPixel = ovmap.getPixelFromCoordinateInternal(
        getBottomRight(extent)
      );

      const boxWidth = Math.abs(topLeftPixel[0] - bottomRightPixel[0]);
      const boxHeight = Math.abs(topLeftPixel[1] - bottomRightPixel[1]);

      const ovmapWidth = ovmapSize[0];
      const ovmapHeight = ovmapSize[1];

      if (
        boxWidth < ovmapWidth * MIN_RATIO ||
        boxHeight < ovmapHeight * MIN_RATIO ||
        boxWidth > ovmapWidth * MAX_RATIO ||
        boxHeight > ovmapHeight * MAX_RATIO
      ) {
        this.resetExtent_();
      } else if (!containsExtent(ovextent, extent)) {
        this.recenter_();
      }
    }

    /**
     * Reset the overview map extent to half calculated min and max ratio times
     * the extent of the main map.
     * @private
     */
    resetExtent_() {

      const map = this.getMap();
      const ovmap = this.ovmap_;

      const mapSize = /** @type {import("../size.js").Size} */ (map.getSize());

      const view = map.getView();
      const extent = view.calculateExtentInternal(mapSize);

      const ovview = ovmap.getView();

      // get how many times the current map overview could hold different
      // box sizes using the min and max ratio, pick the step in the middle used
      // to calculate the extent from the main map to set it to the overview map,
      const steps = Math.log(MAX_RATIO / MIN_RATIO) / Math.LN2;
      const ratio = 1 / (Math.pow(2, steps / 2) * MIN_RATIO);
      scaleFromCenter(extent, ratio);
      ovview.fitInternal(fromExtent(extent));
    }

    /**
     * Set the center of the overview map to the map center without changing its
     * resolution.
     * @private
     */
    recenter_() {
      const map = this.getMap();
      const ovmap = this.ovmap_;

      const view = map.getView();

      const ovview = ovmap.getView();

      ovview.setCenterInternal(view.getCenterInternal());
    }

    /**
     * Update the box using the main map extent
     * @private
     */
    updateBox_() {
      const map = this.getMap();
      const ovmap = this.ovmap_;

      if (!map.isRendered() || !ovmap.isRendered()) {
        return;
      }

      const mapSize = /** @type {import("../size.js").Size} */ (map.getSize());

      const view = map.getView();

      const ovview = ovmap.getView();

      const rotation = this.rotateWithView_ ? 0 : -view.getRotation();

      const overlay = this.boxOverlay_;
      const box = this.boxOverlay_.getElement();
      const center = view.getCenterInternal();
      const resolution = view.getResolution();
      const ovresolution = ovview.getResolution();
      const width = (mapSize[0] * resolution) / ovresolution;
      const height = (mapSize[1] * resolution) / ovresolution;

      // set position using center coordinates
      overlay.setPosition(center);

      // set box size calculated from map extent size and overview map resolution
      if (box) {
        box.style.width = width + 'px';
        box.style.height = height + 'px';
        const transform = 'rotate(' + rotation + 'rad)';
        box.style.transform = transform;
      }
    }

    /**
     * @private
     */
    updateBoxAfterOvmapIsRendered_() {
      if (this.ovmapPostrenderKey_) {
        return;
      }
      this.ovmapPostrenderKey_ = listenOnce(
        this.ovmap_,
        MapEventType.POSTRENDER,
        function (event) {
          delete this.ovmapPostrenderKey_;
          this.updateBox_();
        },
        this
      );
    }

    /**
     * @param {MouseEvent} event The event to handle
     * @private
     */
    handleClick_(event) {
      event.preventDefault();
      this.handleToggle_();
    }

    /**
     * @private
     */
    handleToggle_() {
      this.element.classList.toggle(CLASS_COLLAPSED);
      if (this.collapsed_) {
        replaceNode(this.collapseLabel_, this.label_);
      } else {
        replaceNode(this.label_, this.collapseLabel_);
      }
      this.collapsed_ = !this.collapsed_;

      // manage overview map if it had not been rendered before and control
      // is expanded
      const ovmap = this.ovmap_;
      if (!this.collapsed_) {
        if (ovmap.isRendered()) {
          this.viewExtent_ = undefined;
          ovmap.render();
          return;
        }
        ovmap.updateSize();
        this.resetExtent_();
        this.updateBoxAfterOvmapIsRendered_();
      }
    }

    /**
     * Return `true` if the overview map is collapsible, `false` otherwise.
     * @return {boolean} True if the widget is collapsible.
     * @api
     */
    getCollapsible() {
      return this.collapsible_;
    }

    /**
     * Set whether the overview map should be collapsible.
     * @param {boolean} collapsible True if the widget is collapsible.
     * @api
     */
    setCollapsible(collapsible) {
      if (this.collapsible_ === collapsible) {
        return;
      }
      this.collapsible_ = collapsible;
      this.element.classList.toggle('ol-uncollapsible');
      if (!collapsible && this.collapsed_) {
        this.handleToggle_();
      }
    }

    /**
     * Collapse or expand the overview map according to the passed parameter. Will
     * not do anything if the overview map isn't collapsible or if the current
     * collapsed state is already the one requested.
     * @param {boolean} collapsed True if the widget is collapsed.
     * @api
     */
    setCollapsed(collapsed) {
      if (!this.collapsible_ || this.collapsed_ === collapsed) {
        return;
      }
      this.handleToggle_();
    }

    /**
     * Determine if the overview map is collapsed.
     * @return {boolean} The overview map is collapsed.
     * @api
     */
    getCollapsed() {
      return this.collapsed_;
    }

    /**
     * Return `true` if the overview map view can rotate, `false` otherwise.
     * @return {boolean} True if the control view can rotate.
     * @api
     */
    getRotateWithView() {
      return this.rotateWithView_;
    }

    /**
     * Set whether the overview map view should rotate with the main map view.
     * @param {boolean} rotateWithView True if the control view should rotate.
     * @api
     */
    setRotateWithView(rotateWithView) {
      if (this.rotateWithView_ === rotateWithView) {
        return;
      }
      this.rotateWithView_ = rotateWithView;
      if (this.getMap().getView().getRotation() !== 0) {
        if (this.rotateWithView_) {
          this.handleRotationChanged_();
        } else {
          this.ovmap_.getView().setRotation(0);
        }
        this.viewExtent_ = undefined;
        this.validateExtent_();
        this.updateBox_();
      }
    }

    /**
     * Return the overview map.
     * @return {import("../Map.js").default} Overview map.
     * @api
     */
    getOverviewMap() {
      return this.ovmap_;
    }

    /**
     * Update the overview map element.
     * @param {import("../MapEvent.js").default} mapEvent Map event.
     * @override
     */
    render(mapEvent) {
      this.validateExtent_();
      this.updateBox_();
    }
  }

  /**
   * @module ol/control/ScaleLine
   */

  /**
   * @type {string}
   */
  const UNITS_PROP = 'units';

  /**
   * @typedef {'degrees' | 'imperial' | 'nautical' | 'metric' | 'us'} Units
   * Units for the scale line.
   */

  /**
   * @const
   * @type {Array<number>}
   */
  const LEADING_DIGITS = [1, 2, 5];

  /**
   * @const
   * @type {number}
   */
  const DEFAULT_DPI = 25.4 / 0.28;

  /***
   * @template Return
   * @typedef {import("../Observable").OnSignature<import("../Observable").EventTypes, import("../events/Event.js").default, Return> &
   *   import("../Observable").OnSignature<import("../ObjectEventType").Types|
   *     'change:units', import("../Object").ObjectEvent, Return> &
   *   import("../Observable").CombinedOnSignature<import("../Observable").EventTypes|import("../ObjectEventType").Types
   *     |'change:units', Return>} ScaleLineOnSignature
   */

  /**
   * @typedef {Object} Options
   * @property {string} [className] CSS class name. The default is `ol-scale-bar` when configured with
   * `bar: true`. Otherwise the default is `ol-scale-line`.
   * @property {number} [minWidth=64] Minimum width in pixels at the OGC default dpi. The width will be
   * adjusted to match the dpi used.
   * @property {number} [maxWidth] Maximum width in pixels at the OGC default dpi. The width will be
   * adjusted to match the dpi used.
   * @property {function(import("../MapEvent.js").default):void} [render] Function called when the control
   * should be re-rendered. This is called in a `requestAnimationFrame` callback.
   * @property {HTMLElement|string} [target] Specify a target if you want the control
   * to be rendered outside of the map's viewport.
   * @property {Units} [units='metric'] Units.
   * @property {boolean} [bar=false] Render scalebars instead of a line.
   * @property {number} [steps=4] Number of steps the scalebar should use. Use even numbers
   * for best results. Only applies when `bar` is `true`.
   * @property {boolean} [text=false] Render the text scale above of the scalebar. Only applies
   * when `bar` is `true`.
   * @property {number|undefined} [dpi=undefined] dpi of output device such as printer. Only applies
   * when `bar` is `true`. If undefined the OGC default screen pixel size of 0.28mm will be assumed.
   */

  /**
   * @classdesc
   * A control displaying rough y-axis distances, calculated for the center of the
   * viewport. For conformal projections (e.g. EPSG:3857, the default view
   * projection in OpenLayers), the scale is valid for all directions.
   * No scale line will be shown when the y-axis distance of a pixel at the
   * viewport center cannot be calculated in the view projection.
   * By default the scale line will show in the bottom left portion of the map,
   * but this can be changed by using the css selector `.ol-scale-line`.
   * When specifying `bar` as `true`, a scalebar will be rendered instead
   * of a scaleline.
   *
   * @api
   */
  class ScaleLine extends Control {
    /**
     * @param {Options} [options] Scale line options.
     */
    constructor(options) {
      options = options ? options : {};

      const element = document.createElement('div');
      element.style.pointerEvents = 'none';

      super({
        element: element,
        render: options.render,
        target: options.target,
      });

      /***
       * @type {ScaleLineOnSignature<import("../events").EventsKey>}
       */
      this.on;

      /***
       * @type {ScaleLineOnSignature<import("../events").EventsKey>}
       */
      this.once;

      /***
       * @type {ScaleLineOnSignature<void>}
       */
      this.un;

      const className =
        options.className !== undefined
          ? options.className
          : options.bar
          ? 'ol-scale-bar'
          : 'ol-scale-line';

      /**
       * @private
       * @type {HTMLElement}
       */
      this.innerElement_ = document.createElement('div');
      this.innerElement_.className = className + '-inner';

      this.element.className = className + ' ' + CLASS_UNSELECTABLE;
      this.element.appendChild(this.innerElement_);

      /**
       * @private
       * @type {?import("../View.js").State}
       */
      this.viewState_ = null;

      /**
       * @private
       * @type {number}
       */
      this.minWidth_ = options.minWidth !== undefined ? options.minWidth : 64;

      /**
       * @private
       * @type {number|undefined}
       */
      this.maxWidth_ = options.maxWidth;

      /**
       * @private
       * @type {boolean}
       */
      this.renderedVisible_ = false;

      /**
       * @private
       * @type {number|undefined}
       */
      this.renderedWidth_ = undefined;

      /**
       * @private
       * @type {string}
       */
      this.renderedHTML_ = '';

      this.addChangeListener(UNITS_PROP, this.handleUnitsChanged_);

      this.setUnits(options.units || 'metric');

      /**
       * @private
       * @type {boolean}
       */
      this.scaleBar_ = options.bar || false;

      /**
       * @private
       * @type {number}
       */
      this.scaleBarSteps_ = options.steps || 4;

      /**
       * @private
       * @type {boolean}
       */
      this.scaleBarText_ = options.text || false;

      /**
       * @private
       * @type {number|undefined}
       */
      this.dpi_ = options.dpi || undefined;
    }

    /**
     * Return the units to use in the scale line.
     * @return {Units} The units
     * to use in the scale line.
     * @observable
     * @api
     */
    getUnits() {
      return this.get(UNITS_PROP);
    }

    /**
     * @private
     */
    handleUnitsChanged_() {
      this.updateElement_();
    }

    /**
     * Set the units to use in the scale line.
     * @param {Units} units The units to use in the scale line.
     * @observable
     * @api
     */
    setUnits(units) {
      this.set(UNITS_PROP, units);
    }

    /**
     * Specify the dpi of output device such as printer.
     * @param {number|undefined} dpi The dpi of output device.
     * @api
     */
    setDpi(dpi) {
      this.dpi_ = dpi;
    }

    /**
     * @private
     */
    updateElement_() {
      const viewState = this.viewState_;

      if (!viewState) {
        if (this.renderedVisible_) {
          this.element.style.display = 'none';
          this.renderedVisible_ = false;
        }
        return;
      }

      const center = viewState.center;
      const projection = viewState.projection;
      const units = this.getUnits();
      const pointResolutionUnits = units == 'degrees' ? 'degrees' : 'm';
      let pointResolution = getPointResolution(
        projection,
        viewState.resolution,
        center,
        pointResolutionUnits
      );

      const minWidth =
        (this.minWidth_ * (this.dpi_ || DEFAULT_DPI)) / DEFAULT_DPI;

      const maxWidth =
        this.maxWidth_ !== undefined
          ? (this.maxWidth_ * (this.dpi_ || DEFAULT_DPI)) / DEFAULT_DPI
          : undefined;

      let nominalCount = minWidth * pointResolution;
      let suffix = '';
      if (units == 'degrees') {
        const metersPerDegree = METERS_PER_UNIT$1.degrees;
        nominalCount *= metersPerDegree;
        if (nominalCount < metersPerDegree / 60) {
          suffix = '\u2033'; // seconds
          pointResolution *= 3600;
        } else if (nominalCount < metersPerDegree) {
          suffix = '\u2032'; // minutes
          pointResolution *= 60;
        } else {
          suffix = '\u00b0'; // degrees
        }
      } else if (units == 'imperial') {
        if (nominalCount < 0.9144) {
          suffix = 'in';
          pointResolution /= 0.0254;
        } else if (nominalCount < 1609.344) {
          suffix = 'ft';
          pointResolution /= 0.3048;
        } else {
          suffix = 'mi';
          pointResolution /= 1609.344;
        }
      } else if (units == 'nautical') {
        pointResolution /= 1852;
        suffix = 'NM';
      } else if (units == 'metric') {
        if (nominalCount < 0.001) {
          suffix = 'μm';
          pointResolution *= 1000000;
        } else if (nominalCount < 1) {
          suffix = 'mm';
          pointResolution *= 1000;
        } else if (nominalCount < 1000) {
          suffix = 'm';
        } else {
          suffix = 'km';
          pointResolution /= 1000;
        }
      } else if (units == 'us') {
        if (nominalCount < 0.9144) {
          suffix = 'in';
          pointResolution *= 39.37;
        } else if (nominalCount < 1609.344) {
          suffix = 'ft';
          pointResolution /= 0.30480061;
        } else {
          suffix = 'mi';
          pointResolution /= 1609.3472;
        }
      } else {
        assert(false, 33); // Invalid units
      }

      let i = 3 * Math.floor(Math.log(minWidth * pointResolution) / Math.log(10));
      let count, width, decimalCount;
      let previousCount, previousWidth, previousDecimalCount;
      while (true) {
        decimalCount = Math.floor(i / 3);
        const decimal = Math.pow(10, decimalCount);
        count = LEADING_DIGITS[((i % 3) + 3) % 3] * decimal;
        width = Math.round(count / pointResolution);
        if (isNaN(width)) {
          this.element.style.display = 'none';
          this.renderedVisible_ = false;
          return;
        }
        if (maxWidth !== undefined && width >= maxWidth) {
          count = previousCount;
          width = previousWidth;
          decimalCount = previousDecimalCount;
          break;
        } else if (width >= minWidth) {
          break;
        }
        previousCount = count;
        previousWidth = width;
        previousDecimalCount = decimalCount;
        ++i;
      }
      const html = this.scaleBar_
        ? this.createScaleBar(width, count, suffix)
        : count.toFixed(decimalCount < 0 ? -decimalCount : 0) + ' ' + suffix;

      if (this.renderedHTML_ != html) {
        this.innerElement_.innerHTML = html;
        this.renderedHTML_ = html;
      }

      if (this.renderedWidth_ != width) {
        this.innerElement_.style.width = width + 'px';
        this.renderedWidth_ = width;
      }

      if (!this.renderedVisible_) {
        this.element.style.display = '';
        this.renderedVisible_ = true;
      }
    }

    /**
     * @private
     * @param {number} width The current width of the scalebar.
     * @param {number} scale The current scale.
     * @param {string} suffix The suffix to append to the scale text.
     * @return {string} The stringified HTML of the scalebar.
     */
    createScaleBar(width, scale, suffix) {
      const resolutionScale = this.getScaleForResolution();
      const mapScale =
        resolutionScale < 1
          ? Math.round(1 / resolutionScale).toLocaleString() + ' : 1'
          : '1 : ' + Math.round(resolutionScale).toLocaleString();
      const steps = this.scaleBarSteps_;
      const stepWidth = width / steps;
      const scaleSteps = [this.createMarker('absolute')];
      for (let i = 0; i < steps; ++i) {
        const cls =
          i % 2 === 0 ? 'ol-scale-singlebar-odd' : 'ol-scale-singlebar-even';
        scaleSteps.push(
          '<div>' +
            '<div ' +
            `class="ol-scale-singlebar ${cls}" ` +
            `style="width: ${stepWidth}px;"` +
            '>' +
            '</div>' +
            this.createMarker('relative') +
            // render text every second step, except when only 2 steps
            (i % 2 === 0 || steps === 2
              ? this.createStepText(i, width, false, scale, suffix)
              : '') +
            '</div>'
        );
      }
      // render text at the end
      scaleSteps.push(this.createStepText(steps, width, true, scale, suffix));

      const scaleBarText = this.scaleBarText_
        ? `<div class="ol-scale-text" style="width: ${width}px;">` +
          mapScale +
          '</div>'
        : '';
      return scaleBarText + scaleSteps.join('');
    }

    /**
     * Creates a marker at given position
     * @param {'absolute'|'relative'} position The position, absolute or relative
     * @return {string} The stringified div containing the marker
     */
    createMarker(position) {
      const top = position === 'absolute' ? 3 : -10;
      return (
        '<div ' +
        'class="ol-scale-step-marker" ' +
        `style="position: ${position}; top: ${top}px;"` +
        '></div>'
      );
    }

    /**
     * Creates the label for a marker marker at given position
     * @param {number} i The iterator
     * @param {number} width The width the scalebar will currently use
     * @param {boolean} isLast Flag indicating if we add the last step text
     * @param {number} scale The current scale for the whole scalebar
     * @param {string} suffix The suffix for the scale
     * @return {string} The stringified div containing the step text
     */
    createStepText(i, width, isLast, scale, suffix) {
      const length =
        i === 0 ? 0 : Math.round((scale / this.scaleBarSteps_) * i * 100) / 100;
      const lengthString = length + (i === 0 ? '' : ' ' + suffix);
      const margin = i === 0 ? -3 : (width / this.scaleBarSteps_) * -1;
      const minWidth = i === 0 ? 0 : (width / this.scaleBarSteps_) * 2;
      return (
        '<div ' +
        'class="ol-scale-step-text" ' +
        'style="' +
        `margin-left: ${margin}px;` +
        `text-align: ${i === 0 ? 'left' : 'center'};` +
        `min-width: ${minWidth}px;` +
        `left: ${isLast ? width + 'px' : 'unset'};` +
        '">' +
        lengthString +
        '</div>'
      );
    }

    /**
     * Returns the appropriate scale for the given resolution and units.
     * @return {number} The appropriate scale.
     */
    getScaleForResolution() {
      const resolution = getPointResolution(
        this.viewState_.projection,
        this.viewState_.resolution,
        this.viewState_.center,
        'm'
      );
      const dpi = this.dpi_ || DEFAULT_DPI;
      const inchesPerMeter = 1000 / 25.4;
      return resolution * inchesPerMeter * dpi;
    }

    /**
     * Update the scale line element.
     * @param {import("../MapEvent.js").default} mapEvent Map event.
     * @override
     */
    render(mapEvent) {
      const frameState = mapEvent.frameState;
      if (!frameState) {
        this.viewState_ = null;
      } else {
        this.viewState_ = frameState.viewState;
      }
      this.updateElement_();
    }
  }

  /**
   * @module ol/control/ZoomSlider
   */

  /**
   * The enum for available directions.
   *
   * @enum {number}
   */
  const Direction = {
    VERTICAL: 0,
    HORIZONTAL: 1,
  };

  /**
   * @typedef {Object} Options
   * @property {string} [className='ol-zoomslider'] CSS class name.
   * @property {number} [duration=200] Animation duration in milliseconds.
   * @property {function(import("../MapEvent.js").default):void} [render] Function called when the control
   * should be re-rendered. This is called in a `requestAnimationFrame` callback.
   */

  /**
   * @classdesc
   * A slider type of control for zooming.
   *
   * Example:
   *
   *     map.addControl(new ZoomSlider());
   *
   * @api
   */
  class ZoomSlider extends Control {
    /**
     * @param {Options} [options] Zoom slider options.
     */
    constructor(options) {
      options = options ? options : {};

      super({
        element: document.createElement('div'),
        render: options.render,
      });

      /**
       * @type {!Array<import("../events.js").EventsKey>}
       * @private
       */
      this.dragListenerKeys_ = [];

      /**
       * Will hold the current resolution of the view.
       *
       * @type {number|undefined}
       * @private
       */
      this.currentResolution_ = undefined;

      /**
       * The direction of the slider. Will be determined from actual display of the
       * container and defaults to Direction.VERTICAL.
       *
       * @type {Direction}
       * @private
       */
      this.direction_ = Direction.VERTICAL;

      /**
       * @type {boolean}
       * @private
       */
      this.dragging_;

      /**
       * @type {number}
       * @private
       */
      this.heightLimit_ = 0;

      /**
       * @type {number}
       * @private
       */
      this.widthLimit_ = 0;

      /**
       * @type {number|undefined}
       * @private
       */
      this.startX_;

      /**
       * @type {number|undefined}
       * @private
       */
      this.startY_;

      /**
       * The calculated thumb size (border box plus margins).  Set when initSlider_
       * is called.
       * @type {import("../size.js").Size}
       * @private
       */
      this.thumbSize_ = null;

      /**
       * Whether the slider is initialized.
       * @type {boolean}
       * @private
       */
      this.sliderInitialized_ = false;

      /**
       * @type {number}
       * @private
       */
      this.duration_ = options.duration !== undefined ? options.duration : 200;

      const className =
        options.className !== undefined ? options.className : 'ol-zoomslider';
      const thumbElement = document.createElement('button');
      thumbElement.setAttribute('type', 'button');
      thumbElement.className = className + '-thumb ' + CLASS_UNSELECTABLE;
      const containerElement = this.element;
      containerElement.className =
        className + ' ' + CLASS_UNSELECTABLE + ' ' + CLASS_CONTROL;
      containerElement.appendChild(thumbElement);

      containerElement.addEventListener(
        PointerEventType.POINTERDOWN,
        this.handleDraggerStart_.bind(this),
        false
      );
      containerElement.addEventListener(
        PointerEventType.POINTERMOVE,
        this.handleDraggerDrag_.bind(this),
        false
      );
      containerElement.addEventListener(
        PointerEventType.POINTERUP,
        this.handleDraggerEnd_.bind(this),
        false
      );

      containerElement.addEventListener(
        EventType$1.CLICK,
        this.handleContainerClick_.bind(this),
        false
      );
      thumbElement.addEventListener(EventType$1.CLICK, stopPropagation, false);
    }

    /**
     * Remove the control from its current map and attach it to the new map.
     * Pass `null` to just remove the control from the current map.
     * Subclasses may set up event handlers to get notified about changes to
     * the map here.
     * @param {import("../Map.js").default|null} map Map.
     * @api
     */
    setMap(map) {
      super.setMap(map);
      if (map) {
        map.render();
      }
    }

    /**
     * Initializes the slider element. This will determine and set this controls
     * direction_ and also constrain the dragging of the thumb to always be within
     * the bounds of the container.
     *
     * @return {boolean} Initialization successful
     * @private
     */
    initSlider_() {
      const container = this.element;
      let containerWidth = container.offsetWidth;
      let containerHeight = container.offsetHeight;
      if (containerWidth === 0 && containerHeight === 0) {
        return (this.sliderInitialized_ = false);
      }

      const containerStyle = getComputedStyle(container);
      containerWidth -=
        parseFloat(containerStyle['paddingRight']) +
        parseFloat(containerStyle['paddingLeft']);
      containerHeight -=
        parseFloat(containerStyle['paddingTop']) +
        parseFloat(containerStyle['paddingBottom']);
      const thumb = /** @type {HTMLElement} */ (container.firstElementChild);
      const thumbStyle = getComputedStyle(thumb);
      const thumbWidth =
        thumb.offsetWidth +
        parseFloat(thumbStyle['marginRight']) +
        parseFloat(thumbStyle['marginLeft']);
      const thumbHeight =
        thumb.offsetHeight +
        parseFloat(thumbStyle['marginTop']) +
        parseFloat(thumbStyle['marginBottom']);
      this.thumbSize_ = [thumbWidth, thumbHeight];

      if (containerWidth > containerHeight) {
        this.direction_ = Direction.HORIZONTAL;
        this.widthLimit_ = containerWidth - thumbWidth;
      } else {
        this.direction_ = Direction.VERTICAL;
        this.heightLimit_ = containerHeight - thumbHeight;
      }
      return (this.sliderInitialized_ = true);
    }

    /**
     * @param {PointerEvent} event The browser event to handle.
     * @private
     */
    handleContainerClick_(event) {
      const view = this.getMap().getView();

      const relativePosition = this.getRelativePosition_(
        event.offsetX - this.thumbSize_[0] / 2,
        event.offsetY - this.thumbSize_[1] / 2
      );

      const resolution = this.getResolutionForPosition_(relativePosition);
      const zoom = view.getConstrainedZoom(view.getZoomForResolution(resolution));

      view.animateInternal({
        zoom: zoom,
        duration: this.duration_,
        easing: easeOut,
      });
    }

    /**
     * Handle dragger start events.
     * @param {PointerEvent} event The drag event.
     * @private
     */
    handleDraggerStart_(event) {
      if (!this.dragging_ && event.target === this.element.firstElementChild) {
        const element = /** @type {HTMLElement} */ (
          this.element.firstElementChild
        );
        this.getMap().getView().beginInteraction();
        this.startX_ = event.clientX - parseFloat(element.style.left);
        this.startY_ = event.clientY - parseFloat(element.style.top);
        this.dragging_ = true;

        if (this.dragListenerKeys_.length === 0) {
          const drag = this.handleDraggerDrag_;
          const end = this.handleDraggerEnd_;
          const doc = this.getMap().getOwnerDocument();
          this.dragListenerKeys_.push(
            listen(doc, PointerEventType.POINTERMOVE, drag, this),
            listen(doc, PointerEventType.POINTERUP, end, this)
          );
        }
      }
    }

    /**
     * Handle dragger drag events.
     *
     * @param {PointerEvent} event The drag event.
     * @private
     */
    handleDraggerDrag_(event) {
      if (this.dragging_) {
        const deltaX = event.clientX - this.startX_;
        const deltaY = event.clientY - this.startY_;
        const relativePosition = this.getRelativePosition_(deltaX, deltaY);
        this.currentResolution_ =
          this.getResolutionForPosition_(relativePosition);
        this.getMap().getView().setResolution(this.currentResolution_);
      }
    }

    /**
     * Handle dragger end events.
     * @param {PointerEvent} event The drag event.
     * @private
     */
    handleDraggerEnd_(event) {
      if (this.dragging_) {
        const view = this.getMap().getView();
        view.endInteraction();

        this.dragging_ = false;
        this.startX_ = undefined;
        this.startY_ = undefined;
        this.dragListenerKeys_.forEach(unlistenByKey);
        this.dragListenerKeys_.length = 0;
      }
    }

    /**
     * Positions the thumb inside its container according to the given resolution.
     *
     * @param {number} res The res.
     * @private
     */
    setThumbPosition_(res) {
      const position = this.getPositionForResolution_(res);
      const thumb = /** @type {HTMLElement} */ (this.element.firstElementChild);

      if (this.direction_ == Direction.HORIZONTAL) {
        thumb.style.left = this.widthLimit_ * position + 'px';
      } else {
        thumb.style.top = this.heightLimit_ * position + 'px';
      }
    }

    /**
     * Calculates the relative position of the thumb given x and y offsets.  The
     * relative position scales from 0 to 1.  The x and y offsets are assumed to be
     * in pixel units within the dragger limits.
     *
     * @param {number} x Pixel position relative to the left of the slider.
     * @param {number} y Pixel position relative to the top of the slider.
     * @return {number} The relative position of the thumb.
     * @private
     */
    getRelativePosition_(x, y) {
      let amount;
      if (this.direction_ === Direction.HORIZONTAL) {
        amount = x / this.widthLimit_;
      } else {
        amount = y / this.heightLimit_;
      }
      return clamp(amount, 0, 1);
    }

    /**
     * Calculates the corresponding resolution of the thumb given its relative
     * position (where 0 is the minimum and 1 is the maximum).
     *
     * @param {number} position The relative position of the thumb.
     * @return {number} The corresponding resolution.
     * @private
     */
    getResolutionForPosition_(position) {
      const fn = this.getMap().getView().getResolutionForValueFunction();
      return fn(1 - position);
    }

    /**
     * Determines the relative position of the slider for the given resolution.  A
     * relative position of 0 corresponds to the minimum view resolution.  A
     * relative position of 1 corresponds to the maximum view resolution.
     *
     * @param {number} res The resolution.
     * @return {number} The relative position value (between 0 and 1).
     * @private
     */
    getPositionForResolution_(res) {
      const fn = this.getMap().getView().getValueForResolutionFunction();
      return clamp(1 - fn(res), 0, 1);
    }

    /**
     * Update the zoomslider element.
     * @param {import("../MapEvent.js").default} mapEvent Map event.
     * @override
     */
    render(mapEvent) {
      if (!mapEvent.frameState) {
        return;
      }
      if (!this.sliderInitialized_ && !this.initSlider_()) {
        return;
      }
      const res = mapEvent.frameState.viewState.resolution;
      this.currentResolution_ = res;
      this.setThumbPosition_(res);
    }
  }

  /**
   * @module ol/control/ZoomToExtent
   */

  /**
   * @typedef {Object} Options
   * @property {string} [className='ol-zoom-extent'] Class name.
   * @property {HTMLElement|string} [target] Specify a target if you want the control
   * to be rendered outside of the map's viewport.
   * @property {string|HTMLElement} [label='E'] Text label to use for the button.
   * Instead of text, also an element (e.g. a `span` element) can be used.
   * @property {string} [tipLabel='Fit to extent'] Text label to use for the button tip.
   * @property {import("../extent.js").Extent} [extent] The extent to zoom to. If undefined the validity
   * extent of the view projection is used.
   */

  /**
   * @classdesc
   * A button control which, when pressed, changes the map view to a specific
   * extent. To style this control use the css selector `.ol-zoom-extent`.
   *
   * @api
   */
  class ZoomToExtent extends Control {
    /**
     * @param {Options} [options] Options.
     */
    constructor(options) {
      options = options ? options : {};

      super({
        element: document.createElement('div'),
        target: options.target,
      });

      /**
       * @type {?import("../extent.js").Extent|null}
       * @protected
       */
      this.extent = options.extent ? options.extent : null;

      const className =
        options.className !== undefined ? options.className : 'ol-zoom-extent';

      const label = options.label !== undefined ? options.label : 'E';
      const tipLabel =
        options.tipLabel !== undefined ? options.tipLabel : 'Fit to extent';
      const button = document.createElement('button');
      button.setAttribute('type', 'button');
      button.title = tipLabel;
      button.appendChild(
        typeof label === 'string' ? document.createTextNode(label) : label
      );

      button.addEventListener(
        EventType$1.CLICK,
        this.handleClick_.bind(this),
        false
      );

      const cssClasses =
        className + ' ' + CLASS_UNSELECTABLE + ' ' + CLASS_CONTROL;
      const element = this.element;
      element.className = cssClasses;
      element.appendChild(button);
    }

    /**
     * @param {MouseEvent} event The event to handle
     * @private
     */
    handleClick_(event) {
      event.preventDefault();
      this.handleZoomToExtent();
    }

    /**
     * @protected
     */
    handleZoomToExtent() {
      const map = this.getMap();
      const view = map.getView();
      const extent = !this.extent
        ? view.getProjection().getExtent()
        : this.extent;
      view.fitInternal(fromExtent(extent));
    }
  }

  var ol = {};
  ol.AssertionError = AssertionError;
  ol.Collection = Collection;
  ol.Collection.CollectionEvent = CollectionEvent;
  ol.DataTile = DataTile;
  ol.Disposable = Disposable;
  ol.Feature = Feature;
  ol.Feature.createStyleFunction = createStyleFunction;
  ol.Geolocation = Geolocation;
  ol.Image = ImageWrapper;
  ol.Image.listenImage = listenImage;
  ol.ImageBase = ImageBase;
  ol.ImageCanvas = ImageCanvas;
  ol.ImageTile = ImageTile;
  ol.Kinetic = Kinetic;
  ol.Map = Map;
  ol.MapBrowserEvent = MapBrowserEvent;
  ol.MapBrowserEventHandler = MapBrowserEventHandler;
  ol.MapEvent = MapEvent;
  ol.Object = BaseObject;
  ol.Object.ObjectEvent = ObjectEvent;
  ol.Observable = Observable;
  ol.Observable.unByKey = unByKey;
  ol.Overlay = Overlay;
  ol.Tile = Tile;
  ol.TileCache = TileCache;
  ol.TileQueue = TileQueue;
  ol.TileQueue.getTilePriority = getTilePriority;
  ol.TileRange = TileRange;
  ol.TileRange.createOrUpdate = createOrUpdate;
  ol.VectorRenderTile = VectorRenderTile;
  ol.VectorTile = VectorTile$1;
  ol.View = View;
  ol.View.createCenterConstraint = createCenterConstraint;
  ol.View.createResolutionConstraint = createResolutionConstraint;
  ol.View.createRotationConstraint = createRotationConstraint;
  ol.View.isNoopAnimation = isNoopAnimation;
  ol.array = {};
  ol.array.binarySearch = binarySearch$1;
  ol.array.equals = equals$3;
  ol.array.extend = extend$2;
  ol.array.isSorted = isSorted;
  ol.array.linearFindNearest = linearFindNearest;
  ol.array.numberSafeCompareFunction = numberSafeCompareFunction;
  ol.array.remove = remove$1;
  ol.array.reverseSubArray = reverseSubArray;
  ol.array.stableSort = stableSort;
  ol.asserts = {};
  ol.asserts.assert = assert;
  ol.centerconstraint = {};
  ol.centerconstraint.createExtent = createExtent;
  ol.centerconstraint.none = none$1;
  ol.color = {};
  ol.color.asArray = asArray;
  ol.color.asString = asString;
  ol.color.fromString = fromString;
  ol.color.isStringColor = isStringColor;
  ol.color.normalize = normalize;
  ol.color.toString = toString$2;
  ol.colorlike = {};
  ol.colorlike.asColorLike = asColorLike;
  ol.control = {};
  ol.control.Attribution = Attribution;
  ol.control.Control = Control;
  ol.control.FullScreen = FullScreen;
  ol.control.MousePosition = MousePosition;
  ol.control.OverviewMap = OverviewMap;
  ol.control.Rotate = Rotate;
  ol.control.ScaleLine = ScaleLine;
  ol.control.Zoom = Zoom;
  ol.control.ZoomSlider = ZoomSlider;
  ol.control.ZoomToExtent = ZoomToExtent;
  ol.control.defaults = {};
  ol.control.defaults.defaults = defaults$1;
  ol.coordinate = {};
  ol.coordinate.add = add;
  ol.coordinate.closestOnCircle = closestOnCircle;
  ol.coordinate.closestOnSegment = closestOnSegment;
  ol.coordinate.createStringXY = createStringXY;
  ol.coordinate.degreesToStringHDMS = degreesToStringHDMS;
  ol.coordinate.distance = distance;
  ol.coordinate.equals = equals$1;
  ol.coordinate.format = format;
  ol.coordinate.getWorldsAway = getWorldsAway;
  ol.coordinate.rotate = rotate$1;
  ol.coordinate.scale = scale$2;
  ol.coordinate.squaredDistance = squaredDistance;
  ol.coordinate.squaredDistanceToSegment = squaredDistanceToSegment;
  ol.coordinate.toStringHDMS = toStringHDMS;
  ol.coordinate.toStringXY = toStringXY;
  ol.coordinate.wrapX = wrapX$1;
  ol.css = {};
  ol.css.CLASS_COLLAPSED = CLASS_COLLAPSED;
  ol.css.CLASS_CONTROL = CLASS_CONTROL;
  ol.css.CLASS_HIDDEN = CLASS_HIDDEN;
  ol.css.CLASS_SELECTABLE = CLASS_SELECTABLE;
  ol.css.CLASS_UNSELECTABLE = CLASS_UNSELECTABLE;
  ol.css.CLASS_UNSUPPORTED = CLASS_UNSUPPORTED;
  ol.css.getFontParameters = getFontParameters;
  ol.dom = {};
  ol.dom.createCanvasContext2D = createCanvasContext2D;
  ol.dom.outerHeight = outerHeight;
  ol.dom.outerWidth = outerWidth;
  ol.dom.releaseCanvas = releaseCanvas$1;
  ol.dom.removeChildren = removeChildren;
  ol.dom.removeNode = removeNode$1;
  ol.dom.replaceChildren = replaceChildren;
  ol.dom.replaceNode = replaceNode;
  ol.easing = {};
  ol.easing.easeIn = easeIn;
  ol.easing.easeOut = easeOut;
  ol.easing.inAndOut = inAndOut;
  ol.easing.linear = linear;
  ol.easing.upAndDown = upAndDown;
  ol.events = {};
  ol.events.Event = BaseEvent;
  ol.events.Event.preventDefault = preventDefault;
  ol.events.Event.stopPropagation = stopPropagation;
  ol.events.Target = Target;
  ol.events.condition = {};
  ol.events.condition.all = all$1;
  ol.events.condition.altKeyOnly = altKeyOnly;
  ol.events.condition.altShiftKeysOnly = altShiftKeysOnly;
  ol.events.condition.always = always;
  ol.events.condition.click = click;
  ol.events.condition.doubleClick = doubleClick;
  ol.events.condition.focus = focus;
  ol.events.condition.focusWithTabindex = focusWithTabindex;
  ol.events.condition.mouseActionButton = mouseActionButton;
  ol.events.condition.mouseOnly = mouseOnly;
  ol.events.condition.never = never;
  ol.events.condition.noModifierKeys = noModifierKeys;
  ol.events.condition.penOnly = penOnly;
  ol.events.condition.platformModifierKeyOnly = platformModifierKeyOnly;
  ol.events.condition.pointerMove = pointerMove;
  ol.events.condition.primaryAction = primaryAction;
  ol.events.condition.shiftKeyOnly = shiftKeyOnly;
  ol.events.condition.singleClick = singleClick;
  ol.events.condition.targetNotEditable = targetNotEditable;
  ol.events.condition.touchOnly = touchOnly;
  ol.events.listen = listen;
  ol.events.listenOnce = listenOnce;
  ol.events.unlistenByKey = unlistenByKey;
  ol.extent = {};
  ol.extent.applyTransform = applyTransform;
  ol.extent.approximatelyEquals = approximatelyEquals;
  ol.extent.boundingExtent = boundingExtent;
  ol.extent.buffer = buffer$1;
  ol.extent.clone = clone;
  ol.extent.closestSquaredDistanceXY = closestSquaredDistanceXY;
  ol.extent.containsCoordinate = containsCoordinate;
  ol.extent.containsExtent = containsExtent;
  ol.extent.containsXY = containsXY;
  ol.extent.coordinateRelationship = coordinateRelationship;
  ol.extent.createEmpty = createEmpty;
  ol.extent.createOrUpdate = createOrUpdate$2;
  ol.extent.createOrUpdateEmpty = createOrUpdateEmpty;
  ol.extent.createOrUpdateFromCoordinate = createOrUpdateFromCoordinate;
  ol.extent.createOrUpdateFromCoordinates = createOrUpdateFromCoordinates;
  ol.extent.createOrUpdateFromFlatCoordinates = createOrUpdateFromFlatCoordinates;
  ol.extent.createOrUpdateFromRings = createOrUpdateFromRings;
  ol.extent.equals = equals$2;
  ol.extent.extend = extend$1;
  ol.extent.extendCoordinate = extendCoordinate;
  ol.extent.extendCoordinates = extendCoordinates;
  ol.extent.extendFlatCoordinates = extendFlatCoordinates;
  ol.extent.extendRings = extendRings;
  ol.extent.extendXY = extendXY;
  ol.extent.forEachCorner = forEachCorner;
  ol.extent.getArea = getArea$1;
  ol.extent.getBottomLeft = getBottomLeft;
  ol.extent.getBottomRight = getBottomRight;
  ol.extent.getCenter = getCenter;
  ol.extent.getCorner = getCorner;
  ol.extent.getEnlargedArea = getEnlargedArea;
  ol.extent.getForViewAndSize = getForViewAndSize;
  ol.extent.getHeight = getHeight;
  ol.extent.getIntersection = getIntersection;
  ol.extent.getIntersectionArea = getIntersectionArea;
  ol.extent.getMargin = getMargin;
  ol.extent.getRotatedViewport = getRotatedViewport;
  ol.extent.getSize = getSize;
  ol.extent.getTopLeft = getTopLeft;
  ol.extent.getTopRight = getTopRight;
  ol.extent.getWidth = getWidth;
  ol.extent.intersects = intersects$3;
  ol.extent.intersectsSegment = intersectsSegment;
  ol.extent.isEmpty = isEmpty;
  ol.extent.returnOrUpdate = returnOrUpdate;
  ol.extent.scaleFromCenter = scaleFromCenter;
  ol.extent.wrapAndSliceX = wrapAndSliceX;
  ol.extent.wrapX = wrapX$2;
  ol.featureloader = {};
  ol.featureloader.loadFeaturesXhr = loadFeaturesXhr;
  ol.featureloader.setWithCredentials = setWithCredentials;
  ol.featureloader.xhr = xhr;
  ol.format = {};
  ol.format.EsriJSON = EsriJSON;
  ol.format.Feature = FeatureFormat;
  ol.format.Feature.transformExtentWithOptions = transformExtentWithOptions;
  ol.format.Feature.transformGeometryWithOptions = transformGeometryWithOptions;
  ol.format.GML = GML;
  ol.format.GML2 = GML2;
  ol.format.GML3 = GML3;
  ol.format.GML32 = GML32;
  ol.format.GMLBase = GMLBase;
  ol.format.GMLBase.GMLNS = GMLNS;
  ol.format.GPX = GPX;
  ol.format.GeoJSON = GeoJSON;
  ol.format.IGC = IGC;
  ol.format.IIIFInfo = IIIFInfo;
  ol.format.JSONFeature = JSONFeature;
  ol.format.KML = KML;
  ol.format.KML.getDefaultFillStyle = getDefaultFillStyle;
  ol.format.KML.getDefaultImageStyle = getDefaultImageStyle;
  ol.format.KML.getDefaultStrokeStyle = getDefaultStrokeStyle;
  ol.format.KML.getDefaultStyle = getDefaultStyle;
  ol.format.KML.getDefaultStyleArray = getDefaultStyleArray;
  ol.format.KML.getDefaultTextStyle = getDefaultTextStyle;
  ol.format.KML.readFlatCoordinates = readFlatCoordinates;
  ol.format.MVT = MVT;
  ol.format.OSMXML = OSMXML;
  ol.format.OWS = OWS;
  ol.format.Polyline = Polyline;
  ol.format.Polyline.decodeDeltas = decodeDeltas;
  ol.format.Polyline.decodeFloats = decodeFloats;
  ol.format.Polyline.decodeSignedIntegers = decodeSignedIntegers;
  ol.format.Polyline.decodeUnsignedIntegers = decodeUnsignedIntegers;
  ol.format.Polyline.encodeDeltas = encodeDeltas;
  ol.format.Polyline.encodeFloats = encodeFloats;
  ol.format.Polyline.encodeSignedIntegers = encodeSignedIntegers;
  ol.format.Polyline.encodeUnsignedInteger = encodeUnsignedInteger;
  ol.format.Polyline.encodeUnsignedIntegers = encodeUnsignedIntegers;
  ol.format.TextFeature = TextFeature;
  ol.format.TopoJSON = TopoJSON;
  ol.format.WFS = WFS;
  ol.format.WFS.writeFilter = writeFilter;
  ol.format.WKB = WKB;
  ol.format.WKT = WKT;
  ol.format.WMSCapabilities = WMSCapabilities;
  ol.format.WMSGetFeatureInfo = WMSGetFeatureInfo;
  ol.format.WMTSCapabilities = WMTSCapabilities;
  ol.format.XML = XML;
  ol.format.XMLFeature = XMLFeature;
  ol.format.filter = {};
  ol.format.filter.And = And;
  ol.format.filter.Bbox = Bbox;
  ol.format.filter.Comparison = Comparison;
  ol.format.filter.ComparisonBinary = ComparisonBinary;
  ol.format.filter.Contains = Contains;
  ol.format.filter.DWithin = DWithin;
  ol.format.filter.Disjoint = Disjoint;
  ol.format.filter.During = During;
  ol.format.filter.EqualTo = EqualTo;
  ol.format.filter.Filter = Filter;
  ol.format.filter.GreaterThan = GreaterThan;
  ol.format.filter.GreaterThanOrEqualTo = GreaterThanOrEqualTo;
  ol.format.filter.Intersects = Intersects;
  ol.format.filter.IsBetween = IsBetween;
  ol.format.filter.IsLike = IsLike;
  ol.format.filter.IsNull = IsNull;
  ol.format.filter.LessThan = LessThan;
  ol.format.filter.LessThanOrEqualTo = LessThanOrEqualTo;
  ol.format.filter.LogicalNary = LogicalNary;
  ol.format.filter.Not = Not;
  ol.format.filter.NotEqualTo = NotEqualTo;
  ol.format.filter.Or = Or;
  ol.format.filter.ResourceId = ResourceId;
  ol.format.filter.Spatial = Spatial;
  ol.format.filter.Within = Within;
  ol.format.filter.and = and;
  ol.format.filter.bbox = bbox;
  ol.format.filter.between = between;
  ol.format.filter.contains = contains;
  ol.format.filter.disjoint = disjoint;
  ol.format.filter.during = during;
  ol.format.filter.dwithin = dwithin;
  ol.format.filter.equalTo = equalTo;
  ol.format.filter.greaterThan = greaterThan;
  ol.format.filter.greaterThanOrEqualTo = greaterThanOrEqualTo;
  ol.format.filter.intersects = intersects;
  ol.format.filter.isNull = isNull;
  ol.format.filter.lessThan = lessThan;
  ol.format.filter.lessThanOrEqualTo = lessThanOrEqualTo;
  ol.format.filter.like = like;
  ol.format.filter.not = not;
  ol.format.filter.notEqualTo = notEqualTo;
  ol.format.filter.or = or;
  ol.format.filter.resourceId = resourceId;
  ol.format.filter.within = within;
  ol.format.xlink = {};
  ol.format.xlink.readHref = readHref;
  ol.format.xsd = {};
  ol.format.xsd.readBoolean = readBoolean;
  ol.format.xsd.readBooleanString = readBooleanString;
  ol.format.xsd.readDateTime = readDateTime;
  ol.format.xsd.readDecimal = readDecimal;
  ol.format.xsd.readDecimalString = readDecimalString;
  ol.format.xsd.readNonNegativeIntegerString = readNonNegativeIntegerString;
  ol.format.xsd.readPositiveInteger = readPositiveInteger;
  ol.format.xsd.readString = readString;
  ol.format.xsd.writeBooleanTextNode = writeBooleanTextNode;
  ol.format.xsd.writeCDATASection = writeCDATASection;
  ol.format.xsd.writeDateTimeTextNode = writeDateTimeTextNode;
  ol.format.xsd.writeDecimalTextNode = writeDecimalTextNode;
  ol.format.xsd.writeNonNegativeIntegerTextNode = writeNonNegativeIntegerTextNode;
  ol.format.xsd.writeStringTextNode = writeStringTextNode;
  ol.functions = {};
  ol.functions.FALSE = FALSE;
  ol.functions.TRUE = TRUE;
  ol.functions.VOID = VOID;
  ol.functions.memoizeOne = memoizeOne;
  ol.functions.toPromise = toPromise;
  ol.geom = {};
  ol.geom.Circle = Circle;
  ol.geom.Geometry = Geometry;
  ol.geom.GeometryCollection = GeometryCollection;
  ol.geom.LineString = LineString;
  ol.geom.LinearRing = LinearRing;
  ol.geom.MultiLineString = MultiLineString;
  ol.geom.MultiPoint = MultiPoint;
  ol.geom.MultiPolygon = MultiPolygon;
  ol.geom.Point = Point;
  ol.geom.Polygon = Polygon;
  ol.geom.Polygon.circular = circular;
  ol.geom.Polygon.fromCircle = fromCircle;
  ol.geom.Polygon.fromExtent = fromExtent;
  ol.geom.Polygon.makeRegular = makeRegular;
  ol.geom.SimpleGeometry = SimpleGeometry;
  ol.geom.SimpleGeometry.getStrideForLayout = getStrideForLayout;
  ol.geom.SimpleGeometry.transformGeom2D = transformGeom2D;
  ol.geom.flat = {};
  ol.geom.flat.area = {};
  ol.geom.flat.area.linearRing = linearRing;
  ol.geom.flat.area.linearRings = linearRings;
  ol.geom.flat.area.linearRingss = linearRingss$1;
  ol.geom.flat.center = {};
  ol.geom.flat.center.linearRingss = linearRingss;
  ol.geom.flat.closest = {};
  ol.geom.flat.closest.arrayMaxSquaredDelta = arrayMaxSquaredDelta;
  ol.geom.flat.closest.assignClosestArrayPoint = assignClosestArrayPoint;
  ol.geom.flat.closest.assignClosestMultiArrayPoint = assignClosestMultiArrayPoint;
  ol.geom.flat.closest.assignClosestPoint = assignClosestPoint;
  ol.geom.flat.closest.maxSquaredDelta = maxSquaredDelta;
  ol.geom.flat.closest.multiArrayMaxSquaredDelta = multiArrayMaxSquaredDelta;
  ol.geom.flat.contains = {};
  ol.geom.flat.contains.linearRingContainsExtent = linearRingContainsExtent;
  ol.geom.flat.contains.linearRingContainsXY = linearRingContainsXY;
  ol.geom.flat.contains.linearRingsContainsXY = linearRingsContainsXY;
  ol.geom.flat.contains.linearRingssContainsXY = linearRingssContainsXY;
  ol.geom.flat.deflate = {};
  ol.geom.flat.deflate.deflateCoordinate = deflateCoordinate;
  ol.geom.flat.deflate.deflateCoordinates = deflateCoordinates;
  ol.geom.flat.deflate.deflateCoordinatesArray = deflateCoordinatesArray;
  ol.geom.flat.deflate.deflateMultiCoordinatesArray = deflateMultiCoordinatesArray;
  ol.geom.flat.flip = {};
  ol.geom.flat.flip.flipXY = flipXY;
  ol.geom.flat.geodesic = {};
  ol.geom.flat.geodesic.greatCircleArc = greatCircleArc;
  ol.geom.flat.geodesic.meridian = meridian;
  ol.geom.flat.geodesic.parallel = parallel;
  ol.geom.flat.inflate = {};
  ol.geom.flat.inflate.inflateCoordinates = inflateCoordinates;
  ol.geom.flat.inflate.inflateCoordinatesArray = inflateCoordinatesArray;
  ol.geom.flat.inflate.inflateMultiCoordinatesArray = inflateMultiCoordinatesArray;
  ol.geom.flat.interiorpoint = {};
  ol.geom.flat.interiorpoint.getInteriorPointOfArray = getInteriorPointOfArray;
  ol.geom.flat.interiorpoint.getInteriorPointsOfMultiArray = getInteriorPointsOfMultiArray;
  ol.geom.flat.interpolate = {};
  ol.geom.flat.interpolate.interpolatePoint = interpolatePoint;
  ol.geom.flat.interpolate.lineStringCoordinateAtM = lineStringCoordinateAtM;
  ol.geom.flat.interpolate.lineStringsCoordinateAtM = lineStringsCoordinateAtM;
  ol.geom.flat.intersectsextent = {};
  ol.geom.flat.intersectsextent.intersectsLineString = intersectsLineString;
  ol.geom.flat.intersectsextent.intersectsLineStringArray = intersectsLineStringArray;
  ol.geom.flat.intersectsextent.intersectsLinearRing = intersectsLinearRing;
  ol.geom.flat.intersectsextent.intersectsLinearRingArray = intersectsLinearRingArray;
  ol.geom.flat.intersectsextent.intersectsLinearRingMultiArray = intersectsLinearRingMultiArray;
  ol.geom.flat.length = {};
  ol.geom.flat.length.lineStringLength = lineStringLength;
  ol.geom.flat.length.linearRingLength = linearRingLength;
  ol.geom.flat.orient = {};
  ol.geom.flat.orient.inflateEnds = inflateEnds;
  ol.geom.flat.orient.linearRingIsClockwise = linearRingIsClockwise;
  ol.geom.flat.orient.linearRingsAreOriented = linearRingsAreOriented;
  ol.geom.flat.orient.linearRingssAreOriented = linearRingssAreOriented;
  ol.geom.flat.orient.orientLinearRings = orientLinearRings;
  ol.geom.flat.orient.orientLinearRingsArray = orientLinearRingsArray;
  ol.geom.flat.reverse = {};
  ol.geom.flat.reverse.coordinates = coordinates;
  ol.geom.flat.segments = {};
  ol.geom.flat.segments.forEach = forEach;
  ol.geom.flat.simplify = {};
  ol.geom.flat.simplify.douglasPeucker = douglasPeucker;
  ol.geom.flat.simplify.douglasPeuckerArray = douglasPeuckerArray;
  ol.geom.flat.simplify.douglasPeuckerMultiArray = douglasPeuckerMultiArray;
  ol.geom.flat.simplify.quantize = quantize;
  ol.geom.flat.simplify.quantizeArray = quantizeArray;
  ol.geom.flat.simplify.quantizeMultiArray = quantizeMultiArray;
  ol.geom.flat.simplify.radialDistance = radialDistance;
  ol.geom.flat.simplify.simplifyLineString = simplifyLineString;
  ol.geom.flat.simplify.snap = snap;
  ol.geom.flat.straightchunk = {};
  ol.geom.flat.straightchunk.matchingChunk = matchingChunk;
  ol.geom.flat.textpath = {};
  ol.geom.flat.textpath.drawTextOnPath = drawTextOnPath;
  ol.geom.flat.topology = {};
  ol.geom.flat.topology.lineStringIsClosed = lineStringIsClosed;
  ol.geom.flat.transform = {};
  ol.geom.flat.transform.rotate = rotate;
  ol.geom.flat.transform.scale = scale$1;
  ol.geom.flat.transform.transform2D = transform2D;
  ol.geom.flat.transform.translate = translate;
  ol.has = {};
  ol.has.DEVICE_PIXEL_RATIO = DEVICE_PIXEL_RATIO;
  ol.has.FIREFOX = FIREFOX;
  ol.has.IMAGE_DECODE = IMAGE_DECODE;
  ol.has.MAC = MAC;
  ol.has.PASSIVE_EVENT_LISTENERS = PASSIVE_EVENT_LISTENERS;
  ol.has.SAFARI = SAFARI;
  ol.has.SAFARI_BUG_237906 = SAFARI_BUG_237906;
  ol.has.WEBKIT = WEBKIT;
  ol.has.WORKER_OFFSCREEN_CANVAS = WORKER_OFFSCREEN_CANVAS;
  ol.interaction = {};
  ol.interaction.DoubleClickZoom = DoubleClickZoom;
  ol.interaction.DragAndDrop = DragAndDrop;
  ol.interaction.DragAndDrop.DragAndDropEvent = DragAndDropEvent;
  ol.interaction.DragBox = DragBox;
  ol.interaction.DragBox.DragBoxEvent = DragBoxEvent;
  ol.interaction.DragPan = DragPan;
  ol.interaction.DragRotate = DragRotate;
  ol.interaction.DragRotateAndZoom = DragRotateAndZoom;
  ol.interaction.DragZoom = DragZoom;
  ol.interaction.Draw = Draw;
  ol.interaction.Draw.DrawEvent = DrawEvent;
  ol.interaction.Draw.createBox = createBox;
  ol.interaction.Draw.createRegularPolygon = createRegularPolygon;
  ol.interaction.Extent = Extent;
  ol.interaction.Extent.ExtentEvent = ExtentEvent;
  ol.interaction.Interaction = Interaction;
  ol.interaction.Interaction.pan = pan;
  ol.interaction.Interaction.zoomByDelta = zoomByDelta;
  ol.interaction.KeyboardPan = KeyboardPan;
  ol.interaction.KeyboardZoom = KeyboardZoom;
  ol.interaction.Link = Link;
  ol.interaction.Modify = Modify;
  ol.interaction.Modify.ModifyEvent = ModifyEvent;
  ol.interaction.MouseWheelZoom = MouseWheelZoom;
  ol.interaction.PinchRotate = PinchRotate;
  ol.interaction.PinchZoom = PinchZoom;
  ol.interaction.Pointer = PointerInteraction;
  ol.interaction.Pointer.centroid = centroid;
  ol.interaction.Select = Select;
  ol.interaction.Select.SelectEvent = SelectEvent;
  ol.interaction.Snap = Snap;
  ol.interaction.Translate = Translate;
  ol.interaction.Translate.TranslateEvent = TranslateEvent;
  ol.interaction.defaults = {};
  ol.interaction.defaults.defaults = defaults;
  ol.layer = {};
  ol.layer.Base = BaseLayer;
  ol.layer.BaseImage = BaseImageLayer;
  ol.layer.BaseTile = BaseTileLayer;
  ol.layer.BaseVector = BaseVectorLayer;
  ol.layer.Graticule = Graticule;
  ol.layer.Group = LayerGroup;
  ol.layer.Group.GroupEvent = GroupEvent;
  ol.layer.Heatmap = Heatmap;
  ol.layer.Image = ImageLayer;
  ol.layer.Layer = Layer;
  ol.layer.Layer.inView = inView;
  ol.layer.MapboxVector = MapboxVectorLayer;
  ol.layer.Tile = TileLayer;
  ol.layer.Vector = VectorLayer;
  ol.layer.VectorImage = VectorImageLayer;
  ol.layer.VectorTile = VectorTileLayer;
  ol.layer.WebGLPoints = WebGLPointsLayer;
  ol.layer.WebGLTile = WebGLTileLayer;
  ol.loadingstrategy = {};
  ol.loadingstrategy.all = all;
  ol.loadingstrategy.bbox = bbox$1;
  ol.loadingstrategy.tile = tile;
  ol.math = {};
  ol.math.ceil = ceil;
  ol.math.clamp = clamp;
  ol.math.floor = floor;
  ol.math.lerp = lerp;
  ol.math.modulo = modulo;
  ol.math.round = round;
  ol.math.solveLinearSystem = solveLinearSystem;
  ol.math.squaredDistance = squaredDistance$1;
  ol.math.squaredSegmentDistance = squaredSegmentDistance;
  ol.math.toDegrees = toDegrees;
  ol.math.toFixed = toFixed;
  ol.math.toRadians = toRadians;
  ol.net = {};
  ol.net.ClientError = ClientError;
  ol.net.ResponseError = ResponseError;
  ol.net.getJSON = getJSON;
  ol.net.jsonp = jsonp;
  ol.net.overrideXHR = overrideXHR;
  ol.net.resolveUrl = resolveUrl;
  ol.net.restoreXHR = restoreXHR;
  ol.obj = {};
  ol.obj.clear = clear$2;
  ol.obj.isEmpty = isEmpty$1;
  ol.proj = {};
  ol.proj.Projection = Projection;
  ol.proj.Units = {};
  ol.proj.Units.METERS_PER_UNIT = METERS_PER_UNIT$1;
  ol.proj.Units.fromCode = fromCode;
  ol.proj.addCommon = addCommon;
  ol.proj.addCoordinateTransforms = addCoordinateTransforms;
  ol.proj.addEquivalentProjections = addEquivalentProjections;
  ol.proj.addEquivalentTransforms = addEquivalentTransforms;
  ol.proj.addProjection = addProjection;
  ol.proj.addProjections = addProjections;
  ol.proj.clearAllProjections = clearAllProjections;
  ol.proj.clearUserProjection = clearUserProjection;
  ol.proj.cloneTransform = cloneTransform;
  ol.proj.createProjection = createProjection;
  ol.proj.createSafeCoordinateTransform = createSafeCoordinateTransform;
  ol.proj.createTransformFromCoordinateTransform = createTransformFromCoordinateTransform;
  ol.proj.disableCoordinateWarning = disableCoordinateWarning;
  ol.proj.epsg3857 = {};
  ol.proj.epsg3857.EXTENT = EXTENT$2;
  ol.proj.epsg3857.HALF_SIZE = HALF_SIZE;
  ol.proj.epsg3857.MAX_SAFE_Y = MAX_SAFE_Y;
  ol.proj.epsg3857.PROJECTIONS = PROJECTIONS$1;
  ol.proj.epsg3857.RADIUS = RADIUS$1;
  ol.proj.epsg3857.WORLD_EXTENT = WORLD_EXTENT;
  ol.proj.epsg3857.fromEPSG4326 = fromEPSG4326;
  ol.proj.epsg3857.toEPSG4326 = toEPSG4326;
  ol.proj.epsg4326 = {};
  ol.proj.epsg4326.EXTENT = EXTENT$1;
  ol.proj.epsg4326.METERS_PER_UNIT = METERS_PER_UNIT;
  ol.proj.epsg4326.PROJECTIONS = PROJECTIONS;
  ol.proj.epsg4326.RADIUS = RADIUS;
  ol.proj.equivalent = equivalent;
  ol.proj.fromLonLat = fromLonLat;
  ol.proj.fromUserCoordinate = fromUserCoordinate;
  ol.proj.fromUserExtent = fromUserExtent;
  ol.proj.fromUserResolution = fromUserResolution;
  ol.proj.get = get$2;
  ol.proj.getPointResolution = getPointResolution;
  ol.proj.getTransform = getTransform;
  ol.proj.getTransformFromProjections = getTransformFromProjections;
  ol.proj.getUserProjection = getUserProjection;
  ol.proj.identityTransform = identityTransform;
  ol.proj.proj4 = {};
  ol.proj.proj4.register = register;
  ol.proj.projections = {};
  ol.proj.projections.add = add$2;
  ol.proj.projections.clear = clear$1;
  ol.proj.projections.get = get$4;
  ol.proj.setUserProjection = setUserProjection;
  ol.proj.toLonLat = toLonLat;
  ol.proj.toUserCoordinate = toUserCoordinate;
  ol.proj.toUserExtent = toUserExtent;
  ol.proj.toUserResolution = toUserResolution;
  ol.proj.transform = transform;
  ol.proj.transformExtent = transformExtent;
  ol.proj.transformWithProjections = transformWithProjections;
  ol.proj.transforms = {};
  ol.proj.transforms.add = add$1;
  ol.proj.transforms.clear = clear;
  ol.proj.transforms.get = get$3;
  ol.proj.transforms.remove = remove;
  ol.proj.useGeographic = useGeographic;
  ol.render = {};
  ol.render.Box = RenderBox;
  ol.render.Event = RenderEvent;
  ol.render.Feature = RenderFeature;
  ol.render.Feature.toFeature = toFeature;
  ol.render.Feature.toGeometry = toGeometry;
  ol.render.VectorContext = VectorContext;
  ol.render.canvas = {};
  ol.render.canvas.Builder = CanvasBuilder;
  ol.render.canvas.BuilderGroup = BuilderGroup;
  ol.render.canvas.Executor = Executor;
  ol.render.canvas.ExecutorGroup = ExecutorGroup;
  ol.render.canvas.ExecutorGroup.getPixelIndexArray = getPixelIndexArray;
  ol.render.canvas.ImageBuilder = CanvasImageBuilder;
  ol.render.canvas.Immediate = CanvasImmediateRenderer;
  ol.render.canvas.Instruction = {};
  ol.render.canvas.Instruction.beginPathInstruction = beginPathInstruction;
  ol.render.canvas.Instruction.closePathInstruction = closePathInstruction;
  ol.render.canvas.Instruction.fillInstruction = fillInstruction;
  ol.render.canvas.Instruction.strokeInstruction = strokeInstruction;
  ol.render.canvas.LineStringBuilder = CanvasLineStringBuilder;
  ol.render.canvas.PolygonBuilder = CanvasPolygonBuilder;
  ol.render.canvas.TextBuilder = CanvasTextBuilder;
  ol.render.canvas.checkedFonts = checkedFonts;
  ol.render.canvas.defaultFillStyle = defaultFillStyle;
  ol.render.canvas.defaultFont = defaultFont;
  ol.render.canvas.defaultLineCap = defaultLineCap;
  ol.render.canvas.defaultLineDash = defaultLineDash;
  ol.render.canvas.defaultLineDashOffset = defaultLineDashOffset;
  ol.render.canvas.defaultLineJoin = defaultLineJoin;
  ol.render.canvas.defaultLineWidth = defaultLineWidth;
  ol.render.canvas.defaultMiterLimit = defaultMiterLimit;
  ol.render.canvas.defaultPadding = defaultPadding;
  ol.render.canvas.defaultStrokeStyle = defaultStrokeStyle;
  ol.render.canvas.defaultTextAlign = defaultTextAlign;
  ol.render.canvas.defaultTextBaseline = defaultTextBaseline;
  ol.render.canvas.drawImageOrLabel = drawImageOrLabel;
  ol.render.canvas.getTextDimensions = getTextDimensions;
  ol.render.canvas.hitdetect = {};
  ol.render.canvas.hitdetect.HIT_DETECT_RESOLUTION = HIT_DETECT_RESOLUTION;
  ol.render.canvas.hitdetect.createHitDetectionImageData = createHitDetectionImageData;
  ol.render.canvas.hitdetect.hitDetect = hitDetect;
  ol.render.canvas.measureAndCacheTextWidth = measureAndCacheTextWidth;
  ol.render.canvas.measureTextHeight = measureTextHeight;
  ol.render.canvas.measureTextWidth = measureTextWidth;
  ol.render.canvas.registerFont = registerFont;
  ol.render.canvas.rotateAtOffset = rotateAtOffset;
  ol.render.canvas.textHeights = textHeights;
  ol.render.getRenderPixel = getRenderPixel;
  ol.render.getVectorContext = getVectorContext;
  ol.render.toContext = toContext;
  ol.render.webgl = {};
  ol.render.webgl.BatchRenderer = AbstractBatchRenderer;
  ol.render.webgl.LineStringBatchRenderer = LineStringBatchRenderer;
  ol.render.webgl.MixedGeometryBatch = MixedGeometryBatch;
  ol.render.webgl.PointBatchRenderer = PointBatchRenderer;
  ol.render.webgl.PolygonBatchRenderer = PolygonBatchRenderer;
  ol.render.webgl.utils = {};
  ol.render.webgl.utils.colorDecodeId = colorDecodeId;
  ol.render.webgl.utils.colorEncodeId = colorEncodeId;
  ol.render.webgl.utils.getBlankImageData = getBlankImageData;
  ol.render.webgl.utils.writeLineSegmentToBuffers = writeLineSegmentToBuffers;
  ol.render.webgl.utils.writePointFeatureToBuffers = writePointFeatureToBuffers;
  ol.render.webgl.utils.writePolygonTrianglesToBuffers = writePolygonTrianglesToBuffers;
  ol.renderer = {};
  ol.renderer.Composite = CompositeMapRenderer;
  ol.renderer.Layer = LayerRenderer;
  ol.renderer.Map = MapRenderer;
  ol.renderer.canvas = {};
  ol.renderer.canvas.ImageLayer = CanvasImageLayerRenderer;
  ol.renderer.canvas.Layer = CanvasLayerRenderer;
  ol.renderer.canvas.Layer.canvasPool = canvasPool;
  ol.renderer.canvas.TileLayer = CanvasTileLayerRenderer;
  ol.renderer.canvas.VectorImageLayer = CanvasVectorImageLayerRenderer;
  ol.renderer.canvas.VectorLayer = CanvasVectorLayerRenderer;
  ol.renderer.canvas.VectorTileLayer = CanvasVectorTileLayerRenderer;
  ol.renderer.vector = {};
  ol.renderer.vector.defaultOrder = defaultOrder;
  ol.renderer.vector.getSquaredTolerance = getSquaredTolerance;
  ol.renderer.vector.getTolerance = getTolerance;
  ol.renderer.vector.renderFeature = renderFeature$1;
  ol.renderer.webgl = {};
  ol.renderer.webgl.Layer = WebGLLayerRenderer;
  ol.renderer.webgl.PointsLayer = WebGLPointsLayerRenderer;
  ol.renderer.webgl.TileLayer = WebGLTileLayerRenderer;
  ol.renderer.webgl.TileLayer.Attributes = Attributes$3;
  ol.renderer.webgl.TileLayer.Uniforms = Uniforms;
  ol.renderer.webgl.VectorLayer = WebGLVectorLayerRenderer;
  ol.renderer.webgl.shaders = {};
  ol.renderer.webgl.shaders.FILL_FRAGMENT_SHADER = FILL_FRAGMENT_SHADER;
  ol.renderer.webgl.shaders.FILL_VERTEX_SHADER = FILL_VERTEX_SHADER;
  ol.renderer.webgl.shaders.POINT_FRAGMENT_SHADER = POINT_FRAGMENT_SHADER;
  ol.renderer.webgl.shaders.POINT_VERTEX_SHADER = POINT_VERTEX_SHADER;
  ol.renderer.webgl.shaders.STROKE_FRAGMENT_SHADER = STROKE_FRAGMENT_SHADER;
  ol.renderer.webgl.shaders.STROKE_VERTEX_SHADER = STROKE_VERTEX_SHADER;
  ol.renderer.webgl.shaders.packColor = packColor;
  ol.reproj = {};
  ol.reproj.Image = ReprojImage;
  ol.reproj.Tile = ReprojTile;
  ol.reproj.Triangulation = Triangulation;
  ol.reproj.calculateSourceExtentResolution = calculateSourceExtentResolution;
  ol.reproj.calculateSourceResolution = calculateSourceResolution;
  ol.reproj.canvasPool = canvasPool$1;
  ol.reproj.common = {};
  ol.reproj.common.ERROR_THRESHOLD = ERROR_THRESHOLD;
  ol.reproj.render = render;
  ol.resolutionconstraint = {};
  ol.resolutionconstraint.createMinMaxResolution = createMinMaxResolution;
  ol.resolutionconstraint.createSnapToPower = createSnapToPower;
  ol.resolutionconstraint.createSnapToResolutions = createSnapToResolutions;
  ol.rotationconstraint = {};
  ol.rotationconstraint.createSnapToN = createSnapToN;
  ol.rotationconstraint.createSnapToZero = createSnapToZero;
  ol.rotationconstraint.disable = disable;
  ol.rotationconstraint.none = none;
  ol.size = {};
  ol.size.buffer = buffer;
  ol.size.hasArea = hasArea;
  ol.size.scale = scale;
  ol.size.toSize = toSize;
  ol.source = {};
  ol.source.BingMaps = BingMaps;
  ol.source.BingMaps.quadKey = quadKey;
  ol.source.CartoDB = CartoDB;
  ol.source.Cluster = Cluster;
  ol.source.DataTile = DataTileSource;
  ol.source.GeoTIFF = GeoTIFFSource;
  ol.source.IIIF = IIIF;
  ol.source.Image = ImageSource;
  ol.source.Image.ImageSourceEvent = ImageSourceEvent;
  ol.source.Image.defaultImageLoadFunction = defaultImageLoadFunction;
  ol.source.ImageArcGISRest = ImageArcGISRest;
  ol.source.ImageCanvas = ImageCanvasSource;
  ol.source.ImageMapGuide = ImageMapGuide;
  ol.source.ImageStatic = Static;
  ol.source.ImageWMS = ImageWMS;
  ol.source.OGCMapTile = OGCMapTile;
  ol.source.OGCVectorTile = OGCVectorTile;
  ol.source.OSM = OSM;
  ol.source.OSM.ATTRIBUTION = ATTRIBUTION;
  ol.source.Raster = RasterSource;
  ol.source.Raster.Processor = Processor;
  ol.source.Raster.RasterSourceEvent = RasterSourceEvent;
  ol.source.Raster.newImageData = newImageData;
  ol.source.Source = Source;
  ol.source.Stamen = Stamen;
  ol.source.Tile = TileSource;
  ol.source.Tile.TileSourceEvent = TileSourceEvent;
  ol.source.TileArcGISRest = TileArcGISRest;
  ol.source.TileDebug = TileDebug;
  ol.source.TileImage = TileImage;
  ol.source.TileJSON = TileJSON;
  ol.source.TileWMS = TileWMS;
  ol.source.UTFGrid = UTFGrid;
  ol.source.UTFGrid.CustomTile = CustomTile;
  ol.source.UrlTile = UrlTile;
  ol.source.Vector = VectorSource;
  ol.source.Vector.VectorSourceEvent = VectorSourceEvent;
  ol.source.VectorTile = VectorTile;
  ol.source.VectorTile.defaultLoadFunction = defaultLoadFunction;
  ol.source.WMTS = WMTS;
  ol.source.WMTS.optionsFromCapabilities = optionsFromCapabilities;
  ol.source.XYZ = XYZ;
  ol.source.Zoomify = Zoomify;
  ol.source.Zoomify.CustomTile = CustomTile$1;
  ol.source.common = {};
  ol.source.common.DEFAULT_WMS_VERSION = DEFAULT_WMS_VERSION;
  ol.source.ogcTileUtil = {};
  ol.source.ogcTileUtil.getMapTileUrlTemplate = getMapTileUrlTemplate;
  ol.source.ogcTileUtil.getTileSetInfo = getTileSetInfo;
  ol.source.ogcTileUtil.getVectorTileUrlTemplate = getVectorTileUrlTemplate;
  ol.source.sourcesFromTileGrid = sourcesFromTileGrid;
  ol.source.wms = {};
  ol.source.wms.DEFAULT_VERSION = DEFAULT_VERSION$1;
  ol.sphere = {};
  ol.sphere.DEFAULT_RADIUS = DEFAULT_RADIUS;
  ol.sphere.getArea = getArea;
  ol.sphere.getDistance = getDistance;
  ol.sphere.getLength = getLength;
  ol.sphere.offset = offset;
  ol.string = {};
  ol.string.compareVersions = compareVersions;
  ol.string.padNumber = padNumber;
  ol.structs = {};
  ol.structs.LRUCache = LRUCache;
  ol.structs.LinkedList = LinkedList;
  ol.structs.PriorityQueue = PriorityQueue;
  ol.structs.PriorityQueue.DROP = DROP;
  ol.structs.RBush = RBush;
  ol.style = {};
  ol.style.Circle = CircleStyle;
  ol.style.Fill = Fill;
  ol.style.Icon = Icon;
  ol.style.IconImage = IconImage;
  ol.style.IconImage.get = get$1;
  ol.style.IconImageCache = IconImageCache;
  ol.style.IconImageCache.shared = shared;
  ol.style.Image = ImageStyle;
  ol.style.RegularShape = RegularShape;
  ol.style.Stroke = Stroke;
  ol.style.Style = Style;
  ol.style.Style.createDefaultStyle = createDefaultStyle;
  ol.style.Style.createEditingStyle = createEditingStyle;
  ol.style.Style.toFunction = toFunction;
  ol.style.Text = Text;
  ol.style.expressions = {};
  ol.style.expressions.Operators = Operators;
  ol.style.expressions.PALETTE_TEXTURE_ARRAY = PALETTE_TEXTURE_ARRAY;
  ol.style.expressions.arrayToGlsl = arrayToGlsl;
  ol.style.expressions.colorToGlsl = colorToGlsl;
  ol.style.expressions.expressionToGlsl = expressionToGlsl;
  ol.style.expressions.getStringNumberEquivalent = getStringNumberEquivalent;
  ol.style.expressions.getValueType = getValueType;
  ol.style.expressions.isTypeUnique = isTypeUnique;
  ol.style.expressions.numberToGlsl = numberToGlsl;
  ol.style.expressions.stringToGlsl = stringToGlsl;
  ol.style.expressions.uniformNameForVariable = uniformNameForVariable;
  ol.style.flat = {};
  ol.style.flat.toStyle = toStyle;
  ol.tilecoord = {};
  ol.tilecoord.createOrUpdate = createOrUpdate$1;
  ol.tilecoord.fromKey = fromKey;
  ol.tilecoord.getCacheKeyForTileKey = getCacheKeyForTileKey;
  ol.tilecoord.getKey = getKey;
  ol.tilecoord.getKeyZXY = getKeyZXY;
  ol.tilecoord.hash = hash;
  ol.tilecoord.withinExtentAndZ = withinExtentAndZ;
  ol.tilegrid = {};
  ol.tilegrid.TileGrid = TileGrid;
  ol.tilegrid.WMTS = WMTSTileGrid;
  ol.tilegrid.WMTS.createFromCapabilitiesMatrixSet = createFromCapabilitiesMatrixSet;
  ol.tilegrid.common = {};
  ol.tilegrid.common.DEFAULT_MAX_ZOOM = DEFAULT_MAX_ZOOM;
  ol.tilegrid.common.DEFAULT_TILE_SIZE = DEFAULT_TILE_SIZE;
  ol.tilegrid.createForExtent = createForExtent;
  ol.tilegrid.createForProjection = createForProjection;
  ol.tilegrid.createXYZ = createXYZ;
  ol.tilegrid.extentFromProjection = extentFromProjection;
  ol.tilegrid.getForProjection = getForProjection;
  ol.tilegrid.wrapX = wrapX;
  ol.tileurlfunction = {};
  ol.tileurlfunction.createFromTemplate = createFromTemplate;
  ol.tileurlfunction.createFromTemplates = createFromTemplates;
  ol.tileurlfunction.createFromTileUrlFunctions = createFromTileUrlFunctions;
  ol.tileurlfunction.expandUrl = expandUrl;
  ol.tileurlfunction.nullTileUrlFunction = nullTileUrlFunction;
  ol.transform = {};
  ol.transform.apply = apply;
  ol.transform.compose = compose;
  ol.transform.composeCssTransform = composeCssTransform;
  ol.transform.create = create$2;
  ol.transform.determinant = determinant;
  ol.transform.invert = invert;
  ol.transform.makeInverse = makeInverse;
  ol.transform.makeScale = makeScale;
  ol.transform.multiply = multiply;
  ol.transform.reset = reset;
  ol.transform.rotate = rotate$2;
  ol.transform.scale = scale$3;
  ol.transform.set = set;
  ol.transform.setFromArray = setFromArray;
  ol.transform.toString = toString$3;
  ol.transform.translate = translate$1;
  ol.uri = {};
  ol.uri.appendParams = appendParams;
  ol.util = {};
  ol.util.VERSION = VERSION;
  ol.util.abstract = abstract;
  ol.util.getUid = getUid;
  ol.vec = {};
  ol.vec.mat4 = {};
  ol.vec.mat4.create = create$1;
  ol.vec.mat4.fromTransform = fromTransform;
  ol.webgl = {};
  ol.webgl.ARRAY_BUFFER = ARRAY_BUFFER;
  ol.webgl.Buffer = WebGLArrayBuffer;
  ol.webgl.Buffer.getArrayClassForType = getArrayClassForType;
  ol.webgl.DYNAMIC_DRAW = DYNAMIC_DRAW;
  ol.webgl.ELEMENT_ARRAY_BUFFER = ELEMENT_ARRAY_BUFFER;
  ol.webgl.FLOAT = FLOAT;
  ol.webgl.Helper = WebGLHelper;
  ol.webgl.Helper.computeAttributesStride = computeAttributesStride;
  ol.webgl.PaletteTexture = PaletteTexture;
  ol.webgl.PostProcessingPass = WebGLPostProcessingPass;
  ol.webgl.RenderTarget = WebGLRenderTarget;
  ol.webgl.STATIC_DRAW = STATIC_DRAW;
  ol.webgl.STREAM_DRAW = STREAM_DRAW;
  ol.webgl.ShaderBuilder = {};
  ol.webgl.ShaderBuilder.ShaderBuilder = ShaderBuilder;
  ol.webgl.ShaderBuilder.parseLiteralStyle = parseLiteralStyle;
  ol.webgl.TileTexture = TileTexture;
  ol.webgl.UNSIGNED_BYTE = UNSIGNED_BYTE;
  ol.webgl.UNSIGNED_INT = UNSIGNED_INT;
  ol.webgl.UNSIGNED_SHORT = UNSIGNED_SHORT;
  ol.webgl.getContext = getContext;
  ol.webgl.getSupportedExtensions = getSupportedExtensions;
  ol.xml = {};
  ol.xml.OBJECT_PROPERTY_NODE_FACTORY = OBJECT_PROPERTY_NODE_FACTORY;
  ol.xml.XML_SCHEMA_INSTANCE_URI = XML_SCHEMA_INSTANCE_URI;
  ol.xml.createElementNS = createElementNS;
  ol.xml.getAllTextContent = getAllTextContent;
  ol.xml.getAllTextContent_ = getAllTextContent_;
  ol.xml.getAttributeNS = getAttributeNS;
  ol.xml.getDocument = getDocument;
  ol.xml.getXMLSerializer = getXMLSerializer;
  ol.xml.isDocument = isDocument;
  ol.xml.makeArrayExtender = makeArrayExtender;
  ol.xml.makeArrayPusher = makeArrayPusher;
  ol.xml.makeArraySerializer = makeArraySerializer;
  ol.xml.makeChildAppender = makeChildAppender;
  ol.xml.makeObjectPropertyPusher = makeObjectPropertyPusher;
  ol.xml.makeObjectPropertySetter = makeObjectPropertySetter;
  ol.xml.makeReplacer = makeReplacer;
  ol.xml.makeSequence = makeSequence;
  ol.xml.makeSimpleNodeFactory = makeSimpleNodeFactory;
  ol.xml.makeStructureNS = makeStructureNS;
  ol.xml.parse = parse;
  ol.xml.parseNode = parseNode;
  ol.xml.pushParseAndPop = pushParseAndPop;
  ol.xml.pushSerializeAndPop = pushSerializeAndPop;
  ol.xml.registerDocument = registerDocument;
  ol.xml.registerXMLSerializer = registerXMLSerializer;
  ol.xml.serialize = serialize;

  return ol;

}));
